/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*
// This code deals with image compression decompression
//
// There are many short long inconsistencies in here that ought to
// be cleaned up
*/

#include "rafport.h"
#include "entity.h"

#include "dafs.h"
#include "t4.h"
#include "g3states.h"
#include "smemory.h"
#include "comp.h"

enum {
	STEOL,
	STBIT,
	STG3,
	STG4
};

struct DecodeState {
	unsigned char *stream,*buf;
	short data;			/* current i/o byte */
	int4 amt,bit,a0,type,color,mode,bufsize,runlen;
	int4 len,eols,state,x,st,runt,outoff;
	long width;
	int4 run1,umode,g32dst;
	short blank[4],*code,*ref;
};

struct EncodeState {
	unsigned long data;
	unsigned long *stream,*start;
	short bitpos;
	iDAFSCompType type;
	short blank[4];
	int4 width;
};

#define	BITCASE(b)			\
case b:				\
code <<= 1;			\
if (data & (1<<(7-b))) code |= 1;\
sp->len++;				\
if (code > 0) { bit = b+1; break; }


/*
 * CCITT Group 3 FAX Encoding.
 */

static const tableentry eolcode = {12,0x1};		/* 000000000001 */
static const tableentry horizcode = {3, 0x1};	/* 001 */
static const tableentry passcode = {4, 0x1};	/* 0001 */
static const tableentry vcodes[7] = {
	{7, 0x03},		/* 0000 011 VR3 */
	{6, 0x03},		/* 0000 11 	VR2 */
	{3, 0x03},		/* 011 		VR1 */
	{1, 0x1},		/* 1 		VR0 */
	{3, 0x2},		/* 010 		VL1 */
	{6, 0x02},		/* 0000 10 	VL2	*/
	{7, 0x02}		/* 0000 010 VL3	*/
};



static void putcodeswap(EncodeState *sp, tableentry const *te);
static void putcodenorm(EncodeState *sp, tableentry const *te);
typedef void (* PutCodeFunction) (EncodeState *sp, tableentry const *te);
PutCodeFunction putcode = 0;

void i_InitComp(void)
{
	short swap = 1;
	if (*(char*)&swap == 1)
		putcode = putcodeswap;
	else putcode = putcodenorm;
}

static int4 skiptoeol(DecodeState *sp)
{
	int4 bit = sp->bit;
	/* bogus sp->data is short, always has been -- tom */
	short data = sp->data;
	int4 code = 0;
	/*
	 * Our handling of ``bit'' is painful because the rest of the code does not
	 * maintain it as exactly the bit offset in the current data byte (bit == 0
	 * means refill the data byte). Thus we have to be careful on entry and exit
	 * to insure that we maintain a value that's understandable elsewhere in the
	 * decoding logic.
	 */
	if (bit == 0)		/* force refill */
		bit = 8;
	for (;;) {
		switch (bit) {
		  again:	BITCASE(0);
			BITCASE(1);
			BITCASE(2);
			BITCASE(3);
			BITCASE(4);
			BITCASE(5);
			BITCASE(6);
			BITCASE(7);
		default:
			sp->bit = 0;
			if (!sp->amt--)
				return COMPMORE;
			data = *sp->stream++;
			goto again;
		}
		if (sp->len >= 12 && code == EOL)
			break;
		code = sp->len = 0;
	}
	sp->bit = bit > 7 ? 0 : bit;	/* force refill */
	sp->data = data;
	return 0;
}

/*
 * Decode a run
 */
static int4 decode_run(DecodeState *sp)
{
	short action;
	
	for (;;) {
		if (!sp->bit) {
			if (!sp->amt--)
				return COMPMORE;
			sp->data = *sp->stream++;
		}
		action = RlAction1D[sp->state][sp->data];
		sp->state = RlNextState1D[sp->state][sp->data];
		if (action == ACT_INCOMP) {
			sp->bit = 0;
			continue;
		}
		if (action == ACT_INVALID)
			return (G3CODE_INVALID);
		if (action == ACT_EOL)
			return (G3CODE_EOL);
		sp->bit = sp->state;
		action = RlWhiteCodes[action - sp->runt].runlen;
		sp->runlen += action;
		if (action < 64)
			return (sp->runlen);
		sp->state += sp->st;
	}
	/* NOTREACHED */
}

/*
 * Process one row of 1d Huffman-encoded data.
 */
static int4 Decode1D(DecodeState *sp,short *buf)
{
	int4 runlen;
	
	for (;;) {
		runlen = decode_run(sp);
		switch (runlen) {
		case COMPMORE:
		case G3CODE_EOF:
		case G3CODE_EOL:	/* premature end-of-line code */
			return runlen;	/* try to resynchronize... */
		case G3CODE_INVALID:	/* invalid code */
			/*
			 * An invalid code was encountered. Flush the remainder of
			 * the line and allow the caller to decide whether or not to
			 * continue.  Note that this only works if we have a G3 image
			 * with EOL markers.
			 */
			break;
		}
		sp->x += runlen;
		if (sp->x >= sp->width) {
			/* bogus long assigned to short */
			if (!sp->color) 
				buf[sp->outoff++] = sp->width;
			buf[sp->outoff++] = sp->width;
			buf[sp->outoff++] = -1;
			break;
		}
		buf[sp->outoff++] = sp->x;
		if (sp->color = !sp->color) {
			sp->st = 8;
			sp->runt = ACT_BRUNT;
		} else {
			sp->st = 0;
			sp->runt = ACT_WRUNT;
		}
		sp->runlen = 0;
		sp->state = sp->bit + sp->st;
	}
	return 0;
}

/*
// bogus -- why does this return a long when a short value
// is used all over the place
*/
#if DO_INLINE
inline
#endif
static int4 finddiff(short *ref,int4 x,long w,int4 color)
{
	if (x == w) return w;
	if (color) {
		if (!x) ref++;
		else ref--;
	}
	while (*ref >= 0 && *ref <= x) ref+=2;
	if (*ref < 0) return w;
	return (int4)*ref;
}

/*
 * Return the next uncompressed mode code word.
 */
static int4 decode_uncomp_code(DecodeState *sp)
{
	short code;
	
	do {
		if (sp->bit == 0 || sp->bit > 7) {
			if (!sp->amt--)
				return COMPMORE;
			sp->data = *sp->stream++;
		}
		code = RlUncompAction[sp->bit][sp->data];
		sp->bit = RlUncompNextState[sp->bit][sp->data];
	} while (code == ACT_INCOMP);
	return (code);
}

/*
 * Return the next bit in the input stream.  This is
 * used to extract 2D tag values and the color tag
 * at the end of a terminating uncompressed data code.
 */
static int4 nextbit(DecodeState *sp)
{
	int4 bit;
	
	if (sp->bit == 0) {
		if (!sp->amt--)
			return COMPMORE;
		sp->data = *sp->stream++;
	}
	bit = sp->data & (0x80 >> sp->bit);
	if (++(sp->bit) > 7)
		sp->bit = 0;
	return (bit);
}

/*
 * Process one row of 2d encoded data.
 */
static int4 Decode2D(DecodeState *sp, short *buf,short *ref)
{
	int4 b1, b2,run2;		/* for horizontal mode */
	
	for (;;) {
		switch (sp->mode) {
		case MODE_NULL:
			break;
		case MODE_PASS:
			b1 = finddiff(ref, sp->a0, sp->width, sp->color);
			b2 = finddiff(ref, b1, sp->width, !sp->color);
			if (sp->color) {
				if (sp->a0 <= 0) {
					buf[sp->outoff++] = 0;
					buf[sp->outoff++] = b2; /* bogus another long assigned to short */
				} else if (buf[sp->outoff-1] == sp->a0) {
					buf[sp->outoff-1] = b2;
				} else {
					buf[sp->outoff++] = sp->a0;
					buf[sp->outoff++] = b2;
				}
			}
			sp->a0 = b2;
			break;
		case MODE_HORIZ:
			if (sp->color) {
				sp->st = 8;
				sp->runt = ACT_BRUNT;
			} else {
				sp->st = 0;
				sp->runt = ACT_WRUNT;
			}
			sp->state = sp->bit + sp->st;
			sp->runlen = 0;
			sp->mode = MODE_H1;
		case MODE_H1:
			sp->run1 = decode_run(sp);
			if (sp->run1 < 0)
				return sp->run1;
			if (sp->color) {
				sp->st = 0;
				sp->runt = ACT_WRUNT;
			} else {
				sp->st = 8;
				sp->runt = ACT_BRUNT;
			}
			sp->state = sp->bit + sp->st;
			sp->runlen = 0;
			sp->mode = MODE_H2;
		case MODE_H2:
			run2 = decode_run(sp);
			if (run2 < 0)
				return run2;
			/*
			 * Do the appropriate fill.  Note that we exit this logic
			 * with the same color that we enter with since we do 2
			 * fills.  This explains the somewhat obscure logic below.
			 */
			if (sp->a0 < 0)
				sp->a0 = 0;
			if (sp->a0 + sp->run1 > sp->width)
				sp->run1 = sp->width - sp->a0;
			if (sp->color) {
				if (sp->a0 && buf[sp->outoff-1] == sp->a0)
					/* bogus -- another long assigned to short */
					buf[sp->outoff-1] = sp->run1 + sp->a0;
				else {
					buf[sp->outoff++] = sp->a0;
					buf[sp->outoff++] = sp->run1 + sp->a0;
				}
			}
			sp->a0 += sp->run1;
			if (sp->a0 + run2 > sp->width)
				run2 = sp->width - sp->a0;
			if (!sp->color) {
				if (sp->a0 && buf[sp->outoff-1] == sp->a0)
					buf[sp->outoff-1] = run2 + sp->a0;
				else {
					buf[sp->outoff++] = sp->a0;
					buf[sp->outoff++] = run2 + sp->a0;
				}
			}
			sp->a0 += run2;
			break;
		case MODE_VERT_V0:
		case MODE_VERT_VR1:
		case MODE_VERT_VR2:
		case MODE_VERT_VR3:
		case MODE_VERT_VL1:
		case MODE_VERT_VL2:
		case MODE_VERT_VL3:
			b1 = finddiff(ref, sp->a0, sp->width, sp->color);
			b1 += sp->mode - MODE_VERT_V0;
			if (sp->color) {
				if (sp->a0 <= 0) {
					buf[sp->outoff++] = 0;
					buf[sp->outoff++] = b1; /* bogus -- another long assigned to short */
				} else if (buf[sp->outoff-1] == sp->a0)
					buf[sp->outoff-1] = b1;
				else {
					buf[sp->outoff++] = sp->a0;
					buf[sp->outoff++] = b1;
				}
			}
			sp->color = !sp->color;
			sp->a0 = b1;
			break;
		case MODE_UNCOMP:
			/*
			 * Uncompressed mode: select from the special set of code
			 * words.
			 */
			if (sp->a0 < 0)
				sp->a0 = 0;
			do {
				sp->umode = decode_uncomp_code(sp);
				switch (sp->umode) {
				case UNCOMP_RUN1:
				case UNCOMP_RUN2:
				case UNCOMP_RUN3:
				case UNCOMP_RUN4:
				case UNCOMP_RUN5:
					sp->run1 = sp->umode - UNCOMP_RUN0;
					/* bogus -- another long assigned to short */
					buf[sp->outoff++] = sp->a0 + sp->run1 - 1;
					buf[sp->outoff++] = sp->a0 + sp->run1;
					sp->a0 += sp->run1;
					break;
				case UNCOMP_RUN6:
					sp->a0 += 5;
					break;
				case UNCOMP_TRUN0:
				case UNCOMP_TRUN1:
				case UNCOMP_TRUN2:
				case UNCOMP_TRUN3:
				case UNCOMP_TRUN4:
					sp->run1 = sp->umode - UNCOMP_TRUN0;
					sp->a0 += sp->run1;
					sp->color = nextbit(sp);
					if (sp->color < 0)
						return sp->color;
					break;
				case UNCOMP_INVALID:
					break;
				case UNCOMP_EOF:
				case G3CODE_EOF:
					return G3CODE_EOF;
				}
			} while (sp->umode < UNCOMP_EXIT);
			break;
		case MODE_ERROR_1: /* should be able to resync on 3-2d images */
		case MODE_ERROR:
		default:
			return -1;
		}
		sp->mode = MODE_NULL;
		if (sp->a0 >= sp->width)
			break;
		if (sp->bit == 0 || sp->bit > 7) {
			if (!sp->amt--)
				return COMPMORE;
			sp->data = *sp->stream++;
		}
		sp->mode = RlMode2D[sp->bit][sp->data];
		sp->bit = RlNextState2D[sp->bit][sp->data];
	}
	if (buf[sp->outoff-1] != sp->width) {
		/* bogus -- another long assigned to short */
		buf[sp->outoff++] = sp->width;
		buf[sp->outoff++] = sp->width;
	}
	buf[sp->outoff++] = -1;
	return 0;
}

DecodeStatePtr DecodeAlloc(void)
{
	return (DecodeStatePtr)smalloc(sizeof(DecodeState));
}

void DecodeInit(DecodeStatePtr sp,iImagePtr im,iDAFSCompType type)
{
	sp->g32dst = STEOL;
	sp->bit = 0;
	sp->a0 = -1;
	sp->type = type;
	sp->color = 0;
	sp->mode = MODE_EOL;
	sp->len = 0;
	sp->width = im->width;
	sp->outoff = 1;
	/* bogus blank is short[] im->width is long */
	sp->blank[0] = 0;
	sp->blank[1] = sp->blank[2] = im->width;
	sp->blank[3] = -1;
}

void DecodeSetCodeRef(DecodeStatePtr sp,short *code,short *ref)
{
	sp->code = code;
	if (!ref) ref = sp->blank;
	sp->ref = ref;
}

int Decode(DecodeStatePtr sp)
{
	int4 done;
	int nextst;
	
	switch (sp->type) {
	case CCITT32D:
		switch (sp->g32dst) {
		case STEOL:
			if (done = skiptoeol(sp))
				return done;
			sp->g32dst = STBIT;
		case STBIT:
			nextst = nextbit(sp);
			if (nextst < 0) return nextst;
			if (nextst) {
				sp->g32dst = STG3;
				sp->mode = MODE_NULL;
				sp->runt = ACT_WRUNT;
				sp->runlen = 0;
				sp->st = 0;
				sp->state = sp->bit;
				sp->color = 0;
				sp->x = 0;
				*sp->code = 0;
				sp->outoff = 1;
			} else {
				sp->g32dst = STG4;
				sp->mode = MODE_NULL;
				*sp->code = 0;
				sp->a0 = -1;
				sp->outoff = 1;
				sp->color = 0;
			}
			break;
		}
		switch (sp->g32dst) {
		case STG3:
			if (done = Decode1D(sp,sp->code))
				return done;
			sp->mode = MODE_EOL;
			sp->len = 0;
			sp->g32dst = STEOL;
			break;
		case STG4:
			if (done = Decode2D(sp,sp->code,sp->ref+1))
				return done;
			sp->mode = MODE_EOL;
			sp->g32dst = STEOL;
			break;
		}
		break;
	case CCITT3:
		if (sp->mode == MODE_EOL) {
			if (done = skiptoeol(sp))
				return done;
			sp->mode = MODE_NULL;
			sp->runt = ACT_WRUNT;
			sp->runlen = 0;
			sp->st = 0;
			sp->state = sp->bit;
			sp->color = 0;
			sp->x = 0;
			*sp->code = 0;
			sp->outoff = 1;
		}
		if (done = Decode1D(sp,sp->code))
			return done;
		sp->mode = MODE_EOL;
		sp->len = 0;
		break;
	case CCITT4:
		if (sp->mode == MODE_EOL) {
			sp->mode = MODE_NULL;
			*sp->code = 0;
			sp->a0 = -1;
			sp->outoff = 1;
			sp->color = 0;
		}
		if (done = Decode2D(sp,sp->code,sp->ref+1))
			return done;
		sp->mode = MODE_EOL;
		break;
	}
	return sp->outoff;
}

void DecodeFree(DecodeStatePtr dp)
{
	smfree(dp);
}

void DecodeSetData(DecodeStatePtr sp,char *buf,int amt)
{
	sp->buf = sp->stream = (unsigned char*)buf;
	sp->bufsize = sp->amt = amt;
}

void DecodeFinish(DecodeStatePtr sp,iImagePtr im)
{
/* no need to suck up the remaining eols from the stream */
	RlLinePtr cr;
	int4 i,w,n;

	w = im->width;
	cr = im->d.im.rl; /* go back and fix the -1's */
	for (i=0;i<im->height;i++,cr++) {
		n = cr->size - 2;
		im->d.im.trncnt += n << 1;
		if (cr->run[n].dn != cr->run[n].up)
			im->d.im.trncnt += 1;
	}
#if 0
/* There isn't much of a reason to suck up the remaining eols. */
	int4 done;
	
	switch (sp->type) {
	case CCITT3:
		if (sp->mode != MODE_EOD) {
			sp->eols = 6;
			sp->mode = MODE_EOD;
			sp->len = 0;
		}
		while (sp->eols) {
			if (done = skiptoeol(sp))
				return done;
			sp->len = 0;
			sp->eols--;
		}
		break;
	case CCITT4:
		if (sp->mode != MODE_EOD) {
			sp->eols = 6;
			sp->mode = MODE_EOD;
			sp->len = 0;
		}
		while (sp->eols) {
			if (done = skiptoeol(sp))
				return done;
			sp->len = 0;
			sp->eols--;
		}
		break;
	}
	return 0;
#endif
}

/*
 * Write a variable-length bit-value to
 * the output stream.  Values are
 * assumed to be at most 16 bits.
 */
static void putcodenorm(EncodeState *sp,tableentry const *te)
{
	int t;

	if (te->length < (uint4) sp->bitpos) {
		sp->data |= (int4)te->code << (sp->bitpos - te->length);
		sp->bitpos -= te->length;
	} else if (te->length > (uint4) sp->bitpos) {
		t = te->length - sp->bitpos;
		*sp->stream++ = sp->data | ((int4)te->code >> t);
		sp->bitpos = 32 - t;
		sp->data = (int4)te->code << sp->bitpos;
	} else {
		*sp->stream++ = sp->data | te->code;
		sp->data = 0;
		sp->bitpos = 32;
	}
}

static void putcodeswap(EncodeState *sp, tableentry const *te)
{
	int4 t;
	char *p,*q;
	
	if (te->length < (uint4) sp->bitpos) {
		sp->data |= (int4)te->code << (sp->bitpos - te->length);
		sp->bitpos -= te->length;
	} else if (te->length > (uint4) sp->bitpos) {
		t = te->length - sp->bitpos;
		sp->data |= ((int4)te->code >> t);
		p = (char*)sp->stream++;
		q = (char*)&sp->data;
		*p++ = q[3];*p++ = q[2];*p++ = q[1];*p++ = q[0];
		/* bogus bitpos is short, t is long */
		sp->bitpos = 32 - t;
		sp->data = (int4)te->code << sp->bitpos;
	} else {
		sp->data |= te->code;
		p = (char*)sp->stream++;
		q = (char*)&sp->data;
		*p++ = q[3];*p++ = q[2];*p++ = q[1];*p++ = q[0];
		sp->data = 0;
		sp->bitpos = 32;
	}
}

/*
 * Write the sequence of codes that describes
 * the specified span of zero's or one's.  The
 * appropriate table that holds the make-up and
 * terminating codes is supplied.
 */
static void putspan(EncodeState *sp,int4 span,tableentry const *tab)
{
	while (span >= 2624) {
		tableentry const *te = &tab[63 + (2560 >> 6)];
		putcode(sp, te);
		span -= te->runlen;
	}
	if (span >= 64) {
		tableentry const *te = &tab[63 + (span >> 6)];
		putcode(sp, te);
		span -= te->runlen;
	}
	putcode(sp, &tab[span]);
}

/*
 * Reset encoding state at the start of a strip.
 */
 
EncodeStatePtr EncodeAlloc(void)
{
	return (EncodeStatePtr)smalloc(sizeof(EncodeState));
}

void EncodeFree(EncodeStatePtr dp)
{
	smfree(dp);
}

void EncodeSetData(EncodeStatePtr es,char *buf,int len)
{
	es->start = es->stream = (unsigned long *)buf;
}

int EncodeGetAmt(EncodeStatePtr es)
{
	return (long)es->stream - (long)es->start;
}

void EncodeInit(EncodeStatePtr sp,iImagePtr im,iDAFSCompType type)
{
	sp->width = im->width;
	sp->bitpos = 32;
	sp->data = 0;
	sp->type = type;
	sp->blank[0] = 0;
	/* bogus blank is short, width is long -- tom */
	sp->blank[1] = sp->blank[2] = im->width;
	sp->blank[3] = -1;
}
/*
 * 1d-encode a row of pixels.  The encoding is
 * a sequence of all-white or all-black spans
 * of pixels encoded with Huffman codes.
 */
static void Encode1D(EncodeState *sp,short *bp)
{
	int4 last,span;
	
	last = 0;
	while (*bp >= 0) {
		span = -last;
		span += (last = *bp++);
		putspan(sp, span, RlWhiteCodes);
		span = -last;
		span += (last = *bp++);
		if (!span) break;
		putspan(sp, span, RlBlackCodes);
	}
}

/*
 * 2d-encode a row of pixels.  Consult the CCITT
 * documentation for the algorithm.
 */
static void Encode2D(EncodeState *sp,short *bp,short *rpp)
{
	int4 a0 = 0;
	int4 a1 = *bp;
	int4 a2,b2,b1;
	int4 color = 0;
	short *rp = rpp;
	
	for (;;) {
		b1 = *rp;
		b2 = rp[1];
		if (b1 < 0) b2 = b1 = sp->width;
		else if (b2 < 0) b2 = b1;
		if (b2 >= a1) {
			int4 d = b1 - a1;
			if (!(-3 <= d && d <= 3)) {	/* horizontal mode */
				a2 = bp[1];
				if (a2 < 0) a2 = sp->width;
				putcode(sp, &horizcode);
				if (a0 + a1 == 0 || !color) {
					putspan(sp, a1 - a0, RlWhiteCodes);
					putspan(sp, a2 - a1, RlBlackCodes);
				} else {
					putspan(sp, a1 - a0, RlBlackCodes);
					putspan(sp, a2 - a1, RlWhiteCodes);
				}
				a0 = a2;
				bp += 2;
				a1 = *bp;
				if (a1 < 0) a1 = sp->width;
			} else {/* vertical mode */
				putcode(sp, &vcodes[d + 3]);
				a0 = a1;
				bp++;
				a1 = *bp;
				if (a1 < 0) a1 = sp->width;
#if 0
				if (color = !color) rp--;
				else rp++;
#else
				color = !color;
				if (*rp >= 0 && *rp < a0) rp++;
				else rp--;
#endif
			}
		} else {	/* pass mode */
			putcode(sp, &passcode);
			a0 = b2;
		}
		if (a0 >= sp->width) return;
		while (*rp >= 0 && *rp <= a0)
			rp += 2;
	}
}
/*
 * Encode a buffer of pixels.
 */
void Encode(EncodeStatePtr sp,short *bp,short *ref)
{
	bp++;
	if (sp->type == CCITT3) {
		putcode(sp,&eolcode);
		Encode1D(sp, bp);
	} else {
		if (!ref)ref = sp->blank;
		Encode2D(sp, bp, ref+1);
	}
}

void EncodeFlush(EncodeStatePtr sp,iImagePtr im,int willdie)
{
	int4 i;
	short swap = 1;
	char *p,*q;
	
	switch (sp->type) {
	case CCITT3:
		for (i=0;i<6;i++)
			putcode(sp,&eolcode);
		break;
	case CCITT4:
		for (i=0;i<2;i++)
			putcode(sp,&eolcode);
		break;
	default: break;
	}
	if (sp->bitpos < 32) {
		if (*(char*)&swap == 1) {
			p = (char*)sp->stream++;
			q = (char*)&sp->data;
			*p++ = q[3];*p++ = q[2];*p++ = q[1];*p++ = q[0];
		} else *sp->stream++ = sp->data;
	}
}

