/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "rafutil.h"

#include "dafs.h"


struct ErrorStruct {
	iDAFSError error;
	char *error_string;
};

typedef struct ErrorStruct ErrorStruct;


/* 
// here are where error structs are used
// 32 * 8 = 256 bytes
*/
ErrorStruct error_array[] = {
	{DAFSOK,"OK"},
	{MallocFailError,"Out of memory"},
	{NonExistentFileError,"This file does not exist"},
	{WriteLockedFileError,"This file is write locked"},
	{ProtectedFileError,"This file is protected"},
	{DiskFullError,"The disk is full"},
	{DAFSWriteModeError,"The DAFSWriteMode is incorrect"},
	{WrongPropType,"Wrong type of property"},
	{WriteError,"There was an error in writing"},
	{ReadError,"There was an error in reading"},
	{BadFileType,"The file type was incorrect"},
	{UserStop,"There was a user stop"},
	{NonExistentProp,"That property does not exist"},
	{UnknownDAFSError,"Unknown error"},
	{NotEnoughData,"There was not enough data to do this operation"},
	{NotInitialized,"Something wasn't initialized"},
	{BadRotError,"Invalid rotation"},
	{CantRotateError,"Couldn't rotate entity"},
	{BadScaleError,"Invalid value for scale image"},
	{WrongImageFormat,"Wrong image format"},
	{WrongFileType, "Wrong file type"},
	{AlreadyOrEntity ,"The entity was already an iOr entity"},
	{DiffImageType ,"The images are different type"},
	{BadBoundsError, "The bounds given were bad"},
	{BadEntityPassed,"A bad entity was passed"},
	{CantLimitError,"Can't limit posset to nothing"},
	{OutOfBoundsError,"Parameter out of bounds"},
	{PosSetShareedError,"Posset can't be set on Shareed entities"},
	{FailedTest,"Routine failed to pass the test"},
	{NoMoreData,"i_GetDataFromImage has no more data to give"},
	{CantWrite32D,"Dafslib can't compress to CCITT Group 3 2-D."},
	{CantParseString,"String or UString passed as input could not be parsed."},
	{0,0}
};
 

Dll const char * APIENTRY
i_GetError(iDAFSError error)
{	
	int4 i;
	for (i=FirstError;i<LastError;i++)
		if(error == error_array[i+ABS(FirstError)].error)
			return error_array[i+ABS(FirstError)].error_string;
	return "Not a DAFS Error Code";
}

