/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "c-lib.h"
#include "str-lib.h"

#include "rafutil.h"
#include "entity.h"
#include "rlimage.h"

#include "dafs.h"
#include "smemory.h"

#define CHUNK 8192
#define MAXWID 32767

extern int sprintf(char *, const char *, ...);


static void
flushCachedImage(iImagePtr sp)
{
	iImagePtr cur;
	if (sp->type == Image && (cur = sp->d.im.next)) {
		sp->d.im.next = 0;
		freeimage(cur);
	}
}


static int4 
bitmfillline(RafFile  in, char *inbuf, int2 need, int4 *count)
{
	char *dp = (char *) in;
	dp += need * (*count)++;
	memcpy(inbuf, dp, need);
	return 0;
}


Dll iImagePtr APIENTRY
i_BitMap2Image(iBox b, int2 dw, char *dp)
{
	iImagePtr np, np2;
	iDAFSError e;
	
	if ((dw << 3) > MAXWID) return 0;
	
	dp += dw * b.y;
	if (e = do0d(&np,(RafFile ) dp, dw << 3, b.h, bitmfillline,0))
		return 0;
	b.y=0;
	np2 = rlgetzone(np, b);
	freeimage(np);
	if(np2){
		np2->use++;
		np2->dpi=0;
	}
	return np2;
}

Dll iDAFSError APIENTRY
i_Image2BitMap(iImagePtr sp, iBox b, iFixedPoint mag, int2 *dataWid, char **bits)
{
	char *tiram;
	int4 i, j,size;
	int2 offx,offy;
	iImagePtr saved = 0;
	iBox temp;
	int2 run[4];
	iDAFSError e=DAFSOK;

	if (!sp)
		return DAFSOK;

	if (e = i_UncompressImage(sp)) return e;
	if (b.w) {
		saved = sp;
		temp.x = (int2)(i_INT2_TO_FP(b.x) / mag);
		temp.y = (int2)(i_INT2_TO_FP(b.y) / mag);
		temp.w = (int2)(i_INT2_TO_FP(b.w) / mag);
		temp.h = (int2)(i_INT2_TO_FP(b.h) / mag);
		temp.w += 2;
		temp.h += 2;
		if (temp.w + temp.x > sp->width)
			temp.w = sp->width - temp.x;
		if (temp.h + temp.y > sp->height)
			temp.h = sp->height - temp.y;
		if (temp.h < 1 || temp.w < 1)
			return BadBoundsError;
		sp = rlgetzone(sp, temp);
		if (!sp)
			return 0;
	}
	*dataWid = ((b.w + 0x1f) & ~0x1f) >> 3;
	size = ((int4)b.h * *dataWid) + 4;
#ifdef SYS_OS_WIN16
	if (size > 0xffffL) halt("This routine can't return a HUGE array\n");
#endif
	if (!*bits) {
		*bits = (char *) smalloc(size);
		/* 4 more for Run2Ras */
		if (!*bits)
			return MallocFailError;
	}
	tiram = *bits;
	offy = b.y - i_FP_TO_INT2(temp.y * mag);
	offx = b.x - i_FP_TO_INT2(temp.x * mag);
	for (i = offy; i < (b.h + offy); i++) {
		j = (int2)(i_INT2_TO_FP(i) / mag);
		if (j >= sp->height) {
			run[0] = 0;
			run[1] = b.w;
			run[2] = b.w;
			run[3] = -1;
			for (; i < (b.h + offy); i++) {
				Run2Ras(&run[1], tiram);
				tiram += *dataWid;
			}
			break;
		}
		Run2RasMul((int2 *) sp->d.im.rl[j].run, tiram, mag, offx);
		tiram += *dataWid;
	}
	if (saved)
		freeimage(sp);
	return e;
}

Dll iImagePtr APIENTRY
i_CloneEntityImage(const iEntityPtr en)
{
	iBox bnds;
	iImagePtr im,t;
	im = i_GetImage(en);
	if (!im)
		return 0;
	if (i_UncompressImage(im)) return 0;
	bnds = i_GetEnclosingBox(en);
	if (bnds.x < 0)
		return 0;
	if (bnds.y < 0)
		return 0;
	if (bnds.x + bnds.w > im->width)
		return 0;
	if (bnds.y + bnds.h > im->height)
		return 0;
	t = rlgetzone(im,bnds);

	if(t)
		t->use++;
	return t;
}

Dll int4 APIENTRY
i_GetImageWidth(const iImagePtr image)
{
	if (!image)
		return -1;
	return image->width;
}

Dll int4 APIENTRY
i_GetImageHeight(const iImagePtr image)
{
	if (!image)
		return -1;
	return image->height;
}

Dll iBox APIENTRY
i_GetImageBounds(const iImagePtr image) {
	iBox box;
	box.x=0;
	box.y=0;
	box.h=0;
	box.w=0;
	if (image){
		box.w = image->width;
		box.h = image->height;
	}
	return box;
}

Dll int2 APIENTRY
i_GetImageRes(const iImagePtr image)
{
	if (!image)
		return -1;
	return (image->dpi);
}

Dll void APIENTRY
i_SetImageRes(const iImagePtr image, int2 resolution)
{
	if (image)
		image->dpi = resolution;
}



Dll iDAFSError APIENTRY
i_RotateImage(iImagePtr image, int2 rot)
{
	iImagePtr sp,next,prev;
	iDAFSError e;
	iImage temp;


	while(rot>180)
		rot-=360;
	while(rot<=-180)
		rot+=360;
	
	if (!rot)
		return DAFSOK;
		
	if (rot!=90 && rot!=-90)
		return BadRotError;

	if (e=i_UncompressImage(image)) return e;

	for (sp=image->d.im.next; sp && sp->d.im.rot!=rot; sp=sp->d.im.next)
		;

	if (!sp) {
		sp = rotate(image, rot);
		if (!sp)
			return MallocFailError;
		next=image->d.im.next;
		prev=0;
	} else {
		next=sp->d.im.next;
		prev=sp->d.im.prev;
		/* remove sp from list */
		if (next)
			next->d.im.prev=prev;
		prev->d.im.next=next;
	}

	temp=*sp;
	*sp=*image;
	*image=temp;
	image->id=sp->id;
	image->use=sp->use;

	/* straighten out pointers */
	if (next || prev) {
		image->d.im.prev=0;
		image->d.im.next=sp;
		sp->d.im.prev=image;
		if (sp->d.im.next)
			sp->d.im.next->d.im.prev=sp;
		if (next)
			next->d.im.prev=sp;
			
	} else {
		image->d.im.next=sp;
		sp->d.im.prev = image;
	}

	for (sp=image; sp; sp=sp->d.im.next) {
		sp->d.im.rot-=rot;
		while(sp->d.im.rot>180)
			sp->d.im.rot-=360;
		while(sp->d.im.rot<=-180)
			sp->d.im.rot+=360;
	}

#if 0
	image->d.im.prev = sp;
	image->d.im.rot = -rot;
	sp->d.im.rot = 0;
#endif
	return DAFSOK;
}


static int4 count_image(iEntityPtr en,iImagePtr img)
{
	iImagePtr img1;
	int4 count=0;
	iEntityPtr cur;
	
	img1=i_GetImage(en);
	if (img1==img)
		count++;
	for (cur=i_GetChildEntity(en); cur; cur=i_GetNextEntity(cur))
		count+=count_image(cur,img);
	return count;
}

static int4 ok_images(iEntityPtr en, iEntityPtr par)
{
	iImagePtr img;
	int4 ok=1;
	iEntityPtr cur;
	
	img=i_GetImage(en);
	if (!img || img->id)
		for (cur=i_GetChildEntity(en); cur && ok; cur=i_GetNextEntity(cur)) {
			ok=ok && ok_images(cur,par);
		}
	else {
		if (count_image(par,img)!=img->use)
			return 0;
		img->id=1;
	}
	return ok;
}


Dll iDAFSError APIENTRY
i_InvertImage(iImagePtr sp)
{
	RlLinePtr l;
	int4 row,size,i;
	int2 *p,*q, *temp, w;
	iDAFSError e;
	
	if (e = i_UncompressImage(sp)) return e;
	flushCachedImage(sp);
	w=sp->width;
	l = sp->d.im.rl;
	for (row=0;row<sp->height;l++,row++) {
		size = l->size;
		if (l->run[0].up) {
			if (l->run[size-2].up != w) {
				l->size++;
				if (l->size > l->asize) {
					temp = (int2*)smrealloc(l->code,SL + (ST * (l->asize+1)));
					if (!temp) return MallocFailError;
					l->asize++;
					l->code = temp;
					l->run = (transition*)&l->code[1];
				}
				l->run[size-1].up = w;
				l->run[size-1].dn = w;
				l->run[size].up = -1;
			}
			for (p=&l->run[size-2].dn,q=p-1;*p;) {
				*p-- = *q--;
				*p-- = *q--;
			}
			*p = 0;
		} else {
			for (p=&l->run[0].up,q=p+1,i=0;*q<w;) {
				*p++ = *q++;
				*p++ = *q++;
				i++;
			}
			if (*p == w) {
				*p = -1;
				l->size--;
			} else{
				*p = w;
			}
			if(size>i+2&&*q==*(q+1)){
				l->size--;
				q++;
				*q++=-1;
				*q=0;
			}
		}
	}
	return DAFSOK;
}

Dll iDAFSError APIENTRY
i_MirrorImage(iImagePtr sp)
{
	RlLinePtr l1,l2;
	RlLine temp;
	iDAFSError e;
	
	if (e = i_UncompressImage(sp)) return e;
	flushCachedImage(sp);
	l1 = sp->d.im.rl;
	l2 = l1 + sp->height - 1;
	while (l2 > l1) {
		temp = *l1;
		*l1 = *l2;
		*l2 = temp;
		l1++;
		l2--;
		
	}
	return DAFSOK;
}
