/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */


#include "rafport.h"
#include "rafutil.h"

#include "str-lib.h"
#include "smemory.h"
#include "entity.h"
#include "dafs.h"


struct cornertype {
	unsigned short x,y;
	int2 corner;
};

struct points {
	int2 n;
	struct cornertype *point;
};


enum corner_type {TL, TR, BL, BR};
#ifdef SYS_OS_WIN16
typedef int4 corner_type;
#else
typedef enum corner_type corner_type;
#endif


typedef struct cornertype cornertype;
typedef struct points points;
typedef struct cornertype *cornertypeptr;
typedef struct points *pointsPtr;


Dll iDAFSError APIENTRY
i_ExpandOrthogon(iOrthogonPtr original)
{
	iOrthogonPtr poly;
	iPoint *temp;
	iDAFSError e=DAFSOK;

	if (original && (original->n == 2)){
		poly=i_NewOrthogon(2);
		if (!poly)
			return MallocFailError;
			
		poly->point[0] = original->point[0];
		poly->point[1] = original->point[1];
		temp = (iPoint*)smrealloc(original->point,4*sizeof(iPoint));
		if (!temp) {
			e=MallocFailError;
			goto error;
		} else
			original->point=temp;
		original->n = 4;
		original->point[0].x = original->point[3].x = poly->point[0].x;
		original->point[0].y = original->point[1].y = poly->point[0].y;
		original->point[1].x = original->point[2].x = poly->point[1].x+poly->point[0].x;
		original->point[2].y = original->point[3].y = poly->point[1].y+poly->point[0].y;
error:
		i_DisposeOrthogon(poly);
	}
	return e;
}

/*
// bogus this is the most unbelieveably long function -- 269 lines
*/
Dll iDAFSError APIENTRY
i_BoxList2OrthogonList(iBoxListPtr reg, iOrthogonListPtr *bp)
{
	corner_type mode;
	pointsPtr corners;
	iOrthogonPtr polygon;
	iOrthogonListPtr border;
	int2 i,k;
	int4 index=0, index2=1, stop=0, closest1_i,closest2_i;
	int4 *used, start=0;
	static char table[4][5] = {{2,1,3,4,0},{3,2,4,1,0},{1,2,4,3,0},{2,4,3,1,0}};
	cornertypeptr current, new, closest1, closest2;
	cornertype min={0,0,4}, max={0xFFFF,0xFFFF,4};
	cornertypeptr temp;
	
	corners = (pointsPtr)smalloc(sizeof(points));
	corners->n=4*reg->numRects;
	corners->point = (cornertype*)smalloc(sizeof(cornertype)*corners->n);
	used = (int4*)smalloc(sizeof(int4)*corners->n);
	temp = corners->point;
	for(i=0; i<corners->n; i+=4){
		for (k=0; k<4; k++){
			temp[i+k].corner=k;
			used[i+k]=0;
		}
		temp[i].x = temp[i+2].x = reg->rects[index].x1;
		temp[i].y = temp[i+1].y = reg->rects[index].y1;
		temp[i+2].y = temp[i+3].y = reg->rects[index].y2;
		temp[i+1].x = temp[i+3].x = reg->rects[index++].x2;
	}

	*bp = NULL;
	border = i_NewOrthogonList();
	if(!border)
		return MallocFailError;

	for(; !stop; ) {
		index=start;
		polygon = i_NewOrthogon(corners->n);
		if (!polygon)
			return MallocFailError;
		polygon->point[0].x = corners->point[start].x;
		polygon->point[0].y = corners->point[start].y;
		polygon->n = 1;
		mode = (corner_type)corners->point[index].corner;
		for (k=1; (index2!=start); k++){
			current = &(corners->point[index]);
			index2 = index;
			switch(mode){
			case TL:	
				closest1 = &max;
				closest2 = &max;
				for (i=0; i<corners->n; i++){
					new = &(corners->point[i]);
					if ((i!= index) && (current->y==new->y))
						if (new->corner<BL){
							if (new->x > current->x)
								if ((new->x < closest1->x)){
									closest1_i = i;
									closest1 = new;
								} else {
									if (new->x == closest1->x){
										if (table[TL][new->corner] > 
											table[TL][closest1->corner]){
											closest1_i = i;
											closest1 = new;
										}
									}
								}
						} else
							if (new->x > current->x)
								if ((new->x < closest2->x)){
									closest2_i = i;
									closest2 = new;
								} else {
									if (new->x == closest2->x){
										if (table[TL][new->corner] > 
											table[TL][closest2->corner]){
											closest2_i = i;
											closest2 = new;
										}
									}
								}
				}
				if (closest2->corner==BR)
					mode=BR;
				else {
					if (closest2->corner<4&&closest2->x<=closest1->x){
						index2=closest2_i;
						mode = closest2->corner;
					} else {
						index2=closest1_i;
						mode = closest1->corner;
					}
				}
				break;
				
			case TR:	closest1 = &max;
				closest2 = &max;
				for (i=0; i<corners->n; i++){
					new = &(corners->point[i]);
					if ((i!= index) && (current->x==new->x))
						if (new->corner==TL || new->corner==BL){
							if (new->y > current->y)
								if (new->y < closest1->y){
									closest1_i = i;
									closest1 = new;
								} else {
									if (new->y == closest1->y){
										if (table[TR][new->corner] > 
											table[TR][closest1->corner]){
											closest1_i = i;
											closest1 = new;
										}
									}
								}
						} else
							if (new->y > current->y)
								if ((new->y < closest2->y)){
									closest2_i = i;
									closest2 = new;
								} else {
									if (new->y == closest2->y){
										if (table[TR][new->corner] > 
											table[TR][closest2->corner]){
											closest2_i = i;
											closest2 = new;
										}
									}
								}
					
					
				}
				if (closest1->corner==BL)
					mode=BL;
				else {
					if (closest1->corner<4&&closest1->y<=closest2->y){
						index2=closest1_i;
						mode = closest1->corner;
					} else {
						index2=closest2_i;
						mode = closest2->corner;
					}
				}
				break;
				
			case BR:	closest1 = &min;
				closest2 = &min;
				for (i=0; i<corners->n; i++){
					new = &(corners->point[i]);
					if ((i!= index) && (current->y==new->y))
						if (new->corner<BL){
							if (new->x < current->x)
								if (new->x > closest1->x){
									closest1_i = i;
									closest1 = new;
								} else 
									if (new->x == closest1->x){
										if (table[BR][new->corner] > 
											table[BR][closest1->corner]){
											closest1_i = i;
											closest1 = new;
										}
									}
						} else
							if (new->x < current->x)
								if ((new->x > closest2->x)){
									closest2_i = i;
									closest2 = new;
								} else {
									if (new->x == closest2->x){
										if (table[BR][new->corner] > 
											table[BR][closest2->corner]){
											closest2_i = i;
											closest2 = new;
										}
									}
								}
				}
				if (closest1->corner==TL)
					mode=TL;
				else {
					if (closest1->corner<4&&closest1->x>=closest2->x){
						index2=closest1_i;
						mode = closest1->corner;
					} else {
						index2=closest2_i;
						mode = closest2->corner;
					}
				}
				break;
				
			case BL:	closest1 = &min;
				closest2 = &min;
				for (i=0; i<corners->n; i++){
					new = &(corners->point[i]);
					if ((i!= index) && (current->x==new->x))
						if (new->corner==1 || new->corner==3){
							if (new->y < current->y)
								if (new->y > closest1->y){
									closest1_i = i;
									closest1 = new;
								} else 
									if (new->y == closest1->y){
										if (table[BL][new->corner] > 
											table[BL][closest1->corner]){
											closest1_i = i;
											closest1 = new;
										}
									}
						} else
							if (new->y < current->y)
								if ((new->y > closest2->y)){
									closest2_i = i;
									closest2 = new;
								} else {
									if (new->y == closest2->y){
										if (table[BL][new->corner] > 
											table[BL][closest2->corner]){
											closest2_i = i;
											closest2 = new;
										}
									}
								}
				}
				if (closest1->corner==TR)
					mode=TR;
				else {
					if (closest1->corner<4&&closest1->y>=closest2->y){
						index2=closest1_i;
						mode = closest1->corner;
					} else {
						index2=closest2_i;
						mode = closest2->corner;
					}
				}
				break;
			}
			if((used[index2])&&(index!=index2)){
				used[index2]=-1;
				start=index=index2;
			} else
				if ((index2!=start)&&(index!=index2)){
					polygon->point[polygon->n].x=	corners->point[index2].x;
					polygon->point[polygon->n++].y=	corners->point[index2].y;
					used[index2]=1;
					index = index2;
				}
		}
		if (used[index]<0){
			i_DisposeOrthogon(polygon);
			stop=1;
		} else {
			used[start]=1;
			if (polygon->n>2){
				if (border->num_poly++)
					border->poly = (iOrthogonPtr*)smrealloc(border->poly,sizeof(iOrthogonPtr)*border->num_poly);
				else
					border->poly = (iOrthogonPtr*)smalloc(sizeof(iOrthogonPtr));
				border->poly[border->num_poly-1] = polygon;
			} else {
				i_DisposeOrthogon(polygon);
			}
			for (start=0; (start<corners->n) && used[start];start++);
			if (start==corners->n || used[start]){
				stop=1;
			} else
				index=start;
		}
	}
	smfree(corners->point);
	smfree(corners);
	smfree(used);

	*bp = border;
	return DAFSOK;
}					


Dll void APIENTRY
i_CompressOrthogon(iOrthogonPtr org)
{
	iOrthogonPtr poly;
	
	if (org && (org->n==4)){
		poly=i_CopyOrthogon(org);
		org->point = (iPoint*)smrealloc(org->point,2*sizeof(iPoint));
		org->n = 2;
		org->point[0].x = poly->point[0].x;
		org->point[0].y = poly->point[0].y;
		org->point[1].x = poly->point[3].x-poly->point[0].x;
		org->point[1].y = poly->point[3].y-poly->point[0].y;
		i_DisposeOrthogon(poly);
	}
}

Dll iDAFSError APIENTRY
i_Box2Orthogon(iBox box, iOrthogonPtr *ortho)
{
	iOrthogonPtr poly;
	
	*ortho = NULL;

	if (!box.w){
		if(!(poly = i_NewOrthogon(0)))
			return MallocFailError;
	}
	else {
		if(!(poly = i_NewOrthogon(4)))
			return MallocFailError;
		if (poly){
			poly->point[0].x = poly->point[3].x = box.x;
			poly->point[0].y = poly->point[1].y = box.y;
			poly->point[1].x = poly->point[2].x = box.x+box.w;
			poly->point[3].y = poly->point[2].y = box.y+box.h;
		}
	}

	*ortho = poly;
	return DAFSOK;
}

Dll void APIENTRY
i_Orthogon2Box(iOrthogonPtr poly, iBox *box)
{
	iBox b={0,0,0,0};
	int4 i;
	int4 n=poly->n;
	iPoint *pt;
	short maxx,maxy,minx,miny;

	if (n<2){
		*box = b;
		return;
	}
		
	if (n == 2){
		b.x=poly->point[0].x;
		b.y=poly->point[0].y;
		b.w=poly->point[1].x;
		b.h=poly->point[1].y;
	} else {
		pt=poly->point;
		maxx=pt->x;
		minx=pt->x;
		maxy=pt->y;
		miny=pt->y;
		
		for (i=1,pt++; i<n; i++,pt++){
			if (pt->x>maxx)
				maxx = pt->x;
			else if (pt->x<minx)
				minx = pt->x;
			
			if (pt->y>maxy)
				maxy = pt->y;
			else if (pt->y<miny)
				miny = pt->y;
		}
		b.x = minx;
		b.y = miny;
		b.w = maxx-minx;
		b.h = maxy-miny;
	}

	*box = b;
}

Dll void APIENTRY
i_OrthogonList2Box(iOrthogonListPtr border, iBox *ret_box)
{
	short maxx=0,maxy=0;
	iBox b, box;
	int4 i;

	b.x=0x7fff;
	b.y=0x7fff;
	for (i=0; i<border->num_poly; i++){
		i_Orthogon2Box(border->poly[i], &box);
		if (maxx<box.x+box.w)
			maxx=box.x+box.w;
		if (maxy<box.y+box.h)
			maxy=box.y+box.h;
		if (b.y>box.y)
			b.y=box.y;
		if (b.x>box.x)
			b.x=box.x;
	}
	b.w=maxx-b.x;
	b.h=maxy-b.y;

	*ret_box = b;
}

Dll iDAFSError APIENTRY
i_Box2OrthogonList(iBox box, iOrthogonListPtr *orthos)
{
	iDAFSError e;
	iOrthogonListPtr border = 0;
	
	*orthos = NULL;
	if (box.w) {
		if(!(border = i_NewOrthogonList()))
			return MallocFailError;
		
		if(!(border->poly = (iOrthogonArrayPtr) smalloc(sizeof(iOrthogonPtr)))) {
			i_DisposeOrthogonList(border);
			return MallocFailError;
		}

		if(e = i_Box2Orthogon(box, &border->poly[0])){
			i_DisposeOrthogon(border->poly[0]);
			i_DisposeOrthogonList(border);
			return e;
		}

		border->num_poly=1;
	}

	*orthos = border;
	return DAFSOK;
}

/*	This returns the intersection of two boxes.  If they don't intersect
	the width of the returned box will be 0.
	*/
Dll void APIENTRY
i_IntersectBox(iBox a,iBox b,iBox *c)
{	iBox temp;

	if (a.x < b.x) temp.x = b.x;
	else temp.x = a.x;
	if (a.y < b.y) temp.y = b.y;
	else temp.y = a.y;
	if (a.x + a.w > b.x + b.w) temp.w = b.x + b.w;
	else temp.w = a.x + a.w;
	temp.w -= temp.x;
	if (temp.w <= 0) {
		temp.w = 0;
		*c=temp;
		return;
	}
	if (a.y + a.h > b.y + b.h) temp.h = b.y + b.h;
	else temp.h = a.y + a.h;
	temp.h -= temp.y;
	if (temp.h <= 0) temp.w = 0;
	*c=temp;
}

Dll void APIENTRY
i_UnionBox(iBox b1,iBox b2, iBox *ret)
{
	iBox ob;	
	
	if (!b1.w){
		*ret=b2;
		 return;
	}
	if (!b2.w){
		*ret=b1;
		return;
	}
	ob = b1;
	ob.w += ob.x;
	ob.h += ob.y;
	if (b2.x < ob.x)
		ob.x = b2.x;
	if (b2.y < ob.y)
		ob.y = b2.y;
	if (b2.x + b2.w > ob.w)
		ob.w = b2.x + b2.w;
	if (b2.y + b2.h > ob.h)
		ob.h = b2.y + b2.h;
	ob.w -= ob.x;
	ob.h -= ob.y;
	*ret=ob;
}

