/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "c-lib.h"
#include "str-lib.h"

#include "rafutil.h"
#include "entity.h"
#include "rlimage.h"

#include "dafs.h"
#include "smemory.h"

#define CHUNK 8192
#define MAXWID 32767

extern int sprintf(char *, const char *, ...);

static iDAFSError readheader(RafFile  in, int2 *xp, int2 *yp, int2 *wp, int2 *hp, int2 *comp,int2 *rp)
{
	char buf[50], c;
	int4 e, i;
	int2 x,y,w,h,r;
	do {
		e = riRead(buf, 1, 1, in);
		if (e != 1) return ReadError;
		if (buf[0] == 0xc) {
			*comp = -1; /* REOF */
			return DAFSOK;
		}
	} while (buf[0] == 0xb);
	w = 2560;
	h = 0;
	x = 0;
	y = 0;
	*comp = buf[0] - '0';
	
	e = riRead(buf, 1, 1, in);
	if (e != 1) return ReadError;
	if (buf[0] == 0xa) return DAFSOK;
	
	buf[5] = 0;
	e = riRead(&buf[1], 1, 4, in);
	if (e != 4) return ReadError;
	x = atoi(buf);
	
	e = riRead(buf, 1, 5, in);
	if (e != 5) return ReadError;
	y = atoi(buf);
	
	e = riRead(buf, 1, 5, in);
	if (e != 5) return ReadError;
	w = atoi(buf);
	
	e = riRead(buf, 1, 5, in);
	if (e != 5) return ReadError;
	h = atoi(buf);
	
	for (i=0;i<50;i++) {
		e = riRead(&buf[i],1,1,in);
		if (e != 1) return ReadError;
		if (buf[i] == ',' || buf[i] == 0xa)
			break;
	}
	c = buf[i];
	if (*comp < 3) {
		buf[i - 1] = 0;
		r = atoi(&buf[i-3]);
	} else if (*comp == 3) {
		buf[i] = 0;
		r = atoi(&buf[i-2]);
	} else
		r=4; /* in response to possibly not initializing r */
	if (!r) r=4;
	x /= r;
	y /= r;
	w /= r;
	h /= r;
	if (xp) *xp=x;
	if (yp) *yp=y;
	if (wp) *wp=w;
	if (hp) *hp=h;
	if (rp) *rp=r;
	while (c != 0xa) {
		e = riRead(&c, 1, 1, in);
		if (e != 1) return ReadError;
	}
	return DAFSOK;
}

iDAFSError newchunk(iImagePtr np,uint2 buflen)
{
	int cur = np->d.rw.nchunks;
	iChunkPtr temp;
	iDAFSError e = DAFSOK;

	if (np->d.rw.chunks)
		temp = (iChunkPtr)smrealloc(np->d.rw.chunks,sizeof(iChunk) * (cur + 1));
	else
		temp = (iChunkPtr)smalloc(sizeof(iChunk));
	if (!temp) { e = MallocFailError; goto error; }
	np->d.rw.chunks = temp;

	np->d.rw.chunks[cur].data = smalloc(buflen);
	if (!np->d.rw.chunks[cur].data) { e = MallocFailError; goto error; }
	np->d.rw.chunks[cur].len = buflen;
	np->d.rw.nchunks = ++cur;
error:
	return e;
}

static int4 
pdafillline(RafFile  in, char *inbuf, int2 need, int4 *count)
{
	char buf[5];
	int4 e,n;
	int2 have = 0;
	while (need) {
		if (!*count) {
			e = riRead(buf, 1, 1, in);
			if (e != 1)
				return -1;
			if (buf[0] < '0' || buf[0] > '9')
				return -1;
			e = riRead(&buf[1], 1, 3, in);
			if (e != 3)
				return -1;
			buf[4] = 0;
			*count = atoi(buf);
		}
		n = *count;
		if (n > need)
			n = need;

		e = riRead(&inbuf[have], 1, n, in);
		if (e != n)
			return -1;
		have += (int2)n;
		need -= (int2)n;
		*count -= (int2)n;
	}
	return 0;
}

static iDAFSError 
rawDAFSB(iImagePtr *npp,RafFile  in, int2 w, int2 h, int *swap)
{
	int4 buflen,*strips = 0,rowsper;
	int2 len,cur=0,nstrips,sw;
	uint2 sizes;
	iImagePtr np=0;
	iToken t;
	iDAFSError e=DAFSOK;
	
	np = newrawimage(w, h,CCITT4);
	if (!np) { e = MallocFailError; goto error; }
	
	e = i_ReadToken(in, &t, swap);
	
	switch(t&iTMask) {
	case LongArray:
		e = i_ReadItemAlloc(in,t,&sizes,(void**)&strips,swap);
		nstrips = sizes/sizeof(int4);

		e = i_ReadToken(in, &t, swap); /* this used to be a long */
		if ((t&iTMask) != ShortNShort) {e = ReadError; goto error; }
		if (sw = *swap) *swap = 0; /* turn off swapping */
		e = i_ReadItem(in,t,sizeof(int4),&sizes,(void*)&rowsper,swap);
		if (*swap = sw) rowsper = i_SwapL(rowsper);

		e = i_ReadToken(in, &t, swap);
		if ((t&iTMask) != CCITT4Chunk) {e = ReadError; goto error; }
		if ((e = i_ReadSize(in, t, &buflen,swap)))
			return 0;
		while (buflen) {
			len = strips[cur];
			if (e = newchunk(np,len))
				goto error;
			if (len != riRead(np->d.rw.chunks[cur].data,1,len,in))
				{ e = ReadError; goto error; }
			buflen -= len;
			cur++;
		}
		e = i_ReadToken(in, &t, swap);
		if ((t&iTMask) != EndImage) {e = ReadError; goto error; }
		if (e = i_uncompressimage(np,rowsper))
			goto error;
		break;
	case CCITT4Chunk: 
		 do {
			if (e = i_ReadSize(in, t, &buflen,swap))
				goto error;
			while (buflen) {
				if (buflen > CHUNK)
					len = CHUNK;
				else len = (int2)buflen;
				if (e = newchunk(np,len))
					goto error;
				if (len != riRead(np->d.rw.chunks[cur].data,1,len,in))
					{ e = ReadError; goto error; }
				buflen -= len;
				cur++;
			}
			e = i_ReadToken(in, &t, swap);
			if(e)
				goto error;
		} while ((t&iTMask) != EndImage);
		break;
	}
	*npp = np;
	np = 0;
error:
	if (strips) smfree(strips);
	if (np) freeimage(np);
	return e;
}

static iDAFSError rawpda(iImagePtr *npp,RafFile  in, int2 w, int2 h,int2 comp,int2 res)
{
	int2 cur=0,buflen;
	iImagePtr np=0;
	char buf[5];
	iDAFSError e = DAFSOK;
	
	np = newrawimage(w, h,comp);
	if (!np) { e = MallocFailError; goto error; }
	np->dpi = res;
	if (riRead(buf, 1, 1, in) != 1)  { e = ReadError; goto error; }
	
	while (buf[0] >= '0' && buf[0] <= '9') {
		if (riRead(&buf[1], 1, 3, in) != 3) { e = ReadError; goto error; }
		buf[4] = 0;
		buflen = atoi(buf);

		if (e = newchunk(np,buflen))
			goto error;
		
		if (riRead(np->d.rw.chunks[cur].data, 1, buflen, in) != buflen)
			{ e = ReadError; goto error; }
		cur++;
		if (riRead(buf, 1, 1, in) != 1)  { e = ReadError; goto error; }
	}
	*npp = np;
	np = 0;
error:
	if (np) freeimage(np);
	return e;
}

iDAFSError 
do0d(iImagePtr *npp,RafFile  in, int2 w, int2 h,
	 int4 (*fill) (RafFile , char *, int2, int4 *),int4 invert)
{
	int4 count,oddwid;
	int2 w8,size,row;
	RlLinePtr l;
	char *inbuf=0;
	int2 *code=0;
	iImagePtr np=0;
	iDAFSError e = DAFSOK;
	RlLinePtr temp;
		
	w8 = (w+7) >> 3;
	oddwid = w != (w8 << 3);
	inbuf = smalloc(w8);
	if (!inbuf) { e = MallocFailError; goto error; }
	code = (int2 *) smalloc(sizeof(int2) * (w + 4));
	if (!code) { e = MallocFailError; goto error; }
	count = 0;
	if (!h)
		h = 4096;
	np = rlnewimage(w, h, 0);
	if (!np) { e = MallocFailError; goto error; }
	l = np->d.im.rl;
	for (row = 0; row < h && !(*fill) (in, inbuf, w8, &count); l++,row++) {
		size = (Ras2Run(inbuf, code, w8,0) >> 1) + 1;
		l->code = (int2 *) smalloc(SL + (ST * size));
		if (!l->code) { e = MallocFailError; goto error; }
		l->run = (transition*)&l->code[1];
		memcpy((char *) l->run, (char *) code, size * ST);
		l->code[0] = 0;
		l->size = l->asize = size;
		l->last = 0;
		np->d.im.trncnt += (size - 2) << 1;
		if (l->run[size - 2].dn != l->run[size - 2].up)
			np->d.im.trncnt += 1;
		if (oddwid) {
			size--;
			while (size--) {
				if (l->run[size].dn < w) {
					l->run[size+1].up = w;
					l->run[size+1].dn = w;
					l->run[size+2].up = -1;
					size += 3;
					break;
				}
				if (l->run[size].up < w) {
					l->run[size].dn = w;
					l->run[size+1].up = -1;
					size += 2;
					break;
				}
			}
			if (size > 0)
				l->size = size;
			else {
				l->run[0].up = w;
				l->run[0].dn = w;
			}
		}
	}
	if (row != h) {
		temp = (RlLinePtr ) smrealloc(np->d.im.rl, sizeof(RlLine) * row);
		if (!temp) {
			e=MallocFailError;
			goto error;
		}
		np->d.im.rl = temp;
		np->height = row;
	}
	if (invert) {
		if (e = i_InvertImage(np))
			goto error;
	}
	*npp = np;
	np = 0;
error:
	if (inbuf) smfree(inbuf);
	if (code) smfree(code);
	if (np) freeimage(np);
	return e;
}

static int4 
readpda(iImagePtr * npp, const char *fn)
{
	RafFile in=0;
	int2 comp, len,w,h,res=0;
	iImagePtr np=0;
	char buf[5];
	iDAFSError e = DAFSOK;
	
	in = riOpen(fn, "rb");
	if (!in) { e = ReadError; goto error; }
	if (e = readheader(in, 0, 0, &w, &h, &comp,&res))
		goto error;
	while (comp >= 0 && !np) {
		res = 1200/res;
		switch (comp) {
		case 0:
			if (e = do0d(&np,in, w, h, pdafillline,0))
				goto error;
			i_SetImageRes(np, res);
			break;
		case 1:
		case 2:
			if (e = rawpda(&np,in, w, h,comp,res))
				goto error;
			break;
		case 3:
			buf[4] = 0;
			for (;;) {
				len = riRead(buf, 1, 1, in);
				if (len != 1) { e = ReadError; goto error; }
				if (buf[0] < '0' || buf[0] > '9')
					break;
				len = riRead(&buf[1], 1, 3, in);
				if (len != 3) { e = ReadError; goto error; }
				if (riSeek(in, atoi(buf), 1)) { e = ReadError; goto error; }
			}
			break;
		}
		if (e = readheader(in, 0, 0, &w, &h, &comp,&res))
			goto error;
	}
	*npp = np;
	np = 0;
error:
	if (in) riClose(in);
	if (np) freeimage(np);
	return e;
}

static iDAFSError writepdahead(iImagePtr sp,RafFile fo,int2 comp,int2 x,int2 y,int2 w,int2 h,char *user)
{
	iDAFSError e = DAFSOK;
	int4 ok = 1;
	int2 res = 1200/sp->dpi;
	char buffer[30]; /* I figure 27 is minimum */

	if ((res * sp->dpi) != 1200) res = 1;
	h *= res;
	x *= res;
	y *= res;
	w *= res;
	if (user){
		sprintf(buffer, "%hd%05hd%05hd%05hd%05hd0%02hd0,",
				comp, x,y,w,h, res);
		ok  = ok && riPuts(buffer, fo) != REOF;
		ok  = ok && riPuts(user, fo) != REOF;
		ok  = ok && riPuts("\n", fo) != REOF;
	} else {
		sprintf(buffer, "%hd%05hd%05hd%05hd%05hd0%02hd0\n",
				comp, x,y,w,h, res);
		ok = riPuts(buffer, fo) != REOF;
	}
	if (!ok) { e = WriteError; goto error; }
error:
	return e;
}

static iDAFSError writeunpda(iImagePtr sp, RafFile  fo, int2 x, int2 y, char *user)
	
{
	iDAFSError e = DAFSOK;
	int2 amt,chunk,h,w,w8,minwrite,i;
	char len[5], *bufpos,*buf=0;
	int2 *lcode;
	char int_buffer[8];
	
	h = sp->height & ~15; /* round down to 16 to make pda happy */
	w = (sp->width + 31) & ~31; /* round up to 32 to make us and pda happy */
	if (e = writepdahead(sp,fo,0,x,y,w,h,user))
		goto error;
	chunk = 1020;
	sprintf(len, "%04ld", chunk);
	minwrite = w;
	if (minwrite < chunk)
		minwrite = chunk;
	buf = smalloc(minwrite * 2);
	if (!buf) { e = MallocFailError; goto error; }
	bufpos = buf;
	w8 = w >> 3;
	lcode = 0;
	amt = 0;
	for (i = 0; i < h; i++) {
		Run2Ras((int2 *)sp->d.im.rl[i].run, bufpos);
		bufpos += w8;
		amt += w8;
		if (amt > minwrite) {
			bufpos = buf;
			while (amt > chunk) {
				if (riWrite(len, 1, 4, fo) != 4) { e = WriteError; goto error; }
				if (riWrite(bufpos, 1, chunk, fo) != chunk) { e = WriteError; goto error; }
				amt -= chunk;
				bufpos += chunk;
			}
			memcpy(buf, bufpos, amt);
			bufpos = buf + amt;
		}
		lcode = (int2 *)sp->d.im.rl[i].code;
	}
	bufpos = buf;
	while (amt > chunk) {
		if (riWrite(len, 1, 4, fo) != 4) { e = WriteError; goto error; }
		if (riWrite(bufpos, 1, chunk, fo) != chunk) { e = WriteError; goto error; }
		amt -= chunk;
		bufpos += chunk;
	}
	sprintf(int_buffer,"%04ld", amt);
	if (riPuts(int_buffer, fo) == REOF) { e = WriteError; goto error; }
	if (riWrite(bufpos, 1, amt, fo) != amt) { e = WriteError; goto error; }
	len[0] = 0xb;
	len[1] = 0xc;
	if (riWrite(len, 1, 2, fo) != 2) { e = WriteError; goto error; }
error:
	if (buf) smfree(buf);
	return e;
}

static iDAFSError writerawpda(iImagePtr sp, RafFile  fo, int2 comp, int2 x, int2 y, char *user)
{
	char *dp, len[5];
	int4 dl,amt,i;
	iDAFSError e = DAFSOK;
	
	if (e = writepdahead(sp,fo,comp,x,y,sp->width,sp->height & ~0xf,user))
		goto error;
	
	for (i=0;i<sp->d.rw.nchunks;i++) {
		dp = sp->d.rw.chunks[i].data;
		dl = sp->d.rw.chunks[i].len;
		while (dl) {
			amt = dl;
			if (amt > 1020) amt = 1020;
			sprintf(len,"%04ld",amt);
			if (riWrite(len, 1, 4, fo) != 4) { e = WriteError; goto error; }
			if (riWrite(dp,1,amt,fo) != amt) { e = WriteError; goto error; }
			dl -= amt;
			dp += amt;
		}
	}

	len[0] = 0xb;
	len[1] = 0xc;
	if (riWrite(len, 1, 2, fo) != 2) { e = WriteError; goto error; }
error:
	return e;
}

static iDAFSError writepda(iImagePtr sp, const char *fn, int2 comp, char *user)
{
	RafFile fo=0;
	iDAFSError e = DAFSOK;

	fo = riOpen(fn, "wb");
	if (!fo) { e = WriteError; goto error; }
	if (comp) {
		if (e = i_CompressImage(sp,comp))
			goto error;
		if (e = writerawpda(sp, fo,comp,0,0,user))
			goto error;
	} else {
		if (e = writeunpda(sp, fo,0,0,user))
			goto error;
	}
error:
	if (fo) riClose(fo);
	return e;
}

Dll iDAFSError APIENTRY
i_AppendPdaZone(iImagePtr sp, const char *fn, iDAFSCompType comp, int2 x, int2 y, char *user)
{
	RafFile fo=0;
	iDAFSError e = DAFSOK;

	fo = riOpen(fn, "r+b");
	if (!fo) { e = WriteError; goto error; }
	if (riSeek(fo, -1, 2)) { e = WriteError; goto error; }

	if (comp) {
		if (e = i_CompressImage(sp,comp))
			goto error;
		if (e = writerawpda(sp, fo,comp,x,y,user))
			goto error;
	} else {
		if (e = writeunpda(sp, fo,x,y,user))
			goto error;
	}
error:
	if (fo) riClose(fo);
	return e;
}

iDAFSError WriteDAFSBImage(RafFile  fo, iImagePtr sp)
{
	iDAFSError e=DAFSOK;
	iBox bnds;
	int4 i,res;

	if (e = i_WriteItem(fo, BeginImage | iOneSize, 4, &sp->id))
		goto error;
	if (sp->tag)
		if (e = i_WriteItem(fo,AString,strlen(sp->tag)+1,sp->tag))
			goto error;
	if (sp->tag && gWriteImageProc) {
		if (e = gWriteImageProc(sp->tag,sp))
			goto error;
	} else {
		bnds.x = 0;
		bnds.y = 0;
		bnds.w = (int2) sp->width;
		bnds.h = (int2) sp->height;
		res = sp->dpi;
		if (e = i_WriteItem(fo, Long, sizeof(int4), &res))
			goto error;
		if (e = i_WriteItem(fo, SBox, sizeof(iBox), &bnds))
			goto error;
		if (e = i_CompressImage(sp,CCITT4))
			goto error;
		
		for (i=0;i<sp->d.rw.nchunks;i++) {
			if (e = i_WriteItem(fo,CCITT4Chunk,sp->d.rw.chunks[i].len,sp->d.rw.chunks[i].data))
				goto error;
		}
	}
	if (e = i_WriteItem(fo, EndImage, 0, 0))
		goto error;
error:
	return e;
}

iDAFSError 
ReadDAFSBImage(RafFile  fo, iImagePtr * imp, iGlobid *gid)
{
	iDAFSError e = DAFSOK;
	iBox bnds;
	iLBox lbnds;
	char *s;
	iToken t;
	uint2 size;
	int4 id, res=0;
	void **ims;
	iImagePtr im=0;
	
	*imp = 0;
	if (e = i_ReadItem(fo, BeginImage | iOneSize, sizeof(id), &size, &id, &gid->DAFSBSwap))
		goto error;
	if (e = i_ReadToken(fo, &t, &gid->DAFSBSwap))
		goto error;
	s = 0;
	if ((t & iTMask) == AString) {
		if (e = i_ReadItemAlloc(fo,t,&size,(void**)&s,&gid->DAFSBSwap))
			goto error;
		if (e = i_ReadToken(fo, &t, &gid->DAFSBSwap))
			goto error;
	}
	if (s && gReadImageProc) {
		if ((t&iTMask) != EndImage) {e = ReadError; goto error; }
		if (e = gReadImageProc(s,&im))
			goto error;
	} else {
		while (((t & iTMask) != Box)&&((t & iTMask) != SBox)) {
			if ((t & iTMask) == Long) {
				if (e = i_ReadItem(fo, t, sizeof(res),&size, &res, &gid->DAFSBSwap))
					goto error;
			} else {
				if (e = i_ReadItem(fo, t, 0, &size, 0,&gid->DAFSBSwap))
					goto error;
			}
			if (e = i_ReadToken(fo, &t, &gid->DAFSBSwap))
				goto error;
		}
		if ((t&iTMask)==Box){
			if (e = i_ReadItem(fo, t, sizeof(lbnds), &size, &lbnds, &gid->DAFSBSwap))
				goto error;
			bnds.x=(int2)lbnds.x;
			bnds.y=(int2)lbnds.y;
			bnds.w=(int2)lbnds.w;
			bnds.h=(int2)lbnds.h;

		} else{
			if (e = i_ReadItem(fo, t, sizeof(bnds), &size, &bnds, &gid->DAFSBSwap))
				goto error;
		}	
		if (e = rawDAFSB(&im,fo, bnds.w, bnds.h, &gid->DAFSBSwap))
			goto error;
		im->dpi = (int2)res;
	}
	im->id = id;
	im->tag = s;
	if (gid->im_id <= id) {
		gid->im_id = id+1;
		if (!gid->ims)
			ims = (void **)smalloc(sizeof(iImagePtr) * gid->im_id);
		else
			ims = (void **)smrealloc(gid->ims,sizeof(iImagePtr) * gid->im_id);
		if (!ims) { e = MallocFailError; goto error; }
		gid->ims = ims;
	}
	gid->ims[id] = (void*)im;
	*imp = im;
	im = 0;
error:
	if (im) freeimage(im);
	return e;
}

Dll int4 APIENTRY
i_ImageExt(const char *fn)
{
	int4 len = strlen(fn);
	if (len < 4)
		return 0;
	if (!strcmp(".pda", &fn[len - 4]))
		return 1;
	if (!strcmp(".PDA", &fn[len - 4]))
		return 1;
	if (!strcmp(".tif", &fn[len - 4]))
		return 1;
	if (!strcmp(".TIF", &fn[len - 4]))
		return 1;
	if (!strcmp(".raw", &fn[len - 4]))
		return 1;
	if (!strcmp(".RAW", &fn[len - 4]))
		return 1;
	if (len < 5)
		return 0;
	if (!strcmp(".tiff", &fn[len - 5]))
		return 1;
	if (!strcmp(".TIFF", &fn[len - 5]))
		return 1;
	
	return 0;
}

Dll iDAFSError APIENTRY
i_ReadImage(const char *fn, iImagePtr * npp)
{
	iDAFSError err = DAFSOK;
	int4 len = strlen(fn);
	*npp = 0;
#if IMPLEMENT_TIFF
	if (len < 4)
		err = tiffread(npp, fn);
	else
#endif
	if ((!strcmp(".pda", &fn[len - 4]))||
		(!strcmp(".PDA", &fn[len - 4])))
		err = readpda(npp, fn);
#if IMPLEMENT_TIFF
	else if ((!strcmp(".tif", &fn[len - 4]))||
			 (!strcmp(".TIF", &fn[len - 4])))
		err = tiffread(npp, fn);
	else err = tiffread(npp, fn);
#endif
	if(*npp)
		(*npp)->use++;
	return err;
}

Dll iDAFSError APIENTRY
i_WriteImage(const char *fn, iImagePtr sp, iDAFSCompType comp)
{
	int4 len = strlen(fn);
#if IMPLEMENT_TIFF
	if (len < 4)
		return writetiff(sp, fn, comp);
#endif
	if (!strcmp(".pda", &fn[len - 4]))
		return writepda(sp, fn, comp, 0);
#if IMPLEMENT_TIFF
	else if (!strcmp(".tif", &fn[len - 4]))
		return writetiff(sp, fn, comp);
	return writetiff(sp, fn, comp);
#endif
}

