/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "smemory.h"
#include "c-lib.h"

#include "entity.h"
#include "dafs.h"
#include "dafstype.h"



Dll int4 APIENTRY
i_HorzCompare(iEntityPtr *one, iEntityPtr *two)
{
	iBox oneb,twob;
	
	oneb=i_GetEnclosingBox(*one);
	twob=i_GetEnclosingBox(*two);

	return (oneb.x - twob.x);
	
}

Dll int4 APIENTRY
i_VertCompare(iEntityPtr *one, iEntityPtr *two)
{	
	iBox oneb,twob;
 
 	oneb=i_GetEnclosingBox(*one);
	twob=i_GetEnclosingBox(*two);

	return (oneb.y - twob.y);
}

Dll int4 APIENTRY
i_VHCompare(iEntityPtr *one, iEntityPtr *two)
{	
	iBox b1,b2;

 
 	b1=i_GetEnclosingBox(*one);
	b2=i_GetEnclosingBox(*two);
	b1.h += b1.y;
	b2.h += b2.y;
	if (b2.y >= b1.y && b2.y <= b1.h) return b1.x - b2.x;
	if (b2.h >= b1.y && b2.h <= b1.h) return b1.x - b2.x;
	if (b1.y >= b2.y && b1.y <= b2.h) return b1.x - b2.x;
	if (b1.h >= b2.y && b1.h <= b2.h) return b1.x - b2.x;
	return b1.y - b2.y;
}

static void 
myhsort(iEntityPtr *base, int4 n, int4 (APIENTRY *compare)(iEntityPtr *,iEntityPtr *))
{
	iEntityPtr *cur, temp,*ptr1, *ptr2, *end;

	if (n<=1)
		return;

	cur=base+(n/2);
	end=base+n-1;
	for (;;) {
		if (cur>base) {
			cur--;
			temp=*cur;
		} else {
			temp=*end;
			*end=*base;
			end--;
			if (end==base) {
				*base=temp;
				break;
			}
		}
		ptr1=cur;
		ptr2=base+(cur-base)*2+1;
		while (ptr2<=end) {
			if (ptr2<end && compare(ptr2,ptr2+1)<0)
				ptr2++;
			if (compare(&temp,ptr2)<0) {
				*ptr1=*ptr2;
				ptr1=ptr2;
				ptr2=base+(ptr2-base)*2+1;
			} else
				ptr2=end+1;
		}
		*ptr1=temp;
	}
}

static iDAFSError 
reorder(iEntityPtr *order,iEntityPtr en, int4 size)
{

	int4 i;
	iDAFSError err=DAFSOK;
	
	for(i=0;i<size;i++)
		i_MoveToBack(order[i],en);
	return err;
}

Dll iDAFSError APIENTRY
i_Sort(iEntityPtr en, iCompareProc proc)
{
	iEntityPtr cur;
	int4 nkids,i;
	iEntityPtr *entities;
	iDAFSError err=DAFSOK;
	
	nkids=i_CountChildren(en);
	if (!nkids) return DAFSOK;
	entities=(iEntityPtr*)smalloc(nkids*sizeof(iEntityPtr));
	if (!entities) return MallocFailError;
	for(i=0,cur=i_GetChildEntity(en);i<nkids;cur=i_GetNextEntity(cur),i++)
	{
		entities[i]=cur;
	}
#if 0
	qsort((void*)entities,nkids,sizeof(iEntityPtr),*proc);
#else
	myhsort((void*)entities,nkids,*proc);
#endif	
	reorder(entities,en,nkids);
	smfree(entities);
		
	return err;

}

