/* kills all entities with a specific prop */

#include <stdio.h>
#include <stdlib.h>
#include "smemory.h"
#include "ifc.h"
#include "utf.h"
#include "dafs.h"

static char *prop_name, *prop_text;
static int num_deleted = 0;

ifc myargs[] = {
  {(char **) &prop_name, 0, (int) "", 0, IFC_STR, "p", 0},
  {(char **) &prop_text, 0, (int) "", 0, IFC_STR, "c", 0},
  {0, 0, 0, 1, IFC_NA, 0, 0}
};

static void
usage(char *p)
{
  fprintf(stderr, "usage: %s -p <property name> -c <property content> <dafsfiles>\n", p);
  fprintf(stderr, "-p\tthe name of the property\n");
  fprintf(stderr, "-c\tthe content of the property\n");
  exit(1);
}

static void
killProp(iEntityPtr en, char *name, char *content)
{
  iPropPtr prop;
  int old_deleted = num_deleted;
  
  if(prop = i_FindProp(en, name)) {
    iDAFSError err;
    char *p_string;
    iUCode *p_ustring;
    long p_long;

    switch(i_GetPropType(prop)) {
    case iPString:
      if(!(err = i_GetStringProp(en, name, &p_string))) {
	if(!*content || !strcmp(content, p_string)) {
	  i_DisposeEntity(en);
	  num_deleted++;
	}
      }
      break;
    case iPUString:
      if(!(err = i_GetUStringProp(en, name, &p_ustring))) {
	iUCode *ucontent = (iUCode *) smalloc((strlen(content) + 1) * sizeof(iUCode));
	utf2ucs(ucontent, content);

	if(!*ucontent || !ucscmp(ucontent, p_ustring)) {
	  i_DisposeEntity(en);
	  num_deleted++;
	}
      
	smfree(ucontent);
      }
      break;
    case iPBool:
      if(!(err = i_GetBoolProp(en, name, &p_long))) {
	if(p_long && !strcmp(content, "true")) {
	  i_DisposeEntity(en);
	  num_deleted++;
	}
      }
      break;
    case iPLong:
      if(!(err = i_GetLongProp(en, name, &p_long))) {
	if(!*content || p_long == atoi(content)) {
	  i_DisposeEntity(en);
	  num_deleted++;
	}
      }
      break;
    case iPUserData:
    case iPUserIData:
      if(*content)
	fprintf(stderr, "Property is iPUserData, unable to read contents\n");
      else {
	i_DisposeEntity(en);
	num_deleted++;
      }
      break;
    default:
      fprintf(stderr, "Property isn't of a valid type");
    }
	  
    if(err)
      fprintf(stderr, "There was an error\n");
  }

  if(!prop || num_deleted == old_deleted) {
    iEntityPtr child, next;

    for(child = i_GetChildEntity(en); child; child = next) {
      next = i_GetNextEntity(child);
      killProp(child, name, content);
    }
  }
}

int main(int argc, char *argv[])
{
  if(processargs(argc, argv, myargs))
    usage(argv[0]);

  if(*prop_name) {
    iEntityPtr doc;
    iDAFSError err;
    int i = 1, 
        prev_deleted = num_deleted;

    printf("Deleting entities with property %s of content %s\n", prop_name, prop_text);
    i_InitLib();

    for(i = 1; i < argc; i++) {
      if(argv[i]) {
	if(err = i_ReadEntity(argv[i], &doc))
	  fprintf(stderr, "Couldn't read file %s; skipping it\n", argv[i]);
	else {
	  killProp(doc, prop_name, prop_text);
	  if(num_deleted > prev_deleted) {
	    printf("writing file %s (%i entities deleted)\n", argv[i], num_deleted - prev_deleted);
	    i_WriteEntity(argv[i], doc, DAFSB);
	  }
	  i_DisposeEntity(doc);
	  prev_deleted = num_deleted;
	}
      }
    }

    if(num_deleted)
      printf("Deleted %i entities\n", num_deleted);
    else
      printf("No entities were deleted\n");

    i_ExitLib();
  }



  return 0;
}
