
#include <stdio.h>
#include "rafport.h"
#include "rafutil.h"
#include "dafs.h"
#include "ifc.h"
#include "smemory.h"

static long doall;

ifc myargs[] = {
	{(char**)&doall,0,0,1,IFC_FLAG,"a",0},
	{0, 0, 0, 1, IFC_NA, 0, 0}
};

static void
usage(char *p)
{
	fprintf(stderr, "usage: %s -a <DAFSB files>\n",p);
	fprintf(stderr,"\t-a\tdumps all the DATA not just the first line\n");
	exit(-1);
}

int
main(int c, char *v[])
{
	int i,j,e,num;
	RafFile fi;
	char t,*dp;
	long size;
	iBox *bp;
	iLBox *lbp;
	iPosSet *ps;
	iOrthogonPtr border;
	int swap;

	if (processargs(c, v, myargs))
		usage(v[0]);
	for (i=1;i<c;i++) {
		if (!v[i]) continue;
		fi = riOpen(v[i],"rb");
		if (!fi) continue;
		for (;;) {
			e = i_Read1Token(fi,&t);
			if (riEof(fi)) break;
			if (e) 
				halt("Error: read token %d\n",e);
			dp = 0;
			if (e = i_ReadItemAllocH(fi,t,&size,(void**)&dp,&swap))
				halt("Error: read item alloc %d\n",e);
			switch (t & iTMask) {
			case NullTok:
				printf("NullTok"); break;
			case BeginEntity:
				printf("BeginEntity  %06ld: %d",size,*(int*)dp);
				break;
			case BeginImage:
				printf("BeginImage   %06ld: %d",size,*(int*)dp);
				break;
			case EndEntity:
				printf("EndEntity"); break;
			case Box:
				printf("Box (longs)  %06ld: ",size);
				lbp = (iLBox *)dp;
				printf("%ld %ld %ld %ld",lbp->x,lbp->y,lbp->w,lbp->h);
				break;
			case SBox:
				printf("Box          %06ld: ",size);
				bp = (iBox*) dp;
				printf("%d %d %d %d",bp->x,bp->y,bp->w,bp->h);
				break;
			case TString:
				printf("TString      %06ld: ",size);
				printf("%s",dp);
				break;
			case AString:
				printf("AString      %06ld: ",size);
				printf("%s",dp);
				break;
			case DString:
				printf("DString      %06ld: ",size);
				printf("%s",dp);
				break;
			case EndImage:
				printf("EndImage"); break;
			case BeginProp:
				printf("BeginProp"); break;
			case EndProp:
				printf("EndProp"); break;
			case Data:
				printf("Data         %06ld: ",size);
				e = (doall?(size/4):4);
				for (j=0;j < e;j++) {
					if (j && !(j % 4)) printf("\n             %06lx: ",j*4);
					printf("0x%02x%02x%02x%02x ",
						dp[(j * 4) + 0] & 0xff,
						dp[(j * 4) + 1] & 0xff,
						dp[(j * 4) + 2] & 0xff,
						dp[(j * 4) + 3] & 0xff);
				}
				break;
			case IData:
				printf("IData        %06ld: ",size);
				e = (doall?(size/4):4);
				for (j=0;j< e;j++) {
					if (j && !(j % 4)) printf("\n             %06lx: ",j*4);
					printf("0x%02x%02x%02x%02x ",
						dp[(j * 4) + 0] & 0xff,
						dp[(j * 4) + 1] & 0xff,
						dp[(j * 4) + 2] & 0xff,
						dp[(j * 4) + 3] & 0xff);
				}
				break;
			case CCITT4Chunk:
				printf("CCITT4Chunk  %06ld: ",size);
				for (j=0;j < 4 && j<(size/4);j++) {
					if (!(j%4)) {
						printf("\n%08x: ",j * 4);
					}
					printf("0x%08lx ",((long*)dp)[j]);
				}
				break;
			case Long:
				printf("Long         %06ld: %d",size,*(int*)dp);
				break;
			case ShortNShort:
				printf("ShortNShort  %06ld: %d, %d",size,((short*)dp)[0],((short*)dp)[1]);
				break;
			case Short:
				printf("Short        %06ld: %d",size,*(short*)dp);
				break;
			case LongArray:
				num=size/sizeof(long);
				printf("LongArray    %06ld: ",size);
				for (j=0;j<num; j++)
					printf("%ld ",((long*)dp)[j]);
				break;	
			case ShortArray:
				printf("ShortArray   %06ld: ",size);
				num=size/sizeof(short);
				for(j=0;j<num;j++) {
					int c = ((short*)dp)[j];
					if(c < 255 && isprint(c) && !isspace(c))
						printf("%c ", c);
					else
						printf("U+%04x ", c);
				}
#if 0
					printf("%d ",((short*)dp)[j]);						
#endif
				break;
			case Byte:
				printf("Byte         %06ld: %d",size,*(char*)dp);
				break;
			case BTrue:
				printf("BTrue"); break;
			case BFalse:
				printf("BFalse"); break;
			case BorrowEntity:
				printf("BorrowEntity %06ld: %d",size,*(int*)dp);
				break;
			case DPosSet:
				printf("DPosSet      %06ld:",size); 
				ps=(iPosSet*)dp;
				if (!size) printf(" non-rec");
				for(j=0;j<size/sizeof(iPosSet);j++) {
					if (ps[j].chr < ' ')
						printf(" 0x%02x~%d",
							ps[j].chr,ps[j].conf);
					else
						printf(" %c~%d",
							ps[j].chr,ps[j].conf);
				}
				break;
			case D1PosSet:
				ps=(iPosSet*)dp;
				printf("D1PosSet     %06ld: ",size);
				if (ps->chr < ' ' || ps->chr > 127) 
					printf("0x%04x~%d",ps->chr,ps->conf);
				else printf("%c~%d",ps->chr,ps->conf);
				break;
			case ImageID:
				printf("ImageID      %06ld: %d",size,*(int*)dp);
				break;
			case ByteOrder:
				printf("ByteOrder    %06ld: %s",size,
					(*(short*)dp == 1?"same":"swapped"));
				swap = (*(short*)dp == 0x100);
				break;
			case Poly:
				border = i_NewOrthogon(0);
				printf("Poly        %06ld: %d\n",size,*(int*)dp);
				border->point = (iPoint*)dp;
				border->n = size/sizeof(iPoint);
				for (j=0; j<border->n; j++)
					printf("   %d: %d %d\n",j, border->point[j].x,
						border->point[j].y); 
				i_DisposeOrthogon(border);
				break;   
			default:
				printf("0x%02x         %06ld:",t,size);
			}
			printf("\n");
		}
		riClose(fi);
	}

	return 0;
}
