
/*
 * Copyright (c) 1994, 1995  RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"
#include "canna.h"
#include "Application.h"

/*
// a sui generis object 
// Application wide resources
// that don't belong in MainWindow
// like the app_shell
// and the Japanese server
*/

Application::Application(
	XtAppContext input_app_context,
	Widget input_app_shell, 
	aString input_app_name,
	aString input_canna_host,
	aString input_canna_lib_dir)
	: app_context(input_app_context),
	  app_shell(input_app_shell),
	  app_name(input_app_name),
	  nihon_initialized(FALSE), 
	  canna_host(input_canna_host),
	  tried_once(FALSE),
	  canna_lib_dir(input_canna_lib_dir)
{
}

Application::~Application()
{
	XtDestroyWidget(this->app_shell);
}

/*
// return TRUE on success
// return FALSE on failure, CANNA code may also have supplied
// a warning message
*/
int
Application::initializeNihon(iUCode *warning)
{
	int init_value;
	warning[0] = 0;
	const char *server_name;

	if(this->tried_once)
		return this->nihon_initialized;

	this->tried_once = TRUE;

	if(nihon_initialized == FALSE) {
		// use the canna_host supplied on the command line first
		// then use the value of the CANNAHOST environment variable
		// then use null -- the canna code will assume the local host
		if(this->canna_host == ""){
			// you cannot free the return value of getenv
			server_name = getenv("CANNAHOST");
		} else 
			server_name = this->canna_host.string();
		init_value = nihonInit(server_name, canna_lib_dir.string(), warning);
		this->nihon_initialized = init_value;
	}

	return this->nihon_initialized;
}

int
Application::nihonInitializedP()
{
	return nihon_initialized;
}


