
#include "rafport.h"

#include <Xm/MessageB.h>
#include <Xm/RowColumn.h>
#include <Xm/Form.h>
#include <Xm/LabelG.h>
#include <Xm/TextF.h>
#include <Xm/Text.h>

#include "c-type.h"
#include "smemory.h"

#include "illum.h"

#include "misc/XArg.h"

#include "MainWindow.h"
#include "Application.h"
#include "EntityDialog.h"
#include "BoxEditDialog.h"
#include "Document.h"

/* because Motif include files aren't understood by GDB/G++ */
struct xmtb{
    char *ptr;                  /* Pointer to data. */
    int length;                 /* Number of bytes of data. */
    XmTextFormat format;       /* Representations format */
};

typedef struct xmtb xmtb;

struct xmtvcbs
{
    int reason;
    XEvent  *event;
    Boolean doit;
    long currInsert, newInsert;
    long startPos, endPos;
    xmtb *text;
};

typedef struct xmtvcbs xmtvcbs;

extern void make_string(Widget w,xmtvcbs *cbs,char *num);
extern void i_SetPBBBounds(iEntityPtr entity, iBox pbb);
extern iBox i_GetPBBBounds(iEntityPtr entity);


#define ishex(x) ((x >= '0' && x <= '9') || \
					(x >= 'a' && x <= 'f') || \
					(x >= 'A' && x <= 'F'))


static void 
boxOKCB(Widget w, BoxEditDialog *sl, void *call)
{
  sl->okCB((XmAnyCallbackStruct *)call);
}

static void 
boxCancelCB(Widget w, BoxEditDialog *sl, void *call)
{
  sl->cancelCB((XmAnyCallbackStruct *)call);
}


static void 
setCoordField(Widget w, iBox *b, int i)
{
	char *string = (char*)smalloc(10 * sizeof(char));
	
	switch (i) {
	case 0:
		sprintf(string, "%d", b->x);
		break;
	case 1:
		sprintf(string, "%d", b->y);
		break;
	case 2:
		sprintf(string, "%d", b->w);
		break;
	case 3:
		sprintf(string, "%d", b->h);
		break;
	}
	XmTextFieldSetString(w, string);
	smfree(string);
}

BoxEditDialog::BoxEditDialog(aString input_title, EntityDialog *input_ed)
  : ModalDialog(input_ed->mw, 
				input_ed->app_shell, 
				input_ed->mw->app->app_context),
	title(input_title),
	ed(input_ed),
	entity(input_ed->entity)

{
#if 0
  this->createDialog();
#endif
}

iBox
BoxEditDialog::getBounds(void)
{
  iBox b;
  b = i_GetEnclosingBox(this->entity);
  if(b.w == 0) {
	b.x = b.y = b.w = b.h = 0;
  }
  return b;
}

void
BoxEditDialog::setBounds(iBox b)
{

  if(b.w == 0 || b.h == 0)
	b.w = b.x = b.y = b.h = 0;
  i_SetBox(entity, b);
}


void
verify_coords(Widget w,  void *user_data, xmtvcbs *cbs)
{
	int len = 0,i,hex;
	char num[20],*p;

	make_string(w,cbs,num);
	p = num;
	if (hex = !strncmp(p,"0x",2)) p += 2;
	while (*p == '0') p++;
	len = strlen(p);
	if (!len) return;
	cbs->doit = 0;
	if ((hex && len > 4) || len > 5)
		return;
	for (i=0;i<len;i++)
		if ((hex && !ishex(p[i])) || (!hex && !isdigit(p[i])))
			return;
	sscanf(p,(hex?"%x":"%d"),&len);
	if (len > 0xffff)
		return;
	cbs->doit = 1;
}

void 
BoxEditDialog::create()
{
  XmString ok_string;
  XmString cancel_string;
  XmString title_string;
  int i;
  Widget rowcol,form,*text_w;
  char *labels[] = {"X pos:", "Y pos:", "Width:","Height:"};
  XArg args;
  iBox bounds = this->getBounds();
  ok_string = XmStringCreateLocalized("OK");
  cancel_string = XmStringCreateLocalized("Cancel");
  title_string = XmStringCreateLocalized((char *) title.string());
  
  args.setArg(XmNokLabelString, ok_string);
  args.setArg(XmNcancelLabelString, cancel_string);
  args.setArg(XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL);
  args.setArg(XmNdialogTitle, title_string);
	
  this->dialog = XmCreateTemplateDialog(this->app_shell, 
								  "Edit Box",
								  args.getArgs(), 
								  args.getSize());
  XmStringFree(ok_string);
  XmStringFree(cancel_string);
  XmStringFree(title_string);
	
  text_w = (Widget*)smalloc(sizeof(Widget)*NUM_BOUNDS_FIELDS);
  rowcol = XtVaCreateWidget("rowcol",
							xmRowColumnWidgetClass, this->dialog, NULL);
  for (i = 0; i < NUM_BOUNDS_FIELDS; i++) {
	form = XtVaCreateWidget("form", xmFormWidgetClass, rowcol,
							XmNfractionBase, 		10,
							NULL);
	XtVaCreateManagedWidget(labels[i],
							xmLabelGadgetClass, 	form,
							XmNtopAttachment,		XmATTACH_FORM,
							XmNbottomAttachment,	XmATTACH_FORM,
							XmNleftAttachment,		XmATTACH_FORM,
							XmNrightAttachment,		XmATTACH_POSITION,
							XmNrightPosition,		3,
							XmNalignment,			XmALIGNMENT_END,
							NULL);
	text_w[i] = XtVaCreateManagedWidget("text_w",
										xmTextFieldWidgetClass,	form,
										XmNtraversalOn,			True,
										XmNrightAttachment,		XmATTACH_FORM,
										XmNleftAttachment,		XmATTACH_POSITION,
										XmNleftPosition,		4,
										NULL);
	XtAddCallback(text_w[i], XmNmodifyVerifyCallback,
				  (XtCallbackProc) verify_coords, 0);

	setCoordField(text_w[i], &bounds, i);
	XtManageChild(form);
  }

  XtAddCallback(this->dialog, XmNokCallback, boxOKCB, this);
  XtAddCallback(this->dialog, XmNcancelCallback, boxCancelCB, this);

  XtManageChild(rowcol);
  this->text_boxes = text_w;
}	



static void
getBox(Widget *text_w, iBox *b)
{
	iBox box;
	box.x = atoi((const char*)XmTextGetString(text_w[0]));
	box.y = atoi((const char*)XmTextGetString(text_w[1]));
	box.w = atoi((const char*)XmTextGetString(text_w[2]));
	box.h = atoi((const char*)XmTextGetString(text_w[3]));
	*b = box;
}

void
BoxEditDialog::okCB(XmAnyCallbackStruct *cbs)
{
  this->done = TRUE;
  //this->changed = TRUE;

  iBox box;

  getBox(this->text_boxes, &box);
  this->setBounds(box);
  this->mw->document->setDirty();
  this->hide();
}

void
BoxEditDialog::cancelCB(XmAnyCallbackStruct *cbs)
{
  this->done = TRUE;
  //this->changed = FALSE;

  this->hide();
}

void
BoxEditDialog::showModal(iEntityPtr input_entity)
{
  if(this->dialog == 0)
	this->create();
  this->entity = input_entity;
  iBox bounds = this->getBounds();
  int i;
  for(i = 0; i < 4; i++)
  	setCoordField(this->text_boxes[i], &bounds, i);

  this->ModalDialog::showModal();
}

PBBEditDialog::PBBEditDialog(aString input_title, EntityDialog *input_ed)
  : BoxEditDialog(input_title, input_ed)
{}

iBox
PBBEditDialog::getBounds(void)
{
  iBox b;
  b = i_GetPBBBounds(this->entity);
  if(b.w == 0) {
	b.x = b.y = b.w = b.h = 0;
  }
  return b;
}

void
PBBEditDialog::setBounds(iBox b)
{

  if(b.w == 0 || b.h == 0)
	b.w = b.h = b.x = b.y = 0;
  i_SetPBBBounds(entity, b);
}

BoxEditDialog::~BoxEditDialog()
{
  XtDestroyWidget(this->dialog);
  smfree(this->text_boxes);
}

PBBEditDialog::~PBBEditDialog()
{}
