
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*
// The smallest element in OOC mode
*/

#include "rafport.h"

#include "illum.h"
#include "smemory.h"
#include "rafutil.h" // for min and max

#include "MainWindow.h"
#include "Mode.h"
#include "ViewCanvas.h"
#include "EntityAux.h"
#include "OOCPage.h"
#include "Document.h"
#include "misc/Unicode.h"
#include "misc/bb.h"
#include "ucs.h"
#include "CE.h"
#include "fudge.h"

extern void i_SetPBBBounds(iEntityPtr entity, iBox pbb);
extern iBox i_GetPBBBounds(iEntityPtr entity);

int 
bloatOOCBox(iBox *b,int type)
{
	if (type == vtFLAG) {
		b->x-=FLAG_FUDGE_X;
		b->w+=2*FLAG_FUDGE_X;
		b->y-=FLAG_FUDGE_Y;
		b->h+=2*FLAG_FUDGE_Y;
		b->y = MAX(0, b->y);
		b->x = MAX(0, b->x);
		return FLAG_FUDGE_X || FLAG_FUDGE_Y;
	} else {
		b->x-=OOC_FUDGE_X;
		b->w+=2*OOC_FUDGE_X;
		b->y-=OOC_FUDGE_Y;
		b->h+=2*OOC_FUDGE_Y;
		b->y = MAX(0, b->y);
		b->x = MAX(0, b->x);
		return OOC_FUDGE_X || OOC_FUDGE_Y;
	}
}


CE::CE(OOCPage *input_page, iEntityPtr input_entity)
	: EntityFace(input_page, input_entity), x(0), y(0), sub_page(0)
{
}

CE::~CE()
{
}

void
CE::deleteAndUpdate()
{
	this->page->removeFace(this);
	
	DLList *l = &this->page->face_list;
	OOCPage *page = (OOCPage *) this->page;
			
	// check if page is empty
	if (dllistEmpty(&page->face_list))
		page->should_delete = TRUE;
	// check if sub-page is empty
	else if(page->num_pages > 1){
		CE *ce;

		for(Ref r = dllistFirst(l); r; dllistNext(l, &r)){
			ce = (CE *) dllistContents(l, r);
			if(ce->sub_page == page->current_page)
				break;
		}

		if(!r){
			int new_page;
			if(page->current_page == page->num_pages-1)
				page->mode->forward();
			else {
				new_page = page->current_page;
				page->calcCells();		   
				page->setPage(new_page);
			}
		}
	}
	
	eaSetFace(this->entity, page->mode_type, 0);
	delete this;
}

void
CE::deleteCheap()
{
	delete this;
}

/* 
// This is the real code that actually draws
// it is too long
// it was derived from the equivalent code in
// image mode
//
*/
void
CE::redisplay(Display *display, Pixmap pixmap,
			  ViewCanvas *canvas, 
			  int x_offset, int y_offset,
			  int cell_width, int cell_height)
{
	int drawbb,selected;
	int mag_exp = this->page->getMagExp();
	int mag_image_exp = mag_exp + 16;
	char *bits = 0;
	int2 dw;
	iImagePtr image;
	XImage *x_image;
	GC use_gc;
	iBox box;
	iBox image_bb = i_GetEnclosingBox(this->entity);
	drawbb = bloatOOCBox(&image_bb,((OOCPage *) this->page)->mode_type);
	bbScaleUp(image_bb, mag_exp);

	// blank the scratch pixmap
	XFillRectangle(display, pixmap, canvas->blankGC(),
				   x_offset, y_offset, cell_width, cell_height);
	
	int delta_x=(cell_width-image_bb.w)/2 - image_bb.x;
	int delta_y=(cell_height-image_bb.h)/2 - image_bb.y;
	
	image = i_GetImage(this->entity);
	
	// bogus may be I should push the mag on the off chance
	// that the mags might be different in different windows
	
	i_Image2BitMap(image, image_bb, 1<<mag_image_exp, &dw, &bits);

	x_image = XCreateImage(display, 0, 1, XYBitmap,0,
						   bits,dw * 8, image_bb.h,8,dw);
	x_image->byte_order = MSBFirst;
	x_image->bitmap_bit_order = MSBFirst;
	_XInitImageFuncPtrs(x_image);


	XPutImage(display, pixmap, canvas->copyGC(), x_image, 
			  0, 0, 
			  x_offset + image_bb.x + delta_x, y_offset + image_bb.y + delta_y,
			  image_bb.w, image_bb.h);

	// we want to free the image ourselves, since the memory
	// library might add a header
	x_image->data=0;

	XDestroyImage(x_image);
	smfree(bits);

	if(selected = (getEntitySelection(entity) == scITEM)) {
		use_gc = canvas->thickBlueGC();
	} else {
		use_gc = canvas->blueGC();
	}

	box = i_GetPBBBounds(this->entity);
	if (box.w) {
		bbScaleUp(box, mag_exp);
		// the pitch bounding box
		XDrawRectangle(display, pixmap, canvas->redGC(), 
					   x_offset + box.x + delta_x - BOX_FUDGE,
					   y_offset + box.y + delta_y - BOX_FUDGE,
					   box.w + BOX_FUDGE, 
					   box.h + BOX_FUDGE);
	}
	if (drawbb) 
		// the character bounding box
		XDrawRectangle(display, pixmap, canvas->grayGC(), 
			   x_offset + image_bb.x + delta_x - BOX_FUDGE,
			   y_offset + image_bb.y + delta_y - BOX_FUDGE,
			   image_bb.w + BOX_FUDGE, 
			   image_bb.h + BOX_FUDGE);
	box = i_GetEnclosingBox(this->entity);
	iATag tag;
	void *data;
	iOrthogonPtr poly, newpoly, grownpoly;
	int maxpoints=100, i, j;
	iOrthogonListPtr border2;
	iOrthogonListPtr border=0;
	XRectangle rect;
	if (!i_GetArea(this->entity, &data, &tag)){
		switch(tag){
		
		case iABoxList:  /* change to iAOrthogonList and fall through  */
			if(i_ForceAreaType(entity, iAOrthogonList))
				break;
			if (i_GetArea(this->entity, &data, &tag))
				break;
		case iAOrthogonList:
			border=(iOrthogonList*)data;		
			newpoly = i_NewOrthogon(maxpoints+1);
			grownpoly = i_NewOrthogon(maxpoints+1);
			if (border->num_poly){
				border2=i_CopyOrthogonList(border);
				for(i=0; i<border->num_poly; i++){
					poly=border2->poly[i];
					polyScaleUp(poly, mag_exp);
					if(poly->n){
						if (poly->n > maxpoints){
							maxpoints = grownpoly->n = poly->n;
							newpoly->point = 
								(iPoint*)smrealloc(newpoly->point,
												   sizeof(iPoint)*(poly->n+1));
							grownpoly->point = 
								(iPoint*)smrealloc(grownpoly->point,
												   sizeof(iPoint)*(poly->n+1));
						}
						for (j=0; j<poly->n; j++) {
							newpoly->point[j].x = 
								(short)(x_offset+poly->point[j].x+delta_x);
							newpoly->point[j].y = 
								(short)(y_offset+poly->point[j].y+delta_y);
						}
						newpoly->n = poly->n;
						i_GrowOrthogon(newpoly, grownpoly, BOX_FUDGE);
						grownpoly->point[grownpoly->n].x = grownpoly->point[0].x;
						grownpoly->point[grownpoly->n++].y = grownpoly->point[0].y;
						XDrawLines(display, pixmap, use_gc, (XPoint*)grownpoly->point, 
								   grownpoly->n, CoordModeOrigin);
					}
				}
				i_DisposeOrthogonList(border2);
			}
			i_DisposeOrthogon(newpoly);
			i_DisposeOrthogon(grownpoly);
			break;
				
		case iABox:
			box=*(iBox*)data;
			bbScaleUp(box, mag_exp);
			rect.x = x_offset + box.x +delta_x-BOX_FUDGE;
			rect.y = y_offset + box.y +delta_y-BOX_FUDGE;
			rect.width = box.w +BOX_FUDGE;
			rect.height = box.h +BOX_FUDGE;
			XDrawRectangles(display, pixmap, use_gc,&rect, 1);
			break;
		}
	}
	if(((OOCPage *) this->page)->mode_type == vtFLAG) {
		iUCode buf[2] = {' ', '\0'};
		iUCode *text = 0;
		iUCode *s = i_AllocUText(this->entity);
		int editing = ((OOCPage *) this->page)->edit_record;

		if(s && s[0])
			text = s;
		else if (selected && editing) {
			text = buf;
		}
		
		if (text && text[0]) {
			int x,y,tw,th,ascent;
			this->page->getCanvas()->measureText(text,ucslen(text),
												  &tw,&th,&ascent);
			x = (cell_width - tw) / 2;
			y = ((cell_height + image_bb.h) / 2) + th;
			if (selected && !editing)
				drawUnicode(display,pixmap,
							this->page->getCanvas()->inverseBlackGC(),
							x_offset+x,y_offset+y,text,ucslen(text));
			else {
				int tw,ch,th;
				this->page->getCanvas()->measureText(text, 1, &tw, &ch, &ascent);
				this->page->getCanvas()->measureText(text, ucslen(text), &tw, &th, &ascent);
				drawUnicode(display,pixmap,
							this->page->getCanvas()->copyGC(),
							x_offset+x,y_offset+y,text,ucslen(text));
				if(selected) {
					const aString rcolor = "black";
					GC bgc = this->page->getCanvas()->colorGC(rcolor.string());
					XDrawRectangle(display, pixmap, bgc,
								   x_offset + x - RECT_GAP,
								   y_offset + y - ascent - RECT_GAP,
								   tw + 2 * RECT_GAP,
								   th + 2 * RECT_GAP);
				}
			}
		} 
		if(s)
			// text was malloced so must be freed
			i_FreeUText(s);
	}	
}


/* 
// can all redisplay either from with CE or OOCPage
// so we need something makes the pixmap and everything
*/
void
CE::redrawCell()
{
	ViewCanvas *vc = this->page->getCanvas();
	OOCPage *opage = (OOCPage *) this->page;
	if(!(this->page->managed && vc)) return;
	
	Widget w = this->page->getCanvas()->canvas;
	if(XtIsManaged(w) == FALSE)
		return;

	if(this->sub_page != opage->current_page)
		return;
	
	Display *display;
	Drawable drawable;
	
	display = XtDisplay(w);
	drawable = RootWindowOfScreen(XtScreen(w));
	int depth = DefaultDepthOfScreen(XtScreen(w));
	int width  = opage->cell_width;
	int height = opage->cell_height;
	
	Pixmap pixmap = XCreatePixmap(display, drawable, width, height, depth);
	this->redisplay(display, pixmap, vc, 0, 0, width, height);
	XCopyArea(display, pixmap, XtWindow(vc->canvas), vc->copyGC(), 
			  0, 0, width, height, this->x, this->y);
	
	XFreePixmap(display, pixmap);
}


/*
// it is used when an
// entity is deleted and we want to blank that area
// of the screen -- and no more, leave it blank
*/
void
CE::blankGhost()
{
	OOCPage *opage = (OOCPage *) this->page;
	ViewCanvas *vc = opage->getCanvas();
	Widget w = vc->canvas;
	Pixmap pixmap;
	if(XtIsManaged(w) == FALSE)
		return;
	
	Display *display = XtDisplay(w);
	Drawable drawable = RootWindowOfScreen(XtScreen(w));
	iLBox refresh_box;
	refresh_box.x = this->x;
	refresh_box.y = this->y;
	refresh_box.w = opage->cell_width;
	refresh_box.h = opage->cell_height;
	
	int depth = DefaultDepthOfScreen(XtScreen(w));
	    
	if(this->page->managed && vc && this->sub_page == opage->current_page){
		pixmap = XCreatePixmap(display, drawable, 
							   refresh_box.w, refresh_box.h,
							   depth);
		XFillRectangle(display, pixmap, 
					   vc->grayGC(),
					   0, 0,
					   refresh_box.w, refresh_box.h);
		XCopyArea(display, pixmap, XtWindow(vc->canvas), 
				  vc->copyGC(),
				  0, 0,
				  refresh_box.w, refresh_box.h,
				  refresh_box.x-vc->pix_hoffset, refresh_box.y-vc->pix_voffset);

		XFreePixmap(display, pixmap);
	}
#if 0
	if(this->page->managed && vc && this->sub_page == opage->current_page){
		// clear because we want to look as if never drawn
		// so should be in background color
		XClearArea(XtDisplay(vc->canvas), XtWindow(vc->canvas),
				   this->x, this->y,
				   opage->cell_width, opage->cell_height,
				   FALSE);
	}
#endif
}

/*
*/
void
CE::assignPosition(int input_sub_page, Position input_x, Position input_y)
{
	this->sub_page = input_sub_page;
	this->x = input_x;
	this->y = input_y;
}

iBox
CE::computeBounds()
{
	iBox bounds;
	OOCPage *opage = (OOCPage *) this->page;
	
	bounds.x=this->x;
	bounds.y=this->y;
	bounds.w=opage->cell_width;
	bounds.h=opage->cell_height;
	return bounds;
}


/*
// these are the messages
*/

/* 
// without knowing what the change will be to
// we should not do anything
*/
void 
CE::willUpdate()
{
}

/*
// I would not have to do anything if it weren't for the move function
*/
void 
CE::willMove()
{
}

/* 
*/
void 
CE::destroy()
{
	this->blankGhost();
	this->deleteAndUpdate();
}

/* 
// we send a message to the page because we may even have to be moved to
// a new page
*/
void 
CE::update()
{
	MainWindow *mw = this->getMainWindow();
	OOCPage *opage = (OOCPage *) this->page;
	iEntityPtr en = this->entity;
	opage->change(en);
	ModeType mode_type = opage->mode_type;
	Mode *mode = (Mode *) opage->mode;

	if (opage->should_delete) {
		if (opage->managed)
			if (!mode->forward())
				if(!mode->back())
					mw->view_canvas->unmanage();
		mw->standardTitle();
		mw->refresh_dialogs();
		//delete page;
	}

	// why are we doing this -- because we can be moved
	// to a new page and 'this' will be deleted so
	// have to get our new face
	// also I have a lot of temporary variables
	// like en so I don't dereference 'this'
	CE *ce;
	ce = (CE *) eaGetFace(en, mode_type);
	if(ce && ce->page->managedP()){
		opage = (OOCPage *) ce->page;
		// the whole page can be changed because 
		// the new characters will get sorted in
		// so we order a whole page refresh
		// this is done for everyone who
		// changes but that is harmless since
		// the multiple refresh orders will be compressed
		opage->refresh();
	}
}


/*
// if we couldn't move we wouldn't have to do this
// we also could make this more directly respond
// than update -- we know for example that we
// will stay in the same page
*/
void 
CE::move()
{
  // bogus do we really handle the fact the Cell sizes could change
  OOCPage *opage = (OOCPage *) this->page;
  if(opage->managedP()) {
	opage->conditionallyInitialiseCanvas();
	opage->refresh();
  }
}

void 
CE::select()
{
	this->redrawCell();
}

void 
CE::unSelect()
{
	this->redrawCell();
}

/*
// bogus - we wouldnt have to do anything if we didnt have to handle the 
// image modes bs
// can recognize happen when we are in mode?
// and why blankGhost?
*/
void 
CE::recognize()
{
	this->update();
}

/*
// entities are created when you type in text mode
// in which case you don't have a bounding box
// or when you draw an entity in image mode
// where it won't have a label right away
// however is there a page for unlabelled glyphs?
// should I be transported to that page
*/
void 
CE::create()
{
	// do nothing
}

MainWindow *
CE::getMainWindow()
{
	return this->page->mode->main_window;
}



