
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"

#include <X11/Xutil.h>

#include <Xm/MwmUtil.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/Form.h>
#include <Xm/PushB.h>
#include <Xm/DialogS.h>

#include "misc/color.h"
#include "misc/XArg.h"
#include "misc/misc.h"
#include "misc/prompt.h"
#include "svpmap.h"
#include "dllist.h"
#include "smemory.h"

#include "MainWindow.h"
#include "Application.h"
#include "ViewCanvas.h"
#include "EntityAux.h" // for conf class
#include "bb.h" // for bbNullP
#include "ColorDialog.h"
#include "Document.h"
#include "rgb.h"
#include "config.h"

#define c_MARGIN 0
#define c_SHADOW_WIDTH 2

void
ColorDialog::create()
{
	XArg args;
	Widget parent = this->app_shell;

	args.reset();
	args.setArg(XmNtitle,  "Color Table");
	args.setArg(XmNnoResize, True);
	//args.setArg(XmNdefaultPosition, FALSE);
	this->dialog = XmCreateFormDialog(parent,
									  "form",
									  args.getArgs(),
									  args.getSize());
	this->updateColors();
	

}

ColorDialog::ColorDialog(MainWindow * input_mw, Widget input_app_shell)
	: Dialog(input_mw, input_app_shell), color_box(0),popper(0)
{
}

ColorDialog::~ColorDialog()
{
	if (this->popper)
		XtDestroyWidget(this->popper);
}

void
ColorDialog::updateColors()
{
	int i;

	Widget *box_w;		
	for(i=1; i_TypeToName(i); i++);
	box_w = (Widget*)smalloc(sizeof(Widget)*i);

	if (!box_w)
		return;

	if(this->color_box){
		XtDestroyWidget(this->color_box);
		this->color_box = 0;
	}
											
	this->color_box = 
		XtVaCreateWidget("colorbox",
						 xmRowColumnWidgetClass, this->dialog, 
						 XmNpacking, XmPACK_COLUMN, 
						 XmNorientation, XmHORIZONTAL,
						 XmNnumColumns,     10, 
						 XmNentryAlignment, XmALIGNMENT_CENTER, 
						 NULL);

	/* simply loop thru the strings creating a widget for each one */
	char *color;

	void *v = this;
	XtCallbackProc cb = (XtCallbackProc) pickIt;


	for (i = 0; !i||i_TypeToName(i); i++) {
		color=getTypeColor((char*)i_TypeToName(i));


#if 0			
		XtVaCreateWidget("form", xmFormWidgetClass, 
						 this->color_box, NULL);
#endif

		Widget b;
		if (!i)
			b = XtVaCreateWidget("<NULL>", 
								 xmPushButtonWidgetClass, this->color_box,
								 XmNmarginWidth, c_MARGIN, 
								 XmNmarginHeight,c_MARGIN, 
								 XmNshadowThickness, c_SHADOW_WIDTH, 
								 XmNshadowType, XmSHADOW_IN, 
								 XmNbackground, getColorByName(this->mw->view_canvas->canvas, color),
								 NULL);
		else
			b = XtVaCreateWidget(i_TypeToName(i), xmPushButtonWidgetClass,
								 this->color_box, 
								 XmNmarginWidth, c_MARGIN, 
								 XmNmarginHeight, c_MARGIN, 
								 XmNshadowThickness, c_SHADOW_WIDTH, 
								 XmNshadowType, XmSHADOW_IN, 
								 XmNbackground, getColorByName(this->mw->view_canvas->canvas, color),
								 NULL);
		XtManageChild(b);
		XtAddCallback(b, 
			  XmNactivateCallback,
			  cb, 
			  v);

		
		box_w[i] = b;
	}
	XtManageChild(this->color_box);


}

void
ColorDialog::update()
{	
	int shown;
	
	if (this->dialog){
		shown = XtIsManaged(this->dialog);
		if (shown){
			this->updateColors();
		}
	}
}

static void selColor(Widget w,
	   char *name,
	   XmAnyCallbackStruct *cbs)
{
	strcpy(name,XtName(w));
}

static void CloseColor(XColor *color,char *name)
{	
	int r,g,b;
	r = (color->red >> 8) & 0xff;
	g = (color->green >> 8) & 0xff;
	b = (color->blue >> 8) & 0xff;
	sprintf(name,"#%02x%02x%02x",r,g,b);
}

void ColorDialog::createPopup()
{
	XColor nc;
	Display *display  = XtDisplay(this->color_box);
	int screen  = DefaultScreen(display);
	Colormap map = DefaultColormap(display, screen);
	Widget b;
	XArg args;
	Widget dialog;
	XmString title=XmStringCreateSimple("Choose Color");
	XmString foo = XmStringCreateSimple("  ");

	args.reset();
	args.setArg(XmNdialogTitle,title);
	args.setArg(XmNdialogStyle,XmDIALOG_FULL_APPLICATION_MODAL);
	this->popper = XmCreateFormDialog(this->mw->app->app_shell,"popup dlog",
					args.getArgs(),args.getSize());
	args.reset();
	args.setArg(XmNpacking,XmPACK_COLUMN);
	args.setArg(XmNorientation,XmHORIZONTAL);
	args.setArg(XmNnumColumns,16);
	args.setArg(XmNentryAlignment,XmALIGNMENT_CENTER);
	dialog = XmCreateRowColumn(this->popper,
									  "Color Picker",
									  args.getArgs(),
									  args.getSize());
	
	for (nc.pixel = 0;nc.pixel < DisplayCells(display,screen);nc.pixel++) {
		XQueryColor(display,map,&nc);
		CloseColor(&nc,this->name);
		b = XtVaCreateWidget(this->name, xmPushButtonWidgetClass,
			 dialog, 
			 XmNmarginWidth, c_MARGIN, 
			 XmNmarginHeight, c_MARGIN, 
			 XmNlabelString, foo,
#if 0
							 XmNshadowThickness, c_SHADOW_WIDTH, 
							 XmNshadowType, XmSHADOW_IN, 
#endif
			 XmNbackground, nc.pixel,
			 NULL);
		XtManageChild(b);
		XtAddCallback(b,XmNactivateCallback,(XtCallbackProc) selColor,(void*)this->name);
	}
	XmStringFree(title);
	XmStringFree(foo);
	XtManageChild(dialog);
	XtManageChild(this->popper);
}

static void pickIt(Widget w,
	   ColorDialog *cd,
	   XmAnyCallbackStruct *cbs)
{

	if (!cd->popper)
		cd->createPopup();
	else XtManageChild(cd->popper);
	cd->name[0] = 0;
	
	XtPopup(XtParent(cd->popper), XtGrabNone);
	XMapRaised(XtDisplay(cd->popper), XtWindow((cd->popper)));
    while (XtIsManaged(cd->popper) && !cd->name[0]) {
        XtAppProcessEvent(cd->mw->app->app_context, XtIMAll);
        XSync(XtDisplay(cd->popper), 0);
    }
    if (!XtIsManaged(cd->popper)) return;

	setTypeColor(XtName(w), cd->name);

	cd->mw->color_view->update();

	cd->mw->redrawImageMode();

	XtUnmanageChild(cd->popper);
}
