
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"

#include "c-type.h"
#include "smemory.h"

#include <Xm/DialogS.h>
#include <X11/Xutil.h>

#include <Xm/Text.h>
#include <Xm/Label.h>
#include <Xm/TextF.h>
#include <Xm/Form.h>
#include <Xm/ScrolledW.h>
#include <Xm/MessageB.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/PanedW.h>
#include <Xm/Frame.h>
#include <Xm/Separator.h>



#include "rafutil.h"
#include "ucs.h"

#include "misc/XArg.h"

#include "Application.h"

#include "ContentDialog.h"

/* because Motif include files aren't understood by GDB/G++ */
struct xmtb{
    char *ptr;                  /* Pointer to data. */
    int length;                 /* Number of bytes of data. */
    XmTextFormat format;       /* Representations format */
};

typedef struct xmtb xmtb;

struct xmtvcbs
{
    int reason;
    XEvent  *event;
    Boolean doit;
    long currInsert, newInsert;
    long startPos, endPos;
    xmtb *text;
};

typedef struct xmtvcbs xmtvcbs;

static void 
cdOkCB(Widget w, ContentDialog *sl, void *call)
{
	sl->okCB((XmFileSelectionBoxCallbackStruct *)call);
}

static void 
cdCancelCB(Widget w, ContentDialog *sl, void *call)
{
	sl->cancelCB((XmFileSelectionBoxCallbackStruct *)call);
}

void make_string(Widget w,xmtvcbs *cbs,char *num)
{
	int i,len,diff;

	len = XmTextGetLastPosition(w);
	strcpy(num,XmTextGetString(w));
	diff = cbs->endPos - cbs->startPos;
	if (diff) {
		for (i=cbs->endPos;num[i];i++)
			num[i - diff] = num[i];
		num[i - diff] = 0;
	}
	len = strlen(num);
	for (i=len;i>= cbs->startPos;i--)
		num[i + cbs->text->length] = num[i];
	for (i=0;i<cbs->text->length;i++)
		num[i + cbs->startPos] = cbs->text->ptr[i];
}

#define ishex(x) ((x >= '0' && x <= '9') || \
					(x >= 'a' && x <= 'f') || \
					(x >= 'A' && x <= 'F'))

void
verify_short(Widget w,  void *user_data, xmtvcbs *cbs)
{
	int len = 0,i,hex;
	char num[20],*p;

	make_string(w,cbs,num);
	p = num;
	if (hex = !strncmp(p,"0x",2)) p += 2;
	while (*p == '0') p++;
	len = strlen(p);
	if (!len) return;
	cbs->doit = 0;
	if ((hex && len > 4) || len > 5)
		return;
	for (i=0;i<len;i++)
		if ((hex && !ishex(p[i])) || (!hex && !isdigit(p[i])))
			return;
	sscanf(p,(hex?"%x":"%d"),&len);
	if (len > 0xffff)
		return;
	cbs->doit = 1;
}

void
verify_value(Widget w,  void *user_data, xmtvcbs *cbs)
{
	int len = 0,i,uni;
	char str[20];

	make_string(w, cbs, str);
	len = strlen(str);
	cbs->doit = 1;
	if (!(uni = len > 1)) return;
	cbs->doit = 0;
	for (i=0;i<len;i++)
		if (!ishex(str[i]))
			return;
	sscanf(str,"%x",&len);
	if (len > 0xffff)
		return;
	cbs->doit = 1;
}

void
verify_long(Widget w,  void *user_data, xmtvcbs *cbs)
{
	int len = 0,i,hex;
	char num[20],temp[20],*p;

	make_string(w,cbs,num);
	p = num;
	if (hex = !strncmp(p,"0x",2)) p += 2;
	while (*p == '0') p++;
	len = strlen(p);
	if (!len) return;
	cbs->doit = 0;
	if ((hex && len > 8) || len > 10)
		return;
	for (i=0;i<len;i++)
		if ((hex && !ishex(p[i])) || (!hex && !isdigit(p[i])))
			return;
	if (!hex) {
		sscanf(p,"%d",&len);
		sprintf(temp,"%d",len);
		if (strcmp(temp,p)) return;
	}
	cbs->doit = 1;
}

void 
cdSetType(Widget    w,
		int tag,
		XtPointer    call_data)
{
	ContentDialog *cd;
	XArg args;

	args.setArg(XmNuserData, &cd);
	args.getValues(w);
	
	cd->setType(tag);
}


ContentDialog::ContentDialog(MainWindow *input_mw, 
							 Widget input_app_shell, 
							 XtAppContext input_app_context,
							 MenuItem *input_menu,
							 int input_max_tag,
							 char *input_title)
	: ModalDialog(input_mw, input_app_shell, input_app_context),
	  changed(FALSE),
	  tag(-1),
	  max_tag(input_max_tag),
	  menu(input_menu),
	  title(input_title),
	  separator(0), frame(0), type_menu(0)
{		
	int i;

	this->content_widgets = (Widget *) smalloc(sizeof(Widget) * this->max_tag);
	for(i = 0; i < this->max_tag; i++){
		this->content_widgets[i] = 0;
	}
}

ContentDialog::~ContentDialog()
{
	smfree(this->content_widgets);
}


void
ContentDialog::createBase()
{
	XArg args;
	XmString xms;
	XmString cancel_string, ok_string;
	args.reset();
	xms = XmStringCreateLocalized("Edit Contents");
	ok_string = XmStringCreateLocalized("OK");
	cancel_string = XmStringCreateLocalized("Cancel");

	args.setArg(XmNdialogTitle, xms);
	args.setArg(XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL);
	args.setArg(XmNcancelLabelString, cancel_string);
	args.setArg(XmNokLabelString, ok_string);
	
	this->dialog = XmCreateTemplateDialog(this->app_shell, 
										   "select Choices",
										   args.getArgs(), 
										   args.getSize());

	// dont be swallowing all returns
	args.reset();
	args.setArg(XmNdefaultButton, 0);
	args.setValues(this->dialog);

	XmStringFree(xms);
	XmStringFree(cancel_string);
	XmStringFree(ok_string);

	XtAddCallback(this->dialog, XmNokCallback, cdOkCB, this);
	XtAddCallback(this->dialog, XmNcancelCallback, cdCancelCB, this);

	Widget rc;


	/* put things in a form */
	args.reset();
	rc = XmCreateForm(this->dialog, "rc", 
							 args.getArgs(), args.getSize());

	/* make the type menu and attach it at the bottom */
	this->type_menu = 
		BuildMenu(rc, XmMENU_OPTION, "Type", 'T', this->menu);

	int i;
	args.reset();
	args.setArg(XmNuserData, this);
	for(i = 0; this->menu[i].result != 0; i++)
		args.setValues(this->menu[i].result);

	args.reset();
	args.setArg(XmNleftAttachment, XmATTACH_FORM);
	args.setArg(XmNrightAttachment, XmATTACH_FORM);
	args.setArg(XmNbottomAttachment, XmATTACH_FORM);
	args.setValues(this->type_menu);


	args.reset();
	args.setArg(XmNleftAttachment, XmATTACH_FORM);
	args.setArg(XmNrightAttachment, XmATTACH_FORM);
	args.setArg(XmNbottomAttachment, XmATTACH_WIDGET);
	args.setArg(XmNbottomWidget, this->type_menu);

	this->separator = XmCreateSeparator(rc, "separator",
								  args.getArgs(), args.getSize());
	
	args.reset();
	args.setArg(XmNleftAttachment, XmATTACH_FORM);
	args.setArg(XmNrightAttachment, XmATTACH_FORM);
	args.setArg(XmNbottomAttachment, XmATTACH_WIDGET);
	args.setArg(XmNbottomWidget, separator);
	args.setArg(XmNtopAttachment, XmATTACH_FORM);

	this->frame = XmCreateScrolledWindow(rc, "frame",
										 args.getArgs(), args.getSize());



	XtManageChild(this->frame);
	XtManageChild(separator);
	XtManageChild(this->type_menu);
	XtManageChild(rc);
}

void
ContentDialog::createDefault()
{
	XArg args;
	XmString xms;
	Widget label;
	int i;

	// default to no method
	xms = XmStringCreateLocalized("No method for editing this data");
	args.reset();
	args.setArg(XmNlabelString, xms);
	label = XmCreateLabel(this->frame, "label",
				  args.getArgs(), args.getSize());
	XmStringFree(xms);

	for(i = 0; i < this->max_tag; i++){
		this->content_widgets[i] = label;
	}
}

void
ContentDialog::create()
{
	this->createBase();
	this->createDefault();
	this->createCustom();
	//this->setType(0);
}



void
ContentDialog::setType(int input_tag)
{
	Widget old_widget;
	Widget new_widget;
	XArg args;
	Widget new_menu;
	Widget old_menu;
	int i;

	if(this->type_menu == 0) {
		this->tag = input_tag;
		return;
	}
	args.reset();
	args.setArg(XmNmenuHistory, &old_menu);
	args.getValues(this->type_menu);
	new_menu = this->menu[input_tag].result;

	if(new_menu != old_menu){
		args.reset();
		args.setArg(XmNmenuHistory, new_menu);
		args.setValues(this->type_menu);
		XtManageChild(this->type_menu);
	}

	new_widget = this->content_widgets[input_tag];
	this->tag = input_tag;
												 
	old_widget = 0;
	for(i = 0; i < this->max_tag; i++){
		old_widget = this->content_widgets[i];
		if(XtIsManaged(old_widget))
			break;
	}
	if(i == this->max_tag)
		old_widget = 0;

	if(old_widget != new_widget){
		if(old_widget)
			XtUnmanageChild(old_widget);
		XtManageChild(new_widget);
		XmScrolledWindowSetAreas(this->frame, 0, 0, new_widget);
	}
}

void
ContentDialog::okCB(XmFileSelectionBoxCallbackStruct *cbs)
{
	this->done = TRUE;
	this->changed = TRUE;

	this->getFromDialog();

	this->hide();
}

void
ContentDialog::cancelCB(XmFileSelectionBoxCallbackStruct *cbs)
{
	this->done = TRUE;
	this->changed = FALSE;

	this->hide();
}

/*
// put the usual event loop in here
*/
void
ContentDialog::showModal()
{
	// tag is set
	// data is set
	if(this->dialog == 0)
		this->create();

	this->setType(tag);

	this->putInDialog();

	this->changed = FALSE;
	XtManageChild(this->dialog);

	this->ModalDialog::showModal();
}
