
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef _EntityAux_h_
#define _EntityAux_h_



#include "illum.h"
#include "uString.h"

class EntityFace;
class MainWindow;

enum SelectionClass { scNOT = 0, scITEM = 1, scRANGE = 2};


enum ConfClass {
	agNONREC = 1,
	agLOWREC = 2,
	agMULTIREC = 4,
	agOZERO = 8,
	agOTHER = 16,
};

class EntityAux;

void        eaSetFace(iEntityPtr, ModeType, EntityFace *);
EntityFace *eaGetFace(iEntityPtr, ModeType);

void setEntitySelection(iEntityPtr);
void setEntitySelection(iEntityPtr, int from, int to);
void setEntitySelection(iEntityPtr, SelectionClass sc, int from, int to);
void clearEntitySelection(iEntityPtr);
SelectionClass  getEntitySelection(const iEntityPtr);
void getEntitySelectionRange(const iEntityPtr, int& from, int& to);
void toggleEntitySelection(iEntityPtr);
uString getFirstOfPosset(const iEntityPtr entity);
void flushTextModeCache();


// gets the controller or creates it
EntityAux *getController(iEntityPtr);
// removes the controller but doesnt free it
void removeController(iEntityPtr);
// get controller if it exists
EntityAux *getControllerMaybe(iEntityPtr);

#if 0
const iUCode *eaGetTextReally(const iEntityPtr entity);
#endif

ConfClass eaGetGlyphClass(const iEntityPtr entity);
ConfClass eaConfClassify(iEntityPtr entity);

int entityFlaggedP(iEntityPtr entity);
void selectText(iEntityPtr entity);

/* message responders */
void imageDestroy(MainWindow *, iImagePtr);

void entityClearFlag(MainWindow *mw, iEntityPtr entity);
void entitySetFlag(MainWindow *mw, iEntityPtr entity);
void entityToggleFlag(MainWindow *mw, iEntityPtr entity);

class EntityAux {
public:
	// selection data
	SelectionClass selection_class;
	short from;     // could these be shorts
	short to;
	ConfClass glyph_class;
	char show; // Hierarcy mode uses this to show the children or not

	// move some stuff to IE.h
	
	short flagged;            // could these be smaller
	// back pointers?
	// or my responsibility to delete?
	// I think the entity gets disposed
	// then the controller get deleted
	// then he informs the modes
	iEntityPtr back_pointer;
	
	EntityFace *faces[NUM_MODE_TYPES];
	
	EntityAux(iEntityPtr);
	~EntityAux();
	
	// recompute cached data
	void init();
};

#endif
