
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"

#include "illum.h"

#include <X11/Xutil.h>

#include <Xm/MwmUtil.h>
#include <Xm/ScrolledW.h>
#include <Xm/Text.h>
#include <Xm/Frame.h>
#include <Xm/Form.h>

#include "misc/color.h"
#include "misc/XArg.h"
#include "misc/misc.h"
#include "svpmap.h"
#include "dllist.h"
#include "smemory.h"

#include "SelectionManager.h"
#include "MainWindow.h"
#include "ViewCanvas.h"
#include "EntityAux.h" // for conf class
#include "bb.h" // for bbNullP
#include "HierDialog.h"

#define MAX_HEIGHT 4
#define LINE_LENGTH 20

void 
HierDialog::addstring(iEntityPtr ent,iEntityPtr child, char** s, int height)
{
	char	*temp, *data=0, *data2, *temp2;
	iEntityPtr cur;
	int i;
	
	temp=(char*)smalloc(256);
	temp[0]='\0';
	for (i=0; i<height; i++)
		temp = strcat(temp, "  ");
	temp=strcat(temp, i_TypeToName(i_GetEntityType(ent)));
	if (ent==child)
		temp=strcat(temp, " <- ");
	else
		temp=strcat(temp, "    ");
	if (i_GetEDataType(ent)==iDString)
		data=(char*)i_GetText(ent,0);
	else if (i_GetEDataType(ent)==iDUniString)
		temp=strcat(temp,"Uni String");
	else if (i_GetEntityType(ent) == iGlyphT) {
		temp2=(char*)smalloc(2);
		temp2[0]=(char)i_GetGlyph(ent, 0);
		temp2[1]='\0';
		temp=strcat(temp,temp2);
		smfree(temp2);
	}
	if (data){
		data2=(char*)smalloc(strlen(data)+1);
		if (!data2){
			smfree(*s);
			*s="OUT OF MEMORY\n";
			return;
		}
		for(i=0; data[i]; i++)
			if (data[i]=='\n')
				data2[i]='$';
			else
				data2[i]=data[i];
		data2[i]='\0';
		temp=strcat(temp, data2);
		smfree(data2);
	}
	temp=strcat(temp, "\n");
	if (!(strlen(temp)+strlen(*s)<this->allocated_size)){
		this->allocated_size+=500;
		*s=(char*)smrealloc(*s,this->allocated_size);
	}
	if (*s)
		*s=strcat(*s, temp);
	height++;
	smfree(temp);
	for (cur = i_GetChildEntity(ent);cur&&height<MAX_HEIGHT&&s;cur = i_GetNextEntity(cur))
		this->addstring(cur,child, s, height);
}

void HierDialog::make_string(iEntityPtr ent, char **s)
{
	iEntityPtr child=0;
	
	child=ent;
	if (!i_GetChildEntity(ent)&&i_GetParentEntity(ent))
		ent=i_GetParentEntity(ent);
	
	*s=(char*)smalloc(500);
	this->allocated_size=500;
	if (*s){
		*s[0]='\0';
		this->addstring(ent,child, s, 0);
	}
	
}
	
	
void
HierDialog::create()
{
	XArg args;
	
	Widget parent = this->app_shell;
		

	args.reset();
	args.setArg(XmNtitle,  "Hierarchy Popup");
	args.setArg(XmNnoResize, True); 
	this->dialog = XmCreateFormDialog(parent,
									  "form",
									  args.getArgs(),
									  args.getSize());
											
	args.reset();
	args.setArg(XmNheight, 500);
	args.setArg(XmNwidth, 190);
	args.setArg(XmNscrollingPolicy, XmAUTOMATIC);
	this->win = XmCreateScrolledWindow(this->dialog,"SW",  
									   args.getArgs(),
									   args.getSize());

	if (!this->text){
		args.reset();
		args.setArg(XmNbackground, 
					getColorByName(this->mw->view_canvas->canvas, "white"));
		args.setArg(XmNforeground, 
					getColorByName(this->mw->view_canvas->canvas,"black")); 
		args.setArg(XmNshadowType, XmSHADOW_OUT);
		Widget frame=XmCreateFrame(this->win, "frame", 		
								   args.getArgs(),
								   args.getSize());
		args.reset();
		this->text = XmCreateText(frame,"text",  
								  args.getArgs(),
								  args.getSize());
			
		args.reset();
		args.setArg(XmNmarginWidth, 20);
		args.setArg(XmNmarginHeight, 15);
		args.setArg(XmNeditable, FALSE);
		args.setArg(XmNresizeHeight, True);
		args.setArg(XmNresizeWidth, True);
		args.setArg(XmNcursorPositionVisible, False);
		args.setArg(XmNbackground, 
					getColorByName(this->mw->view_canvas->canvas, "white"));
		args.setArg(XmNforeground, 
					getColorByName(this->mw->view_canvas->canvas,"black")); 
		args.setValues(this->text);
		XtManageChild(this->text);
		XtManageChild(frame);
		XtManageChild(this->win);
	}
}

HierDialog::HierDialog(MainWindow * input_mw, Widget input_app_shell)
	: Dialog(input_mw, input_app_shell),
	  win(0), text(0), dirtyview(0)
{
}

HierDialog::~HierDialog()
{
}
void
HierDialog::setEntity(iEntityPtr input_entity)
{
	if ((this->entity == input_entity)) return;

	this->dirtyview = 1;
	this->entity = input_entity;

	this->update(0);
}
void
HierDialog::update(iEntityPtr ent)
{	
	char *empty_message = "  *     NO     *\n  *   ENTITY   *\n  *  SELECTED  *\n";
	char *memory_message = "   OUT OFF MEMORY   \n";
	char *s;
	XArg args;
	
	if (!this->text||(!this->dialog || !XtIsManaged(this->dialog)))
		return;
	
	if (!(this->dirtyview || (this->entity && (!ent || (ent == this->entity)))))
		if (this->dirtyview||this->entity)
			return;

	this->dirtyview=0;	
	if (this->entity){
		this->make_string(this->entity, &s);
		if (!s)
			s = memory_message;
	} 
	else
		s = empty_message;

	args.reset();
	args.setArg(XmNvalue, s);
	args.setValues(this->text);
	if ((s!=empty_message)&&(s!=memory_message))
		smfree(s);
	this->allocated_size=0;

}

