
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <X11/Xutil.h>
#include <Xm/Xm.h>

#include "rafutil.h"

#include "svpmap.h"
#include "misc/scroll.h"
#include "misc/bb.h"

#include "EntityAux.h"
#include "MainWindow.h"
#include "ChooseDialog.h"
#include "Document.h"
#include "SelectionManager.h"
#include "ViewCanvas.h"
#include "ImageMode.h"
#include "ImagePage.h"
#include "IE.h"

static char *
makeKey(iImagePtr image, char buffer[24])
{
	sprintf(buffer, "%d", image);
	return buffer;
}

void
ImageMode::createPages(DLList *image_list, DLList *entity_list)
{
	stringVoidPtrMap map;
	int page_count = 0;
	Ref r;
	iEntityPtr entity;
	iImagePtr image;
	char buffer[24];
	char buffer2[24];
	char *key;
	ImagePage *page;
	
	svmapInit(&map, 0, 10001);
	
	for(r = dllistFirst(image_list); r; dllistNext(image_list, &r)){
		image = (iImagePtr) dllistContents(image_list, r);
		key = makeKey(image, buffer);
		
		sprintf(buffer2, "%d", ++page_count);
		page = new ImagePage(image, this, buffer2);
		svmapSet(&map, key, page);
		this->addPage(page);
	}
	
	for(r = dllistFirst(entity_list); r; dllistNext(entity_list, &r)) {
		entity = (iEntityPtr) dllistContents(entity_list, r);
		
		image = i_GetImage(entity);
		if(image) {
			key = makeKey(image, buffer);
			page = (ImagePage *) svmapGet(&map, key);
			page->addMember(entity);
		}
	}
	
	for(r = dllistFirst(image_list); r; dllistNext(image_list, &r)){
		image = (iImagePtr) dllistContents(image_list, r);
		key = makeKey(image, buffer);
		page = (ImagePage *) svmapGet(&map, key);
		
		page->initialise();
	}
	
	svmapCleanup(&map);
}

void ImageMode::imageDeleted(iImagePtr image)
{
	Ref r;
	DLList *fl = &this->page_list;
	ImagePage *page;
	
	for(r = dllistFirst(fl); r; dllistNext(fl, &r)) {
		page = (ImagePage *) dllistContents(fl, r);
		if (page->image==image) {
			page->image=0;
			page->should_delete=TRUE;
		}
	}
}


void
ImageMode::findImagesSub(iEntityPtr entity, stringVoidPtrMap *map, DLList *image_list)
{
	iImagePtr image;
	char buffer[24];
	char *key;
	iEntityPtr child;
	
	image = i_GetImage(entity);
	
	if(image) {
		key = makeKey(image, buffer);
		if(svmapContainsP(map, key) == FALSE){
			svmapSet(map, key, image);
			dllistAppend(image_list, image);
		}
	}
	
	for(child = i_GetChildEntity(entity); child; child = i_GetNextEntity(child)) {
		this->findImagesSub(child, map, image_list);
	}
}

void
ImageMode::findImages(iEntityPtr input_root_entity, DLList *image_list)
{
	stringVoidPtrMap map;
	svmapInit(&map, 0, 10001);
	
	this->findImagesSub(input_root_entity, &map, image_list);
	
	svmapCleanup(&map);
}

void
ImageMode::createPages(iEntityPtr input_root_entity)
{
	DLList entity_list;
	DLList image_list;
	
	dllistCtor(&entity_list);
	dllistCtor(&image_list);
	
	this->findImages(input_root_entity, &image_list);
	this->main_window->chooser->choose(input_root_entity, &entity_list);
	this->createPages(&image_list, &entity_list);
	
	dllistDtor(&image_list);
	dllistDtor(&entity_list);
}

ImageMode::ImageMode(MainWindow *input_main_window,
					 iEntityPtr input_root_entity)
	: Mode(input_main_window)
{
	this->createPages(input_root_entity);
}

ImageMode::~ImageMode()
{
}

void
ImageMode::centerOnSelected(iEntityPtr entity)
{
	IE *ie;
	ie = (IE *) eaGetFace(entity, vtIMAGE);
	DLList *l = &this->page_list;
	Ref r;
	if (!ie) return;
	r = dllistSeek(l, ie->page);
	if(r){
		setCurrentPage(r);
		ie->page->centerOnSelected(entity);
	}      
}

void
ImageMode::reconsider(iEntityPtr entity)
{
	DLList *l = &this->page_list;
	Ref r;
	ImagePage *page;
	iImagePtr image = i_GetImage(entity);
	
	for(r = dllistFirst(l); r; dllistNext(l, &r)) {
		page = (ImagePage *) dllistContents(l, r);
		if(image == page->image)
			break;
	}
	
	if(r == 0)
		halt("couldnt find the page.\n");
	
	page->checkEntity(entity);
}

void
ImageMode::addNewEntity(iEntityPtr new_entity)
{
	// bogus
	DLList *l = &this->page_list;
	Ref r;
	ImagePage *page;
	iImagePtr image = i_GetImage(new_entity);
	
	for(r = dllistFirst(l); r; dllistNext(l, &r)) {
		page = (ImagePage *) dllistContents(l, r);
		if(image == page->image)
			break;
	}
	
	if(r == 0)
		halt("couldnt find the page.\n");
	
	page->addNewMember(new_entity);
	
	if(page->tree.initialized())
		page->tree.addItem(new_entity);
}

void
ImageMode::choose(int buttons_changed)
{
	Ref r;
	DLList *l = &this->page_list;
	ImagePage *page;
	ImagePage *current_page;
	iImagePtr image;
	int4 x, y;
	int managed;

	// find the page in mode and remember its image
	// then carefully junk all the current pages
	
	current_page = (ImagePage *) this->getCurrentPage();
	managed = current_page && current_page->managedP();
	
	if(((current_page && current_page->invalid_guy_added) || buttons_changed) == FALSE)
		return;

	// dont clear -- we used to do this because we really hadn't fix
	// the delete of the page
#if 0
	if(current_page)
		current_page->clear();
#endif
	
	if(managed) {
		image = current_page->image;
		getScrollPositionCenter(current_page->mode->scrolled_window, &x, &y);
		current_page->unmanage();
	} else {
		image = 0;
	}
	
	this->current_page_ref = 0;
	
	for(r = dllistFirst(l); r; dllistNext(l, &r)){
		page = (ImagePage *) dllistContents(l, r);
#if 0	
		// this is a cheese -- it deletes the page -- and delete page
		// doesn't kill the ea->faces 
		// we should check out what changeMode does
		delete page;
#else
		page->purge();
#endif
	}
	dllistClear(l);
	iEntityPtr root_entity = this->main_window->document->root_entity;
	this->createPages(root_entity);
	r = 0;
	// if there used to be an image in mode, let's find it
	// among the new pages
	
	if(managed) {
		for(r = dllistFirst(l); r; dllistNext(l, &r)){
			page = (ImagePage *) dllistContents(l, r);
			if(image == page->image)
				break;
		}
	}
	
	if(managed) {
		// we found the new page that corresponds to the image in the
		// one that used to be in mode
		if(r){
			this->setCurrentPage(r);
			page->getCanvas()->scrollToPosition(page->mode->scrolled_window, x, y);
		} else {
			this->start();
		}
	}

	// go through the list of entities selected
	// and unselect the ones that no longer have IE's
	const DLList *selected;
	DLList copy;
	dllistCtor(&copy);
	selected = this->getSM()->getItems();
	// have to have a copy because we will modify the
	// list in the loop
	dllistAssign(&copy, selected);

	for(r = dllistFirst(&copy); r; dllistNext(&copy, &r)){
		iEntityPtr entity;
		IE *ie;
		entity = (iEntityPtr) dllistContents(&copy, r);
		ie = (IE *) eaGetFace(entity, vtIMAGE);
		if(ie == 0)
			this->getSM()->toggleItem(entity);
	}
	
	dllistDtor(&copy);
	// bogus do I have to do this
	this->main_window->refresh_dialogs();
}

iBox
ImageMode::getBounds(iEntityPtr entity)
{
	iBox bounds;
	
	// -- check that the page matches
	ImagePage *current_page;
	ImagePage *entitys_page;
	IE *ie;
	
	bounds.w = 0;
	ie = (IE *) eaGetFace(entity, vtIMAGE);
	if(ie == 0)
		return bounds;
	entitys_page = (ImagePage *) ie->page;
	current_page = (ImagePage *) this->getCurrentPage();
	
	if(entitys_page == current_page) {
		bounds = i_GetEnclosingBox(entity);
		int mag_exp = current_page->getMagExp();
		
		// account for mag
		bbScaleUp(bounds, mag_exp);
	} else {
		bounds.w = 0;
	}
	
	
	return bounds;
}
