
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */


#ifndef _ImagePage_h_
#define _ImagePage_h_


class IE;

#include "Page.h"
#include "QuadTree.h"
#include "ucs.h"

// this turns on the ability to create new entities with middle
// button
// simplify things tremendously by killing other modes when
// you might invalidate them?
#define NEW_ENTITY

enum ThingGrabbed {
	tgNOTHING,
	tgAll,
	tgUL,
	tgLR,
	tgUR,
	tgLL
};

typedef enum ThingGrabbed ThingGrabbed;

class ImagePage : public Page {
public:
	QuadTree tree;
	iImagePtr image;
	DLList select_list;
	int go;
	Position first_x;
	Position first_y;
	Position last_x;
	Position last_y;
	
	iBox xor_box;
	ThingGrabbed thing;
	
	iBox central_position;
	
	int invalid_guy_added;
	
	void setupNewXor();
	void setupMoveXor();
	void drawXOR(Pixmap drawable,int x_off, int y_off);
	void eraseXOR();
	
	ImagePage(iImagePtr image, Mode *parent, const aString& name);
	virtual ~ImagePage();
	
	void addMember(iEntityPtr member);
	
	// this differs from addMember in that it will search
	// for the right place in the dllist to put member
	// whereas addMember will just insert it at the end
	void addNewMember(iEntityPtr member);
	
	void deleteMember(iEntityPtr member);
	void drawBoxes(Display *display, Drawable drawable, 
				   int x_off, int y_off,
				   DLList *list, int selected);
	
	void drawKnobs(Display *display, Drawable drawable, 
				   int x_off, int y_off,
				   iEntityPtr entity, int selected);
	
	void redrawImages(Display *display, Drawable drawable, 
				      int x_off, int y_off,
				      DLList *list);
	
	void drawText(Display *display, Drawable drawable, 
				  int x_off, int y_off,
				  DLList *list, int selected);

	void drawPitch(Display *display, Drawable drawable, 
				   int x_off, int y_off,
				   DLList *list);
	Ref prevRef(iEntityPtr current);
	
	void deleteEntity(iEntityPtr entity);
	iBox entityBounds(iEntityPtr entity);
	
	// check entity to see it (still?) qualifies -- called after
	// entity is modified
	void checkEntity(iEntityPtr entity);

	void rotateImage();
	
	void refresh(IE *face);
	void refreshBorder(iEntityPtr entity);

	int pressKey(iUCode key_value);
	void reassignValues(const DLList *list, iUCode key_value);
	void begin_edit(iEntityPtr entity);
	int editing(iEntityPtr entity);
	void end_edit(iEntityPtr entity);
	void end_all_edits();

	
	
#ifdef NEW_ENTITY   
	void createChild(iEntityPtr entity);
#endif
	void findEnclosingEntities(int x, int y, DLList *);
	iEntityPtr findTheEntity(DLList *entity_list,int x, int y);
	 
	
	// these are from page
	
	// set mag, resize window
	/*virtual*/ void centerOnSelected(iEntityPtr entity);
	/*virtual*/ void scrollToUpperLeft(iEntityPtr entity);
	/*virtual*/ void scrollToTop(iEntityPtr entity);
	/*virtual*/ void scrollToBottom(iEntityPtr entity);
	/*virtual*/ void scrollToLeft(iEntityPtr entity);
	/*virtual*/ void scrollToRight(iEntityPtr entity);

	/*virtual*/ void _setMag(int mag_exp, int old_mag);   
	
	// save pos, call _setMag, scroll back to pos
	/*virtual*/ void setMag(int mag_exp, int old_mag);
	
	/*virtual*/ void         pressButton(XEvent *event);
	/*virtual*/ void         releaseButton(XEvent *event);
	/*virtual*/ void         moveButton(int state, int x, int y);
	
	/*virtual*/ void manage();
	/*virtual*/ void unmanage();
	
	/*virtual*/ iEntityPtr findEnclosingEntity(int x, int y) {return 0;};
	/*virtual*/ void conditionallyInitialiseCanvas();
	
	/*virtual*/ void redisplay(XmDrawingAreaCallbackStruct *call_data);
	/*virtualvoid resize(XmDrawingAreaCallbackStruct *call_data);*/ 
	
	/*virtual*/ void drawRegionNow(iLBox refresh_box);
	/*virtual*/ int resizeNecessaryP() { return FALSE; }

	/*virtual void setPicture(iBox b);*/
	/*virtual Box getPicture();*/
	
	
	/*virtual*/ void prevEntity();
	/*virtual*/ void nextEntity();
	/*virtual*/ void findNextFlagEntity();
	/*virtual*/ void nextEntity(EntityCheckFunc checker);
	/*virtual*/ void prevEntity(EntityCheckFunc checker);
	/*virtual*/ void firstEntity(EntityCheckFunc checker);
	
	// bogus should have last entity in the virtual function
	// requirements
	void lastEntity(EntityCheckFunc checker);


	// set up the quad tree
	void initialise();
};

iBox
boundsOnTheOriginal(ImagePage *page, iEntityPtr entity);

#define LEFTPAD 10
#define TOPPAD  10



#endif
