
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"

#include <X11/Intrinsic.h>
#include <Xm/Frame.h>
#include <Xm/ScrolledW.h>
#include <Xm/MainW.h>

#include "illum.h"
#include "dafs.h"
#include "dafstype.h"
#include "misc/XArg.h"
#include "svpmap.h"

#include "MainWindow.h"
#include "LTEPage.h"
#include "ViewCanvas.h"

#include "LTEMode.h"



LTEMode::LTEMode(MainWindow *input_mw, iEntityPtr input_entity)
	: Mode(input_mw)
{
	XArg args;

	this->old_frame = this->main_window->current_frame;
	Widget rc = XtParent(this->old_frame);
	XtUnmanageChild(this->old_frame);
	

	args.reset();
	args.setArg(XmNmarginWidth, 2); 
	args.setArg(XmNmarginHeight, 2); 
	args.setArg(XmNshadowThickness, 1); 
	args.setArg(XmNshadowType, XmSHADOW_OUT); 
	this->frame = XmCreateFrame (rc, "lte_frame", 
								 args.getArgs(), args.getSize());

	
	args.reset();
	args.setArg(XmNscrollBarDisplayPolicy, XmSTATIC); 
	args.setArg(XmNscrollingPolicy, XmAUTOMATIC); 
	this->sw = XmCreateScrolledWindow (this->frame, "lte_sw", 
									   args.getArgs(), args.getSize());
		
	this->createPages(input_entity)	;
	XtManageChild(this->sw);
	XtManageChild(this->frame);
}

LTEMode::~LTEMode()
{

  XtManageChild(this->old_frame);
  this->main_window->current_frame = this->old_frame;
  XtDestroyWidget(this->frame);
}

void
LTEMode::createPages(iEntityPtr entity)
{	
  this->the_page = new LTEPage(this, "", entity);
  this->addPage(this->the_page);
}

void
LTEMode::configure()
{
  if(the_page)
	the_page->configure();
}

static char bogus_name[] = "Text";

aString
LTEMode::getName()
{
	return bogus_name;
}

ModeType
LTEMode::getMode()
{
	return vtLTE;
}

void
LTEMode::addNewEntity(iEntityPtr entity)
{
  // we will ignore this?
  this->the_page->addNewEntity(entity);
}


void
LTEMode::centerOnSelected(iEntityPtr entity)
{
  // hmm I don't know
	LTEPage *page;
	page = this->the_page;
	page->centerOnSelected(entity);
}

void
LTEMode::processDelete(void)
{
  this->the_page->processDelete();
}
