
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"

#include "smemory.h"

#include <X11/Xutil.h>

#include <Xm/DialogS.h>
#include <Xm/ToggleBG.h>
#include <Xm/PushBG.h>
#include <Xm/SeparatoG.h>
#include <Xm/RowColumn.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/FileSB.h>
#include <Xm/SelectioB.h>
#include <Xm/SelectioBP.h>

#include "smemory.h"

#ifdef RAF_INTERNAL
#include "form.h" 
#include "formint.h"
#endif

#include "rafutil.h"

#include "misc/XArg.h"
#include "misc/color.h"
#include "misc/misc.h"
#include "misc/prompt.h"

#include "keyboard/KeyBoard.h"

#include "misc/bb.h"
#include "mag.h"

#include "MainWindow.h"
#include "MainWindow.cb.h"
#include "ViewCanvas.h"
#include "ListDialog.h"
#include "SelectionManager.h"
#include "Application.h"
#include "RTEMode.h"


#include "Document.h"


#include "ImageMode.h"
#include "OOCMode.h"
#include "ImagePage.h"
#include "Page.h"
#include "scroll.h"
#include "EntityAux.h"


#include "ChooseDialog.h"
#include "EntityDialog.h"

#include "TypePrefDialog.h"
#include "UniEdit.h"
#include "NihonEdit.h"
#include "FileDialog.h"
#include "config.h"



int PromptSave(Widget w, MainWindow *mw)
{
	int new_file;

	if(mw->document->dirtyP()) {
		int answer;
		aString file_name = mw->document->file_name.filename();
		aString prompt;
		if(file_name == ""){
			file_name = "New document";
			new_file = TRUE;
		} else
			new_file = FALSE;

		prompt = file_name + " not saved. Write first?";
		answer = uiAskUser3(prompt.string(),
						  "Yes", "No", "Cancel", 1);
		switch(answer) {
		case 3:
			return TRUE;
		case 1:
			if(new_file) {
				SaveAsCB(w, mw, 0);
            } else
				if(mw->document->saveFile() == 0)
					return TRUE;
			break;
		}
	}
	return FALSE;
}
int PromptRevert(Widget w, MainWindow *mw)
{
	int new_file;

	if(mw->document->dirtyP()) {
		int answer;
		aString file_name = mw->document->file_name.filename();
		aString prompt;
		if(file_name == ""){
			file_name = "New document";
			new_file = TRUE;
		} else
			new_file = FALSE;

		prompt = "Lose changes to " + file_name +"?";
		answer = uiAskUser2(prompt.string(),
						  "Yes", "Cancel", 1);
		switch(answer) {
		case 2:
			return TRUE;
		case 1:
		default:
			break;
		}
	}
	return FALSE;
}

/*
// Bogus - I want this to jump to the first flagged in the
// ImagePage and I don't know how
*/
void
ClearFlagCB(Widget w,
			MainWindow *mw,
			XtPointer call_data)
{
	// bogosity warning: I called this routine from OOCPage
	// with w=0 and call_data=0, so don't use these!!!!!
	Ref r;
	iEntityPtr entity;
	const DLList *list;
	list = mw->sm->getItems();
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	
	if (sole_entity && entityFlaggedP(sole_entity)) {
		// have to do it backwards, since we might be removing entity from mode,
		// and therefore nextEntity won't do anything!
		mw->nextEntity(entityFlaggedP);
		iEntityPtr entity = mw->sm->getSoleItem();
		if (entity==sole_entity) {
			mw->prevEntity(entityFlaggedP);
		}
		if (sole_entity && sole_entity!=mw->document->root_entity) 
			entityClearFlag(mw,sole_entity);
		
	} else {
		int count=0;
		for(r = dllistFirst(list); r; dllistNext(list, &r)){
			entity = (iEntityPtr) dllistContents(list, r);
			if (entityFlaggedP(entity)) {
				// bogus why not the root guy
				// is this a condition that made sense for delete but not here?
				if (entity && entity!=mw->document->root_entity) entityClearFlag(mw,entity);
				count++;
			}
		}
		if (!count)
			mw->firstEntity(entityFlaggedP);
	}
	Mode *mode=mw->document->getCurrentMode();
	if (mode) {
		Page *page=mode->getCurrentPage();
		if (page && page->should_delete) {
			int center=TRUE;
			if (page->managed)
				if (!mode->forward())
					if(!mode->back()) {
						center=FALSE;
						mw->view_canvas->unmanage();
					}
			if (center)
				mw->centerOnSelected();
			//delete page;
		}
	}
	mw->setSensitive();
}

#ifdef RAF_INTERNAL
void
ReinitializeCB(Widget w,
			   MainWindow *mw,
			   XtPointer call_data)
{
	initializeRecognition();
}
#endif

void 
NextEntityCB(Widget    w,
			 MainWindow *window,
			 XtPointer    call_data)
{
	window->nextEntity();
}

void PrevEntityCB(Widget    w,
				  MainWindow *window,
				  XtPointer    call_data)
{
	window->prevEntity();
}

void 
UpEntityCB(Widget    w,
		   MainWindow *window,
		   XtPointer    call_data)
{
	window->upEntity();
}

void 
DownEntityCB(Widget    w,
			 MainWindow *window,
			 XtPointer    call_data)
{
	window->downEntity();
}

void 
FindNextFlagCB(Widget    w,
			 MainWindow *window,
			 XtPointer    call_data)
{
	window->findNextFlagEntity();
}

void AmericanCB(Widget    w,
				MainWindow *window,
				XtPointer    call_data)
{
	window->kb->setKeyMap(window->american);
	//window->kb->show();
}

void ArabicCB(Widget    w,
			  MainWindow *window,
			  XtPointer    call_data)
{
	window->kb->setKeyMap(window->arabic);
	//window->kb->show();
}

void RussianCB(Widget    w,
			   MainWindow *window,
			   XtPointer    call_data)
{
	window->kb->setKeyMap(window->russian);
	//window->kb->show();
}

void EnglishCB(Widget    w,
			   MainWindow *window,
			   XtPointer    call_data)
{
	window->kb->setKeyMap(window->english);
	//window->kb->show();
}

void GermanCB(Widget    w,
			  MainWindow *window,
			  XtPointer    call_data)
{
	window->kb->setKeyMap(window->german);
	//window->kb->show();
}
void SwedishCB(Widget    w,
			  MainWindow *window,
			  XtPointer    call_data)
{
	window->kb->setKeyMap(window->swedish);
	//window->kb->show();
}
void FrenchCB(Widget    w,
			  MainWindow *window,
			  XtPointer    call_data)
{
	window->kb->setKeyMap(window->french);
	//window->kb->show();
}
void GreekCB(Widget    w,
			 MainWindow *window,
			 XtPointer    call_data)
{
	window->kb->setKeyMap(window->greek);
	//window->kb->show();
}
void MicrCB(Widget    w,
			 MainWindow *window,
			 XtPointer    call_data)
{
	window->kb->setKeyMap(window->micr);
	//window->kb->show();
}

void ItalianCB(Widget    w,
			 MainWindow *window,
			 XtPointer    call_data)
{
	window->kb->setKeyMap(window->italian);
	//window->kb->show();
}

void NorwCB(Widget    w,
			 MainWindow *window,
			 XtPointer    call_data)
{
	window->kb->setKeyMap(window->norw);
	//window->kb->show();
}

void SpanishCB(Widget    w,
			 MainWindow *window,
			 XtPointer    call_data)
{
	window->kb->setKeyMap(window->spanish);
	//window->kb->show();
}

void MagCB(Widget w,
		   Tuple *data,
		   XtPointer call_data)
{
	MainWindow *mw = (MainWindow *) data->first;
	int mag = (int) data->second;
	mw->setMag(mag);
}

void changeTypeCB(Widget w,
		   Tuple *data,
		   XtPointer call_data)
{
	int type=(int)data->second;
	Ref r;
	MainWindow *mw = (MainWindow *) data->first;
	iEntityPtr entity;
	
	const DLList *selection=mw->sm->getItems();
	for(r = dllistFirst(selection); r; dllistNext(selection, &r)){
		entity = (iEntityPtr) dllistContents(selection, r);
		i_SetEntityType(entity,type);
	}
	mw->document->setDirty();
}

/*-------------------------------------------------------------
**  QuitCB      - callback for quit button
*/
void QuitCB (Widget    w,
			 MainWindow *mw,
			 XtPointer    call_data)
{
	extern MemoryInfo mi;
	/*  Terminate the application. */
	
	int cancel=PromptSave(w, mw);
	if(cancel) return;
	
	// could be more efficient by not insisting on
	// closing -- just saving
	writeConfig();
	disposeColorTable();
	if(mi.memory_track_flag != sNoChecks){
		mw->document->closeFile();
		delete mw->app;
		mw->app = 0;
		delete mw;
	}
	
	i_ExitLib();
#ifdef RAF_INTERNAL
	f_ExitLib();
#endif
	miscExit();


	if(mi.memory_track_flag != sNoChecks)
		s_ExitLib();
	
	exit (0);
}



int selCB(MainWindow *mw, char *name)
{
    if (!*name)
		return FALSE;
	
    mw->document->selectPage(name);
    mw->setSensitive();
    mw->standardTitle();
	
    return TRUE;
}

void SelectCB (Widget    w,
			   MainWindow *mw,
			   void *call)
{
	Mode *current_mode = mw->document->getCurrentMode();
	
	XmString title=XmStringCreateLocalized("Select Page");
	XtAppContext xapcntxt = 0;
	if (!mw->page_select) {
		MainWindow *pmw = mw;
		Widget wgt = mw->app->app_shell;
		selCBProc pfnSel = selCB;
		void *pvData = (void *) mw;
		mw->page_select=new ListDialog(pmw,
									   wgt,
									   xapcntxt,
									   title, 
									   pfnSel,
									   pvData);
	}
	
	int list_size,i;
	XmStringTable list;
	current_mode->makePageList(&list,&list_size);
	
	mw->page_select->show(list,list_size);
	
	for(i = 0; i < list_size; i++)
		XmStringFree(list[i]);
	smfree(list);
	
	XmStringFree(title);
	
	
}


void ForwardCB (Widget    w,
				MainWindow *mw,
				void *call)
{
	Mode *current_mode = mw->document->getCurrentMode();
	
	if(current_mode) {
		current_mode->forward();
		mw->setSensitive();
		mw->standardTitle();
	}
}


void BackCB (Widget    w,
			 MainWindow *mw,
			 void *call)
{
	Mode *current_mode = mw->document->getCurrentMode();
	
	if(current_mode) {
		current_mode->back();
		mw->setSensitive();
		mw->standardTitle();
	}
}


static void openFile(Widget w, MainWindow *mw,FileName fn)
{
	int success;
		
	int cancel=PromptSave(w, mw);
	if (cancel) return;
	// close the file inside open only if successful
	
	success = mw->document->openFile(fn, mw->document->current_mode);
	// let the message be handled in openFile where he can be more
	// specific
}


void NextCB (Widget    w,
			 MainWindow *mw,
			 void *call)
{
	FileName fn;

	fn = mw->document->file_name;
	if (fn.nextFile()) {
		// do an error message
		uiWarnUser("No next file." , "OK");
		return;
	}
	openFile(w,mw,fn);
}

void PrevCB (Widget    w,
			 MainWindow *mw,
			 void *call)
{
	FileName fn;
	fn = mw->document->file_name;
	if (fn.prevFile()) {
		uiWarnUser("No previous file." , "OK");
		// do an error message
		return;
	}
	openFile(w,mw,fn);
}


char *about_illum_text[]={
	"           Illuminator 0.10b1", // dont forget to change -h
	"",
	"         Produced for ARPA by",
	"            RAF Technology",
	"      16650 NE 79th St Suite 200",
	"           Redmond, WA 98053",
	"            (206) 867-0700",
	"",
	"Report bugs to illbugs@raf.com",
	"Report other comments to illcomm@raf.com",
	0
};

  
void OpenCB (Widget    w,
  			 MainWindow *mw,
  			 void *call)
{
 	aString directory = mw->document->current_directory;
 	
 	int cancel=PromptSave(w, mw);
 	if (cancel) return;
 	
 	if(mw->open_dialog == 0)
		mw->open_dialog = new OpenDialog(mw, mw->app->app_shell, directory);
    else 
		mw->open_dialog->setDir(directory);

	// bogus should be tricky
 	mw->would_be_mode = mw->document->current_mode;

 	mw->open_dialog->show();
}
void RevertCB(Widget    w,
  			 MainWindow *mw,
  			 void *call)
{	FileName fn; 
	ModeType mode;
 	int cancel=PromptRevert(w, mw);
 	if (cancel) return;
 	int4 hsb_value, vsb_value;
 	
 	fn = mw->document->file_name;
 	mode = mw->document->current_mode;
 	getScrollPositionCenter(mw->scroll_window, &hsb_value, &vsb_value);
 	mw->document->closeFile();
 	mw->document->openFile(fn, mode);
 	mw->view_canvas->scrollToPosition(mw->scroll_window, hsb_value, vsb_value);
}
void AboutCB (Widget    w,
			  MainWindow *mw,
			  void *call)
{
	uiTellUserN(about_illum_text,"  OK  ");
}




void NewCB (Widget    w,
			MainWindow *mw,
			void *call)
{
	int cancel=PromptSave(w, mw);
	if (cancel) return;
	mw->document->newFile();
	// bogus should be tricky
	mw->would_be_mode = vtRTE;
}


void SaveCB (Widget    w,
			 MainWindow *mw,
			 void *call)
{
	if(mw->document->file_name == "")
		SaveAsCB(w, mw, call);
	else
		mw->document->saveFile();
}

void rotateCB(Widget w,
			  MainWindow *mw,
			  void *call)
{
	Mode *mode = mw->document->getCurrentMode();
	
	if (mode->getMode() != vtIMAGE) return;
	ImagePage *page = (ImagePage*)mode->getCurrentPage();
	if (!page)return;
	page->rotateImage();
}

void centerOnSelectedCB (Widget    w,
						 MainWindow *mw,
						 void *call)
{
	mw->centerOnSelected();
}

void deleteCB (Widget    w,
			   MainWindow *mw,
			   void *call)
{
//   mw->deleteSelection();
	Document *doc=mw->document;
	if (!doc) 
		return;
	Mode *mode=doc->getCurrentMode();
	if (!mode) 
		return;

	if(doc->current_mode == vtRTE) {
	  RTEMode *rte_mode = (RTEMode *) mode;
	  rte_mode->processDelete();
	  return;
	}
	Page *page=mode->getCurrentPage();
	if (page) {
		page->pressKey('\b');
		if (page->should_delete) {
			int center=TRUE;
			if (page->managed)
				if (!mode->forward())
					if(!mode->back()) {
						center=FALSE;
						mw->view_canvas->unmanage();
					}
			if (center)
				mw->centerOnSelected();
			mw->standardTitle();
			mw->refresh_dialogs();
			//delete page;
		}
		mw->setSensitive();
	}
}

void editKillWordCB (Widget    w,
					 MainWindow *mw,
					 void *call)
{
//   mw->deleteSelection();
	Document *doc=mw->document;
	if (!doc) 
		return;
	Mode *mode=doc->getCurrentMode();
	if (!mode) 
		return;
	Page *page=mode->getCurrentPage();
	// 20 is apparently DC4 or ctrl-u
	if (page) 
		page->pressKey(20);
}

void
ModeCB(Widget w,
       Tuple *data,
       void *call)
{
	MainWindow *mw  = (MainWindow *) data->first;
	ModeType vt = (ModeType) data->second;
	
	mw->document->changeMode(vt);
	mw->setSensitive();
	mw->standardTitle();
	mw->refresh_dialogs();
}

void ConfigureEH (Widget   shell,
				  MainWindow *mw,
				  XConfigureEvent *event,
				  char *foo)
{	
	Position x, y;

	if (event->type != ConfigureNotify)
		return;
	
	if(event->width != mw->last_width || event->height != mw->last_height){
		mw->last_width = event->width;
		mw->last_height = event->height;
		mw->document->configure();
		mw->standardTitle();
	}
	XtVaGetValues(mw->app->app_shell, XmNx, &x, XmNy, &y, NULL);
	if (x!= mw->last_x  ||  y!= mw->last_y){
		mw->last_x=x;
		mw->last_y=y;
		// this hack not ready for primetime because it can refer
		// to a deleted entity
	}
	
}


#ifdef RAF_INTERNAL
void UnknownTypeCB(Widget    w,
				   MainWindow *window,
				   XtPointer    call_data)
{
	window->document->setMailType(0);
	//window->document->saveFile();
	//NextCB(w, window, call_data);
}

void HandTypeCB(Widget    w,
				MainWindow *window,
				XtPointer    call_data)
{
	window->document->setMailType(HAND_TYPE);
	//window->document->saveFile();
	//NextCB(w, window, call_data);
}


void VariableTypeCB(Widget    w,
					MainWindow *window,
					XtPointer    call_data)
{
	window->document->setMailType(VARIABLE_TYPE);
	//window->document->saveFile();
	//NextCB(w, window, call_data);
}

void FixedTypeCB(Widget    w,
				 MainWindow *window,
				 XtPointer    call_data)
{
	window->document->setMailType(FIXED_TYPE);
	//window->document->saveFile();
	//NextCB(w, window, call_data);
}
#endif


#ifdef RAF_INTERNAL
void LeftPhaseCB (Widget    w,
				  MainWindow *mw,
				  void *call)
{
	iEntityPtr entity = mw->sm->getSoleItem();
	iBox box;
	pandp_t *pp;
	void *vp;
	
	i_GetUserDataProp(entity, "pandp", &vp,0,f_SwapPandP);
	pp = (pandp_t *) vp;
	
	if(pp && entity && 
	   !bbNullP((box = i_GetEnclosingBox(entity)))) {   
		// bogus could be trickier
		if(pp->phase)
			pp->phase--;
		else
			pp->phase = (pp->pitch>>16) - 1;
		
		
		i_SetUserDataProp(entity,
						  "pandp",
						  pp,
						  sizeof(pandp_t));
		mw->document->setDirty();
	}
}


void RightPhaseCB (Widget    w,
				   MainWindow *mw,
				   void *call)
{
	iEntityPtr entity = mw->sm->getSoleItem();
	iBox box;
	pandp_t *pp;
	void *vp;
	
	i_GetUserDataProp(entity, "pandp", &vp,0,f_SwapPandP);
	pp = (pandp_t *) vp;
	
	if(pp && entity && 
	   !bbNullP((box = i_GetEnclosingBox(entity)))) {   
		// bogus could be trickier
		if(pp->phase >= (pp->pitch >> 16) - 1)
			pp->phase = 0;
		else
			pp->phase++;
		
		i_SetUserDataProp(entity,
						  "pandp",
						  pp,
						  sizeof(pandp_t));
		mw->document->setDirty();
	}
}


void DecrDoubleCB (Widget    w,
				   MainWindow *mw,
				   void *call)
{
	DecrPitch(mw,2<<16);
}

void DecrPitchCB (Widget    w,
				  MainWindow *mw,
				  void *call)
{
	DecrPitch(mw,1<<16);
}

void DecrPitch(MainWindow *mw,
			   int incr)
{
	iEntityPtr entity = mw->sm->getSoleItem();
	iBox box;
	pandp_t *pp;
	void *vp;
	iDAFSError err;
	
	err = i_GetUserDataProp(entity, "pandp", &vp,0,f_SwapPandP);
	pp = (pandp_t *) vp;
	
	if(entity && 
	   !bbNullP((box = i_GetEnclosingBox(entity))) && (pp) != 0) {   
		// bogus could be trickier
		int num_cells;
		num_cells = (box.w << 16) / pp->pitch;
		pp->pitch -= incr/num_cells;
		
		i_SetUserDataProp(entity,
						  "pandp",
						  pp,
						  sizeof(pandp_t));
		mw->document->setDirty();
	}
	
}
void IncrPitchCB (Widget    w,
				  MainWindow *mw,
				  void *call)
{
	IncrPitch(mw,1<<16);
}

void IncrDoubleCB (Widget    w,
				   MainWindow *mw,
				   void *call)
{
	IncrPitch(mw,2<<16);
}

void IncrPitch(MainWindow *mw,
			   int incr)
{
	iEntityPtr entity = mw->sm->getSoleItem();
	iBox box;
	pandp_t *pp;
	iDAFSError err;
	void *vp;
	
	err = i_GetUserDataProp(entity, "pandp",&vp,0,f_SwapPandP);
	
	pp = (pandp_t *) vp;
	
	if(entity && 
	   !bbNullP((box = i_GetEnclosingBox(entity))) && (pp) != 0) {   
		// bogus could be trickier
		int num_cells;
		num_cells = (box.w << 16) / pp->pitch;
		pp->pitch += incr/num_cells;
		i_SetUserDataProp(entity,
						  "pandp",
						  pp,
						  sizeof(pandp_t));
		mw->document->setDirty();
	}
	
}


#endif


void KeyboardCB(Widget    w,
				MainWindow *window,
				XtPointer    call_data)
{
	window->kb->show();
}


void ChooseCB(Widget    w,
			  MainWindow *mw,
			  XtPointer    call_data)
{
	mw->chooser->show();
}

void ColorsCB(Widget    w,
			  MainWindow *mw,
			  XtPointer    call_data)
{
	mw->type_pref_view->show();
}

#ifdef DO_POPUPS
void FlagCB(Widget    w,
			  MainWindow *mw,
			  XtPointer    call_data)
{
#ifndef KILL_TEXT_MODE
	TextMode *mode = (TextMode *) mw->document->modes[vtTEXT];
	TextPage *page = mode ? mode->the_page : 0;
	if (page)
		page->togglePopup();
#endif
}
#endif

void TextCB(Widget    w,
			MainWindow *mw,
			XtPointer    call_data)
{
	mw->textToggle();
}

void viewCB(Widget w,
			MainWindow *mw,
			XtPointer call_data)
{
	mw->entity_view->show();
	mw->updateMode(0);
}
void pageDownCB(Widget w,
			MainWindow *mw,
			XtPointer call_data)
{
	XArg args;
	int max,ss;
	Dimension h;
	Widget drawing_a,vsb,hsb;
	Mode *mode = (Mode *) mw->document->getCurrentMode();
	
	args.reset();
	args.setArg(XmNworkWindow,&drawing_a);
	args.setArg(XmNverticalScrollBar,&vsb);
	args.setArg(XmNhorizontalScrollBar,&hsb);
	args.getValues(mode->scrolled_window);
	args.reset();
	args.setArg(XmNmaximum,&max);
	args.setArg(XmNsliderSize,&ss);
	args.getValues(vsb);
	args.reset();
	args.setArg(XmNheight,&h);
	args.getValues(drawing_a);

	mw->view_canvas->pix_voffset += h/2;
	mw->view_canvas->pix_voffset = MIN(mw->view_canvas->pix_voffset,(max-ss));
	args.reset();
	args.setArg(XmNvalue,mw->view_canvas->pix_voffset);
	args.setValues(vsb);
	mw->redrawMode();
	//XmScrolledWindowSetAreas(mode->scrolled_window, hsb, vsb, 0);
}
void pageUpCB(Widget w,
			MainWindow *mw,
			XtPointer call_data)
{
	XArg args;
	int min;
	Dimension h,value;

	Widget drawing_a,vsb,hsb;
	Mode *mode = (Mode *) mw->document->getCurrentMode();
	
	args.reset();
	args.setArg(XmNworkWindow,&drawing_a);
	args.setArg(XmNverticalScrollBar,&vsb);
	args.setArg(XmNhorizontalScrollBar,&hsb);
	args.getValues(mode->scrolled_window);
	args.reset();
	args.setArg(XmNminimum,&min);
	args.getValues(vsb);
	args.reset();
	args.setArg(XmNheight,&h);
	args.getValues(drawing_a);

	value = MAX(mw->view_canvas->pix_voffset - h/2,min);
	mw->view_canvas->pix_voffset = value;
	args.reset();
	args.setArg(XmNvalue,mw->view_canvas->pix_voffset);
	args.setValues(vsb);
	mw->redrawMode();
	//XmScrolledWindowSetAreas(mode->scrolled_window, hsb, vsb, 0);
}
void pageRightCB(Widget w,
			MainWindow *mw,
			XtPointer call_data)
{
	XArg args;
	int max,ss;
	Dimension wi;
	Widget drawing_a,vsb,hsb;
	Mode *mode = (Mode *) mw->document->getCurrentMode();
	
	args.reset();
	args.setArg(XmNworkWindow,&drawing_a);
	args.setArg(XmNverticalScrollBar,&vsb);
	args.setArg(XmNhorizontalScrollBar,&hsb);
	args.getValues(mode->scrolled_window);
	args.reset();
	args.setArg(XmNmaximum,&max);
	args.setArg(XmNsliderSize,&ss);
	args.getValues(hsb);
	args.reset();
	args.setArg(XmNwidth,&wi);
	args.getValues(drawing_a);

	mw->view_canvas->pix_hoffset += wi/2;
	mw->view_canvas->pix_hoffset = MIN(mw->view_canvas->pix_hoffset,(max-ss));
	args.reset();
	args.setArg(XmNvalue,mw->view_canvas->pix_hoffset);
	args.setValues(hsb);
	mw->redrawMode();
	//XmScrolledWindowSetAreas(mode->scrolled_window, hsb, vsb, 0);
}
void pageLeftCB(Widget w,
			MainWindow *mw,
			XtPointer call_data)
{
	XArg args;
	int min;
	Dimension wi,value;

	Widget drawing_a,vsb,hsb;
	Mode *mode = (Mode *) mw->document->getCurrentMode();
	
	args.reset();
	args.setArg(XmNworkWindow,&drawing_a);
	args.setArg(XmNverticalScrollBar,&vsb);
	args.setArg(XmNhorizontalScrollBar,&hsb);
	args.getValues(mode->scrolled_window);
	args.reset();
	args.setArg(XmNminimum,&min);
	args.getValues(hsb);
	args.reset();
	args.setArg(XmNwidth,&wi);
	args.getValues(drawing_a);

	value = MAX(mw->view_canvas->pix_hoffset - wi/2,min);
	mw->view_canvas->pix_hoffset = value;
	args.reset();
	args.setArg(XmNvalue,mw->view_canvas->pix_hoffset);
	args.setValues(hsb);
	mw->redrawMode();
	//XmScrolledWindowSetAreas(mode->scrolled_window, hsb, vsb, 0);
}

#ifdef RAF_INTERNAL

static void
addDudes(Document *doc, iEntityPtr entity)
{
	iEntityPtr child;
	
	doc->addNewEntity(entity, FALSE);
	for(child = i_GetChildEntity(entity); child; child =i_GetNextEntity(child))
   		addDudes(doc, child);
}

extern fCharSetPtr cp;

void recognizeCB(Widget w,
				 MainWindow *mw,
				 XtPointer call_data)
{
	iEntityPtr entity = mw->sm->getSoleItem();
	iEntityPtr child;
	
#if 0
		dllistCtor(l);
		
		for (child=i_GetChildEntity(entity); child; child=i_GetNextEntity(child)) {
			dllistAppend(l, child);
		}
		
		for(r = dllistFirst(l); r; dllistNext(l, &r)){
			child = (iEntityPtr) dllistContents(l, r);
			i_DisposeEntity(child);
		}
		dllistDtor(l);
#endif
	if(i_GetEntityType(entity) == iDocumentT)
		i_SetEntityType(entity,iParagraphT);
	// why disable at exactly this point?
	disableCallbacks();
	
	if (cp) {
		if (i_SetSessionUserPtrProp(i_GetEntityType(entity),"CharSet",
									cp, NULL))
			printf("Failed to set character set in entity\n");
	}

	// bogus looks like lines aren't registered
	// bogus, entity might change!!!
	f_Recognize(&entity);
	
	enableCallbacks();
	
#ifdef STEVE
	int conf=i_GetConf(entity);
	printf("confidence of recognition=%i\n",conf);
#endif
	
	//disableCallbacks();
	for(child = i_GetChildEntity(entity); child; child =i_GetNextEntity(child))
   		addDudes(mw->document, child);
	//enableCallbacks();
	
	i_DoCallBack(mRecognize, entity);
}

void deleteChildrenCB(Widget w,
					  MainWindow *mw,
					  XtPointer call_data)
{
	iEntityPtr entity = mw->sm->getSoleItem();
	iEntityPtr child;
	
	DLList temp;
	DLList *l = &temp;
	Ref r;
	
	dllistCtor(l);
	
	for (child=i_GetChildEntity(entity); child; child=i_GetNextEntity(child)) {
		dllistAppend(l, child);
	}
	
	for(r = dllistFirst(l); r; dllistNext(l, &r)){
		child = (iEntityPtr) dllistContents(l, r);
		i_DisposeEntity(child);
	}
	dllistDtor(l);
}

#endif

#ifdef RAF_INTERNAL

int raf_display=0;

void steves_specialCB(Widget w,
					  MainWindow *mw,
					  XtPointer call_data)
{
	raf_display=(raf_display+1)%4;
	printf("debug value=%i\n",raf_display);
	
// not all that special yet...
}
#endif

static int common_parent(MainWindow *mw, char *warning)
{
	iEntityPtr par,entity;
	int good;
	Ref r;
	const DLList *selection=mw->sm->getItems();
	
	par=0; good=TRUE;
	for(r = dllistFirst(selection); r && good; dllistNext(selection, &r)){
		entity = (iEntityPtr) dllistContents(selection, r);
		if (entity!=mw->document->root_entity) {
			if (!par)
				par=i_GetParentEntity(entity);
			else
				if (par!=i_GetParentEntity(entity)) {
					good=FALSE; /* choke */
					uiWarnUser(warning, "OK");
				}
		} else {
			good=FALSE;
			uiWarnUser("You can't delete the root entity.", "OK");
		}
	}
	return good;
}

static int not_root(MainWindow *mw, char *warning)
{
	iEntityPtr entity;
	int good;
	Ref r;
	const DLList *selection=mw->sm->getItems();
	
	good=TRUE;
	for(r = dllistFirst(selection); r && good; dllistNext(selection, &r)){
		entity = (iEntityPtr) dllistContents(selection, r);
		if (entity==mw->document->root_entity) {
			good=FALSE;
			uiWarnUser(warning, "OK");
		}
	}
	return good;
}


#ifdef RAF_INTERNAL
void deskewCB(Widget w,
			  MainWindow *mw,
			  XtPointer call_data)
{
	Mode *current_mode = mw->document->getCurrentMode();
	Page *current_page = current_mode->getCurrentPage();
	
	if(current_page)
		f_Deskew(mw->document->root_entity,5000,HORZ_DESKEW | VERT_DESKEW);
}

#endif


#ifdef RAF_INTERNAL
void 
pitchGrowHorzCB (Widget    w,
				 MainWindow *mw,
				 void *call)
{
	iDAFSError e;
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	iBox *pbox;
	int image_width;
	iImagePtr image;
	if(sole_entity){
		if (image = i_GetImage(sole_entity)) {
			e = i_GetUserDataProp(sole_entity, "PBB", (void**)&pbox, 0,i_SwapUBox);
			image_width = magUpInt(i_GetImageHeight(image), mw->mag_exp);
			// bogus what does this do?
			int left_edge = ((pbox->w+pbox->x)<<(mw->mag_exp+4))>>4;
		   	
			if(bbNullP(*pbox) == FALSE && left_edge < image_width){
				pbox->w++;
				i_SetUserDataProp(sole_entity, "PBB", (void*)pbox, sizeof(iBox));
				mw->document->getCurrentMode()->getCurrentPage()->track(RIGHT);
				mw->document->setDirty();
			}
		}
	}
}
void 
pitchShrinkHorzCB (Widget    w,
				   MainWindow *mw,
				   void *call)
{
	iDAFSError e;
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	iBox *pbox;
	if(sole_entity){
		e = i_GetUserDataProp(sole_entity, "PBB", (void**)&pbox, 0,i_SwapUBox);
		if(bbNullP(*pbox) == FALSE){
			pbox->w--;
			i_SetUserDataProp(sole_entity, "PBB", 
							  (void*)pbox,sizeof(iBox));
			mw->document->getCurrentMode()->getCurrentPage()->track(RIGHT);
			mw->document->setDirty();
		}
	}
}
void 
pitchMoveRightCB (Widget    w,
				  MainWindow *mw,
				  void *call)
{
	iDAFSError e;
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	iBox *pbox;
	if(sole_entity){
		e = i_GetUserDataProp(sole_entity, "PBB", (void**)&pbox, 0,i_SwapUBox);
		if(bbNullP(*pbox) == FALSE){
			pbox->x++;
			i_SetUserDataProp(sole_entity, "PBB", (void*)pbox,sizeof(iBox));
			mw->document->getCurrentMode()->getCurrentPage()->track(RIGHT);
			mw->document->setDirty();
		}
	}
}
void 
pitchMoveLeftCB (Widget    w,
				 MainWindow *mw,
				 void *call)
{
	iDAFSError e;
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	iBox *pbox;
	if(sole_entity){
		e = i_GetUserDataProp(sole_entity, "PBB", (void**)&pbox, 0,i_SwapUBox);
		if(bbNullP(*pbox) == FALSE){
			pbox->x--;
			i_SetUserDataProp(sole_entity, "PBB", (void*)pbox,sizeof(iBox));
			mw->document->getCurrentMode()->getCurrentPage()->track(LEFT);
			mw->document->setDirty();
		}
	}
}
void 
pitchGrowVertCB (Widget    w,
				 MainWindow *mw,
				 void *call)
{
	iDAFSError e;
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	iBox *pbox;
	int image_height;
	iImagePtr image;
	if(sole_entity){
		if (image = i_GetImage(sole_entity)){
			e = i_GetUserDataProp(sole_entity, "PBB", (void**)&pbox, 0,i_SwapUBox);
			int bottom_edge = ((pbox->y+pbox->h)<<(mw->mag_exp+4))>>4;

			if(bbNullP(*pbox) == FALSE && bottom_edge < image_height){
				pbox->h++;
				i_SetUserDataProp(sole_entity, "PBB", (void*)pbox,sizeof(iBox));
				mw->document->getCurrentMode()->getCurrentPage()->track(BOTTOM);
				mw->document->setDirty();
			}
		}
	}
}
void 
pitchShrinkVertCB (Widget    w,
				   MainWindow *mw,
				   void *call)
{
	iDAFSError e;
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	iBox *pbox;
	if(sole_entity){
		e = i_GetUserDataProp(sole_entity, "PBB", (void**)&pbox, 0,i_SwapUBox);
		if(bbNullP(*pbox) == FALSE){
			pbox->h--;
			i_SetUserDataProp(sole_entity, "PBB", (void*)pbox,sizeof(iBox));
			mw->document->getCurrentMode()->getCurrentPage()->track(BOTTOM);
			mw->document->setDirty();
		}
	}
}
void 
pitchMoveUpCB (Widget    w,
			   MainWindow *mw,
			   void *call)
{
	iDAFSError e;
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	iBox *pbox;
	if(sole_entity){
		e = i_GetUserDataProp(sole_entity, "PBB", (void**)&pbox, 0,i_SwapUBox);
		if(bbNullP(*pbox) == FALSE){
			pbox->y--;
			i_SetUserDataProp(sole_entity, "PBB", (void*)pbox, sizeof(iBox));
			mw->document->getCurrentMode()->getCurrentPage()->track(TOP);
			mw->document->setDirty();
		}
	}
}
void 
pitchMoveDownCB (Widget    w,
				 MainWindow *mw,
				 void *call)
{
	iDAFSError e;
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	iBox *pbox;
	int image_height;
	
	if(sole_entity){
		e = i_GetUserDataProp(sole_entity, "PBB", (void**)&pbox, 0,i_SwapUBox);
		image_height = magUpInt(i_GetImageHeight(i_GetImage(sole_entity)), mw->mag_exp);
		int bottom_edge = ((pbox->h+pbox->y)<<(mw->mag_exp+4))>>4;
	   	
		if(bbNullP(*pbox) == FALSE && bottom_edge < image_height){
			pbox->y++;
			i_SetUserDataProp(sole_entity, "PBB", (void*)pbox,sizeof(iBox));
			mw->document->getCurrentMode()->getCurrentPage()->track(TOP);
			mw->document->setDirty();
		}
	}
}

int joinRecurse(iEntityPtr en, iBox *ub)
{
	iEntityPtr cur,par,next;
	iBox b,ib;	
	par=i_GetParentEntity(en);
	if (!par)
		return 0;
	for (cur=i_GetChildEntity(par); cur; cur=next) {
		next=i_GetNextEntity(cur);
		if (cur==en)
			continue;
		b=i_GetEnclosingBox(cur);
		i_IntersectBox(b,*ub,&ib);
		if (ib.w) {
			i_UnionBox(b,*ub,ub);
			i_DisposeEntity(cur);
			return 1;
		}
	}
	return joinRecurse(par,ub);
	/* if box changes, redo */
}


void joinCB (Widget    w,
			   MainWindow *mw,
			   void *call)
{
	// make a copy of selection for fear of
	// referencing a list as it is changing
	const DLList *selection=mw->sm->getItems();
	int good;
	Ref r;
	iEntityPtr entity,joine,cur,next;
	int dirty=FALSE;
	iBox b,ub;

	r = dllistFirst(selection);
	// we shouldnt be allowed to happen by menu sensitivity
	if (r == 0)
		goto cleanup;
	joine = (iEntityPtr) dllistContents(selection, r);
	/* set joine to any parent that might be selected */
	for (entity=i_GetParentEntity(joine); entity; entity=i_GetParentEntity(entity))
		if (mw->sm->selectedItemP(entity))
			joine=entity;
	ub = i_GetEnclosingBox(joine);

	for(dllistNext(selection, &r); r; dllistNext(selection, &r)){
		entity = (iEntityPtr) dllistContents(selection, r);
		b = i_GetEnclosingBox(entity);
		i_UnionBox(b,ub,&ub);
		dirty = TRUE;
	}
	if (!dirty) 
		goto cleanup;

	good=not_root(mw,"You can't join the root entity.");
	if (good) {
		mw->sm->unselect(joine);
		mw->deleteSelection(FALSE);
		for (cur=i_GetChildEntity(joine); cur; cur=next) {
			next=i_GetNextEntity(cur);
			mw->sm->unselect(cur);
			i_DisposeEntity(cur);
		}
		while(joinRecurse(joine,&ub))
			;
		i_SetBox(joine,ub);
		/* need to fix ancestor's bounding boxes */
		for (cur=i_GetParentEntity(joine); cur; cur=i_GetParentEntity(cur)) {
			b=i_GetEnclosingBox(cur);
			i_UnionBox(b,ub,&ub);
			i_SetBox(cur,ub);
		}
		mw->document->setDirty();
		mw->sm->setItem(joine);
	}

  cleanup:
	;
}

void splitCB (Widget    w,
			   MainWindow *mw,
			   void *call)
{
	Ref r;
	iBox b1,b2;
	iEntityPtr split,new_en;
	
	const DLList *selection=mw->sm->getItems();
	r = dllistFirst(selection);
	// we shouldnt be allowed to happen by menu sensitivity
	if (r == 0)
		goto cleanup;
	split = (iEntityPtr) dllistContents(selection, r);
	b1=i_GetEnclosingBox(split);
	if (b1.w>1) {
		b2=b1;
		b1.w/=2;
		b2.x+=b1.w;
		b2.w-=b1.w;
		i_SetBox(split,b1);
		new_en=i_NewEntityAfter(i_GetEntityType(split),split);
		i_SetBox(new_en,b2);
		mw->document->addNewEntity(new_en, TRUE);
		mw->document->setDirty();
	}

	cleanup:
		;
}

#endif


void delparentCB (Widget    w,
			   MainWindow *mw,
			   void *call)
{
	Ref r;
	iEntityPtr par,cur,next;
	
	if (!common_parent(mw,"All the selected entities must have common parent"))
		return;
	const DLList *selection=mw->sm->getItems();
	r = dllistFirst(selection);
	par = i_GetParentEntity((iEntityPtr) dllistContents(selection, r));
	
	if (par==mw->document->root_entity) {
		uiWarnUser("You can't delete the root entity.", "OK");
	} else {
		for (cur=i_GetChildEntity(par); cur; cur=next) {
			next=i_GetNextEntity(cur);
			i_MoveAfter(cur,par);
		}
		i_DisposeEntity(par);
		mw->document->setDirty();
	}
}
			  
void newparentCB (Widget    w,
			   MainWindow *mw,
			   void *call)
{
	const DLList *selection=mw->sm->getItems();
	int good;
	Ref r;
	iEntityPtr entity,firste,par,cur;
	iBox b,ub,ib;

	r = dllistFirst(selection);
	firste= (iEntityPtr) dllistContents(selection, r);
	ub = i_GetEnclosingBox(firste);
	for(dllistNext(selection, &r); r; dllistNext(selection, &r)){
		entity = (iEntityPtr) dllistContents(selection, r);
		b = i_GetEnclosingBox(entity);
		i_UnionBox(b,ub,&ub);
	}
	good=common_parent(mw,"Entities must have a common parent to put in new parent.");
	if (good) {
		DLList scratch_list;
		dllistCtor(&scratch_list);
		for(r = dllistFirst(selection); r; dllistNext(selection, &r)){
			entity = (iEntityPtr) dllistContents(selection, r);
			dllistAppend(&scratch_list, entity);
		}
		/* wipe out all siblings that intersect */
		for (cur=i_GetChildEntity(i_GetParentEntity(firste)); cur; cur=i_GetNextEntity(cur)) {
			b=i_GetEnclosingBox(cur);
			i_IntersectBox(b,ub,&ib);
			if (ib.w && !dllistSeek(&scratch_list,cur)) {
				i_UnionBox(b,ub,&ub);
				dllistAppend(&scratch_list,cur);
			}
		}

		r = dllistFirst(selection);
		firste = (iEntityPtr) dllistContents(selection, r);
		par=i_NewEntityAfter(i_GetEntityType(firste),firste);
		i_SetBox(par,ub);
		mw->document->addNewEntity(par, TRUE);
		for(r=dllistFirst(&scratch_list); r; dllistNext(&scratch_list, &r)){
			entity=(iEntityPtr) dllistContents(&scratch_list, r);
			i_MoveToBack(entity,par);
		}
		dllistDtor(&scratch_list);
		i_Sort(par,i_VHCompare);
		mw->sm->setItem(par);
		mw->document->setDirty();
		mw->setSensitive();
	}
}

void imagepageCB (Widget    w,
			   MainWindow *mw,
			   void *call)
{
	iEntityPtr root=mw->document->root_entity;
	iEntityPtr cur,next,last;
	iBox b,ub;
	iTypeIndex type;
	
	// have to remember not to delete finishing newline entity
	last=i_GetLastEntity(root);
	if (!last)
		return;
	
	ub.w=0;
	for (cur=i_GetChildEntity(root); cur!=last; cur=next) {
		next=i_GetNextEntity(cur);
		b=i_GetEnclosingBox(cur);
		i_UnionBox(b,ub,&ub);
		mw->sm->unselect(cur);
		i_DisposeEntity(cur);
	}
	if (ub.w) {
		i_NameToType("Non-Text",&type);
		cur=i_NewEntityBefore(type,last);
		i_SetBox(cur,ub);
		mw->document->addNewEntity(cur, TRUE);
	}
}

void 
entityGrowHorzCB (Widget    w,
				  MainWindow *mw,
				  void *call)
{
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	iBox box;
	iImagePtr image;
	int image_width;
	
	if(sole_entity){ 
		if(image = i_GetImage(sole_entity)){ 
			box = i_GetEnclosingBox(sole_entity);
			image_width = magUpInt(i_GetImageWidth(image), mw->mag_exp);
			// bogus -- what does this mean
			int right_edge = ((box.w+box.x)<<(mw->mag_exp+4))>>4;
	   	
			if(bbNullP(box) == FALSE && right_edge < image_width){
				box.w++;
				i_SetBox(sole_entity, box);
				mw->document->getCurrentMode()->getCurrentPage()->track(RIGHT);
				mw->document->setDirty();
			}
		}
	}
}

void 
entityShrinkHorzCB (Widget    w,
					MainWindow *mw,
					void *call)
{   
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	iBox box;
	
	if(sole_entity){
		box = i_GetEnclosingBox(sole_entity);
		if(bbNullP(box) == FALSE && box.w != 1){
			box.w--;
			i_SetBox(sole_entity, box);
			if (!visiblePositionP(mw->scroll_window,box.x + box.w, box.y + box.h))
				mw->document->getCurrentMode()->getCurrentPage()->track(RIGHT);
			mw->document->setDirty();
		}
	}
}


void 
entityMoveRightCB (Widget    w,
				   MainWindow *mw,
				   void *call)
{   
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	iBox box,*pb;
	iImagePtr image;
	int image_width;
	
	if(sole_entity){ 
		if(image = i_GetImage(sole_entity)){ 
			box = i_GetEnclosingBox(sole_entity);
			image_width = magUpInt(i_GetImageWidth(image), mw->mag_exp);
			// this makes sure we never shift by a negative number
			int right_edge = ((box.w+box.x)<<(mw->mag_exp+4))>>4;
	   	
			if(bbNullP(box) == FALSE && right_edge < image_width){
				box.x++;
				if (!i_GetUserDataProp(sole_entity, "PBB",(void**) &pb , 0, i_SwapUBox))
					pb->x--;
				i_SetBox(sole_entity, box);
				if (!visiblePositionP(mw->scroll_window, box.x + box.w, box.y))
					mw->document->getCurrentMode()->getCurrentPage()->track(RIGHT);
				mw->document->setDirty();
			}
		}
	}
}


void 
entityMoveLeftCB (Widget    w,
				  MainWindow *mw,
				  void *call)
{  
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	iBox box,*pb;
	
	if(sole_entity){
		box = i_GetEnclosingBox(sole_entity);
		if(bbNullP(box) == FALSE && box.x != 0){
			box.x--;
			if (!i_GetUserDataProp(sole_entity, "PBB",(void**) &pb , 0, i_SwapUBox))
				pb->x++;
			i_SetBox(sole_entity, box);
			mw->document->getCurrentMode()->getCurrentPage()->track(LEFT);
			mw->document->setDirty();
		}
	}
}


void 
entityGrowVertCB (Widget    w,
				  MainWindow *mw,
				  void *call)
{   
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	iBox box;
	iImagePtr image;
	int image_height;
	
	if(sole_entity){ 
		if(image = i_GetImage(sole_entity)){ 
			image_height = magUpInt(i_GetImageHeight(image), mw->mag_exp);
			box = i_GetEnclosingBox(sole_entity);
			// bogus what does this do?
			int bottom_edge = ((box.h+box.y)<<(mw->mag_exp+4))>>4;
	   	
			if(bbNullP(box) == FALSE && bottom_edge < image_height){
				box.h++;
				i_SetBox(sole_entity, box);
				mw->document->getCurrentMode()->getCurrentPage()->track(BOTTOM);
				mw->document->setDirty();
			}
		}
	}
}


void 
entityShrinkVertCB (Widget    w,
					MainWindow *mw,
					void *call)
{  
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	iBox box;
	
	if(sole_entity){
		box = i_GetEnclosingBox(sole_entity);
		if(bbNullP(box) == FALSE && box.h != 1){
			box.h--;
			i_SetBox(sole_entity, box);
			mw->document->getCurrentMode()->getCurrentPage()->track(BOTTOM);
			mw->document->setDirty();
		}
	}
	
}


void 
entityMoveUpCB (Widget    w,
				MainWindow *mw,
				void *call)
{  
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	iBox box,*pb;
	
	if(sole_entity){
		box = i_GetEnclosingBox(sole_entity);
		if(bbNullP(box) == FALSE && box.y != 0){
			box.y--;
			if (!i_GetUserDataProp(sole_entity, "PBB",(void**) &pb , 0, i_SwapUBox))
				pb->y++;
			i_SetBox(sole_entity, box);
			mw->document->getCurrentMode()->getCurrentPage()->track(TOP);
			mw->document->setDirty();
		}
	}
}


void 
entityMoveDownCB (Widget    w,
				  MainWindow *mw,
				  void *call)
{   
	iEntityPtr sole_entity = mw->sm->getSoleItem();
	iBox box,*pb;
	iImagePtr image;
	int image_height;
	
	if(sole_entity){ 
		if(image = i_GetImage(sole_entity)){ 
			box = i_GetEnclosingBox(sole_entity);
			image_height = magUpInt(i_GetImageHeight(image), mw->mag_exp);
			// bogus what does this do?
			int bottom_edge = ((box.h+box.y)<<(mw->mag_exp+4))>>4;
	   	
			if(bbNullP(box) == FALSE && bottom_edge < image_height){
				box.y++;
				if (!i_GetUserDataProp(sole_entity, "PBB",(void**) &pb , 0, i_SwapUBox))
					pb->y--;
				i_SetBox(sole_entity, box);
				mw->document->getCurrentMode()->getCurrentPage()->track(BOTTOM);
				mw->document->setDirty();
			}
		}
	}
}

#if 0
int selCB(MainWindow *mw, char *name)
{
    if (!*name)
		return FALSE;
	
    mw->document->selectPage(name);
    mw->setSensitive();
    mw->standardTitle();
	
    return TRUE;
}
#endif
void newEntityCB(Widget w,
				 MainWindow *mw,
				 void *call)
{
	iEntityPtr sole_entity =(iEntityPtr)mw->sm->getSoleItem();
	iEntityPtr new_entity;
	iTypeIndex type;
	
	if (!sole_entity) return;
#ifdef RAF_INTERNAL
	switch (i_GetEntityType(sole_entity)){
	case iParagraphT:
		type = iLineT;
		break;
	case iWordT:
		type = iGlyphT;
		break;
	case iDocumentT:
		type = iParagraphT;
		break;
	case iLineT:
		type = iWordT;
		break;
	default:
		type = iGlyphT;
		break;
	}
#else
	i_NameToType("Type",&type);
#endif
	new_entity = i_NewEntity(type,sole_entity);
	i_MoveToFront(new_entity, sole_entity);
	i_SetBox(new_entity,i_GetEnclosingBox(sole_entity));
	mw->sm->unselect(sole_entity);
	i_DoCallBack(mUnSelect,sole_entity);
	mw->sm->setItem(new_entity);
	mw->document->setDirty();
	mw->document->addNewEntity(new_entity,TRUE);
	mw->centerOnSelected();
	mw->setSensitive();
}
void inputUnicodeCB (Widget    w,
					 MainWindow *mw,
					 void *call)
{
	mw->uniedit->show();
}




char *about_japanese_text[]={
	"Illuminator works with a Japanese input server called      ",
	"irohasrv. You are having some problem connecting to it.    ",
	"",
	"First, irohasrv must be running on a machine of the same   ",
	"architecture as the one your illum will run on.            ",
    "Second, you must either run illum on the same machine as   ",
	"irohasrv, or set the CANNAHOST environment variable to that",
	"machine, or supply that machine to the -j option of the    ",
	"illum command line.                                        ",
	"",
	"Restart illum after taking these steps. If they don't work,",
	"contact your system administrator.                         ",
	0
};

void inputJapaneseCB(Widget    w,
					 MainWindow *mw,
					 void *call)
{
	int ok = 0;
	iUCode error_buffer[256];

	if(mw->app->nihonInitializedP() == FALSE)
		ok = mw->app->initializeNihon(error_buffer);
	else
		ok = 1;
	if(ok == FALSE){
#if 0
		WarnUserUcs2(mw->app_shell, error_buffer, "OK");
#else
		uiTellUserN(about_japanese_text,"  OK  ");
#endif
	} else {
		mw->nihon_edit->show();
	}
}


void
entitySortCB(Widget w,
			   MainWindow *mw,
			   XtPointer call_data)
{
	iEntityPtr entity=mw->sm->getSoleItem();
	
	if (entity){
		i_Sort(entity, i_VHCompare); /* error? */
		mw->document->setDirty();
	}
}





void SaveAsCB (Widget    w,
			   MainWindow *mw,
			   void *call)
{
	aString directory = mw->document->current_directory;
	
	if(mw->save_as_dialog == 0){
		mw->save_as_dialog = new SaveAsDialog(mw, mw->app->app_shell,
											  mw->app->app_context,
											  directory);
	} else {
		mw->save_as_dialog->setDir(directory);
	}

	mw->document->would_be_savetype=mw->document->save_type;
	
	mw->save_as_dialog->setSaveType(mw->document->save_type);
	
	mw->save_as_dialog->showModal();
	// will not return until accept or cancel
}


#ifdef RAF_INTERNAL

extern "C" iDAFSError i_whitezone(iImagePtr ip, int x, int y, int w, int h);

void 
whiteEntityCB(Widget w,
			  MainWindow *mw,
			  void *call)
{
	iEntityPtr en = 0;
	iBox b;
	iImagePtr im;

	en = mw->sm->getSoleItem();
	if (en && (im = i_GetImage(en)) && (b = i_GetEnclosingBox(en)).w){
		i_DoCallBack(mWillUpdate, en);
		i_whitezone(im, b.x, b.y, b.w, b.h);
		i_DoCallBack(mUpdate, en);
		mw->document->setDirty();
	}
}

#endif
