
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"


#include <X11/Xutil.h>

#include <Xm/MainW.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
// really used for menus
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/ToggleBG.h>
// really used for creating x window stuff
#include <Xm/RowColumn.h>
#include <Xm/PanedW.h>
#include <Xm/ScrolledW.h>
#include <Xm/ScrollBar.h>

#include "rafutil.h"

#ifdef RAF_INTERNAL
#include "form.h"
#endif

#include "smemory.h"

#include "FileName.h"

#include "misc/XArg.h"
#include "misc/color.h"
#include "misc/misc.h"

#include "keyboard/KeyBoard.h"
#include "keyboard/KeyMap.h"

#include "scroll.h"
#include "misc/bb.h"
#include "misc/prompt.h"



#include "Document.h"
#include "Document.h"
#include "MainWindow.h"
#include "MainWindow.cb.h"
#include "Page.h"
#include "ViewCanvas.h"
#include "ChooseDialog.h"
#include "EntityDialog.h"
#include "TypePrefDialog.h"

#include "SelectionManager.h"
#include "UniEdit.h"
#include "NihonEdit.h"
#include "ListDialog.h"
#include "Application.h"


#include "Mode.h"


char ident1[]="illum v.10";

extern KeyMap *initAmerican();
extern KeyMap *initArabic();
extern KeyMap *initRussian();
extern KeyMap *initEnglish();
extern KeyMap *initGerman();
extern KeyMap *initFrench();
extern KeyMap *initGreek();
extern KeyMap *initMicr();
extern KeyMap *initSwedish();
extern KeyMap *initSpanish();
extern KeyMap *initNorw();
extern KeyMap *initItalian();


#define STD_TYPE_COLOR "blue"
#define STD_TYPE_LENGTH 256


MainWindowMenuData::MainWindowMenuData(MainWindow *mw)
{
	eighth_data = new Tuple(mw,(void *) -3);
	quarter_data = new Tuple(mw,(void *) -2);
	half_data = new Tuple(mw,(void *) -1);
	normal_data = new Tuple(mw,(void *) 0);
	double_data = new Tuple(mw,(void *) 1);
	quadruple_data = new Tuple(mw,(void *) 2);
	octuple_data = new Tuple(mw,(void *) 3);

	sixteen_data = new Tuple(mw,(void *) 4);

	image_data = new Tuple(mw,(void *) vtIMAGE);
	ooc_data = new Tuple(mw,(void *) vtOOC);
	flag_data = new Tuple(mw,(void *) vtFLAG);
	tree_data = new Tuple(mw,(void*) vtTREE);
	rte_data = new Tuple(mw,(void*)vtRTE);
}


MainWindowMenuData::~MainWindowMenuData(MainWindow *mw)
{
	delete eighth_data;
	delete quarter_data;
	delete half_data;
	delete normal_data;
	delete double_data;
	delete quadruple_data;
	delete octuple_data;
	delete sixteen_data;
	
	delete image_data;
	delete ooc_data;
	delete flag_data;
	delete tree_data;

	delete rte_data;
#ifdef RAF_INTERNAL

	delete ideal_data;
	delete damaged_data;
	delete ambiguous_data;
	delete badbox_data;
	delete noise_data;
#endif
}


/*
// Menus
*/

// the order in this list must match the order in the list below
static MenuItem displays[] = {
	{ 0, "Keyboard ...", &xmPushButtonGadgetClass, 'K', 
	  NULL, NULL, KeyboardCB, 0, NULL },
	{ 0, "View Selection ...", &xmPushButtonGadgetClass, 'S', 
	  "Ctrl Shift<Key>V", "ctrl-shift-v", (XtCallbackProc) viewCB, 0, NULL },
	NULL_MENU,
};

// the order in this list must match the order in the list above
enum {
	display_keyboard,
	display_selection,
};

static MenuItem prefs[] = {
	{ 0, "Display in Image mode ...", &xmPushButtonGadgetClass, 'C', 
	  NULL, NULL, ChooseCB, 0, NULL },
	{ 0, "Type Properties ...", &xmPushButtonGadgetClass, 'T', 
	  NULL, NULL, ColorsCB, 0, NULL },
#ifdef DO_POPUPS
	{ 0, "Flag Popups", &xmToggleButtonGadgetClass, 'F', 
	  NULL, NULL, FlagCB, 0, NULL },
#endif
	{ 0, "Text on Image", &xmToggleButtonGadgetClass, 'T', 
	  "Ctrl<Key>k","ctrl-k", TextCB, 0, NULL },
	NULL_MENU,
};

// the order in this list must match the order in the list above
enum {
	prefs_choose,
	prefs_colors,
#ifdef DO_POPUPS
	prefs_popup,
#endif
	prefs_text,
};


#ifdef RAF_INTERNAL

static MenuItem settype[] = {
	{ 0, "Alien", &xmPushButtonGadgetClass, 0,
	  "<Key>F1", "F1", (XtCallbackProc) changeTypeCB, 0, NULL },
	{ 0, "section", &xmPushButtonGadgetClass, 0,
	  "<Key>F2", "F2", (XtCallbackProc) changeTypeCB, 0, NULL },
	{ 0, "para0", &xmPushButtonGadgetClass, 0, 
	  "<Key>F3", "F3", (XtCallbackProc) changeTypeCB, 0, NULL },
	{ 0, "subpara1", &xmPushButtonGadgetClass, 0, 
	  "<Key>F4", "F4", (XtCallbackProc) changeTypeCB, 0, NULL },
	{ 0, "subpara2", &xmPushButtonGadgetClass, 0,
	  "<Key>F5", "F5", (XtCallbackProc) changeTypeCB, 0, NULL },
	{ 0, "subpara3", &xmPushButtonGadgetClass, 0,
	  "<Key>F6", "F6", (XtCallbackProc) changeTypeCB, 0, NULL },
	{ 0, "figure", &xmPushButtonGadgetClass, 0,
	  "<Key>F7", "F7", (XtCallbackProc) changeTypeCB, 0, NULL },
	{ 0, "Non-Text", &xmPushButtonGadgetClass, 0,
	  "<Key>F8", "F8", (XtCallbackProc)changeTypeCB, 0, NULL},
	{ 0, "Paragraph", &xmPushButtonGadgetClass, 0,
	  "<Key>F9", "F9", (XtCallbackProc)changeTypeCB, 0, NULL},
	{ 0, "Column", &xmPushButtonGadgetClass, 0,
	  "<Key>F10", "F10", (XtCallbackProc)changeTypeCB, 0, NULL},
	NULL_MENU,
};

static MenuItem types[] = {
	{ 0, UNKNOWN_TYPE, &xmPushButtonGadgetClass, 'U', 
	  "Ctrl<Key>7", NULL, UnknownTypeCB, 0, NULL },
	{ 0, HAND_TYPE, &xmPushButtonGadgetClass, 'H',
	  "", NULL, (XtCallbackProc) HandTypeCB, 0, NULL },
	{ 0, VARIABLE_TYPE, &xmPushButtonGadgetClass, 'V', 
	  "Ctrl<Key>9", NULL, (XtCallbackProc) VariableTypeCB, 0, NULL },
	{ 0, FIXED_TYPE, &xmPushButtonGadgetClass, 'F', 
	  "Ctrl<Key>0","", (XtCallbackProc) FixedTypeCB, 0, NULL },
	NULL_MENU,	
};
#endif

static MenuItem mags[] = {
	{ 0, "1/8", &xmPushButtonGadgetClass, '\0', 
	  "Shift Ctrl<Key>8", "ctrl-shift-8", (XtCallbackProc) MagCB, 0, NULL },
	{ 0, "1/4", &xmPushButtonGadgetClass, '\0', 
	  "Shift Ctrl<Key>4", "ctrl-shift-4", (XtCallbackProc) MagCB, 0, NULL },
	{ 0, "1/2", &xmPushButtonGadgetClass, '\0', 
	  "Shift Ctrl<Key>2", "ctrl-shift-2", (XtCallbackProc) MagCB, 0, NULL },
	{ 0, " 1 ", &xmPushButtonGadgetClass, '\0', 
	  "Ctrl<Key>1", "ctrl-1", (XtCallbackProc) MagCB, 0, NULL },
	{ 0, " 2 ", &xmPushButtonGadgetClass, '\0', 
	  "Ctrl<Key>2", "ctrl-2", (XtCallbackProc) MagCB, 0, NULL },
	{ 0, " 4 ", &xmPushButtonGadgetClass, '\0', 
	  "Ctrl<Key>4", "ctrl-4", (XtCallbackProc) MagCB, 0, NULL },
	{ 0, " 8 ", &xmPushButtonGadgetClass, '\0', 
	  "Ctrl<Key>8", "ctrl-8", (XtCallbackProc) MagCB, 0, NULL },
#ifdef RAF_INTERNAL
	{ 0, "16 ", &xmPushButtonGadgetClass, '\0', 
	  "", "", (XtCallbackProc) MagCB, 0, NULL },
#endif
	NULL_MENU,
};

static MenuItem keyboards[] = {
	{ 0, "American", &xmPushButtonGadgetClass, 'U', 
	  NULL, NULL, (XtCallbackProc) AmericanCB, 0, NULL },	
	{ 0, "Arabic", &xmPushButtonGadgetClass, 'A',
	  NULL, NULL, (XtCallbackProc) ArabicCB, 0, NULL },
	{ 0, "British", &xmPushButtonGadgetClass, 'B', 
	  NULL, NULL, (XtCallbackProc) EnglishCB, 0, NULL },
	{ 0, "French", &xmPushButtonGadgetClass, 'F', 
	  NULL, NULL, (XtCallbackProc) FrenchCB, 0, NULL },
	{ 0, "German", &xmPushButtonGadgetClass, 'G', 
	  NULL, NULL, (XtCallbackProc) GermanCB, 0, NULL },
	{ 0, "Greek", &xmPushButtonGadgetClass, 'k', 
	  NULL, NULL, (XtCallbackProc) GreekCB, 0, NULL },
	{ 0, "Italian", &xmPushButtonGadgetClass, 'I', 
	  NULL, NULL, (XtCallbackProc) ItalianCB, 0, NULL },
	{ 0, "MICR", &xmPushButtonGadgetClass, 'M', 
	  NULL, NULL, (XtCallbackProc) MicrCB, 0, NULL },
	{ 0, "Norwegian", &xmPushButtonGadgetClass, 'N', 
	  NULL, NULL, (XtCallbackProc) NorwCB, 0, NULL },
	{ 0, "Russian", &xmPushButtonGadgetClass, 'R', 
	  NULL, NULL, (XtCallbackProc) RussianCB, 0, NULL },	
	{ 0, "Spanish", &xmPushButtonGadgetClass, 'p', 
	  NULL, NULL, (XtCallbackProc) SpanishCB, 0, NULL },
	{ 0, "Swedish", &xmPushButtonGadgetClass, 'S', 
	  NULL, NULL, (XtCallbackProc) SwedishCB, 0, NULL },
	NULL_MENU,
};

static MenuItem modes[] = {
	{ 0, "Image", &xmPushButtonGadgetClass, 'I', 
	  "Ctrl<Key>i", "ctrl-i", (XtCallbackProc) ModeCB, 0, NULL },

	{ 0, "OOC", &xmPushButtonGadgetClass, 'O', 
	  "Ctrl<Key>c", "ctrl-c", (XtCallbackProc) ModeCB, 0, NULL },
	{ 0, "Flagged", &xmPushButtonGadgetClass, 'F', 
	  "Ctrl<Key>m", "ctrl-m", (XtCallbackProc) ModeCB, 0, NULL },
	{ 0, "Hierarchy", &xmPushButtonGadgetClass, 'H',
	  "Ctrl<Key>h", "ctrl-h",(XtCallbackProc)  ModeCB, 0, NULL },
	{ 0, "Text",&xmPushButtonGadgetClass, 'R',
	  "Ctrl<Key>t","ctrl-t",(XtCallbackProc) ModeCB,0,NULL},
	NULL_MENU,
};

/*
// Pulldown menus
*/

enum {
	file_new_index,
	file_open_index,
	file_save_index,
	file_saveas_index,
	file_revert_index,
	file_next,
	file_prev,
	file_page_up,
	file_page_down,
#if RAF_INTERNAL
	file_reinit,
#endif
	file_quit,
};

static MenuItem files[] = {
	{ 0, "New", &xmPushButtonGadgetClass, 'e', 
	  0,0, NewCB, 0, NULL },
	{ 0, "Open ...", &xmPushButtonGadgetClass, 'O', 
	  "Ctrl<Key>o", "ctrl-o", OpenCB, 0, NULL },
	{ 0, "Save", &xmPushButtonGadgetClass, 'S', 
	  "Ctrl<Key>s", "ctrl-s", SaveCB, 0, NULL },
	{ 0, "Save As ...", &xmPushButtonGadgetClass, 'a',
	  "Ctrl<Key>a", "ctrl-a", (XtCallbackProc) SaveAsCB, 0, NULL },
	{ 0, "Revert to Saved", &xmPushButtonGadgetClass, 'R', 
	  0, 0, RevertCB, 0, NULL },
	{ 0, "Next File", &xmPushButtonGadgetClass, 'N', 
	  "Ctrl<Key>n", "ctrl-n", (XtCallbackProc) NextCB, 0, NULL },
	{ 0, "Previous File", &xmPushButtonGadgetClass, 'P', 
	  "Ctrl<Key>p", "ctrl-p", (XtCallbackProc) PrevCB, 0, NULL },
	{ 0, "Page Up", &xmPushButtonGadgetClass, 'U', 
	  "<Key>Prior", "PgUp", (XtCallbackProc) pageUpCB, 0, NULL },
	{ 0, "Page Down", &xmPushButtonGadgetClass, 'D', 
	  "<Key>Next", "PgDn", (XtCallbackProc) pageDownCB, 0, NULL },
	{ 0, "Page Right", &xmPushButtonGadgetClass, 'R', 
	  "Ctrl<Key>Prior", "ctrl-PgUp", (XtCallbackProc) pageLeftCB, 0, NULL },
	{ 0, "Page Left", &xmPushButtonGadgetClass, 'L', 
	  "Ctrl<Key>Next", "ctrl-PgDn", (XtCallbackProc) pageRightCB, 0, NULL },
#ifdef RAF_INTERNAL
	{ 0, "Reinitialize", &xmPushButtonGadgetClass, 'R', 
	  0, 0, (XtCallbackProc) ReinitializeCB, 0, NULL },
#endif
	{ 0, "Quit", &xmPushButtonGadgetClass, 'Q', 
	  "Ctrl<Key>q", "ctrl-q", (XtCallbackProc) QuitCB, 0, NULL },
	NULL_MENU,
};

enum EditFunctions {
	edit_delete,
#ifdef RAF_INTERNAL
	edit_join,
	edit_split,
	edit_del_parent,
	edit_new_parent,
	edit_image_page,
#endif
	edit_kill_word,
	edit_left,
	edit_right,
	edit_up,
	edit_down,
	edit_flag,
	edit_new,
	edit_uni,
	edit_nihon,
};

typedef enum EditFunctions EditFunctions;

static MenuItem edits[] = {
	{ 0, "Delete", &xmPushButtonGadgetClass, 'D', 
	  "<Key>Delete", "Delete", (XtCallbackProc) deleteCB, 0, NULL },
#ifdef RAF_INTERNAL
	{ 0, "Join", &xmPushButtonGadgetClass, 'J', 
	  "Ctrl<Key>j", "ctrl-j", (XtCallbackProc) joinCB, 0, NULL },
	{ 0, "Split", &xmPushButtonGadgetClass, 'J', 
	  "", "", (XtCallbackProc) splitCB, 0, NULL },
	{ 0, "Delete Parent", &xmPushButtonGadgetClass, 'J', 
	  "", "", (XtCallbackProc) delparentCB, 0, NULL },
	{ 0, "New Parent", &xmPushButtonGadgetClass, 'J', 
	  "", "", (XtCallbackProc) newparentCB, 0, NULL },
	{ 0, "Make Page Graphic", &xmPushButtonGadgetClass, 'J', 
	  "Ctrl<Key>w", "ctrl-w", (XtCallbackProc) imagepageCB, 0, NULL },
#endif
	{ 0, "Kill Contents", &xmPushButtonGadgetClass, 'K', 
	  "Ctrl<Key>u", "ctrl-u", (XtCallbackProc) editKillWordCB, 0, NULL },
	{ 0, "Previous", &xmPushButtonGadgetClass, 'P', 
	  "<Key>Left", "Left", PrevEntityCB, 0, NULL },
	{ 0, "Next", &xmPushButtonGadgetClass, 'N', 
	  "<Key>Right", "Right", NextEntityCB, 0, NULL },
	{ 0, "Up", &xmPushButtonGadgetClass, 'P', 
	  "<Key>Up", "Up", UpEntityCB, 0, NULL },
	{ 0, "Down", &xmPushButtonGadgetClass, 'N', 
	  "<Key>Down", "Down", DownEntityCB, 0, NULL },
	{ 0, "Find Next Flag", &xmPushButtonGadgetClass, 'F', 
	  "Ctrl<Key>g", "ctrl-g", FindNextFlagCB, 0, NULL },

	{ 0, "New Entity", &xmPushButtonGadgetClass, 'e', 
	  "Ctrl<Key>e", "ctrl-e", newEntityCB, 0, NULL },
	{ 0, "Enter Unicode ...", &xmPushButtonGadgetClass, 'U', 
	  "Ctrl<Key>F9", "ctrl-F9", inputUnicodeCB, 0, NULL },
	{ 0, "Japanese Input ...", &xmPushButtonGadgetClass, 'U', 
	  "Ctrl<Key>F10", "ctrl-F10", inputJapaneseCB, 0, NULL },
	NULL_MENU,
};

#if RAF_INTERNAL
static MenuItem pitches[] = {
	{ 0, "Shift Phase Left", &xmPushButtonGadgetClass, 'L', 
	  "Ctrl<Key>w", "ctrl-w", LeftPhaseCB, 0, NULL },
	{ 0, "Shift Phase Right", &xmPushButtonGadgetClass, 'R', 
	  "", "", RightPhaseCB, 0, NULL },
	{ 0, "Decrement Pitch", &xmPushButtonGadgetClass, 'D', 
	  "Ctrl<Key>y", "ctrl-y", DecrPitchCB, 0, NULL },
	{ 0, "Decrement Double", &xmPushButtonGadgetClass, 'e', 
	  "Shift Ctrl<Key>y", "shift-ctrl-y", DecrDoubleCB, 0, NULL },
	{ 0, "Increment Pitch", &xmPushButtonGadgetClass, 'I',
	  "Ctrl<Key>u", "ctrl-u", IncrPitchCB, 0, NULL },
	{ 0, "Increment Double", &xmPushButtonGadgetClass, 'c',
	  "Shift Ctrl<Key>u", "shift-ctrl-u", IncrDoubleCB, 0, NULL },
	NULL_MENU,
};
static MenuItem editPitch[] = {
	{ 0, "Grow Pitch Right", &xmPushButtonGadgetClass, 0, 
	  "Ctrl<Key>F4", "ctrl-F4", (XtCallbackProc) pitchGrowHorzCB,0, NULL },
	
	{ 0, "Shrink Pitch Right", &xmPushButtonGadgetClass, 0, 
	  "Ctrl<Key>F3", "ctrl-F3", (XtCallbackProc) pitchShrinkHorzCB, 0, NULL },
	{ 0, "Move Pitch Right", &xmPushButtonGadgetClass, 0, 
	  "Shift<Key>F4", "shift-F4", (XtCallbackProc) pitchMoveRightCB, 0, NULL },
	{ 0, "Move Pitch Left", &xmPushButtonGadgetClass, 0, 
	  "Shift<Key>F3", "shift-F3", (XtCallbackProc) pitchMoveLeftCB, 0, NULL },
	
	{ 0, "Grow Pitch Down", &xmPushButtonGadgetClass, 0, 
	  "Ctrl<Key>F2", "ctrl-F2", (XtCallbackProc) pitchGrowVertCB, 0, NULL },


	{ 0, "Shrink Pitch Down", &xmPushButtonGadgetClass, 0, 
	  "Ctrl<Key>F1", "ctrl-F1", (XtCallbackProc) pitchShrinkVertCB, 0, NULL },
	{ 0, "Move Pitch Up", &xmPushButtonGadgetClass, 0, 
	  "Shift <Key>F1", "shift-F1", (XtCallbackProc) pitchMoveUpCB, 0, NULL },

	{ 0, "Move Pitch Down", &xmPushButtonGadgetClass, 0, 
	  "Shift <Key>F2", "shift-F2", (XtCallbackProc) pitchMoveDownCB, 0, NULL },
	NULL_MENU
};
#endif

enum {
	entity_center,
	entity_rotate_image,
#if 0
	entity_clear,
#endif
#if RAF_INTERNAL
	entity_white,
	entity_recognize,
	entity_deskew,
	entity_debug,
	entity_sort,
#endif
	entity_grow_horz,
	entity_shrink_horz,
	entity_move_right,
	entity_move_left,
	entity_grow_vert,
	entity_shrink_vert,
	entity_move_up,
	entity_move_down,
};

static MenuItem entities[] = {
	{ 0, "Center", &xmPushButtonGadgetClass, 'C',
	  "Ctrl<Key>l", "ctrl-l", (XtCallbackProc) centerOnSelectedCB, 0, NULL },
	{ 0, "Rotate Image", &xmPushButtonGadgetClass, 'R',
	  0, 0, (XtCallbackProc) rotateCB, 0, NULL },
#ifdef RAF_INTERNAL
	{ 0, "White Out", &xmPushButtonGadgetClass, 'W', 
	  "", "", (XtCallbackProc) whiteEntityCB, 0, NULL },
	{ 0, "Recognize", &xmPushButtonGadgetClass, 'R', 
	  0, 0, (XtCallbackProc) recognizeCB, 0, NULL },
	{ 0, "Deskew Document", &xmPushButtonGadgetClass, 'R', 
	  "Ctrl<Key>d", "ctrl-d", (XtCallbackProc) deskewCB, 0, NULL },
	{ 0, "Debugging Stuff", &xmPushButtonGadgetClass, 'S', 
	  "Ctrl<Key>z", "ctrl-z", (XtCallbackProc) steves_specialCB, 0, NULL },
	{ 0, "Sort", &xmPushButtonGadgetClass, 0, 
	  "Ctrl<Key>v", "ctrl-v", (XtCallbackProc) entitySortCB,0, NULL },
#endif
	{ 0, "Grow Right", &xmPushButtonGadgetClass, 0, 
	  "Ctrl<Key>Right", "ctrl-Right", (XtCallbackProc) entityGrowHorzCB,0, NULL },
	{ 0, "Shrink Right", &xmPushButtonGadgetClass, 0, 
	  "Ctrl<Key>Left", "ctrl-Left", (XtCallbackProc) entityShrinkHorzCB, 0, NULL },
	{ 0, "Move Right", &xmPushButtonGadgetClass, 0, 
	  "Shift<Key>Right", "shift-Right", (XtCallbackProc) entityMoveRightCB, 0, NULL },
	{ 0, "Move Left", &xmPushButtonGadgetClass, 0, 
	  "Shift<Key>Left", "shift-Left", (XtCallbackProc) entityMoveLeftCB, 0, NULL },
	
	{ 0, "Grow Down", &xmPushButtonGadgetClass, 0, 
	  "Ctrl<Key>Down", "ctrl-Down", (XtCallbackProc) entityGrowVertCB, 0, NULL },
	{ 0, "Shrink Down", &xmPushButtonGadgetClass, 0, 
	  "Ctrl<Key>Up", "ctrl-Up", (XtCallbackProc) entityShrinkVertCB, 0, NULL },
	{ 0, "Move Up", &xmPushButtonGadgetClass, 0, 
	  "Shift <Key>Up", "shift-Up", (XtCallbackProc) entityMoveUpCB, 0, NULL },
	{ 0, "Move Down", &xmPushButtonGadgetClass, 0, 
	  "Shift <Key>Down", "shift-Down", (XtCallbackProc) entityMoveDownCB, 0, NULL },
	
	NULL_MENU,
};


enum {
	page_select_page,
	page_forward,
	page_backward,
};

static MenuItem pages[] = {
	{ 0, "Select ...", &xmPushButtonGadgetClass, 'S', 
	  "Ctrl<Key>x", "ctrl-x", SelectCB, 0, NULL },
	{ 0, "Forward", &xmPushButtonGadgetClass, 'F', 
	  "Ctrl<Key>f", "ctrl-f", ForwardCB, 0, NULL },
	{ 0, "Back", &xmPushButtonGadgetClass, 'B', 
	  "Ctrl<Key>b", "ctrl-b", BackCB, 0, NULL },
	NULL_MENU,
};

static MenuItem help[] = {
	{ 0, "About...", &xmPushButtonGadgetClass, 'A', 
	  0,0, AboutCB, 0, NULL },
	NULL_MENU,
};



/*
// because the data structures cant be initialized with this in 
// them
*/

static void stuffData(MenuItem *menus, void *data)
{
	int i;
	for(i = 0; menus[i].label != NULL; i++) {
		menus[i].callback_data = data;
	}
}


int
keyboardH(void *user_data, KeySym sym, iUCode result)
{
	MainWindow *mw;
	Mode *mode;
	Page *page;
	int return_value = FALSE;
	
	if(result) {
		mw = (MainWindow *) user_data;
	
		// bogus
		// this is a bit gross, but I need to make sure that the deleteCB
		// does exactly the same thing as hitting back space
		if (result=='\b') {
			deleteCB(0,mw,0);
			return_value = TRUE;
		} else {
			mode = mw->document->getCurrentMode();
			if(mode) {
				page = mode->getCurrentPage();
				if(page) {
					return_value = page->pressKey(result);
				}
			}
		}
	}

	return return_value;
}

static void 
chooseSub(void *user_data, int changed)
{
	MainWindow * mw;
	mw = (MainWindow *) user_data;
	mw->choose(changed);
}

MainWindow::MainWindow(Application *input_app,
					   ModeType input_would_be_mode)
	: app(input_app),
	  title(input_app->app_name), 
	  scroll_window(0), 
	  open_dialog(0), page_select(0), save_as_dialog(0),
	  would_be_mode(input_would_be_mode),
	  view_canvas(0), mag_exp(0),
	  last_width(0), last_height(0),
	  last_x(0), last_y(0),
	  mag_option(0), 
#ifdef RAF_INTERNAL
	  type_option(0), 
	  classify_menu(0),
#endif
	  dirty_mark(0),
	  chooser(0),
	  menu_data(this),
	  text_disabled(FALSE),
	  pref_menu(0)
{
	this->sm = new SelectionManager(this);
	this->document = new Document(this);
	
	this->chooser = new ChooseDialog(this, this->app->app_shell, chooseSub, this);
	this->type_pref_view = new TypePrefDialog(this, this->app->app_shell);
	this->entity_view = new EntityDialog(this, this->app->app_shell);

	this->uniedit = new UniEdit(this, this->app->app_shell, keyboardH, this);
	this->nihon_edit = new NihonEdit(this, this->app->app_shell, keyboardH, this);
	
	this->american = initAmerican();
	this->arabic = initArabic();
	this->russian = initRussian();
	this->english = initEnglish();
	this->german = initGerman();
	this->french = initFrench();
	this->greek = initGreek();
	this->micr = initMicr();
	this->swedish=initSwedish();
	this->spanish=initSpanish();
	this->italian=initItalian();
	this->norw=initNorw();
	this->kb = new KeyBoard(app->app_shell, 
							american,
							keyboardH, 
							this);
}

MainWindow::~MainWindow()
{
	if(this->document)
		delete this->document;

	delete this->chooser;
	delete this->entity_view;
	delete this->type_pref_view;

	delete this->uniedit;
	delete this->nihon_edit;
	delete this->kb;
	delete this->view_canvas;
	delete this->sm;

	delete this->american;
	delete this->arabic;
	delete this->russian;
	delete this->swedish;
	delete this->english;
	delete this->german;
	delete this->french;
	delete this->greek;
	delete this->micr;
	delete this->spanish;
	delete this->norw;
	delete this->italian;
	if (this->page_select)
		delete this->page_select;
#ifdef RAF_INTERNAL
	/* ok, this is a bit gross to be deleting this stuff here */
	int i;
	for (i=0; settype[i].label; i++)
		delete (Tuple *)settype[i].callback_data;
#endif
}

void
MainWindow::setTitle(const aString& input_title)
{
	XArg args;
	
	if(this->title != input_title && input_title.length()){
		this->title = input_title;
		args.reset();
		args.setArg(XmNtitle,    (char *) this->title.string());
		args.setArg(XmNiconName, (char *) this->title.string());
		args.setValues(this->app->app_shell);
	}
}

void
MainWindow::changeMode(ModeType input_mode)
{
	XArg args;
	
	args.reset();
	args.setArg(XmNmenuHistory, modes[input_mode].result);
	args.setValues(this->modes_option);
	
	XtManageChild(this->modes_option);
	
	this->document->changeMode(input_mode);
}

void
MainWindow::standardTitle()
{
	aString new_title;
	FileName fn;
	
	Page *page=0;
	Mode *current_mode = this->document->getCurrentMode();
	if(current_mode != 0)
		page = current_mode->getCurrentPage();
	
	
	fn = this->document->file_name;
	
	XArg args;
	/* set the dirt */
	if (1) {
		aString dirt_mark;
		if(this->document->dirtyP())
			dirt_mark = "**";
		else
			dirt_mark = "--";
		
		XmString xm_string;
		
		xm_string = XmStringCreateLocalized((char *) dirt_mark.string());
		args.setArg(XmNlabelString, xm_string);
		args.setValues(this->dirty_mark);
		XmStringFree(xm_string);
	}
	
#ifdef RAF_INTERNAL
	/* set the type */
	if(1) {
		aString type;
		type = this->document->getMailType();
		int i;
		for(i = 0; types[i].label != 0 && type != types[i].label; i++)
			;
		args.reset();	 
		if(types[i].label){
			args.setArg(XmNmenuHistory, types[i].result);
		} else {
			args.setArg(XmNmenuHistory, types[0].result);
		}
		args.setValues(this->type_option);
	}
#endif
	
	/* page number is missing */
	
	if(fn == "")
		new_title = "Untitled";
	else
		new_title = fn.filename();

	if(page) {
		aString page_name = page->getName();
		new_title = new_title + " (" + page_name + ")";
	}
	
	this->setTitle(new_title);
}

void
MainWindow::refresh_dialogs()
{
	if (this->page_select && this->page_select->managedP()) {
		Mode *current_mode = this->document->getCurrentMode();
		int list_size,i;
		XmStringTable list;
		
		current_mode->makePageList(&list,&list_size);
		this->page_select->show(list,list_size);
		
		for(i = 0; i < list_size; i++)
			XmStringFree(list[i]);
		smfree(list);
		
	}

	if (this->entity_view)
		this->entity_view->updateTypes();
	if (this->type_pref_view)
		this->type_pref_view->update(); 		
}

GC grayGC(Widget w)
{
	GC gray_gc;

	XGCValues values;
	int mask;
		
	values.foreground = getColorByName(w, GRAY_COLOR);
	values.background = getColorByName(w, "white");
	mask = GCForeground | GCBackground;
	gray_gc = XCreateGC(XtDisplay(w), 
						XtWindow(w),
						mask,
						&values);
	return gray_gc;
}


Widget
MainWindow::createMenus(Widget main_window)
{
	XArg args;
	iTypeIndex type;
	Widget    menu_bar;  
	
	args.reset();
	menu_bar = XmCreateMenuBar (main_window, "menu_bar", 
								args.getArgs(), args.getSize()); 
	XtManageChild (menu_bar);
	
	stuffData(files, this);
	this->file_menu = 
		BuildMenu(menu_bar, XmMENU_PULLDOWN, "File", 'F', files);
	XtManageChild(this->file_menu);
	
	stuffData(edits, this);
	this->edit_menu = 
		BuildMenu(menu_bar, XmMENU_PULLDOWN, "Edit", 'E', edits);
	XtManageChild(this->edit_menu);
	
	stuffData(displays, this);
	this->display_menu = 
		BuildMenu(menu_bar, XmMENU_PULLDOWN, "Display", 'D', displays);

	XtManageChild(this->display_menu);
	
	stuffData(prefs, this);
	this->pref_menu = 
		BuildMenu(menu_bar, XmMENU_PULLDOWN, "Preferences", 'P', prefs);


#ifdef DO_POPUPS
	args.reset();
	args.setArg(XmNset, TRUE);
	args.setArg(XmNvisibleWhenOff, TRUE);
	args.setValues(prefs[prefs_popup].result);
#endif
	args.reset();
	args.setArg(XmNset, this->text_disabled == FALSE);
	args.setArg(XmNvisibleWhenOff, TRUE);
	args.setValues(prefs[prefs_text].result);

	XtManageChild(this->pref_menu);

	stuffData(pages, this);
	this->page_menu = 
		BuildMenu(menu_bar, XmMENU_PULLDOWN, "Page", 'P', pages);
	XtManageChild(this->page_menu);
	
	stuffData(entities, this);
	this->entity_menu = 
		BuildMenu(menu_bar, XmMENU_PULLDOWN, "Entity", 'n', entities);
	XtManageChild(this->entity_menu);
	
#if RAF_INTERNAL
	stuffData(editPitch, this);
	this->pbb_menu = 
		BuildMenu(menu_bar, XmMENU_PULLDOWN, "PBB", 'B', editPitch);
	XtManageChild(this->pbb_menu);
	
	stuffData(pitches, this);
	this->pitch_menu = 
		BuildMenu(menu_bar, XmMENU_PULLDOWN, "Pitch", 'i', pitches);
	XtManageChild(this->pitch_menu);

	int i;
	for (i=0; settype[i].label; i++) {
		i_NameToType(settype[i].label,&type);
		settype[i].callback_data=new Tuple((void *)this, (void *)type);
	}
	this->settype_menu=
		BuildMenu(menu_bar,XmMENU_PULLDOWN,"Set Type",0,settype);
	XtManageChild(this->settype_menu);

#endif
	
	stuffData(help, this);
	this->help_menu = 
		BuildMenu(menu_bar, XmMENU_PULLDOWN, "Help", 'H', help);
	XtManageChild(this->help_menu);
	
	args.reset();
	args.setArg(XmNmenuHelpWidget, this->help_menu);
	args.setValues(menu_bar);
	
	return menu_bar;
}   

Widget
MainWindow::createControls(Widget main_window)
{
	XArg args;
	XmString foo;
	
	args.reset();
	args.setArg(XmNorientation, XmHORIZONTAL);
	//args.setArg(XmNpacking, XmPACK_COLUMN);
	//args.setArg(XmNnumColumns, 2);
	this->control_panel =
		XmCreateRowColumn(main_window, "controls",
						  args.getArgs(), args.getSize());
	
	args.reset();
	foo = XmStringCreateLocalized("--");
	args.setArg(XmNlabelString, foo);
	this->dirty_mark = XmCreateLabel(this->control_panel, "label",
									 args.getArgs(),
									 args.getSize());
	XmStringFree(foo);
	
	XtManageChild(this->dirty_mark);
	
#ifdef RAF_INTERNAL
	stuffData(types, this);
	this->type_option = 
		BuildMenu(this->control_panel, XmMENU_OPTION, "Type", 'T', types);
	args.reset();
#ifndef TRAVERSE_ON
	args.setArg(XmNtraversalOn, FALSE);
#endif
	args.setValues(this->type_option);
#endif
	
	mags[0].callback_data = this->menu_data.eighth_data;
	mags[1].callback_data = this->menu_data.quarter_data;
	mags[2].callback_data = this->menu_data.half_data;
	mags[3].callback_data = this->menu_data.normal_data;
	mags[4].callback_data = this->menu_data.double_data;
	mags[5].callback_data = this->menu_data.quadruple_data;
	mags[6].callback_data = this->menu_data.octuple_data;
#ifdef RAF_INTERNAL
	mags[7].callback_data = this->menu_data.sixteen_data;
#endif
	//stuffData(mags, this);
	
	this->mag_option = 
		BuildMenu(this->control_panel, XmMENU_OPTION, "Mag", 'a', mags);
	args.reset();
	args.setArg(XmNmenuHistory, mags[3].result);
#ifndef TRAVERSE_ON
	args.setArg(XmNtraversalOn, FALSE);
#endif
	args.setValues(this->mag_option);
	
	//stuffData(modes, this);

	int i = 0;
	modes[i++].callback_data = this->menu_data.image_data;
	modes[i++].callback_data = this->menu_data.ooc_data;
	modes[i++].callback_data = this->menu_data.flag_data;
	modes[i++].callback_data = this->menu_data.tree_data;
	modes[i++].callback_data = this->menu_data.rte_data;
	this->modes_option = 
		BuildMenu(this->control_panel, XmMENU_OPTION, "Modes", 'M', modes);
	args.reset();
	args.setArg(XmNmenuHistory, modes[this->would_be_mode].result);
#ifndef TRAVERSE_ON
	args.setArg(XmNtraversalOn, FALSE);
#endif
	args.setValues(this->modes_option);
	
	stuffData(keyboards, this);
	this->keyboard_option = 
		BuildMenu(this->control_panel, XmMENU_OPTION, "Keyboard", 'K', keyboards);   
	args.reset();
#ifndef TRAVERSE_ON
	args.setArg(XmNtraversalOn, FALSE);
#endif
	args.setValues(this->keyboard_option);
	
#ifdef RAF_INTERNAL   
	XtManageChild(this->type_option);
#endif
	XtManageChild(this->mag_option);
	XtManageChild(this->modes_option);
	XtManageChild(this->keyboard_option);
	XtManageChild(this->control_panel);
	
	return this->control_panel;
}
void MainWindow::printsb(Widget sb){
	int value;
	int min,max,slider_size;
	XtVaGetValues(sb,
		XmNmaximum,&max,
		XmNminimum, &min,
		XmNvalue, &value,
		XmNsliderSize,&slider_size,
		NULL);
	printf("The scrollbar value is %d\n",(int)value);
	printf("XmNminimum is %d\n",min);
	printf("XmNmaximum is %d\n", max );
	printf("XmNsliderSize is %d\n",slider_size);

}

void scrolled_hCB(Widget sb, MainWindow *mw, XmScrollBarCallbackStruct *cbs)
{
	mw->view_canvas->pix_hoffset = cbs->value;
	
	mw->redrawMode();

}
void scrolled_vCB(Widget sb, MainWindow *mw, XmScrollBarCallbackStruct *cbs)
{
	
	mw->view_canvas->pix_voffset = cbs->value;
	
	mw->redrawMode();
}

void
MainWindow::createApplication ()
{
	Widget parent = this->app->app_shell;
	Widget page;    
	Widget swindow;  
	
	XArg args;
	
	args.reset();
	args.setArg(XmNwidth, WINDOW_WIDTH);
	args.setArg(XmNheight,WINDOW_HEIGHT);
	this->main_window = XmCreateMainWindow (parent, "main1", 
									  args.getArgs(), args.getSize());
	XtManageChild (main_window);
	
	this->menu_bar = this->createMenus(main_window);

	/*  Create Page MainWindow and ScrolledWindow in Page.*/

	args.reset();
	Widget rc;
	rc = XmCreatePanedWindow(main_window, "rc", 
							 args.getArgs(), args.getSize());
	XtManageChild(rc);
	
	args.reset();
	args.setArg(XmNmarginWidth, 2); 
	args.setArg(XmNmarginHeight, 2); 
	args.setArg(XmNshadowThickness, 1); 
	args.setArg(XmNshadowType, XmSHADOW_OUT); 
	page = XmCreateFrame (rc, "page", 
						   args.getArgs(), args.getSize());
	XtManageChild (page);
	this->current_frame = page;
	  
	args.reset();
	args.setArg(XmNscrollBarDisplayPolicy, XmSTATIC); 
	args.setArg(XmNscrollingPolicy, XmAPPLICATION_DEFINED); 
	swindow = XmCreateScrolledWindow (page, "swindow", 
									  args.getArgs(), args.getSize());
	XtManageChild (swindow);
	
	/*  Set MainWindow areas and add tab groups */
	
	this->control_panel = this->createControls(main_window);
	
	XmMainWindowSetAreas (main_window, menu_bar, this->control_panel, 
						  NULL, NULL, rc);
	args.reset();
	args.setArg(XmNorientation, XmHORIZONTAL); 
	args.setArg(XmNmaximum, WINDOW_WIDTH); 
	args.setArg(XmNminimum, 0); 
	hsb = XmCreateScrollBar(swindow,"hsb",
		args.getArgs(), args.getSize());
		
	args.reset();
	args.setArg(XmNorientation, XmVERTICAL);
	args.setArg(XmNmaximum, WINDOW_HEIGHT);
	args.setArg(XmNminimum, 0); 
	args.setArg(XmNprocessingDirection,XmMAX_ON_BOTTOM);
	vsb = XmCreateScrollBar(swindow, "vsb",
		args.getArgs(), args.getSize());
	
	args.reset();
	args.setArg(XmNincrement, 10);
	args.setArg(XmNpageIncrement, 10);
	XtSetValues(hsb,args.getArgs(),args.getSize());
	XtSetValues(vsb,args.getArgs(),args.getSize());
	XtManageChild(vsb);
	XtManageChild(hsb);
	
	if (hsb)
		XmAddTabGroup (hsb);
	if (vsb)
		XmAddTabGroup (vsb);
	
	XtAddEventHandler(parent,
					  StructureNotifyMask, 
					  False, 
					  (XtEventHandler) ConfigureEH, 
					  this);
	

	this->scroll_window = swindow;
	this->view_canvas = 
		new ViewCanvas(this->scroll_window, 0,
					   WINDOW_HEIGHT, WINDOW_WIDTH);

				
	XtAddCallback(vsb, XmNvalueChangedCallback,(XtCallbackProc) scrolled_vCB, (XtPointer)this);
	XtAddCallback(hsb, XmNvalueChangedCallback,(XtCallbackProc) scrolled_hCB, (XtPointer)this);
	XtAddCallback(vsb, XmNdragCallback, (XtCallbackProc)scrolled_vCB, (XtPointer)this);
	XtAddCallback(hsb, XmNdragCallback, (XtCallbackProc)scrolled_hCB, (XtPointer)this);
	XtAddCallback(vsb, XmNpageIncrementCallback, (XtCallbackProc)scrolled_vCB, (XtPointer)this);
	XtAddCallback(hsb, XmNpageIncrementCallback, (XtCallbackProc)scrolled_hCB, (XtPointer)this);
	XtAddCallback(vsb, XmNpageDecrementCallback, (XtCallbackProc)scrolled_vCB, (XtPointer)this);
	XtAddCallback(hsb, XmNpageDecrementCallback, (XtCallbackProc)scrolled_hCB, (XtPointer)this);
#if 0
	static char defaultTranslations[] = "<Key>Next: PageDownOrRight(0)";
	XtTranslations trans_table = XtParseTranslationTable(defaultTranslations);
	XtAugmentTranslations(vsb,trans_table);
#endif
	this->setSensitive();
}

void
MainWindow::preRefreshItems(DLList *list)
{
	Ref r;
	iEntityPtr entity;
	for(r = dllistFirst(list); r; dllistNext(list, &r)){
		entity = (iEntityPtr) dllistContents(list, r);
		i_DoCallBack(mUnSelect, entity);
	}
}

void
MainWindow::postRefreshItems(DLList *list)
{
	Ref r;
	iEntityPtr entity;
	for(r = dllistFirst(list); r; dllistNext(list, &r)){
		entity = (iEntityPtr) dllistContents(list, r);
		i_DoCallBack(mSelect, entity);
	}
}

void
MainWindow::setMag(int exp)
{
	int old_mag = this->mag_exp;
	this->mag_exp = exp;
	Mode *current_mode = this->document->getCurrentMode();
	if(current_mode) {
		current_mode->setMag(exp, old_mag);
		this->standardTitle();
	}
}

void
MainWindow::nextEntity()
{
	Mode *current_mode = this->document->getCurrentMode();
	
	if(current_mode) {
		Page *current_page = current_mode->getCurrentPage();
		if(current_page) {
			current_page->nextEntity();
			current_page->track(CENTER);
		}
	}
}


void
MainWindow::prevEntity()
{
	Mode *current_mode = this->document->getCurrentMode();
	
	if(current_mode) {
		Page *current_page = current_mode->getCurrentPage();
		if(current_page) {
			current_page->prevEntity();
			current_page->track(CENTER);
		}
	}
}


void
MainWindow::firstEntity(EntityCheckFunc checker)
{
	Mode *current_mode = this->document->getCurrentMode();
	
	if(current_mode) {
		Page *current_page = current_mode->getCurrentPage();
		if (current_page){
			current_page->firstEntity(checker);
			current_page->track(CENTER);
		}
	}
}

void
MainWindow::nextEntity(EntityCheckFunc checker)
{
	Mode *current_mode = this->document->getCurrentMode();
	
	if(current_mode) {
		Page *current_page = current_mode->getCurrentPage();
		if(current_page) {
			current_page->nextEntity(checker);
			current_page->track(CENTER);
		}
	}
}

void
MainWindow::prevEntity(EntityCheckFunc checker)
{
	Mode *current_mode = this->document->getCurrentMode();
	
	if(current_mode) {
		Page *current_page = current_mode->getCurrentPage();
		if(current_page) {
			current_page->prevEntity(checker);
			current_page->track(CENTER);
		}
	}
}


void
MainWindow::upEntity()
{
	Mode *current_mode = this->document->getCurrentMode();
	
	if(current_mode) {
		Page *current_page = current_mode->getCurrentPage();
		if(current_page ) {
			current_page->upEntity();
			current_page->track(CENTER);
		}
	}
}

void
MainWindow::downEntity()
{
	Mode *current_mode = this->document->getCurrentMode();
	
	if(current_mode) {
		Page *current_page = current_mode->getCurrentPage();
		if(current_page) {
			current_page->downEntity();
			current_page->track(CENTER);
		}
	}
}

void
MainWindow::findNextFlagEntity()
{
	Mode *current_mode = this->document->getCurrentMode();
	
	if(current_mode) {
		Page *current_page = current_mode->getCurrentPage();
		if(current_page) {
			current_page->findNextFlagEntity();
			current_page->track(CENTER);
		}
	}
}

void
MainWindow::choose(int buttons_changed)
{
	int i;
	
	for(i = 0; i < NUM_MODE_TYPES; i++)
		if(document->modes[i])
			document->modes[i]->choose(buttons_changed);
}

void
MainWindow::setSensitive()
{
	int file_loaded = this->document->root_entity != 0;
	int always = 1;
	iEntityPtr entity = this->sm->getSoleItem();
	iEntityPtr from_entity;
	iEntityPtr to_entity;
	int from_index;
	int to_index;
	this->sm->getRange(from_entity, from_index, to_entity, to_index);
	int range_selected = from_entity != 0;
	int one_selected = entity != 0;
	int multi_selected = dllistLength(this->sm->getItems()) > 1;
	int some_selected = one_selected || multi_selected;
	Mode *mode = this->document->getCurrentMode();
	ModeType mode_type = mode ? mode->getMode() : vtRTE;


	int num_pages = mode ? mode->countPages() : 0;
	int show_page_menu = mode && (num_pages >1 || (num_pages > 0 && (mode_type == vtOOC || mode_type == vtFLAG)));
#ifdef RAF_INTERNAL
	iDAFSError err;
	void *foo;
	iBox box;

	int pitch_ops;
	pandp_t *pp;
	err = i_GetUserDataProp(entity, "pandp", &foo,0,f_SwapPandP);
	pp = (pandp_t *) foo;
	
	iTypeIndex type;
	i_NameToType("LineSeg",&type);	
	if(entity && 
	   ((i_GetEntityType(entity) == iLineT) || (i_GetEntityType(entity)==type))&& 
	   !bbNullP((box = i_GetEnclosingBox(entity))) && (pp) != 0) 
		pitch_ops = 1;
	else
		pitch_ops = 0;
#endif
	

	XtSetSensitive(this->mag_option, mode_type != vtRTE && mode_type != vtTREE);

#ifdef RAF_INTERNAL
	XtSetSensitive(this->type_option, file_loaded);
	XtSetSensitive(this->pitch_menu, pitch_ops);
#endif
	XtSetSensitive(this->modes_option, always);
	XtSetSensitive(this->file_menu, always);
	XtSetSensitive(this->display_menu, always);
	XtSetSensitive(this->pref_menu, always);
	XtSetSensitive(this->page_menu, show_page_menu);
	XtSetSensitive(this->entity_menu, file_loaded);

	XtSetSensitive(edits[edit_delete].result, one_selected || range_selected);

#ifdef RAF_INTERNAL
    XtSetSensitive(edits[edit_join].result, multi_selected);
    XtSetSensitive(edits[edit_split].result, one_selected);
	XtSetSensitive(edits[edit_del_parent].result, some_selected);
	XtSetSensitive(edits[edit_new_parent].result, some_selected);
#endif
	XtSetSensitive(edits[edit_kill_word].result, some_selected && (mode_type != vtRTE) && (mode_type != vtIMAGE || this->text_disabled == FALSE));


#ifdef RAF_INTERNAL
	XtSetSensitive(edits[edit_left].result, 
				   file_loaded);
	XtSetSensitive(edits[edit_right].result, 
				   file_loaded);
	XtSetSensitive(edits[edit_up].result, 
				   file_loaded && (mode_type!=vtIMAGE) 
				   );
	XtSetSensitive(edits[edit_down].result, 
				   file_loaded && (mode_type!=vtIMAGE) 
				   );
#else
	XtSetSensitive(edits[edit_left].result, file_loaded);
	XtSetSensitive(edits[edit_right].result, file_loaded);
	XtSetSensitive(edits[edit_up].result, file_loaded && (mode_type!=vtIMAGE));
	XtSetSensitive(edits[edit_down].result, file_loaded && (mode_type!=vtIMAGE));
#endif

	XtSetSensitive(edits[edit_flag].result, file_loaded && (mode_type == vtFLAG) );

	XtSetSensitive(edits[edit_new].result,one_selected &&((mode_type == vtTREE) || (mode_type == vtIMAGE)));
	XtSetSensitive(edits[edit_uni].result, file_loaded && (mode_type != vtTREE));
	XtSetSensitive(edits[edit_nihon].result, file_loaded);
	// bogus up and down missing -- should depend onv iew
	
	XtSetSensitive(entities[entity_center].result, one_selected || range_selected);
	XtSetSensitive(entities[entity_rotate_image].result, mode_type==vtIMAGE);
#if RAF_INTERNAL
	XtSetSensitive(entities[entity_recognize].result, one_selected);
	XtSetSensitive(entities[entity_deskew].result, mode_type==vtIMAGE);

	XtSetSensitive(entities[entity_sort].result, one_selected && mode_type==vtIMAGE);	
#endif

	XtSetSensitive(entities[entity_grow_horz].result, one_selected && mode_type==vtIMAGE);
	XtSetSensitive(entities[entity_shrink_horz].result, one_selected && mode_type==vtIMAGE);
	XtSetSensitive(entities[entity_move_right].result, one_selected && mode_type==vtIMAGE);
	XtSetSensitive(entities[entity_move_left].result, one_selected && mode_type==vtIMAGE);
	XtSetSensitive(entities[entity_grow_vert].result, one_selected && mode_type==vtIMAGE);
	XtSetSensitive(entities[entity_shrink_vert].result, one_selected && mode_type==vtIMAGE);
	XtSetSensitive(entities[entity_move_up].result, one_selected && mode_type==vtIMAGE);
	XtSetSensitive(entities[entity_move_down].result, one_selected && mode_type==vtIMAGE);

	XtSetSensitive(files[file_new_index].result, always);
	XtSetSensitive(files[file_open_index].result, always);
	XtSetSensitive(files[file_save_index].result, file_loaded);
	XtSetSensitive(files[file_saveas_index].result, file_loaded);
	XtSetSensitive(files[file_revert_index].result, file_loaded);
	XtSetSensitive(files[file_next].result, file_loaded);
	XtSetSensitive(files[file_prev].result, file_loaded);
	XtSetSensitive(files[file_page_up].result, file_loaded);
	XtSetSensitive(files[file_page_down].result, file_loaded);
#if RAF_INTERNAL
	XtSetSensitive(files[file_reinit].result, always);
#endif
	XtSetSensitive(files[file_quit].result, always);
	
	
	XtSetSensitive(pages[page_select_page].result, mode && mode->countPages()>1);
	XtSetSensitive(pages[page_forward].result, show_page_menu);
	XtSetSensitive(pages[page_backward].result, show_page_menu);

	XtSetSensitive(prefs[prefs_text].result, mode_type == vtIMAGE);
}

#if 0
#ifdef STEVE
void
MainWindow::SetWatchCursor(Widget w)
{
	static Cursor watch = NULL;
	
	if(!watch)
		watch = XCreateFontCursor(XtDisplay(w),XC_watch);
	
	XDefineCursor(XtDisplay(w),XtWindow(w),watch);
	XmUpdateDisplay(w);
}

void
MainWindow::ResetCursor(Widget w)
{
	XUndefineCursor(XtDisplay(w),XtWindow(w));
	XmUpdateDisplay(w);
}
#endif
#endif

void
MainWindow::deleteSelection(int range)
{
	if(range) {
		iEntityPtr from_entity, to_entity;
		int from_index, to_index;

		this->sm->getRange(from_entity, from_index, to_entity, to_index);

		if(!from_entity) return;
		this->sm->_clear();

		i_Edit(&from_entity, (int4*) &from_index, &to_entity, (int4*) &to_index, '\b');
		this->sm->setRange(from_entity, from_index, from_entity, from_index);

		this->document->setDirty();

	} else {
		const DLList *selection=this->sm->getItems();
		Ref r;
		iEntityPtr entity;
		int dirty=FALSE;

		// here we are avoiding deleting an entity twice by unselecting
		// those guys whose parents are alse selected
		// bogus -- my personal feeling is that we should not
		// be directly affecting the selection
		// rather we should be affecting a copy of the list
		//
		// why -- because we really do want to update the
		// display of all the children
		// or maybe im wrong because unselect doesnt appear
		// to change thier display so we are relying on
		// the delete callback
		for(r = dllistFirst(selection); r; dllistNext(selection, &r)){
			entity = (iEntityPtr) dllistContents(selection, r);
			if (entity!=this->document->root_entity)
				if (this->sm->unselectChildren(entity)) {
								/* unfortunately the selection list might change */
					selection=this->sm->getItems();
					r=dllistFirst(selection);
				}
		}

		selection=this->sm->getItems();
		DLList scratch_list;
		dllistCtor(&scratch_list);
		for(r = dllistFirst(selection); r; dllistNext(selection, &r)){
			entity = (iEntityPtr) dllistContents(selection, r);
			dllistAppend(&scratch_list, entity);
		}

		for(r = dllistFirst(&scratch_list); r; dllistNext(&scratch_list, &r)){
			entity = (iEntityPtr) dllistContents(&scratch_list, r);

			if(entity == this->document->root_entity) {
				uiWarnUser("You can't delete the root entity.", "OK");
			} else {
                                // bogus -- we are counting on clearing selection before
                                // deleting

                                // why do we explicit do the unselecting
                                // is it because we want to avoid call backs sometimes
				this->sm->unselect(entity);
				i_DisposeEntity(entity);
				dirty=TRUE;
			}
		}
		dllistDtor(&scratch_list);
		if (dirty) this->document->setDirty();
	}

	this->setSensitive();
}

void
MainWindow::updateMode(iEntityPtr entity)
{
	if(this->entity_view)
		this->entity_view->updateEntity(entity);
}

void
MainWindow::setModeSelection(iEntityPtr en, int index)
{
	if(this->entity_view)
		this->entity_view->setEntity(en,index);
}

void
MainWindow::setModeCursor(iEntityPtr en, int in)
{
	if(this->entity_view)
		this->entity_view->setCursor(en, in);
}


void
MainWindow::centerOnSelected()
{
	Mode *mode = this->document->getCurrentMode();
	iEntityPtr sole_selection = this->sm->getSoleItem();
	iEntityPtr from_entity = 0;
	int from_index = 0;
	iEntityPtr to_entity = 0;
	int to_index = 0;
	this->sm->getRange(from_entity, from_index, to_entity, to_index);
	if(mode) {
		if(sole_selection)
			mode->centerOnSelected(sole_selection);
		else if((this->sm->getState() == scITEM || 
			 this->sm->getState() == scRANGE) && 
			from_entity)
		  mode->centerOnSelected(from_entity);
	}
}


void
MainWindow::redrawImageMode(void)
{
	Mode *mode = 0;
	Page *page = 0;


	if (this->document->current_mode==vtIMAGE){
		mode = (Mode *) this->document->getCurrentMode();
		if(mode)
		    page = (Page *) mode->getCurrentPage();
		if(page) {
		    iLBox b;

		    Widget sw = mode->scrolled_window;
		    b = visibleRectangle(sw);
		    page->drawRegion(b);
		}
	}
}
void
MainWindow::redrawMode(void)
{
	Mode *mode = 0;
	Page *page = 0;


	mode = (Mode *) this->document->getCurrentMode();
	if(mode)
		page = (Page *) mode->getCurrentPage();
	if(page) {
		iLBox b;

		Widget sw = mode->scrolled_window;
		b = visibleRectangle(sw);
		page->drawRegion(b);
	}
}

void 
MainWindow::textToggle(void)
{
	this->text_disabled = !this->text_disabled;
	this->setSensitive();
	this->redrawImageMode();
}

void 
MainWindow::clearMode(iEntityPtr entity)
{
	if(this->entity_view)
		this->entity_view->clearEntity(entity);
}

void
MainWindow::setDir(aString fn)
{
	this->document->current_directory = fn;
}


int
MainWindow::saveAs(aString fn)
{
	int success;

	// so that it will let us save even if we haven't changed
	this->document->setDirty();
	this->document->file_name = fn;
	// directory resets
	this->document->current_directory = this->document->file_name.directory();
	this->document->setSaveType(this->document->would_be_savetype);
	success = this->document->saveFile();
	// even if it fails, leave the dirty mark since 
	// we have changed the filename
	this->standardTitle();
	
	return success;
}
		
