
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */


#ifndef _MainWindow_h_
#define _MainWindow_h_

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <Xm/Xm.h>

#include "dllist.h"

#include "illum.h"

#include "MenuData.h"

#define WINDOW_HEIGHT 700
#define WINDOW_WIDTH 700

class MainWindowMenuData {
  private:
	Tuple *eighth_data;
	Tuple *quarter_data;
	Tuple *half_data;
	Tuple *normal_data;
	Tuple *double_data;
	Tuple *quadruple_data;
	Tuple *octuple_data;
	Tuple *sixteen_data;
		
	Tuple *image_data;
	Tuple *ooc_data;
	Tuple *flag_data;
	Tuple *hier_data;
	Tuple *tree_data;
#ifdef RAF_INTERNAL
	Tuple *ideal_data;
	Tuple *damaged_data;
	Tuple *ambiguous_data;
	Tuple *badbox_data;
	Tuple *noise_data;
#endif
  Tuple *rte_data;


	friend class MainWindow;
	MainWindowMenuData(MainWindow *parent);
	~MainWindowMenuData();
};
/*
// concentration of nearly all the widgets in the application
// however, not all and  that is unfortunate
// unfortunate to that many of the menu functions
// call a method in the MainWindow
// and yet many do not
*/
class MainWindow {
public:
	Application *app; // back_pointer

	SelectionManager *sm;
	MainWindowMenuData menu_data;
	int mag_exp;
	Document *document;


	aString title;
	
	KeyMap *american;
	KeyMap *arabic;
	KeyMap *russian;
	KeyMap *english;
	KeyMap *german;
	KeyMap *french;
	KeyMap *greek;
	KeyMap *micr;
	KeyMap *swedish;
	KeyMap *spanish;
	KeyMap *italian;
	KeyMap *norw;
	KeyBoard *kb;
	
	ModeType would_be_mode;
	int text_disabled;
	
	/* delete this before app_shell */
	ViewCanvas *view_canvas;
	
	/* should only have to destroy app shell in Application */
	Widget scroll_window;
	Widget hsb;
	Widget vsb;
	
	/* hurley need to unmanage the current frame in order to go to 
		RTE mode */
	Widget current_frame;

	SaveAsDialog *save_as_dialog;
	OpenDialog *open_dialog;

	ListDialog *page_select;

	ChooseDialog *chooser;
	TypePrefDialog *type_pref_view;
	EntityDialog *entity_view;
	UniEdit *uniedit;
	NihonEdit *nihon_edit;
	
	Widget control_panel;
	
	// need to keep a pointer to the main_window and menu_bar so that
	// we can call set MainWindowAreas for the RTE view
	Widget main_window;
	Widget menu_bar;
	
	Widget mag_option;
#ifdef RAF_INTERNAL
	Widget type_option;
	Widget pbb_menu;
#endif
	Widget modes_option;
	Widget keyboard_option;
	Widget file_menu;
	Widget edit_menu;
	Widget display_menu;
	Widget pref_menu;
	Widget page_menu;
	Widget pitch_menu;
	Widget entity_menu;
	Widget help_menu;
#ifdef RAF_INTERNAL
	Widget classify_menu;
	Widget settype_menu;
#endif
	
	Widget dirty_mark;
	
	Dimension last_width;
	Dimension last_height;
	Position last_x;
	Position last_y;
	
	MainWindow(Application *app, ModeType mode);
	~MainWindow(void);
	
	Widget createMenus(Widget main_window);
	Widget createControls(Widget main_window);
	void createApplication (void);
	void standardTitle(void);
	void setTitle(const aString& title);
	void changeMode(ModeType input_mode);
	void refresh_dialogs(void);
		
	/* some of these don't make a lot of sense in here*/
	void updateMode(iEntityPtr entity);
	void clearMode(iEntityPtr entity);
	void setModeSelection(iEntityPtr, int);
	void setModeCursor(iEntityPtr, int);
	
	void centerOnSelected(void);
	void deleteSelection(int do_range);
	
	void setEntityCallbacks(iEntityPtr entity);
	void preRefreshItems(DLList *list);
	void postRefreshItems(DLList *list);

	void setMag(int exp);
	
	void prevEntity(void);
	void nextEntity(void);
	void upEntity(void);
	void downEntity(void);

	void nextEntity(EntityCheckFunc checker);
	void prevEntity(EntityCheckFunc checker);
	// why no last entity
	void firstEntity(EntityCheckFunc checker);
	// why no last entity
	void lastEntity(EntityCheckFunc checker);

	void findNextFlagEntity(void);
	void inputUnicodeCB(void);
	
	void printsb(Widget sb);
	//that was a debuger
	
	
	void setSensitive(void);
	
	void choose(int buttons_changed);
	void textToggle(void);
	void setDir(aString fn);
	int saveAs(aString fn);
	void redrawMode(void);
	void redrawImageMode(void);
	friend void scrolled_vCB(Widget sb, MainWindow *mw, XmScrollBarCallbackStruct *cbs);
	friend void scrolled_hCB(Widget sb, MainWindow *mw, XmScrollBarCallbackStruct *cbs);
};

#endif
