
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#include "rafport.h"

#include <Xm/Label.h>     
#include <Xm/SelectioB.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/List.h>
#include <Xm/RowColumn.h>
#include <Xm/TextF.h>
#include <Xm/Text.h>

#include "illum.h"

#include "smemory.h"

#include "misc/XArg.h"
#include "misc/misc.h"
#include "misc/Unicode.h"
#include "ucs.h"
#include "c-type.h"

#include "canna.h"
#include "NihonEdit.h"


NihonEdit::NihonEdit(MainWindow *input_mw,
					 Widget input_app_shell,
					 KeyBoardHandler input_key_h,
					 void *input_user_data)
	: Dialog(input_mw, input_app_shell),
	  key_h(input_key_h),
      user_data(input_user_data),
	  mode_menu(0),
	  text_input(0),
	  ok_button(0),
	  apply_button(0),
	  item_list(0),
	  waiting_for_reset(FALSE)
{
	this->old_mode_line[0] = 0;
}

NihonEdit::~NihonEdit()
{
	this->disposeItems();
}


static void do_returnCB(Widget w, void *ue, void *call)
{
	((NihonEdit *)ue)->returnCB();
}

static void do_nothingCB(Widget w, void *ue, void *call)
{
}

static void do_restartCB(Widget w, void *ue, void *call)
{
	((NihonEdit *)ue)->restartCB();
}

static void do_spaceCB(Widget w, void *ue, void *call)
{
	((NihonEdit *)ue)->spaceCB();
}

static void do_tabCB(Widget w, void *ue, void *call)
{
	((NihonEdit *)ue)->tabCB();
}

static void do_applyCB(Widget w, void *ue, void *call)
{
	((NihonEdit *)ue)->applyCB((XmFileSelectionBoxCallbackStruct *)call);
}

static void do_okCB(Widget w, void *ue, void *call)
{
	((NihonEdit *)ue)->okCB((XmFileSelectionBoxCallbackStruct *)call);
}

static void do_closeCB(Widget w, void *ue, void *call)
{
	((NihonEdit *)ue)->closeCB((XmFileSelectionBoxCallbackStruct *)call);
}

#if 0
static void do_upCB(Widget w, void *ue, void *call)
{
	((NihonEdit *)ue)->upCB((XmFileSelectionBoxCallbackStruct *)call);
}

static void do_downCB(Widget w, void *ue, void *call)
{
	((NihonEdit *)ue)->downCB((XmFileSelectionBoxCallbackStruct *)call);
}
#endif

static void 
addButton(Widget w,char *str,
		  XtCallbackProc cb, NihonEdit *ue, 
		  char *accel)
{
	Widget n;
	XmString xm_string;
	XArg args;

	args.reset();
	if (accel) {
		args.setArg(XmNaccelerator, accel);
	}
	xm_string = XmStringCreateLocalized(str);
	args.setArg(XmNlabelString, xm_string);
	args.setArg(XmNalignment, XmALIGNMENT_CENTER);
	n = XmCreatePushButton(w, str,args.getArgs(), args.getSize());
	XmStringFree(xm_string);
	XtManageChild(n);

	if (cb)
		XtAddCallback(n, XmNactivateCallback, cb, ue);
}

static void 
addCellLabel(Widget w,Widget *np,
			 char *str, XmString xm_string,
			 int align)
{
	Widget n;
	XArg args;

	args.reset();
	args.setArg(XmNlabelString, xm_string);
	args.setArg(XmNalignment, align);
	n = XmCreateLabel(w, str,args.getArgs(), args.getSize());
	XtManageChild(n);
	if (np) *np = n;
}

static void addCellLabel(Widget w,Widget *np,
						 char *str,int align)
{
	XmString xm_string;
	xm_string = XmStringCreateLocalized(str);
	addCellLabel(w,np,str, xm_string, align);
	XmStringFree(xm_string);
}

static void addCellLabel(Widget w,Widget *np,
						 iUCode *str,int align)
{
	XmString xm_string;
	XmFontList font_list = 0;
	xm_string = makeCompoundString(XtDisplay(w), str, ucslen(str), &font_list);
	addCellLabel(w,np,"bogus", xm_string, align);
	if(font_list) {
		XArg args;
		args.reset();
		args.setArg(XmNfontList, font_list);
		args.setValues(*np);
	}
	XmStringFree(xm_string);
}

extern void getJapanese(char *s, iUCode ***list, int *n);

static void
suckSpaces(iUCode **list, int count)
{
	int i;
	iUCode *s;
	iUCode buf[256];
	iUCode *p;

	for(i = 0; i < count; i++) {
		for(p = buf, s = list[i]; *s; s++)
			if(*s != ' '){
				*p++ = *s;
			}
		*p = '\0';
		ucscpy(list[i], buf);
	}
}


void
stuffChoices(NihonEdit *ue)
{
	suckSpaces(ue->item_list, ue->item_count);

	int i;
	XmString xm_string;
	XmFontList font_list;

	if (ue->item_count) {
		for (i=0; i<ue->item_count; i++) {
			xm_string=makeCompoundString(XtDisplay(ue->text_input),
										 ue->item_list[i],
		                            	 ucslen(ue->item_list[i]),&font_list);
			if(font_list) {
				XArg args;
				args.reset();
				args.setArg(XmNfontList, font_list);
				args.setValues(ue->choice_list);
			}
			XmListAddItemUnselected(ue->choice_list,xm_string,0);
			XmStringFree(xm_string);
		}
		XmListSelectPos(ue->choice_list,1,TRUE);
		ue->enableOK();
	} else {
		ue->disableOK();	
	} 	

}

static void
flushChoices(NihonEdit *ue)
{
	ue->disposeItems();
	XmListDeleteAllItems(ue->choice_list);
}

static void 
value_changed(Widget w, NihonEdit *ue, 
			  XmTextVerifyCallbackStruct *cbs)
{
	char *s;

	s = XmTextFieldGetString(w);
	flushChoices(ue);

	getJapanese(s,&ue->item_list,&ue->item_count);


	stuffChoices(ue);
	ue->updateModeLine();
}

void
processOneChar(Widget w, NihonEdit *ue, char ch)
{
	char s[2];

	flushChoices(ue);

	s[0] = ch;
	s[1] = 0;

	getJapanese(s,&ue->item_list,&ue->item_count);

	stuffChoices(ue);
	ue->updateModeLine();
}

static void 
verify_input(Widget w, NihonEdit *ne, XmTextVerifyCallbackStruct *cbs)
{
	int i,len,ok;
	char c;
	char *s = XmTextFieldGetString(w);

	if ((cbs->event != NULL) && // programmer gets to do whatever
		((cbs->text->ptr==NULL) || // no deletes
		 (cbs->startPos < strlen(s)))) // no middle inserts 
		{ 
			cbs->doit=FALSE;
			return;
		} else {

			len=cbs->text->length;

			ok = ne->waiting_for_reset == FALSE;
			for (i=0; ok && i<len; i++) {
				c=cbs->text->ptr[i];
				ok = isprint(c);
			}
			if (!ok)
				cbs->doit=FALSE;
		}

	XtFree(s);
}


static void click_twice(Widget w, NihonEdit *ue, XmListCallbackStruct *cbs)
{
	ue->insertText(cbs->item_position-1);
}

void NihonEdit::disableOK()
{
}

void NihonEdit::enableOK()
{
}

int
NihonEdit::insertText(int p)
{
	int ok;
	ok = TRUE;

	iUCode *str=this->item_list[p];
	while (ok && *str) {
		ok = key_h(this->user_data,0,*str++);
	}
	return ok;
}


void NihonEdit::disposeItems()
{
	int i;
	if (item_list) {
		for (i=0; i<item_count; i++)
			smfree(item_list[i]);
		smfree(item_list);
	}
}


void NihonEdit::closeCB(XmFileSelectionBoxCallbackStruct *call)
{
	this->hide();
}

void NihonEdit::applyCB(XmFileSelectionBoxCallbackStruct *call)
{
	this->returnCB();
}

void NihonEdit::okCB(XmFileSelectionBoxCallbackStruct *call)
{
	if(this->returnCB())
		this->hide();
}

#if 0
void NihonEdit::upCB(XmFileSelectionBoxCallbackStruct *call)
{
	int has_selected,pos_cnt,*pos_list;
	has_selected=XmListGetSelectedPos(this->choice_list,&pos_list,&pos_cnt);

	if (has_selected && pos_cnt==1 && pos_list[0]!=1) {
		XmListSelectPos(this->choice_list,pos_list[0]-1,True);
		XmListSetPos(this->choice_list,pos_list[0]-1);
		XtFree((char *)pos_list);
	}
}

void NihonEdit::downCB(XmFileSelectionBoxCallbackStruct *call)
{
	int has_selected,pos_cnt,*pos_list;
	has_selected=XmListGetSelectedPos(this->choice_list,&pos_list,&pos_cnt);

	if (has_selected && pos_cnt==1) {
		XmListSelectPos(this->choice_list,pos_list[0]+1,True);
		XmListSetPos(this->choice_list,pos_list[0]+1);
		XtFree((char *)pos_list);
	}
}
#endif

static MenuItem DummyMenu[] = {
	{ 0, "OK", &xmPushButtonGadgetClass, 0,
	  "<Key>Return",0, do_returnCB, 0, NULL },
	{ 0, "Ok", &xmPushButtonGadgetClass, 0,
	  "<Key>Delete",0, do_nothingCB, 0, NULL },
	{ 0, "Ok", &xmPushButtonGadgetClass, 0,
	  "<Key>BackSpace",0, do_nothingCB, 0, NULL },
	{ 0, "Ok", &xmPushButtonGadgetClass, 0,
	  "Ctrl<Key>u",0, do_restartCB, 0, NULL },
	{ 0, "Stop", &xmPushButtonGadgetClass, 0,
	  "<Key>space",0, do_spaceCB, 0, NULL },
	{ 0, "Foo", &xmPushButtonGadgetClass, 0,
	  "<Key>Tab",0, do_tabCB, 0, NULL },
	{ 0, "Apply", &xmPushButtonGadgetClass, 0,
	  "Ctrl<Key>C",0, do_closeCB, 0, NULL },
#if 0
	{ 0, "Up", &xmPushButtonGadgetClass, 0, 
	  "<Key>Up", 0, do_upCB, 0, NULL },
	{ 0, "Down", &xmPushButtonGadgetClass, 0, 
	  "<Key>Down", 0, do_downCB, 0, NULL },
#endif
	NULL_MENU,
};

static void 
stuffData(MenuItem *menus, void *data)
{
	int i;
	for(i = 0; menus[i].label != NULL; i++) {
		menus[i].callback_data = data;
	}
}

void 
NihonEdit::create()
{
	XArg args;
	Widget parent = this->app_shell;
	XmString xm_string;
	Widget w,choice_form,button_form,b_rowcol,i_rowcol,input_form;

	args.reset();
	args.setArg(XmNtitle, "Input Japanese");
	
	stuffData(DummyMenu, this);
	this->dialog =
		XmCreateFormDialog(parent,"NihonEdit",args.getArgs(),args.getSize());

	this->dummy_menu = 
		BuildMenu(this->dialog, XmMENU_POPUP, "Buttons", 'F', DummyMenu);

	args.reset();
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);

	args.setArg(XmNshadowType, XmSHADOW_ETCHED_IN);
	/*args.setArg(XmNleftAttachment,XmATTACH_FORM);*/
	args.setArg(XmNbottomAttachment,XmATTACH_FORM);
	button_form = XmCreateFrame(this->dialog,"frame",args.getArgs(),args.getSize());
	XtManageChild(button_form);

	args.reset();
	args.setArg(XmNshadowType, XmSHADOW_ETCHED_IN);
	args.setArg(XmNtopAttachment,XmATTACH_FORM);
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	input_form = XmCreateFrame(this->dialog,"input",args.getArgs(),args.getSize());
	XtManageChild(input_form);

	args.reset();
	/*args.setArg(XmNtopAttachment,XmATTACH_FORM);*/
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	args.setArg(XmNtopAttachment,XmATTACH_WIDGET);
	args.setArg(XmNtopWidget,input_form);
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNbottomAttachment,XmATTACH_WIDGET);
	args.setArg(XmNbottomWidget,button_form);
	choice_form = XmCreateForm(this->dialog,"choices",args.getArgs(), args.getSize());
	XtManageChild(choice_form);

	args.reset();
	args.setArg(XmNpacking,XmPACK_COLUMN);
	args.setArg(XmNnumColumns,3);
	args.setArg(XmNorientation,XmVERTICAL);
	args.setArg(XmNpacking,XmPACK_COLUMN);
	args.setArg(XmNisAligned,FALSE);
	args.setArg(XmNentryAlignment,XmALIGNMENT_CENTER);
	b_rowcol = XmCreateRowColumn(button_form,"Buttons",
				args.getArgs(),args.getSize());
	XtManageChild(b_rowcol);

	addButton(b_rowcol,"OK",do_okCB,this,0);
	addButton(b_rowcol,"Apply",do_applyCB,this,0);
	addButton(b_rowcol,"Close",do_closeCB,this,0);

	args.reset();
	xm_string = XmStringCreateLocalized("Choices");
	args.setArg(XmNresizable,False);
	args.setArg(XmNalignment, XmALIGNMENT_BEGINNING);
	args.setArg(XmNlabelString, xm_string);
	args.setArg(XmNtopAttachment,XmATTACH_FORM);
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	w = XmCreateLabel(choice_form, "name",args.getArgs(), args.getSize());
	XmStringFree(xm_string);
	XtManageChild(w);

	args.reset();
	args.setArg(XmNlistSizePolicy,XmCONSTANT);
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	args.setArg(XmNbottomAttachment,XmATTACH_FORM);
	args.setArg(XmNtopAttachment,XmATTACH_WIDGET);
	args.setArg(XmNtopWidget,w);
	args.setArg(XmNvisibleItemCount,3);
	iUCode glyph=' ';
	XmFontList font_list=0;
	xm_string = makeCompoundString(XtDisplay(choice_form),&glyph,1,&font_list);
	if (font_list)
		args.setArg(XmNfontList, font_list);
	XmStringFree(xm_string);
	
	this->choice_list = XmCreateScrolledList(choice_form,"Properties",
			args.getArgs(),args.getSize());

	XtAddCallback(this->choice_list,XmNdefaultActionCallback,click_twice,this);
	XtManageChild(this->choice_list);

	args.reset();
	args.setArg(XmNpacking,XmPACK_COLUMN);
	args.setArg(XmNnumColumns,2);
	args.setArg(XmNorientation,XmVERTICAL);
	args.setArg(XmNisAligned,FALSE);
	i_rowcol = XmCreateRowColumn(input_form,"Buttons",args.getArgs(),args.getSize());
	XtManageChild(i_rowcol);

	addCellLabel(i_rowcol,0,"Input",XmALIGNMENT_CENTER);

	args.reset();
/*	args.setArg(XmNleftAttachment,XmATTACH_WIDGET);
	args.setArg(XmNleftWidget,w);*/
	args.setArg(XmNmaxLength,20);
	args.setArg(XmNcolumns,20);
	args.setArg(XmNmarginHeight,2);
	args.setArg(XmNmarginWidth,2);
	this->text_input=XmCreateTextField(i_rowcol,"text_input",args.getArgs(),args.getSize());
	XtAddCallback(this->text_input,XmNmodifyVerifyCallback,verify_input,this);
	XtAddCallback(this->text_input,XmNvalueChangedCallback,value_changed,this);
	XtManageChild(this->text_input);

	addCellLabel(i_rowcol,0,"Mode",XmALIGNMENT_CENTER);
	iUCode buffer[64];
	nihonModeName(buffer);
	addCellLabel(i_rowcol,&this->mode_menu,buffer,XmALIGNMENT_CENTER);
	XtSetSensitive(this->dialog,TRUE);

}

void NihonEdit::show()
{
	this->Dialog::show();
	this->restartCB();
}


void
NihonEdit::updateModeLine()
{
	iUCode buffer[64];
	XmString xm_string;
	XmFontList font_list;
	XArg args;

	nihonModeName(buffer);
	if(ucscmp(buffer, this->old_mode_line) == 0)
		return;
	ucscpy(this->old_mode_line, buffer);
	xm_string = makeCompoundString(XtDisplay(this->mode_menu), 
								   buffer,
								   ucslen(buffer),
								   &font_list);
	
	args.reset();
	args.setArg(XmNlabelString, xm_string);
	if(font_list)
		args.setArg(XmNfontList, font_list);
	args.setValues(this->mode_menu);

	XmStringFree(xm_string);
}

int
NihonEdit::returnCB()
{
	int has_selected,pos_cnt,*pos_list;
	int worked = FALSE;
	has_selected=XmListGetSelectedPos(this->choice_list,&pos_list,&pos_cnt);

	if (has_selected && pos_cnt==1) {
		if(this->insertText(pos_list[0]-1))
			worked = TRUE;
	} else {
		worked = TRUE;
	}
			 
	if(worked){
		XtFree((char *)pos_list);
		processOneChar(this->dialog, this, '\n');
		this->waiting_for_reset = FALSE;
		XmTextSetString(this->text_input,"");
	}
	
	return worked;
}


void
NihonEdit::spaceCB()
{
	char *s;

	s = XmTextFieldGetString(this->text_input);
	// bogus -- must I free
	if(s == 0 || *s == '\0'){
	} else {
		this->waiting_for_reset = TRUE;
		processOneChar(this->dialog, this, ' ');
	}

	if(s)
		XtFree(s);
}

void
NihonEdit::tabCB()
{
	// here is my theory
	// check and see if there is any text
	// if there is non, then send a, send \t and send \n
	// this should move things along and result in nothing being in there

	// next, after pressing tab, check the mode line, if it is 1/2width
	// kana or fullwidth ASCII press on
	char *s;
	
	s = XmTextFieldGetString(this->text_input);
	// bogus -- must I free
	if(s == 0 || *s == '\0'){
		char t[3][3] = {"a", "\t", "\n"};
		int i, l;

		l = 3;
		for(i = 0; i < l; i++)
			getJapanese(t[i],&this->item_list,&this->item_count);
		flushChoices(this);
		this->updateModeLine();
	} else {
		processOneChar(this->dialog, this, '\t');
	}

	if(s)
		XtFree(s);
}

void
NihonEdit::restartCB()
{
	for(int i = 0; i < 2; i++){
		char s[] = "\n";

		flushChoices(this);

		getJapanese(s,&this->item_list,&this->item_count);
		this->waiting_for_reset = FALSE;
		XmTextSetString(this->text_input,"");
	}
}
