
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "illum.h"

#include <stdio.h>

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <Xm/Xm.h>

#include "rafutil.h"
#include "smemory.h"

#include "misc/color.h"
#include "misc/XArg.h"

#include "keyboard/KeyBoard.h"
#include "keyboard/KeyMap.h"

#include "text/text.h"

#include "EntityAux.h"
#include "MainWindow.h"
#include "OOCView.h"
#include "ViewCanvas.h"
#include "Document.h"

#include "CE.h"
#include "OOCFrame.h"
#include "SelectionManager.h"

#include "bb.h"
#include "mag.h"
#include "scroll.h"


// *******************************************************
// This pragma forces inclusion of debugging code
// from the named header file or header or same name
#ifdef __GNUC__
#pragma implementation
#endif


OOCFrame::OOCFrame(
	const aString& input_e_name,
	const aString& input_name, 
	View *input_parent,int type,
	int etype)
	: Frame(input_name, input_parent),
	  e_name(input_e_name),
	  cell_width(0), cell_height(0),
	  parent_height(0), parent_width(0),
	  num_rows(0), num_cols(0),
	  edit_entity(0),edit_record(0),
	  num_pages(0), current_page(0)
{
	this->view_type=type;
	this->entity_type=etype;
}

OOCFrame::~OOCFrame()
{
	// face list gets it in frame
}

void
OOCFrame::emptied()
{
	if(this->im_deleted == FALSE) {
		this->im_deleted = TRUE;
		this->view->emptied(this);
	}
}

void
OOCFrame::addMember(iEntityPtr member)
{
	CE *ce;
	ce = new CE(this, member);
	this->addFace(ce);
	eaSetFace(member, this->view_type, ce);
	// hmm this doesnt seem to redisplay stuff
}

void
OOCFrame::sortMember(iEntityPtr member, iEntityPtr prev)
{
	// get a reference to the member
	DLList *l = &this->face_list;
	CE *member_ce = NULL, 
	   *prev_ce   = NULL;

	// search through list for entities
	for(Ref r = dllistFirst(l); r; dllistNext(l, &r)){
		if(((CE *) dllistContents(l, r))->entity == member)
			member_ce = (CE *) dllistContents(l, r);
		if(((CE *) dllistContents(l, r))->entity == prev)
			prev_ce   = (CE *) dllistContents(l, r);
		if((prev_ce || !prev)  && member_ce)
			break;
	}
	
	// detach member
	r  = dllistSeek(l, member_ce);
	dllistRemove(l, r);
	
	// place the member in the correct position
	if(prev_ce){
		r = dllistSeek(l, prev_ce);
		dllistInsertAfter(l, r, member_ce);
	}
	else
		dllistPrepend(l, member_ce);
}


iEntityPtr
OOCFrame::getFirstMember()
{
	Ref r = dllistFirst(&this->face_list);
	CE *ce;
	iEntityPtr entity = 0;
	
	if (r) {
		ce = (CE *) dllistContents(&this->face_list, r);
		entity =  ce->entity;
	} 
	
	return entity;
}

extern int bloatOOCBox(iBox *b,int type);
extern iBox i_GetPBBBounds(iEntityPtr entity);

void
OOCFrame::calcMaxDims(int *max_width_ptr, int *max_height_ptr)
{
	Ref r;
	int max_width = 0;
	int max_height = 0;
	DLList *l = &this->face_list;
	CE *ce;
	int w, h,m;
	iBox box,pbb;
	
	for(r = dllistFirst(l); r; dllistNext(l, &r)){
		ce = (CE *) dllistContents(l, r);
		
		pbb = i_GetPBBBounds(ce->entity);
		box = i_GetBounds(ce->entity);
		bloatOOCBox(&box,view_type);
		if(!bbNullP(pbb)){
			int cx,cy;
			cx = box.x + box.w / 2;
			cy = box.y + box.h / 2;
			w = MAX(box.w / 2, pbb.x + pbb.w - cx);
			h = MAX(box.h / 2, pbb.y + pbb.h - cy);
			w = MAX(w,cx - pbb.x);
			h = MAX(w,cy - pbb.y);
			w *= 2;
			h *= 2;
		} else {
			w = box.w;
			h = box.h;
		}

		m = this->getMagExp();
		w = magUpInt(w, m);
		h = magUpInt(h, m);

		iUCode *text = i_AllocUniText(ce->entity);
		
		if(this->view_type == vtFLAG && text){
			int tw, th;
			int l;
			int as;
			tw = 0;
			th = 0;

			l = ucslen(text);

			this->getCanvas()->measureText(text, l, &tw, &th, &as);
			w = MAX(w, tw);
			h += th;
		}

		i_FreeUniText(text);

		if(w > max_width)
			max_width = w;
		if(h > max_height)
			max_height = h;
	}
	
	*max_width_ptr = max_width;
	*max_height_ptr = max_height;
}

void
OOCFrame::calcAndSetCellDims()
{
	int wi, he;
	Widget ww;
	int max_width;
	int max_height;


	ww = XtParent(this->getCanvas()->canvas);
	getWidgetSize(ww, &wi, &he);

	this->calcMaxDims(&max_width, &max_height);
	
	this->cell_width = max_width + 2 * CELL_MARGIN;
	this->cell_height = max_height + 2 * CELL_MARGIN;

	if(this->cell_width > wi)
		this->cell_width = wi - 5;
}

void
OOCFrame::conditionallyInitialiseCanvas()
{
	int wi, he;
	Widget ww;

	this->calcAndSetCellDims();

	ww = XtParent(this->getCanvas()->canvas);
	getWidgetSize(ww, &wi, &he);

	this->parent_height = he;
	this->parent_width = wi;

	this->calcCells();
	
	this->getCanvas()->setSize(this->width, this->height);
}

void 
OOCFrame::calcCells()
{
	int h;
	int num_members = dllistLength(&this->face_list);
#define MAX_HEIGHT (1<<15)
	
	this->num_cols = (this->parent_width-5) / this->cell_width;
	if (this->num_cols<1) this->num_cols=1;
	if(this->num_cols != 0) {
		this->num_rows = num_members / this->num_cols + 
			((num_members % num_cols == 0) ? 0 : 1);
	} else {
		this->num_rows = 0;
	}
	
	this->num_rows_page = (this->parent_height-5) / this->cell_height;
	this->num_rows_page = MAX(this->num_rows_page, 1);

	this->num_pages =  this->num_rows / this->num_rows_page + 
		((this->num_rows % this->num_rows_page == 0) ? 0 : 1);

	this->width = this->parent_width - 5;
	h = MAX(this->num_rows_page * this->cell_height, this->parent_height - 5);

	this->height = h;
	
	// here is where I should make the assignments
	int i, j, k;
	Ref r;
	DLList *l = &this->face_list;
	CE *ce;

	// sort CE info for each element in list
	for(k = 0, r = dllistFirst(l); k < this->num_pages && r; k++)
		for(j = 0; (j < this->num_rows_page) && r; j++)
			for(i = 0; i<this->num_cols  && r; i++, dllistNext(l, &r)){
				ce = (CE *) dllistContents(l, r);
				ce->assignPosition(k,this->cell_width*i,this->cell_height*j);
			}
}

void
OOCFrame::redisplay (XmDrawingAreaCallbackStruct *call_data)
{
	static Region region = NULL;
	
	Widget w = this->getCanvas()->canvas;
	XEvent *event = (XEvent *) call_data->event;
	
	if(XtIsManaged(w) == FALSE)
		return;
	
	if(!region)
		region = XCreateRegion();
	
	XtAddExposureToRegion(event, region);
	
	if(event->xexpose.count != 0)
		return;
	
	iBox refresh_box;
	
	XRectangle rect;
	XClipBox(region, &rect);
	XDestroyRegion(region);
	region = 0;
	
	refresh_box.x = rect.x;
	refresh_box.y = rect.y;
	refresh_box.w = rect.width;
	refresh_box.h = rect.height;
	
	if(refresh_box.w == 0 || refresh_box.h == 0)
		return;
	
	this->drawRegion(refresh_box);
	
}

void
OOCFrame::drawEntity(iEntityPtr entity)
{
	iBox refresh_box;
	CE *ce = (CE *) eaGetFace(entity, this->view_type);
	if (!ce) return;
	refresh_box.x=ce->x;
	refresh_box.y=ce->y;
	refresh_box.w=this->cell_width;
	refresh_box.h=this->cell_height;
	this->drawRegion(refresh_box);
	
}

void
OOCFrame::drawRegionNow(iBox refresh_box)
{
	Ref r;
	Pixmap pixmap;
	DLList *l = &this->face_list;
	
	Widget w = this->getCanvas()->canvas;
	if(XtIsManaged(w) == FALSE)
		return;
	
	Display *display = XtDisplay(w);
	Drawable drawable = RootWindowOfScreen(XtScreen(w));
	
	int depth = DefaultDepthOfScreen(XtScreen(w));
	    
	pixmap = XCreatePixmap(display, drawable, 
						   refresh_box.w, refresh_box.h,
						   depth);
	XFillRectangle(display, pixmap, 
				   this->getCanvas()->grayGC(),
				   0, 0,
				   refresh_box.w, refresh_box.h);

	
	for(r = dllistFirst(l); r; dllistNext(l, &r)){
		CE *ce;
		ce = (CE *) dllistContents(l, r);
		if(ce->sub_page != this->current_page)
			continue;
		iBox cell_box;
		cell_box.x = ce->x;
		cell_box.y = ce->y;
		cell_box.w = this->cell_width;
		cell_box.h = this->cell_height;
		
		if(overlapBoxes(cell_box, refresh_box)) {
			ce->redisplay(display, pixmap, this->getCanvas(), 
						  ce->x - refresh_box.x, ce->y - refresh_box.y,
						  this->cell_width, this->cell_height);
		} 
	}

	XCopyArea(display, pixmap, XtWindow(this->getCanvas()->canvas), 
			  this->getCanvas()->copyGC(),
			  0, 0,
			  refresh_box.w, refresh_box.h,
			  refresh_box.x, refresh_box.y);
	XFreePixmap(display, pixmap);
}

/*
// bogus there is similar code in initial canvas
*/
void
OOCFrame::configure()
{
	int wi, he;
	Widget ww = XtParent(this->getCanvas()->canvas);
	getWidgetSize(ww, &wi, &he);
	
	if(this->parent_width == 0 && this->parent_height == 0) {
		// special case because this is your first existence?
		this->parent_width = wi;
		this->parent_height = he;
		this->calcCells();
		
		if(this->managedP()) {
			this->unmanage();
			this->manage();
		}
		
		// force a sync here?
		return;
	} else 
		if (this->parent_width == wi && this->parent_height == he) {
			// do nothing since you are the right size
		} else {
			// resize
			this->parent_width = wi;
			this->parent_height = he;
			this->calcCells();
			
			if(this->managedP()) {
				this->unmanage();
				this->manage();
			}
			
			return;
		}
}

iEntityPtr
OOCFrame::findEnclosingEntity(int x, int y)
{   
	DLList *l = &this->face_list;
	CE *ce;
	Ref r;
	
	for(r = dllistFirst(l); r; dllistNext(l, &r)){
		ce = (CE *) dllistContents(l, r);
		if(ce->sub_page != this->current_page)
			continue;
		int cell_x = ce->x;
		int cell_y = ce->y;
		if(x >= cell_x  &&
		   x <= (cell_x + this->cell_width) &&
		   y >= cell_y &&
		   y <= (cell_y + this->cell_height)){
			return ce->entity;
		}
	}
	
	return 0;
}

aString getOOCFrameName(iEntityPtr entity);

void
OOCFrame::refresh()
{
	if(this->getCanvas()){
		iBox refresh_box;
		refresh_box.x = 0;
		refresh_box.y = 0;
		refresh_box.w = this->parent_width;
		refresh_box.h = this->parent_height;

		this->drawRegion(refresh_box);
	}
}

void
OOCFrame::change(iEntityPtr entity)
{
	// for example -- type is changed on a flagged entity in flag view
	// or value changed on glyph in ooc mode
	int change_frame = FALSE;

	// say the type changes in OOC, or the flagged is turned of in flag mode
	int remove_from_view = FALSE;

	// say the text is changed in flag mode
	int changed_value = FALSE;

	aString en_name;

	switch(this->view_type){
	  case vtOOC:
		  remove_from_view = i_GetEntityType(entity) != iGlyphT;
		  en_name = getOOCFrameName(entity);
		  change_frame = en_name != this->e_name;
		  break;
	  case vtFLAG:
		  remove_from_view = entityFlaggedP(entity) == FALSE;
		  en_name = i_TypeToName(i_GetEntityType(entity));
		  change_frame = (en_name!=this->e_name);
		  changed_value = !remove_from_view && !change_frame;
	}

	DLList m;
	CE *ce = (CE *) eaGetFace(entity, this->view_type);

	if(remove_from_view || change_frame) {
		ce->blankGhost();
		// will remove me from the 2 lists
		ce->deleteAndUpdate();
		if(change_frame && !remove_from_view){
			OOCView *od = (OOCView *) this->view;
			dllistCtor(&m);
			dllistAppend(&m, entity);
			od->mergeInList(&m);
			dllistDtor(&m);
				
			ce = (CE *) eaGetFace(entity, this->view_type);
			if (ce) {
				if(ce->frame->managedP()) {
					ce->frame->conditionallyInitialiseCanvas();
				}
			}
		}
	} else if(changed_value)
		ce->redrawCell();
}

void
OOCFrame::begin_edit(iEntityPtr entity)
{
	if (this->edit_record) {
		this->end_edit();
	}
	if (entity) {
		i_StartObjEdit(entity,&this->edit_record);
		this->edit_entity=entity;
	}
}

void
OOCFrame::end_edit()
{
	int delete_me;
	if (this->edit_record) {
		i_EndObjEdit(this->edit_entity,this->edit_record,&delete_me);
		if (delete_me) {
			this->edit_entity=0; // deleteFlagged will call end_edit, so
			this->edit_record=0; // if we don't clear these, we get into trouble
			this->deleteFlagged();
			return; // must return here - the frame itself might be deleted!
		} 
	}
	this->edit_entity=0;
	this->edit_record=0;
}


void
OOCFrame::setItem(iEntityPtr entity)
{
	this->end_edit();
	
	this->getSM()->setItem(entity);
	this->track();
}

iEntityPtr
OOCFrame::getSoleItem()
{
#if 0
	if (this->edit_entity)
		return this->edit_entity;
	else
#endif
	return this->getSM()->getSoleItem();
}

void
OOCFrame::toggleItem(iEntityPtr entity)
{
	this->end_edit();
	this->getSM()->toggleItem(entity);
}

void
OOCFrame::clear()
{
	this->end_edit();
	this->getSM()->clear();
}

const DLList *
OOCFrame::getItems()
{
	return this->getSM()->getItems();
}

static int TRUE_func(iEntityPtr entity)
{
	return TRUE;
}

void
OOCFrame::nextEntity()
{
	this->nextEntity(TRUE_func);
}

void
OOCFrame::prevEntity()
{
	this->prevEntity(TRUE_func);
}

void
OOCFrame::firstEntity()
{
	this->firstEntity(TRUE_func);
}

int
OOCFrame::entityInFrameP(iEntityPtr entity)
{
	CE *ce = (CE *) eaGetFace(entity, this->view_type);
	return ce && (ce->frame == this);
}



void
OOCFrame::downOrUp(oocDir dir)
{
	// bogus: cheap and easy implementation
	CE *ce=0;
	int x_pos;
	int found=0;
	Ref r;
	iEntityPtr sole_entity;
	
	const DLList *selection=this->getSM()->getItems();
	switch(dir) {
	case oocDOWN:
		r = dllistLast(selection);
		break;
	case oocUP:
		r = dllistFirst(selection);
		break;
	}

	if (r)
		sole_entity = (iEntityPtr) dllistContents(selection, r);
	else
		sole_entity=0;
	if (sole_entity) ce = (CE *) eaGetFace(sole_entity, this->view_type);
	if (ce) {
		x_pos=ce->x;
		DLList *l = &this->face_list;
		Ref r = dllistSeek(l, ce);
		if (r) do {
			switch(dir) {
			case oocDOWN:
				dllistNext(l, &r);
				break;
			case oocUP:
				dllistPrev(l, &r);
				break;
			}
			if (!r) break;
			ce = (CE *) dllistContents(l, r);
			if (ce && ce->x==x_pos) found=TRUE;
		} while (r && !found);
		if (found) {
			this->setItem(ce->entity);
		}
	} else {
		switch(dir) {
		case oocDOWN:
			this->firstEntity(TRUE_func);
			break;
		case oocUP:
			this->lastEntity(TRUE_func);
			break;
		}
	}
}


void
OOCFrame::downEntity()
{
	this->downOrUp(oocDOWN);
}

void
OOCFrame::upEntity()
{
	this->downOrUp(oocUP);
}

void
OOCFrame::nextEntity(EntityCheckFunc checker)
{
	iEntityPtr sole_entity;
	iEntityPtr right_entity=0;
	int good=0, next_entity_found=0;
	
	Ref r;
	CE *ce;
	
	const DLList *selection=this->getSM()->getItems();
	r = dllistLast(selection);
	
	if (r)
		sole_entity = (iEntityPtr) dllistContents(selection, r);
	else
		sole_entity = 0;
	
	if(sole_entity){
		DLList *l = &this->face_list;
		ce = (CE *) eaGetFace(sole_entity, this->view_type);
		Ref r = dllistSeek(l, ce);
		while(r && !good) {
			dllistNext(l, &r);
			if(r){
				ce  = (CE *) dllistContents(l, r);
				if (ce) {
					right_entity = ce->entity;
					good=checker(right_entity);
				}
			}
			if (right_entity && good) {
				this->setItem(right_entity);
				next_entity_found = TRUE;
			}
		}
		// wrap to beginning 
		if (!next_entity_found) this->firstEntity(checker);
	} else {
		this->firstEntity(checker);
	}

	this->refresh();
}

void
OOCFrame::prevEntity(EntityCheckFunc checker)
{
	iEntityPtr sole_entity;
	iEntityPtr left_entity=0;
	int good=0;
	
	Ref r;
	
	const DLList *selection=this->getSM()->getItems();
	r = dllistFirst(selection);
	if (r)
		sole_entity = (iEntityPtr) dllistContents(selection, r);
	else
		sole_entity=0;
	if(sole_entity){
		CE *ce;
		DLList *l = &this->face_list;
		ce = (CE *) eaGetFace(sole_entity, this->view_type);
		Ref r = dllistSeek(l, ce);
		while(r && !good) {
			dllistPrev(l, &r);
			if(r){
				ce = (CE *)  dllistContents(l, r);
				if (ce) {
					left_entity = ce->entity;
					good = checker(left_entity);
				}
			}
		}
		if (left_entity && good) this->setItem(left_entity);
	} else {
		this->lastEntity(checker);
	}
}

void
OOCFrame::firstEntity(EntityCheckFunc checker)
{
	iEntityPtr right_entity=0;
	int good=0;
	CE * ce;
	
	DLList *l = &this->face_list;
	Ref r = dllistFirst(l);
	while(r && !good) {
		ce = (CE *) dllistContents(l, r);
		if (ce){
			right_entity = ce->entity;
			good = checker(right_entity);
		}
		dllistNext(l, &r);
	}
	if (right_entity && good) this->setItem(right_entity);
}

void
OOCFrame::lastEntity(EntityCheckFunc checker)
{
	iEntityPtr right_entity=0;
	int good=0;
	CE *ce;
	
	DLList *l = &this->face_list;
	Ref r = dllistLast(l);
	while(r && !good) {
		ce = (CE *) dllistContents(l, r);
		if (ce) {
			right_entity = ce->entity;
			good = checker(right_entity);
		}
		dllistPrev(l, &r);
	}
	if (right_entity && good) this->setItem(right_entity);
}

void
OOCFrame::findNextFlagEntity()
{
	iEntityPtr sole_item = this->getSM()->getSoleItem();
 
	this->nextEntity(entityFlaggedP);
	if(sole_item)
		entityClearFlag(this->view->main_window,sole_item);
}

void 
OOCFrame::setMag(int mag_exp, int old_mag)
{
	this->_setMag(mag_exp, old_mag);
	if(this->managedP()){
		this->unmanage();
		this->manage();
	}
}

void 
OOCFrame::_setMag(int mag_exp, int old_mag)
{
	this->conditionallyInitialiseCanvas();
}



void
OOCFrame::centerOnSelected(iEntityPtr entity)
{
	CE *ce;
	
	ce = (CE *) eaGetFace(entity, this->view_type);
	if(ce && (ce->frame == this)) {
		this->setPage(ce->sub_page);
		Position x;
		Position y;
		
		x = ce->x + this->cell_width/2;
		y = ce->y + this->cell_height/2;
		scrollToPosition(this->view->scrolled_window, x, y);
	}
}

void
OOCFrame::deleteFlagged()
{
	iEntityPtr sole_entity = this->getSoleItem();
	iEntityPtr left_entity=0;
	CE *ce;
	if(sole_entity){
		DLList *l = &this->face_list;
		ce = (CE *) eaGetFace(sole_entity, this->view_type);
		Ref r = dllistSeek(l, ce);
		if(r) {
			dllistNext(l, &r);
			if(r){
				ce = (CE *) dllistContents(l, r);
				left_entity = ce->entity;
			}
		}
		if (!left_entity) {
			ce = (CE *) eaGetFace(sole_entity, this->view_type);
			Ref r = dllistSeek(l, ce);
			if(r) {
				dllistPrev(l, &r);
				if(r){
					// ok, so it's not the left entity any more
					ce = (CE *) dllistContents(l, r);
					left_entity = ce->entity;
				}
			}
		}
	}
	this->end_edit();
	this->view->main_window->deleteSelection(FALSE);

	
	// it's possible for the frame itself to be deleted, in this case
	// we can't call these routines!
	if (!this->im_deleted) {
		iBox foo;
		foo.w = 10000;
		foo.h = 10000;
		foo.x = 0;
		foo.y = 0;
		this->drawRegion(foo);
		if (left_entity) this->setItem(left_entity);
	}
}

// bogus bogus bogus!!!
// okk bbut whaay is it bogus -- Tom
void
ClearFlagCB(Widget w,MainWindow *mw,XtPointer call_data);


int
OOCFrame::pressKeyFlagged(iUCode key_value)
{
	// bogus
	if(key_value == '\0'  || key_value == '\t' || this->getSM()->getState() == scNOT) {
		ringTheBell();
		return 0;
	}

	if (key_value=='\n') {
		// pretty bogus huh?
		ClearFlagCB(0,this->view->main_window,0);
		return 0;
	}
	// bogus why does this make sense here?
	this->track();

	// isnt this where I want to use
	// reassign values so I dont lose the bounding box?

	iEntityPtr entity, sole_entity;
	MainWindow *mw = this->view->main_window;
	sole_entity=mw->sm->getSoleItem();

	if(sole_entity == 0) {
		if(key_value == '\b') {
			this->view->main_window->deleteSelection(FALSE);
			return 1;
		} else if(key_value == 20) {
			killContentsForSelection();
			return 1;
		} else {
			ringTheBell();
			return 0;
		}
	}

	iDTag tag=i_GetEDataType(sole_entity);

	if (!this->edit_record)
		this->begin_edit(sole_entity);

	int delete_me;

	if(key_value < ' ' && (key_value != '\b' && key_value != 20)) {
		ringTheBell();
		return 0;
	}

	i_ObjEdit(this->edit_entity,this->edit_record,key_value,&delete_me);
	if (delete_me) {
		this->deleteFlagged();
	} else {
		this->drawEntity(this->edit_entity);

		this->view->main_window->document->setDirty();
		if ((tag==iD1PosSet)||(tag==iD1PosSet)){
			mw->nextEntity(entityFlaggedP); 
			entity = mw->sm->getSoleItem();
			if (entity==sole_entity){
				mw->prevEntity(entityFlaggedP);
			}

			entityClearFlag(mw,sole_entity); 
		}
	}

	// bogus what does this return value do?
	return 1;
}

int
OOCFrame::pressKeyOOC(iUCode key_value)
{
	if (key_value=='\b') {
		this->view->main_window->deleteSelection(FALSE);
	} else if(key_value == 20) {
		killContentsForSelection();
	} else if (key_value>=' ') {
		this->reassignValues(this->getItems(), key_value);
	}

	// bogus what does this return value do?
	return 1;
}

int
OOCFrame::pressKey(iUCode key_value)
{
	int retval;
	View *view = this->view;
	MainWindow *mw = view->main_window;


	if (this->view_type==vtOOC)
		retval = this->pressKeyOOC(key_value);
	else
		retval = this->pressKeyFlagged(key_value);
	

	if (this->should_delete) {
		if (this->managed)
			if (!view->forward())
				if(!view->back())
					mw->view_canvas->unmanage();
		mw->standardTitle();
		mw->refresh_dialogs();
		//delete frame;
	}

	return retval;
}

void
OOCFrame::manage()
{
	this->managed = TRUE;
	this->setPage(0);
	this->_setMag(this->view->main_window->mag_exp, 
				  this->view->main_window->mag_exp);
	if(this->getCanvas()){
		this->getCanvas()->setClient(this);
		this->getCanvas()->manage();
		this->syncSize();
	}
	if (this->view_type==vtFLAG) {
		iEntityPtr sole_entity = this->getSoleItem();
		if (sole_entity == 0)
			// other_wise assume we will be moved to the
			// sole guy
			this->firstEntity();
	}
}

void
OOCFrame::unmanage()
{
	if (this->view_type==vtFLAG)
		this->end_edit();
	
	this->managed = FALSE;
	if(this->getCanvas()){
		this->getCanvas()->setClient(0);
		this->getCanvas()->unmanage();
		this->syncSize();
	}
}



void
OOCFrame::pressButton(XEvent *event)
{
    iEntityPtr entity;
	
    switch(event->xbutton.button) {
	case Button1:
		entity = this->findEnclosingEntity(event->xbutton.x, event->xbutton.y);
		if(entity != 0) {
			this->end_edit();
			entity = this->findEnclosingEntity(event->xbutton.x, event->xbutton.y);
			if(entity) {
				if(event->xbutton.state & ShiftMask){
					this->toggleItem(entity);
				} else {
					this->setItem(entity);
				}
			}
		} else {
			this->clear();
		}
		break;
	case Button2:
		break;
    }
}

void
OOCFrame::setPage(int input_current_page)
{ 
	this->current_page = input_current_page; 
	this->refresh();
	this->view->main_window->standardTitle();
}

aString
OOCFrame::getName()
{
	aString total_name;
	char buffer[64];
	if(this->num_pages == 1)
		total_name = this->name;
	else {
		sprintf(buffer, " [%d of %d]", this->current_page + 1, this->num_pages);
		total_name = this->name + buffer;
	}

	return total_name;
}

void
OOCFrame::track()
{
	iEntityPtr sole;
	iBox bounds;
	CE *sole_ce;
	
	View *view = this->view;

	sole = this->getSM()->getSoleItem();
	if (sole) {
		bounds = view->getBounds(sole);
		if(bbNullP(bounds) == FALSE) {
			sole_ce = (CE *) eaGetFace(sole, this->view_type);

			// get position of sole_ce
			DLList *l = &this->face_list;
			Ref r = dllistFirst(l);
			int index = 0;

			while((CE *) dllistContents(l, r) != sole_ce){
				index++;
				dllistNext(l, &r);
			}

			view->main_window->document->last_ooc_index = index;
			if(this->current_page != sole_ce->sub_page)
				this->setPage(sole_ce->sub_page);
		}
	}
}
