
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"

#include <stdio.h>

#include "illum.h"

#include "misc/defines.h"

#include "OOCPage.h"
#include "EntityAux.h"

#include "OOCMode.h"
#include "MainWindow.h"
#include "Document.h"
#include "CE.h"
#include "misc/bb.h"
#include "uString.h"
#include "ViewCanvas.h"
#include "SelectionManager.h"
#include "Xm/Form.h"
#include "Xm/Frame.h"
#include "Xm/Label.h"
#include "misc/Unicode.h"

extern "C" int compareCETree(void *a, void *b);
extern "C" int compareEntTree(void *a, void *b);


OOCMode::OOCMode(MainWindow *input_main_window, iEntityPtr root_entity,int type, int flagged)
	: Mode(input_main_window)
{
	this->ooc_type=type;
	this->ooc_flagged=flagged;
	if (this->ooc_type==iGlyphT) this->mode_type=vtOOC;
	else this->mode_type=vtFLAG;
	status_bar = 0;
	
	// create a status bar
	if(this->mode_type == vtOOC) {
	  Widget paned_window = XtParent(this->main_window->current_frame);
	  XtVaSetValues(paned_window, 
			XmNsashHeight, 1,
			XmNsashWidth, 1,
			XmNsashShadowThickness, 0,
			XmNseparatorOn, False,
			XmNspacing, 0,
			NULL);

	  Widget frame;
	  status_bar = XtVaCreateManagedWidget("sb_form", xmFormWidgetClass, paned_window,
					       XmNpaneMinimum, 30,
					       XmNpaneMaximum, 30,
					       NULL);
	  
	  frame = XtVaCreateManagedWidget("sb_frame", xmFrameWidgetClass, status_bar,
					  XmNshadowType, XmSHADOW_IN,
					  XmNtopAttachment, XmATTACH_FORM,
					  XmNbottomAttachment, XmATTACH_FORM,
					  XmNleftAttachment, XmATTACH_FORM,
					  NULL);

	  XmString status_text = XmStringCreateLocalized(" ");

	  status_title = XtVaCreateManagedWidget("status_bar", xmLabelWidgetClass, frame,
						 XmNlabelString, status_text,
						 NULL);

	  frame = XtVaCreateManagedWidget("sb_frame", xmFrameWidgetClass, status_bar,
					  XmNshadowType, XmSHADOW_IN,
					  XmNtopAttachment, XmATTACH_FORM,
					  XmNbottomAttachment, XmATTACH_FORM,
					  XmNrightAttachment, XmATTACH_FORM,
					  XmNleftAttachment, XmATTACH_WIDGET,
					  XmNleftWidget, frame,
					  NULL);

	  status_prompt = XtVaCreateManagedWidget("status_bar", xmLabelWidgetClass, frame,
						  XmNlabelString, status_text,
						  XmNrecomputeSize, False,
						  XmNalignment, XmALIGNMENT_BEGINNING,
						  NULL);

	  XmStringFree(status_text);
	}

	this->createPages(root_entity);
}

OOCMode::~OOCMode()
{
	if(status_bar) {
    	XtUnmanageChild(status_bar);
		XtDestroyWidget(status_bar);
	}

	// not b---- because we have no local members
}

void 
OOCMode::chooseMembers(iEntityPtr entity, DLList *m)
{
	iEntityPtr child;
	int ok;
	iBox bounds;
	
	if (!(entity = pickpath(entity))) 
		return;
	bounds = i_GetEnclosingBox(entity);
	int null_bounds = bbNullP(bounds);	

	ok = (null_bounds==0);
	if (ok && this->ooc_type)    ok = (i_GetEntityType(entity)==this->ooc_type);
	if (ok && this->ooc_flagged) ok = entityFlaggedP(entity);
	if (ok) dllistAppend(m, entity);

	for(child = i_GetChildEntity(entity); child; child = i_GetNextEntity(child))
		this->chooseMembers(child, m);
}

int4
compareEnts(VoidPtr a, VoidPtr b)
{
	iEntityPtr entity_a = (iEntityPtr) a;
	iEntityPtr entity_b = (iEntityPtr) b;
	
	int result;
	uString ua(getFirstOfPosset(entity_a));
	uString ub(getFirstOfPosset(entity_b));
	
	if (ua < ub) result=-1;
	else if (ua == ub) result=0;
	else result=1;
	
	return result;
}

int4
compareEntType(VoidPtr a, VoidPtr b)
{
	int int_a = i_GetEntityType((iEntityPtr) a);
	int int_b = i_GetEntityType((iEntityPtr) b);
	
	if (int_a<int_b) return -1;
	else if (int_a==int_b) return 0;
	else return 1;
}

int4
comparePageType(VoidPtr a, VoidPtr b)
{
	iEntityPtr ent_a = ((OOCPage *) a)->getFirstMember();
	iEntityPtr ent_b = ((OOCPage *) b)->getFirstMember();
	if (ent_a==ent_b) 
		return 0; 
	// really just handling case where ent_a=ent_b=0
	if (ent_a==0) return -1;
	if (ent_b==0) return +1;
	return compareEntType(ent_a,ent_b);
}

int4 
comparePage(VoidPtr a, VoidPtr b)
{
	iEntityPtr ent_a = ((OOCPage *) a)->getFirstMember(),
	           ent_b = ((OOCPage *) b)->getFirstMember();

	if(ent_a == ent_b)
		return 0;
	if(ent_a == 0) 
		return -1;
	if(ent_b == 0)
		return 1;

	return compareEnts(ent_a, ent_b);
}
void
OOCMode::setMag(int mag_exp, int old_mag)
{
	Page *page;
	DLList *selection = (DLList*)this->getSM()->getItems();

	page = getCurrentPage();

	if (!dllistEmpty(selection)) {
		Ref r = dllistFirst(selection);
		iEntityPtr entity = (iEntityPtr)dllistContents(selection,r);
		this->centerOnSelected(entity);
	} else if (page) {
		page->setMag(mag_exp, old_mag);
	}
	
}

void
OOCMode::createPages(iEntityPtr root_entity)
{
	DLList m;
	
	dllistCtor(&m);
	this->chooseMembers(root_entity, &m);
	
	this->createList(&m);
	dllistDtor(&m);
}

// in flag mode, we want to preserve the order,
// therefore we can't sort as in OOC mode
void OOCMode::mergeInListFlag(DLList *m)
{
	Ref p_r;
	Ref e_r;
	OOCPage *p;
	aString p_name;
	aString e_name;
	iEntityPtr first_entity;
	
	for(e_r = dllistFirst(m); e_r; dllistNext(m, &e_r)) {
		iEntityPtr entity = (iEntityPtr) dllistContents(m, e_r);
		for (p_r = dllistFirst(&this->page_list); p_r; dllistNext(&this->page_list,&p_r)) {
			p = (OOCPage *) dllistContents(&this->page_list, p_r);
			first_entity=p->getFirstMember();
			if (compareEntType(entity,first_entity)==0) break;
		}
		if (!p_r) {
			// new page
			e_name = i_TypeToName(i_GetEntityType(entity));
			p_name = e_name + ": "+"flagged";
			
			
			p = new OOCPage(e_name, p_name, 
							this, this->mode_type,
							i_GetEntityType(entity));
			first_entity=entity;
			this->addPage(p);
			
		}
		p->addMember(entity);
	}
	if (p) {
		Page *page = this->getCurrentPage();
		dllistSort(&this->page_list, comparePageType);
		this->current_page_ref = dllistSeek(&this->page_list, page);
	}
}

aString getOOCPageName(iEntityPtr entity)
{
	const iUCode *c;
	uString us(getFirstOfPosset(entity));
	aString as;
	char name_str[8];
	
	for(c = us.string(); *c; c++)
		if (*c<32 || *c>127) {
			sprintf(name_str,"U+%x",*c);
			as += name_str;
		} 
		else {
			name_str[0] = *c;
			name_str[1] = 0;
			as += name_str;
		}

	return as;
}

void
OOCMode::mergeInList(DLList *m)
{
	Ref e_r;
	Ref p_r;
	OOCPage *p;
	aString e_name;
	int4 (*sorter)(VoidPtr,VoidPtr);
	
	if (this->ooc_flagged) {
		this->mergeInListFlag(m);
		return;
	}
	
	if (this->ooc_type==iGlyphT) 
		sorter=compareEnts;
	else sorter=compareEntType;
	dllistSort(m, sorter);
	
	p_r = dllistFirst(&this->page_list);
	if(p_r)
		p = (OOCPage *) dllistContents(&this->page_list, p_r);
	else
		p = 0;
	
	for(e_r = dllistFirst(m); e_r; dllistNext(m, &e_r)) {
		iEntityPtr entity = (iEntityPtr) dllistContents(m, e_r);
		if (this->ooc_type!=iGlyphT) {
			e_name = i_TypeToName(i_GetEntityType(entity));
		} else {
			e_name=getOOCPageName(entity);
		}

		while(p_r && sorter(p->getFirstMember(),entity) < 0) {
			dllistNext(&this->page_list, &p_r);
			if(p_r)
				p = (OOCPage *) dllistContents(&this->page_list, p_r);
			else 
				p = 0;
		}
		
		if(p && (e_name == p->e_name)){
			iEntityPtr prev;

			p->addMember(entity);
			if(this->ooc_type != iGlyphT)
				prev = i_FindEntityBack(entity, i_GetEntityType(entity), ALL, NULL);
			else{
				int top_char = i_GetGlyph(entity, NULL);
				prev = i_FindEntityBack(entity, iGlyphT, POSSET, (void *) top_char);
			}
			p->sortMember(entity, prev);
		}
		else {
			aString p_name;
			if(e_name == "")
				p_name = "Non-rec";
			else
				p_name = e_name;
			p = new OOCPage(e_name, p_name, this, this->mode_type,iGlyphT);
			p_r = this->addPageBefore(p_r, p);
			p->addMember(entity);
		}
	}
}

static const char *
makeKey(iEntityPtr entity, OOCMode *that)
{
	static aString key;

	if (that->ooc_type != iGlyphT) 
		key = i_TypeToName(i_GetEntityType(entity)); 
	else 
		key = getOOCPageName(entity);

	return key.string();
}

void
OOCMode::createList(DLList *entity_list)
{
	// blowout if flagged
	if (this->ooc_flagged) {
		this->mergeInListFlag(entity_list);
		return;
	}

	// create a hash table of the page_list as entries
	stringVoidPtrMap map;
	svmapInit(&map, 0, 10001);

	Ref r;
	OOCPage *page;
	aString key;
	iEntityPtr entity;

	for(r = dllistFirst(entity_list); r; dllistNext(entity_list, &r)) {
		entity = (iEntityPtr) dllistContents(entity_list, r);

		key = makeKey(entity, this);
		if(svmapContainsP(&map, key.string())) {
			page = (OOCPage *) svmapGet(&map, key.string());
			page->addMember(entity);
		}
		else {
			aString f_name, e_name = key;
			if(e_name == "")
				f_name = "Non-rec";
			else
				f_name = e_name;

			page = new OOCPage(e_name, f_name, this, this->mode_type, iGlyphT);
			svmapSet(&map, key.string(), page);
			page->addMember(entity);
			dllistAppend(&this->page_list, page);
		}
	}

	// preserve the current page reference
	Page *current_page = this->getCurrentPage();
	if(this->ooc_type != iGlyphT)
		dllistSort(&this->page_list, comparePageType);
	else
		dllistSort(&this->page_list, comparePage);
	this->current_page_ref = dllistSeek(&this->page_list, current_page);

	svmapCleanup(&map);
}

void
OOCMode::configure()
{
	OOCPage *page;
	DLList *fl = &this->page_list;
	Ref r;
	
	r = this->current_page_ref;
	if(r != 0){
		page = (OOCPage *) dllistContents(fl, r);
		page->configure();
	}
}

void
OOCMode::addNewEntity(iEntityPtr new_entity)
{
	DLList l;
	dllistCtor(&l);
	dllistAppend(&l, new_entity);
	this->mergeInList(&l);
	dllistDtor(&l);
}

void
OOCMode::choose()
{
	
}



void
OOCMode::centerOnSelected(iEntityPtr entity)
{
	// bogus, find the page
	// I'm on not the current page
	
	OOCPage *page;
	CE *ce;
	ce = (CE *) eaGetFace(entity, this->mode_type);
	page = (OOCPage *) (ce ? ce->page : 0);
	DLList *l = &this->page_list;
	Ref r;
	r = dllistSeek(l, page);
	if(r){
		this->setCurrentPage(r);
#if 1
		page->centerOnSelected(entity);
#endif
	}      
}

iBox
OOCMode::getBounds(iEntityPtr entity)
{
	iBox bounds;
	
	// -- check that the page matches
	bounds.w=0;
	
	OOCPage *current_page;
	OOCPage *entitys_page;
	CE *ce;
	
	ce = (CE *) eaGetFace(entity, this->mode_type);
	if (ce) {
		entitys_page = (OOCPage *) ce->page;
		current_page = (OOCPage *) this->getCurrentPage();
		
		if(entitys_page == current_page) {
			bounds=ce->computeBounds();
		}
	}
	
	return bounds;
}


void
OOCMode::setCurrentPage(Ref r)
{
	OOCPage *page;
	this->Mode::setCurrentPage(r);
	
	page = (OOCPage *) this->getCurrentPage();
	if(page) {
	  this->main_window->document->last_ooc_page = page->name;
	}
}


int
OOCMode::forward()
{
	Ref r;
	
	OOCPage *current_page = (OOCPage *) this->getCurrentPage();
	if(current_page->current_page + 1 < current_page->num_pages) {
		current_page->setPage(current_page->current_page + 1);
		return TRUE;
	}

	r=this->forwardPageRef();
	if(r) {
		this->setCurrentPage(r);
		return TRUE;
	} else
		return FALSE;
}

int
OOCMode::back()
{
	Ref r;
	
	OOCPage *current_page = (OOCPage *) this->getCurrentPage();
	if(current_page->current_page - 1 >= 0) {
		current_page->setPage(current_page->current_page - 1);
		return TRUE;
	}
	
	r=this->backwardPageRef();
	if(r) {
		this->setCurrentPage(r);
		current_page = (OOCPage *) this->getCurrentPage();
		current_page->setPage(current_page->num_pages - 1);
		return TRUE;
	} else
		return FALSE;
}


