
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef _OOCPage_h_
#define _OOCPage_h_



#include "Page.h"
#include "text/EntityFace.h"

enum oocDir { oocUP, oocDOWN };

class OOCPage : public Page {
public:
	int parent_width;
	int parent_height;
	int cell_width;
	int cell_height;
	int num_pages;
	int num_cols;
	int num_rows;
	int num_rows_page;
	int current_page;
	iEditRecordPtr edit_record;
	ModeType mode_type;
	int entity_type;
	aString e_name;
	iEntityPtr alt_en;
	int alt_next;
	
	OOCPage(
	    const aString& e_name, 
	    const aString& name, Mode *parent, int type,
	    int etype);
	virtual ~OOCPage();
	void sortMember(iEntityPtr member, iEntityPtr prev);
	void addMember(iEntityPtr member);
	void configure();
	void emptied();
	
	/*virtual*/ void setItem(iEntityPtr);
	/*virtual*/ iEntityPtr getSoleItem();
	/*virtual*/ void toggleItem(iEntityPtr);
	/*virtual*/ void clear();
	/*virtual*/ const DLList *getItems();
	
	iEntityPtr getFirstMember();
	
	void calcMaxDims(int *max_width_ptr, int *max_height_ptr);
	void calcCells();
	void calcAndSetCellDims();

	void setTexts(const DLList *list, iUCode *key_values);
	
	void refresh();
	
	// not different from draw entity
	//void refresh(iEntityPtr);
	void change(iEntityPtr);
	void begin_edit(iEntityPtr);
	void end_edit();
	int entityInPageP(iEntityPtr entity);
	void deleteFlagged();
	
	/*virtual*/ void         pressButton(XEvent *event);
	
	/*virtual*/ void manage();
	/*virtual*/ void unmanage();
	 
	/*virtual*/ void prevEntity();
	/*virtual*/ void nextEntity();
	void firstEntity();
	/*virtual*/ void upEntity();
	/*virtual*/ void downEntity();
	/*virtual*/ void findNextFlagEntity();
	/*virtual*/ void nextEntity(EntityCheckFunc checker);
	/*virtual*/ void prevEntity(EntityCheckFunc checker);
	/*virtual*/ void firstEntity(EntityCheckFunc checker);
	/*virtual*/ void lastEntity(EntityCheckFunc checker);
	/*virtual*/ void setMag(int mag_exp, int old_mag);
	/*virtual*/ void _setMag(int mag_exp, int old_mag);
	/*virtual*/ iEntityPtr findEnclosingEntity(int x, int y);
	
	/*virtual*/ void conditionallyInitialiseCanvas();
	/*virtual*/ void redisplay(XmDrawingAreaCallbackStruct *call_data);
	/*virtual void resize(XmDrawingAreaCallbackStruct *call_data);*/
	/*virtual*/ int pressKey(iUCode key_value);
	int pressKeyFlagged(iUCode key_value);
	int pressKeyOOC(iUCode key_value);
	
	/*virtual*/ void centerOnSelected(iEntityPtr entity);
	/*virtual*/ void scrollToUpperLeft(iEntityPtr entity){};
	/*virtual*/ void scrollToTop(iEntityPtr entity){};
	/*virtual*/ void scrollToRight(iEntityPtr entity){};
	/*virtual*/ void scrollToLeft(iEntityPtr entity){};
	/*virtual*/ void scrollToBottom(iEntityPtr entity){};
	void drawEntity(iEntityPtr entity);
	
	/*virtual*/ void drawRegionNow(iLBox box);
	/*virtual*/ int resizeNecessaryP() { return FALSE; }
	/*virtual void setPicture(iBox b);*/
	/*virtual iBox getPicture();*/
	
	
	void setPage(int input_current_page);
	/*virtual*/ aString getName();

	/*virtual*/ void track(Focus focus);
  private:
	void downOrUp(oocDir dir);
	void findAlternative(void);
};


#endif
