
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */


#ifndef _Page_h_
#define _Page_h_


#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <X11/Xutil.h>

#include <Xm/DrawingA.h>

#include "illum.h"
#include "dllist.h"
#include "aString.h"
enum Focus {
	TOP,
	BOTTOM,
	LEFT,
	RIGHT,
	CENTER
};

typedef enum Focus Focus;

class Mode;
class ViewCanvas;
class SelectionManager;
class EntityFace;

class Page {
public:
	// back_pointer
	Mode *mode;
	iLBox picture;
	
	// these are entity faces
	// dyn
	DLList face_list;
	
  public:
	// window stuff
	aString name;
	int managed;
	int4  width;
	int4  height;
	
	// tricky delete response
	int im_deleted, should_delete;
	
	iEntityPtr edit_entity;
	
	// delayed drawing workproc
	iLBox redraw_region;
	int redraw_procid;
public:
	Page(const aString& name, Mode *parent);
	virtual ~Page();
	
	// for delayed drawing
	void addRedrawRegion(iLBox *b);
	void flushRedrawRegion();
	void drawRegion(iLBox box);
	virtual void drawRegionNow(iLBox box) = 0;
	virtual int resizeNecessaryP() = 0;
	virtual void setPicture(iLBox b);
	virtual iLBox getPicture();
	void setViewPort(iBox b);
	iBox getViewPort();
	
	ViewCanvas *getCanvas();
	void configureCanvas(Page *, int w, int h);
	// x widget stuff
	virtual void conditionallyInitialiseCanvas() = 0;
	// create pixmap -- how strange, why here?
	Pixmap createPixmap(char *bits, int dw, iBox bounds);
	
	virtual void manage();
	virtual void unmanage();
	int managedP();
	
	void syncSize();
	
	// "mirrors" to selection manager calls
	// allows you to manage your selection
	// on your own
	virtual void setItem(iEntityPtr);
	virtual void toggleItem(iEntityPtr);
	virtual void clear();
	virtual iEntityPtr getSoleItem();
	virtual const DLList *getItems();
	
	// handlers and callbacks
	 
	int getMagExp();
	virtual void setMag(int mag_exp, int old_mag_exp) = 0;
	virtual void _setMag(int mag_exp, int old_mag_exp) = 0;
	
	// x event handling
	virtual void redisplay(XmDrawingAreaCallbackStruct *call_data) = 0;
	virtual void resize(XmDrawingAreaCallbackStruct *call_data);
	virtual void         pressButton(XEvent *event);
	virtual void         releaseButton(XEvent *event);
	virtual void         moveButton(int state, int x, int y);
	virtual int  pressKey(iUCode c);
	virtual iEntityPtr findEnclosingEntity(int x, int y) = 0;
	
	// actions
	void reassignValues(const DLList *list, iUCode key_value);
	void deleteSelection();
	
	void removeFace(EntityFace *face);
	virtual void addFace(EntityFace *face);
	
	// for handline left right and arrow keys
	virtual void prevEntity() {}
	virtual void nextEntity() {}
	virtual void upEntity() {}
	virtual void downEntity() {}
	virtual void findNextFlagEntity() {}
	virtual void nextEntity(EntityCheckFunc checker) {}
	virtual void prevEntity(EntityCheckFunc checker) {}
	virtual void firstEntity(EntityCheckFunc checker) {}
	
	// for handling and tracking selection
	virtual void centerOnSelected(iEntityPtr entity) {}
	virtual void scrollToUpperLeft(iEntityPtr entity){};
	virtual void scrollToLeft(iEntityPtr entity) {}
	virtual void scrollToRight(iEntityPtr entity) {}
	virtual void scrollToTop(iEntityPtr entity) {}
	virtual void scrollToBottom(iEntityPtr entity) {}
	SelectionManager *getSM();
	virtual void track(Focus focus);

	void purge();

	virtual aString getName();
};

#endif
