
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"

#include "c-type.h"
#include "smemory.h"

#include <Xm/DialogS.h>
#include <X11/Xutil.h>

#include <Xm/Text.h>
#include <Xm/Label.h>
#include <Xm/TextF.h>
#include <Xm/Form.h>
#include <Xm/ScrolledW.h>
#include <Xm/MessageB.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/PanedW.h>
#include <Xm/Frame.h>
#include <Xm/Separator.h>


#include "misc/misc.h"
#include "rafutil.h"
#include "ucs.h"

#include "misc/XArg.h"

#include "Application.h"

#include "PropContentDialog.h"


static MenuItem ptypes[] = {
	{ 0, "Shouldn't be seen", &xmPushButtonGadgetClass, 'x', 
	  NULL, NULL, (XtCallbackProc) cdSetType, (void *) iPAny, NULL },
	{ 0, "No Data", &xmPushButtonGadgetClass, 'N', 
	  NULL, NULL, (XtCallbackProc) cdSetType, (void *) iPNo, NULL },
	{ 0, "Boolean", &xmPushButtonGadgetClass, 'B', 
	  NULL, NULL, (XtCallbackProc) cdSetType, (void *) iPBool, NULL },
	{ 0, "Integer", &xmPushButtonGadgetClass, 'I', 
	  NULL, NULL, (XtCallbackProc) cdSetType, (void *) iPLong, NULL },
	{ 0, "String", &xmPushButtonGadgetClass, 'S', 
	  NULL, NULL, (XtCallbackProc) cdSetType, (void *) iPString, NULL },
	{ 0, "Unicode String", &xmPushButtonGadgetClass, 'U', 
	  NULL, NULL, (XtCallbackProc) cdSetType, (void *) iPUString, NULL },
	{ 0, "User Data", &xmPushButtonGadgetClass, 'D', 
	  NULL, NULL, (XtCallbackProc) cdSetType, (void *) iPUserData, NULL },
	{ 0, "shouldn't be seen", &xmPushButtonGadgetClass, 'y', 
	  NULL, NULL, (XtCallbackProc) cdSetType, (void *) iPUserIData, NULL },
	{ 0, "Transient Data", &xmPushButtonGadgetClass, 'T', 
	  NULL, NULL, (XtCallbackProc) cdSetType, (void *) iPUserPtr, NULL },
	NULL_MENU,	
};


PropContentDialog::PropContentDialog(MainWindow *input_mw, 
									 Widget input_app_shell, 
									 XtAppContext input_app_context)
	: ContentDialog(input_mw, input_app_shell, input_app_context,
					ptypes, MAX_PTAG, "Property Contents"),
	  entity(0),
	  prop(0),
	  name(0),
	  name_field(0),
	  integer_widget(0)
{		
}

PropContentDialog::~PropContentDialog()
{
}


void
PropContentDialog::createCustom()
{
	XArg args;
	Widget rc;
	Widget label;
	XmString xms;

	XtUnmanageChild(this->menu[iPAny].result);		
	XtUnmanageChild(this->menu[iPUserIData].result);

	args.reset();
	args.setArg(XmNorientation, XmHORIZONTAL);
	args.setArg(XmNpacking, XmPACK_COLUMN);
	args.setArg(XmNnumColumns, 1);
	rc = XmCreateRowColumn(XtParent(this->frame), "rc", 
						   args.getArgs(), args.getSize());

	args.reset();
	xms = XmStringCreateLocalized("Name:");
	args.setArg(XmNlabelString, xms);
	label = XmCreateLabel(rc, "label",
						  args.getArgs(), args.getSize());
	XmStringFree(xms);
	XtManageChild(label);

	args.reset();
	args.setArg(XmNeditable, FALSE);
	this->name_field = XmCreateTextField(rc, "name_field",
						  args.getArgs(), args.getSize());
	XtManageChild(this->name_field);


	// make a space for the name of the property
	args.reset();
	args.setArg(XmNbottomWidget, rc);
	args.setValues(this->frame);

	args.reset();
	args.setArg(XmNleftAttachment, XmATTACH_FORM);
	args.setArg(XmNleftAttachment, XmATTACH_FORM);
	args.setArg(XmNbottomAttachment, XmATTACH_WIDGET);
	args.setArg(XmNbottomWidget, this->separator);
	args.setValues(rc);

	XtManageChild(rc);


	// No data
	xms = XmStringCreateLocalized("No data");
	args.reset();
	args.setArg(XmNlabelString, xms);
	label = XmCreateLabel(this->frame, "label",
				  args.getArgs(), args.getSize());
	XmStringFree(xms);
	this->content_widgets[iPNo] = label;

	Widget text;
	// String
	args.reset();
	args.setArg(XmNeditMode, XmMULTI_LINE_EDIT);
	text = XmCreateText(this->frame, "text",
						args.getArgs(), args.getSize());
	this->content_widgets[iPString] = text;

	// Integer
	args.reset();
	args.setArg(XmNorientation, XmHORIZONTAL);
	args.setArg(XmNpacking, XmPACK_COLUMN);
	args.setArg(XmNnumColumns, 2);
	rc = XmCreateRowColumn(this->frame, "rc", 
						   args.getArgs(), args.getSize());

	args.reset();
	xms = XmStringCreateLocalized("Integer:");
	args.setArg(XmNlabelString, xms);
	label = XmCreateLabel(rc, "label",
						  args.getArgs(), args.getSize());
	XmStringFree(xms);
	XtManageChild(label);

	args.reset();
	args.setArg(XmNmaxLength, 5);
	args.setArg(XmNcolumns, 5);
	args.setArg(XmNvalue, "0");

	this->integer_widget = XmCreateTextField(rc, "ps", 
											 args.getArgs(), args.getSize());
	XtAddCallback(this->integer_widget, XmNmodifyVerifyCallback,
				  (XtCallbackProc) verify_short, this);
	XtManageChild(this->integer_widget);

	this->content_widgets[iPLong] = rc;

	// because we haven't implemented this functionality yet
	Widget ok_button;
	ok_button = XmMessageBoxGetChild(this->dialog,  XmDIALOG_OK_BUTTON);
	XtSetSensitive(ok_button, FALSE);
}

void
PropContentDialog::emptyDialog()
{
	XmTextFieldSetString(this->name_field, "");
}

void
PropContentDialog::putInDialog()
{
	char *s;
	s = this->name.string();
	XmTextFieldSetString(this->name_field, s);
}


void
PropContentDialog::getFromDialog()
{
}

/*
// put the usual event loop in here
*/
int
PropContentDialog::showModal(iEntityPtr entity, iPropPtr prop, char *name)
{
	this->entity = entity;
	this->prop = prop;
	this->name = name;
	this->tag = i_GetPropType(prop);
	
	this->ContentDialog::showModal();
	return this->changed;
}

