
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef _QuadTree_h_
#define _QuadTree_h_

#include "illum.h"
#include "dllist.h"
#include "svpmap.h"

class QuadLeaf;
class QuadNode;
class QuadTree;
class ImageView;
typedef QuadNode *QuadNodePtr;

class QuadLeaf {
	friend class QuadNode;
private:
	// a list of faces
	DLList list;
public:
	QuadLeaf();
	~QuadLeaf();
	
	void addItem(VoidPtr item);
	void deleteItem(VoidPtr item);
	void addMyGuys(stringVoidPtrMap *return_map);
	void print(int indent_level);
};

class QuadNode {
private:
	iBox bounds;

	// it contains 1 or the other of these

	// a list of nodes
	QuadNodePtr *quarters;

	// a leaf
	QuadLeaf *leaf;

public:   
	QuadNode(const iBox& dimensions, int depth);
	~QuadNode();
	
	void addItem(VoidPtr, const iBox& bounds);
	void deleteItem(VoidPtr, const iBox& bounds);
	
	void findOverlappingItems(const iBox& box, stringVoidPtrMap *return_map);
	void print(int indent_level);
	void debugFind(VoidPtr item, const iBox& bounds);
};



typedef iBox (*BoundsFunction)(void *user_data, void *item);

class QuadTree {
private:
	void *user_data;
	// a single node
	QuadNode *node;

	BoundsFunction function;

	friend class QuadLeaf;
	friend class QuadNode;
public:
	QuadTree(BoundsFunction function, void *user_data);
	~QuadTree();
	
	int initialized() { return this->node != 0; }
	void createTree(iBox dimensions, int depth);
	void addItems(DLList *);
	void addItem(VoidPtr);
	void deleteItems(DLList *);
	void deleteItem(VoidPtr);
	
	void findOverlappingItems(const iBox& box, DLList *return_list);
	void print();
	void debugFind(VoidPtr item, int obey_bounds);
};

#endif
