
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#define CORRECT_LINE_DEFINITION

#include "rafport.h"

#include "illum.h"

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <X11/cursorfont.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/BulletinB.h>

#include "rafutil.h"
#include "smemory.h"
#include "misc/Unicode.h"
#include "misc/XArg.h"
#include "misc/misc.h" 
#include "scroll.h"
#include "font.h"
#include "dafs.h"
#include "misc/bb.h"

#include "rte/ConstrainedRectangle.h"

#include "MainWindow.h"

#include "ViewCanvas.h"
#include "Mode.h"
#include "RTEMode.h"
#include "EntityAux.h"
#include "MainWindow.h"
#include "Document.h"
#include "SelectionManager.h"

#include "RTElement.h"
#include "RTEUniEditor.h"


#include "config.h"
#include "RTEPage.h"


#define SCREEN_DPI 75
#define DEFAULT_IMAGE_RES 300
#define DEFAULT_MAGNIFICATION 4
#define MAX_IMAGE_RES 600
#define DEFAULT_PAGE_WIDTH (8*SCREEN_DPI)
#define DEFAULT_PAGE_HEIGHT (10*SCREEN_DPI)
#define DEFAULT_EDITOR_WIDTH 1
#define DEFAULT_EDITOR_HEIGHT 1

RTEPage::RTEPage(Mode *input_mode, 
				 const aString& label, 
				 iEntityPtr input_root_entity)
	: Page(label, input_mode),
	  ConstrainedRectangle(0, 1, 1, 1, 1, LeftToRight, TopToBottom, 0),
	  initialized_canvas(0),
	  x_expansion_factor(1.0),
	  y_expansion_factor(1.0), 
	  root_entity(input_root_entity),
	  active_editor(0)
{
	XArg args;
	int image_res, page_width, page_height;
	RTEMode *mode = (RTEMode*)this->mode;
	iImagePtr image = i_GetImage(this->root_entity);
	
	if(image) {
		image_res = MAX(0,MIN(i_GetImageRes(image),MAX_IMAGE_RES));;
		if(image_res == 0)
			image_res = DEFAULT_IMAGE_RES;
		this->magnification = MAX(1, image_res / SCREEN_DPI);
		page_width = i_GetImageWidth(image)/this->magnification;
		page_height = i_GetImageHeight(image)/this->magnification;
	} else {
		this->magnification = DEFAULT_MAGNIFICATION;
		iBox b = i_GetEnclosingBox(this->root_entity);
		page_width = b.w/this->magnification;
		page_height = b.h/this->magnification;
	}
	
	if(page_width == 0)
		page_width = DEFAULT_PAGE_WIDTH;
	if(page_height == 0)
		page_height = DEFAULT_PAGE_HEIGHT;
		
	this->moveTo(0, 0);

	Position canvas_x, canvas_y;
	Dimension canvas_w, canvas_h;
	Pixel background;
	args.reset();
	args.setArg(XmNx, &canvas_x);
	args.setArg(XmNy, &canvas_y);
	args.setArg(XmNwidth, &canvas_w);
	args.setArg(XmNheight, &canvas_h);
 	args.setArg(XmNbackground, &background);
	args.getValues(mode->main_window->view_canvas->canvas);

	args.reset();
	args.setArg(XmNx, canvas_x);
	args.setArg(XmNy, canvas_y);
	args.setArg(XmNwidth, canvas_w);
	args.setArg(XmNheight, canvas_h);
	args.setArg(XmNbackground, background);
	args.setArg(XmNresizePolicy, XmRESIZE_NONE);
	args.setArg(XmNmarginWidth, 0);
	args.setArg(XmNmarginHeight, 0);
	this->bb = XmCreateBulletinBoard(mode->swindow, "rte_bb", 
	                                 args.getArgs(), args.getSize());
	this->changeSizeTo(page_width, page_height);
}

void 
RTEPage::addMember(iEntityPtr member, int show_cursor)
{
	iBox bounds = i_GetEnclosingBox(member);
	int calc_expansion_factor = TRUE;

	// bogus -- I added this to be able to
	// handle a file like 0003.dr, where the top level
	// entity has no bounding box
	if(bounds.w == 0){
	  iImagePtr image = i_GetImage(member);
	  if(image)
		bounds = i_GetImageBounds(image);
	  if(bounds.w == 0) {
	    calc_expansion_factor = FALSE;
#if 1	    
	    bounds.w = DEFAULT_EDITOR_WIDTH * this->magnification;
	    bounds.h = DEFAULT_EDITOR_HEIGHT * this->magnification;
#else
	    bounds.w = this->width;
	    bounds.h = this->height;
#endif
	  }
	}

	if (bounds.w) {
	  float x_exp_factor, y_exp_factor;
	  const Coord scaled_x = bounds.x / this->magnification,
	    scaled_y = bounds.y / this->magnification,
	    scaled_w = MAX(bounds.w / this->magnification, 1),
	    scaled_h = MAX(bounds.h / this->magnification, 1);
	  RTElement *elem = new RTElement(this, member, scaled_x, scaled_y, scaled_w, scaled_h,
					  &x_exp_factor, &y_exp_factor);
	    if(calc_expansion_factor) {
	        this->x_expansion_factor = MAX(x_exp_factor, this->x_expansion_factor);
		this->y_expansion_factor = MAX(y_exp_factor, this->y_expansion_factor);
	    } else {
	        this->x_expansion_factor = 1.0;  // probably redundant
	        this->y_expansion_factor = 1.0;
	    }
	    this->addFace(elem);
	    eaSetFace(member, vtRTE, elem);
	    if(show_cursor)
	        elem->showCursor();
	}
}

void
RTEPage::expandEditors()
{
	// scale coords of all my UniEditors by my expansion_factor
	if(this->x_expansion_factor > 1.0 || this->y_expansion_factor > 1.0) {
		this->changeSizeTo((Coord)(this->getW()*this->x_expansion_factor), 
		                   (Coord)(this->getH()*this->y_expansion_factor));
		ElementList *el = this->getAllChildren();
		const int squeeze=FALSE, allow_overlap=TRUE;
		for(int i=0; i<el->numElements(); i++) {
			RTElement *e = (RTElement *)el->getItemNumber(i);
			e->moveTo((Coord)(e->getX()*this->x_expansion_factor), 
			          (Coord)(e->getY()*this->y_expansion_factor),
			          squeeze, allow_overlap);
		}
		delete el;
	}
}

RTEPage::~RTEPage()
{
	this->flushRedrawRegion();
	XtDestroyWidget(this->bb);
	dllistDtor(&this->editor_list);
}

int 
RTEPage::changeSizeTo(Coord input_w, Coord input_h, int redraw)
{
	// Update the ConstrainedRectangle part of me
	const int result = this->ParentRectangle::changeSizeTo(input_w, input_h);
	
	// If I successfully changed size as a ConstrainedRectangle (might
	// fail if my parent won't allow me to grow) then update the 
	// Page part of me to agree with ConstrainedRectangle part.
	if(result) {
		// set the width and height members in the Page part of me
		this->width = this->getW();
		this->height = this->getH();
		
		// update my Page bounding box
		XArg args;
		args.reset();
		args.setArg(XmNwidth, this->width);
		args.setArg(XmNheight, this->height);
		args.setValues(this->bb);
	}
	return result;
}

int
RTEPage::resizeNecessaryP()
{
	return 0;
}

void 
RTEPage::conditionallyInitialiseCanvas()
{
}

void 
RTEPage::redisplay(XmDrawingAreaCallbackStruct *call_data)
{
}

void
RTEPage::drawRegionNow(iLBox refresh_box)
{
	ViewCanvas *vc = this->getCanvas();
	int i;
	iLBox draw_box;
	ElementSet *set;
	RTElement *elem;
	
	refresh_box.x -= vc->pix_hoffset;
	refresh_box.y -= vc->pix_voffset;
	iLBox visible_rect = visibleRectangle(this->mode->scrolled_window);
	i_IntersectLBox(refresh_box, visible_rect, &draw_box);
	Rectangle *r = new Rectangle(draw_box.x, draw_box.y, draw_box.w, draw_box.h);
	set = this->getChildrenInRectangle(r);
	delete r;
	for (i = 0; i < set->numElements(); i++) {
		elem = (RTElement*)set->getElementNumber(i);
		if (r->contains(elem))
			elem->drawElement();
		else if (r->overlaps(elem)) {
			Rectangle *intersect = r->intersection(elem);
			intersect->setX(intersect->getX() - elem->getX());
			intersect->setY(intersect->getY() - elem->getY());
			elem->drawRegion(intersect);
			delete intersect;
		}
	}
	delete set;
}

void 
RTEPage::addNewEntity(iEntityPtr entity)
{
}

void
centerIfNecessary(RTEPage *page, iEntityPtr entity)
{
}

iEntityPtr
RTEPage::findEnclosingEntity(int x, int y)
{
	halt("got findEnclosingEntity\n");
	return NULL;
}




void
RTEPage::centerOnSelected(iEntityPtr entity)
{
	RTElement *elem;
	int x, y;
	RTEUniEditor *uni_editor;
	int line, column;
	iEntityPtr par = entity;
	
	
	while (par && !(elem = (RTElement*)eaGetFace(par,vtRTE)))
		par = i_GetParentEntity(par);
	
	if (elem) {
		uni_editor = (RTEUniEditor*)elem->display_widget;
		uni_editor->convertDAFS2Line(entity,0,&line,&column);
		uni_editor->positionToCoord(line, column, cmBEFORE, &x, &y);
		uni_editor->centerOnCoord(x,y);
	}
	
}

int 
RTEPage::pressKey(iUCode key_value)
{
	// All pressKey events are passed to the RTElement which
	// currently has input focus.  Must use the mouse to set
	// focus to a different RTElement.

	return this->active_editor->pressKey(key_value);
}

void
RTEPage::pressButton(XEvent *event)
{
	halt("Got  a press button event\n");
}

void
RTEPage::releaseButton(XEvent *event)
{
  halt("RTEPage got release button.\n");

}

void
RTEPage::moveButton(int state, int x, int y)
{
}

/* these useless magnification functions are virtual requirements */
void 
RTEPage::setMag(int x, int ox)
{
}

void 
RTEPage::_setMag(int x, int ox)
{
}

void
RTEPage::createEditors(void)
{
	DLList *page_list = &this->mode->page_list;
	Ref r;
	int num_editors, show_cursor;
	iEntityPtr entity;
	RTEPage *page = 0;

	dllistCtor(&this->editor_list);

	this->findFieldEntities(this->root_entity, &this->editor_list);

	// count the editors in the list
	num_editors = dllistLength(&this->editor_list);

	if (this->mode->current_page_ref)
		page = (RTEPage*)dllistContents(page_list,this->mode->current_page_ref);
	show_cursor = (num_editors == 1 && page == this);

	for(r = dllistFirst(&this->editor_list); r; 
		  dllistNext(&this->editor_list, &r)){
		entity = (iEntityPtr) dllistContents(&this->editor_list, r);
		this->addMember(entity, show_cursor);
	}
	
}

static int childHasFieldP(iEntityPtr);
static int selfOrChildHasTextP(iEntityPtr);
static int selfOrChildHasFieldP(iEntityPtr);

static
int fieldP(iEntityPtr en)
{
	return getField(en);
	//|| (!childHasFieldP(en) && selfOrChildHasTextP(en));
}

static
int childHasFieldP(iEntityPtr en)
{
	// see if any descendant of en has fieldP
	
	iEntityPtr child;
	int result = FALSE;
	for(child = i_GetChildEntity(en); child && !result; child = i_GetNextEntity(child)) 
			result = selfOrChildHasFieldP(child);
	return result;
}

static
int selfOrChildHasFieldP(iEntityPtr en)
{
	// see if en or any of its descendants have fieldP
	
	return fieldP(en) || childHasFieldP(en);
}

// returns true if the entity contains data that text view will display
static
int textP(iEntityPtr en)
{
	iBox box = i_GetEnclosingBox(en);
	
	switch(i_GetEDataType(en)){
	case iDString:
	case iDUString:
	case iDPosSet:
	case iD1PosSet:
	  return TRUE;
	default:
	  break;
	}

	iTypeIndex type = i_GetEntityType(en);
	if((type == iSpaceT || type == iNewLineT) && box.w && box.h)
	  return TRUE;

	return FALSE;
}

static
int selfOrChildHasTextP(iEntityPtr en)
{
	// see if en or any of its descendants have text
	
	iEntityPtr child;	
	int result = textP(en);
	if(!result) {
		for(child = i_GetChildEntity(en); child && !result; child = i_GetNextEntity(child)) 
			result = selfOrChildHasTextP(child);
	}
	return result;
}


void
RTEPage::findFieldEntitiesSub(iEntityPtr entity, 
							 DLList *field_entity_list)
{
  Ref r;
  iEntityPtr child;

  if(fieldP(entity)) {
	r = dllistSeek(field_entity_list, entity);
	// avoid entering anything twice
	if(r == 0)
	  dllistAppend(field_entity_list, entity);
  } else {
	for(child = i_GetChildEntity(entity); child; child = i_GetNextEntity(child)){
	  this->findFieldEntitiesSub(child, field_entity_list);
	}
  }
}

void
RTEPage::findFieldEntities(iEntityPtr root_entity, 
						   DLList *field_entity_list)
{
  this->findFieldEntitiesSub(root_entity, field_entity_list);
  if(dllistEmpty(field_entity_list))
	dllistAppend(field_entity_list, root_entity);
}

void
RTEPage::setActiveEditor(RTEUniEditor *input_active_editor)
{
  if(this->active_editor && this->active_editor != input_active_editor)
	this->active_editor->clearSelection();
  this->active_editor = input_active_editor;
}



void
RTEPage::nextEntity()
{
  int state = 0;
  if(this->active_editor)
	this->active_editor->right(state);
  else
	ringTheBell();
}

void
RTEPage::prevEntity()
{
  int state = 0;
  if(this->active_editor)
	this->active_editor->left(state);
  else
	ringTheBell();

}


void
RTEPage::upEntity()
{
  int state = 0;
  if(this->active_editor)
	this->active_editor->up(state);
  else
	ringTheBell();

}

void
RTEPage::downEntity()
{
  int state = 0;
  if(this->active_editor)
	this->active_editor->down(state);
  else
	ringTheBell();
}


RTElement *
RTEPage::findElement(iEntityPtr entity)
{
  RTElement *rvalue = 0;
  while(entity) {
	rvalue = (RTElement *) eaGetFace(entity, vtRTE);
	if(rvalue)
	  break;
	entity = i_GetParentEntity(entity);
  }
  return rvalue;
}

void
RTEPage::manage()
{
	DLList *list = this->getSM()->getItems();
	Ref r = dllistFirst(list);

	iEntityPtr entity;
	RTElement *elem;
	if (r) {
	  entity = (iEntityPtr)dllistContents(list, dllistFirst(list));
	  elem = this->findElement(entity);
	}
	if(!r || !elem) {
	  entity = (iEntityPtr) dllistContents(&editor_list, dllistFirst(&editor_list));
	  elem = this->findElement(entity);	
	}
	elem->showCursor();

	RTEMode *mode = (RTEMode*)this->mode;
	this->Page::manage();
	XtManageChild(this->bb);
	XtManageChild(mode->page);

	//bbogus do we need to do this??
	this->mode->main_window->refresh_dialogs();

	//XmProcessTraversal(this->getCanvas()->canvas, 0);
	//this->getCanvas()->setCursor(XC_xterm);
}

void
RTEPage::unmanage()
{
  // we either have legacy or our IllEditor has set things properly
  this->getSM()->convertToItem();
  

  // we are doomed but we might as well do this
  //this->clearPopup();
	RTEMode *mode = (RTEMode*)this->mode;
	this->Page::unmanage();
	XtUnmanageChild(this->bb);
	XtUnmanageChild(mode->page);
}


void
RTEPage::processDelete(void)
{
  // bogus -- what about the case when the selection is not on the
  // current page
  if(this->active_editor)
	this->active_editor->processDelete();
  else
	ringTheBell();
}



void
RTEPage::findNextFlagEntity()
{	
	halt("got next flag entity in RTEPage\n");
}


#if 0
/*
// This is a concept of John's
// that hasn't really been implemented
*/
int 
RTEPage::isVisible(iEntityPtr entity)
{
#if 0
	short x,y;
	Widget sw = this->mode->scrolled_window;
	iBox inner_bounds;
	iBox outer_bounds;
	iBox view_bounds;

	short w, h;
	int visible;
	
	element->getPosition(this->td, 0, x, y);
	element->getDimensions(w, h);

	visible = visiblePositionP(sw, x, y) && visiblePositionP(sw, x+w, y+h);
	if(this->display_popup){
		// would the popup be visible too?
		this->decideImageBounds(element->entity, 
								&inner_bounds,
								&outer_bounds);

		view_bounds =
			this->decideViewBounds(element->entity, 
								   outer_bounds);
		visible = (visible && 
				   visiblePositionP(sw, view_bounds.x, view_bounds.y) && 
				   visiblePositionP(sw, 
									view_bounds.x + view_bounds.w, 
									view_bounds.y + view_bounds.h));
	}

	return visible;
#endif
}
	
#if 0
static int 
entityVisible(iEntityPtr entity)
{
	int is_visible = FALSE;
	TextElement *element = (TextElement *) getTextViewer(entity);
	if(element) {
		TextPage *page = (TextPage *) element->page;
		is_visible = page->isVisible(element);
	}
	return is_visible;
}
#endif

/*
// this is function is necessary because the selection could be at
// the beginning of a word or a paragraph that is selected
*/
iEntityPtr 
findStartEntity(iEntityPtr from_entity)
{
	iEntityPtr start_entity;
	iEntityPtr f, e;
	int fi, ei;
	iEntityPtr curr;


	for(start_entity = from_entity, 
			curr = from_entity ? i_GetParentEntity(from_entity) : 0; 
		curr; 
		curr = i_GetParentEntity(curr)){

		i_Object2Range(curr, 
					   &f, (int4 *) &fi, 
					   &e, (int4 *) &ei);
		if(f == from_entity)
			start_entity = curr;

	}

	return start_entity;
}

void
TextPage::findNextFlagEntitySub()
{
	// is there already an entity stored in the popup
	iEntityPtr from_entity;
	iEntityPtr to_entity;
	int from_index, to_index;
	int object_from_index, object_to_index;
	iEntityPtr flagged_entity = 0;
	int allow_off_page;
	
	iEntityPtr begin_doc_entity;
	iEntityPtr start_entity;
	int had_popup_before = FALSE;

	this->getSM()->getRange(from_entity, from_index, to_entity, to_index);	

	// find the existing popup and clear it
	if(this->flag_entity && from_entity) {
		// so we must compare the bbox of elements
		// and we can't extract from the flag entity until
		// we are sure it hasn't been deleted
		int ancestorP(iEntityPtr child, iEntityPtr would_be_ancestor);
		if(from_entity == this->flag_entity || ancestorP(from_entity, this->flag_entity)){
			iEntityPtr ffe, fte;
			int ffi, fti;
			i_Object2Range(this->flag_entity, 
						   &ffe, (int4 *) &ffi, 
						   &fte, (int4 *) &fti);
			if(ffe == from_entity &&
			   ffi == from_index &&
			   fte == to_entity &&
			   fti == to_index) {
				// before this, the flag was selected using ctrl-g
				// and wasn't edited
				// therefore, we clear this flag and continue from here
				
				// must happen before we clear the popup
				iEntityPtr holdit = this->flag_entity;
				this->clearPopup();
				entityClearFlag(this->mode->main_window, holdit);
				had_popup_before = TRUE;
			}
		}
	}
	begin_doc_entity = this->td->entity;

	// we start our search in the nearest parent
	// who would have the same from_entity
	start_entity = findStartEntity(from_entity);
	// now, if we have no from entity, start at root

	if(start_entity == 0)
		// this is ok because we won't be needing
		// it in the reset last flagged guy
		// at the end
		from_entity = begin_doc_entity;

	for(int i = 0; flagged_entity == 0 && i < 4; i++) {
		// 4 times through
		iEntityPtr loop;
		
		loop = start_entity;

		if(start_entity == begin_doc_entity && (i == 0 || i == 2))
			continue; // the succeeding loop will do this anyway

		if(i == 1 || i == 3) // odd times -- start from the beginning
			loop = begin_doc_entity;

		allow_off_page = (i == 2 || i == 3); // last 2 times allow off page

		
		// now, if the current start_entity is still a flagged entity
		// we select
		if(loop && entityFlaggedP(loop) && (entityVisible(loop) || allow_off_page)) {
			flagged_entity = loop;
		} else { // do an iEntityFindForward
			while(loop) {
				loop = 
					i_FindEntityForward(loop,0, PROPERTY, (void*)FLAG_PROP);
				if(loop && (entityVisible(loop) || allow_off_page))
					break;
			}
			flagged_entity = loop; // loop may be null
		}
	}

	this->flag_entity = flagged_entity;

	if(flagged_entity){
		i_Object2Range(flagged_entity, 
					   &start_entity, (int4 *) &from_index,
					   &to_entity, (int4 *) &to_index);
		from_index = object_from_index;
		to_index = object_to_index;
		this->getSM()->setRangeAdjacent(start_entity, from_index, 
								to_entity, to_index);
		if(entityVisible(flagged_entity) == FALSE) {
		   TextElement *element = (TextElement *) getTextViewer(flagged_entity);
		   if(element)
			   this->centerCursor(element);		
		}
	}

	if(flagged_entity == 0 && had_popup_before){
		// we just cleared the last flag, and don't want anything
		// to be selected now
		// heinous - need to reset selection so it will get refreshed when cleared
		this->getSM()->setRange(from_entity, from_index, to_entity, to_index);
		this->getSM()->clear();
	}
}


void
TextPage::findNextFlagEntity()
{	
	this->findNextFlagEntitySub();
	this->setPopup(this->flag_entity);
}

#if 0
#define DEFAULT_POPUP -1 // popup is shown as default

	  display_popup(DEFAULT_POPUP),
	  popup_image(0)
	this->clearPopup();
	this->setPopup(this->flag_entity);
#endif


/*
// popup stuff
*/

void 
TextPage::clearPopup()
{
	// find the popup
	// order the area to redraw
	if(this->display_popup && this->bounds_on_view.w){
		this->addRedrawRegion(&this->bounds_on_view);
	}
	this->bounds_on_view.w = 0;
	this->outer_image_bounds.w = 0;
	this->inner_image_bounds.w = 0;
	this->popup_image = 0;
	this->flag_entity = 0;
}

void
TextPage::setPopup(iEntityPtr entity)
{
	// if there is a popup area already, clear it
	// set the popup
	// order the area to redraw
	this->clearPopup();
	if(entity){
		this->flag_entity = entity;
		this->decideImageBounds(entity, 
								&this->inner_image_bounds,
								&this->outer_image_bounds);

		this->bounds_on_view = 
			promoteBox(this->decideViewBounds(entity, 
								   this->outer_image_bounds));
		this->popup_image = i_GetImage(entity);
		if(this->display_popup && this->bounds_on_view.w && this->popup_image){
			this->addRedrawRegion(&this->bounds_on_view);
		}
	}
}

void
TextPage::refreshPopup(Display *display, Pixmap pixmap, int off_x, int off_y)
{
	ViewCanvas *vc = this->getCanvas();
	Widget widget = vc->canvas;

	char *bits1 = 0;
	int2 dw1;
	XImage *x_image1 = 0;

	char *bits2 = 0;
	int2 dw2;
	XImage *x_image2 = 0;

	GC gc;

	iBox ibox = this->inner_image_bounds;
	iBox obox = this->outer_image_bounds;

	// actually draw

	int normal_image_mag_exp = 1 << 16;
	if (i_Image2BitMap(this->popup_image, obox,
					   normal_image_mag_exp, &dw1, &bits1)) 
		goto clean_images;
	if (i_Image2BitMap(this->popup_image, ibox,
					   normal_image_mag_exp, &dw2, &bits2))
		goto clean_images;

	// create the image to be the size of the bitmap
	x_image1 = XCreateImage(display, 0, 1, XYBitmap, 
							0, // tricky -- ignore initial bits
							bits1,
							obox.w, 
							obox.h, 
							8, // quanta of scan line
							dw1);

	x_image2 = XCreateImage(display, 0, 1, XYBitmap, 
							0, // tricky -- ignore initial bits
							bits2,
							ibox.w, 
							ibox.h, 
							8, // quanta of scan line
							dw2);

	x_image1->byte_order = MSBFirst;
	x_image1->bitmap_bit_order = MSBFirst;
	x_image2->byte_order = MSBFirst;
	x_image2->bitmap_bit_order = MSBFirst;
	_XInitImageFuncPtrs(x_image1);
	_XInitImageFuncPtrs(x_image2);

	gc = vc->darkgrayGC();

	XPutImage(display, pixmap,
			  gc, x_image1, 
			  0,0, 
			  off_x, off_y,
			  obox.w, obox.h);

	gc = vc->blackGC();

	XPutImage(display, pixmap,
			  gc, x_image2, 
			  0,0, 
			  off_x + ibox.x - obox.x, off_y + ibox.y - obox.y,
			  ibox.w, ibox.h);


	// this is the border
	XDrawLine(display, pixmap, vc->blackGC(),
			  off_x + 0,         off_y + obox.h-1,
			  off_x + obox.w-1,  off_y + obox.h-1);
	XDrawLine(display, pixmap, vc->blackGC(),
			  off_x + obox.w-1, off_y + 0, 
			  off_x + obox.w-1, off_y + obox.h-1);

	XDrawRectangle(display, pixmap, 
				   vc->blueGC(),
				   off_x + 0, off_y + 0, 
				   obox.w-2, obox.h-2); 

clean_images :
		if(bits1){
			smfree(bits1);
		}
	if(x_image1) {
		x_image1->data=0;
		XDestroyImage(x_image1);
	}

	if(bits2){
		smfree(bits2);
	}
	if(x_image2) {
		x_image2->data=0;
		XDestroyImage(x_image2);
	}
}

void
TextPage::decideImageBounds(iEntityPtr entity, 
							 iBox *inner_bounds, iBox *outer_bounds)
{
	iBox ibox = i_GetEnclosingBox(entity);
	iBox obox = ibox;
	iImagePtr image = i_GetImage(entity);

	if(image == 0) {
		obox.w = 0;
		ibox.w = 0;
	} else {
		int iw, ih;
		// no scaling because we always display in mag 1
		iw = i_GetImageWidth(image);
		ih = i_GetImageHeight(image);
		if (obox.x>IMAGE_FRINGE){
			obox.x-=IMAGE_FRINGE;
		} 	else{
			obox.x=0;
		}

		if (obox.y>IMAGE_FRINGE){
			obox.y-=IMAGE_FRINGE;
		} else {
			obox.y = 0;
		}
		obox.w = MIN(iw-obox.x, obox.w+2*IMAGE_FRINGE);
		obox.h = MIN(ih-obox.y, obox.h+2*IMAGE_FRINGE);
	}

	*inner_bounds = ibox;
	*outer_bounds = obox;
}

iBox 
TextPage::decideViewBounds(iEntityPtr entity, iBox image_bounds)
{
	short x, x1, x2, y;
	short w, h;
	iBox b;

	iEntityPtr beg_ent, end_ent;
	int beg_ind, end_ind;

	i_Object2Range(entity, 
				   &beg_ent, (int4 *) &beg_ind, 
				   &end_ent, (int4 *) &end_ind);
	TextElement *begin = (TextElement*)getTextViewer(beg_ent);
	// cause sometime object to range return 0 in
	// that second slot -- bogus
	end_ent = end_ent ? end_ent : beg_ent;
	TextElement *end = (TextElement*)getTextViewer(end_ent);
	begin->getPosition(this->td, beg_ind, x1, y);
	end->getPosition(this->td, end_ind, x2, y);
	begin->getDimensions(w, h);

	w = image_bounds.w;
	x = x1 + ((x2-x1) / 2) - (w/2);
	y += h;
	h = image_bounds.h;

	// hmm should check if both are being violated at once
	if(x < DISTANCE_FROM_EDGE)
		x = DISTANCE_FROM_EDGE;
	if(x + w > this->width - DISTANCE_FROM_EDGE)
		x = this->width - w - DISTANCE_FROM_EDGE;

	b.x = x;
	b.y = y;
	b.w = w;
	b.h = h;

	return b;
}


void 
TextPage::togglePopup(void)
{	
	XArg args;
	iBox box;
	MainWindow *mw=this->mode->main_window;

	this->display_popup = !this->display_popup;
	if (this->flag_entity){
		args.reset();
		XtVaGetValues(mw->scroll_window, 
			XmNheight, &box.h, XmNwidth, &box.w, NULL);

		/* refresh screen  */
		// bogus -- we know what to redraw!
		this->drawRegion(this->bounds_on_view);
	}
}

#endif
