
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef _RTEPage_h_
#define _RTEPage_h_

#include "Page.h"

#include "rte/ConstrainedRectangle.h"

#include "misc/misc.h"
class RTElement;
class RTEUniEditor;

class RTEPage : 
	public Page,
	public ConstrainedRectangle
{
private:
  RTElement *input_focus_region;
  int magnification;  // divide image coords by this to get screen coords
  float x_expansion_factor, y_expansion_factor;
  RTEUniEditor *active_editor;
	
public:
  iEntityPtr root_entity;
  DLList editor_list;
  Widget bb;
  int initialized_canvas;
  RTEPage(Mode *input_mode, const aString& label,iEntityPtr root_entity);
  void addMember(iEntityPtr member, int show_cursor=FALSE);
  void createEditors(void);
  void expandEditors(void);
  void resetExpansionFactor(void) { this->x_expansion_factor = this->y_expansion_factor = 1.0; };
  virtual ~RTEPage();
  void init(iEntityPtr en);

  RTElement *findElement(iEntityPtr entity);
  RTElement *getInputFocusElement(void) { return this->input_focus_region; };

  /*virtual*/ int changeSizeTo(Coord w, Coord h, int redraw=TRUE);
	
  /*virtual*/ void conditionallyInitialiseCanvas();
  /*virtual*/ void redisplay(XmDrawingAreaCallbackStruct *call_data);
  /*virtual void resize(XmDrawingAreaCallbackStruct *call_data); */
  /*virtual*/ iEntityPtr findEnclosingEntity(int x, int y);
	
  /*virtual*/ void addNewEntity(iEntityPtr entity);

  /*virtual*/ void findNextFlagEntity(void);
	 
  /*virtual*/ void centerOnSelected(iEntityPtr entity);
  /*virtual*/ void scrollToUpperLeft(iEntityPtr entity){};
  /*virtual*/ void scrollToTop(iEntityPtr entity){};
  /*virtual*/ void scrollToBottom(iEntityPtr entity){};
  /*virtual*/ void scrollToLeft(iEntityPtr entity){};
  /*virtual*/ void scrollToRight(iEntityPtr entity){};

  // actually ignored

  /*virtual*/ void setMag(int new_mag, int old_mag);
  /*virtual*/ void _setMag(int new_mag, int old_mag);
	
  // for delayed drawing
  /*virtual*/ int resizeNecessaryP(void);
  /*virtual*/ void drawRegionNow(iLBox box);

  /*virtual void setPicture(iLBox b); */
  /*virtual iLBox getPicture(void); */

  /*virtual*/ int pressKey(iUCode key_value);
	
  /*virtual*/ void         pressButton(XEvent *event);
  /*virtual*/ void         releaseButton(XEvent *event);
  /*virtual*/ void         moveButton(int state, int x, int y);


  // correctly implemented from here down
  /*virtual*/ void prevEntity();
  /*virtual*/ void nextEntity();
  /*virtual*/ void upEntity();
  /*virtual*/ void downEntity();

  /*virtual*/ void manage(void);
  /*virtual*/ void unmanage(void);

  // local nonsense 
  void findFieldEntitiesSub(iEntityPtr entity, DLList *entity_list);
  void findFieldEntities(iEntityPtr root_entity, DLList *entity_list);

  void setActiveEditor(RTEUniEditor *active_editor);
  void processDelete(void);
};

#endif

