
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef _RTElement_h_
#define _RTElement_h_

#include "rte/Rectangle.h"
#include "text/EntityFace.h"
#include "lte/UniEditor.h"

class RTEPage;
class IllCanvas;

class RTElement
	: public ChildRectangle,
	 public EntityFace {
protected:
	iEntityPtr entity;
	UniEditor *display_widget;
public:
  IllCanvas *canvas;
	RTEPage *page;
	RTElement(RTEPage *page,iEntityPtr e, Coord x, Coord y, Coord w, Coord h, float *x_expansion_factor, float *y_expansion_factor );
	~RTElement();
	/*virtual*/ int drawElement();
	/*virtual*/ int drawRegion(Rectangle *r);
	/*virtual*/ int drawBorder();
	/*virtual*/ int moveTo(Coord x, Coord y, int squeeze=FALSE, int allow_overlap=FALSE);
	/*virtual*/ int changeSizeTo(Coord w, Coord h, int redraw=TRUE);
	void setPosition(Position x, Position y);
	void setSize(Dimension w, Dimension h);
	void showCursor();
	int pressKey(iUCode key_value);
	int pressButton(XEvent *event);
	int releaseButton(XEvent *event);
	/*virtual*/ void deleteAndUpdate();
	/*virtual*/ void deleteCheap();
	
	/* callbacks -- EntityFace virtuals */
	/*virtual*/ void create();
	
	/*virtual*/ void willUpdate();
	/*virtual*/ void update();
	
	/*virtual*/ void willMove();
	/*virtual*/ void move();
	
	/*virtual*/ void destroy();
	
	/*virtual*/ void select();
	/*virtual*/ void unSelect();
	
	/*virtual*/ void recognize();
	
	/*virtual*/ MainWindow *getMainWindow();
	
	friend class RTEPage;
  friend class RTEUniEditor;
};

#endif
