
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"

#include "MainWindow.h"
#include "Document.h"

#include "SelectionManager.h"

#include "EntityAux.h"


#define CONTROL_POINTS foo

/*
// questions
// how come we need range_selected and sc?
// are the to and from entities stored in the selection list?
//     apparently not -- you can have a to and from without
//     having those entities selected -- they are implied
//     by having a bigger object selected
// what happens in these situations
// one item is selected in image mode
//     switch to text mode
// more than one item selected in image mode
//     switch to text mode
// range selected in text mode
//     switch to image mode
*/


SelectionManager::SelectionManager(MainWindow *input_mw)
	: mw(input_mw),
	  from_entity(0),
	  from_index(-1),
	  to_entity(0),
	  to_index(-1),
	  sc(scNOT),
	  dirty(FALSE),
	  range_selected(FALSE),
	  legacy_from_entity(0),
	  legacy_from_index(-1),
	  legacy_to_entity(0),
	  legacy_to_index(-1)
{
	dllistCtor(&this->selection);
	dllistCtor(&this->legacy_selection);
}

void
SelectionManager::clearLegacy()
{
	dllistClear(&this->legacy_selection);
	legacy_from_entity = 0;
	legacy_from_index = -1;
	legacy_to_entity = 0;
	legacy_to_index = -1;

}

void
SelectionManager::convertToRange(void)
{
  if(this->legacy_from_entity) {
	// the legacy comes from text mode
	// so we must restore it
	this->setRange(this->legacy_from_entity,
				   this->legacy_from_index,
				   this->legacy_to_entity,
				   this->legacy_to_index);
	this->legacy_from_entity = this->from_entity;
	this->legacy_from_index = this->from_index;
	this->legacy_to_entity = this->to_entity;
	this->legacy_to_index = this->to_index;
  } else {
	// we have no legacy from text mode
	// so we must store our current non-text legacy
	// and convert a range from one of the selected entities
	DLList temp_list;
	dllistCtor(&temp_list);
	dllistAssign(&temp_list, &this->selection);
	iEntityPtr e;
	iEntityPtr input_from_entity;
	iEntityPtr input_to_entity;
	int input_from_index;
	int input_to_index;

	const DLList *l = &this->selection;
	Ref r = dllistFirst(l);
	int ancestorP(iEntityPtr child, iEntityPtr would_be_ancestor);

	// we will only use the first item in the selection
	if(r) {
	  e = (iEntityPtr) dllistContents(l, r);
	  if(_iAncestorIsData(e)) {
		// I guess we can't convert to a range
		this->clear();
	  } else {
		i_Object2Range(e,
					   &input_from_entity, (int4 *) &input_from_index,
					   &input_to_entity, (int4 *) &input_to_index);
		this->clear();
		// object to range can screw up and return something
		// that is not a descendant of e
		if(input_from_entity && 
		   input_to_entity && 
		   (input_from_entity == e || ancestorP(input_from_entity, e))
		   ) 
		  { 
			// must make sure both of these guys have
			// entityFaces
			
			// bogus how could this not be the case
			// I am going to remove these tests
			// This has to work for the LTE Mode as well anyway
#if 0
			TextElement *e1;
			TextElement *e2;
			e1 = (TextElement *) 
			  eaGetFace(input_from_entity, vtTEXT);
			e2 = (TextElement *) 
			  eaGetFace(input_to_entity, vtTEXT);
			if(e1 && e2)
			  this->setRange(input_from_entity, 
							 input_from_index, 
							 input_to_entity, 
							 input_to_index);
#else
			this->setRange(input_from_entity, 
						   input_from_index, 
						   input_to_entity, 
						   input_to_index);
#endif
		  }
	  }
	}
	dllistAssign(&this->legacy_selection, &temp_list);
	dllistDtor(&temp_list);
  }

}

void
SelectionManager::convertToItem(void)
{
	if(dllistEmpty(&this->legacy_selection) == FALSE) {
		// the legacy comes from image mode
		Ref r;
		VoidPtr c;
		DLList temp_list;
		dllistCtor(&temp_list);
		dllistAssign(&temp_list, &this->legacy_selection);
		this->clear();
		
		for(r = dllistFirst(&temp_list); r; dllistNext(&temp_list, &r)){
			c = dllistContents(&temp_list, r);
			this->appendItem((iEntityPtr) c);
		}

		dllistAssign(&this->legacy_selection, &temp_list);
		dllistDtor(&temp_list);
	} else {
		iEntityPtr temp_from_entity;
		int temp_from_index;
		iEntityPtr temp_to_entity;
		int temp_to_index;
		temp_from_entity = this->from_entity;
		temp_from_index = this->from_index;
		temp_to_entity = this->to_entity;
		temp_to_index = this->to_index;

		// set the legacy to that of text mode
		iEntityPtr e;

		if(this->from_entity) {
			e = i_Range2Object(this->from_entity, this->from_index,
							   this->to_entity, this->to_index);
			this->clear();
			if(e)
				this->setItem(e);
		}
		this->legacy_from_entity = temp_from_entity;
		this->legacy_from_index = temp_from_index;
		this->legacy_to_entity = temp_to_entity;
		this->legacy_to_index = temp_to_index;
	}
}


static iEntityPtr
i_Range2TreeObject(iEntityPtr bp,int4 beg,iEntityPtr ep,int4 end)
{
	iEntityPtr a,b;
	
	if (bp==ep) 
	  return bp;
	if(end==0)
		ep=i_FindEntityBack(ep,0,DATAEN,0);
	
	for(a=bp;a;a=i_GetParentEntity(a))
		for(b=ep;b;b=i_GetParentEntity(b))
			if(a==b){
			  return a;
			}
	return 0;
}

void
SelectionManager::convertToTreeItem(void)
{
  if(dllistEmpty(&this->legacy_selection) == FALSE) {
	// do nothing -- we are already returning to legacy
  } else {
	// reconvert the legacy from text mode
	iEntityPtr e;

	if(this->legacy_from_entity) {
	  e = i_Range2TreeObject(this->legacy_from_entity, this->legacy_from_index,
						 this->legacy_to_entity, this->legacy_to_index);
	  this->clear();
	  if(e)
		this->setItem(e);
	}
  }
}


SelectionManager::~SelectionManager()
{
	dllistDtor(&this->selection);
}


void
SelectionManager::_clear()
{
	DLList *l = &this->selection;
	Ref r;
	iEntityPtr entity;
	
	for(r = dllistFirst(l); r; dllistNext(l, &r)){
		entity = (iEntityPtr) dllistContents(l, r);
		clearEntitySelection(entity);
	}
	
	dllistClear(&this->selection);
	
	if (this->from_entity)
		clearEntitySelection(this->from_entity);
	if (this->to_entity)
		clearEntitySelection(this->to_entity);
	
	
	// bogus they won't be in the selection list
	// how come I don't have to send a message to these guys
	this->from_entity = 0;
	this->to_entity = 0;
	this->from_index = -1;
	this->to_index = -1;
	this->sc = scNOT;
}

int
SelectionManager::rangeSelectedP()
{
	return this->range_selected;
}

int
SelectionManager::selectedItemP(iEntityPtr item)
{
	return getEntitySelection(item);
}

/* 
// bogus -- how come there isnt a range option
*/
SelectionClass
SelectionManager::getState()
{
	if(dllistEmpty(&this->selection)) {
		if(this->from_entity)
			return scRANGE;
		else 
			return scNOT;
	} else
		return scITEM;
}


/************
// The gets
************/


/* 
// why dont we check range_selected
*/
void
SelectionManager::getRange(iEntityPtr& output_from_entity, 
						   int& output_from_index,
						   iEntityPtr& output_to_entity,   
						   int& output_to_index)
{
	if (this->from_entity) {
		output_from_entity = this->from_entity;
		output_from_index  = this->from_index;
		output_to_entity   = this->to_entity;
		output_to_index    = this->to_index;
	} else {
		DLList *l = &this->selection;
		Ref r=dllistFirst(l);
		if (!r) {
			output_from_entity = 0;
			output_from_index  =-1;
			output_to_entity   = 0;
			output_to_index    =-1;
		} else {
			iEntityPtr entity = (iEntityPtr) dllistContents(l, r);
			i_Object2Range(entity,
						   &output_from_entity,(int4 *) &output_from_index,
						   &output_to_entity,(int4 *) &output_to_index);
		}
		
	}
}

const DLList *
SelectionManager::getItems()
{
	return &this->selection;
}


iEntityPtr
SelectionManager::getSoleItem()
{
	DLList *l = &this->selection;
	return dllistLength(l)==1?(iEntityPtr)dllistContents(l,dllistFirst(l)):0;
}

/******************************
// the puts
******************************/


void
SelectionManager::appendItem(iEntityPtr item)
{
	this->clearLegacy();
	DLList refresh_list;
	
	if(this->selectedItemP(item) == FALSE){
		dllistCtor(&refresh_list);
		dllistAppend(&refresh_list, item);
		
		dllistAppend(&this->selection, item);
		setEntitySelection(item);
		
		// good -- I'm setting the selection before calling the refresh
		this->mw->postRefreshItems(&refresh_list);
		dllistDtor(&refresh_list);

		// we are the only selected and this is equivalent to a setSelection
		if(dllistLength(&this->selection) == 1)
			this->mw->setModeSelection(item,0);

		// we are not the only one selection
		if(dllistLength(&this->selection) != 1)
			this->mw->setModeSelection(0,0);

	}
	this->mw->setSensitive();
}

/*
// how come there is no refresh for the ends of the range?
*/
void
SelectionManager::clear()
{
	this->clearLegacy();
	DLList temp;
	DLList *refresh_list = &temp;
	
	dllistCtor(refresh_list);
	
	if(dllistLength(&this->selection) != 0 || this->from_entity != 0) {
		dllistAssign(refresh_list, &this->selection);
		// good -- I'm calling the refresh before clearing the setting
		// bogus
		this->_clear();
		this->mw->preRefreshItems(refresh_list);
	}
	
	dllistDtor(refresh_list);
	this->mw->setModeSelection(0,0);
	this->mw->setSensitive();
}

/*
// set selection to 0
// is the same as clear selection
*/
void
SelectionManager::setItem(iEntityPtr item)
{
	this->clearLegacy();
	if(item == 0) {
		this->clear();
	} else {
		// why not call clear()?
		// because we might be the one selected and we dont want to flicker
		// or we might be range-selected
		DLList refresh_list;
		if(this->selectedItemP(item) == FALSE || 
		   dllistLength(&this->selection) != 1 || 
		   getEntitySelection(item) == scRANGE){
			dllistCtor(&refresh_list);
			dllistAssign(&refresh_list, &this->selection);
			// good setting refresh before clearing selection
			// bogus
			this->_clear();
			
			this->mw->preRefreshItems(&refresh_list);
			
			dllistClear(&refresh_list);
			
			if(item) {
				// good -- setting selection before sending refresh
				setEntitySelection(item);
				dllistAppend(&this->selection, item);	 
				
				dllistAppend(&refresh_list, item);
				this->mw->postRefreshItems(&refresh_list);
			}
			dllistDtor(&refresh_list);
		}
		this->range_selected=FALSE;
		this->mw->setModeSelection(item,0);
		this->mw->setSensitive();
	}
}

void
SelectionManager::setRange(iEntityPtr input_from_entity, int input_from_index, 
						   iEntityPtr input_to_entity,   int input_to_index)
{	
	this->clearLegacy();
	DLList refresh_list;
	/*
// used to not refresh if item was already selected
// bogus -- should not refresh if the cursor pos
// won't change
*/
	
	dllistCtor(&refresh_list);
	if(this->from_entity)
		dllistAppend(&refresh_list, this->from_entity);
	else
		dllistAssign(&refresh_list, &this->selection);
	
	// good sending refresh before clearing selection
	// one reason why clear() doesnt handle range selections
	// nobody seems to call clear when a range is selected
	this->_clear();
	this->mw->preRefreshItems(&refresh_list);
	
	dllistClear(&refresh_list);
	//already done in _clear()
	//dllistClear(&this->selection);
	
	iEntityPtr entity = 0;
	iEntityPtr object = 0;

	if(input_from_entity != input_to_entity || input_from_index != input_to_index) {
		// ignore thin cursor
		object = i_Range2Object(input_from_entity,input_from_index,
								input_to_entity,input_to_index);
	}

#if 0	
	if (input_from_entity!=0) {
		entity=i_Range2Object(input_from_entity,input_from_index,
							  input_to_entity,input_to_index);
		if (entity && entity != input_from_entity && entity != input_to_entity) {
			setEntitySelection(entity);
			dllistAppend(&this->selection, entity);
			dllistAppend(&refresh_list,entity);
		}
	}
#endif

	// dont override the object selection
	if(input_from_entity != input_to_entity) {

		if(input_from_entity != entity)
			setEntitySelection(input_from_entity, input_from_index, -1);
		else
			setEntitySelection(input_from_entity, scITEM, input_from_index, -1);

		if(input_to_entity != entity)
			setEntitySelection(input_to_entity,   -1, input_to_index);
		else 
			setEntitySelection(input_to_entity, scITEM, -1, input_to_index);

	} else {
		if(input_from_entity != entity)
			setEntitySelection(input_from_entity, input_from_index, input_to_index);
		else
			setEntitySelection(input_from_entity, scITEM, input_from_index, input_to_index);
	}

		
	
	// try to refresh a guy who might have had a control point
#ifdef CONTROL_POINTS
	dllistAppend(&refresh_list,    input_from_entity);
	if(from_entity != to_entity) {
		dllistAppend(&refresh_list,    input_to_entity);
	}
#endif
	
	this->from_entity = input_from_entity;
	this->to_entity =   input_to_entity;
	this->from_index =  input_from_index;
	this->to_index =    input_to_index;
	this->range_selected=TRUE;
	
	// good sending refresh after setting selection
	this->mw->postRefreshItems(&refresh_list);
	dllistDtor(&refresh_list);

	if(object)
		this->mw->setModeSelection(object, input_from_index);
	else
		this->mw->setModeCursor(this->from_entity, this->from_index);
	this->mw->setSensitive();
}

void
SelectionManager::setRangeAdjacent(iEntityPtr input_from_entity, int input_from_index, 
						   		   iEntityPtr input_to_entity,   int input_to_index)
{	
	// Special version of setRange for use by text mode popups.
	// To be used when the input_from_entity and input_to_entity
	// are adjacent.  In this case, the TextElement::select() callback
	// does not need to redraw the whole window. See text mode PR 213.
	
	this->clearLegacy();
	DLList refresh_list;
	/*
// used to not refresh if item was already selected
// bogus -- should not refresh if the cursor pos
// won't change
*/
	
	dllistCtor(&refresh_list);
	if(this->from_entity)
		dllistAppend(&refresh_list, this->from_entity);
	else
		dllistAssign(&refresh_list, &this->selection);
	
	// good sending refresh before clearing selection
	// one reason why clear() doesnt handle range selections
	// nobody seems to call clear when a range is selected
	this->_clear();
	this->mw->preRefreshItems(&refresh_list);
	
	dllistClear(&refresh_list);
	//already done in _clear()
	//dllistClear(&this->selection);
	
	iEntityPtr entity = 0;
	iEntityPtr object = 0;

	if(input_from_entity != input_to_entity || input_from_index != input_to_index) {
		// ignore thin cursor
		object = i_Range2Object(input_from_entity,input_from_index,
								input_to_entity,input_to_index);
	}

	// dont override the object selection
	if(input_from_entity != input_to_entity) {

		// The select() callback is hardwired to look for from_index or
		// to_index == -1, so using -2 avoids call to redrawWindow.
		if(input_from_entity != entity)
			setEntitySelection(input_from_entity, input_from_index, -2);
		else
			setEntitySelection(input_from_entity, scITEM, input_from_index, -2);

		if(input_to_entity != entity)
			setEntitySelection(input_to_entity,   -2, input_to_index);
		else 
			setEntitySelection(input_to_entity, scITEM, -2, input_to_index);

	} else {
		if(input_from_entity != entity)
			setEntitySelection(input_from_entity, input_from_index, input_to_index);
		else
			setEntitySelection(input_from_entity, scITEM, input_from_index, input_to_index);
	}

		
	
	// try to refresh a guy who might have had a control point
#ifdef CONTROL_POINTS
	dllistAppend(&refresh_list,    input_from_entity);
	if(from_entity != to_entity) {
		dllistAppend(&refresh_list,    input_to_entity);
	}
#endif
	
	this->from_entity = input_from_entity;
	this->to_entity =   input_to_entity;
	this->from_index =  input_from_index;
	this->to_index =    input_to_index;
	this->range_selected=TRUE;
	
	// good sending refresh after setting selection
	this->mw->postRefreshItems(&refresh_list);
	dllistDtor(&refresh_list);

	if(object)
		this->mw->setModeSelection(object, input_from_index);
	else
		this->mw->setModeCursor(this->from_entity, this->from_index);
	this->mw->setSensitive();
}


void
SelectionManager::toggleItem(iEntityPtr item)
{
	this->clearLegacy();
	Ref r;
	iEntityPtr entity;
	DLList *l = &this->selection;
	iEntityPtr sole_item;
	DLList refresh_list;
	
	if(this->selectedItemP(item) == FALSE) {
		// special hack to remove the control points
#ifdef CONTROL_POINTS
		sole_item = this->getSoleItem();
		if(sole_item){
			dllistCtor(&refresh_list);
			dllistAppend(&refresh_list, sole_item);
		}
#endif
		this->appendItem(item);
#ifdef CONTROL_POINTS
		if(sole_item){
			this->mw->postRefreshItems(&refresh_list);
			dllistDtor(&refresh_list);
		}
#endif
	} else {
		DLList scratch_list;
		
		dllistCtor(&refresh_list);
		dllistAppend(&refresh_list, item);
		// good send the refresh before clearing the selection
		//bogus
		
		clearEntitySelection(item);
		// make up a list that is the seleciton minus this guy
		dllistCtor(&scratch_list);
		for(r = dllistFirst(l); r; dllistNext(l, &r)){
			entity = (iEntityPtr) dllistContents(l, r);
			if(entity != item)
				dllistAppend(&scratch_list, entity);
		}
		dllistAssign(l, &scratch_list);
		
#ifdef CONTROL_POINTS
		sole_item = this->getSoleItem();
		if(sole_item)
			dllistAppend(&refresh_list, sole_item);
#endif
		this->mw->preRefreshItems(&refresh_list);
		
		dllistDtor(&refresh_list);
		dllistDtor(&scratch_list);
	}
	
	
	// a late addition for hackerys sake and
	// possibly not correct
	if((this->from_entity != 0) || this->to_entity != 0){
		this->from_entity = 0;
		this->to_entity = 0;
		this->from_index = -1;
		this->to_index = -1;
		this->range_selected = FALSE;
	}
	
	this->mw->setModeSelection(this->getSoleItem(), this->from_index);
	this->mw->setSensitive();
}

/*
// why does this guy exist?
// who calls it?
// does it represent a different paradigm than the others
// no callbacks -- called from from MainWindow::deleteSelection
*/
void 
SelectionManager::unselect(iEntityPtr entity)
{
	Ref r = dllistSeek(&this->selection, entity);
	
	if((this->from_entity == entity) || this->to_entity == entity){
		this->from_entity = 0;
		this->to_entity = 0;
		this->from_index = -1;
		this->to_index = -1;
		this->range_selected = FALSE;
		this->clearLegacy();
	}
	
	if (r) {
		clearEntitySelection(entity);
		dllistRemove(&this->selection,r);
		this->clearLegacy();
	}
	
}

int 
SelectionManager::unselectChildren(iEntityPtr entity)
{
	iEntityPtr cur;
	
	int unselected=FALSE;
	for (cur=i_GetChildEntity(entity); cur; cur=i_GetNextEntity(cur)) {
		if (this->unselectChildren(cur))
			unselected=TRUE;
		if (this->selectedItemP(cur)) {
			Ref r = dllistSeek(&this->selection, cur);
			clearEntitySelection(cur);
			dllistRemove(&this->selection,r);
			unselected=TRUE;
		}
	}
	return unselected;
	/* doesn't do screen updates */
	
}

