
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"

#include "rafutil.h"

#include "text/text.h"
#include "text/TextElement.h"
#include "TextPage.h"
#include "EntityAux.h"
#include "Mode.h"
#include "ViewCanvas.h"
#include "misc/scroll.h"
#include "bb.h"

#define CACHE_TEXT_MODE TRUE 
#define CURSOR_WIDTH 3

static void redrawWindow(TextPage *page);

void refreshFlagParent(TextElement *element);

int
chainedPositionCalculate(TextElement *element, TextElement *sib, TextElement *parent)
{
	short new_rx;
	short new_ry;
	
	element->position_valid = TRUE;
	
	new_rx = 0;
	new_ry = 0;
	
	if(parent == 0){
		// you are the root
		// 00 fine
		return 0;
	} else {
		// your relative xy is some function of your sibs dimensions
		// maybe next maybe prev depending on direction
		if(sib) {
			// bogus I know that I want relative to parent
			new_rx = sib->rx;
			new_ry = sib->ry;
			short sib_w;
			short sib_h;
			sib->getDimensions(sib_w, sib_h);
			
			if((sib->pack->flags & pfReset) && sib->pack->orientation == pfVert){
				new_ry += sib_h;
				new_rx = 0;
			} else {
				switch(sib->pack->orientation){
				case pfHoriz:
					new_rx += sib_w;
					break;
				case pfVert:
					new_ry += sib_h;
					break;
				default:
					halt("Illegal orientation.\n");
				}
			}
		}
	}
	if(new_rx == element->rx && new_ry == element->ry){
		return 0;
	} else {
		element->setPosition(new_rx, new_ry);
		return 1;
	}
}


void
changeAndCalculatePosition(TextElement *element)
{
	TextElement *parent = element->getParentElement();
	int do_continue;
	
	if(parent){
		// loop over sibs in relevant direction
		// ogus -- handle reverse direction case
		Ref r;
		TextElement *sib;
		TextElement *prev_sib;
		r = parent->seek(element);
		if(r) {
			do_continue = 1;
			for(prev_sib = element, parent->next(&r); do_continue && r; prev_sib = sib, parent->next(&r)){
				sib = (TextElement *) parent->contents(r);
				do_continue = chainedPositionCalculate(sib, prev_sib, parent);
				//do_continue = 1;
				//sib->position_valid = FALSE;
			}
		}
		changeAndCalculatePosition(parent);
	}
}

void
changeAndCalculate(TextElement *element, int im_deleted)
{
	TextElement *parent = element->getParentElement();
	
	if(im_deleted == FALSE) {
		element->dimension_valid = FALSE;
		element->calculateDimensions();
	} else {
		element->w = 0;
		element->h = 0;
	}
	
	int do_continue;
	
	if(parent){
		// loop over sibs in relevant direction
		// ogus -- handle reverse direction case
		Ref r;
		TextElement *sib;
		TextElement *prev_sib;
		r = parent->seek(element);
		if(r) {
			do_continue = 1;
			for(prev_sib = element, parent->next(&r); do_continue && r; prev_sib = sib, parent->next(&r)){
				sib = (TextElement *) parent->contents(r);
				do_continue = chainedPositionCalculate(sib, prev_sib, parent);
				//do_continue = 1;
				//sib->position_valid = FALSE;
			}
		}
		changeAndCalculate(parent, FALSE);
	}
}


/* 
// should blank
*/
void 
TextElement::willUpdate()
{
//   this->willUpdate();
}

void 
TextElement::willMove()
{
	this->willUpdate();
}

iBox
findBox(TextPage *page, TextElement *element)
{
	iBox box;
	box.w = 0;
	short box_x;
	short box_y;
	short box_w;
	short box_h;
	
	element->getPosition(page->td, 0, box_x, box_y);
	element->getDimensions(box_w, box_h);
	
	// make the box slightly larger to make sure
	// all of the cursor is erased
	box.x = box_x-1;
	box.y = box_y;
	box.w = box_w+1;
	box.h = box_h;
	
	
	if(element->type == iNewLineT){
		box.w = box.x + CURSOR_WIDTH;
		box.x = 0;
	}
	
	return box;
}


void
redrawToEndOfLine(TextElement *element)
{
	TextPage    *page = (TextPage *) element->page;
	const Widget sw = page->mode->scrolled_window;
	const iBox   element_box = findBox(page,element);
	const iLBox   visible_box = visibleRectangle(sw);
	iLBox         redraw_box;
	static int   char_width = charWidth(' ');  // assumes fixed character width

	if(page->managed) {
		redraw_box.x = MAX(visible_box.x,element_box.x-char_width);
		redraw_box.y = element_box.y;
		redraw_box.h = element_box.h;
		redraw_box.w = MIN(visible_box.w, visible_box.x + visible_box.w - redraw_box.x);
		if(redraw_box.w && redraw_box.h)
			page->drawRegion(redraw_box);
	}
}

void 
redrawElement(TextElement *element)
{
	TextPage    *page = (TextPage *) element->page;
	const Widget sw = page->mode->scrolled_window;
	const iBox   element_box = findBox(page,element);
	const iLBox   visible_box = visibleRectangle(sw);
	iLBox         redraw_box;
	static int   char_width = charWidth(' ');  // assumes fixed character width

	if(page->managed) {
		redraw_box.x = element_box.x-char_width;
		redraw_box.y = element_box.y;
		redraw_box.h = element_box.h;
		redraw_box.w = element_box.w + char_width;
		i_IntersectLBox(visible_box, redraw_box, &redraw_box);
		if(redraw_box.w && redraw_box.h)
			page->drawRegion(redraw_box);
	}
}

void 
redrawElementFast(TextElement *element)
{
	TextPage    *page = (TextPage *) element->page;
	iLBox         redraw_box;
	static int   char_width = charWidth(' ');  // assumes fixed character width

	redraw_box = promoteBox(findBox(page,element));
	if(page->managed) {
		redraw_box.x -= char_width;
		redraw_box.w += char_width;
		if(redraw_box.w && redraw_box.h)
			page->drawRegion(redraw_box);
	}
}


void
redrawToEndOfPage(TextElement *element)
{
	TextPage    *page = (TextPage *) element->page;
	const Widget sw = page->mode->scrolled_window;
	const iLBox   visible_box = visibleRectangle(sw);
	const  int   line_height = charHeight('M');

	if(page->managed) {
		iLBox redraw_box;
		redraw_box  = promoteBox(findBox(page, element));
		redraw_box.x = visible_box.x;
		redraw_box.y = MAX(0,MAX(visible_box.y, redraw_box.y-line_height));
		redraw_box.h = visible_box.h + line_height;
		redraw_box.w = visible_box.w;
		if(redraw_box.w && redraw_box.h)
			page->drawRegion(redraw_box);
	}
}


static void redrawElement(TextElement *);

void
redrawWindow(TextPage *page)
{
	
	if(page->managed == FALSE)
		return;
	
	redrawElement(page->td);
}

#if 0
static void
redrawTextElement(TextElement *element)
{
	TextPage *page = (TextPage *) element->page;
	iBox box;
	
	if(page->managed) {
		box = findBox(page, element);
		if(box.w && box.h)
			page->drawRegion(box);
	}
}
#endif

/*
// none of these do any drawing or relaying out because of
// transaction based handling
*/

void
TextElement::create()
{
}


void 
TextElement::destroy()
{
	this->deleteAndUpdate();
}

void 
TextElement::update()
{
	this->initializeText();
}

void 
TextElement::move()
{
	this->update();
}


/*
// should draw directly
// bogus -- transaction based redrawing here too?
*/
void 
TextElement::select()
{
	int from;
	int to;
	
	getEntitySelectionRange(entity, from, to);
	
	if(from == -1 && to == -1)
		return;
	
	// defend again item selection
	if(from == -1 || to == -1) {
		TextPage *page;
		page = (TextPage *) this->page;
		// extends over more than one element
		redrawWindow(page);
	} else {
		// is inside me only
		// if from == to, we have only a thin cursor between
		// characters and none need to be drawn highlighted
		//if (from != to) redrawElement(this); sgs
		redrawElement(this);
	}
}

/*
// should draw directly
// bogus -- transaction based redrawing here too?
*/
void 
TextElement::unSelect()
{
	int from;
	int to;
	
	getEntitySelectionRange(entity, from, to);
	if(from == -1 || to == -1) {
		TextPage *page;
		page = (TextPage *) this->page;
		// extends over more than one element
		redrawWindow(page);
	} else {
		redrawElement(this);
	}
}


void 
TextElement::recognize()
{
}

MainWindow *
TextElement::getMainWindow()
{
	return this->page->mode->main_window;
}

void
TextElement::deleteAndUpdate()
{
	this->page->removeFace(this);

	setTextViewer(this->entity, 0);
	
	delete this;
}

	
#define PRIME 211
#define HASH(n) ( ((int)(n)) % PRIME ) 
struct tvTableEntry { iEntityPtr entity; EntityFace *tv; };
static tvTableEntry tvTable[PRIME];

EntityFace *
getTextViewer(iEntityPtr entity)
{
	const int i=HASH(entity);

	if (CACHE_TEXT_MODE && entity && tvTable[i].tv && (tvTable[i].entity == entity)) {
		return tvTable[i].tv;
	} else {
		EntityAux *ea = getController(entity);
		EntityFace *ef = ea->faces[vtTEXT];
		if (CACHE_TEXT_MODE && ef) {
			tvTable[i].entity = entity;
			tvTable[i].tv = ef; 
		}
		return ef;
	}
}

void
flushTextModeCache()
{
	int i;
	for(i=0; i<PRIME; i++) {
		tvTable[i].entity = (iEntityPtr) 0;
		tvTable[i].tv = (EntityFace *) 0;
	}
}

void
setTextViewer(iEntityPtr entity, EntityFace *face)
{
	// cut out layer of function call
	EntityAux *ea = getController(entity);
	ea->faces[vtTEXT] = face;

	if(face && face->page)
		face->page->addFace(face);
}


void
TextElement::preLayoutCB()
{
#if 0
	printf ("%x\tmoved from\t%d,\t%d\t%dx\t%d\n", this, this->rx, this->ry, this->w, this->h);
#endif
}

void
TextElement::postLayoutCB()
{
#if 0
	printf ("%x\tmoved to\t%d,\t%d\t%dx\t%d\n", this, this->rx, this->ry, this->w, this->h);
#endif
}


/*
// font hackery
*/
static int char_width = 0;
static int char_height = 0;

int values_initialized = FALSE;

void
setFontValues(int input_char_width, int input_char_height)
{
	values_initialized = TRUE;
	
	char_width = input_char_width;
	char_height = input_char_height;
}

void
getFontValues(int & output_char_width, int & output_char_height)
{
	output_char_width=char_width;
	output_char_height=char_height;
}

ViewCanvas *the_vc = 0;

int charHeight(iUCode c)
{
	int h=0, w, ascent;
	if (the_vc)
		the_vc->measureText(&c, 1, &w, &h, &ascent);
	
	return h;
}

int charWidth(iUCode c)
{
	int h, w=0, ascent;
	if(the_vc)
		the_vc->measureText(&c, 1, &w, &h, &ascent);
	return w;
}
