

/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "illum.h"

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <X11/cursorfont.h>

#include "rafutil.h"
#include "smemory.h"

#include "misc/Unicode.h"
#include "misc/XArg.h"

#include "keyboard/KeyBoard.h"
#include "keyboard/KeyMap.h"

#include "text/text.h"
#include "text/TextElement.h"

#include "scroll.h"
#include "bb.h"
#include "font.h"
#include "MainWindow.h"


#include "ViewCanvas.h"
#include "View.h"
#include "EntityAux.h"
#include "MainWindow.h"
#include "Document.h"
#include "SelectionManager.h"

#include "TextFrame.h"
#include "text/TextMargin.h"

#define DISTANCE_FROM_EDGE 5  // keep popup this far from edge of screen
#define POPUP_ADJ 5
#define IMAGE_FRINGE 30  //size of image around entity shown
#define SHADOW_WIDTH 3
#define MARGIN 1



static iUCode uni_space_str[2]={' ',0};


/*
// popup stuff
*/

void 
TextFrame::clearPopup()
{
	// find the popup
	// order the area to redraw
	if(this->display_popup && this->bounds_on_view.w){
		this->addRedrawRegion(&this->bounds_on_view);
	}
	this->bounds_on_view.w = 0;
	this->outer_image_bounds.w = 0;
	this->inner_image_bounds.w = 0;
	this->popup_image = 0;
	this->flag_entity = 0;
}

void
TextFrame::setPopup(iEntityPtr entity)
{
	// if there is a popup area already, clear it
	// set the popup
	// order the area to redraw
	this->clearPopup();
	if(entity){
		this->flag_entity = entity;
		this->decideImageBounds(entity, 
								&this->inner_image_bounds,
								&this->outer_image_bounds);

		this->bounds_on_view = 
			this->decideViewBounds(entity, 
								   this->outer_image_bounds);
		this->popup_image = i_GetImage(entity);
		if(this->display_popup && this->bounds_on_view.w && this->popup_image){
			this->addRedrawRegion(&this->bounds_on_view);
		}
	}
}

void
TextFrame::refreshPopup(Display *display, Pixmap pixmap, int off_x, int off_y)
{
	ViewCanvas *vc = this->getCanvas();
	Widget widget = vc->canvas;

	char *bits1 = 0;
	long dw1;
	XImage *x_image1 = 0;

	char *bits2 = 0;
	long dw2;
	XImage *x_image2 = 0;

	GC gc;

	iBox ibox = this->inner_image_bounds;
	iBox obox = this->outer_image_bounds;

	// actually draw

	int normal_image_mag_exp = 1 << 16;
	if (i_Image2BitMap(this->popup_image, obox,
					   normal_image_mag_exp, &dw1, &bits1)) 
		goto clean_images;
	if (i_Image2BitMap(this->popup_image, ibox,
					   normal_image_mag_exp, &dw2, &bits2))
		goto clean_images;

	// create the image to be the size of the bitmap
	x_image1 = XCreateImage(display, 0, 1, XYBitmap, 
							0, // tricky -- ignore initial bits
							bits1,
							obox.w, 
							obox.h, 
							8, // quanta of scan line
							dw1);

	x_image2 = XCreateImage(display, 0, 1, XYBitmap, 
							0, // tricky -- ignore initial bits
							bits2,
							ibox.w, 
							ibox.h, 
							8, // quanta of scan line
							dw2);

	x_image1->byte_order = MSBFirst;
	x_image1->bitmap_bit_order = MSBFirst;
	x_image2->byte_order = MSBFirst;
	x_image2->bitmap_bit_order = MSBFirst;
	_XInitImageFuncPtrs(x_image1);
	_XInitImageFuncPtrs(x_image2);

	gc = vc->darkgrayGC();

	XPutImage(display, pixmap,
			  gc, x_image1, 
			  0,0, 
			  off_x, off_y,
			  obox.w, obox.h);

	gc = vc->blackGC();

	XPutImage(display, pixmap,
			  gc, x_image2, 
			  0,0, 
			  off_x + ibox.x - obox.x, off_y + ibox.y - obox.y,
			  ibox.w, ibox.h);


	// this is the border
	XDrawLine(display, pixmap, vc->blackGC(),
			  off_x + 0,         off_y + obox.h-1,
			  off_x + obox.w-1,  off_y + obox.h-1);
	XDrawLine(display, pixmap, vc->blackGC(),
			  off_x + obox.w-1, off_y + 0, 
			  off_x + obox.w-1, off_y + obox.h-1);

	XDrawRectangle(display, pixmap, 
				   vc->blueGC(),
				   off_x + 0, off_y + 0, 
				   obox.w-2, obox.h-2); 

clean_images :
		if(bits1){
			smfree(bits1);
		}
	if(x_image1) {
		x_image1->data=0;
		XDestroyImage(x_image1);
	}

	if(bits2){
		smfree(bits2);
	}
	if(x_image2) {
		x_image2->data=0;
		XDestroyImage(x_image2);
	}
}

void
TextFrame::decideImageBounds(iEntityPtr entity, 
							 iBox *inner_bounds, iBox *outer_bounds)
{
	iBox ibox = i_GetBounds(entity);
	iBox obox = ibox;
	iImagePtr image = i_GetImage(entity);

	if(image == 0) {
		obox.w = 0;
		ibox.w = 0;
	} else {
		int iw, ih;
		// no scaling because we always display in mag 1
		iw = i_GetImageWidth(image);
		ih = i_GetImageHeight(image);
		if (obox.x>IMAGE_FRINGE){
			obox.x-=IMAGE_FRINGE;
		} 	else{
			obox.x=0;
		}

		if (obox.y>IMAGE_FRINGE){
			obox.y-=IMAGE_FRINGE;
		} else {
			obox.y = 0;
		}
		obox.w = MIN(iw-obox.x, obox.w+2*IMAGE_FRINGE);
		obox.h = MIN(ih-obox.y, obox.h+2*IMAGE_FRINGE);
	}

	*inner_bounds = ibox;
	*outer_bounds = obox;
}

iBox 
TextFrame::decideViewBounds(iEntityPtr entity, iBox image_bounds)
{
	short x, x1, x2, y;
	short w, h;
	iBox b;

	iEntityPtr beg_ent, end_ent;
	int beg_ind, end_ind;

	i_Object2Range(entity, &beg_ent, &beg_ind, &end_ent, &end_ind);
	TextElement *begin = (TextElement*)getTextViewer(beg_ent);
	// cause sometime object to range return 0 in
	// that second slot -- bogus
	end_ent = end_ent ? end_ent : beg_ent;
	TextElement *end = (TextElement*)getTextViewer(end_ent);
	begin->getPosition(this->td, beg_ind, x1, y);
	end->getPosition(this->td, end_ind, x2, y);
	begin->getDimensions(w, h);

	w = image_bounds.w;
	x = x1 + ((x2-x1) / 2) - (w/2);
	y += h;
	h = image_bounds.h;

	// hmm should check if both are being violated at once
	if(x < DISTANCE_FROM_EDGE)
		x = DISTANCE_FROM_EDGE;
	if(x + w > this->width - DISTANCE_FROM_EDGE)
		x = this->width - w - DISTANCE_FROM_EDGE;

	b.x = x;
	b.y = y;
	b.w = w;
	b.h = h;

	return b;
}


void 
TextFrame::togglePopup(void)
{	
	XArg args;
	iBox box;
	MainWindow *mw=this->view->main_window;

	this->display_popup = !this->display_popup;
	if (this->flag_entity){
		args.reset();
		XtVaGetValues(mw->scroll_window, 
			XmNheight, &box.h, XmNwidth, &box.w, NULL);

		/* refresh screen  */
		// bogus -- we know what to redraw!
		this->drawRegion(this->bounds_on_view);
	}
}



/*
// bogus -- I probably don't have to compress events here
// if I am counting on having them be compressed later by the workproc
*/
void 
TextFrame::redisplay(XmDrawingAreaCallbackStruct *call_data)
{
	if(this->td == 0)
		return;
	
	Widget w = this->getCanvas()->canvas;
	XEvent *event = (XEvent *) call_data->event;
	
	static Region region = NULL;
	
	if(XtIsManaged(w) == FALSE)
		return;
	
	if(!region)
		region = XCreateRegion();
	
	XtAddExposureToRegion(event, region);
	
	if(event->xexpose.count != 0)
		return;
	
	iBox refresh_box;
	
	XRectangle rect;
	XClipBox(region, &rect);
	XDestroyRegion(region);
	region = 0;
	
	refresh_box.x = rect.x;
	refresh_box.y = rect.y;
	refresh_box.w = rect.width;
	refresh_box.h = rect.height;
	
	if(refresh_box.w == 0 || refresh_box.h == 0)
		return;
	
	this->drawRegion(refresh_box);
}


static void
drawArea(TextFrame *frame,
		 Display *display, Pixmap drawable,
		 const iBox& region,
		 int x_off, int y_off)
{
	int truncate = TRUE;
	int ignore_selection = FALSE;

	tfDrawRegion(frame, display, drawable, 
			   region,
			   x_off, y_off, 
			   truncate,
			   ignore_selection);
}



void
TextFrame::drawRegionNow(iBox box)
{
	int x_off, y_off, w, h;
	int add_popups;
	iBox obox;
	
	if(this->td == 0)
		return;

	Widget widget = this->getCanvas()->canvas;
	Display *display = XtDisplay(widget);
	Drawable drawable = RootWindowOfScreen(XtScreen(widget));;
	Pixmap pixmap = XtWindow(widget);
	int depth = DefaultDepthOfScreen(XtScreen(widget));

	// if we want to draw popups and the region overlaps
	// with the popups, we will extend the region to
	// completely overlap with the popups
	// this will simplify drawing the popups

	add_popups = this->display_popup && this->popup_image;

	if(add_popups){
		iBox ibox;
		// redraw the popup if it overlaps the area
		i_IntersectBox(box, this->bounds_on_view, &ibox);
		add_popups = ibox.w;
	}

	obox = box;
	if(add_popups){
		i_UnionBox(obox, this->bounds_on_view, &box);
	}

	x_off = box.x;
	y_off = box.y;
	w = box.w;
	h = box.h;

#define PIXMAP_BUFFERING
	
#ifdef PIXMAP_BUFFERING
	pixmap = XCreatePixmap(display, drawable, box.w, box.h, depth);
	XFillRectangle(display, pixmap, 
				   this->getCanvas()->blankGC(),
				   0, 0,
				   box.w, box.h);
#else
	x_off = 0;
	y_off = 0;
#endif
	drawArea(this, 
			 display, pixmap, 
			 box,
			 x_off, y_off);
	if(add_popups)
		this->refreshPopup(display, pixmap, 
						   bounds_on_view.x - box.x, bounds_on_view.y - box.y);
	
#ifdef PIXMAP_BUFFERING
	// this is the place we finally trim the
	// area to be exactly what the expose wants
	XCopyArea(display, pixmap, XtWindow(widget), this->getCanvas()->copyGC(),
			  0, 0,
			  box.w, box.h,  
			  box.x, box.y);
	XFreePixmap(display, pixmap);
#endif


}


static void
drawElementMaybe(TextElement *element, TextFrame *frame,
				 Display *display, Pixmap pixmap,
				 const iBox& region,
				 int x_off, int y_off,
				 int& selection_on,
				 iEntityPtr from_entity, int from_index,
				 iEntityPtr to_entity, int to_index);

static int 
insideMe(TextElement *element, iEntityPtr entity)
{
	TextElement *child_element = (TextElement *) eaGetFace(entity, vtTYPE);
	while(child_element && child_element != element)
		child_element = child_element->getParentElement();
	return child_element == element;
}

static void
drawElementNot(TextElement *element,
			   TextFrame *frame,
			   int& selection,
			   iEntityPtr from_entity,
			   int from_index,
			   iEntityPtr to_entity,
			   int to_index)
{
	// don't draw anything, just toggle the selection as necessary
	if(from_entity){
		if(insideMe(element, from_entity))
			selection = !selection;
		if(insideMe(element, to_entity))
			selection = !selection;
	}
}

iUCode spaces[1024];

int
measureSome(const iUCode *s, int n)
{
	int w;
	int i;
	
    for(w = 0, i = 0; i < n; i++){
		w += charWidth(s[i]);
	}
	
	return w;
}

static GC
chooseGC(ViewCanvas *vc, int if_flag, int selection_on)
{
	GC use_gc;

	if (if_flag){
		use_gc = selection_on ? vc->inverseFlagGC() : vc->FlagGC();
	}
	else{
		use_gc = selection_on ? vc->inverseBlackGC() : vc->blackGC();
	}

	return use_gc;




}

/*
// what is region?
// what is x_off, y_off?
//
// everything has been made relative to our xy
// except in the case of newline where the x is that
// of the end of the first line
// but everything has been made relative to the 
// x of the successive lines
*/
static void
drawElementMaybe(TextElement *element, TextFrame *frame,
				 Display *display, Pixmap pixmap,
				 const iBox& region,
				 int x_off, int y_off,
				 int& selection_on,
				 iEntityPtr from_entity,
				 int from_index,
				 iEntityPtr to_entity,
				 int to_index)
{
	const int charHeightBlank = charHeight(' ');
	iBox my_region;
	iEntityPtr temp;

	element->getDimensions(my_region.w, my_region.h);
	// my position relative to myself is 0,0 unless I'm the root
       TextElement *parent = element->getParentElement();
	if (parent) {
		my_region.x = my_region.y = 0;
	} else {
		my_region.x = XMARGIN;
		my_region.y = YMARGIN;
	}
	
	int if_flag = (int)i_FindProperty(element->entity, FLAG_PROP);
	for(temp=i_GetParentEntity(element->entity);!if_flag&&temp; 
			temp=i_GetParentEntity(temp))
		if_flag = (int)i_FindProperty(temp, FLAG_PROP);

	if(overlapBoxes(region, my_region)){
		short x,y;
		ViewCanvas *vc = frame->getCanvas();
		int i, from = -1, to = -1;
		const iUCode *s;
		GC use_gc;
		
		// avoid 2 more calls  to get prop
		SelectionClass sc = scNOT;
		if(element->entity == from_entity) {
			sc = scRANGE;
			from = from_index;
		} if(element->entity == to_entity) {
			sc = scRANGE;
			to = to_index;
		}
		
		if((element->pack->flags & pfCounter) && (element->pack->orientation == pfVert)) {
			// we are a newline
			int num_lines = i_GetInt(element->entity,0);
			
			// this code draws a space for each line of the newline entity
			// I don't know why this is necessary (sgs)
			for(i = 0; i < num_lines; i++){
				if(sc == scRANGE && from != to && (from == i || to == i)){
					selection_on = !selection_on;
				}
				
				if(i != 0){
					short e_x, e_y;
					element->getPosition(0,0,e_x,e_y);
					x = -e_x;
				} else
					x = XMARGIN; 
				
				s = uni_space_str;
				
				// my xy is assumed to be 0
				
				y = charHeightBlank * i;
				use_gc = chooseGC(vc, if_flag, selection_on);
#ifdef DRAW_SPACE_FOR_NEWLINE
				drawUnicode(display, pixmap, use_gc, 
							x - x_off,
							y + frame->font_ascent - y_off, 
							s, ucslen(s));
#endif
			}
		} else if (element->text != 0 || ((element->pack->flags & pfCounter) &&  (element->pack->orientation == pfHoriz))) {
			if(element->text == 0){
				// we are a space;
				int num_spaces = i_GetInt(element->entity,0);
				for(i = 0; i < num_spaces; i++)
					spaces[i] = ' ';
				spaces[num_spaces] = '\0';
				s = spaces;
			} else {
				s = element->text;
			}
			
			// my xy is assumed to be 0
			x = y = 0; 
			
			if((sc == scRANGE) && (from!=to)) {
				// we have a range selected

				int less = MIN(from,to), more = MAX(from,to);
				if(less == -1) {
					// there is only 1 flip
					use_gc = chooseGC(vc, if_flag, selection_on);

					drawUnicode(display, pixmap, use_gc, 
								x - x_off,
								y + frame->font_ascent - y_off, 
								s, more);
					selection_on = !selection_on;
					use_gc = chooseGC(vc, if_flag, selection_on);

					drawUnicode(display, pixmap, use_gc, 
								x - x_off + measureSome(s, more),
								y + frame->font_ascent - y_off, 
								s+more, ucslen(s) - more);
				} else {
					// there are 2 flips
					use_gc = chooseGC(vc, if_flag, selection_on);
					drawUnicode(display, pixmap, use_gc, 
								x - x_off,
								y + frame->font_ascent - y_off, 
								s, less);
					selection_on = !selection_on;
					use_gc = chooseGC(vc, if_flag, selection_on);
					drawUnicode(display, pixmap, use_gc, 
								x - x_off + measureSome(s, less),
								y + frame->font_ascent - y_off, 
								s+less, more - less);
					selection_on = !selection_on;
					use_gc = chooseGC(vc, if_flag, selection_on);
					drawUnicode(display, pixmap, use_gc, 
								x - x_off + measureSome(s, more),
								y + frame->font_ascent - y_off, 
								s+more, ucslen(s) - more);
				}
				
			} else {
				use_gc = chooseGC(vc, if_flag, selection_on);
				drawUnicode(display, pixmap, use_gc, 
							x - x_off,
							y+ frame->font_ascent - y_off, 
							s, ucslen(s));
			}
		} else {
			TextElement *child;
			Ref r;
			
			for (r = element->first(); r; element->next(&r)){
				child = (TextElement *) element->contents(r);
				child->getPosition(element,0,x,y);
				iBox new_region = region;
				new_region.x -= x;
				new_region.y -= y;
				drawElementMaybe(child, frame,
								 display, pixmap,
								 new_region,
								 x_off-x, y_off-y,
								 selection_on,
								 from_entity, from_index,
								 to_entity, to_index);
			}
		}
		
	} else {
		// this will compute all the flips that will occur inside
		// of here
		if(from_entity)
			drawElementNot(element, frame, selection_on,
						   from_entity, from_index, to_entity, to_index);
	}
}


void
tfDrawRegion(TextFrame *frame,
		   Display *display, Pixmap pixmap,
		   const iBox& region,              // region is relative to the
								// root
		   int x_off, int y_off,
		   int truncate,
		   int ignore_selection)
{
	const int charHeightBlank = charHeight(' ');
	int selection_on = FALSE;
	iBox truncated_region;
	iBox window_region;
	GC use_gc;
	
	iEntityPtr from_entity;
	iEntityPtr to_entity;
	int from_index;
	int to_index;
	short text_x=0, text_y=0;	
	TextElement *element;
	
	frame->getSM()->getRange(from_entity, from_index,
							 to_entity, to_index);
	
	TextElement *root = frame->td;
	if(truncate) {
		Widget sw = frame->view->scrolled_window;
		window_region = visibleRectangle(sw);
		i_IntersectBox(window_region, region, &truncated_region);
	} else {
		truncated_region = region;
	}
	if(from_entity == to_entity && from_index == to_index){
		from_entity = 0;
		to_entity = 0;
	}
	
	if(truncated_region.w){
		drawElementMaybe(root, frame,
						 display, pixmap,
						 truncated_region,
						 x_off, y_off,
						 selection_on,
						 from_entity, from_index,
						 to_entity, to_index); 
	}

	if(ignore_selection == FALSE)  {
		// bogus -- lots of bogusity for the sake of newline
		
		// special case of a thin cursor
		int fudge;      
		
		iEntityPtr sole_entity;
		int from;
		tfGetThinCursor(frame->getSM(), sole_entity, from);
		
		if(sole_entity){ 
			/* draw that rectangular blob */  
			iBox text_bounds;
			element = (TextElement *) eaGetFace(sole_entity, vtTYPE);
			text_bounds.w = 1;
			element->getPosition(frame->td, 
								 element->type == iNewLineT ? 0 : from, 
								 text_x, text_y);
			if(element->type == iNewLineT){
				if(from == 0)
					; // do nothing -- it is right
				else {
					// bogus -- we depend on our parent
					text_x = XMARGIN;
					text_y = text_y + from * charHeightBlank;
				}
			}
			text_bounds.x = text_x;
			text_bounds.y = text_y+1;
			text_bounds.w = 1;
			text_bounds.h = charHeightBlank-1;
			if(text_bounds.w && overlapBoxes(text_bounds, region)){
				// I think fudge was for the window edge
				// and what I'd rather is 
				// have the 4 pixel boundary
				fudge = 0;
				use_gc = frame->getCanvas()->thickRedGC();
				XDrawRectangle(display, pixmap, use_gc, 
							   text_bounds.x - x_off + fudge,
							   text_bounds.y - y_off, 
							   0, text_bounds.h);
			}
		} 
	}
}
