
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef _TextFrame_h_
#define _TextFrame_h_

#pragma interface

#include "Frame.h"

class TextElement;

class TextFrame : public Frame {
public:

	iEntityPtr flag_entity;
	TextElement *td;
	iEntityPtr root_entity;
	iEntityPtr first_flag;
	int initialized_canvas;


	int font_ascent;

	int goal_column;
	
	int thresh_xl,thresh_yl,oldh,oldw;
	
	TextFrame(View *input_view, const aString& label, iEntityPtr entity);
	virtual ~TextFrame();
	
	void findEntityAndIndex(int x, int y,
							iEntityPtr *entity_return, int *index_return);
	
	/*virtual*/ void conditionallyInitialiseCanvas();
	/*virtual*/ void setMag(int new_mag, int old_mag);
	/*virtual*/ void _setMag(int new_mag, int old_mag);
	/*virtual*/ void redisplay(XmDrawingAreaCallbackStruct *call_data);
	/*virtual*/ iEntityPtr findEnclosingEntity(int x, int y);
	
	/*virtual*/ void addNewEntity(iEntityPtr entity);
	
	/*virtual*/ void prevEntity();
	/*virtual*/ void nextEntity();
	

	/*virtual*/ void findNextFlagEntity();
	void findNextFlagEntitySub();
	
	/*virtual*/ void upEntity();
	/*virtual*/ void downEntity();
	
	/*virtual*/ void centerOnSelected(iEntityPtr entity);
	
	// for delayed drawing
	/*virtual*/ int resizeNecessaryP();
	/*virtual*/ void drawRegionNow(iBox box);
	
	void initialise();
	
	/*virtual*/ void         pressButton(XEvent *event);
	/*virtual*/ void         releaseButton(XEvent *event);
	/*virtual*/ void         moveButton(int state, int x, int y);
	
	/*virtual*/ void manage();
	/*virtual*/ void unmanage();
	
	void getglobalposition(TextElement *element,int index, short *xx, short
	*yy);
	int computeIndex(TextElement *element, Position x, Position y);
	void centerCursor(TextElement *);
	int isVisible(TextElement *);
	void configure();
	
	void deleteChar();
	/*virtual*/ int  pressKey(iUCode key);
	void insertChar(iEntityPtr, char);
	void insertNL(iEntityPtr sole_entity);
	void joinLine(iEntityPtr sole_entity);

    // popup stuff
	int display_popup;

	// when these are null, you have nothing to draw

    // save this because the actual entity might get
	// wiped, this is bit of image to blow up
	iBox inner_image_bounds;
	iBox outer_image_bounds;
	iBox bounds_on_view;  // this is where to draw it this includes
	                      // all the fudges for the shadow
	iImagePtr popup_image;
	
	void clearPopup(void);   // blank the popup and set it to null
	void togglePopup(void);

	void decideImageBounds(iEntityPtr entity,
						   iBox *inner_bounds, iBox *outer_bounds);

	iBox decideViewBounds(iEntityPtr entity, iBox image_bounds);
	
	void setPopup(iEntityPtr entity); //we must set the 2 bounds

    // this gets called by refreshes
	// nobody else draws, ultimately integrated with normal refresh
	void refreshPopup(Display *, Pixmap, int, int);
	
};

/*
// region is relative to the current element
//
// x_off is for translation relative to the pixmap
*/
void
tfDrawRegion(TextFrame *frame,
			 Display *display, Pixmap pixmap,
			 const iBox& region,
			 int x_off, int y_off,
			 int truncate,
			 int ignore_selection);

void
tfGetThinCursor(SelectionManager *sm, iEntityPtr& from_entity, int& from);


#endif
