/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <X11/cursorfont.h>

#include "rafutil.h"
#include "smemory.h"

#include "misc/Unicode.h"
#include "misc/XArg.h"

#include "keyboard/KeyBoard.h"
#include "keyboard/KeyMap.h"

#include "text/text.h"
#include "text/TextElement.h"

#include "misc/scroll.h"
#include "bb.h"
#include "font.h"
#include "MainWindow.h"


#include "ViewCanvas.h"
#include "Mode.h"
#include "EntityAux.h"
#include "MainWindow.h"
#include "Document.h"
#include "SelectionManager.h"

#include "TextPage.h"
#include "text/TextMargin.h"

#define DEFAULT_POPUP -1 // popup is shown as default


#define XBUFFER 50
#define YBUFFER 50 

extern void trackSoleSelection(MainWindow *mw);
void redrawToEndOfLine(TextElement *element);
void redrawToEndOfPage(TextElement *element);
void redrawElement(TextElement *element);
void redrawElementFast(TextElement *element);
extern void changeAndCalculatePosition(TextElement *);
extern void forceRecalc(TextElement *element);
void jiggerEverything(TextPage *page);


void
recomputeEverything(TextElement *element, TextElement *parent,
					TextElement *prev_sib)
{
	static int 
		charWidthBlank = charWidth(' '),
		charHeightBlank = charHeight(' '),
		charHeightM = charHeight('M');

	Ref r;
	TextElement *child;
	TextElement *prev_child;
	iBox box={0,0,0,0};
	iBox child_box;
	
	element->dimension_valid = TRUE;
	
	for(prev_child = 0, r = element->first(); r; prev_child = child, element->next(&r)){
		child = (TextElement *) element->contents(r);
		recomputeEverything(child, element, prev_child);
		//child->getPosition(element,0,child_box.x,child_box.y);
		// bogus could be faster;
		child_box.x = child->rx;
		child_box.y = child->ry;
		child_box.w = child->w;
		child_box.h = child->h;
		i_UnionBox(box,child_box, &box);
	}
	element->w=box.w;
	element->h=box.h;
	
	element->position_valid = TRUE;
	
	element->rx = 0;
	element->ry = 0;
	
	if(parent == 0){
		// you are the root
		// 00 fine
	} else {
		int parent_is_root = !parent->getParentElement();
		if(prev_sib) {
			// your relative xy is some function of your prev_sibs dimensions
			// maybe next maybe prev depending on direction
			if(prev_sib->position_valid == FALSE)
				prev_sib->calculatePosition();
			// if parent is root this will return absolute position
			prev_sib->getPosition(parent, 0, element->rx, element->ry);
			if (parent_is_root)
				element->rx -= XMARGIN;
			short prev_sib_w;
			short prev_sib_h;
			prev_sib->getDimensions(prev_sib_w, prev_sib_h);
			
			if((prev_sib->pack->flags & pfReset) && prev_sib->pack->orientation == pfVert){
				element->ry += prev_sib_h;
				element->rx = 0;
			} else {
				switch(prev_sib->pack->orientation){
				case pfHoriz:
					element->rx += prev_sib_w;
					break;
				case pfVert:
					element->ry += prev_sib_h;
					break;
				default:
					halt("Illegal orientation.\n");
				}
			}
		}
	}
	
	if(element->pack->flags & pfCounter){
		switch(element->pack->orientation) {
		case pfHoriz:
			element->w = charWidthBlank * i_GetShort(element->entity,0);
			element->h = charHeightBlank;
			break;
		case pfVert:
			element->w = charWidthBlank;
			element->h = charHeightBlank * i_GetShort(element->entity,0);
			break;
		default:
			halt("Illegal orientation.\n");
		}
	} else if(element->text != 0) {
		element->h = charHeightM;
		{
			iUCode *s;
			int alt_w;
			alt_w = 0;
			for(s = element->text; *s; s++)
				alt_w += charWidth(*s);
			element->w = alt_w;
		}
	} else {
		// already computed earlier when traversing children
	}
}

void
recomputeEverything(TextElement *element)
{
	Ref r;
	TextElement *parent;
	TextElement *prev_sib = 0;

	parent = element->getParentElement();
	if(parent){
		r = parent->seek(element);
		parent->prev(&r);
		prev_sib = (TextElement *) parent->contents(r);
	}
	recomputeEverything(element, parent, prev_sib);
}


void
jiggerIt(iEntityPtr entity)
{
	TextElement *item;

	item = (TextElement *) eaGetFace(entity, vtTEXT);
	if(item) {
		recomputeEverything(item);
		item->calculatePosition();
	}
}


void
jiggerToEndOfLine(TextElement *element, int from)
{
	if(!element) return;

	const TextPage *page = (TextPage *) element->page;
	const iEntityPtr entity = element->entity;

	if (entity) {

		iEntityPtr start, stop, temp;
		TextElement * item;

		// find entity immediately precceding current one 
		if (!(start = i_FindEntityBack(entity, 0, PROPERTY, ILLUM_PROP)))
			start = page->td->entity;

		// back up one more entity (required to fix PR 210) 
		if (temp = i_FindEntityBack(start, 0, PROPERTY, ILLUM_PROP))
			start = temp;
	
		// find NewLine at end of current line
		if(!(stop = i_FindEntityForward(entity, iNewLineT, ALL, 0)))
			stop = entity;
	
		// update positions from just before current entity to end of line
		jiggerIt(start);
		for (temp=i_FindEntityForward(start,0,PROPERTY,ILLUM_PROP); 
		     temp && (temp != stop); 
		     temp=i_FindEntityForward(temp,0,PROPERTY,ILLUM_PROP))
			jiggerIt(temp);
		if(stop != start) jiggerIt(stop); 

		// update ancestors of element(s) which changed
		// this code goes up the tree from the newline entity
		if( (item=(TextElement *) eaGetFace(stop, vtTEXT)) )
			changeAndCalculatePosition(item);

		page->td->dimension_valid = FALSE;  
	}
}


void
jiggerToEndOfPage(TextElement *element, int from)
{
        TextPage *page = 0;
        iEntityPtr start1=0, start2=0, start3=0, temp;
        iEntityPtr entity = 0;
        TextElement * item;

        if(element) {
                page = (TextPage *) element->page;
        } else {
                return;
        }

        entity = element->entity;

        if (entity){
                // find everything
                if (!(start2 = i_FindEntityBack(entity, iNewLineT, ALL, 0))){
                        start2=entity;
                        start1 = page->td->entity;
                } else {
                        start1 = i_FindEntityBack(start2, iNewLineT, ALL, 0);
                        if(start1 == 0)
                                start1 = page->td->entity;
                }

                start3 = i_FindEntityForward(entity, iNewLineT, ALL, 0);
                if (!start3)
                        start3=entity;
                if (!start1){
                        temp=i_FindEntityBack(start2,0,NOCHILD,0);
                        if (temp)
                                start1=i_GetParentEntity(temp);
                        else
                                start1=start2;
                }

                // use things
                for (temp=start1; temp&&temp!=start3;
                                                  temp=i_FindEntityForward(temp,0,PROPERTY,ILLUM_PROP)){
                        jiggerIt(temp);
                }

                jiggerIt(start3);
                if (temp=i_FindEntityForward(temp,0,PROPERTY,ILLUM_PROP)){
                        for (; temp&&(i_GetEntityType(temp)!=iNewLineT);
                        temp=i_FindEntityForward(temp,0,PROPERTY,ILLUM_PROP))
                                jiggerIt(temp);
                }
                if (temp && i_GetEntityType(temp)==iNewLineT)
                        jiggerIt(temp);

                if(temp)
                        item = (TextElement *) eaGetFace(temp, vtTEXT);
                else
                        item = 0;

                if(item){
                        // hmm we must do this somehow
                        changeAndCalculatePosition(item);
                }

                // ouch -- this is expensive
                page->td->dimension_valid = FALSE;
        }
}


void
jiggerEverything(TextPage *page)
{
	if(page->td)
		recomputeEverything(page->td);
}



void
tfGetThinCursor(SelectionManager *sm, iEntityPtr& from_entity, int& from)
{
	iEntityPtr to_entity;
	from_entity = 0;
	to_entity = 0;
	int to;
	
	sm->getRange(from_entity, from,
				 to_entity, to);
	
	from_entity = (from_entity != to_entity) || (from != to) ? 0 : from_entity;
}


#define RC_FUDGE 6


void
TextPage::initialise()
{
	if(this->td) {
		halt("illum: shouldn't try to reinitialise TextPage.\n");
	}
	
	this->td = 0;
	
	/* this is gross but I might have to create a dummy
	   entity in here
	   */
	disableCallbacks();
	if(root_entity)
		this->td = newParseEntity(root_entity, 0, this);
	enableCallbacks();
}

static iUCode uni_m[2]={'M',0};

TextPage::TextPage(Mode *input_mode, 
					 const aString& label, 
					 iEntityPtr input_root_entity)
	: Page(label, input_mode), 
	  initialized_canvas(0), 
	  root_entity(input_root_entity),
	  td(0),
	  first_flag(0),
	  goal_column(XMARGIN),
	  flag_entity(0),
	  display_popup(DEFAULT_POPUP),
	  thresh_xl(0),
	  thresh_yl(0),
	  popup_image(0)
{

	ViewCanvas *vc;
	
	vc = this->getCanvas();
	
	/* bogus -- get rid of these references */
	if(vc && XtIsRealized(vc->canvas)) {
		int char_width;
		int char_height;
 		vc->measureText(uni_m, 1,
 						&char_width, &char_height, 
 						&this->font_ascent);
		::the_vc = vc;
	}

	// do this sizing stuff after I have set up width of things
	this->initialise();

	this->inner_image_bounds.w = 0;
	this->outer_image_bounds.w = 0;

	this->bounds_on_view.w = 0;
}

TextPage::~TextPage()
{
	this->flushRedrawRegion();

	// now I try to maintain Page's face_list
	// so this should not be necessary
}


#define PAGE_WIDTH WINDOW_WIDTH
#define PAGE_HEIGHT WINDOW_HEIGHT

int
TextPage::resizeNecessaryP()
{
	short w,h;
	
	this->td->getDimensions(w, h);
	Dimension d_w = w;
	Dimension d_h = h;
	this->syncSize();
	if ((d_w + XMARGIN > this->width  || MAX(d_w+XMARGIN,PAGE_WIDTH) < this->width-XBUFFER) ||
	    (d_h + YMARGIN > this->height || MAX(d_h+YMARGIN,PAGE_HEIGHT) < this->height-YBUFFER))
		return 1;

	return 0;
}


void TextPage::configure()
{
	conditionallyInitialiseCanvas();
}



void 
TextPage::conditionallyInitialiseCanvas()
{
	short w = 0;
	short h = 0;
	ViewCanvas *vc;
	if(this->td == 0)
		return;
	
	if(this->initialized_canvas == FALSE) {
		this->initialized_canvas = TRUE;
		// bogus -- guarantees exisitence of canvas
		this->configureCanvas(this, PAGE_WIDTH, PAGE_HEIGHT);
	}
	
	vc = this->getCanvas();
	vc->setClient(this);
	
	if(XtIsRealized(vc->canvas) == FALSE)
		return;
	
	//getWidgetSize(vc->canvas, &this->oldw, &this->oldh);
	getViewableSize(vc->canvas, &this->oldw, &this->oldh);
	this->td->getDimensions(w, h);
	if (w + XMARGIN < this->oldw) {
		if (w + XMARGIN < this->oldw-XBUFFER) {
			this->width = MAX(this->oldw-XBUFFER, PAGE_WIDTH);
		}
	} else {
		this->width =  w + XBUFFER + XMARGIN;
	}
	if (h + YMARGIN < this->oldh) {
		if (h + YMARGIN < this->oldh-YBUFFER) {
			this->height = MAX(this->oldh-YBUFFER, PAGE_HEIGHT);
		}
	} else {
		this->height =  h + YBUFFER + YMARGIN;
	}

	iLBox b = {0,0,w,h};
	this->setPicture(b);
	this->width = MAX(0,MAX(this->width, XBUFFER));
	this->height = MAX(0,MAX(this->height, YBUFFER));

	vc->initializeScrollBarValues();
	vc->setSize(this->width, this->height);

	this->syncSize();
}


iEntityPtr 
TextPage::findEnclosingEntity(int x, int y)
{
	return 0;
}


void
TextPage::findEntityAndIndex(int x, int y,
			      iEntityPtr *entity_return, int *index_return)
{
	iEntityPtr entity = 0;
	int index_var = 0;
	TextElement *element = 0;
	iEntityPtr return_entity = 0;
	
	if(this->td) {
		int over_end = TRUE;
		// bogus -- should do the right thing for empty file
		this->td->getElementAtPosition(x, y, element, index_var, TRUE, over_end);
		if(element)
			return_entity = element->entity;
		if(element)
			entity = element->entity;
		else {
			printf("WARNING: Could not select anything, probably old file?.\n");
		}
	} 
	
	*index_return = index_var;
	*entity_return = entity;
}


/*
// sort of bogus  
// not needed until we create a new function separate from
// typing or allow image and text modes to exist as same time
*/
void 
TextPage::addNewEntity(iEntityPtr entity)
{
}

void
TextPage::prevEntity()
{
	this->clearPopup();
	TextElement *element;
	iEntityPtr entity = 0;
	int index_var = -1;
	
	if(this->td == 0)
		return;
	
	// this code allows me to press left arrow even
	// if a range is selected, not just when a point
	// is selected
	iEntityPtr from_entity;
	iEntityPtr to_entity;
	int from;
	int to;
	this->flag_entity=0;
	this->getSM()->getRange(from_entity, from, to_entity, to);

	// nothing selected so boogie
	if(from_entity == 0)
		return;

	// doesnt have the same problem as next
	// uhh what problem is that?
	entity = from_entity;
	index_var = from;
	
	element = (TextElement *) eaGetFace(entity, vtTEXT);
	short x=0;
	short y=0;
	TextElement *prev_element = element->findPrevElement();

	// if a range is selected, just collapse the range
	if(from_entity!=to_entity||from!=to){
		this->getSM()->setRange(entity, index_var, 
								entity, index_var);
	} else if(index_var > 0) {
		// we can stay inside the current entity
		index_var--;
		this->getSM()->setRange(entity, index_var, 
								entity, index_var);
		//this->centerCursor(element);
	} else {
		// we go to the previous entities end
		element = prev_element;
		if(element){
			int prev_entity_end = prev_element->indexLength() - 1;
			index_var = prev_entity_end;
			this->getSM()->setRange(element->entity, 
									index_var,
									element->entity,
									index_var);
		}
	}

	if(element){
		this->centerCursor(element);
		element->getPosition(this->td, index_var, x, y);
		if(element->type == iNewLineT && index_var != 0)
			x = XMARGIN;
		goal_column = x;
	}
}

/*
// This is a concept of John's
// that hasn't really been implemented
*/
int 
TextPage::isVisible(TextElement *element)
{
	short x,y;
	Widget sw = this->mode->scrolled_window;
	iBox inner_bounds;
	iBox outer_bounds;
	iBox view_bounds;

	short w, h;
	int visible;
	
	element->getPosition(this->td, 0, x, y);
	element->getDimensions(w, h);

	visible = visiblePositionP(sw, x, y) && visiblePositionP(sw, x+w, y+h);
	if(this->display_popup){
		// would the popup be visible too?
		this->decideImageBounds(element->entity, 
								&inner_bounds,
								&outer_bounds);

		view_bounds =
			this->decideViewBounds(element->entity, 
								   outer_bounds);
		visible = (visible && 
				   visiblePositionP(sw, view_bounds.x, view_bounds.y) && 
				   visiblePositionP(sw, 
									view_bounds.x + view_bounds.w, 
									view_bounds.y + view_bounds.h));
	}

	return visible;
}
	

static int 
entityVisible(iEntityPtr entity)
{
	int is_visible = FALSE;
	TextElement *element = (TextElement *) getTextViewer(entity);
	if(element) {
		TextPage *page = (TextPage *) element->page;
		is_visible = page->isVisible(element);
	}
	return is_visible;
}

/*
// this is function is necessary because the selection could be at
// the beginning of a word or a paragraph that is selected
*/
iEntityPtr 
findStartEntity(iEntityPtr from_entity)
{
	iEntityPtr start_entity;
	iEntityPtr f, e;
	int fi, ei;
	iEntityPtr curr;


	for(start_entity = from_entity, 
			curr = from_entity ? i_GetParentEntity(from_entity) : 0; 
		curr; 
		curr = i_GetParentEntity(curr)){

		i_Object2Range(curr, 
					   &f, (int4 *) &fi, 
					   &e, (int4 *) &ei);
		if(f == from_entity)
			start_entity = curr;

	}

	return start_entity;
}

void
TextPage::findNextFlagEntitySub()
{
	// is there already an entity stored in the popup
	iEntityPtr from_entity;
	iEntityPtr to_entity;
	int from_index, to_index;
	int object_from_index, object_to_index;
	iEntityPtr flagged_entity = 0;
	int allow_off_page;
	
	iEntityPtr begin_doc_entity;
	iEntityPtr start_entity;
	int had_popup_before = FALSE;

	this->getSM()->getRange(from_entity, from_index, to_entity, to_index);	

	// find the existing popup and clear it
	if(this->flag_entity && from_entity) {
		// so we must compare the bbox of elements
		// and we can't extract from the flag entity until
		// we are sure it hasn't been deleted
		int ancestorP(iEntityPtr child, iEntityPtr would_be_ancestor);
		if(from_entity == this->flag_entity || ancestorP(from_entity, this->flag_entity)){
			iEntityPtr ffe, fte;
			int ffi, fti;
			i_Object2Range(this->flag_entity, 
						   &ffe, (int4 *) &ffi, 
						   &fte, (int4 *) &fti);
			if(ffe == from_entity &&
			   ffi == from_index &&
			   fte == to_entity &&
			   fti == to_index) {
				// before this, the flag was selected using ctrl-g
				// and wasn't edited
				// therefore, we clear this flag and continue from here
				
				// must happen before we clear the popup
				iEntityPtr holdit = this->flag_entity;
				this->clearPopup();
				entityClearFlag(this->mode->main_window, holdit);
				had_popup_before = TRUE;
			}
		}
	}
	begin_doc_entity = this->td->entity;

	// we start our search in the nearest parent
	// who would have the same from_entity
	start_entity = findStartEntity(from_entity);
	// now, if we have no from entity, start at root

	if(start_entity == 0)
		// this is ok because we won't be needing
		// it in the reset last flagged guy
		// at the end
		from_entity = begin_doc_entity;

	for(int i = 0; flagged_entity == 0 && i < 4; i++) {
		// 4 times through
		iEntityPtr loop;
		
		loop = start_entity;

		if(start_entity == begin_doc_entity && (i == 0 || i == 2))
			continue; // the succeeding loop will do this anyway

		if(i == 1 || i == 3) // odd times -- start from the beginning
			loop = begin_doc_entity;

		allow_off_page = (i == 2 || i == 3); // last 2 times allow off page

		
		// now, if the current start_entity is still a flagged entity
		// we select
		if(loop && entityFlaggedP(loop) && (entityVisible(loop) || allow_off_page)) {
			flagged_entity = loop;
		} else { // do an iEntityFindForward
			while(loop) {
				loop = 
					i_FindEntityForward(loop,0, PROPERTY, (void*)FLAG_PROP);
				if(loop && (entityVisible(loop) || allow_off_page))
					break;
			}
			flagged_entity = loop; // loop may be null
		}
	}

	this->flag_entity = flagged_entity;

	if(flagged_entity){
		i_Object2Range(flagged_entity, 
					   &start_entity, (int4 *) &from_index,
					   &to_entity, (int4 *) &to_index);
		from_index = object_from_index;
		to_index = object_to_index;
		this->getSM()->setRangeAdjacent(start_entity, from_index, 
								to_entity, to_index);
		if(entityVisible(flagged_entity) == FALSE) {
		   TextElement *element = (TextElement *) getTextViewer(flagged_entity);
		   if(element)
			   this->centerCursor(element);		
		}
	}

	if(flagged_entity == 0 && had_popup_before){
		// we just cleared the last flag, and don't want anything
		// to be selected now
		// heinous - need to reset selection so it will get refreshed when cleared
		this->getSM()->setRange(from_entity, from_index, to_entity, to_index);
		this->getSM()->clear();
	}
}


void
TextPage::findNextFlagEntity()
{	
	this->findNextFlagEntitySub();
	this->setPopup(this->flag_entity);
}

/* bogus -- very simplfied for the case of line entities
// and doesnt let you go over boundaries
*/
void
TextPage::nextEntity()
{
	this->clearPopup();
	iEntityPtr entity;
	TextElement *element;
	int index_var = -1;
	
	if(this->td == 0)
		return;

	// arrow key works on range as well as point
	iEntityPtr from_entity;
	iEntityPtr to_entity;
	int from;
	int to;
	this->getSM()->getRange(from_entity, from, to_entity, to);

	// hmm this is a bit different
	// why would there be no to_entity
	if(from_entity == 0)
		return;
	if(to_entity) {
		entity = to_entity;
		index_var = to;
	} else {
		entity = from_entity;
		index_var = from;
	}
	
	this->flag_entity=0;
	element = (TextElement *) eaGetFace(entity, vtTEXT);
	short x;
	short y;
	
	TextElement *next_element = element->findNextElement();
	
	int entity_end = element->indexLength() - 1;
	
	if(from_entity != to_entity || from != to) {
		// just collapse the range
		this->getSM()->setRange(entity, index_var, 
								entity, index_var);
	} else if(index_var < entity_end) {
		index_var++;
		this->getSM()->setRange(entity, index_var, 
								entity, index_var);
		//this->centerCursor(element);
	} else {
		element = next_element;
		if(element){
			index_var = 0;
			//element->getPosition(this->td, index_var, x, y);
			this->getSM()->setRange(element->entity, index_var, 
									element->entity, index_var);
		}
	}
	if(element) {
		this->centerCursor(element);
		element->getPosition(this->td, index_var, x, y);
		if(element->type == iNewLineT && index_var != 0)
			x = XMARGIN;
		goal_column = x;
	}
}

void
TextPage::upEntity()
{
	this->clearPopup();
	iEntityPtr entity;
	
	int index_var = -1;
	
	if(this->td == 0)
		return;
	
	// arrow key works on range as well as point	
	iEntityPtr from_entity;
	iEntityPtr to_entity;
	int from;
	int to;
	this->flag_entity=0;
	this->getSM()->getRange(from_entity, from, to_entity, to);
	if(from_entity == 0)
		return;
	entity = from_entity;
	index_var = from;
	
	if(entity) {
		TextElement *element;
		TextElement *up_element = 0;
		int up_index = 0;
		element = (TextElement *) eaGetFace(entity, vtTEXT);
		short element_x;
		short element_y;
		
		element->getPosition(this->td, index_var, element_x, element_y);
		if(element->type == iNewLineT && index_var != 0)
			element_x = XMARGIN;
		element_y -= charHeight('M');
		element_y = MAX(0 , element_y);
		element_x = this->goal_column;
		int over_end = TRUE, is_root=TRUE;
		this->td->getElementAtPosition(element_x, element_y,
									   up_element, up_index,
									   is_root, over_end);
		this->getSM()->setRange(up_element->entity, up_index, 
								up_element->entity, up_index);
		this->centerCursor(up_element);
		//this->centerOnSelected(up_element->entity);
	}
}

void
TextPage::downEntity()
{
	this->clearPopup();
	iEntityPtr entity;
	
	int index_var = -1;
	
	if(this->td == 0)
		return;

	// arrow key works on range as well as point	
	iEntityPtr from_entity;
	iEntityPtr to_entity;
	int from;
	int to;
	this->flag_entity=0;
	this->getSM()->getRange(from_entity, from, to_entity, to);
	if(from_entity == 0)
		return;
	entity = to_entity;
	index_var = to;
	
	if(entity) {
		TextElement *element;
		TextElement *down_element = 0;
		int down_index = 0;
		element = (TextElement *) eaGetFace(entity, vtTEXT);
		short element_x;
		short element_y;
		
		element->getPosition(this->td, index_var, element_x, element_y);
		if(element->type == iNewLineT && index_var != 0)
			element_x = XMARGIN;
		element_y += charHeight('M');
		element_x = this->goal_column;
		//element_y = MAX(0 , element_y);
		int over_end = TRUE, is_root = TRUE; 
		this->td->getElementAtPosition(element_x, element_y,
									   down_element, down_index,
									   is_root, over_end);
		this->getSM()->setRange(down_element->entity, down_index, 
								down_element->entity, down_index);
		this->centerCursor(down_element);
		//this->centerOnSelected(down_element->entity);
	}
}

iBox findBox(TextPage *, TextElement *element);

/*
// bogus what is the difference between center cursor and center on selected?
*/

void
TextPage::centerOnSelected(iEntityPtr entity)
{
	TextElement *element;
	iBox box;
	
	if(this->td == 0)
		return;
	
	iEntityPtr from_entity;
	int from;
	TextElement *from_element;
	tfGetThinCursor(this->getSM(), from_entity, from);
	if(from_entity){
		from_element = (TextElement *) eaGetFace(from_entity, vtTEXT);
		this->centerCursor(from_element);
		return;
	}
	
	if((element = (TextElement *) eaGetFace(entity, vtTEXT)) && 
	   (box = findBox(this, element)).w) {
		Widget sw = this->mode->scrolled_window;
		if(visiblePositionP(sw, box.x, box.y) == FALSE || 
		   // bogus should be the width of the character at that point
		   visiblePositionP(sw, box.x + charWidth(' '), 
							box.y + charHeight(' ')) == FALSE) 
;
this->getCanvas()->scrollToPosition(sw, 
							 box.x + box.w/2, 
							 box.y + box.h/2);
	}
}


void
TextPage::pressButton(XEvent *event)
{
	TextElement *element;
	iEntityPtr entity;
	int index_var;
	
	if(this->td == 0)
		return;
	this->flag_entity=0; 
	event->xbutton.x += this->getCanvas()->pix_hoffset;
	event->xbutton.y += this->getCanvas()->pix_voffset;
    switch(event->xbutton.button) {
	case Button1:
		goal_column = event->xbutton.x;
		findEntityAndIndex(event->xbutton.x, event->xbutton.y,
								   &entity, &index_var);
		if(entity != 0) {
			// an old idea from image mode that doesnt apply here
			// some day we may want shift click to extend a preexisintg
			// selection as is customary in many applications
#if 0
			if(event->xbutton.state & ShiftMask) ;
#endif
#ifdef TEXTDEBUG
			prlines(entity);
#endif
			element = (TextElement *) eaGetFace(entity, vtTEXT);

			this->clearPopup();
			this->getSM()->setRange(entity, index_var, 
									entity, index_var);

		} else {
			// this should never happen in text Mode
			this->clearPopup();

			this->getSM()->clear();
		}
		break;
	case Button2:
		
		break;
    }
}

void
TextPage::releaseButton(XEvent *event)
{
	TextElement *element;
	iEntityPtr entity;
	int index_var;
	
	if(this->td == 0)
		return;
	
	event->xbutton.x += this->getCanvas()->pix_hoffset;
	event->xbutton.y += this->getCanvas()->pix_voffset;
    switch(event->xbutton.button) {
	case Button1:
		findEntityAndIndex(event->xbutton.x, event->xbutton.y,
								   &entity, &index_var);

		if(entity != 0) {
			if(event->xbutton.state & ShiftMask){
				return;
			} else {
				element = (TextElement *) eaGetFace(entity, vtTEXT);

				int from_index;
				int to_index;
				iEntityPtr from_entity;
				iEntityPtr to_entity;
				this->getSM()->getRange(from_entity, from_index,
										to_entity, to_index);
										
			    // this happened because we got release without a press
				if(from_entity == 0)
					return;
				
				TextElement *from_element;
				from_element = (TextElement *) eaGetFace(from_entity, vtTEXT);
				
				short from_x, from_y;
				short x, y;
				
				element->getPosition(this->td, index_var, 
									 x, y);
				
				from_element->getPosition(this->td, from_index, 
										  from_x, from_y);
				
				if(from_y < y || ((from_y == y) && (from_x < x))){
					to_entity = entity;
					to_index = index_var;
				} else {
					to_entity = from_entity;
					to_index =  from_index;
					from_entity = entity;
					from_index = index_var;
				}
				
				this->getSM()->setRange(from_entity, from_index, 
										to_entity, to_index);
			}
		} 
		break;
	case Button2:
		break;
    }
}

void
TextPage::moveButton(int state, int x, int y)
{
	TextElement *element;
	iEntityPtr entity;
	int index_var;
	
	if(this->td == 0)
		return;
	
	if(state & Button1Mask) {
		findEntityAndIndex(x, y,
				   &entity, &index_var);
		if(entity != 0) {
			if(state & ShiftMask){
				return;
			} else {
				element = (TextElement *) eaGetFace(entity, vtTEXT);

				int from_index;
				int to_index;
				iEntityPtr from_entity;
				iEntityPtr to_entity;
				this->getSM()->getRange(from_entity, from_index,
										to_entity, to_index);
				// bogus how did from_entity get to be zero
				if(from_entity && (to_entity != entity || to_index != index_var)) {
					this->getSM()->setRange(from_entity, from_index, entity, index_var);
					// XmUpdateDisplay(this->getCanvas()->canvas);
				}
			}
		} 
	}
}


int
TextPage::pressKey(iUCode key_value)
{
	if((this->getSM()->getState() == scNOT) || 
	   ((key_value != '\n' && key_value != '\b') &&
		key_value < ' ')) {
		ringTheBell();
		return 0;
	}		

	int from;
	short x, y;
	
	iEntityPtr from_entity, to_entity;
	int from_index, to_index;
	
	if(this->td == 0) {
		return 0;
	}
	
	this->flag_entity=0;
	// we handle this here instead of in 
	// the callback because we need to feed KeyBoard
	// through here -- therefore callbacks call us
	if (key_value == '\b') {
		this->mode->main_window->deleteSelection(TRUE);
		tfGetThinCursor(this->getSM(), from_entity, from);
		if(from_entity){
			TextElement *from_element = 
				(TextElement *) eaGetFace(from_entity, vtTEXT);
			this->centerCursor(from_element);
			from_element->getPosition(this->td, from, x, y);
			this->goal_column = x;
		}
		return 1;
	}
	
	if(key_value == '\0') {
		return 0;
	}

	// trap tab key here so it is not inserted into text strings
	// (it is no longer used to move to next flagged entity)
	if(key_value == '\t') {
		return 0;
	}

	this->getSM()->getRange(from_entity, from_index, to_entity, to_index);
	if(from_entity){
		this->getSM()->_clear();
		i_Edit(&from_entity, (int4 *) &from_index, 
			   &to_entity, (int4 *) &to_index, 
			   key_value); 
		TextElement *element = (TextElement *) eaGetFace(from_entity, vtTEXT);
		const int draw_page = (from_entity != to_entity || 
		 		       from_index != to_index || 
				       key_value == '\n' || 
				       (element->type == iGlyphT) ||
				       (element->type == iNewLineT && from_index != 0));
		const int draw_line = (element->type != iNewLineT);

		if(draw_page)
			jiggerToEndOfPage(element, from_index);
		else
			jiggerToEndOfLine(element, from_index);
		element->getPosition(this->td, from_index, x, y);
		this->goal_column = x;
			
		if(draw_page) 
			redrawToEndOfPage(element);
		else if (draw_line)
			redrawToEndOfLine(element);
		else
			redrawElement(element);
		
		if(key_value == '\n')
			this->clearPopup();
		this->getSM()->setRange(from_entity, from_index, from_entity, from_index);
		this->mode->main_window->document->setDirty();
		// bogus -- delay drawing if window will scroll
		this->centerCursor((TextElement *) eaGetFace(from_entity, vtTEXT));
	} else {
		this->reassignValues(this->getSM()->getItems(), key_value);
	}

	return 1;
}

void
TextPage::centerCursor(TextElement *element)
{
	if(!element || !this->td) return;
	iEntityPtr sole_entity = element->entity;

	int from = -1, to = -2;
	if(getEntitySelection(sole_entity) == scRANGE) 
		getEntitySelectionRange(sole_entity, from, to);
	if(from != to) from = 0;
	
	TextElement *element = (TextElement *) eaGetFace(sole_entity, vtTEXT);

	// determine position of selected character
	short x, y;
	element->getPosition(this->td, from,  x, y);
	if(element->type == iNewLineT && from != 0)
		x = XMARGIN;

	// estimate size of selected character (fixed pitch)
	int cw = charWidth('M'), ch = charHeight('M');
	
	// make sure popup (if any) and selected character are visible
	Widget sw = this->mode->scrolled_window;
	if (this->display_popup) {
		int popup_visible=TRUE, upper_left_visible, lower_right_visible;
		iBox inner_bounds, outer_bounds, view_bounds;

		this->decideImageBounds(element->entity, 
								&inner_bounds,
								&outer_bounds);
		view_bounds =
			this->decideViewBounds(element->entity, 
								   outer_bounds);
		upper_left_visible = visiblePositionP(sw, view_bounds.x, view_bounds.y-ch);
		lower_right_visible = visiblePositionP(sw,
                                                                        view_bounds.x + view_bounds.w,
                                                                        view_bounds.y + view_bounds.h);
		popup_visible = upper_left_visible && lower_right_visible;
	
		if (!popup_visible) {
			if (upper_left_visible)
				this->getCanvas()->scrollToPosition(sw, view_bounds.x + view_bounds.w, view_bounds.y + view_bounds.h);
			else 
				if(lower_right_visible)
					this->getCanvas()->scrollToPosition(sw, view_bounds.x, view_bounds.y-ch);
				else
					// window too small for whole popup - make selected character visible
					this->getCanvas()->scrollToPosition(sw, x, y);
		}
	} 
	// whether or not popups are active, make sure we can see the selected character
	int char_visible = visiblePositionP(sw, x, y) && visiblePositionP(sw, x+cw, y+ch);
	if (!char_visible) {
		// scroll to position of selected character
		this->getCanvas()->scrollToPosition(sw, x, y);
	}
}

void
TextPage::manage()
{
	TextElement *element = 0;
	short x, y;
	iEntityPtr from_entity, to_entity;
	int from_index, to_index;
	this->getSM()->convertToRange();

	this->getSM()->getRange(from_entity, from_index,
							to_entity, to_index);
	if(from_entity)
		element = (TextElement *) eaGetFace(from_entity, vtTEXT);
	
	if(element){
		element->getPosition(this->td, from_index, x, y);
		if(element->type == iNewLineT && from_index != 0)
			x = XMARGIN;
		goal_column = x;
	}

	jiggerEverything(this);
	this->Page::manage();
	XmProcessTraversal(this->getCanvas()->canvas, 0);
	this->getCanvas()->setCursor(XC_xterm);
}

void
TextPage::unmanage()
{
	this->getSM()->convertToItem();
	// we are doomed but we might as well do this
	this->clearPopup();
	
	this->getCanvas()->unsetCursor();
	this->Page::unmanage();
}



/* these useless magnification functions are virtual requirements */
void 
TextPage::setMag(int x, int ox)
{
}

void 
TextPage::_setMag(int x, int ox)
{
}
