
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#define CORRECT_LINE_DEFINITION

#include "rafport.h"

#include "illum.h"
#include "misc/bb.h"

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <X11/cursorfont.h>

#include "rafutil.h"
#include "smemory.h"

#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include "misc/Unicode.h"
#include "misc/XArg.h"
#include "misc/misc.h" 
#include "misc/scroll.h"
#include "font.h"
#include "MainWindow.h"

#include "ViewCanvas.h"
#include "Mode.h"
#include "EntityAux.h"
#include "MainWindow.h"
#include "Document.h"
#include "SelectionManager.h"

#include "TreeElement.h"
#include "TreePage.h"



void TreePage::initTree(iEntityPtr en,int lev,int shown)
{
	TreeElement *elem,*par;

	iEntityPtr cur,parent;
	int x,y;
	
	if (parent = i_GetParentEntity(en)){
		par = (TreeElement*)eaGetFace(parent,vtTREE);
#ifndef IGNORE_ORS
		if (i_GetLogicalType(parent)==iOr)
			x = par->x;
		else
#endif
			x = par->x + (HORZSPACER*cw);
	} else {
		x = 0;
	}
	y = this->nodes * this->vspacer;

#ifndef IGNORE_ORS
	EntityAux *ea;
	if (i_GetLogicalType(en)==iOr){
		ea = getController(en);
		elem = new TreeElement(this,en, x, y,ea->show,lev);
		elem->or_path = 0;
		eaSetFace(en,vtTREE,elem);
		en = getOrChild(en);
		this->nodes--;
	}
#endif
	if (!eaGetFace(en,vtTREE))	{
		EntityAux *ea = getController(en);
		elem = new TreeElement(this,en, x, y,ea->show,lev);
	}
	elem->getBoxTextandWidth();
	
	elem->height = ch;					
#ifndef IGNORE_ORS
	elem->or_path = 0;
#endif
	eaSetFace(en,vtTREE,elem);
	if (shown) {
		if (this->maxlev < lev) this->maxlev = lev;
		this->nodes++;
		if (!i_GetChildEntity(en))
			this->leaves ++;
	}
	for (cur = i_GetChildEntity(en);cur;cur = i_GetNextEntity(cur)){
		this->initTree(cur,lev+1,shown && elem->show);
	}
}

TreePage::TreePage(Mode *input_mode, 
					 const aString& label, 
					 iEntityPtr input_root_entity)
	: Page(label, input_mode),
	  initialized_canvas(0), 
	  root_entity(input_root_entity),
	  maxlev(0),
	  nodes(0),
	  leaves(0)
{
	ViewCanvas *vc;
	iUCode uni_m[2] = { 'M', 0};
	dllistCtor(&this->xor_boxes);
	vc = this->getCanvas();
	vc->measureText(uni_m, 1, &this->cw, &this->ch, &this->asc);
	this->vspacer = ch + VERTSPACER;
	XArg args;
	args.setArg(XmNincrement,vspacer);
	args.setArg(XmNpageIncrement, vspacer * 10);
	args.setValues(input_mode->main_window->vsb);
	this->hspacer = HORZSPACER * cw;
	this->initTree(input_root_entity,0,1);
	this->maxlev++;
	this->height = (this->nodes + 1) * this->vspacer;
	this->xor_box.w = 0;
	this->width = this->maxlev * this->hspacer + MAXBOXWIDTH*cw + (2*LEFTPAD);
	this->stale = 1;
	this->go = 0;
#ifndef IGNORE_ORS
	this->orMenu = 0;
#endif
	this->last_x = this->first_x = -1;
	xor_box.x = xor_box.y = xor_box.w = xor_box.h = 0;
}
#ifdef DEBUG
void prTree(iEntityPtr en)
{
	iEntityPtr cur;
	TreeElement *elem;
	TreePage *page;
	elem = (TreeElement*)eaGetFace(en,vtTREE);	
	page = (TreePage*)elem->page;
#ifndef IGNORE_ORS
	if(i_GetLogicalType(en)==iOr){
		printf("iOr\n");
		en = page->getOrChild(en);
	}
#endif
	elem = (TreeElement*)eaGetFace(en,vtTREE);	
	printf("%s %s lev:%d y:%d\n",i_TypeToName(i_GetEntityType(en)), 
		(i_GetLogicalType(en)==iAnd?"And":"Or"),elem->lev, elem->y);
	for (cur = i_GetChildEntity(en);cur;cur = i_GetNextEntity(cur))
		prTree(cur);
}
#endif	

void 
TreePage::deleteTree(iEntityPtr en)
{

	TreeElement *elem = 0;
	EntityAux *ea;
	iEntityPtr cur;
	
#if 0
	iEntityPtr next;
	if(i_GetLogicalType(en)==iOr){
		next = getOrChild(en);
		elem = (TreeElement*)eaGetFace(en,vtTREE);	
		ea  = getController(en);
		ea->show = elem->show;
		eaSetFace(en,vtTREE,0);
		delete elem;
		en = next;
	}
#endif
	elem = (TreeElement*)eaGetFace(en,vtTREE);	
	if (!elem) return;
	ea  = getController(en);
	ea->show = elem->show;
	eaSetFace(en,vtTREE,0);
	delete elem;
	for (cur = i_GetChildEntity(en);cur;cur = i_GetNextEntity(cur))
		deleteTree(cur);
}
void 
freeBoxlist(DLList *items){
	iLBox *box;
	Ref r;
	for (r = dllistFirst(items); r; dllistNext(items,&r)) {
		box = (iLBox*)dllistContents(items,r);
		smfree(box);
	}
	dllistDtor(items);
}

TreePage::~TreePage()
{
	XArg args;
	args.setArg(XmNincrement,10);
	args.setArg(XmNpageIncrement, 10);
	args.setValues(this->mode->main_window->vsb);
	this->flushRedrawRegion();
	deleteTree(root_entity);
	freeBoxlist(&this->xor_boxes);
}

int
TreePage::resizeNecessaryP()
{
	return 0;
}

void 
TreePage::conditionallyInitialiseCanvas()
{
	ViewCanvas *vc = this->getCanvas();
	iLBox b = {0,0,0,0};
	
	if(this->initialized_canvas == FALSE) {
		this->initialized_canvas = TRUE;
		b.w = this->width;
		b.h = this->height;
		this->setPicture(b);
		this->syncSize();
		vc->initializeScrollBarValues();
	}
	vc->setClient(this);
}

void 
TreePage::redisplay(XmDrawingAreaCallbackStruct *call_data)
{
	ViewCanvas *vc = this->getCanvas();
	Widget w = vc->canvas;
	XEvent *event = (XEvent *) call_data->event;
	
	static Region region = NULL;
	
	if(XtIsManaged(w) == FALSE)
		return;
	
	if(!region)
		region = XCreateRegion();
	
	XtAddExposureToRegion(event, region);
	
	if(event->xexpose.count != 0)
		return;
	
	iLBox refresh_box;
	
	XRectangle rect;
	XClipBox(region, &rect);
	XDestroyRegion(region);
	region = 0;
	
	refresh_box.x = rect.x;
	refresh_box.y = rect.y;
	refresh_box.w = rect.width;
	refresh_box.h = rect.height;
	
	refresh_box.x += vc->pix_hoffset;
	refresh_box.y += vc->pix_voffset;
	if(refresh_box.w == 0 || refresh_box.h == 0)
		return;
	
	this->drawRegionNow(refresh_box);
}
static int 
parentVisibleP(TreeElement *elem, iLBox *bp, int scroll_offset)
{
	iEntityPtr parent = i_GetParentEntity(elem->entity);
	TreeElement *par_elem;
	
	if (parent) {
		par_elem = (TreeElement*)eaGetFace(parent,vtTREE);
		return ((par_elem->y - scroll_offset) < (bp->y + bp->h - scroll_offset));
	} else
		return ((0 - scroll_offset) < (bp->y + bp->h - scroll_offset));
}

int TreePage::drawEntities(Display *display,Pixmap pixmap,
		iLBox *bp,iEntityPtr en)
{
	iEntityPtr cur;
	TreeElement *elem;
	int more=1;
	ViewCanvas *vc = this->getCanvas();
	iLBox visible_rectangle = visibleRectangle(this->mode->scrolled_window);
	int extra_fudge = LEFTPAD + 3;    //add a little to redisplay for line_width

#ifndef IGNORE_ORS
	if (i_GetLogicalType(en)==iOr)
		en = getOrChild(en); 
#endif
	elem  = (TreeElement*)eaGetFace(en,vtTREE);
#if 0
	// bogus -- for debugging only
	if(elem == 0){
	  	elem  = (TreeElement*)eaGetFace(en,vtTREE);
	}
#endif
	eaSetFace(en,vtTREE,elem);  //thisis only a test
	if ((elem->y + ch + extra_fudge - vc->pix_voffset) > bp->y
		&& parentVisibleP(elem,&visible_rectangle, vc->pix_voffset))
		more = elem->drawAnEntity(display,pixmap);
	if (elem->show) {
		for (cur = i_GetChildEntity(en);cur ;cur = i_GetNextEntity(cur))
			more = drawEntities(display,pixmap,bp,cur);
	}
	return more;
}

void
TreePage::refreshSel(iEntityPtr en)
{
	ViewCanvas *vc = this->getCanvas();
	Widget widget = vc->canvas;
	TreeElement *elem,*ce;
	iEntityPtr cur,child;
	
	if(!en) return;
	elem = (TreeElement*)eaGetFace(en,vtTREE);
	for (cur = i_GetParentEntity(en);cur;cur = i_GetParentEntity(cur)) {
#ifndef IGNORE_ORS
		if (i_GetLogicalType(cur) == iOr)
			cur = i_GetParentEntity(cur);
#endif
		ce = (TreeElement*)eaGetFace(cur,vtTREE);
		if (!ce->show) elem = ce;
	}

	int x = elem->x - vc->pix_hoffset - 3;
	int y = elem->y - vc->pix_voffset - 3;

	XFillRectangle(XtDisplay(widget),XtWindow(widget), 
				   vc->blankGC(),
				   x,  y,
				   this->width+6, 3+vspacer); //add two pixels to get rid of thickGC
	elem->drawAnEntity(XtDisplay(widget),XtWindow(widget));

	// bogus help
#ifndef IGNORE_ORS
	if ((child=getOrChild(en)) && elem->show){
		if (i_GetLogicalType(child)==iOr)
			child = getOrChild(child);
		elem = (TreeElement*)eaGetFace(child,vtTREE);
		if (elem)
			elem->drawAnEntity(XtDisplay(widget),XtWindow(widget));
	}
#else
	if ((child=i_GetChildEntity(en)) && elem->show){
	  elem = (TreeElement*)eaGetFace(child,vtTREE);
	  if (elem)
		elem->drawAnEntity(XtDisplay(widget),XtWindow(widget));
	}
#endif
		
}

void
TreePage::drawRegionNow(iLBox refresh_box)
{
	ViewCanvas *vc = this->getCanvas();
	Widget widget = vc->canvas;
	int4 h;
	iLBox b = {0,0,0,0};
	
	refresh_box.x -= vc->pix_hoffset;
	refresh_box.y -= vc->pix_voffset;
	Pixmap pixmap;
	iLBox visible_rect = visibleRectangle(this->mode->main_window->scroll_window);
	int depth = DefaultDepthOfScreen(XtScreen(widget));
	pixmap = XCreatePixmap(XtDisplay(widget), XtWindow(widget), 
						   visible_rect.w, visible_rect.h, 
						   depth);
	
	XFillRectangle(XtDisplay(widget),pixmap, 
				   vc->blankGC(),
				   0, 0,
				   visible_rect.w, visible_rect.h);
	if (stale) {
		this->nodes = 0;
		this->leaves = 0;
		this->maxlev = 0;
		updateTree(root_entity,0);
		this->maxlev++;
		h = (this->nodes + 1) * this->vspacer;
		this->height = h;
		this->width = this->maxlev * this->hspacer + MAXBOXWIDTH*cw + (2*LEFTPAD);
		this->stale = 0;
		b.w = this->width;
		b.h = this->height;
		this->setPicture(b);
		this->syncSize();
	}
	drawEntities(XtDisplay(widget),pixmap,
		&refresh_box,this->root_entity);

	// redraw the XOR
	if(this->go)
		this->drawXor(pixmap);
	XCopyArea(XtDisplay(widget), pixmap, XtWindow(widget), vc->copyGC(),
			  0, refresh_box.y,
			  visible_rect.w, visible_rect.h,  
			  visible_rect.x-vc->pix_hoffset, refresh_box.y);
	XFreePixmap(XtDisplay(widget), pixmap);
}

int TreePage::containsSelection(iEntityPtr en)
{
	iEntityPtr cur;
	
	if (getEntitySelection(en) == scITEM)
		return 1;
	for (cur = i_GetChildEntity(en);cur;cur = i_GetNextEntity(cur))
		if (containsSelection(cur))
			return 1;
	return 0;
}

void
countNodes(TreePage *tp,iEntityPtr en,iEntityPtr my_entity)
{
	iEntityPtr cur;
	
	tp->nodes++;
	for (cur = i_GetParentEntity(my_entity);cur;cur = i_GetPrevEntity(cur)){
		countNodes(tp,cur,cur);
	}
}

void TreePage::updateTree(iEntityPtr en,int lev)
{
	TreeElement *elem,*par_elem;
	iEntityPtr cur,parent;
	int x;
	
	if (parent = i_GetParentEntity(en)){
		par_elem = (TreeElement*)eaGetFace(parent,vtTREE);
		x = par_elem->x + (cw * HORZSPACER);
	} else x = 0;
	x = lev * hspacer;
#ifndef IGNORE_ORS
	TreeElement *or_elem;
	int nodes;
	iEntityPtr prev_or;
	if (i_GetLogicalType(en)==iOr){
		or_elem = (TreeElement*)eaGetFace(en,vtTREE);
		if (or_elem->or_path != or_elem->prev_or_path){
			prev_or = getOrPath(en,or_elem->prev_or_path);
			par_elem = (TreeElement*)eaGetFace(prev_or,vtTREE);
			nodes = this->nodes = par_elem->y/vspacer;
			deleteTree(prev_or);
#ifndef IGNORE_ORS
			en = getOrChild(en);
#endif
			initTree(en,0,or_elem->show);
			this->nodes = nodes;
			or_elem->prev_or_path = or_elem->or_path;
		} else 
			en = getOrChild(en);
	}
#endif
	elem = (TreeElement*)eaGetFace(en,vtTREE);
	if (!elem)
		elem = new TreeElement(this,en, 0, 0,1,lev);
 
	eaSetFace(en,vtTREE,elem);
	elem->y = this->nodes * this->vspacer + TOPPAD;
	elem->x = x + LEFTPAD;
	elem->getBoxTextandWidth();
	elem->height = ch;
	this->nodes++;     
	if (elem->show){
	if (this->maxlev < lev) this->maxlev = lev;
	if (!i_GetChildEntity(en)) this->leaves++;

		for (cur = i_GetChildEntity(en);cur;cur = i_GetNextEntity(cur))
			updateTree(cur,lev+1);
	}
}

#ifndef IGNORE_ORS
iEntityPtr TreePage::getOrPath(iEntityPtr en,int path)
{
	int i;
	iEntityPtr cur;
	for (i=0,cur=i_GetChildEntity(en);cur 
				&& i < path;cur = i_GetNextEntity(cur),i++);
	return cur;
}
#endif

#ifndef IGNORE_ORS
iEntityPtr TreePage::getOrChild(iEntityPtr en)
{
	int i;
	iEntityPtr cur=0;
	TreeElement *elem = (TreeElement*)eaGetFace(en,vtTREE);
	for (i=0,cur=i_GetChildEntity(en);cur 
				&& i < elem->or_path;cur = i_GetNextEntity(cur),i++);
	return cur;
}
#endif
		

iEntityPtr TreePage::FindEntity(iEntityPtr en,int x, int y)
{
	iEntityPtr cur;
	TreeElement *elem;
	
#ifndef IGNORE_ORS
	if(i_GetLogicalType(en)==iOr)
		en = getOrChild(en);
#endif
	elem  = (TreeElement*)eaGetFace(en,vtTREE);
	if (elem->y <= y && ((elem->y + elem->height) >= y) 
		 && (elem->x <= x) && (elem->x + elem->width >= x))
		 return en;
	if (elem->show)
		for (cur = i_GetChildEntity(en);cur; cur = i_GetNextEntity(cur))
			if (en = FindEntity(cur,x,y))
				return en;
	return 0;
}
iEntityPtr TreePage::findClose(iEntityPtr en, int y)
{
	iEntityPtr cur;
	TreeElement *elem;
	
#ifndef IGNORE_ORS
	if(i_GetLogicalType(en)==iOr)
		en = getOrChild(en);
#endif
	elem  = (TreeElement*)eaGetFace(en,vtTREE);
	if (elem->y <= y && ((elem->y + elem->height) >= y))
		 return en;
	if (elem->show)
		for (cur = i_GetChildEntity(en);cur; cur = i_GetNextEntity(cur))
			if (en = findClose(cur,y))
				return en;
	return 0;
}

iEntityPtr 
TreePage::findEnclosingEntity(int x, int y)
{
	return FindEntity(root_entity,x,y);
}
iEntityPtr 
TreePage::findCloseEntity(int y)
{
	return findClose(root_entity,y);
}

void 
TreePage::addNewEntity(iEntityPtr entity)
{
	TreeElement *elem;
	elem = new TreeElement(this,entity, 0, 0,1,0);
	eaSetFace(entity, vtTREE, elem);
	this->stale = 1;
}
int 
parentsShow(TreeElement *elem)
{
	int show = TRUE;
	iEntityPtr parents;

	parents = i_GetParentEntity(elem->entity);
#ifndef IGNORE_ORS
	if (i_GetLogicalType(parents)==iOr)
		parents = i_GetParentEntity(elem->entity);
#endif
	
	while(show && parents){
		elem = (TreeElement *) eaGetFace(parents, vtTREE);
		show = elem && elem->show;
		parents = i_GetParentEntity(parents);
#ifndef IGNORE_ORS
		if (i_GetLogicalType(parents)==iOr)
			parents = i_GetParentEntity(elem->entity);
#endif
	}

	return show;
}

void
centerIfNecessary(TreePage *page, iEntityPtr entity)
{
	TreeElement *elem;
	
	elem = (TreeElement *) eaGetFace(entity, vtTREE);
	if(elem){
		Widget sw = page->mode->main_window->scroll_window;
		if(visiblePositionP(sw, elem->x, elem->y) == FALSE ||
		   visiblePositionP(sw, elem->x, elem->y + page->ch+TOPPAD) == FALSE)
			page->centerOnSelected(entity);
	}
}
static
iEntityPtr getLastEntity(iEntityPtr en)
{
	iEntityPtr last = en;
	
	while (en = i_GetLastEntity(en))
		last = en;
	return last;
}

void
TreePage::prevEntity()
{
	iEntityPtr sole_entity;
	iEntityPtr entity=0;
	Ref r;
	int good = FALSE;
	const DLList *selection=this->getSM()->getItems();

	r = dllistLast(selection);
	if (r)
		sole_entity = (iEntityPtr) dllistContents(selection, r);
	else
		sole_entity=0;

	entity = sole_entity;
	do {
		entity = (sole_entity ? i_GetPrevEntity(entity) 
			: getLastEntity(this->root_entity));
		if(entity){
			TreeElement *elem = (TreeElement *) eaGetFace(entity, vtTREE);
#ifndef IGNORE_ORS
			if(elem && parentsShow(elem)&&i_GetLogicalType(entity)==iAnd) 
				good = TRUE;
#else
			if(elem && parentsShow(elem)) 
				good = TRUE;
#endif
		}
	} while(entity && good == FALSE);
	if (entity && good) {
		this->getSM()->setItem(entity);
		centerIfNecessary(this, entity);
	}
}

void
TreePage::findNextFlagEntity()
{	
}

void
TreePage::nextEntity()
{
	iEntityPtr sole_entity;
	iEntityPtr entity=0;
	Ref r;
	int good = FALSE;
	
	const DLList *selection=this->getSM()->getItems();

	r = dllistLast(selection);
	if (r)
		sole_entity = (iEntityPtr) dllistContents(selection, r);
	else
		sole_entity=0;

	if(1) {
		entity = sole_entity;
		do {
			entity = (sole_entity ? i_GetNextEntity(entity)
					 : this->root_entity);
			if(entity){
				TreeElement *elem = (TreeElement *) eaGetFace(entity, vtTREE);
#ifndef IGNORE_ORS
				if(elem && parentsShow(elem)&&i_GetLogicalType(entity)==iAnd) {
					good = TRUE;
#else
				if(elem && parentsShow(elem)) {
					good = TRUE;
#endif
				}
			}
		} while(entity && good == FALSE);
		if (entity && good) {
			this->getSM()->setItem(entity);
			centerIfNecessary(this, entity);
		}
	} else {
		entity = this->root_entity;
		do {
			if(entity){
				TreeElement *elem = (TreeElement *) eaGetFace(entity, vtTREE);
#ifndef IGNORE_ORS
				if(elem && parentsShow(elem)&&i_GetLogicalType(entity)==iAnd) {
					good = TRUE;
#else
				if(elem && parentsShow(elem)) {
					good = TRUE;
#endif
				}
			}
			entity = i_GetNextEntity(entity);
		} while(entity && good == FALSE);
		if (entity && good) {
			this->getSM()->setItem(entity);
			centerIfNecessary(this, entity);
		}
	}	
	
}


void
TreePage::upEntity()
{

	iEntityPtr sole_entity;
	iEntityPtr left_entity=0;
	Ref r;
	int good = FALSE;

	const DLList *selection=this->getSM()->getItems();

	r = dllistLast(selection);
	if (r)
		sole_entity = (iEntityPtr) dllistContents(selection, r);
	else
		sole_entity=0;

	if(sole_entity) {
		left_entity = sole_entity;
		do {
#ifndef IGNORE_ORS
			left_entity = i_FindEntityBack(left_entity, 0, NOTOR, 0);
#else
			left_entity = i_FindEntityBack(left_entity, 0, ALL, 0);
#endif
			if(left_entity){
				TreeElement *elem = (TreeElement *) eaGetFace(left_entity, vtTREE);
#ifndef IGNORE_ORS
				if(elem && parentsShow(elem)&&i_GetLogicalType(left_entity)==iAnd) 
					good = TRUE;
#else
				if(elem && parentsShow(elem)) 
				  good = TRUE;
#endif

			}
		} while(left_entity && good == FALSE);
		if (left_entity && good) {
			this->getSM()->setItem(left_entity);
			centerIfNecessary(this, left_entity);
		}
	}
}

void
TreePage::downEntity()
{
	iEntityPtr sole_entity;
	iEntityPtr right_entity=0;
	Ref r;
	int good = FALSE;
	
	const DLList *selection=this->getSM()->getItems();

	r = dllistFirst(selection);
	if (r)
		sole_entity = (iEntityPtr) dllistContents(selection, r);
	else
		sole_entity=0;

	if(sole_entity) {
		right_entity = sole_entity;
		do {
#ifndef IGNORE_ORS
			right_entity = i_FindEntityForward(right_entity, 0, NOTOR, 0);
#else
			right_entity = i_FindEntityForward(right_entity, 0, ALL, 0);
#endif
			if(right_entity){
				TreeElement *elem = (TreeElement *) eaGetFace(right_entity, vtTREE);
#ifndef IGNORE_ORS
				if(elem && parentsShow(elem)&&i_GetLogicalType(right_entity)==iAnd) 
				  good = TRUE;
#else
				if(elem && parentsShow(elem)) 
				  good = TRUE;
#endif

			}
		} while(right_entity && good == FALSE);
		if (right_entity && good) {
			this->getSM()->setItem(right_entity);
			centerIfNecessary(this, right_entity);
		} 
	}
}



void
TreePage::centerOnSelected(iEntityPtr entity)
{
	TreeElement *elem,*ce;
	iEntityPtr cur;
	
	elem = (TreeElement*)eaGetFace(entity,vtTREE);
	for (cur = i_GetParentEntity(entity);cur;
							cur = i_GetParentEntity(cur)) {
#ifndef IGNORE_ORS
		if(i_GetLogicalType(cur)==iOr) cur = i_GetParentEntity(cur);
#endif
		ce = (TreeElement*)eaGetFace(cur,vtTREE);
		if (!ce->show) elem = ce;
	}
	this->getCanvas()->scrollToPosition(this->mode->main_window->scroll_window, elem->x, elem->y);
}

void TreePage::changeEntity(iEntityPtr en)
{
	TreeElement *elem = (TreeElement*)eaGetFace(en,vtTREE);
	
	this->syncSize();
	elem->redraw();
}


int 
TreePage::pressKey(iUCode key_value)
{
	if(this->getSM()->getState() == scNOT){
		ringTheBell();
		return 0;
	}		
	if (key_value == '\b') {
		this->mode->main_window->deleteSelection(FALSE);
		return 1;
	} else if(key_value == 20){
		killContentsForSelection();
	} else {
		ringTheBell();
		return 0;
	}
	return 0;
}
void TreePage::unselect()
{
	iEntityPtr selected = this->getSM()->getSoleItem();
	if (selected) refreshSel(selected);
}
	

iEntityPtr
TreePage::getClosest(iEntityPtr en, int x, int y,int *move_pos)
{
#ifndef IGNORE_ORS
	if(i_GetLogicalType(en)==iOr)
		en = getOrChild(en);
#endif
	TreeElement *elem = (TreeElement*)eaGetFace(en,vtTREE);
	iEntityPtr cur,first_entity;
	DLList *selection_list = this->getSM()->getItems();
	first_entity = (iEntityPtr)dllistContents((selection_list),dllistFirst(selection_list));
	int y0 = elem->y - VERTSPACER/2;
	int y1 = elem->y;
	int y2 = elem->y + elem->height;
	int y3 = elem->y + elem->height + VERTSPACER/2;
	if ((y >= y0) && (y <= y1)){
		*move_pos = mBefore;
		return en;
	} else if ((y > y1) && (y < y2)){
		*move_pos = mChild;
		return en;
	} else if ((y >= y2) && (y < y3)){
		*move_pos = mAfter;
			return en;
	}
	if (elem->show)
		for (cur = i_GetChildEntity(en);cur; cur = i_GetNextEntity(cur))
			if (en = getClosest(cur,x,y,move_pos))
				return en;
	return 0;	
}
iEntityPtr
TreePage::getEntityPos(int x, int y, int *move_pos)
{
	return getClosest(this->root_entity, x,y+ch/2,move_pos);
}

void
TreePage::setupXor()
{
	const DLList *items;
	iEntityPtr cur;
	Ref r;
	items = this->getSM()->getItems();
	iLBox bounds;
	iEntityPtr en,par=0;
	int move_pos;
	TreeElement *el,*elem;

	if (!dllistEmpty(&this->xor_boxes))
		freeBoxlist(&this->xor_boxes);
	dllistCtor(&this->xor_boxes);
	for (r = dllistFirst(items); r; dllistNext(items,&r)) {
		cur = (iEntityPtr)dllistContents(items,r);
		elem = (TreeElement*)eaGetFace(cur,vtTREE);
		iLBox *orbox = (iLBox*)smalloc(sizeof(iLBox));
		iLBox box;
		bounds.x = elem->x;
		bounds.y = elem->y;
		bounds.w = elem->width;
		bounds.h = ch;
	
		box.x = 
			this->last_x - this->first_x + bounds.x;
	 	box.y = this->last_y - this->first_y + bounds.y;
		box.w = bounds.w;
		box.h = bounds.h;

		box.x = MAX(0, box.x);
		box.y = MAX(0, box.y);

		box.x = MIN(this->width - box.w, box.x);
		box.y = MIN(this->height - box.h, box.y);
		if (r==dllistFirst(items))
			this->xor_box = box;
		else if (box.y < this->xor_box.y)
			this->xor_box = box;
		*orbox = box;
		dllistAppend(&this->xor_boxes, orbox);
	}
	r = dllistFirst(&this->xor_boxes);
	//this->xor_box = *(iLBox *)dllistContents(&this->xor_boxes, r);
	en = getEntityPos(xor_box.x,xor_box.y,&move_pos);
	if (en)
		switch(move_pos){
			case mChild:
				par = en;
				break;
			case mBefore:
				par = i_GetParentEntity(en);
				break;
			case mAfter:
				par = i_GetParentEntity(en);
			case mNone:
			break;
		}
#ifndef IGNORE_ORS
	if(i_GetLogicalType(par)==iOr) par = i_GetParentEntity(par);
#endif
	if (par) {
		el = (TreeElement*)eaGetFace(par,vtTREE);
		xor_x = xor_box.x;
		xor_y = xor_box.y + ch/2;
		par_xor_x = el->x;
		par_xor_y = xor_y;
	} else {
		par_xor_x = xor_x = 0;
		par_xor_y = xor_y = 0;
		
	}
}
static int4
compare_less_than(void *a, void *b)
{
	iEntityPtr entity_a = (iEntityPtr)a;
	iEntityPtr entity_b = (iEntityPtr)b;
	TreeElement *elem_a = (TreeElement*)eaGetFace(entity_a,vtTREE);
	TreeElement *elem_b = (TreeElement*)eaGetFace(entity_b,vtTREE);
	
	return (elem_b->y - elem_a->y);
}
static int4
compare_greater_than(void *a, void *b)
{
	iEntityPtr entity_a = (iEntityPtr)a;
	iEntityPtr entity_b = (iEntityPtr)b;
	TreeElement *elem_a = (TreeElement*)eaGetFace(entity_a,vtTREE);
	TreeElement *elem_b = (TreeElement*)eaGetFace(entity_b,vtTREE);
	
	return (elem_a->y - elem_b->y);
}

void
TreePage::moveEntity(iEntityPtr entity,int move_pos)
{
	DLList *items = (DLList*)this->getSM()->getItems();
	iEntityPtr the_entity;

	Ref r;

	switch(move_pos){
		case mChild:
			dllistSort(items, compare_less_than);
			break;
		case mAfter:
			dllistSort(items, compare_less_than);
			break;
		case mBefore:
			dllistSort(items, compare_greater_than);
			break;
		case mNone:
			break;
	}
	for (r = dllistFirst(items);r;dllistNext(items,&r)) {
		the_entity = (iEntityPtr)dllistContents(items,r);
		switch (move_pos){
		case mChild:
			if (i_GetChildEntity(entity) == the_entity)
				return;
			i_MoveToFront(the_entity,entity);
			break;
		case mAfter:
			if (i_GetNextEntity(entity) == the_entity)
				return;
			i_MoveAfter(the_entity,entity);
			break;
		case mBefore:
			if (i_GetPrevEntity(entity) == the_entity)
				return;
			i_MoveBefore(the_entity,entity);
			break;
		}
	}
	this->mode->main_window->document->setDirty();
	
	this->stale = 1;
}


void
TreePage::drawXor(Pixmap drawable)
{
	ViewCanvas *vc = this->getCanvas();
	Widget widget = this->getCanvas()->canvas;
	Display *display = XtDisplay(widget);
	GC gc;
	gc = this->getCanvas()->xorGC();
	iLBox box;
	iLBox vr = visibleRectangle(this->mode->scrolled_window);
	
	Ref r;
	for (r = dllistFirst(&this->xor_boxes);r;dllistNext(&xor_boxes,&r)){
		box = *(iLBox *)dllistContents(&this->xor_boxes,r);
		iLBox draw_box = box;
		i_IntersectLBox(vr, draw_box, &draw_box);
		XDrawRectangle(display, drawable, gc,
					   draw_box.x - vc->pix_hoffset, (draw_box.y - vc->pix_voffset),
					   draw_box.w, draw_box.h);
	}
	if (xor_x != par_xor_x)
		XDrawLine(display, drawable, gc,
					   par_xor_x - vc->pix_hoffset, par_xor_y -vc->pix_voffset,
					   xor_x - vc->pix_hoffset, xor_y - vc->pix_voffset);
}
void
TreePage::eraseXor()
{
	Pixmap pixmap = XtWindow(this->getCanvas()->canvas);
	drawXor(pixmap);
}

struct ReturnData {
	TreePage *tree_page;
	int num;
	iEntityPtr entity;
};
typedef ReturnData *ReturnDataPtr;

#ifndef IGNORE_ORS
void
DestroyOrMenu(Widget w, TreePage *page, char *callbackstruct)
{
	page->freeOrMenu();
	XtDestroyWidget(page->orMenu);
	page->orMenu = 0;
}
#endif
	
void
changePathCB(Widget w,
             ReturnDataPtr return_data,
             XtPointer call_data)
{
	TreePage *page = return_data->tree_page;
	iEntityPtr en;
	en = return_data->entity;
	TreeElement *elem = (TreeElement*)eaGetFace(i_GetParentEntity(en),vtTREE);

	elem->prev_or_path = elem->or_path;
	elem->or_path = return_data->num;
	page->stale = 1;
	page->clear();
	page->drawRegionNow(visibleRectangle(page->mode->main_window->scroll_window));
#ifndef IGNORE_ORS
//bbogus this should use grab button to grab the input event instead 
//of destroying the menu here
	DestroyOrMenu(page->orMenu,page,0);
#endif
}

#ifndef IGNORE_ORS
static void
alloclabel(iEntityPtr en,iUCode **label,int check)
{
	iUATablePtr table = (iUATablePtr)smalloc(sizeof(iUATable));
	iUAEntry *entries = (iUAEntry*)smalloc(sizeof(iUAEntry));
	iUCode *new_str, *p;
	int j;
	char *text;

	table->entries = entries;
	table->num = 1;
	table->entries->index = 10;
	table->entries->string = "\\n";
	char *s = i_AllocASCIIText(en,table); 
	text = s;
	if(strlen(s) > 30){
		for (j = 29; j > 26; j--)
			s[j] = '.';
		s[30] = 0;
	} 
	p = new_str = (iUCode*)smalloc((strlen(s)+2)*sizeof(iUCode));
	*p++ = (check) ? 0x2713 : ' ';
	while (*s)
		*p++ = *s++;
	*p = 0;
	*label = new_str;
	smfree(text);
	smfree(entries);
	smfree(table);
}
#endif

int countKids(iEntityPtr en)
{
	int count=0;
	
	iEntityPtr cur;
	for (cur = i_GetChildEntity(en); cur; cur = i_GetNextEntity(cur)){
		count++;
	}
	return count;
}

#ifndef IGNORE_ORS
static void 
createMenuOptions(TreePage *tp,iEntityPtr en,XmFontList *list)
{	
	XmStringMenuItem *items;
	int i;
	iUCode *s;
	iEntityPtr par,cur;

	par = i_GetParentEntity(en);
	TreeElement *par_elem = (TreeElement*)eaGetFace(par,vtTREE);
	int count = countKids(par);
	ReturnDataPtr return_data;
	items = (XmStringMenuItem*)smalloc(sizeof(XmStringMenuItem)*(count+1));
	count = 0;
	for (i = 0,cur=i_GetChildEntity(par);cur;cur=i_GetNextEntity(cur), i++){
		return_data = (ReturnDataPtr)smalloc(sizeof(ReturnData));
		return_data->tree_page = tp;
		return_data->num = i;
		return_data->entity = cur;
		items[i].result = 0;
		alloclabel(cur,&s,par_elem->or_path == count);
		items[i].label = makeCompoundString(XtDisplay(tp->getCanvas()->canvas),s,ucslen(s),list);
		smfree(s);
		items[i].klass = &xmPushButtonGadgetClass;     
		items[i].mnemonic = '\0';     
		items[i].accelerator = NULL;  
		items[i].accel_text = NULL;    
		items[i].callback = (XtCallbackProc)changePathCB;   
		items[i].callback_data = return_data; 
		items[i].subitems = NULL;
		count ++;
	}	
	items[count].result = 0;
	items[count].label = NULL;        
	items[count].klass = 0;     
	items[count].mnemonic = '\0';     
	items[count].accelerator = NULL;  
	items[count].accel_text = NULL;    
	items[count].callback = 0;   
	items[count].callback_data = 0; 
	items[count].subitems = NULL;
	tp->items = items;
}
#endif

#ifndef IGNORE_ORS
void 
TreePage::setupOrMenu(iEntityPtr en)
{
	Widget menu;
	XmFontList list=0;
	XmString xm_string;
	createMenuOptions(this,en,&list);
	xm_string = XmStringCreateLocalized("Select OR path");
	menu = BuildXmStringMenu(this->getCanvas()->canvas, XmMENU_POPUP,xm_string, 'S', list,this->items);
	XmStringFree(xm_string);
#if 0
	XtAddCallback(menu,
		XmNdisarmCallback,
		DestroyOrMenu, this);
#endif
	this->orMenu = menu;
	
}	
#endif

#ifndef IGNORE_ORS
void 
TreePage::freeOrMenu()
{	
	int i;
	if (!this->items) return;
	for(i = 0; this->items[i].label; i++){
		XmStringFree(this->items[i].label);
		this->items[i].label = 0;
		smfree(this->items[i].callback_data);
		this->items[i].callback_data = 0;
	}
	smfree(this->items);
	this->items = 0;
}
#endif

void
TreePage::pressButton(XEvent *event)
{
    iEntityPtr entity,sole_entity= this->getSM()->getSoleItem();
	TreeElement *elem;
	this->go = 0;
	ViewCanvas *vc = this->getCanvas();
	event->xbutton.x += vc->pix_hoffset;
	event->xbutton.y += vc->pix_voffset;	
    switch(event->xbutton.button) {
	case Button1:
#ifndef IGNORE_ORS
		if (this->orMenu)
			DestroyOrMenu(this->orMenu,this,0);
#endif
		this->last_x = this->first_x = -1;
		entity = this->findCloseEntity(event->xbutton.y);
		if(entity != 0) {
			if(event->xbutton.state & ShiftMask){
				this->toggleItem(entity);
				elem = (TreeElement*)eaGetFace(entity,vtTREE);
				if (elem->inBox(event->xbutton.x,event->xbutton.y)){
					this->last_x = this->first_x = event->xbutton.x;
					this->last_y = this->first_y = event->xbutton.y;
					this->setupXor();
				}
			} else {
				elem = (TreeElement*)eaGetFace(entity,vtTREE);
				if (elem->inTri(event->xbutton.x,event->xbutton.y)) {
					if(entity && i_GetChildEntity(entity)){
						elem->show = !elem->show;
						this->changeEntity(entity);
					}
				}
#ifndef IGNORE_ORS
				else if (elem->inRectangle(event->xbutton.x,event->xbutton.y)
						&&i_GetLogicalType(i_GetParentEntity(entity))==iOr) {
					setupOrMenu(entity);
					this->stale = 1;
					Widget widget=this->orMenu;
					XButtonEvent button_event=event->xbutton;
					XmMenuPosition(widget,&button_event);
					XtManageChild(this->orMenu);
				} 
#endif
				else if (elem->inBox(event->xbutton.x,event->xbutton.y)){
					if (sole_entity)
						this->unselect();
					DLList *items = (DLList*)this->getSM()->getItems();
					if (items && !dllistSeek(items, entity))
						this->setItem(entity);
					this->last_x = this->first_x = event->xbutton.x;
					this->last_y = this->first_y = event->xbutton.y;
					this->setupXor();
				}
			}
		} else {
			this->clear();
		}
		break;
	case Button3:
		break;
	case Button2:
		break;
    }
}
int
isParent(DLList *par_list, iEntityPtr en)
{
	iEntityPtr cur,par;
	Ref r;
	
	for(r = dllistFirst(par_list); r ; dllistNext(par_list, &r)){
		par = (iEntityPtr)dllistContents(par_list,r);
		if (par==en) return 0;
		for(cur = en;cur;cur=i_GetParentEntity(cur))
			if(par==cur)
				return 1;
	}
	return 0;
}
void
TreePage::releaseButton(XEvent *event)
{
    iEntityPtr entity,sole_entity;
    DLList *selection;
	TreeElement *elem;
	ViewCanvas *vc = this->getCanvas();
	event->xbutton.x += vc->pix_hoffset;
	event->xbutton.y += vc->pix_voffset;	
	int move_pos;
    switch(event->xbutton.button) {
	case Button2:
	case Button3:
		break;
	case Button1:
		if (!this->go) 
			return;    //we were not doing a move
		this->eraseXor();
		entity = this->getEntityPos(xor_box.x, xor_box.y,&move_pos);
		if (!entity) return;
		elem = (TreeElement*)eaGetFace(entity,vtTREE);
		selection = (DLList*)this->getSM()->getItems();
		sole_entity =(iEntityPtr)this->getSM()->getSoleItem();
		if(entity && !dllistSeek(selection,entity)) {
			if (xor_x != par_xor_x && !isParent(selection,entity))
				this->moveEntity(entity,move_pos);
			this->stale = 1;
		} else if (entity==sole_entity)
			this->stale = 1;
		break;
    }
	this->first_x =  this->last_x = this->first_y = this->last_y = -1;
	this->go = 0;
    this->xor_box.w = 0;
	if (stale) {
		sole_entity = (iEntityPtr)dllistContents(selection,dllistFirst(selection));
		elem = (TreeElement*)eaGetFace(sole_entity,vtTREE);
		this->drawRegionNow(visibleRectangle(this->mode->scrolled_window));
		if(elem && visiblePositionP(this->mode->scrolled_window, elem->x, elem->y) == FALSE ||
	  		visiblePositionP(this->mode->scrolled_window, elem->x, elem->y + this->ch+TOPPAD) == FALSE)
			this->centerOnSelected(sole_entity);
	}

}

void
TreePage::moveButton(int state, int x, int y)
{
	Pixmap pixmap = XtWindow(this->getCanvas()->canvas);
	if((this->first_x == -1) && (this->last_x == -1)) 
		return;
	if(state & Button1Mask) {
		if (this->first_x != -1)
			this->eraseXor();
		this->last_x = x;
		this->last_y = y;
		
		this->setupXor();
		this->drawXor(pixmap);
		this->go = 1;
		
	}
}

void
TreePage::manage()
{
  this->getSM()->convertToTreeItem();
  this->Page::manage();
}

void
TreePage::unmanage()
{
  this->Page::unmanage();
  this->getSM()->convertToItem();
}

/* these useless magnification functions are virtual requirements */
void 
TreePage::setMag(int x, int ox)
{
}

void 
TreePage::_setMag(int x, int ox)
{
}
