
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef _TreePage_h_
#define _TreePage_h_

#include "Page.h"
#include "misc/misc.h"

#define VERTSPACER 10
#define HORZSPACER 3
#define MAXBOXCHARS 20
#define MAXBOXWIDTH 30
#define LEFTPAD 40
#define TOPPAD 20

struct iUAEntry {
	iUCode index;
	char *string;
};
struct iUATable {
	int4 num;
	iUAEntry *entries;
};

class TreeElement;
enum MovePos {mChild, mBefore, mAfter,mNone};

class TreePage : public Page {
  public:
	int stale;
#ifndef IGNORE_ORS
	Widget orMenu;
#endif
	XmStringMenuItem *items;
	int ch,cw,asc;
private:
	void initTree(iEntityPtr en,int lev,int shown);
	void updateTree(iEntityPtr en,int lev);
	int drawEntities(Display *display,Pixmap pixmap,
		iLBox *bp,iEntityPtr en);
	iEntityPtr FindEntity(iEntityPtr,int,int);
	iEntityPtr findCloseEntity(int);
	iEntityPtr findClose(iEntityPtr,int);
	iEntityPtr getEntityPos(int,int,int*);
	iEntityPtr getClosest(iEntityPtr,int,int,int*);
#ifndef IGNORE_ORS
	iEntityPtr getOrPath(iEntityPtr,int);
#endif
	void deleteTree(iEntityPtr en);

public:
#ifndef IGNORE_ORS
	iEntityPtr getOrChild(iEntityPtr);
#endif
	iEntityPtr root_entity;
	int initialized_canvas,go;
	int last_x, first_x,last_y,first_y;
	iLBox xor_box;
	DLList xor_boxes;
	int xor_x, xor_y,par_xor_x,par_xor_y;
	int maxlev,vspacer,nodes,leaves,hspacer;

	TreePage(Mode *input_mode, const aString& label, iEntityPtr entity);
	virtual ~TreePage();
	void moveEntity(iEntityPtr,int);
	void setupXor();
#ifndef IGNORE_ORS
	void setupOrMenu(iEntityPtr);
	void freeOrMenu();
#endif
	void drawXor(Pixmap pixmap);
	void eraseXor();
	void refreshSel(iEntityPtr en);
	int containsSelection(iEntityPtr en);
	void changeEntity(iEntityPtr en);
	void unselect();
	/*virtual*/ void conditionallyInitialiseCanvas();
	/*virtual*/ void setMag(int new_mag, int old_mag);
	/*virtual*/ void _setMag(int new_mag, int old_mag);
	/*virtual*/ void redisplay(XmDrawingAreaCallbackStruct *call_data);
	/*virtual void resize(XmDrawingAreaCallbackStruct *call_data); */
	/*virtual*/ iEntityPtr findEnclosingEntity(int x, int y);
	
	/*virtual*/ void addNewEntity(iEntityPtr entity);
	
	/*virtual*/ void prevEntity();
	/*virtual*/ void nextEntity();

	/*virtual*/ void findNextFlagEntity();
	
	/*virtual*/ void upEntity();
	/*virtual*/ void downEntity();
	 
	/*virtual*/ void centerOnSelected(iEntityPtr entity);
	/*virtual*/ void scrollToUpperLeft(iEntityPtr entity){};
	/*virtual*/ void scrollToTop(iEntityPtr entity){};
	/*virtual*/ void scrollToBottom(iEntityPtr entity){};
	/*virtual*/ void scrollToLeft(iEntityPtr entity){};
	/*virtual*/ void scrollToRight(iEntityPtr entity){};
	
	// for delayed drawing
	/*virtual*/ int resizeNecessaryP();
	/*virtual*/ void drawRegionNow(iLBox box);

	/*virtual void setPicture(iLBox b); */
	/*virtual iLBox getPicture(); */
	
	/*virtual*/ int pressKey(iUCode key_value);
	
	/*virtual*/ void         pressButton(XEvent *event);
	/*virtual*/ void         releaseButton(XEvent *event);
	/*virtual*/ void         moveButton(int state, int x, int y);
	
	/*virtual*/ void manage();
	/*virtual*/ void unmanage();
};

#endif
