
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"

#include <X11/Xutil.h>

#include <Xm/MwmUtil.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/SelectioB.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/TextF.h>
#include <Xm/LabelG.h>
#include <Xm/Form.h>
#include <Xm/PushB.h>
#include "misc/Cmap.h"
#include <Xm/ToggleBG.h>
#include <Xm/DialogS.h>
#include "misc/color.h"
#include "misc/XArg.h"
#include "misc/misc.h"
#include "misc/prompt.h"
#include "svpmap.h"
#include "dllist.h"
#include "smemory.h"

#include "MainWindow.h"
#include "Application.h"
#include "ViewCanvas.h"
#include "EntityAux.h" // for conf class
#include "misc/bb.h" // for bbNullP
#include "TypePrefDialog.h"
#include "Document.h"
#include "rgb.h"
#include "config.h"

#define c_MARGIN 0
#define c_SHADOW_WIDTH 2

// Margins calculated by trial and error, to center buttons.
// CreateActionArea always left justifies.
#define BUTTON_AREA_MARGIN 140
#define BUTTON_BOX_MARGIN 148

#define TOGGLE_AREA_MARGIN 0

void
TypePrefDialog::create()
{
	XArg args;
	Widget parent = this->app_shell;

	args.reset();
	args.setArg(XmNtitle,  "Select DAFS Entity Type");
	args.setArg(XmNnoResize, True);
	args.setArg(XmNresizePolicy, XmRESIZE_NONE);
	//args.setArg(XmNdefaultPosition, FALSE);
	this->dialog = XmCreateFormDialog(parent,
									  "form",
									  args.getArgs(),
									  args.getSize());
	this->updateColors();
}

TypePrefDialog::TypePrefDialog(MainWindow * input_mw, Widget input_app_shell)
	: Dialog(input_mw, input_app_shell), color_box(0),colors(0),button_box(0),popper(0)
{

}

TypePrefDialog::~TypePrefDialog()
{
	if (this->popper)
		XtDestroyWidget(this->popper);
	if (this->colors)
		smfree(this->colors);
}

struct Choice {
	const char *name;
	Widget widget;
	char   original_position;
	char   current_position;
	char   default_position;
};

typedef struct Choice Choice;

static Choice pref_choices[] = {
	{"Place on its own page in Text mode", NULL, 1, 1, 0},
	{"Edit as separate region in Text mode", NULL, 1, 1, 0},
	{"Display in Image mode", NULL, 1, 1, 1},
};

static void
dialog_done_pushed(Widget w, void *user_data, XmAnyCallbackStruct *cbs)
{
	TypePrefDialog *tpd = (TypePrefDialog *) user_data;
	tpd->hide();
}

static void
popup_ok_pushed(Widget w, void *user_data, XmAnyCallbackStruct *cbs)
{
	TypePrefDialog *tpd = (TypePrefDialog *) user_data;
	XtUnmanageChild(tpd->popper);
}

static void
popup_cancel_pushed(Widget w, void *user_data, XmAnyCallbackStruct *cbs)
{
	// undo changes to pref_choices
	pref_choices[0].current_position = pref_choices[0].original_position;
	pref_choices[1].current_position = pref_choices[1].original_position;
	pref_choices[2].current_position = pref_choices[2].original_position;
	TypePrefDialog *tpd = (TypePrefDialog *) user_data;
	tpd->name[0] = 0;
	tpd->type[0] = 0;
	XtUnmanageChild(tpd->popper);
}
static void
popup_show_only(Widget w, void *user_data, XmAnyCallbackStruct *cbs)
{
	TypePrefDialog *tpd = (TypePrefDialog *) user_data;
	
	turn_off_others(tpd->type,tpd->mw->document->root_entity);
	XtUnmanageChild(tpd->popper);
}
static ActionAreaItem dialog_action_items[] = {
	{ "Done",     dialog_done_pushed,     NULL          }
};

static ActionAreaItem popup_action_items[] = {
#ifdef RAF_INTERNAL
	{ "Show Only This Type", popup_show_only, 	NULL			},	
#endif
	{ "OK",     popup_ok_pushed,     NULL          },
	{ "Cancel",     popup_cancel_pushed,     NULL          }
};

static void
initializePrefChoices(char *type_name)
{
	pref_choices[0].original_position = pref_choices[0].current_position = getTypePage(type_name);	
	pref_choices[1].original_position = pref_choices[1].current_position = getTypeField(type_name);	
	pref_choices[2].original_position = pref_choices[2].current_position = getTypeVisible(type_name);	

}

static void
prefToggleMe(Widget w, int ind, XmToggleButtonCallbackStruct *state)
{
	pref_choices[ind].current_position = state->set;
}

void
TypePrefDialog::updateColors()
{
	if(this->color_box){
		XtDestroyWidget(this->color_box);
		this->color_box = 0;
	}
	if(this->button_box) {
		XtDestroyWidget(this->button_box);
		this->button_box = 0;
	}
											
	this->color_box = 
		XtVaCreateWidget("colorbox",
						 xmRowColumnWidgetClass, this->dialog, 
						 XmNpacking, XmPACK_COLUMN, 
						 XmNorientation, XmHORIZONTAL,
						 XmNnumColumns,     10, 
						 XmNentryAlignment, XmALIGNMENT_CENTER, 
						 NULL);
	
	/* simply loop thru the strings creating a widget for each one */
	int i;
	const char *color, *name;
	void *v = this;
	XtCallbackProc cb = (XtCallbackProc) pickIt;

	for (i = 0; !i||i_TypeToName(i); i++) {
		Widget b;
		color=getTypeColor((char*)i_TypeToName(i));
		if (!i)
			b = XtVaCreateWidget("<NULL>", 
								 xmPushButtonWidgetClass, this->color_box,
								 XmNmarginWidth, c_MARGIN, 
								 XmNmarginHeight,c_MARGIN, 
								 XmNshadowThickness, c_SHADOW_WIDTH, 
								 XmNshadowType, XmSHADOW_IN, 
								 XmNbackground, getColorByName(this->mw->view_canvas->canvas, color),
								 NULL);
		else {
			name = i_TypeToName(i);
			if (!strcmp(name, "<NULL>"))	
				continue;
			b = XtVaCreateWidget(name, xmPushButtonWidgetClass,
								 this->color_box, 
								 XmNmarginWidth, c_MARGIN, 
								 XmNmarginHeight, c_MARGIN, 
								 XmNshadowThickness, c_SHADOW_WIDTH, 
								 XmNshadowType, XmSHADOW_IN, 
								 XmNbackground, getColorByName(this->mw->view_canvas->canvas, color),
								 NULL);
		}
		XtManageChild(b);
		XtAddCallback(b, 
			  XmNactivateCallback,
			  cb, 
			  v);
	}
	
	// give the button(s) a separate area
	this->button_box = 
		XtVaCreateWidget("buttonbox",
						 xmRowColumnWidgetClass, this->dialog, 
						 XmNpacking, XmPACK_COLUMN, 
						 XmNorientation, XmHORIZONTAL,
				         XmNmarginWidth, BUTTON_BOX_MARGIN, 
						 XmNbottomAttachment, XmATTACH_FORM,
						 XmNleftAttachment, XmATTACH_FORM,
						 XmNrightAttachment, XmATTACH_FORM,
						 XmNtopAttachment, XmATTACH_WIDGET,
						 XmNtopWidget, color_box, 
						 XmNnumColumns,     1, 
						 XmNentryAlignment, XmALIGNMENT_CENTER, 
						 NULL);
						 
	for(i=0; i<XtNumber(dialog_action_items); i++)
		dialog_action_items[i].data = this;
	
	CreateActionArea(button_box, dialog_action_items, XtNumber(dialog_action_items));
	XtManageChild(this->button_box);
	XtManageChild(this->color_box);
}

void
TypePrefDialog::update()
{	
	int shown;
	
	if (this->dialog){
		shown = XtIsManaged(this->dialog);
		if (shown){
			this->updateColors();
		}
	}
}

static void
setPopperLabel(Widget p, char *type)
{
	XArg args;

	XmString title=XmStringCreateSimple(type);
	args.reset();
	args.setArg(XmNdialogTitle,title);
	XtSetValues(p, args.getArgs(), args.getSize());
}

static void
setWidgetColor(TypePrefDialog *tpd, Widget w, char *color)
{
	XArg args;
	args.reset();
	args.setArg(XmNbackground, getColorByName(tpd->mw->view_canvas->canvas, color));
	XtSetValues(w, args.getArgs(), args.getSize());
}

static
static void CloseColor(XColor *color,char *name)
{	
	int r,g,b;
	r = (color->red >> 8) & 0xff;
	g = (color->green >> 8) & 0xff;
	b = (color->blue >> 8) & 0xff;
	sprintf(name,"#%02x%02x%02x",r,g,b);
}
static void 
selColor(Widget w,
	   TypePrefDialog *tpf,
	   CmapCallbackStruct *ccs)
{
	CloseColor(&ccs->color,tpf->name);
	
}
static void
setToggles()
{
	int i;
	XArg args;
	
	for(i=0; i<XtNumber(pref_choices); i++) {
		args.reset();
		args.setArg(XmNset, pref_choices[i].original_position);
		XtSetValues(pref_choices[i].widget, args.getArgs(), args.getSize());
	}
}
void TypePrefDialog::createPopup()
{
	XColor nc;
	int i;
	Display *display  = XtDisplay(this->color_box);
	ViewCanvas *view_canvas = this->mw->view_canvas;
	int screen  = DefaultScreen(display);
	Colormap map = DefaultColormap(display, screen);
	XArg args;
	Widget toggle_area, button_area;
	XmString title=XmStringCreateSimple("Select Color and Set Display Options");
	XmString foo = XmStringCreateSimple("   ");

	args.reset();
	args.setArg(XmNdialogTitle,title);
	args.setArg(XmNdialogStyle,XmDIALOG_FULL_APPLICATION_MODAL);
	this->popper = XmCreateFormDialog(this->mw->app->app_shell,"popup dlog",
					args.getArgs(),args.getSize());


	n_cells = DisplayCells(display,screen);					  
	
	this->colors = (Pixel*)smalloc(n_cells * sizeof(Pixel));
	 
	for (nc.pixel = 0;nc.pixel < n_cells;nc.pixel++) {
		XQueryColor(display,map,&nc);
		CloseColor(&nc,this->name);
		this->colors[nc.pixel] = nc.pixel;
	}
	args.reset();
	args.setArg(XmNtopAttachment,XmATTACH_FORM);
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	args.setArg(XmNtopAttachment,XmATTACH_FORM);
	args.setArg(XtNmappingTable, (XtPointer)this->colors);
	args.setArg(XmNuserData,(XtPointer)this);
	args.setArg(XtNlastIndex, n_cells - 1);
	args.setArg(XtNfirstIndex, 0);
	color_area = XtCreateManagedWidget("ColorWidget",
									  cmapWidgetClass,
									  this->popper,
									  args.getArgs(),
									  args.getSize());
	XtAddCallback(color_area, XtNchangeCallback, 
				  (XtCallbackProc) selColor, this);
									  
	args.reset();
	args.setArg(XmNpacking,XmPACK_COLUMN);
	args.setArg(XmNorientation,XmVERTICAL);
	args.setArg(XmNnumColumns,1);
	args.setArg(XmNmarginWidth, TOGGLE_AREA_MARGIN); 
	args.setArg(XmNentryAlignment,XmALIGNMENT_BEGINNING);
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	args.setArg(XmNtopAttachment,XmATTACH_WIDGET);
	args.setArg(XmNtopWidget,color_area);
	toggle_area = XmCreateRowColumn(this->popper,
									  "Property Toggles",
									  args.getArgs(),
									  args.getSize());		

	args.reset();
	args.setArg(XmNpacking,XmPACK_COLUMN);
	args.setArg(XmNorientation,XmHORIZONTAL);
	args.setArg(XmNnumColumns,1);
	args.setArg(XmNmarginWidth, BUTTON_AREA_MARGIN); 
	args.setArg(XmNentryAlignment,XmALIGNMENT_END);
	args.setArg(XmNbottomAttachment,XmATTACH_FORM);
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	args.setArg(XmNtopAttachment,XmATTACH_WIDGET);
	args.setArg(XmNtopWidget,toggle_area);
	button_area = XmCreateRowColumn(this->popper,
									  "Action Buttons",
									  args.getArgs(),
									  args.getSize());	
									  
	
	for (i = 0; i < XtNumber(pref_choices); i++) {
		Widget b;
		b = XtVaCreateManagedWidget(pref_choices[i].name,
									xmToggleButtonGadgetClass, 
									toggle_area, 
									XmNset, pref_choices[i].original_position,
									NULL);
		pref_choices[i].widget = b;
		XtAddCallback(b, XmNvalueChangedCallback, 
					  (XtCallbackProc) prefToggleMe, (void *) i);
	}	
	
	for(i=0; i<XtNumber(popup_action_items); i++)
		popup_action_items[i].data = this;

	CreateActionArea(button_area, popup_action_items, XtNumber(popup_action_items));
	XmStringFree(title);
	XmStringFree(foo);
	XtManageChild(color_area);
	XtManageChild(toggle_area);
	XtManageChild(button_area);
	XtManageChild(this->popper);
}

static void pickIt(Widget w,
	   TypePrefDialog *tpd,
	   XmAnyCallbackStruct *cbs)
{
	int buttons_changed = 0;
	
	initializePrefChoices(XtName(w));

	if (!tpd->popper) 
		tpd->createPopup();
	else {
		setToggles();
		XtManageChild(tpd->popper);
	}
	setPopperLabel(tpd->popper, XtName(w));
	strcpy(tpd->type, XtName(w));
		
	tpd->name[0] = 0; // used to pass back picked color
	XtPopup(XtParent(tpd->popper), XtGrabNone);
	XMapRaised(XtDisplay(tpd->popper), XtWindow((tpd->popper)));
    while (XtIsManaged(tpd->popper)) {
        XtAppProcessEvent(tpd->mw->app->app_context, XtIMAll);
        XSync(XtDisplay(tpd->popper), 0);
    }

	if(tpd->name[0]) { // if color was picked
		setTypeColor(XtName(w), tpd->name);
		setWidgetColor(tpd, w, tpd->name);
	}
		
	if(pref_choices[0].current_position != pref_choices[0].original_position){
		buttons_changed = 1;	
		if(pref_choices[0].current_position)
			setTypePage(XtName(w), TRUE);
		else
			setTypePage(XtName(w), FALSE);
	}	
	if(pref_choices[1].current_position != pref_choices[1].original_position){
		buttons_changed = 1;	
		if(pref_choices[1].current_position)
			setTypeField(XtName(w), TRUE);
		else
			setTypeField(XtName(w), FALSE);
	}
	
	if(pref_choices[2].current_position != pref_choices[2].original_position){
		buttons_changed = 1;	
		if(pref_choices[2].current_position)
			setTypeVisible(XtName(w), TRUE);
		else
			setTypeVisible(XtName(w), FALSE);	
	}

	tpd->mw->choose(buttons_changed);
	tpd->mw->redrawImageMode();

	XtUnmanageChild(tpd->popper);
}
