#include <Xm/CutPaste.h>
#include <Xm/DrawingA.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <X11/cursorfont.h>
#include <Xm/RowColumn.h>
#include <stdlib.h>
#include <stdio.h>
#include "ucs.h"
#include "utf.h"
#include "smemory.h"
#include "uString.h"
#include "misc/Unicode.h"
#include "keyboard/KeyMap.h"
#include "UniWidget.h"

static int  
max(int a, int b){
	return a > b ? a : b;
}

static int
min(int a, int b){
	return a < b ? a : b;
}

// forward declarations 
void uniInputCallback(Widget widget, UniWidget *uni, XmDrawingAreaCallbackStruct *cbs);
void uniExposeCallback(Widget widget, UniWidget *uni, XmDrawingAreaCallbackStruct *cbs);
void uniResizeCallback(Widget widget, UniWidget *uni, XmDrawingAreaCallbackStruct *cbs);
void uniMotionHandler(Widget widget, UniWidget *uni, XEvent *event);

UniWidget::UniWidget(Widget parent, KeyMap *km)
{
	// initialize data 
	insertion = base_length = 0;
	overstrike = 0;
	selection = selecting = FALSE;

	// start up drawing area widget
	da = XtVaCreateWidget("drawing_a", 
						  xmDrawingAreaWidgetClass, parent, 
						  NULL);
	XtAddEventHandler(da, Button1MotionMask, False, (XtEventHandler) uniMotionHandler, (XtPointer) this);
	XtAddCallback(da, XmNinputCallback, (XtCallbackProc) uniInputCallback, (XtPointer) this);
	XtAddCallback(da, XmNexposeCallback, (XtCallbackProc) uniExposeCallback, (XtPointer) this);
	XtAddCallback(da, XmNresizeCallback, (XtCallbackProc) uniResizeCallback, (XtPointer) this);

	XGCValues gcv;
	gcv.foreground = BlackPixelOfScreen(XtScreen(da));
	XtVaGetValues(da, XmNbackground, &background_color, NULL);
	gcv.background = background_color;

	gc = XCreateGC(XtDisplay(da), RootWindowOfScreen(XtScreen(da)), 
				   GCForeground | GCBackground, &gcv);
	XtVaSetValues(da, XmNuserData, gc, NULL);	
	
//	cursor = XCreateFontCursor(XtDisplay(da), XC_xterm);
//	XDefineCursor(XtDisplay(da), XtWindow(da), cursor);

	// get information - height and ascent from letter
	changeKeyMap(km);

	// if direction is reversed change gravity
//	if(direction == REVERSE){
//		XSetWindowAttributes attrs;
//		attrs.bit_gravity = NorthEastGravity;
//		XChangeWindowAttributes(XtDisplay(da), XtWindow(da), CWBitGravity, &attrs);
//	}

	// initialize lines array
	lines_length = 1;
	line = 0;
	lines = (uString **) smalloc(sizeof(uString *));
	lines[line] = new uString;
}
	
UniWidget::~UniWidget()
{
	for(int i = 0; i < lines_length; i++)
		delete lines[i];
	XFreeCursor(XtDisplay(da), cursor);
	XFreeGC(XtDisplay(da), gc);
}

static void
setRectangle(XRectangle *r, short x, short y, unsigned short w, unsigned short h)
{
	r->x = x;
	r->y = y;
	r->width  = w;
	r->height = h;
}

void
UniWidget::setString(const iUCode *string)
{
	for(int i = 0; i < lines_length; i++)
		delete lines[i];

	lines_length = 1;
	line = 0;
	lines = (uString **) smalloc(sizeof(uString *));
	lines[line] = new uString(string);

	insertion = base_length = raw2base(lines[line]->length());
}

iUCode *
UniWidget::getString(void)
{
	int size = 0;
	for(int i = 0; i < lines_length; i++)
		size += lines[i]->length() + 1;

	iUCode *string = (iUCode *) smalloc(sizeof(iUCode) * size),
		   *end;
	end = string;
	for(i = 0; i < lines_length; i++){
		ucsncpy(end, lines[i]->string(), lines[i]->length());
		end += lines[i]->length();
		*end = (iUCode) '\n';
		end++;
	}
	end[-1] = 0;

	return string;
}

void
UniWidget::changeKeyMap(KeyMap *km){
	keymap = km;

	iUCode letter;
	keymap->getInfo(&direction, &letter);
	measureUnicode(XtDisplay(da), &letter, 1, &width, &height, &ascent);
}

void
UniWidget::drawBorder(void)
{
	int draw_x, draw_y;
	unsigned int draw_width, draw_height, draw_border, draw_depth;
	Drawable draw_root;
	XGetGeometry(XtDisplay(da), XtWindow(da), &draw_root, &draw_x, &draw_y, 
				 &draw_width, &draw_height, &draw_border, &draw_depth); 

	XSetForeground(XtDisplay(da), gc, background_color);
	XFillRectangle(XtDisplay(da), XtWindow(da), gc, 0, 0, draw_width, Y_MARGIN*height/4);
	XFillRectangle(XtDisplay(da), XtWindow(da), gc, 0, 0, X_MARGIN*width/4, draw_height);
	XSetForeground(XtDisplay(da), gc, BlackPixelOfScreen(XtScreen(da)));
}

void
UniWidget::draw(void)
{
	draw(0, lines_length-1);
}

void
UniWidget::draw(int initial, int final)
{	
	// create pixmap to draw into
	Pixmap pixmap = XCreatePixmap(XtDisplay(da), RootWindowOfScreen(XtScreen(da)), 
								  draw_width, draw_height, DefaultDepthOfScreen(XtScreen(da)));

	// blank out pixmap
	XSetForeground(XtDisplay(da), gc, background_color);
	XFillRectangle(XtDisplay(da), pixmap, gc, 0, 0, draw_width, draw_height);
	XSetForeground(XtDisplay(da), gc, BlackPixelOfScreen(XtScreen(da)));

	// loop over all the lines
	if(initial < 0)
		initial = 0;
	if(final >= lines_length)
		final = lines_length-1;
	for(int i = initial; i <= final; i++){
		// convert unicode characters to glyphs
		uString glyphs = uniContext(*lines[i]);
		
		int x, 
			y = (4*i + Y_MARGIN) * height/4, 
			pos = 0, 
			num_base = 0;
		
		if(direction == NORMAL)
			x = X_MARGIN*width/4;
		else
			x = draw_width - (X_MARGIN+4)*width/4;

		while(pos <= glyphs.length()){ 
			// draw cursor, position determined by base (ie. spacing) characters
			if(i == line && !(selecting || selection) && insertion == num_base)
				if(overstrike)
					XDrawLine(XtDisplay(da), pixmap, gc, x, y+ascent, x+width, y+ascent);
				else
					XDrawLine(XtDisplay(da), pixmap, gc, x, y+ascent, x, y);
			
			// end processing at length
			if(pos < glyphs.length()){
				iUCode glyph = glyphs[pos];
				
				// update position
				int glyph_base, glyph_ascent, glyph_height, glyph_width = 0;
				if(glyph_base = keymap->spacing(glyph))
					measureUnicode(XtDisplay(da), &glyph, 1, &glyph_width, &glyph_height, &glyph_ascent);
				
				if(direction == REVERSE)
				x -= glyph_width;
				
				// invert colors and blank background
				if((selecting || selection) && 
				   ((i >= start_line && i <= stop_line) ||
					(i >= stop_line && i <= start_line)) &&
				   ((num_base >= start_pos && num_base <= stop_pos) ||
					(num_base >= stop_pos && num_base <= start_pos))){					
					XFillRectangle(XtDisplay(da), pixmap, gc, x, y, glyph_width, glyph_height);
					XSetForeground(XtDisplay(da), gc, background_color);
				}

				// draw glyph(s)
				do{
					drawUnicodeChar(XtDisplay(da), pixmap, gc, x, y+glyph_ascent, glyph);	
					glyph = glyphs[++pos];
				}while(glyph && !keymap->spacing(glyph));
				XSetForeground(XtDisplay(da), gc, BlackPixelOfScreen(XtScreen(da)));
				
				if(direction == NORMAL)
					x += glyph_width;
				
				num_base += glyph_base;
			}
			else
				pos++;
		}
	}
		
	int final_pixel;
	if(final == lines_length-1)
		final_pixel = draw_height - (4*initial+Y_MARGIN)*height/4;
	else
		final_pixel = (final-initial+1)*height;
	XCopyArea(XtDisplay(da), pixmap, XtWindow(da), gc, 0, (4*initial+Y_MARGIN)*height/4, 
			  draw_width, final_pixel, 0, (4*initial+Y_MARGIN)*height/4);
	XFreePixmap(XtDisplay(da), pixmap);
}

unsigned 
UniWidget::base2raw(unsigned base)
{
	unsigned raw = 0;
	
	base++;
	while(base > 0)
		base -= keymap->spacing((*lines[line])[raw++]);

	return raw - 1;
}

unsigned 
UniWidget::raw2base(unsigned raw)
{
	unsigned base = 0;

	raw++;
	while(raw)
		base += keymap->spacing((*lines[line])[raw--]);

	return base - 1;
}

unsigned 
UniWidget::nonspacing(unsigned raw)
{
	unsigned non_spacing = 0;

	while(!keymap->spacing((*lines[line])[raw+non_spacing]))
		non_spacing++;

	return non_spacing;
}

unsigned 
UniWidget::pos2line(int y)
{
	int line_pos;
	
	line_pos = (y - Y_MARGIN*height/4)/height;
	if(line_pos < 0)
		line_pos = 0;
	if(line_pos >= lines_length)
		line_pos = lines_length-1;

	return line_pos;
}

void
UniWidget::pos2base(int x, int y, int *line_pos, int *base_pos)
{
	*line_pos = pos2line(y);

	// determine base position within line
	int w, h, a, line_width;
	*base_pos = 0;

	uString glyphs = uniContext(*lines[*line_pos]);

	if(direction == NORMAL)
		line_width = X_MARGIN*width/4;
	else
		line_width = draw_width - (X_MARGIN+4)*width/4;

	for(int i = 0; i < glyphs.length(); i++){
		if((direction == NORMAL && x < line_width) ||
		   (direction == REVERSE && x > line_width))
			return;
		
		int b;
		if(b = keymap->spacing(glyphs[i])){
			measureUnicode(XtDisplay(da), &glyphs[i], 1, &w, &h, &a);
			if(direction == NORMAL)
				line_width += w;
			else
				line_width -= w;
			*base_pos += b;
		}
	}
}

void
UniWidget::ResizeCallback(XmDrawingAreaCallbackStruct *cbs)
{
	Dimension width, height;
	
	XtVaGetValues(da, XmNwidth, &width, XmNheight, &height, NULL);
	draw_width  = width;
	draw_height = height;
	draw_x = draw_y = 0;
}

void  
UniWidget::ExposeCallback(XmDrawingAreaCallbackStruct *cbs)
{
	XEvent *event = cbs->event;

	int initial = pos2line(event->xexpose.y),
		final   = pos2line(event->xexpose.y+event->xexpose.height);
	drawBorder();
	draw(initial, final);
}

void
UniWidget::killLine(int i)
{
	if(i < 0 || i >= lines_length)
		return;

	delete lines[i];

	for(int j = i; j < lines_length; j++)
		lines[j] = lines[j+1];

	lines_length--;
	lines = (uString **) smrealloc(lines, sizeof(uString *)*lines_length);
}

void
UniWidget::newLine(void){
	lines_length++;
	lines = (uString **) smrealloc(lines, sizeof(uString *)*lines_length);
	lines[lines_length-1] = new uString;
}

void 
UniWidget::InputCallback(XmDrawingAreaCallbackStruct *cbs)
{
	XEvent *event = cbs->event;

	switch(event->xany.type) {
	  case ButtonPress: 
		switch(event->xbutton.button){
		  case 1:
			xstart = event->xbutton.x;
			ystart = event->xbutton.y;
			selecting = TRUE;
			selection = FALSE;
			break;

		  case 2:
			int status;
			char buffer[1024];
			
			while((status = XmClipboardRetrieve(XtDisplay(da), XtWindow(da), "STRING",
												buffer, sizeof buffer, NULL, NULL)) == ClipboardLocked);

			if(status == ClipboardSuccess){
				iUCode u_buffer[1024];
				utf2ucs(u_buffer, buffer);

				int raw = base2raw(insertion),
					length = ucslen(u_buffer);
				lines[line]->move(raw, length);
				for(int i = 0; i < length; i++){
					base_length += keymap->spacing(u_buffer[i]);
					lines[line][raw+i] = u_buffer[i];					
				}

				draw();
			}
			break;
		}
		break;
		
	  case ButtonRelease:
		if(event->xbutton.button == 1){
			selecting = FALSE;
			if(!selection){
				pos2base(xstart, ystart, &line, &insertion);
				base_length = raw2base(lines[line]->length());
			}
			else {
				insertion = stop_pos;
				if(stop_pos < start_pos){
					int temp  = stop_pos;
					stop_pos  = start_pos;
					start_pos = temp;
				}
				if(stop_pos == base_length)
					stop_pos--;

				// copy the data to the clipboard
				int status;
				unsigned long item_id = 0;
				XmString label = XmStringCreateSimple("Unicode Widget");
				while((status = XmClipboardStartCopy(XtDisplay(da), XtWindow(da), label, CurrentTime, 
													 NULL, NULL, &item_id)) == ClipboardLocked);
				XmStringFree(label);
				
				unsigned start = base2raw(start_pos),
					     stop  = base2raw(stop_pos + 1);
				stop += nonspacing(stop);

				iUCode *u_buffer = (iUCode *) smalloc((stop-start+1)*sizeof(iUCode));
				ucsncpy(u_buffer, lines[line]->string()+start, stop-start);
				u_buffer[stop-start] = 0;
				char *buffer = (char *) smalloc((stop-start+1)*4);  // max of 4 bytes per unicode chatacter in utf			
				ucs2utf(buffer, u_buffer);
				
				while((status = XmClipboardCopy(XtDisplay(da), XtWindow(da), item_id, "STRING",
												buffer, strlen(buffer)+1, 0, NULL)) == ClipboardLocked);

				smfree(buffer);
				smfree(u_buffer);
				
				while((status = XmClipboardEndCopy(XtDisplay(da), XtWindow(da), item_id)) == ClipboardLocked);
			}

			draw();
		}
		break;
		
	  case KeyPress: {
		char key_value;
		KeySym ks;
		
		int ctrl_on = event->xkey.state & ControlMask,
			alt_on  = event->xkey.state & Mod1Mask;

		XLookupString(&event->xkey,
					  &key_value,
					  1,
					  &ks,
					  NULL);

		// blow off early if ks is a modifier or if selecting
		if(ks >= XK_Shift_L && ks <= XK_Hyper_R)
			break;

		unsigned start, stop;
		if(selection){
			start = base2raw(start_pos);
			stop  = base2raw(stop_pos + 1);
			stop += nonspacing(stop);
		}
		else
			start = stop = 0;

		switch(ks){
		  case XK_BackSpace:
			  if(selection){
				  insertion    = start_pos;
				  base_length -= stop_pos - start_pos + 1;
			  }
			  else if(insertion){
				  start = base2raw(insertion - 1);
				  stop  = base2raw(insertion);
				  stop += nonspacing(stop);
				  insertion--;
				  base_length--;
			  }
			  else if(line){
				  insertion = raw2base(lines[line-1]->length());
				  *lines[line-1] += *lines[line];
				  killLine(line);
				  line--;				  
				  base_length = raw2base(lines[line]->length());
				  break;
			  }

			  if(stop != start)
				  lines[line]->remove(start, stop - start);
			  break;	  
		
		  case XK_Delete:
			  if(selection){
				  insertion    = start_pos;
				  base_length -= stop_pos - start_pos + 1;				  
			  }
			  else if(insertion < base_length){
				  start = base2raw(insertion);
				  stop  = base2raw(insertion + 1);
				  stop += nonspacing(stop);
				  base_length--;
			  }
			  else if(line < lines_length-1){
				  *lines[line] += *lines[line+1];
				  killLine(line+1);
				  base_length = raw2base(lines[line]->length());
				  break;
			  }

			  if(stop != start)
				  lines[line]->remove(start, stop - start);
			  break;
		  
		  case XK_Left: 
			  if(!selection){				  
				  if(direction == NORMAL){
					  if(insertion)
						  insertion--;
					  else if(line){
						  line--;
						  insertion = base_length = raw2base(lines[line]->length());						  
					  }
				  }
				  else 
					  if(insertion < base_length)
						  insertion++;
					  else if(line+1 < lines_length){
						  insertion = 0;
						  line++;
						  base_length = raw2base(lines[line]->length());
					  }
			  }
			  break;

		  case XK_Right:
			  if(!selection){
				  if(direction == NORMAL){
					  if(insertion < base_length)
						  insertion++;
					  else if(line+1 < lines_length){
						  insertion = 0;
						  line++;
						  base_length = raw2base(lines[line]->length());
					  }
				  }
				  else 
					  if(insertion)
						  insertion--;
					  else if(line){
						  line--;
						  insertion = base_length = raw2base(lines[line]->length());						  
					  }
			  }
			  break;

		  case XK_Insert:
			  overstrike = !overstrike;
			  break;

		  case XK_End:
			  if(ctrl_on && line < lines_length){
				  line = lines_length-1;
				  base_length = raw2base(lines[line]->length());
			  }
			  insertion = base_length;
			  break;

  		  case XK_Home:
			  if(ctrl_on && line){
				  line = 0;
				  base_length = raw2base(lines[line]->length());
			  }
			  insertion = 0;
			  break;

		  case XK_Up:
			  if(line){
				  line--;
				  base_length = raw2base(lines[line]->length());
				  if(insertion > base_length)
					  insertion = base_length;
			  }
			  break;

		  case XK_Down:
			  if(line < lines_length-1){
				  line++;
				  base_length = raw2base(lines[line]->length());
				  if(insertion > base_length)
					  insertion = base_length;
			  }
			  break;

  		  case XK_Return:
			  line++;
			  if(line == lines_length)
				  newLine();
			  insertion = 0;
			  base_length = raw2base(lines[line]->length());
			  break;

		  default:
			  iUCode char_value;
              
              // convert keysym to unicode
			  if(alt_on)
				  char_value = keymap->lookupControlValue(ks);
			  else
				  char_value = keymap->lookupValue(ks);

			  // add new character
			  if(char_value){
				  if(insertion == base_length)
					  *lines[line] += char_value;
				  else{
					  int raw = base2raw(insertion);
					  lines[line]->move(raw);
					  (*lines[line])[raw] = char_value;
				  }
				  
				  int char_base = keymap->spacing(char_value);
				  base_length += char_base;
				  insertion += char_base;
			  }
			  else
				  XBell(XtDisplay(da), 0);
		}

		selection = FALSE;
		draw(line-1, line+1);

		break;

	  case KeyRelease:

		break;
	  }
	}
}

void
UniWidget::MotionHandler(XEvent *event)
{
	// must have pressed the button
	if(selecting){
		int x = event->xmotion.x,
			y = event->xmotion.y;

		// must move the cursor at least a third of a character
		if(abs(xstart-x) > width/3 || abs(ystart-y) > height/3){
			if(!selection){
				int line_pos, base_pos;
				pos2base(xstart, ystart, &line_pos, &base_pos);
				start_pos = stop_pos = base_pos;
				start_line = stop_line = line_pos;
				selection = TRUE;
			}
			
			// only redraw if the selected area has changed
			int move_pos, move_line;
			pos2base(x, y, &move_line, &move_pos);
			
			if(move_pos != stop_pos || move_line != stop_line){
				stop_pos  = move_pos;				
				stop_line = move_line;
				draw();
			}
		}
	}
}

void 
uniMotionHandler(Widget w, UniWidget *uni, XEvent *event)
{
	uni->MotionHandler(event);
}

void
uniExposeCallback(Widget widget, UniWidget *uni, XmDrawingAreaCallbackStruct *cbs)
{
	uni->ExposeCallback(cbs);
}

void
uniResizeCallback(Widget widget, UniWidget *uni, XmDrawingAreaCallbackStruct *cbs)
{
	uni->ResizeCallback(cbs);
}

void 
uniInputCallback(Widget widget, UniWidget *uni, XmDrawingAreaCallbackStruct *cbs)
{
	uni->InputCallback(cbs);
}

// for testing purposes
#if 0
extern KeyMap *initArabic(void);

int
main(int argc, char *argv[])
{
	Widget toplevel;
	XtAppContext app;

	KeyMap *arabic = initArabic();

	toplevel = XtVaAppInitialize(&app, "Demos", NULL, 0, &argc, argv, NULL, 
								 XmNwidth, 400,
								 XmNheight, 300,
								 NULL);

	UniWidget uni(toplevel, arabic);
	XtAppMainLoop(app);

	return 0;
}
#endif









