
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */


#ifndef _View_h_
#define _View_h_

#pragma interface

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <X11/Xutil.h>

#include "illum.h"

#include "aString.h"
#include "dllist.h"
#include "misc/stringVoidPtrMap.h"



class View {
public:
	// back pointers
	MainWindow *main_window;
	Widget scrolled_window;
	
	// my data
	DLList frame_list;
	Ref current_frame_ref;
	
	View(MainWindow *main_window);
	virtual ~View();
	Frame *getCurrentFrame();
	
	void setParentWidget(Widget parent);
	void setMag(int mag_exp, int old_mag);
	
	void show();
	void hide();
	void emptied(Frame *f);
	
	virtual void setCurrentFrame(Ref r);
	void start();

	virtual int forward();
	virtual int back();

	int countFrames();
	Ref backwardFrameRef();
	Ref forwardFrameRef();
	
	int managedP();
	void makeFrameList(XmStringTable *list, int *list_size);
	
	virtual void createFrames(iEntityPtr root_entity) = 0;
	virtual void imageDeleted(iImagePtr image) {};
	virtual aString getName() = 0;
	virtual ViewType getMode() = 0;
	virtual void configure() {}
	virtual void centerOnSelected(iEntityPtr entity) {};
	virtual iBox getBounds(iEntityPtr entity);
	
	virtual void addNewEntity(iEntityPtr new_entity) = 0;
	
	virtual void choose(int buttons_changed) {}
	
	SelectionManager *getSM();

	void purge();
	void addFrame(Frame *);
	Ref addFrameBefore(Ref p_r, Frame *frame);
};

#endif
