
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */


#ifndef _ViewCanvas_h_
#define _ViewCanvas_h_

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <X11/Xutil.h>

#include <Xm/DrawingA.h>

class Page;

class ViewCanvas {
public:
	int proc_id;    
	int follow;    
private:
	int release_expected;
	int ascent;
	int char_height;
	int char_width;
	GC blank_gc;
	GC blue_gc;
	GC inverse_blue_gc;
	GC thick_blue_gc;
	
	GC grot_gc;
	GC inverse_grot_gc;
	GC thick_grot_gc;
	
	GC black_gc;
	GC inverse_black_gc;
	GC red_gc;
	GC thick_red_gc;
	GC green_gc;
	GC dark_gray_gc;
	GC gray_gc;
	GC xor_gc;
	GC	flag_gc;
	GC inverse_flag_gc;
	GC copy_gc;

	GC gc_ptr;
	GC inverse_gc_ptr;
	GC thick_gc_ptr;
	GC medium_gc_ptr;
	
	int saveshape;
public:
	int pix_voffset;
	int pix_hoffset;
	iBox view_port;
	GContext font;	
	XFontSet font_set;
	XFontStruct* font_struct;
	Page *client;
	Widget canvas;
	
	void setClient(Page *client);
	ViewCanvas(Widget parent, Page *client, int w, int h);
	~ViewCanvas();
	
	GContext getFont();
	void setSize(int w, int h);
	void initializeScrollBarValues();
	
	int managedP();
	void manage();
	void unmanage();
	
	void setCursor(int cursor);
	void unsetCursor();

	void setFont(XFontStruct *font);
	void releaseFont();
	
	GC colorGC(const char *color);
	GC inverseGC(const char *color);
	GC thickGC(const char *color);
	GC mediumGC(const char *color);
	
	GC blankGC();
	GC blueGC();
	GC inverseBlueGC();
	GC thickBlueGC();
	GC grotGC();
	GC inverseGrotGC();
	GC thickGrotGC();
	GC blackGC();
	GC inverseBlackGC();
	GC redGC();
	GC thickRedGC();
	GC greenGC();
	GC grayGC();
	GC darkgrayGC();
	GC xorGC();
	GC FlagGC();
	GC inverseFlagGC();
	
	GC copyGC();
	
	void activateGrab();
	void deactivateGrab();
	
	void scrollToPosition(Widget sw, int4 x, int4 y);
	
	friend void redisplayCB (Widget         w,
							 ViewCanvas *vc,
							 XmDrawingAreaCallbackStruct *call_data);
	friend void resizeCB (Widget         w,
							 ViewCanvas *vc,
							 XmDrawingAreaCallbackStruct *call_data);
	friend void pressButtonEH(Widget          w,
							  ViewCanvas      *vc,
							  XEvent         *event);
	friend void releaseButtonEH(Widget          w,
								ViewCanvas      *vc,
								XEvent         *event);
	friend void moveButtonEH(Widget          w,
							 ViewCanvas      *vc,
							 XEvent         *event);
	friend void pressKeyEH(Widget          w,
						   ViewCanvas      *vc,
						   XEvent         *event);
	
	void measureText(const iUCode *s, int l, int *tw, int *th, int *ascent);
};


#endif
