
/*
 * Copyright (c) 1994, 1995  RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "rafutil.h"

#include "illum.h"
#include "c-lib.h"

#include "jrkanji.h"
#include "RK.h"
#include "tcs_raf.h"
#include "canna.h" // obey our own prototypes

#include <stdio.h>

static char **warning;

extern "C" int jrKanjiControl(...);
extern "C" int jrKanjiString(...);

static void 
cannaModeKeys()
{
#define CANNAWORKBUFSIZE 32
	char xxx[CANNAWORKBUFSIZE];
	int i, n;
	/* I believe we will find out what characters do which functions */
	n = jrKanjiControl(0, KC_MODEKEYS, xxx);
	/* I belive mule did this backwards so to use cons */
	//printf("cannaModeKeys:");
	for(i = n; i > 0;){
		--i;
		//printf("%c", xxx[i]);
	}
	//printf("\n");
}

static int
initCanna(char *server_name, char *canna_lib_dir, iUCode *u_warning)
{
	int res;
	/* means space between items (words?) */
	int kugiri = 1; /* ʸڤ򤹤뤫 */
	int return_value = 1;

	rafInitCanna(canna_lib_dir);

	jrKanjiControl(0, KC_SETSERVERNAME, (char *) server_name);
	jrKanjiControl(0, KC_SETINITFILENAME, (char *) 0);

	/* why a char ** */
	warning = (char **)0;
	res = jrKanjiControl(0, KC_INITIALIZE, (char *)&warning);

	if(warning){
		jis2unicode(warning[0], u_warning);
#if 0
		for(warning; *warning; warning++)
			printf("%s\n", *warning);
#endif
		return_value = 0;
	}
	if(res == -1){
#if 0
		/* the error is stuck in jrKanjiError */
		fprintf(stderr, "lets try and print out the error.\n");
#endif
		return_value = 0;
	}

#if 1
	jrKanjiControl(0, KC_SETBUNSETSUKUGIRI, (char *)kugiri);
    jrKanjiControl(0, KC_SETWIDTH, (char *)78);
	/* for whatever reason -- commented out in mule */

	/* means inhibit half-width kana */
#if 1
    jrKanjiControl(0, KC_INHIBITHANKAKUKANA, (char *)1);
#endif
	
	/* yomi == read */
    jrKanjiControl(0, KC_YOMIINFO, (char *)2); /* : ޻ޤ֤ */
#endif
	/* slavishly imitate mule here */
	cannaModeKeys();

	return return_value;
}

#define KEYTOSTRSIZE 2048
static unsigned char buf[KEYTOSTRSIZE];

#if 0
static void
printStatus(jrKanjiStatus *ks);
#endif

/* "Change Japanese pre-edit mode." */
static void
cannaChangeMode(int num)
{
	jrKanjiStatusWithValue ksv;
	jrKanjiStatus ks;

	ksv.buffer = (unsigned char *)buf;
	ksv.bytes_buffer = KEYTOSTRSIZE;
	ksv.ks = &ks;
	ksv.val = num;
	jrKanjiControl(0, KC_CHANGEMODE,  (char *)&ksv);
#if 0
	printStatus(&ks);
#endif
	/* bogus -- store results knows a lot about processing
	   output, for now we will just print stuff
	   */
#if 0	
	val = storeResults(buf, ksv.val, ksv.ks);
	return val;
#endif
}

/*
// this is interesting
// perhaps this is the string that should sit on the mode line
// and says 1/2 width full width and so on
*/
static void cannaQueryMode(unsigned char buffer[])
{
  jrKanjiControl(0, KC_QUERYMODE, buffer);
}

static void 
cannaDoFunction(int num, int ch)
{
  jrKanjiStatusWithValue ksv;
  jrKanjiStatus ks;

  if(ch == 0)
    *buf = '@';
  else {
	  *buf = ch;
  }

  ksv.buffer = (unsigned char *)buf;
  ksv.bytes_buffer = KEYTOSTRSIZE;
  ksv.ks = &ks;
  ksv.val = num;
  jrKanjiControl(0, KC_DO,  (char *)&ksv);
  //printStatus(&ks);
#if 0
  Lisp_Object val;
  val = storeResults(buf, ksv.val, ksv.ks);
  return val;
#endif
}

int try_initializing = FALSE;
int succeed_initializing = FALSE;

int 
nihonInit(char *canna_server, char *canna_lib_dir, iUCode *warning)
{
	int value;

	value = initCanna(canna_server, canna_lib_dir, warning);
	if(value){
		cannaChangeMode(0);
		cannaDoFunction(4, 0);
	}
	
	return value;
}

void
nihonModeName(iUCode *uni_buffer)
{
	unsigned char euc_buffer[256];
	cannaQueryMode(euc_buffer);
	jis2unicode((char *) euc_buffer, uni_buffer);
}

void 
getJapanese(char *src,iUCode ***list_ptr, int *n_ptr)
{
	iUCode **list;
	int n;
	jrKanjiStatus ks;
	int len;
	int l = strlen(src);
	int i;
	unsigned char euc_output_buffer[64];

	list = 0;
	n = 0;

	if(l != 0) {
		memset((char *) &ks, 0, sizeof(ks));

		for(i = l - 1; i < l; i++){
			len = jrKanjiString(0, src[i], buf, KEYTOSTRSIZE, &ks);
		}
		if(ks.length) {
			n = 1;
			list = (iUCode **) smalloc(sizeof(iUCode *) * n);
			list[0] = (iUCode *) smalloc(sizeof(iUCode) * 64);
			strncpy(euc_output_buffer, ks.echoStr, ks.length);
			euc_output_buffer[ks.length] = 0;
			jis2unicode(euc_output_buffer, list[0]);
		}
	}

	*list_ptr = list;
	*n_ptr = n;
}

#if 0
static void printStatus(jrKanjiStatus *ks)
{
	int i;
	printf("ks->echoStr:");
	for(i =0; i < ks->length; i++)
		printf("%c", ks->echoStr[i]);
	printf("\n");
	printf("ks->length: %d\n", ks->length);
	printf("ks->revPos: %d\n", ks->revPos);
	printf("ks->revLen: %d\n", ks->revLen);
	printf("ks->info: %d\n", ks->info);
	printf("ks->mode: %d\n", ks->mode);

	printf("ks->gline.line:");
	for(i =0; i < ks->gline.length; i++)
		printf("%c", ks->gline.line[i]);
	printf("\n");
	printf("ks->gline.length: %d\n", ks->gline.length);
	printf("ks->gline.revPos: %d\n", ks->gline.revPos);
	printf("ks->gline.revLen: %d\n", ks->gline.revLen);
}
#endif
