
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*
// contains the main for illum
// arg handling and x windows initialization
*/

#include "rafport.h"

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <X11/Xutil.h>

#include "illum.h"

#include "smemory.h"
#include "ifc.h"
#include "rafutil.h"
#include "misc/misc.h"
#include "misc/XArg.h"
#include "misc/prompt.h"
#include "aString.h"

#include "Application.h"

#include "Document.h"
#include "MainWindow.h"
#include "ViewCanvas.h"
#include "SelectionManager.h"
#include "config.h"

#ifdef RAF_INTERNAL
#include "form.h"
#endif

static char version[] = "0.10b1";// dont forget to change about box
static char figname[256];
static ModeType global_mode;
static char *figfile,*canna_host;
static char *canna_lib_dir;
static int memcheck;

#ifdef RAF_INTERNAL
char *ellipse_file;
#endif

ifc myargs[] = {
	{(char **) &figfile, 0, 0, 0, IFC_STR, "c","ILLUMFIG"},
#ifdef RAF_INTERNAL
	{(char **) &ellipse_file, 0,(int)"data.e", 0, IFC_STR, "D",0},
#endif
	{(char **) &canna_host, 0, 0, 0, IFC_STR, "j",0},
	{(char **) &canna_lib_dir, 0, 0, 0, IFC_STR, "l",0},
	{(char **) &memcheck,0,0,999,IFC_INT,"m",0},
	{(char **) &global_mode, 0, vtIMAGE, vtTREE, IFC_FLAG, "H",0},
	{(char **) &global_mode, 0, vtIMAGE, vtOOC, IFC_FLAG, "O",0},
	{(char **) &global_mode, 0, vtIMAGE, vtFLAG, IFC_FLAG, "M",0},
	{(char **) &global_mode, 0, vtIMAGE, vtRTE, IFC_FLAG, "T",0},
	{0, 0, 0, 1, IFC_NA, 0, 0}
};

static void
usage()
{
	fprintf(stderr, "Illuminator (version %s):\n", version);
#ifdef RAF_INTERNAL
	fprintf(stderr, "\t[-D <ellipse_file>]\n");
#endif
	fprintf(stderr, "\t[-c <config_file>]\n");
	fprintf(stderr, "\t[-h] # print this page and exit\n");
	fprintf(stderr, "\t[-j <canna_host>]\n");
	fprintf(stderr, "\t[-l <canna_lib_path>]\n");

	fprintf(stderr, "\t[-H] # open in hierarchy mode\n");
	fprintf(stderr, "\t[-M] # open in flag mode\n");
	fprintf(stderr, "\t[-O] # open in out of context mode\n");
	fprintf(stderr, "\t[-T] # open in text mode\n");
	fprintf(stderr, "\t[file_name]\n");
	fprintf(stderr, "\n");
}

extern "C" void prall(iEntityPtr);

void foo()
{
#if 0
	prall(0);
#endif
}


MainWindow *the_window = 0;
Application* the_application;
void changeType(Widget,XEvent*,String*,Cardinal*);

void changeType(Widget w,XEvent *event, String *params, Cardinal *num_params)
{
	unsigned char type;
	Ref r;
	iEntityPtr entity;
	
	if (!params[0]) return;
	i_NameToType(params[0],&type);;
	const DLList *selection=the_window->sm->getItems();
	for(r = dllistFirst(selection); r; dllistNext(selection, &r)){
		entity = (iEntityPtr) dllistContents(selection, r);
		i_SetEntityType(entity,type);
	}
	the_window->document->setDirty();
}	

int4
memoryFailure(void *user_data)
{
	//char *file_name;

	fprintf(stderr, "Illuminator ran out of memory.\n");

	if(user_data)		
		smfree(user_data);

	if(the_window){
		aString as = the_window->document->file_name + "~";
		i_WriteEntity(as.string(), the_window->document->root_entity, DAFSB);
		fprintf(stderr, "Saved emergency file %s. It may still be corrupted.\n", as.string());
		// find the document and the file_name
		// save the sucker somewhere
		exit(1);
	}
	return 0;
}

#if 0
#define ENABLE_EDITRES
#endif

#ifdef ENABLE_EDITRES
extern "C" void _XEditResCheckMessages();
#endif

int 
main(int  argc,
     char **argv)
{
	XtAppContext app_context = 0; 

	ArgList      args;    /* argument list */
	
	Widget    app_shell;  /*  ApplicationShell  */
	char *app_name;
	char *file_name;
	int success = 1;
	int i;

	char *app_title;

	app_title = "Illuminator";
	
	/*  Initialize toolkit */
	app_shell = XtAppInitialize(&app_context, app_title, NULL, 0, &argc,
								argv, NULL, args, 0);
	

	if (processargs(argc, argv,myargs)) {
		usage();
		exit(-1);
	}

	// must fix our ability to use s_InitLib from the beginning
	s_InitLib(memcheck,memoryFailure,malloc(20000));

	miscInit(app_shell);
	app_name = argv[0];
	
	i_InitLib();
	uiInit(app_context, app_shell);

	the_application = new Application(app_context, app_shell, 
									  app_title, 
									  canna_host,
									  canna_lib_dir);

	
#ifdef RAF_INTERNAL
	initializeRecognition();
#endif
	for (i=1;i<argc && !argv[i];i++) ;
	
	if(i < argc)
		file_name = argv[i];
	else 
		file_name = 0;
	
	if (!file_name) global_mode=vtRTE;

	// find out about fonts
	// app_shell
	extern int unicodeCheckForFonts(Widget app_shell);

	// returns 0 on success
	if(unicodeCheckForFonts(app_shell)) {
		fprintf(stderr, "Illum could not find the necessary fonts to show Unicode.\n");
		fprintf(stderr, "Consult your installation instructions.\n");
		// print out error message
	}


	// shouldn't hardwire opening image
	the_window = new MainWindow(the_application, global_mode);
	i_SetCallBack(entityCB, the_window);
	the_window->createApplication();

static XtActionsRec actions[] = {
			{"changeType", changeType},
			};

	XtAppAddActions(app_context,actions,1);

	if (!figfile) {
		figfile = figname;
		sprintf(figname,"%s/%s",getenv("HOME"), ILLUM_PREFS_FILE);
	}

	readConfig(figfile);
	initializeText();

	// if you do this after you start the document
	// necessary things don't happen in the text mode
	XtRealizeWidget (app_shell);

	if(file_name) {
		FileName foo;
		foo = file_name;
		success = the_window->document->openFile(foo, global_mode);
		if(success == 0) {
			the_window->document->newFile();
		}
		// let the error message occur in openFile where it
		// can be more specific
	} else {
		the_window->document->newFile();
	}

#ifdef ENABLE_EDITRES
    XtAddEventHandler(app_shell, (EventMask)0, True,
					  _XEditResCheckMessages, NULL);
#endif

	/*  Get and dispatch events. */
	XtAppMainLoop (app_context);
	
	// this stuff never does get executed. it should really be done in the
	// exit command
	i_ExitLib();
#ifdef RAF_INTERNAL
	f_ExitLib();
#endif
	miscExit();
	s_ExitLib();
	
	return 0;
}

