
/* 
// Copyright 1994 RAF Technology
// Internal Use Only
*/
#include "rafport.h"

#include "illum.h"
#include "rafgraph.h"
#include "rlsums.h"

#include "ViewCanvas.h"
#include "smemory.h"
#include "Mode.h"
#include "MainWindow.h"

#ifdef RAF_INTERNAL
#include "form.h"
#include "formint.h"
#endif

inline static int Magnify_int(const int i, const int mag)
{
	return (i*mag)>>16;
}

static iBox Magnify_iBox(const iBox & box, const int mag)
{
	iBox new_box;

	new_box.x=Magnify_int(box.x,mag);
	new_box.y=Magnify_int(box.y,mag);
	new_box.w=Magnify_int(box.w,mag);
	new_box.h=Magnify_int(box.h,mag);
	return new_box;
}

#ifdef STEVE
#include "/raf/linus/home/steve/src/illum/sps_pitch.cc"
#endif

#ifdef STEVE
#define GRAPH_BOTTOM 155
#else
#define GRAPH_BOTTOM 0
#endif

#define GRAPH_RIGHT 30

#define MIN_VERT (16<<16)
#define MAX_VERT (50<<16)

#if STEVE
short *GetCuts(iBox b,short *sumbuf, pandp_t *pptr, int *ncuts)
{

	int maxcuts = (b.w / (pptr->pitch>>16)) + 3;
	int bsize = maxcuts * sizeof(short);
	short *cuts=(short *)smalloc(bsize);
	short *jcuts=(short *)smalloc(bsize);

	*ncuts = makecuts(b.w, cuts, maxcuts, pptr);
	jigglecuts(jcuts, cuts, *ncuts, sumbuf, b.w, ((pptr->pitch+(1<<15))/4) >>16,(pptr->pitch+(1<<15))>>16);

	smfree(cuts);
	return jcuts;
}

void print_pitch(char *str, pandp_t *pp, long blood)
{
	printf("%s pitch=%f, phase=%f, conf=%i, blood=%ld\n", str,
		pp->pitch/65536.0, pp->phase/65536.0, pp->conf, blood);
}
#endif

void MakeHorizSumsGraph(ImagePage *image,iEntityPtr entity)
{
	short *graph_array;
	int map_width;

	if (!entity) return;
#if RAF_INTERNAL

	Widget w = image->getCanvas()->canvas;
	Display *display;
	Pixmap pixmap;


	int depth = DefaultDepthOfScreen(XtScreen(w));
	display = XtDisplay(w);
	Drawable drawable = RootWindowOfScreen(XtScreen(w));
	int mag = 1 << (16 + image->getMagExp());

	iBox b = i_GetEnclosingBox(entity);
	iBox mag_bounds=Magnify_iBox(b,mag);
	int plot_height=Magnify_int(b.h,mag);

	map_width=Magnify_int(100,mag);

	iImagePtr image_ptr;

	image_ptr = i_GetImage(entity);
	graph_array=(short *)smalloc( sizeof(short)*b.h);
	if (!graph_array) return;
	i_hspx(image_ptr,b,graph_array);
		
	pixmap = XCreatePixmap(display, drawable,map_width+GRAPH_RIGHT, plot_height, depth);
	XFillRectangle(display,pixmap,image->getCanvas()->blankGC(),0,0,map_width+GRAPH_RIGHT,plot_height);


	DrawVertGraph(display,pixmap,image->getCanvas()->blackGC(),graph_array,b.h,
	          0,0,map_width,plot_height);
#ifdef STEVE
	int i;
	int conf,res;
	pandp_t pp;
	iDAFSError e;
	
	e = f_FindPP(b.h, graph_array,MIN_VERT,MAX_VERT,1,&pp);
	if (!e)
		{
			print_pitch("ALG1 ",&pp,calc_blood(graph_array,b.h,pp));
			for (i=pp.phase; i<b.w<<16; i+=pp.pitch)
				{
					XDrawLine(display,pixmap,image->getCanvas()->blackGC(),
						map_width+13,Magnify_int(i>>16,mag),map_width+20,Magnify_int(i>>16,mag));

				}
		}
#endif
	smfree(graph_array);
	
	XCopyArea(display, pixmap, XtWindow(w), image->getCanvas()->copyGC(),
	    0, 0, map_width+GRAPH_RIGHT, plot_height,LEFTPAD+mag_bounds.x+mag_bounds.w+10,TOPPAD+mag_bounds.y);
	XFreePixmap(display, pixmap);
#endif
}

void MakeRAFGraph(ImagePage *image,iEntityPtr entity)
{

	short *graph_array;
	int i,map_width;
	ViewCanvas *vc=image->getCanvas();
#ifdef STEVE
	pandp_t pp;
#endif

	if (!entity) return;

	iBox b=i_GetEnclosingBox(entity);
#ifdef STEVE
	graph_array=MakeGraphArray(entity);
	if (!graph_array) return;
#else
	graph_array=(short *)smalloc(b.w*sizeof(short));
	for (i=0; i<b.w; i++)
		graph_array[i]=i;
#endif
	Widget w = vc->canvas;
	Display *display;
	Pixmap pixmap;

	int depth = DefaultDepthOfScreen(XtScreen(w));
	display = XtDisplay(w);
	Drawable drawable = RootWindowOfScreen(XtScreen(w));
	int mag = 1 << (16 + image->getMagExp());

	printf("box = (%i, %i) (%i, %i)\n",b.x,b.y, b.w, b.h);

	iBox mag_bounds=Magnify_iBox(b,mag);
	int plot_height=Magnify_int(100,mag);

	map_width=mag_bounds.w;

	pixmap = XCreatePixmap(display, drawable,map_width, plot_height+GRAPH_BOTTOM, depth);
	XFillRectangle(display,pixmap,image->getCanvas()->blankGC(),0,0,map_width,plot_height+GRAPH_BOTTOM);

	DrawGraph(display,pixmap,image->getCanvas()->blackGC(),graph_array,b.w,
	          0,0,mag_bounds.w,plot_height);
#ifdef STEVE
	iDAFSError err;
	pandp_t *pp_ptr;
	
	err=i_GetUserDataProp(entity, "pandp",(void **)&pp_ptr,0,f_SwapPandP);
	if (pp_ptr && !err)
		{
			pp=*pp_ptr;
			if (pp.phase<pp.pitch>>16) pp.phase=pp.phase<<16;	// change this to my conventions
			print_pitch("GT ",&pp,calc_blood(graph_array,b.w,pp));
			for (i=pp.phase; i<b.w<<16; i+=pp.pitch)
				{
					XDrawLine(display,pixmap,image->getCanvas()->blackGC(),
						Magnify_int(i>>16,mag),plot_height+3,Magnify_int(i>>16,mag),plot_height+10);

				}
		}
	
	long min_pitch,max_pitch;

	void *prm;
	prm=f_AllocLineSegParams(entity);
	get_min_max_pitch(prm,b.h,&min_pitch,&max_pitch);
	f_FreeLineSegParams(prm);
	
	err=f_FindPP(b.w,graph_array,min_pitch,max_pitch,1,&pp);

	print_pitch("ALG1 ",&pp,calc_blood(graph_array,b.w,pp));
	if (!err && pp.conf!=0)
		for (i=pp.phase; i<b.w<<16; i+=pp.pitch) {
			XDrawLine(display,pixmap,image->getCanvas()->blueGC(),
				   Magnify_int(i>>16,mag),plot_height+13,Magnify_int(i>>16,mag),plot_height+20);
		}

	prm=f_AllocLineSegParams(entity);
	get_min_max_pitch(prm,b.h,&min_pitch,&max_pitch);
	f_FreeLineSegParams(prm);

	err=f_FindPP(b.w,graph_array,min_pitch,max_pitch,2,&pp);

	print_pitch("ALG2 ",&pp,calc_blood(graph_array,b.w,pp));
	if (!err && pp.conf!=0)
		for (i=pp.phase; i<b.w<<16; i+=pp.pitch) {
			XDrawLine(display,pixmap,image->getCanvas()->blueGC(),
				   Magnify_int(i>>16,mag),plot_height+23,Magnify_int(i>>16,mag),plot_height+30);
		}


	int ncuts;
	short *cuts;
	cuts=GetCuts(b,graph_array,&pp,&ncuts);
	for (i=0; i<=ncuts; i++) {
		XDrawLine(display,pixmap,image->getCanvas()->blueGC(),
			   Magnify_int(cuts[i],mag),plot_height+33,Magnify_int(cuts[i],mag),plot_height+40);
	}
	smfree(cuts);
#endif

	smfree(graph_array);
	
	XCopyArea(display, pixmap, XtWindow(w), image->getCanvas()->copyGC(),
	    0, 0, map_width, plot_height+GRAPH_BOTTOM,LEFTPAD+mag_bounds.x-vc->pix_hoffset,TOPPAD+mag_bounds.y+mag_bounds.h+10-vc->pix_voffset);
	XFreePixmap(display, pixmap);
}




void DrawGraph(Display *display, Drawable drawable, GC gc,
               short *graph_array, int num_points,
               int left, int top, int width, int height)
{
	int graph_min,graph_max,graph_scale,graph_y,graph_orig,i;
	int xmag;

	xmag=(width<<16)/num_points;
	int line_width=Magnify_int(1,xmag);
	if (!line_width) line_width=1;


	XDrawRectangle(display,drawable,gc,left,top,width-1,height-1);

	graph_min=graph_array[0]; graph_max=graph_array[0];
	for (i=1; i<num_points; i++)
		{
			if (graph_array[i]<graph_min) graph_min=graph_array[i];
			if (graph_array[i]>graph_max) graph_max=graph_array[i];
		}  

	if (graph_min>0) graph_min=0;
	if (graph_max<0) graph_max=0;
	graph_scale=(height<<16)/(graph_max-graph_min+1);

	graph_orig=(((height-1)<<16)+graph_min*graph_scale)>>16;

	// draw origin
	XDrawLine(display,drawable,gc,0,graph_orig,width-1,graph_orig);
	
	for (i=0; i<num_points; i++)
		{
			graph_y=(((height-1)<<16)-(graph_array[i]-graph_min)*graph_scale)>>16;
			if (graph_orig>graph_y)
				XFillRectangle(display,drawable,gc,
			        	   Magnify_int(i,xmag)+left,graph_y+top,line_width,graph_orig-graph_y);
			else
				XFillRectangle(display,drawable,gc,
				    	   Magnify_int(i,xmag)+left,graph_orig+top,line_width,graph_y-graph_orig);
    	}
}

void DrawVertGraph(Display *display, Drawable drawable, GC gc,
               short *graph_array, int num_points,
               int left, int top, int width, int height)
{
	int graph_min,graph_max,graph_scale,graph_x,graph_orig,i;
	int ymag;

	ymag=(height<<16)/num_points;
	int line_width=Magnify_int(1,ymag);
	if (!line_width) line_width=1;


	XDrawRectangle(display,drawable,gc,left,top,width-1,height-1);

	graph_min=graph_array[0]; graph_max=graph_array[0];
	for (i=1; i<num_points; i++)
		{
			if (graph_array[i]<graph_min) graph_min=graph_array[i];
			if (graph_array[i]>graph_max) graph_max=graph_array[i];
		}  

	if (graph_min>0) graph_min=0;
	if (graph_max<0) graph_max=0;
	graph_scale=(width<<16)/(graph_max-graph_min+1);

	graph_orig=(((width-1)<<16)+graph_min*graph_scale)>>16;

	// draw origin
	XDrawLine(display,drawable,gc,graph_orig,0,graph_orig,height-1);
	
	for (i=0; i<num_points; i++)
		{
			graph_x=(((width-1)<<16)-(graph_array[i]-graph_min)*graph_scale)>>16;
			if (graph_orig>graph_x)
				XFillRectangle(display,drawable,gc,
			        	   graph_x+left,Magnify_int(i,ymag)+top,graph_orig-graph_x,line_width);
			else
				XFillRectangle(display,drawable,gc,
				    	   graph_orig+left,Magnify_int(i,ymag)+top,graph_x-graph_orig,line_width);
    	}
}

