
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#ifndef _aString_h_
#define _aString_h_


#include <stdlib.h>
#include <stddef.h>
#ifdef USE_STREAM
#include <iostream.h>
#include <strstream.h>
#endif
#include <assert.h>
#include <stdio.h>
#include <float.h>
#include <string.h>
#include <math.h>

extern const char *EmptyString;

void *operator new(size_t size);
void  operator delete(void *data);

double random01(void); // generate a random number between 0 and 1


/* ---------------------------------
 * dup_string 
 *  returns a copy of the input string
 *  allocated
 *  will work even if in is NULL
 ---------------------------------- */
char *dup_string(const char *in);


unsigned long NextPower2(unsigned long x);

/*
// Name aString
*/


struct uString;

struct aString {
	aString(const uString&);
	aString(const char *s = NULL);
	aString(const aString &s);
	~aString(void);
#if 0
	inline aString *Self(void)  { return(this); }
	inline const aString *Self(void) const  { return(this); }
#endif
	const char *string(void) const ;
	int   length(void) const ;
	aString &operator =(const aString &c); 
	aString &operator =(const char *c); 
	aString &operator +=(const aString &c);
	aString &operator +=(const char *c); 
	int   operator !(void) const; 
	int   operator !=(const aString &c) const; 
	int   operator !=(const  char *c) const; 
	int   operator ==(const aString &c) const; 
	int   operator ==(const char *c) const; 
	// <= only compares to shortest length
	int   operator <=(const aString &c) const; 
	int   operator <=(const char *c) const; 
	
	int   operator <(const aString &c) const; 
	int   operator <(const char *c) const; 
	int   operator >(const aString &c) const; 
	int   operator >(const char *c) const; 
	int   operator >=(const aString &c) const; 
	int   operator >=(const char *c) const; 
	
	aString operator +(const aString &c) const; 
	aString operator +(const char *c) const; 
	
	/* puts unicode into the supplied string */
	void putUnicode(uString&) const;
	
	/* tells me if this is straight 7 bit ascii or not */
	int utfP() const;
	
	/* puts in myself the utf ttranform of the uString */
	void getUtf(const uString&);
	
	/* tells me what the length of a converted unicode string would be */
	int uLength() const;
	
#ifdef USE_STREAM
	friend ostream& operator << (ostream& o,const aString &Y);
#endif
	
	friend  int   operator ==(const  char *c,const aString &s); 
	friend  int   operator !=(const  char *c,const aString &s); 
	friend  int   operator <=(const  char *c,const aString &s); 
	friend  int   operator <(const  char *c,const aString &s); 
	friend  int   operator >=(const  char *c,const aString &s); 
	friend  int   operator >(const  char *c,const aString &s); 
	friend  aString  operator +(const  char *c,const aString &s);
	// 	friend  const char *operator =(const  char *c,const aString &s); 
	
public:
	void clear(void);
	void extend(unsigned newsize);
	
	unsigned allocated;
	unsigned len;
	char *data;
};


#endif _H_aString

