/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*
// Main include file for DAFS
*/

#ifndef _H_dafs_
#define _H_dafs_

#include "rafport.h"
#include "rafio.h"    /* for RafFile -- used to be stdio.h for FILE * */
#include "dafstype.h"
#include "ucs.h"

#ifndef _H_rlimage_
typedef struct iImage *iImagePtr;
#endif

typedef int4 (APIENTRY *iCompareProc)(iEntityPtr *one, iEntityPtr *two);
typedef void (APIENTRY *iUserDataProc)(void *p);

struct iUAEntry;
typedef struct iUAEntry iUAEntry;
typedef struct iUAEntry *iUAEntryPtr;

struct iUATable;
typedef struct iUATable iUATable;
typedef struct iUATable *iUATablePtr;


#ifdef __cplusplus
extern "C" {
#endif
	
/*-----------------------------------------------------------------------*/
/*    Please make no changes before consulting with Tom                  */
/*-----------------------------------------------------------------------*/

/* library entry and exit */
Dll iDAFSError APIENTRY i_InitLib(void);
Dll void APIENTRY i_ExitLib(void);

/* creating and destroying entities */
Dll iEntityPtr APIENTRY i_NewEntity(const iTypeIndex type,
					iEntityPtr parent);
Dll iEntityPtr APIENTRY i_NewEntityAfter(const iTypeIndex type,
					 iEntityPtr sibling);
Dll iEntityPtr APIENTRY i_NewEntityBefore(const iTypeIndex type,
					  iEntityPtr sibling);
Dll iDAFSError APIENTRY i_CopyEntity(iEntityPtr orig, 
									 iEntityPtr parent_of_copy, 
									 iEntityPtr *new_entity_ptr);
Dll iDAFSError APIENTRY i_CopyOneEntity(iEntityPtr orig, 
									 iEntityPtr parent_of_copy, 
									 iEntityPtr *new_entity_ptr);

Dll       void APIENTRY i_DisposeEntity(iEntityPtr entity);

/* accessing entity content */
Dll iDTag APIENTRY i_GetEDataType(iEntityPtr entity);
Dll  void APIENTRY i_SetNoData(iEntityPtr entity);

Dll short APIENTRY i_GetGlyph(const iEntityPtr entity, short* conf);
Dll  void APIENTRY i_SetGlyph(iEntityPtr entity,short chr, short conf);
Dll  int2 APIENTRY i_GetShort(const iEntityPtr entity, short* conf);
Dll  void APIENTRY i_SetShort(iEntityPtr entity,int2 num, short conf);
Dll   const char * APIENTRY i_GetText(const iEntityPtr entity,short *conf);
Dll     iDAFSError APIENTRY i_SetText(iEntityPtr entity,const char *text,short conf);
Dll const iUCode * APIENTRY i_GetUText(const iEntityPtr entity,short *conf);
Dll     iDAFSError APIENTRY i_SetUText(iEntityPtr entity,const iUCode *text,short conf);
Dll          short APIENTRY i_GetConf(const iEntityPtr entity);
Dll           void APIENTRY i_SetConf(iEntityPtr entity,short conf);
Dll iPosSetPtr APIENTRY i_GetPosSet(iEntityPtr entity, int2 *n);
Dll iDAFSError APIENTRY i_SetPosSet(iEntityPtr entity,short n,const iPosSetPtr chrs);

/* comments */
Dll const char * APIENTRY i_GetComment(const iEntityPtr entity);
Dll iDAFSError APIENTRY i_SetComment(iEntityPtr entity,const char *comment);


/* entity types e.g "Line", "Document" */
Dll   iDAFSError APIENTRY i_NameToType(const char *type, iTypeIndex *val);
Dll const char * APIENTRY i_TypeToName(iTypeIndex type);
Dll   iTypeIndex APIENTRY i_GetEntityType(const iEntityPtr entity);
Dll         void APIENTRY i_SetEntityType(iEntityPtr entity,iTypeIndex type);

/* Conjunction Type */
Dll         void APIENTRY i_SetLogicalType(iEntityPtr entity, iLogicalType type);
Dll iLogicalType APIENTRY i_GetLogicalType(iEntityPtr entity);


/* traversing the hierarchy */
Dll iEntityPtr APIENTRY i_GetNextEntity(const iEntityPtr entity);
Dll iEntityPtr APIENTRY i_GetPrevEntity(const iEntityPtr entity);
Dll iEntityPtr APIENTRY i_GetChildEntity(const iEntityPtr entity);
Dll iEntityPtr APIENTRY i_GetLastEntity(const iEntityPtr entity);
Dll iEntityPtr APIENTRY i_GetParentEntity(const iEntityPtr entity);
Dll iEntityPtr APIENTRY i_GetParentPath(iEntityPtr home, 
					iEntityPtr target);
Dll iEntityPtr APIENTRY i_GetParentOtherPath(iEntityPtr home, 
					 iEntityPtr target);


/* searching routines */
Dll iEntityPtr	APIENTRY i_FindEntityForward(iEntityPtr home,
					 iTypeIndex type, 
					 iSearchType method, 
					 void *name);
Dll iEntityPtr	APIENTRY i_FindEntityBack(iEntityPtr home,
					  iTypeIndex type, 
					  iSearchType method, void *name);


/* rearranging the hierarchy */
Dll void APIENTRY i_MoveToFront(iEntityPtr entity,iEntityPtr new_parent);
Dll void APIENTRY i_MoveToBack(iEntityPtr entity, iEntityPtr new_parent);
Dll void APIENTRY i_MoveAfter(iEntityPtr entity, iEntityPtr PrevSibling);
Dll void APIENTRY i_MoveBefore(iEntityPtr entity,iEntityPtr NextSibling);


/* callbacks */
Dll void APIENTRY i_SetCallBack(iCallBackProc callBack,void *userData);
Dll void APIENTRY i_GetCallBack(iCallBackProc *callBack,void **userData);
Dll iDAFSError APIENTRY i_DoCallBack(iMess mess,iEntityPtr entity);


/* error handling */
Dll const char * APIENTRY i_GetError(iDAFSError error);


/* convenience routine that sorts children , plus sample compare funcs */
Dll iDAFSError APIENTRY i_Sort(iEntityPtr, iCompareProc);
Dll int4 APIENTRY i_VertCompare(iEntityPtr *one, iEntityPtr *two);
Dll int4 APIENTRY i_HorzCompare(iEntityPtr *one, iEntityPtr *two);

/* dafs structure to text conversion */
Dll iDAFSError APIENTRY i_ReadUATable(iUATablePtr *table, char *filename);
Dll       void APIENTRY i_DisposeUATable(iUATablePtr table);
Dll     char * APIENTRY i_AllocASCIIText(iEntityPtr en, iUATablePtr table);
Dll       void APIENTRY i_FreeASCIIText(char *text);
Dll   iUCode * APIENTRY i_AllocUText(iEntityPtr en);
Dll       void APIENTRY i_FreeUText(iUCode *text);

/* gathering information on the hierarchy */
Dll int4 APIENTRY i_CountChildren(iEntityPtr en);

Dll int4 APIENTRY i_CountPrev(iEntityPtr cur);
Dll int4 APIENTRY i_CountNext(iEntityPtr cur);



/* property functions */

/* set an entity's prop */
Dll iDAFSError APIENTRY i_SetBoolProp(iEntityPtr entity,const char *name,int data);
Dll iDAFSError APIENTRY i_SetLongProp(iEntityPtr entity,const char *name,int4 data);
Dll iDAFSError APIENTRY i_SetStringProp(iEntityPtr entity,const char *name,char *data);
Dll iDAFSError APIENTRY i_SetUStringProp(iEntityPtr entity,const char *name,iUCode *data);
Dll iDAFSError APIENTRY i_SetUserDataProp(iEntityPtr entity,const char *name,void *data,uint2 size);
Dll iDAFSError APIENTRY i_SetUserPtrProp(iEntityPtr entity,const char *name,void *data,iUserPtrFree callback);

/* associate property with a type */
Dll iDAFSError APIENTRY i_SetSessionBoolProp(iTypeIndex itype,const char *name,int data);
Dll iDAFSError APIENTRY i_SetSessionLongProp(iTypeIndex itype,const char *name,int4 data);
Dll iDAFSError APIENTRY i_SetSessionStringProp(iTypeIndex itype,const char *name,char *data);
Dll iDAFSError APIENTRY i_SetSessionUStringProp(iTypeIndex itype,const char *name,iUCode *data);
Dll iDAFSError APIENTRY i_SetSessionUserDataProp(iTypeIndex itype,const char *name,void *data,uint2 size);
Dll iDAFSError APIENTRY i_SetSessionUserPtrProp(iTypeIndex itype,const char *name,void *data,iUserPtrFree callback);



/* getting the property */
Dll iDAFSError APIENTRY i_GetLongProp(const iEntityPtr entity,
				  const char *name,long *i);
Dll iDAFSError APIENTRY i_GetBoolProp(const iEntityPtr entity,
				  const char *name,int4 *ip);
Dll iDAFSError APIENTRY i_GetUserPtrProp(const iEntityPtr entity,
					 const char *name, void **d,iUserPtrFree *callp);
Dll iDAFSError APIENTRY i_GetUserDataProp(const iEntityPtr entity,
					  const char *name, 
					  void **d, long *size, 
					  iUserDataProc ConvertProc);
Dll iDAFSError APIENTRY i_GetStringProp(const iEntityPtr entity,
					const char *name, char **ip);
Dll iDAFSError APIENTRY i_GetUStringProp(const iEntityPtr entity,
					 const char *name,iUCode **ip);

/* property attributes */
Dll       long APIENTRY i_GetPropSize(const iPropPtr prop);
Dll      iPTag APIENTRY i_GetPropType(const iPropPtr prop);
Dll const char * APIENTRY i_GetPropName(const iPropPtr prop);


/* destroying the properties */
Dll void APIENTRY i_DisposeProp(const iEntityPtr entity,const char *name);
Dll void APIENTRY i_DisposeSessionProp(iTypeIndex itype,const char *name);
Dll void APIENTRY i_DisposeAllProps(const iEntityPtr entity);
Dll void APIENTRY i_DisposeAllSessionProps(void);

/* traversing and moving properties */
Dll iDAFSError APIENTRY i_MoveProp(iEntityPtr entity,
				   iEntityPtr toen,const char *name);
Dll iDAFSError APIENTRY i_ChangePropName(iEntityPtr entity, const char *from_name,
					const char *to_name);
Dll   iPropPtr APIENTRY i_FindProp(const iEntityPtr entity, 
				   const char *name); 
Dll   iPropPtr APIENTRY i_FindPropNoTypes(const iEntityPtr entity, 
					  const char *name);
Dll   iPropPtr APIENTRY i_FirstProp(const iEntityPtr entity);
Dll   iPropPtr APIENTRY i_NextProp(const iEntityPtr entity, 
				   const iPropPtr prev);
Dll iDAFSError APIENTRY i_MoveAllProps(iEntityPtr from_en, iEntityPtr to_en);

Dll iDAFSError APIENTRY i_CopyAllProps(iEntityPtr from_en, iEntityPtr to_en);


/* regular expression routines */
Dll iDAFSError APIENTRY i_MakeARegex(iRegexPtr *rep, const char *str);
Dll iDAFSError APIENTRY i_MakeURegex(iRegexPtr *rep, const iUCode *str);
Dll int4 APIENTRY i_CheckARegex(const iRegexPtr re, const char *str, int4 *start, int4 *end);
Dll int4 APIENTRY i_CheckURegex(const iRegexPtr re, const iUCode *str, int4 *start, int4 *end);
Dll iDAFSError APIENTRY i_WriteRegex(const iRegexPtr re, char *filename);
Dll iDAFSError APIENTRY i_ReadRegex(iRegexPtr *rep, char *filename);
Dll iDAFSError APIENTRY i_ReadRegexFile(char *type,iRegexPtr *rep, RafFile file, int *swap);
Dll iDAFSError APIENTRY i_WriteRegexFile(const iRegexPtr re, RafFile file);
Dll void APIENTRY i_DisposeRegex(iRegexPtr re);

/* bounds */
/* a box is x,y,w,h 
   an Orthogon is a polygon in which successive segments are orthogonal */
/* BoxList is remarkable in that, internally, boxes never overlap */
Dll        iBox APIENTRY i_GetEnclosingBox(const iEntityPtr entity);
Dll iDAFSError APIENTRY i_GetArea(const iEntityPtr entity, 
								  void **data, iATag *tag);
Dll iDAFSError APIENTRY i_ForceAreaType(const iEntityPtr entity, iATag tag);

Dll iDAFSError  APIENTRY i_SetBox(iEntityPtr entity, iBox bounds);
Dll iDAFSError APIENTRY i_SetBoxList(iEntityPtr entity, iBoxListPtr BoxList);
Dll iDAFSError APIENTRY i_SetOrthogonList(iEntityPtr entity, 
					  iOrthogonListPtr ortho_list);

/* convertors */
Dll iDAFSError APIENTRY i_BoxList2OrthogonList(iBoxListPtr reg, 
											   iOrthogonListPtr *ortho_list);
Dll iDAFSError APIENTRY i_Box2OrthogonList(iBox box, iOrthogonListPtr *orthos);
Dll iDAFSError APIENTRY i_Box2BoxList(iBox box, iBoxListPtr *boxes);
Dll iDAFSError APIENTRY i_OrthogonList2BoxList(iOrthogonListPtr ortho_list, iBoxListPtr *boxes);
Dll iDAFSError APIENTRY i_Box2Orthogon(iBox box, iOrthogonPtr *ortho);
Dll iDAFSError APIENTRY i_Orthogon2BoxList(iOrthogonPtr or, iBoxListPtr *boxes);

/* lossy convertors */
Dll void APIENTRY i_BoxList2Box(iBoxListPtr reg, iBox *box);
Dll void APIENTRY i_OrthogonList2Box(iOrthogonListPtr ortho_list, iBox *box);
Dll void APIENTRY i_Orthogon2Box(iOrthogonPtr ortho, iBox *box);


/* box functions */
/* no new, dispose, copy necessary because this is a trivial structure */
Dll void APIENTRY i_IntersectBox(iBox a,iBox b,iBox *c);

Dll void APIENTRY i_UnionBox(iBox b1,iBox b2, iBox *c);

/* orthogonlist functions */
Dll iOrthogonListPtr APIENTRY i_NewOrthogonList(void);
Dll iOrthogonListPtr APIENTRY i_CopyOrthogonList(iOrthogonListPtr original);
Dll       void APIENTRY i_DisposeOrthogonList(iOrthogonListPtr o_list);

Dll iDAFSError APIENTRY i_UnionOrthogonList(const iOrthogonListPtr o_list1, 
					    const iOrthogonListPtr o_list2,
					    iOrthogonListPtr result);
Dll iDAFSError APIENTRY i_IntersectOrthogonList(const iOrthogonListPtr one, 
						const iOrthogonListPtr two,
						iOrthogonListPtr ret);
Dll iDAFSError APIENTRY i_SubtractOrthogonList(iOrthogonListPtr shapes, 
					    iOrthogonListPtr bites);

/* orthogon functions */
Dll   iOrthogonPtr APIENTRY i_NewOrthogon(int2 num);
Dll       void APIENTRY i_DisposeOrthogon(iOrthogonPtr ortho);
Dll   iOrthogonPtr APIENTRY i_CopyOrthogon(iOrthogonPtr shape);
Dll iDAFSError APIENTRY i_UnionOrthogon(const iOrthogonPtr o1, 
				const iOrthogonPtr o2,
				iOrthogonListPtr o_list);
Dll iDAFSError APIENTRY i_IntersectOrthogon(const iOrthogonPtr ortho1, 
					const iOrthogonPtr ortho2,
					iOrthogonListPtr o_list);
Dll void APIENTRY i_TranslateOrthogon(iOrthogonPtr ortho, short x, short y);


/* box list functions */
Dll iBoxListPtr APIENTRY i_NewBoxList(void);
Dll void APIENTRY i_DisposeBoxList(iBoxListPtr);
Dll iBoxListPtr APIENTRY i_CopyBoxList(iBoxListPtr one);
Dll iDAFSError APIENTRY i_UnionBoxList(iBoxListPtr, iBoxListPtr, iBoxListPtr);
Dll iDAFSError APIENTRY i_IntersectBoxList(iBoxListPtr, 
										   iBoxListPtr, iBoxListPtr);
Dll iDAFSError APIENTRY i_SubtractBoxList(iBoxListPtr, 
										  iBoxListPtr, iBoxListPtr);
Dll iDAFSError APIENTRY i_TranslateBoxList(iBoxListPtr reg, short x, short y);


/* image routines */
Dll      void APIENTRY i_SetImage(iEntityPtr entity,iImagePtr image);
Dll iImagePtr APIENTRY i_GetImage(const iEntityPtr entity);
Dll iImagePtr APIENTRY i_CloneEntityImage(const iEntityPtr entity);

/* helps track an independantly stored image */
Dll void APIENTRY i_SetImageCallBack(iImageWriteProc wp,iImageReadProc rp);
Dll void APIENTRY i_GetImageCallBack(iImageWriteProc *wp,iImageReadProc *rp);
Dll char * APIENTRY i_GetImageTag(iImagePtr sp);
Dll iDAFSError APIENTRY i_SetImageTag(iImagePtr sp,char *tag);

Dll iImagePtr APIENTRY i_NewImage(int2 w,int2 h,int2 res,iDAFSCompType comp);
Dll iDAFSError APIENTRY i_AddDataToImage(iImagePtr,char *,int2);
Dll iDAFSError APIENTRY i_GetDataFromImage(iImagePtr,int2,char **,int2 *);

/* these should probably have Image in the name */
Dll iDAFSError APIENTRY i_UncompressImage(iImagePtr im);
Dll iDAFSError APIENTRY i_CompressImage(iImagePtr im,iDAFSCompType comp);

Dll iBox APIENTRY i_GetImageBounds(const iImagePtr image);

Dll      int4 APIENTRY i_GetImageWidth(const iImagePtr image);
Dll      int4 APIENTRY i_GetImageHeight(const iImagePtr image);
Dll      int2 APIENTRY i_GetImageRes(const iImagePtr image);
Dll      void APIENTRY i_SetImageRes(const iImagePtr image, int2 resolution);

Dll  iImagePtr APIENTRY i_BitMap2Image(iBox bounds,int2 dw, char *dp);
/* should have 2 versions -- one that allocates and one that doesnt */
Dll iDAFSError APIENTRY i_Image2BitMap(iImagePtr image,iBox bounds,
				   iFixedPoint mag_exp,
				   int2 *dataWid,char **bits);
Dll iDAFSError APIENTRY i_RotateImage(iImagePtr image,short rotation);

Dll       int4 APIENTRY i_ImageExt(const char *fileName);

Dll iDAFSError APIENTRY i_ReadImage(const char *fileName,iImagePtr *image);
Dll iDAFSError APIENTRY i_WriteImage(const char *fileName,iImagePtr image,
				 iDAFSCompType comp);

Dll       void APIENTRY i_DisposeImage(iImagePtr image);

/* high level read write */
/* the weird asymmetry of ReadEntityNoIm */
Dll iDAFSError APIENTRY i_ReadEntity(const char *fileName,
				 iEntityPtr *entityPtr);
Dll iDAFSError APIENTRY i_ReadEntityNoIm(const char *fileName,
					 iEntityPtr *entityPtr);
Dll iDAFSError APIENTRY i_WriteEntity(const char *fileName,
				  const iEntityPtr entity,
				  iDAFSWriteMode mode);
Dll iDAFSError APIENTRY i_ReadEntityF(RafFile fd,
				 iEntityPtr *entityPtr);
Dll iDAFSError APIENTRY i_WriteEntityF(RafFile fd,
				  const iEntityPtr entity,
				  iDAFSWriteMode mode);

/* low level reading and writing */
Dll iDAFSError APIENTRY i_Read1Token(RafFile fo,iToken *t);
Dll iDAFSError APIENTRY i_ReadToken(RafFile fo,iToken *t,int *swap);
Dll iDAFSError APIENTRY i_ReadSize(RafFile fo,iToken t,
				   long *size, int *swap);
Dll iDAFSError APIENTRY i_ReadItem(RafFile fo,iToken t, uint2 space,
				   uint2 *size,void *data, int *swap);

Dll iDAFSError APIENTRY i_ReadItemAlloc(RafFile fo,iToken t,
					uint2 *size,void **data, int *swap);
Dll iDAFSError APIENTRY i_ReadItemAllocH(RafFile fo,iToken t,
					long *size, void HUGEPTR **data, 
                                        int *swap);

Dll iDAFSError APIENTRY i_WriteItem(RafFile fo,iToken t,
				uint2 size,const void *data);
Dll iDAFSError APIENTRY i_WriteItemH(RafFile fo,iToken t,
				long size,const void HUGEPTR *data);


/* swapping of standard data */
Dll short APIENTRY i_SwapS(short s);
Dll long  APIENTRY i_SwapL(long l);
Dll void APIENTRY i_SwapLArray(long *array,int4 num);
Dll void APIENTRY i_SwapSArray(short *array,int4 num);
Dll void APIENTRY i_SwapBox(iBox *b);
Dll void APIENTRY i_SwapLBox(iLBox *b);
Dll void APIENTRY i_SwapPos(iPosSetPtr p, int4 num);
Dll void APIENTRY i_SwapUShort(void*);
Dll void APIENTRY i_SwapULong(void*);
Dll void APIENTRY i_SwapUBox(void* box);
Dll void APIENTRY i_SwapULBox(void* box);


/* -------------------------------------- */
/* UNDOCUMENTED SECTION */
/* -------------------------------------- */

/* Shares */
Dll int4 APIENTRY i_CountShares(iEntityPtr en);
Dll iDAFSError APIENTRY i_ShareEntity(iEntityPtr entity,
									  iEntityPtr parent,iEntityPtr *epp);

iEntityPtr i_GetShares(iEntityPtr entity, iEntityPtr prev_borrow);

/* bounds */
/* odd balls */
iDAFSError i_AppendBox(iEntityPtr entity, iBox box);
iDAFSError i_AppendOrthogon(iEntityPtr entity, 
					 iOrthogonPtr ortho);


/* something useful for illum ? */
Dll       void APIENTRY i_GrowOrthogon(iOrthogonPtr ortho, 
									   iOrthogonPtr northo, short size);

/* should go away */
Dll iDAFSError APIENTRY i_AddOrthogon2OrthogonList(iOrthogonListPtr o_list, 
						   iOrthogonPtr ortho);
Dll iDAFSError APIENTRY i_UnionBoxWithBoxList(iBox*, iBoxListPtr, iBoxListPtr);

/* should be an internal only function */
void i_DisposeArea(iEntityPtr entity);



/* swapping */
void i_SwapXPoints(iPoint *point,int4 n);
void i_SwapUPoly(void* border); 


/* sample compare functions for i_Sort */
/* not exported because it doesnt always work */
Dll int4 APIENTRY i_VHCompare(iEntityPtr *one, iEntityPtr *two);

Dll iDAFSError APIENTRY i_OrEntity(iEntityPtr *entity);

/* image manipulation */

Dll iDAFSError APIENTRY i_ScaleImage(iImagePtr img, iFixedPoint xmag, iFixedPoint ymag);
Dll iDAFSError APIENTRY rlxscale(iImagePtr sp, iFixedPoint xmul, iImagePtr *result);
Dll iDAFSError APIENTRY rlyscale(iImagePtr sp, iFixedPoint ymul, iImagePtr *result);
Dll iDAFSError APIENTRY i_InvertImage(iImagePtr im);
Dll iDAFSError APIENTRY i_MirrorImage(iImagePtr sp);
Dll iDAFSError APIENTRY i_ConcatImage(iImagePtr left,iImagePtr right);
/* not always reliable */
Dll      iBox APIENTRY i_GetMinBounds(iImagePtr im);


/* lame trick for sharing glyphs */
/* will probably be changed */
Dll     int2 * APIENTRY i_GetPosLimit(iEntityPtr entity, int2 *size);
Dll iDAFSError APIENTRY i_SetPosLimit(iEntityPtr entity, int2 *pos, int2 size);
Dll iDAFSError APIENTRY i_RemPosLimit(iEntityPtr entity, int2 index);
Dll iDAFSError APIENTRY i_AddPosLimit(iEntityPtr entity, int2 pos);
Dll iDAFSError APIENTRY i_TrimPosLimit(iEntityPtr entity);
Dll void       APIENTRY i_NoPosLimit(iEntityPtr entity);
Dll iPosSetPtr APIENTRY i_GetTrimPosSet(iEntityPtr cur,iPosSetPtr temp, int2 *vp, unsigned short thresh);
Dll short      APIENTRY i_GetLimitGlyph(iEntityPtr cur);



#ifdef __cplusplus
};
#endif

extern const char *i_Types[];


enum iTypeInt_e {
	iNullT,
	iDocumentT,
	iParagraphT,
	iLineT,
	iWordT,
	iGlyphT,
	iZoneT,
	iRegionT,
	iSpaceT,
	iNewLineT,
	iCommentT,
	iNTypesT /* leave last  */
};


#define iDocument "Document"
#define iNewLine "NewLine"
#define iParagraph "Paragraph"
#define iRegion "Region"
#define iSpace "Space"
#define iLine "Line"
#define iZone "Zone"
#define iWord "Word"
#define iGlyph "Glyph"
#define iComment "Comment"

#define COLOR_NAME "color"
#define FLAG_PROP "Flag"

#endif







