/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef _H_entity_
#define _H_entity_

/*
// This file contains all the stuff we wanted to hide from the
// end programmer
*/

#include "dafstype.h"
#include "rlimage.h"
#include "region.h"
#include "svpmap.h"
#include "svbmap.h"

#include "rafio.h"  /* for RafFile */

#ifndef __cplusplus
typedef union iDataUnion iDataUnion;
typedef struct iTypes iTypes;
typedef struct iData iData;
typedef struct iTypeTable iTypeTable;
typedef union iAreaData iAreaData;
#endif

typedef struct svbmap iProp;

typedef struct iTypes *iTypesPtr;
typedef struct iData *iDataPtr;
typedef struct iTypeTable *iTypeTablePtr;
typedef struct iGlobid iGlobid;

struct iEntity {
	struct iEntity *next,*prev;
	struct iData *prdata;
	int2 *whichpos;
	int2 num;
	struct iData *data;
};

union iDataUnion {
	iUCode *unitext;
	char *text;
	iPosSet *posp;
	iPosSet pos;
	void *user;
};


struct iOrthogonList {
	int2 num_poly;
	iOrthogonArrayPtr poly;
};

union iAreaData {
	iBoxListPtr box_list;
	iOrthogonListPtr brdr;
	iBox box;
};

struct iArea {
	iATag     a_type;
	iAreaData a;
};

struct iGlobid {
	long en_id;
	long im_id;
	int4 im_flag;
	int DAFSBSwap;
	void **ims;
} ;

struct iData {
	iChildType ct;
	iEntityPtr child,last;
	long id;
	int use;
	struct iArea area;
	iImagePtr im;
	void *userPtr;
	iPropPtr props;
	iDTag dt;
	short num;
	iTypeIndex type;
	iEntityPtr *matrs;
	iDataUnion d;
};

struct iEditRecord {
	int4 pos,dirty;
	iUCode *save;
	iEntityPtr esave;
};
 

struct iTypes {
	char *name;
	iPropPtr props; 
};

struct iTypeTable {
	iTypesPtr table;
	iTypeIndex size;
};


/* data */
/* I'm glad we have the above comment */
extern iTypeTable i_GlobalType;
extern void *gUserData;
extern iCallBackProc gCallBack;
extern iImageWriteProc gWriteImageProc;
extern iImageReadProc gReadImageProc;

/* 
// bogus the fact that we need to tell C++ to use C linkage is a sign that
// we did not hide these functions from illum
// and probably should not have hid them from the end programmer
*/
#ifdef __cplusplus
extern "C" {
#endif


/* bogus -- obviously not really hidden */


iDAFSError rotimg(iImagePtr,iImagePtr,int,int);

Dll iImagePtr APIENTRY rotate(iImagePtr,short);

 iImagePtr rlgetzone(iImagePtr,iBox b);
 iImagePtr rlnewimage(int2,int2,int2);
      void Run2Ras(short *,char *);
      int2 Ras2Run(char *,short *,int2,int2);

      void MatrCallBack(iMess, void *, iEntityPtr );
iDAFSError DisposeProperty(iPropPtr *start, iPropPtr prop);
iDAFSError i_PNameToType(iTypeTablePtr tab,const char *type, iTypeIndex *val);
const char *i_PTypeToName(iTypeTablePtr tab,iTypeIndex type);
      void i_DisposeTypeTab(iTypeTablePtr types);
iDAFSError i_InitTab(iTypeTablePtr tab);
iDAFSError SetProperty(iEntityPtr en, const char *name,
					   iPTag type, uint2 size, void *data,int fromdisk);
iDAFSError SetTypeProperty(iTypeIndex et,const char *name,
						   iPTag type,uint2 size,void *data,int4 fromdisk);
      void Run2RasMul(short *run,char *outbuf,iFixedPoint mul,int2 off);
iDAFSError ReadDAFSBImage(RafFile  fo, iImagePtr * im, iGlobid *gid);
      void i_FreeContents(iEntityPtr entity);
iDAFSError SetBoundsType(iEntityPtr, iATag tag);
      int4 i_CountMem(iImagePtr ip);
iDAFSError WriteDAFSBImage(RafFile  fo, iImagePtr im);
      void ClearID(const iEntityPtr me);
iDAFSError WriteHeader(RafFile fo,iDAFSWriteMode mode,iEntityPtr doc);
iDAFSError ReadHeader(RafFile fo,iDAFSWriteMode *mode);

/* in read */
iDAFSError newchunk(iImagePtr,uint2);
 iImagePtr newrawimage(int2,int2,int2);
/* in rlmain */
void freeimage(iImagePtr);

      void i_InitComp(void);
      void i_InitRasNRun(void);

      Dll iDAFSError APIENTRY i_InitTypeProp(void);
      iDAFSError _WriteEntity(RafFile fo, const iEntityPtr entity, 
							  iDAFSWriteMode mode);
int4 IsHit(iEntityPtr target,iTypeIndex type, iSearchType method, void *name);

#define IMPLEMENT_TIFF 1

#ifndef IMPLEMENT_TIFF
#define IMPLEMENT_TIFF 1
#endif

#if IMPLEMENT_TIFF
iDAFSError tiffread(iImagePtr * npp, const char *fn);
iDAFSError writetiff(iImagePtr sp, const char *fn, int4 comp);
#endif


iDAFSError 
do0d(iImagePtr *npp,RafFile  in, int2 w, int2 h,
	 int4 (*fill) (RafFile , char *, int2, int4 *),int4 invert);

Dll iDAFSError APIENTRY 
i_AppendPdaZone(iImagePtr sp, const char *fn, iDAFSCompType comp, int2 x, int2 y, char *user);

iDAFSError FindUserPtrProp(iEntityPtr en,char *name,void **d);



#ifdef __cplusplus
};
#endif

#endif
