
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */


#ifndef _UniEditor_h_
#define _UniEditor_h_

// margins in 1/4 units 
//#define X_MARGIN 2
//#define Y_MARGIN 2

#include "Xm/Xm.h"
#include "dafstype.h"
#include "uString.h"
#include "misc/Unicode.h"
#include "LineList.h"
#include "Line.h"

#include "misc/CanvasClient.h"

/*
// class which encapsulates a drawingarea widget 
// and callbacks for unicode input
*/

enum ueSelectionType { stNONE, stCURSOR, stSELECTION };
// cMode must map to uDirection in Unicode.h
enum cMode { cmAFTER = 1, cmBEFORE };

class KeyMap;
class Canvas;

class UniEditor : public CanvasClient
{
  friend class RTElement;
public:
  // Widget da;
  iEntityPtr root_entity;
protected:
  LineList lines;   // actual data in here, maybe shouldn't be

  // X-window stuff

  GC gc, cursor_gc;
  Pixel background_color;    // allows us to switch GC quickly
  Cursor cursor;             // The X mouse cursor, not text cursor
  int entity_mode;           // flag if entities are enabled 
  int char_width, char_height, ascent; // information about a typical character
  int overstrike;            // flag about overstrike/insert mode
  int x_margin, y_margin;    // margins in pixels

  // selection
  int xstart, ystart;        // initial x coords for button press

  // either the beginning of selection, or the text cursor
  int begin_line, begin_col;
  int end_line, end_col;
  int goal_col;
  ueSelectionType selection_type;

  cMode cursor_mode;
  uDirection justification_direction;

  int selecting; // flag in the middle of mouse selection

#if 0
  // bogus -- should die
  int draw_x, draw_y;
  unsigned  draw_width, draw_height;
#endif

  // flag that says whether to have resize callback, or to
  // resize when we feel like it
  int scroll_flag;

  unsigned pos2line(int y);
  unsigned pos2base(int click_x, int line_pos);

  void drawAll(void);
  void drawRange(int initial, int final);
  void drawBorder(void);

  // height that the canvas wants to be
  int canvas_width;
  int canvas_height;




  void scrollIfNecessary(void);
  
  virtual int isVisible(int x, int y);
  virtual void calculateVisible(int *first, int *last);

  void selectionDrawRegion(int *first, int *last);
public:
  virtual void centerOnCoord(int x, int y);
  UniEditor(Canvas *canvas, int scroll_flag);
  virtual ~UniEditor();

  virtual void setSize(int w, int h);
  void setXMargin(int margin) { x_margin = margin; };
  void setYMargin(int margin) { y_margin = margin; };
  void coordToPosition(int x, int y, int *line, int *column);
  void positionToCoord(int line, int column, cMode cm, int *x, int *y);

  // callbacks and handlers
  void inputCallback(XmDrawingAreaCallbackStruct *cbs);


  void exposeCallback(XmDrawingAreaCallbackStruct *cbs);

  void motionHandler(XEvent *event);

  // access routines
  Widget getWidget(void);

  // input and output
  // may require resize
  void setString(const iUCode *string);
  iUCode *getString(void);
  iUCode *getString(int first_line, int first_col, 
                    int second_line, int second_col);
  void setEntity(iEntityPtr root);

  // editting
  virtual void insertAtSelection(iUCode c, int *, int *);
  virtual void deleteAtSelection(int *, int *);
  virtual void backspaceAtSelection(int *, int *);
  void blammoSelection(void);

  // cursor
  ueSelectionType getSelectionType(void);
  int getCursor(int *line, int *column, cMode *mode = 0);
  // this is whether the cursor is arabi or english
  void toggleCursorMode(void);
  void setCursorMode(cMode mode) { cursor_mode = mode; drawRange(begin_line, begin_line); };
  int getSelection(int *begin_line, int *begin_column, int *end_line, int *end_column);
  // may require repositioning
  virtual int setCursor(int line, int column);
  virtual int setSelection(int begin_line, int begin_column, int end_line, int end_column);
  virtual int clearSelection(void);
  void moveCursor(uDirection);

  int lastLine() { return lines.size-1; }
  int lastCol(int l) { return lines.line[l]->baseLength();}

  virtual KeyMap *getKM() = 0;
  void lookAtKeyMap();


  void processKeyPress(XEvent *event);

  // virtual requirements of the CanvasClient class
  /*virtual*/ void resize(XmDrawingAreaCallbackStruct *cbs);
  /*virtual*/ int resizeNecessaryP(void);
  /*virtual*/ int pressKey(iUCode c);
  /*virtual*/ iLBox getDesiredSize(void);

  /*virtual*/ void up(int state);
  /*virtual*/ void down(int state);
  /*virtual*/ void left(int state);
  /*virtual*/ void right(int state);
  /*virtual*/ void home(int state);
  /*virtual*/ void end(int state);
  /*virtual*/ void myReturn();
  /*virtual*/ void backspace();
  /*virtual*/ void myDelete();
  /*virtual*/ void insert();

  /*virtual*/ int validP() { return TRUE; }
  /*virtual*/ void drawRegionNow(struct iLBox);
  /*virtual*/ void conditionallyResize();

  /*virtual*/ void pressButton(XEvent *event);
  /*virtual*/ void releaseButton(XEvent *event);
  // funny args for the sake of follow
  virtual void moveButton(int state, int x, int y);

public:
  void prSelf();
  void pr();
  void compare(UniEditor *ue);
};

#endif



