
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */


#ifndef _CanvasClient_h_
#define _CanvasClient_h_


#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <X11/Xutil.h>

#include <Xm/DrawingA.h>

#include "dllist.h"
#include "aString.h"
#include "dafstype.h"

class Canvas;

class CanvasClient {
public:
  Canvas *canvas;

  int desired_width;
  int desired_height;
public:
  CanvasClient(Canvas *canvas);
  virtual ~CanvasClient();

  virtual void drawRegionNow(iLBox box) = 0;
  virtual int resizeNecessaryP() = 0;
  virtual iLBox getDesiredSize(void) = 0;
	
  Canvas *getCanvas();
  int activeP();
	
  // handlers and callbacks

  // why I have to handle this instead of canvas
  // because some want to resist, some can't resist
  // too complicated OOC Mode, LTE
  virtual void resize(XmDrawingAreaCallbackStruct *call_data) = 0;
  // because I might want to do something complicated
  // in image mode -- for moving the xor box?
  virtual void redisplay(XmDrawingAreaCallbackStruct *call_data);

  virtual void conditionallyResize(void) = 0;

  virtual void         pressButton(XEvent *event) = 0;
  virtual void         releaseButton(XEvent *event) = 0;
  virtual void         moveButton(int state, int x, int y) = 0;

  virtual int  pressKey(iUCode c) = 0;
	
  // for handline left right and arrow keys
  virtual void left(int state) = 0;
  virtual void right(int state) = 0;
  virtual void up(int state) = 0;
  virtual void down(int state) = 0;
  virtual void home(int state) = 0;
  virtual void end(int state) = 0;
  virtual void myReturn() = 0;
  virtual void backspace() = 0;
  virtual void myDelete() = 0;
  virtual void insert() = 0;
  	
  virtual int validP() = 0;
};

#endif
