
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef _misc_h_
#define _misc_h_

#include <Xm/Xm.h>

struct MenuItem {
	Widget result;
	const char  *label;         /* the label for the item */
	WidgetClass *klass;         /* pushbutton, label, separator... */
	char         mnemonic;      /* mnemonic; NULL if none */
	const char  *accelerator;   /* accelerator; NULL if none */
	const char  *accel_text;    /* to be converted to compound string */
	XtCallbackProc callback;   /* routine to call; NULL if none */
	XtPointer    callback_data; /* client_data for callback() */
	struct MenuItem *subitems; /* pullright menu items, if not NULL */
} ;
struct XmStringMenuItem {
	Widget result;
	XmString	label;         /* the label for the item */
	WidgetClass *klass;         /* pushbutton, label, separator... */
	char         mnemonic;      /* mnemonic; NULL if none */
	const char  *accelerator;   /* accelerator; NULL if none */
	const char  *accel_text;    /* to be converted to compound string */
	XtCallbackProc callback;   /* routine to call; NULL if none */
	XtPointer    callback_data; /* client_data for callback() */
	struct XmStringMenuItem *subitems; /* pullright menu items, if not NULL */
} ;

typedef struct MenuItem MenuItem;
typedef struct XmStringMenuItem XmStringMenuItem;

#define NULL_MENU { 0, NULL, 0, '\0', NULL, NULL, 0, 0, NULL }


struct ActionAreaItem {
    char *label;
    XtCallbackProc callback;
    void *data;
};

typedef struct ActionAreaItem ActionAreaItem;

#ifdef __cplusplus
extern "C" {
#endif
	Widget
		BuildMenu(Widget parent,
				  int menu_type,
				  char *menu_title, 
				  char menu_mnemonic,
				  MenuItem *items);
	
	Widget
		BuildXmStringMenu(Widget parent,
				  int menu_type,
				  XmString menu_title, 
				  char menu_mnemonic,
				  XmFontList font_list,
				  XmStringMenuItem *items);
	
	
	Widget
		CreateActionArea(
			Widget parent,
			ActionAreaItem *actions,
			int num_actions);
	
	void miscInit(Widget toplevel);
	void miscExit();
	
#ifdef __cplusplus
};
#endif



#endif
