
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef _rafport_h_
#define _rafport_h_

/*
// This file contains definitions for portablility
//
// first, there are a few C versus C++ compatibility issues
//
// second there are system differences
// we attempt to determine the system environment 
// and set some convenient defines and
// second attempt to define a few things like byter order
// 
// system differences are broken down like this:
//

// COMPILER        OS       ARCH    LIBTYPE
//
// MSC             LINUX    INTEL   DLL (for windows)
// GNU             BSD      SPARC
// WATCOM          SYSV     M68K
//                 DOS      AMD29K
//                 WIN16
//                 WIN32
// XL              AIX      PR
// HIGHC           SOLARIS
//                 SUNOS
// 
// and so there are defines like
// SYS_COMPILER_GNU
// SYS_OS_DOS
// SYS_ARCH_MOTOROLA
// 
*/

#ifdef  SYSTYPE_SVR4
    #define SYS_OS_IRIX
    #define SYS_ARCH_MIPS
#endif

#ifdef SYS_PR
	#define SYS_OS_AIX
	#define SYS_COMPILER_XL
	#define SYS_ARCH_PR
	#undef SYS_PR
#endif

#ifdef SYS_TRUESCAN
	#define SYS_COMPILER_UNKNOWN
	#define SYS_ARCH_MOTOROLA
	#define SYS_OS_UNKNOWN
	#undef SYS_TRUESCAN
#endif

#ifdef _MSC_VER
	#define SYS_COMPILER_MSC
	#define INTERNAL_MSDOS
#endif

#ifdef __WATCOMC__
	#define SYS_COMPILER_WATCOM
	#ifdef __WINDOWS__
		#define SYS_OS_WIN16
	#else
		#define SYS_OS_WIN32
	#endif
	#define INTERNAL_MSDOS
#endif

#ifdef INTERNAL_MSDOS
	#define SYS_ARCH_INTEL
	#define SYS_OS_DOS
#endif

#ifdef _WINDOWS
	#ifdef WIN32
		#define SYS_OS_WIN32
	#else
		#define SYS_OS_WIN16
	#endif
#endif

#if defined(SYS_OS_WIN32) || defined(SYS_OS_WIN16) /* assert windows */
	#include "windows.h"

	#ifdef SYS_OS_WIN32

		#ifdef SYS_LIBTYPE_DLL
			#define Dll __declspec(dllexport)
		#else /* not dll lib */
			#ifdef SYS_LIBTYPE_STAT
				#define Dll
			#else
				#define Dll __declspec(dllimport)
			#endif
		#endif

	#endif

	#ifdef SYS_OS_WIN16

		#ifdef _WINDLL
			#define Dll
			#define APIENTRY WINAPI __export
		#else
			#define Dll
			#define APIENTRY WINAPI
		#endif

	#endif

#else /* assert not windows */

	#define Dll
	#define APIENTRY

#endif /* not windows */


#ifdef __linux__
	#define SYS_OS_LINUX
	#define SYS_ARCH_INTEL
#endif

#ifdef _EPI
	#define SYS_OS_YARC
	#define SYS_ARCH_AMD29K
	#define SYS_COMPILER_EPI
#endif

#ifdef NeXT
	#define SYS_ARCH_INTEL
	#define SYS_OS_MACH
#endif

#ifdef sparc
	#define SYS_ARCH_SPARC
#ifdef __unix__

#ifdef __svr4__
	#define SYS_OS_SOLARIS
#else
	#define SYS_OS_SUNOS
#endif

#endif
#endif

#ifdef __GNUC__
	#define SYS_COMPILER_GNU
	#ifndef __STDC__
		#define const
	#endif
#endif

#ifdef __HIGHC__
	#define SYS_COMPILER_HIGHC
#endif

/* add whatever compilers can use inline keyword here */
#if defined(SYS_COMPILER_GNU) || defined(SYS_COMPILER_HIGHC)
	#define DO_INLINE 1
#else
	#define DO_INLINE 0
#endif
#ifdef SYS_COMPILER_HIGHC
	#define inline _Inline
	#define DO_INLINE 1
#endif

#if !defined(SYS_OS_DOS) && !defined(SYS_OS_YARC)
	#define huge
#endif

#ifdef SYS_OS_WIN16
	#define HUGEPTR huge
#else
	#define HUGEPTR
#endif

/*
// here are some system specific special hacks
*/

/*
// prevent XL (the compiler on the NCS machine)
// from using its own idea of ucs routines since
// it thinks it is 4 bytes
*/

#ifdef SYS_COMPILER_XL
	#define __H_CTYPE
	#define _H_STRING
#endif

/*
// prevent VC++ from using its own definitions of uc string routines 
// since they conflict with our own
*/

#ifdef SYS_COMPILER_MSC
	#define _WSTRING_DEFINED
#endif

/* 
// another problem on the ncs machine of a conflicting type definition
*/
#ifndef SYS_OS_AIX
typedef unsigned char uchar;
#endif

/*
// for the TIFF library
*/
#ifdef SYS_COMPILER_MSC
	#define BSDTYPES
#endif

typedef void *VoidPtr;
typedef void *Ref;

typedef short int2;
typedef unsigned short  uint2;
typedef long int4;
typedef unsigned long   uint4;

#if 0
#if defined(__STDC__)  
typedef signed char byte;
#else
typedef char byte;
#endif
typedef unsigned char ubyte;
#endif

typedef int (*CompareFunction)(const void*,const void*);

#endif




