
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef _smemory_h_
#define _smemory_h_

/*
// Our version of memory allocation
// all of our code will call these. But some things won't:
// new and delete operators in C++ and the X-windows code,
// unless you do some extra tricks.
//
// It has many options that allow you to track memory usage
// if you are writing a c program, you initialize the library
// in the main function, but if you are running a C++ program,
// there is a chance some contructor will call smalloc before the
// main function is reached, and memory tracking will be confused.
// It is safer to break in smalloc in the debugger and then set
// the tracking flag
//
// the names smalloc, smfree go back to Calera and RAF's first projects
// that integrated Calera technology
*/

#include "rafport.h"
	
enum sMemFlags {
	sNoChecks,
	sCountTot,
	sTrackAll,
	sCheckClobber,
	sNoReuse,
	sClearFree,
	sCheckClobberAlways,
} ;

typedef enum sMemFlags sMemFlags;

struct hdr {
	long size;
	long alloc_num;
	long dummy; /* need to make hdr multiple of 8 bytes */
	struct hdr *next;
};

typedef struct hdr hdr;


struct MemoryInfo {
	sMemFlags memory_track_flag;
	void *badptr;
	int4 badsize;
	long memlimit;
	long totmem;
	long maxmem;
	long nsmallocs;
	long stophere;
	long sizelimit;
	long failhere;
};

typedef struct MemoryInfo  MemoryInfo;

#define MAX_SMALLOC 31000


#ifdef __cplusplus
extern "C" {
#endif

typedef int4 (APIENTRY *MemoryFailureCallback)(void *);
	
	long s_getsize(void *q);
Dll void APIENTRY s_InitLib(sMemFlags flags, 
								MemoryFailureCallback callback, 
								void *user_data);
Dll void APIENTRY smfree(void *ptr);
Dll	void * APIENTRY smalloc(int4 size);
Dll	void * APIENTRY smalloc0(int4 size);
Dll	void * APIENTRY smrealloc(void *ptr, int4 size);
Dll void APIENTRY s_SetMemFlags(int4 num);
Dll	void APIENTRY s_ExitLib(void);

#ifdef SYS_OS_WIN16
    void smfreeH(void HUGEPTR *);
    void HUGEPTR * smallocH(int4 size);
    void HUGEPTR * smreallocH(void HUGEPTR *, int4 size);
#else
#define smfreeH(p) smfree(p)
#define smallocH(s) smalloc(s)
#define smreallocH(p, s) smrealloc(p, s)
#endif

	int4 checkptr(void *ptr);
	void prlists(void);

    void smpurge(void);

#ifdef DEBUG
    void smstophere(int4 order);
    void add_clobber(char **ptr, int4 size);
    long gettotmem(void);
    void setmemlimit(int4 limit);
    void setmemlimit(int4 limit);
    int4 clobbered(char *ptr, long size);
    void check_all_clobber(void);
    void smstat(int4 i);
    int4 free_written_to(unsigned char *ptr, long size);
    void clear_mem(unsigned char *ptr, long size);
    void smprmem(void);
    void check_not_free(hdr *q);
    void s_print_list(int4 i);
#if QUICK_LIST
    void print_non_free(void);
#endif
    void setmemspace(long space);
#endif
	
	
#ifdef __cplusplus
};
#endif

#endif
