
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#ifndef _stringVoidPtrMap_h_
#define _stringVoidPtrMap_h_

/*
// My version of a map that uses a hash table
// I based the function calls on the interface to the GNU C++ lib
// and the guts on simple fibonacci hashing
*/

#include "rafport.h"

struct svpMapRecord {
	char *key;
	int4 deleted;
	VoidPtr value;
};

#ifndef __cplusplus
typedef struct svpMapRecord svpMapRecord;
#endif

struct stringVoidPtrMap {
	int4 array_size;
	int4 count;
	svpMapRecord *array;
	VoidPtr default_value;
};

#ifndef __cplusplus
typedef struct stringVoidPtrMap stringVoidPtrMap;
#endif

#ifdef __cplusplus
extern "C" {
#endif
	
	stringVoidPtrMap *svmapAlloc(void);
	void svmapFree(stringVoidPtrMap *that);
	
	void svmapInit(stringVoidPtrMap *that, VoidPtr default_value, int4 size);
	void svmapCleanup(stringVoidPtrMap *that);
	
	int4 svmapEmptyP(stringVoidPtrMap *that);
	int4 svmapLength(stringVoidPtrMap *that);
	VoidPtr svmapGet(stringVoidPtrMap *that, const char *key);
	void svmapSet(stringVoidPtrMap *that, const char *key, VoidPtr value);
	
	int4 svmapContainsP(stringVoidPtrMap *that, const char *key);
	void svmapDelete(stringVoidPtrMap *that, const char *key);
	void svmapClear(stringVoidPtrMap *that);
	VoidPtr svmapDefault(stringVoidPtrMap *that);
	char *svmapKey(stringVoidPtrMap *that, Ref r);
	VoidPtr svmapContents(stringVoidPtrMap *that, Ref r);
	Ref svmapFirst(stringVoidPtrMap *that);
	Ref svmapNext(stringVoidPtrMap *that, Ref prev);
	Ref svmapSeek(stringVoidPtrMap *that, const char *key);
	
#ifdef __cplusplus
};
#endif

#endif
