
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#ifndef _uString_h_
#define _uString_h_



#include <stddef.h>
#include <stdlib.h>
#ifdef USE_STREAM
#include <iostream.h>
#include <strstream.h>
#endif
#include <assert.h>
#include <stdio.h>
#include <float.h>
#include <string.h>
#include <math.h>

#include "aString.h"

extern "C" {
#include "ucs.h"
};

extern const iUCode *EmptyUString;

void *operator new(size_t size);
void  operator delete(void *data);

double random01(void); // generate a random number between 0 and 1


struct aString;

struct uString
{
	void putUtf(aString&) const;
	int asciiP();
	void getUnicode(const aString&);
	int utfLength();
	
	uString(const aString&);
	uString(void);
	uString(const iUCode *s);
	uString(iUCode c);
	uString(const uString& s);
	uString(const char *s);
    ~uString(void);

	const uString &ClassName(void) const; 
	int   ClassMember(const uString& TestString) const;

	const iUCode *string(void) const ;
	int   length(void) const ;
	void  remove(unsigned start, unsigned length=1);
	void  move(unsigned start, unsigned length=1);
        uString mid(int start, int stop);
        void  reverse(void);

	iUCode  operator[](unsigned index) const;
	iUCode &operator[](unsigned index);
	uString &operator =(const uString &c); 
	uString &operator =(const iUCode *c); 
	uString &operator +=(const uString &c);
	uString &operator +=(const iUCode *c); 
	uString &operator --(void);
#if 0
	inline const iUCode *operator()(int ignore = 0) const { if(data) return(data); return(EmptyUString);  }
#endif
	int   operator !(void) const; 
	int   operator !=(const uString &c) const; 
	int   operator !=(const  iUCode *c) const; 
	int   operator ==(const uString &c) const; 
	int   operator ==(const iUCode *c) const; 
	int   operator <=(const uString &c) const; 
	int   operator <=(const iUCode *c) const; 
	
	int   operator <(const uString &c) const; 
	int   operator <(const iUCode *c) const; 
	int   operator >(const uString &c) const; 
	int   operator >(const iUCode *c) const; 
	int   operator >=(const uString &c) const; 
	int   operator >=(const iUCode *c) const; 
	
	uString operator +(const uString &c) const; 
	uString operator +(const iUCode *c) const; 
	
#ifdef USE_STREAM
	friend ostream& operator << (ostream& o,const uString &Y);
#endif
	
	friend  int   operator ==(const  iUCode *c,const uString &s); 
	friend  int   operator !=(const  iUCode *c,const uString &s); 
	friend  int   operator <=(const  iUCode *c,const uString &s); 
	friend  int   operator <(const  iUCode *c,const uString &s); 
	friend  int   operator >=(const  iUCode *c,const uString &s); 
	friend  int   operator >(const  iUCode *c,const uString &s); 
	friend  uString  operator +(const  iUCode *c,const uString &s);
	// 	friend  const iUCode *operator =(const  iUCode *c,const uString &s); 
	
protected:
    void clear(void);
    void extend(unsigned newsize);
	
	unsigned allocated;
	unsigned len;
	iUCode *data;
};


#endif _H_uString

