
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "str-lib.h"

#include "KeyMap.h"

iUCode uNULL[1]={0};

KeyMap::KeyMap(aString input_name) 
	: name(input_name)
{
	svmapInit(&this->map, 0, 1000);
	svmapInit(&this->control_map, 0, 1000);
	// append myself to list of keyboards
}

KeyMap::~KeyMap()
{
	// bogus delete records
	Ref r;
	KeyRecord *record;
	
	for(r = svmapFirst(&this->map); r; r = svmapNext(&this->map, r)){
		record = (KeyRecord *) svmapContents(&this->map, r);
		delete record;
	}
	
	for(r = svmapFirst(&this->control_map); r; r = svmapNext(&this->control_map, r)){
		record = (KeyRecord *) svmapContents(&this->control_map, r);
		delete record;
	}
	
	svmapCleanup(&this->map);
	svmapCleanup(&this->control_map);
}

static 
void keyFromSym(KeySym sym, aString& key)
{
	char buffer[20];
	
	sprintf(buffer, "%d", sym);
	key = buffer;
}

void 
KeyMap::setTable(KeySym sym, uString name, iUCode value, uString uvalue)
{
	KeyRecord *record = new KeyRecord();
	aString key;
	record->sym = sym;
	record->normal.name = name;
	record->normal.value = value;
	record->normal.uvalue = uvalue;
	keyFromSym(sym, key);
	if(svmapContainsP(&this->map, key.string()))
		halt("Duplicate keys!!\n");
	svmapSet(&this->map, key.string(), record);
}

void
KeyMap::setTable(KeySym sym, iUCode value)
{
	KeyRecord *record = new KeyRecord();
	aString key;
	record->sym = sym;
	record->normal.name = uString(value);
	record->normal.value = value;
	record->normal.uvalue = uString(value);
	keyFromSym(sym, key);
	if(svmapContainsP(&this->map, key.string()))
		halt("Duplicate keys!!\n");
	svmapSet(&this->map, key.string(), record);	
}

void 
KeyMap::setControlTable(KeySym sym, uString name, iUCode value, uString uvalue)
{
	KeyRecord *record = new KeyRecord();
	aString key;
	record->sym = sym;
	record->normal.name = name;
	record->normal.value = value;
	record->normal.uvalue = uvalue;
	keyFromSym(sym, key);
	if(svmapContainsP(&this->control_map, key.string()))
		halt("Duplicate keys!!\n");
	svmapSet(&this->control_map, key.string(), record);
}

void
KeyMap::setControlTable(KeySym sym, iUCode value)
{
	KeyRecord *record = new KeyRecord();
	aString key;
	record->sym = sym;
	record->normal.name = uString(value);
	record->normal.value = value;
	record->normal.uvalue = uString(value);
	keyFromSym(sym, key);
	if(svmapContainsP(&this->control_map, key.string()))
		halt("Duplicate keys!!\n");
	svmapSet(&this->control_map, key.string(), record);	
}



void 
KeyMap::disableControlTable(KeySym sym)
{
	aString key;
	keyFromSym(sym, key);
	if(svmapContainsP(&this->control_map, key.string()) == FALSE)
	  halt("Disable Non-existent key!!\n");

	KeyRecord *record;
	record = (KeyRecord *) svmapGet(&this->control_map, key.string());
	record->sym = sym;
	record->normal.name = uString();
	record->normal.value = '\0';
	record->normal.uvalue = uString();
}


void 
KeyMap::disableTable(KeySym sym)
{
	aString key;
	keyFromSym(sym, key);
	if(svmapContainsP(&this->map, key.string()) == FALSE)
	  halt("Disable Non-existent key!!\n");

	KeyRecord *record;
	record = (KeyRecord *) svmapGet(&this->map, key.string());
	record->sym = sym;
	record->normal.name = uString();
	record->normal.value = '\0';
	record->normal.uvalue = uString();
}


iUCode
KeyMap::lookupValue(KeySym sym)
{
	KeyRecord *rec;
	aString key;
	const char *key_s;
	keyFromSym(sym, key);
	key_s = key.string();
	
	if (svmapContainsP(&this->map, key_s)) {
		rec = (KeyRecord *) svmapGet(&this->map, key_s);
		return rec->normal.value;
	}
	return '\0';
}

uString
KeyMap::lookupUValue(KeySym sym)
{
	KeyRecord *record;
	aString key;
	keyFromSym(sym, key);
	if (svmapContainsP(&this->map, key.string())) {
		record = (KeyRecord *) svmapGet(&this->map, key.string());
		return record->normal.uvalue;
	}
	return uNULL;
}

uString
KeyMap::lookupName(KeySym sym)
{
	KeyRecord *record;
	aString key;
	keyFromSym(sym, key);
	if (svmapContainsP(&this->map, key.string())) {
		record = (KeyRecord *) svmapGet(&this->map, key.string());
		return record->normal.name;
	}
	return "";
}

iUCode
KeyMap::lookupControlValue(KeySym sym)
{
	KeyRecord *record;
	aString key;
	keyFromSym(sym, key);
	if (svmapContainsP(&this->control_map, key.string())) {
		record = (KeyRecord *) svmapGet(&this->control_map, key.string());
		return record->normal.value;
	}
	return '\0';
}

uString
KeyMap::lookupControlUValue(KeySym sym)
{
	KeyRecord *record;
	aString key;
	keyFromSym(sym, key);
	if (svmapContainsP(&this->control_map, key.string())) {
		record = (KeyRecord *) svmapGet(&this->control_map, key.string());
		return record->normal.uvalue;
	}
	return uNULL;
}

uString
KeyMap::lookupControlName(KeySym sym)
{
	KeyRecord *record;
	aString key;
	keyFromSym(sym, key);
	if (svmapContainsP(&this->control_map, key.string())) {
		record = (KeyRecord *) svmapGet(&this->control_map, key.string());
		return record->normal.name;
	}
	return "";
}

// default behavior is to just return the unicode string
uString 
KeyMap::context(const uString &line)
{
	uString glyphs = line;
	return glyphs;
}

// skip over the general diacritical marks
int 
KeyMap::spacing(iUCode character)
{
	if(character < 0x0300 || character > 0x0341)
		return TRUE;
	else
		return FALSE;
}

void
KeyMap::getInfo(int *direction, iUCode *letter)
{
	*direction = NORMAL;
	*letter = 'M';
}
