
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#include "rafport.h"
#include "str-lib.h"

#include <X11/keysym.h>

#include "KeyMap.h"

#define ESC 27

class American : public KeyMap
{
public:
	American(aString input_name);
	uString context(uString line);
};

American::American(aString input_name) 
	: KeyMap(input_name)
{
	return;
}

static iUCode 
grave(iUCode chr)
{
	switch(chr){
	  case 0x0061: // a
		return 0x00E0;
	  case 0x0065: // e
		return 0x00E8;
	  default:
		return chr;
	}
}

static iUCode 
acute(iUCode chr)
{
	switch(chr){
	  case 0x0061: // a
		return 0x00E1;
	  case 0x0065: // e
		return 0x00E9;
	  default:
		return chr;
	}
}

uString
American::context(uString line)
{
	int i, length;
	uString glyphs;

	length = line.length();
	for(i = 0; i < length; i++)
		switch(line[i]){
		  case 0x0300:
			--glyphs;
			glyphs += grave(line[i-1]);
			break;
		  case 0x0301:
			--glyphs;
			glyphs += acute(line[i-1]);
			break;
		  default:
			glyphs += line[i];
		}

	return glyphs;
}	

KeyMap *
initAmerican()
{
    American *american;
    KeyMap *keymap;
	
    american = new American("American");
    keymap = american;
    keymap->setTable(XK_backslash, uString('\\'),  '\\', uString('\\'));
    keymap->setTable(XK_Delete, uString("Del"), '\b', uString());
	
    keymap->setTable(XK_Escape, uString("Esc"), '\e', uString());
    keymap->setTable(XK_1, uString('1'), '1', uString('1'));
    keymap->setTable(XK_2, uString('2'), '2', uString('2'));
    keymap->setTable(XK_3, uString('3'), '3', uString('3'));
    keymap->setTable(XK_4, uString('4'), '4', uString('4'));
    keymap->setTable(XK_5, uString('5'), '5', uString('5'));
    keymap->setTable(XK_6, uString('6'), '6', uString('6'));
    keymap->setTable(XK_7, uString('7'), '7', uString('7'));
    keymap->setTable(XK_8, uString('8'), '8', uString('8'));
    keymap->setTable(XK_9, uString('9'), '9', uString('9'));
    keymap->setTable(XK_0, uString('0'), '0', uString('0'));
    keymap->setTable(XK_minus, uString('-'), '-', uString('-'));
    keymap->setTable(XK_equal, uString('='), '=', uString('='));
    keymap->setTable(XK_BackSpace, uString("BS"), '\b', uString());
	
    keymap->setTable(XK_Tab, uString("Tab"), '\t', uString('\t'));
    keymap->setTable(XK_q, uString('q'), 'q', uString('q'));
    keymap->setTable(XK_w, uString('w'), 'w', uString('w'));
    keymap->setTable(XK_e, uString('e'), 'e', uString('e'));
    keymap->setTable(XK_r, uString('r'), 'r', uString('r'));
    keymap->setTable(XK_t, uString('t'), 't', uString('t'));
    keymap->setTable(XK_y, uString('y'), 'y', uString('y'));
    keymap->setTable(XK_u, uString('u'), 'u', uString('u'));
    keymap->setTable(XK_i, uString('i'), 'i', uString('i'));
    keymap->setTable(XK_o, uString('o'), 'o', uString('o'));
    keymap->setTable(XK_p, uString('p'), 'p', uString('p'));
    keymap->setTable(XK_bracketleft, uString('['), '[', uString('['));
    keymap->setTable(XK_bracketright, uString(']'), ']', uString(']'));
    keymap->setTable(XK_Return, uString("Ret"), '\n', uString('\n'));
	
    keymap->setTable(XK_Control_L, uString("Control"), '\0', uString());
    keymap->setTable(XK_a, uString('a'), 'a', uString('a'));
    keymap->setTable(XK_s, uString('s'), 's', uString('s'));
    keymap->setTable(XK_d, uString('d'), 'd', uString('d'));
    keymap->setTable(XK_f, uString('f'), 'f', uString('f'));
    keymap->setTable(XK_g, uString('g'), 'g', uString('g'));
    keymap->setTable(XK_h, uString('h'), 'h', uString('h'));
    keymap->setTable(XK_j, uString('j'), 'j', uString('j'));
    keymap->setTable(XK_k, uString('k'), 'k', uString('k'));
    keymap->setTable(XK_l, uString('l'), 'l', uString('l'));
    keymap->setTable(XK_semicolon, uString(';'), ';', uString(';'));
    keymap->setTable(XK_quoteright, uString('\''), '\'', uString('\''));
    keymap->setTable(XK_quoteleft, uString('`'), '`', uString('`'));
	
    keymap->setTable(XK_Shift_L, uString("Shift"), '\0', uString());
    keymap->setTable(XK_z, uString('z'), 'z', uString('z'));
    keymap->setTable(XK_x, uString('x'), 'x', uString('x'));
    keymap->setTable(XK_c, uString('c'), 'c', uString('c'));
    keymap->setTable(XK_v, uString('v'), 'v', uString('v'));
    keymap->setTable(XK_b, uString('b'), 'b', uString('b'));
    keymap->setTable(XK_n, uString('n'), 'n', uString('n'));
    keymap->setTable(XK_m, uString('m'), 'm', uString('m'));
    keymap->setTable(XK_less, uString('<'), '<', uString('<'));
    keymap->setTable(XK_greater, uString('>'), '>', uString('>'));
    keymap->setTable(XK_question, uString('?'), '?', uString('?'));
    keymap->setTable(XK_Shift_R, uString("Shift"), '\0', uString());
    keymap->setTable(XK_Linefeed, uString("LF"), '\0', uString());
	
    keymap->setTable(XK_Caps_Lock, uString(), '\0', uString());
    keymap->setTable(XK_Alt_L, uString("Alt"), '\0', uString());
    keymap->setTable(XK_Meta_L, uString("Meta"), '\0', uString());
    keymap->setTable(XK_space, uString(), ' ', uString(' '));
    keymap->setTable(XK_Meta_R, uString("Meta"), '\0', uString());
    keymap->setTable(XK_Alt_R, uString(), '\0', uString());
    keymap->setTable(XK_Super_R, uString(), '\0', uString());
	
    keymap->setTable(XK_Control_R, uString(), '\0', uString("xo"));
	
    keymap->setTable(XK_bar, uString('|'), '|', uString('|'));
	
    keymap->setTable(XK_exclam, uString('!'), '!', uString('!'));
    keymap->setTable(XK_at, uString('@'), '@', uString('@'));
    keymap->setTable(XK_numbersign, uString('#'), '#', uString('#'));
    keymap->setTable(XK_dollar, uString('$'), '$', uString('$'));
    keymap->setTable(XK_percent, uString('%'), '%', uString('%'));
    keymap->setTable(XK_asciicircum, uString('^'), '^', uString('^'));
    keymap->setTable(XK_ampersand, uString('&'), '&', uString('&'));
    keymap->setTable(XK_asterisk, uString('*'), '*', uString('*'));
    keymap->setTable(XK_parenleft, uString('('), '(', uString('('));
    keymap->setTable(XK_parenright, uString(')'), ')', uString(')'));
    keymap->setTable(XK_underscore, uString('_'), '_', uString('_'));
    keymap->setTable(XK_plus, uString('+'), '+', uString('+'));
	
    keymap->setTable(XK_Q, uString('Q'), 'Q', uString('Q'));
    keymap->setTable(XK_W, uString('W'), 'W', uString('W'));
    keymap->setTable(XK_E, uString('E'), 'E', uString('E'));
    keymap->setTable(XK_R, uString('R'), 'R', uString('R'));
    keymap->setTable(XK_T, uString('T'), 'T', uString('T'));
    keymap->setTable(XK_Y, uString('Y'), 'Y', uString('Y'));
    keymap->setTable(XK_U, uString('U'), 'U', uString('U'));
    keymap->setTable(XK_I, uString('I'), 'I', uString('I'));
    keymap->setTable(XK_O, uString('O'), 'O', uString('O'));
    keymap->setTable(XK_P, uString('P'), 'P', uString('P'));
    keymap->setTable(XK_braceleft, uString('{'), '{', uString('{'));
    keymap->setTable(XK_braceright, uString('}'), '}', uString('}'));
	
    keymap->setTable(XK_A, uString('A'), 'A', uString('A'));
    keymap->setTable(XK_S, uString('S'), 'S', uString('S'));
    keymap->setTable(XK_D, uString('D'), 'D', uString('D'));
    keymap->setTable(XK_F, uString('F'), 'F', uString('F'));
    keymap->setTable(XK_G, uString('G'), 'G', uString('G'));
    keymap->setTable(XK_H, uString('H'), 'H', uString('H'));
    keymap->setTable(XK_J, uString('J'), 'J', uString('J'));
    keymap->setTable(XK_K, uString('K'), 'K', uString('K'));
    keymap->setTable(XK_L, uString('L'), 'L', uString('L'));
    keymap->setTable(XK_colon, uString(':'), ':', uString(':'));
    keymap->setTable(XK_quotedbl, uString('\"'), '\"', uString('\"'));
    keymap->setTable(XK_asciitilde, uString('~'), '~', uString('~'));
	
    keymap->setTable(XK_Z, uString('Z'), 'Z', uString('Z'));
    keymap->setTable(XK_X, uString('X'), 'X', uString('X'));
    keymap->setTable(XK_C, uString('C'), 'C', uString('C'));
    keymap->setTable(XK_V, uString('V'), 'V', uString('V'));
    keymap->setTable(XK_B, uString('B'), 'B', uString('B'));
    keymap->setTable(XK_N, uString('N'), 'N', uString('N'));
    keymap->setTable(XK_M, uString('M'), 'M', uString('M'));
    keymap->setTable(XK_comma, uString(','), ',', uString(','));
    keymap->setTable(XK_period, uString('.'), '.', uString('.'));
    keymap->setTable(XK_slash, uString('/'), '/', uString('/'));

	keymap->setControlTable(XK_quoteleft, uString('`'), 0x0300, uString(0x0300));
	keymap->setControlTable(XK_quoteright, uString('\''), 0x0301, uString(0x0301));
	keymap->setControlTable(XK_asciitilde, uString(0x00aa), 0x00aa, uString(0x00aa));
	keymap->setControlTable(XK_1, uString(0x00b9),0x00b9, uString(0x00b9));	
	keymap->setControlTable(XK_2, uString(0x00b2),0x00b2, uString(0x00b2));	
	keymap->setControlTable(XK_3, uString(0x00b3),0x00b3, uString(0x00b3));	
	keymap->setControlTable(XK_4, uString(0x00a2),0x00a2, uString(0x00a2));	
	keymap->setControlTable(XK_5, uString(0x00a3),0x00a3, uString(0x00a3));	
	keymap->setControlTable(XK_6, uString(0x00a5),0x00a5, uString(0x00a5));	
	keymap->setControlTable(XK_7, uString(0x00a4),0x00a4, uString(0x00a4));	
	keymap->setControlTable(XK_8, uString(0x0192),0x0192, uString(0x0192));	
	keymap->setControlTable(XK_9, uString(0x00b0),0x00b0, uString(0x00b0));	
	keymap->setControlTable(XK_0, uString(0x00d7),0x00d7, uString(0x00d7));	
	keymap->setControlTable(XK_minus, uString(0x00f7),0x00f7, uString(0x00f7));	
	keymap->setControlTable(XK_equal, uString(0x00b1),0x00b1, uString(0x00b1));
	keymap->setControlTable(XK_backslash, uString(0x2015),0x2015, uString(0x2015));
	keymap->setControlTable(XK_asciicircum, uString(0x2020), 0x2020, uString(0x2020));
	keymap->setControlTable(XK_ampersand, uString(0x2021), 0x2021, uString(0x2021));
		
	keymap->setControlTable(XK_BackSpace, uString(0x00ad),0x00ad, uString(0x00ad));	
	
	keymap->setControlTable(XK_q, uString(0x00BC),0x00BC, uString(0x00BC));	
	keymap->setControlTable(XK_w, uString(0x00BD),0x00BD, uString(0x00BD));	
	keymap->setControlTable(XK_e, uString(0x00BE),0x00BE, uString(0x00BE));	
	
	keymap->setControlTable(XK_i, uString(0x2120),0x2120, uString(0x2120));	
	keymap->setControlTable(XK_o, uString(0x00f8),0x00f8, uString(0x00f8));	
	keymap->setControlTable(XK_p, uString(0x00a9),0x00a9, uString(0x00a9));	
	keymap->setControlTable(XK_bracketleft, uString(0x00ae),0x00ae, uString(0x00ae));	
	keymap->setControlTable(XK_bracketright, uString(0x2122),0x002122, uString(0x2122));	

	keymap->setControlTable(XK_a, uString(0x00e5),0x00e5, uString(0x00e5));	
	keymap->setControlTable(XK_s, uString(0x0161),0x0161, uString(0x0161));	
	keymap->setControlTable(XK_d, uString(0x00f0),0x00f0, uString(0x00f0));	
	keymap->setControlTable(XK_f, uString(0x2022),0x2022, uString(0x2022));	
	keymap->setControlTable(XK_g, uString(0x00b7),0x00b7, uString(0x00b7));	
	keymap->setControlTable(XK_h, uString(0x201a),0x201a, uString(0x201a));	
	keymap->setControlTable(XK_j, uString(0x201e),0x201e, uString(0x201e));	
	keymap->setControlTable(XK_k, uString(0x2018),0x2018, uString(0x2018));	
	keymap->setControlTable(XK_l, uString(0x2019),0x2019, uString(0x2019));	
	keymap->setControlTable(XK_semicolon, uString(0x201c),0x201c, uString(0x201c));	
//	keymap->setControlTable(XK_quoteright, uString(0x201d),0x201d, uString(0x201d));	
//	keymap->setControlTable(XK_quoteleft, uString(0x00ba), 0x00ba, uString(0x00ba));

	keymap->setControlTable(XK_v, uString(0x00ac),0x00ac, uString(0x00ac));	
	keymap->setControlTable(XK_b, uString(0x2026),0x2026, uString(0x2026));	
	keymap->setControlTable(XK_n, uString(0x00a6),0x00a6, uString(0x00a6));	
	keymap->setControlTable(XK_m, uString(0x00b5),0x00b5, uString(0x00b5));	
	keymap->setControlTable(XK_comma, uString(0x2039),0x2039, uString(0x2039));	
	keymap->setControlTable(XK_period, uString(0x203a),0x203a, uString(0x203a));	
	keymap->setControlTable(XK_slash, uString(0x00fe),0x00fe, uString(0x00fe));
	
	keymap->setControlTable(XK_exclam, uString(0x00a1),0x00a1, uString(0x00a1));
	keymap->setControlTable(XK_at, uString(0x00bf),0x00bf, uString(0x00bf));
	keymap->setControlTable(XK_numbersign, uString(0x00b6),0x00b6,
	uString(0x00b6));
	keymap->setControlTable(XK_dollar, uString(0x00a7),0x00a7, uString(0x00a7));
	keymap->setControlTable(XK_percent, uString(0x2030),0x2030, uString(0x2030));
	
	keymap->setControlTable(XK_O, uString(0x00D8),0x00D8, uString(0x00D8));	

	keymap->setControlTable(XK_A, uString(0x00c5),0x00c5, uString(0x00c5));	
	keymap->setControlTable(XK_S, uString(0x0160),0x0160, uString(0x0160));	
	keymap->setControlTable(XK_D, uString(0x00d0),0x00d0, uString(0x00d0));	
	
	keymap->setControlTable(XK_less, uString(0x00ab),0x00ab, uString(0x00ab));	
	keymap->setControlTable(XK_greater, uString(0x00bb),0x00bb, uString(0x00bb));	
	keymap->setControlTable(XK_question, uString(0x00de), 0x00de, uString(0x00de));
	keymap->setControlTable(XK_bar, uString(0x2013), 0x2013, uString(0x2013));
    return keymap;
}


