
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"

#include <stdlib.h>
#include <stdio.h>

#include <Xm/CutPaste.h>
#include <Xm/DrawingA.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <X11/cursorfont.h>
#include <Xm/RowColumn.h>

#include "rafutil.h"
#include "smemory.h"

#include "uString.h"
#include "misc/Unicode.h"
#include "keyboard/KeyMap.h"

#include "dafs.h"
#include "ucs.h"
#include "utf.h"

#include "misc/oscroll.h"
#include "Line.h"
#include "LineList.h"

#include "UniEditor.h"

void
UniEditor::drawBorder(void)
{
  Widget da = this->getWidget();
  
  if(XtIsManaged(da) == FALSE)
    return;

  int draw_x, draw_y;
  unsigned int draw_w, draw_h;
  unsigned draw_border, draw_depth;
  Drawable draw_root;

  XGetGeometry(XtDisplay(da), XtWindow(da), &draw_root, &draw_x, &draw_y, 
			   &draw_w, &draw_h, &draw_border, &draw_depth); 

  XSetForeground(XtDisplay(da), gc, background_color);
  XFillRectangle(XtDisplay(da), XtWindow(da), gc, 0, 0, 
		 canvas_width, y_margin);
  XFillRectangle(XtDisplay(da), XtWindow(da), gc, 0, 0, 
		 x_margin, canvas_height);
  XSetForeground(XtDisplay(da), gc, 
		 BlackPixelOfScreen(XtScreen(da)));
}

void
UniEditor::drawAll(void)
{
	this->drawRange(0, lines.size - 1);
}

void
UniEditor::drawRange(int initial, int final)
{	
  Widget da = this->getWidget();

	if(XtIsManaged(da) == FALSE || initial == -1 || 
	   XtIsRealized(da) == FALSE)
		return;

	this->lookAtKeyMap();

	int left_line, left_col, right_line, right_col;
	cMode cursor_place;
	uDirection cursor_dir;
	ueSelectionType st = this->getSelectionType();

	int first_visible;
	int last_visible;
	this->calculateVisible(&first_visible, &last_visible);
	initial = MAX(first_visible, initial);
	final = MIN(final, last_visible);

	if(st == stSELECTION) {
		this->getSelection(&left_line, &left_col,
						   &right_line, &right_col);
	} else if (st == stCURSOR) {    
		this->getCursor(&left_line, &left_col, &cursor_place);
		cursor_dir = lines.line[left_line]->direction(left_col);
		if(cursor_dir == dNONE)
		  cursor_dir = justification_direction;
	}
  
	// determine the visible area
	iBox window;
	if(scroll_flag) {
	  Widget clip_window = XtParent(da),
	         scrolled_window = XtParent(clip_window);
	  window = visibleRectangleAuto(scrolled_window);
	}
	else {
	  window.x = window.y = 0;
	  window.w = canvas_width;
	  window.h = canvas_height;
	}	

	// determine the area of the lines
	iBox range;
	range.x = 0;
	range.w = canvas_width;
	range.y = initial*this->char_height + y_margin;
	if(final == lines.size - 1)
	  range.h = canvas_height;
	else
	  range.h = (final - initial + 1) * this->char_height;
	  
	// intersect them
	i_IntersectBox(window, range, &window);

	if(window.w == 0)
	  return;

	// create pixmap to draw into
	Pixmap pixmap = XCreatePixmap(XtDisplay(da), 
				      RootWindowOfScreen(XtScreen(da)), 
				      window.w, window.h, 
				      DefaultDepthOfScreen(XtScreen(da)));

	// blank out pixmap
	XSetForeground(XtDisplay(da), gc, background_color);
	XFillRectangle(XtDisplay(da), pixmap, gc, 0, 0, window.w, window.h);
	XSetForeground(XtDisplay(da), gc, BlackPixelOfScreen(XtScreen(da)));

	// loop over lines  
	for(int line_index = initial; line_index <= final; line_index++){

		Line *line = lines.line[line_index];
    
		// convert unicode characters to glyphs
		int x, 
		    y = line_index*this->char_height + y_margin - window.y, 
		    pos = 0, 
		    num_base = 0;
		
		if(this->justification_direction == dLEFT)
			x = x_margin - window.x;
		else
			x = canvas_width - x_margin + this->char_width;
    
    
		int cl;
		cl =  line->contextLength();
		while(pos <= cl) {
		  // loop over base context characters
		  num_base = line->context2base(pos);
      
		  // draw cursor, position determined by base 
		  // (ie. spacing) characters
		  if(st == stCURSOR && cursor_mode == cursor_dir &&
			 line_index == left_line && left_col == num_base) {
			if(overstrike)
			  XDrawLine(XtDisplay(da), pixmap, cursor_gc, 
						x, y+ascent, x+this->char_width, y+ascent);
			else{
			  XDrawLine(XtDisplay(da), pixmap, cursor_gc, 
						x, y+ascent, x, y);
			  if(cursor_dir == dLEFT)
				XDrawLine(XtDisplay(da), pixmap, cursor_gc, 
						  x, y+ascent, x+this->char_width/2, y+ascent);
			  else
				XDrawLine(XtDisplay(da), pixmap, cursor_gc, 
						  x, y+ascent, x-this->char_width/2, y+ascent);
			}
		  }
      
		  int draw_inverted_newline;
		  // end processing at length
		  draw_inverted_newline = ((cl == pos) && (st == stSELECTION));
		  draw_inverted_newline = (draw_inverted_newline && 
			((line_index > left_line) 
			 ? 1
			 : (line_index == left_line) && (pos >= left_col)) &&
			((line_index < right_line) 
			 ? 1
			 : (line_index == right_line) && (pos < right_col)));

			// draw a halfwidth inverted box to represent
			// newline if inverted/selected
		  if(draw_inverted_newline)
			XFillRectangle(XtDisplay(da), pixmap, gc, 
						   x, y, this->char_width/2, this->char_height);
      
		  if(pos < cl){
			iUCode glyph = line->getContext(pos);
				
			// update position
			int glyph_base, glyph_ascent, glyph_height, glyph_width = 0;
			if(glyph_base = uniSpacing(glyph))
			  measureUnicode(XtDisplay(da), &glyph, 1, 
							 &glyph_width, &glyph_height, &glyph_ascent);
	
			if(this->justification_direction == dRIGHT)
			  x -= glyph_width;
	
			// invert colors and blank background
			int invert = FALSE;
	
			if(st == stSELECTION) {
			  invert = ((line_index > left_line ||
						 (line_index == left_line && num_base >= left_col)) &&
						(line_index < right_line ||
						 (line_index == right_line && num_base < right_col)));
			}
			if(invert) {
			  XFillRectangle(XtDisplay(da), pixmap, gc, 
							 x, y, glyph_width, this->char_height);
			  XSetForeground(XtDisplay(da), gc, background_color);
			}
	
			// draw glyph(s)
			do{
			  drawUnicodeChar(XtDisplay(da), pixmap, gc, 
							  x, y+glyph_ascent, glyph);	
			  glyph = line->getContext(++pos);
			} while(glyph && !uniSpacing(glyph));
			XSetForeground(XtDisplay(da), gc, 
						   BlackPixelOfScreen(XtScreen(da)));
	
			if(this->justification_direction == dLEFT)
			  x += glyph_width;
		  } else
			pos++;

			// draw cursor, position determined by base 
			// (ie. spacing) characters
		  if(st == stCURSOR && cursor_mode != cursor_dir &&
			 line_index == left_line && left_col == num_base) {
			if(overstrike)
			  XDrawLine(XtDisplay(da), pixmap, cursor_gc, 
						x, y+ascent, x+this->char_width, y+ascent);
			else{
			  XDrawLine(XtDisplay(da), pixmap, cursor_gc, 
						x, y+ascent, x, y);
			  if(cursor_dir == dLEFT)
				XDrawLine(XtDisplay(da),pixmap, cursor_gc, 
						  x+this->char_width/2, y+ascent, x, y+ascent);
			  else	    
				XDrawLine(XtDisplay(da), pixmap, cursor_gc, 
						  x-this->char_width/2, y+ascent, x, y+ascent);
			}
		  }
		}
	}

	XCopyArea(XtDisplay(da), pixmap, XtWindow(da), gc,
		  0, 0,
		  window.w, window.h,
		  window.x, window.y);
	
	XFreePixmap(XtDisplay(da), pixmap);
}

