
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"

#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <Xm/ScrolledW.h>
#include <Xm/ScrollBar.h>

#include "misc/color.h"
#include "misc/XArg.h"

#include "rafport.h"
#include "rafutil.h"
#include "CanvasClient.h"
#include "keyboard/KeyBoard.h"
#include "keyboard/KeyMap.h"

#include "Canvas.h"

#include "misc/scroll.h"
#include "misc/Unicode.h"
#include "bb.h"

#if 0

#define FOLLOW_INCREMENT 30

static char
follow(Canvas *vc)
{
	Window root;
	Window child;
	int root_x, root_y, px, py;
	int4 h_value,v_value;
	unsigned int mask;
#if 0
	int mag_exp = vc->client->getMagExp();
#endif
	/* make a call to xquerypointer and feed your client*/
	XQueryPointer(XtDisplay(vc->canvas_widget), XtWindow(vc->canvas_widget),
				  &root, &child,
				  &root_x, &root_y,
				  &px, &py,
				  &mask);
	
	if (mask&Button1Mask || mask&Button3Mask || mask&Button2Mask) {
		//Widget ww = XtParent(vc->canvas_widget);
		Widget sw = XtParent(vc->canvas_widget);
		iLBox vr;
		
		vr = visibleRectangle(sw);
		getScrollBarValues(sw,&h_value,&v_value);
		px += h_value;
		py += v_value;
		int4 c_x;
		int4 c_y;
		getScrollPositionCenter(sw, &c_x, &c_y);
		if(px < vr.x) {
			c_x-=FOLLOW_INCREMENT;
			px = vr.x;
		} else if(px > (vr.x + vr.w)) {
			c_x+=FOLLOW_INCREMENT;
			px = vr.x + vr.w;
		}
		
		if(py < vr.y){
			c_y-=FOLLOW_INCREMENT;
			py = vr.y;
		} else if(py > (vr.y + vr.h)) {
			c_y+=FOLLOW_INCREMENT;
			py = vr.y + vr.h;
		}
		
		vc->client->moveButton(mask, px, py);	 
		vc->scrollToPosition(sw, c_x, c_y);
		}
	return 0;
}



static void
activateFollow(Canvas *vc)
{
	vc->follow = 1;
	if(vc->follow_procid)
		return;
	vc->follow_procid = 
		XtAppAddWorkProc(vc->app_context,
						 follow, vc);
}

static void
deactivateFollow(Canvas *vc)
{
	vc->follow = 0;
	if(vc->follow_procid == 0)
		return;
	XtRemoveWorkProc(vc->follow_procid);
	vc->follow_procid = 0;
	
}
#endif

void redisplayCB (Widget         w,
				  Canvas *vc,
				  XmDrawingAreaCallbackStruct *call_data)
{
  if(vc->client && vc->client->validP()) {
	// calculate the area and make the adjustments
	vc->client->redisplay(call_data);
  }
}
void resizeCB (Widget         w,
			   Canvas *vc,
			   XmDrawingAreaCallbackStruct *call_data)
{
	if(vc->client && vc->client->validP())
		vc->client->resize(call_data);
}

void pressButtonEH(Widget          w,
				   Canvas      *vc,
				   XEvent         *event)
{
	vc->release_expected = TRUE;
	if(vc->client && vc->client->validP())
		vc->client->pressButton(event);
}

void releaseButtonEH(Widget          w,
					 Canvas      *vc,
					 XEvent         *event)
{
	if(vc->client && vc->client->validP())
	  vc->client->releaseButton(event);
	//deactivateFollow(vc);
}

void moveButtonEH(Widget          w,
				  Canvas      *vc,
				  XEvent         *event)
{
	
  if(!vc->client || !vc->client->validP())
	return;

  int px = event->xmotion.x;
  int py = event->xmotion.y;

#if 0
  Widget sw = XtParent(vc->canvas_widget);
  iLBox vr;
  int4 h_value,v_value;

	vr = visibleRectangle(sw);
	int px = event->xmotion.x + vc->pix_hoffset;
	int py = event->xmotion.y + vc->pix_voffset;
	getScrollBarValues(sw,&h_value,&v_value);
	if((px >= vr.x && px <= (vr.x + vr.w)) &&
	   (py >= vr.y && py <= (vr.y + vr.h))) {
		deactivateFollow(vc);
	} else {
		int4 c_x;
		int4 c_y;
		getScrollPositionCenter(sw, &c_x, &c_y);
		if(px < vr.x) {
			c_x-=FOLLOW_INCREMENT;
			px = vr.x;
		} else if(px > (vr.x + vr.w)) {
			c_x+=FOLLOW_INCREMENT;
			//c_x++;
			px = vr.x + vr.w;
		}
		
		
		if(py < vr.y){
			c_y-=FOLLOW_INCREMENT;
			//c_y--;
			py = vr.y;
		} else if(py > vr.y + vr.h) {
			c_y+=FOLLOW_INCREMENT;
			py = vr.y + vr.h;
		}
		
		vc->client->moveButton(event->xmotion.state, px, py);	 
		vc->scrollToPosition(sw, c_x, c_y);
		
		activateFollow(vc);
		//printf("outside\n");
		
		return;
	}
#endif	
	vc->client->moveButton(event->xmotion.state, px, py);
}

void pressKeyEH(Widget w,
				Canvas *vc,
				XEvent *event)
{
  char key_value;
  iUCode char_value;
  int n;
  KeySym ks;
  int state;
  int alt_on;
  KeyMap *km = vc->getKM();
	
  if(!vc->client || !vc->client->validP())
	return;
	
  state = event->xkey.state;
  alt_on = state & Mod1Mask;

  n = XLookupString(&event->xkey,
					&key_value,
					1,
					&ks,
					0);

  switch(ks){
  case XK_BackSpace:
	vc->client->backspace();
	break;	  
  case XK_Delete:
	vc->client->myDelete();
	break;
  case XK_Insert:
	vc->client->insert();
	break;
  case XK_Left: 
	vc->client->left(state);
	break;
  case XK_Right:
	vc->client->right(state);
	break;
  case XK_Up:
	vc->client->up(state);
	break;
  case XK_Down:
	vc->client->down(state);
	break;
  case XK_End:
	vc->client->end(state);
	break;
  case XK_Home:
	vc->client->home(state);
	break;
  case XK_Return:
	vc->client->myReturn();
	break;
  default:
	if(state & ControlMask)
	  return;
	if(alt_on)
	  char_value = km->lookupControlValue(ks);
	else
	  char_value = km->lookupValue(ks);
	if (char_value)
	  vc->client->pressKey(char_value);
  }
}



Canvas::Canvas(ScrollWindow *input_sw, XtAppContext input_app_context, Widget parent, 
			   CanvasClient *input_client, int width, int height)
	: ScrollClient(input_sw),
      app_context(input_app_context),
	  follow_procid(0),
	  client(input_client),
	  follow(0),
	  redraw_procid(0)
{
	XArg args;
	
	this->redraw_region.w = 0;
	this->canvas_widget = XtCreateWidget("canvas", 
								  xmDrawingAreaWidgetClass, 
								  parent, args.getArgs(), args.getSize());
#if 0
	args.reset();
	args.setArg(XmNworkWindow,this->canvas_widget);
	args.setValues(parent);
#endif

#if 0	
	iBox b;
	b.x = b.y = 0;
	b.w = width;
	b.h = height;
	this->view_port=b;
#endif
	
	XtAddCallback(this->canvas_widget, 
				  XmNexposeCallback, 
				  redisplayCB, 
				  (XtPointer)this); 
	XtAddCallback(this->canvas_widget, 
				  XmNresizeCallback, 
				  resizeCB, 
				  (XtPointer)this); 
	
	XtAddEventHandler(this->canvas_widget, ButtonPressMask, FALSE,
					  (XtEventHandler) pressButtonEH, this);
	XtAddEventHandler(this->canvas_widget, ButtonReleaseMask, FALSE,
					  (XtEventHandler) releaseButtonEH, this);
	XtAddEventHandler(this->canvas_widget, ButtonMotionMask, FALSE,
					  (XtEventHandler) moveButtonEH, this);
	
	XtAddEventHandler(this->canvas_widget, KeyPressMask, FALSE, 
					  (XtEventHandler) pressKeyEH, this); 
	
}
	

Canvas::~Canvas()
{
	XtDestroyWidget(this->canvas_widget);
}

int 
Canvas::managedP() 
{ 
	return XtIsManaged(this->canvas_widget); 
}

void
Canvas::manage()
{
  XtManageChild(this->canvas_widget);
}

void
Canvas::unmanage()
{
	XtUnmanageChild(this->canvas_widget);
}

void
Canvas::setClient(CanvasClient *input_client)
{
	this->client = input_client;
}

iLBox Canvas::getVisibleArea(void)
{
  int w, h;
  iLBox value;
  getWidgetSize(this->canvas_widget, &w, &h);
  value.x = 0;  
  value.y = 0;  
  value.w = w;
  value.h = h;
  
  return value;
}


char 
executeRedrawRegion(Canvas *canvas)
{
  if(canvas->managedP() && canvas->client) {
	if(canvas->client->resizeNecessaryP()){
	  // potentially dangerous loop here
	  canvas->client->conditionallyResize();
	  // this says -- call this workproc again because
	  // we didnt do our work -- we resized
	  // the canvas so that things would fit
	  // and we want to come back here when the xwindows
	  // stuff is done
	  return 0;
	} else {
	  // only draw in the visible area
	  iLBox window_region = canvas->getVisibleArea();
	  iLBox truncated_region;
	  
	  i_IntersectLBox(window_region, canvas->redraw_region, &truncated_region);
	  if (truncated_region.w > 0 && truncated_region.h > 0)
		canvas->client->drawRegionNow(truncated_region);
	}
  }
	
  canvas->redraw_region.w = 0;
  return 1; /* says not to call this workproc again */
}

void
Canvas::addRedrawRegion(iLBox *b)
{
  if (this->redraw_region.w) {
	i_UnionLBox(*b, this->redraw_region, &this->redraw_region);
  } else {
	this->redraw_region = *b;
	this->redraw_procid = 
	  XtAppAddWorkProc(this->app_context,
					   executeRedrawRegion, this);
  }
}

void 
Canvas::flushRedrawRegion()
{
  if (this->redraw_region.w == 0) 
	return;
  XtRemoveWorkProc(this->redraw_procid);
  this->redraw_region.w = 0;
}


void
Canvas::drawRegion(iLBox box)
{
  if (this->follow)
	this->client->drawRegionNow(box);
  else
	this->addRedrawRegion(&box);
}

void
Canvas::setSize(int width, int height)
{
  XArg args;

  args.setArg(XmNwidth, width);
  args.setArg(XmNheight, height);
  args.setValues(canvas_widget);
}
