
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "rafutil.h"

#include "dafs.h"
#include <X11/Intrinsic.h> 
#include <Xm/Xm.h>
#include "smemory.h"
#include "ucs.h"
#include "uString.h"
#include "Unicode.h"
#include "XArg.h"

typedef struct font_mapping {
	iUCode start,end;	
	int offset;
	char *font_name;
	XFontStruct *font;
} font_mapping;

/* this code certainly be more general and complete */
font_mapping font_map[]={
	{0x0000,0x00ff, 0, "etllatin",(XFontStruct *)0},
	{0x0100,0x017f, 0, "rafeuro", (XFontStruct *)0},
	{0x0180,0x01ff, 0, "rafextd", (XFontStruct *)0},
	{0x0386,0x03f5, 128, "mgreek",(XFontStruct *)0},
	{0x0400,0x045f, 0, "etlcyr",(XFontStruct *)0},
	{0x0600,0x0669, 0, "arabic",(XFontStruct *)0},
	{0x2000,0x2043, 0, "mpunc",(XFontStruct *)0},
	{0x2100,0x2138, 0, "mletter",(XFontStruct *)0},
	{0x2446,0x2449,  128, "micr",(XFontStruct *)0},
	{0x2700,0x2713,  0, "check",(XFontStruct *)0},

	{0x3000, 0x303f, 128,  "jis3000.16", (XFontStruct *)0},
	{0x3041, 0x309e, 128,  "jishira", (XFontStruct *)0},
	{0x30a1, 0x30fe, 128,  "jiskata", (XFontStruct *)0},


	{0x3100, 0x317f, 128,  "hang-3100", (XFontStruct *)0},
	{0x3180, 0x31ff, 128,  "hang-3180", (XFontStruct *)0},
	{0x3400, 0x347f, 128,  "hang-3400", (XFontStruct *)0},
	{0x3480, 0x34ff, 128,  "hang-3480", (XFontStruct *)0},
	{0x3500, 0x357f, 128,  "hang-3500", (XFontStruct *)0},
	{0x3580, 0x35ff, 128,  "hang-3580", (XFontStruct *)0},
	{0x3600, 0x367f, 128,  "hang-3600", (XFontStruct *)0},
	{0x3680, 0x36ff, 128,  "hang-3680", (XFontStruct *)0},
	{0x3700, 0x377f, 128,  "hang-3700", (XFontStruct *)0},
	{0x3780, 0x37ff, 128,  "hang-3780", (XFontStruct *)0},
	{0x3800, 0x387f, 128,  "hang-3800", (XFontStruct *)0},
	{0x3880, 0x38ff, 128,  "hang-3880", (XFontStruct *)0},
	{0x3900, 0x397f, 128,  "hang-3400", (XFontStruct *)0},
	{0x3980, 0x39ff, 128,  "hang-3480", (XFontStruct *)0},
	{0x3a00, 0x3a7f, 128,  "hang-3900", (XFontStruct *)0},
	{0x3a80, 0x3aff, 128,  "hang-3980", (XFontStruct *)0},
	{0x3b00, 0x3b7f, 128,  "hang-3a00", (XFontStruct *)0},
	{0x3b80, 0x3bff, 128,  "hang-3a80", (XFontStruct *)0},
	{0x3c00, 0x3c7f, 128,  "hang-3b00", (XFontStruct *)0},
	{0x3c80, 0x3cff, 128,  "hang-3b80", (XFontStruct *)0},
	{0x3d00, 0x3d7f, 128,  "hang-3d00", (XFontStruct *)0},


	{0x4e00, 0x4e7f, 128,  "jis4e00.16", (XFontStruct *)0},
	{0x4e80, 0x4eff, 128,  "jis4e80.16", (XFontStruct *)0},
	{0x4f00, 0x4f7f, 128,  "jis4f00.16", (XFontStruct *)0},
	{0x4f80, 0x4fff, 128,  "jis4f80.16", (XFontStruct *)0},
	{0x5000, 0x507f, 128,  "jis5000.16", (XFontStruct *)0},
	{0x5080, 0x50ff, 128,  "jis5080.16", (XFontStruct *)0},
	{0x5100, 0x517f, 128,  "jis5100.16", (XFontStruct *)0},
	{0x5180, 0x51ff, 128,  "jis5180.16", (XFontStruct *)0},
	{0x5200, 0x527f, 128,  "jis5200.16", (XFontStruct *)0},
	{0x5280, 0x52ff, 128,  "jis5280.16", (XFontStruct *)0},
	{0x5300, 0x537f, 128,  "jis5300.16", (XFontStruct *)0},
	{0x5380, 0x53ff, 128,  "jis5380.16", (XFontStruct *)0},
	{0x5400, 0x547f, 128,  "jis5400.16", (XFontStruct *)0},
	{0x5480, 0x54ff, 128,  "jis5480.16", (XFontStruct *)0},
	{0x5500, 0x557f, 128,  "jis5500.16", (XFontStruct *)0},
	{0x5580, 0x55ff, 128,  "jis5580.16", (XFontStruct *)0},
	{0x5600, 0x567f, 128,  "jis5600.16", (XFontStruct *)0},
	{0x5680, 0x56ff, 128,  "jis5680.16", (XFontStruct *)0},
	{0x5700, 0x577f, 128,  "jis5700.16", (XFontStruct *)0},
	{0x5780, 0x57ff, 128,  "jis5780.16", (XFontStruct *)0},
	{0x5800, 0x587f, 128,  "jis5800.16", (XFontStruct *)0},
	{0x5880, 0x58ff, 128,  "jis5880.16", (XFontStruct *)0},
	{0x5900, 0x597f, 128,  "jis5900.16", (XFontStruct *)0},
	{0x5980, 0x59ff, 128,  "jis5980.16", (XFontStruct *)0},
	{0x5a00, 0x5a7f, 128,  "jis5a00.16", (XFontStruct *)0},
	{0x5a80, 0x5aff, 128,  "jis5a80.16", (XFontStruct *)0},
	{0x5b00, 0x5b7f, 128,  "jis5b00.16", (XFontStruct *)0},
	{0x5b80, 0x5bff, 128,  "jis5b80.16", (XFontStruct *)0},
	{0x5c00, 0x5c7f, 128,  "jis5c00.16", (XFontStruct *)0},
	{0x5c80, 0x5cff, 128,  "jis5c80.16", (XFontStruct *)0},
	{0x5d00, 0x5d7f, 128,  "jis5d00.16", (XFontStruct *)0},
	{0x5d80, 0x5dff, 128,  "jis5d80.16", (XFontStruct *)0},
	{0x5e00, 0x5e7f, 128,  "jis5e00.16", (XFontStruct *)0},
	{0x5e80, 0x5eff, 128,  "jis5e80.16", (XFontStruct *)0},
	{0x5f00, 0x5f7f, 128,  "jis5f00.16", (XFontStruct *)0},
	{0x5f80, 0x5fff, 128,  "jis5f80.16", (XFontStruct *)0},
	{0x6000, 0x607f, 128,  "jis6000.16", (XFontStruct *)0},
	{0x6080, 0x60ff, 128,  "jis6080.16", (XFontStruct *)0},
	{0x6100, 0x617f, 128,  "jis6100.16", (XFontStruct *)0},
	{0x6180, 0x61ff, 128,  "jis6180.16", (XFontStruct *)0},
	{0x6200, 0x627f, 128,  "jis6200.16", (XFontStruct *)0},
	{0x6280, 0x62ff, 128,  "jis6280.16", (XFontStruct *)0},
	{0x6300, 0x637f, 128,  "jis6300.16", (XFontStruct *)0},
	{0x6380, 0x63ff, 128,  "jis6380.16", (XFontStruct *)0},
	{0x6400, 0x647f, 128,  "jis6400.16", (XFontStruct *)0},
	{0x6480, 0x64ff, 128,  "jis6480.16", (XFontStruct *)0},
	{0x6500, 0x657f, 128,  "jis6500.16", (XFontStruct *)0},
	{0x6580, 0x65ff, 128,  "jis6580.16", (XFontStruct *)0},
	{0x6600, 0x667f, 128,  "jis6600.16", (XFontStruct *)0},
	{0x6680, 0x66ff, 128,  "jis6680.16", (XFontStruct *)0},
	{0x6700, 0x677f, 128,  "jis6700.16", (XFontStruct *)0},
	{0x6780, 0x67ff, 128,  "jis6780.16", (XFontStruct *)0},
	{0x6800, 0x687f, 128,  "jis6800.16", (XFontStruct *)0},
	{0x6880, 0x68ff, 128,  "jis6880.16", (XFontStruct *)0},
	{0x6900, 0x697f, 128,  "jis6900.16", (XFontStruct *)0},
	{0x6980, 0x69ff, 128,  "jis6980.16", (XFontStruct *)0},
	{0x6a00, 0x6a7f, 128,  "jis6a00.16", (XFontStruct *)0},
	{0x6a80, 0x6aff, 128,  "jis6a80.16", (XFontStruct *)0},
	{0x6b00, 0x6b7f, 128,  "jis6b00.16", (XFontStruct *)0},
	{0x6b80, 0x6bff, 128,  "jis6b80.16", (XFontStruct *)0},
	{0x6c00, 0x6c7f, 128,  "jis6c00.16", (XFontStruct *)0},
	{0x6c80, 0x6cff, 128,  "jis6c80.16", (XFontStruct *)0},
	{0x6d00, 0x6d7f, 128,  "jis6d00.16", (XFontStruct *)0},
	{0x6d80, 0x6dff, 128,  "jis6d80.16", (XFontStruct *)0},
	{0x6e00, 0x6e7f, 128,  "jis6e00.16", (XFontStruct *)0},
	{0x6e80, 0x6eff, 128,  "jis6e80.16", (XFontStruct *)0},
	{0x6f00, 0x6f7f, 128,  "jis6f00.16", (XFontStruct *)0},
	{0x6f80, 0x6fff, 128,  "jis6f80.16", (XFontStruct *)0},
	{0x7000, 0x707f, 128,  "jis7000.16", (XFontStruct *)0},
	{0x7080, 0x70ff, 128,  "jis7080.16", (XFontStruct *)0},
	{0x7100, 0x717f, 128,  "jis7100.16", (XFontStruct *)0},
	{0x7180, 0x71ff, 128,  "jis7180.16", (XFontStruct *)0},
	{0x7200, 0x727f, 128,  "jis7200.16", (XFontStruct *)0},
	{0x7280, 0x72ff, 128,  "jis7280.16", (XFontStruct *)0},
	{0x7300, 0x737f, 128,  "jis7300.16", (XFontStruct *)0},
	{0x7380, 0x73ff, 128,  "jis7380.16", (XFontStruct *)0},
	{0x7400, 0x747f, 128,  "jis7400.16", (XFontStruct *)0},
	{0x7480, 0x74ff, 128,  "jis7480.16", (XFontStruct *)0},
	{0x7500, 0x757f, 128,  "jis7500.16", (XFontStruct *)0},
	{0x7580, 0x75ff, 128,  "jis7580.16", (XFontStruct *)0},
	{0x7600, 0x767f, 128,  "jis7600.16", (XFontStruct *)0},
	{0x7680, 0x76ff, 128,  "jis7680.16", (XFontStruct *)0},
	{0x7700, 0x777f, 128,  "jis7700.16", (XFontStruct *)0},
	{0x7780, 0x77ff, 128,  "jis7780.16", (XFontStruct *)0},
	{0x7800, 0x787f, 128,  "jis7800.16", (XFontStruct *)0},
	{0x7880, 0x78ff, 128,  "jis7880.16", (XFontStruct *)0},
	{0x7900, 0x797f, 128,  "jis7900.16", (XFontStruct *)0},
	{0x7980, 0x79ff, 128,  "jis7980.16", (XFontStruct *)0},
	{0x7a00, 0x7a7f, 128,  "jis7a00.16", (XFontStruct *)0},
	{0x7a80, 0x7aff, 128,  "jis7a80.16", (XFontStruct *)0},
	{0x7b00, 0x7b7f, 128,  "jis7b00.16", (XFontStruct *)0},
	{0x7b80, 0x7bff, 128,  "jis7b80.16", (XFontStruct *)0},
	{0x7c00, 0x7c7f, 128,  "jis7c00.16", (XFontStruct *)0},
	{0x7c80, 0x7cff, 128,  "jis7c80.16", (XFontStruct *)0},
	{0x7d00, 0x7d7f, 128,  "jis7d00.16", (XFontStruct *)0},
	{0x7d80, 0x7dff, 128,  "jis7d80.16", (XFontStruct *)0},
	{0x7e00, 0x7e7f, 128,  "jis7e00.16", (XFontStruct *)0},
	{0x7e80, 0x7eff, 128,  "jis7e80.16", (XFontStruct *)0},
	{0x7f00, 0x7f7f, 128,  "jis7f00.16", (XFontStruct *)0},
	{0x7f80, 0x7fff, 128,  "jis7f80.16", (XFontStruct *)0},
	{0x8000, 0x807f, 128,  "jis8000.16", (XFontStruct *)0},
	{0x8080, 0x80ff, 128,  "jis8080.16", (XFontStruct *)0},
	{0x8100, 0x817f, 128,  "jis8100.16", (XFontStruct *)0},
	{0x8180, 0x81ff, 128,  "jis8180.16", (XFontStruct *)0},
	{0x8200, 0x827f, 128,  "jis8200.16", (XFontStruct *)0},
	{0x8280, 0x82ff, 128,  "jis8280.16", (XFontStruct *)0},
	{0x8300, 0x837f, 128,  "jis8300.16", (XFontStruct *)0},
	{0x8380, 0x83ff, 128,  "jis8380.16", (XFontStruct *)0},
	{0x8400, 0x847f, 128,  "jis8400.16", (XFontStruct *)0},
	{0x8480, 0x84ff, 128,  "jis8480.16", (XFontStruct *)0},
	{0x8500, 0x857f, 128,  "jis8500.16", (XFontStruct *)0},
	{0x8580, 0x85ff, 128,  "jis8580.16", (XFontStruct *)0},
	{0x8600, 0x867f, 128,  "jis8600.16", (XFontStruct *)0},
	{0x8680, 0x86ff, 128,  "jis8680.16", (XFontStruct *)0},
	{0x8700, 0x877f, 128,  "jis8700.16", (XFontStruct *)0},
	{0x8780, 0x87ff, 128,  "jis8780.16", (XFontStruct *)0},
	{0x8800, 0x887f, 128,  "jis8800.16", (XFontStruct *)0},
	{0x8880, 0x88ff, 128,  "jis8880.16", (XFontStruct *)0},
	{0x8900, 0x897f, 128,  "jis8900.16", (XFontStruct *)0},
	{0x8980, 0x89ff, 128,  "jis8980.16", (XFontStruct *)0},
	{0x8a00, 0x8a7f, 128,  "jis8a00.16", (XFontStruct *)0},
	{0x8a80, 0x8aff, 128,  "jis8a80.16", (XFontStruct *)0},
	{0x8b00, 0x8b7f, 128,  "jis8b00.16", (XFontStruct *)0},
	{0x8b80, 0x8bff, 128,  "jis8b80.16", (XFontStruct *)0},
	{0x8c00, 0x8c7f, 128,  "jis8c00.16", (XFontStruct *)0},
	{0x8c80, 0x8cff, 128,  "jis8c80.16", (XFontStruct *)0},
	{0x8d00, 0x8d7f, 128,  "jis8d00.16", (XFontStruct *)0},
	{0x8d80, 0x8dff, 128,  "jis8d80.16", (XFontStruct *)0},
	{0x8e00, 0x8e7f, 128,  "jis8e00.16", (XFontStruct *)0},
	{0x8e80, 0x8eff, 128,  "jis8e80.16", (XFontStruct *)0},
	{0x8f00, 0x8f7f, 128,  "jis8f00.16", (XFontStruct *)0},
	{0x8f80, 0x8fff, 128,  "jis8f80.16", (XFontStruct *)0},
	{0x9000, 0x907f, 128,  "jis9000.16", (XFontStruct *)0},
	{0x9080, 0x90ff, 128,  "jis9080.16", (XFontStruct *)0},
	{0x9100, 0x917f, 128,  "jis9100.16", (XFontStruct *)0},
	{0x9180, 0x91ff, 128,  "jis9180.16", (XFontStruct *)0},
	{0x9200, 0x927f, 128,  "jis9200.16", (XFontStruct *)0},
	{0x9280, 0x92ff, 128,  "jis9280.16", (XFontStruct *)0},
	{0x9300, 0x937f, 128,  "jis9300.16", (XFontStruct *)0},
	{0x9380, 0x93ff, 128,  "jis9380.16", (XFontStruct *)0},
	{0x9400, 0x947f, 128,  "jis9400.16", (XFontStruct *)0},
	{0x9480, 0x94ff, 128,  "jis9480.16", (XFontStruct *)0},
	{0x9500, 0x957f, 128,  "jis9500.16", (XFontStruct *)0},
	{0x9580, 0x95ff, 128,  "jis9580.16", (XFontStruct *)0},
	{0x9600, 0x967f, 128,  "jis9600.16", (XFontStruct *)0},
	{0x9680, 0x96ff, 128,  "jis9680.16", (XFontStruct *)0},
	{0x9700, 0x977f, 128,  "jis9700.16", (XFontStruct *)0},
	{0x9780, 0x97ff, 128,  "jis9780.16", (XFontStruct *)0},
	{0x9800, 0x987f, 128,  "jis9800.16", (XFontStruct *)0},
	{0x9880, 0x98ff, 128,  "jis9880.16", (XFontStruct *)0},
	{0x9900, 0x997f, 128,  "jis9900.16", (XFontStruct *)0},
	{0x9980, 0x99ff, 128,  "jis9980.16", (XFontStruct *)0},
	{0x9a00, 0x9a7f, 128,  "jis9a00.16", (XFontStruct *)0},
	{0x9a80, 0x9aff, 128,  "jis9a80.16", (XFontStruct *)0},
	{0x9b00, 0x9b7f, 128,  "jis9b00.16", (XFontStruct *)0},
	{0x9b80, 0x9bff, 128,  "jis9b80.16", (XFontStruct *)0},
	{0x9c00, 0x9c7f, 128,  "jis9c00.16", (XFontStruct *)0},
	{0x9c80, 0x9cff, 128,  "jis9c80.16", (XFontStruct *)0},
	{0x9d00, 0x9d7f, 128,  "jis9d00.16", (XFontStruct *)0},
	{0x9d80, 0x9dff, 128,  "jis9d80.16", (XFontStruct *)0},
	{0x9e00, 0x9e7f, 128,  "jis9e00.16", (XFontStruct *)0},
	{0x9e80, 0x9eff, 128,  "jis9e80.16", (XFontStruct *)0},
	{0x9f00, 0x9f7f, 128,  "jis9f00.16", (XFontStruct *)0},
	{0x9f80, 0x9fff, 128,  "jis9f80.16", (XFontStruct *)0},
	{0xfe00, 0xfefc, 0, "c-arabic", (XFontStruct *)0},
//	{0xfee0, 0xff5e, 0, "etlfwa", (XFontStruct *)0},	
	{0,0xffff, 0, 0,0}
};

static XmFontList unicode_list=0;


static void add_font_to_list(int i)
{
	XmFontListEntry entry;
	
	if (font_map[i].font) {
		entry=XmFontListEntryCreate(font_map[i].font_name,
									XmFONT_IS_FONT,
									font_map[i].font);
		if (!entry) return;
		unicode_list=XmFontListAppendEntry(unicode_list,entry);
		XmFontListEntryFree(&entry);
	}
}

unsigned char
mapIndex(int font_index, iUCode c)
{
        const iUCode tab = 9, blank = 32;

        if (c == tab) c = blank;
	return c - font_map[font_index].start + font_map[font_index].offset;
}

static int find_font(Display *display,iUCode c)
{
	int i;
	
	for (i=0; font_map[i].end!=0xffff; i++) {
		if (c<=font_map[i].end) break;
	}
	if (font_map[i].end==0xffff)
		return -1;
	if (c>=font_map[i].start) {
		if (!font_map[i].font) {
			font_map[i].font=XLoadQueryFont(display,font_map[i].font_name);
			add_font_to_list(i);
		}
		
		if (font_map[i].font)
			return i;
		else if (c<=font_map[0].end) { // use default ascii table if can't find font
			font_map[0].font=XLoadQueryFont(display,"9x15");
			add_font_to_list(0);
			if (!font_map[0].font)
				halt("Can't load first font\n");
			return 0;
		} else
			return -1;
	}
	else
		return -1;
}

/*
// return 0 on success
*/
int
unicodeCheckForFonts(Widget app_shell)
{
	XmString xm_string;
	iUCode s[2] = {'a', 0};
	XmFontList list = 0;
	xm_string = makeCompoundString(XtDisplay(app_shell),
								   s, 1, &list);
	
	if(list) {
		XArg args;
		args.reset();
		args.setArg(XmNfontList, list);
		args.setValues(app_shell);
	}

	XmStringFree(xm_string);

	return list == 0;
}


// bogus hard coded limit
// bogus, there is a maximum number of items which can be sent to an XServer,
// should be able to deal with that case

#define MAX_ITEMS 100
#define CANT_DISPLAY 1

/*
// changed so that it returns how far it got into the ptr before
// it had too many font changes
// so you can do something like
//
int progress;
while ((progress = MakeItemList(string, length, items)) < length) {
   use(items);
   length -= progress;
   string += length;
}
//
*/
static int
MakeItemList(Display *display, const iUCode *string, 
			 int length, char *str_buffer,
             XTextItem *items, XFontStruct **fs, 
			 int max_items, int *nitems)
{
	int i,n,start_index,cur_font,font_index;
	
	if (!length) return length;
	
	cur_font=find_font(display,string[0]);
	if (cur_font<0)
		cur_font=find_font(display,CANT_DISPLAY);
	
	n=0;
	start_index=0;
	*nitems=0;
	for (i=0; i<length; i++, string++) {
		font_index=find_font(display,*string);

		if (font_index<0) {
			font_index=find_font(display,CANT_DISPLAY);
			/* should do something here if still can't find font */
			str_buffer[i]=mapIndex(font_index, CANT_DISPLAY);
		} else 
			str_buffer[i]=mapIndex(font_index, *string);
		
		if (font_index!=cur_font) {
			items[*nitems].nchars=n;
			items[*nitems].chars= &str_buffer[start_index];
			items[*nitems].delta=0;
			items[*nitems].font=font_map[cur_font].font->fid;
			fs[*nitems]=font_map[cur_font].font;
			
			(*nitems)++;
			start_index=i;
			n=0;
			
			cur_font=font_index;
			if (*nitems+1==max_items) {
				return i;
			}
		}
		n++;
	}


	/* the trailing case */
	items[*nitems].nchars=n;
	items[*nitems].chars= &str_buffer[start_index];
	items[*nitems].delta=0;
	items[*nitems].font=font_map[cur_font].font->fid;
	fs[*nitems]=font_map[cur_font].font;
	(*nitems)++;

	return i;
}

static void 
measuresub(XFontStruct **fs, int nitems,XTextItem *items, 
		   int *tw, int *th, int *ascent, int *descent)
{
	int i;
	int font_ascent,font_descent,direction;
	XCharStruct overall;
	
	*tw=0;
	*th=0;
	*ascent=0;
	*descent=0;
	
	for (i=0; i<nitems; i++) {
		XTextExtents (fs[i],items[i].chars,items[i].nchars,
					  &direction,&font_ascent,&font_descent,&overall);
		// bogus - ignoring direction
		*tw += overall.width;
		if (font_ascent + font_descent>*th)
			*th=font_ascent+font_descent;
		if (font_ascent>*ascent)
			*ascent=font_ascent;
		if (font_descent>*descent)
			*descent=font_descent;
	}
	
}

/*
// bogus we must add assumption that we will only get one line 
// because we have to make MakeItemList iterative
*/
#define FIXED_BUFFER_MAX 128

extern "C" void
drawUnicodeChar(Display *display, Drawable drawable, GC gc,
				int x, int y, const iUCode uni_char)
{
	// load the appropiate font
	char buffer;
	int font_index = find_font(display, uni_char);
	if(font_index < 0){
		font_index = find_font(display, CANT_DISPLAY);
		buffer = mapIndex(font_index, CANT_DISPLAY);
	}
	else 
		buffer = mapIndex(font_index, uni_char);
				
	// change the font
	XGCValues old_values, new_values;
	XGetGCValues(display, gc, GCFont, &old_values);
	new_values.font = font_map[font_index].font->fid;
	XChangeGC(display, gc, GCFont, &new_values);
	
	// draw the character
	XDrawString(display, drawable, gc, x, y, &buffer, 1);

	// change the font back
//	XChangeGC(display, gc, GCFont, &old_values);
}

static void 
drawUnicodeLine(Display *display, Drawable drawable, GC gc,
				 int x,int y, const iUCode *string, int length, 
				int *h_ptr, int blank)
{
	XTextItem items[MAX_ITEMS];
	XFontStruct *fs[MAX_ITEMS];
	XGCValues values;
	int progress;
        char fixed_buffer[FIXED_BUFFER_MAX];
        int use_fixed_buffer=TRUE;
	
	char *str_buffer=fixed_buffer;
	int nitems,h,w,ascent,descent;
	
	if (!length) return;
	
	if(length > FIXED_BUFFER_MAX) {
		use_fixed_buffer = FALSE;
		str_buffer=(char *)smalloc(length*sizeof(char));
	}
	
	progress = 0;
	while(progress < length){
		string += progress;
		length -= progress;
		progress = MakeItemList(display,string,length,str_buffer,
								items,fs,MAX_ITEMS,&nitems);
	
		// I'm annoyed that I have to measure the text and blank the background
		// I wish there were a XDrawImageText routine...
		measuresub(fs,nitems,items,&w,&h,&ascent,&descent);
	
		if(blank) {
			XGetGCValues(display,gc,GCForeground | GCBackground, &values);
			XSetForeground(display,gc,values.background);
			XSetBackground(display,gc,values.foreground);
			
			XFillRectangle(display,drawable,gc,x,y-h+descent,w,h);
			
			XSetForeground(display,gc,values.foreground);
			XSetBackground(display,gc,values.background);
		}
	
		XDrawText(display,drawable,gc,x,y,items,nitems);
		x += w;
	} 
	if (!use_fixed_buffer) smfree(str_buffer);
	*h_ptr = h;
}

extern "C" void 
drawUnicode(Display *display, Drawable drawable, GC gc,
			int x, int y, const iUCode *string, int length)
{
	/* function doesn't make a copy of the string */
	int i = 0, j = 0;
	int tx, ty;
	tx = x;
	ty = y;
	int h;

	i = 0;
	while(i < length){
		j = 0;
		while(i+j < length && string[i+j] != '\n') j++;

	    drawUnicodeLine(display, drawable, gc, tx, ty, string+i, j, &h, TRUE);

		i  += j+1;
		ty += h;
	}
}

extern "C" void
drawUnicodeText(Display *display, Drawable drawable, GC gc,
			int x, int y, const iUCode *string, int length)
{
	/* function doesn't make a copy of the string */
	int i = 0, j = 0;
	int tx, ty;
	tx = x;
	ty = y;
	int h;

	i = 0;
	while(i < length){
		j = 0;
		while(i+j < length && string[i+j] != '\n') j++;

	    drawUnicodeLine(display, drawable, gc, tx, ty, string+i, j, &h, FALSE);

		i  += j+1;
		ty += h;
	}
}

static void
measureUnicodeLine(Display *display, const iUCode *s, 
			   int l, int *tw, int *th, int *ascent)
{
	if(s == 0 || *s == 0){
		*tw = 0;
		*th = 0;
		*ascent = 0;
	} else {
		int nitems;
	
		XTextItem items[MAX_ITEMS];
		XFontStruct *fs[MAX_ITEMS];
		int descent;
		int progress;
		int one_w;

		char *temp;
		temp=(char *)smalloc(l*sizeof(char));
		if (!temp){
			/* out of memory */
			// bogus -- we need to return an error
			return;
		}
		*tw = 0;

		progress = 0;
		while(progress < l){
			l -= progress;
			s += progress;
			progress = MakeItemList(display,s,l,temp,items,fs,MAX_ITEMS,&nitems);
			measuresub(fs,nitems,items,&one_w,th,ascent,&descent);
			*tw += one_w;
		}

		smfree(temp);
	}
}

extern "C" void
measureUnicode(Display *display, const iUCode *s, 
			   int l, int *tw, int *th, int *ascent)
{
	int i, j;
	int lw;
	int lh;
	*tw = 0;
	*th = 0;
	*ascent = 0;

	i = 0;
	while(i < l){
		j = 0;
		while(i+j < l && s[i+j] != '\n') j++;
	   
		measureUnicodeLine(display, s+i, j, &lw, &lh, ascent);

		i   += j+1;
		*th += lh;
		*tw  = MAX(lw, *tw);
	}
}

extern "C" void 
measureUnicodePos(Display *display, const iUCode *s, 
				  int l, int *tw, int *th, int *ascent)
{
	int i, j;
	int lw;
	int lh;
	*tw = 0;
	*th = 0;
	*ascent = 0;

	i = 0;
	while(i < l){
		j = 0;
		while(i+j < l && s[i+j] != '\n') j++;
	   
		measureUnicodeLine(display, s+i, j, &lw, &lh, ascent);

		i   += j+1;
		*th += lh;
		*tw  = lw;
	}
}

extern "C" int
unicodeDrawableP(Display *display, const iUCode *s)
{
	int drawable = TRUE;
	while(drawable && *s){
		drawable = find_font(display,*s++) >= 0;
	}
	return drawable;
}

/*
// bogus -- no way to make a compound string using 16 bit fonts
// not related to make item list so we dont have font change problem
*/
extern "C" XmString
makeCompoundString(Display *display,const iUCode *string, int length, XmFontList *list)
{
	// bogus - assuming L to R direction for now
	unsigned char *str_buffer;
	int i,n,start_index,cur_font,font_index;
	XmString temp1,temp2;
	
	if (!length)
		return NULL;
	
	str_buffer=(unsigned char *)smalloc(sizeof(unsigned char)*(length+1));
	if (!str_buffer)
		return NULL;
	
	XmString result=NULL;
	
	cur_font=find_font(display,string[0]);
	if (cur_font<0)
		cur_font=find_font(display,CANT_DISPLAY);
	
	n=0;
	start_index=0;
	for (i=0; i<length; i++, string++) {
		font_index=find_font(display,*string);
		int value;

		if (font_index<0) {
			font_index=find_font(display,CANT_DISPLAY);
			/* bogus - should do something here if still can't find font */
			value = mapIndex(font_index, CANT_DISPLAY);
		} else 
			value = mapIndex(font_index, *string);

		str_buffer[n] = value;

		if (font_index!=cur_font) {
			str_buffer[n]='\0';
			
			if (result) {
				temp1=XmStringCreate((char *) str_buffer,font_map[cur_font].font_name);
				temp2=XmStringConcat(result,temp1);
				XmStringFree(result);
				XmStringFree(temp1);
				result=temp2;
			} else {
				result=XmStringCreate((char *)str_buffer,font_map[cur_font].font_name);
			}
			
			start_index=i;
			n=0;
			str_buffer[n++] = value;
			cur_font=font_index;
		} else 
			n++;
	}

	/* the trailing case */
	str_buffer[n]='\0';
			
	if (result) {
		temp1=XmStringCreate((char *) str_buffer,font_map[cur_font].font_name);
		temp2=XmStringConcat(result,temp1);
		XmStringFree(result);
		XmStringFree(temp1);
		result=temp2;
	} else {
		result=XmStringCreate((char *)str_buffer,font_map[cur_font].font_name);
	}

	smfree(str_buffer);
	*list=unicode_list;
	return result;
}

/* routines to parse unicode, handling ligatures, 
   composed character subsitution, and contextual forms */

#define NONE           0
#define RIGHT          1
#define LEFT           2
#define DUAL           3
#define TRANSPARENT    4
#define RIGHT_CAUSING  8
#define LEFT_CAUSING  16

// final, initial, medial
static iUCode forms[224][3] = {
	{ 0, 0, 0 },                   // 0x0620
	{ 0, 0, 0 },
	{ 0x82, 0, 0 },
	{ 0x84, 0, 0 },
	{ 0x86, 0, 0 },
	{ 0x88, 0, 0 },	
	{ 0x8C, 0x8A, 0x8B },
	{ 0x8E, 0, 0 },

	{ 0x92, 0x90, 0x91 },          // 0x0628 BAA
	{ 0x94, 0, 0 },
	{ 0x98, 0x96, 0x97 },
	{ 0x9C, 0x9A, 0x9B },	
	{ 0xA0, 0x9E, 0x9F },
	{ 0xA4, 0xA2, 0xA3 },
	{ 0xA8, 0xA6, 0xA7 },
	{ 0xAA, 0, 0 },

	{ 0xAC, 0, 0 },                // 0x0630 THAL
	{ 0xAE, 0, 0 },	
	{ 0xB0, 0, 0 },
	{ 0xB4, 0xB2, 0xB3 },
	{ 0xB8, 0xB6, 0xB7 },
	{ 0xBC, 0xBA, 0xBB },
	{ 0xC0, 0xBE, 0xBF },
	{ 0xC4, 0xC2, 0xC3 },	

	{ 0xC8, 0xC6, 0xC7 },          // 0x0638 DHAH
	{ 0xCC, 0xCA, 0xCB },
	{ 0xD0, 0xCE, 0xCF },
	{ 0, 0, 0 },
	{ 0, 0, 0 },
	{ 0, 0, 0 },
	{ 0, 0, 0 },
	{ 0, 0, 0 },

	{ 0, 0, 0 },                   // 0x0640 TATWEEL
	{ 0xD4, 0xD2, 0xD3 },
	{ 0xD8, 0xD6, 0xD7 },
	{ 0xDC, 0xDA, 0xDB },
	{ 0xE0, 0xDE, 0xDF },
	{ 0xE4, 0xE2, 0xE3 },
	{ 0xE8, 0xE6, 0xE7 },
	{ 0xEC, 0xEA, 0xEB },

	{ 0xEE, 0, 0 },	               // 0x0648 WAW
	{ 0xF0, 0, 0 },	
	{ 0xF4, 0xF2, 0xF3 }
};

static unsigned char linking[224];

static void 
formTables(void){
	memset(linking, 0, 224);

	for(int i = 0x0; i < 0x2A; i++){
		if(forms[i][1])
			linking[i] = RIGHT_CAUSING | LEFT_CAUSING | RIGHT | LEFT;
		else if(forms[i][0])
			linking[i] = LEFT_CAUSING | RIGHT;
	}
			
	linking[0x20] = RIGHT_CAUSING | LEFT_CAUSING;
	
	for(i = 0x2B; i <= 0x35; i++)
		linking[i] = TRANSPARENT;
}

uString
uniContext(const uString &line)
{
	unsigned i, length;
	uString glyphs;

	int cur_linking, next_linking, 
		prev_linking = NONE;

	static int initialized = FALSE;
	if(!initialized)
		formTables();

	i = 0;
	length = line.length();
	while(i < length){ 		
		iUCode glyph = line[i];
		uString marks;

		if((glyph & 0xFF00) == 0x0600){
			// handle arabic contextual forms		
			// determine next base character
			unsigned j = 1;
			while(line[i+j] && (linking[line[i+j]-0x0620] & TRANSPARENT)){
				/* handle the cases where a font exists to represent
				   a base character and a transparent mark */
				iUCode mark = line[i+j];

				if(glyph == 0x0627 && mark == 0x0653)       // maddah on alef
					glyph = 0x0622;
				else if(glyph == 0x0627 && mark == 0x0654)  // hamzah on alef
					glyph = 0x0623;
				else if(glyph == 0x0627 && mark == 0x0655)  // hamzah below alef
					glyph = 0x0625;
				else if(glyph == 0x0648 && mark == 0x0654)  // hamzah on waw
					glyph = 0x0624;
				else if(glyph == 0x064A && mark == 0x0654)  // hamzah on ya
					glyph = 0x0626;	
				else 
					marks += mark;

				j++;
			}
			
			// determine form
			if(line[i+j])
				next_linking = linking[line[i+j]-0x0620];
			else
				next_linking = NONE;
			cur_linking  = linking[glyph-0x0620];
			
			int form = NONE;
			if(cur_linking & RIGHT && prev_linking & RIGHT_CAUSING)
				form |= RIGHT;
			if(cur_linking & LEFT && next_linking & LEFT_CAUSING)
				form |= LEFT;
			prev_linking = cur_linking;			
			
			// handle ligatures
			if(glyph == 0x0644 && line[i+j] == 0x0627){
				if(form == DUAL)
					glyph = (iUCode) 0xFEFC;
				else
					glyph = (iUCode) 0xFEFB;
				
				j++;
				while(linking[line[i+j]-0x0620] & TRANSPARENT) j++;
				prev_linking = form = NONE;
			}			

			if(form) {
				uString temp((forms[glyph-0x0620][form-1] + 0xFE00));
				glyphs += temp;
			} else
				glyphs += glyph;
			glyphs +=  marks;

			i += j;
		}
		else if(glyph >= 0x0300 && glyph < 0x0370){
			// general diacrital marks			
			switch(glyph){
			case 0x0300: // accent grave
				switch(line[i-1]){
				case 0x0061: // a
					--glyphs;
					glyphs += 0x00E0;
					break;
				case 0x0065: // e
					--glyphs;
					glyphs += 0x00E8;
					break;
				case 0x0041: // A
					--glyphs;
					glyphs += 0x00C0;
					break;
				case 0x0045: // E
					--glyphs;
					glyphs += 0x00C8;
					break;
				default:
					glyphs += glyph;
					break;
				}
				break;
			case 0x0301:
				switch(line[i-1]){
				case 0x0061: // a
					--glyphs;
					glyphs += 0x00E1;
					break;
				case 0x0065: // e
					--glyphs;
					glyphs += 0x00E9;
					break;
				case 0x0041: // A
					--glyphs;
					glyphs += 0x00C1;
					break;
				case 0x0045: // E
					--glyphs;
					glyphs += 0x00C9;
					break;
				default:
					glyphs += glyph;
					break;
				}
				break;
			default:
				glyphs += glyph;
			}
			i++;
		}	
		else{
			glyphs += glyph;
			prev_linking = NONE;
			i++;
		}		
	}


	return glyphs;
}	

int 
uniSpacing(iUCode character)
{
	// Arabic dicrital marks
	if(character >= 0x064B && character <= 0x0655)
		return 0;

	if(character == 0xFEFB || character == 0xFEFC)
		return 2;

	// diacritical marks
	if(character >= 0x0300 && character < 0x0370)
		return 0;

	return 1;
}

uDirection
uniOrder(const iUCode glyph)
{
	// Arabic, Hewbrew, and compatability Arabic
	if((glyph >= 0x0590 && glyph < 0x0700) ||
	   (glyph >= 0xFE70 && glyph < 0xFEFD))
		return dRIGHT;

	// space, punctuation, etc.
	if((glyph >= 0x0020 && glyph < 0x0030) ||
	   (glyph >= 0x003A && glyph < 0x0041) ||
	   (glyph >= 0x005B && glyph < 0x0061) ||
	   (glyph >= 0x007B && glyph < 0x007F))
		return dNONE;

	return dLEFT;
}

static void
uniReverse(iUCode *ucs, int n)
{
    int i, j;
    iUCode uc;
    for (i = 0, j = n - 1; i < j; i++, j--) {
		uc = ucs[i];
		ucs[i] = ucs[j];
		ucs[j] = uc;
    }
}

static void
uniStrip(iUCode *contextual, iUCode *indices, int &start, int &stop, int strip = TRUE)
{
  // remove leading and trailing neutrals
  if(strip) {
    for(; !uniOrder(contextual[start]) && uniSpacing(contextual[start]); start++);
    for(; !uniOrder(contextual[stop-1]) && uniSpacing(contextual[stop-1]); stop--);
  }
  
  // reverse all nonspacing marks so that when reversed again follow base character
  for(int j = start+1; j < stop; j++) {
    if(!uniSpacing(contextual[j])) {
      int k = 1;
      while(j+k < stop && !uniSpacing(contextual[j+k]))
	k++;
      uniReverse(contextual+j-1, k+1);
      if(indices)
	uniReverse(indices+j-1, k+1);
      j += k; // j gets incremented by one automatically
    }
  }  
}

/* parses and modifies a unicode string for display using L-R routines,
   ignores unicode directional tokens */
iUCode *
uniDirection(iUCode *contextual, iUCode **ip)
{
  int i, length = ucslen(contextual);
  iUCode *indices = NULL;
  if(ip){
    // generate a string which indicates which base character is represented by which glyph
    indices = (iUCode *) smalloc((length+2)*sizeof(iUCode));
    int base = -1;
    for(i = 0; i <= length; i++) {
      base += uniSpacing(contextual[i]);
      indices[i] = base;
    }
    indices[i] = 0;
  }
  
  if(length){    
    int prev_index = 0;
    uDirection dir, prev_dir = uniOrder(contextual[0]);
    
    i = 0;
    while(i < length && !prev_dir)
      prev_dir = uniOrder(contextual[i++]);
    if(i == length)
      prev_dir = dLEFT;

    for(i = 0; i < length; i++){
      dir = uniOrder(contextual[i]);
      if(dir && dir != prev_dir) {
	if(prev_dir == dRIGHT) {
	  uniStrip(contextual, indices, prev_index, i);

	  // reverse right to left text
	  uniReverse(contextual+prev_index, i - prev_index);
	  if(indices)
	    uniReverse(indices+prev_index, i - prev_index);
	}
	
	prev_dir = dir;
	prev_index = i;
      }
    }

    if(dir == dRIGHT || (dir == dNONE && prev_dir == dRIGHT)) {
      if(prev_index > 0)
	uniStrip(contextual, indices, prev_index, i);      
      else
	uniStrip(contextual, indices, prev_index, i, FALSE);

      uniReverse(contextual+prev_index, i - prev_index);
      if(indices)
	uniReverse(indices+prev_index, i - prev_index);
    }
  }

  if(ip) 
    *ip = indices;
  
  return contextual;
}
