
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
// check for errors on mallocs
#include <stdlib.h>
#include <X11/Intrinsic.h>

#include "smemory.h"

#include "XArg.h"

#define DEFAULT_SIZE 20

XArg::XArg()
{
	this->size = 0;
	this->max_size = DEFAULT_SIZE;
	this->array = (Arg *) smalloc(sizeof(Arg) * this->max_size);
}

XArg::~XArg()
{
	smfree(this->array);
}

Arg *
XArg::getArgs()
{
	return this->array;
}

int
XArg::getSize()
{
	return this->size;
}

void
XArg::reset()
{
	this->size = 0;
}

void
XArg::setArg(char *label, void *value)
{
	if(this->size == this->max_size)
		this->doubleSize();
	XtSetArg(this->array[this->size], label, value), this->size++;
}

void
XArg::setArg(char *label, int value)
{
	if(this->size == this->max_size)
		this->doubleSize();
	XtSetArg(this->array[this->size], label, value),   this->size++;
}

void
XArg::doubleSize()
{
	Arg *old_array;
	int old_size = this->size;
	int i;
	
	this->max_size *= 2;
	old_array = this->array;
	this->size = 0;
	this->array = (Arg *) smalloc(sizeof(Arg) * this->max_size);
	
	for(i = 0; i < old_size; i++)
		this->setArg(old_array[i].name, (void *) old_array[i].value);
}


void
XArg::setValues(Widget w)
{
	XtSetValues(w, this->getArgs(), this->getSize());
}

void
XArg::getValues(Widget w)
{
	XtGetValues(w, this->getArgs(), this->getSize());
}
