
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"

#include "rafutil.h" // for MIN and MAX

#include "dafstype.h"

/*
// bb stuff
*/

int
bbNullP(const iBox& box) 
{
	return box.w == 0;
}

void 
polyScaleUp(iOrthogonPtr poly, int mag_exp)
{
	int num, i;

	num=poly->n;
	if (num==2)
		num--;

	if (mag_exp < 0){
		mag_exp=-mag_exp;
		for(i=0; i<num; i++){
			poly->point[i].x >>=mag_exp;
			poly->point[i].y >>=mag_exp;
		}
	} else
		for(i=0; i<num; i++){
			poly->point[i].x <<=mag_exp;
			poly->point[i].y <<=mag_exp;
		}
}	

void
bbScaleUp(iBox& cbb, int mag_exp)
{
	if (mag_exp < 0) {
		mag_exp = -mag_exp;
		cbb.x >>= mag_exp;
		cbb.y >>= mag_exp;
		cbb.w >>= mag_exp;
		cbb.h >>= mag_exp;
	} else {
		cbb.x <<= mag_exp;
		cbb.y <<= mag_exp;
		cbb.w <<= mag_exp;
		cbb.h <<= mag_exp;
	}
}

void
bbScaleDown(iBox& cbb, int mag_exp)
{
	if (mag_exp < 0) {
		mag_exp = -mag_exp;
		cbb.x <<= mag_exp;
		cbb.y <<= mag_exp;
		cbb.w <<= mag_exp;
		cbb.h <<= mag_exp;
	} else {
		cbb.x >>= mag_exp;
		cbb.y >>= mag_exp;
		cbb.w >>= mag_exp;
		cbb.h >>= mag_exp;
	}
}

iLBox promoteBox(iBox box)
{
   iLBox lbox;
   lbox.x = box.x;
   lbox.y = box.y;
   lbox.w = box.w;
   lbox.h = box.h;
	
	return lbox;
}   
iBox demoteBox(iLBox lbox)
{
   iBox box;
   box.x = lbox.x;
   box.y = lbox.y;
   box.w = lbox.w;
   box.h = lbox.h;
	
	return box;
}   


int
overlapBoxes(const iBox& left, const iBox& right)
{
	return ( (MIN(left.x + left.w, right.x + right.w) - MAX(left.x, right.x) > 0) && 
	         (MIN(left.y + left.h, right.y + right.h) - MAX(left.y, right.y) > 0) );
}

int
equalBoxes(const iBox& left, const iBox& right)
{
	return((left.x == right.x) &&
		   (left.y == right.y) &&
		   (left.w == right.w) &&
		   (left.h == right.h));
}
int
equalLBoxes(const iLBox& left, const iLBox& right)
{
	return((left.x == right.x) &&
		   (left.y == right.y) &&
		   (left.w == right.w) &&
		   (left.h == right.h));
}
void setLBox(iLBox& rbox, int x, int y, int w, int h)
{
	rbox.x = x;
	rbox.y = y;
	rbox.w = w;
	rbox.h = h;
}

void
i_UnionLBox(iLBox b1,iLBox b2, iLBox *ret)
{
	iLBox ob;	
	
	if (!b1.w){
		*ret=b2;
		 return;
	}
	if (!b2.w){
		*ret=b1;
		return;
	}
	ob = b1;
	ob.w += ob.x;
	ob.h += ob.y;
	if (b2.x < ob.x)
		ob.x = b2.x;
	if (b2.y < ob.y)
		ob.y = b2.y;
	if (b2.x + b2.w > ob.w)
		ob.w = b2.x + b2.w;
	if (b2.y + b2.h > ob.h)
		ob.h = b2.y + b2.h;
	ob.w -= ob.x;
	ob.h -= ob.y;
	*ret=ob;
}
void
i_IntersectLBox(iLBox a,iLBox b,iLBox *c)
{	iLBox temp;

	if (a.x < b.x) temp.x = b.x;
	else temp.x = a.x;
	if (a.y < b.y) temp.y = b.y;
	else temp.y = a.y;
	if (a.x + a.w > b.x + b.w) temp.w = b.x + b.w;
	else temp.w = a.x + a.w;
	temp.w -= temp.x;
	if (temp.w < 0) {
		temp.w = 0;
		*c=temp;
		return;
	}
	if (a.y + a.h > b.y + b.h) temp.h = b.y + b.h;
	else temp.h = a.y + a.h;
	temp.h -= temp.y;
	if (temp.h < 0) temp.w = 0;
	*c=temp;
}
