
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <stdlib.h>
#include <stdio.h>

#include "rafutil.h"
#include "color.h"
#include "svpmap.h"

/*
// bogus need better error checking
*/

extern stringVoidPtrMap *color_name_map;
extern int color_depth;

extern "C" Pixel
getColorByName(Widget w,
			   const char  *name)
{
	Display *display  = XtDisplay(w);
	int      screen  = DefaultScreen(display);
	Colormap map = DefaultColormap(display, screen);
	XColor   color, ignore;
	Pixel p;

	if(color_name_map == 0) {
		halt("call miscInit.\n");
	}
	
	if(svmapContainsP(color_name_map, name) == FALSE) {
		if(color_depth == 1){
			// bogus -- should be more robust about this string
			if(strcmp(name, "white") == 0 || strcmp(name, "gray88") == 0)
				p =  WhitePixel(display, screen);
			else 
				p =  BlackPixel(display, screen);
		} else {
			int ok;
			ok = XAllocNamedColor(display, map, name, &color, &ignore);

			if(ok == FALSE) {
				fprintf(stderr, "Warning: Couldn't allocate color %s\n", name);
				p =  BlackPixel(display, screen);
			} else
				p = color.pixel;
		}

		svmapSet(color_name_map, name, (VoidPtr) p);
	}
	
	return (Pixel) svmapGet(color_name_map, name);
}
