
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*
// originally modelled on code by Dan Heller
// a notice follows
*/

/* Written by Dan Heller.  Copyright 1991, O'Reilly && Associates.
 * This program is freely distributable without licensing fees and
 * is provided without guarantee or warrantee expressed or implied.
 * This program is -not- in the public domain.
 */

/*
// wraps up methods to send messages warnings and prompts
*/

#include <Xm/DialogS.h>
#include <Xm/SelectioB.h>
#include <Xm/RowColumn.h>
#include <Xm/MessageB.h>
#include <Xm/PushB.h>

#include "dafs.h"
#include "ucs.h"
#include "Unicode.h"

#define YES 1
#define NO  2
#define THIRD 3

/* Generalize the question/answer process by creating a data structure
 * that has the necessary labels, questions and everything needed to
 * execute a command.
 */
typedef struct {
    char *label;    /* label for pushbutton used to invoke cmd */
    char *question; /* question for dialog box to confirm cmd */
    char *yes;      /* what the "ok" button says */
    char *no;       /* what the "cancel" button says */
    int   dflt;     /* which should be the default answer */
    char *cmd;      /* actual command to execute (using system()) */
} QandA;


static XtAppContext static_app_context = 0;
static Widget       static_app_shell = 0;


void 
uiInit(XtAppContext app_context, Widget app_shell)
{
	static_app_context = app_context;
	static_app_shell   = app_shell;
}


/*
 * uiAskUser2() -- a generalized routine that asks the user a question
 * and returns a response.  Parameters are: the question, the labels
 * for the "Yes" and "No" buttons, and the default selection to use.
 */

int
uiAskUser2(char *question,
		   char *ans1, 
		   char *ans2,
		   int default_ans)
{
    static Widget dialog; /* static to avoid multiple creation */
    XmString text, yes, no;
    extern void response();
    static int answer = 0;
    answer = 0;
	
    if (!dialog) {
        XmString title;
		Arg args[2];
		int n = 0;
		
		n = 0;
		title = XmStringCreateLocalized("Question");
		XtSetArg(args[n], XmNdialogTitle, title), n++;
		
        dialog = XmCreateQuestionDialog(static_app_shell, "dialog", args, n);
		XmStringFree(title);
		
        XtVaSetValues(dialog,
					  XmNdialogStyle,        XmDIALOG_FULL_APPLICATION_MODAL,
					  
					  NULL);
        XtSetSensitive(
            XmMessageBoxGetChild(dialog, XmDIALOG_HELP_BUTTON),
            False);
        XtAddCallback(dialog, XmNokCallback, response, &answer);
        XtAddCallback(dialog, XmNcancelCallback, response, &answer);
    }
    text = XmStringCreateLocalized(question);
    yes = XmStringCreateLocalized(ans1);
    no = XmStringCreateLocalized(ans2);
    XtVaSetValues(dialog,
				  XmNmessageString,      text,
				  XmNokLabelString,      yes,
				  XmNcancelLabelString,  no,
				  XmNdefaultButtonType,  default_ans == YES?
				  XmDIALOG_OK_BUTTON : XmDIALOG_CANCEL_BUTTON,
				  NULL);
    XmStringFree(text);
    XmStringFree(yes);
    XmStringFree(no);
    XtManageChild(dialog);
	
    while (answer == 0) {
        XtAppProcessEvent(static_app_context, XtIMAll);
        XSync(XtDisplay(dialog), 0);
    }
	
    XtUnmanageChild(dialog);
    XSync(XtDisplay(dialog), 0);
    XmUpdateDisplay(dialog);
	
    return answer;
}

/*
 * AskUser3() -- a generalized routine that asks the user a question
 * and returns a response.  Parameters are: the question, the labels
 * for the "Yes" and "No"  and "Third" buttons, 
 * and the default selection to use.
 */
int
uiAskUser3(char *question,
		   char *ans1, 
		   char *ans2,
		   char *ans3,
		   int default_ans)
{
    static Widget dialog; /* static to avoid multiple creation */
    XmString text, yes, no, third;
    extern void response();
    static int answer = 0;
    answer = 0;
	
    if (!dialog) {
        XmString title;
		Arg args[2];
		int n = 0;
		
		n = 0;
		title = XmStringCreateLocalized("Question");
		XtSetArg(args[n], XmNdialogTitle, title), n++;
		
        dialog = XmCreateQuestionDialog(static_app_shell, "dialog", args, n);
		XmStringFree(title);
        XtVaSetValues(dialog,
					  XmNdialogStyle,        XmDIALOG_FULL_APPLICATION_MODAL,
					  NULL);
        XtAddCallback(dialog, XmNokCallback, response, &answer);
        XtAddCallback(dialog, XmNcancelCallback, response, &answer);
        XtAddCallback(dialog, XmNhelpCallback, response, &answer);
    }
    text = XmStringCreateLocalized(question);
    yes = XmStringCreateLocalized(ans1);
    no = XmStringCreateLocalized(ans2);
    third = XmStringCreateLocalized(ans3);


    XtVaSetValues(dialog,
				  XmNmessageString,      text,
				  XmNokLabelString,      yes,
				  XmNcancelLabelString,  no,
				  XmNhelpLabelString,  third,
				  XmNdefaultButtonType,  default_ans == YES?
				  XmDIALOG_OK_BUTTON : XmDIALOG_CANCEL_BUTTON,
				  NULL);
    XmStringFree(text);
    XmStringFree(yes);
    XmStringFree(no);
    XmStringFree(third);
    XtManageChild(dialog);
	
    while (answer == 0) {
        XtAppProcessEvent(static_app_context, XtIMAll);
        XSync(XtDisplay(dialog), 0);
    }
	
    XtUnmanageChild(dialog);
    XSync(XtDisplay(dialog), 0);
    XmUpdateDisplay(dialog);
	
    return answer;
}

void
uiTellUser(char *statement,
		   char *confirm)
{
    static Widget dialog; /* static to avoid multiple creation */
    XmString text, yes;
    extern void response();
    static int answer = 0;
    answer = 0;
	
    if (!dialog) {
        XmString title;
		Arg args[2];
		int n = 0;
		
		n = 0;
		title = XmStringCreateLocalized("Information");
		XtSetArg(args[n], XmNdialogTitle, title), n++;
		
		dialog = XmCreateMessageDialog(static_app_shell, "dialog", args, n);
		
		XmStringFree(title);
		
		XtVaSetValues(dialog,
					  XmNdialogStyle,        XmDIALOG_FULL_APPLICATION_MODAL,
					  NULL);
		XtUnmanageChild(XmMessageBoxGetChild(dialog,XmDIALOG_CANCEL_BUTTON));
		XtUnmanageChild(XmMessageBoxGetChild(dialog,XmDIALOG_HELP_BUTTON));
		XtAddCallback(dialog, XmNokCallback, response, &answer);
		/*XtAddCallback(dialog, XmNcancelCallback, response, &answer);*/
    }
    text = XmStringCreateLocalized(statement);
    yes = XmStringCreateLocalized(confirm);
	
    XtVaSetValues(dialog,
				  XmNmessageString,      text,
				  XmNokLabelString,      yes,
				  XmNdefaultButtonType,  XmDIALOG_OK_BUTTON,
				  NULL);
    XmStringFree(text);
    XmStringFree(yes);
    XtManageChild(dialog);
	
    while (answer == 0) {
		XtAppProcessEvent(static_app_context, XtIMAll);
		XSync(XtDisplay(dialog), 0);
    }
	
    XtUnmanageChild(dialog);
    XSync(XtDisplay(dialog), 0);
    XmUpdateDisplay(dialog);
}

XmString xs_str_array_to_xmstr(char *cs[])
{
	XmString xmstr;
	int i=0;
	
	xmstr=NULL;
	
	while (*cs) {
		if (i!=0)
			xmstr=XmStringConcat(xmstr,XmStringSeparatorCreate());
		xmstr=XmStringConcat(xmstr,XmStringCreate(*cs,XmSTRING_DEFAULT_CHARSET));
		cs++;
		i++;
	}
	return xmstr;
}

void
uiTellUserN(char **statements,
			char *confirm)
{
    static Widget dialog; /* static to avoid multiple creation */
    XmString text, yes;
    extern void response();
    static int answer = 0;
    answer = 0;
	
    text = xs_str_array_to_xmstr(statements);
    if (!text) return;
	
    if (!dialog) {
        XmString title;
		Arg args[2];
		int n = 0;
		
		n = 0;
		title = XmStringCreateLocalized("Information");
		XtSetArg(args[n], XmNdialogTitle, title), n++;
		
		dialog = XmCreateMessageDialog(static_app_shell, "dialog", args, n);
		
		XmStringFree(title);
		
		XtVaSetValues(dialog,
					  XmNdialogStyle,        XmDIALOG_FULL_APPLICATION_MODAL,
					  NULL);
		XtUnmanageChild(XmMessageBoxGetChild(dialog,XmDIALOG_CANCEL_BUTTON));
		XtUnmanageChild(XmMessageBoxGetChild(dialog,XmDIALOG_HELP_BUTTON));
		XtAddCallback(dialog, XmNokCallback, response, &answer);
    }
    yes = XmStringCreateLocalized(confirm);
	
    XtVaSetValues(dialog,
				  XmNmessageString,      text,
				  XmNokLabelString,      yes,
				  XmNdefaultButtonType,  XmDIALOG_OK_BUTTON,
				  NULL);
    XmStringFree(text);
    XmStringFree(yes);
    XtManageChild(dialog);
	
    while (answer == 0) {
		XtAppProcessEvent(static_app_context, XtIMAll);
		XSync(XtDisplay(dialog), 0);
    }
	
    XtUnmanageChild(dialog);
    XSync(XtDisplay(dialog), 0);
    XmUpdateDisplay(dialog);
}

void
uiWarnUser(char *statement,
		   char *confirm)
{
    static Widget dialog; /* static to avoid multiple creation */
    XmString text, yes;
    extern void response();
    static int answer = 0;
    answer = 0;
	
    if (!dialog) {
		
        XmString title;
		Arg args[2];
		int n = 0;
		
		n = 0;
		title = XmStringCreateLocalized("Warning");
		XtSetArg(args[n], XmNdialogTitle, title), n++;
		
		dialog = XmCreateErrorDialog(static_app_shell, "dialog", args, n);
		
		XmStringFree(title);
		
		XtVaSetValues(dialog,
					  XmNdialogStyle,        XmDIALOG_FULL_APPLICATION_MODAL,
					  NULL);
		XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));
		XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_CANCEL_BUTTON));
		XtAddCallback(dialog, XmNokCallback, response, &answer);
    }
    text = XmStringCreateLocalized(statement);
    yes = XmStringCreateLocalized(confirm);
	
    XtVaSetValues(dialog,
				  XmNmessageString,      text,
				  XmNokLabelString,      yes,
				  XmNdefaultButtonType,  XmDIALOG_OK_BUTTON,
				  NULL);
    XmStringFree(text);
    XmStringFree(yes);
    XtManageChild(dialog);
	
    while (answer == 0) {
		XtAppProcessEvent(static_app_context, XtIMAll);
		XSync(XtDisplay(dialog), 0);
    }
	
    XtUnmanageChild(dialog);
    XSync(XtDisplay(dialog), 0);
    XmUpdateDisplay(dialog);
}

void
uiWarnUserUni(iUCode *statement,
			  char *confirm)
{
    static Widget dialog; /* static to avoid multiple creation */
    XmString text, yes;
    extern void response();
    static int answer = 0;
	XmFontList font_list;

    answer = 0;
	
    if (!dialog) {
		
        XmString title;
		Arg args[2];
		int n = 0;
		
		n = 0;
		title = XmStringCreateLocalized("Warning");
		XtSetArg(args[n], XmNdialogTitle, title), n++;
		
		dialog = XmCreateErrorDialog(static_app_shell, "dialog", args, n);
		
		XmStringFree(title);
		
		XtVaSetValues(dialog,
					  XmNdialogStyle,        XmDIALOG_FULL_APPLICATION_MODAL,
					  NULL);
		XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));
		XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_CANCEL_BUTTON));
		XtAddCallback(dialog, XmNokCallback, response, &answer);
    }
    text = makeCompoundString(XtDisplay(dialog), statement, ucslen(statement), &font_list);
    yes = XmStringCreateLocalized(confirm);
	
	if(font_list){
		XtVaSetValues(dialog,
					  XmNmessageString,      text,
					  XmNokLabelString,      yes,
					  XmNdefaultButtonType,  XmDIALOG_OK_BUTTON,
					  NULL);
		XtVaSetValues(XmMessageBoxGetChild(dialog, XmDIALOG_MESSAGE_LABEL),
					  XmNfontList, font_list,
					  NULL);
	} else {
		XtVaSetValues(dialog,
					  XmNmessageString,      text,
					  XmNokLabelString,      yes,
					  XmNdefaultButtonType,  XmDIALOG_OK_BUTTON,
					  NULL);
	}
    XmStringFree(text);
    XmStringFree(yes);
    XtManageChild(dialog);
	
    while (answer == 0) {
		XtAppProcessEvent(static_app_context, XtIMAll);
		XSync(XtDisplay(dialog), 0);
    }
	
    XtUnmanageChild(dialog);
    XSync(XtDisplay(dialog), 0);
    XmUpdateDisplay(dialog);
}

/* response() --The user made some sort of response to the
 * question posed in any one of the above routines
 */
void
response(Widget w,
		 int *answer,
		 XmAnyCallbackStruct *cbs)
{
	if (cbs->reason == XmCR_OK)
		*answer = YES;
	else if (cbs->reason == XmCR_CANCEL)
		*answer = NO;
	else if (cbs->reason == XmCR_HELP)
		*answer = THIRD;
}
