
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
/*
// modelled on gnu map class
// bogus -- no growth function or overflow condition checking
*/

#include "rafport.h"

#include "rafutil.h"

#include "smemory.h"

#include "hash.h"
#include "stringVoidPtrMap.h"

/* bogus -- I want to choose a prime -- and choose another when I grow */
#define DEFAULT_ARRAY_SIZE 1013

stringVoidPtrMap *
svmapAlloc()
{
	return (stringVoidPtrMap *) smalloc(sizeof(stringVoidPtrMap));
}

void 
svmapFree(stringVoidPtrMap *that)
{
	smfree(that);
}

void 
svmapInit(stringVoidPtrMap *that, VoidPtr input_default_value, int input_array_size)
{
	int i;
	that->array_size = input_array_size ? input_array_size : DEFAULT_ARRAY_SIZE;
	that->count = 0;
	that->array = (MapRecord *) smalloc(sizeof(MapRecord) * that->array_size);
	that->default_value = input_default_value;
	
	for(i = 0; i < that->array_size; i++){
		that->array[i].key = 0;
		that->array[i].deleted = FALSE;
		that->array[i].value = that->default_value;
	}
}

void 
svmapCleanup(stringVoidPtrMap *that)
{
	svmapClear(that);
	smfree(that->array);
	that->array = 0;
	that->array_size = 0;
	that->count = 0;
}

int 
svmapEmptyP(stringVoidPtrMap *that)
{
	return that->count == 0;
}

int 
svmapLength(stringVoidPtrMap *that)
{
	return that->count;
}

static MapRecord *
indexByKeyFib(stringVoidPtrMap *that, const char *key, 
			  int i, int fib0, int fib1)
{
	int fib2;
	MapRecord *record = that->array + i;
	
	if(record->key){
		/* there is an entry here, do the strings match */
		if(stringEqual(key, record->key))
			return record;
		else {
			/* this is where we thought we would be but somebody 
			   else has hashed to this spot, the search continues
			   */
			fib2 = fib0 + fib1;
			return indexByKeyFib(that, key, (i + fib2) % that->array_size, fib1, fib2); 
		}
	} else {
		return record;
	}
}

static MapRecord *
indexByKey(stringVoidPtrMap *that, const char *key)
{
	int initial_guess;
	
	initial_guess = stringHash(key) % that->array_size;
	return indexByKeyFib(that, key, initial_guess, 0, 1);
}

VoidPtr 
svmapGet(stringVoidPtrMap *that, const char *key)
{
	MapRecord *record;
	
	record = indexByKey(that, key);
	if(record->key == 0 || record->deleted)
		halt("svmap: accessed non-existent entry %s.\n", key);
	
	return record->value;
}


char *
myStrdup(const char *original)
{
	char *s;
	s = smalloc(strlen(original) + 1);
	if(s)
		strcpy(s, original);
	/* bogus, log an error */
	return s;
}

void 
svmapSet(stringVoidPtrMap *that, const char *key, VoidPtr value)
{
	MapRecord *record;
	
	record = indexByKey(that, key);
	if(record->key == 0){
		/* bogus make an overflow check */
		record->key = myStrdup(key);
		that->count++;
	} else if(record->deleted){
		record->deleted = FALSE;
		that->count++;
	}
	
	record->value = value;
}

int
svmapContainsP(stringVoidPtrMap *that, const char *key)
{
	MapRecord *record;
	
	record = indexByKey(that, key);
	return (record->key != 0) && (record->deleted == FALSE);
}

void 
svmapDelete(stringVoidPtrMap *that, const char *key)
{
	MapRecord *record;
	
	record = indexByKey(that, key);
	if(record->key == 0 || record->deleted)
		halt("svmap: attempted to delete non-existent entry %s\n", key);
	
	record->deleted = TRUE;
	that->count--;
}

void 
svmapClear(stringVoidPtrMap *that)
{
	int i;
	
	for(i = 0; i < that->array_size; i++)
		if(that->array[i].key){
			smfree(that->array[i].key);
			that->array[i].key = 0;
			that->array[i].value = that->default_value;
			that->array[i].deleted = FALSE;
		}
	that->count = 0;
}

VoidPtr 
svmapDefault(stringVoidPtrMap *that)
{
	return that->default_value;
}

char *
svmapKey(stringVoidPtrMap *that, Ref r)
{
	MapRecord *record = (MapRecord *) r;
	
	return record->key;
}

VoidPtr 
svmapContents(stringVoidPtrMap *that, Ref r)
{
	MapRecord *record = (MapRecord *) r;
	
	return record->value;
}

Ref 
svmapFirst(stringVoidPtrMap *that)
{
	int i;
	MapRecord *record;
	
	for(i = 0; i < that->array_size; i++){
		record = that->array + i;
		if(record->key && (record->deleted == FALSE))
			return (Ref) record;
	}
	
	return 0;
}

Ref 
svmapNext(stringVoidPtrMap *that, Ref prev)
{
	int i;
	MapRecord *record;
	
	record = (MapRecord *) prev;
	for(i = (record - that->array) + 1; i < that->array_size; i++){
		record = that->array + i;
		if(record->key && (record->deleted == FALSE))
			return (Ref) record;
	}
	
	return 0;
}

Ref 
svmapSeek(stringVoidPtrMap *that, const char *key)
{
	MapRecord *record;
	
	record = indexByKey(that, key);
	if(record->key == 0 || record->deleted)
		return 0;
	else
		return record;
}

