
#include <stdio.h>
#include "rafport.h"

#include "rafutil.h"
#include "smemory.h"
#include "c-type.h"
#include "str-lib.h"
#include "c-lib.h"

#include "ifc.h"
#include "cyrconv.h"
#include "dafs.h"

static iTypeIndex aPara0,aChapter,aSection,aSubpara1,aSubpara2;
static iTypeIndex aSubpara3,aSubpara4;
	
static int memcheck,stripleading,debug;
#if 0
static int beg, end;
#endif
static char *outfile,*tempfile,*imagefile;

static ifc myargs[] = {
	{(char **) &debug,0,0,1,IFC_FLAG,"d",0},
	{(char **) &imagefile,0,0,1,IFC_STR,"i",0},
	{(char **) &memcheck,0,0,10,IFC_INT,"m",0},
	{(char **) &outfile,0,0,0,IFC_STR,"o",0},
	{(char **) &stripleading,0,1,0,IFC_FLAG,"s",0},
	{(char **) &tempfile,0,0,0,IFC_STR,"t",0},
	{0, 0,0, 1, IFC_NA, 0, 0}
};

static void
usage(char *p)
{
   fprintf(stderr, "Usage: %s [-s] [-t file] [-i file] [-o file] text.pda\n",p);
   fprintf(stderr, "\t-i file\timage file name.\n");
   fprintf(stderr, "\t-o file\toutput file name.\n");
   fprintf(stderr, "\t-s\tdon't strip leading spaces and newlines\n");
   fprintf(stderr, "\t-t file\tmerge with file.\n");
   exit(-1);
}

static unsigned char *csiparse(unsigned char *bp, int *code, int *arr)
{
	unsigned char *cbp, save;
	for (cbp = bp;; cbp++) {
		if (*cbp == ';') {
			save = *cbp;
			*cbp = 0;
			*arr++ = atoi(bp);
			*cbp = save;
			bp = cbp + 1;
		} else if (*cbp < '0' || *cbp > '9')
			break;
	}
	*code = *cbp;
	*cbp = 0;
	*arr++ = atoi(bp);
	*cbp = *code;
	bp = cbp + 1;
	*arr++ = 0;
	if (*code == ' ') {
		*code = cbp[1];
		bp++;
	}
	return bp;
}

static int 
zoneheader(FILE * in, short *xp, short *yp, short *wp, short *hp, int *comp,int *rp)
{
	char buf[50], c;
	int e, i,x,y,w,h,r;
	do {
		e = fread(buf, 1, 1, in);
		if (e != 1 || feof(in) || buf[0] == 0xc)
			return 0;
	} while (buf[0] == 0xb);
	w = 2560;
	h = 0;
	x = 0;
	y = 0;
	*comp = buf[0] - '0';

	e = fread(buf, 1, 1, in);
	if (e != 1 || buf[0] == 0xa)
		return 1;

	buf[5] = 0;
	e = fread(&buf[1], 1, 4, in);
	if (e != 4)
		return 0;
	x = atoi(buf);
	e = fread(buf, 1, 5, in);
	if (e != 5)
		return 0;
	y = atoi(buf);

	e = fread(buf, 1, 5, in);
	if (e != 5)
		return 0;
	w = atoi(buf);
	e = fread(buf, 1, 5, in);
	if (e != 5)
		return 0;
	h = atoi(buf);
	for (i=0;i<50;i++) {
		e = fread(&buf[i],1,1,in);
		if (e != 1) return 0;
		if (buf[i] == ',' || buf[i] == 0xa)
			break;
	}
	c = buf[i];
	if (*comp < 3) {
		buf[i - 1] = 0;
		r = atoi(&buf[i-3]);
	} else if (*comp == 3) {
		buf[i] = 0;
		r = atoi(&buf[i-2]);
	}
	if (!r) r=4;
	x /= r;
	y /= r;
	w /= r;
	h /= r;
	if (xp) *xp=x;
	if (yp) *yp=y;
	if (wp) *wp=w;
	if (hp) *hp=h;
	if (rp) *rp=r;
	while (c != 0xa) {
		e = fread(&c, 1, 1, in);
		if (e != 1 || feof(in))
			return 0;
	}
	return 1;
}

static void MinLeadSpace(iEntityPtr en)
{
	iEntityPtr line,next;
	int min,n,type,first,len,alen;
	iUCode *uc;

	min = 0x7fffffff;
	first = 1;
	for (line=i_GetChildEntity(en);line;line = i_GetNextEntity(line)) {
		type = i_GetEntityType(line);
		if (type == iNewLineT) {
			first = 1;
		} else if (first) {
			uc = (iUCode*)i_GetUText(line);
			if (!uc) return;
			n = 0;
			while (*uc++ == ' ') n++;
			if (n < min) min = n;
			first = 0;
		}
	}
	first = 1;
	alen = 300;
	uc = (iUCode *)smalloc((alen+1) * sizeof(iUCode));
	if (!uc) 
		halt("malloc failed\n");
	for (line=i_GetChildEntity(en);line;line = next) {
		next = i_GetNextEntity(line);
		type = i_GetEntityType(line);
		if (type == iNewLineT) {
			first = 1;
		} else if (first) {
			len = ucslen(i_GetUText(line));
			if (len > min) {
				if (len >= alen) {
					alen = len + 300;
					uc = (iUCode*)smrealloc(uc,alen);
					if (!uc) halt("malloc failed\n");
				}
				ucscpy(uc,i_GetUText(line));
				i_SetUText(line,uc+min);
			} else
				i_DisposeEntity(line);
			first = 0;
		}
	}
	smfree(uc);
}

static int sameBox(iEntityPtr en,iBox cb)
{
	iBox b;
	
	b = i_GetEnclosingBox(en);
	if (b.x != cb.x) return 0;
	if (b.y != cb.y) return 0;
	if (b.w != cb.w) return 0;
	if (b.h != cb.h) return 0;
	return 1;
}

static short *bagsps(short *in)
{
	while (*in == ' ') in++;
	return in;
}

static int countchrs(iEntityPtr lp,int first)
{
	int dt;
	short *sp;
	
	if (i_GetChildEntity(lp))
		return i_CountChildren(lp);
	dt = i_GetEDataType(lp);
	if (dt == iDPosSet || dt == iD1PosSet)
		return 1;
	if (dt != iDUString) halt("should be here\n");
	if (first) sp = bagsps(i_GetUText(lp));
	else sp = i_GetUText(lp);
	return ucslen(sp);
}	

static iDAFSError 
fixforw(iEntityPtr lp,iBox lb,int cnt,int first)
{
	iBox b;
	int pos,xe;
	iDAFSError e;
	
	xe = lb.x;
	for (;lp;lp = i_GetNextEntity(lp)) {
		if (i_GetEntityType(lp) == iNewLineT) return DAFSOK;
		b = i_GetEnclosingBox(lp);
		if (b.h) return DAFSOK;
		pos = countchrs(lp,first);
		first = 0;
		b = lb;
		b.x = xe;
		xe = lb.x + ((lb.w * pos) / cnt);
		b.w = xe - b.x;
		i_SetBox(lp,b);
		if (e = i_SetUserPtrProp(lp,"I Was Here",0,NULL))
			return e;
	}
	return DAFSOK;
}

static iDAFSError 
fixbbs(iEntityPtr lp,iBox lb)
{
	iBox b,tb;
	int num,lastpos,first,t;
	iEntityPtr wp,prev;
	iDAFSError e;

	num = 0;
	if (!lb.h) {
		lb.h = 40;
		lb.y -= 40;
	}
	lastpos = lb.x + lb.w;
	for (wp=lp;wp;wp = prev) {
		t = i_GetEntityType(wp);
		if (t == iNewLineT) break;
		b = i_GetEnclosingBox(wp);
		if (!i_GetUserPtrProp(wp,"I Was Here",0)) break;
		prev = i_GetPrevEntity(wp);
		first = prev && (i_GetEntityType(prev) == iNewLineT);
		if (b.h && b.w) {
			if (num) {
				tb = lb;
				tb.x = b.x + b.w;
				tb.w = lastpos - tb.x;
				if (e = fixforw(i_GetNextEntity(wp),tb,num,0))
					return e;
			}
			num = 0;
			lastpos = b.x;
		} else {
			num += countchrs(wp,first);
		}
	}
	if (num) {
		tb = lb;
		tb.w = lastpos - tb.x;
		if (e = fixforw(i_GetNextEntity(wp),tb,num,1))
			return e;
	}
	return DAFSOK;
}

static iDAFSError 
parsetext(unsigned char *bp, int size, iEntityPtr curz,int res)
{
/*************************************************************************
	 	bp -- byte pointer used to move through the image
		size--total amount of bytes to parse
		curz--current zone to be parsed
		res --resolution in units of 1200/dpi
**************************************************************************/
	int arr[5], code, num,numsame;
	long lx,lw;
	iBox cb,wb,lb,zb;
	unsigned char *ebp;
	iEntityPtr wp,lp,cp,lasten;
	short ch, conf,flagc,flagw;
	const iUCode *text;
	FILE *fo;

	if (debug && !(fo = fopen("fred","w")))
		halt("error making fred\n");

	zb=i_GetEnclosingBox(curz);
	lb.h = 0;
	cb.h = 0;
	wb.h = 0;
	flagc = 0;
	flagw = 0;
	lasten = 0;
	numsame = 0;
	for (ebp = bp + size;bp< ebp;) {
		switch (*bp) {
		/* newline byte */
		case 0xa:
			num = 0;
			while (*bp == 0xa) {
				num++;
				bp++;
				if (debug) fprintf(fo,"\n");
			}
			lp = i_NewEntity(iNewLineT,curz);
			if (!lp) return MallocFailError;
			i_SetShort(lp, num,0);
			lp = 0;
			break;
		/* space byte */
		case ' ':
			wp = 0;
			if (!lp) {
				lp = i_NewEntity(iLineT,curz);
				if (!lp) return MallocFailError;
			}
			while (*bp == ' ') {
				i_AddUni(lp,-1,' ');
				bp++;
				if (debug) fprintf(fo," ");
			}
			break;
		/* csi parse byte */
		case 0x9b:
			bp++;
			bp = csiparse(bp, &code, arr);
			switch (code) {
			case 0x43:	/* Point size */
			case 0x6d:	/* bold, italic ... */
				break;
			case 0x70:	/* char mark */
				flagc = 1;
				break;

			case 0x72:	/* word mark */
				flagw = 1;
				if (debug) fprintf(fo,"\\");
				break;
			case 0x60:	/* horiz position */
				if (debug) fprintf(fo,"|");
			    lb.w=(arr[0]/res+zb.x)-lb.x;
			    if (!lp) lp = wp;
			    fixbbs(i_GetLastEntity(curz),lb);
				lp = 0;
			    break;
			case 0x66:     /* horiz and vert position */
				if (debug) fprintf(fo,"%%");
			/* line bounding box */
				lb.y=arr[0]/res+zb.y;
				lb.x=arr[1]/res+zb.x;
				lb.h=0;
				break;
			case 0x71:
				if (debug) fprintf(fo,"$");
			/* character bounding box */
			/* cbb arr[0],arr[1],arr[2],arr[3] */
				cb.x=arr[0]/(res);
				cb.y=arr[1]/(res);
				cb.w=arr[2]/(res);
				cb.h=arr[3]/(res);
				if (cb.y < lb.y) {
					lb.h += lb.y-cb.y;
					lb.y = cb.y;
				}
				if (cb.y+cb.h > lb.y+lb.h)
					lb.h += cb.y+cb.h - (lb.y+lb.h);
				break;
			case 0x73:	
				if (debug) fprintf(fo,"+");
			/* word bounding box */
			/* wbb arr[0],arr[1],arr[2],arr[3] */
				wb.x=arr[0]/(res);
				wb.y=arr[1]/(res);
				wb.w=arr[2]/(res);
				wb.h=arr[3]/(res);
				if (wb.y < lb.y) {
					lb.h += lb.y-wb.y;
					lb.y = wb.y;
				}
				if (wb.y+wb.h > lb.y+lb.h)
					lb.h += wb.y+wb.h - (lb.y+lb.h);
				break;
			}
			break;
		default:
			if (*bp < 128) ch = *bp;
			else ch = isotable[*bp - 128].uni_code;
			if (debug) fprintf(fo,"%c",ch);
			bp++;
			if (flagw) {
				wp = i_NewEntity(iLineT,curz);
				if (!wp) return MallocFailError;
				i_SetLongProp(wp,"Flag",0);
				if (wb.h) i_SetBox(wp,wb);
				wb.h = 0;
				flagw = 0;
				lp = 0;
			}
			if (flagc) {
				lp = 0;
				if (wp) {
					if (text = i_GetUText(wp)) {
						while (*text) {
							cp = i_NewEntity(iGlyphT,wp);
							i_SetGlyph(cp,*text++,0);
						}
						i_SetUText(wp,0);
					}
					cp = i_NewEntity(iGlyphT,wp);
				} else {
					cp = i_NewEntity(iGlyphT,curz);
				}
				i_SetLongProp(cp,"Flag",0);
		    	if (cb.h) {
		    		i_SetBox(cp,cb);
			    	if (!lasten) lasten = cp;
			    	else if (sameBox(lasten,cb)) {
			    		numsame++;
			    	} else if (numsame) {
			    		numsame++;
			    		cb = i_GetEnclosingBox(lasten);
			    		lw = cb.w << 16;
			    		lw /= numsame;
			    		cb.w = lw >> 16;
			    		lx = cb.x << 16;
			    		for (;lasten && numsame--;lasten = i_GetNextEntity(lasten)) {
			    			cb.x = lx >> 16;
			    			i_SetBox(lasten,cb);
			    			lx += lw;
			    		}
			    		lasten = cp;
			    		numsame = 0;
			    	} else lasten = cp;
					cb.h = 0;
			    }
			    flagc = 0;
				i_SetGlyph(cp, ch,0);
		    } else {
				if (!lp && !wp) {
					lp = i_NewEntity(iLineT,curz);
					if (!lp) return MallocFailError;
				}
				if (wp) {
					if (i_GetChildEntity(wp)) {
						cp = i_NewEntity(iGlyphT,wp);
						i_SetGlyph(cp, ch,0);
					} else i_AddUni(wp,-1,ch);
				} else i_AddUni(lp,-1,ch);
		    }
			break;
		}
	}
	if (stripleading) {
		MinLeadSpace(curz);
		lp = i_GetChildEntity(curz);
		if (lp && (i_GetEntityType(lp) == iNewLineT))
			i_DisposeEntity(lp);
	}
	lp = i_GetLastEntity(curz);
	if (lp && i_GetEntityType(lp) == iNewLineT) {
		i_SetShort(lp,i_GetShort(lp, &conf) + 1, conf);
	} else {
		lp = i_NewEntity(iNewLineT,curz);
		if (!lp) return MallocFailError;
		i_SetShort(lp,1,0);
	}
	if (debug) fclose(fo);
	return DAFSOK;
}

static iDAFSError 
convert_PDA(const char *fn, iEntityPtr doc)
{
	FILE *in;
	int comp, amt, allcd, size,res=0;

	iEntityPtr curz;
	unsigned char buf[5], *bp;
	iBox b={0,0,0,0};
	iDAFSError e;

	in = fopen(fn, "rb");
	if (!in)
		return NonExistentFileError;
	while (zoneheader(in, &b.x, &b.y, &b.w, &b.h, &comp,&res)) {
		switch (comp) {
		case 0:
		case 1:
		case 2:
			buf[4] = 0;
			for (;;) {
				fread(buf, 1, 1, in);
				if (buf[0] < '0' || buf[0] > '9')
					break;
				fread(&buf[1], 1, 3, in);
				amt = atoi(buf);
				fseek(in,amt,1);
			}
			break;
		case 3:
			curz = i_NewEntity(iZoneT, doc);
			i_SetBox(curz, b);
			buf[4] = 0;
			allcd = 20000;
			bp = (unsigned char *) smalloc(allcd);
			if (!bp)
				return MallocFailError;
			size = 0;
			for (; bp;) {
				fread(buf, 1, 1, in);
				if (buf[0] < '0' || buf[0] > '9')
					break;
				fread(&buf[1], 1, 3, in);
				amt = atoi(buf);
				if (amt + size > allcd) {
					allcd += 20000 + amt;
					bp = (unsigned char *) smrealloc(bp, allcd);
					if (!bp)
						return MallocFailError;
				}
				fread(bp + size, 1, amt, in);
				size += amt;
			}
			e = parsetext(bp, size, curz,res);
			smfree(bp);
			if (e) return e;
			break;
		}
	}
	fclose(in);
	return DAFSOK;
}

static iDAFSError getthem(iEntityPtr pd,iEntityPtr en,int *foundp)
{
	iDAFSError e;
	iEntityPtr cur,next;
	iBox b,pb,nb;
	int an,ap,found;

	if (foundp) {
		*foundp = 0;
		b = i_GetEnclosingBox(en);
		pb = i_GetEnclosingBox(pd);
		ap = pb.w * pb.h;
		if (!ap) return DAFSOK;
		i_IntersectBox(b,pb,&nb);
		an = nb.w * nb.h;
		if (an < ap / 2) return DAFSOK;
		if (foundp) *foundp = 1;
		next = i_GetNextEntity(pd);
		i_MoveToBack(pd,en);

		while (next) {
			pd = next;
			next = i_GetNextEntity(pd);
			b = i_GetEnclosingBox(pd);
			if (b.w && b.h)
				break;
			if (i_GetEntityType(pd) == iNewLineT)
				i_SetShort(pd,1,0);
			i_MoveToBack(pd,en);
		}
	} else if (i_GetEntityType(pd) == iZoneT) {
		for (cur=i_GetChildEntity(pd);cur;) {
			if (e = getthem(cur,en,&found))
				return e;
			if (found) cur = i_GetChildEntity(pd);
			else cur = i_GetNextEntity(cur);
		}
	} else for (cur=i_GetChildEntity(pd);cur;cur = i_GetNextEntity(cur))
		if (e = getthem(cur,en,0))
			return e;
	return DAFSOK;
}

static int 
aGoodType(iEntityPtr en)
{
	int type;

	type=i_GetEntityType(en);
	return (type==aChapter)||(type==aSection)||
		(type==aSubpara1)||(type==iParagraphT)||
		(type==aSubpara2)||(type==aSubpara3)||
		(type==aSubpara4)||(type==aPara0);
}

static iDAFSError mergedafs(iEntityPtr pd,iEntityPtr en)
{
	iDAFSError e;
	iEntityPtr cur;
	
	if (aGoodType(en)) {
		while (cur = i_GetChildEntity(en))
			i_DisposeEntity(cur);
		if (e = getthem(pd,en,0))
			return e;
		cur = i_GetLastEntity(en);
		if (cur && (i_GetEntityType(cur) != iNewLineT)) {
			cur = i_NewEntity(iNewLineT,en);
			if (!cur) return MallocFailError;
			i_SetShort(cur,1,255);
		}
	} else for (cur=i_GetChildEntity(en);cur;cur = i_GetNextEntity(cur))
		if (e = mergedafs(pd,cur))
			return e;
	return DAFSOK;
}

static void
setallimg(iEntityPtr doc,iImagePtr img)
{
	iEntityPtr cur;
	i_SetImage(doc,img);
	for (cur=i_GetChildEntity(doc);cur;cur = i_GetNextEntity(cur))
		setallimg(cur,img);
}

static iDAFSError stripsp(iEntityPtr doc)
{
	iDAFSError e;
	iEntityPtr cur,prev,next;
	char *cp;
	short *sp;
	int4 beg,end;
	
	if (i_GetEDataType(doc) == iDString) {
		if (prev = i_GetPrevEntity(doc)) {
			if (i_GetEntityType(prev) != iNewLineT)
				return DAFSOK;
		}
		beg = end = 0;
		cp = (char*)i_GetText(doc);
		while (*cp && *cp == ' ') cp++,end++;
		if (end) {
			if (end == strlen(i_GetText(doc))) {
				i_DisposeEntity(doc);
			} else {
				cur = doc;
				i_DeleteSelection(&doc,&beg,&cur,&end);
			}
		}
	} else if (i_GetEDataType(doc) == iDUString) {
		beg = end = 0;
		if (prev = i_GetPrevEntity(doc)) {
			if (i_GetEntityType(prev) != iNewLineT)
				return DAFSOK;
		}
		sp = (short*)i_GetUText(doc);
		while (*sp && *sp == ' ') sp++,end++;
		if (end) {
			if (end == ucslen(i_GetUText(doc))) {
				i_DisposeEntity(doc);
			} else {
				cur = doc;
				i_DeleteSelection(&doc,&beg,&cur,&end);
			}
		}
	} else for (cur=i_GetChildEntity(doc);cur;cur = next) {
		next = i_GetNextEntity(cur);
		if (e = stripsp(cur))
			return e;
	}
	return DAFSOK;
}

int
main(int c, char *v[])
{
	int i;
	iDAFSError e;
	iImagePtr image;
	iEntityPtr doc,en;
	char name[256];
	char *extv[] = {".dr",".pda",""};

	printf("pda2dafs doesn't do select zones properly.\n");
	
	if (processargs(c, v, myargs))
		usage(v[0]);
	s_InitLib(memcheck, 0, 0);
	i_InitLib();

	i_NameToType("chapter", &aChapter);
	i_NameToType("section", &aSection);
	i_NameToType("para0", &aPara0);
	i_NameToType("subpara1", &aSubpara1);
	i_NameToType("subpara2", &aSubpara2);
	i_NameToType("subpara3", &aSubpara3);
	i_NameToType("subpara4", &aSubpara4);
	for (i = 1; i < c && !v[i]; i++) ;
	image = 0;
	printf("reading %s\n",v[i]);
	doc = i_NewEntity(iDocumentT,0);
	if (e = convert_PDA(v[i],doc)) halt("couldn't open %s\n", v[i]);
	if (tempfile) {
		if (e = i_ReadEntity(tempfile,&en))
			halt("failed to read %s\n",tempfile);
		image = i_GetImage(en);
		if (e = mergedafs(doc,en))
			halt("failed mergedafs with %s\n",i_GetError(e));
		if (e = stripsp(en))
			halt("failed to strip spaces with %s\n",i_GetError(e));
#if 0
		if (debug) pprall(doc);
#endif
		i_DisposeEntity(doc);
		doc = en;
	}
	if (imagefile) {
		if (e = i_ReadImage(imagefile,&image))
			halt("couldn't open %s\n",imagefile);
	}
	if (image) {
		i_SetImage(doc,image);
		i_SetBox(doc,i_GetImageBounds(image));
		setallimg(doc,i_GetImage(doc));
	}
	stripext(c,v,extv);
	if (!outfile) {
		sprintf(name,"%s.dr",v[i]);
		printf("writing %s\n",name);
		outfile = name;
	}
	e = i_WriteEntity(outfile,doc,DAFSB);
	if (e) halt("couldn't write %s\n",outfile);
	i_DisposeEntity(doc);

	i_ExitLib();
	s_ExitLib();
	return 0;
}
