
#include "RTEView.h"
#include "ConstrainedRectangle.h"


// ------------------------- ConstraniedRectangle methods -----------------------

ConstrainedRectangle::ConstrainedRectangle(Coord x, Coord y, Coord w, Coord h,
                    int canMove, int canGrowUp,
                    int canGrowDown, int canGrowLeft, int canGrowRight, 
                    Direction primaryGrowthDir, Direction secondaryGrowthDir,
                    Coord spacing) :
	PackedRectangle(x, y, w, h, primaryGrowthDir, secondaryGrowthDir, spacing)
{
	assert(canMove      == TRUE || canMove      == FALSE);
	assert(canGrowUp    == TRUE || canGrowUp    == FALSE);
	assert(canGrowDown  == TRUE || canGrowDown  == FALSE);
	assert(canGrowLeft  == TRUE || canGrowLeft  == FALSE);
	assert(canGrowRight == TRUE || canGrowRight == FALSE);

	can_move = canMove;
	can_grow_up = canGrowUp;
	can_grow_down = canGrowDown;
	can_grow_left = canGrowLeft;
	can_grow_right = canGrowRight;
	primary_growth_dir = primaryGrowthDir;
	secondary_growth_dir = secondaryGrowthDir;
}

ConstrainedRectangle::ConstrainedRectangle(int canMove, int canGrowUp,
                    int canGrowDown, int canGrowLeft, int canGrowRight, 
                    Direction primaryGrowthDir, Direction secondaryGrowthDir,
                    Coord spacing) :
	PackedRectangle(primaryGrowthDir, secondaryGrowthDir, spacing)
{
	assert(canMove      == TRUE || canMove      == FALSE);
	assert(canGrowUp    == TRUE || canGrowUp    == FALSE);
	assert(canGrowDown  == TRUE || canGrowDown  == FALSE);
	assert(canGrowLeft  == TRUE || canGrowLeft  == FALSE);
	assert(canGrowRight == TRUE || canGrowRight == FALSE);

	can_move = canMove;
	can_grow_up = canGrowUp;
	can_grow_down = canGrowDown;
	can_grow_left = canGrowLeft;
	can_grow_right = canGrowRight;
	primary_growth_dir = primaryGrowthDir;
	secondary_growth_dir = secondaryGrowthDir;
}

ConstrainedRectangle::ConstrainedRectangle(Coord x, Coord y, Coord w, Coord h,
		Direction primaryGrowthDir, Direction secondaryGrowthDir,
		Coord spacing) :
	PackedRectangle(x, y, w, h, primaryGrowthDir, secondaryGrowthDir, spacing)
{
	// This defines various default values
	can_move = TRUE;
	can_grow_up = FALSE;
	can_grow_down = FALSE;
	can_grow_left = FALSE;
	can_grow_right = FALSE;
	primary_growth_dir = DEFAULT_PRIM_READ_ORDER;
	secondary_growth_dir = DEFAULT_SCND_READ_ORDER;
}

ConstrainedRectangle::ConstrainedRectangle(Direction primaryGrowthDir,
		Direction secondaryGrowthDir, Coord spacing) :
	PackedRectangle(primaryGrowthDir, secondaryGrowthDir, spacing)
{
	// This defines various default values
	can_move = TRUE;
	can_grow_up = FALSE;
	can_grow_down = FALSE;
	can_grow_left = FALSE;
	can_grow_right = FALSE;
	primary_growth_dir = DEFAULT_PRIM_READ_ORDER;
	secondary_growth_dir = DEFAULT_SCND_READ_ORDER;
}

ConstrainedRectangle::ConstrainedRectangle()
{
	// This defines various default values
	can_move = TRUE;
	can_grow_up = FALSE;
	can_grow_down = FALSE;
	can_grow_left = FALSE;
	can_grow_right = FALSE;
	primary_growth_dir = DEFAULT_PRIM_READ_ORDER;
	secondary_growth_dir = DEFAULT_SCND_READ_ORDER;
}

#if 0
ConstrainedRectangle::ConstrainedRectangle(Coord x, Coord y, Coord w, Coord h)
	: PackedRectangle(x, y, w, h)
{
	// This defines various default values
	can_move = TRUE;
	can_grow_up = FALSE;
	can_grow_down = FALSE;
	can_grow_left = FALSE;
	can_grow_right = FALSE;
	primary_growth_dir = DEFAULT_PRIM_READ_ORDER;
	secondary_growth_dir = DEFAULT_SCND_READ_ORDER;
}	
#endif

int ConstrainedRectangle::allocateArea(Coord x1, Coord y1, Coord x2, Coord y2, 
								  ChildRectangle *cr)
{
	int success = FALSE;
	ChildRectangle *r = new ChildRectangle(x1, y1, x2-x1+1, y2-y1+1);

	if(r) {
		if(this->element_list->insertItem(r)) {
			success = ConstrainedRectangle::allocateArea(r, cr);
			this->element_list->removeItem(r);
			delete r; r = 0;
		}
	}
	return success;
}

int ConstrainedRectangle::allocateArea(ChildRectangle *r, ChildRectangle *cr)
{
	// allocate the rectangular area r for the child Rectangle cr

	int success;
	
	assert(this->isParentOf(cr));
	assert(this->isParentOf(r));
	if(r->area() == 0)
		success = TRUE;  // null area allocation
	else {
		if(!this->contains(r)) this->growToContain(r);
		// Call allocateArea even if growToContain failed
		// so that the area will be cleared.
		success = PackedRectangle::allocateArea(r, cr);
	}
	return success;
}


int ConstrainedRectangle::canGrow(Direction dir)
{
	int result=FALSE; 
	switch (dir) {
		case LeftToRight: result = can_grow_right; break;
		case RightToLeft: result = can_grow_left;  break;
		case TopToBottom: result = can_grow_down;  break;
		case BottomToTop: result = can_grow_up;    break;
		default: halt("called with illegal direction"); break;
	}
	assert(result == TRUE || result == FALSE);
	return result;
}	


#if 0
Rectangle * ConstrainedRectangle::areaUnion(Rectangle *r)
{
	// Assume r's coords are relative to me.
	// Construct a rectangle nr with coords relative to my parent
	// such that r is contained by nr when inserted in nr at
	// its specified coords relative to nr.

	Rectangle *nr=this->ParentRectangle::areaUnion(r);
	if(this->parent != NULL) {
		switch(this->parent->getPrimaryWritingDir()) {
			case LeftToRight:
			case RightToLeft:
				const Coord dh = nr->h - this->h;
				if(dh > 0) {
					const Coord top_room = this->parent->getSpacing() - this->h % this->parent->getSpacing(),
				            dy = MIN(nr->getY(),MIN(top_room, dh));
					nr->setY(nr->getY() - dy);
				}
				break;
			case TopToBottom:
			case BottomToTop:
				break;  // do nothing
			case RTENone:
				break;  // do nothing
		}
	}
	return nr;
}
#endif

Rectangle * ConstrainedRectangle::areaUnion(Rectangle *r)
{
	// Assume r's coords are relative to me.
	// Construct a rectangle nr with coords relative to my parent
	// such that r is contained by nr when inserted in nr at
	// its specified coords relative to nr.

	return this->ParentRectangle::areaUnion(r);
}

Coord smgt(Coord, Coord);  // heinous - put this declaration into RTEMisc.h

int ConstrainedRectangle::calculateGrowthDeltas(Rectangle*r, Coord *pdx, Coord *pdy, 
                                                Coord *pdw, Coord *pdh)
{
	int result = FALSE;

	// calculate what I want to grow to
	Rectangle *rtest = this->areaUnion(r);

	if(rtest) {
		// note that dx and dy will never be positive because rtest contains me
		// and dw >= -dx >= 0 and dh >= -dw >= 0 for the same reason
		Coord dx = rtest->x - this->x,
	      	      dy = rtest->y - this->y,
	      	      dw = rtest->w - this->w,
	      	      dh = rtest->h - this->h;
		assert(dx <= 0 && dy <= 0 && dw >= -dx && dh >= -dw);

		// If I'm growing in the secondary writing direction, make sure I grow
		// by a multiple of my line spacing
		Coord ddw, ddh;
		Coord spacing;
		if(this->spacing)
		  spacing = this->spacing;
		else switch(secondary_writing_direction) {
		case LeftToRight:
		case RightToLeft:
		  spacing = 1;
		  break;
		case TopToBottom:
		case BottomToTop:
		  spacing = 1;
		  break;
		default:
		  break;
		}

		switch(this->secondary_writing_direction) {
			case LeftToRight:
			case RightToLeft:
				if(dw > 0) {
					if(!this->spacing || this->w % this->spacing == 0)
						ddw = 0;
					else
						ddw = this->spacing - (this->w % this->spacing);
					dw = ddw + smgt(MAX(0,dw-ddw), spacing);
					assert(this->w + dw >= rtest->w);
					rtest->w = this->w + dw;
					assert(rtest->w % spacing == 0);
				}
				break;
			case TopToBottom:
			case BottomToTop:
				if(dh > 0) {
					if(!this->spacing || this->h % this->spacing == 0)
						ddh = 0;
					else
						ddh = this->spacing - (this->h % this->spacing);
					dh = ddh + smgt(MAX(0,dh-ddh), spacing);
					assert(this->h + dh >= rtest->h);
					rtest->h = this->h + dh;
					assert(rtest->h % spacing == 0);
				}
				break;
			default:
				halt("no secondary writing direction\n");
				break;
		}
		delete rtest;
		*pdx = dx;  *pdy = dy;  *pdw = dw;  *pdh = dh;
		result = TRUE;
	}
	return result;
}

int ConstrainedRectangle::checkIfCanGrow(Coord dx, Coord dy, Coord dw, Coord dh)
{
	const int want_grow_left  = (dx < 0),
	          want_grow_right = (dw > -dx),
	          want_grow_up    = (dy < 0),
	          want_grow_down  = (dh > -dy);

	// see of my constraints allow me to do it
	const int can_grow = (!want_grow_left  || this->can_grow_left ) &&
	                     (!want_grow_right || this->can_grow_right) &&
	                     (!want_grow_up    || this->can_grow_up   ) &&
	                     (!want_grow_down  || this->can_grow_down );
	return can_grow;
}

int ConstrainedRectangle::canGrowToContain(Rectangle *r)
{
	Coord dx, dy, dw, dh;
	this->calculateGrowthDeltas(r, &dx, &dy, &dw, &dh);
	return this->checkIfCanGrow(dx, dy, dw, dh);
}

int ConstrainedRectangle::growToContain(Rectangle *r, int shift_children)
{
	// Assume r's x and y coords. relative to me are already set.
	// Grow myself so that r fits inside me when positioned at
	// its specified x and y relative to me.
	
	int result;
	Coord dx, dy, dw, dh;
	
	this->calculateGrowthDeltas(r, &dx, &dy, &dw, &dh);
	if(dx == 0 && dy == 0 && dw == 0 && dh == 0) {
		result = TRUE; // nothing to do
	} else {
		result = FALSE;
		int shifted = FALSE;
		if(this->checkIfCanGrow(dx, dy, dw, dh)) {
			// if(dx || dy) shifted = this->moveBy(dx, dy);
			if(dx || dy) shifted = TRUE;
			// Redraw myself only if I'm shrinking.  If my children move around,
			// they will redraw themselves.
			const int redraw = (dw < 0 || dh < 0);
			// use ChildRectangle method to avoid repacking
			// heinous - won't work in line wrap mode
			// result = this->ChildRectangle::changeSizeTo(this->getW()+dw, this->getH()+dh, redraw);
			result = this->changeSizeTo(this->getW()+dw, this->getH()+dh, redraw);
			if(shifted) {
				// shift my children by the opposite offset so they wind
				// up in the same location
				const int squeeze=FALSE, allow_overlap=TRUE; 
				int child_shifted = r->moveBy(-dx, -dy, squeeze, allow_overlap);
				for(int i=0; shift_children && child_shifted && i<this->numChildren(); i++) {
					ChildRectangle *c = this->getChildNumber(i);
					if(c->positionIsValid())
						child_shifted = c->moveBy(-dx, -dy, squeeze, allow_overlap);
				}
				if(!child_shifted) result = FALSE;
			}
		}
	}

	assert(this->getType() == RootElementT || this->parent != NULL);
	return result;
}

