
#include "smemory.h"
#include "Compare.h"
#include "Generic.h"

// -------- General purpose template routines used by various classes -------

#if 0
void swap(void *a, void *b)
{
	void temp = *a;
	*a = *b;
	*b = temp;
}
#endif

int simpleArrayInsert(void ***a, void *e, int *size)
{
	// This version of arrayInsert uses the simpleBSearch which compares
	// items by pointer value rather than packing order.
	
	int success;
	assert(isSorted(*a, *size));
	if(*a == NULL) {
		const int loc = 0;
		success = listInsert(a, e, loc, size);
	} else {
		assert(*size > 0);
		int found, loc = simpleBSearch(*a, e, *size, &found);
		if(found) 
			success = TRUE;  
		else
			success = listInsert(a, e, loc, size);
	}
	assert(isSorted(*a, *size));
	assert((success == TRUE) || (success == FALSE));
	return success;
}

int simpleBSearch(void **a, void *e, int size, int *found)
{
	// General purpose routine which searches for a thing *e in an array of
	// comparable things.  If *e is found, return the index where found.  If not,
	// return the location where it should be inserted in the list to maintain
	// sorted order.
	
	assert((size >= 0) && isSorted(a, size));
	
	int loc;
	*found = FALSE;
	if(size == 0) {
		loc = 0;
		// *found = FALSE;
	} else {
		if (e < a[0]) {
			loc = 0;
			// *found = FALSE;
		} else if (e == a[0]) {
			*found = TRUE;
			loc = 0;
		} else if (size == 1 && e > a[0]) {
			loc = 1;
			//*found = FALSE;
		} else if (e > a[size-1]) {
			loc = size;
			//*found = FALSE;
		} else if (e == a[size-1]) {
			*found = TRUE;
			loc = size-1;
		} else {
			int start=0, end=size-1, shift=0;
			do{
				int middle=(start+end+shift)/2;
				if(e == a[middle]) {
					*found = TRUE;
					loc = middle;
				} else if (e > a[middle]) {
					if(start == middle) shift++;
					start = middle;
				} else {
					if(end == middle) shift++;
					end = middle;
				}			
				/* handle special case where middle
				   jumps between start and end */
				if(shift == 2) start = end = middle;		
			} while(!*found && start < end);
			if(!*found) loc = start;
		}
	}
	
	assert(loc >= 0 && loc <= size);
	if(*found) 
		assert(e == a[loc]);
	else
		assert( (loc == 0 || (e >= a[loc-1])) &&
		        (loc == size || (e <= a[loc])) );
	
	return loc;
}

int arrayInsert(void ***a, void *e, int *size, CompareFunction compare)
{
	int success = TRUE+1;
	assert(isSorted(*a, *size));
	if(*a == NULL) {
		int loc = 0;
		success = listInsert(a, e, loc, size);
	} else {
		assert(*size > 0);
		int found, loc = bSearch(*a, e, *size, &found, compare);
		if(found && (*a)[loc] != e) {
			int lloc = loc +1;
			while(lloc < *size && (*a)[lloc] != e && 
			      compare((*a)[lloc], e) != GREATER_THAN)
				lloc++;
			found = (lloc < *size && (*a)[lloc] == e);
		}
		if(found) 
			success = TRUE;  
		else
			success = listInsert(a, e, loc, size);
	}
	assert(isSorted(*a, *size));
	assert((success == TRUE) || (success == FALSE));
	return success;
}

int bSearch(void **a, void *e, int size, int *found, CompareFunction compare)
{
	// General purpose routine which searches for a thing *e in an array of
	// comparable things.  If *e is found, return the index where found.  If not,
	// return the location where it should be inserted in the list to maintain
	// sorted order.
	
	// When searching for an interval in a list of intervals, if found is true
	// loc will be the index of the first interval which has some overlap with
	// the sought interval.
	
	assert((size >= 0) && isSorted(a, size));
	
	int loc=-1; // initial bogus value
	*found = FALSE;
	
	if(size == 0) 
		loc = 0;
	else {
		int start=0, middle, end=size-1, diff = compare(e, a[start]);
		if (diff == EQUAL || diff == IS_CONTAINED || diff == CONTAINS || diff == OVERLAPS) {
			*found = TRUE;
			loc = start;
		} else if (diff == LESS_THAN)
			loc = start;
		else if ((diff == GREATER_THAN) && (end == start))
			loc = end+1;
		else if ((diff=compare(e, a[end])) == GREATER_THAN)
			loc = end+1;
		else if (diff == EQUAL || diff == IS_CONTAINED || diff == CONTAINS || diff == OVERLAPS) {
			loc = end;
			*found = TRUE;
		}
		else {
			int shift=0;
			start = 0;
			do{
				middle=(start+end+shift)/2;
				diff=compare(e, a[middle]);
				if(diff == EQUAL || diff == IS_CONTAINED || diff == CONTAINS || 
				   diff == OVERLAPS) {
					*found = TRUE;
					loc = middle;
				}
				else if(diff == GREATER_THAN){
					if(start == middle) shift++;
					start = middle;
				}
				else {
					if(end == middle) shift++;
					end = middle;
				}
				
				/* handle special case where middle
				   jumps between start and end */
				if(shift == 2) start = end = middle;
				
			} while(!*found && start < end);
			if(!*found) loc = start;
		}
		
		// In case the list has duplicates, make sure we return the 
		// location of the first element which is equal to e.
		if(*found && (void *)a[loc] != (void *)e)
			while(loc > 0 && compare(e, a[loc-1]) != GREATER_THAN)
				loc--;
	}
	
	assert(loc >= 0 && loc <= size);
	if(*found) {
		assert(compare(e,a[loc]) == EQUAL || compare(e,a[loc]) == IS_CONTAINED ||
		       compare(e,a[loc]) == CONTAINS || compare(e,a[loc]) == OVERLAPS);
		if(compare(e, a[loc]) == OVERLAPS || compare(e, a[loc]) == CONTAINS)
			assert(loc == 0 || compare(e, a[loc-1]) == GREATER_THAN);
	} else {
		assert( (loc == 0 || compare(e, a[loc-1]) != LESS_THAN) &&
		        (loc == size || compare(e, a[loc]) != GREATER_THAN) );
	}
	
	return loc;
}


int listInsert(void ***a, void *e, int loc, int *size)
{
	int success = TRUE+1;
	assert(isSorted(*a, *size) && loc <= *size);
	if(*a == NULL) {
		assert(*size == 0);
		*a = (void **) smalloc(sizeof(void *));
		if(*a == NULL) {
			success = FALSE;
		} else {
			**a = e;
			*size = 1;
			success = TRUE;
		}
	} else {
		assert(*size > 0); 
		void **new_a = (void **) smrealloc(*a, (*size +1) * sizeof(void *));
		if(new_a == NULL) {
			// don't corrupt the existing list
			assert(isSorted(*a, *size)); // is existing list OK?
			success = FALSE;
		} else {
			for (int i=*size; i>loc; i--)
				new_a[i] = new_a[i-1];
			new_a[loc] = e;
			*a = new_a;
			(*size)++;
			success = TRUE;
		}
	}
	assert(isSorted(*a, *size));
	assert((success == TRUE) || (success == FALSE));
	return success;
}



int isSorted(void **a, int size, CompareFunction compare)
{
	// make sure the array a[] of pointers to type T
	// is sorted in ascending order with no duplicates
	
	assert(size >= 0);
	for(int i=1; i<size; i++) {
		const int c = compare(a[i-1],a[i]);
		if(c == GREATER_THAN)
			return FALSE;
		for(int j=i+1; j<size; j++)
			if(a[j] == a[i])
				return FALSE;
	}
	return TRUE;
}


 void listDelete(void ***a, int loc, int *size)
{
	assert(*a != NULL && *size > 0);
	int i;
	void **aa;
	for (i=loc, aa = *a; i < *size-1; i++)
		aa[i] = aa[i+1];
	(*size)--;
	if(*size > 0) {
		*a = (void **) smrealloc(*a, *size * sizeof(void *));
		if(*a == NULL) halt("smrealloc failed when size was decreasing");
	} else {
		smfree(*a); *a = NULL;	
	}
	assert(isSorted(*a, *size));
}
