
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/* use indice not index because it a macro on some systems -- yech */

#include "stdio.h"


#include "rafport.h"
#include "memdbg.h"

#include "rafutil.h"
#include "c-lib.h"

#include "smemory.h"


#define N_HEADS 200
#define N_FREE_HEADS 500

#ifndef SYS_OS_LINUX
#define QUICK_LIST 1
#endif

#ifndef DO_INLINE
#define DO_INLINE 0
#endif

static hdr *head[N_HEADS];
static hdr *free_head[N_FREE_HEADS];

MemoryInfo mi = { 
	sNoChecks,
    0,
	-1,
	0,
	0,
	0,
	0,
	-1,
	0,
	0
};

int4 errorFlag;
int4 sReturnError=0;

/* random pattern to be stuck in the beginning and end to
   check for clobbering */
#define N_RAND 8
static char start_pat[N_RAND]={'S','t','a','r','t','1','2','3'};
static char end_pat[N_RAND]={'F','i','n','i','s','h','9','8'};
static char *clob_str[]={"","before","after"};

static MemoryFailureCallback fail_callback = 0;
static void *fail_user_data = 0;

static int4 memory_has_initialized = FALSE;

#if DO_INLINE
#define INLINE inline
#else
#define INLINE
#endif

#ifdef SYS_OS_WIN16
#include <windowsx.h>
#endif

Dll void APIENTRY
s_InitLib(sMemFlags st, 
		  MemoryFailureCallback callback, 
		  void *user_data)
{	
	if(memory_has_initialized == FALSE) {
		long indice;
		mi.memory_track_flag = st;

		fail_callback = callback;
		fail_user_data = user_data;

		for (indice=0; indice<N_HEADS; indice++) {
			head[indice]=0;
			free_head[indice]=0;
		}
		mi.nsmallocs=0;
		memory_has_initialized = TRUE;
	}
}

void 
smstophere(int4 order)
{
	mi.stophere = order;
}

void 
add_clobber(char **ptr, int4 size)
{
	int4 i;
	size+=sizeof(hdr)+N_RAND;
	for (i=0; i<N_RAND; i++) {
		(*ptr)[i]=start_pat[i];
		(*ptr)[i+size]=end_pat[i];
	}
	(*ptr)+=N_RAND;
}

long gettotmem(void)
{
	return mi.totmem;
}

void setmemlimit(int4 limit)
{
	mi.memlimit=limit;
}

void setmemspace(long space)
{
	mi.memlimit=mi.totmem+space;
}

void setsizelimit(long limit)
{
	mi.sizelimit=limit;
}

int4 clobbered(char *ptr, long size)
{
	int4 i;
	char *ptr1,*ptr2;
	ptr1=ptr-N_RAND-sizeof(hdr);
	ptr2=ptr+size;
	for (i=0; i<N_RAND; i++) {
		if (*ptr1++!=start_pat[i])
			return 1;
		if (*ptr2++!=end_pat[i])
			return 2;
	}
	return 0;
}

void check_all_clobber(void)
{
	hdr *p;
	long indice;
	int4 trigger=0;
	int4 cl;
	
	if (mi.memory_track_flag < sTrackAll) return;
	for (indice=0; indice<N_HEADS; indice++)
		for (p = head[indice]; p; p = p->next) {
			if ((cl=clobbered((char *)&p[1],p->size))) {
				debugf("%p, size=%ld, nsmalloc=%ld clobbered %s\n",
					   &p[1], p->size, 
					   p->alloc_num, clob_str[cl]);
				trigger=1;
			}
		}
	if (trigger)
		*(char *)0=0;
}


void smstat(int4 i)
{
	if (mi.memory_track_flag == sNoChecks) return;
	debugf("total mem %ld, at %ld\n", mi.totmem,i);
}


int4 free_written_to(unsigned char *ptr, long size)
{
	int4 i;
	for (i=0; i<size; i++,ptr++)
		if (*ptr!=0xFF) return 1;
	return 0;
	
}

void clear_mem(unsigned char *ptr, long size)
{
	int4 i;
	for (i=0; i<size; i++,ptr++)
		*ptr=0xFF;
}

void
smprmem(void)
{
	hdr *p;
	int4 cl;
	long indice;
	int4 trigger=0;
	
	if (mi.memory_track_flag < sTrackAll) return;
	for (indice=0; indice<N_HEADS; indice++)
		for (p = head[indice]; p; p = p->next) {
			debugf("address:%p size:%ld order:%ld",
				   &p[1], p->size, p->alloc_num);
			trigger=1;
			if (mi.memory_track_flag>=sCheckClobber)
				if ((cl=clobbered((char *)&p[1],p->size)))
					debugf(" (clobbered %s)",clob_str[cl]);
			debugf("\n");
		}
	
	
	if (mi.memory_track_flag>=sClearFree)
		for (indice=0; indice<N_HEADS; indice++)
			for (p = free_head[indice]; p; p = p->next)
				if (free_written_to((unsigned char *)&p[1],p->size))
					debugf("free memory %p, %ld written to\n",
						   &p[1], p->size);
	
	if (!trigger) debugf("All memory was freed at the end\n");
	debugf("maximum memory allocated so far %ld\n",mi.maxmem);

}

#if 0
static int4 
checkmem()
{
	hdr *p;
	long tot;
	long indice;
	
	if (mi.memory_track_flag < sTrackAll) return 0;
	for (indice=0; indice<N_HEADS; indice++)
		for (p = head[indice]; p; p = p->next)
			tot += p->size;
	return tot != mi.totmem;
}
#endif

static void 
myRemove(hdr * q, int4 freeing)
{	
	long indice;
	hdr *p;

	indice = q->size/4;
	if (indice>=N_HEADS)
		indice = N_HEADS-1;

	if (!head[indice])
		*(char *) 0 = 0;

	if (q == head[indice]) {
		head[indice] = q->next;
	} else {
		for (p = head[indice]; p->next && p->next != q; p = p->next)
			;
	
	/* crash me if q isnt in the list */
		if (p->next != q) {
			debugf("freeing ptr %p twice\n",(q+1));
			*(char *) 0 = 0;
		}
		p->next = q->next;
	}

	if (freeing && mi.memory_track_flag>=sNoReuse) {
		q->next = free_head[indice];
		free_head[indice] = q;
	}
}

#ifdef QUICK_LIST

#if DO_INLINE
static inline
#endif
long get_index(long size)
{
	return size;
}

#if DO_INLINE
static inline
#endif
long get_rounded_index(long size)
{
	return size;
}

#if DO_INLINE
inline
#endif
static void free_mem(hdr *h)
{
	long p=get_index(h->size);
	if (p>=N_FREE_HEADS)
		INTERNAL_FREE(h);
	else {
		h->next = free_head[p];
		free_head[p] = h;
	}
	
}

static int4 num_up=10;
#if DO_INLINE
inline
#endif
static hdr *reuse_mem(int4 size)
{
	hdr *h;
	int4 i;
	
	long p=get_rounded_index(size);
	if (p>=N_FREE_HEADS)
		return 0;
	for (i=0; i<num_up && p<N_FREE_HEADS; i++,p++) {
		if ((h=free_head[p])) {
			free_head[p]=h->next;
			return h;
		}
	}
	return 0;
	
}
#endif

void check_not_free(hdr *q)
{
	long indice;
	hdr *p;
	
	indice = q->size/4;
	if (indice>=N_HEADS)
		indice = N_HEADS-1;

	for (p = free_head[indice]; p && p != q; p = p->next)
		;
	
	/* crash me if q is in the list */
	if (p) {
		/* bogus under DOS do we get the whole pointer */
		debugf("freeing ptr %p twice\n",(q+1));
		*(char *) 0 = 0;
	}

}

Dll void APIENTRY smfree(void *ptr)
{
	hdr *q;
	int4 cl;

	s_InitLib(0,0,0);

	if (!ptr){
		/* x will free 0 ptrs and thinks thats ok */
		/* apparently ANSI says that it is OK - MAV */
		return;
	}

	
	if (mi.memory_track_flag == sNoChecks) {
#ifndef QUICK_LIST
		INTERNAL_FREE(ptr);
#else
		q=(hdr *)ptr;
		q--;
		free_mem(q);
#endif
	} else {
		q = (hdr *) ((long) ptr - sizeof(hdr));
		if (mi.memory_track_flag>=sCheckClobber) {
			if ((cl=clobbered((char *)ptr,q->size))) {
				debugf("%p (# %ld): clobbered data size:%ld %s allocated pointer\n", 
					   ptr, (long)q->alloc_num, (long)q->size, clob_str[cl]);
				*(char *)0=0;
			}
		}
		if (mi.memory_track_flag>=sCheckClobber)
			check_not_free(q);

		if (mi.memory_track_flag>=sCheckClobberAlways)
			check_all_clobber();
		if (mi.memory_track_flag >= sTrackAll)
			myRemove(q,1);
		mi.totmem -= q->size;
		if (mi.memory_track_flag<sNoReuse) {
			if (mi.memory_track_flag>=sCheckClobber)
				INTERNAL_FREE((void *)(((long)q)-N_RAND));
			else
				INTERNAL_FREE(q);
		} else {
			if (mi.memory_track_flag>=sClearFree)
				clear_mem(ptr,q->size);
		}
	}
}

/* This routine is meant to be called at debugging time */
long s_getsize(void *q)
{
	hdr *p;
	
	p=q;
	if(!p) return 0;
	else {
		p--;
		return p->size;
	}
}

void s_print_list(int4 i)
{
	hdr *p;
	/* bogus under DOS do we get the whole pointer*/
	for (p = free_head[i]; p; p = p->next)
		debugf("%p %ld  ",p,p->size);
	debugf("\n");
}


#if QUICK_LIST
void print_non_free(void)
{
	int4 i,num;
	long tot;
	hdr *p;
	int4 len[N_FREE_HEADS],length;
	
	if (mi.memory_track_flag == sNoChecks) {
		tot=0;
		num=0;
		for (i=0; i<N_FREE_HEADS; i++) {
			length=0;
			for (p = free_head[i]; p; p = p->next) {
				tot += p->size;
				length++;
				num++;
				if (get_index(p->size)!=i)
					debugf("problem in memory?\n");
			}
			len[i]=length;
		}
		
		debugf("total free but not free memory = %ld, num blocks not free =%ld\n",tot,num);
#ifdef STEVE		
		for (i=0; i<N_FREE_HEADS; i++) {
			debugf("%i  ",len[i]);
		}
		debugf("\n");
#endif		
		
		
		
	}
}
#endif


void smpurge(void)
{
#ifdef QUICK_LIST
	int4 i;
	hdr *p,*next;

	s_InitLib(0,0,0);
	
	if (mi.memory_track_flag<=sNoChecks) {
		
		for (i=0; i<N_FREE_HEADS; i++) {
			for (p = free_head[i]; p; p = next) {
				next=p->next;
				INTERNAL_FREE(p);
			}
			free_head[i]=0;
		}
	}	
#endif
}

Dll void APIENTRY
s_SetMemFlags(int4 num)
{
	if(!num) {
		sReturnError=0;
		errorFlag=0;
	}else {
		errorFlag = num;
		sReturnError = 1;
		mi.failhere = num;
	}
}

static int MAX_INCREMENT = 10000;
#if 1
#define TEST_FOR_HUGE
#endif

Dll void * APIENTRY smalloc(int4 size)
{
	void *r;
	hdr *q;
	long indice;

	s_InitLib(0,0,0);

#ifdef TEST_FOR_HUGE
	if(size >= MAX_INCREMENT)
		indice = 0;
#endif
	
	if (mi.memory_track_flag == sNoChecks) {
		/* should find in list */
#ifndef QUICK_LIST
		if(sReturnError && (rand()==size))
			return 0;
		q = INTERNAL_MALLOC(size);
		if (!q) {
			debugf("Failed malloc size %ld, n mallocs %ld\n",size,mi.nsmallocs);
			if(fail_callback)
				fail_callback(fail_user_data);
		}
		return q;
#else
		if(sReturnError && mi.nsmallocs==errorFlag)
			return 0;
		q = reuse_mem(size);
		if (!q) {
			q = (hdr *) INTERNAL_MALLOC(size+sizeof(hdr));
			if (!q) {
				smpurge();
				q = (hdr *) INTERNAL_MALLOC(size+sizeof(hdr));
				if (!q) {
					debugf("Failed malloc size %ld, n mallocs %ld\n",size,mi.nsmallocs);
					if(fail_callback)
						fail_callback(fail_user_data);
					return 0;
				}
			}
			q->size=size;
		}
		
		q->alloc_num=++mi.nsmallocs;
		return (void *)&q[1];
#endif
		
	}
	if (mi.nsmallocs+1 == mi.failhere)
		return 0;

	if (mi.memory_track_flag>=sCheckClobber)
		q = (hdr *) (INTERNAL_MALLOC(size + sizeof(hdr)+N_RAND*2));
	else
		q = (hdr *) INTERNAL_MALLOC(size + sizeof(hdr));
	
	if (!q) {
		debugf("Failed malloc size %ld, n mallocs %ld\n",size,mi.nsmallocs);
		if(fail_callback)
			fail_callback(fail_user_data);
		return 0;
	}
		
	if (mi.memory_track_flag>=sCheckClobber)
		add_clobber((char **)&q,size);
	r = (void *) &q[1];
	q->size = size;
	q->alloc_num=++mi.nsmallocs;
	if (q->alloc_num == mi.stophere)
		debugf("I found #%d\n",mi.stophere);
	indice = size/4;
	if (indice>=N_HEADS)
		indice = N_HEADS-1;
	q->next = head[indice];
	head[indice] = q;
	mi.totmem += size;	
	
	if(mi.sizelimit&&size>mi.sizelimit)
		halt("Sizelimit excedded\n");
	
	if(mi.memlimit&&mi.totmem>mi.memlimit){
		smfree(r);
		r=0;
	}
	if (mi.totmem > mi.maxmem) {
		mi.maxmem = mi.totmem;
	}
	if (mi.memory_track_flag>=sCheckClobberAlways)
		check_all_clobber();

	return r;
}

Dll void * APIENTRY smalloc0(int4 size)
{
	void *ptr;
	ptr=smalloc(size);
	if (ptr) memset(ptr,0,size);
	return ptr;
}


Dll void * APIENTRY
smrealloc(void *ptr, int4 size)
{
	hdr *q, *p;
	void *r;
	long indice;
	int4 cl;

	s_InitLib(0,0,0);

#ifdef TEST_FOR_HUGE
	if(size >= MAX_INCREMENT)
		indice = 0;
#endif
	
	if (mi.memory_track_flag == sNoChecks) {
#ifndef QUICK_LIST
		q=INTERNAL_REALLOC(ptr,size);

		if (!q) {
			debugf("returning 0 from smrealloc, size=%i\n",size);
			if(fail_callback)
				fail_callback(fail_user_data);

			return 0;
		}
		return q;
#else
		q=((hdr *)ptr)-1;
		
		if (q->size>=size)
			return ptr;
		p=reuse_mem(size);
		if (p) {
			memcpy((char *)(p+1),(char *)(q+1),(size_t)q->size);
			free_mem(q);
			return (void *)(p+1);
		} else {
			q = (hdr *)INTERNAL_REALLOC(q, size+sizeof(hdr));
			if (!q) {
				smpurge();
				q = (hdr *)INTERNAL_REALLOC(q, size+sizeof(hdr));

				if (!q) {
					debugf("returning 0 from smrealloc, size=%li\n",size);
					if(fail_callback)
						fail_callback(fail_user_data);

					return 0;
				}
			}
			q->size=size;
			return (void *)(q+1);
		}
#endif
	}
	
	q = (hdr *) ((long) ptr - sizeof(hdr));
	
	mi.totmem -= q->size;
	if (mi.memory_track_flag >= sTrackAll)
		myRemove(q,0);
	if (mi.memory_track_flag>=sCheckClobber) {
		if ((cl=clobbered((char *)ptr,q->size))) {
			halt("%p (# %ld): clobbered data size:%ld %s allocated pointer\n", 
					   ptr, (long)q->alloc_num, (long)q->size, clob_str[cl]);
		}
		q = (hdr *)(((long)ptr)-sizeof(hdr)-N_RAND);
		p = (hdr *) INTERNAL_REALLOC(q, size + sizeof(hdr)+N_RAND*2);
	} else
		p = (hdr *) INTERNAL_REALLOC(q, size + sizeof(hdr));
	
	if (!p) {
		debugf("returning 0 from smrealloc, size=%li\n",size);
		if(fail_callback)
			fail_callback(fail_user_data);
		return 0;
	}
	
	if (mi.memory_track_flag>=sCheckClobber)
		add_clobber((char **)&p,size);
	
	r = (void *) &p[1];
	if (r == mi.badptr)
		debugf("hi there\n");
	if (r == mi.badptr && size == mi.badsize)
		debugf("ho there\n");
	p->size = size;
	indice = size/4;
	if (indice>=N_HEADS)
		indice = N_HEADS-1;	
	p->next = head[indice];
	head[indice] = p;
	mi.totmem += size;	
	if(mi.memlimit&&mi.totmem>mi.memlimit){
		smfree(r);
		r=0;
	}
	if (mi.totmem > mi.maxmem) {
		mi.maxmem = mi.totmem;
	}

	return r;
}

void prlists(void)
{	
    /* This dumps the lists of memory blocks		*/
	long indice;
	int4 total=0, largest=0;
	long largestindex=0;
	int4 length=0;
	hdr *p;
	
	if (mi.memory_track_flag < sTrackAll) return;
	
	debugf("\n");
	for (indice=0; indice<N_HEADS; indice++){
		length=0;
		for (p=head[indice];p ;p = p->next) 
			length++;
		if (length>largest){
			largest=length;
			largestindex = indice;
		}
		total += length;
		debugf("%ld | ", length);
	}
	debugf("\n%ld Memory Blocks \n",total);
	debugf("Index %ld is the longest at %ld\n",largestindex, largest);
}

int4 checkptr(void *ptr)
{
	hdr *q,*p;
	long indice;
	
	if (!ptr) return 0;
	if (mi.memory_track_flag < sTrackAll) return 0;
	q = (hdr *) ((long) ptr - sizeof(hdr));
	indice = q->size/4;
	if (indice>=N_HEADS)
		indice = N_HEADS-1;
	
	for (p=head[indice];p && (p != q);p = p->next) ;
	if (!p) return -1;
	return 0;
}

Dll void APIENTRY s_ExitLib(void)
{
	smpurge();
	if (mi.memory_track_flag == sCountTot) smstat(0);
	if (mi.memory_track_flag >= sTrackAll) smprmem();
}

#ifdef SYS_OS_WIN16

void HUGEPTR *
smallocH(int4 size)
{
	void HUGEPTR *value;
	
	value = GlobalAllocPtr(GMEM_MOVEABLE, size);

	return value;
}

void HUGEPTR *
smreallocH(void *orig, int4 size)
{
	void HUGEPTR *value;

	value = GlobalReAllocPtr(orig, size, GMEM_MOVEABLE);
	return value;
}

void
smfreeH(void *ptr)
{
	GlobalFreePtr(ptr);
}

#endif
