
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
/*
// This file implements all the members of the PriorityQueue ADT
*/

#include "rafport.h"

#include "c-lib.h"
#include "rafutil.h"

#include "pq.h"

#define DEFAULT_SIZE 128

static void
pqUpHeap(PriorityQueue *that, int element_index)
{
	VoidPtr new_value = that->array[element_index];
	
	for(;(that->comparison_function(that->array[element_index >> 1], 
									new_value) < 0)
			 && (element_index  > 0); element_index >>= 1)
		that->array[element_index] = that->array[element_index >> 1];
	
	that->array[element_index] = new_value;
}

static void
pqDownHeap(PriorityQueue *that, int k)
{
	int j;
	
	VoidPtr v;
	v = that->array[k];
	
	while(k < that->num_elements / 2){
		j = (k > 0) ? k << 1 : 1;
		
		/*
		  // - 1 because we are considering adding 1 to it
		  */
		if((j < (that->num_elements - 1)) &&
		   (that->comparison_function(that->array[j], that->array[j + 1]) < 0))
			j++;
		
		if(that->comparison_function(v, that->array[j]) >= 0)
			break;
		
		that->array[k] = that->array[j];
		k = j;
	}
	
	that->array[k] = v;
}

void
pqCtor(PriorityQueue * that, ComparisonFunction input_comparison_function)
{
	that->comparison_function = input_comparison_function;
	that->max_elements = DEFAULT_SIZE;
	that->num_elements = 0;
	that->array = (VoidPtr *) smalloc(that->max_elements * sizeof(VoidPtr));
}

void
pqDtor(PriorityQueue *that) 
{ 
	smfree(that->array);
}

VoidPtr
pqFront(PriorityQueue *that)
{
	if(pqEmpty(that))
		halt("PriorityQueue::pqFront: accessed empty queue.\n");
	
	return that->array[0];
}

void
pqEnqueue(PriorityQueue *that, VoidPtr new_item)
{
	if(that->num_elements == that->max_elements){
		that->max_elements <<= 1;
		that->array = (VoidPtr *) 
			smrealloc(that->array, that->max_elements * sizeof(VoidPtr));
	}
	that->array[that->num_elements] = new_item;
	pqUpHeap(that, that->num_elements++);
}

VoidPtr
pqDequeue(PriorityQueue *that)
{
	if(pqEmpty(that))
		halt("PriorityQueue::pqDequeue: accessed empty queue.\n");      
	else {
		VoidPtr ret_val = pqFront(that);
		
		that->array[0] = that->array[--that->num_elements];
		pqDownHeap(that, 0);
		
		return ret_val;
	}
	
	/* NOT REACHED */
	return 0;
}

void
pqRemoveFront(PriorityQueue *that)
{
	if(pqEmpty(that)) 
		halt("PriorityQueue::pqRemoveFront: accessed empty queue.\n");
	else {
		that->array[0] = that->array[--that->num_elements];
		pqDownHeap(that, 0);
	}
}


int
pqEmpty(PriorityQueue *that)
{ 
	return that->num_elements == 0;
}

int
pqFull(PriorityQueue *that)
{ 
	return FALSE;
}

int
pqLength(PriorityQueue *that)
{ 
	return that->num_elements;
}

void
pqClear(PriorityQueue *that)
{ 
	that->num_elements = 0;
}


