
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/********************************************************************\
*                                                                    *
*  File:      utdll16.c                                              *
*                                                                    *
*  Comments:  16-bit side of the Universal Thunk                     *
*                                                                    *
*  Functions: LibMain                                                *
*             UTInit                                                 *
*             UTProc                                                 *
*             WEP                                                    *
*                                                                    *
\********************************************************************/

#include "thunk.h"

enum {  
	GETHWND = RUTILSTART,  /* to get the handle of the stub program */
	riReadT,
	riWriteT,
	riOpenT,
	riCloseT,
	riSeekT,
	riTellT,
	riGetcT,
	riGetsT,
	riPutsT,
	riEofT,
	s_InitLibT,
	s_ExitLibT,
	smallocT,
	smfreeT,
	smalloc0T,
	smreallocT,
	doallocT,

	s_SetMemFlagsT,
	s_MemCallbackT,
};

#ifdef W32SUT_16
static UT16CBPROC gpUTCBack = NULL;
static HWND ghWnd = NULL;
static HINSTANCE ghInst;
static MemoryFailureCallback mem_cb;
static void *mem_data=0;

/********************************************************************\
* Function: int FAR PASCAL LibMain( HINSTANCE, WORD, WORD, LPSTR )   *
*                                                                    *
*  Purpose: DLL entry point                                          *
*                                                                    *
* Comments: Loads the Win32 stub application                         *
*                                                                    *
*                                                                    *
\********************************************************************/
							   
int FAR PASCAL __export LibMain( HINSTANCE hInst, WORD wDataSeg, WORD cbHeapSize, LPSTR lpszCmdLine )
{
   WORD result;
   ghInst = hInst;

	result=WinExec( "formstub.exe", SW_SHOW );
	if (result<=32) goto error;
	return 1;
error:
	return 0;
}

/********************************************************************\
* Function: DWORD FAR PASCAL UTInit( UT16CBPROC, LPVOID )            *
*                                                                    *
*  Purpose: Universal Thunk initialization function                  *
*                                                                    *
* Comments: Called when UTRegister() is called by utdll32.dll. The   *
*           window handle for the Win32 stub program is obtained and *
*           saved in a global for cleanup purposes later on.         *
*           FreeLibrary() is also called for cleanup purposes, by    *
*           reducing the reference count (the DLL is also referenced *
*           in the 32-bit side of the Universal Thunk through the    *
*           call to UTRegister().                                    *
*                                                                    *
*                                                                    *
\********************************************************************/

int t1[]={1,1};
int t12[]={2,1,1};
int t123[]={3,1,1,1};       
int t1234[]={4,1,1,1,1};  
int t125[]={5,1,1,0,0,1};
int t13[]={3,1,0,1};
int t2[]={2,0,1};      
int t23[]={3,0,1,1};
int t234[]={4,0,1,1,1};    
int t2345[]={5,0,1,1,1,1};
int t245[]={5,0,1,0,1,1};    
int t24[]={4,0,1,0,1};
int t25[]={5,0,1,0,0,1};          
int t235[]={5,0,1,1,0,1};          
int t3[]={3,0,0,1};                          
int t34[]={4,0,0,1,1};
int t345[]={5,0,0,1,1,1};
int t456[]={6,0,0,0,1,1,1}; 
int t1356[]={6,1,0,1,0,1,1}; 
int t1345[]={5,1,0,1,1,1}; 
int t135[]={5,1,0,1,0,1}; 
int t136[]={6,1,0,1,0,0,1}; 
int t4[]={4,0,0,0,1};                      
int t5[]={5,0,0,0,0,1};

DWORD FAR PASCAL _export
UTInit( UT16CBPROC pfnUT16CallBack, LPVOID lpBuff )
{
	gpUTCBack = pfnUT16CallBack;
	ghWnd = (*gpUTCBack)(NULL, GETHWND, NULL);
	FreeLibrary( ghInst );

	return TRUE;
}

DWORD
s_CSwitch(LPDWORD lp,DWORD dwFunc)
{
	switch (dwFunc) {
	case s_MemCallbackT:
		return mem_cb(mem_data);
	default:
		halt("bad callback function selector in rutil\n");
	}
	return 0;
}

DWORD FAR PASCAL _export
UTProc(LPVOID lpBuf, DWORD dwFunc)
{
	if (dwFunc < RUTILSTART) halt("bad function number\n");
	if (dwFunc < DAFSTART) return s_CSwitch((LPDWORD)lpBuf,dwFunc);
	if (dwFunc < FORMSTART) return i_CSwitch((LPDWORD)lpBuf,dwFunc);
	return f_CSwitch((LPDWORD)lpBuf,dwFunc);
}

DWORD
thunkIt(int Fnc, int *thunks, DWORD Args[])
{
	DWORD  rtn;

	/* arbitrarily set to a max of 12 arguments */
	LPVOID TransList[12];
	if(thunks){
		int i,j,Num;
		Num=thunks[0];
		thunks++;
				
		for(i = 0,j=0; i < Num; i++) {
			if (thunks[i]) {
				TransList[j] = &Args[i];
				j++;
			}
		}
		TransList[j] = NULL;
		rtn = (*gpUTCBack) (Args, Fnc, TransList);
	} else {
		TransList[0]=0;
		rtn = (*gpUTCBack) (Args, Fnc, TransList);
	}
	return rtn;
}

int FAR PASCAL _export
WEP(int nParam)
{
	if (ghWnd)
		PostMessage(ghWnd, WM_CLOSE, 0, 0);
	return TRUE;
}    

Dll void APIENTRY doalloc(void)
{
	DWORD Args[2];
	uint4 **foo=0,*p;
	uint4 i,quan;

	Args[0]=(DWORD)&foo;
	Args[1]=(DWORD)&quan;
	thunkIt(doallocT,t12,Args);
	debugf("Alloced %ld\n",quan);
	foo = (uint4**)UTLinearToSelectorOffset((void*)foo);
	for (i=0;i<quan;i++) {
		p = (uint4*)UTLinearToSelectorOffset((void*)foo[i]);
		if (!p) {
			debugf("failed thunk on %ld\n",i);
			break;
		}
		if (*p != (uint4)foo[i]) {
			debugf("Data is not right on %ld\n",i);
			break;
		}
	}
	foo = (uint4**)UTSelectorOffsetToLinear((void*)foo);
	Args[0]=(DWORD)&foo;
	Args[1]=(DWORD)&quan;
	thunkIt(doallocT,t12,Args);
}

Dll int4  APIENTRY riRead(char *ptr, int4 size, int4 nitems, RafFile f)
{
	DWORD   Args[4];
	Args[0]=(DWORD)ptr;
	Args[1]=(DWORD)size;
	Args[2]=(DWORD)nitems;
	Args[3]=(DWORD)f;
	return (int4) thunkIt(riReadT,t1,Args);
}              

Dll int4  APIENTRY riWrite(const char *ptr, int4 size, int4 nitems, RafFile f)
{
	DWORD   Args[4];
	Args[0]=(DWORD)ptr;
	Args[1]=(DWORD)size;
	Args[2]=(DWORD)nitems;
	Args[3]=(DWORD)f;
	return (int4) thunkIt(riWriteT,t1,Args);
}

Dll RafFile  APIENTRY riOpen(const char *filename, const char *type)
{
	DWORD   Args[2];
	Args[0]=(DWORD)filename;
	Args[1]=(DWORD)type;
	return (RafFile) thunkIt(riOpenT,t12,Args);
}

Dll int4  APIENTRY riClose(RafFile f)
{
	DWORD   Args[1];
	Args[0]=(DWORD)f;
	return (int4) thunkIt(riCloseT, NULL, Args);
}    


Dll int4  APIENTRY riSeek(RafFile handle, long offset, int4 whence)
{
	DWORD   Args[3];
	Args[0]=(DWORD)handle;
	Args[1]=(DWORD)offset;
	Args[2]=(DWORD)whence;
	return (int4) thunkIt(riSeekT,NULL,Args);
} 


Dll long  APIENTRY riTell(RafFile handle)
{
	DWORD   Args[1];
	Args[0]=(DWORD)handle;
	return (long) thunkIt(riTellT,NULL,Args);
} 

Dll int4  APIENTRY riGetc(RafFile handle)
{
	DWORD   Args[1];
	Args[0]=(DWORD)handle;
	return (int4) thunkIt(riGetcT,NULL,Args);
}

Dll char * APIENTRY riGets(char *s, int4 size, RafFile handle)
{
	DWORD   Args[3];
	Args[0]=(DWORD)s;
	Args[1]=(DWORD)size;
	Args[2]=(DWORD)handle;
	return (char *) UTLinearToSelectorOffset((LPBYTE)thunkIt(riGetsT,t1,Args));
}

Dll int4  APIENTRY riPuts(char *s, RafFile handle)
{
	DWORD   Args[2];
	Args[0]=(DWORD)s;
	Args[1]=(DWORD)handle;
	return (int4) thunkIt(riPutsT,t1,Args);
}

Dll int4  APIENTRY riEof(RafFile handle)
{
	DWORD   Args[1];
	Args[0]=(DWORD)handle;
	return (int4) thunkIt(riEofT,NULL,Args);
} 

Dll void APIENTRY s_InitLib(sMemFlags flags, 
								MemoryFailureCallback callback, 
								void *user_data)
{
	DWORD Args[3];
	
	mem_cb = callback;
	mem_data = user_data;
	Args[0] = (DWORD)flags;
	Args[1] = (DWORD)callback;
	Args[2] = (DWORD)user_data;
	thunkIt(s_InitLibT,NULL,Args);
}

Dll	void APIENTRY s_ExitLib(void)
{
	thunkIt(s_ExitLibT,NULL,NULL);
}

Dll void APIENTRY smfree(void *ptr)
{
	DWORD Args[1];
	
	Args[0] = (DWORD)ptr;
	thunkIt(smfreeT,t1,Args);
}

Dll	void * APIENTRY smalloc(int4 size)
{
	DWORD Args[1];
	
	Args[0] = (DWORD)size;
	return UTLinearToSelectorOffset((void*)thunkIt(smallocT,NULL,Args));
}

Dll	void * APIENTRY smalloc0(int4 size)
{
	DWORD Args[1];
	
	Args[0] = (DWORD)size;
	return UTLinearToSelectorOffset((void*)thunkIt(smalloc0T,NULL,Args));
}

Dll	void * APIENTRY smrealloc(void *ptr, int4 size)
{
	DWORD Args[2];
	
	Args[0] = (DWORD)ptr;
	Args[1] = (DWORD)size;
	return UTLinearToSelectorOffset((void*)thunkIt(smreallocT,t1,Args));
}

Dll void APIENTRY
s_SetMemFlags(int4 num)
{
	DWORD Args[1];
	
	Args[0] = (DWORD)num;
	thunkIt(s_SetMemFlagsT,NULL,Args);
}

#endif
#ifdef W32SUT_32
HWND     ghWnd;
HANDLE   ghDLL;
UT32PROC pfnUTProc = NULL;

DWORD APIENTRY UTCBProc( LPVOID, DWORD );


BOOL APIENTRY 
DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved)
{
   static int cAttached = 0;

   switch (dwReason)
   {
      case DLL_PROCESS_ATTACH:
	 if( !cAttached )
	    ghDLL = hDLL;
	 cAttached++;
	 break;

      case DLL_THREAD_ATTACH:
	 break;

      case DLL_THREAD_DETACH:
	 break;

      case DLL_PROCESS_DETACH:
	 --cAttached;
	 if( !cAttached ) {
	    UTUnRegister( hDLL );
	 }
	 break;
   }

   return TRUE;
}

__declspec(dllexport) void APIENTRY
InitThunk(HWND hWnd)
{
   char buffer[50]; 
   ghWnd = hWnd;

   if(UTRegister( ghDLL,
		   "form_t16.dll",
		   "UTInit",
		   "UTProc",
		   &pfnUTProc,
		   UTCBProc,
		   NULL))     
		return;
	wsprintf(buffer, "InitThunk: error %i\n", GetLastError()); 
	OutputDebugString(buffer);
}

DWORD CBThunkIt(DWORD Args[],int func)
{
	return (*pfnUTProc)(Args,func,NULL);
}

int4 APIENTRY mem_cb(void *data)
{
	DWORD Args[1];
	
	Args[0] = (DWORD)data;
	return CBThunkIt(Args,s_MemCallbackT);
}

Dll void APIENTRY doalloc(uint4 ***fred,uint4 *quanp)
{
	uint4 **foo;
	uint4 i,quan;
#define QUAN 4000

/*	debugf("Doing the test %lx, %lx\n",fred,quanp);
*/	if (fred && *fred) {
/*		debugf("Freeing %ld\n",*quanp);
*/		foo = *fred;
		for (i=0;i<*quanp;i++)
			smfree(foo[i]);
		smfree(foo);
		return;
	}
	foo = smalloc(QUAN * sizeof(uint4*));
	if (!foo) {
		OutputDebugString("failed to alloc foo\n");
		return;
	}
	for (i=0;i<QUAN;i++) {
		foo[i] = smalloc(3000);
		if (!foo[i]) break;
		*foo[i] = (uint4)foo[i];
	}
	quan = i;
	if (i != QUAN) debugf("failed at %ld\n",quan);
	if (quanp) *quanp = i;
	if (fred) *fred = foo;
	else {
		for (i=0;i<quan;i++)
			smfree(foo[i]);
		smfree(foo);
	}
}

DWORD
s_Switch(LPDWORD lp, DWORD dwFunc) 
{ 
	switch( dwFunc ) {
	case GETHWND:
		return (DWORD) ghWnd;
	case riReadT:
		return (DWORD) riRead((char *)lp[0],( int4 )lp[1],( int4 )lp[2], (RafFile)lp[3]);  
	case riWriteT:
		return (DWORD) riWrite((const char *)lp[0],( int4 )lp[1],( int4 )lp[2], (RafFile)lp[3]); 
	case riOpenT:
		return (DWORD) riOpen((const char *)lp[0],( const char *)lp[1]);        
	case riCloseT:
		return (DWORD) riClose((RafFile)lp[0]);                              
	case riSeekT:
		return (DWORD) riSeek((RafFile )lp[0],( long )lp[1],( int4 )lp[2]); 
	case riTellT:
		return (DWORD) riTell((RafFile )lp[0]); 
	case riGetcT:
		return (DWORD) riGetc((RafFile )lp[0]);
	case riGetsT:
		return (DWORD) riGets((char *)lp[0],( int4 )lp[1],( RafFile )lp[2]); 
	case riPutsT:
		return (DWORD) riPuts((char *)lp[0],( RafFile )lp[1]); 
	case riEofT:
		return (DWORD) riEof((RafFile )lp[0]); 
	case s_InitLibT:
		if (lp[1])
			s_InitLib((sMemFlags)lp[0],mem_cb,0);
		else s_InitLib((sMemFlags)lp[0],0,0);
		break;
	case s_ExitLibT:
		s_ExitLib();
		break;
	case smallocT:
		return (DWORD)smalloc(lp[0]);
	case smalloc0T:
		return (DWORD)smalloc0(lp[0]);
	case smreallocT:
		return (DWORD)smrealloc((void*)lp[0],lp[1]);
	case smfreeT:
		smfree((void*)lp[0]);
		break;
	case s_SetMemFlagsT:
		s_SetMemFlags((int4)lp[0]);
		break;
	case doallocT:
		doalloc((uint4***)lp[0],(uint4*)lp[1]);
		break;
	default:
		halt("bad function selector in rutil range\n");
   } 
   return TRUE; 
} 

DWORD APIENTRY
UTCBProc(LPVOID lpBuf, DWORD dwFunc) 
{
	if (dwFunc < RUTILSTART) halt("bad function number\n");
	if (dwFunc < DAFSTART) return s_Switch((LPDWORD)lpBuf,dwFunc);
	if (dwFunc < FORMSTART) return i_Switch((LPDWORD)lpBuf,dwFunc);
	return f_Switch((LPDWORD)lpBuf,dwFunc);
}
#endif
