
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*
// This code was designed to get around the 64 k limit under windows
// I am not sure it is used anymore
*/


/*
// 16 bit windows is differently abled because its fwrite is
// like this
// size_t fwrite(char *, size_t, size_t, FILE *)
// where size_t is only 2 bytes
// and we want to feed in 4 bytes
*/

#include "rafport.h"

#if defined(SYS_OS_WIN16) || defined(MAVERICK)

#include <stdio.h>
#include "rafutil.h"
#include "rafio.h"
#include "rafios.h"


#define MAX_SHORT 1 << 14


int4
streamRead(char HUGEPTR *ptr, int4 size, int4 nitems, RafFile handle)
{
	FILE *stream = (FILE *) handle;
	int4 got_bytes;
	int4 i;
	int4 incr = MAX_SHORT;
	int4 last_incr;
	int4 read_incr;
	int4 want_bytes = size * nitems;
	

   
	for(i = 0, got_bytes = 0; i+incr < want_bytes; i += incr){
		read_incr = (int4) fread(ptr + i, (size_t) 1, 
								 (size_t) incr, stream);
		got_bytes += read_incr;
	}
	last_incr = want_bytes - i;
	if(last_incr) {
		read_incr = (int4) fread(ptr + i, (size_t) 1, 
								
 (size_t) last_incr, stream);
		got_bytes += read_incr;
	}

	return got_bytes / size;
}

int4
streamWrite(const char HUGEPTR *ptr, int4 size, int4 nitems, RafFile handle)
{
	FILE *stream = (FILE *) handle;
	int4 got_bytes;
	int4 i;
	int4 incr = MAX_SHORT;
	int4 last_incr;
	int4 write_incr;
	int4 want_bytes = size * nitems;
	

   
	for(i = 0, got_bytes = 0; i+incr < want_bytes; i += incr){
		write_incr = (int4) fwrite(ptr + i, (size_t) 1, 
								 (size_t) incr, stream);
		got_bytes += write_incr;
	}
	last_incr = want_bytes - i;
	if(last_incr) {
		write_incr = (int4) fwrite(ptr + i, (size_t) 1, 
								 (size_t) last_incr, stream);
		got_bytes += write_incr;
	}

	return got_bytes / size;
}

RafFile
streamOpen(const char *filename, const char *type)
{
	FILE *stream;
	stream = fopen(filename, type);
	return (RafFile) stream;
}

int4
streamClose(RafFile handle)
{
		FILE *stream = (FILE *) handle;
		return fclose(stream);
}

int4
streamGetc(RafFile handle)
{
		FILE *stream = (FILE *) handle;
		return fgetc(stream);
}	

char *
streamGets(char HUGEPTR *s, int4 size, RafFile handle)
{
	/* remember, this is for win16 */
	char *p = 0;
	char *op; /* basically the value of the first fgets that works
				 will be s if not 0, but may be 0 if no fgets works */
	FILE *stream = (FILE *) handle;
	const int4 incr = 30000;
	int4 i;
	int done = FALSE; /* fgets returns nothing, or returns less than what
						 we asked for */

	if(size < incr)
		op = fgets(s, (size_t) size, stream);
	else {
		for(i = 0; i + incr < size && !done; i += incr){
			p = fgets(s + i, (size_t) incr, stream);
			if(p){
				/* if the first time fgets worked, assign op */
				op = op ? op : p;
				/* we got less then we wanted so we must have found a newline */
				if((int4)strlen(p) != incr - 1)
					done = TRUE;
			} else {
				done = TRUE;
			}
		}
		if(!done && (i + incr != size))
			/* op must already be assigned so return of this fgets not 
			   important */
			fgets(s + i, (size_t) (size - i), stream);
	}

	/* can't just return s because at least one fgets has to work */
	return op;
}

int4
streamPuts(char HUGEPTR *s, RafFile handle)
{
		FILE *stream = (FILE *) handle;
		return fputs(s, stream);
}

int4 
streamSeek(RafFile handle, long offset, int4 whence)
{
		FILE *stream = (FILE *) handle;
		return fseek(stream, offset, (int) whence);
}

long 
streamTell(RafFile handle)
{
	FILE *stream = (FILE *) handle;
	return ftell(stream);
}

int4
streamEof(RafFile handle)
{
	FILE *stream = (FILE *) handle;
	return feof(stream);
}

#endif
