
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*
// looks to be more RAF memory stuff, I don't know what the relation is
// to the memory.c file
*/


#include "rafport.h"
#include "rafutil.h"
#include "memdbg.h"

#ifdef USE_RAF_MALLOC

struct memhead {
	uint4 size;
	struct memhead *next,*prev;
};

#define QUAN 0x10L
#if 0
#define BIGGIE 0 /* (QUAN << 17) 2 Meg */
#define BIGCUT 0 /* (BIGGIE - QUAN - sizeof(uint4)) */
#else
#define BIGGIE (QUAN << 11) /* 32K */
#define BIGCUT (BIGGIE - QUAN - sizeof(uint4))
#endif

typedef struct memhead memhead,*memheadptr;

static memhead freelist = {0,&freelist,&freelist};
static int nchunks=0;

static void dumpheap(void)
{
	memheadptr p;
	uint4 tot = 0;
	
	for (p=freelist.next;p != &freelist;p = p->next) {
		debugf("0x%08lx at 0x%08lx\n",p->size,p);
		tot += p->size;
	}
	debugf("total = 0x%lx\n",tot);
}

static void freehead(memheadptr p,int checkkill)
{
	memheadptr q;
	int i;

	if (checkkill && p->size == BIGGIE) {
		nchunks--;
		free(p);
		return;
	}
	if (freelist.next > p && freelist.next->size) {
		p->prev = &freelist;
		p->next = freelist.next;
		q = &freelist;
	} else {
		for (q=freelist.next;q != &freelist && q->next < p;q = q->next) ;
		if (!q->size) { /* we are at freelist tack it on the end */
			p->prev = q->prev;
			p->next = q;
			q = p->prev;
		} else {
			p->prev = q;
			p->next = q->next;
		}
	}
	p->prev->next = p;
	p->next->prev = p;
	for (i=0;i<2;i++) {
		if (q->size && ((memheadptr)((uint4)q + q->size) == p)) {
			q->size += p->size;
			if (q->size > BIGGIE) halt("It's also too big\n");
			q->next = p->next;
			q->next->prev = q;
		} else
			q = p;
		p = q->next;
		if (q->size > BIGGIE) halt("It's too big\n");
		if (checkkill && (q->size == BIGGIE)) {
			q->next->prev = q->prev;
			q->prev->next = q->next;
			nchunks--;
			free(q);
			break;
		}
	}
}

static memheadptr allochead(uint4 s)
{
	memheadptr p,q;

	for (p = freelist.next;p != &freelist && p->size < s;p = p->next) ;
	if (p == &freelist) {
		p = malloc(BIGGIE+1); /* plus 1 so that if the os does put them */
		if (!p) {
			return 0;		/* next to each other I won't coalece them */
		}
		nchunks++;
		p->size = BIGGIE;
		freehead(p,0);
	}
	if (p->size == s) {
		p->prev->next = p->next;
		p->next->prev = p->prev;
	} else {
		q = (memheadptr)((uint4)p + s);
		q->size = p->size - s;
		q->prev = p->prev;
		q->next = p->next;
		q->prev->next = q;
		q->next->prev = q;
	}
	return p;
}

void *raf_malloc(uint4 s)
{
	memheadptr p;

	s = (s + sizeof(p->size) + QUAN-1) & ~(QUAN-1);
	if (s >= BIGCUT) {
		if (p = malloc(s)) nchunks++;
	} else
		p = allochead(s);
	if (!p) return 0;
	p->size = s;
	return (void*)&p->next;
}

void raf_free(void * oldp)
{
	memheadptr p = (memheadptr)((uint4)oldp - sizeof(uint4));
	if (p->size >= BIGCUT) {
		nchunks--;
		free(p);
	} else 
		freehead(p,1);
}

void *raf_realloc(void *oldp, uint4 s)
{
	memheadptr p = (memheadptr)((uint4)oldp - sizeof(uint4));
	void *new;

	if ((p->size-sizeof(p->size)) >= s)
		return oldp; /* we don't recover memory by shrinking */
	if (!(new = raf_malloc(s))) return 0;
	memcpy(new,oldp,p->size-4);
	if (p->size >= BIGCUT) {
		nchunks--;
		free(p);
	} else freehead(p,1);
	return new;
}

#endif
