
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "util/port.h"

#include "util/defines.h"
#include "util/util.h"

#include "set.h"

int 
setEqual(Set *that, Set *other)
{
	int ret_val;
	Ref r;
	VoidPtr item;
	
	if(setLength(that) != setLength(other))
		ret_val = FALSE;
	else {
		ret_val = TRUE;
		for(r = setFirst(that); r != 0 && ret_val; setNext(that, &r)){
			item = setContents(that, r);
			ret_val = setContains(other, item);
		}
	}
	
	return ret_val;
}

int 
setNotEqual(Set *that, Set *other)
{
	return !setEqual(that, other);
}

int 
setSubset(Set *that, Set *other)
{
	int ret_val;
	Ref r;
	VoidPtr item;
	
	if(setLength(that) > setLength(other))
		ret_val = FALSE;
	else {
		ret_val = TRUE;
		for(r = setFirst(that); r != 0 && ret_val; setNext(that, &r)){
			item = setContents(that, r);
			ret_val = setContains(other, item);
		}
	}
	
	return ret_val;
}

void 
setAddSet(Set *that, Set *other)
{
	Ref r;
	VoidPtr item;
	
	for(r = setFirst(other); r != 0; setNext(other, &r)){
		item = setContents(other, r);
		setAdd(that, item);
	}
}

void 
setSubtractSet(Set *that, Set *other)
{
	Ref r;
	VoidPtr item;
	
	for(r = setFirst(other); r != 0; setNext(other, &r)){
		item = setContents(other, r);
		if(setContains(that, item))
			setDelete(that, item);
	}
}

void 
setSubtractIntersectionComplement(Set *that, Set *other)
{
	Set copy_of_that;
	Ref r;
	VoidPtr item;
	
	setCtor(&copy_of_that);
	setAdd(&copy_of_that, that);
	
	for(r = setFirst(&copy_of_that); r != 0; setNext(&copy_of_that, &r)){
		item = setContents(&copy_of_that, r);
		if(setContains(other, item) ==  FALSE)
			setDelete(that, item);
	}
}

