
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*  This interprets the switches given on the command line
	
    For the most part all the command line switches can also be set with
    environment variables.
	*/

#include <stdio.h>
#include "rafport.h"
#include "rafio.h"
#include "c-lib.h"
#include "c-type.h"
#include "str-lib.h"
#include "rhash.h"
#include "ifc.h"

#define halt fool_preprocessor_again
#define debugf fool_preprocessor_once_more
#include "rafutil.h"
#undef halt
#undef debugf

void halt(char *s,long a, long b,long c, long d, long e);
void debugf(char *s,long a, long b,long c, long d, long e, long f, long g, long h, long i, long j, long k, long l, long m, long n);
extern char *getenv(const char *);

static int4 
settype(ifc *cp, char *p)
{
	int4 bad=0;
	long t;
	float f;

	switch (cp->type) {
	case IFC_CHAR:
		t = *(char *) cp->val = *p;
		bad = t < cp->lval || t > cp->fval;
		break;
	case IFC_SHORT:
		t = *(short *) cp->val = *p;
		bad = t < cp->lval || t > cp->fval;
		break;
	case IFC_INT:
		t = *(int4 *) cp->val = atoi(p);
		bad = t < cp->lval || t > cp->fval;
		break;
	case IFC_FLAG:
		t = *(int4 *) cp->val = atoi(p);
		break;
	case IFC_HEX:
#if defined(SYS_OS_WIN16) || defined(MAVERICK)
		t = *(int4 *) cp->val = strtol(p, (char **) NULL, 16);
#else
		sscanf(p,"%x", (unsigned int *) cp->val);
		t = *(int*)cp->val;
#endif
		bad = t < cp->lval || t > cp->fval;
		break;
	case IFC_LONG:
		t = *(long *) cp->val = atoi(p);
		bad = t < cp->lval || t > cp->fval;
		break;
	case IFC_FLOAT:
		/* bogus atof returns double */
		f = *(float*)cp->val = (float) atof(p);
		bad = f < (float)atof((char*)cp->lval) || f > (float)atof((char*)cp->fval);
		break;
	case IFC_STR:
		*cp->val = p;
		bad = 0;
		break;
	}
	return bad;
}

int4 stripext(int4 c,char *v[],char *ev[])
{
	int4 i,j,elen,which,count;
	
	which = -1;
	for (count=0,i=1;i<c;i++)
		if (v[i]) count++;
	for (j=0;*ev[j];j++) {
		elen = strlen(ev[j]);
		for (i=1;i<c;i++) {
			if (!v[i]) continue;
			if (!strcmp(&v[i][strlen(v[i]) - elen], ev[j])) {
				v[i][strlen(v[i]) - elen] = 0;
				which = j;
			}
		}
	}
	return which;
}

int4 processargs(int4 c, char *v[], ifc * ip)
{
	ifc *cp;
	int4 i,bad;
	size_t len;
	char *p;
	cp = ip;

	bad = 0;
	while (cp->type != IFC_NA) {
		switch(cp->type){
		case IFC_CHAR:
			*(char *) cp->val = (char) cp->init;
			break;
		case IFC_SHORT:
			*(short *) cp->val = (short) cp->init;
			break;
		  case IFC_LONG:
			  *(long *) cp->val = cp->init;
			  break;
		case IFC_FLOAT:
			/* bogus atof returns double */
			*(float*) cp->val = (float) atof((char*)cp->init);
			break;
		default:
			*(int4 *) cp->val = (int4) cp->init;
			break;
		}
		if (cp->env && (p = getenv(cp->env)))
			bad |= settype(cp, p);
		cp++;
	}
	for (i = 1; i < c; i++) {
		if (v[i][0] == '-') {
			cp = ip;
			while (cp->type != IFC_NA) {
				len = strlen(cp->sw);
				if (!strncmp(&v[i][1], cp->sw, len) &&
					(cp->type != IFC_FLAG || !v[i][len+1]))
					break;
				cp++;
			}
			if (cp->type == IFC_NA)
				return -1;
			if (cp->type != IFC_FLAG) {
				if (v[i][1 + len])
					bad |= settype(cp, &v[i][1 + len]);
				else {
					v[i++] = 0;
					bad |= settype(cp, v[i]);
				}
			} else
				*(int4 *) cp->val = (int4) cp->fval;
			v[i] = 0;
		}
	}
	return bad;
}

#if !defined(SYS_OS_WIN16) && !defined(SYS_OS_WIN32)
static void
printError(char *s,long a, long b,long c, long d, long e)
{
    fprintf(stderr,s,a,b,c,d,e);
    fflush(stderr);
}

void
debugf(char *s,long a, long b,long c, long d, long e, long f, long g, long h, long i, long j, long k, long l, long m, long n)
{
    fprintf(stdout,s,a,b,c,d,e,f,g,h,i,j,k,l,m,n);
    fflush(stdout);
}

#else
#include <windows.h>

static void
printError(char *s,long a, long b,long c, long d, long e)
{
	char buffer[1024];
	wsprintf(buffer, s, a, b, c, d, e);
	MessageBox(NULL, buffer, "Error", MB_OK);
}

void
debugf(s,a,b,c,d,e,f,g,h,i)
char *s;
long a,b,c,d,e,f,g,h,i;
{
	char buffer[1024];
	wsprintf(buffer, s, a,b,c,d,e,f,g,h,i);
	OutputDebugString(buffer);
}

#endif

void
halt(char *s,long a, long b,long c, long d, long e)
{
    printError(s,a,b,c,d,e);
#if !defined(SYS_OS_WIN16)
    exit(1);
#else
	{
		/* crash me */
		char *s = 0;
		*s = 0;
	}
#endif
}

int4
fileSizeInBytes(const char *file_name)
{
	RafFile file;
	long length;
	
#ifdef SYS_TRUESCAN
	halt("fileSizeInBytes: no file operations on TRUE SCAN.\n");
#else
	file = riOpen(file_name, "r");
	if(file != 0) {
		riSeek(file, 0, 2);
		length = riTell(file);
		riClose(file);
	} else {
		length = -1;
	}
#endif
	/* probably has to be long */
	return (int4) length;
}

