
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include "FileName.h"

FileName::FileName()
{
}

FileName::FileName(const FileName &c)
{
	this->extend(c.length());
	len = c.length();
	strcpy(data,c.data);
}


FileName& FileName::operator =(const char *c)
{
	if(!c) 
		c = EmptyString;
	
	unsigned l = strlen(c);
	this->extend(l + 1);
	len = l;
	strcpy(data,c);
	
	return *this;
}

FileName& FileName::operator =(const aString &c)
{
	if(this == &c)
		return *this;
	this->extend(c.length() + 1);
	len = c.length();
	strcpy(data,c.data);
	
	return *this;
}

FileName& FileName::operator =(const FileName &c)
{
	if(this == &c)
		return *this;
	this->extend(c.length() + 1);
	len = c.length();
	strcpy(data,c.data);
	
	return *this;
}


static const char *
backfind(const char *data,int len,char c,char not)
{
	const char *p;
	
	if (!len)
		return 0;
	
	p = data + len - 1;
	while (*p != c && (!not || *p != not) && p != data) p--;
	if (*p != c) p = 0;
	return p;
}

int
FileName::existsP() const
{
	struct stat buf;
	int errcode;
	errcode = stat(data, &buf);
	return (errcode == 0);
}

static int isdir(char *p)
{
	struct stat buf;
	
	if (!stat(p, &buf))
		return S_ISDIR(buf.st_mode);
	return 0;
}

int
FileName::directoryP() const
{
	return isdir(data);
}

int
FileName::relativeP() const
{
	return (*data != '/' && *data != '~');
}

aString
FileName::extension() const
{
	const char *p;
	
	p = backfind(data,len,'.','/');
	if (p) p++;
	return aString(p);
}

aString
FileName::filename() const
{
	const char *p;
	
	p = backfind(data,len,'/',0);
	if (p) p++;
	else p = data;
	return aString(p);
}

aString
FileName::directory() const
{
	// must show this to John
	// as originally written destroys original
	// modifies original temporarily
	
	// why is that const doesnt stop this from happening
	// because data was protected -- not *data
	aString a;
	char *p;
	
	p = backfind(data,len,'/',0);
	if (!p) return aString(".");
	
	// replace with a / with null so that a sees only directory
	*p = 0;
	a = data;
	
	// replace this character which I know to be a /
	*p = '/';
	
	return a;
}

int
FileName::nextprev(int dir)
{
	DIR *dirp;
	struct dirent *dp;
	int cmp;
	aString best,file,ext;
	const char *p;
	
	file = directory();
	ext = extension();
	dirp = opendir(file.string());
	if (!dirp) return -1;
	file = filename();
	while (dp = readdir(dirp)) {
		p = backfind(dp->d_name,strlen(dp->d_name),'.','/');
		if ((!ext.len && !p) || (ext.len && p && !strcmp(ext.string(),p+1))) {
			cmp = strcmp(dp->d_name,file.string());
			if ((dir && (cmp > 0) && (!best.len || (strcmp(dp->d_name,best.string())) < 0)) ||
				(!dir && (cmp < 0) && (!best.len || (strcmp(dp->d_name,best.string()) > 0)))) {
				if (isdir(dp->d_name)) continue;
				best = dp->d_name;
			}
		}
	}
	closedir(dirp);
	if (!best.len) return 1;
	file = this->directory() + "/";
	file += best.string();
	*this = file;
	return 0;
}

int
FileName::nextFile()
{
	return nextprev(1);
}

int
FileName::prevFile()
{
	return nextprev(0);
}


