
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#include "path.h"

extern "C" int split_path(const char *path,char *filename, char *dirname)
{
	if(!path || !path[0]) {
		filename[0] = dirname[0] = 0;
		return(1); // error
	}
	int len = strlen(path);
	assert(len < MAX_PATH_LENGTH);
	
	// No directory specified
	if(!strchr(path,'/')) {
		strcpy(dirname,".");
		strcpy(filename,path);
		return(0);
	}
	
	strcpy(dirname,path);
	len--;
	while(len >= 0) {
		if(dirname[len] == '/') {
			strcpy(filename,&dirname[len + 1]);
			dirname[len] = 0;
			return(0);
		}
		len--;
	}
	// No official directory so dirname = "."
	strcpy(filename,dirname); 
	strcpy(filename,"."); 
	return(0);
}

extern "C" int split_ext(const char *filename,char *basename,char *extname)
{
	if(!filename) {
		basename[0] = extname[0] = 0;
		return(1); // error
	}
	int len = strlen(filename);
	assert(len < MAX_PATH_LENGTH);
	
	strcpy(basename,filename);
	extname[0] = 0;
	len--;
	while(len >= 0) {
		if(basename[len] == '/') break; // reached directory so stop
		if(basename[len] == '.') {
			strcpy(extname,&basename[len + 1]);
			basename[len] = 0;
			return(0);
		}
		len--;
	}   
	return(0);
}

extern "C" int  get_path_body(const char *pathname,char *basename)
{
	char dir[MAX_PATH_LENGTH];
	char file[MAX_PATH_LENGTH];
	char ext[MAX_PATH_LENGTH];
	if(split_path(pathname,file,dir)) return(1);
	return(split_ext(file,basename,ext));
}

extern "C" int  get_path_ext(const char *pathname,char *ext)
{
	char dir[MAX_PATH_LENGTH];
	char file[MAX_PATH_LENGTH];
	char basename[MAX_PATH_LENGTH];
	if(split_path(pathname,file,dir)) return(1);
	return(split_ext(file,basename,ext));
}

extern "C" void add_file_extension(const char *filename,const char *extname,char *outname)
{
	assert(outname);
	sprintf(outname,"%s.%s",filename,extname);
}


extern "C" void add_body_extension(const char *filename,const char *extname,char *outname)
{
	char ext[MAX_PATH_LENGTH];
	char body[MAX_PATH_LENGTH];
	assert(outname);
	split_ext(filename,body,ext);
	sprintf(outname,"%s.%s",body,extname);
}

aString PathBody(const char *pathname)
{
	char body[MAX_PATH_LENGTH];
	body[0] = 0;
	get_path_body(pathname,body);
	return(aString(body));
}

aString FileBody(const char *pathname)
{
	char filename[MAX_PATH_LENGTH];
	char dirname[MAX_PATH_LENGTH];
	char body[MAX_PATH_LENGTH];
	filename[0] = dirname[0] = body[0] = 0;
	split_path(pathname,filename, dirname);
	if(filename[0])
		get_path_body(filename,body);
	return(aString(body));
}

aString PathDirectory(const char *pathname)
{
	char filename[MAX_PATH_LENGTH];
	char dirname[MAX_PATH_LENGTH];
	
	filename[0] = dirname[0] = 0;
	split_path(pathname,filename, dirname);
	return(aString(dirname));
}

aString PathExtension(const char *pathname)
{
	char filename[MAX_PATH_LENGTH];
	char extname[MAX_PATH_LENGTH];
	
	filename[0] = extname[0] = 0;
	split_ext(pathname,filename, extname);
	return(aString(extname));
}
