
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#include <stddef.h>
#include <assert.h>

#include "smemory.h"

#include "uString.h"
#include "aString.h"

#ifndef toupper
#define toupper(x) (((x) >= 'a' && (x) <= 'z') ? ((x) + 'A' - 'a') : (x))

#define tolower(x) (((x) >= 'A' && (x) <= 'Z') ? ((x) + 'a' - 'A') : (x))
#endif toupper
#ifndef min
#define min(x,y) ((x<y) ? x : y)
#define max(x,y) ((x>y) ? x : y)
#endif

const iUCode *EmptyUString = (iUCode *) L"";


#include "utf.h"

void
uString::putUtf(aString &as) const
{
	char buffer[256];
	ucs2utf(buffer, this->string());
	as = buffer;
}

int
uString::asciiP()
{
	int i;
	
	for(i = 0; i < len; i++)
		if(data[i] & 0xff80)
			return 0;
	return 1;
}

void
uString::getUnicode(const aString &as)
{
	as.putUnicode(*this);
}

int
uString::utfLength()
{
	aString as;
	as.getUtf(*this);
	return as.length();
}

uString::uString(const aString &as)
	: allocated(0),
	  len(0),
	  data(NULL)
{
	this->getUnicode(as);
}

uString::uString(const char *s)
	: allocated(0),
	  data(NULL),
	  len(0)
{
	aString as=s;
	this->getUnicode(as);
}

uString::~uString(void)
{
	if (data)
		smfree(data);
	//I have trouble with delete[]
}

uString::uString(iUCode c)
	: allocated(0),
	  data(NULL),
	  len(0)
{
	this->extend(1);
	len = 1;
	data[0] = c;
	data[1] = 0;
}

uString::uString(void)
	: allocated(0),
	  data(NULL),
	  len(0)
{
	this->extend(1);
	len = 0;
	data[0] = 0;
}

uString::uString(const uString &c)
	: allocated(0),
	  data(NULL),
	  len(0)
{
	this->extend(c.length());
	len = c.length();
	ucscpy(data, c.data);
}
uString::uString(const iUCode *c)
	: allocated(0),
	  data(NULL),
	  len(0)
{
	if (!c)
		c = EmptyUString;
	
	len = ucslen(c);
	this->extend(len);
	ucscpy(data, c);
}

void
uString::move(unsigned start, unsigned length)
{
	// check arguments
	if(start > len) 
		return;

	unsigned newlen = len + length;
	this->extend(newlen);

	// add 1 for the null terminator
	memmove(data+start+length, data+start, sizeof(iUCode)*(len-start+1));

	len += length;
}

void
uString::remove(unsigned start, unsigned length)
{
	// check arguments
	if(start > len) 
		return;
	if(start+length > len)
		length = len - start;

	// add 1 for the null terminator
	memmove(data+start, data+start+length, sizeof(iUCode)*(len-length-start+1));

	len -= length;
}

void
uString::reverse(void)
{
       for(int i = 0; i < len/2; i++) {
                iUCode temp = data[i];
		data[i] = data[len-i];
		data[len-i] = temp;
       }
}

uString
uString::mid(int start, int stop) 
{
      uString mid_string;
      int mid_length = stop-start+1;
      mid_string.extend(mid_length+1);
      ucsncpy(mid_string.data, data+start, mid_length);
      mid_string.len = mid_length;

      return mid_string;
}  

iUCode &
uString::operator[](unsigned index)
{
	if(index < len)
		return *(data+index);
	else
		return *(data+len);
}

iUCode
uString::operator[](unsigned index) const
{
	if(index < len)
		return *(data+index);
	else
		return *(data+len);
}

uString&
uString::operator = (const iUCode *c) 
{
	if (!c)
		c = EmptyUString;
	
	unsigned l = ucslen(c);
	this->extend(l);
	len = l;
	ucscpy(data, c);
	
	return (*this);
}

uString& uString::operator = (const uString &c) 
{
	
	if (this == &c)
		return *this;
	this->extend(c.length());
	len = c.length();
	ucscpy(data, c.data);
	
	return (*this);
}

int
uString::operator != (const uString &c) const
{
	return ucscmp(c.data, data);
}

int uString::operator != (const iUCode *c) const
{
	if (!c)
		c = EmptyUString;
	
	return (ucscmp(c, data));
}


int uString::operator == (const uString& c) const
{
	return !ucscmp(c.data, data);
}

int uString::operator == (const iUCode *c) const
{
	if (!c)
		c = EmptyUString;
	return (!ucscmp(c, data));
}


int uString::operator <= (const iUCode *c) const
{
	if (!c)
		c = EmptyUString;
	return (ucscmp(data, c) <= 0);
}

int uString::operator < (const iUCode *c) const
{
	if (!c)
		c = EmptyUString;
	
	return (ucscmp(data, c) < 0);
}

int uString::operator > (const iUCode *c) const
{
	if (!c)
		c = EmptyUString;
	return (ucscmp(data, c) > 0);
}

int uString::operator >= (const iUCode *c) const
{
	if (!c)
		c = EmptyUString;
	return (ucscmp(data, c) >= 0);
}

int uString::operator <= (const uString &c) const
{
	return (ucscmp(data, c.data) <= 0);
}

int uString::operator < (const uString &c) const
{
	return (ucscmp(data, c.data) < 0);
}

int uString::operator > (const uString &c) const
{
	return (ucscmp(data, c.data) > 0);
}

int uString::operator >= (const uString &c) const
{
	return (ucscmp(data, c.data) >= 0);
}

uString uString::operator + (const uString& c) const
{
	uString a;
	a = *this;
	a += c;
	return a;
}


uString uString::operator + (const iUCode *c) const
{
	uString a;
	a = *this;
	a += c;
	return a;
}

uString &uString::operator -- (void) {
	if(len) len--;
	
	data[len] = 0;

	return (*this);
}

uString &uString::operator += (const uString &c) {
	unsigned newlen = len + c.length();
	if (newlen == len)
		return (*this);
	//+=null string
	this->extend(newlen);
	if (&c != this) {
		ucscpy(&data[len], c.data);
	} else {
		//special handling for +=self
		ucsncpy(&data[len], data, len);
		data[newlen] = 0;
	}
	len = newlen;
	return (*this);
	
}

uString &uString::operator += (const iUCode *c) {
	if (!c)
		return (*this);
	
	unsigned newlen = len + ucslen(c);
	if (newlen == len)
		return (*this);
	//+=null string
	if (newlen > allocated)
		this->extend(newlen);
	
	ucscpy(&data[len], c);
	len = newlen;
	return (*this);
}



void uString::clear(void)
{
	len = 0;
	this->extend(1);
	data[0] = '\0';
}

#ifndef max
#define max(x,y) ((x>y) ? x : y)
#endif

/*
// There were some bugs related to not asking for enough memory
// from extend -- confusiion about how to count the null
// therefore -- I always add at least 1 to requested length
// and report allocated as 1 less than actual
//
*/
void
uString::extend(unsigned newsize)
{
	// allocated stores the space for characters less null
	if (allocated >= newsize && allocated != 0)
		return;
	//big enough
	
	newsize = NextPower2(newsize + 1);
	newsize = max(newsize, 4);
	if (data) {
		data = (iUCode *) smrealloc(data, sizeof(iUCode) * newsize);
	} else {
		data = (iUCode *) smalloc(sizeof(iUCode) * newsize);
		data[0] = '\0';
	}
	
	allocated = newsize - 1;
}



//*************************************************************
//-Name Operators
//*************************************************************
int operator == (const iUCode *c, const uString& s) {
	if (!c)
		c = EmptyUString;
	return (ucscmp(c, s.data));
}

int operator < (const iUCode *c, const uString& s) {
	if (!c)
		c = EmptyUString;
	return (ucscmp(c, s.data) < 0);
}

int operator > (const iUCode *c, const uString &s) {
	if (!c)
		c = EmptyUString;
	return (ucscmp(c, s.data) > 0);
}

int operator >= (const iUCode *c, const uString &s) {
	if (!c)
		c = EmptyUString;
	return (ucscmp(c, s.data) >= 0);
}

int operator <= (const iUCode *c, const uString &s) {
	if (!c)
		c = EmptyUString;
	return (ucscmp(c, s.data) <= 0);
	return 0;
}

int operator != (const iUCode *c, const uString &s) {
	if (!c)
		c = EmptyUString;
	return (!ucscmp(c, s.data));
}


uString operator + (const iUCode *c, const uString &s) {
	uString r = c;
	r += s;
	return (r);
}


const iUCode *uString::string(void) const  { if(data) return(data); return(EmptyUString);  }	
int   uString::length(void) const { return(len);} 

