
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*
// The main part of representing text entities
// it's main faults -- assumes concentric entities instead of
// flowing entities
//
// what does this mean -- I assumed the whole content of a line
// was contained inside a line entity
// instead, lines were like were words, 
// the actual boundrys were caused by newlines
//
// so right now a lot of this code is wrong
//
// 941201 -- I believe I have now corrected this
// by using the reset flag more intelligently and
// looking to your sib for positional information
*/
#include "rafport.h"
#include "text.h"

#include "rafutil.h"
#include "smemory.h"
#include "ucs.h"
#include "TextElement.h"
#include "dafs.h"
#include "TextMargin.h"

/*
// this method is called when the entity is first parsed
// and whenever we are informed that our contents are
// changed
// it converts the dafs form of our data into a form that
// makes it easy for us to display and manipulate
// bogus -- have we considered the tradeoffs 
// of caching this info versus calculating it when needed
// have we tried to minimize the amount of processing
// maybe we need to reduce the processing or convince
// dafslib to store things in a friendlier manner
*/
void
TextElement::initializeText()
{
	iUCode *ubuf;
	char *buf;
	int t;
	
	if(this->text)
		smfree(this->text);

	// bogus -- under what circumstances
	// is there no entity
	// I seem to recall that it was possible because
	// I had some hack for the top level entity
	// and still haven't addressed multiple pages in 
	// a doc
	if(this->entity) {
		t = i_GetEDataType(this->entity);
		switch (t) {
		case iDPosSet:
		case iD1PosSet:
			ubuf=(iUCode *)smalloc(sizeof(iUCode)*2);
			if (!ubuf) halt("Can't allocate memory");
			ubuf[0] = i_GetGlyph(this->entity,0);
			if (ubuf[0] == '\t')
				ubuf[0] = ' ';
			else if (ubuf[0] == '\0')
				ubuf[0]='^';
			ubuf[1] = '\0';
			this->text = ubuf;
			break;
		case iDUString:
			ubuf = (iUCode*)i_GetUText(this->entity);
			if (ubuf) {
				this->text=(iUCode *)smalloc(sizeof(iUCode)*(ucslen(ubuf)+1));
				ucscpy(this->text,ubuf);
			} else 
				this->text = 0;
			break;
		case iDString:
			buf = (char*)i_GetText(this->entity);
			if (buf) {
				this->text=(iUCode *)smalloc(sizeof(iUCode)*(strlen(buf)+1));
				for (ubuf = this->text;*ubuf++ = *buf++;) ;
			} else 
				this->text = 0;
			break;
		}
	} else
		this->text=0;
}

// The application using this class must supply a function
// for measuring text.
static textMeasureFunction measureText;
extern void initializeMeasureFunction(textMeasureFunction f)
{
	measureText = f;
}

TextElement::TextElement(Page *input_page,
						 iEntityPtr input_entity, 
						 int input_type,
						 TextElement *input_parent, 
						 iPackType *input_pack)
	: EntityFace(input_page, input_entity),
	  type(input_type), 
	  position_valid(FALSE),
	  dimension_valid(FALSE),
	  pack(input_pack),
	  rx(0), ry(0), w(0), h(0),
	  text(0) // will really be set in initialize text
{
	setTextViewer(input_entity, this);
	
	// this is just a double check that was removed for 
	// efficiency's sake
#if 0
	int has_pack=!i_GetUserDataProp();
	
	if(has_pack == FALSE) {
		halt("didnt have packing for an element.\n");
	}
#endif
	this->initializeText();
}

TextElement::~TextElement()
{
	if (this->text) {
		smfree(this->text);
	}
}

void 
TextElement::deleteCheap()
{
	delete this;
}

/*
// deleteAndUpdate removed to illum and dafs2text because in the
// former case it had to know too much about illum
*/

/*
// hmm, what is this for?
*/
TextElement *
TextElement::findPrevElement()
{
	TextElement *current;
	TextElement *parent;
	TextElement *prev = 0;
	TextElement *l;
	
	current = this;
	parent = current->getParentElement();
	
	while(parent) {
		l = parent;
		Ref r = l->seek(current);
		l->prev(&r);
		if(r != 0) {
			prev = (TextElement *) l->contents( r);
			break;
		}
		current = parent;
		parent = current->getParentElement();
	}
	if(prev)
		while((l = prev)->length() != 0)
			prev = (TextElement *) l->contents( l->last());
	
	return prev;
}

TextElement *
TextElement::findNextElement()
{
	TextElement *current;
	TextElement *parent;
	TextElement *next = 0;
	TextElement *l;
	
	current = this;
	parent = current->getParentElement();
	
	while(parent) {
		l = parent;
		Ref r = l->seek(current);
		l->next( &r);
		if(r != 0) {
			next = (TextElement *) l->contents( r);
			break;
		}
		current = parent;
		parent = current->getParentElement();
	}
	
	if(next)
		while((l = next)->length() != 0)
			next = (TextElement *) l->contents( l->first());
	
	return next;
}

/*
// hmm, this seems in the style of Steve
// what is terminal? what is purpose?
*/
static TextElement *
last_terminal_element(TextElement *root)
{
	Ref r;
	TextElement *child;
	r = root->last();
	if(r) {
		child = (TextElement *) root->contents(r);
		return last_terminal_element(child);
	} else {
		return root;
	}
}

/*
// hmm is seems that a terminal is a counter
// or an element with text
// it seems like a terminal is a guy that we want to allow
// the cursor to be in
*/
static int 
is_terminal_element(TextElement *element)
{
	return (element->pack->flags & pfCounter) || (element->text != 0);
}

/*
// ok this is an aux to the function that tells where your mouse
// clicked
// you have to return not only an element but an index.
// in the case of a line of text or a word, the index is straightforward
// in the case of a Space or NewLine it means the number of the
// spaces from the left or the number of lines from the top of the newline
//
// bogus should replace charWidth and so on with max line height
// or with specific char height depending on whether we have VP
*/
static int 
terminalIndex(TextElement *element, int x, int y)
{
	int return_index=0;
	int max=0;
	if(element->pack->flags & pfCounter){
		max=i_GetShort(element->entity,0)-1;
		switch(element->pack->orientation) {
		case pfHoriz:
			// ok, we give space a fixed width 
			return_index = x/charWidth('M');
			break;
		case pfVert:
			return_index=y/charHeight('M');
			break;
		default:
			halt("Illegal orientation.\n");
		}
	} else if(element->text != 0){
		const iUCode *str=element->text;
		max=ucslen(str)-1;

		// now, we believe that an element can have 
		// newlines in it, we have to face the fact
		// that packing doesn't describe the text running in
		// it -- paras pack vertically, but their text runs
		// l to r t to b

		// first calculate the line

		int c;
		c = *str;
		int cheight = charHeight(c);

		// as long as we are not on the right line
		// advance
		while(*str && y >= cheight) {
			c = *str++;
			// increment the line when you see a newline
			if(c == '\n') {
				y = y - charWidth(c);
			}
			// advance no matter what, you know you are not on this line
			return_index++;
		}
		// now we are on the right line, advance as long as you can see
		// aren't at the x position, but stop if you are going onto a newline
		while(*str && x >= 0){
			c = *str++;
			x = x - charWidth(c);
			if(x >= 0)
				return_index++;
			if(c == '\n')
				break;
		}
	}
	if (return_index>max) 
		return_index=max;
	if (return_index < 0)
		return_index = 0;

	return return_index;
}

void TextElement::getElementAtPosition(int x, int y, 
				       TextElementPtr& return_element,
				       int& return_index,
				       int is_root, int over_end)
{
	// Get the element at position (offset) x, y relative to me.
	// Return element pointer and index of its sub-element at
	// the specified location.  

	// is_root means this is the first call (not recursion)

	// over_end means return the last terminal of a horizontally
	// packed element if we are to its right 

	short height, width;  // size of me
	Ref r;
	TextElement *child;
	short child_x, child_y, lastHit_x, lastHit_y, tx=0, ty=0;
	static    TextElementPtr lastHitTerminal=NULL;
	TextElement *parent = this->getParentElement();

	// don't allow negative x or y in root call
	if (is_root) {
		x -= XMARGIN; y -= YMARGIN;
		if (x<0	) x=0;
		if (y<0) y=0;
	}
	this->getDimensions(width, height);

	const int in_vert_range   = ( (y >= 0) && (y < height) ),
	          in_horiz_range  = ( (x >= 0) && (x < width)  ),
	          in_element      = ( in_vert_range && in_horiz_range ),
		  horiz_packed    = ( (this->pack->flags & pfCounter) &&
				      (this->pack->orientation == pfHoriz) ),
		  vert_packed     = ( (this->pack->flags & pfCounter) &&
	 		              (this->pack->orientation == pfVert) );
	if (is_root) {
		x += XMARGIN; y += YMARGIN;
	}

	if ( in_element && is_terminal_element(this) ) {
		// hit a terminal element
		return_element = this;
		return_index = terminalIndex( return_element, x, y );
		lastHitTerminal = this;	
		if ( return_element->text ) {
			int tw, th, ascent;
			// determine width of character hit
			measureText(return_element->text + return_index, 1, &tw, &th, &ascent); 
			const int in_right_half = (x > tw*return_index + tw/2); // fixed pitch!
			// if hit right half of character, move to next
			if (in_right_half) {
				if (return_index < ucslen(return_element->text)-1)
					// increment return_index if in right half
					return_index = terminalIndex( return_element, x + tw/2, y );
				else {
					iEntityPtr next_entity = i_GetNextEntity(this->entity);
					if (next_entity) {
						TextElement *next = (TextElement *) getTextViewer(next_entity);
						if (next) {
							TextElement *temp_return_element=0;
							int temp_return_index;
	
							// warning - only valid for fixed character width
							next->getElementAtPosition(x - width + tw/2,y, 
											temp_return_element,
											temp_return_index,
											FALSE,TRUE);
							if (temp_return_element) {
								return_element = temp_return_element;
								return_index = temp_return_index;
							}
						}
					}
				}
			}
		}
	} else if ( is_root && y >= height ) {
		// at or off the bottom
		return_element = last_terminal_element(this);
		if ( return_element ) 
			return_index = terminalIndex( return_element, x, y );
	} else if ( over_end /* && vert_packed */ && in_vert_range && is_terminal_element(this)) { 
		// off the end in the some direction 
		return_element = this;
		return_index = terminalIndex( return_element, x, y);
	} else if (in_element || is_root) {
		// Either within a non-terminal element or root call.
		// Call recursively for each child, trying for hit
		// within the child element.
#ifdef LHT_OPTIMIZATION

		// Optimization: on root call, try starting search with the
		// last hit terminal element
		if (is_root && lastHitTerminal) {
			lastHitTerminal->getPosition(this, 0, lastHit_x, lastHit_y);
			lastHitTerminal->getElementAtPosition(x - lastHit_x, y - lastHit_y,
				 	              	      return_element,
						      	      return_index,
					  	      	      FALSE,FALSE);
			// if (return_element) printf("Hay!  The optimization worked.\n");
		}
#endif

		for(r = this->first(); r && !return_element; this->next(&r)){
			child = (TextElement *) this->contents(r);
			child->getPosition(this, 0, child_x, child_y);
			// because positions stored are relative not absolute
			child->getElementAtPosition(x - child_x, y - child_y, 
							return_element,
							return_index,
							FALSE,FALSE);
		}
	}

	if ( !return_element && is_root && over_end ) {
		// Try for over end hit in child element.  Check in reverse order
		// to get last over end hit on a line.
		for(r = this->last(); r && (return_element == 0); this->prev(&r)){
			child = (TextElement *) this->contents(r);
			child->getPosition(this, 0, child_x, child_y);
			// because positions stored are relative not absolute
			child->getElementAtPosition(x - child_x, y - child_y, 
							return_element,
							return_index,
							FALSE,TRUE);
		}
	}

	if (!return_element && is_root) {
		// I think this is bad news, and should not happen.
		return_element=last_terminal_element(this);
		if(return_element->type == iNewLineT){
			return_index = i_GetShort(return_element->entity, 0) - 1;
		} else {
			if(return_element->type == iSpaceT) 
				return_index = i_GetShort(return_element->entity, 0) - 1;
			else
				return_index = ucslen(return_element->text) - 1;
		}
	}
}

void
TextElement::setPosition(short new_rx, short new_ry)
{
	this->position_valid = TRUE;
	if((new_rx != this->rx) || (new_ry != this->ry)){
		this->preLayoutCB();
		this->rx = new_rx;
		this->ry = new_ry;
		this->postLayoutCB();
	}
}


void
TextElement::calculatePosition(TextElement *parent)
{
	short new_rx = 0;
	short new_ry = 0;
	
	if (parent) {
		// your relative xy is some function of your sibs dimensions
		// maybe next maybe prev depending on direction
		Ref r;
		int parent_is_root = !parent->getParentElement();
		TextElement *sib;
		r = parent->seek(this);
		parent->prev(&r);
		if(r){
			sib = (TextElement *) parent->contents(r);
			// if parent is root, this returns absolute position
			// need to subtract x margin in this case
			sib->getPosition(parent, 0, new_rx, new_ry);
			if (parent_is_root) 
				new_rx -= XMARGIN;

			short sib_w;
			short sib_h;
			sib->getDimensions(sib_w, sib_h);
			
			if((sib->pack->flags & pfReset) && sib->pack->orientation == pfVert){
				new_ry += sib_h;
				new_rx = 0;
			} else {
				switch(sib->pack->orientation){
				case pfHoriz:
					new_rx += sib_w;
					break;
				case pfVert:
					new_ry += sib_h;
					break;
				default:
					halt("Illegal orientation.\n");
				}
			}
		}
	}

	this->setPosition(new_rx, new_ry);
}

/*
// why call it index var? because gdb got it confused with a function
// of the same name
*/
void
TextElement::getPosition(TextElement *root, int index_var, short& ax, short& ay)
{
	TextElement *parent = this->getParentElement();

	if (!parent) {
		ax = XMARGIN;
		ay = YMARGIN;
	} else if (this == root) {
		ax = ay = 0;
	} else {
		if(this->position_valid == FALSE) {
			calculatePosition(parent);
		}
		// if parent is root, this returns XMARGIN
		if(parent)
			parent->getPosition(root, 0, ax, ay);
		else {
			// suggested by purify, I think it is right
			ax = 0;
			ay = 0;
		}
	
		ax += rx;
		ay += ry;
		if (index_var!=0) {
			if(this->pack->flags & pfCounter){
				int max=i_GetShort(this->entity,0);
				// index_var is allowed to equal end point?
				if (index_var>max) halt("index %d too big, must be <= %d\n", index_var, max);
				switch(this->pack->orientation) {
				case pfHoriz:
					// bogus -- lets try var pitch
					ax += charWidth('M') * index_var;
				
					break;
				case pfVert:
					// bogus assume fixed height
					ay+= charHeight('M') * index_var;
					break;
				default:
					halt("Illegal orientation.\n");
				}
			} else if(this->text != 0) {
				int max = ucslen(this->text);
				if (index_var>max) 
					halt("index too big\n");

				int i;

				for (i=0; this->text[i] && i<index_var; i++) {
					if(this->text[i] == '\n')
						ay += charHeight(this->text[i]);
					else {
						ax += charWidth(this->text[i]);
					}
				}
			} else {
				halt("indexing unknown index type\n");
			}
		}
	
		// bogus add in some calculation for the index_var
	}
}

void
TextElement::setDimensions(short new_w, short new_h)
{
	this->dimension_valid = TRUE;
	
	if(new_w != this->w || new_h != this->h){
		this->preLayoutCB();
		this->w = new_w;
		this->h = new_h;
		this->postLayoutCB();
	}
}

void
TextElement::calculateDimensions()
{

	Ref r;
	TextElement *child;
	short new_w;
	short new_h;
	new_w = 0;
	new_h = 0;
	
	if(this->pack->flags & pfCounter){
		switch(this->pack->orientation) {
		case pfHoriz:
			// ok -- spaces are fixed pitch
			new_w += charWidth('M') * i_GetShort(this->entity,0);
			new_h = charHeight('M');
			break;
		case pfVert:
			// ok -- newlines are fixed pitch
			new_w = charWidth('M');
			new_h +=  charHeight('M') * i_GetShort(this->entity,0);
			break;
		default:
			halt("Illegal orientation.\n");
		}
	} else if(this->text != 0) {
		int num_nls=0;
		iUCode *s;
		int frag_w;

		for(s = this->text; *s; s++)
			num_nls += *s == '\n' ?1 : 0;

		new_h = charHeight(this->text[0] * (num_nls + 1));

		s = this->text;
		while(*s){
			frag_w = 0;
			do {
				frag_w += charWidth(*s++);
			} while(*s && *s != '\n');
			new_w = MAX(new_w, frag_w);
			if(*s) s++;
		}
	} else {
		iBox box={0,0,0,0};
		iBox child_box;
		short child_x;
		short child_y;
		short child_w;
		short child_h;
		TextElement *parent = this->getParentElement();
		
		for(r = this->first(); r; this->next(&r)){
			child = (TextElement *) this->contents(r);
			child->getPosition(this,0,child_x,child_y);
			child->getDimensions(child_w, child_h);
			child_box.x = child_x;
			child_box.y = child_y;
			child_box.w = child_w;	 
			child_box.h = child_h;
#if 0
			child_box.w = MAX(1, child_box.w);
			child_box.h = MAX(1, child_box.h);
#endif
			i_UnionBox(box,child_box, &box);
		}
		new_w=box.w;
		new_h=box.h;
	}

	this->setDimensions(new_w, new_h);
}

void
TextElement::getDimensions(short& rw, short& rh)
{
	if(this->dimension_valid == FALSE){
		this->calculateDimensions();
	}
	
	rw = this->w;
	rh = this->h;
}

void
TextElement::change()
{
	this->dimension_valid = FALSE;
	TextElement *parent = this->getParentElement();
	
	if(parent){
		// loop over sibs in relevant direction
		// ogus -- handle reverse direction case
		Ref r;
		TextElement *sib;
		r = parent->seek(this);
		if(r) {
			for(parent->next(&r); r; parent->next(&r)){
				sib = (TextElement *) parent->contents(r);
				sib->position_valid = FALSE;
			}
		}
		parent->change();
	}
}

int
TextElement::indexLength()
{
	if(this->pack->flags & pfCounter)
		return i_GetShort(this->entity,0);
	else
		return ucslen(this->text);
}


TextElement *
TextElement::getParentElement()
{
	iEntityPtr current;
	TextElement *parent_element = 0;
	
	for(current = i_GetParentEntity(this->entity); 
		current && ((parent_element = (TextElement *) getTextViewer(current)) == 0);
		current = i_GetParentEntity(current))
		;
	return parent_element;
}


extern iEntityPtr pickpath(iEntityPtr doc);

int
ancestorP(iEntityPtr child, iEntityPtr would_be_ancestor)
{
	for(child = i_GetParentEntity(child); child; child = i_GetParentEntity(child))
		if(child == would_be_ancestor)
			return TRUE;

	return FALSE;
}


iEntityPtr 
TextElement::getNextChild(iEntityPtr prev_child)
{
	iEntityPtr target;
	int found = 0;
	// use find entity forward
	
	target = prev_child ? prev_child : this->entity;
	while(target = 
		  i_FindEntityForward(target, 0, ALL, 0)) {
		// get past the children of the prev_child -- they don't count
		if(prev_child && ancestorP(target, prev_child) == TRUE){
			continue;
		}

		// we must keep checking to be sure we arent looking at people
		// at our same laterally level or higher 
		// or no longer our descendants
		if(ancestorP(target, this->entity) == FALSE)
			break;

		if(getTextViewer(target) != 0) {
			found = 1;
			break;
		}
	}
	
	if(!found)
		target = 0;

	return target;
}

iEntityPtr 
TextElement::getPrevChild(iEntityPtr next_child)
{
	iEntityPtr target;
	int found = 0;
	// use find entity back

	target = next_child ? next_child : this->entity;

	if(next_child == 0){
		iEntityPtr last_descendant;
		iEntityPtr penultimate_descendant;
		last_descendant = i_GetLastEntity(this->entity);
		if(last_descendant) {
			for(penultimate_descendant = i_GetLastEntity(last_descendant); 
				penultimate_descendant; 
				penultimate_descendant = i_GetLastEntity(last_descendant)){
				last_descendant = penultimate_descendant;
			}
		}
		
		target = last_descendant;
		
		if(target == 0 || getTextViewer(target) != 0) {
			found = 1;
		} else {
			next_child = last_descendant;
		}

	}
	
	if(!found) {
		while(target = 
			  i_FindEntityBack(target, 0, ALL, 0)) {
			// get past the children of the next_child -- they don't count
			if(next_child && ancestorP(target, next_child) == TRUE){
				continue;
			}

			// we must keep checking to be sure we arent looking at people
			// at our same laterally level or higher 
			// or no longer our descendants
			if(ancestorP(target, this->entity) == FALSE)
				break;

			if(getTextViewer(target) != 0) {
				found = 1;
				break;
			}
		}
	}
	
	if(!found)
		target = 0;

	return target;
}

Ref 
TextElement::first()
{
	//
	iEntityPtr next_entity;
	TextElement *next_element;
	
	next_entity = this->getNextChild(0);
	if(next_entity) {
		next_element = (TextElement *) getTextViewer(next_entity);
		if(next_element == 0)
			halt("no element for entity.\n");
	} else
		next_element = 0;
	
	return (Ref) next_element;
}

void 
TextElement::next(Ref* r_ptr)
{
	iEntityPtr next_entity;
	TextElement *current_element;
	iEntityPtr current_entity;
	TextElement *next_element;
	
	current_element = (TextElement *) *r_ptr;
	current_entity = current_element->entity;
	
	next_entity = this->getNextChild(current_entity);
	if(next_entity) {
		next_element = (TextElement *) getTextViewer(next_entity);
		if(next_element == 0)
			halt("no element for entity.\n");
	} else
		next_element = 0;
	
	*r_ptr = next_element;
}

void 
TextElement::prev(Ref* r_ptr)
{
	iEntityPtr prev_entity;
	TextElement *current_element;
	iEntityPtr current_entity;
	TextElement *prev_element;
	
	current_element = (TextElement *) *r_ptr;
	current_entity = current_element->entity;
	
	prev_entity = this->getPrevChild(current_entity);
	if(prev_entity) {
		prev_element = (TextElement *) getTextViewer(prev_entity);
		if(prev_element == 0)
			halt("no element for entity.\n");
	} else
		prev_element = 0;
	
	*r_ptr = prev_element;
}

Ref 
TextElement::last()
{
	iEntityPtr prev_entity;
	TextElement *prev_element;
	
	prev_entity = this->getPrevChild(0);
	if(prev_entity) {
		prev_element = (TextElement *) getTextViewer(prev_entity);
		if(prev_element == 0)
			halt("no element for entity.\n");
	} else
		prev_element = 0;
	
	return (Ref) prev_element;
}

int
TextElement::length()
{
	// nasty;
	int count;
	iEntityPtr child;
	child = 0;
	for(count = 0, child = this->getNextChild(child); child; count++, child = this->getNextChild(child))
		;
	
	return count;
}


void
textfriend() {}
