/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "rafutil.h"
#include "ucs.h"
#include "utf.h"
/* for the definition of strcat */
#include "str-lib.h"
/* for the definition of qsort */
#include "c-lib.h"

#include "smemory.h"
#include "rlimage.h"
#include "entity.h"
#include "dafs.h"

#ifndef DEBUG
#define TURN_OFF_DEBUG
#define DEBUG 1
#endif

#include "debug.h"

#ifdef TURN_OFF_DEBUG
#undef DEBUG
#endif

struct UnicodeNameTable {
   unsigned short index;
   char *name;
};

typedef struct UnicodeNameTable UnicodeNameTable;

static char *chrtostr(char p)
{
	static char buf[2];
	switch (p) {
	case 0:
		return "NR";
	case 1:
		return "NZ";
	case ' ':
		return "SP";
	default:
		buf[0] = p;
		buf[1] = 0;
		return buf;
	}
}

static void 
prprops(char *buf,iEntityPtr en,char *sep)
{
	iPTag prop_type;
	iPropPtr pr;
	iDAFSError err;
	const char *name;
	int4 int_value;
	char *str;
	iUCode *ustr;
	void *prop_value;
	iBox *bp;
	char buffer[100];

	*buf= 0;
	for (pr = i_FirstProp(en);pr;pr = i_NextProp(en,pr)) {
		name = i_GetPropName(pr);
		strcat(buf,name);
		strcat(buf," ");
		prop_type = i_GetPropType(pr);
		if (prop_type == iPLong){
			err = i_GetLongProp(en, name, (void*)&int_value);
			sprintf(buffer,"%ld",int_value);
			strcat(buf,buffer);
		} else if ((prop_type == iPUserData)||(prop_type == iPUserIData)){
			err = i_GetUserDataProp(en, name, &prop_value,0,0);
			if (strcmp(name,"PBB"))
				sprintf(buffer,"0x%x",(unsigned int)prop_value);
			else {
				err = i_GetUserDataProp(en, name, &prop_value,0,(iUserDataProc)i_SwapUBox);
				bp = (iBox*)prop_value;
				sprintf(buffer,"%d %d %d %d",bp->x,bp->y,
					bp->w,bp->h);
			}
			strcat(buf,buffer);
		} else if (prop_type == iPString) {
			err = i_GetStringProp(en,name,&str);
			strcat(buf,str);
		} else if (prop_type == iPUString) {
			err = i_GetUStringProp(en,name,&ustr);
			ucs2utf(buffer, ustr);
			strcat(buf,buffer);
		}
		strcat(buf,sep);
		
	}
}

void 
printUniText(iEntityPtr en)
{
	short *ut;
	int4 i;
	ut=(short*)i_GetUText(en,0);
	for(i=0;i<ucslen((const unsigned short*)ut);i++) {
		debugf("%c",(char)ut[i]);
	}
	debugf("\n");
}

#define PRNTXT 256

static void prtxt(FILE * fo,const char *s)
{
	char buf[PRNTXT+1];
	
	if (strlen(s) <= PRNTXT) 
		fprintf(fo,s);
	else {
		memcpy(buf,s,PRNTXT);
		buf[PRNTXT] = 0;
		fprintf(fo,"%s ...",buf);
	}
}

void 
prenfo(iEntityPtr en,FILE * fo)
{
	iBox b;
	int2 valid, n;
	int4 i, index;
	int2 *pos;
	iUCode *ut;
	iPosSetPtr ps;
	char buffer[256];
	
	b = i_GetEnclosingBox(en);
	fprintf(fo,"%s\n",i_TypeToName(i_GetEntityType(en)));
	prprops(buffer,en,"\n");
	fprintf(fo,buffer);
	pos = i_GetPosLimit(en, &n);
	if (n){
		fprintf(fo,"whichpos");
		for (index=0;index<n;index++)	
			fprintf(fo,", %ld",pos[index]);
		fprintf(fo,"\n");
	}
	fprintf(fo,"%d %d %d %d ",b.x,b.y,b.w,b.h);
	switch (i_GetEDataType(en)){
	case iDShort:
		valid=i_GetShort(en,0);
		for(i=0;i<valid;i++){
			if(i_GetEntityType(en)==iNewLineT)
				fprintf(fo,"\n");
			else
				fprintf(fo," ");
		}
		break;
	case iDPosSet:
	case iD1PosSet:
		ps = i_GetPosSet(en,&valid);
		for (i=0;i<valid;i++)
			fprintf(fo,"%s~%d ",chrtostr((char) ps[i].chr),ps[i].conf);
		break;
	case iDUString:
		ut=(iUCode*)i_GetUText(en,0);
		for(i=0;i<ucslen(ut);i++) {
			fprintf(fo,"%c",(char)ut[i]);
		}
		break;
	case iDString:
		prtxt(fo,i_GetText(en,0));
		break;
	default:
		break;
	}
	fprintf(fo,"\n");
	if (i_GetChildEntity(en))
		fprintf(fo,"has %ld %s children\n",i_CountChildren(en),
				(i_GetLogicalType(en) == iOr?"OR":"AND"));
	else fprintf(fo,"No Children\n");
}

void 
pren(iEntityPtr en)
{
	prenfo(en,stdout);
}

static void prentity(iEntityPtr en,FILE * fo)
{
	if (!en) return;
	fprintf(fo,",");
	if (i_GetLogicalType(en) == iOr) fprintf(fo,"Or ");
	fprintf(fo,"%s %ld ",i_TypeToName(i_GetEntityType(en)),i_CountPrev(en));
	prentity(i_GetParentEntity(en),fo);
}

void 
prme(char *pr,iEntityPtr en,FILE * fo)
{
	int4 iswrd;
	
	if(!fo) fo=stderr;
	fprintf(fo,pr);
	prentity(en,fo);
	if (!i_GetBoolProp(en,"In Dict",&iswrd))
		fprintf(fo,"In Dict ");
	fprintf(fo,"\n");
}

void 
i_PrintWord(iEntityPtr word,FILE * fo)
{
	iEntityPtr cur;
	iPosSetPtr glyphs;
	int2 n, num, conf;
	int2 *which;
	
	if (cur = i_GetChildEntity(word)) {
		for (;cur;cur = i_GetNextEntity(cur)) {
			if (i_GetLogicalType(cur)==iOr) {
				iEntityPtr best,loop;
				int2 bestconf;
				best=0; bestconf=-999;
				for (loop=i_GetChildEntity(cur); loop; loop=i_GetNextEntity(loop)) {
					conf=i_GetConf(cur);
					if (conf>bestconf) {
						bestconf=conf;
						best=loop;
					}
				}
				if (best) i_PrintWord(best,fo);
			} else {
				i_PrintWord(cur,fo);
			}
		}
	} else {
		if (i_GetEntityType(word) == iSpaceT)
			fprintf(fo," ");
		else if (i_GetEntityType(word) == iNewLineT)
			fprintf(fo,"\\n");
		else {
			glyphs = i_GetPosSet(word, &n);
			which = i_GetPosLimit(word, &num);
			if (!n) fprintf(fo,"^");
			else if (num) fprintf(fo,"%s",chrtostr((char) glyphs[which[0]].chr));
			else fprintf(fo,"%s",chrtostr((char) glyphs[0].chr));
		}
	}
}

static int4 
comp(char *a,char *b)
{
	return *a - *b;
}

static int4 isone(char p)
{
	return (p >= 'A' && p <= 'Z') ||
		(p >= 'a' && p <= 'z') ||
		(p >= '0' && p <= '9');
}

void 
fpruc(FILE *fo, iUCode uc)
{
    char string[MAXUNILENGTH+2];

#if 0
	uc2subs(uc, string);
    fprintf(fo,"%s", string);
#else
	if(uc < 128) {
	  if(uc != '\n')
		fprintf(fo,"%c", uc);
	  else
		fprintf(fo, "\\n");
	} else {
	  uc2subs(uc, string);
	  fprintf(fo,"%s", string);
	}
#endif
}

void 
fprucs(FILE *fo, iUCode *s)
{
	int4 ucslength;
	int4 i;

	ucslength = ucslen((const unsigned short*)s);

	for(i=0;i<ucslength;i++)
	    fpruc(fo,s[i]);
}

void 
pruc(iUCode uc)
{
	fpruc(stdout,uc);
}

void 
prucs(iUCode *s)
{
	fprucs(stdout,s);
}


void 
printchset(char *p,FILE * fo)
{
	int4 len,i,j;
	
	len = strlen(p);
	qsort(p,
		  len,
		  1,
		  (void*)comp);
	for (i=0;i<len;i++) {
		fprintf(fo,"%s",chrtostr(p[i]));
		for (j=i+1;j<len && isone(p[i]) && isone(p[j]) && ((p[i]+(j-i)) == p[j]);j++) ;
		if (j-i > 3) {
			fprintf(fo,"-%s",chrtostr(p[j-1]));
			i = j-1;
		}
	}
}

static void 
bprlongglyph(iEntityPtr en,FILE * fo)
{
	iPosSetPtr glyphs;
	int2 n, num,conf;
	int4 i;
	int2 *which;
	iUCode *ustr;
	char buf[256],*p;
	iDTag t;
	
	t=i_GetEDataType(en);
	
	switch (t) {
		case iDPosSet:
		case iD1PosSet:
			glyphs = i_GetPosSet(en, &n);
			if (n) {
				which = i_GetPosLimit(en, &num);
				if (num) {
					p = buf;
					for (i=0;i<num;i++)
						/* bogus this won't work with Unicode */
						*p++ = (char ) glyphs[which[i]].chr;
					*p = 0;
					printchset(buf,fo);
					fprintf(fo,"(");
				}
				p = buf;
				for (i = 0; i < n; i++)   
					/* bogus this won't work with unicode */
					*p++ = (char) glyphs[i].chr;
				*p = 0;
				printchset(buf,fo);
				if (num) fprintf(fo,")");
			} else 
				fprintf(fo,chrtostr(0));
			break;
		
		case iDUString:
			ustr=(iUCode*)i_GetUText(en,&conf);
			fprucs(fo,ustr);
			fprintf(fo,"~%i",conf);
			break;
		
		default:
			fprintf(fo,chrtostr(0));
	}

}


static void 
prlongglyph(iEntityPtr en,FILE * fo)
{
	iBox b;

	b=i_GetEnclosingBox(en);
	fprintf(fo,"{%d,%d,%d,%d} ",b.x,b.y,b.w,b.h);	
	bprlongglyph(en,fo);
}


static void 
prshortglyph(iEntityPtr en,FILE * fo)
{
	iPosSetPtr glyphs;
	int2 n, num;
	int2 *which;
	
	glyphs = i_GetPosSet(en, &n);
	if (n) {
		which = i_GetPosLimit(en, &num);
		if (num)
			fprintf(fo,"L %d %s of %d",num,chrtostr((char) glyphs[which[0]].chr),n);
		else fprintf(fo,"No Limit on %d %s",n,chrtostr((char ) glyphs[0].chr));
	} else 
		fprintf(fo,chrtostr(0));
}

static void 
bprallwtab(iEntityPtr en,int4 tab, int4 level,int4 max_child,
		   void (*prglyph)(iEntityPtr,FILE*),FILE * fo)
{
	char *p = "             ";
	iEntityPtr cur;
	char buffer[256];
	int2 valid,conf;
	int4 i;
	iPosSetPtr ps;
	const iUCode *text;
	iBox b;
	
	if (tab) fprintf(fo,&p[strlen(p)-tab]);
	if (level)
		fprintf(fo,"0x%08x, ",(unsigned int)en);
	fprintf(fo,"%s, ",i_TypeToName(i_GetEntityType(en)));
	b=i_GetEnclosingBox(en);
	if (b.w) {
		fprintf(fo,"{%i,%i,%i,%i} ",b.x,b.y,b.w,b.h);
	}
	prprops(buffer,en,", ");
	fprintf(fo,buffer);
	if (i_GetEDataType(en) == iDString) {
		prtxt(fo,i_GetText(en,&conf));
	} else if (i_GetEDataType(en) == iDUString) {
		text = i_GetUText(en,&conf);
		fprucs(fo,(iUCode*)text);
		fprintf(fo," ~ %d",conf);
	} else if (i_GetEntityType(en) == iGlyphT && i_GetLogicalType(en)==iAnd) {
		prglyph(en,fo);
		ps=i_GetPosSet(en,&valid);
		fprintf(fo," (");
		for(i=0;i<valid;i++) {
			fprintf(fo,"%s",chrtostr((char) ps[i].chr));
			if (i+1==valid || (ps[i+1].conf!=ps[i].conf)) {
				fprintf(fo,"~%d",ps[i].conf);
				if (i+1!=valid) fprintf(fo," ");
			}
		}
		fprintf(fo,")");
	} else {
		if (i_GetEntityType(en) == iWordT) i_PrintWord(en,fo);
		else if (i_GetEntityType(en) == iSpaceT || i_GetEntityType(en) == iNewLineT)
			fprintf(fo,"= %d", i_GetShort(en,0));
		fprintf(fo,"%s ",(i_GetLogicalType(en) == iOr?" Or":" And"));
		if (i_GetConf(en)>=0) fprintf(fo," conf=%i",i_GetConf(en));
	}
	fprintf(fo,"\n");
	if (max_child == i_GetEntityType(en)) return;
	for (cur = i_GetChildEntity(en);cur;cur = i_GetNextEntity(cur))
		bprallwtab(cur,tab+1,level,max_child,prglyph,fo);
}

static void 
prallwtab(iEntityPtr en,int4 tab, int4 level,int4 max_child,
		  void (*prglyph)(iEntityPtr,FILE*),FILE * fo)
{
	char *p = "             ";
	iEntityPtr cur;
	char buffer[4096];
	int4 i;
	int2 conf;
	const iUCode *text;
	
	if (tab) fprintf(fo,&p[strlen(p)-tab]);
	if (level)
	 	fprintf(fo,"0x%08x, ",(unsigned int)en);
	fprintf(fo,"%s, ",i_TypeToName(i_GetEntityType(en)));
	prprops(buffer,en,", ");
	fprintf(fo,buffer);
	if (i_GetEDataType(en) == iDString) {
		prtxt(fo,i_GetText(en,&conf));
        fprintf((FILE*)"~%d",(const char *)conf);
    } else if (i_GetEDataType(en) == iDUString) {
        text = i_GetUText(en,&conf);
        fprucs(fo,(iUCode*)text);
        fprintf(fo," ~ %d",conf);
	} else if (i_GetEntityType(en) == iGlyphT && i_GetLogicalType(en)==iAnd) {
		prglyph(en,fo);
	} else {
		if (i_GetEntityType(en) == iWordT) i_PrintWord(en,fo);
		else if (i_GetEntityType(en) == iSpaceT || i_GetEntityType(en) == iNewLineT) 
			fprintf(fo,"= %ld", i_GetShort(en,0));
		fprintf(fo,"%s ",(i_GetLogicalType(en) == iOr?" Or":" And"));
		if (i_GetConf(en)>=0) fprintf(fo," conf=%i",i_GetConf(en));
	}
	fprintf(fo,"\n");
	if (max_child == i_GetEntityType(en)) return;
	for (cur = i_GetChildEntity(en);cur;cur = i_GetNextEntity(cur))
		prallwtab(cur,tab+1,level,max_child,prglyph,fo);
}

void 
prball(iEntityPtr en)
{
	bprallwtab(en,0,0,999,prlongglyph,stdout);
}

void 
fprall(iEntityPtr en)
{
	FILE * fo;
	
	fo = fopen("fred","w");
	prallwtab(en,0,0,999,prlongglyph,fo);
	fclose(fo);
}

void 
fpprall(iEntityPtr en)
{
	FILE * fo;
	
	fo = fopen("fred","w");
	prallwtab(en,0,1,999,prlongglyph,fo);
	fclose(fo);
}

void 
pprall(iEntityPtr en)
{
	prallwtab(en,0,1,999,prlongglyph,stdout);
}

void 
prall(iEntityPtr en)
{
	prallwtab(en,0,0,999,prlongglyph,stdout);
}

void 
prchild(iEntityPtr en)
{
	prallwtab(en,0,1,1,prlongglyph,stdout);
}

void 
prnchild(iEntityPtr en,int4 n)
{
	prallwtab(en,0,1,n,prlongglyph,stdout);
}

void 
prsall(iEntityPtr en)
{
	prallwtab(en,0,1,999,prshortglyph,stdout);
}
static int4 countmem(iEntityPtr en)
{
	int4 tot;
	iEntityPtr cur;
	iPropPtr pr;
	iPData *data;

	
	tot=s_getsize(en);
	if (en->num)
		tot += s_getsize(en->whichpos);
	if (en->data->use > 1) {
		if (en->data->id)
			return tot;
		tot += s_getsize(en->data->matrs);
	}
	en->data->id = 1;
	if (en->data->im && !en->data->im->id) {
		en->data->im->id = 1;
		tot += i_CountMem(en->data->im);
	}
	for (pr = i_FirstProp(en); pr; pr = i_NextProp(en,pr)) {
		data = (iPData *) pr->data;

		switch (data->type) {
		case iPString:
			if (data->val.s)
				tot += s_getsize(data->val.s);
			break;
		case iPUserData:
		case iPUserIData:
			if (data->size && data->val.v)
				tot += s_getsize(data->val.v);
		default:
			break;
		}
		tot += s_getsize(pr);
	}
	switch (en->data->dt) {
	case iDString:
		if (en->data->d.at.text)
			tot += s_getsize(en->data->d.at.text);
		break;
	case iDUString:
		if (en->data->d.ut.unitext)
			tot += s_getsize(en->data->d.ut.unitext);
		break;
	case iDPosSet:
		if (en->data->d.pp.len && en->data->d.pp.posp)
			tot += s_getsize(en->data->d.pp.posp);
		break;
	case iDUserData:
		if (en->data->d.ud.user && en->data->d.ud.len)
			tot += s_getsize(en->data->d.ud.user);
		break;
	default: break;
	}
	for (cur=i_GetChildEntity(en);cur;cur=i_GetNextEntity(cur))
		tot+=countmem(cur);
	return tot;
}

int4 
printmem(iEntityPtr en)
{
	ClearID(en);
	return countmem(en);
}

void 
prpoly(iOrthogonPtr poly)
{	
	int4 i;
	if (poly && poly->n){
	if (poly->n==2)
		debugf("x:%d y:%d w:%d h:%d\n", 
			   poly->point[0].x, poly->point[0].y,
			   poly->point[1].x, poly->point[1].y);
	else 
		for (i=0; i<poly->n; i++)
			debugf("Point %ld: %d,%d\n", 
				   i, poly->point[i].x, poly->point[i].y);
}
else
	debugf("bad polygon\n");
}

void 
prborder(iOrthogonListPtr border)
{
	int4 i;
	if (border){
		if (!border->num_poly)
			debugf("No polys in border\n");
		else {
			for(i=0; i<border->num_poly; i++){
				debugf("#%ld\n", i);
				prpoly(border->poly[i]);
			}	
			debugf("\n");
		}
	}
}

static void 
_prsa(FILE * fo,short *array, int4 n)
{
	int4 i;
	for (i=0; i<n; i++)
		fprintf(fo,"%i ",array[i]);
	fprintf(fo,"\n");
}

static void 
_prla(FILE * fo,long *array, int4 n)
{
	int4 i;
	
	for (i=0; i<n; i++)
		fprintf(fo,"%ld ",array[i]);
	fprintf(fo,"\n");
}

static void 
_prca(FILE * fo,char *array, int4 n)
{
	int4 i;
	
	for (i=0; i<n; i++)
		fprintf(fo,"%02x ",array[i] & 0xff);
	fprintf(fo,"\n");
}

void 
prsa(short *array, int4 n)
{
	_prsa(stdout,array,n);
}

void 
prla(long *array, int4 n)
{
	_prla(stdout,array,n);
}

void 
prca(char *array, int4 n)
{
	_prca(stdout,array,n);
}


void 
fprsa(short *array, int4 n)
{
	FILE * fo = fopen("fred","w");
	_prsa(fo,array,n);
	fclose(fo);
}

void 
fprla(long *array, int4 n)
{
	FILE * fo = fopen("fred","w");
	_prla(fo,array,n);
	fclose(fo);
}

void 
fprca(char *array, int4 n)
{
	FILE * fo = fopen("fred","w");
	_prca(fo,array,n);
	fclose(fo);
}

void 
prregion(iBoxListPtr rgn)
{
	BOX b,b1;
	char *image, *temp;
	int4 rect,w,h,i,j,img_size;
	
	b=rgn->extents;
	w=b.x2-b.x1;
	h=b.y2-b.y1;
	img_size=w*h;
	
	image=smalloc(sizeof(char)*img_size);
	for (i=0; i<img_size; i++) {
		image[i]='.';
	}

	for (rect=0; rect<rgn->numRects; rect++) {
		b1=rgn->rects[rect];
		for (j=b1.y1-b.y1; j<b1.y2-b.y1; j++) {
			temp=image+(j*w);
			for (i=b1.x1-b.x1; i<b1.x2-b.x1; i++) {
				temp[i]='0'+rect;
			}
		}
	}
	for (j=0; j<h; j++) {
		temp=image+j*w;
		for (i=0; i<w; i++)
			debugf("%c",temp[i]);
		debugf("\n");
	}
	smfree(image);
	debugf("bounding box:(%i,%i) (%i,%i)\n",b.x1,b.y1,w,h);
}

#define MAX_LINE 200
static void 
rlshowline(RlLinePtr s,int4 low, int4 high)
{
	int4 i,j,count;
	transition *t;
	
	i = s->size;
	t = s->run;
	count=0;
	while (--i && count<MAX_LINE) {
		for (j=t[-1].dn; j<(t->up) && count<MAX_LINE; j++) {
			if (j<low)
				continue;
			if (j>=high)
				break;
			count++;
			debugf("0");
		}
		for (j=t->up; j<(t->dn) && count<MAX_LINE; j++) {
			if (j<low)
				continue;
			if (j>=high)
				break;
			count++;
			debugf("1");
		}
		t++;
	}
	if (count==MAX_LINE) debugf("+");
	debugf("\n");
}

void
rlshowclipped(iImagePtr sp, iBox b)
{
	int4 i;
	iImagePtr foo = 0;
	iDAFSError e;
	
	e=i_UncompressImage(sp);
	if(!e) {
	debugf("%lx %lx %s\n", (long)sp,(long)sp->d.im.rl,(sp->type == Image?"Image":"Bit Map"));
	if (b.w > 20000) {
		debugf("height=%ld, width=%ld\n",sp->height,sp->width);
		return;
	}
	}
	
	for (i = b.y; i < sp->height && i<b.y+b.h; i++)
		rlshowline(&sp->d.im.rl[i],b.x,b.x+b.w);
	if (foo) i_DisposeImage(foo);
	return;
}

void
rlshowimage(iImagePtr sp)
{
	iBox b;
	b.w=(short) sp->width;
	b.h=(short) sp->height;
	b.x=0;
	b.y=0;
	
	rlshowclipped(sp,b);
	
}


static void rlprintline(RlLinePtr s,FILE * fo)
{
	int4 i;
	fprintf(fo,"%lx ", (unsigned long)s);
	for (i = 0; i < s->size; i++)
		fprintf(fo,"%d %d  ", s->run[i].up, s->run[i].dn);
	fprintf(fo,"\n");
}

int4 
rlprintimage(iImagePtr sp)
{
	int4 i;
	FILE * fo;
	
	fo = fopen("fred","w");
	if (!fo) return -1;
	fprintf(fo,"%lx %lx\n", (unsigned long)sp, (unsigned long)sp->d.im.rl);
	for (i = 0; i < sp->height; i++)
		rlprintline(&sp->d.im.rl[i],fo);
	fclose(fo);
	return 0;
}

static void prchildtype(iLogicalType ct)
{
	switch(ct){
		case iAnd:
			debugf("iAnd\n");
			break;
		case iOr:	
			debugf("iOr\n");
			break;
		default:
			debugf("unknown\n");
			break;
	}
}

static char *gettag(iDTag tag)
{
	switch(tag) {
		case iDNo:
			return "iDNo";
			break;
		case iDPosSet:
			return"iDPosSet";
			break;
		case iDString:
			return"iDString";
			break;
		case iD1PosSet:
			return"iD1PosSet";
			break;
		case iDUserData:
			return"iDUserData";
			break;
		case iDConf:
			return"iDConf";
			break;
		case iDShort:
			return"iDShort";
			break;
		case iDUString:
			return"iDUString";
			break;
		default:
			return"Unkown";
			break;
	}
}

static void prpropguts(iPropPtr props)
{
	iPData *data;
	if(props) {
		data = (iPData *) props->data;
		debugf("name=%s,data.type=%s, data.size=%li, data.val=%lx\n",
			   props->name,gettag(data->type),
			   data->size, (unsigned long) data->val.i);
	}
}

static void pridata(iData *data)
{
	iPropPtr p;
	
	debugf("child type=");
	prchildtype(data->ct);
	debugf("child=%p, last=%p, id=%li, use=%li\n",
		   data->child,data->last,data->id,data->use);
	debugf("image=%p, userPtr=%p, props=%p, tag=%s",
		   data->im,data->userPtr,data->props,gettag(data->dt));
	debugf("len=%i, type=%i, matrs=%p, data=%p\n",
		   data->d.ud.len,data->type,data->matrs,data->d.ud.user);
	debugf("properties:\n");
	for (p=i_FirstPropPtr(data->props); 
		 p; 
		 p=i_NextPropPtr(data->props,p))
		prpropguts(p);
}



void 
prguts(iEntityPtr en)
{
	if (!en)
		return;
		
	debugf("me=%p, next=%p, prev=%p, prdata=%p, "
		   "whichpos=%p, num=%li, data=%p\n",
		   en,en->next,en->prev,en->prdata,en->whichpos,en->num,en->data);
	if (en->prdata)
		pridata(en->prdata);
	if (en->prdata)
		pridata(en->prdata);
	if (en->data)
		pridata(en->data);
}


/* used in dafstest */
int4 
getGlobalTypeSize(void)
{
	return i_GlobalType.size;
}



/* global data */

/*
// global data
// size is 6 * 13000 which is greater than 64 k
// no because lots are commented out -- only 
// 1300 lines are really in there
//
// must either be huge or broken up
// why is it not static?
*/

static UnicodeNameTable uni_table[] = {

	{ 0x0000 , "<control>" },
	{ 0x0001 , "<control>" },
	{ 0x0002 , "<control>" },
	{ 0x0003 , "<control>" },
	{ 0x0004 , "<control>" },
	{ 0x0005 , "<control>" },
	{ 0x0006 , "<control>" },
	{ 0x0007 , "<control>" },
	{ 0x0008 , "<control>" },
	{ 0x0009 , "<control>" },
	{ 0x000A , "<control>" },
	{ 0x000B , "<control>" },
	{ 0x000C , "<control>" },
	{ 0x000D , "<control>" },
	{ 0x000E , "<control>" },
	{ 0x000F , "<control>" },
	{ 0x0010 , "<control>" },
	{ 0x0011 , "<control>" },
	{ 0x0012 , "<control>" },
	{ 0x0013 , "<control>" },
	{ 0x0014 , "<control>" },
	{ 0x0015 , "<control>" },
	{ 0x0016 , "<control>" },
	{ 0x0017 , "<control>" },
	{ 0x0018 , "<control>" },
	{ 0x0019 , "<control>" },
	{ 0x001A , "<control>" },
	{ 0x001B , "<control>" },
	{ 0x001C , "<control>" },
	{ 0x001D , "<control>" },
	{ 0x001E , "<control>" },
	{ 0x001F , "<control>" },
	{ 0x0020 , " " },
	{ 0x0021 , "!" },
	{ 0x0022 , "\"" },
	{ 0x0023 , "#" },
	{ 0x0024 , "$" },
	{ 0x0025 , "%" },
	{ 0x0026 , "&" },
	{ 0x0027 , "'" },
	{ 0x0028 , "(" },
	{ 0x0029 , ")" },
	{ 0x002A , "*" },
	{ 0x002B , "+" },
	{ 0x002C , "," },
	{ 0x002D , "-" },
	{ 0x002E , "." },
	{ 0x002F , "/" },
	{ 0x0030 , "0" },
	{ 0x0031 , "1" },
	{ 0x0032 , "2" },
	{ 0x0033 , "3" },
	{ 0x0034 , "4" },
	{ 0x0035 , "5" },
	{ 0x0036 , "6" },
	{ 0x0037 , "7" },
	{ 0x0038 , "8" },
	{ 0x0039 , "9" },
	{ 0x003A , ":" },
	{ 0x003B , ";" },
	{ 0x003C , "<" },
	{ 0x003D , "=" },
	{ 0x003E , ">" },
	{ 0x003F , "?" },
	{ 0x0040 , "@" },
	{ 0x0041 , "A" },
	{ 0x0042 , "B" },
	{ 0x0043 , "C" },
	{ 0x0044 , "D" },
	{ 0x0045 , "E" },
	{ 0x0046 , "F" },
	{ 0x0047 , "G" },
	{ 0x0048 , "H" },
	{ 0x0049 , "I" },
	{ 0x004A , "J" },
	{ 0x004B , "K" },
	{ 0x004C , "L" },
	{ 0x004D , "M" },
	{ 0x004E , "N" },
	{ 0x004F , "O" },
	{ 0x0050 , "P" },
	{ 0x0051 , "Q" },
	{ 0x0052 , "R" },
	{ 0x0053 , "S" },
	{ 0x0054 , "T" },
	{ 0x0055 , "U" },
	{ 0x0056 , "V" },
	{ 0x0057 , "W" },
	{ 0x0058 , "X" },
	{ 0x0059 , "Y" },
	{ 0x005A , "Z" },
	{ 0x005B , "[" },
	{ 0x005C , "\\" },
	{ 0x005D , "]" },
	{ 0x005E , "^" },
	{ 0x005F , "_" },
	{ 0x0060 , "`" },
	{ 0x0061 , "a" },
	{ 0x0062 , "b" },
	{ 0x0063 , "c" },
	{ 0x0064 , "d" },
	{ 0x0065 , "e" },
	{ 0x0066 , "f" },
	{ 0x0067 , "g" },
	{ 0x0068 , "h" },
	{ 0x0069 , "i" },
	{ 0x006A , "j" },
	{ 0x006B , "k" },
	{ 0x006C , "l" },
	{ 0x006D , "m" },
	{ 0x006E , "n" },
	{ 0x006F , "o" },
	{ 0x0070 , "p" },
	{ 0x0071 , "q" },
	{ 0x0072 , "r" },
	{ 0x0073 , "s" },
	{ 0x0074 , "t" },
	{ 0x0075 , "u" },
	{ 0x0076 , "v" },
	{ 0x0077 , "w" },
	{ 0x0078 , "x" },
	{ 0x0079 , "y" },
	{ 0x007A , "z" },
	{ 0x007B , "{" },
	{ 0x007C , "|" },
	{ 0x007D , "}" },
	{ 0x007E , "~" },
#if 0
	{ 0x0080 , "<control>" },
	{ 0x0081 , "<control>" },
	{ 0x0082 , "<control>" },
	{ 0x0083 , "<control>" },
	{ 0x0084 , "<control>" },
	{ 0x0085 , "<control>" },
	{ 0x0086 , "<control>" },
	{ 0x0087 , "<control>" },
	{ 0x0088 , "<control>" },
	{ 0x0089 , "<control>" },
	{ 0x008A , "<control>" },
	{ 0x008B , "<control>" },
	{ 0x008C , "<control>" },
	{ 0x008D , "<control>" },
	{ 0x008E , "<control>" },
	{ 0x008F , "<control>" },
	{ 0x0090 , "<control>" },
	{ 0x0091 , "<control>" },
	{ 0x0092 , "<control>" },
	{ 0x0093 , "<control>" },
	{ 0x0094 , "<control>" },
	{ 0x0095 , "<control>" },
	{ 0x0096 , "<control>" },
	{ 0x0097 , "<control>" },
	{ 0x0098 , "<control>" },
	{ 0x0099 , "<control>" },
	{ 0x009A , "<control>" },
	{ 0x009B , "<control>" },
	{ 0x009C , "<control>" },
	{ 0x009D , "<control>" },
	{ 0x009E , "<control>" },
	{ 0x009F , "<control>" },
#endif
	{ 0x00A0 , "NO-BREAK SPACE" },
	{ 0x00A1 , "INVERTED EXCLAMATION MARK" },
	{ 0x00A2 , "CENT SIGN" },
	{ 0x00A3 , "POUND SIGN" },
	{ 0x00A4 , "CURRENCY SIGN" },
	{ 0x00A5 , "YEN SIGN" },
	{ 0x00A6 , "BROKEN BAR" },
	{ 0x00A7 , "SECTION SIGN" },
	{ 0x00A8 , "DIAERESIS" },
	{ 0x00A9 , "COPYRIGHT SIGN" },
	{ 0x00AA , "FEMININE ORDINAL INDICATOR" },
	{ 0x00AB , "LEFT-POINTING DOUBLE ANGLE QUOTATION MARK" },
	{ 0x00AC , "NOT SIGN" },
	{ 0x00AD , "SOFT HYPHEN" },
	{ 0x00AE , "REGISTERED SIGN" },
	{ 0x00AF , "MACRON" },
	{ 0x00B0 , "DEGREE SIGN" },
	{ 0x00B1 , "PLUS-MINUS SIGN" },
	{ 0x00B2 , "SUPERSCRIPT TWO" },
	{ 0x00B3 , "SUPERSCRIPT THREE" },
	{ 0x00B4 , "ACUTE ACCENT" },
	{ 0x00B5 , "MICRO SIGN" },
	{ 0x00B6 , "PILCROW SIGN" },
	{ 0x00B7 , "MIDDLE DOT" },
	{ 0x00B8 , "CEDILLA" },
	{ 0x00B9 , "SUPERSCRIPT ONE" },
	{ 0x00BA , "MASCULINE ORDINAL INDICATOR" },
	{ 0x00BB , "RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK" },
	{ 0x00BC , "VULGAR FRACTION ONE QUARTER" },
	{ 0x00BD , "VULGAR FRACTION ONE HALF" },
	{ 0x00BE , "VULGAR FRACTION THREE QUARTERS" },
	{ 0x00BF , "INVERTED QUESTION MARK" },
	{ 0x00C0 , "LATIN CAPITAL LETTER A WITH GRAVE" },
	{ 0x00C1 , "LATIN CAPITAL LETTER A WITH ACUTE" },
	{ 0x00C2 , "LATIN CAPITAL LETTER A WITH CIRCUMFLEX" },
	{ 0x00C3 , "LATIN CAPITAL LETTER A WITH TILDE" },
	{ 0x00C4 , "LATIN CAPITAL LETTER A WITH DIAERESIS" },
	{ 0x00C5 , "LATIN CAPITAL LETTER A WITH RING ABOVE" },
	{ 0x00C6 , "AE" },
	{ 0x00C7 , "LATIN CAPITAL LETTER C WITH CEDILLA" },
	{ 0x00C8 , "LATIN CAPITAL LETTER E WITH GRAVE" },
	{ 0x00C9 , "LATIN CAPITAL LETTER E WITH ACUTE" },
	{ 0x00CA , "LATIN CAPITAL LETTER E WITH CIRCUMFLEX" },
	{ 0x00CB , "LATIN CAPITAL LETTER E WITH DIAERESIS" },
	{ 0x00CC , "LATIN CAPITAL LETTER I WITH GRAVE" },
	{ 0x00CD , "LATIN CAPITAL LETTER I WITH ACUTE" },
	{ 0x00CE , "LATIN CAPITAL LETTER I WITH CIRCUMFLEX" },
	{ 0x00CF , "LATIN CAPITAL LETTER I WITH DIAERESIS" },
	{ 0x00D0 , "LATIN CAPITAL LETTER ETH" },
	{ 0x00D1 , "LATIN CAPITAL LETTER N WITH TILDE" },
	{ 0x00D2 , "LATIN CAPITAL LETTER O WITH GRAVE" },
	{ 0x00D3 , "LATIN CAPITAL LETTER O WITH ACUTE" },
	{ 0x00D4 , "LATIN CAPITAL LETTER O WITH CIRCUMFLEX" },
	{ 0x00D5 , "LATIN CAPITAL LETTER O WITH TILDE" },
	{ 0x00D6 , "LATIN CAPITAL LETTER O WITH DIAERESIS" },
	{ 0x00D7 , "MULTIPLICATION SIGN" },
	{ 0x00D8 , "LATIN CAPITAL LETTER O WITH STROKE" },
	{ 0x00D9 , "LATIN CAPITAL LETTER U WITH GRAVE" },
	{ 0x00DA , "LATIN CAPITAL LETTER U WITH ACUTE" },
	{ 0x00DB , "LATIN CAPITAL LETTER U WITH CIRCUMFLEX" },
	{ 0x00DC , "LATIN CAPITAL LETTER U WITH DIAERESIS" },
	{ 0x00DD , "LATIN CAPITAL LETTER Y WITH ACUTE" },
	{ 0x00DE , "LATIN CAPITAL LETTER THORN" },
	{ 0x00DF , "LATIN SMALL LETTER SHARP S" },
	{ 0x00E0 , "LATIN SMALL LETTER A WITH GRAVE" },
	{ 0x00E1 , "LATIN SMALL LETTER A WITH ACUTE" },
	{ 0x00E2 , "LATIN SMALL LETTER A WITH CIRCUMFLEX" },
	{ 0x00E3 , "LATIN SMALL LETTER A WITH TILDE" },
	{ 0x00E4 , "LATIN SMALL LETTER A WITH DIAERESIS" },
	{ 0x00E5 , "LATIN SMALL LETTER A WITH RING ABOVE" },
	{ 0x00E6 , "ae" },
	{ 0x00E7 , "LATIN SMALL LETTER C WITH CEDILLA" },
	{ 0x00E8 , "LATIN SMALL LETTER E WITH GRAVE" },
	{ 0x00E9 , "LATIN SMALL LETTER E WITH ACUTE" },
	{ 0x00EA , "LATIN SMALL LETTER E WITH CIRCUMFLEX" },
	{ 0x00EB , "LATIN SMALL LETTER E WITH DIAERESIS" },
	{ 0x00EC , "LATIN SMALL LETTER I WITH GRAVE" },
	{ 0x00ED , "LATIN SMALL LETTER I WITH ACUTE" },
	{ 0x00EE , "LATIN SMALL LETTER I WITH CIRCUMFLEX" },
	{ 0x00EF , "LATIN SMALL LETTER I WITH DIAERESIS" },
	{ 0x00F0 , "LATIN SMALL LETTER ETH" },
	{ 0x00F1 , "LATIN SMALL LETTER N WITH TILDE" },
	{ 0x00F2 , "LATIN SMALL LETTER O WITH GRAVE" },
	{ 0x00F3 , "LATIN SMALL LETTER O WITH ACUTE" },
	{ 0x00F4 , "LATIN SMALL LETTER O WITH CIRCUMFLEX" },
	{ 0x00F5 , "LATIN SMALL LETTER O WITH TILDE" },
	{ 0x00F6 , "LATIN SMALL LETTER O WITH DIAERESIS" },
	{ 0x00F7 , "DIVISION SIGN" },
	{ 0x00F8 , "LATIN SMALL LETTER O WITH STROKE" },
	{ 0x00F9 , "LATIN SMALL LETTER U WITH GRAVE" },
	{ 0x00FA , "LATIN SMALL LETTER U WITH ACUTE" },
	{ 0x00FB , "LATIN SMALL LETTER U WITH CIRCUMFLEX" },
	{ 0x00FC , "LATIN SMALL LETTER U WITH DIAERESIS" },
	{ 0x00FD , "LATIN SMALL LETTER Y WITH ACUTE" },
	{ 0x00FE , "LATIN SMALL LETTER THORN" },
	{ 0x00FF , "LATIN SMALL LETTER Y WITH DIAERESIS" },
	{ 0x0100 , "LATIN CAPITAL LETTER A WITH MACRON" },
	{ 0x0101 , "LATIN SMALL LETTER A WITH MACRON" },
	{ 0x0102 , "LATIN CAPITAL LETTER A WITH BREVE" },
	{ 0x0103 , "LATIN SMALL LETTER A WITH BREVE" },
	{ 0x0104 , "LATIN CAPITAL LETTER A WITH OGONEK" },
	{ 0x0105 , "LATIN SMALL LETTER A WITH OGONEK" },
	{ 0x0106 , "LATIN CAPITAL LETTER C WITH ACUTE" },
	{ 0x0107 , "LATIN SMALL LETTER C WITH ACUTE" },
	{ 0x0108 , "LATIN CAPITAL LETTER C WITH CIRCUMFLEX" },
	{ 0x0109 , "LATIN SMALL LETTER C WITH CIRCUMFLEX" },
	{ 0x010A , "LATIN CAPITAL LETTER C WITH DOT ABOVE" },
	{ 0x010B , "LATIN SMALL LETTER C WITH DOT ABOVE" },
	{ 0x010C , "LATIN CAPITAL LETTER C WITH CARON" },
	{ 0x010D , "LATIN SMALL LETTER C WITH CARON" },
	{ 0x010E , "LATIN CAPITAL LETTER D WITH CARON" },
	{ 0x010F , "LATIN SMALL LETTER D WITH CARON" },
	{ 0x0110 , "LATIN CAPITAL LETTER D WITH STROKE" },
	{ 0x0111 , "LATIN SMALL LETTER D WITH STROKE" },
	{ 0x0112 , "LATIN CAPITAL LETTER E WITH MACRON" },
	{ 0x0113 , "LATIN SMALL LETTER E WITH MACRON" },
	{ 0x0114 , "LATIN CAPITAL LETTER E WITH BREVE" },
	{ 0x0115 , "LATIN SMALL LETTER E WITH BREVE" },
	{ 0x0116 , "LATIN CAPITAL LETTER E WITH DOT ABOVE" },
	{ 0x0117 , "LATIN SMALL LETTER E WITH DOT ABOVE" },
	{ 0x0118 , "LATIN CAPITAL LETTER E WITH OGONEK" },
	{ 0x0119 , "LATIN SMALL LETTER E WITH OGONEK" },
	{ 0x011A , "LATIN CAPITAL LETTER E WITH CARON" },
	{ 0x011B , "LATIN SMALL LETTER E WITH CARON" },
	{ 0x011C , "LATIN CAPITAL LETTER G WITH CIRCUMFLEX" },
	{ 0x011D , "LATIN SMALL LETTER G WITH CIRCUMFLEX" },
	{ 0x011E , "LATIN CAPITAL LETTER G WITH BREVE" },
	{ 0x011F , "LATIN SMALL LETTER G WITH BREVE" },
	{ 0x0120 , "LATIN CAPITAL LETTER G WITH DOT ABOVE" },
	{ 0x0121 , "LATIN SMALL LETTER G WITH DOT ABOVE" },
	{ 0x0122 , "LATIN CAPITAL LETTER G WITH CEDILLA" },
	{ 0x0123 , "LATIN SMALL LETTER G WITH CEDILLA" },
	{ 0x0124 , "LATIN CAPITAL LETTER H WITH CIRCUMFLEX" },
	{ 0x0125 , "LATIN SMALL LETTER H WITH CIRCUMFLEX" },
	{ 0x0126 , "LATIN CAPITAL LETTER H WITH STROKE" },
	{ 0x0127 , "LATIN SMALL LETTER H WITH STROKE" },
	{ 0x0128 , "LATIN CAPITAL LETTER I WITH TILDE" },
	{ 0x0129 , "LATIN SMALL LETTER I WITH TILDE" },
	{ 0x012A , "LATIN CAPITAL LETTER I WITH MACRON" },
	{ 0x012B , "LATIN SMALL LETTER I WITH MACRON" },
	{ 0x012C , "LATIN CAPITAL LETTER I WITH BREVE" },
	{ 0x012D , "LATIN SMALL LETTER I WITH BREVE" },
	{ 0x012E , "LATIN CAPITAL LETTER I WITH OGONEK" },
	{ 0x012F , "LATIN SMALL LETTER I WITH OGONEK" },
	{ 0x0130 , "LATIN CAPITAL LETTER I WITH DOT ABOVE" },
	{ 0x0131 , "LATIN SMALL LETTER DOTLESS I" },
	{ 0x0132 , "LATIN CAPITAL LIGATURE IJ" },
	{ 0x0133 , "LATIN SMALL LIGATURE IJ" },
	{ 0x0134 , "LATIN CAPITAL LETTER J WITH CIRCUMFLEX" },
	{ 0x0135 , "LATIN SMALL LETTER J WITH CIRCUMFLEX" },
	{ 0x0136 , "LATIN CAPITAL LETTER K WITH CEDILLA" },
	{ 0x0137 , "LATIN SMALL LETTER K WITH CEDILLA" },
	{ 0x0138 , "LATIN SMALL LETTER KRA" },
	{ 0x0139 , "LATIN CAPITAL LETTER L WITH ACUTE" },
	{ 0x013A , "LATIN SMALL LETTER L WITH ACUTE" },
	{ 0x013B , "LATIN CAPITAL LETTER L WITH CEDILLA" },
	{ 0x013C , "LATIN SMALL LETTER L WITH CEDILLA" },
	{ 0x013D , "LATIN CAPITAL LETTER L WITH CARON" },
	{ 0x013E , "LATIN SMALL LETTER L WITH CARON" },
	{ 0x013F , "LATIN CAPITAL LETTER L WITH MIDDLE DOT" },
	{ 0x0140 , "LATIN SMALL LETTER L WITH MIDDLE DOT" },
	{ 0x0141 , "LATIN CAPITAL LETTER L WITH STROKE" },
	{ 0x0142 , "LATIN SMALL LETTER L WITH STROKE" },
	{ 0x0143 , "LATIN CAPITAL LETTER N WITH ACUTE" },
	{ 0x0144 , "LATIN SMALL LETTER N WITH ACUTE" },
	{ 0x0145 , "LATIN CAPITAL LETTER N WITH CEDILLA" },
	{ 0x0146 , "LATIN SMALL LETTER N WITH CEDILLA" },
	{ 0x0147 , "LATIN CAPITAL LETTER N WITH CARON" },
	{ 0x0148 , "LATIN SMALL LETTER N WITH CARON" },
	{ 0x0149 , "LATIN SMALL LETTER N PRECEDED BY APOSTROPHE" },
	{ 0x014A , "LATIN CAPITAL LETTER ENG" },
	{ 0x014B , "LATIN SMALL LETTER ENG" },
	{ 0x014C , "LATIN CAPITAL LETTER O WITH MACRON" },
	{ 0x014D , "LATIN SMALL LETTER O WITH MACRON" },
	{ 0x014E , "LATIN CAPITAL LETTER O WITH BREVE" },
	{ 0x014F , "LATIN SMALL LETTER O WITH BREVE" },
	{ 0x0150 , "LATIN CAPITAL LETTER O WITH DOUBLE ACUTE" },
	{ 0x0151 , "LATIN SMALL LETTER O WITH DOUBLE ACUTE" },
	{ 0x0152 , "LATIN CAPITAL LIGATURE OE" },
	{ 0x0153 , "LATIN SMALL LIGATURE OE" },
	{ 0x0154 , "LATIN CAPITAL LETTER R WITH ACUTE" },
	{ 0x0155 , "LATIN SMALL LETTER R WITH ACUTE" },
	{ 0x0156 , "LATIN CAPITAL LETTER R WITH CEDILLA" },
	{ 0x0157 , "LATIN SMALL LETTER R WITH CEDILLA" },
	{ 0x0158 , "LATIN CAPITAL LETTER R WITH CARON" },
	{ 0x0159 , "LATIN SMALL LETTER R WITH CARON" },
	{ 0x015A , "LATIN CAPITAL LETTER S WITH ACUTE" },
	{ 0x015B , "LATIN SMALL LETTER S WITH ACUTE" },
	{ 0x015C , "LATIN CAPITAL LETTER S WITH CIRCUMFLEX" },
	{ 0x015D , "LATIN SMALL LETTER S WITH CIRCUMFLEX" },
	{ 0x015E , "LATIN CAPITAL LETTER S WITH CEDILLA" },
	{ 0x015F , "LATIN SMALL LETTER S WITH CEDILLA" },
	{ 0x0160 , "LATIN CAPITAL LETTER S WITH CARON" },
	{ 0x0161 , "LATIN SMALL LETTER S WITH CARON" },
	{ 0x0162 , "LATIN CAPITAL LETTER T WITH CEDILLA" },
	{ 0x0163 , "LATIN SMALL LETTER T WITH CEDILLA" },
	{ 0x0164 , "LATIN CAPITAL LETTER T WITH CARON" },
	{ 0x0165 , "LATIN SMALL LETTER T WITH CARON" },
	{ 0x0166 , "LATIN CAPITAL LETTER T WITH STROKE" },
	{ 0x0167 , "LATIN SMALL LETTER T WITH STROKE" },
	{ 0x0168 , "LATIN CAPITAL LETTER U WITH TILDE" },
	{ 0x0169 , "LATIN SMALL LETTER U WITH TILDE" },
	{ 0x016A , "LATIN CAPITAL LETTER U WITH MACRON" },
	{ 0x016B , "LATIN SMALL LETTER U WITH MACRON" },
	{ 0x016C , "LATIN CAPITAL LETTER U WITH BREVE" },
	{ 0x016D , "LATIN SMALL LETTER U WITH BREVE" },
	{ 0x016E , "LATIN CAPITAL LETTER U WITH RING ABOVE" },
	{ 0x016F , "LATIN SMALL LETTER U WITH RING ABOVE" },
	{ 0x0170 , "LATIN CAPITAL LETTER U WITH DOUBLE ACUTE" },
	{ 0x0171 , "LATIN SMALL LETTER U WITH DOUBLE ACUTE" },
	{ 0x0172 , "LATIN CAPITAL LETTER U WITH OGONEK" },
	{ 0x0173 , "LATIN SMALL LETTER U WITH OGONEK" },
	{ 0x0174 , "LATIN CAPITAL LETTER W WITH CIRCUMFLEX" },
	{ 0x0175 , "LATIN SMALL LETTER W WITH CIRCUMFLEX" },
	{ 0x0176 , "LATIN CAPITAL LETTER Y WITH CIRCUMFLEX" },
	{ 0x0177 , "LATIN SMALL LETTER Y WITH CIRCUMFLEX" },
	{ 0x0178 , "LATIN CAPITAL LETTER Y WITH DIAERESIS" },
	{ 0x0179 , "LATIN CAPITAL LETTER Z WITH ACUTE" },
	{ 0x017A , "LATIN SMALL LETTER Z WITH ACUTE" },
	{ 0x017B , "LATIN CAPITAL LETTER Z WITH DOT ABOVE" },
	{ 0x017C , "LATIN SMALL LETTER Z WITH DOT ABOVE" },
	{ 0x017D , "LATIN CAPITAL LETTER Z WITH CARON" },
	{ 0x017E , "LATIN SMALL LETTER Z WITH CARON" },
	{ 0x017F , "LATIN SMALL LETTER LONG S" },
	{ 0x0180 , "LATIN SMALL LETTER B WITH STROKE" },
	{ 0x0181 , "LATIN CAPITAL LETTER B WITH HOOK" },
	{ 0x0182 , "LATIN CAPITAL LETTER B WITH TOPBAR" },
	{ 0x0183 , "LATIN SMALL LETTER B WITH TOPBAR" },
	{ 0x0184 , "LATIN CAPITAL LETTER TONE SIX" },
	{ 0x0185 , "LATIN SMALL LETTER TONE SIX" },
	{ 0x0186 , "LATIN CAPITAL LETTER OPEN O" },
	{ 0x0187 , "LATIN CAPITAL LETTER C WITH HOOK" },
	{ 0x0188 , "LATIN SMALL LETTER C WITH HOOK" },
	{ 0x0189 , "LATIN CAPITAL LETTER AFRICAN D" },
	{ 0x018A , "LATIN CAPITAL LETTER D WITH HOOK" },
	{ 0x018B , "LATIN CAPITAL LETTER D WITH TOPBAR" },
	{ 0x018C , "LATIN SMALL LETTER D WITH TOPBAR" },
	{ 0x018D , "LATIN SMALL LETTER TURNED DELTA" },
	{ 0x018E , "LATIN CAPITAL LETTER REVERSED E" },
	{ 0x018F , "LATIN CAPITAL LETTER SCHWA" },
	{ 0x0190 , "LATIN CAPITAL LETTER OPEN E" },
	{ 0x0191 , "LATIN CAPITAL LETTER F WITH HOOK" },
	{ 0x0192 , "LATIN SMALL LETTER F WITH HOOK" },
	{ 0x0193 , "LATIN CAPITAL LETTER G WITH HOOK" },
	{ 0x0194 , "LATIN CAPITAL LETTER GAMMA" },
	{ 0x0195 , "LATIN SMALL LETTER HV" },
	{ 0x0196 , "LATIN CAPITAL LETTER IOTA" },
	{ 0x0197 , "LATIN CAPITAL LETTER I WITH STROKE" },
	{ 0x0198 , "LATIN CAPITAL LETTER K WITH HOOK" },
	{ 0x0199 , "LATIN SMALL LETTER K WITH HOOK" },
	{ 0x019A , "LATIN SMALL LETTER L WITH BAR" },
	{ 0x019B , "LATIN SMALL LETTER LAMBDA WITH STROKE" },
	{ 0x019C , "LATIN CAPITAL LETTER TURNED M" },
	{ 0x019D , "LATIN CAPITAL LETTER N WITH LEFT HOOK" },
	{ 0x019E , "LATIN SMALL LETTER N WITH LONG RIGHT LEG" },
	{ 0x019F , "LATIN CAPITAL LETTER O WITH MIDDLE TILDE" },
	{ 0x01A0 , "LATIN CAPITAL LETTER O WITH HORN" },
	{ 0x01A1 , "LATIN SMALL LETTER O WITH HORN" },
	{ 0x01A2 , "LATIN CAPITAL LETTER OI" },
	{ 0x01A3 , "LATIN SMALL LETTER OI" },
	{ 0x01A4 , "LATIN CAPITAL LETTER P WITH HOOK" },
	{ 0x01A5 , "LATIN SMALL LETTER P WITH HOOK" },
	{ 0x01A6 , "LATIN LETTER YR" },
	{ 0x01A7 , "LATIN CAPITAL LETTER TONE TWO" },
	{ 0x01A8 , "LATIN SMALL LETTER TONE TWO" },
	{ 0x01A9 , "LATIN CAPITAL LETTER ESH" },
	{ 0x01AA , "LATIN LETTER REVERSED ESH LOOP" },
	{ 0x01AB , "LATIN SMALL LETTER T WITH PALATAL HOOK" },
	{ 0x01AC , "LATIN CAPITAL LETTER T WITH HOOK" },
	{ 0x01AD , "LATIN SMALL LETTER T WITH HOOK" },
	{ 0x01AE , "LATIN CAPITAL LETTER T WITH RETROFLEX HOOK" },
	{ 0x01AF , "LATIN CAPITAL LETTER U WITH HORN" },
	{ 0x01B0 , "LATIN SMALL LETTER U WITH HORN" },
	{ 0x01B1 , "LATIN CAPITAL LETTER UPSILON" },
	{ 0x01B2 , "LATIN CAPITAL LETTER V WITH HOOK" },
	{ 0x01B3 , "LATIN CAPITAL LETTER Y WITH HOOK" },
	{ 0x01B4 , "LATIN SMALL LETTER Y WITH HOOK" },
	{ 0x01B5 , "LATIN CAPITAL LETTER Z WITH STROKE" },
	{ 0x01B6 , "LATIN SMALL LETTER Z WITH STROKE" },
	{ 0x01B7 , "LATIN CAPITAL LETTER EZH" },
	{ 0x01B8 , "LATIN CAPITAL LETTER EZH REVERSED" },
	{ 0x01B9 , "LATIN SMALL LETTER EZH REVERSED" },
	{ 0x01BA , "LATIN SMALL LETTER EZH WITH TAIL" },
	{ 0x01BB , "LATIN LETTER TWO WITH STROKE" },
	{ 0x01BC , "LATIN CAPITAL LETTER TONE FIVE" },
	{ 0x01BD , "LATIN SMALL LETTER TONE FIVE" },
	{ 0x01BE , "LATIN LETTER INVERTED GLOTTAL STOP WITH STROKE" },
	{ 0x01BF , "LATIN LETTER WYNN" },
	{ 0x01C0 , "LATIN LETTER DENTAL CLICK" },
	{ 0x01C1 , "LATIN LETTER LATERAL CLICK" },
	{ 0x01C2 , "LATIN LETTER ALVEOLAR CLICK" },
	{ 0x01C3 , "LATIN LETTER RETROFLEX CLICK" },
	{ 0x01C4 , "LATIN CAPITAL LETTER DZ WITH CARON" },
	{ 0x01C5 , "LATIN CAPITAL LETTER D WITH SMALL LETTER Z WITH CARON" },
	{ 0x01C6 , "LATIN SMALL LETTER DZ WITH CARON" },
	{ 0x01C7 , "LATIN CAPITAL LETTER LJ" },
	{ 0x01C8 , "LATIN CAPITAL LETTER L WITH SMALL LETTER J" },
	{ 0x01C9 , "LATIN SMALL LETTER LJ" },
	{ 0x01CA , "LATIN CAPITAL LETTER NJ" },
	{ 0x01CB , "LATIN CAPITAL LETTER N WITH SMALL LETTER J" },
	{ 0x01CC , "LATIN SMALL LETTER NJ" },
	{ 0x01CD , "LATIN CAPITAL LETTER A WITH CARON" },
	{ 0x01CE , "LATIN SMALL LETTER A WITH CARON" },
	{ 0x01CF , "LATIN CAPITAL LETTER I WITH CARON" },
	{ 0x01D0 , "LATIN SMALL LETTER I WITH CARON" },
	{ 0x01D1 , "LATIN CAPITAL LETTER O WITH CARON" },
	{ 0x01D2 , "LATIN SMALL LETTER O WITH CARON" },
	{ 0x01D3 , "LATIN CAPITAL LETTER U WITH CARON" },
	{ 0x01D4 , "LATIN SMALL LETTER U WITH CARON" },
	{ 0x01D5 , "LATIN CAPITAL LETTER U WITH DIAERESIS AND MACRON" },
	{ 0x01D6 , "LATIN SMALL LETTER U WITH DIAERESIS AND MACRON" },
	{ 0x01D7 , "LATIN CAPITAL LETTER U WITH DIAERESIS AND ACUTE" },
	{ 0x01D8 , "LATIN SMALL LETTER U WITH DIAERESIS AND ACUTE" },
	{ 0x01D9 , "LATIN CAPITAL LETTER U WITH DIAERESIS AND CARON" },
	{ 0x01DA , "LATIN SMALL LETTER U WITH DIAERESIS AND CARON" },
	{ 0x01DB , "LATIN CAPITAL LETTER U WITH DIAERESIS AND GRAVE" },
	{ 0x01DC , "LATIN SMALL LETTER U WITH DIAERESIS AND GRAVE" },
	{ 0x01DD , "LATIN SMALL LETTER TURNED E" },
	{ 0x01DE , "LATIN CAPITAL LETTER A WITH DIAERESIS AND MACRON" },
	{ 0x01DF , "LATIN SMALL LETTER A WITH DIAERESIS AND MACRON" },
	{ 0x01E0 , "LATIN CAPITAL LETTER A WITH DOT ABOVE AND MACRON" },
	{ 0x01E1 , "LATIN SMALL LETTER A WITH DOT ABOVE AND MACRON" },
	{ 0x01E2 , "LATIN CAPITAL LIGATURE AE WITH MACRON" },
	{ 0x01E3 , "LATIN SMALL LIGATURE AE WITH MACRON" },
	{ 0x01E4 , "LATIN CAPITAL LETTER G WITH STROKE" },
	{ 0x01E5 , "LATIN SMALL LETTER G WITH STROKE" },
	{ 0x01E6 , "LATIN CAPITAL LETTER G WITH CARON" },
	{ 0x01E7 , "LATIN SMALL LETTER G WITH CARON" },
	{ 0x01E8 , "LATIN CAPITAL LETTER K WITH CARON" },
	{ 0x01E9 , "LATIN SMALL LETTER K WITH CARON" },
	{ 0x01EA , "LATIN CAPITAL LETTER O WITH OGONEK" },
	{ 0x01EB , "LATIN SMALL LETTER O WITH OGONEK" },
	{ 0x01EC , "LATIN CAPITAL LETTER O WITH OGONEK AND MACRON" },
	{ 0x01ED , "LATIN SMALL LETTER O WITH OGONEK AND MACRON" },
	{ 0x01EE , "LATIN CAPITAL LETTER EZH WITH CARON" },
	{ 0x01EF , "LATIN SMALL LETTER EZH WITH CARON" },
	{ 0x01F0 , "LATIN SMALL LETTER J WITH CARON" },
	{ 0x01F1 , "LATIN CAPITAL LETTER DZ" },
	{ 0x01F2 , "LATIN CAPITAL LETTER D WITH SMALL LETTER Z" },
	{ 0x01F3 , "LATIN SMALL LETTER DZ" },
	{ 0x01F4 , "LATIN CAPITAL LETTER G WITH ACUTE" },
	{ 0x01F5 , "LATIN SMALL LETTER G WITH ACUTE" },
	{ 0x01FA , "LATIN CAPITAL LETTER A WITH RING ABOVE AND ACUTE" },
	{ 0x01FB , "LATIN SMALL LETTER A WITH RING ABOVE AND ACUTE" },
	{ 0x01FC , "LATIN CAPITAL LIGATURE AE WITH ACUTE" },
	{ 0x01FD , "LATIN SMALL LIGATURE AE WITH ACUTE" },
	{ 0x01FE , "LATIN CAPITAL LETTER O WITH STROKE AND ACUTE" },
	{ 0x01FF , "LATIN SMALL LETTER O WITH STROKE AND ACUTE" },
	{ 0x0200 , "LATIN CAPITAL LETTER A WITH DOUBLE GRAVE" },
	{ 0x0201 , "LATIN SMALL LETTER A WITH DOUBLE GRAVE" },
	{ 0x0202 , "LATIN CAPITAL LETTER A WITH INVERTED BREVE" },
	{ 0x0203 , "LATIN SMALL LETTER A WITH INVERTED BREVE" },
	{ 0x0204 , "LATIN CAPITAL LETTER E WITH DOUBLE GRAVE" },
	{ 0x0205 , "LATIN SMALL LETTER E WITH DOUBLE GRAVE" },
	{ 0x0206 , "LATIN CAPITAL LETTER E WITH INVERTED BREVE" },
	{ 0x0207 , "LATIN SMALL LETTER E WITH INVERTED BREVE" },
	{ 0x0208 , "LATIN CAPITAL LETTER I WITH DOUBLE GRAVE" },
	{ 0x0209 , "LATIN SMALL LETTER I WITH DOUBLE GRAVE" },
	{ 0x020A , "LATIN CAPITAL LETTER I WITH INVERTED BREVE" },
	{ 0x020B , "LATIN SMALL LETTER I WITH INVERTED BREVE" },
	{ 0x020C , "LATIN CAPITAL LETTER O WITH DOUBLE GRAVE" },
	{ 0x020D , "LATIN SMALL LETTER O WITH DOUBLE GRAVE" },
	{ 0x020E , "LATIN CAPITAL LETTER O WITH INVERTED BREVE" },
	{ 0x020F , "LATIN SMALL LETTER O WITH INVERTED BREVE" },
	{ 0x0210 , "LATIN CAPITAL LETTER R WITH DOUBLE GRAVE" },
	{ 0x0211 , "LATIN SMALL LETTER R WITH DOUBLE GRAVE" },
	{ 0x0212 , "LATIN CAPITAL LETTER R WITH INVERTED BREVE" },
	{ 0x0213 , "LATIN SMALL LETTER R WITH INVERTED BREVE" },
	{ 0x0214 , "LATIN CAPITAL LETTER U WITH DOUBLE GRAVE" },
	{ 0x0215 , "LATIN SMALL LETTER U WITH DOUBLE GRAVE" },
	{ 0x0216 , "LATIN CAPITAL LETTER U WITH INVERTED BREVE" },
	{ 0x0217 , "LATIN SMALL LETTER U WITH INVERTED BREVE" },
	{ 0x0250 , "LATIN SMALL LETTER TURNED A" },
	{ 0x0251 , "LATIN SMALL LETTER ALPHA" },
	{ 0x0252 , "LATIN SMALL LETTER TURNED ALPHA" },
	{ 0x0253 , "LATIN SMALL LETTER B WITH HOOK" },
	{ 0x0254 , "LATIN SMALL LETTER OPEN O" },
	{ 0x0255 , "LATIN SMALL LETTER C WITH CURL" },
	{ 0x0256 , "LATIN SMALL LETTER D WITH TAIL" },
	{ 0x0257 , "LATIN SMALL LETTER D WITH HOOK" },
	{ 0x0258 , "LATIN SMALL LETTER REVERSED E" },
	{ 0x0259 , "LATIN SMALL LETTER SCHWA" },
	{ 0x025A , "LATIN SMALL LETTER SCHWA WITH HOOK" },
	{ 0x025B , "LATIN SMALL LETTER OPEN E" },
	{ 0x025C , "LATIN SMALL LETTER REVERSED OPEN E" },
	{ 0x025D , "LATIN SMALL LETTER REVERSED OPEN E WITH HOOK" },
	{ 0x025E , "LATIN SMALL LETTER CLOSED REVERSED OPEN E" },
	{ 0x025F , "LATIN SMALL LETTER DOTLESS J WITH STROKE" },
	{ 0x0260 , "LATIN SMALL LETTER G WITH HOOK" },
	{ 0x0261 , "LATIN SMALL LETTER SCRIPT G" },
	{ 0x0262 , "LATIN LETTER SMALL CAPITAL G" },
	{ 0x0263 , "LATIN SMALL LETTER GAMMA" },
	{ 0x0264 , "LATIN SMALL LETTER RAMS HORN" },
	{ 0x0265 , "LATIN SMALL LETTER TURNED H" },
	{ 0x0266 , "LATIN SMALL LETTER H WITH HOOK" },
	{ 0x0267 , "LATIN SMALL LETTER HENG WITH HOOK" },
	{ 0x0268 , "LATIN SMALL LETTER I WITH STROKE" },
	{ 0x0269 , "LATIN SMALL LETTER IOTA" },
	{ 0x026A , "LATIN LETTER SMALL CAPITAL I" },
	{ 0x026B , "LATIN SMALL LETTER L WITH MIDDLE TILDE" },
	{ 0x026C , "LATIN SMALL LETTER L WITH BELT" },
	{ 0x026D , "LATIN SMALL LETTER L WITH RETROFLEX HOOK" },
	{ 0x026E , "LATIN SMALL LETTER LEZH" },
	{ 0x026F , "LATIN SMALL LETTER TURNED M" },
	{ 0x0270 , "LATIN SMALL LETTER TURNED M WITH LONG LEG" },
	{ 0x0271 , "LATIN SMALL LETTER M WITH HOOK" },
	{ 0x0272 , "LATIN SMALL LETTER N WITH LEFT HOOK" },
	{ 0x0273 , "LATIN SMALL LETTER N WITH RETROFLEX HOOK" },
	{ 0x0274 , "LATIN LETTER SMALL CAPITAL N" },
	{ 0x0275 , "LATIN SMALL LETTER BARRED O" },
	{ 0x0276 , "LATIN LETTER SMALL CAPITAL OE" },
	{ 0x0277 , "LATIN SMALL LETTER CLOSED OMEGA" },
	{ 0x0278 , "LATIN SMALL LETTER PHI" },
	{ 0x0279 , "LATIN SMALL LETTER TURNED R" },
	{ 0x027A , "LATIN SMALL LETTER TURNED R WITH LONG LEG" },
	{ 0x027B , "LATIN SMALL LETTER TURNED R WITH HOOK" },
	{ 0x027C , "LATIN SMALL LETTER R WITH LONG LEG" },
	{ 0x027D , "LATIN SMALL LETTER R WITH TAIL" },
	{ 0x027E , "LATIN SMALL LETTER R WITH FISHHOOK" },
	{ 0x027F , "LATIN SMALL LETTER REVERSED R WITH FISHHOOK" },
	{ 0x0280 , "LATIN LETTER SMALL CAPITAL R" },
	{ 0x0281 , "LATIN LETTER SMALL CAPITAL INVERTED R" },
	{ 0x0282 , "LATIN SMALL LETTER S WITH HOOK" },
	{ 0x0283 , "LATIN SMALL LETTER ESH" },
	{ 0x0284 , "LATIN SMALL LETTER DOTLESS J WITH STROKE AND HOOK" },
	{ 0x0285 , "LATIN SMALL LETTER SQUAT REVERSED ESH" },
	{ 0x0286 , "LATIN SMALL LETTER ESH WITH CURL" },
	{ 0x0287 , "LATIN SMALL LETTER TURNED T" },
	{ 0x0288 , "LATIN SMALL LETTER T WITH RETROFLEX HOOK" },
	{ 0x0289 , "LATIN SMALL LETTER U BAR" },
	{ 0x028A , "LATIN SMALL LETTER UPSILON" },
	{ 0x028B , "LATIN SMALL LETTER V WITH HOOK" },
	{ 0x028C , "LATIN SMALL LETTER TURNED V" },
	{ 0x028D , "LATIN SMALL LETTER TURNED W" },
	{ 0x028E , "LATIN SMALL LETTER TURNED Y" },
	{ 0x028F , "LATIN LETTER SMALL CAPITAL Y" },
	{ 0x0290 , "LATIN SMALL LETTER Z WITH RETROFLEX HOOK" },
	{ 0x0291 , "LATIN SMALL LETTER Z WITH CURL" },
	{ 0x0292 , "LATIN SMALL LETTER EZH" },
	{ 0x0293 , "LATIN SMALL LETTER EZH WITH CURL" },
	{ 0x0294 , "LATIN LETTER GLOTTAL STOP" },
	{ 0x0295 , "LATIN LETTER PHARYNGEAL VOICED FRICATIVE" },
	{ 0x0296 , "LATIN LETTER INVERTED GLOTTAL STOP" },
	{ 0x0297 , "LATIN LETTER STRETCHED C" },
	{ 0x0298 , "LATIN LETTER BILABIAL CLICK" },
	{ 0x0299 , "LATIN LETTER SMALL CAPITAL B" },
	{ 0x029A , "LATIN SMALL LETTER CLOSED OPEN E" },
	{ 0x029B , "LATIN LETTER SMALL CAPITAL G WITH HOOK" },
	{ 0x029C , "LATIN LETTER SMALL CAPITAL H" },
	{ 0x029D , "LATIN SMALL LETTER J WITH CROSSED-TAIL" },
	{ 0x029E , "LATIN SMALL LETTER TURNED K" },
	{ 0x029F , "LATIN LETTER SMALL CAPITAL L" },
	{ 0x02A0 , "LATIN SMALL LETTER Q WITH HOOK" },
	{ 0x02A1 , "LATIN LETTER GLOTTAL STOP WITH STROKE" },
	{ 0x02A2 , "LATIN LETTER REVERSED GLOTTAL STOP WITH STROKE" },
	{ 0x02A3 , "LATIN SMALL LETTER DZ DIGRAPH" },
	{ 0x02A4 , "LATIN SMALL LETTER DEZH DIGRAPH" },
	{ 0x02A5 , "LATIN SMALL LETTER DZ DIGRAPH WITH CURL" },
	{ 0x02A6 , "LATIN SMALL LETTER TS DIGRAPH" },
	{ 0x02A7 , "LATIN SMALL LETTER TESH DIGRAPH" },
	{ 0x02A8 , "LATIN SMALL LETTER TC DIGRAPH WITH CURL" },
#if 0
	{ 0x02B0 , "MODIFIER LETTER SMALL H" },
	{ 0x02B1 , "MODIFIER LETTER SMALL H WITH HOOK" },
	{ 0x02B2 , "MODIFIER LETTER SMALL J" },
	{ 0x02B3 , "MODIFIER LETTER SMALL R" },
	{ 0x02B4 , "MODIFIER LETTER SMALL TURNED R" },
	{ 0x02B5 , "MODIFIER LETTER SMALL TURNED R WITH HOOK" },
	{ 0x02B6 , "MODIFIER LETTER SMALL CAPITAL INVERTED R" },
	{ 0x02B7 , "MODIFIER LETTER SMALL W" },
	{ 0x02B8 , "MODIFIER LETTER SMALL Y" },
	{ 0x02B9 , "MODIFIER LETTER PRIME" },
	{ 0x02BA , "MODIFIER LETTER DOUBLE PRIME" },
	{ 0x02BB , "MODIFIER LETTER TURNED COMMA" },
	{ 0x02BC , "MODIFIER LETTER APOSTROPHE" },
	{ 0x02BD , "MODIFIER LETTER REVERSED COMMA" },
	{ 0x02BE , "MODIFIER LETTER RIGHT HALF RING" },
	{ 0x02BF , "MODIFIER LETTER LEFT HALF RING" },
	{ 0x02C0 , "MODIFIER LETTER GLOTTAL STOP" },
	{ 0x02C1 , "MODIFIER LETTER REVERSED GLOTTAL STOP" },
	{ 0x02C2 , "MODIFIER LETTER LEFT ARROWHEAD" },
	{ 0x02C3 , "MODIFIER LETTER RIGHT ARROWHEAD" },
	{ 0x02C4 , "MODIFIER LETTER UP ARROWHEAD" },
	{ 0x02C5 , "MODIFIER LETTER DOWN ARROWHEAD" },
	{ 0x02C6 , "MODIFIER LETTER CIRCUMFLEX ACCENT" },
	{ 0x02C7 , "CARON" },
	{ 0x02C8 , "MODIFIER LETTER VERTICAL LINE" },
	{ 0x02C9 , "MODIFIER LETTER MACRON" },
	{ 0x02CA , "MODIFIER LETTER ACUTE ACCENT" },
	{ 0x02CB , "MODIFIER LETTER GRAVE ACCENT" },
	{ 0x02CC , "MODIFIER LETTER LOW VERTICAL LINE" },
	{ 0x02CD , "MODIFIER LETTER LOW MACRON" },
	{ 0x02CE , "MODIFIER LETTER LOW GRAVE ACCENT" },
	{ 0x02CF , "MODIFIER LETTER LOW ACUTE ACCENT" },
	{ 0x02D0 , "MODIFIER LETTER TRIANGULAR COLON" },
	{ 0x02D1 , "MODIFIER LETTER HALF TRIANGULAR COLON" },
	{ 0x02D2 , "MODIFIER LETTER CENTRED RIGHT HALF RING" },
	{ 0x02D3 , "MODIFIER LETTER CENTRED LEFT HALF RING" },
	{ 0x02D4 , "MODIFIER LETTER UP TACK" },
	{ 0x02D5 , "MODIFIER LETTER DOWN TACK" },
	{ 0x02D6 , "MODIFIER LETTER PLUS SIGN" },
	{ 0x02D7 , "MODIFIER LETTER MINUS SIGN" },
	{ 0x02D8 , "BREVE" },
	{ 0x02D9 , "DOT ABOVE" },
	{ 0x02DA , "RING ABOVE" },
	{ 0x02DB , "OGONEK" },
	{ 0x02DC , "SMALL TILDE" },
	{ 0x02DD , "DOUBLE ACUTE ACCENT" },
	{ 0x02DE , "MODIFIER LETTER RHOTIC HOOK" },
	{ 0x02E0 , "MODIFIER LETTER SMALL GAMMA" },
	{ 0x02E1 , "MODIFIER LETTER SMALL L" },
	{ 0x02E2 , "MODIFIER LETTER SMALL S" },
	{ 0x02E3 , "MODIFIER LETTER SMALL X" },
	{ 0x02E4 , "MODIFIER LETTER SMALL REVERSED GLOTTAL STOP" },
	{ 0x02E5 , "MODIFIER LETTER EXTRA-HIGH TONE BAR" },
	{ 0x02E6 , "MODIFIER LETTER HIGH TONE BAR" },
	{ 0x02E7 , "MODIFIER LETTER MID TONE BAR" },
	{ 0x02E8 , "MODIFIER LETTER LOW TONE BAR" },
	{ 0x02E9 , "MODIFIER LETTER EXTRA-LOW TONE BAR" },
	{ 0x0300 , "COMBINING GRAVE ACCENT" },
	{ 0x0301 , "COMBINING ACUTE ACCENT" },
	{ 0x0302 , "COMBINING CIRCUMFLEX ACCENT" },
	{ 0x0303 , "COMBINING TILDE" },
	{ 0x0304 , "COMBINING MACRON" },
	{ 0x0305 , "COMBINING OVERLINE" },
	{ 0x0306 , "COMBINING BREVE" },
	{ 0x0307 , "COMBINING DOT ABOVE" },
	{ 0x0308 , "COMBINING DIAERESIS" },
	{ 0x0309 , "COMBINING HOOK ABOVE" },
	{ 0x030A , "COMBINING RING ABOVE" },
	{ 0x030B , "COMBINING DOUBLE ACUTE ACCENT" },
	{ 0x030C , "COMBINING CARON" },
	{ 0x030D , "COMBINING VERTICAL LINE ABOVE" },
	{ 0x030E , "COMBINING DOUBLE VERTICAL LINE ABOVE" },
	{ 0x030F , "COMBINING DOUBLE GRAVE ACCENT" },
	{ 0x0310 , "COMBINING CANDRABINDU" },
	{ 0x0311 , "COMBINING INVERTED BREVE" },
	{ 0x0312 , "COMBINING TURNED COMMA ABOVE" },
	{ 0x0313 , "COMBINING COMMA ABOVE" },
	{ 0x0314 , "COMBINING REVERSED COMMA ABOVE" },
	{ 0x0315 , "COMBINING COMMA ABOVE RIGHT" },
	{ 0x0316 , "COMBINING GRAVE ACCENT BELOW" },
	{ 0x0317 , "COMBINING ACUTE ACCENT BELOW" },
	{ 0x0318 , "COMBINING LEFT TACK BELOW" },
	{ 0x0319 , "COMBINING RIGHT TACK BELOW" },
	{ 0x031A , "COMBINING LEFT ANGLE ABOVE" },
	{ 0x031B , "COMBINING HORN" },
	{ 0x031C , "COMBINING LEFT HALF RING BELOW" },
	{ 0x031D , "COMBINING UP TACK BELOW" },
	{ 0x031E , "COMBINING DOWN TACK BELOW" },
	{ 0x031F , "COMBINING PLUS SIGN BELOW" },
	{ 0x0320 , "COMBINING MINUS SIGN BELOW" },
	{ 0x0321 , "COMBINING PALATALIZED HOOK BELOW" },
	{ 0x0322 , "COMBINING RETROFLEX HOOK BELOW" },
	{ 0x0323 , "COMBINING DOT BELOW" },
	{ 0x0324 , "COMBINING DIAERESIS BELOW" },
	{ 0x0325 , "COMBINING RING BELOW" },
	{ 0x0326 , "COMBINING COMMA BELOW" },
	{ 0x0327 , "COMBINING CEDILLA" },
	{ 0x0328 , "COMBINING OGONEK" },
	{ 0x0329 , "COMBINING VERTICAL LINE BELOW" },
	{ 0x032A , "COMBINING BRIDGE BELOW" },
	{ 0x032B , "COMBINING INVERTED DOUBLE ARCH BELOW" },
	{ 0x032C , "COMBINING CARON BELOW" },
	{ 0x032D , "COMBINING CIRCUMFLEX ACCENT BELOW" },
	{ 0x032E , "COMBINING BREVE BELOW" },
	{ 0x032F , "COMBINING INVERTED BREVE BELOW" },
	{ 0x0330 , "COMBINING TILDE BELOW" },
	{ 0x0331 , "COMBINING MACRON BELOW" },
	{ 0x0332 , "COMBINING LOW LINE" },
	{ 0x0333 , "COMBINING DOUBLE LOW LINE" },
	{ 0x0334 , "COMBINING TILDE OVERLAY" },
	{ 0x0335 , "COMBINING SHORT STROKE OVERLAY" },
	{ 0x0336 , "COMBINING LONG STROKE OVERLAY" },
	{ 0x0337 , "COMBINING SHORT SOLIDUS OVERLAY" },
	{ 0x0338 , "COMBINING LONG SOLIDUS OVERLAY" },
	{ 0x0339 , "COMBINING RIGHT HALF RING BELOW" },
	{ 0x033A , "COMBINING INVERTED BRIDGE BELOW" },
	{ 0x033B , "COMBINING SQUARE BELOW" },
	{ 0x033C , "COMBINING SEAGULL BELOW" },
	{ 0x033D , "COMBINING X ABOVE" },
	{ 0x033E , "COMBINING VERTICAL TILDE" },
	{ 0x033F , "COMBINING DOUBLE OVERLINE" },
	{ 0x0340 , "COMBINING GRAVE TONE MARK" },
	{ 0x0341 , "COMBINING ACUTE TONE MARK" },
	{ 0x0342 , "COMBINING GREEK PERISPOMENI" },
	{ 0x0343 , "COMBINING GREEK KORONIS" },
	{ 0x0344 , "COMBINING GREEK DIALYTIKA TONOS" },
	{ 0x0345 , "COMBINING GREEK YPOGEGRAMMENI" },
	{ 0x0360 , "COMBINING DOUBLE TILDE" },
	{ 0x0361 , "COMBINING DOUBLE INVERTED BREVE" },
#endif
	{ 0x0374 , "GREEK NUMERAL SIGN" },
	{ 0x0375 , "GREEK LOWER NUMERAL SIGN" },
	{ 0x037A , "GREEK YPOGEGRAMMENI" },
	{ 0x037E , "GREEK QUESTION MARK" },
	{ 0x0384 , "GREEK TONOS" },
	{ 0x0385 , "GREEK DIALYTIKA TONOS" },
	{ 0x0386 , "GREEK CAPITAL LETTER ALPHA WITH TONOS" },
	{ 0x0387 , "GREEK ANO TELEIA" },
	{ 0x0388 , "GREEK CAPITAL LETTER EPSILON WITH TONOS" },
	{ 0x0389 , "GREEK CAPITAL LETTER ETA WITH TONOS" },
	{ 0x038A , "GREEK CAPITAL LETTER IOTA WITH TONOS" },
	{ 0x038C , "GREEK CAPITAL LETTER OMICRON WITH TONOS" },
	{ 0x038E , "GREEK CAPITAL LETTER UPSILON WITH TONOS" },
	{ 0x038F , "GREEK CAPITAL LETTER OMEGA WITH TONOS" },
	{ 0x0390 , "GREEK SMALL LETTER IOTA WITH DIALYTIKA AND TONOS" },
	{ 0x0391 , "GREEK CAPITAL LETTER ALPHA" },
	{ 0x0392 , "GREEK CAPITAL LETTER BETA" },
	{ 0x0393 , "GREEK CAPITAL LETTER GAMMA" },
	{ 0x0394 , "GREEK CAPITAL LETTER DELTA" },
	{ 0x0395 , "GREEK CAPITAL LETTER EPSILON" },
	{ 0x0396 , "GREEK CAPITAL LETTER ZETA" },
	{ 0x0397 , "GREEK CAPITAL LETTER ETA" },
	{ 0x0398 , "GREEK CAPITAL LETTER THETA" },
	{ 0x0399 , "GREEK CAPITAL LETTER IOTA" },
	{ 0x039A , "GREEK CAPITAL LETTER KAPPA" },
	{ 0x039B , "GREEK CAPITAL LETTER LAMDA" },
	{ 0x039C , "GREEK CAPITAL LETTER MU" },
	{ 0x039D , "GREEK CAPITAL LETTER NU" },
	{ 0x039E , "GREEK CAPITAL LETTER XI" },
	{ 0x039F , "GREEK CAPITAL LETTER OMICRON" },
	{ 0x03A0 , "GREEK CAPITAL LETTER PI" },
	{ 0x03A1 , "GREEK CAPITAL LETTER RHO" },
	{ 0x03A3 , "GREEK CAPITAL LETTER SIGMA" },
	{ 0x03A4 , "GREEK CAPITAL LETTER TAU" },
	{ 0x03A5 , "GREEK CAPITAL LETTER UPSILON" },
	{ 0x03A6 , "GREEK CAPITAL LETTER PHI" },
	{ 0x03A7 , "GREEK CAPITAL LETTER CHI" },
	{ 0x03A8 , "GREEK CAPITAL LETTER PSI" },
	{ 0x03A9 , "GREEK CAPITAL LETTER OMEGA" },
	{ 0x03AA , "GREEK CAPITAL LETTER IOTA WITH DIALYTIKA" },
	{ 0x03AB , "GREEK CAPITAL LETTER UPSILON WITH DIALYTIKA" },
	{ 0x03AC , "GREEK SMALL LETTER ALPHA WITH TONOS" },
	{ 0x03AD , "GREEK SMALL LETTER EPSILON WITH TONOS" },
	{ 0x03AE , "GREEK SMALL LETTER ETA WITH TONOS" },
	{ 0x03AF , "GREEK SMALL LETTER IOTA WITH TONOS" },
	{ 0x03B0 , "GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND TONOS" },
	{ 0x03B1 , "GREEK SMALL LETTER ALPHA" },
	{ 0x03B2 , "GREEK SMALL LETTER BETA" },
	{ 0x03B3 , "GREEK SMALL LETTER GAMMA" },
	{ 0x03B4 , "GREEK SMALL LETTER DELTA" },
	{ 0x03B5 , "GREEK SMALL LETTER EPSILON" },
	{ 0x03B6 , "GREEK SMALL LETTER ZETA" },
	{ 0x03B7 , "GREEK SMALL LETTER ETA" },
	{ 0x03B8 , "GREEK SMALL LETTER THETA" },
	{ 0x03B9 , "GREEK SMALL LETTER IOTA" },
	{ 0x03BA , "GREEK SMALL LETTER KAPPA" },
	{ 0x03BB , "GREEK SMALL LETTER LAMDA" },
	{ 0x03BC , "GREEK SMALL LETTER MU" },
	{ 0x03BD , "GREEK SMALL LETTER NU" },
	{ 0x03BE , "GREEK SMALL LETTER XI" },
	{ 0x03BF , "GREEK SMALL LETTER OMICRON" },
	{ 0x03C0 , "GREEK SMALL LETTER PI" },
	{ 0x03C1 , "GREEK SMALL LETTER RHO" },
	{ 0x03C2 , "GREEK SMALL LETTER FINAL SIGMA" },
	{ 0x03C3 , "GREEK SMALL LETTER SIGMA" },
	{ 0x03C4 , "GREEK SMALL LETTER TAU" },
	{ 0x03C5 , "GREEK SMALL LETTER UPSILON" },
	{ 0x03C6 , "GREEK SMALL LETTER PHI" },
	{ 0x03C7 , "GREEK SMALL LETTER CHI" },
	{ 0x03C8 , "GREEK SMALL LETTER PSI" },
	{ 0x03C9 , "GREEK SMALL LETTER OMEGA" },
	{ 0x03CA , "GREEK SMALL LETTER IOTA WITH DIALYTIKA" },
	{ 0x03CB , "GREEK SMALL LETTER UPSILON WITH DIALYTIKA" },
	{ 0x03CC , "GREEK SMALL LETTER OMICRON WITH TONOS" },
	{ 0x03CD , "GREEK SMALL LETTER UPSILON WITH TONOS" },
	{ 0x03CE , "GREEK SMALL LETTER OMEGA WITH TONOS" },
	{ 0x03D0 , "GREEK BETA SYMBOL" },
	{ 0x03D1 , "GREEK THETA SYMBOL" },
	{ 0x03D2 , "GREEK UPSILON WITH HOOK SYMBOL" },
	{ 0x03D3 , "GREEK UPSILON WITH ACUTE AND HOOK SYMBOL" },
	{ 0x03D4 , "GREEK UPSILON WITH DIAERESIS AND HOOK SYMBOL" },
	{ 0x03D5 , "GREEK PHI SYMBOL" },
	{ 0x03D6 , "GREEK PI SYMBOL" },
	{ 0x03DA , "GREEK LETTER STIGMA" },
	{ 0x03DC , "GREEK LETTER DIGAMMA" },
	{ 0x03DE , "GREEK LETTER KOPPA" },
	{ 0x03E0 , "GREEK LETTER SAMPI" },
#if 0
	{ 0x03E2 , "COPTIC CAPITAL LETTER SHEI" },
	{ 0x03E3 , "COPTIC SMALL LETTER SHEI" },
	{ 0x03E4 , "COPTIC CAPITAL LETTER FEI" },
	{ 0x03E5 , "COPTIC SMALL LETTER FEI" },
	{ 0x03E6 , "COPTIC CAPITAL LETTER KHEI" },
	{ 0x03E7 , "COPTIC SMALL LETTER KHEI" },
	{ 0x03E8 , "COPTIC CAPITAL LETTER HORI" },
	{ 0x03E9 , "COPTIC SMALL LETTER HORI" },
	{ 0x03EA , "COPTIC CAPITAL LETTER GANGIA" },
	{ 0x03EB , "COPTIC SMALL LETTER GANGIA" },
	{ 0x03EC , "COPTIC CAPITAL LETTER SHIMA" },
	{ 0x03ED , "COPTIC SMALL LETTER SHIMA" },
	{ 0x03EE , "COPTIC CAPITAL LETTER DEI" },
	{ 0x03EF , "COPTIC SMALL LETTER DEI" },
#endif
	{ 0x03F0 , "GREEK KAPPA SYMBOL" },
	{ 0x03F1 , "GREEK RHO SYMBOL" },
	{ 0x03F2 , "GREEK LUNATE SIGMA SYMBOL" },
	{ 0x03F3 , "GREEK LETTER YOT" },
	{ 0x0401 , "CYRILLIC CAPITAL LETTER IO" },
	{ 0x0402 , "CYRILLIC CAPITAL LETTER DJE" },
	{ 0x0403 , "CYRILLIC CAPITAL LETTER GJE" },
	{ 0x0404 , "CYRILLIC CAPITAL LETTER UKRAINIAN IE" },
	{ 0x0405 , "CYRILLIC CAPITAL LETTER DZE" },
	{ 0x0406 , "CYRILLIC CAPITAL LETTER BYELORUSSIAN-UKRAINIAN I" },
	{ 0x0407 , "CYRILLIC CAPITAL LETTER YI" },
	{ 0x0408 , "CYRILLIC CAPITAL LETTER JE" },
	{ 0x0409 , "CYRILLIC CAPITAL LETTER LJE" },
	{ 0x040A , "CYRILLIC CAPITAL LETTER NJE" },
	{ 0x040B , "CYRILLIC CAPITAL LETTER TSHE" },
	{ 0x040C , "CYRILLIC CAPITAL LETTER KJE" },
	{ 0x040E , "CYRILLIC CAPITAL LETTER SHORT U" },
	{ 0x040F , "CYRILLIC CAPITAL LETTER DZHE" },
	{ 0x0410 , "CYRILLIC CAPITAL LETTER A" },
	{ 0x0411 , "CYRILLIC CAPITAL LETTER BE" },
	{ 0x0412 , "CYRILLIC CAPITAL LETTER VE" },
	{ 0x0413 , "CYRILLIC CAPITAL LETTER GHE" },
	{ 0x0414 , "CYRILLIC CAPITAL LETTER DE" },
	{ 0x0415 , "CYRILLIC CAPITAL LETTER IE" },
	{ 0x0416 , "CYRILLIC CAPITAL LETTER ZHE" },
	{ 0x0417 , "CYRILLIC CAPITAL LETTER ZE" },
	{ 0x0418 , "CYRILLIC CAPITAL LETTER I" },
	{ 0x0419 , "CYRILLIC CAPITAL LETTER SHORT I" },
	{ 0x041A , "CYRILLIC CAPITAL LETTER KA" },
	{ 0x041B , "CYRILLIC CAPITAL LETTER EL" },
	{ 0x041C , "CYRILLIC CAPITAL LETTER EM" },
	{ 0x041D , "CYRILLIC CAPITAL LETTER EN" },
	{ 0x041E , "CYRILLIC CAPITAL LETTER O" },
	{ 0x041F , "CYRILLIC CAPITAL LETTER PE" },
	{ 0x0420 , "CYRILLIC CAPITAL LETTER ER" },
	{ 0x0421 , "CYRILLIC CAPITAL LETTER ES" },
	{ 0x0422 , "CYRILLIC CAPITAL LETTER TE" },
	{ 0x0423 , "CYRILLIC CAPITAL LETTER U" },
	{ 0x0424 , "CYRILLIC CAPITAL LETTER EF" },
	{ 0x0425 , "CYRILLIC CAPITAL LETTER HA" },
	{ 0x0426 , "CYRILLIC CAPITAL LETTER TSE" },
	{ 0x0427 , "CYRILLIC CAPITAL LETTER CHE" },
	{ 0x0428 , "CYRILLIC CAPITAL LETTER SHA" },
	{ 0x0429 , "CYRILLIC CAPITAL LETTER SHCHA" },
	{ 0x042A , "CYRILLIC CAPITAL LETTER HARD SIGN" },
	{ 0x042B , "CYRILLIC CAPITAL LETTER YERU" },
	{ 0x042C , "CYRILLIC CAPITAL LETTER SOFT SIGN" },
	{ 0x042D , "CYRILLIC CAPITAL LETTER E" },
	{ 0x042E , "CYRILLIC CAPITAL LETTER YU" },
	{ 0x042F , "CYRILLIC CAPITAL LETTER YA" },
	{ 0x0430 , "CYRILLIC SMALL LETTER A" },
	{ 0x0431 , "CYRILLIC SMALL LETTER BE" },
	{ 0x0432 , "CYRILLIC SMALL LETTER VE" },
	{ 0x0433 , "CYRILLIC SMALL LETTER GHE" },
	{ 0x0434 , "CYRILLIC SMALL LETTER DE" },
	{ 0x0435 , "CYRILLIC SMALL LETTER IE" },
	{ 0x0436 , "CYRILLIC SMALL LETTER ZHE" },
	{ 0x0437 , "CYRILLIC SMALL LETTER ZE" },
	{ 0x0438 , "CYRILLIC SMALL LETTER I" },
	{ 0x0439 , "CYRILLIC SMALL LETTER SHORT I" },
	{ 0x043A , "CYRILLIC SMALL LETTER KA" },
	{ 0x043B , "CYRILLIC SMALL LETTER EL" },
	{ 0x043C , "CYRILLIC SMALL LETTER EM" },
	{ 0x043D , "CYRILLIC SMALL LETTER EN" },
	{ 0x043E , "CYRILLIC SMALL LETTER O" },
	{ 0x043F , "CYRILLIC SMALL LETTER PE" },
	{ 0x0440 , "CYRILLIC SMALL LETTER ER" },
	{ 0x0441 , "CYRILLIC SMALL LETTER ES" },
	{ 0x0442 , "CYRILLIC SMALL LETTER TE" },
	{ 0x0443 , "CYRILLIC SMALL LETTER U" },
	{ 0x0444 , "CYRILLIC SMALL LETTER EF" },
	{ 0x0445 , "CYRILLIC SMALL LETTER HA" },
	{ 0x0446 , "CYRILLIC SMALL LETTER TSE" },
	{ 0x0447 , "CYRILLIC SMALL LETTER CHE" },
	{ 0x0448 , "CYRILLIC SMALL LETTER SHA" },
	{ 0x0449 , "CYRILLIC SMALL LETTER SHCHA" },
	{ 0x044A , "CYRILLIC SMALL LETTER HARD SIGN" },
	{ 0x044B , "CYRILLIC SMALL LETTER YERU" },
	{ 0x044C , "CYRILLIC SMALL LETTER SOFT SIGN" },
	{ 0x044D , "CYRILLIC SMALL LETTER E" },
	{ 0x044E , "CYRILLIC SMALL LETTER YU" },
	{ 0x044F , "CYRILLIC SMALL LETTER YA" },
	{ 0x0451 , "CYRILLIC SMALL LETTER IO" },
	{ 0x0452 , "CYRILLIC SMALL LETTER DJE" },
	{ 0x0453 , "CYRILLIC SMALL LETTER GJE" },
	{ 0x0454 , "CYRILLIC SMALL LETTER UKRAINIAN IE" },
	{ 0x0455 , "CYRILLIC SMALL LETTER DZE" },
	{ 0x0456 , "CYRILLIC SMALL LETTER BYELORUSSIAN-UKRAINIAN I" },
	{ 0x0457 , "CYRILLIC SMALL LETTER YI" },
	{ 0x0458 , "CYRILLIC SMALL LETTER JE" },
	{ 0x0459 , "CYRILLIC SMALL LETTER LJE" },
	{ 0x045A , "CYRILLIC SMALL LETTER NJE" },
	{ 0x045B , "CYRILLIC SMALL LETTER TSHE" },
	{ 0x045C , "CYRILLIC SMALL LETTER KJE" },
	{ 0x045E , "CYRILLIC SMALL LETTER SHORT U" },
	{ 0x045F , "CYRILLIC SMALL LETTER DZHE" },
	{ 0x0460 , "CYRILLIC CAPITAL LETTER OMEGA" },
	{ 0x0461 , "CYRILLIC SMALL LETTER OMEGA" },
	{ 0x0462 , "CYRILLIC CAPITAL LETTER YAT" },
	{ 0x0463 , "CYRILLIC SMALL LETTER YAT" },
	{ 0x0464 , "CYRILLIC CAPITAL LETTER IOTIFIED E" },
	{ 0x0465 , "CYRILLIC SMALL LETTER IOTIFIED E" },
	{ 0x0466 , "CYRILLIC CAPITAL LETTER LITTLE YUS" },
	{ 0x0467 , "CYRILLIC SMALL LETTER LITTLE YUS" },
	{ 0x0468 , "CYRILLIC CAPITAL LETTER IOTIFIED LITTLE YUS" },
	{ 0x0469 , "CYRILLIC SMALL LETTER IOTIFIED LITTLE YUS" },
	{ 0x046A , "CYRILLIC CAPITAL LETTER BIG YUS" },
	{ 0x046B , "CYRILLIC SMALL LETTER BIG YUS" },
	{ 0x046C , "CYRILLIC CAPITAL LETTER IOTIFIED BIG YUS" },
	{ 0x046D , "CYRILLIC SMALL LETTER IOTIFIED BIG YUS" },
	{ 0x046E , "CYRILLIC CAPITAL LETTER KSI" },
	{ 0x046F , "CYRILLIC SMALL LETTER KSI" },
	{ 0x0470 , "CYRILLIC CAPITAL LETTER PSI" },
	{ 0x0471 , "CYRILLIC SMALL LETTER PSI" },
	{ 0x0472 , "CYRILLIC CAPITAL LETTER FITA" },
	{ 0x0473 , "CYRILLIC SMALL LETTER FITA" },
	{ 0x0474 , "CYRILLIC CAPITAL LETTER IZHITSA" },
	{ 0x0475 , "CYRILLIC SMALL LETTER IZHITSA" },
	{ 0x0476 , "CYRILLIC CAPITAL LETTER IZHITSA WITH DOUBLE GRAVE ACCENT" },
	{ 0x0477 , "CYRILLIC SMALL LETTER IZHITSA WITH DOUBLE GRAVE ACCENT" },
	{ 0x0478 , "CYRILLIC CAPITAL LETTER UK" },
	{ 0x0479 , "CYRILLIC SMALL LETTER UK" },
	{ 0x047A , "CYRILLIC CAPITAL LETTER ROUND OMEGA" },
	{ 0x047B , "CYRILLIC SMALL LETTER ROUND OMEGA" },
	{ 0x047C , "CYRILLIC CAPITAL LETTER OMEGA WITH TITLO" },
	{ 0x047D , "CYRILLIC SMALL LETTER OMEGA WITH TITLO" },
	{ 0x047E , "CYRILLIC CAPITAL LETTER OT" },
	{ 0x047F , "CYRILLIC SMALL LETTER OT" },
	{ 0x0480 , "CYRILLIC CAPITAL LETTER KOPPA" },
	{ 0x0481 , "CYRILLIC SMALL LETTER KOPPA" },
	{ 0x0482 , "CYRILLIC THOUSANDS SIGN" },
	{ 0x0483 , "COMBINING CYRILLIC TITLO" },
	{ 0x0484 , "COMBINING CYRILLIC PALATALIZATION" },
	{ 0x0485 , "COMBINING CYRILLIC DASIA PNEUMATA" },
	{ 0x0486 , "COMBINING CYRILLIC PSILI PNEUMATA" },
	{ 0x0490 , "CYRILLIC CAPITAL LETTER GHE WITH UPTURN" },
	{ 0x0491 , "CYRILLIC SMALL LETTER GHE WITH UPTURN" },
	{ 0x0492 , "CYRILLIC CAPITAL LETTER GHE WITH STROKE" },
	{ 0x0493 , "CYRILLIC SMALL LETTER GHE WITH STROKE" },
	{ 0x0494 , "CYRILLIC CAPITAL LETTER GHE WITH MIDDLE HOOK" },
	{ 0x0495 , "CYRILLIC SMALL LETTER GHE WITH MIDDLE HOOK" },
	{ 0x0496 , "CYRILLIC CAPITAL LETTER ZHE WITH DESCENDER" },
	{ 0x0497 , "CYRILLIC SMALL LETTER ZHE WITH DESCENDER" },
	{ 0x0498 , "CYRILLIC CAPITAL LETTER ZE WITH DESCENDER" },
	{ 0x0499 , "CYRILLIC SMALL LETTER ZE WITH DESCENDER" },
	{ 0x049A , "CYRILLIC CAPITAL LETTER KA WITH DESCENDER" },
	{ 0x049B , "CYRILLIC SMALL LETTER KA WITH DESCENDER" },
	{ 0x049C , "CYRILLIC CAPITAL LETTER KA WITH VERTICAL STROKE" },
	{ 0x049D , "CYRILLIC SMALL LETTER KA WITH VERTICAL STROKE" },
	{ 0x049E , "CYRILLIC CAPITAL LETTER KA WITH STROKE" },
	{ 0x049F , "CYRILLIC SMALL LETTER KA WITH STROKE" },
	{ 0x04A0 , "CYRILLIC CAPITAL LETTER BASHKIR KA" },
	{ 0x04A1 , "CYRILLIC SMALL LETTER BASHKIR KA" },
	{ 0x04A2 , "CYRILLIC CAPITAL LETTER EN WITH DESCENDER" },
	{ 0x04A3 , "CYRILLIC SMALL LETTER EN WITH DESCENDER" },
	{ 0x04A4 , "CYRILLIC CAPITAL LIGATURE EN GHE" },
	{ 0x04A5 , "CYRILLIC SMALL LIGATURE EN GHE" },
	{ 0x04A6 , "CYRILLIC CAPITAL LETTER PE WITH MIDDLE HOOK" },
	{ 0x04A7 , "CYRILLIC SMALL LETTER PE WITH MIDDLE HOOK" },
	{ 0x04A8 , "CYRILLIC CAPITAL LETTER ABKHASIAN HA" },
	{ 0x04A9 , "CYRILLIC SMALL LETTER ABKHASIAN HA" },
	{ 0x04AA , "CYRILLIC CAPITAL LETTER ES WITH DESCENDER" },
	{ 0x04AB , "CYRILLIC SMALL LETTER ES WITH DESCENDER" },
	{ 0x04AC , "CYRILLIC CAPITAL LETTER TE WITH DESCENDER" },
	{ 0x04AD , "CYRILLIC SMALL LETTER TE WITH DESCENDER" },
	{ 0x04AE , "CYRILLIC CAPITAL LETTER STRAIGHT U" },
	{ 0x04AF , "CYRILLIC SMALL LETTER STRAIGHT U" },
	{ 0x04B0 , "CYRILLIC CAPITAL LETTER STRAIGHT U WITH STROKE" },
	{ 0x04B1 , "CYRILLIC SMALL LETTER STRAIGHT U WITH STROKE" },
	{ 0x04B2 , "CYRILLIC CAPITAL LETTER HA WITH DESCENDER" },
	{ 0x04B3 , "CYRILLIC SMALL LETTER HA WITH DESCENDER" },
	{ 0x04B4 , "CYRILLIC CAPITAL LIGATURE TE TSE" },
	{ 0x04B5 , "CYRILLIC SMALL LIGATURE TE TSE" },
	{ 0x04B6 , "CYRILLIC CAPITAL LETTER CHE WITH DESCENDER" },
	{ 0x04B7 , "CYRILLIC SMALL LETTER CHE WITH DESCENDER" },
	{ 0x04B8 , "CYRILLIC CAPITAL LETTER CHE WITH VERTICAL STROKE" },
	{ 0x04B9 , "CYRILLIC SMALL LETTER CHE WITH VERTICAL STROKE" },
	{ 0x04BA , "CYRILLIC CAPITAL LETTER SHHA" },
	{ 0x04BB , "CYRILLIC SMALL LETTER SHHA" },
	{ 0x04BC , "CYRILLIC CAPITAL LETTER ABKHASIAN CHE" },
	{ 0x04BD , "CYRILLIC SMALL LETTER ABKHASIAN CHE" },
	{ 0x04BE , "CYRILLIC CAPITAL LETTER ABKHASIAN CHE WITH DESCENDER" },
	{ 0x04BF , "CYRILLIC SMALL LETTER ABKHASIAN CHE WITH DESCENDER" },
	{ 0x04C0 , "CYRILLIC LETTER PALOCHKA" },
	{ 0x04C1 , "CYRILLIC CAPITAL LETTER ZHE WITH BREVE" },
	{ 0x04C2 , "CYRILLIC SMALL LETTER ZHE WITH BREVE" },
	{ 0x04C3 , "CYRILLIC CAPITAL LETTER KA WITH HOOK" },
	{ 0x04C4 , "CYRILLIC SMALL LETTER KA WITH HOOK" },
	{ 0x04C7 , "CYRILLIC CAPITAL LETTER EN WITH HOOK" },
	{ 0x04C8 , "CYRILLIC SMALL LETTER EN WITH HOOK" },
	{ 0x04CB , "CYRILLIC CAPITAL LETTER KHAKASSIAN CHE" },
	{ 0x04CC , "CYRILLIC SMALL LETTER KHAKASSIAN CHE" },
	{ 0x04D0 , "CYRILLIC CAPITAL LETTER A WITH BREVE" },
	{ 0x04D1 , "CYRILLIC SMALL LETTER A WITH BREVE" },
	{ 0x04D2 , "CYRILLIC CAPITAL LETTER A WITH DIAERESIS" },
	{ 0x04D3 , "CYRILLIC SMALL LETTER A WITH DIAERESIS" },
	{ 0x04D4 , "CYRILLIC CAPITAL LIGATURE A IE" },
	{ 0x04D5 , "CYRILLIC SMALL LIGATURE A IE" },
	{ 0x04D6 , "CYRILLIC CAPITAL LETTER IE WITH BREVE" },
	{ 0x04D7 , "CYRILLIC SMALL LETTER IE WITH BREVE" },
	{ 0x04D8 , "CYRILLIC CAPITAL LETTER SCHWA" },
	{ 0x04D9 , "CYRILLIC SMALL LETTER SCHWA" },
	{ 0x04DA , "CYRILLIC CAPITAL LETTER SCHWA WITH DIAERESIS" },
	{ 0x04DB , "CYRILLIC SMALL LETTER SCHWA WITH DIAERESIS" },
	{ 0x04DC , "CYRILLIC CAPITAL LETTER ZHE WITH DIAERESIS" },
	{ 0x04DD , "CYRILLIC SMALL LETTER ZHE WITH DIAERESIS" },
	{ 0x04DE , "CYRILLIC CAPITAL LETTER ZE WITH DIAERESIS" },
	{ 0x04DF , "CYRILLIC SMALL LETTER ZE WITH DIAERESIS" },
	{ 0x04E0 , "CYRILLIC CAPITAL LETTER ABKHASIAN DZE" },
	{ 0x04E1 , "CYRILLIC SMALL LETTER ABKHASIAN DZE" },
	{ 0x04E2 , "CYRILLIC CAPITAL LETTER I WITH MACRON" },
	{ 0x04E3 , "CYRILLIC SMALL LETTER I WITH MACRON" },
	{ 0x04E4 , "CYRILLIC CAPITAL LETTER I WITH DIAERESIS" },
	{ 0x04E5 , "CYRILLIC SMALL LETTER I WITH DIAERESIS" },
	{ 0x04E6 , "CYRILLIC CAPITAL LETTER O WITH DIAERESIS" },
	{ 0x04E7 , "CYRILLIC SMALL LETTER O WITH DIAERESIS" },
	{ 0x04E8 , "CYRILLIC CAPITAL LETTER BARRED O" },
	{ 0x04E9 , "CYRILLIC SMALL LETTER BARRED O" },
	{ 0x04EA , "CYRILLIC CAPITAL LETTER BARRED O WITH DIAERESIS" },
	{ 0x04EB , "CYRILLIC SMALL LETTER BARRED O WITH DIAERESIS" },
	{ 0x04EE , "CYRILLIC CAPITAL LETTER U WITH MACRON" },
	{ 0x04EF , "CYRILLIC SMALL LETTER U WITH MACRON" },
	{ 0x04F0 , "CYRILLIC CAPITAL LETTER U WITH DIAERESIS" },
	{ 0x04F1 , "CYRILLIC SMALL LETTER U WITH DIAERESIS" },
	{ 0x04F2 , "CYRILLIC CAPITAL LETTER U WITH DOUBLE ACUTE" },
	{ 0x04F3 , "CYRILLIC SMALL LETTER U WITH DOUBLE ACUTE" },
	{ 0x04F4 , "CYRILLIC CAPITAL LETTER CHE WITH DIAERESIS" },
	{ 0x04F5 , "CYRILLIC SMALL LETTER CHE WITH DIAERESIS" },
	{ 0x04F8 , "CYRILLIC CAPITAL LETTER YERU WITH DIAERESIS" },
	{ 0x04F9 , "CYRILLIC SMALL LETTER YERU WITH DIAERESIS" },
#if 0
	{ 0x0531 , "ARMENIAN CAPITAL LETTER AYB" },
	{ 0x0532 , "ARMENIAN CAPITAL LETTER BEN" },
	{ 0x0533 , "ARMENIAN CAPITAL LETTER GIM" },
	{ 0x0534 , "ARMENIAN CAPITAL LETTER DA" },
	{ 0x0535 , "ARMENIAN CAPITAL LETTER ECH" },
	{ 0x0536 , "ARMENIAN CAPITAL LETTER ZA" },
	{ 0x0537 , "ARMENIAN CAPITAL LETTER EH" },
	{ 0x0538 , "ARMENIAN CAPITAL LETTER ET" },
	{ 0x0539 , "ARMENIAN CAPITAL LETTER TO" },
	{ 0x053A , "ARMENIAN CAPITAL LETTER ZHE" },
	{ 0x053B , "ARMENIAN CAPITAL LETTER INI" },
	{ 0x053C , "ARMENIAN CAPITAL LETTER LIWN" },
	{ 0x053D , "ARMENIAN CAPITAL LETTER XEH" },
	{ 0x053E , "ARMENIAN CAPITAL LETTER CA" },
	{ 0x053F , "ARMENIAN CAPITAL LETTER KEN" },
	{ 0x0540 , "ARMENIAN CAPITAL LETTER HO" },
	{ 0x0541 , "ARMENIAN CAPITAL LETTER JA" },
	{ 0x0542 , "ARMENIAN CAPITAL LETTER GHAD" },
	{ 0x0543 , "ARMENIAN CAPITAL LETTER CHEH" },
	{ 0x0544 , "ARMENIAN CAPITAL LETTER MEN" },
	{ 0x0545 , "ARMENIAN CAPITAL LETTER YI" },
	{ 0x0546 , "ARMENIAN CAPITAL LETTER NOW" },
	{ 0x0547 , "ARMENIAN CAPITAL LETTER SHA" },
	{ 0x0548 , "ARMENIAN CAPITAL LETTER VO" },
	{ 0x0549 , "ARMENIAN CAPITAL LETTER CHA" },
	{ 0x054A , "ARMENIAN CAPITAL LETTER PEH" },
	{ 0x054B , "ARMENIAN CAPITAL LETTER JHEH" },
	{ 0x054C , "ARMENIAN CAPITAL LETTER RA" },
	{ 0x054D , "ARMENIAN CAPITAL LETTER SEH" },
	{ 0x054E , "ARMENIAN CAPITAL LETTER VEW" },
	{ 0x054F , "ARMENIAN CAPITAL LETTER TIWN" },
	{ 0x0550 , "ARMENIAN CAPITAL LETTER REH" },
	{ 0x0551 , "ARMENIAN CAPITAL LETTER CO" },
	{ 0x0552 , "ARMENIAN CAPITAL LETTER YIWN" },
	{ 0x0553 , "ARMENIAN CAPITAL LETTER PIWR" },
	{ 0x0554 , "ARMENIAN CAPITAL LETTER KEH" },
	{ 0x0555 , "ARMENIAN CAPITAL LETTER OH" },
	{ 0x0556 , "ARMENIAN CAPITAL LETTER FEH" },
	{ 0x0559 , "ARMENIAN MODIFIER LETTER LEFT HALF RING" },
	{ 0x055A , "ARMENIAN APOSTROPHE" },
	{ 0x055B , "ARMENIAN EMPHASIS MARK" },
	{ 0x055C , "ARMENIAN EXCLAMATION MARK" },
	{ 0x055D , "ARMENIAN COMMA" },
	{ 0x055E , "ARMENIAN QUESTION MARK" },
	{ 0x055F , "ARMENIAN ABBREVIATION MARK" },
	{ 0x0561 , "ARMENIAN SMALL LETTER AYB" },
	{ 0x0562 , "ARMENIAN SMALL LETTER BEN" },
	{ 0x0563 , "ARMENIAN SMALL LETTER GIM" },
	{ 0x0564 , "ARMENIAN SMALL LETTER DA" },
	{ 0x0565 , "ARMENIAN SMALL LETTER ECH" },
	{ 0x0566 , "ARMENIAN SMALL LETTER ZA" },
	{ 0x0567 , "ARMENIAN SMALL LETTER EH" },
	{ 0x0568 , "ARMENIAN SMALL LETTER ET" },
	{ 0x0569 , "ARMENIAN SMALL LETTER TO" },
	{ 0x056A , "ARMENIAN SMALL LETTER ZHE" },
	{ 0x056B , "ARMENIAN SMALL LETTER INI" },
	{ 0x056C , "ARMENIAN SMALL LETTER LIWN" },
	{ 0x056D , "ARMENIAN SMALL LETTER XEH" },
	{ 0x056E , "ARMENIAN SMALL LETTER CA" },
	{ 0x056F , "ARMENIAN SMALL LETTER KEN" },
	{ 0x0570 , "ARMENIAN SMALL LETTER HO" },
	{ 0x0571 , "ARMENIAN SMALL LETTER JA" },
	{ 0x0572 , "ARMENIAN SMALL LETTER GHAD" },
	{ 0x0573 , "ARMENIAN SMALL LETTER CHEH" },
	{ 0x0574 , "ARMENIAN SMALL LETTER MEN" },
	{ 0x0575 , "ARMENIAN SMALL LETTER YI" },
	{ 0x0576 , "ARMENIAN SMALL LETTER NOW" },
	{ 0x0577 , "ARMENIAN SMALL LETTER SHA" },
	{ 0x0578 , "ARMENIAN SMALL LETTER VO" },
	{ 0x0579 , "ARMENIAN SMALL LETTER CHA" },
	{ 0x057A , "ARMENIAN SMALL LETTER PEH" },
	{ 0x057B , "ARMENIAN SMALL LETTER JHEH" },
	{ 0x057C , "ARMENIAN SMALL LETTER RA" },
	{ 0x057D , "ARMENIAN SMALL LETTER SEH" },
	{ 0x057E , "ARMENIAN SMALL LETTER VEW" },
	{ 0x057F , "ARMENIAN SMALL LETTER TIWN" },
	{ 0x0580 , "ARMENIAN SMALL LETTER REH" },
	{ 0x0581 , "ARMENIAN SMALL LETTER CO" },
	{ 0x0582 , "ARMENIAN SMALL LETTER YIWN" },
	{ 0x0583 , "ARMENIAN SMALL LETTER PIWR" },
	{ 0x0584 , "ARMENIAN SMALL LETTER KEH" },
	{ 0x0585 , "ARMENIAN SMALL LETTER OH" },
	{ 0x0586 , "ARMENIAN SMALL LETTER FEH" },
	{ 0x0587 , "ARMENIAN SMALL LIGATURE ECH YIWN" },
	{ 0x0589 , "ARMENIAN FULL STOP" },
	{ 0x05B0 , "HEBREW POINT SHEVA" },
	{ 0x05B1 , "HEBREW POINT HATAF SEGOL" },
	{ 0x05B2 , "HEBREW POINT HATAF PATAH" },
	{ 0x05B3 , "HEBREW POINT HATAF QAMATS" },
	{ 0x05B4 , "HEBREW POINT HIRIQ" },
	{ 0x05B5 , "HEBREW POINT TSERE" },
	{ 0x05B6 , "HEBREW POINT SEGOL" },
	{ 0x05B7 , "HEBREW POINT PATAH" },
	{ 0x05B8 , "HEBREW POINT QAMATS" },
	{ 0x05B9 , "HEBREW POINT HOLAM" },
	{ 0x05BB , "HEBREW POINT QUBUTS" },
	{ 0x05BC , "HEBREW POINT DAGESH OR MAPIQ" },
	{ 0x05BD , "HEBREW POINT METEG" },
	{ 0x05BE , "HEBREW PUNCTUATION MAQAF" },
	{ 0x05BF , "HEBREW POINT RAFE" },
	{ 0x05C0 , "HEBREW PUNCTUATION PASEQ" },
	{ 0x05C1 , "HEBREW POINT SHIN DOT" },
	{ 0x05C2 , "HEBREW POINT SIN DOT" },
	{ 0x05C3 , "HEBREW PUNCTUATION SOF PASUQ" },
	{ 0x05D0 , "HEBREW LETTER ALEF" },
	{ 0x05D1 , "HEBREW LETTER BET" },
	{ 0x05D2 , "HEBREW LETTER GIMEL" },
	{ 0x05D3 , "HEBREW LETTER DALET" },
	{ 0x05D4 , "HEBREW LETTER HE" },
	{ 0x05D5 , "HEBREW LETTER VAV" },
	{ 0x05D6 , "HEBREW LETTER ZAYIN" },
	{ 0x05D7 , "HEBREW LETTER HET" },
	{ 0x05D8 , "HEBREW LETTER TET" },
	{ 0x05D9 , "HEBREW LETTER YOD" },
	{ 0x05DA , "HEBREW LETTER FINAL KAF" },
	{ 0x05DB , "HEBREW LETTER KAF" },
	{ 0x05DC , "HEBREW LETTER LAMED" },
	{ 0x05DD , "HEBREW LETTER FINAL MEM" },
	{ 0x05DE , "HEBREW LETTER MEM" },
	{ 0x05DF , "HEBREW LETTER FINAL NUN" },
	{ 0x05E0 , "HEBREW LETTER NUN" },
	{ 0x05E1 , "HEBREW LETTER SAMEKH" },
	{ 0x05E2 , "HEBREW LETTER AYIN" },
	{ 0x05E3 , "HEBREW LETTER FINAL PE" },
	{ 0x05E4 , "HEBREW LETTER PE" },
	{ 0x05E5 , "HEBREW LETTER FINAL TSADI" },
	{ 0x05E6 , "HEBREW LETTER TSADI" },
	{ 0x05E7 , "HEBREW LETTER QOF" },
	{ 0x05E8 , "HEBREW LETTER RESH" },
	{ 0x05E9 , "HEBREW LETTER SHIN" },
	{ 0x05EA , "HEBREW LETTER TAV" },
	{ 0x05F0 , "HEBREW LIGATURE YIDDISH DOUBLE VAV" },
	{ 0x05F1 , "HEBREW LIGATURE YIDDISH VAV YOD" },
	{ 0x05F2 , "HEBREW LIGATURE YIDDISH DOUBLE YOD" },
	{ 0x05F3 , "HEBREW PUNCTUATION GERESH" },
	{ 0x05F4 , "HEBREW PUNCTUATION GERSHAYIM" },
	{ 0x060C , "ARABIC COMMA" },
	{ 0x061B , "ARABIC SEMICOLON" },
	{ 0x061F , "ARABIC QUESTION MARK" },
	{ 0x0621 , "ARABIC LETTER HAMZA" },
	{ 0x0622 , "ARABIC LETTER ALEF WITH MADDA ABOVE" },
	{ 0x0623 , "ARABIC LETTER ALEF WITH HAMZA ABOVE" },
	{ 0x0624 , "ARABIC LETTER WAW WITH HAMZA ABOVE" },
	{ 0x0625 , "ARABIC LETTER ALEF WITH HAMZA BELOW" },
	{ 0x0626 , "ARABIC LETTER YEH WITH HAMZA ABOVE" },
	{ 0x0627 , "ARABIC LETTER ALEF" },
	{ 0x0628 , "ARABIC LETTER BEH" },
	{ 0x0629 , "ARABIC LETTER TEH MARBUTA" },
	{ 0x062A , "ARABIC LETTER TEH" },
	{ 0x062B , "ARABIC LETTER THEH" },
	{ 0x062C , "ARABIC LETTER JEEM" },
	{ 0x062D , "ARABIC LETTER HAH" },
	{ 0x062E , "ARABIC LETTER KHAH" },
	{ 0x062F , "ARABIC LETTER DAL" },
	{ 0x0630 , "ARABIC LETTER THAL" },
	{ 0x0631 , "ARABIC LETTER REH" },
	{ 0x0632 , "ARABIC LETTER ZAIN" },
	{ 0x0633 , "ARABIC LETTER SEEN" },
	{ 0x0634 , "ARABIC LETTER SHEEN" },
	{ 0x0635 , "ARABIC LETTER SAD" },
	{ 0x0636 , "ARABIC LETTER DAD" },
	{ 0x0637 , "ARABIC LETTER TAH" },
	{ 0x0638 , "ARABIC LETTER ZAH" },
	{ 0x0639 , "ARABIC LETTER AIN" },
	{ 0x063A , "ARABIC LETTER GHAIN" },
	{ 0x0640 , "ARABIC TATWEEL" },
	{ 0x0641 , "ARABIC LETTER FEH" },
	{ 0x0642 , "ARABIC LETTER QAF" },
	{ 0x0643 , "ARABIC LETTER KAF" },
	{ 0x0644 , "ARABIC LETTER LAM" },
	{ 0x0645 , "ARABIC LETTER MEEM" },
	{ 0x0646 , "ARABIC LETTER NOON" },
	{ 0x0647 , "ARABIC LETTER HEH" },
	{ 0x0648 , "ARABIC LETTER WAW" },
	{ 0x0649 , "ARABIC LETTER ALEF MAKSURA" },
	{ 0x064A , "ARABIC LETTER YEH" },
	{ 0x064B , "ARABIC FATHATAN" },
	{ 0x064C , "ARABIC DAMMATAN" },
	{ 0x064D , "ARABIC KASRATAN" },
	{ 0x064E , "ARABIC FATHA" },
	{ 0x064F , "ARABIC DAMMA" },
	{ 0x0650 , "ARABIC KASRA" },
	{ 0x0651 , "ARABIC SHADDA" },
	{ 0x0652 , "ARABIC SUKUN" },
	{ 0x0660 , "ARABIC-INDIC DIGIT ZERO" },
	{ 0x0661 , "ARABIC-INDIC DIGIT ONE" },
	{ 0x0662 , "ARABIC-INDIC DIGIT TWO" },
	{ 0x0663 , "ARABIC-INDIC DIGIT THREE" },
	{ 0x0664 , "ARABIC-INDIC DIGIT FOUR" },
	{ 0x0665 , "ARABIC-INDIC DIGIT FIVE" },
	{ 0x0666 , "ARABIC-INDIC DIGIT SIX" },
	{ 0x0667 , "ARABIC-INDIC DIGIT SEVEN" },
	{ 0x0668 , "ARABIC-INDIC DIGIT EIGHT" },
	{ 0x0669 , "ARABIC-INDIC DIGIT NINE" },
	{ 0x066A , "ARABIC PERCENT SIGN" },
	{ 0x066B , "ARABIC DECIMAL SEPARATOR" },
	{ 0x066C , "ARABIC THOUSANDS SEPARATOR" },
	{ 0x066D , "ARABIC FIVE POINTED STAR" },
	{ 0x0670 , "ARABIC LETTER SUPERSCRIPT ALEF" },
	{ 0x0671 , "ARABIC LETTER ALEF WASLA" },
	{ 0x0672 , "ARABIC LETTER ALEF WITH WAVY HAMZA ABOVE" },
	{ 0x0673 , "ARABIC LETTER ALEF WITH WAVY HAMZA BELOW" },
	{ 0x0674 , "ARABIC LETTER HIGH HAMZA" },
	{ 0x0675 , "ARABIC LETTER HIGH HAMZA ALEF" },
	{ 0x0676 , "ARABIC LETTER HIGH HAMZA WAW" },
	{ 0x0677 , "ARABIC LETTER U WITH HAMZA ABOVE" },
	{ 0x0678 , "ARABIC LETTER HIGH HAMZA YEH" },
	{ 0x0679 , "ARABIC LETTER TTEH" },
	{ 0x067A , "ARABIC LETTER TTEHEH" },
	{ 0x067B , "ARABIC LETTER BEEH" },
	{ 0x067C , "ARABIC LETTER TEH WITH RING" },
	{ 0x067D , "ARABIC LETTER TEH WITH THREE DOTS ABOVE DOWNWARDS" },
	{ 0x067E , "ARABIC LETTER PEH" },
	{ 0x067F , "ARABIC LETTER TEHEH" },
	{ 0x0680 , "ARABIC LETTER BEHEH" },
	{ 0x0681 , "ARABIC LETTER HAH WITH HAMZA ABOVE" },
	{ 0x0682 , "ARABIC LETTER HAH WITH TWO DOTS VERTICAL ABOVE" },
	{ 0x0683 , "ARABIC LETTER NYEH" },
	{ 0x0684 , "ARABIC LETTER DYEH" },
	{ 0x0685 , "ARABIC LETTER HAH WITH THREE DOTS ABOVE" },
	{ 0x0686 , "ARABIC LETTER TCHEH" },
	{ 0x0687 , "ARABIC LETTER TCHEHEH" },
	{ 0x0688 , "ARABIC LETTER DDAL" },
	{ 0x0689 , "ARABIC LETTER DAL WITH RING" },
	{ 0x068A , "ARABIC LETTER DAL WITH DOT BELOW" },
	{ 0x068B , "ARABIC LETTER DAL WITH DOT BELOW AND SMALL TAH" },
	{ 0x068C , "ARABIC LETTER DAHAL" },
	{ 0x068D , "ARABIC LETTER DDAHAL" },
	{ 0x068E , "ARABIC LETTER DUL" },
	{ 0x068F , "ARABIC LETTER DAL WITH THREE DOTS ABOVE DOWNWARDS" },
	{ 0x0690 , "ARABIC LETTER DAL WITH FOUR DOTS ABOVE" },
	{ 0x0691 , "ARABIC LETTER RREH" },
	{ 0x0692 , "ARABIC LETTER REH WITH SMALL V" },
	{ 0x0693 , "ARABIC LETTER REH WITH RING" },
	{ 0x0694 , "ARABIC LETTER REH WITH DOT BELOW" },
	{ 0x0695 , "ARABIC LETTER REH WITH SMALL V BELOW" },
	{ 0x0696 , "ARABIC LETTER REH WITH DOT BELOW AND DOT ABOVE" },
	{ 0x0697 , "ARABIC LETTER REH WITH TWO DOTS ABOVE" },
	{ 0x0698 , "ARABIC LETTER JEH" },
	{ 0x0699 , "ARABIC LETTER REH WITH FOUR DOTS ABOVE" },
	{ 0x069A , "ARABIC LETTER SEEN WITH DOT BELOW AND DOT ABOVE" },
	{ 0x069B , "ARABIC LETTER SEEN WITH THREE DOTS BELOW" },
	{ 0x069C , "ARABIC LETTER SEEN WITH THREE DOTS BELOW AND THREE DOTS ABOVE" },
	{ 0x069D , "ARABIC LETTER SAD WITH TWO DOTS BELOW" },
	{ 0x069E , "ARABIC LETTER SAD WITH THREE DOTS ABOVE" },
	{ 0x069F , "ARABIC LETTER TAH WITH THREE DOTS ABOVE" },
	{ 0x06A0 , "ARABIC LETTER AIN WITH THREE DOTS ABOVE" },
	{ 0x06A1 , "ARABIC LETTER DOTLESS FEH" },
	{ 0x06A2 , "ARABIC LETTER FEH WITH DOT MOVED BELOW" },
	{ 0x06A3 , "ARABIC LETTER FEH WITH DOT BELOW" },
	{ 0x06A4 , "ARABIC LETTER VEH" },
	{ 0x06A5 , "ARABIC LETTER FEH WITH THREE DOTS BELOW" },
	{ 0x06A6 , "ARABIC LETTER PEHEH" },
	{ 0x06A7 , "ARABIC LETTER QAF WITH DOT ABOVE" },
	{ 0x06A8 , "ARABIC LETTER QAF WITH THREE DOTS ABOVE" },
	{ 0x06A9 , "ARABIC LETTER KEHEH" },
	{ 0x06AA , "ARABIC LETTER SWASH KAF" },
	{ 0x06AB , "ARABIC LETTER KAF WITH RING" },
	{ 0x06AC , "ARABIC LETTER KAF WITH DOT ABOVE" },
	{ 0x06AD , "ARABIC LETTER NG" },
	{ 0x06AE , "ARABIC LETTER KAF WITH THREE DOTS BELOW" },
	{ 0x06AF , "ARABIC LETTER GAF" },
	{ 0x06B0 , "ARABIC LETTER GAF WITH RING" },
	{ 0x06B1 , "ARABIC LETTER NGOEH" },
	{ 0x06B2 , "ARABIC LETTER GAF WITH TWO DOTS BELOW" },
	{ 0x06B3 , "ARABIC LETTER GUEH" },
	{ 0x06B4 , "ARABIC LETTER GAF WITH THREE DOTS ABOVE" },
	{ 0x06B5 , "ARABIC LETTER LAM WITH SMALL V" },
	{ 0x06B6 , "ARABIC LETTER LAM WITH DOT ABOVE" },
	{ 0x06B7 , "ARABIC LETTER LAM WITH THREE DOTS ABOVE" },
	{ 0x06BA , "ARABIC LETTER NOON GHUNNA" },
	{ 0x06BB , "ARABIC LETTER RNOON" },
	{ 0x06BC , "ARABIC LETTER NOON WITH RING" },
	{ 0x06BD , "ARABIC LETTER NOON WITH THREE DOTS ABOVE" },
	{ 0x06BE , "ARABIC LETTER HEH DOACHASHMEE" },
	{ 0x06C0 , "ARABIC LETTER HEH WITH YEH ABOVE" },
	{ 0x06C1 , "ARABIC LETTER HEH GOAL" },
	{ 0x06C2 , "ARABIC LETTER HEH GOAL WITH HAMZA ABOVE" },
	{ 0x06C3 , "ARABIC LETTER TEH MARBUTA GOAL" },
	{ 0x06C4 , "ARABIC LETTER WAW WITH RING" },
	{ 0x06C5 , "ARABIC LETTER KIRGHIZ OE" },
	{ 0x06C6 , "ARABIC LETTER OE" },
	{ 0x06C7 , "ARABIC LETTER U" },
	{ 0x06C8 , "ARABIC LETTER YU" },
	{ 0x06C9 , "ARABIC LETTER KIRGHIZ YU" },
	{ 0x06CA , "ARABIC LETTER WAW WITH TWO DOTS ABOVE" },
	{ 0x06CB , "ARABIC LETTER VE" },
	{ 0x06CC , "ARABIC LETTER FARSI YEH" },
	{ 0x06CD , "ARABIC LETTER YEH WITH TAIL" },
	{ 0x06CE , "ARABIC LETTER YEH WITH SMALL V" },
	{ 0x06D0 , "ARABIC LETTER E" },
	{ 0x06D1 , "ARABIC LETTER YEH WITH THREE DOTS BELOW" },
	{ 0x06D2 , "ARABIC LETTER YEH BARREE" },
	{ 0x06D3 , "ARABIC LETTER YEH BARREE WITH HAMZA ABOVE" },
	{ 0x06D4 , "ARABIC FULL STOP" },
	{ 0x06D5 , "ARABIC LETTER AE" },
	{ 0x06D6 , "ARABIC SMALL HIGH LIGATURE SAD WITH LAM WITH ALEF MAKSURA" },
	{ 0x06D7 , "ARABIC SMALL HIGH LIGATURE QAF WITH LAM WITH ALEF MAKSURA" },
	{ 0x06D8 , "ARABIC SMALL HIGH MEEM INITIAL FORM" },
	{ 0x06D9 , "ARABIC SMALL HIGH LAM ALEF" },
	{ 0x06DA , "ARABIC SMALL HIGH JEEM" },
	{ 0x06DB , "ARABIC SMALL HIGH THREE DOTS" },
	{ 0x06DC , "ARABIC SMALL HIGH SEEN" },
	{ 0x06DD , "ARABIC END OF AYAH" },
	{ 0x06DE , "ARABIC START OF RUB EL HIZB" },
	{ 0x06DF , "ARABIC SMALL HIGH ROUNDED ZERO" },
	{ 0x06E0 , "ARABIC SMALL HIGH UPRIGHT RECTANGULAR ZERO" },
	{ 0x06E1 , "ARABIC SMALL HIGH DOTLESS HEAD OF KHAH" },
	{ 0x06E2 , "ARABIC SMALL HIGH MEEM ISOLATED FORM" },
	{ 0x06E3 , "ARABIC SMALL LOW SEEN" },
	{ 0x06E4 , "ARABIC SMALL HIGH MADDA" },
	{ 0x06E5 , "ARABIC SMALL WAW" },
	{ 0x06E6 , "ARABIC SMALL YEH" },
	{ 0x06E7 , "ARABIC SMALL HIGH YEH" },
	{ 0x06E8 , "ARABIC SMALL HIGH NOON" },
	{ 0x06E9 , "ARABIC PLACE OF SAJDAH" },
	{ 0x06EA , "ARABIC EMPTY CENTRE LOW STOP" },
	{ 0x06EB , "ARABIC EMPTY CENTRE HIGH STOP" },
	{ 0x06EC , "ARABIC ROUNDED HIGH STOP WITH FILLED CENTRE" },
	{ 0x06ED , "ARABIC SMALL LOW MEEM" },
	{ 0x06F0 , "EXTENDED ARABIC-INDIC DIGIT ZERO" },
	{ 0x06F1 , "EXTENDED ARABIC-INDIC DIGIT ONE" },
	{ 0x06F2 , "EXTENDED ARABIC-INDIC DIGIT TWO" },
	{ 0x06F3 , "EXTENDED ARABIC-INDIC DIGIT THREE" },
	{ 0x06F4 , "EXTENDED ARABIC-INDIC DIGIT FOUR" },
	{ 0x06F5 , "EXTENDED ARABIC-INDIC DIGIT FIVE" },
	{ 0x06F6 , "EXTENDED ARABIC-INDIC DIGIT SIX" },
	{ 0x06F7 , "EXTENDED ARABIC-INDIC DIGIT SEVEN" },
	{ 0x06F8 , "EXTENDED ARABIC-INDIC DIGIT EIGHT" },
	{ 0x06F9 , "EXTENDED ARABIC-INDIC DIGIT NINE" },
	{ 0x0901 , "DEVANAGARI SIGN CANDRABINDU" },
	{ 0x0902 , "DEVANAGARI SIGN ANUSVARA" },
	{ 0x0903 , "DEVANAGARI SIGN VISARGA" },
	{ 0x0905 , "DEVANAGARI LETTER A" },
	{ 0x0906 , "DEVANAGARI LETTER AA" },
	{ 0x0907 , "DEVANAGARI LETTER I" },
	{ 0x0908 , "DEVANAGARI LETTER II" },
	{ 0x0909 , "DEVANAGARI LETTER U" },
	{ 0x090A , "DEVANAGARI LETTER UU" },
	{ 0x090B , "DEVANAGARI LETTER VOCALIC R" },
	{ 0x090C , "DEVANAGARI LETTER VOCALIC L" },
	{ 0x090D , "DEVANAGARI LETTER CANDRA E" },
	{ 0x090E , "DEVANAGARI LETTER SHORT E" },
	{ 0x090F , "DEVANAGARI LETTER E" },
	{ 0x0910 , "DEVANAGARI LETTER AI" },
	{ 0x0911 , "DEVANAGARI LETTER CANDRA O" },
	{ 0x0912 , "DEVANAGARI LETTER SHORT O" },
	{ 0x0913 , "DEVANAGARI LETTER O" },
	{ 0x0914 , "DEVANAGARI LETTER AU" },
	{ 0x0915 , "DEVANAGARI LETTER KA" },
	{ 0x0916 , "DEVANAGARI LETTER KHA" },
	{ 0x0917 , "DEVANAGARI LETTER GA" },
	{ 0x0918 , "DEVANAGARI LETTER GHA" },
	{ 0x0919 , "DEVANAGARI LETTER NGA" },
	{ 0x091A , "DEVANAGARI LETTER CA" },
	{ 0x091B , "DEVANAGARI LETTER CHA" },
	{ 0x091C , "DEVANAGARI LETTER JA" },
	{ 0x091D , "DEVANAGARI LETTER JHA" },
	{ 0x091E , "DEVANAGARI LETTER NYA" },
	{ 0x091F , "DEVANAGARI LETTER TTA" },
	{ 0x0920 , "DEVANAGARI LETTER TTHA" },
	{ 0x0921 , "DEVANAGARI LETTER DDA" },
	{ 0x0922 , "DEVANAGARI LETTER DDHA" },
	{ 0x0923 , "DEVANAGARI LETTER NNA" },
	{ 0x0924 , "DEVANAGARI LETTER TA" },
	{ 0x0925 , "DEVANAGARI LETTER THA" },
	{ 0x0926 , "DEVANAGARI LETTER DA" },
	{ 0x0927 , "DEVANAGARI LETTER DHA" },
	{ 0x0928 , "DEVANAGARI LETTER NA" },
	{ 0x0929 , "DEVANAGARI LETTER NNNA" },
	{ 0x092A , "DEVANAGARI LETTER PA" },
	{ 0x092B , "DEVANAGARI LETTER PHA" },
	{ 0x092C , "DEVANAGARI LETTER BA" },
	{ 0x092D , "DEVANAGARI LETTER BHA" },
	{ 0x092E , "DEVANAGARI LETTER MA" },
	{ 0x092F , "DEVANAGARI LETTER YA" },
	{ 0x0930 , "DEVANAGARI LETTER RA" },
	{ 0x0931 , "DEVANAGARI LETTER RRA" },
	{ 0x0932 , "DEVANAGARI LETTER LA" },
	{ 0x0933 , "DEVANAGARI LETTER LLA" },
	{ 0x0934 , "DEVANAGARI LETTER LLLA" },
	{ 0x0935 , "DEVANAGARI LETTER VA" },
	{ 0x0936 , "DEVANAGARI LETTER SHA" },
	{ 0x0937 , "DEVANAGARI LETTER SSA" },
	{ 0x0938 , "DEVANAGARI LETTER SA" },
	{ 0x0939 , "DEVANAGARI LETTER HA" },
	{ 0x093C , "DEVANAGARI SIGN NUKTA" },
	{ 0x093D , "DEVANAGARI SIGN AVAGRAHA" },
	{ 0x093E , "DEVANAGARI VOWEL SIGN AA" },
	{ 0x093F , "DEVANAGARI VOWEL SIGN I" },
	{ 0x0940 , "DEVANAGARI VOWEL SIGN II" },
	{ 0x0941 , "DEVANAGARI VOWEL SIGN U" },
	{ 0x0942 , "DEVANAGARI VOWEL SIGN UU" },
	{ 0x0943 , "DEVANAGARI VOWEL SIGN VOCALIC R" },
	{ 0x0944 , "DEVANAGARI VOWEL SIGN VOCALIC RR" },
	{ 0x0945 , "DEVANAGARI VOWEL SIGN CANDRA E" },
	{ 0x0946 , "DEVANAGARI VOWEL SIGN SHORT E" },
	{ 0x0947 , "DEVANAGARI VOWEL SIGN E" },
	{ 0x0948 , "DEVANAGARI VOWEL SIGN AI" },
	{ 0x0949 , "DEVANAGARI VOWEL SIGN CANDRA O" },
	{ 0x094A , "DEVANAGARI VOWEL SIGN SHORT O" },
	{ 0x094B , "DEVANAGARI VOWEL SIGN O" },
	{ 0x094C , "DEVANAGARI VOWEL SIGN AU" },
	{ 0x094D , "DEVANAGARI SIGN VIRAMA" },
	{ 0x0950 , "DEVANAGARI OM" },
	{ 0x0951 , "DEVANAGARI STRESS SIGN UDATTA" },
	{ 0x0952 , "DEVANAGARI STRESS SIGN ANUDATTA" },
	{ 0x0953 , "DEVANAGARI GRAVE ACCENT" },
	{ 0x0954 , "DEVANAGARI ACUTE ACCENT" },
	{ 0x0958 , "DEVANAGARI LETTER QA" },
	{ 0x0959 , "DEVANAGARI LETTER KHHA" },
	{ 0x095A , "DEVANAGARI LETTER GHHA" },
	{ 0x095B , "DEVANAGARI LETTER ZA" },
	{ 0x095C , "DEVANAGARI LETTER DDDHA" },
	{ 0x095D , "DEVANAGARI LETTER RHA" },
	{ 0x095E , "DEVANAGARI LETTER FA" },
	{ 0x095F , "DEVANAGARI LETTER YYA" },
	{ 0x0960 , "DEVANAGARI LETTER VOCALIC RR" },
	{ 0x0961 , "DEVANAGARI LETTER VOCALIC LL" },
	{ 0x0962 , "DEVANAGARI VOWEL SIGN VOCALIC L" },
	{ 0x0963 , "DEVANAGARI VOWEL SIGN VOCALIC LL" },
	{ 0x0964 , "DEVANAGARI DANDA" },
	{ 0x0965 , "DEVANAGARI DOUBLE DANDA" },
	{ 0x0966 , "DEVANAGARI DIGIT ZERO" },
	{ 0x0967 , "DEVANAGARI DIGIT ONE" },
	{ 0x0968 , "DEVANAGARI DIGIT TWO" },
	{ 0x0969 , "DEVANAGARI DIGIT THREE" },
	{ 0x096A , "DEVANAGARI DIGIT FOUR" },
	{ 0x096B , "DEVANAGARI DIGIT FIVE" },
	{ 0x096C , "DEVANAGARI DIGIT SIX" },
	{ 0x096D , "DEVANAGARI DIGIT SEVEN" },
	{ 0x096E , "DEVANAGARI DIGIT EIGHT" },
	{ 0x096F , "DEVANAGARI DIGIT NINE" },
	{ 0x0970 , "DEVANAGARI ABBREVIATION SIGN" },
	{ 0x0981 , "BENGALI SIGN CANDRABINDU" },
	{ 0x0982 , "BENGALI SIGN ANUSVARA" },
	{ 0x0983 , "BENGALI SIGN VISARGA" },
	{ 0x0985 , "BENGALI LETTER A" },
	{ 0x0986 , "BENGALI LETTER AA" },
	{ 0x0987 , "BENGALI LETTER I" },
	{ 0x0988 , "BENGALI LETTER II" },
	{ 0x0989 , "BENGALI LETTER U" },
	{ 0x098A , "BENGALI LETTER UU" },
	{ 0x098B , "BENGALI LETTER VOCALIC R" },
	{ 0x098C , "BENGALI LETTER VOCALIC L" },
	{ 0x098F , "BENGALI LETTER E" },
	{ 0x0990 , "BENGALI LETTER AI" },
	{ 0x0993 , "BENGALI LETTER O" },
	{ 0x0994 , "BENGALI LETTER AU" },
	{ 0x0995 , "BENGALI LETTER KA" },
	{ 0x0996 , "BENGALI LETTER KHA" },
	{ 0x0997 , "BENGALI LETTER GA" },
	{ 0x0998 , "BENGALI LETTER GHA" },
	{ 0x0999 , "BENGALI LETTER NGA" },
	{ 0x099A , "BENGALI LETTER CA" },
	{ 0x099B , "BENGALI LETTER CHA" },
	{ 0x099C , "BENGALI LETTER JA" },
	{ 0x099D , "BENGALI LETTER JHA" },
	{ 0x099E , "BENGALI LETTER NYA" },
	{ 0x099F , "BENGALI LETTER TTA" },
	{ 0x09A0 , "BENGALI LETTER TTHA" },
	{ 0x09A1 , "BENGALI LETTER DDA" },
	{ 0x09A2 , "BENGALI LETTER DDHA" },
	{ 0x09A3 , "BENGALI LETTER NNA" },
	{ 0x09A4 , "BENGALI LETTER TA" },
	{ 0x09A5 , "BENGALI LETTER THA" },
	{ 0x09A6 , "BENGALI LETTER DA" },
	{ 0x09A7 , "BENGALI LETTER DHA" },
	{ 0x09A8 , "BENGALI LETTER NA" },
	{ 0x09AA , "BENGALI LETTER PA" },
	{ 0x09AB , "BENGALI LETTER PHA" },
	{ 0x09AC , "BENGALI LETTER BA" },
	{ 0x09AD , "BENGALI LETTER BHA" },
	{ 0x09AE , "BENGALI LETTER MA" },
	{ 0x09AF , "BENGALI LETTER YA" },
	{ 0x09B0 , "BENGALI LETTER RA" },
	{ 0x09B2 , "BENGALI LETTER LA" },
	{ 0x09B6 , "BENGALI LETTER SHA" },
	{ 0x09B7 , "BENGALI LETTER SSA" },
	{ 0x09B8 , "BENGALI LETTER SA" },
	{ 0x09B9 , "BENGALI LETTER HA" },
	{ 0x09BC , "BENGALI SIGN NUKTA" },
	{ 0x09BE , "BENGALI VOWEL SIGN AA" },
	{ 0x09BF , "BENGALI VOWEL SIGN I" },
	{ 0x09C0 , "BENGALI VOWEL SIGN II" },
	{ 0x09C1 , "BENGALI VOWEL SIGN U" },
	{ 0x09C2 , "BENGALI VOWEL SIGN UU" },
	{ 0x09C3 , "BENGALI VOWEL SIGN VOCALIC R" },
	{ 0x09C4 , "BENGALI VOWEL SIGN VOCALIC RR" },
	{ 0x09C7 , "BENGALI VOWEL SIGN E" },
	{ 0x09C8 , "BENGALI VOWEL SIGN AI" },
	{ 0x09CB , "BENGALI VOWEL SIGN O" },
	{ 0x09CC , "BENGALI VOWEL SIGN AU" },
	{ 0x09CD , "BENGALI SIGN VIRAMA" },
	{ 0x09D7 , "BENGALI AU LENGTH MARK" },
	{ 0x09DC , "BENGALI LETTER RRA" },
	{ 0x09DD , "BENGALI LETTER RHA" },
	{ 0x09DF , "BENGALI LETTER YYA" },
	{ 0x09E0 , "BENGALI LETTER VOCALIC RR" },
	{ 0x09E1 , "BENGALI LETTER VOCALIC LL" },
	{ 0x09E2 , "BENGALI VOWEL SIGN VOCALIC L" },
	{ 0x09E3 , "BENGALI VOWEL SIGN VOCALIC LL" },
	{ 0x09E6 , "BENGALI DIGIT ZERO" },
	{ 0x09E7 , "BENGALI DIGIT ONE" },
	{ 0x09E8 , "BENGALI DIGIT TWO" },
	{ 0x09E9 , "BENGALI DIGIT THREE" },
	{ 0x09EA , "BENGALI DIGIT FOUR" },
	{ 0x09EB , "BENGALI DIGIT FIVE" },
	{ 0x09EC , "BENGALI DIGIT SIX" },
	{ 0x09ED , "BENGALI DIGIT SEVEN" },
	{ 0x09EE , "BENGALI DIGIT EIGHT" },
	{ 0x09EF , "BENGALI DIGIT NINE" },
	{ 0x09F0 , "BENGALI LETTER RA WITH MIDDLE DIAGONAL" },
	{ 0x09F1 , "BENGALI LETTER RA WITH LOWER DIAGONAL" },
	{ 0x09F2 , "BENGALI RUPEE MARK" },
	{ 0x09F3 , "BENGALI RUPEE SIGN" },
	{ 0x09F4 , "BENGALI CURRENCY NUMERATOR ONE" },
	{ 0x09F5 , "BENGALI CURRENCY NUMERATOR TWO" },
	{ 0x09F6 , "BENGALI CURRENCY NUMERATOR THREE" },
	{ 0x09F7 , "BENGALI CURRENCY NUMERATOR FOUR" },
	{ 0x09F8 , "BENGALI CURRENCY NUMERATOR ONE LESS THAN THE DENOMINATOR" },
	{ 0x09F9 , "BENGALI CURRENCY DENOMINATOR SIXTEEN" },
	{ 0x09FA , "BENGALI ISSHAR" },
	{ 0x0A02 , "GURMUKHI SIGN BINDI" },
	{ 0x0A05 , "GURMUKHI LETTER A" },
	{ 0x0A06 , "GURMUKHI LETTER AA" },
	{ 0x0A07 , "GURMUKHI LETTER I" },
	{ 0x0A08 , "GURMUKHI LETTER II" },
	{ 0x0A09 , "GURMUKHI LETTER U" },
	{ 0x0A0A , "GURMUKHI LETTER UU" },
	{ 0x0A0F , "GURMUKHI LETTER EE" },
	{ 0x0A10 , "GURMUKHI LETTER AI" },
	{ 0x0A13 , "GURMUKHI LETTER OO" },
	{ 0x0A14 , "GURMUKHI LETTER AU" },
	{ 0x0A15 , "GURMUKHI LETTER KA" },
	{ 0x0A16 , "GURMUKHI LETTER KHA" },
	{ 0x0A17 , "GURMUKHI LETTER GA" },
	{ 0x0A18 , "GURMUKHI LETTER GHA" },
	{ 0x0A19 , "GURMUKHI LETTER NGA" },
	{ 0x0A1A , "GURMUKHI LETTER CA" },
	{ 0x0A1B , "GURMUKHI LETTER CHA" },
	{ 0x0A1C , "GURMUKHI LETTER JA" },
	{ 0x0A1D , "GURMUKHI LETTER JHA" },
	{ 0x0A1E , "GURMUKHI LETTER NYA" },
	{ 0x0A1F , "GURMUKHI LETTER TTA" },
	{ 0x0A20 , "GURMUKHI LETTER TTHA" },
	{ 0x0A21 , "GURMUKHI LETTER DDA" },
	{ 0x0A22 , "GURMUKHI LETTER DDHA" },
	{ 0x0A23 , "GURMUKHI LETTER NNA" },
	{ 0x0A24 , "GURMUKHI LETTER TA" },
	{ 0x0A25 , "GURMUKHI LETTER THA" },
	{ 0x0A26 , "GURMUKHI LETTER DA" },
	{ 0x0A27 , "GURMUKHI LETTER DHA" },
	{ 0x0A28 , "GURMUKHI LETTER NA" },
	{ 0x0A2A , "GURMUKHI LETTER PA" },
	{ 0x0A2B , "GURMUKHI LETTER PHA" },
	{ 0x0A2C , "GURMUKHI LETTER BA" },
	{ 0x0A2D , "GURMUKHI LETTER BHA" },
	{ 0x0A2E , "GURMUKHI LETTER MA" },
	{ 0x0A2F , "GURMUKHI LETTER YA" },
	{ 0x0A30 , "GURMUKHI LETTER RA" },
	{ 0x0A32 , "GURMUKHI LETTER LA" },
	{ 0x0A33 , "GURMUKHI LETTER LLA" },
	{ 0x0A35 , "GURMUKHI LETTER VA" },
	{ 0x0A36 , "GURMUKHI LETTER SHA" },
	{ 0x0A38 , "GURMUKHI LETTER SA" },
	{ 0x0A39 , "GURMUKHI LETTER HA" },
	{ 0x0A3C , "GURMUKHI SIGN NUKTA" },
	{ 0x0A3E , "GURMUKHI VOWEL SIGN AA" },
	{ 0x0A3F , "GURMUKHI VOWEL SIGN I" },
	{ 0x0A40 , "GURMUKHI VOWEL SIGN II" },
	{ 0x0A41 , "GURMUKHI VOWEL SIGN U" },
	{ 0x0A42 , "GURMUKHI VOWEL SIGN UU" },
	{ 0x0A47 , "GURMUKHI VOWEL SIGN EE" },
	{ 0x0A48 , "GURMUKHI VOWEL SIGN AI" },
	{ 0x0A4B , "GURMUKHI VOWEL SIGN OO" },
	{ 0x0A4C , "GURMUKHI VOWEL SIGN AU" },
	{ 0x0A4D , "GURMUKHI SIGN VIRAMA" },
	{ 0x0A59 , "GURMUKHI LETTER KHHA" },
	{ 0x0A5A , "GURMUKHI LETTER GHHA" },
	{ 0x0A5B , "GURMUKHI LETTER ZA" },
	{ 0x0A5C , "GURMUKHI LETTER RRA" },
	{ 0x0A5E , "GURMUKHI LETTER FA" },
	{ 0x0A66 , "GURMUKHI DIGIT ZERO" },
	{ 0x0A67 , "GURMUKHI DIGIT ONE" },
	{ 0x0A68 , "GURMUKHI DIGIT TWO" },
	{ 0x0A69 , "GURMUKHI DIGIT THREE" },
	{ 0x0A6A , "GURMUKHI DIGIT FOUR" },
	{ 0x0A6B , "GURMUKHI DIGIT FIVE" },
	{ 0x0A6C , "GURMUKHI DIGIT SIX" },
	{ 0x0A6D , "GURMUKHI DIGIT SEVEN" },
	{ 0x0A6E , "GURMUKHI DIGIT EIGHT" },
	{ 0x0A6F , "GURMUKHI DIGIT NINE" },
	{ 0x0A70 , "GURMUKHI TIPPI" },
	{ 0x0A71 , "GURMUKHI ADDAK" },
	{ 0x0A72 , "GURMUKHI IRI" },
	{ 0x0A73 , "GURMUKHI URA" },
	{ 0x0A74 , "GURMUKHI EK ONKAR" },
	{ 0x0A81 , "GUJARATI SIGN CANDRABINDU" },
	{ 0x0A82 , "GUJARATI SIGN ANUSVARA" },
	{ 0x0A83 , "GUJARATI SIGN VISARGA" },
	{ 0x0A85 , "GUJARATI LETTER A" },
	{ 0x0A86 , "GUJARATI LETTER AA" },
	{ 0x0A87 , "GUJARATI LETTER I" },
	{ 0x0A88 , "GUJARATI LETTER II" },
	{ 0x0A89 , "GUJARATI LETTER U" },
	{ 0x0A8A , "GUJARATI LETTER UU" },
	{ 0x0A8B , "GUJARATI LETTER VOCALIC R" },
	{ 0x0A8D , "GUJARATI VOWEL CANDRA E" },
	{ 0x0A8F , "GUJARATI LETTER E" },
	{ 0x0A90 , "GUJARATI LETTER AI" },
	{ 0x0A91 , "GUJARATI VOWEL CANDRA O" },
	{ 0x0A93 , "GUJARATI LETTER O" },
	{ 0x0A94 , "GUJARATI LETTER AU" },
	{ 0x0A95 , "GUJARATI LETTER KA" },
	{ 0x0A96 , "GUJARATI LETTER KHA" },
	{ 0x0A97 , "GUJARATI LETTER GA" },
	{ 0x0A98 , "GUJARATI LETTER GHA" },
	{ 0x0A99 , "GUJARATI LETTER NGA" },
	{ 0x0A9A , "GUJARATI LETTER CA" },
	{ 0x0A9B , "GUJARATI LETTER CHA" },
	{ 0x0A9C , "GUJARATI LETTER JA" },
	{ 0x0A9D , "GUJARATI LETTER JHA" },
	{ 0x0A9E , "GUJARATI LETTER NYA" },
	{ 0x0A9F , "GUJARATI LETTER TTA" },
	{ 0x0AA0 , "GUJARATI LETTER TTHA" },
	{ 0x0AA1 , "GUJARATI LETTER DDA" },
	{ 0x0AA2 , "GUJARATI LETTER DDHA" },
	{ 0x0AA3 , "GUJARATI LETTER NNA" },
	{ 0x0AA4 , "GUJARATI LETTER TA" },
	{ 0x0AA5 , "GUJARATI LETTER THA" },
	{ 0x0AA6 , "GUJARATI LETTER DA" },
	{ 0x0AA7 , "GUJARATI LETTER DHA" },
	{ 0x0AA8 , "GUJARATI LETTER NA" },
	{ 0x0AAA , "GUJARATI LETTER PA" },
	{ 0x0AAB , "GUJARATI LETTER PHA" },
	{ 0x0AAC , "GUJARATI LETTER BA" },
	{ 0x0AAD , "GUJARATI LETTER BHA" },
	{ 0x0AAE , "GUJARATI LETTER MA" },
	{ 0x0AAF , "GUJARATI LETTER YA" },
	{ 0x0AB0 , "GUJARATI LETTER RA" },
	{ 0x0AB2 , "GUJARATI LETTER LA" },
	{ 0x0AB3 , "GUJARATI LETTER LLA" },
	{ 0x0AB5 , "GUJARATI LETTER VA" },
	{ 0x0AB6 , "GUJARATI LETTER SHA" },
	{ 0x0AB7 , "GUJARATI LETTER SSA" },
	{ 0x0AB8 , "GUJARATI LETTER SA" },
	{ 0x0AB9 , "GUJARATI LETTER HA" },
	{ 0x0ABC , "GUJARATI SIGN NUKTA" },
	{ 0x0ABD , "GUJARATI SIGN AVAGRAHA" },
	{ 0x0ABE , "GUJARATI VOWEL SIGN AA" },
	{ 0x0ABF , "GUJARATI VOWEL SIGN I" },
	{ 0x0AC0 , "GUJARATI VOWEL SIGN II" },
	{ 0x0AC1 , "GUJARATI VOWEL SIGN U" },
	{ 0x0AC2 , "GUJARATI VOWEL SIGN UU" },
	{ 0x0AC3 , "GUJARATI VOWEL SIGN VOCALIC R" },
	{ 0x0AC4 , "GUJARATI VOWEL SIGN VOCALIC RR" },
	{ 0x0AC5 , "GUJARATI VOWEL SIGN CANDRA E" },
	{ 0x0AC7 , "GUJARATI VOWEL SIGN E" },
	{ 0x0AC8 , "GUJARATI VOWEL SIGN AI" },
	{ 0x0AC9 , "GUJARATI VOWEL SIGN CANDRA O" },
	{ 0x0ACB , "GUJARATI VOWEL SIGN O" },
	{ 0x0ACC , "GUJARATI VOWEL SIGN AU" },
	{ 0x0ACD , "GUJARATI SIGN VIRAMA" },
	{ 0x0AD0 , "GUJARATI OM" },
	{ 0x0AE0 , "GUJARATI LETTER VOCALIC RR" },
	{ 0x0AE6 , "GUJARATI DIGIT ZERO" },
	{ 0x0AE7 , "GUJARATI DIGIT ONE" },
	{ 0x0AE8 , "GUJARATI DIGIT TWO" },
	{ 0x0AE9 , "GUJARATI DIGIT THREE" },
	{ 0x0AEA , "GUJARATI DIGIT FOUR" },
	{ 0x0AEB , "GUJARATI DIGIT FIVE" },
	{ 0x0AEC , "GUJARATI DIGIT SIX" },
	{ 0x0AED , "GUJARATI DIGIT SEVEN" },
	{ 0x0AEE , "GUJARATI DIGIT EIGHT" },
	{ 0x0AEF , "GUJARATI DIGIT NINE" },
	{ 0x0B01 , "ORIYA SIGN CANDRABINDU" },
	{ 0x0B02 , "ORIYA SIGN ANUSVARA" },
	{ 0x0B03 , "ORIYA SIGN VISARGA" },
	{ 0x0B05 , "ORIYA LETTER A" },
	{ 0x0B06 , "ORIYA LETTER AA" },
	{ 0x0B07 , "ORIYA LETTER I" },
	{ 0x0B08 , "ORIYA LETTER II" },
	{ 0x0B09 , "ORIYA LETTER U" },
	{ 0x0B0A , "ORIYA LETTER UU" },
	{ 0x0B0B , "ORIYA LETTER VOCALIC R" },
	{ 0x0B0C , "ORIYA LETTER VOCALIC L" },
	{ 0x0B0F , "ORIYA LETTER E" },
	{ 0x0B10 , "ORIYA LETTER AI" },
	{ 0x0B13 , "ORIYA LETTER O" },
	{ 0x0B14 , "ORIYA LETTER AU" },
	{ 0x0B15 , "ORIYA LETTER KA" },
	{ 0x0B16 , "ORIYA LETTER KHA" },
	{ 0x0B17 , "ORIYA LETTER GA" },
	{ 0x0B18 , "ORIYA LETTER GHA" },
	{ 0x0B19 , "ORIYA LETTER NGA" },
	{ 0x0B1A , "ORIYA LETTER CA" },
	{ 0x0B1B , "ORIYA LETTER CHA" },
	{ 0x0B1C , "ORIYA LETTER JA" },
	{ 0x0B1D , "ORIYA LETTER JHA" },
	{ 0x0B1E , "ORIYA LETTER NYA" },
	{ 0x0B1F , "ORIYA LETTER TTA" },
	{ 0x0B20 , "ORIYA LETTER TTHA" },
	{ 0x0B21 , "ORIYA LETTER DDA" },
	{ 0x0B22 , "ORIYA LETTER DDHA" },
	{ 0x0B23 , "ORIYA LETTER NNA" },
	{ 0x0B24 , "ORIYA LETTER TA" },
	{ 0x0B25 , "ORIYA LETTER THA" },
	{ 0x0B26 , "ORIYA LETTER DA" },
	{ 0x0B27 , "ORIYA LETTER DHA" },
	{ 0x0B28 , "ORIYA LETTER NA" },
	{ 0x0B2A , "ORIYA LETTER PA" },
	{ 0x0B2B , "ORIYA LETTER PHA" },
	{ 0x0B2C , "ORIYA LETTER BA" },
	{ 0x0B2D , "ORIYA LETTER BHA" },
	{ 0x0B2E , "ORIYA LETTER MA" },
	{ 0x0B2F , "ORIYA LETTER YA" },
	{ 0x0B30 , "ORIYA LETTER RA" },
	{ 0x0B32 , "ORIYA LETTER LA" },
	{ 0x0B33 , "ORIYA LETTER LLA" },
	{ 0x0B36 , "ORIYA LETTER SHA" },
	{ 0x0B37 , "ORIYA LETTER SSA" },
	{ 0x0B38 , "ORIYA LETTER SA" },
	{ 0x0B39 , "ORIYA LETTER HA" },
	{ 0x0B3C , "ORIYA SIGN NUKTA" },
	{ 0x0B3D , "ORIYA SIGN AVAGRAHA" },
	{ 0x0B3E , "ORIYA VOWEL SIGN AA" },
	{ 0x0B3F , "ORIYA VOWEL SIGN I" },
	{ 0x0B40 , "ORIYA VOWEL SIGN II" },
	{ 0x0B41 , "ORIYA VOWEL SIGN U" },
	{ 0x0B42 , "ORIYA VOWEL SIGN UU" },
	{ 0x0B43 , "ORIYA VOWEL SIGN VOCALIC R" },
	{ 0x0B47 , "ORIYA VOWEL SIGN E" },
	{ 0x0B48 , "ORIYA VOWEL SIGN AI" },
	{ 0x0B4B , "ORIYA VOWEL SIGN O" },
	{ 0x0B4C , "ORIYA VOWEL SIGN AU" },
	{ 0x0B4D , "ORIYA SIGN VIRAMA" },
	{ 0x0B56 , "ORIYA AI LENGTH MARK" },
	{ 0x0B57 , "ORIYA AU LENGTH MARK" },
	{ 0x0B5C , "ORIYA LETTER RRA" },
	{ 0x0B5D , "ORIYA LETTER RHA" },
	{ 0x0B5F , "ORIYA LETTER YYA" },
	{ 0x0B60 , "ORIYA LETTER VOCALIC RR" },
	{ 0x0B61 , "ORIYA LETTER VOCALIC LL" },
	{ 0x0B66 , "ORIYA DIGIT ZERO" },
	{ 0x0B67 , "ORIYA DIGIT ONE" },
	{ 0x0B68 , "ORIYA DIGIT TWO" },
	{ 0x0B69 , "ORIYA DIGIT THREE" },
	{ 0x0B6A , "ORIYA DIGIT FOUR" },
	{ 0x0B6B , "ORIYA DIGIT FIVE" },
	{ 0x0B6C , "ORIYA DIGIT SIX" },
	{ 0x0B6D , "ORIYA DIGIT SEVEN" },
	{ 0x0B6E , "ORIYA DIGIT EIGHT" },
	{ 0x0B6F , "ORIYA DIGIT NINE" },
	{ 0x0B70 , "ORIYA ISSHAR" },
	{ 0x0B82 , "TAMIL SIGN ANUSVARA" },
	{ 0x0B83 , "TAMIL SIGN VISARGA" },
	{ 0x0B85 , "TAMIL LETTER A" },
	{ 0x0B86 , "TAMIL LETTER AA" },
	{ 0x0B87 , "TAMIL LETTER I" },
	{ 0x0B88 , "TAMIL LETTER II" },
	{ 0x0B89 , "TAMIL LETTER U" },
	{ 0x0B8A , "TAMIL LETTER UU" },
	{ 0x0B8E , "TAMIL LETTER E" },
	{ 0x0B8F , "TAMIL LETTER EE" },
	{ 0x0B90 , "TAMIL LETTER AI" },
	{ 0x0B92 , "TAMIL LETTER O" },
	{ 0x0B93 , "TAMIL LETTER OO" },
	{ 0x0B94 , "TAMIL LETTER AU" },
	{ 0x0B95 , "TAMIL LETTER KA" },
	{ 0x0B99 , "TAMIL LETTER NGA" },
	{ 0x0B9A , "TAMIL LETTER CA" },
	{ 0x0B9C , "TAMIL LETTER JA" },
	{ 0x0B9E , "TAMIL LETTER NYA" },
	{ 0x0B9F , "TAMIL LETTER TTA" },
	{ 0x0BA3 , "TAMIL LETTER NNA" },
	{ 0x0BA4 , "TAMIL LETTER TA" },
	{ 0x0BA8 , "TAMIL LETTER NA" },
	{ 0x0BA9 , "TAMIL LETTER NNNA" },
	{ 0x0BAA , "TAMIL LETTER PA" },
	{ 0x0BAE , "TAMIL LETTER MA" },
	{ 0x0BAF , "TAMIL LETTER YA" },
	{ 0x0BB0 , "TAMIL LETTER RA" },
	{ 0x0BB1 , "TAMIL LETTER RRA" },
	{ 0x0BB2 , "TAMIL LETTER LA" },
	{ 0x0BB3 , "TAMIL LETTER LLA" },
	{ 0x0BB4 , "TAMIL LETTER LLLA" },
	{ 0x0BB5 , "TAMIL LETTER VA" },
	{ 0x0BB7 , "TAMIL LETTER SSA" },
	{ 0x0BB8 , "TAMIL LETTER SA" },
	{ 0x0BB9 , "TAMIL LETTER HA" },
	{ 0x0BBE , "TAMIL VOWEL SIGN AA" },
	{ 0x0BBF , "TAMIL VOWEL SIGN I" },
	{ 0x0BC0 , "TAMIL VOWEL SIGN II" },
	{ 0x0BC1 , "TAMIL VOWEL SIGN U" },
	{ 0x0BC2 , "TAMIL VOWEL SIGN UU" },
	{ 0x0BC6 , "TAMIL VOWEL SIGN E" },
	{ 0x0BC7 , "TAMIL VOWEL SIGN EE" },
	{ 0x0BC8 , "TAMIL VOWEL SIGN AI" },
	{ 0x0BCA , "TAMIL VOWEL SIGN O" },
	{ 0x0BCB , "TAMIL VOWEL SIGN OO" },
	{ 0x0BCC , "TAMIL VOWEL SIGN AU" },
	{ 0x0BCD , "TAMIL SIGN VIRAMA" },
	{ 0x0BD7 , "TAMIL AU LENGTH MARK" },
	{ 0x0BE7 , "TAMIL DIGIT ONE" },
	{ 0x0BE8 , "TAMIL DIGIT TWO" },
	{ 0x0BE9 , "TAMIL DIGIT THREE" },
	{ 0x0BEA , "TAMIL DIGIT FOUR" },
	{ 0x0BEB , "TAMIL DIGIT FIVE" },
	{ 0x0BEC , "TAMIL DIGIT SIX" },
	{ 0x0BED , "TAMIL DIGIT SEVEN" },
	{ 0x0BEE , "TAMIL DIGIT EIGHT" },
	{ 0x0BEF , "TAMIL DIGIT NINE" },
	{ 0x0BF0 , "TAMIL NUMBER TEN" },
	{ 0x0BF1 , "TAMIL NUMBER ONE HUNDRED" },
	{ 0x0BF2 , "TAMIL NUMBER ONE THOUSAND" },
	{ 0x0C01 , "TELUGU SIGN CANDRABINDU" },
	{ 0x0C02 , "TELUGU SIGN ANUSVARA" },
	{ 0x0C03 , "TELUGU SIGN VISARGA" },
	{ 0x0C05 , "TELUGU LETTER A" },
	{ 0x0C06 , "TELUGU LETTER AA" },
	{ 0x0C07 , "TELUGU LETTER I" },
	{ 0x0C08 , "TELUGU LETTER II" },
	{ 0x0C09 , "TELUGU LETTER U" },
	{ 0x0C0A , "TELUGU LETTER UU" },
	{ 0x0C0B , "TELUGU LETTER VOCALIC R" },
	{ 0x0C0C , "TELUGU LETTER VOCALIC L" },
	{ 0x0C0E , "TELUGU LETTER E" },
	{ 0x0C0F , "TELUGU LETTER EE" },
	{ 0x0C10 , "TELUGU LETTER AI" },
	{ 0x0C12 , "TELUGU LETTER O" },
	{ 0x0C13 , "TELUGU LETTER OO" },
	{ 0x0C14 , "TELUGU LETTER AU" },
	{ 0x0C15 , "TELUGU LETTER KA" },
	{ 0x0C16 , "TELUGU LETTER KHA" },
	{ 0x0C17 , "TELUGU LETTER GA" },
	{ 0x0C18 , "TELUGU LETTER GHA" },
	{ 0x0C19 , "TELUGU LETTER NGA" },
	{ 0x0C1A , "TELUGU LETTER CA" },
	{ 0x0C1B , "TELUGU LETTER CHA" },
	{ 0x0C1C , "TELUGU LETTER JA" },
	{ 0x0C1D , "TELUGU LETTER JHA" },
	{ 0x0C1E , "TELUGU LETTER NYA" },
	{ 0x0C1F , "TELUGU LETTER TTA" },
	{ 0x0C20 , "TELUGU LETTER TTHA" },
	{ 0x0C21 , "TELUGU LETTER DDA" },
	{ 0x0C22 , "TELUGU LETTER DDHA" },
	{ 0x0C23 , "TELUGU LETTER NNA" },
	{ 0x0C24 , "TELUGU LETTER TA" },
	{ 0x0C25 , "TELUGU LETTER THA" },
	{ 0x0C26 , "TELUGU LETTER DA" },
	{ 0x0C27 , "TELUGU LETTER DHA" },
	{ 0x0C28 , "TELUGU LETTER NA" },
	{ 0x0C2A , "TELUGU LETTER PA" },
	{ 0x0C2B , "TELUGU LETTER PHA" },
	{ 0x0C2C , "TELUGU LETTER BA" },
	{ 0x0C2D , "TELUGU LETTER BHA" },
	{ 0x0C2E , "TELUGU LETTER MA" },
	{ 0x0C2F , "TELUGU LETTER YA" },
	{ 0x0C30 , "TELUGU LETTER RA" },
	{ 0x0C31 , "TELUGU LETTER RRA" },
	{ 0x0C32 , "TELUGU LETTER LA" },
	{ 0x0C33 , "TELUGU LETTER LLA" },
	{ 0x0C35 , "TELUGU LETTER VA" },
	{ 0x0C36 , "TELUGU LETTER SHA" },
	{ 0x0C37 , "TELUGU LETTER SSA" },
	{ 0x0C38 , "TELUGU LETTER SA" },
	{ 0x0C39 , "TELUGU LETTER HA" },
	{ 0x0C3E , "TELUGU VOWEL SIGN AA" },
	{ 0x0C3F , "TELUGU VOWEL SIGN I" },
	{ 0x0C40 , "TELUGU VOWEL SIGN II" },
	{ 0x0C41 , "TELUGU VOWEL SIGN U" },
	{ 0x0C42 , "TELUGU VOWEL SIGN UU" },
	{ 0x0C43 , "TELUGU VOWEL SIGN VOCALIC R" },
	{ 0x0C44 , "TELUGU VOWEL SIGN VOCALIC RR" },
	{ 0x0C46 , "TELUGU VOWEL SIGN E" },
	{ 0x0C47 , "TELUGU VOWEL SIGN EE" },
	{ 0x0C48 , "TELUGU VOWEL SIGN AI" },
	{ 0x0C4A , "TELUGU VOWEL SIGN O" },
	{ 0x0C4B , "TELUGU VOWEL SIGN OO" },
	{ 0x0C4C , "TELUGU VOWEL SIGN AU" },
	{ 0x0C4D , "TELUGU SIGN VIRAMA" },
	{ 0x0C55 , "TELUGU LENGTH MARK" },
	{ 0x0C56 , "TELUGU AI LENGTH MARK" },
	{ 0x0C60 , "TELUGU LETTER VOCALIC RR" },
	{ 0x0C61 , "TELUGU LETTER VOCALIC LL" },
	{ 0x0C66 , "TELUGU DIGIT ZERO" },
	{ 0x0C67 , "TELUGU DIGIT ONE" },
	{ 0x0C68 , "TELUGU DIGIT TWO" },
	{ 0x0C69 , "TELUGU DIGIT THREE" },
	{ 0x0C6A , "TELUGU DIGIT FOUR" },
	{ 0x0C6B , "TELUGU DIGIT FIVE" },
	{ 0x0C6C , "TELUGU DIGIT SIX" },
	{ 0x0C6D , "TELUGU DIGIT SEVEN" },
	{ 0x0C6E , "TELUGU DIGIT EIGHT" },
	{ 0x0C6F , "TELUGU DIGIT NINE" },
	{ 0x0C82 , "KANNADA SIGN ANUSVARA" },
	{ 0x0C83 , "KANNADA SIGN VISARGA" },
	{ 0x0C85 , "KANNADA LETTER A" },
	{ 0x0C86 , "KANNADA LETTER AA" },
	{ 0x0C87 , "KANNADA LETTER I" },
	{ 0x0C88 , "KANNADA LETTER II" },
	{ 0x0C89 , "KANNADA LETTER U" },
	{ 0x0C8A , "KANNADA LETTER UU" },
	{ 0x0C8B , "KANNADA LETTER VOCALIC R" },
	{ 0x0C8C , "KANNADA LETTER VOCALIC L" },
	{ 0x0C8E , "KANNADA LETTER E" },
	{ 0x0C8F , "KANNADA LETTER EE" },
	{ 0x0C90 , "KANNADA LETTER AI" },
	{ 0x0C92 , "KANNADA LETTER O" },
	{ 0x0C93 , "KANNADA LETTER OO" },
	{ 0x0C94 , "KANNADA LETTER AU" },
	{ 0x0C95 , "KANNADA LETTER KA" },
	{ 0x0C96 , "KANNADA LETTER KHA" },
	{ 0x0C97 , "KANNADA LETTER GA" },
	{ 0x0C98 , "KANNADA LETTER GHA" },
	{ 0x0C99 , "KANNADA LETTER NGA" },
	{ 0x0C9A , "KANNADA LETTER CA" },
	{ 0x0C9B , "KANNADA LETTER CHA" },
	{ 0x0C9C , "KANNADA LETTER JA" },
	{ 0x0C9D , "KANNADA LETTER JHA" },
	{ 0x0C9E , "KANNADA LETTER NYA" },
	{ 0x0C9F , "KANNADA LETTER TTA" },
	{ 0x0CA0 , "KANNADA LETTER TTHA" },
	{ 0x0CA1 , "KANNADA LETTER DDA" },
	{ 0x0CA2 , "KANNADA LETTER DDHA" },
	{ 0x0CA3 , "KANNADA LETTER NNA" },
	{ 0x0CA4 , "KANNADA LETTER TA" },
	{ 0x0CA5 , "KANNADA LETTER THA" },
	{ 0x0CA6 , "KANNADA LETTER DA" },
	{ 0x0CA7 , "KANNADA LETTER DHA" },
	{ 0x0CA8 , "KANNADA LETTER NA" },
	{ 0x0CAA , "KANNADA LETTER PA" },
	{ 0x0CAB , "KANNADA LETTER PHA" },
	{ 0x0CAC , "KANNADA LETTER BA" },
	{ 0x0CAD , "KANNADA LETTER BHA" },
	{ 0x0CAE , "KANNADA LETTER MA" },
	{ 0x0CAF , "KANNADA LETTER YA" },
	{ 0x0CB0 , "KANNADA LETTER RA" },
	{ 0x0CB1 , "KANNADA LETTER RRA" },
	{ 0x0CB2 , "KANNADA LETTER LA" },
	{ 0x0CB3 , "KANNADA LETTER LLA" },
	{ 0x0CB5 , "KANNADA LETTER VA" },
	{ 0x0CB6 , "KANNADA LETTER SHA" },
	{ 0x0CB7 , "KANNADA LETTER SSA" },
	{ 0x0CB8 , "KANNADA LETTER SA" },
	{ 0x0CB9 , "KANNADA LETTER HA" },
	{ 0x0CBE , "KANNADA VOWEL SIGN AA" },
	{ 0x0CBF , "KANNADA VOWEL SIGN I" },
	{ 0x0CC0 , "KANNADA VOWEL SIGN II" },
	{ 0x0CC1 , "KANNADA VOWEL SIGN U" },
	{ 0x0CC2 , "KANNADA VOWEL SIGN UU" },
	{ 0x0CC3 , "KANNADA VOWEL SIGN VOCALIC R" },
	{ 0x0CC4 , "KANNADA VOWEL SIGN VOCALIC RR" },
	{ 0x0CC6 , "KANNADA VOWEL SIGN E" },
	{ 0x0CC7 , "KANNADA VOWEL SIGN EE" },
	{ 0x0CC8 , "KANNADA VOWEL SIGN AI" },
	{ 0x0CCA , "KANNADA VOWEL SIGN O" },
	{ 0x0CCB , "KANNADA VOWEL SIGN OO" },
	{ 0x0CCC , "KANNADA VOWEL SIGN AU" },
	{ 0x0CCD , "KANNADA SIGN VIRAMA" },
	{ 0x0CD5 , "KANNADA LENGTH MARK" },
	{ 0x0CD6 , "KANNADA AI LENGTH MARK" },
	{ 0x0CDE , "KANNADA LETTER FA" },
	{ 0x0CE0 , "KANNADA LETTER VOCALIC RR" },
	{ 0x0CE1 , "KANNADA LETTER VOCALIC LL" },
	{ 0x0CE6 , "KANNADA DIGIT ZERO" },
	{ 0x0CE7 , "KANNADA DIGIT ONE" },
	{ 0x0CE8 , "KANNADA DIGIT TWO" },
	{ 0x0CE9 , "KANNADA DIGIT THREE" },
	{ 0x0CEA , "KANNADA DIGIT FOUR" },
	{ 0x0CEB , "KANNADA DIGIT FIVE" },
	{ 0x0CEC , "KANNADA DIGIT SIX" },
	{ 0x0CED , "KANNADA DIGIT SEVEN" },
	{ 0x0CEE , "KANNADA DIGIT EIGHT" },
	{ 0x0CEF , "KANNADA DIGIT NINE" },
	{ 0x0D02 , "MALAYALAM SIGN ANUSVARA" },
	{ 0x0D03 , "MALAYALAM SIGN VISARGA" },
	{ 0x0D05 , "MALAYALAM LETTER A" },
	{ 0x0D06 , "MALAYALAM LETTER AA" },
	{ 0x0D07 , "MALAYALAM LETTER I" },
	{ 0x0D08 , "MALAYALAM LETTER II" },
	{ 0x0D09 , "MALAYALAM LETTER U" },
	{ 0x0D0A , "MALAYALAM LETTER UU" },
	{ 0x0D0B , "MALAYALAM LETTER VOCALIC R" },
	{ 0x0D0C , "MALAYALAM LETTER VOCALIC L" },
	{ 0x0D0E , "MALAYALAM LETTER E" },
	{ 0x0D0F , "MALAYALAM LETTER EE" },
	{ 0x0D10 , "MALAYALAM LETTER AI" },
	{ 0x0D12 , "MALAYALAM LETTER O" },
	{ 0x0D13 , "MALAYALAM LETTER OO" },
	{ 0x0D14 , "MALAYALAM LETTER AU" },
	{ 0x0D15 , "MALAYALAM LETTER KA" },
	{ 0x0D16 , "MALAYALAM LETTER KHA" },
	{ 0x0D17 , "MALAYALAM LETTER GA" },
	{ 0x0D18 , "MALAYALAM LETTER GHA" },
	{ 0x0D19 , "MALAYALAM LETTER NGA" },
	{ 0x0D1A , "MALAYALAM LETTER CA" },
	{ 0x0D1B , "MALAYALAM LETTER CHA" },
	{ 0x0D1C , "MALAYALAM LETTER JA" },
	{ 0x0D1D , "MALAYALAM LETTER JHA" },
	{ 0x0D1E , "MALAYALAM LETTER NYA" },
	{ 0x0D1F , "MALAYALAM LETTER TTA" },
	{ 0x0D20 , "MALAYALAM LETTER TTHA" },
	{ 0x0D21 , "MALAYALAM LETTER DDA" },
	{ 0x0D22 , "MALAYALAM LETTER DDHA" },
	{ 0x0D23 , "MALAYALAM LETTER NNA" },
	{ 0x0D24 , "MALAYALAM LETTER TA" },
	{ 0x0D25 , "MALAYALAM LETTER THA" },
	{ 0x0D26 , "MALAYALAM LETTER DA" },
	{ 0x0D27 , "MALAYALAM LETTER DHA" },
	{ 0x0D28 , "MALAYALAM LETTER NA" },
	{ 0x0D2A , "MALAYALAM LETTER PA" },
	{ 0x0D2B , "MALAYALAM LETTER PHA" },
	{ 0x0D2C , "MALAYALAM LETTER BA" },
	{ 0x0D2D , "MALAYALAM LETTER BHA" },
	{ 0x0D2E , "MALAYALAM LETTER MA" },
	{ 0x0D2F , "MALAYALAM LETTER YA" },
	{ 0x0D30 , "MALAYALAM LETTER RA" },
	{ 0x0D31 , "MALAYALAM LETTER RRA" },
	{ 0x0D32 , "MALAYALAM LETTER LA" },
	{ 0x0D33 , "MALAYALAM LETTER LLA" },
	{ 0x0D34 , "MALAYALAM LETTER LLLA" },
	{ 0x0D35 , "MALAYALAM LETTER VA" },
	{ 0x0D36 , "MALAYALAM LETTER SHA" },
	{ 0x0D37 , "MALAYALAM LETTER SSA" },
	{ 0x0D38 , "MALAYALAM LETTER SA" },
	{ 0x0D39 , "MALAYALAM LETTER HA" },
	{ 0x0D3E , "MALAYALAM VOWEL SIGN AA" },
	{ 0x0D3F , "MALAYALAM VOWEL SIGN I" },
	{ 0x0D40 , "MALAYALAM VOWEL SIGN II" },
	{ 0x0D41 , "MALAYALAM VOWEL SIGN U" },
	{ 0x0D42 , "MALAYALAM VOWEL SIGN UU" },
	{ 0x0D43 , "MALAYALAM VOWEL SIGN VOCALIC R" },
	{ 0x0D46 , "MALAYALAM VOWEL SIGN E" },
	{ 0x0D47 , "MALAYALAM VOWEL SIGN EE" },
	{ 0x0D48 , "MALAYALAM VOWEL SIGN AI" },
	{ 0x0D4A , "MALAYALAM VOWEL SIGN O" },
	{ 0x0D4B , "MALAYALAM VOWEL SIGN OO" },
	{ 0x0D4C , "MALAYALAM VOWEL SIGN AU" },
	{ 0x0D4D , "MALAYALAM SIGN VIRAMA" },
	{ 0x0D57 , "MALAYALAM AU LENGTH MARK" },
	{ 0x0D60 , "MALAYALAM LETTER VOCALIC RR" },
	{ 0x0D61 , "MALAYALAM LETTER VOCALIC LL" },
	{ 0x0D66 , "MALAYALAM DIGIT ZERO" },
	{ 0x0D67 , "MALAYALAM DIGIT ONE" },
	{ 0x0D68 , "MALAYALAM DIGIT TWO" },
	{ 0x0D69 , "MALAYALAM DIGIT THREE" },
	{ 0x0D6A , "MALAYALAM DIGIT FOUR" },
	{ 0x0D6B , "MALAYALAM DIGIT FIVE" },
	{ 0x0D6C , "MALAYALAM DIGIT SIX" },
	{ 0x0D6D , "MALAYALAM DIGIT SEVEN" },
	{ 0x0D6E , "MALAYALAM DIGIT EIGHT" },
	{ 0x0D6F , "MALAYALAM DIGIT NINE" },
	{ 0x0E01 , "THAI CHARACTER KO KAI" },
	{ 0x0E02 , "THAI CHARACTER KHO KHAI" },
	{ 0x0E03 , "THAI CHARACTER KHO KHUAT" },
	{ 0x0E04 , "THAI CHARACTER KHO KHWAI" },
	{ 0x0E05 , "THAI CHARACTER KHO KHON" },
	{ 0x0E06 , "THAI CHARACTER KHO RAKHANG" },
	{ 0x0E07 , "THAI CHARACTER NGO NGU" },
	{ 0x0E08 , "THAI CHARACTER CHO CHAN" },
	{ 0x0E09 , "THAI CHARACTER CHO CHING" },
	{ 0x0E0A , "THAI CHARACTER CHO CHANG" },
	{ 0x0E0B , "THAI CHARACTER SO SO" },
	{ 0x0E0C , "THAI CHARACTER CHO CHOE" },
	{ 0x0E0D , "THAI CHARACTER YO YING" },
	{ 0x0E0E , "THAI CHARACTER DO CHADA" },
	{ 0x0E0F , "THAI CHARACTER TO PATAK" },
	{ 0x0E10 , "THAI CHARACTER THO THAN" },
	{ 0x0E11 , "THAI CHARACTER THO NANGMONTHO" },
	{ 0x0E12 , "THAI CHARACTER THO PHUTHAO" },
	{ 0x0E13 , "THAI CHARACTER NO NEN" },
	{ 0x0E14 , "THAI CHARACTER DO DEK" },
	{ 0x0E15 , "THAI CHARACTER TO TAO" },
	{ 0x0E16 , "THAI CHARACTER THO THUNG" },
	{ 0x0E17 , "THAI CHARACTER THO THAHAN" },
	{ 0x0E18 , "THAI CHARACTER THO THONG" },
	{ 0x0E19 , "THAI CHARACTER NO NU" },
	{ 0x0E1A , "THAI CHARACTER BO BAIMAI" },
	{ 0x0E1B , "THAI CHARACTER PO PLA" },
	{ 0x0E1C , "THAI CHARACTER PHO PHUNG" },
	{ 0x0E1D , "THAI CHARACTER FO FA" },
	{ 0x0E1E , "THAI CHARACTER PHO PHAN" },
	{ 0x0E1F , "THAI CHARACTER FO FAN" },
	{ 0x0E20 , "THAI CHARACTER PHO SAMPHAO" },
	{ 0x0E21 , "THAI CHARACTER MO MA" },
	{ 0x0E22 , "THAI CHARACTER YO YAK" },
	{ 0x0E23 , "THAI CHARACTER RO RUA" },
	{ 0x0E24 , "THAI CHARACTER RU" },
	{ 0x0E25 , "THAI CHARACTER LO LING" },
	{ 0x0E26 , "THAI CHARACTER LU" },
	{ 0x0E27 , "THAI CHARACTER WO WAEN" },
	{ 0x0E28 , "THAI CHARACTER SO SALA" },
	{ 0x0E29 , "THAI CHARACTER SO RUSI" },
	{ 0x0E2A , "THAI CHARACTER SO SUA" },
	{ 0x0E2B , "THAI CHARACTER HO HIP" },
	{ 0x0E2C , "THAI CHARACTER LO CHULA" },
	{ 0x0E2D , "THAI CHARACTER O ANG" },
	{ 0x0E2E , "THAI CHARACTER HO NOKHUK" },
	{ 0x0E2F , "THAI CHARACTER PAIYANNOI" },
	{ 0x0E30 , "THAI CHARACTER SARA A" },
	{ 0x0E31 , "THAI CHARACTER MAI HAN-AKAT" },
	{ 0x0E32 , "THAI CHARACTER SARA AA" },
	{ 0x0E33 , "THAI CHARACTER SARA AM" },
	{ 0x0E34 , "THAI CHARACTER SARA I" },
	{ 0x0E35 , "THAI CHARACTER SARA II" },
	{ 0x0E36 , "THAI CHARACTER SARA UE" },
	{ 0x0E37 , "THAI CHARACTER SARA UEE" },
	{ 0x0E38 , "THAI CHARACTER SARA U" },
	{ 0x0E39 , "THAI CHARACTER SARA UU" },
	{ 0x0E3A , "THAI CHARACTER PHINTHU" },
	{ 0x0E3F , "THAI CURRENCY SYMBOL BAHT" },
	{ 0x0E40 , "THAI CHARACTER SARA E" },
	{ 0x0E41 , "THAI CHARACTER SARA AE" },
	{ 0x0E42 , "THAI CHARACTER SARA O" },
	{ 0x0E43 , "THAI CHARACTER SARA AI MAIMUAN" },
	{ 0x0E44 , "THAI CHARACTER SARA AI MAIMALAI" },
	{ 0x0E45 , "THAI CHARACTER LAKKHANGYAO" },
	{ 0x0E46 , "THAI CHARACTER MAIYAMOK" },
	{ 0x0E47 , "THAI CHARACTER MAITAIKHU" },
	{ 0x0E48 , "THAI CHARACTER MAI EK" },
	{ 0x0E49 , "THAI CHARACTER MAI THO" },
	{ 0x0E4A , "THAI CHARACTER MAI TRI" },
	{ 0x0E4B , "THAI CHARACTER MAI CHATTAWA" },
	{ 0x0E4C , "THAI CHARACTER THANTHAKHAT" },
	{ 0x0E4D , "THAI CHARACTER NIKHAHIT" },
	{ 0x0E4E , "THAI CHARACTER YAMAKKAN" },
	{ 0x0E4F , "THAI CHARACTER FONGMAN" },
	{ 0x0E50 , "THAI DIGIT ZERO" },
	{ 0x0E51 , "THAI DIGIT ONE" },
	{ 0x0E52 , "THAI DIGIT TWO" },
	{ 0x0E53 , "THAI DIGIT THREE" },
	{ 0x0E54 , "THAI DIGIT FOUR" },
	{ 0x0E55 , "THAI DIGIT FIVE" },
	{ 0x0E56 , "THAI DIGIT SIX" },
	{ 0x0E57 , "THAI DIGIT SEVEN" },
	{ 0x0E58 , "THAI DIGIT EIGHT" },
	{ 0x0E59 , "THAI DIGIT NINE" },
	{ 0x0E5A , "THAI CHARACTER ANGKHANKHU" },
	{ 0x0E5B , "THAI CHARACTER KHOMUT" },
	{ 0x0E81 , "LAO LETTER KO" },
	{ 0x0E82 , "LAO LETTER KHO SUNG" },
	{ 0x0E84 , "LAO LETTER KHO TAM" },
	{ 0x0E87 , "LAO LETTER NGO" },
	{ 0x0E88 , "LAO LETTER CO" },
	{ 0x0E8A , "LAO LETTER SO TAM" },
	{ 0x0E8D , "LAO LETTER NYO" },
	{ 0x0E94 , "LAO LETTER DO" },
	{ 0x0E95 , "LAO LETTER TO" },
	{ 0x0E96 , "LAO LETTER THO SUNG" },
	{ 0x0E97 , "LAO LETTER THO TAM" },
	{ 0x0E99 , "LAO LETTER NO" },
	{ 0x0E9A , "LAO LETTER BO" },
	{ 0x0E9B , "LAO LETTER PO" },
	{ 0x0E9C , "LAO LETTER PHO SUNG" },
	{ 0x0E9D , "LAO LETTER FO TAM" },
	{ 0x0E9E , "LAO LETTER PHO TAM" },
	{ 0x0E9F , "LAO LETTER FO SUNG" },
	{ 0x0EA1 , "LAO LETTER MO" },
	{ 0x0EA2 , "LAO LETTER YO" },
	{ 0x0EA3 , "LAO LETTER LO LING" },
	{ 0x0EA5 , "LAO LETTER LO LOOT" },
	{ 0x0EA7 , "LAO LETTER WO" },
	{ 0x0EAA , "LAO LETTER SO SUNG" },
	{ 0x0EAB , "LAO LETTER HO SUNG" },
	{ 0x0EAD , "LAO LETTER O" },
	{ 0x0EAE , "LAO LETTER HO TAM" },
	{ 0x0EAF , "LAO ELLIPSIS" },
	{ 0x0EB0 , "LAO VOWEL SIGN A" },
	{ 0x0EB1 , "LAO VOWEL SIGN MAI KAN" },
	{ 0x0EB2 , "LAO VOWEL SIGN AA" },
	{ 0x0EB3 , "LAO VOWEL SIGN AM" },
	{ 0x0EB4 , "LAO VOWEL SIGN I" },
	{ 0x0EB5 , "LAO VOWEL SIGN II" },
	{ 0x0EB6 , "LAO VOWEL SIGN Y" },
	{ 0x0EB7 , "LAO VOWEL SIGN YY" },
	{ 0x0EB8 , "LAO VOWEL SIGN U" },
	{ 0x0EB9 , "LAO VOWEL SIGN UU" },
	{ 0x0EBB , "LAO VOWEL SIGN MAI KON" },
	{ 0x0EBC , "LAO SEMIVOWEL SIGN LO" },
	{ 0x0EBD , "LAO SEMIVOWEL SIGN NYO" },
	{ 0x0EC0 , "LAO VOWEL SIGN E" },
	{ 0x0EC1 , "LAO VOWEL SIGN EI" },
	{ 0x0EC2 , "LAO VOWEL SIGN O" },
	{ 0x0EC3 , "LAO VOWEL SIGN AY" },
	{ 0x0EC4 , "LAO VOWEL SIGN AI" },
	{ 0x0EC6 , "LAO KO LA" },
	{ 0x0EC8 , "LAO TONE MAI EK" },
	{ 0x0EC9 , "LAO TONE MAI THO" },
	{ 0x0ECA , "LAO TONE MAI TI" },
	{ 0x0ECB , "LAO TONE MAI CATAWA" },
	{ 0x0ECC , "LAO CANCELLATION MARK" },
	{ 0x0ECD , "LAO NIGGAHITA" },
	{ 0x0ED0 , "LAO DIGIT ZERO" },
	{ 0x0ED1 , "LAO DIGIT ONE" },
	{ 0x0ED2 , "LAO DIGIT TWO" },
	{ 0x0ED3 , "LAO DIGIT THREE" },
	{ 0x0ED4 , "LAO DIGIT FOUR" },
	{ 0x0ED5 , "LAO DIGIT FIVE" },
	{ 0x0ED6 , "LAO DIGIT SIX" },
	{ 0x0ED7 , "LAO DIGIT SEVEN" },
	{ 0x0ED8 , "LAO DIGIT EIGHT" },
	{ 0x0ED9 , "LAO DIGIT NINE" },
	{ 0x0EDC , "LAO HO NO" },
	{ 0x0EDD , "LAO HO MO" },
	{ 0x10A0 , "GEORGIAN CAPITAL LETTER AN" },
	{ 0x10A1 , "GEORGIAN CAPITAL LETTER BAN" },
	{ 0x10A2 , "GEORGIAN CAPITAL LETTER GAN" },
	{ 0x10A3 , "GEORGIAN CAPITAL LETTER DON" },
	{ 0x10A4 , "GEORGIAN CAPITAL LETTER EN" },
	{ 0x10A5 , "GEORGIAN CAPITAL LETTER VIN" },
	{ 0x10A6 , "GEORGIAN CAPITAL LETTER ZEN" },
	{ 0x10A7 , "GEORGIAN CAPITAL LETTER TAN" },
	{ 0x10A8 , "GEORGIAN CAPITAL LETTER IN" },
	{ 0x10A9 , "GEORGIAN CAPITAL LETTER KAN" },
	{ 0x10AA , "GEORGIAN CAPITAL LETTER LAS" },
	{ 0x10AB , "GEORGIAN CAPITAL LETTER MAN" },
	{ 0x10AC , "GEORGIAN CAPITAL LETTER NAR" },
	{ 0x10AD , "GEORGIAN CAPITAL LETTER ON" },
	{ 0x10AE , "GEORGIAN CAPITAL LETTER PAR" },
	{ 0x10AF , "GEORGIAN CAPITAL LETTER ZHAR" },
	{ 0x10B0 , "GEORGIAN CAPITAL LETTER RAE" },
	{ 0x10B1 , "GEORGIAN CAPITAL LETTER SAN" },
	{ 0x10B2 , "GEORGIAN CAPITAL LETTER TAR" },
	{ 0x10B3 , "GEORGIAN CAPITAL LETTER UN" },
	{ 0x10B4 , "GEORGIAN CAPITAL LETTER PHAR" },
	{ 0x10B5 , "GEORGIAN CAPITAL LETTER KHAR" },
	{ 0x10B6 , "GEORGIAN CAPITAL LETTER GHAN" },
	{ 0x10B7 , "GEORGIAN CAPITAL LETTER QAR" },
	{ 0x10B8 , "GEORGIAN CAPITAL LETTER SHIN" },
	{ 0x10B9 , "GEORGIAN CAPITAL LETTER CHIN" },
	{ 0x10BA , "GEORGIAN CAPITAL LETTER CAN" },
	{ 0x10BB , "GEORGIAN CAPITAL LETTER JIL" },
	{ 0x10BC , "GEORGIAN CAPITAL LETTER CIL" },
	{ 0x10BD , "GEORGIAN CAPITAL LETTER CHAR" },
	{ 0x10BE , "GEORGIAN CAPITAL LETTER XAN" },
	{ 0x10BF , "GEORGIAN CAPITAL LETTER JHAN" },
	{ 0x10C0 , "GEORGIAN CAPITAL LETTER HAE" },
	{ 0x10C1 , "GEORGIAN CAPITAL LETTER HE" },
	{ 0x10C2 , "GEORGIAN CAPITAL LETTER HIE" },
	{ 0x10C3 , "GEORGIAN CAPITAL LETTER WE" },
	{ 0x10C4 , "GEORGIAN CAPITAL LETTER HAR" },
	{ 0x10C5 , "GEORGIAN CAPITAL LETTER HOE" },
	{ 0x10D0 , "GEORGIAN LETTER AN" },
	{ 0x10D1 , "GEORGIAN LETTER BAN" },
	{ 0x10D2 , "GEORGIAN LETTER GAN" },
	{ 0x10D3 , "GEORGIAN LETTER DON" },
	{ 0x10D4 , "GEORGIAN LETTER EN" },
	{ 0x10D5 , "GEORGIAN LETTER VIN" },
	{ 0x10D6 , "GEORGIAN LETTER ZEN" },
	{ 0x10D7 , "GEORGIAN LETTER TAN" },
	{ 0x10D8 , "GEORGIAN LETTER IN" },
	{ 0x10D9 , "GEORGIAN LETTER KAN" },
	{ 0x10DA , "GEORGIAN LETTER LAS" },
	{ 0x10DB , "GEORGIAN LETTER MAN" },
	{ 0x10DC , "GEORGIAN LETTER NAR" },
	{ 0x10DD , "GEORGIAN LETTER ON" },
	{ 0x10DE , "GEORGIAN LETTER PAR" },
	{ 0x10DF , "GEORGIAN LETTER ZHAR" },
	{ 0x10E0 , "GEORGIAN LETTER RAE" },
	{ 0x10E1 , "GEORGIAN LETTER SAN" },
	{ 0x10E2 , "GEORGIAN LETTER TAR" },
	{ 0x10E3 , "GEORGIAN LETTER UN" },
	{ 0x10E4 , "GEORGIAN LETTER PHAR" },
	{ 0x10E5 , "GEORGIAN LETTER KHAR" },
	{ 0x10E6 , "GEORGIAN LETTER GHAN" },
	{ 0x10E7 , "GEORGIAN LETTER QAR" },
	{ 0x10E8 , "GEORGIAN LETTER SHIN" },
	{ 0x10E9 , "GEORGIAN LETTER CHIN" },
	{ 0x10EA , "GEORGIAN LETTER CAN" },
	{ 0x10EB , "GEORGIAN LETTER JIL" },
	{ 0x10EC , "GEORGIAN LETTER CIL" },
	{ 0x10ED , "GEORGIAN LETTER CHAR" },
	{ 0x10EE , "GEORGIAN LETTER XAN" },
	{ 0x10EF , "GEORGIAN LETTER JHAN" },
	{ 0x10F0 , "GEORGIAN LETTER HAE" },
	{ 0x10F1 , "GEORGIAN LETTER HE" },
	{ 0x10F2 , "GEORGIAN LETTER HIE" },
	{ 0x10F3 , "GEORGIAN LETTER WE" },
	{ 0x10F4 , "GEORGIAN LETTER HAR" },
	{ 0x10F5 , "GEORGIAN LETTER HOE" },
	{ 0x10F6 , "GEORGIAN LETTER FI" },
	{ 0x10FB , "GEORGIAN PARAGRAPH SEPARATOR" },
	{ 0x1100 , "HANGUL CHOSEONG KIYEOK" },
	{ 0x1101 , "HANGUL CHOSEONG SSANGKIYEOK" },
	{ 0x1102 , "HANGUL CHOSEONG NIEUN" },
	{ 0x1103 , "HANGUL CHOSEONG TIKEUT" },
	{ 0x1104 , "HANGUL CHOSEONG SSANGTIKEUT" },
	{ 0x1105 , "HANGUL CHOSEONG RIEUL" },
	{ 0x1106 , "HANGUL CHOSEONG MIEUM" },
	{ 0x1107 , "HANGUL CHOSEONG PIEUP" },
	{ 0x1108 , "HANGUL CHOSEONG SSANGPIEUP" },
	{ 0x1109 , "HANGUL CHOSEONG SIOS" },
	{ 0x110A , "HANGUL CHOSEONG SSANGSIOS" },
	{ 0x110B , "HANGUL CHOSEONG IEUNG" },
	{ 0x110C , "HANGUL CHOSEONG CIEUC" },
	{ 0x110D , "HANGUL CHOSEONG SSANGCIEUC" },
	{ 0x110E , "HANGUL CHOSEONG CHIEUCH" },
	{ 0x110F , "HANGUL CHOSEONG KHIEUKH" },
	{ 0x1110 , "HANGUL CHOSEONG THIEUTH" },
	{ 0x1111 , "HANGUL CHOSEONG PHIEUPH" },
	{ 0x1112 , "HANGUL CHOSEONG HIEUH" },
	{ 0x1113 , "HANGUL CHOSEONG NIEUN-KIYEOK" },
	{ 0x1114 , "HANGUL CHOSEONG SSANGNIEUN" },
	{ 0x1115 , "HANGUL CHOSEONG NIEUN-TIKEUT" },
	{ 0x1116 , "HANGUL CHOSEONG NIEUN-PIEUP" },
	{ 0x1117 , "HANGUL CHOSEONG TIKEUT-KIYEOK" },
	{ 0x1118 , "HANGUL CHOSEONG RIEUL-NIEUN" },
	{ 0x1119 , "HANGUL CHOSEONG SSANGRIEUL" },
	{ 0x111A , "HANGUL CHOSEONG RIEUL-HIEUH" },
	{ 0x111B , "HANGUL CHOSEONG KAPYEOUNRIEUL" },
	{ 0x111C , "HANGUL CHOSEONG MIEUM-PIEUP" },
	{ 0x111D , "HANGUL CHOSEONG KAPYEOUNMIEUM" },
	{ 0x111E , "HANGUL CHOSEONG PIEUP-KIYEOK" },
	{ 0x111F , "HANGUL CHOSEONG PIEUP-NIEUN" },
	{ 0x1120 , "HANGUL CHOSEONG PIEUP-TIKEUT" },
	{ 0x1121 , "HANGUL CHOSEONG PIEUP-SIOS" },
	{ 0x1122 , "HANGUL CHOSEONG PIEUP-SIOS-KIYEOK" },
	{ 0x1123 , "HANGUL CHOSEONG PIEUP-SIOS-TIKEUT" },
	{ 0x1124 , "HANGUL CHOSEONG PIEUP-SIOS-PIEUP" },
	{ 0x1125 , "HANGUL CHOSEONG PIEUP-SSANGSIOS" },
	{ 0x1126 , "HANGUL CHOSEONG PIEUP-SIOS-CIEUC" },
	{ 0x1127 , "HANGUL CHOSEONG PIEUP-CIEUC" },
	{ 0x1128 , "HANGUL CHOSEONG PIEUP-CHIEUCH" },
	{ 0x1129 , "HANGUL CHOSEONG PIEUP-THIEUTH" },
	{ 0x112A , "HANGUL CHOSEONG PIEUP-PHIEUPH" },
	{ 0x112B , "HANGUL CHOSEONG KAPYEOUNPIEUP" },
	{ 0x112C , "HANGUL CHOSEONG KAPYEOUNSSANGPIEUP" },
	{ 0x112D , "HANGUL CHOSEONG SIOS-KIYEOK" },
	{ 0x112E , "HANGUL CHOSEONG SIOS-NIEUN" },
	{ 0x112F , "HANGUL CHOSEONG SIOS-TIKEUT" },
	{ 0x1130 , "HANGUL CHOSEONG SIOS-RIEUL" },
	{ 0x1131 , "HANGUL CHOSEONG SIOS-MIEUM" },
	{ 0x1132 , "HANGUL CHOSEONG SIOS-PIEUP" },
	{ 0x1133 , "HANGUL CHOSEONG SIOS-PIEUP-KIYEOK" },
	{ 0x1134 , "HANGUL CHOSEONG SIOS-SSANGSIOS" },
	{ 0x1135 , "HANGUL CHOSEONG SIOS-IEUNG" },
	{ 0x1136 , "HANGUL CHOSEONG SIOS-CIEUC" },
	{ 0x1137 , "HANGUL CHOSEONG SIOS-CHIEUCH" },
	{ 0x1138 , "HANGUL CHOSEONG SIOS-KHIEUKH" },
	{ 0x1139 , "HANGUL CHOSEONG SIOS-THIEUTH" },
	{ 0x113A , "HANGUL CHOSEONG SIOS-PHIEUPH" },
	{ 0x113B , "HANGUL CHOSEONG SIOS-HIEUH" },
	{ 0x113C , "HANGUL CHOSEONG CHITUEUMSIOS" },
	{ 0x113D , "HANGUL CHOSEONG CHITUEUMSSANGSIOS" },
	{ 0x113E , "HANGUL CHOSEONG CEONGCHIEUMSIOS" },
	{ 0x113F , "HANGUL CHOSEONG CEONGCHIEUMSSANGSIOS" },
	{ 0x1140 , "HANGUL CHOSEONG PANSIOS" },
	{ 0x1141 , "HANGUL CHOSEONG IEUNG-KIYEOK" },
	{ 0x1142 , "HANGUL CHOSEONG IEUNG-TIKEUT" },
	{ 0x1143 , "HANGUL CHOSEONG IEUNG-MIEUM" },
	{ 0x1144 , "HANGUL CHOSEONG IEUNG-PIEUP" },
	{ 0x1145 , "HANGUL CHOSEONG IEUNG-SIOS" },
	{ 0x1146 , "HANGUL CHOSEONG IEUNG-PANSIOS" },
	{ 0x1147 , "HANGUL CHOSEONG SSANGIEUNG" },
	{ 0x1148 , "HANGUL CHOSEONG IEUNG-CIEUC" },
	{ 0x1149 , "HANGUL CHOSEONG IEUNG-CHIEUCH" },
	{ 0x114A , "HANGUL CHOSEONG IEUNG-THIEUTH" },
	{ 0x114B , "HANGUL CHOSEONG IEUNG-PHIEUPH" },
	{ 0x114C , "HANGUL CHOSEONG YESIEUNG" },
	{ 0x114D , "HANGUL CHOSEONG CIEUC-IEUNG" },
	{ 0x114E , "HANGUL CHOSEONG CHITUEUMCIEUC" },
	{ 0x114F , "HANGUL CHOSEONG CHITUEUMSSANGCIEUC" },
	{ 0x1150 , "HANGUL CHOSEONG CEONGCHIEUMCIEUC" },
	{ 0x1151 , "HANGUL CHOSEONG CEONGCHIEUMSSANGCIEUC" },
	{ 0x1152 , "HANGUL CHOSEONG CHIEUCH-KHIEUKH" },
	{ 0x1153 , "HANGUL CHOSEONG CHIEUCH-HIEUH" },
	{ 0x1154 , "HANGUL CHOSEONG CHITUEUMCHIEUCH" },
	{ 0x1155 , "HANGUL CHOSEONG CEONGCHIEUMCHIEUCH" },
	{ 0x1156 , "HANGUL CHOSEONG PHIEUPH-PIEUP" },
	{ 0x1157 , "HANGUL CHOSEONG KAPYEOUNPHIEUPH" },
	{ 0x1158 , "HANGUL CHOSEONG SSANGHIEUH" },
	{ 0x1159 , "HANGUL CHOSEONG YEORINHIEUH" },
	{ 0x115F , "HANGUL CHOSEONG FILLER" },
	{ 0x1160 , "HANGUL JUNGSEONG FILLER" },
	{ 0x1161 , "HANGUL JUNGSEONG A" },
	{ 0x1162 , "HANGUL JUNGSEONG AE" },
	{ 0x1163 , "HANGUL JUNGSEONG YA" },
	{ 0x1164 , "HANGUL JUNGSEONG YAE" },
	{ 0x1165 , "HANGUL JUNGSEONG EO" },
	{ 0x1166 , "HANGUL JUNGSEONG E" },
	{ 0x1167 , "HANGUL JUNGSEONG YEO" },
	{ 0x1168 , "HANGUL JUNGSEONG YE" },
	{ 0x1169 , "HANGUL JUNGSEONG O" },
	{ 0x116A , "HANGUL JUNGSEONG WA" },
	{ 0x116B , "HANGUL JUNGSEONG WAE" },
	{ 0x116C , "HANGUL JUNGSEONG OE" },
	{ 0x116D , "HANGUL JUNGSEONG YO" },
	{ 0x116E , "HANGUL JUNGSEONG U" },
	{ 0x116F , "HANGUL JUNGSEONG WEO" },
	{ 0x1170 , "HANGUL JUNGSEONG WE" },
	{ 0x1171 , "HANGUL JUNGSEONG WI" },
	{ 0x1172 , "HANGUL JUNGSEONG YU" },
	{ 0x1173 , "HANGUL JUNGSEONG EU" },
	{ 0x1174 , "HANGUL JUNGSEONG YI" },
	{ 0x1175 , "HANGUL JUNGSEONG I" },
	{ 0x1176 , "HANGUL JUNGSEONG A-O" },
	{ 0x1177 , "HANGUL JUNGSEONG A-U" },
	{ 0x1178 , "HANGUL JUNGSEONG YA-O" },
	{ 0x1179 , "HANGUL JUNGSEONG YA-YO" },
	{ 0x117A , "HANGUL JUNGSEONG EO-O" },
	{ 0x117B , "HANGUL JUNGSEONG EO-U" },
	{ 0x117C , "HANGUL JUNGSEONG EO-EU" },
	{ 0x117D , "HANGUL JUNGSEONG YEO-O" },
	{ 0x117E , "HANGUL JUNGSEONG YEO-U" },
	{ 0x117F , "HANGUL JUNGSEONG O-EO" },
	{ 0x1180 , "HANGUL JUNGSEONG O-E" },
	{ 0x1181 , "HANGUL JUNGSEONG O-YE" },
	{ 0x1182 , "HANGUL JUNGSEONG O-O" },
	{ 0x1183 , "HANGUL JUNGSEONG O-U" },
	{ 0x1184 , "HANGUL JUNGSEONG YO-YA" },
	{ 0x1185 , "HANGUL JUNGSEONG YO-YAE" },
	{ 0x1186 , "HANGUL JUNGSEONG YO-YEO" },
	{ 0x1187 , "HANGUL JUNGSEONG YO-O" },
	{ 0x1188 , "HANGUL JUNGSEONG YO-I" },
	{ 0x1189 , "HANGUL JUNGSEONG U-A" },
	{ 0x118A , "HANGUL JUNGSEONG U-AE" },
	{ 0x118B , "HANGUL JUNGSEONG U-EO-EU" },
	{ 0x118C , "HANGUL JUNGSEONG U-YE" },
	{ 0x118D , "HANGUL JUNGSEONG U-U" },
	{ 0x118E , "HANGUL JUNGSEONG YU-A" },
	{ 0x118F , "HANGUL JUNGSEONG YU-EO" },
	{ 0x1190 , "HANGUL JUNGSEONG YU-E" },
	{ 0x1191 , "HANGUL JUNGSEONG YU-YEO" },
	{ 0x1192 , "HANGUL JUNGSEONG YU-YE" },
	{ 0x1193 , "HANGUL JUNGSEONG YU-U" },
	{ 0x1194 , "HANGUL JUNGSEONG YU-I" },
	{ 0x1195 , "HANGUL JUNGSEONG EU-U" },
	{ 0x1196 , "HANGUL JUNGSEONG EU-EU" },
	{ 0x1197 , "HANGUL JUNGSEONG YI-U" },
	{ 0x1198 , "HANGUL JUNGSEONG I-A" },
	{ 0x1199 , "HANGUL JUNGSEONG I-YA" },
	{ 0x119A , "HANGUL JUNGSEONG I-O" },
	{ 0x119B , "HANGUL JUNGSEONG I-U" },
	{ 0x119C , "HANGUL JUNGSEONG I-EU" },
	{ 0x119D , "HANGUL JUNGSEONG I-ARAEA" },
	{ 0x119E , "HANGUL JUNGSEONG ARAEA" },
	{ 0x119F , "HANGUL JUNGSEONG ARAEA-EO" },
	{ 0x11A0 , "HANGUL JUNGSEONG ARAEA-U" },
	{ 0x11A1 , "HANGUL JUNGSEONG ARAEA-I" },
	{ 0x11A2 , "HANGUL JUNGSEONG SSANGARAEA" },
	{ 0x11A8 , "HANGUL JONGSEONG KIYEOK" },
	{ 0x11A9 , "HANGUL JONGSEONG SSANGKIYEOK" },
	{ 0x11AA , "HANGUL JONGSEONG KIYEOK-SIOS" },
	{ 0x11AB , "HANGUL JONGSEONG NIEUN" },
	{ 0x11AC , "HANGUL JONGSEONG NIEUN-CIEUC" },
	{ 0x11AD , "HANGUL JONGSEONG NIEUN-HIEUH" },
	{ 0x11AE , "HANGUL JONGSEONG TIKEUT" },
	{ 0x11AF , "HANGUL JONGSEONG RIEUL" },
	{ 0x11B0 , "HANGUL JONGSEONG RIEUL-KIYEOK" },
	{ 0x11B1 , "HANGUL JONGSEONG RIEUL-MIEUM" },
	{ 0x11B2 , "HANGUL JONGSEONG RIEUL-PIEUP" },
	{ 0x11B3 , "HANGUL JONGSEONG RIEUL-SIOS" },
	{ 0x11B4 , "HANGUL JONGSEONG RIEUL-THIEUTH" },
	{ 0x11B5 , "HANGUL JONGSEONG RIEUL-PHIEUPH" },
	{ 0x11B6 , "HANGUL JONGSEONG RIEUL-HIEUH" },
	{ 0x11B7 , "HANGUL JONGSEONG MIEUM" },
	{ 0x11B8 , "HANGUL JONGSEONG PIEUP" },
	{ 0x11B9 , "HANGUL JONGSEONG PIEUP-SIOS" },
	{ 0x11BA , "HANGUL JONGSEONG SIOS" },
	{ 0x11BB , "HANGUL JONGSEONG SSANGSIOS" },
	{ 0x11BC , "HANGUL JONGSEONG IEUNG" },
	{ 0x11BD , "HANGUL JONGSEONG CIEUC" },
	{ 0x11BE , "HANGUL JONGSEONG CHIEUCH" },
	{ 0x11BF , "HANGUL JONGSEONG KHIEUKH" },
	{ 0x11C0 , "HANGUL JONGSEONG THIEUTH" },
	{ 0x11C1 , "HANGUL JONGSEONG PHIEUPH" },
	{ 0x11C2 , "HANGUL JONGSEONG HIEUH" },
	{ 0x11C3 , "HANGUL JONGSEONG KIYEOK-RIEUL" },
	{ 0x11C4 , "HANGUL JONGSEONG KIYEOK-SIOS-KIYEOK" },
	{ 0x11C5 , "HANGUL JONGSEONG NIEUN-KIYEOK" },
	{ 0x11C6 , "HANGUL JONGSEONG NIEUN-TIKEUT" },
	{ 0x11C7 , "HANGUL JONGSEONG NIEUN-SIOS" },
	{ 0x11C8 , "HANGUL JONGSEONG NIEUN-PANSIOS" },
	{ 0x11C9 , "HANGUL JONGSEONG NIEUN-THIEUTH" },
	{ 0x11CA , "HANGUL JONGSEONG TIKEUT-KIYEOK" },
	{ 0x11CB , "HANGUL JONGSEONG TIKEUT-RIEUL" },
	{ 0x11CC , "HANGUL JONGSEONG RIEUL-KIYEOK-SIOS" },
	{ 0x11CD , "HANGUL JONGSEONG RIEUL-NIEUN" },
	{ 0x11CE , "HANGUL JONGSEONG RIEUL-TIKEUT" },
	{ 0x11CF , "HANGUL JONGSEONG RIEUL-TIKEUT-HIEUH" },
	{ 0x11D0 , "HANGUL JONGSEONG SSANGRIEUL" },
	{ 0x11D1 , "HANGUL JONGSEONG RIEUL-MIEUM-KIYEOK" },
	{ 0x11D2 , "HANGUL JONGSEONG RIEUL-MIEUM-SIOS" },
	{ 0x11D3 , "HANGUL JONGSEONG RIEUL-PIEUP-SIOS" },
	{ 0x11D4 , "HANGUL JONGSEONG RIEUL-PIEUP-HIEUH" },
	{ 0x11D5 , "HANGUL JONGSEONG RIEUL-KAPYEOUNPIEUP" },
	{ 0x11D6 , "HANGUL JONGSEONG RIEUL-SSANGSIOS" },
	{ 0x11D7 , "HANGUL JONGSEONG RIEUL-PANSIOS" },
	{ 0x11D8 , "HANGUL JONGSEONG RIEUL-KHIEUKH" },
	{ 0x11D9 , "HANGUL JONGSEONG RIEUL-YEORINHIEUH" },
	{ 0x11DA , "HANGUL JONGSEONG MIEUM-KIYEOK" },
	{ 0x11DB , "HANGUL JONGSEONG MIEUM-RIEUL" },
	{ 0x11DC , "HANGUL JONGSEONG MIEUM-PIEUP" },
	{ 0x11DD , "HANGUL JONGSEONG MIEUM-SIOS" },
	{ 0x11DE , "HANGUL JONGSEONG MIEUM-SSANGSIOS" },
	{ 0x11DF , "HANGUL JONGSEONG MIEUM-PANSIOS" },
	{ 0x11E0 , "HANGUL JONGSEONG MIEUM-CHIEUCH" },
	{ 0x11E1 , "HANGUL JONGSEONG MIEUM-HIEUH" },
	{ 0x11E2 , "HANGUL JONGSEONG KAPYEOUNMIEUM" },
	{ 0x11E3 , "HANGUL JONGSEONG PIEUP-RIEUL" },
	{ 0x11E4 , "HANGUL JONGSEONG PIEUP-PHIEUPH" },
	{ 0x11E5 , "HANGUL JONGSEONG PIEUP-HIEUH" },
	{ 0x11E6 , "HANGUL JONGSEONG KAPYEOUNPIEUP" },
	{ 0x11E7 , "HANGUL JONGSEONG SIOS-KIYEOK" },
	{ 0x11E8 , "HANGUL JONGSEONG SIOS-TIKEUT" },
	{ 0x11E9 , "HANGUL JONGSEONG SIOS-RIEUL" },
	{ 0x11EA , "HANGUL JONGSEONG SIOS-PIEUP" },
	{ 0x11EB , "HANGUL JONGSEONG PANSIOS" },
	{ 0x11EC , "HANGUL JONGSEONG IEUNG-KIYEOK" },
	{ 0x11ED , "HANGUL JONGSEONG IEUNG-SSANGKIYEOK" },
	{ 0x11EE , "HANGUL JONGSEONG SSANGIEUNG" },
	{ 0x11EF , "HANGUL JONGSEONG IEUNG-KHIEUKH" },
	{ 0x11F0 , "HANGUL JONGSEONG YESIEUNG" },
	{ 0x11F1 , "HANGUL JONGSEONG YESIEUNG-SIOS" },
	{ 0x11F2 , "HANGUL JONGSEONG YESIEUNG-PANSIOS" },
	{ 0x11F3 , "HANGUL JONGSEONG PHIEUPH-PIEUP" },
	{ 0x11F4 , "HANGUL JONGSEONG KAPYEOUNPHIEUPH" },
	{ 0x11F5 , "HANGUL JONGSEONG HIEUH-NIEUN" },
	{ 0x11F6 , "HANGUL JONGSEONG HIEUH-RIEUL" },
	{ 0x11F7 , "HANGUL JONGSEONG HIEUH-MIEUM" },
	{ 0x11F8 , "HANGUL JONGSEONG HIEUH-PIEUP" },
	{ 0x11F9 , "HANGUL JONGSEONG YEORINHIEUH" },
#endif
#if 0
	{ 0x1E00 , "LATIN CAPITAL LETTER A WITH RING BELOW" },
	{ 0x1E01 , "LATIN SMALL LETTER A WITH RING BELOW" },
	{ 0x1E02 , "LATIN CAPITAL LETTER B WITH DOT ABOVE" },
	{ 0x1E03 , "LATIN SMALL LETTER B WITH DOT ABOVE" },
	{ 0x1E04 , "LATIN CAPITAL LETTER B WITH DOT BELOW" },
	{ 0x1E05 , "LATIN SMALL LETTER B WITH DOT BELOW" },
	{ 0x1E06 , "LATIN CAPITAL LETTER B WITH LINE BELOW" },
	{ 0x1E07 , "LATIN SMALL LETTER B WITH LINE BELOW" },
	{ 0x1E08 , "LATIN CAPITAL LETTER C WITH CEDILLA AND ACUTE" },
	{ 0x1E09 , "LATIN SMALL LETTER C WITH CEDILLA AND ACUTE" },
	{ 0x1E0A , "LATIN CAPITAL LETTER D WITH DOT ABOVE" },
	{ 0x1E0B , "LATIN SMALL LETTER D WITH DOT ABOVE" },
	{ 0x1E0C , "LATIN CAPITAL LETTER D WITH DOT BELOW" },
	{ 0x1E0D , "LATIN SMALL LETTER D WITH DOT BELOW" },
	{ 0x1E0E , "LATIN CAPITAL LETTER D WITH LINE BELOW" },
	{ 0x1E0F , "LATIN SMALL LETTER D WITH LINE BELOW" },
	{ 0x1E10 , "LATIN CAPITAL LETTER D WITH CEDILLA" },
	{ 0x1E11 , "LATIN SMALL LETTER D WITH CEDILLA" },
	{ 0x1E12 , "LATIN CAPITAL LETTER D WITH CIRCUMFLEX BELOW" },
	{ 0x1E13 , "LATIN SMALL LETTER D WITH CIRCUMFLEX BELOW" },
	{ 0x1E14 , "LATIN CAPITAL LETTER E WITH MACRON AND GRAVE" },
	{ 0x1E15 , "LATIN SMALL LETTER E WITH MACRON AND GRAVE" },
	{ 0x1E16 , "LATIN CAPITAL LETTER E WITH MACRON AND ACUTE" },
	{ 0x1E17 , "LATIN SMALL LETTER E WITH MACRON AND ACUTE" },
	{ 0x1E18 , "LATIN CAPITAL LETTER E WITH CIRCUMFLEX BELOW" },
	{ 0x1E19 , "LATIN SMALL LETTER E WITH CIRCUMFLEX BELOW" },
	{ 0x1E1A , "LATIN CAPITAL LETTER E WITH TILDE BELOW" },
	{ 0x1E1B , "LATIN SMALL LETTER E WITH TILDE BELOW" },
	{ 0x1E1C , "LATIN CAPITAL LETTER E WITH CEDILLA AND BREVE" },
	{ 0x1E1D , "LATIN SMALL LETTER E WITH CEDILLA AND BREVE" },
	{ 0x1E1E , "LATIN CAPITAL LETTER F WITH DOT ABOVE" },
	{ 0x1E1F , "LATIN SMALL LETTER F WITH DOT ABOVE" },
	{ 0x1E20 , "LATIN CAPITAL LETTER G WITH MACRON" },
	{ 0x1E21 , "LATIN SMALL LETTER G WITH MACRON" },
	{ 0x1E22 , "LATIN CAPITAL LETTER H WITH DOT ABOVE" },
	{ 0x1E23 , "LATIN SMALL LETTER H WITH DOT ABOVE" },
	{ 0x1E24 , "LATIN CAPITAL LETTER H WITH DOT BELOW" },
	{ 0x1E25 , "LATIN SMALL LETTER H WITH DOT BELOW" },
	{ 0x1E26 , "LATIN CAPITAL LETTER H WITH DIAERESIS" },
	{ 0x1E27 , "LATIN SMALL LETTER H WITH DIAERESIS" },
	{ 0x1E28 , "LATIN CAPITAL LETTER H WITH CEDILLA" },
	{ 0x1E29 , "LATIN SMALL LETTER H WITH CEDILLA" },
	{ 0x1E2A , "LATIN CAPITAL LETTER H WITH BREVE BELOW" },
	{ 0x1E2B , "LATIN SMALL LETTER H WITH BREVE BELOW" },
	{ 0x1E2C , "LATIN CAPITAL LETTER I WITH TILDE BELOW" },
	{ 0x1E2D , "LATIN SMALL LETTER I WITH TILDE BELOW" },
	{ 0x1E2E , "LATIN CAPITAL LETTER I WITH DIAERESIS AND ACUTE" },
	{ 0x1E2F , "LATIN SMALL LETTER I WITH DIAERESIS AND ACUTE" },
	{ 0x1E30 , "LATIN CAPITAL LETTER K WITH ACUTE" },
	{ 0x1E31 , "LATIN SMALL LETTER K WITH ACUTE" },
	{ 0x1E32 , "LATIN CAPITAL LETTER K WITH DOT BELOW" },
	{ 0x1E33 , "LATIN SMALL LETTER K WITH DOT BELOW" },
	{ 0x1E34 , "LATIN CAPITAL LETTER K WITH LINE BELOW" },
	{ 0x1E35 , "LATIN SMALL LETTER K WITH LINE BELOW" },
	{ 0x1E36 , "LATIN CAPITAL LETTER L WITH DOT BELOW" },
	{ 0x1E37 , "LATIN SMALL LETTER L WITH DOT BELOW" },
	{ 0x1E38 , "LATIN CAPITAL LETTER L WITH DOT BELOW AND MACRON" },
	{ 0x1E39 , "LATIN SMALL LETTER L WITH DOT BELOW AND MACRON" },
	{ 0x1E3A , "LATIN CAPITAL LETTER L WITH LINE BELOW" },
	{ 0x1E3B , "LATIN SMALL LETTER L WITH LINE BELOW" },
	{ 0x1E3C , "LATIN CAPITAL LETTER L WITH CIRCUMFLEX BELOW" },
	{ 0x1E3D , "LATIN SMALL LETTER L WITH CIRCUMFLEX BELOW" },
	{ 0x1E3E , "LATIN CAPITAL LETTER M WITH ACUTE" },
	{ 0x1E3F , "LATIN SMALL LETTER M WITH ACUTE" },
	{ 0x1E40 , "LATIN CAPITAL LETTER M WITH DOT ABOVE" },
	{ 0x1E41 , "LATIN SMALL LETTER M WITH DOT ABOVE" },
	{ 0x1E42 , "LATIN CAPITAL LETTER M WITH DOT BELOW" },
	{ 0x1E43 , "LATIN SMALL LETTER M WITH DOT BELOW" },
	{ 0x1E44 , "LATIN CAPITAL LETTER N WITH DOT ABOVE" },
	{ 0x1E45 , "LATIN SMALL LETTER N WITH DOT ABOVE" },
	{ 0x1E46 , "LATIN CAPITAL LETTER N WITH DOT BELOW" },
	{ 0x1E47 , "LATIN SMALL LETTER N WITH DOT BELOW" },
	{ 0x1E48 , "LATIN CAPITAL LETTER N WITH LINE BELOW" },
	{ 0x1E49 , "LATIN SMALL LETTER N WITH LINE BELOW" },
	{ 0x1E4A , "LATIN CAPITAL LETTER N WITH CIRCUMFLEX BELOW" },
	{ 0x1E4B , "LATIN SMALL LETTER N WITH CIRCUMFLEX BELOW" },
	{ 0x1E4C , "LATIN CAPITAL LETTER O WITH TILDE AND ACUTE" },
	{ 0x1E4D , "LATIN SMALL LETTER O WITH TILDE AND ACUTE" },
	{ 0x1E4E , "LATIN CAPITAL LETTER O WITH TILDE AND DIAERESIS" },
	{ 0x1E4F , "LATIN SMALL LETTER O WITH TILDE AND DIAERESIS" },
	{ 0x1E50 , "LATIN CAPITAL LETTER O WITH MACRON AND GRAVE" },
	{ 0x1E51 , "LATIN SMALL LETTER O WITH MACRON AND GRAVE" },
	{ 0x1E52 , "LATIN CAPITAL LETTER O WITH MACRON AND ACUTE" },
	{ 0x1E53 , "LATIN SMALL LETTER O WITH MACRON AND ACUTE" },
	{ 0x1E54 , "LATIN CAPITAL LETTER P WITH ACUTE" },
	{ 0x1E55 , "LATIN SMALL LETTER P WITH ACUTE" },
	{ 0x1E56 , "LATIN CAPITAL LETTER P WITH DOT ABOVE" },
	{ 0x1E57 , "LATIN SMALL LETTER P WITH DOT ABOVE" },
	{ 0x1E58 , "LATIN CAPITAL LETTER R WITH DOT ABOVE" },
	{ 0x1E59 , "LATIN SMALL LETTER R WITH DOT ABOVE" },
	{ 0x1E5A , "LATIN CAPITAL LETTER R WITH DOT BELOW" },
	{ 0x1E5B , "LATIN SMALL LETTER R WITH DOT BELOW" },
	{ 0x1E5C , "LATIN CAPITAL LETTER R WITH DOT BELOW AND MACRON" },
	{ 0x1E5D , "LATIN SMALL LETTER R WITH DOT BELOW AND MACRON" },
	{ 0x1E5E , "LATIN CAPITAL LETTER R WITH LINE BELOW" },
	{ 0x1E5F , "LATIN SMALL LETTER R WITH LINE BELOW" },
	{ 0x1E60 , "LATIN CAPITAL LETTER S WITH DOT ABOVE" },
	{ 0x1E61 , "LATIN SMALL LETTER S WITH DOT ABOVE" },
	{ 0x1E62 , "LATIN CAPITAL LETTER S WITH DOT BELOW" },
	{ 0x1E63 , "LATIN SMALL LETTER S WITH DOT BELOW" },
	{ 0x1E64 , "LATIN CAPITAL LETTER S WITH ACUTE AND DOT ABOVE" },
	{ 0x1E65 , "LATIN SMALL LETTER S WITH ACUTE AND DOT ABOVE" },
	{ 0x1E66 , "LATIN CAPITAL LETTER S WITH CARON AND DOT ABOVE" },
	{ 0x1E67 , "LATIN SMALL LETTER S WITH CARON AND DOT ABOVE" },
	{ 0x1E68 , "LATIN CAPITAL LETTER S WITH DOT BELOW AND DOT ABOVE" },
	{ 0x1E69 , "LATIN SMALL LETTER S WITH DOT BELOW AND DOT ABOVE" },
	{ 0x1E6A , "LATIN CAPITAL LETTER T WITH DOT ABOVE" },
	{ 0x1E6B , "LATIN SMALL LETTER T WITH DOT ABOVE" },
	{ 0x1E6C , "LATIN CAPITAL LETTER T WITH DOT BELOW" },
	{ 0x1E6D , "LATIN SMALL LETTER T WITH DOT BELOW" },
	{ 0x1E6E , "LATIN CAPITAL LETTER T WITH LINE BELOW" },
	{ 0x1E6F , "LATIN SMALL LETTER T WITH LINE BELOW" },
	{ 0x1E70 , "LATIN CAPITAL LETTER T WITH CIRCUMFLEX BELOW" },
	{ 0x1E71 , "LATIN SMALL LETTER T WITH CIRCUMFLEX BELOW" },
	{ 0x1E72 , "LATIN CAPITAL LETTER U WITH DIAERESIS BELOW" },
	{ 0x1E73 , "LATIN SMALL LETTER U WITH DIAERESIS BELOW" },
	{ 0x1E74 , "LATIN CAPITAL LETTER U WITH TILDE BELOW" },
	{ 0x1E75 , "LATIN SMALL LETTER U WITH TILDE BELOW" },
	{ 0x1E76 , "LATIN CAPITAL LETTER U WITH CIRCUMFLEX BELOW" },
	{ 0x1E77 , "LATIN SMALL LETTER U WITH CIRCUMFLEX BELOW" },
	{ 0x1E78 , "LATIN CAPITAL LETTER U WITH TILDE AND ACUTE" },
	{ 0x1E79 , "LATIN SMALL LETTER U WITH TILDE AND ACUTE" },
	{ 0x1E7A , "LATIN CAPITAL LETTER U WITH MACRON AND DIAERESIS" },
	{ 0x1E7B , "LATIN SMALL LETTER U WITH MACRON AND DIAERESIS" },
	{ 0x1E7C , "LATIN CAPITAL LETTER V WITH TILDE" },
	{ 0x1E7D , "LATIN SMALL LETTER V WITH TILDE" },
	{ 0x1E7E , "LATIN CAPITAL LETTER V WITH DOT BELOW" },
	{ 0x1E7F , "LATIN SMALL LETTER V WITH DOT BELOW" },
	{ 0x1E80 , "LATIN CAPITAL LETTER W WITH GRAVE" },
	{ 0x1E81 , "LATIN SMALL LETTER W WITH GRAVE" },
	{ 0x1E82 , "LATIN CAPITAL LETTER W WITH ACUTE" },
	{ 0x1E83 , "LATIN SMALL LETTER W WITH ACUTE" },
	{ 0x1E84 , "LATIN CAPITAL LETTER W WITH DIAERESIS" },
	{ 0x1E85 , "LATIN SMALL LETTER W WITH DIAERESIS" },
	{ 0x1E86 , "LATIN CAPITAL LETTER W WITH DOT ABOVE" },
	{ 0x1E87 , "LATIN SMALL LETTER W WITH DOT ABOVE" },
	{ 0x1E88 , "LATIN CAPITAL LETTER W WITH DOT BELOW" },
	{ 0x1E89 , "LATIN SMALL LETTER W WITH DOT BELOW" },
	{ 0x1E8A , "LATIN CAPITAL LETTER X WITH DOT ABOVE" },
	{ 0x1E8B , "LATIN SMALL LETTER X WITH DOT ABOVE" },
	{ 0x1E8C , "LATIN CAPITAL LETTER X WITH DIAERESIS" },
	{ 0x1E8D , "LATIN SMALL LETTER X WITH DIAERESIS" },
	{ 0x1E8E , "LATIN CAPITAL LETTER Y WITH DOT ABOVE" },
	{ 0x1E8F , "LATIN SMALL LETTER Y WITH DOT ABOVE" },
	{ 0x1E90 , "LATIN CAPITAL LETTER Z WITH CIRCUMFLEX" },
	{ 0x1E91 , "LATIN SMALL LETTER Z WITH CIRCUMFLEX" },
	{ 0x1E92 , "LATIN CAPITAL LETTER Z WITH DOT BELOW" },
	{ 0x1E93 , "LATIN SMALL LETTER Z WITH DOT BELOW" },
	{ 0x1E94 , "LATIN CAPITAL LETTER Z WITH LINE BELOW" },
	{ 0x1E95 , "LATIN SMALL LETTER Z WITH LINE BELOW" },
	{ 0x1E96 , "LATIN SMALL LETTER H WITH LINE BELOW" },
	{ 0x1E97 , "LATIN SMALL LETTER T WITH DIAERESIS" },
	{ 0x1E98 , "LATIN SMALL LETTER W WITH RING ABOVE" },
	{ 0x1E99 , "LATIN SMALL LETTER Y WITH RING ABOVE" },
	{ 0x1E9A , "LATIN SMALL LETTER A WITH RIGHT HALF RING" },
	{ 0x1EA0 , "LATIN CAPITAL LETTER A WITH DOT BELOW" },
	{ 0x1EA1 , "LATIN SMALL LETTER A WITH DOT BELOW" },
	{ 0x1EA2 , "LATIN CAPITAL LETTER A WITH HOOK ABOVE" },
	{ 0x1EA3 , "LATIN SMALL LETTER A WITH HOOK ABOVE" },
	{ 0x1EA4 , "LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND ACUTE" },
	{ 0x1EA5 , "LATIN SMALL LETTER A WITH CIRCUMFLEX AND ACUTE" },
	{ 0x1EA6 , "LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND GRAVE" },
	{ 0x1EA7 , "LATIN SMALL LETTER A WITH CIRCUMFLEX AND GRAVE" },
	{ 0x1EA8 , "LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND HOOK ABOVE" },
	{ 0x1EA9 , "LATIN SMALL LETTER A WITH CIRCUMFLEX AND HOOK ABOVE" },
	{ 0x1EAA , "LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND TILDE" },
	{ 0x1EAB , "LATIN SMALL LETTER A WITH CIRCUMFLEX AND TILDE" },
	{ 0x1EAC , "LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND DOT BELOW" },
	{ 0x1EAD , "LATIN SMALL LETTER A WITH CIRCUMFLEX AND DOT BELOW" },
	{ 0x1EAE , "LATIN CAPITAL LETTER A WITH BREVE AND ACUTE" },
	{ 0x1EAF , "LATIN SMALL LETTER A WITH BREVE AND ACUTE" },
	{ 0x1EB0 , "LATIN CAPITAL LETTER A WITH BREVE AND GRAVE" },
	{ 0x1EB1 , "LATIN SMALL LETTER A WITH BREVE AND GRAVE" },
	{ 0x1EB2 , "LATIN CAPITAL LETTER A WITH BREVE AND HOOK ABOVE" },
	{ 0x1EB3 , "LATIN SMALL LETTER A WITH BREVE AND HOOK ABOVE" },
	{ 0x1EB4 , "LATIN CAPITAL LETTER A WITH BREVE AND TILDE" },
	{ 0x1EB5 , "LATIN SMALL LETTER A WITH BREVE AND TILDE" },
	{ 0x1EB6 , "LATIN CAPITAL LETTER A WITH BREVE AND DOT BELOW" },
	{ 0x1EB7 , "LATIN SMALL LETTER A WITH BREVE AND DOT BELOW" },
	{ 0x1EB8 , "LATIN CAPITAL LETTER E WITH DOT BELOW" },
	{ 0x1EB9 , "LATIN SMALL LETTER E WITH DOT BELOW" },
	{ 0x1EBA , "LATIN CAPITAL LETTER E WITH HOOK ABOVE" },
	{ 0x1EBB , "LATIN SMALL LETTER E WITH HOOK ABOVE" },
	{ 0x1EBC , "LATIN CAPITAL LETTER E WITH TILDE" },
	{ 0x1EBD , "LATIN SMALL LETTER E WITH TILDE" },
	{ 0x1EBE , "LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND ACUTE" },
	{ 0x1EBF , "LATIN SMALL LETTER E WITH CIRCUMFLEX AND ACUTE" },
	{ 0x1EC0 , "LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND GRAVE" },
	{ 0x1EC1 , "LATIN SMALL LETTER E WITH CIRCUMFLEX AND GRAVE" },
	{ 0x1EC2 , "LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND HOOK ABOVE" },
	{ 0x1EC3 , "LATIN SMALL LETTER E WITH CIRCUMFLEX AND HOOK ABOVE" },
	{ 0x1EC4 , "LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND TILDE" },
	{ 0x1EC5 , "LATIN SMALL LETTER E WITH CIRCUMFLEX AND TILDE" },
	{ 0x1EC6 , "LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND DOT BELOW" },
	{ 0x1EC7 , "LATIN SMALL LETTER E WITH CIRCUMFLEX AND DOT BELOW" },
	{ 0x1EC8 , "LATIN CAPITAL LETTER I WITH HOOK ABOVE" },
	{ 0x1EC9 , "LATIN SMALL LETTER I WITH HOOK ABOVE" },
	{ 0x1ECA , "LATIN CAPITAL LETTER I WITH DOT BELOW" },
	{ 0x1ECB , "LATIN SMALL LETTER I WITH DOT BELOW" },
	{ 0x1ECC , "LATIN CAPITAL LETTER O WITH DOT BELOW" },
	{ 0x1ECD , "LATIN SMALL LETTER O WITH DOT BELOW" },
	{ 0x1ECE , "LATIN CAPITAL LETTER O WITH HOOK ABOVE" },
	{ 0x1ECF , "LATIN SMALL LETTER O WITH HOOK ABOVE" },
	{ 0x1ED0 , "LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND ACUTE" },
	{ 0x1ED1 , "LATIN SMALL LETTER O WITH CIRCUMFLEX AND ACUTE" },
	{ 0x1ED2 , "LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND GRAVE" },
	{ 0x1ED3 , "LATIN SMALL LETTER O WITH CIRCUMFLEX AND GRAVE" },
	{ 0x1ED4 , "LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND HOOK ABOVE" },
	{ 0x1ED5 , "LATIN SMALL LETTER O WITH CIRCUMFLEX AND HOOK ABOVE" },
	{ 0x1ED6 , "LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND TILDE" },
	{ 0x1ED7 , "LATIN SMALL LETTER O WITH CIRCUMFLEX AND TILDE" },
	{ 0x1ED8 , "LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND DOT BELOW" },
	{ 0x1ED9 , "LATIN SMALL LETTER O WITH CIRCUMFLEX AND DOT BELOW" },
	{ 0x1EDA , "LATIN CAPITAL LETTER O WITH HORN AND ACUTE" },
	{ 0x1EDB , "LATIN SMALL LETTER O WITH HORN AND ACUTE" },
	{ 0x1EDC , "LATIN CAPITAL LETTER O WITH HORN AND GRAVE" },
	{ 0x1EDD , "LATIN SMALL LETTER O WITH HORN AND GRAVE" },
	{ 0x1EDE , "LATIN CAPITAL LETTER O WITH HORN AND HOOK ABOVE" },
	{ 0x1EDF , "LATIN SMALL LETTER O WITH HORN AND HOOK ABOVE" },
	{ 0x1EE0 , "LATIN CAPITAL LETTER O WITH HORN AND TILDE" },
	{ 0x1EE1 , "LATIN SMALL LETTER O WITH HORN AND TILDE" },
	{ 0x1EE2 , "LATIN CAPITAL LETTER O WITH HORN AND DOT BELOW" },
	{ 0x1EE3 , "LATIN SMALL LETTER O WITH HORN AND DOT BELOW" },
	{ 0x1EE4 , "LATIN CAPITAL LETTER U WITH DOT BELOW" },
	{ 0x1EE5 , "LATIN SMALL LETTER U WITH DOT BELOW" },
	{ 0x1EE6 , "LATIN CAPITAL LETTER U WITH HOOK ABOVE" },
	{ 0x1EE7 , "LATIN SMALL LETTER U WITH HOOK ABOVE" },
	{ 0x1EE8 , "LATIN CAPITAL LETTER U WITH HORN AND ACUTE" },
	{ 0x1EE9 , "LATIN SMALL LETTER U WITH HORN AND ACUTE" },
	{ 0x1EEA , "LATIN CAPITAL LETTER U WITH HORN AND GRAVE" },
	{ 0x1EEB , "LATIN SMALL LETTER U WITH HORN AND GRAVE" },
	{ 0x1EEC , "LATIN CAPITAL LETTER U WITH HORN AND HOOK ABOVE" },
	{ 0x1EED , "LATIN SMALL LETTER U WITH HORN AND HOOK ABOVE" },
	{ 0x1EEE , "LATIN CAPITAL LETTER U WITH HORN AND TILDE" },
	{ 0x1EEF , "LATIN SMALL LETTER U WITH HORN AND TILDE" },
	{ 0x1EF0 , "LATIN CAPITAL LETTER U WITH HORN AND DOT BELOW" },
	{ 0x1EF1 , "LATIN SMALL LETTER U WITH HORN AND DOT BELOW" },
	{ 0x1EF2 , "LATIN CAPITAL LETTER Y WITH GRAVE" },
	{ 0x1EF3 , "LATIN SMALL LETTER Y WITH GRAVE" },
	{ 0x1EF4 , "LATIN CAPITAL LETTER Y WITH DOT BELOW" },
	{ 0x1EF5 , "LATIN SMALL LETTER Y WITH DOT BELOW" },
	{ 0x1EF6 , "LATIN CAPITAL LETTER Y WITH HOOK ABOVE" },
	{ 0x1EF7 , "LATIN SMALL LETTER Y WITH HOOK ABOVE" },
	{ 0x1EF8 , "LATIN CAPITAL LETTER Y WITH TILDE" },
	{ 0x1EF9 , "LATIN SMALL LETTER Y WITH TILDE" },
#endif
#if 0
	{ 0x1F00 , "GREEK SMALL LETTER ALPHA WITH PSILI" },
	{ 0x1F01 , "GREEK SMALL LETTER ALPHA WITH DASIA" },
	{ 0x1F02 , "GREEK SMALL LETTER ALPHA WITH PSILI AND VARIA" },
	{ 0x1F03 , "GREEK SMALL LETTER ALPHA WITH DASIA AND VARIA" },
	{ 0x1F04 , "GREEK SMALL LETTER ALPHA WITH PSILI AND OXIA" },
	{ 0x1F05 , "GREEK SMALL LETTER ALPHA WITH DASIA AND OXIA" },
	{ 0x1F06 , "GREEK SMALL LETTER ALPHA WITH PSILI AND PERISPOMENI" },
	{ 0x1F07 , "GREEK SMALL LETTER ALPHA WITH DASIA AND PERISPOMENI" },
	{ 0x1F08 , "GREEK CAPITAL LETTER ALPHA WITH PSILI" },
	{ 0x1F09 , "GREEK CAPITAL LETTER ALPHA WITH DASIA" },
	{ 0x1F0A , "GREEK CAPITAL LETTER ALPHA WITH PSILI AND VARIA" },
	{ 0x1F0B , "GREEK CAPITAL LETTER ALPHA WITH DASIA AND VARIA" },
	{ 0x1F0C , "GREEK CAPITAL LETTER ALPHA WITH PSILI AND OXIA" },
	{ 0x1F0D , "GREEK CAPITAL LETTER ALPHA WITH DASIA AND OXIA" },
	{ 0x1F0E , "GREEK CAPITAL LETTER ALPHA WITH PSILI AND PERISPOMENI" },
	{ 0x1F0F , "GREEK CAPITAL LETTER ALPHA WITH DASIA AND PERISPOMENI" },
	{ 0x1F10 , "GREEK SMALL LETTER EPSILON WITH PSILI" },
	{ 0x1F11 , "GREEK SMALL LETTER EPSILON WITH DASIA" },
	{ 0x1F12 , "GREEK SMALL LETTER EPSILON WITH PSILI AND VARIA" },
	{ 0x1F13 , "GREEK SMALL LETTER EPSILON WITH DASIA AND VARIA" },
	{ 0x1F14 , "GREEK SMALL LETTER EPSILON WITH PSILI AND OXIA" },
	{ 0x1F15 , "GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA" },
	{ 0x1F18 , "GREEK CAPITAL LETTER EPSILON WITH PSILI" },
	{ 0x1F19 , "GREEK CAPITAL LETTER EPSILON WITH DASIA" },
	{ 0x1F1A , "GREEK CAPITAL LETTER EPSILON WITH PSILI AND VARIA" },
	{ 0x1F1B , "GREEK CAPITAL LETTER EPSILON WITH DASIA AND VARIA" },
	{ 0x1F1C , "GREEK CAPITAL LETTER EPSILON WITH PSILI AND OXIA" },
	{ 0x1F1D , "GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA" },
	{ 0x1F20 , "GREEK SMALL LETTER ETA WITH PSILI" },
	{ 0x1F21 , "GREEK SMALL LETTER ETA WITH DASIA" },
	{ 0x1F22 , "GREEK SMALL LETTER ETA WITH PSILI AND VARIA" },
	{ 0x1F23 , "GREEK SMALL LETTER ETA WITH DASIA AND VARIA" },
	{ 0x1F24 , "GREEK SMALL LETTER ETA WITH PSILI AND OXIA" },
	{ 0x1F25 , "GREEK SMALL LETTER ETA WITH DASIA AND OXIA" },
	{ 0x1F26 , "GREEK SMALL LETTER ETA WITH PSILI AND PERISPOMENI" },
	{ 0x1F27 , "GREEK SMALL LETTER ETA WITH DASIA AND PERISPOMENI" },
	{ 0x1F28 , "GREEK CAPITAL LETTER ETA WITH PSILI" },
	{ 0x1F29 , "GREEK CAPITAL LETTER ETA WITH DASIA" },
	{ 0x1F2A , "GREEK CAPITAL LETTER ETA WITH PSILI AND VARIA" },
	{ 0x1F2B , "GREEK CAPITAL LETTER ETA WITH DASIA AND VARIA" },
	{ 0x1F2C , "GREEK CAPITAL LETTER ETA WITH PSILI AND OXIA" },
	{ 0x1F2D , "GREEK CAPITAL LETTER ETA WITH DASIA AND OXIA" },
	{ 0x1F2E , "GREEK CAPITAL LETTER ETA WITH PSILI AND PERISPOMENI" },
	{ 0x1F2F , "GREEK CAPITAL LETTER ETA WITH DASIA AND PERISPOMENI" },
	{ 0x1F30 , "GREEK SMALL LETTER IOTA WITH PSILI" },
	{ 0x1F31 , "GREEK SMALL LETTER IOTA WITH DASIA" },
	{ 0x1F32 , "GREEK SMALL LETTER IOTA WITH PSILI AND VARIA" },
	{ 0x1F33 , "GREEK SMALL LETTER IOTA WITH DASIA AND VARIA" },
	{ 0x1F34 , "GREEK SMALL LETTER IOTA WITH PSILI AND OXIA" },
	{ 0x1F35 , "GREEK SMALL LETTER IOTA WITH DASIA AND OXIA" },
	{ 0x1F36 , "GREEK SMALL LETTER IOTA WITH PSILI AND PERISPOMENI" },
	{ 0x1F37 , "GREEK SMALL LETTER IOTA WITH DASIA AND PERISPOMENI" },
	{ 0x1F38 , "GREEK CAPITAL LETTER IOTA WITH PSILI" },
	{ 0x1F39 , "GREEK CAPITAL LETTER IOTA WITH DASIA" },
	{ 0x1F3A , "GREEK CAPITAL LETTER IOTA WITH PSILI AND VARIA" },
	{ 0x1F3B , "GREEK CAPITAL LETTER IOTA WITH DASIA AND VARIA" },
	{ 0x1F3C , "GREEK CAPITAL LETTER IOTA WITH PSILI AND OXIA" },
	{ 0x1F3D , "GREEK CAPITAL LETTER IOTA WITH DASIA AND OXIA" },
	{ 0x1F3E , "GREEK CAPITAL LETTER IOTA WITH PSILI AND PERISPOMENI" },
	{ 0x1F3F , "GREEK CAPITAL LETTER IOTA WITH DASIA AND PERISPOMENI" },
	{ 0x1F40 , "GREEK SMALL LETTER OMICRON WITH PSILI" },
	{ 0x1F41 , "GREEK SMALL LETTER OMICRON WITH DASIA" },
	{ 0x1F42 , "GREEK SMALL LETTER OMICRON WITH PSILI AND VARIA" },
	{ 0x1F43 , "GREEK SMALL LETTER OMICRON WITH DASIA AND VARIA" },
	{ 0x1F44 , "GREEK SMALL LETTER OMICRON WITH PSILI AND OXIA" },
	{ 0x1F45 , "GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA" },
	{ 0x1F48 , "GREEK CAPITAL LETTER OMICRON WITH PSILI" },
	{ 0x1F49 , "GREEK CAPITAL LETTER OMICRON WITH DASIA" },
	{ 0x1F4A , "GREEK CAPITAL LETTER OMICRON WITH PSILI AND VARIA" },
	{ 0x1F4B , "GREEK CAPITAL LETTER OMICRON WITH DASIA AND VARIA" },
	{ 0x1F4C , "GREEK CAPITAL LETTER OMICRON WITH PSILI AND OXIA" },
	{ 0x1F4D , "GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA" },
	{ 0x1F50 , "GREEK SMALL LETTER UPSILON WITH PSILI" },
	{ 0x1F51 , "GREEK SMALL LETTER UPSILON WITH DASIA" },
	{ 0x1F52 , "GREEK SMALL LETTER UPSILON WITH PSILI AND VARIA" },
	{ 0x1F53 , "GREEK SMALL LETTER UPSILON WITH DASIA AND VARIA" },
	{ 0x1F54 , "GREEK SMALL LETTER UPSILON WITH PSILI AND OXIA" },
	{ 0x1F55 , "GREEK SMALL LETTER UPSILON WITH DASIA AND OXIA" },
	{ 0x1F56 , "GREEK SMALL LETTER UPSILON WITH PSILI AND PERISPOMENI" },
	{ 0x1F57 , "GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI" },
	{ 0x1F59 , "GREEK CAPITAL LETTER UPSILON WITH DASIA" },
	{ 0x1F5B , "GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA" },
	{ 0x1F5D , "GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA" },
	{ 0x1F5F , "GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI" },
	{ 0x1F60 , "GREEK SMALL LETTER OMEGA WITH PSILI" },
	{ 0x1F61 , "GREEK SMALL LETTER OMEGA WITH DASIA" },
	{ 0x1F62 , "GREEK SMALL LETTER OMEGA WITH PSILI AND VARIA" },
	{ 0x1F63 , "GREEK SMALL LETTER OMEGA WITH DASIA AND VARIA" },
	{ 0x1F64 , "GREEK SMALL LETTER OMEGA WITH PSILI AND OXIA" },
	{ 0x1F65 , "GREEK SMALL LETTER OMEGA WITH DASIA AND OXIA" },
	{ 0x1F66 , "GREEK SMALL LETTER OMEGA WITH PSILI AND PERISPOMENI" },
	{ 0x1F67 , "GREEK SMALL LETTER OMEGA WITH DASIA AND PERISPOMENI" },
	{ 0x1F68 , "GREEK CAPITAL LETTER OMEGA WITH PSILI" },
	{ 0x1F69 , "GREEK CAPITAL LETTER OMEGA WITH DASIA" },
	{ 0x1F6A , "GREEK CAPITAL LETTER OMEGA WITH PSILI AND VARIA" },
	{ 0x1F6B , "GREEK CAPITAL LETTER OMEGA WITH DASIA AND VARIA" },
	{ 0x1F6C , "GREEK CAPITAL LETTER OMEGA WITH PSILI AND OXIA" },
	{ 0x1F6D , "GREEK CAPITAL LETTER OMEGA WITH DASIA AND OXIA" },
	{ 0x1F6E , "GREEK CAPITAL LETTER OMEGA WITH PSILI AND PERISPOMENI" },
	{ 0x1F6F , "GREEK CAPITAL LETTER OMEGA WITH DASIA AND PERISPOMENI" },
	{ 0x1F70 , "GREEK SMALL LETTER ALPHA WITH VARIA" },
	{ 0x1F71 , "GREEK SMALL LETTER ALPHA WITH OXIA" },
	{ 0x1F72 , "GREEK SMALL LETTER EPSILON WITH VARIA" },
	{ 0x1F73 , "GREEK SMALL LETTER EPSILON WITH OXIA" },
	{ 0x1F74 , "GREEK SMALL LETTER ETA WITH VARIA" },
	{ 0x1F75 , "GREEK SMALL LETTER ETA WITH OXIA" },
	{ 0x1F76 , "GREEK SMALL LETTER IOTA WITH VARIA" },
	{ 0x1F77 , "GREEK SMALL LETTER IOTA WITH OXIA" },
	{ 0x1F78 , "GREEK SMALL LETTER OMICRON WITH VARIA" },
	{ 0x1F79 , "GREEK SMALL LETTER OMICRON WITH OXIA" },
	{ 0x1F7A , "GREEK SMALL LETTER UPSILON WITH VARIA" },
	{ 0x1F7B , "GREEK SMALL LETTER UPSILON WITH OXIA" },
	{ 0x1F7C , "GREEK SMALL LETTER OMEGA WITH VARIA" },
	{ 0x1F7D , "GREEK SMALL LETTER OMEGA WITH OXIA" },
	{ 0x1F80 , "GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI" },
	{ 0x1F81 , "GREEK SMALL LETTER ALPHA WITH DASIA AND YPOGEGRAMMENI" },
	{ 0x1F82 , "GREEK SMALL LETTER ALPHA WITH PSILI AND VARIA AND YPOGEGRAMMENI" },
	{ 0x1F83 , "GREEK SMALL LETTER ALPHA WITH DASIA AND VARIA AND YPOGEGRAMMENI" },
	{ 0x1F84 , "GREEK SMALL LETTER ALPHA WITH PSILI AND OXIA AND YPOGEGRAMMENI" },
	{ 0x1F85 , "GREEK SMALL LETTER ALPHA WITH DASIA AND OXIA AND YPOGEGRAMMENI" },
	{ 0x1F86 , "GREEK SMALL LETTER ALPHA WITH PSILI AND PERISPOMENI AND YPOGEGRAMMENI" },
	{ 0x1F87 , "GREEK SMALL LETTER ALPHA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI" },
	{ 0x1F88 , "GREEK CAPITAL LETTER ALPHA WITH PSILI AND PROSGEGRAMMENI" },
	{ 0x1F89 , "GREEK CAPITAL LETTER ALPHA WITH DASIA AND PROSGEGRAMMENI" },
	{ 0x1F8A , "GREEK CAPITAL LETTER ALPHA WITH PSILI AND VARIA AND PROSGEGRAMMENI" },
	{ 0x1F8B , "GREEK CAPITAL LETTER ALPHA WITH DASIA AND VARIA AND PROSGEGRAMMENI" },
	{ 0x1F8C , "GREEK CAPITAL LETTER ALPHA WITH PSILI AND OXIA AND PROSGEGRAMMENI" },
	{ 0x1F8D , "GREEK CAPITAL LETTER ALPHA WITH DASIA AND OXIA AND PROSGEGRAMMENI" },
	{ 0x1F8E , "GREEK CAPITAL LETTER ALPHA WITH PSILI AND PERISPOMENI AND PROSGEGRAMMENI" },
	{ 0x1F8F , "GREEK CAPITAL LETTER ALPHA WITH DASIA AND PERISPOMENI AND PROSGEGRAMMENI" },
	{ 0x1F90 , "GREEK SMALL LETTER ETA WITH PSILI AND YPOGEGRAMMENI" },
	{ 0x1F91 , "GREEK SMALL LETTER ETA WITH DASIA AND YPOGEGRAMMENI" },
	{ 0x1F92 , "GREEK SMALL LETTER ETA WITH PSILI AND VARIA AND YPOGEGRAMMENI" },
	{ 0x1F93 , "GREEK SMALL LETTER ETA WITH DASIA AND VARIA AND YPOGEGRAMMENI" },
	{ 0x1F94 , "GREEK SMALL LETTER ETA WITH PSILI AND OXIA AND YPOGEGRAMMENI" },
	{ 0x1F95 , "GREEK SMALL LETTER ETA WITH DASIA AND OXIA AND YPOGEGRAMMENI" },
	{ 0x1F96 , "GREEK SMALL LETTER ETA WITH PSILI AND PERISPOMENI AND YPOGEGRAMMENI" },
	{ 0x1F97 , "GREEK SMALL LETTER ETA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI" },
	{ 0x1F98 , "GREEK CAPITAL LETTER ETA WITH PSILI AND PROSGEGRAMMENI" },
	{ 0x1F99 , "GREEK CAPITAL LETTER ETA WITH DASIA AND PROSGEGRAMMENI" },
	{ 0x1F9A , "GREEK CAPITAL LETTER ETA WITH PSILI AND VARIA AND PROSGEGRAMMENI" },
	{ 0x1F9B , "GREEK CAPITAL LETTER ETA WITH DASIA AND VARIA AND PROSGEGRAMMENI" },
	{ 0x1F9C , "GREEK CAPITAL LETTER ETA WITH PSILI AND OXIA AND PROSGEGRAMMENI" },
	{ 0x1F9D , "GREEK CAPITAL LETTER ETA WITH DASIA AND OXIA AND PROSGEGRAMMENI" },
	{ 0x1F9E , "GREEK CAPITAL LETTER ETA WITH PSILI AND PERISPOMENI AND PROSGEGRAMMENI" },
	{ 0x1F9F , "GREEK CAPITAL LETTER ETA WITH DASIA AND PERISPOMENI AND PROSGEGRAMMENI" },
	{ 0x1FA0 , "GREEK SMALL LETTER OMEGA WITH PSILI AND YPOGEGRAMMENI" },
	{ 0x1FA1 , "GREEK SMALL LETTER OMEGA WITH DASIA AND YPOGEGRAMMENI" },
	{ 0x1FA2 , "GREEK SMALL LETTER OMEGA WITH PSILI AND VARIA AND YPOGEGRAMMENI" },
	{ 0x1FA3 , "GREEK SMALL LETTER OMEGA WITH DASIA AND VARIA AND YPOGEGRAMMENI" },
	{ 0x1FA4 , "GREEK SMALL LETTER OMEGA WITH PSILI AND OXIA AND YPOGEGRAMMENI" },
	{ 0x1FA5 , "GREEK SMALL LETTER OMEGA WITH DASIA AND OXIA AND YPOGEGRAMMENI" },
	{ 0x1FA6 , "GREEK SMALL LETTER OMEGA WITH PSILI AND PERISPOMENI AND YPOGEGRAMMENI" },
	{ 0x1FA7 , "GREEK SMALL LETTER OMEGA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI" },
	{ 0x1FA8 , "GREEK CAPITAL LETTER OMEGA WITH PSILI AND PROSGEGRAMMENI" },
	{ 0x1FA9 , "GREEK CAPITAL LETTER OMEGA WITH DASIA AND PROSGEGRAMMENI" },
	{ 0x1FAA , "GREEK CAPITAL LETTER OMEGA WITH PSILI AND VARIA AND PROSGEGRAMMENI" },
	{ 0x1FAB , "GREEK CAPITAL LETTER OMEGA WITH DASIA AND VARIA AND PROSGEGRAMMENI" },
	{ 0x1FAC , "GREEK CAPITAL LETTER OMEGA WITH PSILI AND OXIA AND PROSGEGRAMMENI" },
	{ 0x1FAD , "GREEK CAPITAL LETTER OMEGA WITH DASIA AND OXIA AND PROSGEGRAMMENI" },
	{ 0x1FAE , "GREEK CAPITAL LETTER OMEGA WITH PSILI AND PERISPOMENI AND PROSGEGRAMMENI" },
	{ 0x1FAF , "GREEK CAPITAL LETTER OMEGA WITH DASIA AND PERISPOMENI AND PROSGEGRAMMENI" },
	{ 0x1FB0 , "GREEK SMALL LETTER ALPHA WITH VRACHY" },
	{ 0x1FB1 , "GREEK SMALL LETTER ALPHA WITH MACRON" },
	{ 0x1FB2 , "GREEK SMALL LETTER ALPHA WITH VARIA AND YPOGEGRAMMENI" },
	{ 0x1FB3 , "GREEK SMALL LETTER ALPHA WITH YPOGEGRAMMENI" },
	{ 0x1FB4 , "GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI" },
	{ 0x1FB6 , "GREEK SMALL LETTER ALPHA WITH PERISPOMENI" },
	{ 0x1FB7 , "GREEK SMALL LETTER ALPHA WITH PERISPOMENI AND YPOGEGRAMMENI" },
	{ 0x1FB8 , "GREEK CAPITAL LETTER ALPHA WITH VRACHY" },
	{ 0x1FB9 , "GREEK CAPITAL LETTER ALPHA WITH MACRON" },
	{ 0x1FBA , "GREEK CAPITAL LETTER ALPHA WITH VARIA" },
	{ 0x1FBB , "GREEK CAPITAL LETTER ALPHA WITH OXIA" },
	{ 0x1FBC , "GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI" },
	{ 0x1FBD , "GREEK KORONIS" },
	{ 0x1FBE , "GREEK PROSGEGRAMMENI" },
	{ 0x1FBF , "GREEK PSILI" },
	{ 0x1FC0 , "GREEK PERISPOMENI" },
	{ 0x1FC1 , "GREEK DIALYTIKA AND PERISPOMENI" },
	{ 0x1FC2 , "GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI" },
	{ 0x1FC3 , "GREEK SMALL LETTER ETA WITH YPOGEGRAMMENI" },
	{ 0x1FC4 , "GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI" },
	{ 0x1FC6 , "GREEK SMALL LETTER ETA WITH PERISPOMENI" },
	{ 0x1FC7 , "GREEK SMALL LETTER ETA WITH PERISPOMENI AND YPOGEGRAMMENI" },
	{ 0x1FC8 , "GREEK CAPITAL LETTER EPSILON WITH VARIA" },
	{ 0x1FC9 , "GREEK CAPITAL LETTER EPSILON WITH OXIA" },
	{ 0x1FCA , "GREEK CAPITAL LETTER ETA WITH VARIA" },
	{ 0x1FCB , "GREEK CAPITAL LETTER ETA WITH OXIA" },
	{ 0x1FCC , "GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI" },
	{ 0x1FCD , "GREEK PSILI AND VARIA" },
	{ 0x1FCE , "GREEK PSILI AND OXIA" },
	{ 0x1FCF , "GREEK PSILI AND PERISPOMENI" },
	{ 0x1FD0 , "GREEK SMALL LETTER IOTA WITH VRACHY" },
	{ 0x1FD1 , "GREEK SMALL LETTER IOTA WITH MACRON" },
	{ 0x1FD2 , "GREEK SMALL LETTER IOTA WITH DIALYTIKA AND VARIA" },
	{ 0x1FD3 , "GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA" },
	{ 0x1FD6 , "GREEK SMALL LETTER IOTA WITH PERISPOMENI" },
	{ 0x1FD7 , "GREEK SMALL LETTER IOTA WITH DIALYTIKA AND PERISPOMENI" },
	{ 0x1FD8 , "GREEK CAPITAL LETTER IOTA WITH VRACHY" },
	{ 0x1FD9 , "GREEK CAPITAL LETTER IOTA WITH MACRON" },
	{ 0x1FDA , "GREEK CAPITAL LETTER IOTA WITH VARIA" },
	{ 0x1FDB , "GREEK CAPITAL LETTER IOTA WITH OXIA" },
	{ 0x1FDD , "GREEK DASIA AND VARIA" },
	{ 0x1FDE , "GREEK DASIA AND OXIA" },
	{ 0x1FDF , "GREEK DASIA AND PERISPOMENI" },
	{ 0x1FE0 , "GREEK SMALL LETTER UPSILON WITH VRACHY" },
	{ 0x1FE1 , "GREEK SMALL LETTER UPSILON WITH MACRON" },
	{ 0x1FE2 , "GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND VARIA" },
	{ 0x1FE3 , "GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND OXIA" },
	{ 0x1FE4 , "GREEK SMALL LETTER RHO WITH PSILI" },
	{ 0x1FE5 , "GREEK SMALL LETTER RHO WITH DASIA" },
	{ 0x1FE6 , "GREEK SMALL LETTER UPSILON WITH PERISPOMENI" },
	{ 0x1FE7 , "GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND PERISPOMENI" },
	{ 0x1FE8 , "GREEK CAPITAL LETTER UPSILON WITH VRACHY" },
	{ 0x1FE9 , "GREEK CAPITAL LETTER UPSILON WITH MACRON" },
	{ 0x1FEA , "GREEK CAPITAL LETTER UPSILON WITH VARIA" },
	{ 0x1FEB , "GREEK CAPITAL LETTER UPSILON WITH OXIA" },
	{ 0x1FEC , "GREEK CAPITAL LETTER RHO WITH DASIA" },
	{ 0x1FED , "GREEK DIALYTIKA AND VARIA" },
	{ 0x1FEE , "GREEK DIALYTIKA AND OXIA" },
	{ 0x1FEF , "GREEK VARIA" },
	{ 0x1FF2 , "GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI" },
	{ 0x1FF3 , "GREEK SMALL LETTER OMEGA WITH YPOGEGRAMMENI" },
	{ 0x1FF4 , "GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI" },
	{ 0x1FF6 , "GREEK SMALL LETTER OMEGA WITH PERISPOMENI" },
	{ 0x1FF7 , "GREEK SMALL LETTER OMEGA WITH PERISPOMENI AND YPOGEGRAMMENI" },
	{ 0x1FF8 , "GREEK CAPITAL LETTER OMICRON WITH VARIA" },
	{ 0x1FF9 , "GREEK CAPITAL LETTER OMICRON WITH OXIA" },
	{ 0x1FFA , "GREEK CAPITAL LETTER OMEGA WITH VARIA" },
	{ 0x1FFB , "GREEK CAPITAL LETTER OMEGA WITH OXIA" },
	{ 0x1FFC , "GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI" },
	{ 0x1FFD , "GREEK OXIA" },
	{ 0x1FFE , "GREEK DASIA" },
#endif
#if 0
	{ 0x2000 , "EN QUAD" },
	{ 0x2001 , "EM QUAD" },
	{ 0x2002 , "EN SPACE" },
	{ 0x2003 , "EM SPACE" },
	{ 0x2004 , "THREE-PER-EM SPACE" },
	{ 0x2005 , "FOUR-PER-EM SPACE" },
	{ 0x2006 , "SIX-PER-EM SPACE" },
	{ 0x2007 , "FIGURE SPACE" },
	{ 0x2008 , "PUNCTUATION SPACE" },
	{ 0x2009 , "THIN SPACE" },
	{ 0x200A , "HAIR SPACE" },
	{ 0x200B , "ZERO WIDTH SPACE" },
	{ 0x200C , "ZERO WIDTH NON-JOINER" },
	{ 0x200D , "ZERO WIDTH JOINER" },
	{ 0x200E , "LEFT-TO-RIGHT MARK" },
	{ 0x200F , "RIGHT-TO-LEFT MARK" },
	{ 0x2010 , "HYPHEN" },
	{ 0x2011 , "NON-BREAKING HYPHEN" },
	{ 0x2012 , "FIGURE DASH" },
	{ 0x2013 , "EN DASH" },
	{ 0x2014 , "EM DASH" },
	{ 0x2015 , "HORIZONTAL BAR" },
	{ 0x2016 , "DOUBLE VERTICAL LINE" },
	{ 0x2017 , "DOUBLE LOW LINE" },
	{ 0x2018 , "LEFT SINGLE QUOTATION MARK" },
	{ 0x2019 , "RIGHT SINGLE QUOTATION MARK" },
	{ 0x201A , "SINGLE LOW-9 QUOTATION MARK" },
	{ 0x201B , "SINGLE HIGH-REVERSED-9 QUOTATION MARK" },
	{ 0x201C , "LEFT DOUBLE QUOTATION MARK" },
	{ 0x201D , "RIGHT DOUBLE QUOTATION MARK" },
	{ 0x201E , "DOUBLE LOW-9 QUOTATION MARK" },
	{ 0x201F , "DOUBLE HIGH-REVERSED-9 QUOTATION MARK" },
	{ 0x2020 , "DAGGER" },
	{ 0x2021 , "DOUBLE DAGGER" },
	{ 0x2022 , "BULLET" },
	{ 0x2023 , "TRIANGULAR BULLET" },
	{ 0x2024 , "ONE DOT LEADER" },
	{ 0x2025 , "TWO DOT LEADER" },
	{ 0x2026 , "HORIZONTAL ELLIPSIS" },
	{ 0x2027 , "HYPHENATION POINT" },
	{ 0x2028 , "LINE SEPARATOR" },
	{ 0x2029 , "PARAGRAPH SEPARATOR" },
	{ 0x202A , "LEFT-TO-RIGHT EMBEDDING" },
	{ 0x202B , "RIGHT-TO-LEFT EMBEDDING" },
	{ 0x202C , "POP DIRECTIONAL FORMATTING" },
	{ 0x202D , "LEFT-TO-RIGHT OVERRIDE" },
	{ 0x202E , "RIGHT-TO-LEFT OVERRIDE" },
	{ 0x2030 , "PER MILLE SIGN" },
	{ 0x2031 , "PER TEN THOUSAND SIGN" },
	{ 0x2032 , "PRIME" },
	{ 0x2033 , "DOUBLE PRIME" },
	{ 0x2034 , "TRIPLE PRIME" },
	{ 0x2035 , "REVERSED PRIME" },
	{ 0x2036 , "REVERSED DOUBLE PRIME" },
	{ 0x2037 , "REVERSED TRIPLE PRIME" },
	{ 0x2038 , "CARET" },
	{ 0x2039 , "SINGLE LEFT-POINTING ANGLE QUOTATION MARK" },
	{ 0x203A , "SINGLE RIGHT-POINTING ANGLE QUOTATION MARK" },
	{ 0x203B , "REFERENCE MARK" },
	{ 0x203C , "DOUBLE EXCLAMATION MARK" },
	{ 0x203D , "INTERROBANG" },
	{ 0x203E , "OVERLINE" },
	{ 0x203F , "UNDERTIE" },
	{ 0x2040 , "CHARACTER TIE" },
	{ 0x2041 , "CARET INSERTION POINT" },
	{ 0x2042 , "ASTERISM" },
	{ 0x2043 , "HYPHEN BULLET" },
	{ 0x2044 , "FRACTION SLASH" },
	{ 0x2045 , "LEFT SQUARE BRACKET WITH QUILL" },
	{ 0x2046 , "RIGHT SQUARE BRACKET WITH QUILL" },
	{ 0x206A , "INHIBIT SYMMETRIC SWAPPING" },
	{ 0x206B , "ACTIVATE SYMMETRIC SWAPPING" },
	{ 0x206C , "INHIBIT ARABIC FORM SHAPING" },
	{ 0x206D , "ACTIVATE ARABIC FORM SHAPING" },
	{ 0x206E , "NATIONAL DIGIT SHAPES" },
	{ 0x206F , "NOMINAL DIGIT SHAPES" },
	{ 0x2070 , "SUPERSCRIPT ZERO" },
	{ 0x2074 , "SUPERSCRIPT FOUR" },
	{ 0x2075 , "SUPERSCRIPT FIVE" },
	{ 0x2076 , "SUPERSCRIPT SIX" },
	{ 0x2077 , "SUPERSCRIPT SEVEN" },
	{ 0x2078 , "SUPERSCRIPT EIGHT" },
	{ 0x2079 , "SUPERSCRIPT NINE" },
	{ 0x207A , "SUPERSCRIPT PLUS SIGN" },
	{ 0x207B , "SUPERSCRIPT MINUS" },
	{ 0x207C , "SUPERSCRIPT EQUALS SIGN" },
	{ 0x207D , "SUPERSCRIPT LEFT PARENTHESIS" },
	{ 0x207E , "SUPERSCRIPT RIGHT PARENTHESIS" },
	{ 0x207F , "SUPERSCRIPT LATIN SMALL LETTER N" },
	{ 0x2080 , "SUBSCRIPT ZERO" },
	{ 0x2081 , "SUBSCRIPT ONE" },
	{ 0x2082 , "SUBSCRIPT TWO" },
	{ 0x2083 , "SUBSCRIPT THREE" },
	{ 0x2084 , "SUBSCRIPT FOUR" },
	{ 0x2085 , "SUBSCRIPT FIVE" },
	{ 0x2086 , "SUBSCRIPT SIX" },
	{ 0x2087 , "SUBSCRIPT SEVEN" },
	{ 0x2088 , "SUBSCRIPT EIGHT" },
	{ 0x2089 , "SUBSCRIPT NINE" },
	{ 0x208A , "SUBSCRIPT PLUS SIGN" },
	{ 0x208B , "SUBSCRIPT MINUS" },
	{ 0x208C , "SUBSCRIPT EQUALS SIGN" },
	{ 0x208D , "SUBSCRIPT LEFT PARENTHESIS" },
	{ 0x208E , "SUBSCRIPT RIGHT PARENTHESIS" },
	{ 0x20A0 , "EURO-CURRENCY SIGN" },
	{ 0x20A1 , "COLON SIGN" },
	{ 0x20A2 , "CRUZEIRO SIGN" },
	{ 0x20A3 , "FRENCH FRANC SIGN" },
	{ 0x20A4 , "LIRA SIGN" },
	{ 0x20A5 , "MILL SIGN" },
	{ 0x20A6 , "NAIRA SIGN" },
	{ 0x20A7 , "PESETA SIGN" },
	{ 0x20A8 , "RUPEE SIGN" },
	{ 0x20A9 , "WON SIGN" },
	{ 0x20AA , "NEW SHEQEL SIGN" },
	{ 0x20D0 , "COMBINING LEFT HARPOON ABOVE" },
	{ 0x20D1 , "COMBINING RIGHT HARPOON ABOVE" },
	{ 0x20D2 , "COMBINING LONG VERTICAL LINE OVERLAY" },
	{ 0x20D3 , "COMBINING SHORT VERTICAL LINE OVERLAY" },
	{ 0x20D4 , "COMBINING ANTICLOCKWISE ARROW ABOVE" },
	{ 0x20D5 , "COMBINING CLOCKWISE ARROW ABOVE" },
	{ 0x20D6 , "COMBINING LEFT ARROW ABOVE" },
	{ 0x20D7 , "COMBINING RIGHT ARROW ABOVE" },
	{ 0x20D8 , "COMBINING RING OVERLAY" },
	{ 0x20D9 , "COMBINING CLOCKWISE RING OVERLAY" },
	{ 0x20DA , "COMBINING ANTICLOCKWISE RING OVERLAY" },
	{ 0x20DB , "COMBINING THREE DOTS ABOVE" },
	{ 0x20DC , "COMBINING FOUR DOTS ABOVE" },
	{ 0x20DD , "COMBINING ENCLOSING CIRCLE" },
	{ 0x20DE , "COMBINING ENCLOSING SQUARE" },
	{ 0x20DF , "COMBINING ENCLOSING DIAMOND" },
	{ 0x20E0 , "COMBINING ENCLOSING CIRCLE BACKSLASH" },
	{ 0x20E1 , "COMBINING LEFT RIGHT ARROW ABOVE" },
	{ 0x2100 , "ACCOUNT OF" },
	{ 0x2101 , "ADDRESSED TO THE SUBJECT" },
	{ 0x2102 , "DOUBLE-STRUCK CAPITAL C" },
	{ 0x2103 , "DEGREE CELSIUS" },
	{ 0x2104 , "CENTRE LINE SYMBOL" },
	{ 0x2105 , "CARE OF" },
	{ 0x2106 , "CADA UNA" },
	{ 0x2107 , "EULER CONSTANT" },
	{ 0x2108 , "SCRUPLE" },
	{ 0x2109 , "DEGREE FAHRENHEIT" },
	{ 0x210A , "SCRIPT SMALL G" },
	{ 0x210B , "SCRIPT CAPITAL H" },
	{ 0x210C , "BLACK-LETTER CAPITAL H" },
	{ 0x210D , "DOUBLE-STRUCK CAPITAL H" },
	{ 0x210E , "PLANCK CONSTANT" },
	{ 0x210F , "PLANCK CONSTANT OVER TWO PI" },
	{ 0x2110 , "SCRIPT CAPITAL I" },
	{ 0x2111 , "BLACK-LETTER CAPITAL I" },
	{ 0x2112 , "SCRIPT CAPITAL L" },
	{ 0x2113 , "SCRIPT SMALL L" },
	{ 0x2114 , "L B BAR SYMBOL" },
	{ 0x2115 , "DOUBLE-STRUCK CAPITAL N" },
	{ 0x2116 , "NUMERO SIGN" },
	{ 0x2117 , "SOUND RECORDING COPYRIGHT" },
	{ 0x2118 , "SCRIPT CAPITAL P" },
	{ 0x2119 , "DOUBLE-STRUCK CAPITAL P" },
	{ 0x211A , "DOUBLE-STRUCK CAPITAL Q" },
	{ 0x211B , "SCRIPT CAPITAL R" },
	{ 0x211C , "BLACK-LETTER CAPITAL R" },
	{ 0x211D , "DOUBLE-STRUCK CAPITAL R" },
	{ 0x211E , "PRESCRIPTION TAKE" },
	{ 0x211F , "RESPONSE" },
	{ 0x2120 , "SERVICE MARK" },
	{ 0x2121 , "TELEPHONE SIGN" },
	{ 0x2122 , "TRADE MARK SIGN" },
	{ 0x2123 , "VERSICLE" },
	{ 0x2124 , "DOUBLE-STRUCK CAPITAL Z" },
	{ 0x2125 , "OUNCE SIGN" },
	{ 0x2126 , "OHM SIGN" },
	{ 0x2127 , "INVERTED OHM SIGN" },
	{ 0x2128 , "BLACK-LETTER CAPITAL Z" },
	{ 0x2129 , "TURNED GREEK SMALL LETTER IOTA" },
	{ 0x212A , "KELVIN SIGN" },
	{ 0x212B , "ANGSTROM SIGN" },
	{ 0x212C , "SCRIPT CAPITAL B" },
	{ 0x212D , "BLACK-LETTER CAPITAL C" },
	{ 0x212E , "ESTIMATED SYMBOL" },
	{ 0x212F , "SCRIPT SMALL E" },
	{ 0x2130 , "SCRIPT CAPITAL E" },
	{ 0x2131 , "SCRIPT CAPITAL F" },
	{ 0x2132 , "TURNED CAPITAL F" },
	{ 0x2133 , "SCRIPT CAPITAL M" },
	{ 0x2134 , "SCRIPT SMALL O" },
	{ 0x2135 , "ALEF SYMBOL" },
	{ 0x2136 , "BET SYMBOL" },
	{ 0x2137 , "GIMEL SYMBOL" },
	{ 0x2138 , "DALET SYMBOL" },
	{ 0x2153 , "VULGAR FRACTION ONE THIRD" },
	{ 0x2154 , "VULGAR FRACTION TWO THIRDS" },
	{ 0x2155 , "VULGAR FRACTION ONE FIFTH" },
	{ 0x2156 , "VULGAR FRACTION TWO FIFTHS" },
	{ 0x2157 , "VULGAR FRACTION THREE FIFTHS" },
	{ 0x2158 , "VULGAR FRACTION FOUR FIFTHS" },
	{ 0x2159 , "VULGAR FRACTION ONE SIXTH" },
	{ 0x215A , "VULGAR FRACTION FIVE SIXTHS" },
	{ 0x215B , "VULGAR FRACTION ONE EIGHTH" },
	{ 0x215C , "VULGAR FRACTION THREE EIGHTHS" },
	{ 0x215D , "VULGAR FRACTION FIVE EIGHTHS" },
	{ 0x215E , "VULGAR FRACTION SEVEN EIGHTHS" },
	{ 0x215F , "FRACTION NUMERATOR ONE" },
	{ 0x2160 , "ROMAN NUMERAL ONE" },
	{ 0x2161 , "ROMAN NUMERAL TWO" },
	{ 0x2162 , "ROMAN NUMERAL THREE" },
	{ 0x2163 , "ROMAN NUMERAL FOUR" },
	{ 0x2164 , "ROMAN NUMERAL FIVE" },
	{ 0x2165 , "ROMAN NUMERAL SIX" },
	{ 0x2166 , "ROMAN NUMERAL SEVEN" },
	{ 0x2167 , "ROMAN NUMERAL EIGHT" },
	{ 0x2168 , "ROMAN NUMERAL NINE" },
	{ 0x2169 , "ROMAN NUMERAL TEN" },
	{ 0x216A , "ROMAN NUMERAL ELEVEN" },
	{ 0x216B , "ROMAN NUMERAL TWELVE" },
	{ 0x216C , "ROMAN NUMERAL FIFTY" },
	{ 0x216D , "ROMAN NUMERAL ONE HUNDRED" },
	{ 0x216E , "ROMAN NUMERAL FIVE HUNDRED" },
	{ 0x216F , "ROMAN NUMERAL ONE THOUSAND" },
	{ 0x2170 , "SMALL ROMAN NUMERAL ONE" },
	{ 0x2171 , "SMALL ROMAN NUMERAL TWO" },
	{ 0x2172 , "SMALL ROMAN NUMERAL THREE" },
	{ 0x2173 , "SMALL ROMAN NUMERAL FOUR" },
	{ 0x2174 , "SMALL ROMAN NUMERAL FIVE" },
	{ 0x2175 , "SMALL ROMAN NUMERAL SIX" },
	{ 0x2176 , "SMALL ROMAN NUMERAL SEVEN" },
	{ 0x2177 , "SMALL ROMAN NUMERAL EIGHT" },
	{ 0x2178 , "SMALL ROMAN NUMERAL NINE" },
	{ 0x2179 , "SMALL ROMAN NUMERAL TEN" },
	{ 0x217A , "SMALL ROMAN NUMERAL ELEVEN" },
	{ 0x217B , "SMALL ROMAN NUMERAL TWELVE" },
	{ 0x217C , "SMALL ROMAN NUMERAL FIFTY" },
	{ 0x217D , "SMALL ROMAN NUMERAL ONE HUNDRED" },
	{ 0x217E , "SMALL ROMAN NUMERAL FIVE HUNDRED" },
	{ 0x217F , "SMALL ROMAN NUMERAL ONE THOUSAND" },
	{ 0x2180 , "ROMAN NUMERAL ONE THOUSAND C D" },
	{ 0x2181 , "ROMAN NUMERAL FIVE THOUSAND" },
	{ 0x2182 , "ROMAN NUMERAL TEN THOUSAND" },
	{ 0x2190 , "LEFTWARDS ARROW" },
	{ 0x2191 , "UPWARDS ARROW" },
	{ 0x2192 , "RIGHTWARDS ARROW" },
	{ 0x2193 , "DOWNWARDS ARROW" },
	{ 0x2194 , "LEFT RIGHT ARROW" },
	{ 0x2195 , "UP DOWN ARROW" },
	{ 0x2196 , "NORTH WEST ARROW" },
	{ 0x2197 , "NORTH EAST ARROW" },
	{ 0x2198 , "SOUTH EAST ARROW" },
	{ 0x2199 , "SOUTH WEST ARROW" },
	{ 0x219A , "LEFTWARDS ARROW WITH STROKE" },
	{ 0x219B , "RIGHTWARDS ARROW WITH STROKE" },
	{ 0x219C , "LEFTWARDS WAVE ARROW" },
	{ 0x219D , "RIGHTWARDS WAVE ARROW" },
	{ 0x219E , "LEFTWARDS TWO HEADED ARROW" },
	{ 0x219F , "UPWARDS TWO HEADED ARROW" },
	{ 0x21A0 , "RIGHTWARDS TWO HEADED ARROW" },
	{ 0x21A1 , "DOWNWARDS TWO HEADED ARROW" },
	{ 0x21A2 , "LEFTWARDS ARROW WITH TAIL" },
	{ 0x21A3 , "RIGHTWARDS ARROW WITH TAIL" },
	{ 0x21A4 , "LEFTWARDS ARROW FROM BAR" },
	{ 0x21A5 , "UPWARDS ARROW FROM BAR" },
	{ 0x21A6 , "RIGHTWARDS ARROW FROM BAR" },
	{ 0x21A7 , "DOWNWARDS ARROW FROM BAR" },
	{ 0x21A8 , "UP DOWN ARROW WITH BASE" },
	{ 0x21A9 , "LEFTWARDS ARROW WITH HOOK" },
	{ 0x21AA , "RIGHTWARDS ARROW WITH HOOK" },
	{ 0x21AB , "LEFTWARDS ARROW WITH LOOP" },
	{ 0x21AC , "RIGHTWARDS ARROW WITH LOOP" },
	{ 0x21AD , "LEFT RIGHT WAVE ARROW" },
	{ 0x21AE , "LEFT RIGHT ARROW WITH STROKE" },
	{ 0x21AF , "DOWNWARDS ZIGZAG ARROW" },
	{ 0x21B0 , "UPWARDS ARROW WITH TIP LEFTWARDS" },
	{ 0x21B1 , "UPWARDS ARROW WITH TIP RIGHTWARDS" },
	{ 0x21B2 , "DOWNWARDS ARROW WITH TIP LEFTWARDS" },
	{ 0x21B3 , "DOWNWARDS ARROW WITH TIP RIGHTWARDS" },
	{ 0x21B4 , "RIGHTWARDS ARROW WITH CORNER DOWNWARDS" },
	{ 0x21B5 , "DOWNWARDS ARROW WITH CORNER LEFTWARDS" },
	{ 0x21B6 , "ANTICLOCKWISE TOP SEMICIRCLE ARROW" },
	{ 0x21B7 , "CLOCKWISE TOP SEMICIRCLE ARROW" },
	{ 0x21B8 , "NORTH WEST ARROW TO LONG BAR" },
	{ 0x21B9 , "LEFTWARDS ARROW TO BAR OVER RIGHTWARDS ARROW TO BAR" },
	{ 0x21BA , "ANTICLOCKWISE OPEN CIRCLE ARROW" },
	{ 0x21BB , "CLOCKWISE OPEN CIRCLE ARROW" },
	{ 0x21BC , "LEFTWARDS HARPOON WITH BARB UPWARDS" },
	{ 0x21BD , "LEFTWARDS HARPOON WITH BARB DOWNWARDS" },
	{ 0x21BE , "UPWARDS HARPOON WITH BARB RIGHTWARDS" },
	{ 0x21BF , "UPWARDS HARPOON WITH BARB LEFTWARDS" },
	{ 0x21C0 , "RIGHTWARDS HARPOON WITH BARB UPWARDS" },
	{ 0x21C1 , "RIGHTWARDS HARPOON WITH BARB DOWNWARDS" },
	{ 0x21C2 , "DOWNWARDS HARPOON WITH BARB RIGHTWARDS" },
	{ 0x21C3 , "DOWNWARDS HARPOON WITH BARB LEFTWARDS" },
	{ 0x21C4 , "RIGHTWARDS ARROW OVER LEFTWARDS ARROW" },
	{ 0x21C5 , "UPWARDS ARROW LEFTWARDS OF DOWNWARDS ARROW" },
	{ 0x21C6 , "LEFTWARDS ARROW OVER RIGHTWARDS ARROW" },
	{ 0x21C7 , "LEFTWARDS PAIRED ARROWS" },
	{ 0x21C8 , "UPWARDS PAIRED ARROWS" },
	{ 0x21C9 , "RIGHTWARDS PAIRED ARROWS" },
	{ 0x21CA , "DOWNWARDS PAIRED ARROWS" },
	{ 0x21CB , "LEFTWARDS HARPOON OVER RIGHTWARDS HARPOON" },
	{ 0x21CC , "RIGHTWARDS HARPOON OVER LEFTWARDS HARPOON" },
	{ 0x21CD , "LEFTWARDS DOUBLE ARROW WITH STROKE" },
	{ 0x21CE , "LEFT RIGHT DOUBLE ARROW WITH STROKE" },
	{ 0x21CF , "RIGHTWARDS DOUBLE ARROW WITH STROKE" },
	{ 0x21D0 , "LEFTWARDS DOUBLE ARROW" },
	{ 0x21D1 , "UPWARDS DOUBLE ARROW" },
	{ 0x21D2 , "RIGHTWARDS DOUBLE ARROW" },
	{ 0x21D3 , "DOWNWARDS DOUBLE ARROW" },
	{ 0x21D4 , "LEFT RIGHT DOUBLE ARROW" },
	{ 0x21D5 , "UP DOWN DOUBLE ARROW" },
	{ 0x21D6 , "NORTH WEST DOUBLE ARROW" },
	{ 0x21D7 , "NORTH EAST DOUBLE ARROW" },
	{ 0x21D8 , "SOUTH EAST DOUBLE ARROW" },
	{ 0x21D9 , "SOUTH WEST DOUBLE ARROW" },
	{ 0x21DA , "LEFTWARDS TRIPLE ARROW" },
	{ 0x21DB , "RIGHTWARDS TRIPLE ARROW" },
	{ 0x21DC , "LEFTWARDS SQUIGGLE ARROW" },
	{ 0x21DD , "RIGHTWARDS SQUIGGLE ARROW" },
	{ 0x21DE , "UPWARDS ARROW WITH DOUBLE STROKE" },
	{ 0x21DF , "DOWNWARDS ARROW WITH DOUBLE STROKE" },
	{ 0x21E0 , "LEFTWARDS DASHED ARROW" },
	{ 0x21E1 , "UPWARDS DASHED ARROW" },
	{ 0x21E2 , "RIGHTWARDS DASHED ARROW" },
	{ 0x21E3 , "DOWNWARDS DASHED ARROW" },
	{ 0x21E4 , "LEFTWARDS ARROW TO BAR" },
	{ 0x21E5 , "RIGHTWARDS ARROW TO BAR" },
	{ 0x21E6 , "LEFTWARDS WHITE ARROW" },
	{ 0x21E7 , "UPWARDS WHITE ARROW" },
	{ 0x21E8 , "RIGHTWARDS WHITE ARROW" },
	{ 0x21E9 , "DOWNWARDS WHITE ARROW" },
	{ 0x21EA , "UPWARDS WHITE ARROW FROM BAR" },
	{ 0x2200 , "FOR ALL" },
	{ 0x2201 , "COMPLEMENT" },
	{ 0x2202 , "PARTIAL DIFFERENTIAL" },
	{ 0x2203 , "THERE EXISTS" },
	{ 0x2204 , "THERE DOES NOT EXIST" },
	{ 0x2205 , "EMPTY SET" },
	{ 0x2206 , "INCREMENT" },
	{ 0x2207 , "NABLA" },
	{ 0x2208 , "ELEMENT OF" },
	{ 0x2209 , "NOT AN ELEMENT OF" },
	{ 0x220A , "SMALL ELEMENT OF" },
	{ 0x220B , "CONTAINS AS MEMBER" },
	{ 0x220C , "DOES NOT CONTAIN AS MEMBER" },
	{ 0x220D , "SMALL CONTAINS AS MEMBER" },
	{ 0x220E , "END OF PROOF" },
	{ 0x220F , "N-ARY PRODUCT" },
	{ 0x2210 , "N-ARY COPRODUCT" },
	{ 0x2211 , "N-ARY SUMMATION" },
	{ 0x2212 , "MINUS SIGN" },
	{ 0x2213 , "MINUS-OR-PLUS SIGN" },
	{ 0x2214 , "DOT PLUS" },
	{ 0x2215 , "DIVISION SLASH" },
	{ 0x2216 , "SET MINUS" },
	{ 0x2217 , "ASTERISK OPERATOR" },
	{ 0x2218 , "RING OPERATOR" },
	{ 0x2219 , "BULLET OPERATOR" },
	{ 0x221A , "SQUARE ROOT" },
	{ 0x221B , "CUBE ROOT" },
	{ 0x221C , "FOURTH ROOT" },
	{ 0x221D , "PROPORTIONAL TO" },
	{ 0x221E , "INFINITY" },
	{ 0x221F , "RIGHT ANGLE" },
	{ 0x2220 , "ANGLE" },
	{ 0x2221 , "MEASURED ANGLE" },
	{ 0x2222 , "SPHERICAL ANGLE" },
	{ 0x2223 , "DIVIDES" },
	{ 0x2224 , "DOES NOT DIVIDE" },
	{ 0x2225 , "PARALLEL TO" },
	{ 0x2226 , "NOT PARALLEL TO" },
	{ 0x2227 , "LOGICAL AND" },
	{ 0x2228 , "LOGICAL OR" },
	{ 0x2229 , "INTERSECTION" },
	{ 0x222A , "UNION" },
	{ 0x222B , "INTEGRAL" },
	{ 0x222C , "DOUBLE INTEGRAL" },
	{ 0x222D , "TRIPLE INTEGRAL" },
	{ 0x222E , "CONTOUR INTEGRAL" },
	{ 0x222F , "SURFACE INTEGRAL" },
	{ 0x2230 , "VOLUME INTEGRAL" },
	{ 0x2231 , "CLOCKWISE INTEGRAL" },
	{ 0x2232 , "CLOCKWISE CONTOUR INTEGRAL" },
	{ 0x2233 , "ANTICLOCKWISE CONTOUR INTEGRAL" },
	{ 0x2234 , "THEREFORE" },
	{ 0x2235 , "BECAUSE" },
	{ 0x2236 , "RATIO" },
	{ 0x2237 , "PROPORTION" },
	{ 0x2238 , "DOT MINUS" },
	{ 0x2239 , "EXCESS" },
	{ 0x223A , "GEOMETRIC PROPORTION" },
	{ 0x223B , "HOMOTHETIC" },
	{ 0x223C , "TILDE OPERATOR" },
	{ 0x223D , "REVERSED TILDE" },
	{ 0x223E , "INVERTED LAZY S" },
	{ 0x223F , "SINE WAVE" },
	{ 0x2240 , "WREATH PRODUCT" },
	{ 0x2241 , "NOT TILDE" },
	{ 0x2242 , "MINUS TILDE" },
	{ 0x2243 , "ASYMPTOTICALLY EQUAL TO" },
	{ 0x2244 , "NOT ASYMPTOTICALLY EQUAL TO" },
	{ 0x2245 , "APPROXIMATELY EQUAL TO" },
	{ 0x2246 , "APPROXIMATELY BUT NOT ACTUALLY EQUAL TO" },
	{ 0x2247 , "NEITHER APPROXIMATELY NOR ACTUALLY EQUAL TO" },
	{ 0x2248 , "ALMOST EQUAL TO" },
	{ 0x2249 , "NOT ALMOST EQUAL TO" },
	{ 0x224A , "ALMOST EQUAL OR EQUAL TO" },
	{ 0x224B , "TRIPLE TILDE" },
	{ 0x224C , "ALL EQUAL TO" },
	{ 0x224D , "EQUIVALENT TO" },
	{ 0x224E , "GEOMETRICALLY EQUIVALENT TO" },
	{ 0x224F , "DIFFERENCE BETWEEN" },
	{ 0x2250 , "APPROACHES THE LIMIT" },
	{ 0x2251 , "GEOMETRICALLY EQUAL TO" },
	{ 0x2252 , "APPROXIMATELY EQUAL TO OR THE IMAGE OF" },
	{ 0x2253 , "IMAGE OF OR APPROXIMATELY EQUAL TO" },
	{ 0x2254 , "COLON EQUALS" },
	{ 0x2255 , "EQUALS COLON" },
	{ 0x2256 , "RING IN EQUAL TO" },
	{ 0x2257 , "RING EQUAL TO" },
	{ 0x2258 , "CORRESPONDS TO" },
	{ 0x2259 , "ESTIMATES" },
	{ 0x225A , "EQUIANGULAR TO" },
	{ 0x225B , "STAR EQUALS" },
	{ 0x225C , "DELTA EQUAL TO" },
	{ 0x225D , "EQUAL TO BY DEFINITION" },
	{ 0x225E , "MEASURED BY" },
	{ 0x225F , "QUESTIONED EQUAL TO" },
	{ 0x2260 , "NOT EQUAL TO" },
	{ 0x2261 , "IDENTICAL TO" },
	{ 0x2262 , "NOT IDENTICAL TO" },
	{ 0x2263 , "STRICTLY EQUIVALENT TO" },
	{ 0x2264 , "LESS-THAN OR EQUAL TO" },
	{ 0x2265 , "GREATER-THAN OR EQUAL TO" },
	{ 0x2266 , "LESS-THAN OVER EQUAL TO" },
	{ 0x2267 , "GREATER-THAN OVER EQUAL TO" },
	{ 0x2268 , "LESS-THAN BUT NOT EQUAL TO" },
	{ 0x2269 , "GREATER-THAN BUT NOT EQUAL TO" },
	{ 0x226A , "MUCH LESS-THAN" },
	{ 0x226B , "MUCH GREATER-THAN" },
	{ 0x226C , "BETWEEN" },
	{ 0x226D , "NOT EQUIVALENT TO" },
	{ 0x226E , "NOT LESS-THAN" },
	{ 0x226F , "NOT GREATER-THAN" },
	{ 0x2270 , "NEITHER LESS-THAN NOR EQUAL TO" },
	{ 0x2271 , "NEITHER GREATER-THAN NOR EQUAL TO" },
	{ 0x2272 , "LESS-THAN OR EQUIVALENT TO" },
	{ 0x2273 , "GREATER-THAN OR EQUIVALENT TO" },
	{ 0x2274 , "NEITHER LESS-THAN NOR EQUIVALENT TO" },
	{ 0x2275 , "NEITHER GREATER-THAN NOR EQUIVALENT TO" },
	{ 0x2276 , "LESS-THAN OR GREATER-THAN" },
	{ 0x2277 , "GREATER-THAN OR LESS-THAN" },
	{ 0x2278 , "NEITHER LESS-THAN NOR GREATER-THAN" },
	{ 0x2279 , "NEITHER GREATER-THAN NOR LESS-THAN" },
	{ 0x227A , "PRECEDES" },
	{ 0x227B , "SUCCEEDS" },
	{ 0x227C , "PRECEDES OR EQUAL TO" },
	{ 0x227D , "SUCCEEDS OR EQUAL TO" },
	{ 0x227E , "PRECEDES OR EQUIVALENT TO" },
	{ 0x227F , "SUCCEEDS OR EQUIVALENT TO" },
	{ 0x2280 , "DOES NOT PRECEDE" },
	{ 0x2281 , "DOES NOT SUCCEED" },
	{ 0x2282 , "SUBSET OF" },
	{ 0x2283 , "SUPERSET OF" },
	{ 0x2284 , "NOT A SUBSET OF" },
	{ 0x2285 , "NOT A SUPERSET OF" },
	{ 0x2286 , "SUBSET OF OR EQUAL TO" },
	{ 0x2287 , "SUPERSET OF OR EQUAL TO" },
	{ 0x2288 , "NEITHER A SUBSET OF NOR EQUAL TO" },
	{ 0x2289 , "NEITHER A SUPERSET OF NOR EQUAL TO" },
	{ 0x228A , "SUBSET OF WITH NOT EQUAL TO" },
	{ 0x228B , "SUPERSET OF WITH NOT EQUAL TO" },
	{ 0x228C , "MULTISET" },
	{ 0x228D , "MULTISET MULTIPLICATION" },
	{ 0x228E , "MULTISET UNION" },
	{ 0x228F , "SQUARE IMAGE OF" },
	{ 0x2290 , "SQUARE ORIGINAL OF" },
	{ 0x2291 , "SQUARE IMAGE OF OR EQUAL TO" },
	{ 0x2292 , "SQUARE ORIGINAL OF OR EQUAL TO" },
	{ 0x2293 , "SQUARE CAP" },
	{ 0x2294 , "SQUARE CUP" },
	{ 0x2295 , "CIRCLED PLUS" },
	{ 0x2296 , "CIRCLED MINUS" },
	{ 0x2297 , "CIRCLED TIMES" },
	{ 0x2298 , "CIRCLED DIVISION SLASH" },
	{ 0x2299 , "CIRCLED DOT OPERATOR" },
	{ 0x229A , "CIRCLED RING OPERATOR" },
	{ 0x229B , "CIRCLED ASTERISK OPERATOR" },
	{ 0x229C , "CIRCLED EQUALS" },
	{ 0x229D , "CIRCLED DASH" },
	{ 0x229E , "SQUARED PLUS" },
	{ 0x229F , "SQUARED MINUS" },
	{ 0x22A0 , "SQUARED TIMES" },
	{ 0x22A1 , "SQUARED DOT OPERATOR" },
	{ 0x22A2 , "RIGHT TACK" },
	{ 0x22A3 , "LEFT TACK" },
	{ 0x22A4 , "DOWN TACK" },
	{ 0x22A5 , "UP TACK" },
	{ 0x22A6 , "ASSERTION" },
	{ 0x22A7 , "MODELS" },
	{ 0x22A8 , "TRUE" },
	{ 0x22A9 , "FORCES" },
	{ 0x22AA , "TRIPLE VERTICAL BAR RIGHT TURNSTILE" },
	{ 0x22AB , "DOUBLE VERTICAL BAR DOUBLE RIGHT TURNSTILE" },
	{ 0x22AC , "DOES NOT PROVE" },
	{ 0x22AD , "NOT TRUE" },
	{ 0x22AE , "DOES NOT FORCE" },
	{ 0x22AF , "NEGATED DOUBLE VERTICAL BAR DOUBLE RIGHT TURNSTILE" },
	{ 0x22B0 , "PRECEDES UNDER RELATION" },
	{ 0x22B1 , "SUCCEEDS UNDER RELATION" },
	{ 0x22B2 , "NORMAL SUBGROUP OF" },
	{ 0x22B3 , "CONTAINS AS NORMAL SUBGROUP" },
	{ 0x22B4 , "NORMAL SUBGROUP OF OR EQUAL TO" },
	{ 0x22B5 , "CONTAINS AS NORMAL SUBGROUP OR EQUAL TO" },
	{ 0x22B6 , "ORIGINAL OF" },
	{ 0x22B7 , "IMAGE OF" },
	{ 0x22B8 , "MULTIMAP" },
	{ 0x22B9 , "HERMITIAN CONJUGATE MATRIX" },
	{ 0x22BA , "INTERCALATE" },
	{ 0x22BB , "XOR" },
	{ 0x22BC , "NAND" },
	{ 0x22BD , "NOR" },
	{ 0x22BE , "RIGHT ANGLE WITH ARC" },
	{ 0x22BF , "RIGHT TRIANGLE" },
	{ 0x22C0 , "N-ARY LOGICAL AND" },
	{ 0x22C1 , "N-ARY LOGICAL OR" },
	{ 0x22C2 , "N-ARY INTERSECTION" },
	{ 0x22C3 , "N-ARY UNION" },
	{ 0x22C4 , "DIAMOND OPERATOR" },
	{ 0x22C5 , "DOT OPERATOR" },
	{ 0x22C6 , "STAR OPERATOR" },
	{ 0x22C7 , "DIVISION TIMES" },
	{ 0x22C8 , "BOWTIE" },
	{ 0x22C9 , "LEFT NORMAL FACTOR SEMIDIRECT PRODUCT" },
	{ 0x22CA , "RIGHT NORMAL FACTOR SEMIDIRECT PRODUCT" },
	{ 0x22CB , "LEFT SEMIDIRECT PRODUCT" },
	{ 0x22CC , "RIGHT SEMIDIRECT PRODUCT" },
	{ 0x22CD , "REVERSED TILDE EQUALS" },
	{ 0x22CE , "CURLY LOGICAL OR" },
	{ 0x22CF , "CURLY LOGICAL AND" },
	{ 0x22D0 , "DOUBLE SUBSET" },
	{ 0x22D1 , "DOUBLE SUPERSET" },
	{ 0x22D2 , "DOUBLE INTERSECTION" },
	{ 0x22D3 , "DOUBLE UNION" },
	{ 0x22D4 , "PITCHFORK" },
	{ 0x22D5 , "EQUAL AND PARALLEL TO" },
	{ 0x22D6 , "LESS-THAN WITH DOT" },
	{ 0x22D7 , "GREATER-THAN WITH DOT" },
	{ 0x22D8 , "VERY MUCH LESS-THAN" },
	{ 0x22D9 , "VERY MUCH GREATER-THAN" },
	{ 0x22DA , "LESS-THAN EQUAL TO OR GREATER-THAN" },
	{ 0x22DB , "GREATER-THAN EQUAL TO OR LESS-THAN" },
	{ 0x22DC , "EQUAL TO OR LESS-THAN" },
	{ 0x22DD , "EQUAL TO OR GREATER-THAN" },
	{ 0x22DE , "EQUAL TO OR PRECEDES" },
	{ 0x22DF , "EQUAL TO OR SUCCEEDS" },
	{ 0x22E0 , "DOES NOT PRECEDE OR EQUAL" },
	{ 0x22E1 , "DOES NOT SUCCEED OR EQUAL" },
	{ 0x22E2 , "NOT SQUARE IMAGE OF OR EQUAL TO" },
	{ 0x22E3 , "NOT SQUARE ORIGINAL OF OR EQUAL TO" },
	{ 0x22E4 , "SQUARE IMAGE OF OR NOT EQUAL TO" },
	{ 0x22E5 , "SQUARE ORIGINAL OF OR NOT EQUAL TO" },
	{ 0x22E6 , "LESS-THAN BUT NOT EQUIVALENT TO" },
	{ 0x22E7 , "GREATER-THAN BUT NOT EQUIVALENT TO" },
	{ 0x22E8 , "PRECEDES BUT NOT EQUIVALENT TO" },
	{ 0x22E9 , "SUCCEEDS BUT NOT EQUIVALENT TO" },
	{ 0x22EA , "NOT NORMAL SUBGROUP OF" },
	{ 0x22EB , "DOES NOT CONTAIN AS NORMAL SUBGROUP" },
	{ 0x22EC , "NOT NORMAL SUBGROUP OF OR EQUAL TO" },
	{ 0x22ED , "DOES NOT CONTAIN AS NORMAL SUBGROUP OR EQUAL" },
	{ 0x22EE , "VERTICAL ELLIPSIS" },
	{ 0x22EF , "MIDLINE HORIZONTAL ELLIPSIS" },
	{ 0x22F0 , "UP RIGHT DIAGONAL ELLIPSIS" },
	{ 0x22F1 , "DOWN RIGHT DIAGONAL ELLIPSIS" },
	{ 0x2300 , "DIAMETER SIGN" },
	{ 0x2302 , "HOUSE" },
	{ 0x2303 , "UP ARROWHEAD" },
	{ 0x2304 , "DOWN ARROWHEAD" },
	{ 0x2305 , "PROJECTIVE" },
	{ 0x2306 , "PERSPECTIVE" },
	{ 0x2307 , "WAVY LINE" },
	{ 0x2308 , "LEFT CEILING" },
	{ 0x2309 , "RIGHT CEILING" },
	{ 0x230A , "LEFT FLOOR" },
	{ 0x230B , "RIGHT FLOOR" },
	{ 0x230C , "BOTTOM RIGHT CROP" },
	{ 0x230D , "BOTTOM LEFT CROP" },
	{ 0x230E , "TOP RIGHT CROP" },
	{ 0x230F , "TOP LEFT CROP" },
	{ 0x2310 , "REVERSED NOT SIGN" },
	{ 0x2311 , "SQUARE LOZENGE" },
	{ 0x2312 , "ARC" },
	{ 0x2313 , "SEGMENT" },
	{ 0x2314 , "SECTOR" },
	{ 0x2315 , "TELEPHONE RECORDER" },
	{ 0x2316 , "POSITION INDICATOR" },
	{ 0x2317 , "VIEWDATA SQUARE" },
	{ 0x2318 , "PLACE OF INTEREST SIGN" },
	{ 0x2319 , "TURNED NOT SIGN" },
	{ 0x231A , "WATCH" },
	{ 0x231B , "HOURGLASS" },
	{ 0x231C , "TOP LEFT CORNER" },
	{ 0x231D , "TOP RIGHT CORNER" },
	{ 0x231E , "BOTTOM LEFT CORNER" },
	{ 0x231F , "BOTTOM RIGHT CORNER" },
	{ 0x2320 , "TOP HALF INTEGRAL" },
	{ 0x2321 , "BOTTOM HALF INTEGRAL" },
	{ 0x2322 , "FROWN" },
	{ 0x2323 , "SMILE" },
	{ 0x2324 , "UP ARROWHEAD BETWEEN TWO HORIZONTAL BARS" },
	{ 0x2325 , "OPTION KEY" },
	{ 0x2326 , "ERASE TO THE RIGHT" },
	{ 0x2327 , "X IN A RECTANGLE BOX" },
	{ 0x2328 , "KEYBOARD" },
	{ 0x2329 , "LEFT-POINTING ANGLE BRACKET" },
	{ 0x232A , "RIGHT-POINTING ANGLE BRACKET" },
	{ 0x232B , "ERASE TO THE LEFT" },
	{ 0x232C , "BENZENE RING" },
	{ 0x232D , "CYLINDRICITY" },
	{ 0x232E , "ALL AROUND-PROFILE" },
	{ 0x232F , "SYMMETRY" },
	{ 0x2330 , "TOTAL RUNOUT" },
	{ 0x2331 , "DIMENSION ORIGIN" },
	{ 0x2332 , "CONICAL TAPER" },
	{ 0x2333 , "SLOPE" },
	{ 0x2334 , "COUNTERBORE" },
	{ 0x2335 , "COUNTERSINK" },
	{ 0x2336 , "APL FUNCTIONAL SYMBOL I-BEAM" },
	{ 0x2337 , "APL FUNCTIONAL SYMBOL SQUISH QUAD" },
	{ 0x2338 , "APL FUNCTIONAL SYMBOL QUAD EQUAL" },
	{ 0x2339 , "APL FUNCTIONAL SYMBOL QUAD DIVIDE" },
	{ 0x233A , "APL FUNCTIONAL SYMBOL QUAD DIAMOND" },
	{ 0x233B , "APL FUNCTIONAL SYMBOL QUAD JOT" },
	{ 0x233C , "APL FUNCTIONAL SYMBOL QUAD CIRCLE" },
	{ 0x233D , "APL FUNCTIONAL SYMBOL CIRCLE STILE" },
	{ 0x233E , "APL FUNCTIONAL SYMBOL CIRCLE JOT" },
	{ 0x233F , "APL FUNCTIONAL SYMBOL SLASH BAR" },
	{ 0x2340 , "APL FUNCTIONAL SYMBOL BACKSLASH BAR" },
	{ 0x2341 , "APL FUNCTIONAL SYMBOL QUAD SLASH" },
	{ 0x2342 , "APL FUNCTIONAL SYMBOL QUAD BACKSLASH" },
	{ 0x2343 , "APL FUNCTIONAL SYMBOL QUAD LESS-THAN" },
	{ 0x2344 , "APL FUNCTIONAL SYMBOL QUAD GREATER-THAN" },
	{ 0x2345 , "APL FUNCTIONAL SYMBOL LEFTWARDS VANE" },
	{ 0x2346 , "APL FUNCTIONAL SYMBOL RIGHTWARDS VANE" },
	{ 0x2347 , "APL FUNCTIONAL SYMBOL QUAD LEFTWARDS ARROW" },
	{ 0x2348 , "APL FUNCTIONAL SYMBOL QUAD RIGHTWARDS ARROW" },
	{ 0x2349 , "APL FUNCTIONAL SYMBOL CIRCLE BACKSLASH" },
	{ 0x234A , "APL FUNCTIONAL SYMBOL DOWN TACK UNDERBAR" },
	{ 0x234B , "APL FUNCTIONAL SYMBOL DELTA STILE" },
	{ 0x234C , "APL FUNCTIONAL SYMBOL QUAD DOWN CARET" },
	{ 0x234D , "APL FUNCTIONAL SYMBOL QUAD DELTA" },
	{ 0x234E , "APL FUNCTIONAL SYMBOL DOWN TACK JOT" },
	{ 0x234F , "APL FUNCTIONAL SYMBOL UPWARDS VANE" },
	{ 0x2350 , "APL FUNCTIONAL SYMBOL QUAD UPWARDS ARROW" },
	{ 0x2351 , "APL FUNCTIONAL SYMBOL UP TACK OVERBAR" },
	{ 0x2352 , "APL FUNCTIONAL SYMBOL DEL STILE" },
	{ 0x2353 , "APL FUNCTIONAL SYMBOL QUAD UP CARET" },
	{ 0x2354 , "APL FUNCTIONAL SYMBOL QUAD DEL" },
	{ 0x2355 , "APL FUNCTIONAL SYMBOL UP TACK JOT" },
	{ 0x2356 , "APL FUNCTIONAL SYMBOL DOWNWARDS VANE" },
	{ 0x2357 , "APL FUNCTIONAL SYMBOL QUAD DOWNWARDS ARROW" },
	{ 0x2358 , "APL FUNCTIONAL SYMBOL QUOTE UNDERBAR" },
	{ 0x2359 , "APL FUNCTIONAL SYMBOL DELTA UNDERBAR" },
	{ 0x235A , "APL FUNCTIONAL SYMBOL DIAMOND UNDERBAR" },
	{ 0x235B , "APL FUNCTIONAL SYMBOL JOT UNDERBAR" },
	{ 0x235C , "APL FUNCTIONAL SYMBOL CIRCLE UNDERBAR" },
	{ 0x235D , "APL FUNCTIONAL SYMBOL UP SHOE JOT" },
	{ 0x235E , "APL FUNCTIONAL SYMBOL QUOTE QUAD" },
	{ 0x235F , "APL FUNCTIONAL SYMBOL CIRCLE STAR" },
	{ 0x2360 , "APL FUNCTIONAL SYMBOL QUAD COLON" },
	{ 0x2361 , "APL FUNCTIONAL SYMBOL UP TACK DIAERESIS" },
	{ 0x2362 , "APL FUNCTIONAL SYMBOL DEL DIAERESIS" },
	{ 0x2363 , "APL FUNCTIONAL SYMBOL STAR DIAERESIS" },
	{ 0x2364 , "APL FUNCTIONAL SYMBOL JOT DIAERESIS" },
	{ 0x2365 , "APL FUNCTIONAL SYMBOL CIRCLE DIAERESIS" },
	{ 0x2366 , "APL FUNCTIONAL SYMBOL DOWN SHOE STILE" },
	{ 0x2367 , "APL FUNCTIONAL SYMBOL LEFT SHOE STILE" },
	{ 0x2368 , "APL FUNCTIONAL SYMBOL TILDE DIAERESIS" },
	{ 0x2369 , "APL FUNCTIONAL SYMBOL GREATER-THAN DIAERESIS" },
	{ 0x236A , "APL FUNCTIONAL SYMBOL COMMA BAR" },
	{ 0x236B , "APL FUNCTIONAL SYMBOL DEL TILDE" },
	{ 0x236C , "APL FUNCTIONAL SYMBOL ZILDE" },
	{ 0x236D , "APL FUNCTIONAL SYMBOL STILE TILDE" },
	{ 0x236E , "APL FUNCTIONAL SYMBOL SEMICOLON UNDERBAR" },
	{ 0x236F , "APL FUNCTIONAL SYMBOL QUAD NOT EQUAL" },
	{ 0x2370 , "APL FUNCTIONAL SYMBOL QUAD QUESTION" },
	{ 0x2371 , "APL FUNCTIONAL SYMBOL DOWN CARET TILDE" },
	{ 0x2372 , "APL FUNCTIONAL SYMBOL UP CARET TILDE" },
	{ 0x2373 , "APL FUNCTIONAL SYMBOL IOTA" },
	{ 0x2374 , "APL FUNCTIONAL SYMBOL RHO" },
	{ 0x2375 , "APL FUNCTIONAL SYMBOL OMEGA" },
	{ 0x2376 , "APL FUNCTIONAL SYMBOL ALPHA UNDERBAR" },
	{ 0x2377 , "APL FUNCTIONAL SYMBOL EPSILON UNDERBAR" },
	{ 0x2378 , "APL FUNCTIONAL SYMBOL IOTA UNDERBAR" },
	{ 0x2379 , "APL FUNCTIONAL SYMBOL OMEGA UNDERBAR" },
	{ 0x237A , "APL FUNCTIONAL SYMBOL ALPHA" },
	{ 0x2400 , "SYMBOL FOR NULL" },
	{ 0x2401 , "SYMBOL FOR START OF HEADING" },
	{ 0x2402 , "SYMBOL FOR START OF TEXT" },
	{ 0x2403 , "SYMBOL FOR END OF TEXT" },
	{ 0x2404 , "SYMBOL FOR END OF TRANSMISSION" },
	{ 0x2405 , "SYMBOL FOR ENQUIRY" },
	{ 0x2406 , "SYMBOL FOR ACKNOWLEDGE" },
	{ 0x2407 , "SYMBOL FOR BELL" },
	{ 0x2408 , "SYMBOL FOR BACKSPACE" },
	{ 0x2409 , "SYMBOL FOR HORIZONTAL TABULATION" },
	{ 0x240A , "SYMBOL FOR LINE FEED" },
	{ 0x240B , "SYMBOL FOR VERTICAL TABULATION" },
	{ 0x240C , "SYMBOL FOR FORM FEED" },
	{ 0x240D , "SYMBOL FOR CARRIAGE RETURN" },
	{ 0x240E , "SYMBOL FOR SHIFT OUT" },
	{ 0x240F , "SYMBOL FOR SHIFT IN" },
	{ 0x2410 , "SYMBOL FOR DATA LINK ESCAPE" },
	{ 0x2411 , "SYMBOL FOR DEVICE CONTROL ONE" },
	{ 0x2412 , "SYMBOL FOR DEVICE CONTROL TWO" },
	{ 0x2413 , "SYMBOL FOR DEVICE CONTROL THREE" },
	{ 0x2414 , "SYMBOL FOR DEVICE CONTROL FOUR" },
	{ 0x2415 , "SYMBOL FOR NEGATIVE ACKNOWLEDGE" },
	{ 0x2416 , "SYMBOL FOR SYNCHRONOUS IDLE" },
	{ 0x2417 , "SYMBOL FOR END OF TRANSMISSION BLOCK" },
	{ 0x2418 , "SYMBOL FOR CANCEL" },
	{ 0x2419 , "SYMBOL FOR END OF MEDIUM" },
	{ 0x241A , "SYMBOL FOR SUBSTITUTE" },
	{ 0x241B , "SYMBOL FOR ESCAPE" },
	{ 0x241C , "SYMBOL FOR FILE SEPARATOR" },
	{ 0x241D , "SYMBOL FOR GROUP SEPARATOR" },
	{ 0x241E , "SYMBOL FOR RECORD SEPARATOR" },
	{ 0x241F , "SYMBOL FOR UNIT SEPARATOR" },
	{ 0x2420 , "SYMBOL FOR SPACE" },
	{ 0x2421 , "SYMBOL FOR DELETE" },
	{ 0x2422 , "BLANK SYMBOL" },
	{ 0x2423 , "OPEN BOX" },
	{ 0x2424 , "SYMBOL FOR NEWLINE" },
#endif
	{ 0x2440 , "OCR HOOK" },
	{ 0x2441 , "OCR CHAIR" },
	{ 0x2442 , "OCR FORK" },
	{ 0x2443 , "OCR INVERTED FORK" },
	{ 0x2444 , "OCR BELT BUCKLE" },
	{ 0x2445 , "OCR BOW TIE" },
	{ 0x2446 , "MICR 'c'" },
	{ 0x2447 , "MICR 'i'" },
	{ 0x2448 , "MICR 'r'" },
	{ 0x2449 , "MICR 'm'" },
	{ 0x244A , "OCR DOUBLE BACKSLASH" },
#if 0
	{ 0x2460 , "CIRCLED DIGIT ONE" },
	{ 0x2461 , "CIRCLED DIGIT TWO" },
	{ 0x2462 , "CIRCLED DIGIT THREE" },
	{ 0x2463 , "CIRCLED DIGIT FOUR" },
	{ 0x2464 , "CIRCLED DIGIT FIVE" },
	{ 0x2465 , "CIRCLED DIGIT SIX" },
	{ 0x2466 , "CIRCLED DIGIT SEVEN" },
	{ 0x2467 , "CIRCLED DIGIT EIGHT" },
	{ 0x2468 , "CIRCLED DIGIT NINE" },
	{ 0x2469 , "CIRCLED NUMBER TEN" },
	{ 0x246A , "CIRCLED NUMBER ELEVEN" },
	{ 0x246B , "CIRCLED NUMBER TWELVE" },
	{ 0x246C , "CIRCLED NUMBER THIRTEEN" },
	{ 0x246D , "CIRCLED NUMBER FOURTEEN" },
	{ 0x246E , "CIRCLED NUMBER FIFTEEN" },
	{ 0x246F , "CIRCLED NUMBER SIXTEEN" },
	{ 0x2470 , "CIRCLED NUMBER SEVENTEEN" },
	{ 0x2471 , "CIRCLED NUMBER EIGHTEEN" },
	{ 0x2472 , "CIRCLED NUMBER NINETEEN" },
	{ 0x2473 , "CIRCLED NUMBER TWENTY" },
	{ 0x2474 , "PARENTHESIZED DIGIT ONE" },
	{ 0x2475 , "PARENTHESIZED DIGIT TWO" },
	{ 0x2476 , "PARENTHESIZED DIGIT THREE" },
	{ 0x2477 , "PARENTHESIZED DIGIT FOUR" },
	{ 0x2478 , "PARENTHESIZED DIGIT FIVE" },
	{ 0x2479 , "PARENTHESIZED DIGIT SIX" },
	{ 0x247A , "PARENTHESIZED DIGIT SEVEN" },
	{ 0x247B , "PARENTHESIZED DIGIT EIGHT" },
	{ 0x247C , "PARENTHESIZED DIGIT NINE" },
	{ 0x247D , "PARENTHESIZED NUMBER TEN" },
	{ 0x247E , "PARENTHESIZED NUMBER ELEVEN" },
	{ 0x247F , "PARENTHESIZED NUMBER TWELVE" },
	{ 0x2480 , "PARENTHESIZED NUMBER THIRTEEN" },
	{ 0x2481 , "PARENTHESIZED NUMBER FOURTEEN" },
	{ 0x2482 , "PARENTHESIZED NUMBER FIFTEEN" },
	{ 0x2483 , "PARENTHESIZED NUMBER SIXTEEN" },
	{ 0x2484 , "PARENTHESIZED NUMBER SEVENTEEN" },
	{ 0x2485 , "PARENTHESIZED NUMBER EIGHTEEN" },
	{ 0x2486 , "PARENTHESIZED NUMBER NINETEEN" },
	{ 0x2487 , "PARENTHESIZED NUMBER TWENTY" },
	{ 0x2488 , "DIGIT ONE FULL STOP" },
	{ 0x2489 , "DIGIT TWO FULL STOP" },
	{ 0x248A , "DIGIT THREE FULL STOP" },
	{ 0x248B , "DIGIT FOUR FULL STOP" },
	{ 0x248C , "DIGIT FIVE FULL STOP" },
	{ 0x248D , "DIGIT SIX FULL STOP" },
	{ 0x248E , "DIGIT SEVEN FULL STOP" },
	{ 0x248F , "DIGIT EIGHT FULL STOP" },
	{ 0x2490 , "DIGIT NINE FULL STOP" },
	{ 0x2491 , "NUMBER TEN FULL STOP" },
	{ 0x2492 , "NUMBER ELEVEN FULL STOP" },
	{ 0x2493 , "NUMBER TWELVE FULL STOP" },
	{ 0x2494 , "NUMBER THIRTEEN FULL STOP" },
	{ 0x2495 , "NUMBER FOURTEEN FULL STOP" },
	{ 0x2496 , "NUMBER FIFTEEN FULL STOP" },
	{ 0x2497 , "NUMBER SIXTEEN FULL STOP" },
	{ 0x2498 , "NUMBER SEVENTEEN FULL STOP" },
	{ 0x2499 , "NUMBER EIGHTEEN FULL STOP" },
	{ 0x249A , "NUMBER NINETEEN FULL STOP" },
	{ 0x249B , "NUMBER TWENTY FULL STOP" },
	{ 0x249C , "PARENTHESIZED LATIN SMALL LETTER A" },
	{ 0x249D , "PARENTHESIZED LATIN SMALL LETTER B" },
	{ 0x249E , "PARENTHESIZED LATIN SMALL LETTER C" },
	{ 0x249F , "PARENTHESIZED LATIN SMALL LETTER D" },
	{ 0x24A0 , "PARENTHESIZED LATIN SMALL LETTER E" },
	{ 0x24A1 , "PARENTHESIZED LATIN SMALL LETTER F" },
	{ 0x24A2 , "PARENTHESIZED LATIN SMALL LETTER G" },
	{ 0x24A3 , "PARENTHESIZED LATIN SMALL LETTER H" },
	{ 0x24A4 , "PARENTHESIZED LATIN SMALL LETTER I" },
	{ 0x24A5 , "PARENTHESIZED LATIN SMALL LETTER J" },
	{ 0x24A6 , "PARENTHESIZED LATIN SMALL LETTER K" },
	{ 0x24A7 , "PARENTHESIZED LATIN SMALL LETTER L" },
	{ 0x24A8 , "PARENTHESIZED LATIN SMALL LETTER M" },
	{ 0x24A9 , "PARENTHESIZED LATIN SMALL LETTER N" },
	{ 0x24AA , "PARENTHESIZED LATIN SMALL LETTER O" },
	{ 0x24AB , "PARENTHESIZED LATIN SMALL LETTER P" },
	{ 0x24AC , "PARENTHESIZED LATIN SMALL LETTER Q" },
	{ 0x24AD , "PARENTHESIZED LATIN SMALL LETTER R" },
	{ 0x24AE , "PARENTHESIZED LATIN SMALL LETTER S" },
	{ 0x24AF , "PARENTHESIZED LATIN SMALL LETTER T" },
	{ 0x24B0 , "PARENTHESIZED LATIN SMALL LETTER U" },
	{ 0x24B1 , "PARENTHESIZED LATIN SMALL LETTER V" },
	{ 0x24B2 , "PARENTHESIZED LATIN SMALL LETTER W" },
	{ 0x24B3 , "PARENTHESIZED LATIN SMALL LETTER X" },
	{ 0x24B4 , "PARENTHESIZED LATIN SMALL LETTER Y" },
	{ 0x24B5 , "PARENTHESIZED LATIN SMALL LETTER Z" },
	{ 0x24B6 , "CIRCLED LATIN CAPITAL LETTER A" },
	{ 0x24B7 , "CIRCLED LATIN CAPITAL LETTER B" },
	{ 0x24B8 , "CIRCLED LATIN CAPITAL LETTER C" },
	{ 0x24B9 , "CIRCLED LATIN CAPITAL LETTER D" },
	{ 0x24BA , "CIRCLED LATIN CAPITAL LETTER E" },
	{ 0x24BB , "CIRCLED LATIN CAPITAL LETTER F" },
	{ 0x24BC , "CIRCLED LATIN CAPITAL LETTER G" },
	{ 0x24BD , "CIRCLED LATIN CAPITAL LETTER H" },
	{ 0x24BE , "CIRCLED LATIN CAPITAL LETTER I" },
	{ 0x24BF , "CIRCLED LATIN CAPITAL LETTER J" },
	{ 0x24C0 , "CIRCLED LATIN CAPITAL LETTER K" },
	{ 0x24C1 , "CIRCLED LATIN CAPITAL LETTER L" },
	{ 0x24C2 , "CIRCLED LATIN CAPITAL LETTER M" },
	{ 0x24C3 , "CIRCLED LATIN CAPITAL LETTER N" },
	{ 0x24C4 , "CIRCLED LATIN CAPITAL LETTER O" },
	{ 0x24C5 , "CIRCLED LATIN CAPITAL LETTER P" },
	{ 0x24C6 , "CIRCLED LATIN CAPITAL LETTER Q" },
	{ 0x24C7 , "CIRCLED LATIN CAPITAL LETTER R" },
	{ 0x24C8 , "CIRCLED LATIN CAPITAL LETTER S" },
	{ 0x24C9 , "CIRCLED LATIN CAPITAL LETTER T" },
	{ 0x24CA , "CIRCLED LATIN CAPITAL LETTER U" },
	{ 0x24CB , "CIRCLED LATIN CAPITAL LETTER V" },
	{ 0x24CC , "CIRCLED LATIN CAPITAL LETTER W" },
	{ 0x24CD , "CIRCLED LATIN CAPITAL LETTER X" },
	{ 0x24CE , "CIRCLED LATIN CAPITAL LETTER Y" },
	{ 0x24CF , "CIRCLED LATIN CAPITAL LETTER Z" },
	{ 0x24D0 , "CIRCLED LATIN SMALL LETTER A" },
	{ 0x24D1 , "CIRCLED LATIN SMALL LETTER B" },
	{ 0x24D2 , "CIRCLED LATIN SMALL LETTER C" },
	{ 0x24D3 , "CIRCLED LATIN SMALL LETTER D" },
	{ 0x24D4 , "CIRCLED LATIN SMALL LETTER E" },
	{ 0x24D5 , "CIRCLED LATIN SMALL LETTER F" },
	{ 0x24D6 , "CIRCLED LATIN SMALL LETTER G" },
	{ 0x24D7 , "CIRCLED LATIN SMALL LETTER H" },
	{ 0x24D8 , "CIRCLED LATIN SMALL LETTER I" },
	{ 0x24D9 , "CIRCLED LATIN SMALL LETTER J" },
	{ 0x24DA , "CIRCLED LATIN SMALL LETTER K" },
	{ 0x24DB , "CIRCLED LATIN SMALL LETTER L" },
	{ 0x24DC , "CIRCLED LATIN SMALL LETTER M" },
	{ 0x24DD , "CIRCLED LATIN SMALL LETTER N" },
	{ 0x24DE , "CIRCLED LATIN SMALL LETTER O" },
	{ 0x24DF , "CIRCLED LATIN SMALL LETTER P" },
	{ 0x24E0 , "CIRCLED LATIN SMALL LETTER Q" },
	{ 0x24E1 , "CIRCLED LATIN SMALL LETTER R" },
	{ 0x24E2 , "CIRCLED LATIN SMALL LETTER S" },
	{ 0x24E3 , "CIRCLED LATIN SMALL LETTER T" },
	{ 0x24E4 , "CIRCLED LATIN SMALL LETTER U" },
	{ 0x24E5 , "CIRCLED LATIN SMALL LETTER V" },
	{ 0x24E6 , "CIRCLED LATIN SMALL LETTER W" },
	{ 0x24E7 , "CIRCLED LATIN SMALL LETTER X" },
	{ 0x24E8 , "CIRCLED LATIN SMALL LETTER Y" },
	{ 0x24E9 , "CIRCLED LATIN SMALL LETTER Z" },
	{ 0x24EA , "CIRCLED DIGIT ZERO" },
	{ 0x2500 , "BOX DRAWINGS LIGHT HORIZONTAL" },
	{ 0x2501 , "BOX DRAWINGS HEAVY HORIZONTAL" },
	{ 0x2502 , "BOX DRAWINGS LIGHT VERTICAL" },
	{ 0x2503 , "BOX DRAWINGS HEAVY VERTICAL" },
	{ 0x2504 , "BOX DRAWINGS LIGHT TRIPLE DASH HORIZONTAL" },
	{ 0x2505 , "BOX DRAWINGS HEAVY TRIPLE DASH HORIZONTAL" },
	{ 0x2506 , "BOX DRAWINGS LIGHT TRIPLE DASH VERTICAL" },
	{ 0x2507 , "BOX DRAWINGS HEAVY TRIPLE DASH VERTICAL" },
	{ 0x2508 , "BOX DRAWINGS LIGHT QUADRUPLE DASH HORIZONTAL" },
	{ 0x2509 , "BOX DRAWINGS HEAVY QUADRUPLE DASH HORIZONTAL" },
	{ 0x250A , "BOX DRAWINGS LIGHT QUADRUPLE DASH VERTICAL" },
	{ 0x250B , "BOX DRAWINGS HEAVY QUADRUPLE DASH VERTICAL" },
	{ 0x250C , "BOX DRAWINGS LIGHT DOWN AND RIGHT" },
	{ 0x250D , "BOX DRAWINGS DOWN LIGHT AND RIGHT HEAVY" },
	{ 0x250E , "BOX DRAWINGS DOWN HEAVY AND RIGHT LIGHT" },
	{ 0x250F , "BOX DRAWINGS HEAVY DOWN AND RIGHT" },
	{ 0x2510 , "BOX DRAWINGS LIGHT DOWN AND LEFT" },
	{ 0x2511 , "BOX DRAWINGS DOWN LIGHT AND LEFT HEAVY" },
	{ 0x2512 , "BOX DRAWINGS DOWN HEAVY AND LEFT LIGHT" },
	{ 0x2513 , "BOX DRAWINGS HEAVY DOWN AND LEFT" },
	{ 0x2514 , "BOX DRAWINGS LIGHT UP AND RIGHT" },
	{ 0x2515 , "BOX DRAWINGS UP LIGHT AND RIGHT HEAVY" },
	{ 0x2516 , "BOX DRAWINGS UP HEAVY AND RIGHT LIGHT" },
	{ 0x2517 , "BOX DRAWINGS HEAVY UP AND RIGHT" },
	{ 0x2518 , "BOX DRAWINGS LIGHT UP AND LEFT" },
	{ 0x2519 , "BOX DRAWINGS UP LIGHT AND LEFT HEAVY" },
	{ 0x251A , "BOX DRAWINGS UP HEAVY AND LEFT LIGHT" },
	{ 0x251B , "BOX DRAWINGS HEAVY UP AND LEFT" },
	{ 0x251C , "BOX DRAWINGS LIGHT VERTICAL AND RIGHT" },
	{ 0x251D , "BOX DRAWINGS VERTICAL LIGHT AND RIGHT HEAVY" },
	{ 0x251E , "BOX DRAWINGS UP HEAVY AND RIGHT DOWN LIGHT" },
	{ 0x251F , "BOX DRAWINGS DOWN HEAVY AND RIGHT UP LIGHT" },
	{ 0x2520 , "BOX DRAWINGS VERTICAL HEAVY AND RIGHT LIGHT" },
	{ 0x2521 , "BOX DRAWINGS DOWN LIGHT AND RIGHT UP HEAVY" },
	{ 0x2522 , "BOX DRAWINGS UP LIGHT AND RIGHT DOWN HEAVY" },
	{ 0x2523 , "BOX DRAWINGS HEAVY VERTICAL AND RIGHT" },
	{ 0x2524 , "BOX DRAWINGS LIGHT VERTICAL AND LEFT" },
	{ 0x2525 , "BOX DRAWINGS VERTICAL LIGHT AND LEFT HEAVY" },
	{ 0x2526 , "BOX DRAWINGS UP HEAVY AND LEFT DOWN LIGHT" },
	{ 0x2527 , "BOX DRAWINGS DOWN HEAVY AND LEFT UP LIGHT" },
	{ 0x2528 , "BOX DRAWINGS VERTICAL HEAVY AND LEFT LIGHT" },
	{ 0x2529 , "BOX DRAWINGS DOWN LIGHT AND LEFT UP HEAVY" },
	{ 0x252A , "BOX DRAWINGS UP LIGHT AND LEFT DOWN HEAVY" },
	{ 0x252B , "BOX DRAWINGS HEAVY VERTICAL AND LEFT" },
	{ 0x252C , "BOX DRAWINGS LIGHT DOWN AND HORIZONTAL" },
	{ 0x252D , "BOX DRAWINGS LEFT HEAVY AND RIGHT DOWN LIGHT" },
	{ 0x252E , "BOX DRAWINGS RIGHT HEAVY AND LEFT DOWN LIGHT" },
	{ 0x252F , "BOX DRAWINGS DOWN LIGHT AND HORIZONTAL HEAVY" },
	{ 0x2530 , "BOX DRAWINGS DOWN HEAVY AND HORIZONTAL LIGHT" },
	{ 0x2531 , "BOX DRAWINGS RIGHT LIGHT AND LEFT DOWN HEAVY" },
	{ 0x2532 , "BOX DRAWINGS LEFT LIGHT AND RIGHT DOWN HEAVY" },
	{ 0x2533 , "BOX DRAWINGS HEAVY DOWN AND HORIZONTAL" },
	{ 0x2534 , "BOX DRAWINGS LIGHT UP AND HORIZONTAL" },
	{ 0x2535 , "BOX DRAWINGS LEFT HEAVY AND RIGHT UP LIGHT" },
	{ 0x2536 , "BOX DRAWINGS RIGHT HEAVY AND LEFT UP LIGHT" },
	{ 0x2537 , "BOX DRAWINGS UP LIGHT AND HORIZONTAL HEAVY" },
	{ 0x2538 , "BOX DRAWINGS UP HEAVY AND HORIZONTAL LIGHT" },
	{ 0x2539 , "BOX DRAWINGS RIGHT LIGHT AND LEFT UP HEAVY" },
	{ 0x253A , "BOX DRAWINGS LEFT LIGHT AND RIGHT UP HEAVY" },
	{ 0x253B , "BOX DRAWINGS HEAVY UP AND HORIZONTAL" },
	{ 0x253C , "BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL" },
	{ 0x253D , "BOX DRAWINGS LEFT HEAVY AND RIGHT VERTICAL LIGHT" },
	{ 0x253E , "BOX DRAWINGS RIGHT HEAVY AND LEFT VERTICAL LIGHT" },
	{ 0x253F , "BOX DRAWINGS VERTICAL LIGHT AND HORIZONTAL HEAVY" },
	{ 0x2540 , "BOX DRAWINGS UP HEAVY AND DOWN HORIZONTAL LIGHT" },
	{ 0x2541 , "BOX DRAWINGS DOWN HEAVY AND UP HORIZONTAL LIGHT" },
	{ 0x2542 , "BOX DRAWINGS VERTICAL HEAVY AND HORIZONTAL LIGHT" },
	{ 0x2543 , "BOX DRAWINGS LEFT UP HEAVY AND RIGHT DOWN LIGHT" },
	{ 0x2544 , "BOX DRAWINGS RIGHT UP HEAVY AND LEFT DOWN LIGHT" },
	{ 0x2545 , "BOX DRAWINGS LEFT DOWN HEAVY AND RIGHT UP LIGHT" },
	{ 0x2546 , "BOX DRAWINGS RIGHT DOWN HEAVY AND LEFT UP LIGHT" },
	{ 0x2547 , "BOX DRAWINGS DOWN LIGHT AND UP HORIZONTAL HEAVY" },
	{ 0x2548 , "BOX DRAWINGS UP LIGHT AND DOWN HORIZONTAL HEAVY" },
	{ 0x2549 , "BOX DRAWINGS RIGHT LIGHT AND LEFT VERTICAL HEAVY" },
	{ 0x254A , "BOX DRAWINGS LEFT LIGHT AND RIGHT VERTICAL HEAVY" },
	{ 0x254B , "BOX DRAWINGS HEAVY VERTICAL AND HORIZONTAL" },
	{ 0x254C , "BOX DRAWINGS LIGHT DOUBLE DASH HORIZONTAL" },
	{ 0x254D , "BOX DRAWINGS HEAVY DOUBLE DASH HORIZONTAL" },
	{ 0x254E , "BOX DRAWINGS LIGHT DOUBLE DASH VERTICAL" },
	{ 0x254F , "BOX DRAWINGS HEAVY DOUBLE DASH VERTICAL" },
	{ 0x2550 , "BOX DRAWINGS DOUBLE HORIZONTAL" },
	{ 0x2551 , "BOX DRAWINGS DOUBLE VERTICAL" },
	{ 0x2552 , "BOX DRAWINGS DOWN SINGLE AND RIGHT DOUBLE" },
	{ 0x2553 , "BOX DRAWINGS DOWN DOUBLE AND RIGHT SINGLE" },
	{ 0x2554 , "BOX DRAWINGS DOUBLE DOWN AND RIGHT" },
	{ 0x2555 , "BOX DRAWINGS DOWN SINGLE AND LEFT DOUBLE" },
	{ 0x2556 , "BOX DRAWINGS DOWN DOUBLE AND LEFT SINGLE" },
	{ 0x2557 , "BOX DRAWINGS DOUBLE DOWN AND LEFT" },
	{ 0x2558 , "BOX DRAWINGS UP SINGLE AND RIGHT DOUBLE" },
	{ 0x2559 , "BOX DRAWINGS UP DOUBLE AND RIGHT SINGLE" },
	{ 0x255A , "BOX DRAWINGS DOUBLE UP AND RIGHT" },
	{ 0x255B , "BOX DRAWINGS UP SINGLE AND LEFT DOUBLE" },
	{ 0x255C , "BOX DRAWINGS UP DOUBLE AND LEFT SINGLE" },
	{ 0x255D , "BOX DRAWINGS DOUBLE UP AND LEFT" },
	{ 0x255E , "BOX DRAWINGS VERTICAL SINGLE AND RIGHT DOUBLE" },
	{ 0x255F , "BOX DRAWINGS VERTICAL DOUBLE AND RIGHT SINGLE" },
	{ 0x2560 , "BOX DRAWINGS DOUBLE VERTICAL AND RIGHT" },
	{ 0x2561 , "BOX DRAWINGS VERTICAL SINGLE AND LEFT DOUBLE" },
	{ 0x2562 , "BOX DRAWINGS VERTICAL DOUBLE AND LEFT SINGLE" },
	{ 0x2563 , "BOX DRAWINGS DOUBLE VERTICAL AND LEFT" },
	{ 0x2564 , "BOX DRAWINGS DOWN SINGLE AND HORIZONTAL DOUBLE" },
	{ 0x2565 , "BOX DRAWINGS DOWN DOUBLE AND HORIZONTAL SINGLE" },
	{ 0x2566 , "BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL" },
	{ 0x2567 , "BOX DRAWINGS UP SINGLE AND HORIZONTAL DOUBLE" },
	{ 0x2568 , "BOX DRAWINGS UP DOUBLE AND HORIZONTAL SINGLE" },
	{ 0x2569 , "BOX DRAWINGS DOUBLE UP AND HORIZONTAL" },
	{ 0x256A , "BOX DRAWINGS VERTICAL SINGLE AND HORIZONTAL DOUBLE" },
	{ 0x256B , "BOX DRAWINGS VERTICAL DOUBLE AND HORIZONTAL SINGLE" },
	{ 0x256C , "BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL" },
	{ 0x256D , "BOX DRAWINGS LIGHT ARC DOWN AND RIGHT" },
	{ 0x256E , "BOX DRAWINGS LIGHT ARC DOWN AND LEFT" },
	{ 0x256F , "BOX DRAWINGS LIGHT ARC UP AND LEFT" },
	{ 0x2570 , "BOX DRAWINGS LIGHT ARC UP AND RIGHT" },
	{ 0x2571 , "BOX DRAWINGS LIGHT DIAGONAL UPPER RIGHT TO LOWER LEFT" },
	{ 0x2572 , "BOX DRAWINGS LIGHT DIAGONAL UPPER LEFT TO LOWER RIGHT" },
	{ 0x2573 , "BOX DRAWINGS LIGHT DIAGONAL CROSS" },
	{ 0x2574 , "BOX DRAWINGS LIGHT LEFT" },
	{ 0x2575 , "BOX DRAWINGS LIGHT UP" },
	{ 0x2576 , "BOX DRAWINGS LIGHT RIGHT" },
	{ 0x2577 , "BOX DRAWINGS LIGHT DOWN" },
	{ 0x2578 , "BOX DRAWINGS HEAVY LEFT" },
	{ 0x2579 , "BOX DRAWINGS HEAVY UP" },
	{ 0x257A , "BOX DRAWINGS HEAVY RIGHT" },
	{ 0x257B , "BOX DRAWINGS HEAVY DOWN" },
	{ 0x257C , "BOX DRAWINGS LIGHT LEFT AND HEAVY RIGHT" },
	{ 0x257D , "BOX DRAWINGS LIGHT UP AND HEAVY DOWN" },
	{ 0x257E , "BOX DRAWINGS HEAVY LEFT AND LIGHT RIGHT" },
	{ 0x257F , "BOX DRAWINGS HEAVY UP AND LIGHT DOWN" },
	{ 0x2580 , "UPPER HALF BLOCK" },
	{ 0x2581 , "LOWER ONE EIGHTH BLOCK" },
	{ 0x2582 , "LOWER ONE QUARTER BLOCK" },
	{ 0x2583 , "LOWER THREE EIGHTHS BLOCK" },
	{ 0x2584 , "LOWER HALF BLOCK" },
	{ 0x2585 , "LOWER FIVE EIGHTHS BLOCK" },
	{ 0x2586 , "LOWER THREE QUARTERS BLOCK" },
	{ 0x2587 , "LOWER SEVEN EIGHTHS BLOCK" },
	{ 0x2588 , "FULL BLOCK" },
	{ 0x2589 , "LEFT SEVEN EIGHTHS BLOCK" },
	{ 0x258A , "LEFT THREE QUARTERS BLOCK" },
	{ 0x258B , "LEFT FIVE EIGHTHS BLOCK" },
	{ 0x258C , "LEFT HALF BLOCK" },
	{ 0x258D , "LEFT THREE EIGHTHS BLOCK" },
	{ 0x258E , "LEFT ONE QUARTER BLOCK" },
	{ 0x258F , "LEFT ONE EIGHTH BLOCK" },
	{ 0x2590 , "RIGHT HALF BLOCK" },
	{ 0x2591 , "LIGHT SHADE" },
	{ 0x2592 , "MEDIUM SHADE" },
	{ 0x2593 , "DARK SHADE" },
	{ 0x2594 , "UPPER ONE EIGHTH BLOCK" },
	{ 0x2595 , "RIGHT ONE EIGHTH BLOCK" },
	{ 0x25A0 , "BLACK SQUARE" },
	{ 0x25A1 , "WHITE SQUARE" },
	{ 0x25A2 , "WHITE SQUARE WITH ROUNDED CORNERS" },
	{ 0x25A3 , "WHITE SQUARE CONTAINING BLACK SMALL SQUARE" },
	{ 0x25A4 , "SQUARE WITH HORIZONTAL FILL" },
	{ 0x25A5 , "SQUARE WITH VERTICAL FILL" },
	{ 0x25A6 , "SQUARE WITH ORTHOGONAL CROSSHATCH FILL" },
	{ 0x25A7 , "SQUARE WITH UPPER LEFT TO LOWER RIGHT FILL" },
	{ 0x25A8 , "SQUARE WITH UPPER RIGHT TO LOWER LEFT FILL" },
	{ 0x25A9 , "SQUARE WITH DIAGONAL CROSSHATCH FILL" },
	{ 0x25AA , "BLACK SMALL SQUARE" },
	{ 0x25AB , "WHITE SMALL SQUARE" },
	{ 0x25AC , "BLACK RECTANGLE" },
	{ 0x25AD , "WHITE RECTANGLE" },
	{ 0x25AE , "BLACK VERTICAL RECTANGLE" },
	{ 0x25AF , "WHITE VERTICAL RECTANGLE" },
	{ 0x25B0 , "BLACK PARALLELOGRAM" },
	{ 0x25B1 , "WHITE PARALLELOGRAM" },
	{ 0x25B2 , "BLACK UP-POINTING TRIANGLE" },
	{ 0x25B3 , "WHITE UP-POINTING TRIANGLE" },
	{ 0x25B4 , "BLACK UP-POINTING SMALL TRIANGLE" },
	{ 0x25B5 , "WHITE UP-POINTING SMALL TRIANGLE" },
	{ 0x25B6 , "BLACK RIGHT-POINTING TRIANGLE" },
	{ 0x25B7 , "WHITE RIGHT-POINTING TRIANGLE" },
	{ 0x25B8 , "BLACK RIGHT-POINTING SMALL TRIANGLE" },
	{ 0x25B9 , "WHITE RIGHT-POINTING SMALL TRIANGLE" },
	{ 0x25BA , "BLACK RIGHT-POINTING POINTER" },
	{ 0x25BB , "WHITE RIGHT-POINTING POINTER" },
	{ 0x25BC , "BLACK DOWN-POINTING TRIANGLE" },
	{ 0x25BD , "WHITE DOWN-POINTING TRIANGLE" },
	{ 0x25BE , "BLACK DOWN-POINTING SMALL TRIANGLE" },
	{ 0x25BF , "WHITE DOWN-POINTING SMALL TRIANGLE" },
	{ 0x25C0 , "BLACK LEFT-POINTING TRIANGLE" },
	{ 0x25C1 , "WHITE LEFT-POINTING TRIANGLE" },
	{ 0x25C2 , "BLACK LEFT-POINTING SMALL TRIANGLE" },
	{ 0x25C3 , "WHITE LEFT-POINTING SMALL TRIANGLE" },
	{ 0x25C4 , "BLACK LEFT-POINTING POINTER" },
	{ 0x25C5 , "WHITE LEFT-POINTING POINTER" },
	{ 0x25C6 , "BLACK DIAMOND" },
	{ 0x25C7 , "WHITE DIAMOND" },
	{ 0x25C8 , "WHITE DIAMOND CONTAINING BLACK SMALL DIAMOND" },
	{ 0x25C9 , "FISHEYE" },
	{ 0x25CA , "LOZENGE" },
	{ 0x25CB , "WHITE CIRCLE" },
	{ 0x25CC , "DOTTED CIRCLE" },
	{ 0x25CD , "CIRCLE WITH VERTICAL FILL" },
	{ 0x25CE , "BULLSEYE" },
	{ 0x25CF , "BLACK CIRCLE" },
	{ 0x25D0 , "CIRCLE WITH LEFT HALF BLACK" },
	{ 0x25D1 , "CIRCLE WITH RIGHT HALF BLACK" },
	{ 0x25D2 , "CIRCLE WITH LOWER HALF BLACK" },
	{ 0x25D3 , "CIRCLE WITH UPPER HALF BLACK" },
	{ 0x25D4 , "CIRCLE WITH UPPER RIGHT QUADRANT BLACK" },
	{ 0x25D5 , "CIRCLE WITH ALL BUT UPPER LEFT QUADRANT BLACK" },
	{ 0x25D6 , "LEFT HALF BLACK CIRCLE" },
	{ 0x25D7 , "RIGHT HALF BLACK CIRCLE" },
	{ 0x25D8 , "INVERSE BULLET" },
	{ 0x25D9 , "INVERSE WHITE CIRCLE" },
	{ 0x25DA , "UPPER HALF INVERSE WHITE CIRCLE" },
	{ 0x25DB , "LOWER HALF INVERSE WHITE CIRCLE" },
	{ 0x25DC , "UPPER LEFT QUADRANT CIRCULAR ARC" },
	{ 0x25DD , "UPPER RIGHT QUADRANT CIRCULAR ARC" },
	{ 0x25DE , "LOWER RIGHT QUADRANT CIRCULAR ARC" },
	{ 0x25DF , "LOWER LEFT QUADRANT CIRCULAR ARC" },
	{ 0x25E0 , "UPPER HALF CIRCLE" },
	{ 0x25E1 , "LOWER HALF CIRCLE" },
	{ 0x25E2 , "BLACK LOWER RIGHT TRIANGLE" },
	{ 0x25E3 , "BLACK LOWER LEFT TRIANGLE" },
	{ 0x25E4 , "BLACK UPPER LEFT TRIANGLE" },
	{ 0x25E5 , "BLACK UPPER RIGHT TRIANGLE" },
	{ 0x25E6 , "WHITE BULLET" },
	{ 0x25E7 , "SQUARE WITH LEFT HALF BLACK" },
	{ 0x25E8 , "SQUARE WITH RIGHT HALF BLACK" },
	{ 0x25E9 , "SQUARE WITH UPPER LEFT DIAGONAL HALF BLACK" },
	{ 0x25EA , "SQUARE WITH LOWER RIGHT DIAGONAL HALF BLACK" },
	{ 0x25EB , "WHITE SQUARE WITH VERTICAL BISECTING LINE" },
	{ 0x25EC , "WHITE UP-POINTING TRIANGLE WITH DOT" },
	{ 0x25ED , "UP-POINTING TRIANGLE WITH LEFT HALF BLACK" },
	{ 0x25EE , "UP-POINTING TRIANGLE WITH RIGHT HALF BLACK" },
	{ 0x25EF , "LARGE CIRCLE" },
	{ 0x2600 , "BLACK SUN WITH RAYS" },
	{ 0x2601 , "CLOUD" },
	{ 0x2602 , "UMBRELLA" },
	{ 0x2603 , "SNOWMAN" },
	{ 0x2604 , "COMET" },
	{ 0x2605 , "BLACK STAR" },
	{ 0x2606 , "WHITE STAR" },
	{ 0x2607 , "LIGHTNING" },
	{ 0x2608 , "THUNDERSTORM" },
	{ 0x2609 , "SUN" },
	{ 0x260A , "ASCENDING NODE" },
	{ 0x260B , "DESCENDING NODE" },
	{ 0x260C , "CONJUNCTION" },
	{ 0x260D , "OPPOSITION" },
	{ 0x260E , "BLACK TELEPHONE" },
	{ 0x260F , "WHITE TELEPHONE" },
	{ 0x2610 , "BALLOT BOX" },
	{ 0x2611 , "BALLOT BOX WITH CHECK" },
	{ 0x2612 , "BALLOT BOX WITH X" },
	{ 0x2613 , "SALTIRE" },
	{ 0x261A , "BLACK LEFT POINTING INDEX" },
	{ 0x261B , "BLACK RIGHT POINTING INDEX" },
	{ 0x261C , "WHITE LEFT POINTING INDEX" },
	{ 0x261D , "WHITE UP POINTING INDEX" },
	{ 0x261E , "WHITE RIGHT POINTING INDEX" },
	{ 0x261F , "WHITE DOWN POINTING INDEX" },
	{ 0x2620 , "SKULL AND CROSSBONES" },
	{ 0x2621 , "CAUTION SIGN" },
	{ 0x2622 , "RADIOACTIVE SIGN" },
	{ 0x2623 , "BIOHAZARD SIGN" },
	{ 0x2624 , "CADUCEUS" },
	{ 0x2625 , "ANKH" },
	{ 0x2626 , "ORTHODOX CROSS" },
	{ 0x2627 , "CHI RHO" },
	{ 0x2628 , "CROSS OF LORRAINE" },
	{ 0x2629 , "CROSS OF JERUSALEM" },
	{ 0x262A , "STAR AND CRESCENT" },
	{ 0x262B , "FARSI SYMBOL" },
	{ 0x262C , "ADI SHAKTI" },
	{ 0x262D , "HAMMER AND SICKLE" },
	{ 0x262E , "PEACE SYMBOL" },
	{ 0x262F , "YIN YANG" },
	{ 0x2630 , "TRIGRAM FOR HEAVEN" },
	{ 0x2631 , "TRIGRAM FOR LAKE" },
	{ 0x2632 , "TRIGRAM FOR FIRE" },
	{ 0x2633 , "TRIGRAM FOR THUNDER" },
	{ 0x2634 , "TRIGRAM FOR WIND" },
	{ 0x2635 , "TRIGRAM FOR WATER" },
	{ 0x2636 , "TRIGRAM FOR MOUNTAIN" },
	{ 0x2637 , "TRIGRAM FOR EARTH" },
	{ 0x2638 , "WHEEL OF DHARMA" },
	{ 0x2639 , "WHITE FROWNING FACE" },
	{ 0x263A , "WHITE SMILING FACE" },
	{ 0x263B , "BLACK SMILING FACE" },
	{ 0x263C , "WHITE SUN WITH RAYS" },
	{ 0x263D , "FIRST QUARTER MOON" },
	{ 0x263E , "LAST QUARTER MOON" },
	{ 0x263F , "MERCURY" },
	{ 0x2640 , "FEMALE SIGN" },
	{ 0x2641 , "EARTH" },
	{ 0x2642 , "MALE SIGN" },
	{ 0x2643 , "JUPITER" },
	{ 0x2644 , "SATURN" },
	{ 0x2645 , "URANUS" },
	{ 0x2646 , "NEPTUNE" },
	{ 0x2647 , "PLUTO" },
	{ 0x2648 , "ARIES" },
	{ 0x2649 , "TAURUS" },
	{ 0x264A , "GEMINI" },
	{ 0x264B , "CANCER" },
	{ 0x264C , "LEO" },
	{ 0x264D , "VIRGO" },
	{ 0x264E , "LIBRA" },
	{ 0x264F , "SCORPIUS" },
	{ 0x2650 , "SAGITTARIUS" },
	{ 0x2651 , "CAPRICORN" },
	{ 0x2652 , "AQUARIUS" },
	{ 0x2653 , "PISCES" },
	{ 0x2654 , "WHITE CHESS KING" },
	{ 0x2655 , "WHITE CHESS QUEEN" },
	{ 0x2656 , "WHITE CHESS ROOK" },
	{ 0x2657 , "WHITE CHESS BISHOP" },
	{ 0x2658 , "WHITE CHESS KNIGHT" },
	{ 0x2659 , "WHITE CHESS PAWN" },
	{ 0x265A , "BLACK CHESS KING" },
	{ 0x265B , "BLACK CHESS QUEEN" },
	{ 0x265C , "BLACK CHESS ROOK" },
	{ 0x265D , "BLACK CHESS BISHOP" },
	{ 0x265E , "BLACK CHESS KNIGHT" },
	{ 0x265F , "BLACK CHESS PAWN" },
	{ 0x2660 , "BLACK SPADE SUIT" },
	{ 0x2661 , "WHITE HEART SUIT" },
	{ 0x2662 , "WHITE DIAMOND SUIT" },
	{ 0x2663 , "BLACK CLUB SUIT" },
	{ 0x2664 , "WHITE SPADE SUIT" },
	{ 0x2665 , "BLACK HEART SUIT" },
	{ 0x2666 , "BLACK DIAMOND SUIT" },
	{ 0x2667 , "WHITE CLUB SUIT" },
	{ 0x2668 , "HOT SPRINGS" },
	{ 0x2669 , "QUARTER NOTE" },
	{ 0x266A , "EIGHTH NOTE" },
	{ 0x266B , "BEAMED EIGHTH NOTES" },
	{ 0x266C , "BEAMED SIXTEENTH NOTES" },
	{ 0x266D , "MUSIC FLAT SIGN" },
	{ 0x266E , "MUSIC NATURAL SIGN" },
	{ 0x266F , "MUSIC SHARP SIGN" },
	{ 0x2701 , "UPPER BLADE SCISSORS" },
	{ 0x2702 , "BLACK SCISSORS" },
	{ 0x2703 , "LOWER BLADE SCISSORS" },
	{ 0x2704 , "WHITE SCISSORS" },
	{ 0x2706 , "TELEPHONE LOCATION SIGN" },
	{ 0x2707 , "TAPE DRIVE" },
	{ 0x2708 , "AIRPLANE" },
	{ 0x2709 , "ENVELOPE" },
	{ 0x270C , "VICTORY HAND" },
	{ 0x270D , "WRITING HAND" },
	{ 0x270E , "LOWER RIGHT PENCIL" },
	{ 0x270F , "PENCIL" },
	{ 0x2710 , "UPPER RIGHT PENCIL" },
	{ 0x2711 , "WHITE NIB" },
	{ 0x2712 , "BLACK NIB" },
	{ 0x2713 , "CHECK MARK" },
	{ 0x2714 , "HEAVY CHECK MARK" },
	{ 0x2715 , "MULTIPLICATION X" },
	{ 0x2716 , "HEAVY MULTIPLICATION X" },
	{ 0x2717 , "BALLOT X" },
	{ 0x2718 , "HEAVY BALLOT X" },
	{ 0x2719 , "OUTLINED GREEK CROSS" },
	{ 0x271A , "HEAVY GREEK CROSS" },
	{ 0x271B , "OPEN CENTRE CROSS" },
	{ 0x271C , "HEAVY OPEN CENTRE CROSS" },
	{ 0x271D , "LATIN CROSS" },
	{ 0x271E , "SHADOWED WHITE LATIN CROSS" },
	{ 0x271F , "OUTLINED LATIN CROSS" },
	{ 0x2720 , "MALTESE CROSS" },
	{ 0x2721 , "STAR OF DAVID" },
	{ 0x2722 , "FOUR TEARDROP-SPOKED ASTERISK" },
	{ 0x2723 , "FOUR BALLOON-SPOKED ASTERISK" },
	{ 0x2724 , "HEAVY FOUR BALLOON-SPOKED ASTERISK" },
	{ 0x2725 , "FOUR CLUB-SPOKED ASTERISK" },
	{ 0x2726 , "BLACK FOUR POINTED STAR" },
	{ 0x2727 , "WHITE FOUR POINTED STAR" },
	{ 0x2729 , "STRESS OUTLINED WHITE STAR" },
	{ 0x272A , "CIRCLED WHITE STAR" },
	{ 0x272B , "OPEN CENTRE BLACK STAR" },
	{ 0x272C , "BLACK CENTRE WHITE STAR" },
	{ 0x272D , "OUTLINED BLACK STAR" },
	{ 0x272E , "HEAVY OUTLINED BLACK STAR" },
	{ 0x272F , "PINWHEEL STAR" },
	{ 0x2730 , "SHADOWED WHITE STAR" },
	{ 0x2731 , "HEAVY ASTERISK" },
	{ 0x2732 , "OPEN CENTRE ASTERISK" },
	{ 0x2733 , "EIGHT SPOKED ASTERISK" },
	{ 0x2734 , "EIGHT POINTED BLACK STAR" },
	{ 0x2735 , "EIGHT POINTED PINWHEEL STAR" },
	{ 0x2736 , "SIX POINTED BLACK STAR" },
	{ 0x2737 , "EIGHT POINTED RECTILINEAR BLACK STAR" },
	{ 0x2738 , "HEAVY EIGHT POINTED RECTILINEAR BLACK STAR" },
	{ 0x2739 , "TWELVE POINTED BLACK STAR" },
	{ 0x273A , "SIXTEEN POINTED ASTERISK" },
	{ 0x273B , "TEARDROP-SPOKED ASTERISK" },
	{ 0x273C , "OPEN CENTRE TEARDROP-SPOKED ASTERISK" },
	{ 0x273D , "HEAVY TEARDROP-SPOKED ASTERISK" },
	{ 0x273E , "SIX PETALLED BLACK AND WHITE FLORETTE" },
	{ 0x273F , "BLACK FLORETTE" },
	{ 0x2740 , "WHITE FLORETTE" },
	{ 0x2741 , "EIGHT PETALLED OUTLINED BLACK FLORETTE" },
	{ 0x2742 , "CIRCLED OPEN CENTRE EIGHT POINTED STAR" },
	{ 0x2743 , "HEAVY TEARDROP-SPOKED PINWHEEL ASTERISK" },
	{ 0x2744 , "SNOWFLAKE" },
	{ 0x2745 , "TIGHT TRIFOLIATE SNOWFLAKE" },
	{ 0x2746 , "HEAVY CHEVRON SNOWFLAKE" },
	{ 0x2747 , "SPARKLE" },
	{ 0x2748 , "HEAVY SPARKLE" },
	{ 0x2749 , "BALLOON-SPOKED ASTERISK" },
	{ 0x274A , "EIGHT TEARDROP-SPOKED PROPELLER ASTERISK" },
	{ 0x274B , "HEAVY EIGHT TEARDROP-SPOKED PROPELLER ASTERISK" },
	{ 0x274D , "SHADOWED WHITE CIRCLE" },
	{ 0x274F , "LOWER RIGHT DROP-SHADOWED WHITE SQUARE" },
	{ 0x2750 , "UPPER RIGHT DROP-SHADOWED WHITE SQUARE" },
	{ 0x2751 , "LOWER RIGHT SHADOWED WHITE SQUARE" },
	{ 0x2752 , "UPPER RIGHT SHADOWED WHITE SQUARE" },
	{ 0x2756 , "BLACK DIAMOND MINUS WHITE X" },
	{ 0x2758 , "LIGHT VERTICAL BAR" },
	{ 0x2759 , "MEDIUM VERTICAL BAR" },
	{ 0x275A , "HEAVY VERTICAL BAR" },
	{ 0x275B , "HEAVY SINGLE TURNED COMMA QUOTATION MARK ORNAMENT" },
	{ 0x275C , "HEAVY SINGLE COMMA QUOTATION MARK ORNAMENT" },
	{ 0x275D , "HEAVY DOUBLE TURNED COMMA QUOTATION MARK ORNAMENT" },
	{ 0x275E , "HEAVY DOUBLE COMMA QUOTATION MARK ORNAMENT" },
	{ 0x2761 , "CURVED STEM PARAGRAPH SIGN ORNAMENT" },
	{ 0x2762 , "HEAVY EXCLAMATION MARK ORNAMENT" },
	{ 0x2763 , "HEAVY HEART EXCLAMATION MARK ORNAMENT" },
	{ 0x2764 , "HEAVY BLACK HEART" },
	{ 0x2765 , "ROTATED HEAVY BLACK HEART BULLET" },
	{ 0x2766 , "FLORAL HEART" },
	{ 0x2767 , "ROTATED FLORAL HEART BULLET" },
	{ 0x2776 , "DINGBAT NEGATIVE CIRCLED DIGIT ONE" },
	{ 0x2777 , "DINGBAT NEGATIVE CIRCLED DIGIT TWO" },
	{ 0x2778 , "DINGBAT NEGATIVE CIRCLED DIGIT THREE" },
	{ 0x2779 , "DINGBAT NEGATIVE CIRCLED DIGIT FOUR" },
	{ 0x277A , "DINGBAT NEGATIVE CIRCLED DIGIT FIVE" },
	{ 0x277B , "DINGBAT NEGATIVE CIRCLED DIGIT SIX" },
	{ 0x277C , "DINGBAT NEGATIVE CIRCLED DIGIT SEVEN" },
	{ 0x277D , "DINGBAT NEGATIVE CIRCLED DIGIT EIGHT" },
	{ 0x277E , "DINGBAT NEGATIVE CIRCLED DIGIT NINE" },
	{ 0x277F , "DINGBAT NEGATIVE CIRCLED NUMBER TEN" },
	{ 0x2780 , "DINGBAT CIRCLED SANS-SERIF DIGIT ONE" },
	{ 0x2781 , "DINGBAT CIRCLED SANS-SERIF DIGIT TWO" },
	{ 0x2782 , "DINGBAT CIRCLED SANS-SERIF DIGIT THREE" },
	{ 0x2783 , "DINGBAT CIRCLED SANS-SERIF DIGIT FOUR" },
	{ 0x2784 , "DINGBAT CIRCLED SANS-SERIF DIGIT FIVE" },
	{ 0x2785 , "DINGBAT CIRCLED SANS-SERIF DIGIT SIX" },
	{ 0x2786 , "DINGBAT CIRCLED SANS-SERIF DIGIT SEVEN" },
	{ 0x2787 , "DINGBAT CIRCLED SANS-SERIF DIGIT EIGHT" },
	{ 0x2788 , "DINGBAT CIRCLED SANS-SERIF DIGIT NINE" },
	{ 0x2789 , "DINGBAT CIRCLED SANS-SERIF NUMBER TEN" },
	{ 0x278A , "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT ONE" },
	{ 0x278B , "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT TWO" },
	{ 0x278C , "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT THREE" },
	{ 0x278D , "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT FOUR" },
	{ 0x278E , "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT FIVE" },
	{ 0x278F , "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT SIX" },
	{ 0x2790 , "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT SEVEN" },
	{ 0x2791 , "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT EIGHT" },
	{ 0x2792 , "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT NINE" },
	{ 0x2793 , "DINGBAT NEGATIVE CIRCLED SANS-SERIF NUMBER TEN" },
	{ 0x2794 , "HEAVY WIDE-HEADED RIGHTWARDS ARROW" },
	{ 0x2798 , "HEAVY SOUTH EAST ARROW" },
	{ 0x2799 , "HEAVY RIGHTWARDS ARROW" },
	{ 0x279A , "HEAVY NORTH EAST ARROW" },
	{ 0x279B , "DRAFTING POINT RIGHTWARDS ARROW" },
	{ 0x279C , "HEAVY ROUND-TIPPED RIGHTWARDS ARROW" },
	{ 0x279D , "TRIANGLE-HEADED RIGHTWARDS ARROW" },
	{ 0x279E , "HEAVY TRIANGLE-HEADED RIGHTWARDS ARROW" },
	{ 0x279F , "DASHED TRIANGLE-HEADED RIGHTWARDS ARROW" },
	{ 0x27A0 , "HEAVY DASHED TRIANGLE-HEADED RIGHTWARDS ARROW" },
	{ 0x27A1 , "BLACK RIGHTWARDS ARROW" },
	{ 0x27A2 , "THREE-D TOP-LIGHTED RIGHTWARDS ARROWHEAD" },
	{ 0x27A3 , "THREE-D BOTTOM-LIGHTED RIGHTWARDS ARROWHEAD" },
	{ 0x27A4 , "BLACK RIGHTWARDS ARROWHEAD" },
	{ 0x27A5 , "HEAVY BLACK CURVED DOWNWARDS AND RIGHTWARDS ARROW" },
	{ 0x27A6 , "HEAVY BLACK CURVED UPWARDS AND RIGHTWARDS ARROW" },
	{ 0x27A7 , "SQUAT BLACK RIGHTWARDS ARROW" },
	{ 0x27A8 , "HEAVY CONCAVE-POINTED BLACK RIGHTWARDS ARROW" },
	{ 0x27A9 , "RIGHT-SHADED WHITE RIGHTWARDS ARROW" },
	{ 0x27AA , "LEFT-SHADED WHITE RIGHTWARDS ARROW" },
	{ 0x27AB , "BACK-TILTED SHADOWED WHITE RIGHTWARDS ARROW" },
	{ 0x27AC , "FRONT-TILTED SHADOWED WHITE RIGHTWARDS ARROW" },
	{ 0x27AD , "HEAVY LOWER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW" },
	{ 0x27AE , "HEAVY UPPER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW" },
	{ 0x27AF , "NOTCHED LOWER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW" },
	{ 0x27B1 , "NOTCHED UPPER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW" },
	{ 0x27B2 , "CIRCLED HEAVY WHITE RIGHTWARDS ARROW" },
	{ 0x27B3 , "WHITE-FEATHERED RIGHTWARDS ARROW" },
	{ 0x27B4 , "BLACK-FEATHERED SOUTH EAST ARROW" },
	{ 0x27B5 , "BLACK-FEATHERED RIGHTWARDS ARROW" },
	{ 0x27B6 , "BLACK-FEATHERED NORTH EAST ARROW" },
	{ 0x27B7 , "HEAVY BLACK-FEATHERED SOUTH EAST ARROW" },
	{ 0x27B8 , "HEAVY BLACK-FEATHERED RIGHTWARDS ARROW" },
	{ 0x27B9 , "HEAVY BLACK-FEATHERED NORTH EAST ARROW" },
	{ 0x27BA , "TEARDROP-BARBED RIGHTWARDS ARROW" },
	{ 0x27BB , "HEAVY TEARDROP-SHANKED RIGHTWARDS ARROW" },
	{ 0x27BC , "WEDGE-TAILED RIGHTWARDS ARROW" },
	{ 0x27BD , "HEAVY WEDGE-TAILED RIGHTWARDS ARROW" },
	{ 0x27BE , "OPEN-OUTLINED RIGHTWARDS ARROW" },
	{ 0x3000 , "IDEOGRAPHIC SPACE" },
	{ 0x3001 , "IDEOGRAPHIC COMMA" },
	{ 0x3002 , "IDEOGRAPHIC FULL STOP" },
	{ 0x3003 , "DITTO MARK" },
	{ 0x3004 , "JAPANESE INDUSTRIAL STANDARD SYMBOL" },
	{ 0x3005 , "IDEOGRAPHIC ITERATION MARK" },
	{ 0x3006 , "IDEOGRAPHIC CLOSING MARK" },
	{ 0x3007 , "IDEOGRAPHIC NUMBER ZERO" },
	{ 0x3008 , "LEFT ANGLE BRACKET" },
	{ 0x3009 , "RIGHT ANGLE BRACKET" },
	{ 0x300A , "LEFT DOUBLE ANGLE BRACKET" },
	{ 0x300B , "RIGHT DOUBLE ANGLE BRACKET" },
	{ 0x300C , "LEFT CORNER BRACKET" },
	{ 0x300D , "RIGHT CORNER BRACKET" },
	{ 0x300E , "LEFT WHITE CORNER BRACKET" },
	{ 0x300F , "RIGHT WHITE CORNER BRACKET" },
	{ 0x3010 , "LEFT BLACK LENTICULAR BRACKET" },
	{ 0x3011 , "RIGHT BLACK LENTICULAR BRACKET" },
	{ 0x3012 , "POSTAL MARK" },
	{ 0x3013 , "GETA MARK" },
	{ 0x3014 , "LEFT TORTOISE SHELL BRACKET" },
	{ 0x3015 , "RIGHT TORTOISE SHELL BRACKET" },
	{ 0x3016 , "LEFT WHITE LENTICULAR BRACKET" },
	{ 0x3017 , "RIGHT WHITE LENTICULAR BRACKET" },
	{ 0x3018 , "LEFT WHITE TORTOISE SHELL BRACKET" },
	{ 0x3019 , "RIGHT WHITE TORTOISE SHELL BRACKET" },
	{ 0x301A , "LEFT WHITE SQUARE BRACKET" },
	{ 0x301B , "RIGHT WHITE SQUARE BRACKET" },
	{ 0x301C , "WAVE DASH" },
	{ 0x301D , "REVERSED DOUBLE PRIME QUOTATION MARK" },
	{ 0x301E , "DOUBLE PRIME QUOTATION MARK" },
	{ 0x301F , "LOW DOUBLE PRIME QUOTATION MARK" },
	{ 0x3020 , "POSTAL MARK FACE" },
	{ 0x3021 , "HANGZHOU NUMERAL ONE" },
	{ 0x3022 , "HANGZHOU NUMERAL TWO" },
	{ 0x3023 , "HANGZHOU NUMERAL THREE" },
	{ 0x3024 , "HANGZHOU NUMERAL FOUR" },
	{ 0x3025 , "HANGZHOU NUMERAL FIVE" },
	{ 0x3026 , "HANGZHOU NUMERAL SIX" },
	{ 0x3027 , "HANGZHOU NUMERAL SEVEN" },
	{ 0x3028 , "HANGZHOU NUMERAL EIGHT" },
	{ 0x3029 , "HANGZHOU NUMERAL NINE" },
	{ 0x302A , "IDEOGRAPHIC LEVEL TONE MARK" },
	{ 0x302B , "IDEOGRAPHIC RISING TONE MARK" },
	{ 0x302C , "IDEOGRAPHIC DEPARTING TONE MARK" },
	{ 0x302D , "IDEOGRAPHIC ENTERING TONE MARK" },
	{ 0x302E , "HANGUL SINGLE DOT TONE MARK" },
	{ 0x302F , "HANGUL DOUBLE DOT TONE MARK" },
	{ 0x3030 , "WAVY DASH" },
	{ 0x3031 , "VERTICAL KANA REPEAT MARK" },
	{ 0x3032 , "VERTICAL KANA REPEAT WITH VOICED SOUND MARK" },
	{ 0x3033 , "VERTICAL KANA REPEAT MARK UPPER HALF" },
	{ 0x3034 , "VERTICAL KANA REPEAT WITH VOICED SOUND MARK UPPER HALF" },
	{ 0x3035 , "VERTICAL KANA REPEAT MARK LOWER HALF" },
	{ 0x3036 , "CIRCLED POSTAL MARK" },
	{ 0x3037 , "IDEOGRAPHIC TELEGRAPH LINE FEED SEPARATOR SYMBOL" },
	{ 0x303F , "IDEOGRAPHIC HALF FILL SPACE" },
#endif
	{ 0x3041 , "HIRAGANA LETTER SMALL A" },
	{ 0x3042 , "HIRAGANA LETTER A" },
	{ 0x3043 , "HIRAGANA LETTER SMALL I" },
	{ 0x3044 , "HIRAGANA LETTER I" },
	{ 0x3045 , "HIRAGANA LETTER SMALL U" },
	{ 0x3046 , "HIRAGANA LETTER U" },
	{ 0x3047 , "HIRAGANA LETTER SMALL E" },
	{ 0x3048 , "HIRAGANA LETTER E" },
	{ 0x3049 , "HIRAGANA LETTER SMALL O" },
	{ 0x304A , "HIRAGANA LETTER O" },
	{ 0x304B , "HIRAGANA LETTER KA" },
	{ 0x304C , "HIRAGANA LETTER GA" },
	{ 0x304D , "HIRAGANA LETTER KI" },
	{ 0x304E , "HIRAGANA LETTER GI" },
	{ 0x304F , "HIRAGANA LETTER KU" },
	{ 0x3050 , "HIRAGANA LETTER GU" },
	{ 0x3051 , "HIRAGANA LETTER KE" },
	{ 0x3052 , "HIRAGANA LETTER GE" },
	{ 0x3053 , "HIRAGANA LETTER KO" },
	{ 0x3054 , "HIRAGANA LETTER GO" },
	{ 0x3055 , "HIRAGANA LETTER SA" },
	{ 0x3056 , "HIRAGANA LETTER ZA" },
	{ 0x3057 , "HIRAGANA LETTER SI" },
	{ 0x3058 , "HIRAGANA LETTER ZI" },
	{ 0x3059 , "HIRAGANA LETTER SU" },
	{ 0x305A , "HIRAGANA LETTER ZU" },
	{ 0x305B , "HIRAGANA LETTER SE" },
	{ 0x305C , "HIRAGANA LETTER ZE" },
	{ 0x305D , "HIRAGANA LETTER SO" },
	{ 0x305E , "HIRAGANA LETTER ZO" },
	{ 0x305F , "HIRAGANA LETTER TA" },
	{ 0x3060 , "HIRAGANA LETTER DA" },
	{ 0x3061 , "HIRAGANA LETTER TI" },
	{ 0x3062 , "HIRAGANA LETTER DI" },
	{ 0x3063 , "HIRAGANA LETTER SMALL TU" },
	{ 0x3064 , "HIRAGANA LETTER TU" },
	{ 0x3065 , "HIRAGANA LETTER DU" },
	{ 0x3066 , "HIRAGANA LETTER TE" },
	{ 0x3067 , "HIRAGANA LETTER DE" },
	{ 0x3068 , "HIRAGANA LETTER TO" },
	{ 0x3069 , "HIRAGANA LETTER DO" },
	{ 0x306A , "HIRAGANA LETTER NA" },
	{ 0x306B , "HIRAGANA LETTER NI" },
	{ 0x306C , "HIRAGANA LETTER NU" },
	{ 0x306D , "HIRAGANA LETTER NE" },
	{ 0x306E , "HIRAGANA LETTER NO" },
	{ 0x306F , "HIRAGANA LETTER HA" },
	{ 0x3070 , "HIRAGANA LETTER BA" },
	{ 0x3071 , "HIRAGANA LETTER PA" },
	{ 0x3072 , "HIRAGANA LETTER HI" },
	{ 0x3073 , "HIRAGANA LETTER BI" },
	{ 0x3074 , "HIRAGANA LETTER PI" },
	{ 0x3075 , "HIRAGANA LETTER HU" },
	{ 0x3076 , "HIRAGANA LETTER BU" },
	{ 0x3077 , "HIRAGANA LETTER PU" },
	{ 0x3078 , "HIRAGANA LETTER HE" },
	{ 0x3079 , "HIRAGANA LETTER BE" },
	{ 0x307A , "HIRAGANA LETTER PE" },
	{ 0x307B , "HIRAGANA LETTER HO" },
	{ 0x307C , "HIRAGANA LETTER BO" },
	{ 0x307D , "HIRAGANA LETTER PO" },
	{ 0x307E , "HIRAGANA LETTER MA" },
	{ 0x307F , "HIRAGANA LETTER MI" },
	{ 0x3080 , "HIRAGANA LETTER MU" },
	{ 0x3081 , "HIRAGANA LETTER ME" },
	{ 0x3082 , "HIRAGANA LETTER MO" },
	{ 0x3083 , "HIRAGANA LETTER SMALL YA" },
	{ 0x3084 , "HIRAGANA LETTER YA" },
	{ 0x3085 , "HIRAGANA LETTER SMALL YU" },
	{ 0x3086 , "HIRAGANA LETTER YU" },
	{ 0x3087 , "HIRAGANA LETTER SMALL YO" },
	{ 0x3088 , "HIRAGANA LETTER YO" },
	{ 0x3089 , "HIRAGANA LETTER RA" },
	{ 0x308A , "HIRAGANA LETTER RI" },
	{ 0x308B , "HIRAGANA LETTER RU" },
	{ 0x308C , "HIRAGANA LETTER RE" },
	{ 0x308D , "HIRAGANA LETTER RO" },
	{ 0x308E , "HIRAGANA LETTER SMALL WA" },
	{ 0x308F , "HIRAGANA LETTER WA" },
	{ 0x3090 , "HIRAGANA LETTER WI" },
	{ 0x3091 , "HIRAGANA LETTER WE" },
	{ 0x3092 , "HIRAGANA LETTER WO" },
	{ 0x3093 , "HIRAGANA LETTER N" },
	{ 0x3094 , "HIRAGANA LETTER VU" },
	{ 0x3099 , "COMBINING KATAKANA-HIRAGANA VOICED SOUND MARK" },
	{ 0x309A , "COMBINING KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK" },
	{ 0x309B , "KATAKANA-HIRAGANA VOICED SOUND MARK" },
	{ 0x309C , "KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK" },
	{ 0x309D , "HIRAGANA ITERATION MARK" },
	{ 0x309E , "HIRAGANA VOICED ITERATION MARK" },
	{ 0x30A1 , "KATAKANA LETTER SMALL A" },
	{ 0x30A2 , "KATAKANA LETTER A" },
	{ 0x30A3 , "KATAKANA LETTER SMALL I" },
	{ 0x30A4 , "KATAKANA LETTER I" },
	{ 0x30A5 , "KATAKANA LETTER SMALL U" },
	{ 0x30A6 , "KATAKANA LETTER U" },
	{ 0x30A7 , "KATAKANA LETTER SMALL E" },
	{ 0x30A8 , "KATAKANA LETTER E" },
	{ 0x30A9 , "KATAKANA LETTER SMALL O" },
	{ 0x30AA , "KATAKANA LETTER O" },
	{ 0x30AB , "KATAKANA LETTER KA" },
	{ 0x30AC , "KATAKANA LETTER GA" },
	{ 0x30AD , "KATAKANA LETTER KI" },
	{ 0x30AE , "KATAKANA LETTER GI" },
	{ 0x30AF , "KATAKANA LETTER KU" },
	{ 0x30B0 , "KATAKANA LETTER GU" },
	{ 0x30B1 , "KATAKANA LETTER KE" },
	{ 0x30B2 , "KATAKANA LETTER GE" },
	{ 0x30B3 , "KATAKANA LETTER KO" },
	{ 0x30B4 , "KATAKANA LETTER GO" },
	{ 0x30B5 , "KATAKANA LETTER SA" },
	{ 0x30B6 , "KATAKANA LETTER ZA" },
	{ 0x30B7 , "KATAKANA LETTER SI" },
	{ 0x30B8 , "KATAKANA LETTER ZI" },
	{ 0x30B9 , "KATAKANA LETTER SU" },
	{ 0x30BA , "KATAKANA LETTER ZU" },
	{ 0x30BB , "KATAKANA LETTER SE" },
	{ 0x30BC , "KATAKANA LETTER ZE" },
	{ 0x30BD , "KATAKANA LETTER SO" },
	{ 0x30BE , "KATAKANA LETTER ZO" },
	{ 0x30BF , "KATAKANA LETTER TA" },
	{ 0x30C0 , "KATAKANA LETTER DA" },
	{ 0x30C1 , "KATAKANA LETTER TI" },
	{ 0x30C2 , "KATAKANA LETTER DI" },
	{ 0x30C3 , "KATAKANA LETTER SMALL TU" },
	{ 0x30C4 , "KATAKANA LETTER TU" },
	{ 0x30C5 , "KATAKANA LETTER DU" },
	{ 0x30C6 , "KATAKANA LETTER TE" },
	{ 0x30C7 , "KATAKANA LETTER DE" },
	{ 0x30C8 , "KATAKANA LETTER TO" },
	{ 0x30C9 , "KATAKANA LETTER DO" },
	{ 0x30CA , "KATAKANA LETTER NA" },
	{ 0x30CB , "KATAKANA LETTER NI" },
	{ 0x30CC , "KATAKANA LETTER NU" },
	{ 0x30CD , "KATAKANA LETTER NE" },
	{ 0x30CE , "KATAKANA LETTER NO" },
	{ 0x30CF , "KATAKANA LETTER HA" },
	{ 0x30D0 , "KATAKANA LETTER BA" },
	{ 0x30D1 , "KATAKANA LETTER PA" },
	{ 0x30D2 , "KATAKANA LETTER HI" },
	{ 0x30D3 , "KATAKANA LETTER BI" },
	{ 0x30D4 , "KATAKANA LETTER PI" },
	{ 0x30D5 , "KATAKANA LETTER HU" },
	{ 0x30D6 , "KATAKANA LETTER BU" },
	{ 0x30D7 , "KATAKANA LETTER PU" },
	{ 0x30D8 , "KATAKANA LETTER HE" },
	{ 0x30D9 , "KATAKANA LETTER BE" },
	{ 0x30DA , "KATAKANA LETTER PE" },
	{ 0x30DB , "KATAKANA LETTER HO" },
	{ 0x30DC , "KATAKANA LETTER BO" },
	{ 0x30DD , "KATAKANA LETTER PO" },
	{ 0x30DE , "KATAKANA LETTER MA" },
	{ 0x30DF , "KATAKANA LETTER MI" },
	{ 0x30E0 , "KATAKANA LETTER MU" },
	{ 0x30E1 , "KATAKANA LETTER ME" },
	{ 0x30E2 , "KATAKANA LETTER MO" },
	{ 0x30E3 , "KATAKANA LETTER SMALL YA" },
	{ 0x30E4 , "KATAKANA LETTER YA" },
	{ 0x30E5 , "KATAKANA LETTER SMALL YU" },
	{ 0x30E6 , "KATAKANA LETTER YU" },
	{ 0x30E7 , "KATAKANA LETTER SMALL YO" },
	{ 0x30E8 , "KATAKANA LETTER YO" },
	{ 0x30E9 , "KATAKANA LETTER RA" },
	{ 0x30EA , "KATAKANA LETTER RI" },
	{ 0x30EB , "KATAKANA LETTER RU" },
	{ 0x30EC , "KATAKANA LETTER RE" },
	{ 0x30ED , "KATAKANA LETTER RO" },
	{ 0x30EE , "KATAKANA LETTER SMALL WA" },
	{ 0x30EF , "KATAKANA LETTER WA" },
	{ 0x30F0 , "KATAKANA LETTER WI" },
	{ 0x30F1 , "KATAKANA LETTER WE" },
	{ 0x30F2 , "KATAKANA LETTER WO" },
	{ 0x30F3 , "KATAKANA LETTER N" },
	{ 0x30F4 , "KATAKANA LETTER VU" },
	{ 0x30F5 , "KATAKANA LETTER SMALL KA" },
	{ 0x30F6 , "KATAKANA LETTER SMALL KE" },
	{ 0x30F7 , "KATAKANA LETTER VA" },
	{ 0x30F8 , "KATAKANA LETTER VI" },
	{ 0x30F9 , "KATAKANA LETTER VE" },
	{ 0x30FA , "KATAKANA LETTER VO" },
	{ 0x30FB , "KATAKANA MIDDLE DOT" },
	{ 0x30FC , "KATAKANA-HIRAGANA PROLONGED SOUND MARK" },
	{ 0x30FD , "KATAKANA ITERATION MARK" },
	{ 0x30FE , "KATAKANA VOICED ITERATION MARK" },
#if 0
	{ 0x3105 , "BOPOMOFO LETTER B" },
	{ 0x3106 , "BOPOMOFO LETTER P" },
	{ 0x3107 , "BOPOMOFO LETTER M" },
	{ 0x3108 , "BOPOMOFO LETTER F" },
	{ 0x3109 , "BOPOMOFO LETTER D" },
	{ 0x310A , "BOPOMOFO LETTER T" },
	{ 0x310B , "BOPOMOFO LETTER N" },
	{ 0x310C , "BOPOMOFO LETTER L" },
	{ 0x310D , "BOPOMOFO LETTER G" },
	{ 0x310E , "BOPOMOFO LETTER K" },
	{ 0x310F , "BOPOMOFO LETTER H" },
	{ 0x3110 , "BOPOMOFO LETTER J" },
	{ 0x3111 , "BOPOMOFO LETTER Q" },
	{ 0x3112 , "BOPOMOFO LETTER X" },
	{ 0x3113 , "BOPOMOFO LETTER ZH" },
	{ 0x3114 , "BOPOMOFO LETTER CH" },
	{ 0x3115 , "BOPOMOFO LETTER SH" },
	{ 0x3116 , "BOPOMOFO LETTER R" },
	{ 0x3117 , "BOPOMOFO LETTER Z" },
	{ 0x3118 , "BOPOMOFO LETTER C" },
	{ 0x3119 , "BOPOMOFO LETTER S" },
	{ 0x311A , "BOPOMOFO LETTER A" },
	{ 0x311B , "BOPOMOFO LETTER O" },
	{ 0x311C , "BOPOMOFO LETTER E" },
	{ 0x311D , "BOPOMOFO LETTER EH" },
	{ 0x311E , "BOPOMOFO LETTER AI" },
	{ 0x311F , "BOPOMOFO LETTER EI" },
	{ 0x3120 , "BOPOMOFO LETTER AU" },
	{ 0x3121 , "BOPOMOFO LETTER OU" },
	{ 0x3122 , "BOPOMOFO LETTER AN" },
	{ 0x3123 , "BOPOMOFO LETTER EN" },
	{ 0x3124 , "BOPOMOFO LETTER ANG" },
	{ 0x3125 , "BOPOMOFO LETTER ENG" },
	{ 0x3126 , "BOPOMOFO LETTER ER" },
	{ 0x3127 , "BOPOMOFO LETTER I" },
	{ 0x3128 , "BOPOMOFO LETTER U" },
	{ 0x3129 , "BOPOMOFO LETTER IU" },
	{ 0x312A , "BOPOMOFO LETTER V" },
	{ 0x312B , "BOPOMOFO LETTER NG" },
	{ 0x312C , "BOPOMOFO LETTER GN" },
	{ 0x3131 , "HANGUL LETTER KIYEOK" },
	{ 0x3132 , "HANGUL LETTER SSANGKIYEOK" },
	{ 0x3133 , "HANGUL LETTER KIYEOK-SIOS" },
	{ 0x3134 , "HANGUL LETTER NIEUN" },
	{ 0x3135 , "HANGUL LETTER NIEUN-CIEUC" },
	{ 0x3136 , "HANGUL LETTER NIEUN-HIEUH" },
	{ 0x3137 , "HANGUL LETTER TIKEUT" },
	{ 0x3138 , "HANGUL LETTER SSANGTIKEUT" },
	{ 0x3139 , "HANGUL LETTER RIEUL" },
	{ 0x313A , "HANGUL LETTER RIEUL-KIYEOK" },
	{ 0x313B , "HANGUL LETTER RIEUL-MIEUM" },
	{ 0x313C , "HANGUL LETTER RIEUL-PIEUP" },
	{ 0x313D , "HANGUL LETTER RIEUL-SIOS" },
	{ 0x313E , "HANGUL LETTER RIEUL-THIEUTH" },
	{ 0x313F , "HANGUL LETTER RIEUL-PHIEUPH" },
	{ 0x3140 , "HANGUL LETTER RIEUL-HIEUH" },
	{ 0x3141 , "HANGUL LETTER MIEUM" },
	{ 0x3142 , "HANGUL LETTER PIEUP" },
	{ 0x3143 , "HANGUL LETTER SSANGPIEUP" },
	{ 0x3144 , "HANGUL LETTER PIEUP-SIOS" },
	{ 0x3145 , "HANGUL LETTER SIOS" },
	{ 0x3146 , "HANGUL LETTER SSANGSIOS" },
	{ 0x3147 , "HANGUL LETTER IEUNG" },
	{ 0x3148 , "HANGUL LETTER CIEUC" },
	{ 0x3149 , "HANGUL LETTER SSANGCIEUC" },
	{ 0x314A , "HANGUL LETTER CHIEUCH" },
	{ 0x314B , "HANGUL LETTER KHIEUKH" },
	{ 0x314C , "HANGUL LETTER THIEUTH" },
	{ 0x314D , "HANGUL LETTER PHIEUPH" },
	{ 0x314E , "HANGUL LETTER HIEUH" },
	{ 0x314F , "HANGUL LETTER A" },
	{ 0x3150 , "HANGUL LETTER AE" },
	{ 0x3151 , "HANGUL LETTER YA" },
	{ 0x3152 , "HANGUL LETTER YAE" },
	{ 0x3153 , "HANGUL LETTER EO" },
	{ 0x3154 , "HANGUL LETTER E" },
	{ 0x3155 , "HANGUL LETTER YEO" },
	{ 0x3156 , "HANGUL LETTER YE" },
	{ 0x3157 , "HANGUL LETTER O" },
	{ 0x3158 , "HANGUL LETTER WA" },
	{ 0x3159 , "HANGUL LETTER WAE" },
	{ 0x315A , "HANGUL LETTER OE" },
	{ 0x315B , "HANGUL LETTER YO" },
	{ 0x315C , "HANGUL LETTER U" },
	{ 0x315D , "HANGUL LETTER WEO" },
	{ 0x315E , "HANGUL LETTER WE" },
	{ 0x315F , "HANGUL LETTER WI" },
	{ 0x3160 , "HANGUL LETTER YU" },
	{ 0x3161 , "HANGUL LETTER EU" },
	{ 0x3162 , "HANGUL LETTER YI" },
	{ 0x3163 , "HANGUL LETTER I" },
	{ 0x3164 , "HANGUL FILLER" },
	{ 0x3165 , "HANGUL LETTER SSANGNIEUN" },
	{ 0x3166 , "HANGUL LETTER NIEUN-TIKEUT" },
	{ 0x3167 , "HANGUL LETTER NIEUN-SIOS" },
	{ 0x3168 , "HANGUL LETTER NIEUN-PANSIOS" },
	{ 0x3169 , "HANGUL LETTER RIEUL-KIYEOK-SIOS" },
	{ 0x316A , "HANGUL LETTER RIEUL-TIKEUT" },
	{ 0x316B , "HANGUL LETTER RIEUL-PIEUP-SIOS" },
	{ 0x316C , "HANGUL LETTER RIEUL-PANSIOS" },
	{ 0x316D , "HANGUL LETTER RIEUL-YEORINHIEUH" },
	{ 0x316E , "HANGUL LETTER MIEUM-PIEUP" },
	{ 0x316F , "HANGUL LETTER MIEUM-SIOS" },
	{ 0x3170 , "HANGUL LETTER MIEUM-PANSIOS" },
	{ 0x3171 , "HANGUL LETTER KAPYEOUNMIEUM" },
	{ 0x3172 , "HANGUL LETTER PIEUP-KIYEOK" },
	{ 0x3173 , "HANGUL LETTER PIEUP-TIKEUT" },
	{ 0x3174 , "HANGUL LETTER PIEUP-SIOS-KIYEOK" },
	{ 0x3175 , "HANGUL LETTER PIEUP-SIOS-TIKEUT" },
	{ 0x3176 , "HANGUL LETTER PIEUP-CIEUC" },
	{ 0x3177 , "HANGUL LETTER PIEUP-THIEUTH" },
	{ 0x3178 , "HANGUL LETTER KAPYEOUNPIEUP" },
	{ 0x3179 , "HANGUL LETTER KAPYEOUNSSANGPIEUP" },
	{ 0x317A , "HANGUL LETTER SIOS-KIYEOK" },
	{ 0x317B , "HANGUL LETTER SIOS-NIEUN" },
	{ 0x317C , "HANGUL LETTER SIOS-TIKEUT" },
	{ 0x317D , "HANGUL LETTER SIOS-PIEUP" },
	{ 0x317E , "HANGUL LETTER SIOS-CIEUC" },
	{ 0x317F , "HANGUL LETTER PANSIOS" },
	{ 0x3180 , "HANGUL LETTER SSANGIEUNG" },
	{ 0x3181 , "HANGUL LETTER YESIEUNG" },
	{ 0x3182 , "HANGUL LETTER YESIEUNG-SIOS" },
	{ 0x3183 , "HANGUL LETTER YESIEUNG-PANSIOS" },
	{ 0x3184 , "HANGUL LETTER KAPYEOUNPHIEUPH" },
	{ 0x3185 , "HANGUL LETTER SSANGHIEUH" },
	{ 0x3186 , "HANGUL LETTER YEORINHIEUH" },
	{ 0x3187 , "HANGUL LETTER YO-YA" },
	{ 0x3188 , "HANGUL LETTER YO-YAE" },
	{ 0x3189 , "HANGUL LETTER YO-I" },
	{ 0x318A , "HANGUL LETTER YU-YEO" },
	{ 0x318B , "HANGUL LETTER YU-YE" },
	{ 0x318C , "HANGUL LETTER YU-I" },
	{ 0x318D , "HANGUL LETTER ARAEA" },
	{ 0x318E , "HANGUL LETTER ARAEAE" },
	{ 0x3190 , "IDEOGRAPHIC ANNOTATION LINKING MARK" },
	{ 0x3191 , "IDEOGRAPHIC ANNOTATION REVERSE MARK" },
	{ 0x3192 , "IDEOGRAPHIC ANNOTATION ONE MARK" },
	{ 0x3193 , "IDEOGRAPHIC ANNOTATION TWO MARK" },
	{ 0x3194 , "IDEOGRAPHIC ANNOTATION THREE MARK" },
	{ 0x3195 , "IDEOGRAPHIC ANNOTATION FOUR MARK" },
	{ 0x3196 , "IDEOGRAPHIC ANNOTATION TOP MARK" },
	{ 0x3197 , "IDEOGRAPHIC ANNOTATION MIDDLE MARK" },
	{ 0x3198 , "IDEOGRAPHIC ANNOTATION BOTTOM MARK" },
	{ 0x3199 , "IDEOGRAPHIC ANNOTATION FIRST MARK" },
	{ 0x319A , "IDEOGRAPHIC ANNOTATION SECOND MARK" },
	{ 0x319B , "IDEOGRAPHIC ANNOTATION THIRD MARK" },
	{ 0x319C , "IDEOGRAPHIC ANNOTATION FOURTH MARK" },
	{ 0x319D , "IDEOGRAPHIC ANNOTATION HEAVEN MARK" },
	{ 0x319E , "IDEOGRAPHIC ANNOTATION EARTH MARK" },
	{ 0x319F , "IDEOGRAPHIC ANNOTATION MAN MARK" },
	{ 0x3200 , "PARENTHESIZED HANGUL KIYEOK" },
	{ 0x3201 , "PARENTHESIZED HANGUL NIEUN" },
	{ 0x3202 , "PARENTHESIZED HANGUL TIKEUT" },
	{ 0x3203 , "PARENTHESIZED HANGUL RIEUL" },
	{ 0x3204 , "PARENTHESIZED HANGUL MIEUM" },
	{ 0x3205 , "PARENTHESIZED HANGUL PIEUP" },
	{ 0x3206 , "PARENTHESIZED HANGUL SIOS" },
	{ 0x3207 , "PARENTHESIZED HANGUL IEUNG" },
	{ 0x3208 , "PARENTHESIZED HANGUL CIEUC" },
	{ 0x3209 , "PARENTHESIZED HANGUL CHIEUCH" },
	{ 0x320A , "PARENTHESIZED HANGUL KHIEUKH" },
	{ 0x320B , "PARENTHESIZED HANGUL THIEUTH" },
	{ 0x320C , "PARENTHESIZED HANGUL PHIEUPH" },
	{ 0x320D , "PARENTHESIZED HANGUL HIEUH" },
	{ 0x320E , "PARENTHESIZED HANGUL KIYEOK A" },
	{ 0x320F , "PARENTHESIZED HANGUL NIEUN A" },
	{ 0x3210 , "PARENTHESIZED HANGUL TIKEUT A" },
	{ 0x3211 , "PARENTHESIZED HANGUL RIEUL A" },
	{ 0x3212 , "PARENTHESIZED HANGUL MIEUM A" },
	{ 0x3213 , "PARENTHESIZED HANGUL PIEUP A" },
	{ 0x3214 , "PARENTHESIZED HANGUL SIOS A" },
	{ 0x3215 , "PARENTHESIZED HANGUL IEUNG A" },
	{ 0x3216 , "PARENTHESIZED HANGUL CIEUC A" },
	{ 0x3217 , "PARENTHESIZED HANGUL CHIEUCH A" },
	{ 0x3218 , "PARENTHESIZED HANGUL KHIEUKH A" },
	{ 0x3219 , "PARENTHESIZED HANGUL THIEUTH A" },
	{ 0x321A , "PARENTHESIZED HANGUL PHIEUPH A" },
	{ 0x321B , "PARENTHESIZED HANGUL HIEUH A" },
	{ 0x321C , "PARENTHESIZED HANGUL CIEUC U" },
	{ 0x3220 , "PARENTHESIZED IDEOGRAPH ONE" },
	{ 0x3221 , "PARENTHESIZED IDEOGRAPH TWO" },
	{ 0x3222 , "PARENTHESIZED IDEOGRAPH THREE" },
	{ 0x3223 , "PARENTHESIZED IDEOGRAPH FOUR" },
	{ 0x3224 , "PARENTHESIZED IDEOGRAPH FIVE" },
	{ 0x3225 , "PARENTHESIZED IDEOGRAPH SIX" },
	{ 0x3226 , "PARENTHESIZED IDEOGRAPH SEVEN" },
	{ 0x3227 , "PARENTHESIZED IDEOGRAPH EIGHT" },
	{ 0x3228 , "PARENTHESIZED IDEOGRAPH NINE" },
	{ 0x3229 , "PARENTHESIZED IDEOGRAPH TEN" },
	{ 0x322A , "PARENTHESIZED IDEOGRAPH MOON" },
	{ 0x322B , "PARENTHESIZED IDEOGRAPH FIRE" },
	{ 0x322C , "PARENTHESIZED IDEOGRAPH WATER" },
	{ 0x322D , "PARENTHESIZED IDEOGRAPH WOOD" },
	{ 0x322E , "PARENTHESIZED IDEOGRAPH METAL" },
	{ 0x322F , "PARENTHESIZED IDEOGRAPH EARTH" },
	{ 0x3230 , "PARENTHESIZED IDEOGRAPH SUN" },
	{ 0x3231 , "PARENTHESIZED IDEOGRAPH STOCK" },
	{ 0x3232 , "PARENTHESIZED IDEOGRAPH HAVE" },
	{ 0x3233 , "PARENTHESIZED IDEOGRAPH SOCIETY" },
	{ 0x3234 , "PARENTHESIZED IDEOGRAPH NAME" },
	{ 0x3235 , "PARENTHESIZED IDEOGRAPH SPECIAL" },
	{ 0x3236 , "PARENTHESIZED IDEOGRAPH FINANCIAL" },
	{ 0x3237 , "PARENTHESIZED IDEOGRAPH CONGRATULATION" },
	{ 0x3238 , "PARENTHESIZED IDEOGRAPH LABOR" },
	{ 0x3239 , "PARENTHESIZED IDEOGRAPH REPRESENT" },
	{ 0x323A , "PARENTHESIZED IDEOGRAPH CALL" },
	{ 0x323B , "PARENTHESIZED IDEOGRAPH STUDY" },
	{ 0x323C , "PARENTHESIZED IDEOGRAPH SUPERVISE" },
	{ 0x323D , "PARENTHESIZED IDEOGRAPH ENTERPRISE" },
	{ 0x323E , "PARENTHESIZED IDEOGRAPH RESOURCE" },
	{ 0x323F , "PARENTHESIZED IDEOGRAPH ALLIANCE" },
	{ 0x3240 , "PARENTHESIZED IDEOGRAPH FESTIVAL" },
	{ 0x3241 , "PARENTHESIZED IDEOGRAPH REST" },
	{ 0x3242 , "PARENTHESIZED IDEOGRAPH SELF" },
	{ 0x3243 , "PARENTHESIZED IDEOGRAPH REACH" },
	{ 0x3260 , "CIRCLED HANGUL KIYEOK" },
	{ 0x3261 , "CIRCLED HANGUL NIEUN" },
	{ 0x3262 , "CIRCLED HANGUL TIKEUT" },
	{ 0x3263 , "CIRCLED HANGUL RIEUL" },
	{ 0x3264 , "CIRCLED HANGUL MIEUM" },
	{ 0x3265 , "CIRCLED HANGUL PIEUP" },
	{ 0x3266 , "CIRCLED HANGUL SIOS" },
	{ 0x3267 , "CIRCLED HANGUL IEUNG" },
	{ 0x3268 , "CIRCLED HANGUL CIEUC" },
	{ 0x3269 , "CIRCLED HANGUL CHIEUCH" },
	{ 0x326A , "CIRCLED HANGUL KHIEUKH" },
	{ 0x326B , "CIRCLED HANGUL THIEUTH" },
	{ 0x326C , "CIRCLED HANGUL PHIEUPH" },
	{ 0x326D , "CIRCLED HANGUL HIEUH" },
	{ 0x326E , "CIRCLED HANGUL KIYEOK A" },
	{ 0x326F , "CIRCLED HANGUL NIEUN A" },
	{ 0x3270 , "CIRCLED HANGUL TIKEUT A" },
	{ 0x3271 , "CIRCLED HANGUL RIEUL A" },
	{ 0x3272 , "CIRCLED HANGUL MIEUM A" },
	{ 0x3273 , "CIRCLED HANGUL PIEUP A" },
	{ 0x3274 , "CIRCLED HANGUL SIOS A" },
	{ 0x3275 , "CIRCLED HANGUL IEUNG A" },
	{ 0x3276 , "CIRCLED HANGUL CIEUC A" },
	{ 0x3277 , "CIRCLED HANGUL CHIEUCH A" },
	{ 0x3278 , "CIRCLED HANGUL KHIEUKH A" },
	{ 0x3279 , "CIRCLED HANGUL THIEUTH A" },
	{ 0x327A , "CIRCLED HANGUL PHIEUPH A" },
	{ 0x327B , "CIRCLED HANGUL HIEUH A" },
	{ 0x327F , "KOREAN STANDARD SYMBOL" },
	{ 0x3280 , "CIRCLED IDEOGRAPH ONE" },
	{ 0x3281 , "CIRCLED IDEOGRAPH TWO" },
	{ 0x3282 , "CIRCLED IDEOGRAPH THREE" },
	{ 0x3283 , "CIRCLED IDEOGRAPH FOUR" },
	{ 0x3284 , "CIRCLED IDEOGRAPH FIVE" },
	{ 0x3285 , "CIRCLED IDEOGRAPH SIX" },
	{ 0x3286 , "CIRCLED IDEOGRAPH SEVEN" },
	{ 0x3287 , "CIRCLED IDEOGRAPH EIGHT" },
	{ 0x3288 , "CIRCLED IDEOGRAPH NINE" },
	{ 0x3289 , "CIRCLED IDEOGRAPH TEN" },
	{ 0x328A , "CIRCLED IDEOGRAPH MOON" },
	{ 0x328B , "CIRCLED IDEOGRAPH FIRE" },
	{ 0x328C , "CIRCLED IDEOGRAPH WATER" },
	{ 0x328D , "CIRCLED IDEOGRAPH WOOD" },
	{ 0x328E , "CIRCLED IDEOGRAPH METAL" },
	{ 0x328F , "CIRCLED IDEOGRAPH EARTH" },
	{ 0x3290 , "CIRCLED IDEOGRAPH SUN" },
	{ 0x3291 , "CIRCLED IDEOGRAPH STOCK" },
	{ 0x3292 , "CIRCLED IDEOGRAPH HAVE" },
	{ 0x3293 , "CIRCLED IDEOGRAPH SOCIETY" },
	{ 0x3294 , "CIRCLED IDEOGRAPH NAME" },
	{ 0x3295 , "CIRCLED IDEOGRAPH SPECIAL" },
	{ 0x3296 , "CIRCLED IDEOGRAPH FINANCIAL" },
	{ 0x3297 , "CIRCLED IDEOGRAPH CONGRATULATION" },
	{ 0x3298 , "CIRCLED IDEOGRAPH LABOR" },
	{ 0x3299 , "CIRCLED IDEOGRAPH SECRET" },
	{ 0x329A , "CIRCLED IDEOGRAPH MALE" },
	{ 0x329B , "CIRCLED IDEOGRAPH FEMALE" },
	{ 0x329C , "CIRCLED IDEOGRAPH SUITABLE" },
	{ 0x329D , "CIRCLED IDEOGRAPH EXCELLENT" },
	{ 0x329E , "CIRCLED IDEOGRAPH PRINT" },
	{ 0x329F , "CIRCLED IDEOGRAPH ATTENTION" },
	{ 0x32A0 , "CIRCLED IDEOGRAPH ITEM" },
	{ 0x32A1 , "CIRCLED IDEOGRAPH REST" },
	{ 0x32A2 , "CIRCLED IDEOGRAPH COPY" },
	{ 0x32A3 , "CIRCLED IDEOGRAPH CORRECT" },
	{ 0x32A4 , "CIRCLED IDEOGRAPH HIGH" },
	{ 0x32A5 , "CIRCLED IDEOGRAPH CENTRE" },
	{ 0x32A6 , "CIRCLED IDEOGRAPH LOW" },
	{ 0x32A7 , "CIRCLED IDEOGRAPH LEFT" },
	{ 0x32A8 , "CIRCLED IDEOGRAPH RIGHT" },
	{ 0x32A9 , "CIRCLED IDEOGRAPH MEDICINE" },
	{ 0x32AA , "CIRCLED IDEOGRAPH RELIGION" },
	{ 0x32AB , "CIRCLED IDEOGRAPH STUDY" },
	{ 0x32AC , "CIRCLED IDEOGRAPH SUPERVISE" },
	{ 0x32AD , "CIRCLED IDEOGRAPH ENTERPRISE" },
	{ 0x32AE , "CIRCLED IDEOGRAPH RESOURCE" },
	{ 0x32AF , "CIRCLED IDEOGRAPH ALLIANCE" },
	{ 0x32B0 , "CIRCLED IDEOGRAPH NIGHT" },
	{ 0x32C0 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR JANUARY" },
	{ 0x32C1 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR FEBRUARY" },
	{ 0x32C2 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR MARCH" },
	{ 0x32C3 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR APRIL" },
	{ 0x32C4 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR MAY" },
	{ 0x32C5 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR JUNE" },
	{ 0x32C6 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR JULY" },
	{ 0x32C7 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR AUGUST" },
	{ 0x32C8 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR SEPTEMBER" },
	{ 0x32C9 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR OCTOBER" },
	{ 0x32CA , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR NOVEMBER" },
	{ 0x32CB , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DECEMBER" },
	{ 0x32D0 , "CIRCLED KATAKANA A" },
	{ 0x32D1 , "CIRCLED KATAKANA I" },
	{ 0x32D2 , "CIRCLED KATAKANA U" },
	{ 0x32D3 , "CIRCLED KATAKANA E" },
	{ 0x32D4 , "CIRCLED KATAKANA O" },
	{ 0x32D5 , "CIRCLED KATAKANA KA" },
	{ 0x32D6 , "CIRCLED KATAKANA KI" },
	{ 0x32D7 , "CIRCLED KATAKANA KU" },
	{ 0x32D8 , "CIRCLED KATAKANA KE" },
	{ 0x32D9 , "CIRCLED KATAKANA KO" },
	{ 0x32DA , "CIRCLED KATAKANA SA" },
	{ 0x32DB , "CIRCLED KATAKANA SI" },
	{ 0x32DC , "CIRCLED KATAKANA SU" },
	{ 0x32DD , "CIRCLED KATAKANA SE" },
	{ 0x32DE , "CIRCLED KATAKANA SO" },
	{ 0x32DF , "CIRCLED KATAKANA TA" },
	{ 0x32E0 , "CIRCLED KATAKANA TI" },
	{ 0x32E1 , "CIRCLED KATAKANA TU" },
	{ 0x32E2 , "CIRCLED KATAKANA TE" },
	{ 0x32E3 , "CIRCLED KATAKANA TO" },
	{ 0x32E4 , "CIRCLED KATAKANA NA" },
	{ 0x32E5 , "CIRCLED KATAKANA NI" },
	{ 0x32E6 , "CIRCLED KATAKANA NU" },
	{ 0x32E7 , "CIRCLED KATAKANA NE" },
	{ 0x32E8 , "CIRCLED KATAKANA NO" },
	{ 0x32E9 , "CIRCLED KATAKANA HA" },
	{ 0x32EA , "CIRCLED KATAKANA HI" },
	{ 0x32EB , "CIRCLED KATAKANA HU" },
	{ 0x32EC , "CIRCLED KATAKANA HE" },
	{ 0x32ED , "CIRCLED KATAKANA HO" },
	{ 0x32EE , "CIRCLED KATAKANA MA" },
	{ 0x32EF , "CIRCLED KATAKANA MI" },
	{ 0x32F0 , "CIRCLED KATAKANA MU" },
	{ 0x32F1 , "CIRCLED KATAKANA ME" },
	{ 0x32F2 , "CIRCLED KATAKANA MO" },
	{ 0x32F3 , "CIRCLED KATAKANA YA" },
	{ 0x32F4 , "CIRCLED KATAKANA YU" },
	{ 0x32F5 , "CIRCLED KATAKANA YO" },
	{ 0x32F6 , "CIRCLED KATAKANA RA" },
	{ 0x32F7 , "CIRCLED KATAKANA RI" },
	{ 0x32F8 , "CIRCLED KATAKANA RU" },
	{ 0x32F9 , "CIRCLED KATAKANA RE" },
	{ 0x32FA , "CIRCLED KATAKANA RO" },
	{ 0x32FB , "CIRCLED KATAKANA WA" },
	{ 0x32FC , "CIRCLED KATAKANA WI" },
	{ 0x32FD , "CIRCLED KATAKANA WE" },
	{ 0x32FE , "CIRCLED KATAKANA WO" },
	{ 0x3300 , "SQUARE APAATO" },
	{ 0x3301 , "SQUARE ARUHUA" },
	{ 0x3302 , "SQUARE ANPEA" },
	{ 0x3303 , "SQUARE AARU" },
	{ 0x3304 , "SQUARE ININGU" },
	{ 0x3305 , "SQUARE INTI" },
	{ 0x3306 , "SQUARE UON" },
	{ 0x3307 , "SQUARE ESUKUUDO" },
	{ 0x3308 , "SQUARE EEKAA" },
	{ 0x3309 , "SQUARE ONSU" },
	{ 0x330A , "SQUARE OOMU" },
	{ 0x330B , "SQUARE KAIRI" },
	{ 0x330C , "SQUARE KARATTO" },
	{ 0x330D , "SQUARE KARORII" },
	{ 0x330E , "SQUARE GARON" },
	{ 0x330F , "SQUARE GANMA" },
	{ 0x3310 , "SQUARE GIGA" },
	{ 0x3311 , "SQUARE GINII" },
	{ 0x3312 , "SQUARE KYURII" },
	{ 0x3313 , "SQUARE GIRUDAA" },
	{ 0x3314 , "SQUARE KIRO" },
	{ 0x3315 , "SQUARE KIROGURAMU" },
	{ 0x3316 , "SQUARE KIROMEETORU" },
	{ 0x3317 , "SQUARE KIROWATTO" },
	{ 0x3318 , "SQUARE GURAMU" },
	{ 0x3319 , "SQUARE GURAMUTON" },
	{ 0x331A , "SQUARE KURUZEIRO" },
	{ 0x331B , "SQUARE KUROONE" },
	{ 0x331C , "SQUARE KEESU" },
	{ 0x331D , "SQUARE KORUNA" },
	{ 0x331E , "SQUARE KOOPO" },
	{ 0x331F , "SQUARE SAIKURU" },
	{ 0x3320 , "SQUARE SANTIIMU" },
	{ 0x3321 , "SQUARE SIRINGU" },
	{ 0x3322 , "SQUARE SENTI" },
	{ 0x3323 , "SQUARE SENTO" },
	{ 0x3324 , "SQUARE DAASU" },
	{ 0x3325 , "SQUARE DESI" },
	{ 0x3326 , "SQUARE DORU" },
	{ 0x3327 , "SQUARE TON" },
	{ 0x3328 , "SQUARE NANO" },
	{ 0x3329 , "SQUARE NOTTO" },
	{ 0x332A , "SQUARE HAITU" },
	{ 0x332B , "SQUARE PAASENTO" },
	{ 0x332C , "SQUARE PAATU" },
	{ 0x332D , "SQUARE BAARERU" },
	{ 0x332E , "SQUARE PIASUTORU" },
	{ 0x332F , "SQUARE PIKURU" },
	{ 0x3330 , "SQUARE PIKO" },
	{ 0x3331 , "SQUARE BIRU" },
	{ 0x3332 , "SQUARE HUARADDO" },
	{ 0x3333 , "SQUARE HUIITO" },
	{ 0x3334 , "SQUARE BUSSYERU" },
	{ 0x3335 , "SQUARE HURAN" },
	{ 0x3336 , "SQUARE HEKUTAARU" },
	{ 0x3337 , "SQUARE PESO" },
	{ 0x3338 , "SQUARE PENIHI" },
	{ 0x3339 , "SQUARE HERUTU" },
	{ 0x333A , "SQUARE PENSU" },
	{ 0x333B , "SQUARE PEEZI" },
	{ 0x333C , "SQUARE BEETA" },
	{ 0x333D , "SQUARE POINTO" },
	{ 0x333E , "SQUARE BORUTO" },
	{ 0x333F , "SQUARE HON" },
	{ 0x3340 , "SQUARE PONDO" },
	{ 0x3341 , "SQUARE HOORU" },
	{ 0x3342 , "SQUARE HOON" },
	{ 0x3343 , "SQUARE MAIKURO" },
	{ 0x3344 , "SQUARE MAIRU" },
	{ 0x3345 , "SQUARE MAHHA" },
	{ 0x3346 , "SQUARE MARUKU" },
	{ 0x3347 , "SQUARE MANSYON" },
	{ 0x3348 , "SQUARE MIKURON" },
	{ 0x3349 , "SQUARE MIRI" },
	{ 0x334A , "SQUARE MIRIBAARU" },
	{ 0x334B , "SQUARE MEGA" },
	{ 0x334C , "SQUARE MEGATON" },
	{ 0x334D , "SQUARE MEETORU" },
	{ 0x334E , "SQUARE YAADO" },
	{ 0x334F , "SQUARE YAARU" },
	{ 0x3350 , "SQUARE YUAN" },
	{ 0x3351 , "SQUARE RITTORU" },
	{ 0x3352 , "SQUARE RIRA" },
	{ 0x3353 , "SQUARE RUPII" },
	{ 0x3354 , "SQUARE RUUBURU" },
	{ 0x3355 , "SQUARE REMU" },
	{ 0x3356 , "SQUARE RENTOGEN" },
	{ 0x3357 , "SQUARE WATTO" },
	{ 0x3358 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR ZERO" },
	{ 0x3359 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR ONE" },
	{ 0x335A , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TWO" },
	{ 0x335B , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR THREE" },
	{ 0x335C , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR FOUR" },
	{ 0x335D , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR FIVE" },
	{ 0x335E , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR SIX" },
	{ 0x335F , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR SEVEN" },
	{ 0x3360 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR EIGHT" },
	{ 0x3361 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR NINE" },
	{ 0x3362 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TEN" },
	{ 0x3363 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR ELEVEN" },
	{ 0x3364 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TWELVE" },
	{ 0x3365 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR THIRTEEN" },
	{ 0x3366 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR FOURTEEN" },
	{ 0x3367 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR FIFTEEN" },
	{ 0x3368 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR SIXTEEN" },
	{ 0x3369 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR SEVENTEEN" },
	{ 0x336A , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR EIGHTEEN" },
	{ 0x336B , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR NINETEEN" },
	{ 0x336C , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TWENTY" },
	{ 0x336D , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TWENTY-ONE" },
	{ 0x336E , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TWENTY-TWO" },
	{ 0x336F , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TWENTY-THREE" },
	{ 0x3370 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TWENTY-FOUR" },
	{ 0x3371 , "SQUARE HPA" },
	{ 0x3372 , "SQUARE DA" },
	{ 0x3373 , "SQUARE AU" },
	{ 0x3374 , "SQUARE BAR" },
	{ 0x3375 , "SQUARE OV" },
	{ 0x3376 , "SQUARE PC" },
	{ 0x337B , "SQUARE ERA NAME HEISEI" },
	{ 0x337C , "SQUARE ERA NAME SYOUWA" },
	{ 0x337D , "SQUARE ERA NAME TAISYOU" },
	{ 0x337E , "SQUARE ERA NAME MEIZI" },
	{ 0x337F , "SQUARE CORPORATION" },
	{ 0x3380 , "SQUARE PA AMPS" },
	{ 0x3381 , "SQUARE NA" },
	{ 0x3382 , "SQUARE MU A" },
	{ 0x3383 , "SQUARE MA" },
	{ 0x3384 , "SQUARE KA" },
	{ 0x3385 , "SQUARE KB" },
	{ 0x3386 , "SQUARE MB" },
	{ 0x3387 , "SQUARE GB" },
	{ 0x3388 , "SQUARE CAL" },
	{ 0x3389 , "SQUARE KCAL" },
	{ 0x338A , "SQUARE PF" },
	{ 0x338B , "SQUARE NF" },
	{ 0x338C , "SQUARE MU F" },
	{ 0x338D , "SQUARE MU G" },
	{ 0x338E , "SQUARE MG" },
	{ 0x338F , "SQUARE KG" },
	{ 0x3390 , "SQUARE HZ" },
	{ 0x3391 , "SQUARE KHZ" },
	{ 0x3392 , "SQUARE MHZ" },
	{ 0x3393 , "SQUARE GHZ" },
	{ 0x3394 , "SQUARE THZ" },
	{ 0x3395 , "SQUARE MU L" },
	{ 0x3396 , "SQUARE ML" },
	{ 0x3397 , "SQUARE DL" },
	{ 0x3398 , "SQUARE KL" },
	{ 0x3399 , "SQUARE FM" },
	{ 0x339A , "SQUARE NM" },
	{ 0x339B , "SQUARE MU M" },
	{ 0x339C , "SQUARE MM" },
	{ 0x339D , "SQUARE CM" },
	{ 0x339E , "SQUARE KM" },
	{ 0x339F , "SQUARE MM SQUARED" },
	{ 0x33A0 , "SQUARE CM SQUARED" },
	{ 0x33A1 , "SQUARE M SQUARED" },
	{ 0x33A2 , "SQUARE KM SQUARED" },
	{ 0x33A3 , "SQUARE MM CUBED" },
	{ 0x33A4 , "SQUARE CM CUBED" },
	{ 0x33A5 , "SQUARE M CUBED" },
	{ 0x33A6 , "SQUARE KM CUBED" },
	{ 0x33A7 , "SQUARE M OVER S" },
	{ 0x33A8 , "SQUARE M OVER S SQUARED" },
	{ 0x33A9 , "SQUARE PA" },
	{ 0x33AA , "SQUARE KPA" },
	{ 0x33AB , "SQUARE MPA" },
	{ 0x33AC , "SQUARE GPA" },
	{ 0x33AD , "SQUARE RAD" },
	{ 0x33AE , "SQUARE RAD OVER S" },
	{ 0x33AF , "SQUARE RAD OVER S SQUARED" },
	{ 0x33B0 , "SQUARE PS" },
	{ 0x33B1 , "SQUARE NS" },
	{ 0x33B2 , "SQUARE MU S" },
	{ 0x33B3 , "SQUARE MS" },
	{ 0x33B4 , "SQUARE PV" },
	{ 0x33B5 , "SQUARE NV" },
	{ 0x33B6 , "SQUARE MU V" },
	{ 0x33B7 , "SQUARE MV" },
	{ 0x33B8 , "SQUARE KV" },
	{ 0x33B9 , "SQUARE MV MEGA" },
	{ 0x33BA , "SQUARE PW" },
	{ 0x33BB , "SQUARE NW" },
	{ 0x33BC , "SQUARE MU W" },
	{ 0x33BD , "SQUARE MW" },
	{ 0x33BE , "SQUARE KW" },
	{ 0x33BF , "SQUARE MW MEGA" },
	{ 0x33C0 , "SQUARE K OHM" },
	{ 0x33C1 , "SQUARE M OHM" },
	{ 0x33C2 , "SQUARE AM" },
	{ 0x33C3 , "SQUARE BQ" },
	{ 0x33C4 , "SQUARE CC" },
	{ 0x33C5 , "SQUARE CD" },
	{ 0x33C6 , "SQUARE C OVER KG" },
	{ 0x33C7 , "SQUARE CO" },
	{ 0x33C8 , "SQUARE DB" },
	{ 0x33C9 , "SQUARE GY" },
	{ 0x33CA , "SQUARE HA" },
	{ 0x33CB , "SQUARE HP" },
	{ 0x33CC , "SQUARE IN" },
	{ 0x33CD , "SQUARE KK" },
	{ 0x33CE , "SQUARE KM CAPITAL" },
	{ 0x33CF , "SQUARE KT" },
	{ 0x33D0 , "SQUARE LM" },
	{ 0x33D1 , "SQUARE LN" },
	{ 0x33D2 , "SQUARE LOG" },
	{ 0x33D3 , "SQUARE LX" },
	{ 0x33D4 , "SQUARE MB SMALL" },
	{ 0x33D5 , "SQUARE MIL" },
	{ 0x33D6 , "SQUARE MOL" },
	{ 0x33D7 , "SQUARE PH" },
	{ 0x33D8 , "SQUARE PM" },
	{ 0x33D9 , "SQUARE PPM" },
	{ 0x33DA , "SQUARE PR" },
	{ 0x33DB , "SQUARE SR" },
	{ 0x33DC , "SQUARE SV" },
	{ 0x33DD , "SQUARE WB" },
	{ 0x33E0 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY ONE" },
	{ 0x33E1 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWO" },
	{ 0x33E2 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY THREE" },
	{ 0x33E3 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY FOUR" },
	{ 0x33E4 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY FIVE" },
	{ 0x33E5 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY SIX" },
	{ 0x33E6 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY SEVEN" },
	{ 0x33E7 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY EIGHT" },
	{ 0x33E8 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY NINE" },
	{ 0x33E9 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TEN" },
	{ 0x33EA , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY ELEVEN" },
	{ 0x33EB , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWELVE" },
	{ 0x33EC , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY THIRTEEN" },
	{ 0x33ED , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY FOURTEEN" },
	{ 0x33EE , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY FIFTEEN" },
	{ 0x33EF , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY SIXTEEN" },
	{ 0x33F0 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY SEVENTEEN" },
	{ 0x33F1 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY EIGHTEEN" },
	{ 0x33F2 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY NINETEEN" },
	{ 0x33F3 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY" },
	{ 0x33F4 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-ONE" },
	{ 0x33F5 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-TWO" },
	{ 0x33F6 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-THREE" },
	{ 0x33F7 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-FOUR" },
	{ 0x33F8 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-FIVE" },
	{ 0x33F9 , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-SIX" },
	{ 0x33FA , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-SEVEN" },
	{ 0x33FB , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-EIGHT" },
	{ 0x33FC , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-NINE" },
	{ 0x33FD , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY THIRTY" },
	{ 0x33FE , "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY THIRTY-ONE" },
	{ 0x3400 , "HANGUL SYLLABLE KIYEOK A" },
	{ 0x3401 , "HANGUL SYLLABLE KIYEOK A KIYEOK" },
	{ 0x3402 , "HANGUL SYLLABLE KIYEOK A NIEUN" },
	{ 0x3403 , "HANGUL SYLLABLE KIYEOK A TIKEUT" },
	{ 0x3404 , "HANGUL SYLLABLE KIYEOK A RIEUL" },
	{ 0x3405 , "HANGUL SYLLABLE KIYEOK A RIEUL-KIYEOK" },
	{ 0x3406 , "HANGUL SYLLABLE KIYEOK A RIEUL-MIEUM" },
	{ 0x3407 , "HANGUL SYLLABLE KIYEOK A MIEUM" },
	{ 0x3408 , "HANGUL SYLLABLE KIYEOK A PIEUP" },
	{ 0x3409 , "HANGUL SYLLABLE KIYEOK A PIEUP-SIOS" },
	{ 0x340A , "HANGUL SYLLABLE KIYEOK A SIOS" },
	{ 0x340B , "HANGUL SYLLABLE KIYEOK A SSANGSIOS" },
	{ 0x340C , "HANGUL SYLLABLE KIYEOK A IEUNG" },
	{ 0x340D , "HANGUL SYLLABLE KIYEOK A CIEUC" },
	{ 0x340E , "HANGUL SYLLABLE KIYEOK A CHIEUCH" },
	{ 0x340F , "HANGUL SYLLABLE KIYEOK A THIEUTH" },
	{ 0x3410 , "HANGUL SYLLABLE KIYEOK A PHIEUPH" },
	{ 0x3411 , "HANGUL SYLLABLE KIYEOK A HIEUH" },
	{ 0x3412 , "HANGUL SYLLABLE KIYEOK AE" },
	{ 0x3413 , "HANGUL SYLLABLE KIYEOK AE KIYEOK" },
	{ 0x3414 , "HANGUL SYLLABLE KIYEOK AE NIEUN" },
	{ 0x3415 , "HANGUL SYLLABLE KIYEOK AE RIEUL" },
	{ 0x3416 , "HANGUL SYLLABLE KIYEOK AE MIEUM" },
	{ 0x3417 , "HANGUL SYLLABLE KIYEOK AE PIEUP" },
	{ 0x3418 , "HANGUL SYLLABLE KIYEOK AE SIOS" },
	{ 0x3419 , "HANGUL SYLLABLE KIYEOK AE SSANGSIOS" },
	{ 0x341A , "HANGUL SYLLABLE KIYEOK AE IEUNG" },
	{ 0x341B , "HANGUL SYLLABLE KIYEOK YA" },
	{ 0x341C , "HANGUL SYLLABLE KIYEOK YA KIYEOK" },
	{ 0x341D , "HANGUL SYLLABLE KIYEOK YA NIEUN" },
	{ 0x341E , "HANGUL SYLLABLE KIYEOK YA RIEUL" },
	{ 0x341F , "HANGUL SYLLABLE KIYEOK YA SIOS" },
	{ 0x3420 , "HANGUL SYLLABLE KIYEOK YA IEUNG" },
	{ 0x3421 , "HANGUL SYLLABLE KIYEOK YAE" },
	{ 0x3422 , "HANGUL SYLLABLE KIYEOK YAE NIEUN" },
	{ 0x3423 , "HANGUL SYLLABLE KIYEOK YAE RIEUL" },
	{ 0x3424 , "HANGUL SYLLABLE KIYEOK EO" },
	{ 0x3425 , "HANGUL SYLLABLE KIYEOK EO KIYEOK" },
	{ 0x3426 , "HANGUL SYLLABLE KIYEOK EO NIEUN" },
	{ 0x3427 , "HANGUL SYLLABLE KIYEOK EO TIKEUT" },
	{ 0x3428 , "HANGUL SYLLABLE KIYEOK EO RIEUL" },
	{ 0x3429 , "HANGUL SYLLABLE KIYEOK EO RIEUL-MIEUM" },
	{ 0x342A , "HANGUL SYLLABLE KIYEOK EO MIEUM" },
	{ 0x342B , "HANGUL SYLLABLE KIYEOK EO PIEUP" },
	{ 0x342C , "HANGUL SYLLABLE KIYEOK EO SIOS" },
	{ 0x342D , "HANGUL SYLLABLE KIYEOK EO SSANGSIOS" },
	{ 0x342E , "HANGUL SYLLABLE KIYEOK EO IEUNG" },
	{ 0x342F , "HANGUL SYLLABLE KIYEOK EO CIEUC" },
	{ 0x3430 , "HANGUL SYLLABLE KIYEOK EO THIEUTH" },
	{ 0x3431 , "HANGUL SYLLABLE KIYEOK EO PHIEUPH" },
	{ 0x3432 , "HANGUL SYLLABLE KIYEOK EO HIEUH" },
	{ 0x3433 , "HANGUL SYLLABLE KIYEOK E" },
	{ 0x3434 , "HANGUL SYLLABLE KIYEOK E NIEUN" },
	{ 0x3435 , "HANGUL SYLLABLE KIYEOK E RIEUL" },
	{ 0x3436 , "HANGUL SYLLABLE KIYEOK E MIEUM" },
	{ 0x3437 , "HANGUL SYLLABLE KIYEOK E PIEUP" },
	{ 0x3438 , "HANGUL SYLLABLE KIYEOK E SIOS" },
	{ 0x3439 , "HANGUL SYLLABLE KIYEOK E SSANGSIOS" },
	{ 0x343A , "HANGUL SYLLABLE KIYEOK E IEUNG" },
	{ 0x343B , "HANGUL SYLLABLE KIYEOK YEO" },
	{ 0x343C , "HANGUL SYLLABLE KIYEOK YEO KIYEOK" },
	{ 0x343D , "HANGUL SYLLABLE KIYEOK YEO SSANGKIYEOK" },
	{ 0x343E , "HANGUL SYLLABLE KIYEOK YEO NIEUN" },
	{ 0x343F , "HANGUL SYLLABLE KIYEOK YEO TIKEUT" },
	{ 0x3440 , "HANGUL SYLLABLE KIYEOK YEO RIEUL" },
	{ 0x3441 , "HANGUL SYLLABLE KIYEOK YEO MIEUM" },
	{ 0x3442 , "HANGUL SYLLABLE KIYEOK YEO PIEUP" },
	{ 0x3443 , "HANGUL SYLLABLE KIYEOK YEO SIOS" },
	{ 0x3444 , "HANGUL SYLLABLE KIYEOK YEO SSANGSIOS" },
	{ 0x3445 , "HANGUL SYLLABLE KIYEOK YEO IEUNG" },
	{ 0x3446 , "HANGUL SYLLABLE KIYEOK YEO THIEUTH" },
	{ 0x3447 , "HANGUL SYLLABLE KIYEOK YE" },
	{ 0x3448 , "HANGUL SYLLABLE KIYEOK YE NIEUN" },
	{ 0x3449 , "HANGUL SYLLABLE KIYEOK YE RIEUL" },
	{ 0x344A , "HANGUL SYLLABLE KIYEOK YE PIEUP" },
	{ 0x344B , "HANGUL SYLLABLE KIYEOK YE SIOS" },
	{ 0x344C , "HANGUL SYLLABLE KIYEOK O" },
	{ 0x344D , "HANGUL SYLLABLE KIYEOK O KIYEOK" },
	{ 0x344E , "HANGUL SYLLABLE KIYEOK O NIEUN" },
	{ 0x344F , "HANGUL SYLLABLE KIYEOK O TIKEUT" },
	{ 0x3450 , "HANGUL SYLLABLE KIYEOK O RIEUL" },
	{ 0x3451 , "HANGUL SYLLABLE KIYEOK O RIEUL-MIEUM" },
	{ 0x3452 , "HANGUL SYLLABLE KIYEOK O RIEUL-SIOS" },
	{ 0x3453 , "HANGUL SYLLABLE KIYEOK O RIEUL-HIEUH" },
	{ 0x3454 , "HANGUL SYLLABLE KIYEOK O MIEUM" },
	{ 0x3455 , "HANGUL SYLLABLE KIYEOK O PIEUP" },
	{ 0x3456 , "HANGUL SYLLABLE KIYEOK O SIOS" },
	{ 0x3457 , "HANGUL SYLLABLE KIYEOK O IEUNG" },
	{ 0x3458 , "HANGUL SYLLABLE KIYEOK O CIEUC" },
	{ 0x3459 , "HANGUL SYLLABLE KIYEOK WA" },
	{ 0x345A , "HANGUL SYLLABLE KIYEOK WA KIYEOK" },
	{ 0x345B , "HANGUL SYLLABLE KIYEOK WA NIEUN" },
	{ 0x345C , "HANGUL SYLLABLE KIYEOK WA RIEUL" },
	{ 0x345D , "HANGUL SYLLABLE KIYEOK WA RIEUL-MIEUM" },
	{ 0x345E , "HANGUL SYLLABLE KIYEOK WA MIEUM" },
	{ 0x345F , "HANGUL SYLLABLE KIYEOK WA PIEUP" },
	{ 0x3460 , "HANGUL SYLLABLE KIYEOK WA SIOS" },
	{ 0x3461 , "HANGUL SYLLABLE KIYEOK WA IEUNG" },
	{ 0x3462 , "HANGUL SYLLABLE KIYEOK WAE" },
	{ 0x3463 , "HANGUL SYLLABLE KIYEOK WAE NIEUN" },
	{ 0x3464 , "HANGUL SYLLABLE KIYEOK WAE RIEUL" },
	{ 0x3465 , "HANGUL SYLLABLE KIYEOK WAE PIEUP" },
	{ 0x3466 , "HANGUL SYLLABLE KIYEOK WAE SSANGSIOS" },
	{ 0x3467 , "HANGUL SYLLABLE KIYEOK WAE IEUNG" },
	{ 0x3468 , "HANGUL SYLLABLE KIYEOK OE" },
	{ 0x3469 , "HANGUL SYLLABLE KIYEOK OE KIYEOK" },
	{ 0x346A , "HANGUL SYLLABLE KIYEOK OE NIEUN" },
	{ 0x346B , "HANGUL SYLLABLE KIYEOK OE RIEUL" },
	{ 0x346C , "HANGUL SYLLABLE KIYEOK OE MIEUM" },
	{ 0x346D , "HANGUL SYLLABLE KIYEOK OE PIEUP" },
	{ 0x346E , "HANGUL SYLLABLE KIYEOK OE SIOS" },
	{ 0x346F , "HANGUL SYLLABLE KIYEOK OE IEUNG" },
	{ 0x3470 , "HANGUL SYLLABLE KIYEOK YO" },
	{ 0x3471 , "HANGUL SYLLABLE KIYEOK YO NIEUN" },
	{ 0x3472 , "HANGUL SYLLABLE KIYEOK YO RIEUL" },
	{ 0x3473 , "HANGUL SYLLABLE KIYEOK YO PIEUP" },
	{ 0x3474 , "HANGUL SYLLABLE KIYEOK YO SIOS" },
	{ 0x3475 , "HANGUL SYLLABLE KIYEOK U" },
	{ 0x3476 , "HANGUL SYLLABLE KIYEOK U KIYEOK" },
	{ 0x3477 , "HANGUL SYLLABLE KIYEOK U NIEUN" },
	{ 0x3478 , "HANGUL SYLLABLE KIYEOK U TIKEUT" },
	{ 0x3479 , "HANGUL SYLLABLE KIYEOK U RIEUL" },
	{ 0x347A , "HANGUL SYLLABLE KIYEOK U RIEUL-KIYEOK" },
	{ 0x347B , "HANGUL SYLLABLE KIYEOK U RIEUL-MIEUM" },
	{ 0x347C , "HANGUL SYLLABLE KIYEOK U RIEUL-HIEUH" },
	{ 0x347D , "HANGUL SYLLABLE KIYEOK U MIEUM" },
	{ 0x347E , "HANGUL SYLLABLE KIYEOK U PIEUP" },
	{ 0x347F , "HANGUL SYLLABLE KIYEOK U SIOS" },
	{ 0x3480 , "HANGUL SYLLABLE KIYEOK U IEUNG" },
	{ 0x3481 , "HANGUL SYLLABLE KIYEOK U CIEUC" },
	{ 0x3482 , "HANGUL SYLLABLE KIYEOK WEO" },
	{ 0x3483 , "HANGUL SYLLABLE KIYEOK WEO KIYEOK" },
	{ 0x3484 , "HANGUL SYLLABLE KIYEOK WEO NIEUN" },
	{ 0x3485 , "HANGUL SYLLABLE KIYEOK WEO RIEUL" },
	{ 0x3486 , "HANGUL SYLLABLE KIYEOK WEO SSANGSIOS" },
	{ 0x3487 , "HANGUL SYLLABLE KIYEOK WEO IEUNG" },
	{ 0x3488 , "HANGUL SYLLABLE KIYEOK WE" },
	{ 0x3489 , "HANGUL SYLLABLE KIYEOK WE SIOS" },
	{ 0x348A , "HANGUL SYLLABLE KIYEOK WI" },
	{ 0x348B , "HANGUL SYLLABLE KIYEOK WI KIYEOK" },
	{ 0x348C , "HANGUL SYLLABLE KIYEOK WI NIEUN" },
	{ 0x348D , "HANGUL SYLLABLE KIYEOK WI RIEUL" },
	{ 0x348E , "HANGUL SYLLABLE KIYEOK WI MIEUM" },
	{ 0x348F , "HANGUL SYLLABLE KIYEOK WI PIEUP" },
	{ 0x3490 , "HANGUL SYLLABLE KIYEOK WI SIOS" },
	{ 0x3491 , "HANGUL SYLLABLE KIYEOK YU" },
	{ 0x3492 , "HANGUL SYLLABLE KIYEOK YU NIEUN" },
	{ 0x3493 , "HANGUL SYLLABLE KIYEOK YU RIEUL" },
	{ 0x3494 , "HANGUL SYLLABLE KIYEOK EU" },
	{ 0x3495 , "HANGUL SYLLABLE KIYEOK EU KIYEOK" },
	{ 0x3496 , "HANGUL SYLLABLE KIYEOK EU NIEUN" },
	{ 0x3497 , "HANGUL SYLLABLE KIYEOK EU TIKEUT" },
	{ 0x3498 , "HANGUL SYLLABLE KIYEOK EU RIEUL" },
	{ 0x3499 , "HANGUL SYLLABLE KIYEOK EU RIEUL-KIYEOK" },
	{ 0x349A , "HANGUL SYLLABLE KIYEOK EU MIEUM" },
	{ 0x349B , "HANGUL SYLLABLE KIYEOK EU PIEUP" },
	{ 0x349C , "HANGUL SYLLABLE KIYEOK EU SIOS" },
	{ 0x349D , "HANGUL SYLLABLE KIYEOK EU IEUNG" },
	{ 0x349E , "HANGUL SYLLABLE KIYEOK YI" },
	{ 0x349F , "HANGUL SYLLABLE KIYEOK I" },
	{ 0x34A0 , "HANGUL SYLLABLE KIYEOK I KIYEOK" },
	{ 0x34A1 , "HANGUL SYLLABLE KIYEOK I NIEUN" },
	{ 0x34A2 , "HANGUL SYLLABLE KIYEOK I TIKEUT" },
	{ 0x34A3 , "HANGUL SYLLABLE KIYEOK I RIEUL" },
	{ 0x34A4 , "HANGUL SYLLABLE KIYEOK I RIEUL-MIEUM" },
	{ 0x34A5 , "HANGUL SYLLABLE KIYEOK I MIEUM" },
	{ 0x34A6 , "HANGUL SYLLABLE KIYEOK I PIEUP" },
	{ 0x34A7 , "HANGUL SYLLABLE KIYEOK I SIOS" },
	{ 0x34A8 , "HANGUL SYLLABLE KIYEOK I IEUNG" },
	{ 0x34A9 , "HANGUL SYLLABLE KIYEOK I CIEUC" },
	{ 0x34AA , "HANGUL SYLLABLE KIYEOK I PHIEUPH" },
	{ 0x34AB , "HANGUL SYLLABLE SSANGKIYEOK A" },
	{ 0x34AC , "HANGUL SYLLABLE SSANGKIYEOK A KIYEOK" },
	{ 0x34AD , "HANGUL SYLLABLE SSANGKIYEOK A SSANGKIYEOK" },
	{ 0x34AE , "HANGUL SYLLABLE SSANGKIYEOK A NIEUN" },
	{ 0x34AF , "HANGUL SYLLABLE SSANGKIYEOK A RIEUL" },
	{ 0x34B0 , "HANGUL SYLLABLE SSANGKIYEOK A RIEUL-MIEUM" },
	{ 0x34B1 , "HANGUL SYLLABLE SSANGKIYEOK A MIEUM" },
	{ 0x34B2 , "HANGUL SYLLABLE SSANGKIYEOK A PIEUP" },
	{ 0x34B3 , "HANGUL SYLLABLE SSANGKIYEOK A SIOS" },
	{ 0x34B4 , "HANGUL SYLLABLE SSANGKIYEOK A SSANGSIOS" },
	{ 0x34B5 , "HANGUL SYLLABLE SSANGKIYEOK A IEUNG" },
	{ 0x34B6 , "HANGUL SYLLABLE SSANGKIYEOK A THIEUTH" },
	{ 0x34B7 , "HANGUL SYLLABLE SSANGKIYEOK AE" },
	{ 0x34B8 , "HANGUL SYLLABLE SSANGKIYEOK AE KIYEOK" },
	{ 0x34B9 , "HANGUL SYLLABLE SSANGKIYEOK AE NIEUN" },
	{ 0x34BA , "HANGUL SYLLABLE SSANGKIYEOK AE RIEUL" },
	{ 0x34BB , "HANGUL SYLLABLE SSANGKIYEOK AE MIEUM" },
	{ 0x34BC , "HANGUL SYLLABLE SSANGKIYEOK AE PIEUP" },
	{ 0x34BD , "HANGUL SYLLABLE SSANGKIYEOK AE SIOS" },
	{ 0x34BE , "HANGUL SYLLABLE SSANGKIYEOK AE SSANGSIOS" },
	{ 0x34BF , "HANGUL SYLLABLE SSANGKIYEOK AE IEUNG" },
	{ 0x34C0 , "HANGUL SYLLABLE SSANGKIYEOK YA" },
	{ 0x34C1 , "HANGUL SYLLABLE SSANGKIYEOK YA KIYEOK" },
	{ 0x34C2 , "HANGUL SYLLABLE SSANGKIYEOK YA RIEUL" },
	{ 0x34C3 , "HANGUL SYLLABLE SSANGKIYEOK EO" },
	{ 0x34C4 , "HANGUL SYLLABLE SSANGKIYEOK EO KIYEOK" },
	{ 0x34C5 , "HANGUL SYLLABLE SSANGKIYEOK EO SSANGKIYEOK" },
	{ 0x34C6 , "HANGUL SYLLABLE SSANGKIYEOK EO NIEUN" },
	{ 0x34C7 , "HANGUL SYLLABLE SSANGKIYEOK EO RIEUL" },
	{ 0x34C8 , "HANGUL SYLLABLE SSANGKIYEOK EO MIEUM" },
	{ 0x34C9 , "HANGUL SYLLABLE SSANGKIYEOK EO PIEUP" },
	{ 0x34CA , "HANGUL SYLLABLE SSANGKIYEOK EO SIOS" },
	{ 0x34CB , "HANGUL SYLLABLE SSANGKIYEOK EO SSANGSIOS" },
	{ 0x34CC , "HANGUL SYLLABLE SSANGKIYEOK EO IEUNG" },
	{ 0x34CD , "HANGUL SYLLABLE SSANGKIYEOK E" },
	{ 0x34CE , "HANGUL SYLLABLE SSANGKIYEOK E KIYEOK" },
	{ 0x34CF , "HANGUL SYLLABLE SSANGKIYEOK E NIEUN" },
	{ 0x34D0 , "HANGUL SYLLABLE SSANGKIYEOK E MIEUM" },
	{ 0x34D1 , "HANGUL SYLLABLE SSANGKIYEOK E SIOS" },
	{ 0x34D2 , "HANGUL SYLLABLE SSANGKIYEOK E IEUNG" },
	{ 0x34D3 , "HANGUL SYLLABLE SSANGKIYEOK YEO" },
	{ 0x34D4 , "HANGUL SYLLABLE SSANGKIYEOK YEO NIEUN" },
	{ 0x34D5 , "HANGUL SYLLABLE SSANGKIYEOK YEO RIEUL" },
	{ 0x34D6 , "HANGUL SYLLABLE SSANGKIYEOK YEO SIOS" },
	{ 0x34D7 , "HANGUL SYLLABLE SSANGKIYEOK YEO SSANGSIOS" },
	{ 0x34D8 , "HANGUL SYLLABLE SSANGKIYEOK YEO THIEUTH" },
	{ 0x34D9 , "HANGUL SYLLABLE SSANGKIYEOK YE" },
	{ 0x34DA , "HANGUL SYLLABLE SSANGKIYEOK O" },
	{ 0x34DB , "HANGUL SYLLABLE SSANGKIYEOK O KIYEOK" },
	{ 0x34DC , "HANGUL SYLLABLE SSANGKIYEOK O NIEUN" },
	{ 0x34DD , "HANGUL SYLLABLE SSANGKIYEOK O NIEUN-HIEUH" },
	{ 0x34DE , "HANGUL SYLLABLE SSANGKIYEOK O RIEUL" },
	{ 0x34DF , "HANGUL SYLLABLE SSANGKIYEOK O MIEUM" },
	{ 0x34E0 , "HANGUL SYLLABLE SSANGKIYEOK O PIEUP" },
	{ 0x34E1 , "HANGUL SYLLABLE SSANGKIYEOK O SIOS" },
	{ 0x34E2 , "HANGUL SYLLABLE SSANGKIYEOK O IEUNG" },
	{ 0x34E3 , "HANGUL SYLLABLE SSANGKIYEOK O CIEUC" },
	{ 0x34E4 , "HANGUL SYLLABLE SSANGKIYEOK O CHIEUCH" },
	{ 0x34E5 , "HANGUL SYLLABLE SSANGKIYEOK WA" },
	{ 0x34E6 , "HANGUL SYLLABLE SSANGKIYEOK WA KIYEOK" },
	{ 0x34E7 , "HANGUL SYLLABLE SSANGKIYEOK WA RIEUL" },
	{ 0x34E8 , "HANGUL SYLLABLE SSANGKIYEOK WA SSANGSIOS" },
	{ 0x34E9 , "HANGUL SYLLABLE SSANGKIYEOK WA IEUNG" },
	{ 0x34EA , "HANGUL SYLLABLE SSANGKIYEOK WAE" },
	{ 0x34EB , "HANGUL SYLLABLE SSANGKIYEOK WAE KIYEOK" },
	{ 0x34EC , "HANGUL SYLLABLE SSANGKIYEOK WAE IEUNG" },
	{ 0x34ED , "HANGUL SYLLABLE SSANGKIYEOK OE" },
	{ 0x34EE , "HANGUL SYLLABLE SSANGKIYEOK OE NIEUN" },
	{ 0x34EF , "HANGUL SYLLABLE SSANGKIYEOK OE RIEUL" },
	{ 0x34F0 , "HANGUL SYLLABLE SSANGKIYEOK OE MIEUM" },
	{ 0x34F1 , "HANGUL SYLLABLE SSANGKIYEOK OE PIEUP" },
	{ 0x34F2 , "HANGUL SYLLABLE SSANGKIYEOK OE IEUNG" },
	{ 0x34F3 , "HANGUL SYLLABLE SSANGKIYEOK YO" },
	{ 0x34F4 , "HANGUL SYLLABLE SSANGKIYEOK U" },
	{ 0x34F5 , "HANGUL SYLLABLE SSANGKIYEOK U KIYEOK" },
	{ 0x34F6 , "HANGUL SYLLABLE SSANGKIYEOK U NIEUN" },
	{ 0x34F7 , "HANGUL SYLLABLE SSANGKIYEOK U RIEUL" },
	{ 0x34F8 , "HANGUL SYLLABLE SSANGKIYEOK U RIEUL-HIEUH" },
	{ 0x34F9 , "HANGUL SYLLABLE SSANGKIYEOK U MIEUM" },
	{ 0x34FA , "HANGUL SYLLABLE SSANGKIYEOK U PIEUP" },
	{ 0x34FB , "HANGUL SYLLABLE SSANGKIYEOK U SIOS" },
	{ 0x34FC , "HANGUL SYLLABLE SSANGKIYEOK U IEUNG" },
	{ 0x34FD , "HANGUL SYLLABLE SSANGKIYEOK U CIEUC" },
	{ 0x34FE , "HANGUL SYLLABLE SSANGKIYEOK WEO" },
	{ 0x34FF , "HANGUL SYLLABLE SSANGKIYEOK WEO RIEUL" },
	{ 0x3500 , "HANGUL SYLLABLE SSANGKIYEOK WEO SSANGSIOS" },
	{ 0x3501 , "HANGUL SYLLABLE SSANGKIYEOK WEO IEUNG" },
	{ 0x3502 , "HANGUL SYLLABLE SSANGKIYEOK WE" },
	{ 0x3503 , "HANGUL SYLLABLE SSANGKIYEOK WE KIYEOK" },
	{ 0x3504 , "HANGUL SYLLABLE SSANGKIYEOK WE NIEUN" },
	{ 0x3505 , "HANGUL SYLLABLE SSANGKIYEOK WE RIEUL" },
	{ 0x3506 , "HANGUL SYLLABLE SSANGKIYEOK WE MIEUM" },
	{ 0x3507 , "HANGUL SYLLABLE SSANGKIYEOK WE PIEUP" },
	{ 0x3508 , "HANGUL SYLLABLE SSANGKIYEOK WE SSANGSIOS" },
	{ 0x3509 , "HANGUL SYLLABLE SSANGKIYEOK WI" },
	{ 0x350A , "HANGUL SYLLABLE SSANGKIYEOK WI NIEUN" },
	{ 0x350B , "HANGUL SYLLABLE SSANGKIYEOK WI RIEUL" },
	{ 0x350C , "HANGUL SYLLABLE SSANGKIYEOK WI MIEUM" },
	{ 0x350D , "HANGUL SYLLABLE SSANGKIYEOK WI PIEUP" },
	{ 0x350E , "HANGUL SYLLABLE SSANGKIYEOK YU" },
	{ 0x350F , "HANGUL SYLLABLE SSANGKIYEOK EU" },
	{ 0x3510 , "HANGUL SYLLABLE SSANGKIYEOK EU KIYEOK" },
	{ 0x3511 , "HANGUL SYLLABLE SSANGKIYEOK EU NIEUN" },
	{ 0x3512 , "HANGUL SYLLABLE SSANGKIYEOK EU NIEUN-HIEUH" },
	{ 0x3513 , "HANGUL SYLLABLE SSANGKIYEOK EU RIEUL" },
	{ 0x3514 , "HANGUL SYLLABLE SSANGKIYEOK EU RIEUL-MIEUM" },
	{ 0x3515 , "HANGUL SYLLABLE SSANGKIYEOK EU RIEUL-HIEUH" },
	{ 0x3516 , "HANGUL SYLLABLE SSANGKIYEOK EU MIEUM" },
	{ 0x3517 , "HANGUL SYLLABLE SSANGKIYEOK EU PIEUP" },
	{ 0x3518 , "HANGUL SYLLABLE SSANGKIYEOK EU SIOS" },
	{ 0x3519 , "HANGUL SYLLABLE SSANGKIYEOK EU IEUNG" },
	{ 0x351A , "HANGUL SYLLABLE SSANGKIYEOK EU THIEUTH" },
	{ 0x351B , "HANGUL SYLLABLE SSANGKIYEOK I" },
	{ 0x351C , "HANGUL SYLLABLE SSANGKIYEOK I KIYEOK" },
	{ 0x351D , "HANGUL SYLLABLE SSANGKIYEOK I NIEUN" },
	{ 0x351E , "HANGUL SYLLABLE SSANGKIYEOK I RIEUL" },
	{ 0x351F , "HANGUL SYLLABLE SSANGKIYEOK I MIEUM" },
	{ 0x3520 , "HANGUL SYLLABLE SSANGKIYEOK I PIEUP" },
	{ 0x3521 , "HANGUL SYLLABLE SSANGKIYEOK I SIOS" },
	{ 0x3522 , "HANGUL SYLLABLE SSANGKIYEOK I IEUNG" },
	{ 0x3523 , "HANGUL SYLLABLE NIEUN A" },
	{ 0x3524 , "HANGUL SYLLABLE NIEUN A KIYEOK" },
	{ 0x3525 , "HANGUL SYLLABLE NIEUN A SSANGKIYEOK" },
	{ 0x3526 , "HANGUL SYLLABLE NIEUN A NIEUN" },
	{ 0x3527 , "HANGUL SYLLABLE NIEUN A TIKEUT" },
	{ 0x3528 , "HANGUL SYLLABLE NIEUN A RIEUL" },
	{ 0x3529 , "HANGUL SYLLABLE NIEUN A RIEUL-KIYEOK" },
	{ 0x352A , "HANGUL SYLLABLE NIEUN A RIEUL-MIEUM" },
	{ 0x352B , "HANGUL SYLLABLE NIEUN A MIEUM" },
	{ 0x352C , "HANGUL SYLLABLE NIEUN A PIEUP" },
	{ 0x352D , "HANGUL SYLLABLE NIEUN A SIOS" },
	{ 0x352E , "HANGUL SYLLABLE NIEUN A SSANGSIOS" },
	{ 0x352F , "HANGUL SYLLABLE NIEUN A IEUNG" },
	{ 0x3530 , "HANGUL SYLLABLE NIEUN A CIEUC" },
	{ 0x3531 , "HANGUL SYLLABLE NIEUN A CHIEUCH" },
	{ 0x3532 , "HANGUL SYLLABLE NIEUN A THIEUTH" },
	{ 0x3533 , "HANGUL SYLLABLE NIEUN A HIEUH" },
	{ 0x3534 , "HANGUL SYLLABLE NIEUN AE" },
	{ 0x3535 , "HANGUL SYLLABLE NIEUN AE KIYEOK" },
	{ 0x3536 , "HANGUL SYLLABLE NIEUN AE NIEUN" },
	{ 0x3537 , "HANGUL SYLLABLE NIEUN AE RIEUL" },
	{ 0x3538 , "HANGUL SYLLABLE NIEUN AE MIEUM" },
	{ 0x3539 , "HANGUL SYLLABLE NIEUN AE PIEUP" },
	{ 0x353A , "HANGUL SYLLABLE NIEUN AE SIOS" },
	{ 0x353B , "HANGUL SYLLABLE NIEUN AE SSANGSIOS" },
	{ 0x353C , "HANGUL SYLLABLE NIEUN AE IEUNG" },
	{ 0x353D , "HANGUL SYLLABLE NIEUN YA" },
	{ 0x353E , "HANGUL SYLLABLE NIEUN YA KIYEOK" },
	{ 0x353F , "HANGUL SYLLABLE NIEUN YA NIEUN" },
	{ 0x3540 , "HANGUL SYLLABLE NIEUN YA RIEUL" },
	{ 0x3541 , "HANGUL SYLLABLE NIEUN YA MIEUM" },
	{ 0x3542 , "HANGUL SYLLABLE NIEUN YA IEUNG" },
	{ 0x3543 , "HANGUL SYLLABLE NIEUN EO" },
	{ 0x3544 , "HANGUL SYLLABLE NIEUN EO KIYEOK" },
	{ 0x3545 , "HANGUL SYLLABLE NIEUN EO KIYEOK-SIOS" },
	{ 0x3546 , "HANGUL SYLLABLE NIEUN EO NIEUN" },
	{ 0x3547 , "HANGUL SYLLABLE NIEUN EO RIEUL" },
	{ 0x3548 , "HANGUL SYLLABLE NIEUN EO RIEUL-MIEUM" },
	{ 0x3549 , "HANGUL SYLLABLE NIEUN EO RIEUL-PIEUP" },
	{ 0x354A , "HANGUL SYLLABLE NIEUN EO MIEUM" },
	{ 0x354B , "HANGUL SYLLABLE NIEUN EO PIEUP" },
	{ 0x354C , "HANGUL SYLLABLE NIEUN EO SIOS" },
	{ 0x354D , "HANGUL SYLLABLE NIEUN EO SSANGSIOS" },
	{ 0x354E , "HANGUL SYLLABLE NIEUN EO IEUNG" },
	{ 0x354F , "HANGUL SYLLABLE NIEUN EO HIEUH" },
	{ 0x3550 , "HANGUL SYLLABLE NIEUN E" },
	{ 0x3551 , "HANGUL SYLLABLE NIEUN E KIYEOK" },
	{ 0x3552 , "HANGUL SYLLABLE NIEUN E NIEUN" },
	{ 0x3553 , "HANGUL SYLLABLE NIEUN E RIEUL" },
	{ 0x3554 , "HANGUL SYLLABLE NIEUN E MIEUM" },
	{ 0x3555 , "HANGUL SYLLABLE NIEUN E PIEUP" },
	{ 0x3556 , "HANGUL SYLLABLE NIEUN E SIOS" },
	{ 0x3557 , "HANGUL SYLLABLE NIEUN E SSANGSIOS" },
	{ 0x3558 , "HANGUL SYLLABLE NIEUN E IEUNG" },
	{ 0x3559 , "HANGUL SYLLABLE NIEUN YEO" },
	{ 0x355A , "HANGUL SYLLABLE NIEUN YEO KIYEOK" },
	{ 0x355B , "HANGUL SYLLABLE NIEUN YEO NIEUN" },
	{ 0x355C , "HANGUL SYLLABLE NIEUN YEO RIEUL" },
	{ 0x355D , "HANGUL SYLLABLE NIEUN YEO MIEUM" },
	{ 0x355E , "HANGUL SYLLABLE NIEUN YEO PIEUP" },
	{ 0x355F , "HANGUL SYLLABLE NIEUN YEO SSANGSIOS" },
	{ 0x3560 , "HANGUL SYLLABLE NIEUN YEO IEUNG" },
	{ 0x3561 , "HANGUL SYLLABLE NIEUN YEO KHIEUKH" },
	{ 0x3562 , "HANGUL SYLLABLE NIEUN YE" },
	{ 0x3563 , "HANGUL SYLLABLE NIEUN YE NIEUN" },
	{ 0x3564 , "HANGUL SYLLABLE NIEUN O" },
	{ 0x3565 , "HANGUL SYLLABLE NIEUN O KIYEOK" },
	{ 0x3566 , "HANGUL SYLLABLE NIEUN O NIEUN" },
	{ 0x3567 , "HANGUL SYLLABLE NIEUN O RIEUL" },
	{ 0x3568 , "HANGUL SYLLABLE NIEUN O RIEUL-MIEUM" },
	{ 0x3569 , "HANGUL SYLLABLE NIEUN O MIEUM" },
	{ 0x356A , "HANGUL SYLLABLE NIEUN O PIEUP" },
	{ 0x356B , "HANGUL SYLLABLE NIEUN O SIOS" },
	{ 0x356C , "HANGUL SYLLABLE NIEUN O IEUNG" },
	{ 0x356D , "HANGUL SYLLABLE NIEUN O PHIEUPH" },
	{ 0x356E , "HANGUL SYLLABLE NIEUN O HIEUH" },
	{ 0x356F , "HANGUL SYLLABLE NIEUN WA" },
	{ 0x3570 , "HANGUL SYLLABLE NIEUN WA NIEUN" },
	{ 0x3571 , "HANGUL SYLLABLE NIEUN WA RIEUL" },
	{ 0x3572 , "HANGUL SYLLABLE NIEUN WA SSANGSIOS" },
	{ 0x3573 , "HANGUL SYLLABLE NIEUN OE" },
	{ 0x3574 , "HANGUL SYLLABLE NIEUN OE NIEUN" },
	{ 0x3575 , "HANGUL SYLLABLE NIEUN OE RIEUL" },
	{ 0x3576 , "HANGUL SYLLABLE NIEUN OE MIEUM" },
	{ 0x3577 , "HANGUL SYLLABLE NIEUN OE PIEUP" },
	{ 0x3578 , "HANGUL SYLLABLE NIEUN OE SIOS" },
	{ 0x3579 , "HANGUL SYLLABLE NIEUN YO" },
	{ 0x357A , "HANGUL SYLLABLE NIEUN YO KIYEOK" },
	{ 0x357B , "HANGUL SYLLABLE NIEUN YO NIEUN" },
	{ 0x357C , "HANGUL SYLLABLE NIEUN YO RIEUL" },
	{ 0x357D , "HANGUL SYLLABLE NIEUN YO PIEUP" },
	{ 0x357E , "HANGUL SYLLABLE NIEUN YO SIOS" },
	{ 0x357F , "HANGUL SYLLABLE NIEUN YO IEUNG" },
	{ 0x3580 , "HANGUL SYLLABLE NIEUN U" },
	{ 0x3581 , "HANGUL SYLLABLE NIEUN U KIYEOK" },
	{ 0x3582 , "HANGUL SYLLABLE NIEUN U NIEUN" },
	{ 0x3583 , "HANGUL SYLLABLE NIEUN U TIKEUT" },
	{ 0x3584 , "HANGUL SYLLABLE NIEUN U RIEUL" },
	{ 0x3585 , "HANGUL SYLLABLE NIEUN U MIEUM" },
	{ 0x3586 , "HANGUL SYLLABLE NIEUN U PIEUP" },
	{ 0x3587 , "HANGUL SYLLABLE NIEUN U SIOS" },
	{ 0x3588 , "HANGUL SYLLABLE NIEUN U IEUNG" },
	{ 0x3589 , "HANGUL SYLLABLE NIEUN WEO" },
	{ 0x358A , "HANGUL SYLLABLE NIEUN WEO SSANGSIOS" },
	{ 0x358B , "HANGUL SYLLABLE NIEUN WE" },
	{ 0x358C , "HANGUL SYLLABLE NIEUN WI" },
	{ 0x358D , "HANGUL SYLLABLE NIEUN WI NIEUN" },
	{ 0x358E , "HANGUL SYLLABLE NIEUN WI RIEUL" },
	{ 0x358F , "HANGUL SYLLABLE NIEUN WI MIEUM" },
	{ 0x3590 , "HANGUL SYLLABLE NIEUN WI PIEUP" },
	{ 0x3591 , "HANGUL SYLLABLE NIEUN YU" },
	{ 0x3592 , "HANGUL SYLLABLE NIEUN YU KIYEOK" },
	{ 0x3593 , "HANGUL SYLLABLE NIEUN YU RIEUL" },
	{ 0x3594 , "HANGUL SYLLABLE NIEUN YU MIEUM" },
	{ 0x3595 , "HANGUL SYLLABLE NIEUN YU PIEUP" },
	{ 0x3596 , "HANGUL SYLLABLE NIEUN YU IEUNG" },
	{ 0x3597 , "HANGUL SYLLABLE NIEUN EU" },
	{ 0x3598 , "HANGUL SYLLABLE NIEUN EU KIYEOK" },
	{ 0x3599 , "HANGUL SYLLABLE NIEUN EU NIEUN" },
	{ 0x359A , "HANGUL SYLLABLE NIEUN EU RIEUL" },
	{ 0x359B , "HANGUL SYLLABLE NIEUN EU RIEUL-KIYEOK" },
	{ 0x359C , "HANGUL SYLLABLE NIEUN EU RIEUL-MIEUM" },
	{ 0x359D , "HANGUL SYLLABLE NIEUN EU MIEUM" },
	{ 0x359E , "HANGUL SYLLABLE NIEUN EU PIEUP" },
	{ 0x359F , "HANGUL SYLLABLE NIEUN EU SIOS" },
	{ 0x35A0 , "HANGUL SYLLABLE NIEUN EU IEUNG" },
	{ 0x35A1 , "HANGUL SYLLABLE NIEUN EU CIEUC" },
	{ 0x35A2 , "HANGUL SYLLABLE NIEUN EU PHIEUPH" },
	{ 0x35A3 , "HANGUL SYLLABLE NIEUN YI" },
	{ 0x35A4 , "HANGUL SYLLABLE NIEUN YI NIEUN" },
	{ 0x35A5 , "HANGUL SYLLABLE NIEUN YI RIEUL" },
	{ 0x35A6 , "HANGUL SYLLABLE NIEUN I" },
	{ 0x35A7 , "HANGUL SYLLABLE NIEUN I KIYEOK" },
	{ 0x35A8 , "HANGUL SYLLABLE NIEUN I NIEUN" },
	{ 0x35A9 , "HANGUL SYLLABLE NIEUN I RIEUL" },
	{ 0x35AA , "HANGUL SYLLABLE NIEUN I RIEUL-MIEUM" },
	{ 0x35AB , "HANGUL SYLLABLE NIEUN I MIEUM" },
	{ 0x35AC , "HANGUL SYLLABLE NIEUN I PIEUP" },
	{ 0x35AD , "HANGUL SYLLABLE NIEUN I SIOS" },
	{ 0x35AE , "HANGUL SYLLABLE NIEUN I IEUNG" },
	{ 0x35AF , "HANGUL SYLLABLE NIEUN I PHIEUPH" },
	{ 0x35B0 , "HANGUL SYLLABLE TIKEUT A" },
	{ 0x35B1 , "HANGUL SYLLABLE TIKEUT A KIYEOK" },
	{ 0x35B2 , "HANGUL SYLLABLE TIKEUT A SSANGKIYEOK" },
	{ 0x35B3 , "HANGUL SYLLABLE TIKEUT A NIEUN" },
	{ 0x35B4 , "HANGUL SYLLABLE TIKEUT A TIKEUT" },
	{ 0x35B5 , "HANGUL SYLLABLE TIKEUT A RIEUL" },
	{ 0x35B6 , "HANGUL SYLLABLE TIKEUT A RIEUL-KIYEOK" },
	{ 0x35B7 , "HANGUL SYLLABLE TIKEUT A RIEUL-MIEUM" },
	{ 0x35B8 , "HANGUL SYLLABLE TIKEUT A RIEUL-PIEUP" },
	{ 0x35B9 , "HANGUL SYLLABLE TIKEUT A RIEUL-HIEUH" },
	{ 0x35BA , "HANGUL SYLLABLE TIKEUT A MIEUM" },
	{ 0x35BB , "HANGUL SYLLABLE TIKEUT A PIEUP" },
	{ 0x35BC , "HANGUL SYLLABLE TIKEUT A SIOS" },
	{ 0x35BD , "HANGUL SYLLABLE TIKEUT A SSANGSIOS" },
	{ 0x35BE , "HANGUL SYLLABLE TIKEUT A IEUNG" },
	{ 0x35BF , "HANGUL SYLLABLE TIKEUT A CIEUC" },
	{ 0x35C0 , "HANGUL SYLLABLE TIKEUT A CHIEUCH" },
	{ 0x35C1 , "HANGUL SYLLABLE TIKEUT A HIEUH" },
	{ 0x35C2 , "HANGUL SYLLABLE TIKEUT AE" },
	{ 0x35C3 , "HANGUL SYLLABLE TIKEUT AE KIYEOK" },
	{ 0x35C4 , "HANGUL SYLLABLE TIKEUT AE NIEUN" },
	{ 0x35C5 , "HANGUL SYLLABLE TIKEUT AE RIEUL" },
	{ 0x35C6 , "HANGUL SYLLABLE TIKEUT AE MIEUM" },
	{ 0x35C7 , "HANGUL SYLLABLE TIKEUT AE PIEUP" },
	{ 0x35C8 , "HANGUL SYLLABLE TIKEUT AE SIOS" },
	{ 0x35C9 , "HANGUL SYLLABLE TIKEUT AE SSANGSIOS" },
	{ 0x35CA , "HANGUL SYLLABLE TIKEUT AE IEUNG" },
	{ 0x35CB , "HANGUL SYLLABLE TIKEUT YA" },
	{ 0x35CC , "HANGUL SYLLABLE TIKEUT EO" },
	{ 0x35CD , "HANGUL SYLLABLE TIKEUT EO KIYEOK" },
	{ 0x35CE , "HANGUL SYLLABLE TIKEUT EO SSANGKIYEOK" },
	{ 0x35CF , "HANGUL SYLLABLE TIKEUT EO NIEUN" },
	{ 0x35D0 , "HANGUL SYLLABLE TIKEUT EO TIKEUT" },
	{ 0x35D1 , "HANGUL SYLLABLE TIKEUT EO RIEUL" },
	{ 0x35D2 , "HANGUL SYLLABLE TIKEUT EO RIEUL-MIEUM" },
	{ 0x35D3 , "HANGUL SYLLABLE TIKEUT EO RIEUL-PIEUP" },
	{ 0x35D4 , "HANGUL SYLLABLE TIKEUT EO MIEUM" },
	{ 0x35D5 , "HANGUL SYLLABLE TIKEUT EO PIEUP" },
	{ 0x35D6 , "HANGUL SYLLABLE TIKEUT EO SIOS" },
	{ 0x35D7 , "HANGUL SYLLABLE TIKEUT EO IEUNG" },
	{ 0x35D8 , "HANGUL SYLLABLE TIKEUT EO CHIEUCH" },
	{ 0x35D9 , "HANGUL SYLLABLE TIKEUT EO PHIEUPH" },
	{ 0x35DA , "HANGUL SYLLABLE TIKEUT E" },
	{ 0x35DB , "HANGUL SYLLABLE TIKEUT E KIYEOK" },
	{ 0x35DC , "HANGUL SYLLABLE TIKEUT E NIEUN" },
	{ 0x35DD , "HANGUL SYLLABLE TIKEUT E RIEUL" },
	{ 0x35DE , "HANGUL SYLLABLE TIKEUT E MIEUM" },
	{ 0x35DF , "HANGUL SYLLABLE TIKEUT E PIEUP" },
	{ 0x35E0 , "HANGUL SYLLABLE TIKEUT E SIOS" },
	{ 0x35E1 , "HANGUL SYLLABLE TIKEUT E SSANGSIOS" },
	{ 0x35E2 , "HANGUL SYLLABLE TIKEUT E IEUNG" },
	{ 0x35E3 , "HANGUL SYLLABLE TIKEUT YEO" },
	{ 0x35E4 , "HANGUL SYLLABLE TIKEUT YEO NIEUN" },
	{ 0x35E5 , "HANGUL SYLLABLE TIKEUT YEO RIEUL" },
	{ 0x35E6 , "HANGUL SYLLABLE TIKEUT YEO SSANGSIOS" },
	{ 0x35E7 , "HANGUL SYLLABLE TIKEUT YEO IEUNG" },
	{ 0x35E8 , "HANGUL SYLLABLE TIKEUT YE" },
	{ 0x35E9 , "HANGUL SYLLABLE TIKEUT YE NIEUN" },
	{ 0x35EA , "HANGUL SYLLABLE TIKEUT O" },
	{ 0x35EB , "HANGUL SYLLABLE TIKEUT O KIYEOK" },
	{ 0x35EC , "HANGUL SYLLABLE TIKEUT O NIEUN" },
	{ 0x35ED , "HANGUL SYLLABLE TIKEUT O TIKEUT" },
	{ 0x35EE , "HANGUL SYLLABLE TIKEUT O RIEUL" },
	{ 0x35EF , "HANGUL SYLLABLE TIKEUT O RIEUL-MIEUM" },
	{ 0x35F0 , "HANGUL SYLLABLE TIKEUT O RIEUL-SIOS" },
	{ 0x35F1 , "HANGUL SYLLABLE TIKEUT O MIEUM" },
	{ 0x35F2 , "HANGUL SYLLABLE TIKEUT O PIEUP" },
	{ 0x35F3 , "HANGUL SYLLABLE TIKEUT O SIOS" },
	{ 0x35F4 , "HANGUL SYLLABLE TIKEUT O IEUNG" },
	{ 0x35F5 , "HANGUL SYLLABLE TIKEUT O CHIEUCH" },
	{ 0x35F6 , "HANGUL SYLLABLE TIKEUT O THIEUTH" },
	{ 0x35F7 , "HANGUL SYLLABLE TIKEUT WA" },
	{ 0x35F8 , "HANGUL SYLLABLE TIKEUT WA NIEUN" },
	{ 0x35F9 , "HANGUL SYLLABLE TIKEUT WA RIEUL" },
	{ 0x35FA , "HANGUL SYLLABLE TIKEUT WAE" },
	{ 0x35FB , "HANGUL SYLLABLE TIKEUT WAE SSANGSIOS" },
	{ 0x35FC , "HANGUL SYLLABLE TIKEUT OE" },
	{ 0x35FD , "HANGUL SYLLABLE TIKEUT OE NIEUN" },
	{ 0x35FE , "HANGUL SYLLABLE TIKEUT OE RIEUL" },
	{ 0x35FF , "HANGUL SYLLABLE TIKEUT OE MIEUM" },
	{ 0x3600 , "HANGUL SYLLABLE TIKEUT OE PIEUP" },
	{ 0x3601 , "HANGUL SYLLABLE TIKEUT OE SIOS" },
	{ 0x3602 , "HANGUL SYLLABLE TIKEUT YO" },
	{ 0x3603 , "HANGUL SYLLABLE TIKEUT U" },
	{ 0x3604 , "HANGUL SYLLABLE TIKEUT U KIYEOK" },
	{ 0x3605 , "HANGUL SYLLABLE TIKEUT U NIEUN" },
	{ 0x3606 , "HANGUL SYLLABLE TIKEUT U RIEUL" },
	{ 0x3607 , "HANGUL SYLLABLE TIKEUT U MIEUM" },
	{ 0x3608 , "HANGUL SYLLABLE TIKEUT U PIEUP" },
	{ 0x3609 , "HANGUL SYLLABLE TIKEUT U SIOS" },
	{ 0x360A , "HANGUL SYLLABLE TIKEUT U IEUNG" },
	{ 0x360B , "HANGUL SYLLABLE TIKEUT WEO" },
	{ 0x360C , "HANGUL SYLLABLE TIKEUT WEO SSANGSIOS" },
	{ 0x360D , "HANGUL SYLLABLE TIKEUT WE" },
	{ 0x360E , "HANGUL SYLLABLE TIKEUT WE IEUNG" },
	{ 0x360F , "HANGUL SYLLABLE TIKEUT WI" },
	{ 0x3610 , "HANGUL SYLLABLE TIKEUT WI NIEUN" },
	{ 0x3611 , "HANGUL SYLLABLE TIKEUT WI RIEUL" },
	{ 0x3612 , "HANGUL SYLLABLE TIKEUT WI PIEUP" },
	{ 0x3613 , "HANGUL SYLLABLE TIKEUT WI SIOS" },
	{ 0x3614 , "HANGUL SYLLABLE TIKEUT WI IEUNG" },
	{ 0x3615 , "HANGUL SYLLABLE TIKEUT YU" },
	{ 0x3616 , "HANGUL SYLLABLE TIKEUT YU NIEUN" },
	{ 0x3617 , "HANGUL SYLLABLE TIKEUT YU RIEUL" },
	{ 0x3618 , "HANGUL SYLLABLE TIKEUT YU MIEUM" },
	{ 0x3619 , "HANGUL SYLLABLE TIKEUT YU IEUNG" },
	{ 0x361A , "HANGUL SYLLABLE TIKEUT EU" },
	{ 0x361B , "HANGUL SYLLABLE TIKEUT EU KIYEOK" },
	{ 0x361C , "HANGUL SYLLABLE TIKEUT EU NIEUN" },
	{ 0x361D , "HANGUL SYLLABLE TIKEUT EU TIKEUT" },
	{ 0x361E , "HANGUL SYLLABLE TIKEUT EU RIEUL" },
	{ 0x361F , "HANGUL SYLLABLE TIKEUT EU RIEUL-MIEUM" },
	{ 0x3620 , "HANGUL SYLLABLE TIKEUT EU MIEUM" },
	{ 0x3621 , "HANGUL SYLLABLE TIKEUT EU PIEUP" },
	{ 0x3622 , "HANGUL SYLLABLE TIKEUT EU SIOS" },
	{ 0x3623 , "HANGUL SYLLABLE TIKEUT EU IEUNG" },
	{ 0x3624 , "HANGUL SYLLABLE TIKEUT YI" },
	{ 0x3625 , "HANGUL SYLLABLE TIKEUT I" },
	{ 0x3626 , "HANGUL SYLLABLE TIKEUT I KIYEOK" },
	{ 0x3627 , "HANGUL SYLLABLE TIKEUT I NIEUN" },
	{ 0x3628 , "HANGUL SYLLABLE TIKEUT I TIKEUT" },
	{ 0x3629 , "HANGUL SYLLABLE TIKEUT I RIEUL" },
	{ 0x362A , "HANGUL SYLLABLE TIKEUT I MIEUM" },
	{ 0x362B , "HANGUL SYLLABLE TIKEUT I PIEUP" },
	{ 0x362C , "HANGUL SYLLABLE TIKEUT I SIOS" },
	{ 0x362D , "HANGUL SYLLABLE TIKEUT I SSANGSIOS" },
	{ 0x362E , "HANGUL SYLLABLE TIKEUT I IEUNG" },
	{ 0x362F , "HANGUL SYLLABLE TIKEUT I CIEUC" },
	{ 0x3630 , "HANGUL SYLLABLE SSANGTIKEUT A" },
	{ 0x3631 , "HANGUL SYLLABLE SSANGTIKEUT A KIYEOK" },
	{ 0x3632 , "HANGUL SYLLABLE SSANGTIKEUT A NIEUN" },
	{ 0x3633 , "HANGUL SYLLABLE SSANGTIKEUT A RIEUL" },
	{ 0x3634 , "HANGUL SYLLABLE SSANGTIKEUT A MIEUM" },
	{ 0x3635 , "HANGUL SYLLABLE SSANGTIKEUT A PIEUP" },
	{ 0x3636 , "HANGUL SYLLABLE SSANGTIKEUT A SIOS" },
	{ 0x3637 , "HANGUL SYLLABLE SSANGTIKEUT A SSANGSIOS" },
	{ 0x3638 , "HANGUL SYLLABLE SSANGTIKEUT A IEUNG" },
	{ 0x3639 , "HANGUL SYLLABLE SSANGTIKEUT A HIEUH" },
	{ 0x363A , "HANGUL SYLLABLE SSANGTIKEUT AE" },
	{ 0x363B , "HANGUL SYLLABLE SSANGTIKEUT AE KIYEOK" },
	{ 0x363C , "HANGUL SYLLABLE SSANGTIKEUT AE NIEUN" },
	{ 0x363D , "HANGUL SYLLABLE SSANGTIKEUT AE RIEUL" },
	{ 0x363E , "HANGUL SYLLABLE SSANGTIKEUT AE MIEUM" },
	{ 0x363F , "HANGUL SYLLABLE SSANGTIKEUT AE PIEUP" },
	{ 0x3640 , "HANGUL SYLLABLE SSANGTIKEUT AE SIOS" },
	{ 0x3641 , "HANGUL SYLLABLE SSANGTIKEUT AE SSANGSIOS" },
	{ 0x3642 , "HANGUL SYLLABLE SSANGTIKEUT AE IEUNG" },
	{ 0x3643 , "HANGUL SYLLABLE SSANGTIKEUT EO" },
	{ 0x3644 , "HANGUL SYLLABLE SSANGTIKEUT EO KIYEOK" },
	{ 0x3645 , "HANGUL SYLLABLE SSANGTIKEUT EO NIEUN" },
	{ 0x3646 , "HANGUL SYLLABLE SSANGTIKEUT EO RIEUL" },
	{ 0x3647 , "HANGUL SYLLABLE SSANGTIKEUT EO RIEUL-MIEUM" },
	{ 0x3648 , "HANGUL SYLLABLE SSANGTIKEUT EO RIEUL-PIEUP" },
	{ 0x3649 , "HANGUL SYLLABLE SSANGTIKEUT EO MIEUM" },
	{ 0x364A , "HANGUL SYLLABLE SSANGTIKEUT EO PIEUP" },
	{ 0x364B , "HANGUL SYLLABLE SSANGTIKEUT EO SIOS" },
	{ 0x364C , "HANGUL SYLLABLE SSANGTIKEUT EO SSANGSIOS" },
	{ 0x364D , "HANGUL SYLLABLE SSANGTIKEUT EO IEUNG" },
	{ 0x364E , "HANGUL SYLLABLE SSANGTIKEUT EO HIEUH" },
	{ 0x364F , "HANGUL SYLLABLE SSANGTIKEUT E" },
	{ 0x3650 , "HANGUL SYLLABLE SSANGTIKEUT E KIYEOK" },
	{ 0x3651 , "HANGUL SYLLABLE SSANGTIKEUT E NIEUN" },
	{ 0x3652 , "HANGUL SYLLABLE SSANGTIKEUT E RIEUL" },
	{ 0x3653 , "HANGUL SYLLABLE SSANGTIKEUT E MIEUM" },
	{ 0x3654 , "HANGUL SYLLABLE SSANGTIKEUT E PIEUP" },
	{ 0x3655 , "HANGUL SYLLABLE SSANGTIKEUT E SIOS" },
	{ 0x3656 , "HANGUL SYLLABLE SSANGTIKEUT E SSANGSIOS" },
	{ 0x3657 , "HANGUL SYLLABLE SSANGTIKEUT E IEUNG" },
	{ 0x3658 , "HANGUL SYLLABLE SSANGTIKEUT YEO" },
	{ 0x3659 , "HANGUL SYLLABLE SSANGTIKEUT YEO SSANGSIOS" },
	{ 0x365A , "HANGUL SYLLABLE SSANGTIKEUT O" },
	{ 0x365B , "HANGUL SYLLABLE SSANGTIKEUT O KIYEOK" },
	{ 0x365C , "HANGUL SYLLABLE SSANGTIKEUT O NIEUN" },
	{ 0x365D , "HANGUL SYLLABLE SSANGTIKEUT O RIEUL" },
	{ 0x365E , "HANGUL SYLLABLE SSANGTIKEUT O IEUNG" },
	{ 0x365F , "HANGUL SYLLABLE SSANGTIKEUT WA" },
	{ 0x3660 , "HANGUL SYLLABLE SSANGTIKEUT WA RIEUL" },
	{ 0x3661 , "HANGUL SYLLABLE SSANGTIKEUT WAE" },
	{ 0x3662 , "HANGUL SYLLABLE SSANGTIKEUT OE" },
	{ 0x3663 , "HANGUL SYLLABLE SSANGTIKEUT OE NIEUN" },
	{ 0x3664 , "HANGUL SYLLABLE SSANGTIKEUT U" },
	{ 0x3665 , "HANGUL SYLLABLE SSANGTIKEUT U KIYEOK" },
	{ 0x3666 , "HANGUL SYLLABLE SSANGTIKEUT U NIEUN" },
	{ 0x3667 , "HANGUL SYLLABLE SSANGTIKEUT U RIEUL" },
	{ 0x3668 , "HANGUL SYLLABLE SSANGTIKEUT U RIEUL-HIEUH" },
	{ 0x3669 , "HANGUL SYLLABLE SSANGTIKEUT U MIEUM" },
	{ 0x366A , "HANGUL SYLLABLE SSANGTIKEUT U IEUNG" },
	{ 0x366B , "HANGUL SYLLABLE SSANGTIKEUT WE" },
	{ 0x366C , "HANGUL SYLLABLE SSANGTIKEUT WI" },
	{ 0x366D , "HANGUL SYLLABLE SSANGTIKEUT WI NIEUN" },
	{ 0x366E , "HANGUL SYLLABLE SSANGTIKEUT WI RIEUL" },
	{ 0x366F , "HANGUL SYLLABLE SSANGTIKEUT WI MIEUM" },
	{ 0x3670 , "HANGUL SYLLABLE SSANGTIKEUT WI PIEUP" },
	{ 0x3671 , "HANGUL SYLLABLE SSANGTIKEUT WI IEUNG" },
	{ 0x3672 , "HANGUL SYLLABLE SSANGTIKEUT EU" },
	{ 0x3673 , "HANGUL SYLLABLE SSANGTIKEUT EU KIYEOK" },
	{ 0x3674 , "HANGUL SYLLABLE SSANGTIKEUT EU NIEUN" },
	{ 0x3675 , "HANGUL SYLLABLE SSANGTIKEUT EU TIKEUT" },
	{ 0x3676 , "HANGUL SYLLABLE SSANGTIKEUT EU RIEUL" },
	{ 0x3677 , "HANGUL SYLLABLE SSANGTIKEUT EU MIEUM" },
	{ 0x3678 , "HANGUL SYLLABLE SSANGTIKEUT EU PIEUP" },
	{ 0x3679 , "HANGUL SYLLABLE SSANGTIKEUT EU SIOS" },
	{ 0x367A , "HANGUL SYLLABLE SSANGTIKEUT YI" },
	{ 0x367B , "HANGUL SYLLABLE SSANGTIKEUT YI NIEUN" },
	{ 0x367C , "HANGUL SYLLABLE SSANGTIKEUT YI RIEUL" },
	{ 0x367D , "HANGUL SYLLABLE SSANGTIKEUT YI MIEUM" },
	{ 0x367E , "HANGUL SYLLABLE SSANGTIKEUT YI PIEUP" },
	{ 0x367F , "HANGUL SYLLABLE SSANGTIKEUT I" },
	{ 0x3680 , "HANGUL SYLLABLE SSANGTIKEUT I NIEUN" },
	{ 0x3681 , "HANGUL SYLLABLE SSANGTIKEUT I RIEUL" },
	{ 0x3682 , "HANGUL SYLLABLE SSANGTIKEUT I MIEUM" },
	{ 0x3683 , "HANGUL SYLLABLE SSANGTIKEUT I PIEUP" },
	{ 0x3684 , "HANGUL SYLLABLE SSANGTIKEUT I SIOS" },
	{ 0x3685 , "HANGUL SYLLABLE SSANGTIKEUT I IEUNG" },
	{ 0x3686 , "HANGUL SYLLABLE RIEUL A" },
	{ 0x3687 , "HANGUL SYLLABLE RIEUL A KIYEOK" },
	{ 0x3688 , "HANGUL SYLLABLE RIEUL A NIEUN" },
	{ 0x3689 , "HANGUL SYLLABLE RIEUL A RIEUL" },
	{ 0x368A , "HANGUL SYLLABLE RIEUL A MIEUM" },
	{ 0x368B , "HANGUL SYLLABLE RIEUL A PIEUP" },
	{ 0x368C , "HANGUL SYLLABLE RIEUL A SIOS" },
	{ 0x368D , "HANGUL SYLLABLE RIEUL A SSANGSIOS" },
	{ 0x368E , "HANGUL SYLLABLE RIEUL A IEUNG" },
	{ 0x368F , "HANGUL SYLLABLE RIEUL A CIEUC" },
	{ 0x3690 , "HANGUL SYLLABLE RIEUL A PHIEUPH" },
	{ 0x3691 , "HANGUL SYLLABLE RIEUL A HIEUH" },
	{ 0x3692 , "HANGUL SYLLABLE RIEUL AE" },
	{ 0x3693 , "HANGUL SYLLABLE RIEUL AE KIYEOK" },
	{ 0x3694 , "HANGUL SYLLABLE RIEUL AE NIEUN" },
	{ 0x3695 , "HANGUL SYLLABLE RIEUL AE RIEUL" },
	{ 0x3696 , "HANGUL SYLLABLE RIEUL AE MIEUM" },
	{ 0x3697 , "HANGUL SYLLABLE RIEUL AE PIEUP" },
	{ 0x3698 , "HANGUL SYLLABLE RIEUL AE SIOS" },
	{ 0x3699 , "HANGUL SYLLABLE RIEUL AE SSANGSIOS" },
	{ 0x369A , "HANGUL SYLLABLE RIEUL AE IEUNG" },
	{ 0x369B , "HANGUL SYLLABLE RIEUL YA" },
	{ 0x369C , "HANGUL SYLLABLE RIEUL YA KIYEOK" },
	{ 0x369D , "HANGUL SYLLABLE RIEUL YA NIEUN" },
	{ 0x369E , "HANGUL SYLLABLE RIEUL YA SIOS" },
	{ 0x369F , "HANGUL SYLLABLE RIEUL YA IEUNG" },
	{ 0x36A0 , "HANGUL SYLLABLE RIEUL EO" },
	{ 0x36A1 , "HANGUL SYLLABLE RIEUL EO KIYEOK" },
	{ 0x36A2 , "HANGUL SYLLABLE RIEUL EO NIEUN" },
	{ 0x36A3 , "HANGUL SYLLABLE RIEUL EO RIEUL" },
	{ 0x36A4 , "HANGUL SYLLABLE RIEUL EO MIEUM" },
	{ 0x36A5 , "HANGUL SYLLABLE RIEUL EO PIEUP" },
	{ 0x36A6 , "HANGUL SYLLABLE RIEUL EO SIOS" },
	{ 0x36A7 , "HANGUL SYLLABLE RIEUL EO SSANGSIOS" },
	{ 0x36A8 , "HANGUL SYLLABLE RIEUL EO IEUNG" },
	{ 0x36A9 , "HANGUL SYLLABLE RIEUL EO HIEUH" },
	{ 0x36AA , "HANGUL SYLLABLE RIEUL E" },
	{ 0x36AB , "HANGUL SYLLABLE RIEUL E KIYEOK" },
	{ 0x36AC , "HANGUL SYLLABLE RIEUL E NIEUN" },
	{ 0x36AD , "HANGUL SYLLABLE RIEUL E RIEUL" },
	{ 0x36AE , "HANGUL SYLLABLE RIEUL E MIEUM" },
	{ 0x36AF , "HANGUL SYLLABLE RIEUL E PIEUP" },
	{ 0x36B0 , "HANGUL SYLLABLE RIEUL E SIOS" },
	{ 0x36B1 , "HANGUL SYLLABLE RIEUL E IEUNG" },
	{ 0x36B2 , "HANGUL SYLLABLE RIEUL YEO" },
	{ 0x36B3 , "HANGUL SYLLABLE RIEUL YEO KIYEOK" },
	{ 0x36B4 , "HANGUL SYLLABLE RIEUL YEO NIEUN" },
	{ 0x36B5 , "HANGUL SYLLABLE RIEUL YEO RIEUL" },
	{ 0x36B6 , "HANGUL SYLLABLE RIEUL YEO MIEUM" },
	{ 0x36B7 , "HANGUL SYLLABLE RIEUL YEO PIEUP" },
	{ 0x36B8 , "HANGUL SYLLABLE RIEUL YEO SIOS" },
	{ 0x36B9 , "HANGUL SYLLABLE RIEUL YEO SSANGSIOS" },
	{ 0x36BA , "HANGUL SYLLABLE RIEUL YEO IEUNG" },
	{ 0x36BB , "HANGUL SYLLABLE RIEUL YE" },
	{ 0x36BC , "HANGUL SYLLABLE RIEUL YE NIEUN" },
	{ 0x36BD , "HANGUL SYLLABLE RIEUL YE PIEUP" },
	{ 0x36BE , "HANGUL SYLLABLE RIEUL YE SIOS" },
	{ 0x36BF , "HANGUL SYLLABLE RIEUL O" },
	{ 0x36C0 , "HANGUL SYLLABLE RIEUL O KIYEOK" },
	{ 0x36C1 , "HANGUL SYLLABLE RIEUL O NIEUN" },
	{ 0x36C2 , "HANGUL SYLLABLE RIEUL O RIEUL" },
	{ 0x36C3 , "HANGUL SYLLABLE RIEUL O MIEUM" },
	{ 0x36C4 , "HANGUL SYLLABLE RIEUL O PIEUP" },
	{ 0x36C5 , "HANGUL SYLLABLE RIEUL O SIOS" },
	{ 0x36C6 , "HANGUL SYLLABLE RIEUL O IEUNG" },
	{ 0x36C7 , "HANGUL SYLLABLE RIEUL WA" },
	{ 0x36C8 , "HANGUL SYLLABLE RIEUL WA NIEUN" },
	{ 0x36C9 , "HANGUL SYLLABLE RIEUL WA IEUNG" },
	{ 0x36CA , "HANGUL SYLLABLE RIEUL WAE SSANGSIOS" },
	{ 0x36CB , "HANGUL SYLLABLE RIEUL OE" },
	{ 0x36CC , "HANGUL SYLLABLE RIEUL OE NIEUN" },
	{ 0x36CD , "HANGUL SYLLABLE RIEUL OE RIEUL" },
	{ 0x36CE , "HANGUL SYLLABLE RIEUL OE MIEUM" },
	{ 0x36CF , "HANGUL SYLLABLE RIEUL OE PIEUP" },
	{ 0x36D0 , "HANGUL SYLLABLE RIEUL OE SIOS" },
	{ 0x36D1 , "HANGUL SYLLABLE RIEUL OE IEUNG" },
	{ 0x36D2 , "HANGUL SYLLABLE RIEUL YO" },
	{ 0x36D3 , "HANGUL SYLLABLE RIEUL YO NIEUN" },
	{ 0x36D4 , "HANGUL SYLLABLE RIEUL YO RIEUL" },
	{ 0x36D5 , "HANGUL SYLLABLE RIEUL YO PIEUP" },
	{ 0x36D6 , "HANGUL SYLLABLE RIEUL YO SIOS" },
	{ 0x36D7 , "HANGUL SYLLABLE RIEUL YO IEUNG" },
	{ 0x36D8 , "HANGUL SYLLABLE RIEUL U" },
	{ 0x36D9 , "HANGUL SYLLABLE RIEUL U KIYEOK" },
	{ 0x36DA , "HANGUL SYLLABLE RIEUL U NIEUN" },
	{ 0x36DB , "HANGUL SYLLABLE RIEUL U RIEUL" },
	{ 0x36DC , "HANGUL SYLLABLE RIEUL U MIEUM" },
	{ 0x36DD , "HANGUL SYLLABLE RIEUL U PIEUP" },
	{ 0x36DE , "HANGUL SYLLABLE RIEUL U SIOS" },
	{ 0x36DF , "HANGUL SYLLABLE RIEUL U IEUNG" },
	{ 0x36E0 , "HANGUL SYLLABLE RIEUL WEO" },
	{ 0x36E1 , "HANGUL SYLLABLE RIEUL WEO SSANGSIOS" },
	{ 0x36E2 , "HANGUL SYLLABLE RIEUL WE" },
	{ 0x36E3 , "HANGUL SYLLABLE RIEUL WI" },
	{ 0x36E4 , "HANGUL SYLLABLE RIEUL WI KIYEOK" },
	{ 0x36E5 , "HANGUL SYLLABLE RIEUL WI NIEUN" },
	{ 0x36E6 , "HANGUL SYLLABLE RIEUL WI RIEUL" },
	{ 0x36E7 , "HANGUL SYLLABLE RIEUL WI MIEUM" },
	{ 0x36E8 , "HANGUL SYLLABLE RIEUL WI SIOS" },
	{ 0x36E9 , "HANGUL SYLLABLE RIEUL WI IEUNG" },
	{ 0x36EA , "HANGUL SYLLABLE RIEUL YU" },
	{ 0x36EB , "HANGUL SYLLABLE RIEUL YU KIYEOK" },
	{ 0x36EC , "HANGUL SYLLABLE RIEUL YU NIEUN" },
	{ 0x36ED , "HANGUL SYLLABLE RIEUL YU RIEUL" },
	{ 0x36EE , "HANGUL SYLLABLE RIEUL YU MIEUM" },
	{ 0x36EF , "HANGUL SYLLABLE RIEUL YU PIEUP" },
	{ 0x36F0 , "HANGUL SYLLABLE RIEUL YU SIOS" },
	{ 0x36F1 , "HANGUL SYLLABLE RIEUL YU IEUNG" },
	{ 0x36F2 , "HANGUL SYLLABLE RIEUL EU" },
	{ 0x36F3 , "HANGUL SYLLABLE RIEUL EU KIYEOK" },
	{ 0x36F4 , "HANGUL SYLLABLE RIEUL EU NIEUN" },
	{ 0x36F5 , "HANGUL SYLLABLE RIEUL EU RIEUL" },
	{ 0x36F6 , "HANGUL SYLLABLE RIEUL EU MIEUM" },
	{ 0x36F7 , "HANGUL SYLLABLE RIEUL EU PIEUP" },
	{ 0x36F8 , "HANGUL SYLLABLE RIEUL EU SIOS" },
	{ 0x36F9 , "HANGUL SYLLABLE RIEUL EU IEUNG" },
	{ 0x36FA , "HANGUL SYLLABLE RIEUL EU CIEUC" },
	{ 0x36FB , "HANGUL SYLLABLE RIEUL EU THIEUTH" },
	{ 0x36FC , "HANGUL SYLLABLE RIEUL EU PHIEUPH" },
	{ 0x36FD , "HANGUL SYLLABLE RIEUL I" },
	{ 0x36FE , "HANGUL SYLLABLE RIEUL I KIYEOK" },
	{ 0x36FF , "HANGUL SYLLABLE RIEUL I NIEUN" },
	{ 0x3700 , "HANGUL SYLLABLE RIEUL I RIEUL" },
	{ 0x3701 , "HANGUL SYLLABLE RIEUL I MIEUM" },
	{ 0x3702 , "HANGUL SYLLABLE RIEUL I PIEUP" },
	{ 0x3703 , "HANGUL SYLLABLE RIEUL I SIOS" },
	{ 0x3704 , "HANGUL SYLLABLE RIEUL I IEUNG" },
	{ 0x3705 , "HANGUL SYLLABLE MIEUM A" },
	{ 0x3706 , "HANGUL SYLLABLE MIEUM A KIYEOK" },
	{ 0x3707 , "HANGUL SYLLABLE MIEUM A NIEUN" },
	{ 0x3708 , "HANGUL SYLLABLE MIEUM A NIEUN-HIEUH" },
	{ 0x3709 , "HANGUL SYLLABLE MIEUM A TIKEUT" },
	{ 0x370A , "HANGUL SYLLABLE MIEUM A RIEUL" },
	{ 0x370B , "HANGUL SYLLABLE MIEUM A RIEUL-KIYEOK" },
	{ 0x370C , "HANGUL SYLLABLE MIEUM A RIEUL-MIEUM" },
	{ 0x370D , "HANGUL SYLLABLE MIEUM A MIEUM" },
	{ 0x370E , "HANGUL SYLLABLE MIEUM A PIEUP" },
	{ 0x370F , "HANGUL SYLLABLE MIEUM A SIOS" },
	{ 0x3710 , "HANGUL SYLLABLE MIEUM A IEUNG" },
	{ 0x3711 , "HANGUL SYLLABLE MIEUM A CIEUC" },
	{ 0x3712 , "HANGUL SYLLABLE MIEUM A THIEUTH" },
	{ 0x3713 , "HANGUL SYLLABLE MIEUM A HIEUH" },
	{ 0x3714 , "HANGUL SYLLABLE MIEUM AE" },
	{ 0x3715 , "HANGUL SYLLABLE MIEUM AE KIYEOK" },
	{ 0x3716 , "HANGUL SYLLABLE MIEUM AE NIEUN" },
	{ 0x3717 , "HANGUL SYLLABLE MIEUM AE RIEUL" },
	{ 0x3718 , "HANGUL SYLLABLE MIEUM AE MIEUM" },
	{ 0x3719 , "HANGUL SYLLABLE MIEUM AE PIEUP" },
	{ 0x371A , "HANGUL SYLLABLE MIEUM AE SIOS" },
	{ 0x371B , "HANGUL SYLLABLE MIEUM AE SSANGSIOS" },
	{ 0x371C , "HANGUL SYLLABLE MIEUM AE IEUNG" },
	{ 0x371D , "HANGUL SYLLABLE MIEUM AE CIEUC" },
	{ 0x371E , "HANGUL SYLLABLE MIEUM YA" },
	{ 0x371F , "HANGUL SYLLABLE MIEUM YA KIYEOK" },
	{ 0x3720 , "HANGUL SYLLABLE MIEUM YA RIEUL" },
	{ 0x3721 , "HANGUL SYLLABLE MIEUM YA IEUNG" },
	{ 0x3722 , "HANGUL SYLLABLE MIEUM EO" },
	{ 0x3723 , "HANGUL SYLLABLE MIEUM EO KIYEOK" },
	{ 0x3724 , "HANGUL SYLLABLE MIEUM EO NIEUN" },
	{ 0x3725 , "HANGUL SYLLABLE MIEUM EO RIEUL" },
	{ 0x3726 , "HANGUL SYLLABLE MIEUM EO RIEUL-MIEUM" },
	{ 0x3727 , "HANGUL SYLLABLE MIEUM EO MIEUM" },
	{ 0x3728 , "HANGUL SYLLABLE MIEUM EO PIEUP" },
	{ 0x3729 , "HANGUL SYLLABLE MIEUM EO SIOS" },
	{ 0x372A , "HANGUL SYLLABLE MIEUM EO IEUNG" },
	{ 0x372B , "HANGUL SYLLABLE MIEUM EO CIEUC" },
	{ 0x372C , "HANGUL SYLLABLE MIEUM EO HIEUH" },
	{ 0x372D , "HANGUL SYLLABLE MIEUM E" },
	{ 0x372E , "HANGUL SYLLABLE MIEUM E KIYEOK" },
	{ 0x372F , "HANGUL SYLLABLE MIEUM E NIEUN" },
	{ 0x3730 , "HANGUL SYLLABLE MIEUM E RIEUL" },
	{ 0x3731 , "HANGUL SYLLABLE MIEUM E MIEUM" },
	{ 0x3732 , "HANGUL SYLLABLE MIEUM E PIEUP" },
	{ 0x3733 , "HANGUL SYLLABLE MIEUM E SIOS" },
	{ 0x3734 , "HANGUL SYLLABLE MIEUM E SSANGSIOS" },
	{ 0x3735 , "HANGUL SYLLABLE MIEUM E IEUNG" },
	{ 0x3736 , "HANGUL SYLLABLE MIEUM YEO" },
	{ 0x3737 , "HANGUL SYLLABLE MIEUM YEO KIYEOK" },
	{ 0x3738 , "HANGUL SYLLABLE MIEUM YEO NIEUN" },
	{ 0x3739 , "HANGUL SYLLABLE MIEUM YEO RIEUL" },
	{ 0x373A , "HANGUL SYLLABLE MIEUM YEO SIOS" },
	{ 0x373B , "HANGUL SYLLABLE MIEUM YEO SSANGSIOS" },
	{ 0x373C , "HANGUL SYLLABLE MIEUM YEO IEUNG" },
	{ 0x373D , "HANGUL SYLLABLE MIEUM YEO CHIEUCH" },
	{ 0x373E , "HANGUL SYLLABLE MIEUM YE" },
	{ 0x373F , "HANGUL SYLLABLE MIEUM O" },
	{ 0x3740 , "HANGUL SYLLABLE MIEUM O KIYEOK" },
	{ 0x3741 , "HANGUL SYLLABLE MIEUM O KIYEOK-SIOS" },
	{ 0x3742 , "HANGUL SYLLABLE MIEUM O NIEUN" },
	{ 0x3743 , "HANGUL SYLLABLE MIEUM O RIEUL" },
	{ 0x3744 , "HANGUL SYLLABLE MIEUM O RIEUL-MIEUM" },
	{ 0x3745 , "HANGUL SYLLABLE MIEUM O MIEUM" },
	{ 0x3746 , "HANGUL SYLLABLE MIEUM O PIEUP" },
	{ 0x3747 , "HANGUL SYLLABLE MIEUM O SIOS" },
	{ 0x3748 , "HANGUL SYLLABLE MIEUM O IEUNG" },
	{ 0x3749 , "HANGUL SYLLABLE MIEUM WA" },
	{ 0x374A , "HANGUL SYLLABLE MIEUM WA NIEUN" },
	{ 0x374B , "HANGUL SYLLABLE MIEUM WA SSANGSIOS" },
	{ 0x374C , "HANGUL SYLLABLE MIEUM WA IEUNG" },
	{ 0x374D , "HANGUL SYLLABLE MIEUM OE" },
	{ 0x374E , "HANGUL SYLLABLE MIEUM OE NIEUN" },
	{ 0x374F , "HANGUL SYLLABLE MIEUM OE RIEUL" },
	{ 0x3750 , "HANGUL SYLLABLE MIEUM OE PIEUP" },
	{ 0x3751 , "HANGUL SYLLABLE MIEUM OE SIOS" },
	{ 0x3752 , "HANGUL SYLLABLE MIEUM OE IEUNG" },
	{ 0x3753 , "HANGUL SYLLABLE MIEUM YO" },
	{ 0x3754 , "HANGUL SYLLABLE MIEUM YO NIEUN" },
	{ 0x3755 , "HANGUL SYLLABLE MIEUM YO RIEUL" },
	{ 0x3756 , "HANGUL SYLLABLE MIEUM YO PIEUP" },
	{ 0x3757 , "HANGUL SYLLABLE MIEUM YO SIOS" },
	{ 0x3758 , "HANGUL SYLLABLE MIEUM U" },
	{ 0x3759 , "HANGUL SYLLABLE MIEUM U KIYEOK" },
	{ 0x375A , "HANGUL SYLLABLE MIEUM U SSANGKIYEOK" },
	{ 0x375B , "HANGUL SYLLABLE MIEUM U NIEUN" },
	{ 0x375C , "HANGUL SYLLABLE MIEUM U TIKEUT" },
	{ 0x375D , "HANGUL SYLLABLE MIEUM U RIEUL" },
	{ 0x375E , "HANGUL SYLLABLE MIEUM U RIEUL-KIYEOK" },
	{ 0x375F , "HANGUL SYLLABLE MIEUM U RIEUL-MIEUM" },
	{ 0x3760 , "HANGUL SYLLABLE MIEUM U MIEUM" },
	{ 0x3761 , "HANGUL SYLLABLE MIEUM U PIEUP" },
	{ 0x3762 , "HANGUL SYLLABLE MIEUM U SIOS" },
	{ 0x3763 , "HANGUL SYLLABLE MIEUM U IEUNG" },
	{ 0x3764 , "HANGUL SYLLABLE MIEUM U THIEUTH" },
	{ 0x3765 , "HANGUL SYLLABLE MIEUM U HIEUH" },
	{ 0x3766 , "HANGUL SYLLABLE MIEUM WEO" },
	{ 0x3767 , "HANGUL SYLLABLE MIEUM WEO NIEUN" },
	{ 0x3768 , "HANGUL SYLLABLE MIEUM WEO RIEUL" },
	{ 0x3769 , "HANGUL SYLLABLE MIEUM WEO PIEUP" },
	{ 0x376A , "HANGUL SYLLABLE MIEUM WEO SIOS" },
	{ 0x376B , "HANGUL SYLLABLE MIEUM WE" },
	{ 0x376C , "HANGUL SYLLABLE MIEUM WI" },
	{ 0x376D , "HANGUL SYLLABLE MIEUM WI NIEUN" },
	{ 0x376E , "HANGUL SYLLABLE MIEUM WI RIEUL" },
	{ 0x376F , "HANGUL SYLLABLE MIEUM YU" },
	{ 0x3770 , "HANGUL SYLLABLE MIEUM YU NIEUN" },
	{ 0x3771 , "HANGUL SYLLABLE MIEUM YU RIEUL" },
	{ 0x3772 , "HANGUL SYLLABLE MIEUM YU MIEUM" },
	{ 0x3773 , "HANGUL SYLLABLE MIEUM YU SIOS" },
	{ 0x3774 , "HANGUL SYLLABLE MIEUM EU" },
	{ 0x3775 , "HANGUL SYLLABLE MIEUM EU NIEUN" },
	{ 0x3776 , "HANGUL SYLLABLE MIEUM EU RIEUL" },
	{ 0x3777 , "HANGUL SYLLABLE MIEUM EU MIEUM" },
	{ 0x3778 , "HANGUL SYLLABLE MIEUM EU SIOS" },
	{ 0x3779 , "HANGUL SYLLABLE MIEUM I" },
	{ 0x377A , "HANGUL SYLLABLE MIEUM I KIYEOK" },
	{ 0x377B , "HANGUL SYLLABLE MIEUM I NIEUN" },
	{ 0x377C , "HANGUL SYLLABLE MIEUM I TIKEUT" },
	{ 0x377D , "HANGUL SYLLABLE MIEUM I RIEUL" },
	{ 0x377E , "HANGUL SYLLABLE MIEUM I RIEUL-MIEUM" },
	{ 0x377F , "HANGUL SYLLABLE MIEUM I MIEUM" },
	{ 0x3780 , "HANGUL SYLLABLE MIEUM I PIEUP" },
	{ 0x3781 , "HANGUL SYLLABLE MIEUM I SIOS" },
	{ 0x3782 , "HANGUL SYLLABLE MIEUM I SSANGSIOS" },
	{ 0x3783 , "HANGUL SYLLABLE MIEUM I IEUNG" },
	{ 0x3784 , "HANGUL SYLLABLE MIEUM I CHIEUCH" },
	{ 0x3785 , "HANGUL SYLLABLE MIEUM I THIEUTH" },
	{ 0x3786 , "HANGUL SYLLABLE PIEUP A" },
	{ 0x3787 , "HANGUL SYLLABLE PIEUP A KIYEOK" },
	{ 0x3788 , "HANGUL SYLLABLE PIEUP A SSANGKIYEOK" },
	{ 0x3789 , "HANGUL SYLLABLE PIEUP A KIYEOK-SIOS" },
	{ 0x378A , "HANGUL SYLLABLE PIEUP A NIEUN" },
	{ 0x378B , "HANGUL SYLLABLE PIEUP A TIKEUT" },
	{ 0x378C , "HANGUL SYLLABLE PIEUP A RIEUL" },
	{ 0x378D , "HANGUL SYLLABLE PIEUP A RIEUL-KIYEOK" },
	{ 0x378E , "HANGUL SYLLABLE PIEUP A RIEUL-MIEUM" },
	{ 0x378F , "HANGUL SYLLABLE PIEUP A RIEUL-PIEUP" },
	{ 0x3790 , "HANGUL SYLLABLE PIEUP A MIEUM" },
	{ 0x3791 , "HANGUL SYLLABLE PIEUP A PIEUP" },
	{ 0x3792 , "HANGUL SYLLABLE PIEUP A SIOS" },
	{ 0x3793 , "HANGUL SYLLABLE PIEUP A IEUNG" },
	{ 0x3794 , "HANGUL SYLLABLE PIEUP A THIEUTH" },
	{ 0x3795 , "HANGUL SYLLABLE PIEUP AE" },
	{ 0x3796 , "HANGUL SYLLABLE PIEUP AE KIYEOK" },
	{ 0x3797 , "HANGUL SYLLABLE PIEUP AE NIEUN" },
	{ 0x3798 , "HANGUL SYLLABLE PIEUP AE RIEUL" },
	{ 0x3799 , "HANGUL SYLLABLE PIEUP AE MIEUM" },
	{ 0x379A , "HANGUL SYLLABLE PIEUP AE PIEUP" },
	{ 0x379B , "HANGUL SYLLABLE PIEUP AE SIOS" },
	{ 0x379C , "HANGUL SYLLABLE PIEUP AE SSANGSIOS" },
	{ 0x379D , "HANGUL SYLLABLE PIEUP AE IEUNG" },
	{ 0x379E , "HANGUL SYLLABLE PIEUP AE THIEUTH" },
	{ 0x379F , "HANGUL SYLLABLE PIEUP YA" },
	{ 0x37A0 , "HANGUL SYLLABLE PIEUP YA KIYEOK" },
	{ 0x37A1 , "HANGUL SYLLABLE PIEUP YA NIEUN" },
	{ 0x37A2 , "HANGUL SYLLABLE PIEUP YA PIEUP" },
	{ 0x37A3 , "HANGUL SYLLABLE PIEUP EO" },
	{ 0x37A4 , "HANGUL SYLLABLE PIEUP EO KIYEOK" },
	{ 0x37A5 , "HANGUL SYLLABLE PIEUP EO NIEUN" },
	{ 0x37A6 , "HANGUL SYLLABLE PIEUP EO TIKEUT" },
	{ 0x37A7 , "HANGUL SYLLABLE PIEUP EO RIEUL" },
	{ 0x37A8 , "HANGUL SYLLABLE PIEUP EO RIEUL-MIEUM" },
	{ 0x37A9 , "HANGUL SYLLABLE PIEUP EO MIEUM" },
	{ 0x37AA , "HANGUL SYLLABLE PIEUP EO PIEUP" },
	{ 0x37AB , "HANGUL SYLLABLE PIEUP EO SIOS" },
	{ 0x37AC , "HANGUL SYLLABLE PIEUP EO IEUNG" },
	{ 0x37AD , "HANGUL SYLLABLE PIEUP EO CIEUC" },
	{ 0x37AE , "HANGUL SYLLABLE PIEUP E" },
	{ 0x37AF , "HANGUL SYLLABLE PIEUP E KIYEOK" },
	{ 0x37B0 , "HANGUL SYLLABLE PIEUP E NIEUN" },
	{ 0x37B1 , "HANGUL SYLLABLE PIEUP E TIKEUT" },
	{ 0x37B2 , "HANGUL SYLLABLE PIEUP E RIEUL" },
	{ 0x37B3 , "HANGUL SYLLABLE PIEUP E MIEUM" },
	{ 0x37B4 , "HANGUL SYLLABLE PIEUP E PIEUP" },
	{ 0x37B5 , "HANGUL SYLLABLE PIEUP E SIOS" },
	{ 0x37B6 , "HANGUL SYLLABLE PIEUP E SSANGSIOS" },
	{ 0x37B7 , "HANGUL SYLLABLE PIEUP E IEUNG" },
	{ 0x37B8 , "HANGUL SYLLABLE PIEUP YEO" },
	{ 0x37B9 , "HANGUL SYLLABLE PIEUP YEO KIYEOK" },
	{ 0x37BA , "HANGUL SYLLABLE PIEUP YEO NIEUN" },
	{ 0x37BB , "HANGUL SYLLABLE PIEUP YEO RIEUL" },
	{ 0x37BC , "HANGUL SYLLABLE PIEUP YEO PIEUP" },
	{ 0x37BD , "HANGUL SYLLABLE PIEUP YEO SIOS" },
	{ 0x37BE , "HANGUL SYLLABLE PIEUP YEO SSANGSIOS" },
	{ 0x37BF , "HANGUL SYLLABLE PIEUP YEO IEUNG" },
	{ 0x37C0 , "HANGUL SYLLABLE PIEUP YEO THIEUTH" },
	{ 0x37C1 , "HANGUL SYLLABLE PIEUP YE" },
	{ 0x37C2 , "HANGUL SYLLABLE PIEUP YE NIEUN" },
	{ 0x37C3 , "HANGUL SYLLABLE PIEUP O" },
	{ 0x37C4 , "HANGUL SYLLABLE PIEUP O KIYEOK" },
	{ 0x37C5 , "HANGUL SYLLABLE PIEUP O SSANGKIYEOK" },
	{ 0x37C6 , "HANGUL SYLLABLE PIEUP O NIEUN" },
	{ 0x37C7 , "HANGUL SYLLABLE PIEUP O RIEUL" },
	{ 0x37C8 , "HANGUL SYLLABLE PIEUP O MIEUM" },
	{ 0x37C9 , "HANGUL SYLLABLE PIEUP O PIEUP" },
	{ 0x37CA , "HANGUL SYLLABLE PIEUP O SIOS" },
	{ 0x37CB , "HANGUL SYLLABLE PIEUP O IEUNG" },
	{ 0x37CC , "HANGUL SYLLABLE PIEUP WA" },
	{ 0x37CD , "HANGUL SYLLABLE PIEUP WA NIEUN" },
	{ 0x37CE , "HANGUL SYLLABLE PIEUP WA SSANGSIOS" },
	{ 0x37CF , "HANGUL SYLLABLE PIEUP WAE" },
	{ 0x37D0 , "HANGUL SYLLABLE PIEUP WAE SSANGSIOS" },
	{ 0x37D1 , "HANGUL SYLLABLE PIEUP OE" },
	{ 0x37D2 , "HANGUL SYLLABLE PIEUP OE KIYEOK" },
	{ 0x37D3 , "HANGUL SYLLABLE PIEUP OE NIEUN" },
	{ 0x37D4 , "HANGUL SYLLABLE PIEUP OE RIEUL" },
	{ 0x37D5 , "HANGUL SYLLABLE PIEUP OE MIEUM" },
	{ 0x37D6 , "HANGUL SYLLABLE PIEUP OE PIEUP" },
	{ 0x37D7 , "HANGUL SYLLABLE PIEUP YO" },
	{ 0x37D8 , "HANGUL SYLLABLE PIEUP YO NIEUN" },
	{ 0x37D9 , "HANGUL SYLLABLE PIEUP U" },
	{ 0x37DA , "HANGUL SYLLABLE PIEUP U KIYEOK" },
	{ 0x37DB , "HANGUL SYLLABLE PIEUP U NIEUN" },
	{ 0x37DC , "HANGUL SYLLABLE PIEUP U TIKEUT" },
	{ 0x37DD , "HANGUL SYLLABLE PIEUP U RIEUL" },
	{ 0x37DE , "HANGUL SYLLABLE PIEUP U RIEUL-KIYEOK" },
	{ 0x37DF , "HANGUL SYLLABLE PIEUP U RIEUL-MIEUM" },
	{ 0x37E0 , "HANGUL SYLLABLE PIEUP U MIEUM" },
	{ 0x37E1 , "HANGUL SYLLABLE PIEUP U PIEUP" },
	{ 0x37E2 , "HANGUL SYLLABLE PIEUP U SIOS" },
	{ 0x37E3 , "HANGUL SYLLABLE PIEUP U IEUNG" },
	{ 0x37E4 , "HANGUL SYLLABLE PIEUP U THIEUTH" },
	{ 0x37E5 , "HANGUL SYLLABLE PIEUP U PHIEUPH" },
	{ 0x37E6 , "HANGUL SYLLABLE PIEUP WEO" },
	{ 0x37E7 , "HANGUL SYLLABLE PIEUP WEO RIEUL" },
	{ 0x37E8 , "HANGUL SYLLABLE PIEUP WEO SSANGSIOS" },
	{ 0x37E9 , "HANGUL SYLLABLE PIEUP WE" },
	{ 0x37EA , "HANGUL SYLLABLE PIEUP WI" },
	{ 0x37EB , "HANGUL SYLLABLE PIEUP WI KIYEOK" },
	{ 0x37EC , "HANGUL SYLLABLE PIEUP WI NIEUN" },
	{ 0x37ED , "HANGUL SYLLABLE PIEUP WI RIEUL" },
	{ 0x37EE , "HANGUL SYLLABLE PIEUP WI IEUNG" },
	{ 0x37EF , "HANGUL SYLLABLE PIEUP YU" },
	{ 0x37F0 , "HANGUL SYLLABLE PIEUP YU NIEUN" },
	{ 0x37F1 , "HANGUL SYLLABLE PIEUP YU RIEUL" },
	{ 0x37F2 , "HANGUL SYLLABLE PIEUP YU MIEUM" },
	{ 0x37F3 , "HANGUL SYLLABLE PIEUP YU SIOS" },
	{ 0x37F4 , "HANGUL SYLLABLE PIEUP YU IEUNG" },
	{ 0x37F5 , "HANGUL SYLLABLE PIEUP EU" },
	{ 0x37F6 , "HANGUL SYLLABLE PIEUP EU KIYEOK" },
	{ 0x37F7 , "HANGUL SYLLABLE PIEUP EU NIEUN" },
	{ 0x37F8 , "HANGUL SYLLABLE PIEUP EU RIEUL" },
	{ 0x37F9 , "HANGUL SYLLABLE PIEUP EU MIEUM" },
	{ 0x37FA , "HANGUL SYLLABLE PIEUP EU PIEUP" },
	{ 0x37FB , "HANGUL SYLLABLE PIEUP EU SIOS" },
	{ 0x37FC , "HANGUL SYLLABLE PIEUP I" },
	{ 0x37FD , "HANGUL SYLLABLE PIEUP I KIYEOK" },
	{ 0x37FE , "HANGUL SYLLABLE PIEUP I NIEUN" },
	{ 0x37FF , "HANGUL SYLLABLE PIEUP I RIEUL" },
	{ 0x3800 , "HANGUL SYLLABLE PIEUP I RIEUL-MIEUM" },
	{ 0x3801 , "HANGUL SYLLABLE PIEUP I MIEUM" },
	{ 0x3802 , "HANGUL SYLLABLE PIEUP I PIEUP" },
	{ 0x3803 , "HANGUL SYLLABLE PIEUP I SIOS" },
	{ 0x3804 , "HANGUL SYLLABLE PIEUP I IEUNG" },
	{ 0x3805 , "HANGUL SYLLABLE PIEUP I CIEUC" },
	{ 0x3806 , "HANGUL SYLLABLE PIEUP I CHIEUCH" },
	{ 0x3807 , "HANGUL SYLLABLE SSANGPIEUP A" },
	{ 0x3808 , "HANGUL SYLLABLE SSANGPIEUP A KIYEOK" },
	{ 0x3809 , "HANGUL SYLLABLE SSANGPIEUP A NIEUN" },
	{ 0x380A , "HANGUL SYLLABLE SSANGPIEUP A RIEUL" },
	{ 0x380B , "HANGUL SYLLABLE SSANGPIEUP A RIEUL-MIEUM" },
	{ 0x380C , "HANGUL SYLLABLE SSANGPIEUP A MIEUM" },
	{ 0x380D , "HANGUL SYLLABLE SSANGPIEUP A PIEUP" },
	{ 0x380E , "HANGUL SYLLABLE SSANGPIEUP A SIOS" },
	{ 0x380F , "HANGUL SYLLABLE SSANGPIEUP A SSANGSIOS" },
	{ 0x3810 , "HANGUL SYLLABLE SSANGPIEUP A IEUNG" },
	{ 0x3811 , "HANGUL SYLLABLE SSANGPIEUP A HIEUH" },
	{ 0x3812 , "HANGUL SYLLABLE SSANGPIEUP AE" },
	{ 0x3813 , "HANGUL SYLLABLE SSANGPIEUP AE KIYEOK" },
	{ 0x3814 , "HANGUL SYLLABLE SSANGPIEUP AE NIEUN" },
	{ 0x3815 , "HANGUL SYLLABLE SSANGPIEUP AE RIEUL" },
	{ 0x3816 , "HANGUL SYLLABLE SSANGPIEUP AE MIEUM" },
	{ 0x3817 , "HANGUL SYLLABLE SSANGPIEUP AE PIEUP" },
	{ 0x3818 , "HANGUL SYLLABLE SSANGPIEUP AE SIOS" },
	{ 0x3819 , "HANGUL SYLLABLE SSANGPIEUP AE SSANGSIOS" },
	{ 0x381A , "HANGUL SYLLABLE SSANGPIEUP AE IEUNG" },
	{ 0x381B , "HANGUL SYLLABLE SSANGPIEUP YA" },
	{ 0x381C , "HANGUL SYLLABLE SSANGPIEUP YA KIYEOK" },
	{ 0x381D , "HANGUL SYLLABLE SSANGPIEUP YA MIEUM" },
	{ 0x381E , "HANGUL SYLLABLE SSANGPIEUP EO" },
	{ 0x381F , "HANGUL SYLLABLE SSANGPIEUP EO KIYEOK" },
	{ 0x3820 , "HANGUL SYLLABLE SSANGPIEUP EO NIEUN" },
	{ 0x3821 , "HANGUL SYLLABLE SSANGPIEUP EO TIKEUT" },
	{ 0x3822 , "HANGUL SYLLABLE SSANGPIEUP EO RIEUL" },
	{ 0x3823 , "HANGUL SYLLABLE SSANGPIEUP EO MIEUM" },
	{ 0x3824 , "HANGUL SYLLABLE SSANGPIEUP EO SIOS" },
	{ 0x3825 , "HANGUL SYLLABLE SSANGPIEUP EO SSANGSIOS" },
	{ 0x3826 , "HANGUL SYLLABLE SSANGPIEUP EO IEUNG" },
	{ 0x3827 , "HANGUL SYLLABLE SSANGPIEUP E" },
	{ 0x3828 , "HANGUL SYLLABLE SSANGPIEUP E IEUNG" },
	{ 0x3829 , "HANGUL SYLLABLE SSANGPIEUP YEO" },
	{ 0x382A , "HANGUL SYLLABLE SSANGPIEUP YEO KIYEOK" },
	{ 0x382B , "HANGUL SYLLABLE SSANGPIEUP YEO MIEUM" },
	{ 0x382C , "HANGUL SYLLABLE SSANGPIEUP YEO PIEUP" },
	{ 0x382D , "HANGUL SYLLABLE SSANGPIEUP YEO SIOS" },
	{ 0x382E , "HANGUL SYLLABLE SSANGPIEUP YEO SSANGSIOS" },
	{ 0x382F , "HANGUL SYLLABLE SSANGPIEUP YEO IEUNG" },
	{ 0x3830 , "HANGUL SYLLABLE SSANGPIEUP O" },
	{ 0x3831 , "HANGUL SYLLABLE SSANGPIEUP O KIYEOK" },
	{ 0x3832 , "HANGUL SYLLABLE SSANGPIEUP O NIEUN" },
	{ 0x3833 , "HANGUL SYLLABLE SSANGPIEUP O RIEUL" },
	{ 0x3834 , "HANGUL SYLLABLE SSANGPIEUP O MIEUM" },
	{ 0x3835 , "HANGUL SYLLABLE SSANGPIEUP O PIEUP" },
	{ 0x3836 , "HANGUL SYLLABLE SSANGPIEUP O IEUNG" },
	{ 0x3837 , "HANGUL SYLLABLE SSANGPIEUP OE" },
	{ 0x3838 , "HANGUL SYLLABLE SSANGPIEUP YO" },
	{ 0x3839 , "HANGUL SYLLABLE SSANGPIEUP YO IEUNG" },
	{ 0x383A , "HANGUL SYLLABLE SSANGPIEUP U" },
	{ 0x383B , "HANGUL SYLLABLE SSANGPIEUP U KIYEOK" },
	{ 0x383C , "HANGUL SYLLABLE SSANGPIEUP U NIEUN" },
	{ 0x383D , "HANGUL SYLLABLE SSANGPIEUP U RIEUL" },
	{ 0x383E , "HANGUL SYLLABLE SSANGPIEUP U MIEUM" },
	{ 0x383F , "HANGUL SYLLABLE SSANGPIEUP U SIOS" },
	{ 0x3840 , "HANGUL SYLLABLE SSANGPIEUP U IEUNG" },
	{ 0x3841 , "HANGUL SYLLABLE SSANGPIEUP YU" },
	{ 0x3842 , "HANGUL SYLLABLE SSANGPIEUP YU IEUNG" },
	{ 0x3843 , "HANGUL SYLLABLE SSANGPIEUP EU" },
	{ 0x3844 , "HANGUL SYLLABLE SSANGPIEUP EU NIEUN" },
	{ 0x3845 , "HANGUL SYLLABLE SSANGPIEUP EU RIEUL" },
	{ 0x3846 , "HANGUL SYLLABLE SSANGPIEUP EU MIEUM" },
	{ 0x3847 , "HANGUL SYLLABLE SSANGPIEUP EU PIEUP" },
	{ 0x3848 , "HANGUL SYLLABLE SSANGPIEUP I" },
	{ 0x3849 , "HANGUL SYLLABLE SSANGPIEUP I KIYEOK" },
	{ 0x384A , "HANGUL SYLLABLE SSANGPIEUP I NIEUN" },
	{ 0x384B , "HANGUL SYLLABLE SSANGPIEUP I RIEUL" },
	{ 0x384C , "HANGUL SYLLABLE SSANGPIEUP I MIEUM" },
	{ 0x384D , "HANGUL SYLLABLE SSANGPIEUP I PIEUP" },
	{ 0x384E , "HANGUL SYLLABLE SSANGPIEUP I SSANGSIOS" },
	{ 0x384F , "HANGUL SYLLABLE SSANGPIEUP I IEUNG" },
	{ 0x3850 , "HANGUL SYLLABLE SIOS A" },
	{ 0x3851 , "HANGUL SYLLABLE SIOS A KIYEOK" },
	{ 0x3852 , "HANGUL SYLLABLE SIOS A KIYEOK-SIOS" },
	{ 0x3853 , "HANGUL SYLLABLE SIOS A NIEUN" },
	{ 0x3854 , "HANGUL SYLLABLE SIOS A TIKEUT" },
	{ 0x3855 , "HANGUL SYLLABLE SIOS A RIEUL" },
	{ 0x3856 , "HANGUL SYLLABLE SIOS A RIEUL-KIYEOK" },
	{ 0x3857 , "HANGUL SYLLABLE SIOS A RIEUL-MIEUM" },
	{ 0x3858 , "HANGUL SYLLABLE SIOS A MIEUM" },
	{ 0x3859 , "HANGUL SYLLABLE SIOS A PIEUP" },
	{ 0x385A , "HANGUL SYLLABLE SIOS A SIOS" },
	{ 0x385B , "HANGUL SYLLABLE SIOS A SSANGSIOS" },
	{ 0x385C , "HANGUL SYLLABLE SIOS A IEUNG" },
	{ 0x385D , "HANGUL SYLLABLE SIOS A THIEUTH" },
	{ 0x385E , "HANGUL SYLLABLE SIOS AE" },
	{ 0x385F , "HANGUL SYLLABLE SIOS AE KIYEOK" },
	{ 0x3860 , "HANGUL SYLLABLE SIOS AE NIEUN" },
	{ 0x3861 , "HANGUL SYLLABLE SIOS AE RIEUL" },
	{ 0x3862 , "HANGUL SYLLABLE SIOS AE MIEUM" },
	{ 0x3863 , "HANGUL SYLLABLE SIOS AE PIEUP" },
	{ 0x3864 , "HANGUL SYLLABLE SIOS AE SIOS" },
	{ 0x3865 , "HANGUL SYLLABLE SIOS AE SSANGSIOS" },
	{ 0x3866 , "HANGUL SYLLABLE SIOS AE IEUNG" },
	{ 0x3867 , "HANGUL SYLLABLE SIOS YA" },
	{ 0x3868 , "HANGUL SYLLABLE SIOS YA KIYEOK" },
	{ 0x3869 , "HANGUL SYLLABLE SIOS YA NIEUN" },
	{ 0x386A , "HANGUL SYLLABLE SIOS YA RIEUL" },
	{ 0x386B , "HANGUL SYLLABLE SIOS YA MIEUM" },
	{ 0x386C , "HANGUL SYLLABLE SIOS YA PIEUP" },
	{ 0x386D , "HANGUL SYLLABLE SIOS YA SIOS" },
	{ 0x386E , "HANGUL SYLLABLE SIOS YA IEUNG" },
	{ 0x386F , "HANGUL SYLLABLE SIOS YAE" },
	{ 0x3870 , "HANGUL SYLLABLE SIOS YAE NIEUN" },
	{ 0x3871 , "HANGUL SYLLABLE SIOS YAE RIEUL" },
	{ 0x3872 , "HANGUL SYLLABLE SIOS YAE MIEUM" },
	{ 0x3873 , "HANGUL SYLLABLE SIOS YAE IEUNG" },
	{ 0x3874 , "HANGUL SYLLABLE SIOS EO" },
	{ 0x3875 , "HANGUL SYLLABLE SIOS EO KIYEOK" },
	{ 0x3876 , "HANGUL SYLLABLE SIOS EO SSANGKIYEOK" },
	{ 0x3877 , "HANGUL SYLLABLE SIOS EO KIYEOK-SIOS" },
	{ 0x3878 , "HANGUL SYLLABLE SIOS EO NIEUN" },
	{ 0x3879 , "HANGUL SYLLABLE SIOS EO TIKEUT" },
	{ 0x387A , "HANGUL SYLLABLE SIOS EO RIEUL" },
	{ 0x387B , "HANGUL SYLLABLE SIOS EO RIEUL-MIEUM" },
	{ 0x387C , "HANGUL SYLLABLE SIOS EO RIEUL-PIEUP" },
	{ 0x387D , "HANGUL SYLLABLE SIOS EO MIEUM" },
	{ 0x387E , "HANGUL SYLLABLE SIOS EO PIEUP" },
	{ 0x387F , "HANGUL SYLLABLE SIOS EO SIOS" },
	{ 0x3880 , "HANGUL SYLLABLE SIOS EO SSANGSIOS" },
	{ 0x3881 , "HANGUL SYLLABLE SIOS EO IEUNG" },
	{ 0x3882 , "HANGUL SYLLABLE SIOS EO PHIEUPH" },
	{ 0x3883 , "HANGUL SYLLABLE SIOS E" },
	{ 0x3884 , "HANGUL SYLLABLE SIOS E KIYEOK" },
	{ 0x3885 , "HANGUL SYLLABLE SIOS E NIEUN" },
	{ 0x3886 , "HANGUL SYLLABLE SIOS E RIEUL" },
	{ 0x3887 , "HANGUL SYLLABLE SIOS E MIEUM" },
	{ 0x3888 , "HANGUL SYLLABLE SIOS E PIEUP" },
	{ 0x3889 , "HANGUL SYLLABLE SIOS E SIOS" },
	{ 0x388A , "HANGUL SYLLABLE SIOS E SSANGSIOS" },
	{ 0x388B , "HANGUL SYLLABLE SIOS E IEUNG" },
	{ 0x388C , "HANGUL SYLLABLE SIOS YEO" },
	{ 0x388D , "HANGUL SYLLABLE SIOS YEO KIYEOK" },
	{ 0x388E , "HANGUL SYLLABLE SIOS YEO NIEUN" },
	{ 0x388F , "HANGUL SYLLABLE SIOS YEO RIEUL" },
	{ 0x3890 , "HANGUL SYLLABLE SIOS YEO MIEUM" },
	{ 0x3891 , "HANGUL SYLLABLE SIOS YEO PIEUP" },
	{ 0x3892 , "HANGUL SYLLABLE SIOS YEO SIOS" },
	{ 0x3893 , "HANGUL SYLLABLE SIOS YEO SSANGSIOS" },
	{ 0x3894 , "HANGUL SYLLABLE SIOS YEO IEUNG" },
	{ 0x3895 , "HANGUL SYLLABLE SIOS YE" },
	{ 0x3896 , "HANGUL SYLLABLE SIOS YE NIEUN" },
	{ 0x3897 , "HANGUL SYLLABLE SIOS YE RIEUL" },
	{ 0x3898 , "HANGUL SYLLABLE SIOS YE IEUNG" },
	{ 0x3899 , "HANGUL SYLLABLE SIOS O" },
	{ 0x389A , "HANGUL SYLLABLE SIOS O KIYEOK" },
	{ 0x389B , "HANGUL SYLLABLE SIOS O SSANGKIYEOK" },
	{ 0x389C , "HANGUL SYLLABLE SIOS O NIEUN" },
	{ 0x389D , "HANGUL SYLLABLE SIOS O RIEUL" },
	{ 0x389E , "HANGUL SYLLABLE SIOS O RIEUL-MIEUM" },
	{ 0x389F , "HANGUL SYLLABLE SIOS O MIEUM" },
	{ 0x38A0 , "HANGUL SYLLABLE SIOS O PIEUP" },
	{ 0x38A1 , "HANGUL SYLLABLE SIOS O SIOS" },
	{ 0x38A2 , "HANGUL SYLLABLE SIOS O IEUNG" },
	{ 0x38A3 , "HANGUL SYLLABLE SIOS O THIEUTH" },
	{ 0x38A4 , "HANGUL SYLLABLE SIOS WA" },
	{ 0x38A5 , "HANGUL SYLLABLE SIOS WA KIYEOK" },
	{ 0x38A6 , "HANGUL SYLLABLE SIOS WA NIEUN" },
	{ 0x38A7 , "HANGUL SYLLABLE SIOS WA RIEUL" },
	{ 0x38A8 , "HANGUL SYLLABLE SIOS WA IEUNG" },
	{ 0x38A9 , "HANGUL SYLLABLE SIOS WAE" },
	{ 0x38AA , "HANGUL SYLLABLE SIOS WAE NIEUN" },
	{ 0x38AB , "HANGUL SYLLABLE SIOS WAE RIEUL" },
	{ 0x38AC , "HANGUL SYLLABLE SIOS WAE MIEUM" },
	{ 0x38AD , "HANGUL SYLLABLE SIOS WAE SIOS" },
	{ 0x38AE , "HANGUL SYLLABLE SIOS WAE SSANGSIOS" },
	{ 0x38AF , "HANGUL SYLLABLE SIOS OE" },
	{ 0x38B0 , "HANGUL SYLLABLE SIOS OE NIEUN" },
	{ 0x38B1 , "HANGUL SYLLABLE SIOS OE RIEUL" },
	{ 0x38B2 , "HANGUL SYLLABLE SIOS OE MIEUM" },
	{ 0x38B3 , "HANGUL SYLLABLE SIOS OE PIEUP" },
	{ 0x38B4 , "HANGUL SYLLABLE SIOS OE SIOS" },
	{ 0x38B5 , "HANGUL SYLLABLE SIOS YO" },
	{ 0x38B6 , "HANGUL SYLLABLE SIOS YO KIYEOK" },
	{ 0x38B7 , "HANGUL SYLLABLE SIOS YO NIEUN" },
	{ 0x38B8 , "HANGUL SYLLABLE SIOS YO RIEUL" },
	{ 0x38B9 , "HANGUL SYLLABLE SIOS YO MIEUM" },
	{ 0x38BA , "HANGUL SYLLABLE SIOS YO PIEUP" },
	{ 0x38BB , "HANGUL SYLLABLE SIOS YO SIOS" },
	{ 0x38BC , "HANGUL SYLLABLE SIOS YO IEUNG" },
	{ 0x38BD , "HANGUL SYLLABLE SIOS U" },
	{ 0x38BE , "HANGUL SYLLABLE SIOS U KIYEOK" },
	{ 0x38BF , "HANGUL SYLLABLE SIOS U NIEUN" },
	{ 0x38C0 , "HANGUL SYLLABLE SIOS U TIKEUT" },
	{ 0x38C1 , "HANGUL SYLLABLE SIOS U RIEUL" },
	{ 0x38C2 , "HANGUL SYLLABLE SIOS U MIEUM" },
	{ 0x38C3 , "HANGUL SYLLABLE SIOS U PIEUP" },
	{ 0x38C4 , "HANGUL SYLLABLE SIOS U SIOS" },
	{ 0x38C5 , "HANGUL SYLLABLE SIOS U IEUNG" },
	{ 0x38C6 , "HANGUL SYLLABLE SIOS U CHIEUCH" },
	{ 0x38C7 , "HANGUL SYLLABLE SIOS U THIEUTH" },
	{ 0x38C8 , "HANGUL SYLLABLE SIOS U PHIEUPH" },
	{ 0x38C9 , "HANGUL SYLLABLE SIOS WEO" },
	{ 0x38CA , "HANGUL SYLLABLE SIOS WEO SSANGSIOS" },
	{ 0x38CB , "HANGUL SYLLABLE SIOS WE" },
	{ 0x38CC , "HANGUL SYLLABLE SIOS WE KIYEOK" },
	{ 0x38CD , "HANGUL SYLLABLE SIOS WE NIEUN" },
	{ 0x38CE , "HANGUL SYLLABLE SIOS WE RIEUL" },
	{ 0x38CF , "HANGUL SYLLABLE SIOS WE MIEUM" },
	{ 0x38D0 , "HANGUL SYLLABLE SIOS WE IEUNG" },
	{ 0x38D1 , "HANGUL SYLLABLE SIOS WI" },
	{ 0x38D2 , "HANGUL SYLLABLE SIOS WI KIYEOK" },
	{ 0x38D3 , "HANGUL SYLLABLE SIOS WI NIEUN" },
	{ 0x38D4 , "HANGUL SYLLABLE SIOS WI RIEUL" },
	{ 0x38D5 , "HANGUL SYLLABLE SIOS WI MIEUM" },
	{ 0x38D6 , "HANGUL SYLLABLE SIOS WI PIEUP" },
	{ 0x38D7 , "HANGUL SYLLABLE SIOS WI SIOS" },
	{ 0x38D8 , "HANGUL SYLLABLE SIOS WI IEUNG" },
	{ 0x38D9 , "HANGUL SYLLABLE SIOS YU" },
	{ 0x38DA , "HANGUL SYLLABLE SIOS YU KIYEOK" },
	{ 0x38DB , "HANGUL SYLLABLE SIOS YU RIEUL" },
	{ 0x38DC , "HANGUL SYLLABLE SIOS YU MIEUM" },
	{ 0x38DD , "HANGUL SYLLABLE SIOS YU SIOS" },
	{ 0x38DE , "HANGUL SYLLABLE SIOS YU IEUNG" },
	{ 0x38DF , "HANGUL SYLLABLE SIOS EU" },
	{ 0x38E0 , "HANGUL SYLLABLE SIOS EU KIYEOK" },
	{ 0x38E1 , "HANGUL SYLLABLE SIOS EU NIEUN" },
	{ 0x38E2 , "HANGUL SYLLABLE SIOS EU RIEUL" },
	{ 0x38E3 , "HANGUL SYLLABLE SIOS EU RIEUL-KIYEOK" },
	{ 0x38E4 , "HANGUL SYLLABLE SIOS EU MIEUM" },
	{ 0x38E5 , "HANGUL SYLLABLE SIOS EU PIEUP" },
	{ 0x38E6 , "HANGUL SYLLABLE SIOS EU SIOS" },
	{ 0x38E7 , "HANGUL SYLLABLE SIOS EU IEUNG" },
	{ 0x38E8 , "HANGUL SYLLABLE SIOS I" },
	{ 0x38E9 , "HANGUL SYLLABLE SIOS I KIYEOK" },
	{ 0x38EA , "HANGUL SYLLABLE SIOS I NIEUN" },
	{ 0x38EB , "HANGUL SYLLABLE SIOS I TIKEUT" },
	{ 0x38EC , "HANGUL SYLLABLE SIOS I RIEUL" },
	{ 0x38ED , "HANGUL SYLLABLE SIOS I RIEUL-HIEUH" },
	{ 0x38EE , "HANGUL SYLLABLE SIOS I MIEUM" },
	{ 0x38EF , "HANGUL SYLLABLE SIOS I PIEUP" },
	{ 0x38F0 , "HANGUL SYLLABLE SIOS I SIOS" },
	{ 0x38F1 , "HANGUL SYLLABLE SIOS I IEUNG" },
	{ 0x38F2 , "HANGUL SYLLABLE SIOS I PHIEUPH" },
	{ 0x38F3 , "HANGUL SYLLABLE SSANGSIOS A" },
	{ 0x38F4 , "HANGUL SYLLABLE SSANGSIOS A KIYEOK" },
	{ 0x38F5 , "HANGUL SYLLABLE SSANGSIOS A KIYEOK-SIOS" },
	{ 0x38F6 , "HANGUL SYLLABLE SSANGSIOS A NIEUN" },
	{ 0x38F7 , "HANGUL SYLLABLE SSANGSIOS A RIEUL" },
	{ 0x38F8 , "HANGUL SYLLABLE SSANGSIOS A MIEUM" },
	{ 0x38F9 , "HANGUL SYLLABLE SSANGSIOS A PIEUP" },
	{ 0x38FA , "HANGUL SYLLABLE SSANGSIOS A SSANGSIOS" },
	{ 0x38FB , "HANGUL SYLLABLE SSANGSIOS A IEUNG" },
	{ 0x38FC , "HANGUL SYLLABLE SSANGSIOS A HIEUH" },
	{ 0x38FD , "HANGUL SYLLABLE SSANGSIOS AE" },
	{ 0x38FE , "HANGUL SYLLABLE SSANGSIOS AE KIYEOK" },
	{ 0x38FF , "HANGUL SYLLABLE SSANGSIOS AE NIEUN" },
	{ 0x3900 , "HANGUL SYLLABLE SSANGSIOS AE RIEUL" },
	{ 0x3901 , "HANGUL SYLLABLE SSANGSIOS AE MIEUM" },
	{ 0x3902 , "HANGUL SYLLABLE SSANGSIOS AE PIEUP" },
	{ 0x3903 , "HANGUL SYLLABLE SSANGSIOS AE SSANGSIOS" },
	{ 0x3904 , "HANGUL SYLLABLE SSANGSIOS AE IEUNG" },
	{ 0x3905 , "HANGUL SYLLABLE SSANGSIOS YA IEUNG" },
	{ 0x3906 , "HANGUL SYLLABLE SSANGSIOS EO" },
	{ 0x3907 , "HANGUL SYLLABLE SSANGSIOS EO KIYEOK" },
	{ 0x3908 , "HANGUL SYLLABLE SSANGSIOS EO NIEUN" },
	{ 0x3909 , "HANGUL SYLLABLE SSANGSIOS EO RIEUL" },
	{ 0x390A , "HANGUL SYLLABLE SSANGSIOS EO RIEUL-MIEUM" },
	{ 0x390B , "HANGUL SYLLABLE SSANGSIOS EO MIEUM" },
	{ 0x390C , "HANGUL SYLLABLE SSANGSIOS EO PIEUP" },
	{ 0x390D , "HANGUL SYLLABLE SSANGSIOS EO SSANGSIOS" },
	{ 0x390E , "HANGUL SYLLABLE SSANGSIOS EO IEUNG" },
	{ 0x390F , "HANGUL SYLLABLE SSANGSIOS E" },
	{ 0x3910 , "HANGUL SYLLABLE SSANGSIOS E NIEUN" },
	{ 0x3911 , "HANGUL SYLLABLE SSANGSIOS E RIEUL" },
	{ 0x3912 , "HANGUL SYLLABLE SSANGSIOS YE NIEUN" },
	{ 0x3913 , "HANGUL SYLLABLE SSANGSIOS O" },
	{ 0x3914 , "HANGUL SYLLABLE SSANGSIOS O KIYEOK" },
	{ 0x3915 , "HANGUL SYLLABLE SSANGSIOS O NIEUN" },
	{ 0x3916 , "HANGUL SYLLABLE SSANGSIOS O TIKEUT" },
	{ 0x3917 , "HANGUL SYLLABLE SSANGSIOS O RIEUL" },
	{ 0x3918 , "HANGUL SYLLABLE SSANGSIOS O RIEUL-MIEUM" },
	{ 0x3919 , "HANGUL SYLLABLE SSANGSIOS O MIEUM" },
	{ 0x391A , "HANGUL SYLLABLE SSANGSIOS O PIEUP" },
	{ 0x391B , "HANGUL SYLLABLE SSANGSIOS O IEUNG" },
	{ 0x391C , "HANGUL SYLLABLE SSANGSIOS WA" },
	{ 0x391D , "HANGUL SYLLABLE SSANGSIOS WA KIYEOK" },
	{ 0x391E , "HANGUL SYLLABLE SSANGSIOS WA NIEUN" },
	{ 0x391F , "HANGUL SYLLABLE SSANGSIOS WA SSANGSIOS" },
	{ 0x3920 , "HANGUL SYLLABLE SSANGSIOS WAE" },
	{ 0x3921 , "HANGUL SYLLABLE SSANGSIOS WAE SSANGSIOS" },
	{ 0x3922 , "HANGUL SYLLABLE SSANGSIOS OE" },
	{ 0x3923 , "HANGUL SYLLABLE SSANGSIOS OE NIEUN" },
	{ 0x3924 , "HANGUL SYLLABLE SSANGSIOS OE RIEUL" },
	{ 0x3925 , "HANGUL SYLLABLE SSANGSIOS OE MIEUM" },
	{ 0x3926 , "HANGUL SYLLABLE SSANGSIOS OE PIEUP" },
	{ 0x3927 , "HANGUL SYLLABLE SSANGSIOS YO" },
	{ 0x3928 , "HANGUL SYLLABLE SSANGSIOS U" },
	{ 0x3929 , "HANGUL SYLLABLE SSANGSIOS U KIYEOK" },
	{ 0x392A , "HANGUL SYLLABLE SSANGSIOS U NIEUN" },
	{ 0x392B , "HANGUL SYLLABLE SSANGSIOS U RIEUL" },
	{ 0x392C , "HANGUL SYLLABLE SSANGSIOS U MIEUM" },
	{ 0x392D , "HANGUL SYLLABLE SSANGSIOS U PIEUP" },
	{ 0x392E , "HANGUL SYLLABLE SSANGSIOS U IEUNG" },
	{ 0x392F , "HANGUL SYLLABLE SSANGSIOS WEO" },
	{ 0x3930 , "HANGUL SYLLABLE SSANGSIOS WEO SSANGSIOS" },
	{ 0x3931 , "HANGUL SYLLABLE SSANGSIOS WE" },
	{ 0x3932 , "HANGUL SYLLABLE SSANGSIOS WI" },
	{ 0x3933 , "HANGUL SYLLABLE SSANGSIOS WI NIEUN" },
	{ 0x3934 , "HANGUL SYLLABLE SSANGSIOS YU IEUNG" },
	{ 0x3935 , "HANGUL SYLLABLE SSANGSIOS EU" },
	{ 0x3936 , "HANGUL SYLLABLE SSANGSIOS EU KIYEOK" },
	{ 0x3937 , "HANGUL SYLLABLE SSANGSIOS EU NIEUN" },
	{ 0x3938 , "HANGUL SYLLABLE SSANGSIOS EU RIEUL" },
	{ 0x3939 , "HANGUL SYLLABLE SSANGSIOS EU RIEUL-MIEUM" },
	{ 0x393A , "HANGUL SYLLABLE SSANGSIOS EU RIEUL-HIEUH" },
	{ 0x393B , "HANGUL SYLLABLE SSANGSIOS EU MIEUM" },
	{ 0x393C , "HANGUL SYLLABLE SSANGSIOS EU PIEUP" },
	{ 0x393D , "HANGUL SYLLABLE SSANGSIOS YI" },
	{ 0x393E , "HANGUL SYLLABLE SSANGSIOS YI NIEUN" },
	{ 0x393F , "HANGUL SYLLABLE SSANGSIOS YI RIEUL" },
	{ 0x3940 , "HANGUL SYLLABLE SSANGSIOS YI MIEUM" },
	{ 0x3941 , "HANGUL SYLLABLE SSANGSIOS I" },
	{ 0x3942 , "HANGUL SYLLABLE SSANGSIOS I KIYEOK" },
	{ 0x3943 , "HANGUL SYLLABLE SSANGSIOS I NIEUN" },
	{ 0x3944 , "HANGUL SYLLABLE SSANGSIOS I RIEUL" },
	{ 0x3945 , "HANGUL SYLLABLE SSANGSIOS I MIEUM" },
	{ 0x3946 , "HANGUL SYLLABLE SSANGSIOS I PIEUP" },
	{ 0x3947 , "HANGUL SYLLABLE SSANGSIOS I SIOS" },
	{ 0x3948 , "HANGUL SYLLABLE SSANGSIOS I IEUNG" },
	{ 0x3949 , "HANGUL SYLLABLE IEUNG A" },
	{ 0x394A , "HANGUL SYLLABLE IEUNG A KIYEOK" },
	{ 0x394B , "HANGUL SYLLABLE IEUNG A NIEUN" },
	{ 0x394C , "HANGUL SYLLABLE IEUNG A NIEUN-CIEUC" },
	{ 0x394D , "HANGUL SYLLABLE IEUNG A NIEUN-HIEUH" },
	{ 0x394E , "HANGUL SYLLABLE IEUNG A RIEUL" },
	{ 0x394F , "HANGUL SYLLABLE IEUNG A RIEUL-KIYEOK" },
	{ 0x3950 , "HANGUL SYLLABLE IEUNG A RIEUL-MIEUM" },
	{ 0x3951 , "HANGUL SYLLABLE IEUNG A RIEUL-HIEUH" },
	{ 0x3952 , "HANGUL SYLLABLE IEUNG A MIEUM" },
	{ 0x3953 , "HANGUL SYLLABLE IEUNG A PIEUP" },
	{ 0x3954 , "HANGUL SYLLABLE IEUNG A SIOS" },
	{ 0x3955 , "HANGUL SYLLABLE IEUNG A SSANGSIOS" },
	{ 0x3956 , "HANGUL SYLLABLE IEUNG A IEUNG" },
	{ 0x3957 , "HANGUL SYLLABLE IEUNG A THIEUTH" },
	{ 0x3958 , "HANGUL SYLLABLE IEUNG A PHIEUPH" },
	{ 0x3959 , "HANGUL SYLLABLE IEUNG AE" },
	{ 0x395A , "HANGUL SYLLABLE IEUNG AE KIYEOK" },
	{ 0x395B , "HANGUL SYLLABLE IEUNG AE NIEUN" },
	{ 0x395C , "HANGUL SYLLABLE IEUNG AE RIEUL" },
	{ 0x395D , "HANGUL SYLLABLE IEUNG AE MIEUM" },
	{ 0x395E , "HANGUL SYLLABLE IEUNG AE PIEUP" },
	{ 0x395F , "HANGUL SYLLABLE IEUNG AE SIOS" },
	{ 0x3960 , "HANGUL SYLLABLE IEUNG AE SSANGSIOS" },
	{ 0x3961 , "HANGUL SYLLABLE IEUNG AE IEUNG" },
	{ 0x3962 , "HANGUL SYLLABLE IEUNG YA" },
	{ 0x3963 , "HANGUL SYLLABLE IEUNG YA KIYEOK" },
	{ 0x3964 , "HANGUL SYLLABLE IEUNG YA NIEUN" },
	{ 0x3965 , "HANGUL SYLLABLE IEUNG YA RIEUL" },
	{ 0x3966 , "HANGUL SYLLABLE IEUNG YA RIEUL-PIEUP" },
	{ 0x3967 , "HANGUL SYLLABLE IEUNG YA MIEUM" },
	{ 0x3968 , "HANGUL SYLLABLE IEUNG YA PIEUP" },
	{ 0x3969 , "HANGUL SYLLABLE IEUNG YA SIOS" },
	{ 0x396A , "HANGUL SYLLABLE IEUNG YA IEUNG" },
	{ 0x396B , "HANGUL SYLLABLE IEUNG YA THIEUTH" },
	{ 0x396C , "HANGUL SYLLABLE IEUNG YA HIEUH" },
	{ 0x396D , "HANGUL SYLLABLE IEUNG YAE" },
	{ 0x396E , "HANGUL SYLLABLE IEUNG YAE NIEUN" },
	{ 0x396F , "HANGUL SYLLABLE IEUNG YAE RIEUL" },
	{ 0x3970 , "HANGUL SYLLABLE IEUNG YAE PIEUP" },
	{ 0x3971 , "HANGUL SYLLABLE IEUNG EO" },
	{ 0x3972 , "HANGUL SYLLABLE IEUNG EO KIYEOK" },
	{ 0x3973 , "HANGUL SYLLABLE IEUNG EO NIEUN" },
	{ 0x3974 , "HANGUL SYLLABLE IEUNG EO NIEUN-CIEUC" },
	{ 0x3975 , "HANGUL SYLLABLE IEUNG EO TIKEUT" },
	{ 0x3976 , "HANGUL SYLLABLE IEUNG EO RIEUL" },
	{ 0x3977 , "HANGUL SYLLABLE IEUNG EO RIEUL-KIYEOK" },
	{ 0x3978 , "HANGUL SYLLABLE IEUNG EO RIEUL-MIEUM" },
	{ 0x3979 , "HANGUL SYLLABLE IEUNG EO MIEUM" },
	{ 0x397A , "HANGUL SYLLABLE IEUNG EO PIEUP" },
	{ 0x397B , "HANGUL SYLLABLE IEUNG EO PIEUP-SIOS" },
	{ 0x397C , "HANGUL SYLLABLE IEUNG EO SIOS" },
	{ 0x397D , "HANGUL SYLLABLE IEUNG EO SSANGSIOS" },
	{ 0x397E , "HANGUL SYLLABLE IEUNG EO IEUNG" },
	{ 0x397F , "HANGUL SYLLABLE IEUNG EO CIEUC" },
	{ 0x3980 , "HANGUL SYLLABLE IEUNG EO KHIEUKH" },
	{ 0x3981 , "HANGUL SYLLABLE IEUNG EO PHIEUPH" },
	{ 0x3982 , "HANGUL SYLLABLE IEUNG E" },
	{ 0x3983 , "HANGUL SYLLABLE IEUNG E KIYEOK" },
	{ 0x3984 , "HANGUL SYLLABLE IEUNG E NIEUN" },
	{ 0x3985 , "HANGUL SYLLABLE IEUNG E RIEUL" },
	{ 0x3986 , "HANGUL SYLLABLE IEUNG E MIEUM" },
	{ 0x3987 , "HANGUL SYLLABLE IEUNG E PIEUP" },
	{ 0x3988 , "HANGUL SYLLABLE IEUNG E SIOS" },
	{ 0x3989 , "HANGUL SYLLABLE IEUNG E IEUNG" },
	{ 0x398A , "HANGUL SYLLABLE IEUNG YEO" },
	{ 0x398B , "HANGUL SYLLABLE IEUNG YEO KIYEOK" },
	{ 0x398C , "HANGUL SYLLABLE IEUNG YEO SSANGKIYEOK" },
	{ 0x398D , "HANGUL SYLLABLE IEUNG YEO NIEUN" },
	{ 0x398E , "HANGUL SYLLABLE IEUNG YEO RIEUL" },
	{ 0x398F , "HANGUL SYLLABLE IEUNG YEO RIEUL-MIEUM" },
	{ 0x3990 , "HANGUL SYLLABLE IEUNG YEO RIEUL-PIEUP" },
	{ 0x3991 , "HANGUL SYLLABLE IEUNG YEO MIEUM" },
	{ 0x3992 , "HANGUL SYLLABLE IEUNG YEO PIEUP" },
	{ 0x3993 , "HANGUL SYLLABLE IEUNG YEO PIEUP-SIOS" },
	{ 0x3994 , "HANGUL SYLLABLE IEUNG YEO SIOS" },
	{ 0x3995 , "HANGUL SYLLABLE IEUNG YEO SSANGSIOS" },
	{ 0x3996 , "HANGUL SYLLABLE IEUNG YEO IEUNG" },
	{ 0x3997 , "HANGUL SYLLABLE IEUNG YEO THIEUTH" },
	{ 0x3998 , "HANGUL SYLLABLE IEUNG YEO PHIEUPH" },
	{ 0x3999 , "HANGUL SYLLABLE IEUNG YEO HIEUH" },
	{ 0x399A , "HANGUL SYLLABLE IEUNG YE" },
	{ 0x399B , "HANGUL SYLLABLE IEUNG YE NIEUN" },
	{ 0x399C , "HANGUL SYLLABLE IEUNG YE RIEUL" },
	{ 0x399D , "HANGUL SYLLABLE IEUNG YE MIEUM" },
	{ 0x399E , "HANGUL SYLLABLE IEUNG YE PIEUP" },
	{ 0x399F , "HANGUL SYLLABLE IEUNG YE SIOS" },
	{ 0x39A0 , "HANGUL SYLLABLE IEUNG YE SSANGSIOS" },
	{ 0x39A1 , "HANGUL SYLLABLE IEUNG O" },
	{ 0x39A2 , "HANGUL SYLLABLE IEUNG O KIYEOK" },
	{ 0x39A3 , "HANGUL SYLLABLE IEUNG O NIEUN" },
	{ 0x39A4 , "HANGUL SYLLABLE IEUNG O RIEUL" },
	{ 0x39A5 , "HANGUL SYLLABLE IEUNG O RIEUL-KIYEOK" },
	{ 0x39A6 , "HANGUL SYLLABLE IEUNG O RIEUL-MIEUM" },
	{ 0x39A7 , "HANGUL SYLLABLE IEUNG O RIEUL-SIOS" },
	{ 0x39A8 , "HANGUL SYLLABLE IEUNG O RIEUL-HIEUH" },
	{ 0x39A9 , "HANGUL SYLLABLE IEUNG O MIEUM" },
	{ 0x39AA , "HANGUL SYLLABLE IEUNG O PIEUP" },
	{ 0x39AB , "HANGUL SYLLABLE IEUNG O SIOS" },
	{ 0x39AC , "HANGUL SYLLABLE IEUNG O IEUNG" },
	{ 0x39AD , "HANGUL SYLLABLE IEUNG O CHIEUCH" },
	{ 0x39AE , "HANGUL SYLLABLE IEUNG WA" },
	{ 0x39AF , "HANGUL SYLLABLE IEUNG WA KIYEOK" },
	{ 0x39B0 , "HANGUL SYLLABLE IEUNG WA NIEUN" },
	{ 0x39B1 , "HANGUL SYLLABLE IEUNG WA RIEUL" },
	{ 0x39B2 , "HANGUL SYLLABLE IEUNG WA MIEUM" },
	{ 0x39B3 , "HANGUL SYLLABLE IEUNG WA PIEUP" },
	{ 0x39B4 , "HANGUL SYLLABLE IEUNG WA SIOS" },
	{ 0x39B5 , "HANGUL SYLLABLE IEUNG WA SSANGSIOS" },
	{ 0x39B6 , "HANGUL SYLLABLE IEUNG WA IEUNG" },
	{ 0x39B7 , "HANGUL SYLLABLE IEUNG WAE" },
	{ 0x39B8 , "HANGUL SYLLABLE IEUNG WAE KIYEOK" },
	{ 0x39B9 , "HANGUL SYLLABLE IEUNG WAE NIEUN" },
	{ 0x39BA , "HANGUL SYLLABLE IEUNG WAE MIEUM" },
	{ 0x39BB , "HANGUL SYLLABLE IEUNG WAE SIOS" },
	{ 0x39BC , "HANGUL SYLLABLE IEUNG WAE IEUNG" },
	{ 0x39BD , "HANGUL SYLLABLE IEUNG OE" },
	{ 0x39BE , "HANGUL SYLLABLE IEUNG OE KIYEOK" },
	{ 0x39BF , "HANGUL SYLLABLE IEUNG OE NIEUN" },
	{ 0x39C0 , "HANGUL SYLLABLE IEUNG OE RIEUL" },
	{ 0x39C1 , "HANGUL SYLLABLE IEUNG OE MIEUM" },
	{ 0x39C2 , "HANGUL SYLLABLE IEUNG OE PIEUP" },
	{ 0x39C3 , "HANGUL SYLLABLE IEUNG OE SIOS" },
	{ 0x39C4 , "HANGUL SYLLABLE IEUNG OE IEUNG" },
	{ 0x39C5 , "HANGUL SYLLABLE IEUNG YO" },
	{ 0x39C6 , "HANGUL SYLLABLE IEUNG YO KIYEOK" },
	{ 0x39C7 , "HANGUL SYLLABLE IEUNG YO NIEUN" },
	{ 0x39C8 , "HANGUL SYLLABLE IEUNG YO RIEUL" },
	{ 0x39C9 , "HANGUL SYLLABLE IEUNG YO MIEUM" },
	{ 0x39CA , "HANGUL SYLLABLE IEUNG YO PIEUP" },
	{ 0x39CB , "HANGUL SYLLABLE IEUNG YO SIOS" },
	{ 0x39CC , "HANGUL SYLLABLE IEUNG YO IEUNG" },
	{ 0x39CD , "HANGUL SYLLABLE IEUNG U" },
	{ 0x39CE , "HANGUL SYLLABLE IEUNG U KIYEOK" },
	{ 0x39CF , "HANGUL SYLLABLE IEUNG U NIEUN" },
	{ 0x39D0 , "HANGUL SYLLABLE IEUNG U RIEUL" },
	{ 0x39D1 , "HANGUL SYLLABLE IEUNG U RIEUL-KIYEOK" },
	{ 0x39D2 , "HANGUL SYLLABLE IEUNG U RIEUL-MIEUM" },
	{ 0x39D3 , "HANGUL SYLLABLE IEUNG U MIEUM" },
	{ 0x39D4 , "HANGUL SYLLABLE IEUNG U PIEUP" },
	{ 0x39D5 , "HANGUL SYLLABLE IEUNG U SIOS" },
	{ 0x39D6 , "HANGUL SYLLABLE IEUNG U IEUNG" },
	{ 0x39D7 , "HANGUL SYLLABLE IEUNG WEO" },
	{ 0x39D8 , "HANGUL SYLLABLE IEUNG WEO KIYEOK" },
	{ 0x39D9 , "HANGUL SYLLABLE IEUNG WEO NIEUN" },
	{ 0x39DA , "HANGUL SYLLABLE IEUNG WEO RIEUL" },
	{ 0x39DB , "HANGUL SYLLABLE IEUNG WEO MIEUM" },
	{ 0x39DC , "HANGUL SYLLABLE IEUNG WEO PIEUP" },
	{ 0x39DD , "HANGUL SYLLABLE IEUNG WEO SSANGSIOS" },
	{ 0x39DE , "HANGUL SYLLABLE IEUNG WEO IEUNG" },
	{ 0x39DF , "HANGUL SYLLABLE IEUNG WE" },
	{ 0x39E0 , "HANGUL SYLLABLE IEUNG WE KIYEOK" },
	{ 0x39E1 , "HANGUL SYLLABLE IEUNG WE NIEUN" },
	{ 0x39E2 , "HANGUL SYLLABLE IEUNG WE RIEUL" },
	{ 0x39E3 , "HANGUL SYLLABLE IEUNG WE MIEUM" },
	{ 0x39E4 , "HANGUL SYLLABLE IEUNG WE PIEUP" },
	{ 0x39E5 , "HANGUL SYLLABLE IEUNG WE IEUNG" },
	{ 0x39E6 , "HANGUL SYLLABLE IEUNG WI" },
	{ 0x39E7 , "HANGUL SYLLABLE IEUNG WI KIYEOK" },
	{ 0x39E8 , "HANGUL SYLLABLE IEUNG WI NIEUN" },
	{ 0x39E9 , "HANGUL SYLLABLE IEUNG WI RIEUL" },
	{ 0x39EA , "HANGUL SYLLABLE IEUNG WI MIEUM" },
	{ 0x39EB , "HANGUL SYLLABLE IEUNG WI PIEUP" },
	{ 0x39EC , "HANGUL SYLLABLE IEUNG WI SIOS" },
	{ 0x39ED , "HANGUL SYLLABLE IEUNG WI IEUNG" },
	{ 0x39EE , "HANGUL SYLLABLE IEUNG YU" },
	{ 0x39EF , "HANGUL SYLLABLE IEUNG YU KIYEOK" },
	{ 0x39F0 , "HANGUL SYLLABLE IEUNG YU NIEUN" },
	{ 0x39F1 , "HANGUL SYLLABLE IEUNG YU RIEUL" },
	{ 0x39F2 , "HANGUL SYLLABLE IEUNG YU MIEUM" },
	{ 0x39F3 , "HANGUL SYLLABLE IEUNG YU PIEUP" },
	{ 0x39F4 , "HANGUL SYLLABLE IEUNG YU SIOS" },
	{ 0x39F5 , "HANGUL SYLLABLE IEUNG YU IEUNG" },
	{ 0x39F6 , "HANGUL SYLLABLE IEUNG YU CHIEUCH" },
	{ 0x39F7 , "HANGUL SYLLABLE IEUNG EU" },
	{ 0x39F8 , "HANGUL SYLLABLE IEUNG EU KIYEOK" },
	{ 0x39F9 , "HANGUL SYLLABLE IEUNG EU NIEUN" },
	{ 0x39FA , "HANGUL SYLLABLE IEUNG EU RIEUL" },
	{ 0x39FB , "HANGUL SYLLABLE IEUNG EU RIEUL-PHIEUPH" },
	{ 0x39FC , "HANGUL SYLLABLE IEUNG EU MIEUM" },
	{ 0x39FD , "HANGUL SYLLABLE IEUNG EU PIEUP" },
	{ 0x39FE , "HANGUL SYLLABLE IEUNG EU SIOS" },
	{ 0x39FF , "HANGUL SYLLABLE IEUNG EU IEUNG" },
	{ 0x3A00 , "HANGUL SYLLABLE IEUNG EU CIEUC" },
	{ 0x3A01 , "HANGUL SYLLABLE IEUNG EU CHIEUCH" },
	{ 0x3A02 , "HANGUL SYLLABLE IEUNG EU KHIEUKH" },
	{ 0x3A03 , "HANGUL SYLLABLE IEUNG EU THIEUTH" },
	{ 0x3A04 , "HANGUL SYLLABLE IEUNG EU PHIEUPH" },
	{ 0x3A05 , "HANGUL SYLLABLE IEUNG EU HIEUH" },
	{ 0x3A06 , "HANGUL SYLLABLE IEUNG YI" },
	{ 0x3A07 , "HANGUL SYLLABLE IEUNG YI NIEUN" },
	{ 0x3A08 , "HANGUL SYLLABLE IEUNG YI RIEUL" },
	{ 0x3A09 , "HANGUL SYLLABLE IEUNG YI MIEUM" },
	{ 0x3A0A , "HANGUL SYLLABLE IEUNG YI SIOS" },
	{ 0x3A0B , "HANGUL SYLLABLE IEUNG I" },
	{ 0x3A0C , "HANGUL SYLLABLE IEUNG I KIYEOK" },
	{ 0x3A0D , "HANGUL SYLLABLE IEUNG I NIEUN" },
	{ 0x3A0E , "HANGUL SYLLABLE IEUNG I RIEUL" },
	{ 0x3A0F , "HANGUL SYLLABLE IEUNG I RIEUL-KIYEOK" },
	{ 0x3A10 , "HANGUL SYLLABLE IEUNG I RIEUL-MIEUM" },
	{ 0x3A11 , "HANGUL SYLLABLE IEUNG I RIEUL-HIEUH" },
	{ 0x3A12 , "HANGUL SYLLABLE IEUNG I MIEUM" },
	{ 0x3A13 , "HANGUL SYLLABLE IEUNG I PIEUP" },
	{ 0x3A14 , "HANGUL SYLLABLE IEUNG I SIOS" },
	{ 0x3A15 , "HANGUL SYLLABLE IEUNG I SSANGSIOS" },
	{ 0x3A16 , "HANGUL SYLLABLE IEUNG I IEUNG" },
	{ 0x3A17 , "HANGUL SYLLABLE IEUNG I CIEUC" },
	{ 0x3A18 , "HANGUL SYLLABLE IEUNG I PHIEUPH" },
	{ 0x3A19 , "HANGUL SYLLABLE CIEUC A" },
	{ 0x3A1A , "HANGUL SYLLABLE CIEUC A KIYEOK" },
	{ 0x3A1B , "HANGUL SYLLABLE CIEUC A NIEUN" },
	{ 0x3A1C , "HANGUL SYLLABLE CIEUC A NIEUN-HIEUH" },
	{ 0x3A1D , "HANGUL SYLLABLE CIEUC A TIKEUT" },
	{ 0x3A1E , "HANGUL SYLLABLE CIEUC A RIEUL" },
	{ 0x3A1F , "HANGUL SYLLABLE CIEUC A RIEUL-MIEUM" },
	{ 0x3A20 , "HANGUL SYLLABLE CIEUC A MIEUM" },
	{ 0x3A21 , "HANGUL SYLLABLE CIEUC A PIEUP" },
	{ 0x3A22 , "HANGUL SYLLABLE CIEUC A SIOS" },
	{ 0x3A23 , "HANGUL SYLLABLE CIEUC A SSANGSIOS" },
	{ 0x3A24 , "HANGUL SYLLABLE CIEUC A IEUNG" },
	{ 0x3A25 , "HANGUL SYLLABLE CIEUC A CIEUC" },
	{ 0x3A26 , "HANGUL SYLLABLE CIEUC AE" },
	{ 0x3A27 , "HANGUL SYLLABLE CIEUC AE KIYEOK" },
	{ 0x3A28 , "HANGUL SYLLABLE CIEUC AE NIEUN" },
	{ 0x3A29 , "HANGUL SYLLABLE CIEUC AE RIEUL" },
	{ 0x3A2A , "HANGUL SYLLABLE CIEUC AE MIEUM" },
	{ 0x3A2B , "HANGUL SYLLABLE CIEUC AE PIEUP" },
	{ 0x3A2C , "HANGUL SYLLABLE CIEUC AE SIOS" },
	{ 0x3A2D , "HANGUL SYLLABLE CIEUC AE SSANGSIOS" },
	{ 0x3A2E , "HANGUL SYLLABLE CIEUC AE IEUNG" },
	{ 0x3A2F , "HANGUL SYLLABLE CIEUC YA" },
	{ 0x3A30 , "HANGUL SYLLABLE CIEUC YA KIYEOK" },
	{ 0x3A31 , "HANGUL SYLLABLE CIEUC YA NIEUN" },
	{ 0x3A32 , "HANGUL SYLLABLE CIEUC YA NIEUN-HIEUH" },
	{ 0x3A33 , "HANGUL SYLLABLE CIEUC YA RIEUL" },
	{ 0x3A34 , "HANGUL SYLLABLE CIEUC YA MIEUM" },
	{ 0x3A35 , "HANGUL SYLLABLE CIEUC YA IEUNG" },
	{ 0x3A36 , "HANGUL SYLLABLE CIEUC YAE" },
	{ 0x3A37 , "HANGUL SYLLABLE CIEUC YAE NIEUN" },
	{ 0x3A38 , "HANGUL SYLLABLE CIEUC YAE RIEUL" },
	{ 0x3A39 , "HANGUL SYLLABLE CIEUC EO" },
	{ 0x3A3A , "HANGUL SYLLABLE CIEUC EO KIYEOK" },
	{ 0x3A3B , "HANGUL SYLLABLE CIEUC EO NIEUN" },
	{ 0x3A3C , "HANGUL SYLLABLE CIEUC EO RIEUL" },
	{ 0x3A3D , "HANGUL SYLLABLE CIEUC EO RIEUL-MIEUM" },
	{ 0x3A3E , "HANGUL SYLLABLE CIEUC EO MIEUM" },
	{ 0x3A3F , "HANGUL SYLLABLE CIEUC EO PIEUP" },
	{ 0x3A40 , "HANGUL SYLLABLE CIEUC EO SIOS" },
	{ 0x3A41 , "HANGUL SYLLABLE CIEUC EO IEUNG" },
	{ 0x3A42 , "HANGUL SYLLABLE CIEUC EO CIEUC" },
	{ 0x3A43 , "HANGUL SYLLABLE CIEUC E" },
	{ 0x3A44 , "HANGUL SYLLABLE CIEUC E KIYEOK" },
	{ 0x3A45 , "HANGUL SYLLABLE CIEUC E NIEUN" },
	{ 0x3A46 , "HANGUL SYLLABLE CIEUC E RIEUL" },
	{ 0x3A47 , "HANGUL SYLLABLE CIEUC E MIEUM" },
	{ 0x3A48 , "HANGUL SYLLABLE CIEUC E PIEUP" },
	{ 0x3A49 , "HANGUL SYLLABLE CIEUC E SIOS" },
	{ 0x3A4A , "HANGUL SYLLABLE CIEUC E IEUNG" },
	{ 0x3A4B , "HANGUL SYLLABLE CIEUC YEO" },
	{ 0x3A4C , "HANGUL SYLLABLE CIEUC YEO NIEUN" },
	{ 0x3A4D , "HANGUL SYLLABLE CIEUC YEO RIEUL" },
	{ 0x3A4E , "HANGUL SYLLABLE CIEUC YEO MIEUM" },
	{ 0x3A4F , "HANGUL SYLLABLE CIEUC YEO PIEUP" },
	{ 0x3A50 , "HANGUL SYLLABLE CIEUC YEO SSANGSIOS" },
	{ 0x3A51 , "HANGUL SYLLABLE CIEUC YEO IEUNG" },
	{ 0x3A52 , "HANGUL SYLLABLE CIEUC YE" },
	{ 0x3A53 , "HANGUL SYLLABLE CIEUC O" },
	{ 0x3A54 , "HANGUL SYLLABLE CIEUC O KIYEOK" },
	{ 0x3A55 , "HANGUL SYLLABLE CIEUC O NIEUN" },
	{ 0x3A56 , "HANGUL SYLLABLE CIEUC O RIEUL" },
	{ 0x3A57 , "HANGUL SYLLABLE CIEUC O RIEUL-MIEUM" },
	{ 0x3A58 , "HANGUL SYLLABLE CIEUC O MIEUM" },
	{ 0x3A59 , "HANGUL SYLLABLE CIEUC O PIEUP" },
	{ 0x3A5A , "HANGUL SYLLABLE CIEUC O SIOS" },
	{ 0x3A5B , "HANGUL SYLLABLE CIEUC O IEUNG" },
	{ 0x3A5C , "HANGUL SYLLABLE CIEUC O CIEUC" },
	{ 0x3A5D , "HANGUL SYLLABLE CIEUC O CHIEUCH" },
	{ 0x3A5E , "HANGUL SYLLABLE CIEUC O HIEUH" },
	{ 0x3A5F , "HANGUL SYLLABLE CIEUC WA" },
	{ 0x3A60 , "HANGUL SYLLABLE CIEUC WA KIYEOK" },
	{ 0x3A61 , "HANGUL SYLLABLE CIEUC WA RIEUL" },
	{ 0x3A62 , "HANGUL SYLLABLE CIEUC WA PIEUP" },
	{ 0x3A63 , "HANGUL SYLLABLE CIEUC WA SIOS" },
	{ 0x3A64 , "HANGUL SYLLABLE CIEUC WA IEUNG" },
	{ 0x3A65 , "HANGUL SYLLABLE CIEUC WAE" },
	{ 0x3A66 , "HANGUL SYLLABLE CIEUC WAE SSANGSIOS" },
	{ 0x3A67 , "HANGUL SYLLABLE CIEUC WAE IEUNG" },
	{ 0x3A68 , "HANGUL SYLLABLE CIEUC OE" },
	{ 0x3A69 , "HANGUL SYLLABLE CIEUC OE NIEUN" },
	{ 0x3A6A , "HANGUL SYLLABLE CIEUC OE RIEUL" },
	{ 0x3A6B , "HANGUL SYLLABLE CIEUC OE MIEUM" },
	{ 0x3A6C , "HANGUL SYLLABLE CIEUC OE PIEUP" },
	{ 0x3A6D , "HANGUL SYLLABLE CIEUC OE SIOS" },
	{ 0x3A6E , "HANGUL SYLLABLE CIEUC OE IEUNG" },
	{ 0x3A6F , "HANGUL SYLLABLE CIEUC YO" },
	{ 0x3A70 , "HANGUL SYLLABLE CIEUC YO KIYEOK" },
	{ 0x3A71 , "HANGUL SYLLABLE CIEUC YO NIEUN" },
	{ 0x3A72 , "HANGUL SYLLABLE CIEUC YO IEUNG" },
	{ 0x3A73 , "HANGUL SYLLABLE CIEUC U" },
	{ 0x3A74 , "HANGUL SYLLABLE CIEUC U KIYEOK" },
	{ 0x3A75 , "HANGUL SYLLABLE CIEUC U NIEUN" },
	{ 0x3A76 , "HANGUL SYLLABLE CIEUC U RIEUL" },
	{ 0x3A77 , "HANGUL SYLLABLE CIEUC U RIEUL-KIYEOK" },
	{ 0x3A78 , "HANGUL SYLLABLE CIEUC U RIEUL-MIEUM" },
	{ 0x3A79 , "HANGUL SYLLABLE CIEUC U MIEUM" },
	{ 0x3A7A , "HANGUL SYLLABLE CIEUC U PIEUP" },
	{ 0x3A7B , "HANGUL SYLLABLE CIEUC U SIOS" },
	{ 0x3A7C , "HANGUL SYLLABLE CIEUC U IEUNG" },
	{ 0x3A7D , "HANGUL SYLLABLE CIEUC WEO" },
	{ 0x3A7E , "HANGUL SYLLABLE CIEUC WEO SSANGSIOS" },
	{ 0x3A7F , "HANGUL SYLLABLE CIEUC WE" },
	{ 0x3A80 , "HANGUL SYLLABLE CIEUC WI" },
	{ 0x3A81 , "HANGUL SYLLABLE CIEUC WI KIYEOK" },
	{ 0x3A82 , "HANGUL SYLLABLE CIEUC WI NIEUN" },
	{ 0x3A83 , "HANGUL SYLLABLE CIEUC WI RIEUL" },
	{ 0x3A84 , "HANGUL SYLLABLE CIEUC WI MIEUM" },
	{ 0x3A85 , "HANGUL SYLLABLE CIEUC WI PIEUP" },
	{ 0x3A86 , "HANGUL SYLLABLE CIEUC WI SIOS" },
	{ 0x3A87 , "HANGUL SYLLABLE CIEUC YU" },
	{ 0x3A88 , "HANGUL SYLLABLE CIEUC YU NIEUN" },
	{ 0x3A89 , "HANGUL SYLLABLE CIEUC YU RIEUL" },
	{ 0x3A8A , "HANGUL SYLLABLE CIEUC YU MIEUM" },
	{ 0x3A8B , "HANGUL SYLLABLE CIEUC EU" },
	{ 0x3A8C , "HANGUL SYLLABLE CIEUC EU KIYEOK" },
	{ 0x3A8D , "HANGUL SYLLABLE CIEUC EU NIEUN" },
	{ 0x3A8E , "HANGUL SYLLABLE CIEUC EU RIEUL" },
	{ 0x3A8F , "HANGUL SYLLABLE CIEUC EU MIEUM" },
	{ 0x3A90 , "HANGUL SYLLABLE CIEUC EU PIEUP" },
	{ 0x3A91 , "HANGUL SYLLABLE CIEUC EU SIOS" },
	{ 0x3A92 , "HANGUL SYLLABLE CIEUC EU IEUNG" },
	{ 0x3A93 , "HANGUL SYLLABLE CIEUC I" },
	{ 0x3A94 , "HANGUL SYLLABLE CIEUC I KIYEOK" },
	{ 0x3A95 , "HANGUL SYLLABLE CIEUC I NIEUN" },
	{ 0x3A96 , "HANGUL SYLLABLE CIEUC I TIKEUT" },
	{ 0x3A97 , "HANGUL SYLLABLE CIEUC I RIEUL" },
	{ 0x3A98 , "HANGUL SYLLABLE CIEUC I RIEUL-MIEUM" },
	{ 0x3A99 , "HANGUL SYLLABLE CIEUC I MIEUM" },
	{ 0x3A9A , "HANGUL SYLLABLE CIEUC I PIEUP" },
	{ 0x3A9B , "HANGUL SYLLABLE CIEUC I SIOS" },
	{ 0x3A9C , "HANGUL SYLLABLE CIEUC I IEUNG" },
	{ 0x3A9D , "HANGUL SYLLABLE CIEUC I CIEUC" },
	{ 0x3A9E , "HANGUL SYLLABLE CIEUC I THIEUTH" },
	{ 0x3A9F , "HANGUL SYLLABLE CIEUC I PHIEUPH" },
	{ 0x3AA0 , "HANGUL SYLLABLE SSANGCIEUC A" },
	{ 0x3AA1 , "HANGUL SYLLABLE SSANGCIEUC A KIYEOK" },
	{ 0x3AA2 , "HANGUL SYLLABLE SSANGCIEUC A NIEUN" },
	{ 0x3AA3 , "HANGUL SYLLABLE SSANGCIEUC A NIEUN-HIEUH" },
	{ 0x3AA4 , "HANGUL SYLLABLE SSANGCIEUC A RIEUL" },
	{ 0x3AA5 , "HANGUL SYLLABLE SSANGCIEUC A RIEUL-PIEUP" },
	{ 0x3AA6 , "HANGUL SYLLABLE SSANGCIEUC A MIEUM" },
	{ 0x3AA7 , "HANGUL SYLLABLE SSANGCIEUC A PIEUP" },
	{ 0x3AA8 , "HANGUL SYLLABLE SSANGCIEUC A SIOS" },
	{ 0x3AA9 , "HANGUL SYLLABLE SSANGCIEUC A SSANGSIOS" },
	{ 0x3AAA , "HANGUL SYLLABLE SSANGCIEUC A IEUNG" },
	{ 0x3AAB , "HANGUL SYLLABLE SSANGCIEUC AE" },
	{ 0x3AAC , "HANGUL SYLLABLE SSANGCIEUC AE KIYEOK" },
	{ 0x3AAD , "HANGUL SYLLABLE SSANGCIEUC AE NIEUN" },
	{ 0x3AAE , "HANGUL SYLLABLE SSANGCIEUC AE RIEUL" },
	{ 0x3AAF , "HANGUL SYLLABLE SSANGCIEUC AE MIEUM" },
	{ 0x3AB0 , "HANGUL SYLLABLE SSANGCIEUC AE PIEUP" },
	{ 0x3AB1 , "HANGUL SYLLABLE SSANGCIEUC AE SIOS" },
	{ 0x3AB2 , "HANGUL SYLLABLE SSANGCIEUC AE SSANGSIOS" },
	{ 0x3AB3 , "HANGUL SYLLABLE SSANGCIEUC AE IEUNG" },
	{ 0x3AB4 , "HANGUL SYLLABLE SSANGCIEUC YA" },
	{ 0x3AB5 , "HANGUL SYLLABLE SSANGCIEUC YA NIEUN" },
	{ 0x3AB6 , "HANGUL SYLLABLE SSANGCIEUC YA IEUNG" },
	{ 0x3AB7 , "HANGUL SYLLABLE SSANGCIEUC EO" },
	{ 0x3AB8 , "HANGUL SYLLABLE SSANGCIEUC EO KIYEOK" },
	{ 0x3AB9 , "HANGUL SYLLABLE SSANGCIEUC EO NIEUN" },
	{ 0x3ABA , "HANGUL SYLLABLE SSANGCIEUC EO RIEUL" },
	{ 0x3ABB , "HANGUL SYLLABLE SSANGCIEUC EO MIEUM" },
	{ 0x3ABC , "HANGUL SYLLABLE SSANGCIEUC EO PIEUP" },
	{ 0x3ABD , "HANGUL SYLLABLE SSANGCIEUC EO SIOS" },
	{ 0x3ABE , "HANGUL SYLLABLE SSANGCIEUC EO SSANGSIOS" },
	{ 0x3ABF , "HANGUL SYLLABLE SSANGCIEUC EO IEUNG" },
	{ 0x3AC0 , "HANGUL SYLLABLE SSANGCIEUC E" },
	{ 0x3AC1 , "HANGUL SYLLABLE SSANGCIEUC E IEUNG" },
	{ 0x3AC2 , "HANGUL SYLLABLE SSANGCIEUC YEO" },
	{ 0x3AC3 , "HANGUL SYLLABLE SSANGCIEUC YEO SSANGSIOS" },
	{ 0x3AC4 , "HANGUL SYLLABLE SSANGCIEUC O" },
	{ 0x3AC5 , "HANGUL SYLLABLE SSANGCIEUC O KIYEOK" },
	{ 0x3AC6 , "HANGUL SYLLABLE SSANGCIEUC O NIEUN" },
	{ 0x3AC7 , "HANGUL SYLLABLE SSANGCIEUC O RIEUL" },
	{ 0x3AC8 , "HANGUL SYLLABLE SSANGCIEUC O MIEUM" },
	{ 0x3AC9 , "HANGUL SYLLABLE SSANGCIEUC O PIEUP" },
	{ 0x3ACA , "HANGUL SYLLABLE SSANGCIEUC O SIOS" },
	{ 0x3ACB , "HANGUL SYLLABLE SSANGCIEUC O IEUNG" },
	{ 0x3ACC , "HANGUL SYLLABLE SSANGCIEUC O CHIEUCH" },
	{ 0x3ACD , "HANGUL SYLLABLE SSANGCIEUC WA" },
	{ 0x3ACE , "HANGUL SYLLABLE SSANGCIEUC WA KIYEOK" },
	{ 0x3ACF , "HANGUL SYLLABLE SSANGCIEUC WA RIEUL" },
	{ 0x3AD0 , "HANGUL SYLLABLE SSANGCIEUC WA SSANGSIOS" },
	{ 0x3AD1 , "HANGUL SYLLABLE SSANGCIEUC WAE" },
	{ 0x3AD2 , "HANGUL SYLLABLE SSANGCIEUC WAE SSANGSIOS" },
	{ 0x3AD3 , "HANGUL SYLLABLE SSANGCIEUC OE" },
	{ 0x3AD4 , "HANGUL SYLLABLE SSANGCIEUC OE NIEUN" },
	{ 0x3AD5 , "HANGUL SYLLABLE SSANGCIEUC OE RIEUL" },
	{ 0x3AD6 , "HANGUL SYLLABLE SSANGCIEUC OE MIEUM" },
	{ 0x3AD7 , "HANGUL SYLLABLE SSANGCIEUC OE PIEUP" },
	{ 0x3AD8 , "HANGUL SYLLABLE SSANGCIEUC YO IEUNG" },
	{ 0x3AD9 , "HANGUL SYLLABLE SSANGCIEUC U" },
	{ 0x3ADA , "HANGUL SYLLABLE SSANGCIEUC U KIYEOK" },
	{ 0x3ADB , "HANGUL SYLLABLE SSANGCIEUC U NIEUN" },
	{ 0x3ADC , "HANGUL SYLLABLE SSANGCIEUC U RIEUL" },
	{ 0x3ADD , "HANGUL SYLLABLE SSANGCIEUC U MIEUM" },
	{ 0x3ADE , "HANGUL SYLLABLE SSANGCIEUC U PIEUP" },
	{ 0x3ADF , "HANGUL SYLLABLE SSANGCIEUC U IEUNG" },
	{ 0x3AE0 , "HANGUL SYLLABLE SSANGCIEUC WEO" },
	{ 0x3AE1 , "HANGUL SYLLABLE SSANGCIEUC WEO SSANGSIOS" },
	{ 0x3AE2 , "HANGUL SYLLABLE SSANGCIEUC WEO IEUNG" },
	{ 0x3AE3 , "HANGUL SYLLABLE SSANGCIEUC WI" },
	{ 0x3AE4 , "HANGUL SYLLABLE SSANGCIEUC YU" },
	{ 0x3AE5 , "HANGUL SYLLABLE SSANGCIEUC EU" },
	{ 0x3AE6 , "HANGUL SYLLABLE SSANGCIEUC EU MIEUM" },
	{ 0x3AE7 , "HANGUL SYLLABLE SSANGCIEUC EU SIOS" },
	{ 0x3AE8 , "HANGUL SYLLABLE SSANGCIEUC EU IEUNG" },
	{ 0x3AE9 , "HANGUL SYLLABLE SSANGCIEUC I" },
	{ 0x3AEA , "HANGUL SYLLABLE SSANGCIEUC I KIYEOK" },
	{ 0x3AEB , "HANGUL SYLLABLE SSANGCIEUC I NIEUN" },
	{ 0x3AEC , "HANGUL SYLLABLE SSANGCIEUC I RIEUL" },
	{ 0x3AED , "HANGUL SYLLABLE SSANGCIEUC I MIEUM" },
	{ 0x3AEE , "HANGUL SYLLABLE SSANGCIEUC I PIEUP" },
	{ 0x3AEF , "HANGUL SYLLABLE SSANGCIEUC I IEUNG" },
	{ 0x3AF0 , "HANGUL SYLLABLE SSANGCIEUC I CIEUC" },
	{ 0x3AF1 , "HANGUL SYLLABLE SSANGCIEUC I HIEUH" },
	{ 0x3AF2 , "HANGUL SYLLABLE CHIEUCH A" },
	{ 0x3AF3 , "HANGUL SYLLABLE CHIEUCH A KIYEOK" },
	{ 0x3AF4 , "HANGUL SYLLABLE CHIEUCH A NIEUN" },
	{ 0x3AF5 , "HANGUL SYLLABLE CHIEUCH A NIEUN-HIEUH" },
	{ 0x3AF6 , "HANGUL SYLLABLE CHIEUCH A RIEUL" },
	{ 0x3AF7 , "HANGUL SYLLABLE CHIEUCH A MIEUM" },
	{ 0x3AF8 , "HANGUL SYLLABLE CHIEUCH A PIEUP" },
	{ 0x3AF9 , "HANGUL SYLLABLE CHIEUCH A SIOS" },
	{ 0x3AFA , "HANGUL SYLLABLE CHIEUCH A SSANGSIOS" },
	{ 0x3AFB , "HANGUL SYLLABLE CHIEUCH A IEUNG" },
	{ 0x3AFC , "HANGUL SYLLABLE CHIEUCH A CIEUC" },
	{ 0x3AFD , "HANGUL SYLLABLE CHIEUCH AE" },
	{ 0x3AFE , "HANGUL SYLLABLE CHIEUCH AE KIYEOK" },
	{ 0x3AFF , "HANGUL SYLLABLE CHIEUCH AE NIEUN" },
	{ 0x3B00 , "HANGUL SYLLABLE CHIEUCH AE RIEUL" },
	{ 0x3B01 , "HANGUL SYLLABLE CHIEUCH AE MIEUM" },
	{ 0x3B02 , "HANGUL SYLLABLE CHIEUCH AE PIEUP" },
	{ 0x3B03 , "HANGUL SYLLABLE CHIEUCH AE SIOS" },
	{ 0x3B04 , "HANGUL SYLLABLE CHIEUCH AE SSANGSIOS" },
	{ 0x3B05 , "HANGUL SYLLABLE CHIEUCH AE IEUNG" },
	{ 0x3B06 , "HANGUL SYLLABLE CHIEUCH YA" },
	{ 0x3B07 , "HANGUL SYLLABLE CHIEUCH YA NIEUN" },
	{ 0x3B08 , "HANGUL SYLLABLE CHIEUCH YA NIEUN-HIEUH" },
	{ 0x3B09 , "HANGUL SYLLABLE CHIEUCH YA RIEUL" },
	{ 0x3B0A , "HANGUL SYLLABLE CHIEUCH YA MIEUM" },
	{ 0x3B0B , "HANGUL SYLLABLE CHIEUCH YA IEUNG" },
	{ 0x3B0C , "HANGUL SYLLABLE CHIEUCH EO" },
	{ 0x3B0D , "HANGUL SYLLABLE CHIEUCH EO KIYEOK" },
	{ 0x3B0E , "HANGUL SYLLABLE CHIEUCH EO NIEUN" },
	{ 0x3B0F , "HANGUL SYLLABLE CHIEUCH EO RIEUL" },
	{ 0x3B10 , "HANGUL SYLLABLE CHIEUCH EO MIEUM" },
	{ 0x3B11 , "HANGUL SYLLABLE CHIEUCH EO PIEUP" },
	{ 0x3B12 , "HANGUL SYLLABLE CHIEUCH EO SIOS" },
	{ 0x3B13 , "HANGUL SYLLABLE CHIEUCH EO SSANGSIOS" },
	{ 0x3B14 , "HANGUL SYLLABLE CHIEUCH EO IEUNG" },
	{ 0x3B15 , "HANGUL SYLLABLE CHIEUCH E" },
	{ 0x3B16 , "HANGUL SYLLABLE CHIEUCH E KIYEOK" },
	{ 0x3B17 , "HANGUL SYLLABLE CHIEUCH E NIEUN" },
	{ 0x3B18 , "HANGUL SYLLABLE CHIEUCH E RIEUL" },
	{ 0x3B19 , "HANGUL SYLLABLE CHIEUCH E MIEUM" },
	{ 0x3B1A , "HANGUL SYLLABLE CHIEUCH E PIEUP" },
	{ 0x3B1B , "HANGUL SYLLABLE CHIEUCH E SIOS" },
	{ 0x3B1C , "HANGUL SYLLABLE CHIEUCH E IEUNG" },
	{ 0x3B1D , "HANGUL SYLLABLE CHIEUCH YEO" },
	{ 0x3B1E , "HANGUL SYLLABLE CHIEUCH YEO NIEUN" },
	{ 0x3B1F , "HANGUL SYLLABLE CHIEUCH YEO SSANGSIOS" },
	{ 0x3B20 , "HANGUL SYLLABLE CHIEUCH YE" },
	{ 0x3B21 , "HANGUL SYLLABLE CHIEUCH YE NIEUN" },
	{ 0x3B22 , "HANGUL SYLLABLE CHIEUCH YE IEUNG" },
	{ 0x3B23 , "HANGUL SYLLABLE CHIEUCH O" },
	{ 0x3B24 , "HANGUL SYLLABLE CHIEUCH O KIYEOK" },
	{ 0x3B25 , "HANGUL SYLLABLE CHIEUCH O NIEUN" },
	{ 0x3B26 , "HANGUL SYLLABLE CHIEUCH O RIEUL" },
	{ 0x3B27 , "HANGUL SYLLABLE CHIEUCH O MIEUM" },
	{ 0x3B28 , "HANGUL SYLLABLE CHIEUCH O PIEUP" },
	{ 0x3B29 , "HANGUL SYLLABLE CHIEUCH O SIOS" },
	{ 0x3B2A , "HANGUL SYLLABLE CHIEUCH O IEUNG" },
	{ 0x3B2B , "HANGUL SYLLABLE CHIEUCH WA" },
	{ 0x3B2C , "HANGUL SYLLABLE CHIEUCH WA NIEUN" },
	{ 0x3B2D , "HANGUL SYLLABLE CHIEUCH WA RIEUL" },
	{ 0x3B2E , "HANGUL SYLLABLE CHIEUCH WA IEUNG" },
	{ 0x3B2F , "HANGUL SYLLABLE CHIEUCH OE" },
	{ 0x3B30 , "HANGUL SYLLABLE CHIEUCH OE NIEUN" },
	{ 0x3B31 , "HANGUL SYLLABLE CHIEUCH OE RIEUL" },
	{ 0x3B32 , "HANGUL SYLLABLE CHIEUCH OE MIEUM" },
	{ 0x3B33 , "HANGUL SYLLABLE CHIEUCH OE PIEUP" },
	{ 0x3B34 , "HANGUL SYLLABLE CHIEUCH OE SIOS" },
	{ 0x3B35 , "HANGUL SYLLABLE CHIEUCH OE IEUNG" },
	{ 0x3B36 , "HANGUL SYLLABLE CHIEUCH YO" },
	{ 0x3B37 , "HANGUL SYLLABLE CHIEUCH YO MIEUM" },
	{ 0x3B38 , "HANGUL SYLLABLE CHIEUCH U" },
	{ 0x3B39 , "HANGUL SYLLABLE CHIEUCH U KIYEOK" },
	{ 0x3B3A , "HANGUL SYLLABLE CHIEUCH U NIEUN" },
	{ 0x3B3B , "HANGUL SYLLABLE CHIEUCH U RIEUL" },
	{ 0x3B3C , "HANGUL SYLLABLE CHIEUCH U MIEUM" },
	{ 0x3B3D , "HANGUL SYLLABLE CHIEUCH U PIEUP" },
	{ 0x3B3E , "HANGUL SYLLABLE CHIEUCH U SIOS" },
	{ 0x3B3F , "HANGUL SYLLABLE CHIEUCH U IEUNG" },
	{ 0x3B40 , "HANGUL SYLLABLE CHIEUCH WEO" },
	{ 0x3B41 , "HANGUL SYLLABLE CHIEUCH WEO SSANGSIOS" },
	{ 0x3B42 , "HANGUL SYLLABLE CHIEUCH WE" },
	{ 0x3B43 , "HANGUL SYLLABLE CHIEUCH WE NIEUN" },
	{ 0x3B44 , "HANGUL SYLLABLE CHIEUCH WI" },
	{ 0x3B45 , "HANGUL SYLLABLE CHIEUCH WI NIEUN" },
	{ 0x3B46 , "HANGUL SYLLABLE CHIEUCH WI RIEUL" },
	{ 0x3B47 , "HANGUL SYLLABLE CHIEUCH WI MIEUM" },
	{ 0x3B48 , "HANGUL SYLLABLE CHIEUCH WI PIEUP" },
	{ 0x3B49 , "HANGUL SYLLABLE CHIEUCH WI SIOS" },
	{ 0x3B4A , "HANGUL SYLLABLE CHIEUCH WI IEUNG" },
	{ 0x3B4B , "HANGUL SYLLABLE CHIEUCH YU" },
	{ 0x3B4C , "HANGUL SYLLABLE CHIEUCH YU NIEUN" },
	{ 0x3B4D , "HANGUL SYLLABLE CHIEUCH YU RIEUL" },
	{ 0x3B4E , "HANGUL SYLLABLE CHIEUCH YU MIEUM" },
	{ 0x3B4F , "HANGUL SYLLABLE CHIEUCH YU IEUNG" },
	{ 0x3B50 , "HANGUL SYLLABLE CHIEUCH EU" },
	{ 0x3B51 , "HANGUL SYLLABLE CHIEUCH EU KIYEOK" },
	{ 0x3B52 , "HANGUL SYLLABLE CHIEUCH EU NIEUN" },
	{ 0x3B53 , "HANGUL SYLLABLE CHIEUCH EU RIEUL" },
	{ 0x3B54 , "HANGUL SYLLABLE CHIEUCH EU MIEUM" },
	{ 0x3B55 , "HANGUL SYLLABLE CHIEUCH EU PIEUP" },
	{ 0x3B56 , "HANGUL SYLLABLE CHIEUCH EU SIOS" },
	{ 0x3B57 , "HANGUL SYLLABLE CHIEUCH EU IEUNG" },
	{ 0x3B58 , "HANGUL SYLLABLE CHIEUCH I" },
	{ 0x3B59 , "HANGUL SYLLABLE CHIEUCH I KIYEOK" },
	{ 0x3B5A , "HANGUL SYLLABLE CHIEUCH I NIEUN" },
	{ 0x3B5B , "HANGUL SYLLABLE CHIEUCH I TIKEUT" },
	{ 0x3B5C , "HANGUL SYLLABLE CHIEUCH I RIEUL" },
	{ 0x3B5D , "HANGUL SYLLABLE CHIEUCH I RIEUL-KIYEOK" },
	{ 0x3B5E , "HANGUL SYLLABLE CHIEUCH I MIEUM" },
	{ 0x3B5F , "HANGUL SYLLABLE CHIEUCH I PIEUP" },
	{ 0x3B60 , "HANGUL SYLLABLE CHIEUCH I SIOS" },
	{ 0x3B61 , "HANGUL SYLLABLE CHIEUCH I IEUNG" },
	{ 0x3B62 , "HANGUL SYLLABLE KHIEUKH A" },
	{ 0x3B63 , "HANGUL SYLLABLE KHIEUKH A KIYEOK" },
	{ 0x3B64 , "HANGUL SYLLABLE KHIEUKH A NIEUN" },
	{ 0x3B65 , "HANGUL SYLLABLE KHIEUKH A RIEUL" },
	{ 0x3B66 , "HANGUL SYLLABLE KHIEUKH A MIEUM" },
	{ 0x3B67 , "HANGUL SYLLABLE KHIEUKH A PIEUP" },
	{ 0x3B68 , "HANGUL SYLLABLE KHIEUKH A SIOS" },
	{ 0x3B69 , "HANGUL SYLLABLE KHIEUKH A IEUNG" },
	{ 0x3B6A , "HANGUL SYLLABLE KHIEUKH AE" },
	{ 0x3B6B , "HANGUL SYLLABLE KHIEUKH AE KIYEOK" },
	{ 0x3B6C , "HANGUL SYLLABLE KHIEUKH AE NIEUN" },
	{ 0x3B6D , "HANGUL SYLLABLE KHIEUKH AE RIEUL" },
	{ 0x3B6E , "HANGUL SYLLABLE KHIEUKH AE MIEUM" },
	{ 0x3B6F , "HANGUL SYLLABLE KHIEUKH AE PIEUP" },
	{ 0x3B70 , "HANGUL SYLLABLE KHIEUKH AE SIOS" },
	{ 0x3B71 , "HANGUL SYLLABLE KHIEUKH AE SSANGSIOS" },
	{ 0x3B72 , "HANGUL SYLLABLE KHIEUKH AE IEUNG" },
	{ 0x3B73 , "HANGUL SYLLABLE KHIEUKH YA" },
	{ 0x3B74 , "HANGUL SYLLABLE KHIEUKH YA KIYEOK" },
	{ 0x3B75 , "HANGUL SYLLABLE KHIEUKH YA IEUNG" },
	{ 0x3B76 , "HANGUL SYLLABLE KHIEUKH EO" },
	{ 0x3B77 , "HANGUL SYLLABLE KHIEUKH EO KIYEOK" },
	{ 0x3B78 , "HANGUL SYLLABLE KHIEUKH EO NIEUN" },
	{ 0x3B79 , "HANGUL SYLLABLE KHIEUKH EO TIKEUT" },
	{ 0x3B7A , "HANGUL SYLLABLE KHIEUKH EO RIEUL" },
	{ 0x3B7B , "HANGUL SYLLABLE KHIEUKH EO MIEUM" },
	{ 0x3B7C , "HANGUL SYLLABLE KHIEUKH EO PIEUP" },
	{ 0x3B7D , "HANGUL SYLLABLE KHIEUKH EO SIOS" },
	{ 0x3B7E , "HANGUL SYLLABLE KHIEUKH EO SSANGSIOS" },
	{ 0x3B7F , "HANGUL SYLLABLE KHIEUKH EO IEUNG" },
	{ 0x3B80 , "HANGUL SYLLABLE KHIEUKH E" },
	{ 0x3B81 , "HANGUL SYLLABLE KHIEUKH E KIYEOK" },
	{ 0x3B82 , "HANGUL SYLLABLE KHIEUKH E NIEUN" },
	{ 0x3B83 , "HANGUL SYLLABLE KHIEUKH E RIEUL" },
	{ 0x3B84 , "HANGUL SYLLABLE KHIEUKH E MIEUM" },
	{ 0x3B85 , "HANGUL SYLLABLE KHIEUKH E PIEUP" },
	{ 0x3B86 , "HANGUL SYLLABLE KHIEUKH E SIOS" },
	{ 0x3B87 , "HANGUL SYLLABLE KHIEUKH E IEUNG" },
	{ 0x3B88 , "HANGUL SYLLABLE KHIEUKH YEO" },
	{ 0x3B89 , "HANGUL SYLLABLE KHIEUKH YEO NIEUN" },
	{ 0x3B8A , "HANGUL SYLLABLE KHIEUKH YEO RIEUL" },
	{ 0x3B8B , "HANGUL SYLLABLE KHIEUKH YEO MIEUM" },
	{ 0x3B8C , "HANGUL SYLLABLE KHIEUKH YEO PIEUP" },
	{ 0x3B8D , "HANGUL SYLLABLE KHIEUKH YEO SIOS" },
	{ 0x3B8E , "HANGUL SYLLABLE KHIEUKH YEO SSANGSIOS" },
	{ 0x3B8F , "HANGUL SYLLABLE KHIEUKH YEO IEUNG" },
	{ 0x3B90 , "HANGUL SYLLABLE KHIEUKH YE" },
	{ 0x3B91 , "HANGUL SYLLABLE KHIEUKH O" },
	{ 0x3B92 , "HANGUL SYLLABLE KHIEUKH O KIYEOK" },
	{ 0x3B93 , "HANGUL SYLLABLE KHIEUKH O NIEUN" },
	{ 0x3B94 , "HANGUL SYLLABLE KHIEUKH O RIEUL" },
	{ 0x3B95 , "HANGUL SYLLABLE KHIEUKH O MIEUM" },
	{ 0x3B96 , "HANGUL SYLLABLE KHIEUKH O PIEUP" },
	{ 0x3B97 , "HANGUL SYLLABLE KHIEUKH O SIOS" },
	{ 0x3B98 , "HANGUL SYLLABLE KHIEUKH O IEUNG" },
	{ 0x3B99 , "HANGUL SYLLABLE KHIEUKH WA" },
	{ 0x3B9A , "HANGUL SYLLABLE KHIEUKH WA KIYEOK" },
	{ 0x3B9B , "HANGUL SYLLABLE KHIEUKH WA NIEUN" },
	{ 0x3B9C , "HANGUL SYLLABLE KHIEUKH WA RIEUL" },
	{ 0x3B9D , "HANGUL SYLLABLE KHIEUKH WA MIEUM" },
	{ 0x3B9E , "HANGUL SYLLABLE KHIEUKH WA IEUNG" },
	{ 0x3B9F , "HANGUL SYLLABLE KHIEUKH WAE" },
	{ 0x3BA0 , "HANGUL SYLLABLE KHIEUKH WAE IEUNG" },
	{ 0x3BA1 , "HANGUL SYLLABLE KHIEUKH OE" },
	{ 0x3BA2 , "HANGUL SYLLABLE KHIEUKH OE RIEUL" },
	{ 0x3BA3 , "HANGUL SYLLABLE KHIEUKH YO" },
	{ 0x3BA4 , "HANGUL SYLLABLE KHIEUKH U" },
	{ 0x3BA5 , "HANGUL SYLLABLE KHIEUKH U KIYEOK" },
	{ 0x3BA6 , "HANGUL SYLLABLE KHIEUKH U NIEUN" },
	{ 0x3BA7 , "HANGUL SYLLABLE KHIEUKH U RIEUL" },
	{ 0x3BA8 , "HANGUL SYLLABLE KHIEUKH U MIEUM" },
	{ 0x3BA9 , "HANGUL SYLLABLE KHIEUKH U PIEUP" },
	{ 0x3BAA , "HANGUL SYLLABLE KHIEUKH U SIOS" },
	{ 0x3BAB , "HANGUL SYLLABLE KHIEUKH U IEUNG" },
	{ 0x3BAC , "HANGUL SYLLABLE KHIEUKH WEO" },
	{ 0x3BAD , "HANGUL SYLLABLE KHIEUKH WEO NIEUN" },
	{ 0x3BAE , "HANGUL SYLLABLE KHIEUKH WEO RIEUL" },
	{ 0x3BAF , "HANGUL SYLLABLE KHIEUKH WEO IEUNG" },
	{ 0x3BB0 , "HANGUL SYLLABLE KHIEUKH WE" },
	{ 0x3BB1 , "HANGUL SYLLABLE KHIEUKH WE IEUNG" },
	{ 0x3BB2 , "HANGUL SYLLABLE KHIEUKH WI" },
	{ 0x3BB3 , "HANGUL SYLLABLE KHIEUKH WI KIYEOK" },
	{ 0x3BB4 , "HANGUL SYLLABLE KHIEUKH WI NIEUN" },
	{ 0x3BB5 , "HANGUL SYLLABLE KHIEUKH WI RIEUL" },
	{ 0x3BB6 , "HANGUL SYLLABLE KHIEUKH WI MIEUM" },
	{ 0x3BB7 , "HANGUL SYLLABLE KHIEUKH WI PIEUP" },
	{ 0x3BB8 , "HANGUL SYLLABLE KHIEUKH WI SIOS" },
	{ 0x3BB9 , "HANGUL SYLLABLE KHIEUKH WI IEUNG" },
	{ 0x3BBA , "HANGUL SYLLABLE KHIEUKH YU" },
	{ 0x3BBB , "HANGUL SYLLABLE KHIEUKH YU NIEUN" },
	{ 0x3BBC , "HANGUL SYLLABLE KHIEUKH YU RIEUL" },
	{ 0x3BBD , "HANGUL SYLLABLE KHIEUKH YU MIEUM" },
	{ 0x3BBE , "HANGUL SYLLABLE KHIEUKH EU" },
	{ 0x3BBF , "HANGUL SYLLABLE KHIEUKH EU KIYEOK" },
	{ 0x3BC0 , "HANGUL SYLLABLE KHIEUKH EU NIEUN" },
	{ 0x3BC1 , "HANGUL SYLLABLE KHIEUKH EU RIEUL" },
	{ 0x3BC2 , "HANGUL SYLLABLE KHIEUKH EU MIEUM" },
	{ 0x3BC3 , "HANGUL SYLLABLE KHIEUKH EU PIEUP" },
	{ 0x3BC4 , "HANGUL SYLLABLE KHIEUKH EU IEUNG" },
	{ 0x3BC5 , "HANGUL SYLLABLE KHIEUKH I" },
	{ 0x3BC6 , "HANGUL SYLLABLE KHIEUKH I KIYEOK" },
	{ 0x3BC7 , "HANGUL SYLLABLE KHIEUKH I NIEUN" },
	{ 0x3BC8 , "HANGUL SYLLABLE KHIEUKH I RIEUL" },
	{ 0x3BC9 , "HANGUL SYLLABLE KHIEUKH I MIEUM" },
	{ 0x3BCA , "HANGUL SYLLABLE KHIEUKH I PIEUP" },
	{ 0x3BCB , "HANGUL SYLLABLE KHIEUKH I SIOS" },
	{ 0x3BCC , "HANGUL SYLLABLE KHIEUKH I IEUNG" },
	{ 0x3BCD , "HANGUL SYLLABLE THIEUTH A" },
	{ 0x3BCE , "HANGUL SYLLABLE THIEUTH A KIYEOK" },
	{ 0x3BCF , "HANGUL SYLLABLE THIEUTH A NIEUN" },
	{ 0x3BD0 , "HANGUL SYLLABLE THIEUTH A RIEUL" },
	{ 0x3BD1 , "HANGUL SYLLABLE THIEUTH A RIEUL-KIYEOK" },
	{ 0x3BD2 , "HANGUL SYLLABLE THIEUTH A MIEUM" },
	{ 0x3BD3 , "HANGUL SYLLABLE THIEUTH A PIEUP" },
	{ 0x3BD4 , "HANGUL SYLLABLE THIEUTH A SIOS" },
	{ 0x3BD5 , "HANGUL SYLLABLE THIEUTH A SSANGSIOS" },
	{ 0x3BD6 , "HANGUL SYLLABLE THIEUTH A IEUNG" },
	{ 0x3BD7 , "HANGUL SYLLABLE THIEUTH AE" },
	{ 0x3BD8 , "HANGUL SYLLABLE THIEUTH AE KIYEOK" },
	{ 0x3BD9 , "HANGUL SYLLABLE THIEUTH AE NIEUN" },
	{ 0x3BDA , "HANGUL SYLLABLE THIEUTH AE RIEUL" },
	{ 0x3BDB , "HANGUL SYLLABLE THIEUTH AE MIEUM" },
	{ 0x3BDC , "HANGUL SYLLABLE THIEUTH AE PIEUP" },
	{ 0x3BDD , "HANGUL SYLLABLE THIEUTH AE SIOS" },
	{ 0x3BDE , "HANGUL SYLLABLE THIEUTH AE SSANGSIOS" },
	{ 0x3BDF , "HANGUL SYLLABLE THIEUTH AE IEUNG" },
	{ 0x3BE0 , "HANGUL SYLLABLE THIEUTH YA" },
	{ 0x3BE1 , "HANGUL SYLLABLE THIEUTH YA IEUNG" },
	{ 0x3BE2 , "HANGUL SYLLABLE THIEUTH EO" },
	{ 0x3BE3 , "HANGUL SYLLABLE THIEUTH EO KIYEOK" },
	{ 0x3BE4 , "HANGUL SYLLABLE THIEUTH EO NIEUN" },
	{ 0x3BE5 , "HANGUL SYLLABLE THIEUTH EO RIEUL" },
	{ 0x3BE6 , "HANGUL SYLLABLE THIEUTH EO RIEUL-MIEUM" },
	{ 0x3BE7 , "HANGUL SYLLABLE THIEUTH EO MIEUM" },
	{ 0x3BE8 , "HANGUL SYLLABLE THIEUTH EO PIEUP" },
	{ 0x3BE9 , "HANGUL SYLLABLE THIEUTH EO SIOS" },
	{ 0x3BEA , "HANGUL SYLLABLE THIEUTH EO SSANGSIOS" },
	{ 0x3BEB , "HANGUL SYLLABLE THIEUTH EO IEUNG" },
	{ 0x3BEC , "HANGUL SYLLABLE THIEUTH E" },
	{ 0x3BED , "HANGUL SYLLABLE THIEUTH E KIYEOK" },
	{ 0x3BEE , "HANGUL SYLLABLE THIEUTH E NIEUN" },
	{ 0x3BEF , "HANGUL SYLLABLE THIEUTH E RIEUL" },
	{ 0x3BF0 , "HANGUL SYLLABLE THIEUTH E MIEUM" },
	{ 0x3BF1 , "HANGUL SYLLABLE THIEUTH E PIEUP" },
	{ 0x3BF2 , "HANGUL SYLLABLE THIEUTH E SIOS" },
	{ 0x3BF3 , "HANGUL SYLLABLE THIEUTH E IEUNG" },
	{ 0x3BF4 , "HANGUL SYLLABLE THIEUTH YEO" },
	{ 0x3BF5 , "HANGUL SYLLABLE THIEUTH YEO NIEUN" },
	{ 0x3BF6 , "HANGUL SYLLABLE THIEUTH YEO SSANGSIOS" },
	{ 0x3BF7 , "HANGUL SYLLABLE THIEUTH YE" },
	{ 0x3BF8 , "HANGUL SYLLABLE THIEUTH YE NIEUN" },
	{ 0x3BF9 , "HANGUL SYLLABLE THIEUTH O" },
	{ 0x3BFA , "HANGUL SYLLABLE THIEUTH O KIYEOK" },
	{ 0x3BFB , "HANGUL SYLLABLE THIEUTH O NIEUN" },
	{ 0x3BFC , "HANGUL SYLLABLE THIEUTH O RIEUL" },
	{ 0x3BFD , "HANGUL SYLLABLE THIEUTH O MIEUM" },
	{ 0x3BFE , "HANGUL SYLLABLE THIEUTH O PIEUP" },
	{ 0x3BFF , "HANGUL SYLLABLE THIEUTH O SIOS" },
	{ 0x3C00 , "HANGUL SYLLABLE THIEUTH O IEUNG" },
	{ 0x3C01 , "HANGUL SYLLABLE THIEUTH O PHIEUPH" },
	{ 0x3C02 , "HANGUL SYLLABLE THIEUTH WA" },
	{ 0x3C03 , "HANGUL SYLLABLE THIEUTH WA NIEUN" },
	{ 0x3C04 , "HANGUL SYLLABLE THIEUTH WAE" },
	{ 0x3C05 , "HANGUL SYLLABLE THIEUTH OE" },
	{ 0x3C06 , "HANGUL SYLLABLE THIEUTH OE NIEUN" },
	{ 0x3C07 , "HANGUL SYLLABLE THIEUTH OE SIOS" },
	{ 0x3C08 , "HANGUL SYLLABLE THIEUTH OE IEUNG" },
	{ 0x3C09 , "HANGUL SYLLABLE THIEUTH YO" },
	{ 0x3C0A , "HANGUL SYLLABLE THIEUTH U" },
	{ 0x3C0B , "HANGUL SYLLABLE THIEUTH U KIYEOK" },
	{ 0x3C0C , "HANGUL SYLLABLE THIEUTH U NIEUN" },
	{ 0x3C0D , "HANGUL SYLLABLE THIEUTH U RIEUL" },
	{ 0x3C0E , "HANGUL SYLLABLE THIEUTH U MIEUM" },
	{ 0x3C0F , "HANGUL SYLLABLE THIEUTH U PIEUP" },
	{ 0x3C10 , "HANGUL SYLLABLE THIEUTH U SIOS" },
	{ 0x3C11 , "HANGUL SYLLABLE THIEUTH U IEUNG" },
	{ 0x3C12 , "HANGUL SYLLABLE THIEUTH WEO" },
	{ 0x3C13 , "HANGUL SYLLABLE THIEUTH WEO SSANGSIOS" },
	{ 0x3C14 , "HANGUL SYLLABLE THIEUTH WE" },
	{ 0x3C15 , "HANGUL SYLLABLE THIEUTH WI" },
	{ 0x3C16 , "HANGUL SYLLABLE THIEUTH WI KIYEOK" },
	{ 0x3C17 , "HANGUL SYLLABLE THIEUTH WI NIEUN" },
	{ 0x3C18 , "HANGUL SYLLABLE THIEUTH WI RIEUL" },
	{ 0x3C19 , "HANGUL SYLLABLE THIEUTH WI MIEUM" },
	{ 0x3C1A , "HANGUL SYLLABLE THIEUTH WI PIEUP" },
	{ 0x3C1B , "HANGUL SYLLABLE THIEUTH WI IEUNG" },
	{ 0x3C1C , "HANGUL SYLLABLE THIEUTH YU" },
	{ 0x3C1D , "HANGUL SYLLABLE THIEUTH YU NIEUN" },
	{ 0x3C1E , "HANGUL SYLLABLE THIEUTH YU RIEUL" },
	{ 0x3C1F , "HANGUL SYLLABLE THIEUTH YU MIEUM" },
	{ 0x3C20 , "HANGUL SYLLABLE THIEUTH YU IEUNG" },
	{ 0x3C21 , "HANGUL SYLLABLE THIEUTH EU" },
	{ 0x3C22 , "HANGUL SYLLABLE THIEUTH EU KIYEOK" },
	{ 0x3C23 , "HANGUL SYLLABLE THIEUTH EU NIEUN" },
	{ 0x3C24 , "HANGUL SYLLABLE THIEUTH EU TIKEUT" },
	{ 0x3C25 , "HANGUL SYLLABLE THIEUTH EU RIEUL" },
	{ 0x3C26 , "HANGUL SYLLABLE THIEUTH EU RIEUL-MIEUM" },
	{ 0x3C27 , "HANGUL SYLLABLE THIEUTH EU MIEUM" },
	{ 0x3C28 , "HANGUL SYLLABLE THIEUTH EU PIEUP" },
	{ 0x3C29 , "HANGUL SYLLABLE THIEUTH EU SIOS" },
	{ 0x3C2A , "HANGUL SYLLABLE THIEUTH YI" },
	{ 0x3C2B , "HANGUL SYLLABLE THIEUTH YI NIEUN" },
	{ 0x3C2C , "HANGUL SYLLABLE THIEUTH YI RIEUL" },
	{ 0x3C2D , "HANGUL SYLLABLE THIEUTH YI MIEUM" },
	{ 0x3C2E , "HANGUL SYLLABLE THIEUTH YI PIEUP" },
	{ 0x3C2F , "HANGUL SYLLABLE THIEUTH I" },
	{ 0x3C30 , "HANGUL SYLLABLE THIEUTH I KIYEOK" },
	{ 0x3C31 , "HANGUL SYLLABLE THIEUTH I NIEUN" },
	{ 0x3C32 , "HANGUL SYLLABLE THIEUTH I RIEUL" },
	{ 0x3C33 , "HANGUL SYLLABLE THIEUTH I MIEUM" },
	{ 0x3C34 , "HANGUL SYLLABLE THIEUTH I PIEUP" },
	{ 0x3C35 , "HANGUL SYLLABLE THIEUTH I SIOS" },
	{ 0x3C36 , "HANGUL SYLLABLE THIEUTH I IEUNG" },
	{ 0x3C37 , "HANGUL SYLLABLE PHIEUPH A" },
	{ 0x3C38 , "HANGUL SYLLABLE PHIEUPH A KIYEOK" },
	{ 0x3C39 , "HANGUL SYLLABLE PHIEUPH A SSANGKIYEOK" },
	{ 0x3C3A , "HANGUL SYLLABLE PHIEUPH A NIEUN" },
	{ 0x3C3B , "HANGUL SYLLABLE PHIEUPH A RIEUL" },
	{ 0x3C3C , "HANGUL SYLLABLE PHIEUPH A RIEUL-MIEUM" },
	{ 0x3C3D , "HANGUL SYLLABLE PHIEUPH A MIEUM" },
	{ 0x3C3E , "HANGUL SYLLABLE PHIEUPH A PIEUP" },
	{ 0x3C3F , "HANGUL SYLLABLE PHIEUPH A SIOS" },
	{ 0x3C40 , "HANGUL SYLLABLE PHIEUPH A SSANGSIOS" },
	{ 0x3C41 , "HANGUL SYLLABLE PHIEUPH A IEUNG" },
	{ 0x3C42 , "HANGUL SYLLABLE PHIEUPH A THIEUTH" },
	{ 0x3C43 , "HANGUL SYLLABLE PHIEUPH AE" },
	{ 0x3C44 , "HANGUL SYLLABLE PHIEUPH AE KIYEOK" },
	{ 0x3C45 , "HANGUL SYLLABLE PHIEUPH AE NIEUN" },
	{ 0x3C46 , "HANGUL SYLLABLE PHIEUPH AE RIEUL" },
	{ 0x3C47 , "HANGUL SYLLABLE PHIEUPH AE MIEUM" },
	{ 0x3C48 , "HANGUL SYLLABLE PHIEUPH AE PIEUP" },
	{ 0x3C49 , "HANGUL SYLLABLE PHIEUPH AE SIOS" },
	{ 0x3C4A , "HANGUL SYLLABLE PHIEUPH AE SSANGSIOS" },
	{ 0x3C4B , "HANGUL SYLLABLE PHIEUPH AE IEUNG" },
	{ 0x3C4C , "HANGUL SYLLABLE PHIEUPH YA" },
	{ 0x3C4D , "HANGUL SYLLABLE PHIEUPH YA KIYEOK" },
	{ 0x3C4E , "HANGUL SYLLABLE PHIEUPH EO" },
	{ 0x3C4F , "HANGUL SYLLABLE PHIEUPH EO KIYEOK" },
	{ 0x3C50 , "HANGUL SYLLABLE PHIEUPH EO NIEUN" },
	{ 0x3C51 , "HANGUL SYLLABLE PHIEUPH EO RIEUL" },
	{ 0x3C52 , "HANGUL SYLLABLE PHIEUPH EO MIEUM" },
	{ 0x3C53 , "HANGUL SYLLABLE PHIEUPH EO PIEUP" },
	{ 0x3C54 , "HANGUL SYLLABLE PHIEUPH EO SIOS" },
	{ 0x3C55 , "HANGUL SYLLABLE PHIEUPH EO SSANGSIOS" },
	{ 0x3C56 , "HANGUL SYLLABLE PHIEUPH EO IEUNG" },
	{ 0x3C57 , "HANGUL SYLLABLE PHIEUPH E" },
	{ 0x3C58 , "HANGUL SYLLABLE PHIEUPH E KIYEOK" },
	{ 0x3C59 , "HANGUL SYLLABLE PHIEUPH E NIEUN" },
	{ 0x3C5A , "HANGUL SYLLABLE PHIEUPH E RIEUL" },
	{ 0x3C5B , "HANGUL SYLLABLE PHIEUPH E MIEUM" },
	{ 0x3C5C , "HANGUL SYLLABLE PHIEUPH E PIEUP" },
	{ 0x3C5D , "HANGUL SYLLABLE PHIEUPH E SIOS" },
	{ 0x3C5E , "HANGUL SYLLABLE PHIEUPH E IEUNG" },
	{ 0x3C5F , "HANGUL SYLLABLE PHIEUPH YEO" },
	{ 0x3C60 , "HANGUL SYLLABLE PHIEUPH YEO NIEUN" },
	{ 0x3C61 , "HANGUL SYLLABLE PHIEUPH YEO RIEUL" },
	{ 0x3C62 , "HANGUL SYLLABLE PHIEUPH YEO MIEUM" },
	{ 0x3C63 , "HANGUL SYLLABLE PHIEUPH YEO PIEUP" },
	{ 0x3C64 , "HANGUL SYLLABLE PHIEUPH YEO SSANGSIOS" },
	{ 0x3C65 , "HANGUL SYLLABLE PHIEUPH YEO IEUNG" },
	{ 0x3C66 , "HANGUL SYLLABLE PHIEUPH YE" },
	{ 0x3C67 , "HANGUL SYLLABLE PHIEUPH YE RIEUL" },
	{ 0x3C68 , "HANGUL SYLLABLE PHIEUPH YE PIEUP" },
	{ 0x3C69 , "HANGUL SYLLABLE PHIEUPH YE SIOS" },
	{ 0x3C6A , "HANGUL SYLLABLE PHIEUPH O" },
	{ 0x3C6B , "HANGUL SYLLABLE PHIEUPH O KIYEOK" },
	{ 0x3C6C , "HANGUL SYLLABLE PHIEUPH O NIEUN" },
	{ 0x3C6D , "HANGUL SYLLABLE PHIEUPH O RIEUL" },
	{ 0x3C6E , "HANGUL SYLLABLE PHIEUPH O MIEUM" },
	{ 0x3C6F , "HANGUL SYLLABLE PHIEUPH O PIEUP" },
	{ 0x3C70 , "HANGUL SYLLABLE PHIEUPH O SIOS" },
	{ 0x3C71 , "HANGUL SYLLABLE PHIEUPH O IEUNG" },
	{ 0x3C72 , "HANGUL SYLLABLE PHIEUPH WA" },
	{ 0x3C73 , "HANGUL SYLLABLE PHIEUPH WA IEUNG" },
	{ 0x3C74 , "HANGUL SYLLABLE PHIEUPH OE" },
	{ 0x3C75 , "HANGUL SYLLABLE PHIEUPH OE NIEUN" },
	{ 0x3C76 , "HANGUL SYLLABLE PHIEUPH YO" },
	{ 0x3C77 , "HANGUL SYLLABLE PHIEUPH YO NIEUN" },
	{ 0x3C78 , "HANGUL SYLLABLE PHIEUPH YO RIEUL" },
	{ 0x3C79 , "HANGUL SYLLABLE PHIEUPH YO PIEUP" },
	{ 0x3C7A , "HANGUL SYLLABLE PHIEUPH YO SIOS" },
	{ 0x3C7B , "HANGUL SYLLABLE PHIEUPH U" },
	{ 0x3C7C , "HANGUL SYLLABLE PHIEUPH U KIYEOK" },
	{ 0x3C7D , "HANGUL SYLLABLE PHIEUPH U NIEUN" },
	{ 0x3C7E , "HANGUL SYLLABLE PHIEUPH U TIKEUT" },
	{ 0x3C7F , "HANGUL SYLLABLE PHIEUPH U RIEUL" },
	{ 0x3C80 , "HANGUL SYLLABLE PHIEUPH U RIEUL-MIEUM" },
	{ 0x3C81 , "HANGUL SYLLABLE PHIEUPH U MIEUM" },
	{ 0x3C82 , "HANGUL SYLLABLE PHIEUPH U PIEUP" },
	{ 0x3C83 , "HANGUL SYLLABLE PHIEUPH U SIOS" },
	{ 0x3C84 , "HANGUL SYLLABLE PHIEUPH U IEUNG" },
	{ 0x3C85 , "HANGUL SYLLABLE PHIEUPH WEO" },
	{ 0x3C86 , "HANGUL SYLLABLE PHIEUPH WEO IEUNG" },
	{ 0x3C87 , "HANGUL SYLLABLE PHIEUPH WI" },
	{ 0x3C88 , "HANGUL SYLLABLE PHIEUPH WI NIEUN" },
	{ 0x3C89 , "HANGUL SYLLABLE PHIEUPH WI RIEUL" },
	{ 0x3C8A , "HANGUL SYLLABLE PHIEUPH WI MIEUM" },
	{ 0x3C8B , "HANGUL SYLLABLE PHIEUPH WI SIOS" },
	{ 0x3C8C , "HANGUL SYLLABLE PHIEUPH YU" },
	{ 0x3C8D , "HANGUL SYLLABLE PHIEUPH YU NIEUN" },
	{ 0x3C8E , "HANGUL SYLLABLE PHIEUPH YU RIEUL" },
	{ 0x3C8F , "HANGUL SYLLABLE PHIEUPH YU MIEUM" },
	{ 0x3C90 , "HANGUL SYLLABLE PHIEUPH YU SIOS" },
	{ 0x3C91 , "HANGUL SYLLABLE PHIEUPH YU IEUNG" },
	{ 0x3C92 , "HANGUL SYLLABLE PHIEUPH EU" },
	{ 0x3C93 , "HANGUL SYLLABLE PHIEUPH EU NIEUN" },
	{ 0x3C94 , "HANGUL SYLLABLE PHIEUPH EU RIEUL" },
	{ 0x3C95 , "HANGUL SYLLABLE PHIEUPH EU MIEUM" },
	{ 0x3C96 , "HANGUL SYLLABLE PHIEUPH EU PIEUP" },
	{ 0x3C97 , "HANGUL SYLLABLE PHIEUPH EU SIOS" },
	{ 0x3C98 , "HANGUL SYLLABLE PHIEUPH I" },
	{ 0x3C99 , "HANGUL SYLLABLE PHIEUPH I KIYEOK" },
	{ 0x3C9A , "HANGUL SYLLABLE PHIEUPH I NIEUN" },
	{ 0x3C9B , "HANGUL SYLLABLE PHIEUPH I RIEUL" },
	{ 0x3C9C , "HANGUL SYLLABLE PHIEUPH I MIEUM" },
	{ 0x3C9D , "HANGUL SYLLABLE PHIEUPH I PIEUP" },
	{ 0x3C9E , "HANGUL SYLLABLE PHIEUPH I SIOS" },
	{ 0x3C9F , "HANGUL SYLLABLE PHIEUPH I IEUNG" },
	{ 0x3CA0 , "HANGUL SYLLABLE HIEUH A" },
	{ 0x3CA1 , "HANGUL SYLLABLE HIEUH A KIYEOK" },
	{ 0x3CA2 , "HANGUL SYLLABLE HIEUH A NIEUN" },
	{ 0x3CA3 , "HANGUL SYLLABLE HIEUH A RIEUL" },
	{ 0x3CA4 , "HANGUL SYLLABLE HIEUH A RIEUL-THIEUTH" },
	{ 0x3CA5 , "HANGUL SYLLABLE HIEUH A MIEUM" },
	{ 0x3CA6 , "HANGUL SYLLABLE HIEUH A PIEUP" },
	{ 0x3CA7 , "HANGUL SYLLABLE HIEUH A SIOS" },
	{ 0x3CA8 , "HANGUL SYLLABLE HIEUH A IEUNG" },
	{ 0x3CA9 , "HANGUL SYLLABLE HIEUH AE" },
	{ 0x3CAA , "HANGUL SYLLABLE HIEUH AE KIYEOK" },
	{ 0x3CAB , "HANGUL SYLLABLE HIEUH AE NIEUN" },
	{ 0x3CAC , "HANGUL SYLLABLE HIEUH AE RIEUL" },
	{ 0x3CAD , "HANGUL SYLLABLE HIEUH AE MIEUM" },
	{ 0x3CAE , "HANGUL SYLLABLE HIEUH AE PIEUP" },
	{ 0x3CAF , "HANGUL SYLLABLE HIEUH AE SIOS" },
	{ 0x3CB0 , "HANGUL SYLLABLE HIEUH AE SSANGSIOS" },
	{ 0x3CB1 , "HANGUL SYLLABLE HIEUH AE IEUNG" },
	{ 0x3CB2 , "HANGUL SYLLABLE HIEUH YA" },
	{ 0x3CB3 , "HANGUL SYLLABLE HIEUH YA IEUNG" },
	{ 0x3CB4 , "HANGUL SYLLABLE HIEUH EO" },
	{ 0x3CB5 , "HANGUL SYLLABLE HIEUH EO KIYEOK" },
	{ 0x3CB6 , "HANGUL SYLLABLE HIEUH EO NIEUN" },
	{ 0x3CB7 , "HANGUL SYLLABLE HIEUH EO RIEUL" },
	{ 0x3CB8 , "HANGUL SYLLABLE HIEUH EO RIEUL-MIEUM" },
	{ 0x3CB9 , "HANGUL SYLLABLE HIEUH EO MIEUM" },
	{ 0x3CBA , "HANGUL SYLLABLE HIEUH EO PIEUP" },
	{ 0x3CBB , "HANGUL SYLLABLE HIEUH EO SIOS" },
	{ 0x3CBC , "HANGUL SYLLABLE HIEUH EO IEUNG" },
	{ 0x3CBD , "HANGUL SYLLABLE HIEUH E" },
	{ 0x3CBE , "HANGUL SYLLABLE HIEUH E KIYEOK" },
	{ 0x3CBF , "HANGUL SYLLABLE HIEUH E NIEUN" },
	{ 0x3CC0 , "HANGUL SYLLABLE HIEUH E RIEUL" },
	{ 0x3CC1 , "HANGUL SYLLABLE HIEUH E MIEUM" },
	{ 0x3CC2 , "HANGUL SYLLABLE HIEUH E PIEUP" },
	{ 0x3CC3 , "HANGUL SYLLABLE HIEUH E SIOS" },
	{ 0x3CC4 , "HANGUL SYLLABLE HIEUH E IEUNG" },
	{ 0x3CC5 , "HANGUL SYLLABLE HIEUH YEO" },
	{ 0x3CC6 , "HANGUL SYLLABLE HIEUH YEO KIYEOK" },
	{ 0x3CC7 , "HANGUL SYLLABLE HIEUH YEO NIEUN" },
	{ 0x3CC8 , "HANGUL SYLLABLE HIEUH YEO RIEUL" },
	{ 0x3CC9 , "HANGUL SYLLABLE HIEUH YEO MIEUM" },
	{ 0x3CCA , "HANGUL SYLLABLE HIEUH YEO PIEUP" },
	{ 0x3CCB , "HANGUL SYLLABLE HIEUH YEO SIOS" },
	{ 0x3CCC , "HANGUL SYLLABLE HIEUH YEO SSANGSIOS" },
	{ 0x3CCD , "HANGUL SYLLABLE HIEUH YEO IEUNG" },
	{ 0x3CCE , "HANGUL SYLLABLE HIEUH YE" },
	{ 0x3CCF , "HANGUL SYLLABLE HIEUH YE NIEUN" },
	{ 0x3CD0 , "HANGUL SYLLABLE HIEUH YE RIEUL" },
	{ 0x3CD1 , "HANGUL SYLLABLE HIEUH YE PIEUP" },
	{ 0x3CD2 , "HANGUL SYLLABLE HIEUH O" },
	{ 0x3CD3 , "HANGUL SYLLABLE HIEUH O KIYEOK" },
	{ 0x3CD4 , "HANGUL SYLLABLE HIEUH O NIEUN" },
	{ 0x3CD5 , "HANGUL SYLLABLE HIEUH O RIEUL" },
	{ 0x3CD6 , "HANGUL SYLLABLE HIEUH O RIEUL-THIEUTH" },
	{ 0x3CD7 , "HANGUL SYLLABLE HIEUH O MIEUM" },
	{ 0x3CD8 , "HANGUL SYLLABLE HIEUH O PIEUP" },
	{ 0x3CD9 , "HANGUL SYLLABLE HIEUH O SIOS" },
	{ 0x3CDA , "HANGUL SYLLABLE HIEUH O IEUNG" },
	{ 0x3CDB , "HANGUL SYLLABLE HIEUH O THIEUTH" },
	{ 0x3CDC , "HANGUL SYLLABLE HIEUH WA" },
	{ 0x3CDD , "HANGUL SYLLABLE HIEUH WA KIYEOK" },
	{ 0x3CDE , "HANGUL SYLLABLE HIEUH WA NIEUN" },
	{ 0x3CDF , "HANGUL SYLLABLE HIEUH WA RIEUL" },
	{ 0x3CE0 , "HANGUL SYLLABLE HIEUH WA SIOS" },
	{ 0x3CE1 , "HANGUL SYLLABLE HIEUH WA IEUNG" },
	{ 0x3CE2 , "HANGUL SYLLABLE HIEUH WAE" },
	{ 0x3CE3 , "HANGUL SYLLABLE HIEUH WAE KIYEOK" },
	{ 0x3CE4 , "HANGUL SYLLABLE HIEUH WAE NIEUN" },
	{ 0x3CE5 , "HANGUL SYLLABLE HIEUH WAE SIOS" },
	{ 0x3CE6 , "HANGUL SYLLABLE HIEUH WAE IEUNG" },
	{ 0x3CE7 , "HANGUL SYLLABLE HIEUH OE" },
	{ 0x3CE8 , "HANGUL SYLLABLE HIEUH OE KIYEOK" },
	{ 0x3CE9 , "HANGUL SYLLABLE HIEUH OE NIEUN" },
	{ 0x3CEA , "HANGUL SYLLABLE HIEUH OE RIEUL" },
	{ 0x3CEB , "HANGUL SYLLABLE HIEUH OE PIEUP" },
	{ 0x3CEC , "HANGUL SYLLABLE HIEUH OE SIOS" },
	{ 0x3CED , "HANGUL SYLLABLE HIEUH OE IEUNG" },
	{ 0x3CEE , "HANGUL SYLLABLE HIEUH YO" },
	{ 0x3CEF , "HANGUL SYLLABLE HIEUH YO NIEUN" },
	{ 0x3CF0 , "HANGUL SYLLABLE HIEUH YO RIEUL" },
	{ 0x3CF1 , "HANGUL SYLLABLE HIEUH YO PIEUP" },
	{ 0x3CF2 , "HANGUL SYLLABLE HIEUH YO SIOS" },
	{ 0x3CF3 , "HANGUL SYLLABLE HIEUH U" },
	{ 0x3CF4 , "HANGUL SYLLABLE HIEUH U KIYEOK" },
	{ 0x3CF5 , "HANGUL SYLLABLE HIEUH U NIEUN" },
	{ 0x3CF6 , "HANGUL SYLLABLE HIEUH U RIEUL" },
	{ 0x3CF7 , "HANGUL SYLLABLE HIEUH U RIEUL-THIEUTH" },
	{ 0x3CF8 , "HANGUL SYLLABLE HIEUH U MIEUM" },
	{ 0x3CF9 , "HANGUL SYLLABLE HIEUH U SIOS" },
	{ 0x3CFA , "HANGUL SYLLABLE HIEUH U IEUNG" },
	{ 0x3CFB , "HANGUL SYLLABLE HIEUH WEO" },
	{ 0x3CFC , "HANGUL SYLLABLE HIEUH WEO NIEUN" },
	{ 0x3CFD , "HANGUL SYLLABLE HIEUH WEO RIEUL" },
	{ 0x3CFE , "HANGUL SYLLABLE HIEUH WEO MIEUM" },
	{ 0x3CFF , "HANGUL SYLLABLE HIEUH WEO IEUNG" },
	{ 0x3D00 , "HANGUL SYLLABLE HIEUH WE" },
	{ 0x3D01 , "HANGUL SYLLABLE HIEUH WE KIYEOK" },
	{ 0x3D02 , "HANGUL SYLLABLE HIEUH WE NIEUN" },
	{ 0x3D03 , "HANGUL SYLLABLE HIEUH WE RIEUL" },
	{ 0x3D04 , "HANGUL SYLLABLE HIEUH WE IEUNG" },
	{ 0x3D05 , "HANGUL SYLLABLE HIEUH WI" },
	{ 0x3D06 , "HANGUL SYLLABLE HIEUH WI KIYEOK" },
	{ 0x3D07 , "HANGUL SYLLABLE HIEUH WI NIEUN" },
	{ 0x3D08 , "HANGUL SYLLABLE HIEUH WI RIEUL" },
	{ 0x3D09 , "HANGUL SYLLABLE HIEUH WI MIEUM" },
	{ 0x3D0A , "HANGUL SYLLABLE HIEUH WI PIEUP" },
	{ 0x3D0B , "HANGUL SYLLABLE HIEUH WI SIOS" },
	{ 0x3D0C , "HANGUL SYLLABLE HIEUH WI IEUNG" },
	{ 0x3D0D , "HANGUL SYLLABLE HIEUH YU" },
	{ 0x3D0E , "HANGUL SYLLABLE HIEUH YU KIYEOK" },
	{ 0x3D0F , "HANGUL SYLLABLE HIEUH YU NIEUN" },
	{ 0x3D10 , "HANGUL SYLLABLE HIEUH YU RIEUL" },
	{ 0x3D11 , "HANGUL SYLLABLE HIEUH YU MIEUM" },
	{ 0x3D12 , "HANGUL SYLLABLE HIEUH YU SIOS" },
	{ 0x3D13 , "HANGUL SYLLABLE HIEUH YU IEUNG" },
	{ 0x3D14 , "HANGUL SYLLABLE HIEUH EU" },
	{ 0x3D15 , "HANGUL SYLLABLE HIEUH EU KIYEOK" },
	{ 0x3D16 , "HANGUL SYLLABLE HIEUH EU NIEUN" },
	{ 0x3D17 , "HANGUL SYLLABLE HIEUH EU NIEUN-HIEUH" },
	{ 0x3D18 , "HANGUL SYLLABLE HIEUH EU TIKEUT" },
	{ 0x3D19 , "HANGUL SYLLABLE HIEUH EU RIEUL" },
	{ 0x3D1A , "HANGUL SYLLABLE HIEUH EU RIEUL-KIYEOK" },
	{ 0x3D1B , "HANGUL SYLLABLE HIEUH EU MIEUM" },
	{ 0x3D1C , "HANGUL SYLLABLE HIEUH EU PIEUP" },
	{ 0x3D1D , "HANGUL SYLLABLE HIEUH EU SIOS" },
	{ 0x3D1E , "HANGUL SYLLABLE HIEUH EU IEUNG" },
	{ 0x3D1F , "HANGUL SYLLABLE HIEUH EU THIEUTH" },
	{ 0x3D20 , "HANGUL SYLLABLE HIEUH YI" },
	{ 0x3D21 , "HANGUL SYLLABLE HIEUH YI NIEUN" },
	{ 0x3D22 , "HANGUL SYLLABLE HIEUH YI RIEUL" },
	{ 0x3D23 , "HANGUL SYLLABLE HIEUH YI MIEUM" },
	{ 0x3D24 , "HANGUL SYLLABLE HIEUH YI PIEUP" },
	{ 0x3D25 , "HANGUL SYLLABLE HIEUH YI IEUNG" },
	{ 0x3D26 , "HANGUL SYLLABLE HIEUH I" },
	{ 0x3D27 , "HANGUL SYLLABLE HIEUH I KIYEOK" },
	{ 0x3D28 , "HANGUL SYLLABLE HIEUH I NIEUN" },
	{ 0x3D29 , "HANGUL SYLLABLE HIEUH I RIEUL" },
	{ 0x3D2A , "HANGUL SYLLABLE HIEUH I MIEUM" },
	{ 0x3D2B , "HANGUL SYLLABLE HIEUH I PIEUP" },
	{ 0x3D2C , "HANGUL SYLLABLE HIEUH I SIOS" },
	{ 0x3D2D , "HANGUL SYLLABLE HIEUH I IEUNG" },
	{ 0x3D2E , "HANGUL SYLLABLE KIYEOK A SSANGKIYEOK" },
	{ 0x3D2F , "HANGUL SYLLABLE KIYEOK A RIEUL-PIEUP" },
	{ 0x3D30 , "HANGUL SYLLABLE KIYEOK A RIEUL-SIOS" },
	{ 0x3D31 , "HANGUL SYLLABLE KIYEOK AE NIEUN-HIEUH" },
	{ 0x3D32 , "HANGUL SYLLABLE KIYEOK AE TIKEUT" },
	{ 0x3D33 , "HANGUL SYLLABLE KIYEOK AE CIEUC" },
	{ 0x3D34 , "HANGUL SYLLABLE KIYEOK AE THIEUTH" },
	{ 0x3D35 , "HANGUL SYLLABLE KIYEOK AE PHIEUPH" },
	{ 0x3D36 , "HANGUL SYLLABLE KIYEOK YA TIKEUT" },
	{ 0x3D37 , "HANGUL SYLLABLE KIYEOK YA RIEUL-KIYEOK" },
	{ 0x3D38 , "HANGUL SYLLABLE KIYEOK YA RIEUL-HIEUH" },
	{ 0x3D39 , "HANGUL SYLLABLE KIYEOK YA MIEUM" },
	{ 0x3D3A , "HANGUL SYLLABLE KIYEOK YA PIEUP" },
	{ 0x3D3B , "HANGUL SYLLABLE KIYEOK YA SSANGSIOS" },
	{ 0x3D3C , "HANGUL SYLLABLE KIYEOK YAE MIEUM" },
	{ 0x3D3D , "HANGUL SYLLABLE KIYEOK YAE PIEUP" },
	{ 0x3D3E , "HANGUL SYLLABLE KIYEOK EO KIYEOK-SIOS" },
	{ 0x3D3F , "HANGUL SYLLABLE KIYEOK EO NIEUN-CIEUC" },
	{ 0x3D40 , "HANGUL SYLLABLE KIYEOK EO RIEUL-KIYEOK" },
	{ 0x3D41 , "HANGUL SYLLABLE KIYEOK EO CHIEUCH" },
	{ 0x3D42 , "HANGUL SYLLABLE KIYEOK E KIYEOK" },
	{ 0x3D43 , "HANGUL SYLLABLE KIYEOK E TIKEUT" },
	{ 0x3D44 , "HANGUL SYLLABLE KIYEOK E THIEUTH" },
	{ 0x3D45 , "HANGUL SYLLABLE KIYEOK E HIEUH" },
	{ 0x3D46 , "HANGUL SYLLABLE KIYEOK YEO RIEUL-KIYEOK" },
	{ 0x3D47 , "HANGUL SYLLABLE KIYEOK YEO RIEUL-SIOS" },
	{ 0x3D48 , "HANGUL SYLLABLE KIYEOK YEO RIEUL-HIEUH" },
	{ 0x3D49 , "HANGUL SYLLABLE KIYEOK YEO CIEUC" },
	{ 0x3D4A , "HANGUL SYLLABLE KIYEOK YEO CHIEUCH" },
	{ 0x3D4B , "HANGUL SYLLABLE KIYEOK YEO PHIEUPH" },
	{ 0x3D4C , "HANGUL SYLLABLE KIYEOK YE KIYEOK" },
	{ 0x3D4D , "HANGUL SYLLABLE KIYEOK YE TIKEUT" },
	{ 0x3D4E , "HANGUL SYLLABLE KIYEOK YE MIEUM" },
	{ 0x3D4F , "HANGUL SYLLABLE KIYEOK YE SSANGSIOS" },
	{ 0x3D50 , "HANGUL SYLLABLE KIYEOK YE IEUNG" },
	{ 0x3D51 , "HANGUL SYLLABLE KIYEOK O RIEUL-KIYEOK" },
	{ 0x3D52 , "HANGUL SYLLABLE KIYEOK O RIEUL-PIEUP" },
	{ 0x3D53 , "HANGUL SYLLABLE KIYEOK O RIEUL-PHIEUPH" },
	{ 0x3D54 , "HANGUL SYLLABLE KIYEOK O CHIEUCH" },
	{ 0x3D55 , "HANGUL SYLLABLE KIYEOK O THIEUTH" },
	{ 0x3D56 , "HANGUL SYLLABLE KIYEOK O PHIEUPH" },
	{ 0x3D57 , "HANGUL SYLLABLE KIYEOK O HIEUH" },
	{ 0x3D58 , "HANGUL SYLLABLE KIYEOK WA TIKEUT" },
	{ 0x3D59 , "HANGUL SYLLABLE KIYEOK WA SSANGSIOS" },
	{ 0x3D5A , "HANGUL SYLLABLE KIYEOK WAE KIYEOK" },
	{ 0x3D5B , "HANGUL SYLLABLE KIYEOK WAE TIKEUT" },
	{ 0x3D5C , "HANGUL SYLLABLE KIYEOK WAE RIEUL-MIEUM" },
	{ 0x3D5D , "HANGUL SYLLABLE KIYEOK WAE MIEUM" },
	{ 0x3D5E , "HANGUL SYLLABLE KIYEOK WAE SIOS" },
	{ 0x3D5F , "HANGUL SYLLABLE KIYEOK OE TIKEUT" },
	{ 0x3D60 , "HANGUL SYLLABLE KIYEOK OE RIEUL-MIEUM" },
	{ 0x3D61 , "HANGUL SYLLABLE KIYEOK OE SSANGSIOS" },
	{ 0x3D62 , "HANGUL SYLLABLE KIYEOK YO KIYEOK" },
	{ 0x3D63 , "HANGUL SYLLABLE KIYEOK YO TIKEUT" },
	{ 0x3D64 , "HANGUL SYLLABLE KIYEOK YO MIEUM" },
	{ 0x3D65 , "HANGUL SYLLABLE KIYEOK YO IEUNG" },
	{ 0x3D66 , "HANGUL SYLLABLE KIYEOK U RIEUL-SIOS" },
	{ 0x3D67 , "HANGUL SYLLABLE KIYEOK U CHIEUCH" },
	{ 0x3D68 , "HANGUL SYLLABLE KIYEOK U PHIEUPH" },
	{ 0x3D69 , "HANGUL SYLLABLE KIYEOK WEO TIKEUT" },
	{ 0x3D6A , "HANGUL SYLLABLE KIYEOK WEO PIEUP" },
	{ 0x3D6B , "HANGUL SYLLABLE KIYEOK WEO SIOS" },
	{ 0x3D6C , "HANGUL SYLLABLE KIYEOK WE KIYEOK" },
	{ 0x3D6D , "HANGUL SYLLABLE KIYEOK WE NIEUN" },
	{ 0x3D6E , "HANGUL SYLLABLE KIYEOK WE TIKEUT" },
	{ 0x3D6F , "HANGUL SYLLABLE KIYEOK WE RIEUL" },
	{ 0x3D70 , "HANGUL SYLLABLE KIYEOK WE MIEUM" },
	{ 0x3D71 , "HANGUL SYLLABLE KIYEOK WE PIEUP" },
	{ 0x3D72 , "HANGUL SYLLABLE KIYEOK WE SSANGSIOS" },
	{ 0x3D73 , "HANGUL SYLLABLE KIYEOK WE IEUNG" },
	{ 0x3D74 , "HANGUL SYLLABLE KIYEOK WI TIKEUT" },
	{ 0x3D75 , "HANGUL SYLLABLE KIYEOK WI RIEUL-MIEUM" },
	{ 0x3D76 , "HANGUL SYLLABLE KIYEOK WI SSANGSIOS" },
	{ 0x3D77 , "HANGUL SYLLABLE KIYEOK WI IEUNG" },
	{ 0x3D78 , "HANGUL SYLLABLE KIYEOK YU KIYEOK" },
	{ 0x3D79 , "HANGUL SYLLABLE KIYEOK YU RIEUL-SIOS" },
	{ 0x3D7A , "HANGUL SYLLABLE KIYEOK YU MIEUM" },
	{ 0x3D7B , "HANGUL SYLLABLE KIYEOK YU PIEUP" },
	{ 0x3D7C , "HANGUL SYLLABLE KIYEOK YU SIOS" },
	{ 0x3D7D , "HANGUL SYLLABLE KIYEOK YU IEUNG" },
	{ 0x3D7E , "HANGUL SYLLABLE KIYEOK EU NIEUN-HIEUH" },
	{ 0x3D7F , "HANGUL SYLLABLE KIYEOK EU RIEUL-MIEUM" },
	{ 0x3D80 , "HANGUL SYLLABLE KIYEOK EU RIEUL-PIEUP" },
	{ 0x3D81 , "HANGUL SYLLABLE KIYEOK EU RIEUL-SIOS" },
	{ 0x3D82 , "HANGUL SYLLABLE KIYEOK EU RIEUL-HIEUH" },
	{ 0x3D83 , "HANGUL SYLLABLE KIYEOK EU CIEUC" },
	{ 0x3D84 , "HANGUL SYLLABLE KIYEOK EU CHIEUCH" },
	{ 0x3D85 , "HANGUL SYLLABLE KIYEOK EU THIEUTH" },
	{ 0x3D86 , "HANGUL SYLLABLE KIYEOK EU PHIEUPH" },
	{ 0x3D87 , "HANGUL SYLLABLE KIYEOK EU HIEUH" },
	{ 0x3D88 , "HANGUL SYLLABLE KIYEOK YI KIYEOK" },
	{ 0x3D89 , "HANGUL SYLLABLE KIYEOK YI NIEUN" },
	{ 0x3D8A , "HANGUL SYLLABLE KIYEOK YI RIEUL" },
	{ 0x3D8B , "HANGUL SYLLABLE KIYEOK YI RIEUL-SIOS" },
	{ 0x3D8C , "HANGUL SYLLABLE KIYEOK YI MIEUM" },
	{ 0x3D8D , "HANGUL SYLLABLE KIYEOK YI PIEUP" },
	{ 0x3D8E , "HANGUL SYLLABLE KIYEOK YI SIOS" },
	{ 0x3D8F , "HANGUL SYLLABLE KIYEOK YI IEUNG" },
	{ 0x3D90 , "HANGUL SYLLABLE KIYEOK I RIEUL-KIYEOK" },
	{ 0x3D91 , "HANGUL SYLLABLE KIYEOK I RIEUL-SIOS" },
	{ 0x3D92 , "HANGUL SYLLABLE KIYEOK I SSANGSIOS" },
	{ 0x3D93 , "HANGUL SYLLABLE KIYEOK I CHIEUCH" },
	{ 0x3D94 , "HANGUL SYLLABLE KIYEOK I THIEUTH" },
	{ 0x3D95 , "HANGUL SYLLABLE KIYEOK I HIEUH" },
	{ 0x3D96 , "HANGUL SYLLABLE SSANGKIYEOK A TIKEUT" },
	{ 0x3D97 , "HANGUL SYLLABLE SSANGKIYEOK A CIEUC" },
	{ 0x3D98 , "HANGUL SYLLABLE SSANGKIYEOK A CHIEUCH" },
	{ 0x3D99 , "HANGUL SYLLABLE SSANGKIYEOK AE SSANGKIYEOK" },
	{ 0x3D9A , "HANGUL SYLLABLE SSANGKIYEOK AE TIKEUT" },
	{ 0x3D9B , "HANGUL SYLLABLE SSANGKIYEOK AE RIEUL-PIEUP" },
	{ 0x3D9C , "HANGUL SYLLABLE SSANGKIYEOK AE RIEUL-PHIEUPH" },
	{ 0x3D9D , "HANGUL SYLLABLE SSANGKIYEOK AE THIEUTH" },
	{ 0x3D9E , "HANGUL SYLLABLE SSANGKIYEOK YA NIEUN" },
	{ 0x3D9F , "HANGUL SYLLABLE SSANGKIYEOK YA TIKEUT" },
	{ 0x3DA0 , "HANGUL SYLLABLE SSANGKIYEOK YA RIEUL-KIYEOK" },
	{ 0x3DA1 , "HANGUL SYLLABLE SSANGKIYEOK YA SIOS" },
	{ 0x3DA2 , "HANGUL SYLLABLE SSANGKIYEOK YA IEUNG" },
	{ 0x3DA3 , "HANGUL SYLLABLE SSANGKIYEOK YAE" },
	{ 0x3DA4 , "HANGUL SYLLABLE SSANGKIYEOK YAE IEUNG" },
	{ 0x3DA5 , "HANGUL SYLLABLE SSANGKIYEOK EO NIEUN-HIEUH" },
	{ 0x3DA6 , "HANGUL SYLLABLE SSANGKIYEOK EO TIKEUT" },
	{ 0x3DA7 , "HANGUL SYLLABLE SSANGKIYEOK EO CIEUC" },
	{ 0x3DA8 , "HANGUL SYLLABLE SSANGKIYEOK EO CHIEUCH" },
	{ 0x3DA9 , "HANGUL SYLLABLE SSANGKIYEOK EO THIEUTH" },
	{ 0x3DAA , "HANGUL SYLLABLE SSANGKIYEOK E SSANGKIYEOK" },
	{ 0x3DAB , "HANGUL SYLLABLE SSANGKIYEOK E TIKEUT" },
	{ 0x3DAC , "HANGUL SYLLABLE SSANGKIYEOK E RIEUL" },
	{ 0x3DAD , "HANGUL SYLLABLE SSANGKIYEOK E PIEUP" },
	{ 0x3DAE , "HANGUL SYLLABLE SSANGKIYEOK E SSANGSIOS" },
	{ 0x3DAF , "HANGUL SYLLABLE SSANGKIYEOK E THIEUTH" },
	{ 0x3DB0 , "HANGUL SYLLABLE SSANGKIYEOK YEO KIYEOK" },
	{ 0x3DB1 , "HANGUL SYLLABLE SSANGKIYEOK YEO TIKEUT" },
	{ 0x3DB2 , "HANGUL SYLLABLE SSANGKIYEOK YEO MIEUM" },
	{ 0x3DB3 , "HANGUL SYLLABLE SSANGKIYEOK YEO PIEUP" },
	{ 0x3DB4 , "HANGUL SYLLABLE SSANGKIYEOK YEO IEUNG" },
	{ 0x3DB5 , "HANGUL SYLLABLE SSANGKIYEOK YE TIKEUT" },
	{ 0x3DB6 , "HANGUL SYLLABLE SSANGKIYEOK YE IEUNG" },
	{ 0x3DB7 , "HANGUL SYLLABLE SSANGKIYEOK O TIKEUT" },
	{ 0x3DB8 , "HANGUL SYLLABLE SSANGKIYEOK O RIEUL-MIEUM" },
	{ 0x3DB9 , "HANGUL SYLLABLE SSANGKIYEOK O RIEUL-SIOS" },
	{ 0x3DBA , "HANGUL SYLLABLE SSANGKIYEOK O RIEUL-HIEUH" },
	{ 0x3DBB , "HANGUL SYLLABLE SSANGKIYEOK O THIEUTH" },
	{ 0x3DBC , "HANGUL SYLLABLE SSANGKIYEOK O HIEUH" },
	{ 0x3DBD , "HANGUL SYLLABLE SSANGKIYEOK WA NIEUN" },
	{ 0x3DBE , "HANGUL SYLLABLE SSANGKIYEOK WA TIKEUT" },
	{ 0x3DBF , "HANGUL SYLLABLE SSANGKIYEOK WA MIEUM" },
	{ 0x3DC0 , "HANGUL SYLLABLE SSANGKIYEOK WA PIEUP" },
	{ 0x3DC1 , "HANGUL SYLLABLE SSANGKIYEOK WA SIOS" },
	{ 0x3DC2 , "HANGUL SYLLABLE SSANGKIYEOK WAE NIEUN" },
	{ 0x3DC3 , "HANGUL SYLLABLE SSANGKIYEOK WAE TIKEUT" },
	{ 0x3DC4 , "HANGUL SYLLABLE SSANGKIYEOK WAE RIEUL" },
	{ 0x3DC5 , "HANGUL SYLLABLE SSANGKIYEOK WAE MIEUM" },
	{ 0x3DC6 , "HANGUL SYLLABLE SSANGKIYEOK WAE PIEUP" },
	{ 0x3DC7 , "HANGUL SYLLABLE SSANGKIYEOK WAE SSANGSIOS" },
	{ 0x3DC8 , "HANGUL SYLLABLE SSANGKIYEOK OE KIYEOK" },
	{ 0x3DC9 , "HANGUL SYLLABLE SSANGKIYEOK OE TIKEUT" },
	{ 0x3DCA , "HANGUL SYLLABLE SSANGKIYEOK OE SIOS" },
	{ 0x3DCB , "HANGUL SYLLABLE SSANGKIYEOK OE SSANGSIOS" },
	{ 0x3DCC , "HANGUL SYLLABLE SSANGKIYEOK YO NIEUN" },
	{ 0x3DCD , "HANGUL SYLLABLE SSANGKIYEOK YO TIKEUT" },
	{ 0x3DCE , "HANGUL SYLLABLE SSANGKIYEOK YO RIEUL" },
	{ 0x3DCF , "HANGUL SYLLABLE SSANGKIYEOK YO MIEUM" },
	{ 0x3DD0 , "HANGUL SYLLABLE SSANGKIYEOK YO PIEUP" },
	{ 0x3DD1 , "HANGUL SYLLABLE SSANGKIYEOK YO CIEUC" },
	{ 0x3DD2 , "HANGUL SYLLABLE SSANGKIYEOK U TIKEUT" },
	{ 0x3DD3 , "HANGUL SYLLABLE SSANGKIYEOK U RIEUL-KIYEOK" },
	{ 0x3DD4 , "HANGUL SYLLABLE SSANGKIYEOK U CHIEUCH" },
	{ 0x3DD5 , "HANGUL SYLLABLE SSANGKIYEOK WEO KIYEOK" },
	{ 0x3DD6 , "HANGUL SYLLABLE SSANGKIYEOK WEO NIEUN" },
	{ 0x3DD7 , "HANGUL SYLLABLE SSANGKIYEOK WEO TIKEUT" },
	{ 0x3DD8 , "HANGUL SYLLABLE SSANGKIYEOK WEO MIEUM" },
	{ 0x3DD9 , "HANGUL SYLLABLE SSANGKIYEOK WEO SIOS" },
	{ 0x3DDA , "HANGUL SYLLABLE SSANGKIYEOK WE TIKEUT" },
	{ 0x3DDB , "HANGUL SYLLABLE SSANGKIYEOK WE SIOS" },
	{ 0x3DDC , "HANGUL SYLLABLE SSANGKIYEOK WE IEUNG" },
	{ 0x3DDD , "HANGUL SYLLABLE SSANGKIYEOK WI KIYEOK" },
	{ 0x3DDE , "HANGUL SYLLABLE SSANGKIYEOK WI TIKEUT" },
	{ 0x3DDF , "HANGUL SYLLABLE SSANGKIYEOK WI SIOS" },
	{ 0x3DE0 , "HANGUL SYLLABLE SSANGKIYEOK WI IEUNG" },
	{ 0x3DE1 , "HANGUL SYLLABLE SSANGKIYEOK YU NIEUN" },
	{ 0x3DE2 , "HANGUL SYLLABLE SSANGKIYEOK YU RIEUL" },
	{ 0x3DE3 , "HANGUL SYLLABLE SSANGKIYEOK YU MIEUM" },
	{ 0x3DE4 , "HANGUL SYLLABLE SSANGKIYEOK YU PIEUP" },
	{ 0x3DE5 , "HANGUL SYLLABLE SSANGKIYEOK EU TIKEUT" },
	{ 0x3DE6 , "HANGUL SYLLABLE SSANGKIYEOK EU RIEUL-KIYEOK" },
	{ 0x3DE7 , "HANGUL SYLLABLE SSANGKIYEOK EU CHIEUCH" },
	{ 0x3DE8 , "HANGUL SYLLABLE SSANGKIYEOK EU HIEUH" },
	{ 0x3DE9 , "HANGUL SYLLABLE SSANGKIYEOK YI" },
	{ 0x3DEA , "HANGUL SYLLABLE SSANGKIYEOK YI KIYEOK" },
	{ 0x3DEB , "HANGUL SYLLABLE SSANGKIYEOK YI TIKEUT" },
	{ 0x3DEC , "HANGUL SYLLABLE SSANGKIYEOK YI RIEUL" },
	{ 0x3DED , "HANGUL SYLLABLE SSANGKIYEOK YI RIEUL-PIEUP" },
	{ 0x3DEE , "HANGUL SYLLABLE SSANGKIYEOK YI SIOS" },
	{ 0x3DEF , "HANGUL SYLLABLE SSANGKIYEOK YI IEUNG" },
	{ 0x3DF0 , "HANGUL SYLLABLE SSANGKIYEOK I TIKEUT" },
	{ 0x3DF1 , "HANGUL SYLLABLE SSANGKIYEOK I RIEUL-HIEUH" },
	{ 0x3DF2 , "HANGUL SYLLABLE SSANGKIYEOK I SSANGSIOS" },
	{ 0x3DF3 , "HANGUL SYLLABLE SSANGKIYEOK I THIEUTH" },
	{ 0x3DF4 , "HANGUL SYLLABLE NIEUN A KIYEOK-SIOS" },
	{ 0x3DF5 , "HANGUL SYLLABLE NIEUN A RIEUL-SIOS" },
	{ 0x3DF6 , "HANGUL SYLLABLE NIEUN A PIEUP-SIOS" },
	{ 0x3DF7 , "HANGUL SYLLABLE NIEUN A KHIEUKH" },
	{ 0x3DF8 , "HANGUL SYLLABLE NIEUN A PHIEUPH" },
	{ 0x3DF9 , "HANGUL SYLLABLE NIEUN AE TIKEUT" },
	{ 0x3DFA , "HANGUL SYLLABLE NIEUN AE PHIEUPH" },
	{ 0x3DFB , "HANGUL SYLLABLE NIEUN YA TIKEUT" },
	{ 0x3DFC , "HANGUL SYLLABLE NIEUN YA PIEUP" },
	{ 0x3DFD , "HANGUL SYLLABLE NIEUN YA SIOS" },
	{ 0x3DFE , "HANGUL SYLLABLE NIEUN YA CIEUC" },
	{ 0x3DFF , "HANGUL SYLLABLE NIEUN YA THIEUTH" },
	{ 0x3E00 , "HANGUL SYLLABLE NIEUN YAE" },
	{ 0x3E01 , "HANGUL SYLLABLE NIEUN YAE IEUNG" },
	{ 0x3E02 , "HANGUL SYLLABLE NIEUN EO SSANGKIYEOK" },
	{ 0x3E03 , "HANGUL SYLLABLE NIEUN EO TIKEUT" },
	{ 0x3E04 , "HANGUL SYLLABLE NIEUN EO RIEUL-HIEUH" },
	{ 0x3E05 , "HANGUL SYLLABLE NIEUN EO CIEUC" },
	{ 0x3E06 , "HANGUL SYLLABLE NIEUN EO KHIEUKH" },
	{ 0x3E07 , "HANGUL SYLLABLE NIEUN EO THIEUTH" },
	{ 0x3E08 , "HANGUL SYLLABLE NIEUN EO PHIEUPH" },
	{ 0x3E09 , "HANGUL SYLLABLE NIEUN E TIKEUT" },
	{ 0x3E0A , "HANGUL SYLLABLE NIEUN E KHIEUKH" },
	{ 0x3E0B , "HANGUL SYLLABLE NIEUN E THIEUTH" },
	{ 0x3E0C , "HANGUL SYLLABLE NIEUN E PHIEUPH" },
	{ 0x3E0D , "HANGUL SYLLABLE NIEUN E HIEUH" },
	{ 0x3E0E , "HANGUL SYLLABLE NIEUN YEO KIYEOK-SIOS" },
	{ 0x3E0F , "HANGUL SYLLABLE NIEUN YEO TIKEUT" },
	{ 0x3E10 , "HANGUL SYLLABLE NIEUN YEO RIEUL-PIEUP" },
	{ 0x3E11 , "HANGUL SYLLABLE NIEUN YEO SIOS" },
	{ 0x3E12 , "HANGUL SYLLABLE NIEUN YEO THIEUTH" },
	{ 0x3E13 , "HANGUL SYLLABLE NIEUN YEO PHIEUPH" },
	{ 0x3E14 , "HANGUL SYLLABLE NIEUN YEO HIEUH" },
	{ 0x3E15 , "HANGUL SYLLABLE NIEUN YE KIYEOK" },
	{ 0x3E16 , "HANGUL SYLLABLE NIEUN YE TIKEUT" },
	{ 0x3E17 , "HANGUL SYLLABLE NIEUN YE RIEUL" },
	{ 0x3E18 , "HANGUL SYLLABLE NIEUN YE MIEUM" },
	{ 0x3E19 , "HANGUL SYLLABLE NIEUN YE PIEUP" },
	{ 0x3E1A , "HANGUL SYLLABLE NIEUN YE SIOS" },
	{ 0x3E1B , "HANGUL SYLLABLE NIEUN YE IEUNG" },
	{ 0x3E1C , "HANGUL SYLLABLE NIEUN O SSANGKIYEOK" },
	{ 0x3E1D , "HANGUL SYLLABLE NIEUN O KIYEOK-SIOS" },
	{ 0x3E1E , "HANGUL SYLLABLE NIEUN O NIEUN-HIEUH" },
	{ 0x3E1F , "HANGUL SYLLABLE NIEUN O TIKEUT" },
	{ 0x3E20 , "HANGUL SYLLABLE NIEUN O CIEUC" },
	{ 0x3E21 , "HANGUL SYLLABLE NIEUN O KHIEUKH" },
	{ 0x3E22 , "HANGUL SYLLABLE NIEUN O THIEUTH" },
	{ 0x3E23 , "HANGUL SYLLABLE NIEUN WA KIYEOK" },
	{ 0x3E24 , "HANGUL SYLLABLE NIEUN WA TIKEUT" },
	{ 0x3E25 , "HANGUL SYLLABLE NIEUN WA MIEUM" },
	{ 0x3E26 , "HANGUL SYLLABLE NIEUN WA PIEUP" },
	{ 0x3E27 , "HANGUL SYLLABLE NIEUN WA SIOS" },
	{ 0x3E28 , "HANGUL SYLLABLE NIEUN WA IEUNG" },
	{ 0x3E29 , "HANGUL SYLLABLE NIEUN WAE" },
	{ 0x3E2A , "HANGUL SYLLABLE NIEUN WAE NIEUN" },
	{ 0x3E2B , "HANGUL SYLLABLE NIEUN WAE RIEUL" },
	{ 0x3E2C , "HANGUL SYLLABLE NIEUN WAE SSANGSIOS" },
	{ 0x3E2D , "HANGUL SYLLABLE NIEUN OE KIYEOK" },
	{ 0x3E2E , "HANGUL SYLLABLE NIEUN OE TIKEUT" },
	{ 0x3E2F , "HANGUL SYLLABLE NIEUN OE SSANGSIOS" },
	{ 0x3E30 , "HANGUL SYLLABLE NIEUN OE IEUNG" },
	{ 0x3E31 , "HANGUL SYLLABLE NIEUN OE PHIEUPH" },
	{ 0x3E32 , "HANGUL SYLLABLE NIEUN YO TIKEUT" },
	{ 0x3E33 , "HANGUL SYLLABLE NIEUN YO MIEUM" },
	{ 0x3E34 , "HANGUL SYLLABLE NIEUN U RIEUL-KIYEOK" },
	{ 0x3E35 , "HANGUL SYLLABLE NIEUN U RIEUL-SIOS" },
	{ 0x3E36 , "HANGUL SYLLABLE NIEUN U RIEUL-HIEUH" },
	{ 0x3E37 , "HANGUL SYLLABLE NIEUN U CHIEUCH" },
	{ 0x3E38 , "HANGUL SYLLABLE NIEUN U PHIEUPH" },
	{ 0x3E39 , "HANGUL SYLLABLE NIEUN WEO KIYEOK" },
	{ 0x3E3A , "HANGUL SYLLABLE NIEUN WEO NIEUN" },
	{ 0x3E3B , "HANGUL SYLLABLE NIEUN WEO TIKEUT" },
	{ 0x3E3C , "HANGUL SYLLABLE NIEUN WEO RIEUL" },
	{ 0x3E3D , "HANGUL SYLLABLE NIEUN WEO MIEUM" },
	{ 0x3E3E , "HANGUL SYLLABLE NIEUN WEO PIEUP" },
	{ 0x3E3F , "HANGUL SYLLABLE NIEUN WEO SIOS" },
	{ 0x3E40 , "HANGUL SYLLABLE NIEUN WEO IEUNG" },
	{ 0x3E41 , "HANGUL SYLLABLE NIEUN WE KIYEOK" },
	{ 0x3E42 , "HANGUL SYLLABLE NIEUN WE NIEUN" },
	{ 0x3E43 , "HANGUL SYLLABLE NIEUN WE TIKEUT" },
	{ 0x3E44 , "HANGUL SYLLABLE NIEUN WE RIEUL" },
	{ 0x3E45 , "HANGUL SYLLABLE NIEUN WE MIEUM" },
	{ 0x3E46 , "HANGUL SYLLABLE NIEUN WE PIEUP" },
	{ 0x3E47 , "HANGUL SYLLABLE NIEUN WE SIOS" },
	{ 0x3E48 , "HANGUL SYLLABLE NIEUN WE SSANGSIOS" },
	{ 0x3E49 , "HANGUL SYLLABLE NIEUN WE IEUNG" },
	{ 0x3E4A , "HANGUL SYLLABLE NIEUN WI KIYEOK" },
	{ 0x3E4B , "HANGUL SYLLABLE NIEUN WI TIKEUT" },
	{ 0x3E4C , "HANGUL SYLLABLE NIEUN WI SIOS" },
	{ 0x3E4D , "HANGUL SYLLABLE NIEUN WI IEUNG" },
	{ 0x3E4E , "HANGUL SYLLABLE NIEUN WI CHIEUCH" },
	{ 0x3E4F , "HANGUL SYLLABLE NIEUN YU NIEUN" },
	{ 0x3E50 , "HANGUL SYLLABLE NIEUN YU TIKEUT" },
	{ 0x3E51 , "HANGUL SYLLABLE NIEUN YU SIOS" },
	{ 0x3E52 , "HANGUL SYLLABLE NIEUN YU CIEUC" },
	{ 0x3E53 , "HANGUL SYLLABLE NIEUN YU CHIEUCH" },
	{ 0x3E54 , "HANGUL SYLLABLE NIEUN EU TIKEUT" },
	{ 0x3E55 , "HANGUL SYLLABLE NIEUN EU RIEUL-SIOS" },
	{ 0x3E56 , "HANGUL SYLLABLE NIEUN EU CHIEUCH" },
	{ 0x3E57 , "HANGUL SYLLABLE NIEUN EU HIEUH" },
	{ 0x3E58 , "HANGUL SYLLABLE NIEUN YI KIYEOK" },
	{ 0x3E59 , "HANGUL SYLLABLE NIEUN YI TIKEUT" },
	{ 0x3E5A , "HANGUL SYLLABLE NIEUN YI MIEUM" },
	{ 0x3E5B , "HANGUL SYLLABLE NIEUN YI PIEUP" },
	{ 0x3E5C , "HANGUL SYLLABLE NIEUN YI SIOS" },
	{ 0x3E5D , "HANGUL SYLLABLE NIEUN YI SSANGSIOS" },
	{ 0x3E5E , "HANGUL SYLLABLE NIEUN YI IEUNG" },
	{ 0x3E5F , "HANGUL SYLLABLE NIEUN I TIKEUT" },
	{ 0x3E60 , "HANGUL SYLLABLE NIEUN I RIEUL-KIYEOK" },
	{ 0x3E61 , "HANGUL SYLLABLE NIEUN I RIEUL-PIEUP" },
	{ 0x3E62 , "HANGUL SYLLABLE NIEUN I RIEUL-SIOS" },
	{ 0x3E63 , "HANGUL SYLLABLE NIEUN I CIEUC" },
	{ 0x3E64 , "HANGUL SYLLABLE NIEUN I KHIEUKH" },
	{ 0x3E65 , "HANGUL SYLLABLE NIEUN I HIEUH" },
	{ 0x3E66 , "HANGUL SYLLABLE TIKEUT A RIEUL-SIOS" },
	{ 0x3E67 , "HANGUL SYLLABLE TIKEUT A RIEUL-PHIEUPH" },
	{ 0x3E68 , "HANGUL SYLLABLE TIKEUT A PIEUP-SIOS" },
	{ 0x3E69 , "HANGUL SYLLABLE TIKEUT A KHIEUKH" },
	{ 0x3E6A , "HANGUL SYLLABLE TIKEUT A THIEUTH" },
	{ 0x3E6B , "HANGUL SYLLABLE TIKEUT AE TIKEUT" },
	{ 0x3E6C , "HANGUL SYLLABLE TIKEUT AE THIEUTH" },
	{ 0x3E6D , "HANGUL SYLLABLE TIKEUT YA KIYEOK" },
	{ 0x3E6E , "HANGUL SYLLABLE TIKEUT YA NIEUN" },
	{ 0x3E6F , "HANGUL SYLLABLE TIKEUT YA RIEUL" },
	{ 0x3E70 , "HANGUL SYLLABLE TIKEUT YA RIEUL-PIEUP" },
	{ 0x3E71 , "HANGUL SYLLABLE TIKEUT YA MIEUM" },
	{ 0x3E72 , "HANGUL SYLLABLE TIKEUT YA PIEUP" },
	{ 0x3E73 , "HANGUL SYLLABLE TIKEUT YA SIOS" },
	{ 0x3E74 , "HANGUL SYLLABLE TIKEUT YA IEUNG" },
	{ 0x3E75 , "HANGUL SYLLABLE TIKEUT YAE" },
	{ 0x3E76 , "HANGUL SYLLABLE TIKEUT YAE NIEUN" },
	{ 0x3E77 , "HANGUL SYLLABLE TIKEUT YAE IEUNG" },
	{ 0x3E78 , "HANGUL SYLLABLE TIKEUT EO NIEUN-CIEUC" },
	{ 0x3E79 , "HANGUL SYLLABLE TIKEUT EO PIEUP-SIOS" },
	{ 0x3E7A , "HANGUL SYLLABLE TIKEUT EO SSANGSIOS" },
	{ 0x3E7B , "HANGUL SYLLABLE TIKEUT EO CIEUC" },
	{ 0x3E7C , "HANGUL SYLLABLE TIKEUT EO THIEUTH" },
	{ 0x3E7D , "HANGUL SYLLABLE TIKEUT EO HIEUH" },
	{ 0x3E7E , "HANGUL SYLLABLE TIKEUT E TIKEUT" },
	{ 0x3E7F , "HANGUL SYLLABLE TIKEUT E CIEUC" },
	{ 0x3E80 , "HANGUL SYLLABLE TIKEUT E PHIEUPH" },
	{ 0x3E81 , "HANGUL SYLLABLE TIKEUT YEO KIYEOK" },
	{ 0x3E82 , "HANGUL SYLLABLE TIKEUT YEO KIYEOK-SIOS" },
	{ 0x3E83 , "HANGUL SYLLABLE TIKEUT YEO TIKEUT" },
	{ 0x3E84 , "HANGUL SYLLABLE TIKEUT YEO RIEUL-SIOS" },
	{ 0x3E85 , "HANGUL SYLLABLE TIKEUT YEO MIEUM" },
	{ 0x3E86 , "HANGUL SYLLABLE TIKEUT YEO PIEUP" },
	{ 0x3E87 , "HANGUL SYLLABLE TIKEUT YEO SIOS" },
	{ 0x3E88 , "HANGUL SYLLABLE TIKEUT YE KIYEOK" },
	{ 0x3E89 , "HANGUL SYLLABLE TIKEUT YE RIEUL" },
	{ 0x3E8A , "HANGUL SYLLABLE TIKEUT YE MIEUM" },
	{ 0x3E8B , "HANGUL SYLLABLE TIKEUT YE PIEUP" },
	{ 0x3E8C , "HANGUL SYLLABLE TIKEUT YE SIOS" },
	{ 0x3E8D , "HANGUL SYLLABLE TIKEUT YE IEUNG" },
	{ 0x3E8E , "HANGUL SYLLABLE TIKEUT O SSANGKIYEOK" },
	{ 0x3E8F , "HANGUL SYLLABLE TIKEUT O KIYEOK-SIOS" },
	{ 0x3E90 , "HANGUL SYLLABLE TIKEUT O RIEUL-PIEUP" },
	{ 0x3E91 , "HANGUL SYLLABLE TIKEUT O RIEUL-HIEUH" },
	{ 0x3E92 , "HANGUL SYLLABLE TIKEUT O CIEUC" },
	{ 0x3E93 , "HANGUL SYLLABLE TIKEUT O KHIEUKH" },
	{ 0x3E94 , "HANGUL SYLLABLE TIKEUT O PHIEUPH" },
	{ 0x3E95 , "HANGUL SYLLABLE TIKEUT O HIEUH" },
	{ 0x3E96 , "HANGUL SYLLABLE TIKEUT WA KIYEOK" },
	{ 0x3E97 , "HANGUL SYLLABLE TIKEUT WA MIEUM" },
	{ 0x3E98 , "HANGUL SYLLABLE TIKEUT WA PIEUP" },
	{ 0x3E99 , "HANGUL SYLLABLE TIKEUT WA SIOS" },
	{ 0x3E9A , "HANGUL SYLLABLE TIKEUT WA SSANGSIOS" },
	{ 0x3E9B , "HANGUL SYLLABLE TIKEUT WA IEUNG" },
	{ 0x3E9C , "HANGUL SYLLABLE TIKEUT WAE NIEUN" },
	{ 0x3E9D , "HANGUL SYLLABLE TIKEUT WAE TIKEUT" },
	{ 0x3E9E , "HANGUL SYLLABLE TIKEUT WAE RIEUL" },
	{ 0x3E9F , "HANGUL SYLLABLE TIKEUT WAE MIEUM" },
	{ 0x3EA0 , "HANGUL SYLLABLE TIKEUT WAE PIEUP" },
	{ 0x3EA1 , "HANGUL SYLLABLE TIKEUT WAE SIOS" },
	{ 0x3EA2 , "HANGUL SYLLABLE TIKEUT OE KIYEOK" },
	{ 0x3EA3 , "HANGUL SYLLABLE TIKEUT OE TIKEUT" },
	{ 0x3EA4 , "HANGUL SYLLABLE TIKEUT OE RIEUL-SIOS" },
	{ 0x3EA5 , "HANGUL SYLLABLE TIKEUT OE SSANGSIOS" },
	{ 0x3EA6 , "HANGUL SYLLABLE TIKEUT OE IEUNG" },
	{ 0x3EA7 , "HANGUL SYLLABLE TIKEUT YO KIYEOK" },
	{ 0x3EA8 , "HANGUL SYLLABLE TIKEUT YO NIEUN" },
	{ 0x3EA9 , "HANGUL SYLLABLE TIKEUT YO TIKEUT" },
	{ 0x3EAA , "HANGUL SYLLABLE TIKEUT YO RIEUL" },
	{ 0x3EAB , "HANGUL SYLLABLE TIKEUT YO MIEUM" },
	{ 0x3EAC , "HANGUL SYLLABLE TIKEUT YO PIEUP" },
	{ 0x3EAD , "HANGUL SYLLABLE TIKEUT YO SIOS" },
	{ 0x3EAE , "HANGUL SYLLABLE TIKEUT YO IEUNG" },
	{ 0x3EAF , "HANGUL SYLLABLE TIKEUT YO HIEUH" },
	{ 0x3EB0 , "HANGUL SYLLABLE TIKEUT U TIKEUT" },
	{ 0x3EB1 , "HANGUL SYLLABLE TIKEUT U RIEUL-KIYEOK" },
	{ 0x3EB2 , "HANGUL SYLLABLE TIKEUT U RIEUL-MIEUM" },
	{ 0x3EB3 , "HANGUL SYLLABLE TIKEUT U RIEUL-PIEUP" },
	{ 0x3EB4 , "HANGUL SYLLABLE TIKEUT U PHIEUPH" },
	{ 0x3EB5 , "HANGUL SYLLABLE TIKEUT WEO KIYEOK" },
	{ 0x3EB6 , "HANGUL SYLLABLE TIKEUT WEO NIEUN" },
	{ 0x3EB7 , "HANGUL SYLLABLE TIKEUT WEO TIKEUT" },
	{ 0x3EB8 , "HANGUL SYLLABLE TIKEUT WEO RIEUL" },
	{ 0x3EB9 , "HANGUL SYLLABLE TIKEUT WEO MIEUM" },
	{ 0x3EBA , "HANGUL SYLLABLE TIKEUT WEO PIEUP" },
	{ 0x3EBB , "HANGUL SYLLABLE TIKEUT WEO SIOS" },
	{ 0x3EBC , "HANGUL SYLLABLE TIKEUT WEO IEUNG" },
	{ 0x3EBD , "HANGUL SYLLABLE TIKEUT WE KIYEOK" },
	{ 0x3EBE , "HANGUL SYLLABLE TIKEUT WE NIEUN" },
	{ 0x3EBF , "HANGUL SYLLABLE TIKEUT WE TIKEUT" },
	{ 0x3EC0 , "HANGUL SYLLABLE TIKEUT WE RIEUL" },
	{ 0x3EC1 , "HANGUL SYLLABLE TIKEUT WE MIEUM" },
	{ 0x3EC2 , "HANGUL SYLLABLE TIKEUT WE PIEUP" },
	{ 0x3EC3 , "HANGUL SYLLABLE TIKEUT WE SIOS" },
	{ 0x3EC4 , "HANGUL SYLLABLE TIKEUT WE SSANGSIOS" },
	{ 0x3EC5 , "HANGUL SYLLABLE TIKEUT WI KIYEOK" },
	{ 0x3EC6 , "HANGUL SYLLABLE TIKEUT WI TIKEUT" },
	{ 0x3EC7 , "HANGUL SYLLABLE TIKEUT WI MIEUM" },
	{ 0x3EC8 , "HANGUL SYLLABLE TIKEUT WI SSANGSIOS" },
	{ 0x3EC9 , "HANGUL SYLLABLE TIKEUT YU KIYEOK" },
	{ 0x3ECA , "HANGUL SYLLABLE TIKEUT YU PIEUP" },
	{ 0x3ECB , "HANGUL SYLLABLE TIKEUT YU SIOS" },
	{ 0x3ECC , "HANGUL SYLLABLE TIKEUT EU RIEUL-KIYEOK" },
	{ 0x3ECD , "HANGUL SYLLABLE TIKEUT EU RIEUL-PIEUP" },
	{ 0x3ECE , "HANGUL SYLLABLE TIKEUT EU RIEUL-SIOS" },
	{ 0x3ECF , "HANGUL SYLLABLE TIKEUT YI KIYEOK" },
	{ 0x3ED0 , "HANGUL SYLLABLE TIKEUT YI NIEUN" },
	{ 0x3ED1 , "HANGUL SYLLABLE TIKEUT YI TIKEUT" },
	{ 0x3ED2 , "HANGUL SYLLABLE TIKEUT YI RIEUL" },
	{ 0x3ED3 , "HANGUL SYLLABLE TIKEUT YI MIEUM" },
	{ 0x3ED4 , "HANGUL SYLLABLE TIKEUT YI PIEUP" },
	{ 0x3ED5 , "HANGUL SYLLABLE TIKEUT YI SIOS" },
	{ 0x3ED6 , "HANGUL SYLLABLE TIKEUT YI IEUNG" },
	{ 0x3ED7 , "HANGUL SYLLABLE TIKEUT I CHIEUCH" },
	{ 0x3ED8 , "HANGUL SYLLABLE TIKEUT I THIEUTH" },
	{ 0x3ED9 , "HANGUL SYLLABLE TIKEUT I PHIEUPH" },
	{ 0x3EDA , "HANGUL SYLLABLE TIKEUT I HIEUH" },
	{ 0x3EDB , "HANGUL SYLLABLE SSANGTIKEUT A SSANGKIYEOK" },
	{ 0x3EDC , "HANGUL SYLLABLE SSANGTIKEUT A TIKEUT" },
	{ 0x3EDD , "HANGUL SYLLABLE SSANGTIKEUT A RIEUL-KIYEOK" },
	{ 0x3EDE , "HANGUL SYLLABLE SSANGTIKEUT A RIEUL-MIEUM" },
	{ 0x3EDF , "HANGUL SYLLABLE SSANGTIKEUT A RIEUL-PIEUP" },
	{ 0x3EE0 , "HANGUL SYLLABLE SSANGTIKEUT A RIEUL-HIEUH" },
	{ 0x3EE1 , "HANGUL SYLLABLE SSANGTIKEUT AE SSANGKIYEOK" },
	{ 0x3EE2 , "HANGUL SYLLABLE SSANGTIKEUT AE TIKEUT" },
	{ 0x3EE3 , "HANGUL SYLLABLE SSANGTIKEUT AE HIEUH" },
	{ 0x3EE4 , "HANGUL SYLLABLE SSANGTIKEUT YA" },
	{ 0x3EE5 , "HANGUL SYLLABLE SSANGTIKEUT YA KIYEOK" },
	{ 0x3EE6 , "HANGUL SYLLABLE SSANGTIKEUT YA NIEUN" },
	{ 0x3EE7 , "HANGUL SYLLABLE SSANGTIKEUT YA IEUNG" },
	{ 0x3EE8 , "HANGUL SYLLABLE SSANGTIKEUT YAE" },
	{ 0x3EE9 , "HANGUL SYLLABLE SSANGTIKEUT YAE NIEUN" },
	{ 0x3EEA , "HANGUL SYLLABLE SSANGTIKEUT EO TIKEUT" },
	{ 0x3EEB , "HANGUL SYLLABLE SSANGTIKEUT EO RIEUL-HIEUH" },
	{ 0x3EEC , "HANGUL SYLLABLE SSANGTIKEUT E TIKEUT" },
	{ 0x3EED , "HANGUL SYLLABLE SSANGTIKEUT YEO KIYEOK" },
	{ 0x3EEE , "HANGUL SYLLABLE SSANGTIKEUT YEO NIEUN" },
	{ 0x3EEF , "HANGUL SYLLABLE SSANGTIKEUT YEO TIKEUT" },
	{ 0x3EF0 , "HANGUL SYLLABLE SSANGTIKEUT YEO MIEUM" },
	{ 0x3EF1 , "HANGUL SYLLABLE SSANGTIKEUT YEO PIEUP" },
	{ 0x3EF2 , "HANGUL SYLLABLE SSANGTIKEUT YEO SIOS" },
	{ 0x3EF3 , "HANGUL SYLLABLE SSANGTIKEUT YEO IEUNG" },
	{ 0x3EF4 , "HANGUL SYLLABLE SSANGTIKEUT YE" },
	{ 0x3EF5 , "HANGUL SYLLABLE SSANGTIKEUT YE NIEUN" },
	{ 0x3EF6 , "HANGUL SYLLABLE SSANGTIKEUT YE PIEUP" },
	{ 0x3EF7 , "HANGUL SYLLABLE SSANGTIKEUT O SSANGKIYEOK" },
	{ 0x3EF8 , "HANGUL SYLLABLE SSANGTIKEUT O TIKEUT" },
	{ 0x3EF9 , "HANGUL SYLLABLE SSANGTIKEUT O RIEUL-KIYEOK" },
	{ 0x3EFA , "HANGUL SYLLABLE SSANGTIKEUT O RIEUL-MIEUM" },
	{ 0x3EFB , "HANGUL SYLLABLE SSANGTIKEUT O RIEUL-HIEUH" },
	{ 0x3EFC , "HANGUL SYLLABLE SSANGTIKEUT O MIEUM" },
	{ 0x3EFD , "HANGUL SYLLABLE SSANGTIKEUT O PIEUP" },
	{ 0x3EFE , "HANGUL SYLLABLE SSANGTIKEUT O SIOS" },
	{ 0x3EFF , "HANGUL SYLLABLE SSANGTIKEUT WA KIYEOK" },
	{ 0x3F00 , "HANGUL SYLLABLE SSANGTIKEUT WA NIEUN" },
	{ 0x3F01 , "HANGUL SYLLABLE SSANGTIKEUT WA IEUNG" },
	{ 0x3F02 , "HANGUL SYLLABLE SSANGTIKEUT WAE KIYEOK" },
	{ 0x3F03 , "HANGUL SYLLABLE SSANGTIKEUT WAE NIEUN" },
	{ 0x3F04 , "HANGUL SYLLABLE SSANGTIKEUT WAE TIKEUT" },
	{ 0x3F05 , "HANGUL SYLLABLE SSANGTIKEUT WAE RIEUL" },
	{ 0x3F06 , "HANGUL SYLLABLE SSANGTIKEUT WAE MIEUM" },
	{ 0x3F07 , "HANGUL SYLLABLE SSANGTIKEUT WAE PIEUP" },
	{ 0x3F08 , "HANGUL SYLLABLE SSANGTIKEUT WAE SIOS" },
	{ 0x3F09 , "HANGUL SYLLABLE SSANGTIKEUT WAE SSANGSIOS" },
	{ 0x3F0A , "HANGUL SYLLABLE SSANGTIKEUT OE KIYEOK" },
	{ 0x3F0B , "HANGUL SYLLABLE SSANGTIKEUT OE TIKEUT" },
	{ 0x3F0C , "HANGUL SYLLABLE SSANGTIKEUT OE RIEUL" },
	{ 0x3F0D , "HANGUL SYLLABLE SSANGTIKEUT OE MIEUM" },
	{ 0x3F0E , "HANGUL SYLLABLE SSANGTIKEUT OE PIEUP" },
	{ 0x3F0F , "HANGUL SYLLABLE SSANGTIKEUT OE SIOS" },
	{ 0x3F10 , "HANGUL SYLLABLE SSANGTIKEUT OE SSANGSIOS" },
	{ 0x3F11 , "HANGUL SYLLABLE SSANGTIKEUT OE IEUNG" },
	{ 0x3F12 , "HANGUL SYLLABLE SSANGTIKEUT YO" },
	{ 0x3F13 , "HANGUL SYLLABLE SSANGTIKEUT YO KIYEOK" },
	{ 0x3F14 , "HANGUL SYLLABLE SSANGTIKEUT U TIKEUT" },
	{ 0x3F15 , "HANGUL SYLLABLE SSANGTIKEUT U RIEUL-MIEUM" },
	{ 0x3F16 , "HANGUL SYLLABLE SSANGTIKEUT U RIEUL-PIEUP" },
	{ 0x3F17 , "HANGUL SYLLABLE SSANGTIKEUT U PIEUP" },
	{ 0x3F18 , "HANGUL SYLLABLE SSANGTIKEUT U SIOS" },
	{ 0x3F19 , "HANGUL SYLLABLE SSANGTIKEUT U THIEUTH" },
	{ 0x3F1A , "HANGUL SYLLABLE SSANGTIKEUT WEO" },
	{ 0x3F1B , "HANGUL SYLLABLE SSANGTIKEUT WEO NIEUN" },
	{ 0x3F1C , "HANGUL SYLLABLE SSANGTIKEUT WEO TIKEUT" },
	{ 0x3F1D , "HANGUL SYLLABLE SSANGTIKEUT WEO SIOS" },
	{ 0x3F1E , "HANGUL SYLLABLE SSANGTIKEUT WE NIEUN" },
	{ 0x3F1F , "HANGUL SYLLABLE SSANGTIKEUT WE TIKEUT" },
	{ 0x3F20 , "HANGUL SYLLABLE SSANGTIKEUT WE RIEUL" },
	{ 0x3F21 , "HANGUL SYLLABLE SSANGTIKEUT WE MIEUM" },
	{ 0x3F22 , "HANGUL SYLLABLE SSANGTIKEUT WE PIEUP" },
	{ 0x3F23 , "HANGUL SYLLABLE SSANGTIKEUT WE SSANGSIOS" },
	{ 0x3F24 , "HANGUL SYLLABLE SSANGTIKEUT WE IEUNG" },
	{ 0x3F25 , "HANGUL SYLLABLE SSANGTIKEUT WI TIKEUT" },
	{ 0x3F26 , "HANGUL SYLLABLE SSANGTIKEUT WI SIOS" },
	{ 0x3F27 , "HANGUL SYLLABLE SSANGTIKEUT YU" },
	{ 0x3F28 , "HANGUL SYLLABLE SSANGTIKEUT YU KIYEOK" },
	{ 0x3F29 , "HANGUL SYLLABLE SSANGTIKEUT YU RIEUL" },
	{ 0x3F2A , "HANGUL SYLLABLE SSANGTIKEUT YU MIEUM" },
	{ 0x3F2B , "HANGUL SYLLABLE SSANGTIKEUT YU IEUNG" },
	{ 0x3F2C , "HANGUL SYLLABLE SSANGTIKEUT EU RIEUL-MIEUM" },
	{ 0x3F2D , "HANGUL SYLLABLE SSANGTIKEUT EU RIEUL-PIEUP" },
	{ 0x3F2E , "HANGUL SYLLABLE SSANGTIKEUT EU RIEUL-HIEUH" },
	{ 0x3F2F , "HANGUL SYLLABLE SSANGTIKEUT EU IEUNG" },
	{ 0x3F30 , "HANGUL SYLLABLE SSANGTIKEUT YI IEUNG" },
	{ 0x3F31 , "HANGUL SYLLABLE SSANGTIKEUT I KIYEOK" },
	{ 0x3F32 , "HANGUL SYLLABLE SSANGTIKEUT I TIKEUT" },
	{ 0x3F33 , "HANGUL SYLLABLE SSANGTIKEUT I HIEUH" },
	{ 0x3F34 , "HANGUL SYLLABLE RIEUL A TIKEUT" },
	{ 0x3F35 , "HANGUL SYLLABLE RIEUL A RIEUL-SIOS" },
	{ 0x3F36 , "HANGUL SYLLABLE RIEUL A CHIEUCH" },
	{ 0x3F37 , "HANGUL SYLLABLE RIEUL A KHIEUKH" },
	{ 0x3F38 , "HANGUL SYLLABLE RIEUL A THIEUTH" },
	{ 0x3F39 , "HANGUL SYLLABLE RIEUL AE TIKEUT" },
	{ 0x3F3A , "HANGUL SYLLABLE RIEUL AE KHIEUKH" },
	{ 0x3F3B , "HANGUL SYLLABLE RIEUL AE THIEUTH" },
	{ 0x3F3C , "HANGUL SYLLABLE RIEUL AE PHIEUPH" },
	{ 0x3F3D , "HANGUL SYLLABLE RIEUL YA TIKEUT" },
	{ 0x3F3E , "HANGUL SYLLABLE RIEUL YA RIEUL" },
	{ 0x3F3F , "HANGUL SYLLABLE RIEUL YA MIEUM" },
	{ 0x3F40 , "HANGUL SYLLABLE RIEUL YA PIEUP" },
	{ 0x3F41 , "HANGUL SYLLABLE RIEUL YAE" },
	{ 0x3F42 , "HANGUL SYLLABLE RIEUL YAE SIOS" },
	{ 0x3F43 , "HANGUL SYLLABLE RIEUL EO NIEUN-HIEUH" },
	{ 0x3F44 , "HANGUL SYLLABLE RIEUL EO TIKEUT" },
	{ 0x3F45 , "HANGUL SYLLABLE RIEUL EO PIEUP-SIOS" },
	{ 0x3F46 , "HANGUL SYLLABLE RIEUL EO CIEUC" },
	{ 0x3F47 , "HANGUL SYLLABLE RIEUL EO KHIEUKH" },
	{ 0x3F48 , "HANGUL SYLLABLE RIEUL EO PHIEUPH" },
	{ 0x3F49 , "HANGUL SYLLABLE RIEUL E TIKEUT" },
	{ 0x3F4A , "HANGUL SYLLABLE RIEUL E SSANGSIOS" },
	{ 0x3F4B , "HANGUL SYLLABLE RIEUL E THIEUTH" },
	{ 0x3F4C , "HANGUL SYLLABLE RIEUL E PHIEUPH" },
	{ 0x3F4D , "HANGUL SYLLABLE RIEUL E HIEUH" },
	{ 0x3F4E , "HANGUL SYLLABLE RIEUL YEO TIKEUT" },
	{ 0x3F4F , "HANGUL SYLLABLE RIEUL YEO RIEUL-SIOS" },
	{ 0x3F50 , "HANGUL SYLLABLE RIEUL YEO KHIEUKH" },
	{ 0x3F51 , "HANGUL SYLLABLE RIEUL YEO PHIEUPH" },
	{ 0x3F52 , "HANGUL SYLLABLE RIEUL YE KIYEOK" },
	{ 0x3F53 , "HANGUL SYLLABLE RIEUL YE TIKEUT" },
	{ 0x3F54 , "HANGUL SYLLABLE RIEUL YE RIEUL" },
	{ 0x3F55 , "HANGUL SYLLABLE RIEUL YE MIEUM" },
	{ 0x3F56 , "HANGUL SYLLABLE RIEUL YE IEUNG" },
	{ 0x3F57 , "HANGUL SYLLABLE RIEUL O TIKEUT" },
	{ 0x3F58 , "HANGUL SYLLABLE RIEUL O RIEUL-SIOS" },
	{ 0x3F59 , "HANGUL SYLLABLE RIEUL O RIEUL-HIEUH" },
	{ 0x3F5A , "HANGUL SYLLABLE RIEUL O KHIEUKH" },
	{ 0x3F5B , "HANGUL SYLLABLE RIEUL O PHIEUPH" },
	{ 0x3F5C , "HANGUL SYLLABLE RIEUL O HIEUH" },
	{ 0x3F5D , "HANGUL SYLLABLE RIEUL WA KIYEOK" },
	{ 0x3F5E , "HANGUL SYLLABLE RIEUL WA RIEUL" },
	{ 0x3F5F , "HANGUL SYLLABLE RIEUL WA MIEUM" },
	{ 0x3F60 , "HANGUL SYLLABLE RIEUL WA PIEUP" },
	{ 0x3F61 , "HANGUL SYLLABLE RIEUL WA SIOS" },
	{ 0x3F62 , "HANGUL SYLLABLE RIEUL WA SSANGSIOS" },
	{ 0x3F63 , "HANGUL SYLLABLE RIEUL WAE" },
	{ 0x3F64 , "HANGUL SYLLABLE RIEUL WAE NIEUN" },
	{ 0x3F65 , "HANGUL SYLLABLE RIEUL WAE TIKEUT" },
	{ 0x3F66 , "HANGUL SYLLABLE RIEUL WAE RIEUL" },
	{ 0x3F67 , "HANGUL SYLLABLE RIEUL WAE SIOS" },
	{ 0x3F68 , "HANGUL SYLLABLE RIEUL OE KIYEOK" },
	{ 0x3F69 , "HANGUL SYLLABLE RIEUL OE NIEUN-CIEUC" },
	{ 0x3F6A , "HANGUL SYLLABLE RIEUL OE TIKEUT" },
	{ 0x3F6B , "HANGUL SYLLABLE RIEUL OE SSANGSIOS" },
	{ 0x3F6C , "HANGUL SYLLABLE RIEUL YO KIYEOK" },
	{ 0x3F6D , "HANGUL SYLLABLE RIEUL YO TIKEUT" },
	{ 0x3F6E , "HANGUL SYLLABLE RIEUL YO MIEUM" },
	{ 0x3F6F , "HANGUL SYLLABLE RIEUL U TIKEUT" },
	{ 0x3F70 , "HANGUL SYLLABLE RIEUL U RIEUL-PIEUP" },
	{ 0x3F71 , "HANGUL SYLLABLE RIEUL U KHIEUKH" },
	{ 0x3F72 , "HANGUL SYLLABLE RIEUL U PHIEUPH" },
	{ 0x3F73 , "HANGUL SYLLABLE RIEUL WEO KIYEOK" },
	{ 0x3F74 , "HANGUL SYLLABLE RIEUL WEO NIEUN" },
	{ 0x3F75 , "HANGUL SYLLABLE RIEUL WEO TIKEUT" },
	{ 0x3F76 , "HANGUL SYLLABLE RIEUL WEO RIEUL" },
	{ 0x3F77 , "HANGUL SYLLABLE RIEUL WEO MIEUM" },
	{ 0x3F78 , "HANGUL SYLLABLE RIEUL WEO PIEUP" },
	{ 0x3F79 , "HANGUL SYLLABLE RIEUL WEO SIOS" },
	{ 0x3F7A , "HANGUL SYLLABLE RIEUL WEO IEUNG" },
	{ 0x3F7B , "HANGUL SYLLABLE RIEUL WE KIYEOK" },
	{ 0x3F7C , "HANGUL SYLLABLE RIEUL WE NIEUN" },
	{ 0x3F7D , "HANGUL SYLLABLE RIEUL WE RIEUL" },
	{ 0x3F7E , "HANGUL SYLLABLE RIEUL WE MIEUM" },
	{ 0x3F7F , "HANGUL SYLLABLE RIEUL WE PIEUP" },
	{ 0x3F80 , "HANGUL SYLLABLE RIEUL WE SIOS" },
	{ 0x3F81 , "HANGUL SYLLABLE RIEUL WE SSANGSIOS" },
	{ 0x3F82 , "HANGUL SYLLABLE RIEUL WE IEUNG" },
	{ 0x3F83 , "HANGUL SYLLABLE RIEUL WI TIKEUT" },
	{ 0x3F84 , "HANGUL SYLLABLE RIEUL WI PIEUP" },
	{ 0x3F85 , "HANGUL SYLLABLE RIEUL WI SSANGSIOS" },
	{ 0x3F86 , "HANGUL SYLLABLE RIEUL YU TIKEUT" },
	{ 0x3F87 , "HANGUL SYLLABLE RIEUL EU TIKEUT" },
	{ 0x3F88 , "HANGUL SYLLABLE RIEUL EU RIEUL-KIYEOK" },
	{ 0x3F89 , "HANGUL SYLLABLE RIEUL EU RIEUL-SIOS" },
	{ 0x3F8A , "HANGUL SYLLABLE RIEUL EU CHIEUCH" },
	{ 0x3F8B , "HANGUL SYLLABLE RIEUL EU HIEUH" },
	{ 0x3F8C , "HANGUL SYLLABLE RIEUL YI" },
	{ 0x3F8D , "HANGUL SYLLABLE RIEUL YI KIYEOK" },
	{ 0x3F8E , "HANGUL SYLLABLE RIEUL YI NIEUN" },
	{ 0x3F8F , "HANGUL SYLLABLE RIEUL YI RIEUL" },
	{ 0x3F90 , "HANGUL SYLLABLE RIEUL YI RIEUL-PHIEUPH" },
	{ 0x3F91 , "HANGUL SYLLABLE RIEUL YI MIEUM" },
	{ 0x3F92 , "HANGUL SYLLABLE RIEUL YI PIEUP" },
	{ 0x3F93 , "HANGUL SYLLABLE RIEUL YI SIOS" },
	{ 0x3F94 , "HANGUL SYLLABLE RIEUL YI IEUNG" },
	{ 0x3F95 , "HANGUL SYLLABLE RIEUL I TIKEUT" },
	{ 0x3F96 , "HANGUL SYLLABLE RIEUL I PIEUP-SIOS" },
	{ 0x3F97 , "HANGUL SYLLABLE RIEUL I SSANGSIOS" },
	{ 0x3F98 , "HANGUL SYLLABLE RIEUL I KHIEUKH" },
	{ 0x3F99 , "HANGUL SYLLABLE RIEUL I PHIEUPH" },
	{ 0x3F9A , "HANGUL SYLLABLE MIEUM A SSANGKIYEOK" },
	{ 0x3F9B , "HANGUL SYLLABLE MIEUM A RIEUL-SIOS" },
	{ 0x3F9C , "HANGUL SYLLABLE MIEUM A SSANGSIOS" },
	{ 0x3F9D , "HANGUL SYLLABLE MIEUM A CHIEUCH" },
	{ 0x3F9E , "HANGUL SYLLABLE MIEUM A KHIEUKH" },
	{ 0x3F9F , "HANGUL SYLLABLE MIEUM A PHIEUPH" },
	{ 0x3FA0 , "HANGUL SYLLABLE MIEUM AE TIKEUT" },
	{ 0x3FA1 , "HANGUL SYLLABLE MIEUM AE RIEUL-KIYEOK" },
	{ 0x3FA2 , "HANGUL SYLLABLE MIEUM AE CHIEUCH" },
	{ 0x3FA3 , "HANGUL SYLLABLE MIEUM AE THIEUTH" },
	{ 0x3FA4 , "HANGUL SYLLABLE MIEUM AE PHIEUPH" },
	{ 0x3FA5 , "HANGUL SYLLABLE MIEUM YA NIEUN" },
	{ 0x3FA6 , "HANGUL SYLLABLE MIEUM YA MIEUM" },
	{ 0x3FA7 , "HANGUL SYLLABLE MIEUM YA PIEUP" },
	{ 0x3FA8 , "HANGUL SYLLABLE MIEUM YA SIOS" },
	{ 0x3FA9 , "HANGUL SYLLABLE MIEUM YA KHIEUKH" },
	{ 0x3FAA , "HANGUL SYLLABLE MIEUM YAE" },
	{ 0x3FAB , "HANGUL SYLLABLE MIEUM EO KIYEOK-SIOS" },
	{ 0x3FAC , "HANGUL SYLLABLE MIEUM EO TIKEUT" },
	{ 0x3FAD , "HANGUL SYLLABLE MIEUM EO RIEUL-KIYEOK" },
	{ 0x3FAE , "HANGUL SYLLABLE MIEUM EO SSANGSIOS" },
	{ 0x3FAF , "HANGUL SYLLABLE MIEUM EO CHIEUCH" },
	{ 0x3FB0 , "HANGUL SYLLABLE MIEUM E TIKEUT" },
	{ 0x3FB1 , "HANGUL SYLLABLE MIEUM E RIEUL-SIOS" },
	{ 0x3FB2 , "HANGUL SYLLABLE MIEUM E CIEUC" },
	{ 0x3FB3 , "HANGUL SYLLABLE MIEUM E CHIEUCH" },
	{ 0x3FB4 , "HANGUL SYLLABLE MIEUM E THIEUTH" },
	{ 0x3FB5 , "HANGUL SYLLABLE MIEUM YEO TIKEUT" },
	{ 0x3FB6 , "HANGUL SYLLABLE MIEUM YEO RIEUL-MIEUM" },
	{ 0x3FB7 , "HANGUL SYLLABLE MIEUM YEO MIEUM" },
	{ 0x3FB8 , "HANGUL SYLLABLE MIEUM YEO PIEUP" },
	{ 0x3FB9 , "HANGUL SYLLABLE MIEUM YEO THIEUTH" },
	{ 0x3FBA , "HANGUL SYLLABLE MIEUM YE KIYEOK" },
	{ 0x3FBB , "HANGUL SYLLABLE MIEUM YE NIEUN" },
	{ 0x3FBC , "HANGUL SYLLABLE MIEUM YE TIKEUT" },
	{ 0x3FBD , "HANGUL SYLLABLE MIEUM YE RIEUL" },
	{ 0x3FBE , "HANGUL SYLLABLE MIEUM YE MIEUM" },
	{ 0x3FBF , "HANGUL SYLLABLE MIEUM YE PIEUP" },
	{ 0x3FC0 , "HANGUL SYLLABLE MIEUM YE SIOS" },
	{ 0x3FC1 , "HANGUL SYLLABLE MIEUM YE IEUNG" },
	{ 0x3FC2 , "HANGUL SYLLABLE MIEUM YE CHIEUCH" },
	{ 0x3FC3 , "HANGUL SYLLABLE MIEUM YE THIEUTH" },
	{ 0x3FC4 , "HANGUL SYLLABLE MIEUM YE PHIEUPH" },
	{ 0x3FC5 , "HANGUL SYLLABLE MIEUM O TIKEUT" },
	{ 0x3FC6 , "HANGUL SYLLABLE MIEUM O RIEUL-KIYEOK" },
	{ 0x3FC7 , "HANGUL SYLLABLE MIEUM O RIEUL-SIOS" },
	{ 0x3FC8 , "HANGUL SYLLABLE MIEUM O CHIEUCH" },
	{ 0x3FC9 , "HANGUL SYLLABLE MIEUM O HIEUH" },
	{ 0x3FCA , "HANGUL SYLLABLE MIEUM WA KIYEOK" },
	{ 0x3FCB , "HANGUL SYLLABLE MIEUM WA TIKEUT" },
	{ 0x3FCC , "HANGUL SYLLABLE MIEUM WA RIEUL" },
	{ 0x3FCD , "HANGUL SYLLABLE MIEUM WA MIEUM" },
	{ 0x3FCE , "HANGUL SYLLABLE MIEUM WA PIEUP" },
	{ 0x3FCF , "HANGUL SYLLABLE MIEUM WA SIOS" },
	{ 0x3FD0 , "HANGUL SYLLABLE MIEUM WAE" },
	{ 0x3FD1 , "HANGUL SYLLABLE MIEUM WAE NIEUN" },
	{ 0x3FD2 , "HANGUL SYLLABLE MIEUM WAE RIEUL" },
	{ 0x3FD3 , "HANGUL SYLLABLE MIEUM WAE PIEUP" },
	{ 0x3FD4 , "HANGUL SYLLABLE MIEUM WAE SSANGSIOS" },
	{ 0x3FD5 , "HANGUL SYLLABLE MIEUM OE KIYEOK" },
	{ 0x3FD6 , "HANGUL SYLLABLE MIEUM OE TIKEUT" },
	{ 0x3FD7 , "HANGUL SYLLABLE MIEUM OE MIEUM" },
	{ 0x3FD8 , "HANGUL SYLLABLE MIEUM YO KIYEOK" },
	{ 0x3FD9 , "HANGUL SYLLABLE MIEUM YO TIKEUT" },
	{ 0x3FDA , "HANGUL SYLLABLE MIEUM YO MIEUM" },
	{ 0x3FDB , "HANGUL SYLLABLE MIEUM YO IEUNG" },
	{ 0x3FDC , "HANGUL SYLLABLE MIEUM U NIEUN-HIEUH" },
	{ 0x3FDD , "HANGUL SYLLABLE MIEUM U RIEUL-SIOS" },
	{ 0x3FDE , "HANGUL SYLLABLE MIEUM U CHIEUCH" },
	{ 0x3FDF , "HANGUL SYLLABLE MIEUM WEO KIYEOK" },
	{ 0x3FE0 , "HANGUL SYLLABLE MIEUM WEO TIKEUT" },
	{ 0x3FE1 , "HANGUL SYLLABLE MIEUM WEO MIEUM" },
	{ 0x3FE2 , "HANGUL SYLLABLE MIEUM WEO SSANGSIOS" },
	{ 0x3FE3 , "HANGUL SYLLABLE MIEUM WEO IEUNG" },
	{ 0x3FE4 , "HANGUL SYLLABLE MIEUM WE KIYEOK" },
	{ 0x3FE5 , "HANGUL SYLLABLE MIEUM WE NIEUN" },
	{ 0x3FE6 , "HANGUL SYLLABLE MIEUM WE RIEUL" },
	{ 0x3FE7 , "HANGUL SYLLABLE MIEUM WE MIEUM" },
	{ 0x3FE8 , "HANGUL SYLLABLE MIEUM WE PIEUP" },
	{ 0x3FE9 , "HANGUL SYLLABLE MIEUM WE SIOS" },
	{ 0x3FEA , "HANGUL SYLLABLE MIEUM WE SSANGSIOS" },
	{ 0x3FEB , "HANGUL SYLLABLE MIEUM WE IEUNG" },
	{ 0x3FEC , "HANGUL SYLLABLE MIEUM WI KIYEOK" },
	{ 0x3FED , "HANGUL SYLLABLE MIEUM WI SSANGKIYEOK" },
	{ 0x3FEE , "HANGUL SYLLABLE MIEUM WI MIEUM" },
	{ 0x3FEF , "HANGUL SYLLABLE MIEUM WI PIEUP" },
	{ 0x3FF0 , "HANGUL SYLLABLE MIEUM WI SIOS" },
	{ 0x3FF1 , "HANGUL SYLLABLE MIEUM WI SSANGSIOS" },
	{ 0x3FF2 , "HANGUL SYLLABLE MIEUM WI IEUNG" },
	{ 0x3FF3 , "HANGUL SYLLABLE MIEUM YU KIYEOK" },
	{ 0x3FF4 , "HANGUL SYLLABLE MIEUM YU TIKEUT" },
	{ 0x3FF5 , "HANGUL SYLLABLE MIEUM YU PIEUP" },
	{ 0x3FF6 , "HANGUL SYLLABLE MIEUM YU IEUNG" },
	{ 0x3FF7 , "HANGUL SYLLABLE MIEUM EU KIYEOK" },
	{ 0x3FF8 , "HANGUL SYLLABLE MIEUM EU TIKEUT" },
	{ 0x3FF9 , "HANGUL SYLLABLE MIEUM EU RIEUL-KIYEOK" },
	{ 0x3FFA , "HANGUL SYLLABLE MIEUM EU RIEUL-SIOS" },
	{ 0x3FFB , "HANGUL SYLLABLE MIEUM EU RIEUL-HIEUH" },
	{ 0x3FFC , "HANGUL SYLLABLE MIEUM EU PIEUP" },
	{ 0x3FFD , "HANGUL SYLLABLE MIEUM EU IEUNG" },
	{ 0x3FFE , "HANGUL SYLLABLE MIEUM EU THIEUTH" },
	{ 0x3FFF , "HANGUL SYLLABLE MIEUM YI" },
	{ 0x4000 , "HANGUL SYLLABLE MIEUM YI KIYEOK" },
	{ 0x4001 , "HANGUL SYLLABLE MIEUM YI NIEUN" },
	{ 0x4002 , "HANGUL SYLLABLE MIEUM YI RIEUL" },
	{ 0x4003 , "HANGUL SYLLABLE MIEUM YI MIEUM" },
	{ 0x4004 , "HANGUL SYLLABLE MIEUM YI PIEUP" },
	{ 0x4005 , "HANGUL SYLLABLE MIEUM YI SIOS" },
	{ 0x4006 , "HANGUL SYLLABLE MIEUM YI IEUNG" },
	{ 0x4007 , "HANGUL SYLLABLE MIEUM YI CIEUC" },
	{ 0x4008 , "HANGUL SYLLABLE MIEUM I RIEUL-KIYEOK" },
	{ 0x4009 , "HANGUL SYLLABLE MIEUM I RIEUL-SIOS" },
	{ 0x400A , "HANGUL SYLLABLE MIEUM I CIEUC" },
	{ 0x400B , "HANGUL SYLLABLE MIEUM I KHIEUKH" },
	{ 0x400C , "HANGUL SYLLABLE PIEUP A RIEUL-SIOS" },
	{ 0x400D , "HANGUL SYLLABLE PIEUP A RIEUL-HIEUH" },
	{ 0x400E , "HANGUL SYLLABLE PIEUP A SSANGSIOS" },
	{ 0x400F , "HANGUL SYLLABLE PIEUP A CHIEUCH" },
	{ 0x4010 , "HANGUL SYLLABLE PIEUP A KHIEUKH" },
	{ 0x4011 , "HANGUL SYLLABLE PIEUP A HIEUH" },
	{ 0x4012 , "HANGUL SYLLABLE PIEUP AE TIKEUT" },
	{ 0x4013 , "HANGUL SYLLABLE PIEUP AE CIEUC" },
	{ 0x4014 , "HANGUL SYLLABLE PIEUP YA RIEUL" },
	{ 0x4015 , "HANGUL SYLLABLE PIEUP YA MIEUM" },
	{ 0x4016 , "HANGUL SYLLABLE PIEUP YA SIOS" },
	{ 0x4017 , "HANGUL SYLLABLE PIEUP YA IEUNG" },
	{ 0x4018 , "HANGUL SYLLABLE PIEUP YA HIEUH" },
	{ 0x4019 , "HANGUL SYLLABLE PIEUP YAE" },
	{ 0x401A , "HANGUL SYLLABLE PIEUP YAE RIEUL" },
	{ 0x401B , "HANGUL SYLLABLE PIEUP YAE RIEUL-HIEUH" },
	{ 0x401C , "HANGUL SYLLABLE PIEUP YAE IEUNG" },
	{ 0x401D , "HANGUL SYLLABLE PIEUP EO SSANGKIYEOK" },
	{ 0x401E , "HANGUL SYLLABLE PIEUP EO RIEUL-KIYEOK" },
	{ 0x401F , "HANGUL SYLLABLE PIEUP EO RIEUL-SIOS" },
	{ 0x4020 , "HANGUL SYLLABLE PIEUP EO SSANGSIOS" },
	{ 0x4021 , "HANGUL SYLLABLE PIEUP EO KHIEUKH" },
	{ 0x4022 , "HANGUL SYLLABLE PIEUP EO THIEUTH" },
	{ 0x4023 , "HANGUL SYLLABLE PIEUP E SSANGKIYEOK" },
	{ 0x4024 , "HANGUL SYLLABLE PIEUP E KHIEUKH" },
	{ 0x4025 , "HANGUL SYLLABLE PIEUP E THIEUTH" },
	{ 0x4026 , "HANGUL SYLLABLE PIEUP YEO TIKEUT" },
	{ 0x4027 , "HANGUL SYLLABLE PIEUP YEO RIEUL-SIOS" },
	{ 0x4028 , "HANGUL SYLLABLE PIEUP YEO MIEUM" },
	{ 0x4029 , "HANGUL SYLLABLE PIEUP YEO CIEUC" },
	{ 0x402A , "HANGUL SYLLABLE PIEUP YEO CHIEUCH" },
	{ 0x402B , "HANGUL SYLLABLE PIEUP YEO KHIEUKH" },
	{ 0x402C , "HANGUL SYLLABLE PIEUP YE KIYEOK" },
	{ 0x402D , "HANGUL SYLLABLE PIEUP YE RIEUL" },
	{ 0x402E , "HANGUL SYLLABLE PIEUP YE MIEUM" },
	{ 0x402F , "HANGUL SYLLABLE PIEUP YE PIEUP" },
	{ 0x4030 , "HANGUL SYLLABLE PIEUP YE SIOS" },
	{ 0x4031 , "HANGUL SYLLABLE PIEUP YE IEUNG" },
	{ 0x4032 , "HANGUL SYLLABLE PIEUP O KIYEOK-SIOS" },
	{ 0x4033 , "HANGUL SYLLABLE PIEUP O TIKEUT" },
	{ 0x4034 , "HANGUL SYLLABLE PIEUP O RIEUL-KIYEOK" },
	{ 0x4035 , "HANGUL SYLLABLE PIEUP O RIEUL-PIEUP" },
	{ 0x4036 , "HANGUL SYLLABLE PIEUP O CIEUC" },
	{ 0x4037 , "HANGUL SYLLABLE PIEUP O CHIEUCH" },
	{ 0x4038 , "HANGUL SYLLABLE PIEUP O THIEUTH" },
	{ 0x4039 , "HANGUL SYLLABLE PIEUP O HIEUH" },
	{ 0x403A , "HANGUL SYLLABLE PIEUP WA KIYEOK" },
	{ 0x403B , "HANGUL SYLLABLE PIEUP WA TIKEUT" },
	{ 0x403C , "HANGUL SYLLABLE PIEUP WA RIEUL" },
	{ 0x403D , "HANGUL SYLLABLE PIEUP WA MIEUM" },
	{ 0x403E , "HANGUL SYLLABLE PIEUP WA PIEUP" },
	{ 0x403F , "HANGUL SYLLABLE PIEUP WA SIOS" },
	{ 0x4040 , "HANGUL SYLLABLE PIEUP WA IEUNG" },
	{ 0x4041 , "HANGUL SYLLABLE PIEUP WAE NIEUN" },
	{ 0x4042 , "HANGUL SYLLABLE PIEUP WAE TIKEUT" },
	{ 0x4043 , "HANGUL SYLLABLE PIEUP WAE RIEUL" },
	{ 0x4044 , "HANGUL SYLLABLE PIEUP OE SSANGKIYEOK" },
	{ 0x4045 , "HANGUL SYLLABLE PIEUP OE TIKEUT" },
	{ 0x4046 , "HANGUL SYLLABLE PIEUP OE SIOS" },
	{ 0x4047 , "HANGUL SYLLABLE PIEUP OE SSANGSIOS" },
	{ 0x4048 , "HANGUL SYLLABLE PIEUP OE IEUNG" },
	{ 0x4049 , "HANGUL SYLLABLE PIEUP YO KIYEOK" },
	{ 0x404A , "HANGUL SYLLABLE PIEUP YO RIEUL" },
	{ 0x404B , "HANGUL SYLLABLE PIEUP YO MIEUM" },
	{ 0x404C , "HANGUL SYLLABLE PIEUP YO PIEUP" },
	{ 0x404D , "HANGUL SYLLABLE PIEUP YO SIOS" },
	{ 0x404E , "HANGUL SYLLABLE PIEUP YO IEUNG" },
	{ 0x404F , "HANGUL SYLLABLE PIEUP U SSANGKIYEOK" },
	{ 0x4050 , "HANGUL SYLLABLE PIEUP U RIEUL-PIEUP" },
	{ 0x4051 , "HANGUL SYLLABLE PIEUP U RIEUL-PHIEUPH" },
	{ 0x4052 , "HANGUL SYLLABLE PIEUP U CIEUC" },
	{ 0x4053 , "HANGUL SYLLABLE PIEUP U CHIEUCH" },
	{ 0x4054 , "HANGUL SYLLABLE PIEUP U KHIEUKH" },
	{ 0x4055 , "HANGUL SYLLABLE PIEUP U HIEUH" },
	{ 0x4056 , "HANGUL SYLLABLE PIEUP WEO KIYEOK" },
	{ 0x4057 , "HANGUL SYLLABLE PIEUP WEO NIEUN" },
	{ 0x4058 , "HANGUL SYLLABLE PIEUP WEO TIKEUT" },
	{ 0x4059 , "HANGUL SYLLABLE PIEUP WEO MIEUM" },
	{ 0x405A , "HANGUL SYLLABLE PIEUP WEO PIEUP" },
	{ 0x405B , "HANGUL SYLLABLE PIEUP WEO SIOS" },
	{ 0x405C , "HANGUL SYLLABLE PIEUP WEO IEUNG" },
	{ 0x405D , "HANGUL SYLLABLE PIEUP WEO KHIEUKH" },
	{ 0x405E , "HANGUL SYLLABLE PIEUP WE KIYEOK" },
	{ 0x405F , "HANGUL SYLLABLE PIEUP WE NIEUN" },
	{ 0x4060 , "HANGUL SYLLABLE PIEUP WE TIKEUT" },
	{ 0x4061 , "HANGUL SYLLABLE PIEUP WE RIEUL" },
	{ 0x4062 , "HANGUL SYLLABLE PIEUP WE MIEUM" },
	{ 0x4063 , "HANGUL SYLLABLE PIEUP WE PIEUP" },
	{ 0x4064 , "HANGUL SYLLABLE PIEUP WE SIOS" },
	{ 0x4065 , "HANGUL SYLLABLE PIEUP WE SSANGSIOS" },
	{ 0x4066 , "HANGUL SYLLABLE PIEUP WE IEUNG" },
	{ 0x4067 , "HANGUL SYLLABLE PIEUP WI TIKEUT" },
	{ 0x4068 , "HANGUL SYLLABLE PIEUP WI MIEUM" },
	{ 0x4069 , "HANGUL SYLLABLE PIEUP WI PIEUP" },
	{ 0x406A , "HANGUL SYLLABLE PIEUP WI SIOS" },
	{ 0x406B , "HANGUL SYLLABLE PIEUP YU KIYEOK" },
	{ 0x406C , "HANGUL SYLLABLE PIEUP YU TIKEUT" },
	{ 0x406D , "HANGUL SYLLABLE PIEUP YU PIEUP" },
	{ 0x406E , "HANGUL SYLLABLE PIEUP EU TIKEUT" },
	{ 0x406F , "HANGUL SYLLABLE PIEUP EU RIEUL-KIYEOK" },
	{ 0x4070 , "HANGUL SYLLABLE PIEUP EU RIEUL-PIEUP" },
	{ 0x4071 , "HANGUL SYLLABLE PIEUP EU RIEUL-SIOS" },
	{ 0x4072 , "HANGUL SYLLABLE PIEUP EU RIEUL-HIEUH" },
	{ 0x4073 , "HANGUL SYLLABLE PIEUP EU IEUNG" },
	{ 0x4074 , "HANGUL SYLLABLE PIEUP EU CHIEUCH" },
	{ 0x4075 , "HANGUL SYLLABLE PIEUP EU THIEUTH" },
	{ 0x4076 , "HANGUL SYLLABLE PIEUP EU HIEUH" },
	{ 0x4077 , "HANGUL SYLLABLE PIEUP YI" },
	{ 0x4078 , "HANGUL SYLLABLE PIEUP YI KIYEOK" },
	{ 0x4079 , "HANGUL SYLLABLE PIEUP YI NIEUN" },
	{ 0x407A , "HANGUL SYLLABLE PIEUP YI RIEUL" },
	{ 0x407B , "HANGUL SYLLABLE PIEUP YI MIEUM" },
	{ 0x407C , "HANGUL SYLLABLE PIEUP YI PIEUP" },
	{ 0x407D , "HANGUL SYLLABLE PIEUP YI SIOS" },
	{ 0x407E , "HANGUL SYLLABLE PIEUP YI IEUNG" },
	{ 0x407F , "HANGUL SYLLABLE PIEUP I TIKEUT" },
	{ 0x4080 , "HANGUL SYLLABLE PIEUP I SSANGSIOS" },
	{ 0x4081 , "HANGUL SYLLABLE PIEUP I KHIEUKH" },
	{ 0x4082 , "HANGUL SYLLABLE PIEUP I THIEUTH" },
	{ 0x4083 , "HANGUL SYLLABLE PIEUP I HIEUH" },
	{ 0x4084 , "HANGUL SYLLABLE SSANGPIEUP A SSANGKIYEOK" },
	{ 0x4085 , "HANGUL SYLLABLE SSANGPIEUP A TIKEUT" },
	{ 0x4086 , "HANGUL SYLLABLE SSANGPIEUP A RIEUL-KIYEOK" },
	{ 0x4087 , "HANGUL SYLLABLE SSANGPIEUP A CIEUC" },
	{ 0x4088 , "HANGUL SYLLABLE SSANGPIEUP A THIEUTH" },
	{ 0x4089 , "HANGUL SYLLABLE SSANGPIEUP AE SSANGKIYEOK" },
	{ 0x408A , "HANGUL SYLLABLE SSANGPIEUP AE TIKEUT" },
	{ 0x408B , "HANGUL SYLLABLE SSANGPIEUP YA NIEUN" },
	{ 0x408C , "HANGUL SYLLABLE SSANGPIEUP YAE" },
	{ 0x408D , "HANGUL SYLLABLE SSANGPIEUP YAE NIEUN" },
	{ 0x408E , "HANGUL SYLLABLE SSANGPIEUP EO PIEUP" },
	{ 0x408F , "HANGUL SYLLABLE SSANGPIEUP EO CIEUC" },
	{ 0x4090 , "HANGUL SYLLABLE SSANGPIEUP E KIYEOK" },
	{ 0x4091 , "HANGUL SYLLABLE SSANGPIEUP E NIEUN" },
	{ 0x4092 , "HANGUL SYLLABLE SSANGPIEUP E TIKEUT" },
	{ 0x4093 , "HANGUL SYLLABLE SSANGPIEUP E RIEUL" },
	{ 0x4094 , "HANGUL SYLLABLE SSANGPIEUP E MIEUM" },
	{ 0x4095 , "HANGUL SYLLABLE SSANGPIEUP E PIEUP" },
	{ 0x4096 , "HANGUL SYLLABLE SSANGPIEUP E SIOS" },
	{ 0x4097 , "HANGUL SYLLABLE SSANGPIEUP YEO NIEUN" },
	{ 0x4098 , "HANGUL SYLLABLE SSANGPIEUP YEO TIKEUT" },
	{ 0x4099 , "HANGUL SYLLABLE SSANGPIEUP YEO RIEUL" },
	{ 0x409A , "HANGUL SYLLABLE SSANGPIEUP YEO CHIEUCH" },
	{ 0x409B , "HANGUL SYLLABLE SSANGPIEUP YEO THIEUTH" },
	{ 0x409C , "HANGUL SYLLABLE SSANGPIEUP YE" },
	{ 0x409D , "HANGUL SYLLABLE SSANGPIEUP YE SIOS" },
	{ 0x409E , "HANGUL SYLLABLE SSANGPIEUP YE SSANGSIOS" },
	{ 0x409F , "HANGUL SYLLABLE SSANGPIEUP YE IEUNG" },
	{ 0x40A0 , "HANGUL SYLLABLE SSANGPIEUP O TIKEUT" },
	{ 0x40A1 , "HANGUL SYLLABLE SSANGPIEUP O SIOS" },
	{ 0x40A2 , "HANGUL SYLLABLE SSANGPIEUP O HIEUH" },
	{ 0x40A3 , "HANGUL SYLLABLE SSANGPIEUP WA" },
	{ 0x40A4 , "HANGUL SYLLABLE SSANGPIEUP WA NIEUN" },
	{ 0x40A5 , "HANGUL SYLLABLE SSANGPIEUP WA TIKEUT" },
	{ 0x40A6 , "HANGUL SYLLABLE SSANGPIEUP WAE" },
	{ 0x40A7 , "HANGUL SYLLABLE SSANGPIEUP WAE TIKEUT" },
	{ 0x40A8 , "HANGUL SYLLABLE SSANGPIEUP OE MIEUM" },
	{ 0x40A9 , "HANGUL SYLLABLE SSANGPIEUP OE PIEUP" },
	{ 0x40AA , "HANGUL SYLLABLE SSANGPIEUP OE KHIEUKH" },
	{ 0x40AB , "HANGUL SYLLABLE SSANGPIEUP YO MIEUM" },
	{ 0x40AC , "HANGUL SYLLABLE SSANGPIEUP YO PIEUP" },
	{ 0x40AD , "HANGUL SYLLABLE SSANGPIEUP U TIKEUT" },
	{ 0x40AE , "HANGUL SYLLABLE SSANGPIEUP U RIEUL-KIYEOK" },
	{ 0x40AF , "HANGUL SYLLABLE SSANGPIEUP U PIEUP" },
	{ 0x40B0 , "HANGUL SYLLABLE SSANGPIEUP WEO" },
	{ 0x40B1 , "HANGUL SYLLABLE SSANGPIEUP WEO TIKEUT" },
	{ 0x40B2 , "HANGUL SYLLABLE SSANGPIEUP WE" },
	{ 0x40B3 , "HANGUL SYLLABLE SSANGPIEUP WI" },
	{ 0x40B4 , "HANGUL SYLLABLE SSANGPIEUP WI KIYEOK" },
	{ 0x40B5 , "HANGUL SYLLABLE SSANGPIEUP YU TIKEUT" },
	{ 0x40B6 , "HANGUL SYLLABLE SSANGPIEUP YU RIEUL" },
	{ 0x40B7 , "HANGUL SYLLABLE SSANGPIEUP EU KIYEOK" },
	{ 0x40B8 , "HANGUL SYLLABLE SSANGPIEUP EU TIKEUT" },
	{ 0x40B9 , "HANGUL SYLLABLE SSANGPIEUP EU SIOS" },
	{ 0x40BA , "HANGUL SYLLABLE SSANGPIEUP YI" },
	{ 0x40BB , "HANGUL SYLLABLE SSANGPIEUP I TIKEUT" },
	{ 0x40BC , "HANGUL SYLLABLE SSANGPIEUP I SIOS" },
	{ 0x40BD , "HANGUL SYLLABLE SSANGPIEUP I CIEUC" },
	{ 0x40BE , "HANGUL SYLLABLE SSANGPIEUP I CHIEUCH" },
	{ 0x40BF , "HANGUL SYLLABLE SSANGPIEUP I HIEUH" },
	{ 0x40C0 , "HANGUL SYLLABLE SIOS A SSANGKIYEOK" },
	{ 0x40C1 , "HANGUL SYLLABLE SIOS A RIEUL-PIEUP" },
	{ 0x40C2 , "HANGUL SYLLABLE SIOS A RIEUL-SIOS" },
	{ 0x40C3 , "HANGUL SYLLABLE SIOS A RIEUL-PHIEUPH" },
	{ 0x40C4 , "HANGUL SYLLABLE SIOS A RIEUL-HIEUH" },
	{ 0x40C5 , "HANGUL SYLLABLE SIOS A CIEUC" },
	{ 0x40C6 , "HANGUL SYLLABLE SIOS A CHIEUCH" },
	{ 0x40C7 , "HANGUL SYLLABLE SIOS A KHIEUKH" },
	{ 0x40C8 , "HANGUL SYLLABLE SIOS A PHIEUPH" },
	{ 0x40C9 , "HANGUL SYLLABLE SIOS A HIEUH" },
	{ 0x40CA , "HANGUL SYLLABLE SIOS AE KIYEOK-SIOS" },
	{ 0x40CB , "HANGUL SYLLABLE SIOS AE TIKEUT" },
	{ 0x40CC , "HANGUL SYLLABLE SIOS AE HIEUH" },
	{ 0x40CD , "HANGUL SYLLABLE SIOS YA TIKEUT" },
	{ 0x40CE , "HANGUL SYLLABLE SIOS YA SSANGSIOS" },
	{ 0x40CF , "HANGUL SYLLABLE SIOS YA CHIEUCH" },
	{ 0x40D0 , "HANGUL SYLLABLE SIOS YA PHIEUPH" },
	{ 0x40D1 , "HANGUL SYLLABLE SIOS YA HIEUH" },
	{ 0x40D2 , "HANGUL SYLLABLE SIOS EO RIEUL-KIYEOK" },
	{ 0x40D3 , "HANGUL SYLLABLE SIOS EO RIEUL-SIOS" },
	{ 0x40D4 , "HANGUL SYLLABLE SIOS EO RIEUL-PHIEUPH" },
	{ 0x40D5 , "HANGUL SYLLABLE SIOS EO KHIEUKH" },
	{ 0x40D6 , "HANGUL SYLLABLE SIOS E TIKEUT" },
	{ 0x40D7 , "HANGUL SYLLABLE SIOS E CIEUC" },
	{ 0x40D8 , "HANGUL SYLLABLE SIOS E THIEUTH" },
	{ 0x40D9 , "HANGUL SYLLABLE SIOS E PHIEUPH" },
	{ 0x40DA , "HANGUL SYLLABLE SIOS YEO KIYEOK-SIOS" },
	{ 0x40DB , "HANGUL SYLLABLE SIOS YEO TIKEUT" },
	{ 0x40DC , "HANGUL SYLLABLE SIOS YEO RIEUL-PIEUP" },
	{ 0x40DD , "HANGUL SYLLABLE SIOS YEO RIEUL-SIOS" },
	{ 0x40DE , "HANGUL SYLLABLE SIOS YE KIYEOK" },
	{ 0x40DF , "HANGUL SYLLABLE SIOS YE KIYEOK-SIOS" },
	{ 0x40E0 , "HANGUL SYLLABLE SIOS YE MIEUM" },
	{ 0x40E1 , "HANGUL SYLLABLE SIOS YE PIEUP" },
	{ 0x40E2 , "HANGUL SYLLABLE SIOS YE SIOS" },
	{ 0x40E3 , "HANGUL SYLLABLE SIOS YE SSANGSIOS" },
	{ 0x40E4 , "HANGUL SYLLABLE SIOS O TIKEUT" },
	{ 0x40E5 , "HANGUL SYLLABLE SIOS O RIEUL-KIYEOK" },
	{ 0x40E6 , "HANGUL SYLLABLE SIOS O RIEUL-PIEUP" },
	{ 0x40E7 , "HANGUL SYLLABLE SIOS O RIEUL-SIOS" },
	{ 0x40E8 , "HANGUL SYLLABLE SIOS O CHIEUCH" },
	{ 0x40E9 , "HANGUL SYLLABLE SIOS O PHIEUPH" },
	{ 0x40EA , "HANGUL SYLLABLE SIOS WA MIEUM" },
	{ 0x40EB , "HANGUL SYLLABLE SIOS WA PIEUP" },
	{ 0x40EC , "HANGUL SYLLABLE SIOS WA SIOS" },
	{ 0x40ED , "HANGUL SYLLABLE SIOS WA SSANGSIOS" },
	{ 0x40EE , "HANGUL SYLLABLE SIOS WAE KIYEOK" },
	{ 0x40EF , "HANGUL SYLLABLE SIOS WAE TIKEUT" },
	{ 0x40F0 , "HANGUL SYLLABLE SIOS WAE PIEUP" },
	{ 0x40F1 , "HANGUL SYLLABLE SIOS WAE IEUNG" },
	{ 0x40F2 , "HANGUL SYLLABLE SIOS OE KIYEOK" },
	{ 0x40F3 , "HANGUL SYLLABLE SIOS OE TIKEUT" },
	{ 0x40F4 , "HANGUL SYLLABLE SIOS OE SSANGSIOS" },
	{ 0x40F5 , "HANGUL SYLLABLE SIOS OE IEUNG" },
	{ 0x40F6 , "HANGUL SYLLABLE SIOS YO TIKEUT" },
	{ 0x40F7 , "HANGUL SYLLABLE SIOS YO PHIEUPH" },
	{ 0x40F8 , "HANGUL SYLLABLE SIOS U RIEUL-KIYEOK" },
	{ 0x40F9 , "HANGUL SYLLABLE SIOS U RIEUL-SIOS" },
	{ 0x40FA , "HANGUL SYLLABLE SIOS U RIEUL-HIEUH" },
	{ 0x40FB , "HANGUL SYLLABLE SIOS U CIEUC" },
	{ 0x40FC , "HANGUL SYLLABLE SIOS U HIEUH" },
	{ 0x40FD , "HANGUL SYLLABLE SIOS WEO KIYEOK" },
	{ 0x40FE , "HANGUL SYLLABLE SIOS WEO NIEUN" },
	{ 0x40FF , "HANGUL SYLLABLE SIOS WEO TIKEUT" },
	{ 0x4100 , "HANGUL SYLLABLE SIOS WEO RIEUL" },
	{ 0x4101 , "HANGUL SYLLABLE SIOS WEO MIEUM" },
	{ 0x4102 , "HANGUL SYLLABLE SIOS WEO PIEUP" },
	{ 0x4103 , "HANGUL SYLLABLE SIOS WEO SIOS" },
	{ 0x4104 , "HANGUL SYLLABLE SIOS WEO IEUNG" },
	{ 0x4105 , "HANGUL SYLLABLE SIOS WE TIKEUT" },
	{ 0x4106 , "HANGUL SYLLABLE SIOS WE PIEUP" },
	{ 0x4107 , "HANGUL SYLLABLE SIOS WE SIOS" },
	{ 0x4108 , "HANGUL SYLLABLE SIOS WE SSANGSIOS" },
	{ 0x4109 , "HANGUL SYLLABLE SIOS WI TIKEUT" },
	{ 0x410A , "HANGUL SYLLABLE SIOS WI PHIEUPH" },
	{ 0x410B , "HANGUL SYLLABLE SIOS YU NIEUN" },
	{ 0x410C , "HANGUL SYLLABLE SIOS YU TIKEUT" },
	{ 0x410D , "HANGUL SYLLABLE SIOS YU PIEUP" },
	{ 0x410E , "HANGUL SYLLABLE SIOS EU TIKEUT" },
	{ 0x410F , "HANGUL SYLLABLE SIOS EU RIEUL-MIEUM" },
	{ 0x4110 , "HANGUL SYLLABLE SIOS EU RIEUL-PIEUP" },
	{ 0x4111 , "HANGUL SYLLABLE SIOS EU RIEUL-SIOS" },
	{ 0x4112 , "HANGUL SYLLABLE SIOS EU RIEUL-PHIEUPH" },
	{ 0x4113 , "HANGUL SYLLABLE SIOS EU RIEUL-HIEUH" },
	{ 0x4114 , "HANGUL SYLLABLE SIOS EU CIEUC" },
	{ 0x4115 , "HANGUL SYLLABLE SIOS EU CHIEUCH" },
	{ 0x4116 , "HANGUL SYLLABLE SIOS EU PHIEUPH" },
	{ 0x4117 , "HANGUL SYLLABLE SIOS YI" },
	{ 0x4118 , "HANGUL SYLLABLE SIOS YI KIYEOK" },
	{ 0x4119 , "HANGUL SYLLABLE SIOS YI NIEUN" },
	{ 0x411A , "HANGUL SYLLABLE SIOS YI RIEUL" },
	{ 0x411B , "HANGUL SYLLABLE SIOS YI MIEUM" },
	{ 0x411C , "HANGUL SYLLABLE SIOS YI PIEUP" },
	{ 0x411D , "HANGUL SYLLABLE SIOS YI SIOS" },
	{ 0x411E , "HANGUL SYLLABLE SIOS YI IEUNG" },
	{ 0x411F , "HANGUL SYLLABLE SIOS I SSANGKIYEOK" },
	{ 0x4120 , "HANGUL SYLLABLE SIOS I NIEUN-HIEUH" },
	{ 0x4121 , "HANGUL SYLLABLE SIOS I RIEUL-KIYEOK" },
	{ 0x4122 , "HANGUL SYLLABLE SIOS I RIEUL-MIEUM" },
	{ 0x4123 , "HANGUL SYLLABLE SIOS I RIEUL-SIOS" },
	{ 0x4124 , "HANGUL SYLLABLE SIOS I SSANGSIOS" },
	{ 0x4125 , "HANGUL SYLLABLE SIOS I CHIEUCH" },
	{ 0x4126 , "HANGUL SYLLABLE SIOS I KHIEUKH" },
	{ 0x4127 , "HANGUL SYLLABLE SSANGSIOS A TIKEUT" },
	{ 0x4128 , "HANGUL SYLLABLE SSANGSIOS A RIEUL-KIYEOK" },
	{ 0x4129 , "HANGUL SYLLABLE SSANGSIOS A RIEUL-MIEUM" },
	{ 0x412A , "HANGUL SYLLABLE SSANGSIOS A SIOS" },
	{ 0x412B , "HANGUL SYLLABLE SSANGSIOS A CIEUC" },
	{ 0x412C , "HANGUL SYLLABLE SSANGSIOS A THIEUTH" },
	{ 0x412D , "HANGUL SYLLABLE SSANGSIOS AE TIKEUT" },
	{ 0x412E , "HANGUL SYLLABLE SSANGSIOS AE SIOS" },
	{ 0x412F , "HANGUL SYLLABLE SSANGSIOS AE HIEUH" },
	{ 0x4130 , "HANGUL SYLLABLE SSANGSIOS YA" },
	{ 0x4131 , "HANGUL SYLLABLE SSANGSIOS YA KIYEOK" },
	{ 0x4132 , "HANGUL SYLLABLE SSANGSIOS YA NIEUN" },
	{ 0x4133 , "HANGUL SYLLABLE SSANGSIOS YA TIKEUT" },
	{ 0x4134 , "HANGUL SYLLABLE SSANGSIOS YA RIEUL" },
	{ 0x4135 , "HANGUL SYLLABLE SSANGSIOS YA MIEUM" },
	{ 0x4136 , "HANGUL SYLLABLE SSANGSIOS YA PIEUP" },
	{ 0x4137 , "HANGUL SYLLABLE SSANGSIOS YA SIOS" },
	{ 0x4138 , "HANGUL SYLLABLE SSANGSIOS YAE" },
	{ 0x4139 , "HANGUL SYLLABLE SSANGSIOS YAE NIEUN" },
	{ 0x413A , "HANGUL SYLLABLE SSANGSIOS YAE RIEUL" },
	{ 0x413B , "HANGUL SYLLABLE SSANGSIOS YAE MIEUM" },
	{ 0x413C , "HANGUL SYLLABLE SSANGSIOS YAE IEUNG" },
	{ 0x413D , "HANGUL SYLLABLE SSANGSIOS EO SSANGKIYEOK" },
	{ 0x413E , "HANGUL SYLLABLE SSANGSIOS EO TIKEUT" },
	{ 0x413F , "HANGUL SYLLABLE SSANGSIOS EO SIOS" },
	{ 0x4140 , "HANGUL SYLLABLE SSANGSIOS EO PHIEUPH" },
	{ 0x4141 , "HANGUL SYLLABLE SSANGSIOS E KIYEOK" },
	{ 0x4142 , "HANGUL SYLLABLE SSANGSIOS E TIKEUT" },
	{ 0x4143 , "HANGUL SYLLABLE SSANGSIOS E MIEUM" },
	{ 0x4144 , "HANGUL SYLLABLE SSANGSIOS E PIEUP" },
	{ 0x4145 , "HANGUL SYLLABLE SSANGSIOS E SIOS" },
	{ 0x4146 , "HANGUL SYLLABLE SSANGSIOS E SSANGSIOS" },
	{ 0x4147 , "HANGUL SYLLABLE SSANGSIOS E IEUNG" },
	{ 0x4148 , "HANGUL SYLLABLE SSANGSIOS E THIEUTH" },
	{ 0x4149 , "HANGUL SYLLABLE SSANGSIOS YEO" },
	{ 0x414A , "HANGUL SYLLABLE SSANGSIOS YEO KIYEOK" },
	{ 0x414B , "HANGUL SYLLABLE SSANGSIOS YEO NIEUN" },
	{ 0x414C , "HANGUL SYLLABLE SSANGSIOS YEO TIKEUT" },
	{ 0x414D , "HANGUL SYLLABLE SSANGSIOS YEO RIEUL" },
	{ 0x414E , "HANGUL SYLLABLE SSANGSIOS YEO MIEUM" },
	{ 0x414F , "HANGUL SYLLABLE SSANGSIOS YEO PIEUP" },
	{ 0x4150 , "HANGUL SYLLABLE SSANGSIOS YEO SIOS" },
	{ 0x4151 , "HANGUL SYLLABLE SSANGSIOS YEO SSANGSIOS" },
	{ 0x4152 , "HANGUL SYLLABLE SSANGSIOS YEO IEUNG" },
	{ 0x4153 , "HANGUL SYLLABLE SSANGSIOS YE" },
	{ 0x4154 , "HANGUL SYLLABLE SSANGSIOS YE KIYEOK" },
	{ 0x4155 , "HANGUL SYLLABLE SSANGSIOS YE MIEUM" },
	{ 0x4156 , "HANGUL SYLLABLE SSANGSIOS YE PIEUP" },
	{ 0x4157 , "HANGUL SYLLABLE SSANGSIOS YE SIOS" },
	{ 0x4158 , "HANGUL SYLLABLE SSANGSIOS YE SSANGSIOS" },
	{ 0x4159 , "HANGUL SYLLABLE SSANGSIOS YE IEUNG" },
	{ 0x415A , "HANGUL SYLLABLE SSANGSIOS O SIOS" },
	{ 0x415B , "HANGUL SYLLABLE SSANGSIOS O THIEUTH" },
	{ 0x415C , "HANGUL SYLLABLE SSANGSIOS WA TIKEUT" },
	{ 0x415D , "HANGUL SYLLABLE SSANGSIOS WA RIEUL" },
	{ 0x415E , "HANGUL SYLLABLE SSANGSIOS WA MIEUM" },
	{ 0x415F , "HANGUL SYLLABLE SSANGSIOS WA PIEUP" },
	{ 0x4160 , "HANGUL SYLLABLE SSANGSIOS WA SIOS" },
	{ 0x4161 , "HANGUL SYLLABLE SSANGSIOS WA IEUNG" },
	{ 0x4162 , "HANGUL SYLLABLE SSANGSIOS WAE KIYEOK" },
	{ 0x4163 , "HANGUL SYLLABLE SSANGSIOS WAE NIEUN" },
	{ 0x4164 , "HANGUL SYLLABLE SSANGSIOS WAE TIKEUT" },
	{ 0x4165 , "HANGUL SYLLABLE SSANGSIOS WAE RIEUL" },
	{ 0x4166 , "HANGUL SYLLABLE SSANGSIOS WAE SIOS" },
	{ 0x4167 , "HANGUL SYLLABLE SSANGSIOS OE KIYEOK" },
	{ 0x4168 , "HANGUL SYLLABLE SSANGSIOS OE TIKEUT" },
	{ 0x4169 , "HANGUL SYLLABLE SSANGSIOS OE SIOS" },
	{ 0x416A , "HANGUL SYLLABLE SSANGSIOS OE SSANGSIOS" },
	{ 0x416B , "HANGUL SYLLABLE SSANGSIOS OE IEUNG" },
	{ 0x416C , "HANGUL SYLLABLE SSANGSIOS YO KIYEOK" },
	{ 0x416D , "HANGUL SYLLABLE SSANGSIOS YO NIEUN" },
	{ 0x416E , "HANGUL SYLLABLE SSANGSIOS YO TIKEUT" },
	{ 0x416F , "HANGUL SYLLABLE SSANGSIOS YO RIEUL" },
	{ 0x4170 , "HANGUL SYLLABLE SSANGSIOS YO MIEUM" },
	{ 0x4171 , "HANGUL SYLLABLE SSANGSIOS YO PIEUP" },
	{ 0x4172 , "HANGUL SYLLABLE SSANGSIOS YO SIOS" },
	{ 0x4173 , "HANGUL SYLLABLE SSANGSIOS YO IEUNG" },
	{ 0x4174 , "HANGUL SYLLABLE SSANGSIOS U TIKEUT" },
	{ 0x4175 , "HANGUL SYLLABLE SSANGSIOS U SIOS" },
	{ 0x4176 , "HANGUL SYLLABLE SSANGSIOS U PHIEUPH" },
	{ 0x4177 , "HANGUL SYLLABLE SSANGSIOS WEO KIYEOK" },
	{ 0x4178 , "HANGUL SYLLABLE SSANGSIOS WEO NIEUN" },
	{ 0x4179 , "HANGUL SYLLABLE SSANGSIOS WEO TIKEUT" },
	{ 0x417A , "HANGUL SYLLABLE SSANGSIOS WEO RIEUL" },
	{ 0x417B , "HANGUL SYLLABLE SSANGSIOS WEO MIEUM" },
	{ 0x417C , "HANGUL SYLLABLE SSANGSIOS WEO PIEUP" },
	{ 0x417D , "HANGUL SYLLABLE SSANGSIOS WEO SIOS" },
	{ 0x417E , "HANGUL SYLLABLE SSANGSIOS WEO IEUNG" },
	{ 0x417F , "HANGUL SYLLABLE SSANGSIOS WE KIYEOK" },
	{ 0x4180 , "HANGUL SYLLABLE SSANGSIOS WE NIEUN" },
	{ 0x4181 , "HANGUL SYLLABLE SSANGSIOS WE TIKEUT" },
	{ 0x4182 , "HANGUL SYLLABLE SSANGSIOS WE RIEUL" },
	{ 0x4183 , "HANGUL SYLLABLE SSANGSIOS WE MIEUM" },
	{ 0x4184 , "HANGUL SYLLABLE SSANGSIOS WE PIEUP" },
	{ 0x4185 , "HANGUL SYLLABLE SSANGSIOS WE SIOS" },
	{ 0x4186 , "HANGUL SYLLABLE SSANGSIOS WE SSANGSIOS" },
	{ 0x4187 , "HANGUL SYLLABLE SSANGSIOS WE IEUNG" },
	{ 0x4188 , "HANGUL SYLLABLE SSANGSIOS WI KIYEOK" },
	{ 0x4189 , "HANGUL SYLLABLE SSANGSIOS WI TIKEUT" },
	{ 0x418A , "HANGUL SYLLABLE SSANGSIOS WI RIEUL" },
	{ 0x418B , "HANGUL SYLLABLE SSANGSIOS WI MIEUM" },
	{ 0x418C , "HANGUL SYLLABLE SSANGSIOS WI PIEUP" },
	{ 0x418D , "HANGUL SYLLABLE SSANGSIOS WI SIOS" },
	{ 0x418E , "HANGUL SYLLABLE SSANGSIOS WI IEUNG" },
	{ 0x418F , "HANGUL SYLLABLE SSANGSIOS WI HIEUH" },
	{ 0x4190 , "HANGUL SYLLABLE SSANGSIOS YU" },
	{ 0x4191 , "HANGUL SYLLABLE SSANGSIOS YU KIYEOK" },
	{ 0x4192 , "HANGUL SYLLABLE SSANGSIOS YU NIEUN" },
	{ 0x4193 , "HANGUL SYLLABLE SSANGSIOS YU TIKEUT" },
	{ 0x4194 , "HANGUL SYLLABLE SSANGSIOS YU RIEUL" },
	{ 0x4195 , "HANGUL SYLLABLE SSANGSIOS YU MIEUM" },
	{ 0x4196 , "HANGUL SYLLABLE SSANGSIOS YU PIEUP" },
	{ 0x4197 , "HANGUL SYLLABLE SSANGSIOS YU SIOS" },
	{ 0x4198 , "HANGUL SYLLABLE SSANGSIOS EU TIKEUT" },
	{ 0x4199 , "HANGUL SYLLABLE SSANGSIOS EU SIOS" },
	{ 0x419A , "HANGUL SYLLABLE SSANGSIOS EU IEUNG" },
	{ 0x419B , "HANGUL SYLLABLE SSANGSIOS YI KIYEOK" },
	{ 0x419C , "HANGUL SYLLABLE SSANGSIOS YI PIEUP" },
	{ 0x419D , "HANGUL SYLLABLE SSANGSIOS YI SIOS" },
	{ 0x419E , "HANGUL SYLLABLE SSANGSIOS YI IEUNG" },
	{ 0x419F , "HANGUL SYLLABLE SSANGSIOS I TIKEUT" },
	{ 0x41A0 , "HANGUL SYLLABLE SSANGSIOS I RIEUL-KIYEOK" },
	{ 0x41A1 , "HANGUL SYLLABLE SSANGSIOS I SSANGSIOS" },
	{ 0x41A2 , "HANGUL SYLLABLE SSANGSIOS I CHIEUCH" },
	{ 0x41A3 , "HANGUL SYLLABLE SSANGSIOS I KHIEUKH" },
	{ 0x41A4 , "HANGUL SYLLABLE SSANGSIOS I HIEUH" },
	{ 0x41A5 , "HANGUL SYLLABLE IEUNG A TIKEUT" },
	{ 0x41A6 , "HANGUL SYLLABLE IEUNG A RIEUL-PIEUP" },
	{ 0x41A7 , "HANGUL SYLLABLE IEUNG A RIEUL-PHIEUPH" },
	{ 0x41A8 , "HANGUL SYLLABLE IEUNG A PIEUP-SIOS" },
	{ 0x41A9 , "HANGUL SYLLABLE IEUNG A CIEUC" },
	{ 0x41AA , "HANGUL SYLLABLE IEUNG A CHIEUCH" },
	{ 0x41AB , "HANGUL SYLLABLE IEUNG A HIEUH" },
	{ 0x41AC , "HANGUL SYLLABLE IEUNG AE TIKEUT" },
	{ 0x41AD , "HANGUL SYLLABLE IEUNG AE THIEUTH" },
	{ 0x41AE , "HANGUL SYLLABLE IEUNG AE PHIEUPH" },
	{ 0x41AF , "HANGUL SYLLABLE IEUNG YA SSANGKIYEOK" },
	{ 0x41B0 , "HANGUL SYLLABLE IEUNG YA TIKEUT" },
	{ 0x41B1 , "HANGUL SYLLABLE IEUNG YA SSANGSIOS" },
	{ 0x41B2 , "HANGUL SYLLABLE IEUNG YA CIEUC" },
	{ 0x41B3 , "HANGUL SYLLABLE IEUNG YA KHIEUKH" },
	{ 0x41B4 , "HANGUL SYLLABLE IEUNG YAE KIYEOK" },
	{ 0x41B5 , "HANGUL SYLLABLE IEUNG YAE TIKEUT" },
	{ 0x41B6 , "HANGUL SYLLABLE IEUNG YAE MIEUM" },
	{ 0x41B7 , "HANGUL SYLLABLE IEUNG YAE SIOS" },
	{ 0x41B8 , "HANGUL SYLLABLE IEUNG YAE SSANGSIOS" },
	{ 0x41B9 , "HANGUL SYLLABLE IEUNG YAE IEUNG" },
	{ 0x41BA , "HANGUL SYLLABLE IEUNG EO SSANGKIYEOK" },
	{ 0x41BB , "HANGUL SYLLABLE IEUNG EO NIEUN-HIEUH" },
	{ 0x41BC , "HANGUL SYLLABLE IEUNG EO RIEUL-PIEUP" },
	{ 0x41BD , "HANGUL SYLLABLE IEUNG EO HIEUH" },
	{ 0x41BE , "HANGUL SYLLABLE IEUNG E TIKEUT" },
	{ 0x41BF , "HANGUL SYLLABLE IEUNG E SSANGSIOS" },
	{ 0x41C0 , "HANGUL SYLLABLE IEUNG E THIEUTH" },
	{ 0x41C1 , "HANGUL SYLLABLE IEUNG E PHIEUPH" },
	{ 0x41C2 , "HANGUL SYLLABLE IEUNG YEO NIEUN-CIEUC" },
	{ 0x41C3 , "HANGUL SYLLABLE IEUNG YEO TIKEUT" },
	{ 0x41C4 , "HANGUL SYLLABLE IEUNG YEO RIEUL-SIOS" },
	{ 0x41C5 , "HANGUL SYLLABLE IEUNG YEO KHIEUKH" },
	{ 0x41C6 , "HANGUL SYLLABLE IEUNG YE KIYEOK" },
	{ 0x41C7 , "HANGUL SYLLABLE IEUNG YE TIKEUT" },
	{ 0x41C8 , "HANGUL SYLLABLE IEUNG YE IEUNG" },
	{ 0x41C9 , "HANGUL SYLLABLE IEUNG YE KHIEUKH" },
	{ 0x41CA , "HANGUL SYLLABLE IEUNG O SSANGKIYEOK" },
	{ 0x41CB , "HANGUL SYLLABLE IEUNG O NIEUN-HIEUH" },
	{ 0x41CC , "HANGUL SYLLABLE IEUNG O TIKEUT" },
	{ 0x41CD , "HANGUL SYLLABLE IEUNG O RIEUL-PIEUP" },
	{ 0x41CE , "HANGUL SYLLABLE IEUNG O RIEUL-PHIEUPH" },
	{ 0x41CF , "HANGUL SYLLABLE IEUNG O CIEUC" },
	{ 0x41D0 , "HANGUL SYLLABLE IEUNG O THIEUTH" },
	{ 0x41D1 , "HANGUL SYLLABLE IEUNG O PHIEUPH" },
	{ 0x41D2 , "HANGUL SYLLABLE IEUNG WA TIKEUT" },
	{ 0x41D3 , "HANGUL SYLLABLE IEUNG WA KHIEUKH" },
	{ 0x41D4 , "HANGUL SYLLABLE IEUNG WA THIEUTH" },
	{ 0x41D5 , "HANGUL SYLLABLE IEUNG WAE TIKEUT" },
	{ 0x41D6 , "HANGUL SYLLABLE IEUNG WAE RIEUL" },
	{ 0x41D7 , "HANGUL SYLLABLE IEUNG WAE PIEUP" },
	{ 0x41D8 , "HANGUL SYLLABLE IEUNG WAE SSANGSIOS" },
	{ 0x41D9 , "HANGUL SYLLABLE IEUNG OE TIKEUT" },
	{ 0x41DA , "HANGUL SYLLABLE IEUNG OE RIEUL-MIEUM" },
	{ 0x41DB , "HANGUL SYLLABLE IEUNG OE SSANGSIOS" },
	{ 0x41DC , "HANGUL SYLLABLE IEUNG OE PHIEUPH" },
	{ 0x41DD , "HANGUL SYLLABLE IEUNG YO TIKEUT" },
	{ 0x41DE , "HANGUL SYLLABLE IEUNG YO RIEUL-KIYEOK" },
	{ 0x41DF , "HANGUL SYLLABLE IEUNG YO KHIEUKH" },
	{ 0x41E0 , "HANGUL SYLLABLE IEUNG U TIKEUT" },
	{ 0x41E1 , "HANGUL SYLLABLE IEUNG U RIEUL-SIOS" },
	{ 0x41E2 , "HANGUL SYLLABLE IEUNG U PIEUP-SIOS" },
	{ 0x41E3 , "HANGUL SYLLABLE IEUNG U CIEUC" },
	{ 0x41E4 , "HANGUL SYLLABLE IEUNG U CHIEUCH" },
	{ 0x41E5 , "HANGUL SYLLABLE IEUNG U THIEUTH" },
	{ 0x41E6 , "HANGUL SYLLABLE IEUNG WEO NIEUN-HIEUH" },
	{ 0x41E7 , "HANGUL SYLLABLE IEUNG WEO TIKEUT" },
	{ 0x41E8 , "HANGUL SYLLABLE IEUNG WEO SIOS" },
	{ 0x41E9 , "HANGUL SYLLABLE IEUNG WEO KHIEUKH" },
	{ 0x41EA , "HANGUL SYLLABLE IEUNG WEO THIEUTH" },
	{ 0x41EB , "HANGUL SYLLABLE IEUNG WE TIKEUT" },
	{ 0x41EC , "HANGUL SYLLABLE IEUNG WE SIOS" },
	{ 0x41ED , "HANGUL SYLLABLE IEUNG WE SSANGSIOS" },
	{ 0x41EE , "HANGUL SYLLABLE IEUNG WE THIEUTH" },
	{ 0x41EF , "HANGUL SYLLABLE IEUNG WI TIKEUT" },
	{ 0x41F0 , "HANGUL SYLLABLE IEUNG WI RIEUL-MIEUM" },
	{ 0x41F1 , "HANGUL SYLLABLE IEUNG WI RIEUL-HIEUH" },
	{ 0x41F2 , "HANGUL SYLLABLE IEUNG WI SSANGSIOS" },
	{ 0x41F3 , "HANGUL SYLLABLE IEUNG WI KHIEUKH" },
	{ 0x41F4 , "HANGUL SYLLABLE IEUNG WI THIEUTH" },
	{ 0x41F5 , "HANGUL SYLLABLE IEUNG YU TIKEUT" },
	{ 0x41F6 , "HANGUL SYLLABLE IEUNG YU CIEUC" },
	{ 0x41F7 , "HANGUL SYLLABLE IEUNG YU KHIEUKH" },
	{ 0x41F8 , "HANGUL SYLLABLE IEUNG YU THIEUTH" },
	{ 0x41F9 , "HANGUL SYLLABLE IEUNG EU TIKEUT" },
	{ 0x41FA , "HANGUL SYLLABLE IEUNG EU RIEUL-KIYEOK" },
	{ 0x41FB , "HANGUL SYLLABLE IEUNG EU RIEUL-MIEUM" },
	{ 0x41FC , "HANGUL SYLLABLE IEUNG EU RIEUL-PIEUP" },
	{ 0x41FD , "HANGUL SYLLABLE IEUNG EU PIEUP-SIOS" },
	{ 0x41FE , "HANGUL SYLLABLE IEUNG YI KIYEOK" },
	{ 0x41FF , "HANGUL SYLLABLE IEUNG YI TIKEUT" },
	{ 0x4200 , "HANGUL SYLLABLE IEUNG YI RIEUL-PHIEUPH" },
	{ 0x4201 , "HANGUL SYLLABLE IEUNG YI PIEUP" },
	{ 0x4202 , "HANGUL SYLLABLE IEUNG YI IEUNG" },
	{ 0x4203 , "HANGUL SYLLABLE IEUNG I TIKEUT" },
	{ 0x4204 , "HANGUL SYLLABLE IEUNG I RIEUL-SIOS" },
	{ 0x4205 , "HANGUL SYLLABLE IEUNG I RIEUL-PHIEUPH" },
	{ 0x4206 , "HANGUL SYLLABLE IEUNG I PIEUP-SIOS" },
	{ 0x4207 , "HANGUL SYLLABLE IEUNG I CHIEUCH" },
	{ 0x4208 , "HANGUL SYLLABLE IEUNG I KHIEUKH" },
	{ 0x4209 , "HANGUL SYLLABLE IEUNG I THIEUTH" },
	{ 0x420A , "HANGUL SYLLABLE IEUNG I HIEUH" },
	{ 0x420B , "HANGUL SYLLABLE CIEUC A KIYEOK-SIOS" },
	{ 0x420C , "HANGUL SYLLABLE CIEUC A RIEUL-KIYEOK" },
	{ 0x420D , "HANGUL SYLLABLE CIEUC A CHIEUCH" },
	{ 0x420E , "HANGUL SYLLABLE CIEUC A THIEUTH" },
	{ 0x420F , "HANGUL SYLLABLE CIEUC A PHIEUPH" },
	{ 0x4210 , "HANGUL SYLLABLE CIEUC A HIEUH" },
	{ 0x4211 , "HANGUL SYLLABLE CIEUC AE NIEUN-HIEUH" },
	{ 0x4212 , "HANGUL SYLLABLE CIEUC AE TIKEUT" },
	{ 0x4213 , "HANGUL SYLLABLE CIEUC AE CIEUC" },
	{ 0x4214 , "HANGUL SYLLABLE CIEUC YA TIKEUT" },
	{ 0x4215 , "HANGUL SYLLABLE CIEUC YA PIEUP" },
	{ 0x4216 , "HANGUL SYLLABLE CIEUC YA SIOS" },
	{ 0x4217 , "HANGUL SYLLABLE CIEUC YAE TIKEUT" },
	{ 0x4218 , "HANGUL SYLLABLE CIEUC YAE MIEUM" },
	{ 0x4219 , "HANGUL SYLLABLE CIEUC YAE PIEUP" },
	{ 0x421A , "HANGUL SYLLABLE CIEUC YAE SIOS" },
	{ 0x421B , "HANGUL SYLLABLE CIEUC YAE IEUNG" },
	{ 0x421C , "HANGUL SYLLABLE CIEUC EO SSANGKIYEOK" },
	{ 0x421D , "HANGUL SYLLABLE CIEUC EO NIEUN-HIEUH" },
	{ 0x421E , "HANGUL SYLLABLE CIEUC EO TIKEUT" },
	{ 0x421F , "HANGUL SYLLABLE CIEUC EO RIEUL-KIYEOK" },
	{ 0x4220 , "HANGUL SYLLABLE CIEUC EO SSANGSIOS" },
	{ 0x4221 , "HANGUL SYLLABLE CIEUC EO THIEUTH" },
	{ 0x4222 , "HANGUL SYLLABLE CIEUC EO HIEUH" },
	{ 0x4223 , "HANGUL SYLLABLE CIEUC E TIKEUT" },
	{ 0x4224 , "HANGUL SYLLABLE CIEUC E SSANGSIOS" },
	{ 0x4225 , "HANGUL SYLLABLE CIEUC E CIEUC" },
	{ 0x4226 , "HANGUL SYLLABLE CIEUC YEO KIYEOK" },
	{ 0x4227 , "HANGUL SYLLABLE CIEUC YEO TIKEUT" },
	{ 0x4228 , "HANGUL SYLLABLE CIEUC YEO RIEUL-KIYEOK" },
	{ 0x4229 , "HANGUL SYLLABLE CIEUC YEO RIEUL-MIEUM" },
	{ 0x422A , "HANGUL SYLLABLE CIEUC YEO RIEUL-PIEUP" },
	{ 0x422B , "HANGUL SYLLABLE CIEUC YEO RIEUL-HIEUH" },
	{ 0x422C , "HANGUL SYLLABLE CIEUC YEO SIOS" },
	{ 0x422D , "HANGUL SYLLABLE CIEUC YEO CIEUC" },
	{ 0x422E , "HANGUL SYLLABLE CIEUC YEO THIEUTH" },
	{ 0x422F , "HANGUL SYLLABLE CIEUC YEO HIEUH" },
	{ 0x4230 , "HANGUL SYLLABLE CIEUC YE KIYEOK" },
	{ 0x4231 , "HANGUL SYLLABLE CIEUC YE NIEUN" },
	{ 0x4232 , "HANGUL SYLLABLE CIEUC YE RIEUL" },
	{ 0x4233 , "HANGUL SYLLABLE CIEUC YE MIEUM" },
	{ 0x4234 , "HANGUL SYLLABLE CIEUC YE PIEUP" },
	{ 0x4235 , "HANGUL SYLLABLE CIEUC YE SIOS" },
	{ 0x4236 , "HANGUL SYLLABLE CIEUC YE IEUNG" },
	{ 0x4237 , "HANGUL SYLLABLE CIEUC O TIKEUT" },
	{ 0x4238 , "HANGUL SYLLABLE CIEUC WA NIEUN" },
	{ 0x4239 , "HANGUL SYLLABLE CIEUC WA NIEUN-HIEUH" },
	{ 0x423A , "HANGUL SYLLABLE CIEUC WA TIKEUT" },
	{ 0x423B , "HANGUL SYLLABLE CIEUC WA RIEUL-KIYEOK" },
	{ 0x423C , "HANGUL SYLLABLE CIEUC WA MIEUM" },
	{ 0x423D , "HANGUL SYLLABLE CIEUC WA SSANGSIOS" },
	{ 0x423E , "HANGUL SYLLABLE CIEUC WAE KIYEOK" },
	{ 0x423F , "HANGUL SYLLABLE CIEUC WAE NIEUN" },
	{ 0x4240 , "HANGUL SYLLABLE CIEUC WAE TIKEUT" },
	{ 0x4241 , "HANGUL SYLLABLE CIEUC WAE RIEUL" },
	{ 0x4242 , "HANGUL SYLLABLE CIEUC WAE MIEUM" },
	{ 0x4243 , "HANGUL SYLLABLE CIEUC WAE SIOS" },
	{ 0x4244 , "HANGUL SYLLABLE CIEUC OE KIYEOK" },
	{ 0x4245 , "HANGUL SYLLABLE CIEUC OE TIKEUT" },
	{ 0x4246 , "HANGUL SYLLABLE CIEUC OE SSANGSIOS" },
	{ 0x4247 , "HANGUL SYLLABLE CIEUC YO TIKEUT" },
	{ 0x4248 , "HANGUL SYLLABLE CIEUC YO RIEUL" },
	{ 0x4249 , "HANGUL SYLLABLE CIEUC YO MIEUM" },
	{ 0x424A , "HANGUL SYLLABLE CIEUC YO PIEUP" },
	{ 0x424B , "HANGUL SYLLABLE CIEUC YO SIOS" },
	{ 0x424C , "HANGUL SYLLABLE CIEUC YO HIEUH" },
	{ 0x424D , "HANGUL SYLLABLE CIEUC U TIKEUT" },
	{ 0x424E , "HANGUL SYLLABLE CIEUC U RIEUL-SIOS" },
	{ 0x424F , "HANGUL SYLLABLE CIEUC U HIEUH" },
	{ 0x4250 , "HANGUL SYLLABLE CIEUC WEO KIYEOK" },
	{ 0x4251 , "HANGUL SYLLABLE CIEUC WEO NIEUN" },
	{ 0x4252 , "HANGUL SYLLABLE CIEUC WEO TIKEUT" },
	{ 0x4253 , "HANGUL SYLLABLE CIEUC WEO RIEUL" },
	{ 0x4254 , "HANGUL SYLLABLE CIEUC WEO MIEUM" },
	{ 0x4255 , "HANGUL SYLLABLE CIEUC WEO PIEUP" },
	{ 0x4256 , "HANGUL SYLLABLE CIEUC WEO SIOS" },
	{ 0x4257 , "HANGUL SYLLABLE CIEUC WEO IEUNG" },
	{ 0x4258 , "HANGUL SYLLABLE CIEUC WE KIYEOK" },
	{ 0x4259 , "HANGUL SYLLABLE CIEUC WE NIEUN" },
	{ 0x425A , "HANGUL SYLLABLE CIEUC WE TIKEUT" },
	{ 0x425B , "HANGUL SYLLABLE CIEUC WE RIEUL" },
	{ 0x425C , "HANGUL SYLLABLE CIEUC WE MIEUM" },
	{ 0x425D , "HANGUL SYLLABLE CIEUC WE PIEUP" },
	{ 0x425E , "HANGUL SYLLABLE CIEUC WE SIOS" },
	{ 0x425F , "HANGUL SYLLABLE CIEUC WE SSANGSIOS" },
	{ 0x4260 , "HANGUL SYLLABLE CIEUC WE IEUNG" },
	{ 0x4261 , "HANGUL SYLLABLE CIEUC WI TIKEUT" },
	{ 0x4262 , "HANGUL SYLLABLE CIEUC WI IEUNG" },
	{ 0x4263 , "HANGUL SYLLABLE CIEUC YU KIYEOK" },
	{ 0x4264 , "HANGUL SYLLABLE CIEUC YU PIEUP" },
	{ 0x4265 , "HANGUL SYLLABLE CIEUC YU SIOS" },
	{ 0x4266 , "HANGUL SYLLABLE CIEUC YU IEUNG" },
	{ 0x4267 , "HANGUL SYLLABLE CIEUC EU TIKEUT" },
	{ 0x4268 , "HANGUL SYLLABLE CIEUC EU RIEUL-KIYEOK" },
	{ 0x4269 , "HANGUL SYLLABLE CIEUC EU RIEUL-MIEUM" },
	{ 0x426A , "HANGUL SYLLABLE CIEUC EU RIEUL-SIOS" },
	{ 0x426B , "HANGUL SYLLABLE CIEUC EU CIEUC" },
	{ 0x426C , "HANGUL SYLLABLE CIEUC YI" },
	{ 0x426D , "HANGUL SYLLABLE CIEUC YI KIYEOK" },
	{ 0x426E , "HANGUL SYLLABLE CIEUC YI NIEUN" },
	{ 0x426F , "HANGUL SYLLABLE CIEUC YI RIEUL" },
	{ 0x4270 , "HANGUL SYLLABLE CIEUC YI MIEUM" },
	{ 0x4271 , "HANGUL SYLLABLE CIEUC YI PIEUP" },
	{ 0x4272 , "HANGUL SYLLABLE CIEUC YI SIOS" },
	{ 0x4273 , "HANGUL SYLLABLE CIEUC YI IEUNG" },
	{ 0x4274 , "HANGUL SYLLABLE CIEUC I RIEUL-HIEUH" },
	{ 0x4275 , "HANGUL SYLLABLE CIEUC I PIEUP-SIOS" },
	{ 0x4276 , "HANGUL SYLLABLE CIEUC I SSANGSIOS" },
	{ 0x4277 , "HANGUL SYLLABLE CIEUC I CHIEUCH" },
	{ 0x4278 , "HANGUL SYLLABLE CIEUC I HIEUH" },
	{ 0x4279 , "HANGUL SYLLABLE SSANGCIEUC A SSANGKIYEOK" },
	{ 0x427A , "HANGUL SYLLABLE SSANGCIEUC A TIKEUT" },
	{ 0x427B , "HANGUL SYLLABLE SSANGCIEUC A RIEUL-SIOS" },
	{ 0x427C , "HANGUL SYLLABLE SSANGCIEUC A HIEUH" },
	{ 0x427D , "HANGUL SYLLABLE SSANGCIEUC AE TIKEUT" },
	{ 0x427E , "HANGUL SYLLABLE SSANGCIEUC YA KIYEOK" },
	{ 0x427F , "HANGUL SYLLABLE SSANGCIEUC YA NIEUN-HIEUH" },
	{ 0x4280 , "HANGUL SYLLABLE SSANGCIEUC YA RIEUL" },
	{ 0x4281 , "HANGUL SYLLABLE SSANGCIEUC YA MIEUM" },
	{ 0x4282 , "HANGUL SYLLABLE SSANGCIEUC YA PIEUP" },
	{ 0x4283 , "HANGUL SYLLABLE SSANGCIEUC YA SIOS" },
	{ 0x4284 , "HANGUL SYLLABLE SSANGCIEUC YA THIEUTH" },
	{ 0x4285 , "HANGUL SYLLABLE SSANGCIEUC YAE" },
	{ 0x4286 , "HANGUL SYLLABLE SSANGCIEUC EO TIKEUT" },
	{ 0x4287 , "HANGUL SYLLABLE SSANGCIEUC EO RIEUL-PIEUP" },
	{ 0x4288 , "HANGUL SYLLABLE SSANGCIEUC EO RIEUL-SIOS" },
	{ 0x4289 , "HANGUL SYLLABLE SSANGCIEUC EO RIEUL-HIEUH" },
	{ 0x428A , "HANGUL SYLLABLE SSANGCIEUC EO HIEUH" },
	{ 0x428B , "HANGUL SYLLABLE SSANGCIEUC E KIYEOK" },
	{ 0x428C , "HANGUL SYLLABLE SSANGCIEUC E NIEUN" },
	{ 0x428D , "HANGUL SYLLABLE SSANGCIEUC E TIKEUT" },
	{ 0x428E , "HANGUL SYLLABLE SSANGCIEUC E RIEUL" },
	{ 0x428F , "HANGUL SYLLABLE SSANGCIEUC E MIEUM" },
	{ 0x4290 , "HANGUL SYLLABLE SSANGCIEUC E PIEUP" },
	{ 0x4291 , "HANGUL SYLLABLE SSANGCIEUC E SIOS" },
	{ 0x4292 , "HANGUL SYLLABLE SSANGCIEUC E SSANGSIOS" },
	{ 0x4293 , "HANGUL SYLLABLE SSANGCIEUC E THIEUTH" },
	{ 0x4294 , "HANGUL SYLLABLE SSANGCIEUC YEO KIYEOK" },
	{ 0x4295 , "HANGUL SYLLABLE SSANGCIEUC YEO NIEUN" },
	{ 0x4296 , "HANGUL SYLLABLE SSANGCIEUC YEO TIKEUT" },
	{ 0x4297 , "HANGUL SYLLABLE SSANGCIEUC YEO RIEUL" },
	{ 0x4298 , "HANGUL SYLLABLE SSANGCIEUC YEO MIEUM" },
	{ 0x4299 , "HANGUL SYLLABLE SSANGCIEUC YEO PIEUP" },
	{ 0x429A , "HANGUL SYLLABLE SSANGCIEUC YEO SIOS" },
	{ 0x429B , "HANGUL SYLLABLE SSANGCIEUC YEO IEUNG" },
	{ 0x429C , "HANGUL SYLLABLE SSANGCIEUC YE" },
	{ 0x429D , "HANGUL SYLLABLE SSANGCIEUC YE KIYEOK" },
	{ 0x429E , "HANGUL SYLLABLE SSANGCIEUC YE NIEUN" },
	{ 0x429F , "HANGUL SYLLABLE SSANGCIEUC YE RIEUL" },
	{ 0x42A0 , "HANGUL SYLLABLE SSANGCIEUC YE MIEUM" },
	{ 0x42A1 , "HANGUL SYLLABLE SSANGCIEUC YE PIEUP" },
	{ 0x42A2 , "HANGUL SYLLABLE SSANGCIEUC YE SIOS" },
	{ 0x42A3 , "HANGUL SYLLABLE SSANGCIEUC YE IEUNG" },
	{ 0x42A4 , "HANGUL SYLLABLE SSANGCIEUC O SSANGKIYEOK" },
	{ 0x42A5 , "HANGUL SYLLABLE SSANGCIEUC O TIKEUT" },
	{ 0x42A6 , "HANGUL SYLLABLE SSANGCIEUC O RIEUL-MIEUM" },
	{ 0x42A7 , "HANGUL SYLLABLE SSANGCIEUC O CIEUC" },
	{ 0x42A8 , "HANGUL SYLLABLE SSANGCIEUC O HIEUH" },
	{ 0x42A9 , "HANGUL SYLLABLE SSANGCIEUC WA NIEUN" },
	{ 0x42AA , "HANGUL SYLLABLE SSANGCIEUC WA TIKEUT" },
	{ 0x42AB , "HANGUL SYLLABLE SSANGCIEUC WA MIEUM" },
	{ 0x42AC , "HANGUL SYLLABLE SSANGCIEUC WA PIEUP" },
	{ 0x42AD , "HANGUL SYLLABLE SSANGCIEUC WA SIOS" },
	{ 0x42AE , "HANGUL SYLLABLE SSANGCIEUC WA IEUNG" },
	{ 0x42AF , "HANGUL SYLLABLE SSANGCIEUC WAE KIYEOK" },
	{ 0x42B0 , "HANGUL SYLLABLE SSANGCIEUC WAE NIEUN" },
	{ 0x42B1 , "HANGUL SYLLABLE SSANGCIEUC WAE TIKEUT" },
	{ 0x42B2 , "HANGUL SYLLABLE SSANGCIEUC WAE RIEUL" },
	{ 0x42B3 , "HANGUL SYLLABLE SSANGCIEUC OE KIYEOK" },
	{ 0x42B4 , "HANGUL SYLLABLE SSANGCIEUC OE TIKEUT" },
	{ 0x42B5 , "HANGUL SYLLABLE SSANGCIEUC OE SIOS" },
	{ 0x42B6 , "HANGUL SYLLABLE SSANGCIEUC OE SSANGSIOS" },
	{ 0x42B7 , "HANGUL SYLLABLE SSANGCIEUC OE IEUNG" },
	{ 0x42B8 , "HANGUL SYLLABLE SSANGCIEUC OE CHIEUCH" },
	{ 0x42B9 , "HANGUL SYLLABLE SSANGCIEUC YO" },
	{ 0x42BA , "HANGUL SYLLABLE SSANGCIEUC YO KIYEOK" },
	{ 0x42BB , "HANGUL SYLLABLE SSANGCIEUC YO NIEUN" },
	{ 0x42BC , "HANGUL SYLLABLE SSANGCIEUC YO RIEUL" },
	{ 0x42BD , "HANGUL SYLLABLE SSANGCIEUC YO MIEUM" },
	{ 0x42BE , "HANGUL SYLLABLE SSANGCIEUC YO PIEUP" },
	{ 0x42BF , "HANGUL SYLLABLE SSANGCIEUC YO SIOS" },
	{ 0x42C0 , "HANGUL SYLLABLE SSANGCIEUC U SSANGKIYEOK" },
	{ 0x42C1 , "HANGUL SYLLABLE SSANGCIEUC U TIKEUT" },
	{ 0x42C2 , "HANGUL SYLLABLE SSANGCIEUC U RIEUL-MIEUM" },
	{ 0x42C3 , "HANGUL SYLLABLE SSANGCIEUC U SIOS" },
	{ 0x42C4 , "HANGUL SYLLABLE SSANGCIEUC WEO KIYEOK" },
	{ 0x42C5 , "HANGUL SYLLABLE SSANGCIEUC WEO NIEUN" },
	{ 0x42C6 , "HANGUL SYLLABLE SSANGCIEUC WEO TIKEUT" },
	{ 0x42C7 , "HANGUL SYLLABLE SSANGCIEUC WEO RIEUL" },
	{ 0x42C8 , "HANGUL SYLLABLE SSANGCIEUC WEO MIEUM" },
	{ 0x42C9 , "HANGUL SYLLABLE SSANGCIEUC WEO PIEUP" },
	{ 0x42CA , "HANGUL SYLLABLE SSANGCIEUC WEO SIOS" },
	{ 0x42CB , "HANGUL SYLLABLE SSANGCIEUC WE" },
	{ 0x42CC , "HANGUL SYLLABLE SSANGCIEUC WE KIYEOK" },
	{ 0x42CD , "HANGUL SYLLABLE SSANGCIEUC WE NIEUN" },
	{ 0x42CE , "HANGUL SYLLABLE SSANGCIEUC WE TIKEUT" },
	{ 0x42CF , "HANGUL SYLLABLE SSANGCIEUC WE RIEUL" },
	{ 0x42D0 , "HANGUL SYLLABLE SSANGCIEUC WE MIEUM" },
	{ 0x42D1 , "HANGUL SYLLABLE SSANGCIEUC WE PIEUP" },
	{ 0x42D2 , "HANGUL SYLLABLE SSANGCIEUC WE SIOS" },
	{ 0x42D3 , "HANGUL SYLLABLE SSANGCIEUC WE IEUNG" },
	{ 0x42D4 , "HANGUL SYLLABLE SSANGCIEUC WI KIYEOK" },
	{ 0x42D5 , "HANGUL SYLLABLE SSANGCIEUC WI NIEUN" },
	{ 0x42D6 , "HANGUL SYLLABLE SSANGCIEUC WI TIKEUT" },
	{ 0x42D7 , "HANGUL SYLLABLE SSANGCIEUC WI RIEUL" },
	{ 0x42D8 , "HANGUL SYLLABLE SSANGCIEUC WI MIEUM" },
	{ 0x42D9 , "HANGUL SYLLABLE SSANGCIEUC WI PIEUP" },
	{ 0x42DA , "HANGUL SYLLABLE SSANGCIEUC WI SIOS" },
	{ 0x42DB , "HANGUL SYLLABLE SSANGCIEUC WI IEUNG" },
	{ 0x42DC , "HANGUL SYLLABLE SSANGCIEUC YU KIYEOK" },
	{ 0x42DD , "HANGUL SYLLABLE SSANGCIEUC YU NIEUN" },
	{ 0x42DE , "HANGUL SYLLABLE SSANGCIEUC YU RIEUL" },
	{ 0x42DF , "HANGUL SYLLABLE SSANGCIEUC YU MIEUM" },
	{ 0x42E0 , "HANGUL SYLLABLE SSANGCIEUC YU PIEUP" },
	{ 0x42E1 , "HANGUL SYLLABLE SSANGCIEUC YU SIOS" },
	{ 0x42E2 , "HANGUL SYLLABLE SSANGCIEUC YU IEUNG" },
	{ 0x42E3 , "HANGUL SYLLABLE SSANGCIEUC EU KIYEOK" },
	{ 0x42E4 , "HANGUL SYLLABLE SSANGCIEUC EU NIEUN" },
	{ 0x42E5 , "HANGUL SYLLABLE SSANGCIEUC EU TIKEUT" },
	{ 0x42E6 , "HANGUL SYLLABLE SSANGCIEUC EU RIEUL" },
	{ 0x42E7 , "HANGUL SYLLABLE SSANGCIEUC EU PIEUP" },
	{ 0x42E8 , "HANGUL SYLLABLE SSANGCIEUC EU CIEUC" },
	{ 0x42E9 , "HANGUL SYLLABLE SSANGCIEUC YI" },
	{ 0x42EA , "HANGUL SYLLABLE SSANGCIEUC YI KIYEOK" },
	{ 0x42EB , "HANGUL SYLLABLE SSANGCIEUC YI NIEUN" },
	{ 0x42EC , "HANGUL SYLLABLE SSANGCIEUC YI RIEUL" },
	{ 0x42ED , "HANGUL SYLLABLE SSANGCIEUC YI MIEUM" },
	{ 0x42EE , "HANGUL SYLLABLE SSANGCIEUC YI PIEUP" },
	{ 0x42EF , "HANGUL SYLLABLE SSANGCIEUC YI SIOS" },
	{ 0x42F0 , "HANGUL SYLLABLE SSANGCIEUC YI IEUNG" },
	{ 0x42F1 , "HANGUL SYLLABLE SSANGCIEUC YI CIEUC" },
	{ 0x42F2 , "HANGUL SYLLABLE SSANGCIEUC I TIKEUT" },
	{ 0x42F3 , "HANGUL SYLLABLE SSANGCIEUC I SIOS" },
	{ 0x42F4 , "HANGUL SYLLABLE SSANGCIEUC I PHIEUPH" },
	{ 0x42F5 , "HANGUL SYLLABLE CHIEUCH A TIKEUT" },
	{ 0x42F6 , "HANGUL SYLLABLE CHIEUCH A RIEUL-KIYEOK" },
	{ 0x42F7 , "HANGUL SYLLABLE CHIEUCH A CHIEUCH" },
	{ 0x42F8 , "HANGUL SYLLABLE CHIEUCH A PHIEUPH" },
	{ 0x42F9 , "HANGUL SYLLABLE CHIEUCH AE TIKEUT" },
	{ 0x42FA , "HANGUL SYLLABLE CHIEUCH AE CHIEUCH" },
	{ 0x42FB , "HANGUL SYLLABLE CHIEUCH AE PHIEUPH" },
	{ 0x42FC , "HANGUL SYLLABLE CHIEUCH YA KIYEOK" },
	{ 0x42FD , "HANGUL SYLLABLE CHIEUCH YA PIEUP" },
	{ 0x42FE , "HANGUL SYLLABLE CHIEUCH YA SIOS" },
	{ 0x42FF , "HANGUL SYLLABLE CHIEUCH YA PHIEUPH" },
	{ 0x4300 , "HANGUL SYLLABLE CHIEUCH YAE" },
	{ 0x4301 , "HANGUL SYLLABLE CHIEUCH YAE IEUNG" },
	{ 0x4302 , "HANGUL SYLLABLE CHIEUCH EO TIKEUT" },
	{ 0x4303 , "HANGUL SYLLABLE CHIEUCH EO RIEUL-HIEUH" },
	{ 0x4304 , "HANGUL SYLLABLE CHIEUCH EO CIEUC" },
	{ 0x4305 , "HANGUL SYLLABLE CHIEUCH EO PHIEUPH" },
	{ 0x4306 , "HANGUL SYLLABLE CHIEUCH E TIKEUT" },
	{ 0x4307 , "HANGUL SYLLABLE CHIEUCH E RIEUL-HIEUH" },
	{ 0x4308 , "HANGUL SYLLABLE CHIEUCH E SSANGSIOS" },
	{ 0x4309 , "HANGUL SYLLABLE CHIEUCH YEO KIYEOK" },
	{ 0x430A , "HANGUL SYLLABLE CHIEUCH YEO TIKEUT" },
	{ 0x430B , "HANGUL SYLLABLE CHIEUCH YEO RIEUL" },
	{ 0x430C , "HANGUL SYLLABLE CHIEUCH YEO MIEUM" },
	{ 0x430D , "HANGUL SYLLABLE CHIEUCH YEO PIEUP" },
	{ 0x430E , "HANGUL SYLLABLE CHIEUCH YEO SIOS" },
	{ 0x430F , "HANGUL SYLLABLE CHIEUCH YEO IEUNG" },
	{ 0x4310 , "HANGUL SYLLABLE CHIEUCH YE KIYEOK" },
	{ 0x4311 , "HANGUL SYLLABLE CHIEUCH YE RIEUL" },
	{ 0x4312 , "HANGUL SYLLABLE CHIEUCH YE MIEUM" },
	{ 0x4313 , "HANGUL SYLLABLE CHIEUCH YE PIEUP" },
	{ 0x4314 , "HANGUL SYLLABLE CHIEUCH YE SIOS" },
	{ 0x4315 , "HANGUL SYLLABLE CHIEUCH O TIKEUT" },
	{ 0x4316 , "HANGUL SYLLABLE CHIEUCH WA KIYEOK" },
	{ 0x4317 , "HANGUL SYLLABLE CHIEUCH WA TIKEUT" },
	{ 0x4318 , "HANGUL SYLLABLE CHIEUCH WA MIEUM" },
	{ 0x4319 , "HANGUL SYLLABLE CHIEUCH WA PIEUP" },
	{ 0x431A , "HANGUL SYLLABLE CHIEUCH WA SIOS" },
	{ 0x431B , "HANGUL SYLLABLE CHIEUCH WAE" },
	{ 0x431C , "HANGUL SYLLABLE CHIEUCH WAE SIOS" },
	{ 0x431D , "HANGUL SYLLABLE CHIEUCH WAE SSANGSIOS" },
	{ 0x431E , "HANGUL SYLLABLE CHIEUCH OE KIYEOK" },
	{ 0x431F , "HANGUL SYLLABLE CHIEUCH OE TIKEUT" },
	{ 0x4320 , "HANGUL SYLLABLE CHIEUCH OE SSANGSIOS" },
	{ 0x4321 , "HANGUL SYLLABLE CHIEUCH YO KIYEOK" },
	{ 0x4322 , "HANGUL SYLLABLE CHIEUCH YO NIEUN" },
	{ 0x4323 , "HANGUL SYLLABLE CHIEUCH YO RIEUL" },
	{ 0x4324 , "HANGUL SYLLABLE CHIEUCH YO PIEUP" },
	{ 0x4325 , "HANGUL SYLLABLE CHIEUCH YO SIOS" },
	{ 0x4326 , "HANGUL SYLLABLE CHIEUCH YO IEUNG" },
	{ 0x4327 , "HANGUL SYLLABLE CHIEUCH U TIKEUT" },
	{ 0x4328 , "HANGUL SYLLABLE CHIEUCH WEO KIYEOK" },
	{ 0x4329 , "HANGUL SYLLABLE CHIEUCH WEO NIEUN" },
	{ 0x432A , "HANGUL SYLLABLE CHIEUCH WEO TIKEUT" },
	{ 0x432B , "HANGUL SYLLABLE CHIEUCH WEO RIEUL" },
	{ 0x432C , "HANGUL SYLLABLE CHIEUCH WEO MIEUM" },
	{ 0x432D , "HANGUL SYLLABLE CHIEUCH WEO PIEUP" },
	{ 0x432E , "HANGUL SYLLABLE CHIEUCH WEO SIOS" },
	{ 0x432F , "HANGUL SYLLABLE CHIEUCH WEO IEUNG" },
	{ 0x4330 , "HANGUL SYLLABLE CHIEUCH WE KIYEOK" },
	{ 0x4331 , "HANGUL SYLLABLE CHIEUCH WE RIEUL" },
	{ 0x4332 , "HANGUL SYLLABLE CHIEUCH WE MIEUM" },
	{ 0x4333 , "HANGUL SYLLABLE CHIEUCH WE PIEUP" },
	{ 0x4334 , "HANGUL SYLLABLE CHIEUCH WE SIOS" },
	{ 0x4335 , "HANGUL SYLLABLE CHIEUCH WE SSANGSIOS" },
	{ 0x4336 , "HANGUL SYLLABLE CHIEUCH WE IEUNG" },
	{ 0x4337 , "HANGUL SYLLABLE CHIEUCH WI KIYEOK" },
	{ 0x4338 , "HANGUL SYLLABLE CHIEUCH WI TIKEUT" },
	{ 0x4339 , "HANGUL SYLLABLE CHIEUCH YU KIYEOK" },
	{ 0x433A , "HANGUL SYLLABLE CHIEUCH YU PIEUP" },
	{ 0x433B , "HANGUL SYLLABLE CHIEUCH YU SIOS" },
	{ 0x433C , "HANGUL SYLLABLE CHIEUCH EU TIKEUT" },
	{ 0x433D , "HANGUL SYLLABLE CHIEUCH EU RIEUL-KIYEOK" },
	{ 0x433E , "HANGUL SYLLABLE CHIEUCH EU RIEUL-SIOS" },
	{ 0x433F , "HANGUL SYLLABLE CHIEUCH YI" },
	{ 0x4340 , "HANGUL SYLLABLE CHIEUCH YI KIYEOK" },
	{ 0x4341 , "HANGUL SYLLABLE CHIEUCH YI NIEUN" },
	{ 0x4342 , "HANGUL SYLLABLE CHIEUCH YI RIEUL" },
	{ 0x4343 , "HANGUL SYLLABLE CHIEUCH YI MIEUM" },
	{ 0x4344 , "HANGUL SYLLABLE CHIEUCH YI PIEUP" },
	{ 0x4345 , "HANGUL SYLLABLE CHIEUCH YI SIOS" },
	{ 0x4346 , "HANGUL SYLLABLE CHIEUCH YI IEUNG" },
	{ 0x4347 , "HANGUL SYLLABLE CHIEUCH I RIEUL-MIEUM" },
	{ 0x4348 , "HANGUL SYLLABLE CHIEUCH I SSANGSIOS" },
	{ 0x4349 , "HANGUL SYLLABLE CHIEUCH I CIEUC" },
	{ 0x434A , "HANGUL SYLLABLE CHIEUCH I KHIEUKH" },
	{ 0x434B , "HANGUL SYLLABLE CHIEUCH I PHIEUPH" },
	{ 0x434C , "HANGUL SYLLABLE KHIEUKH A TIKEUT" },
	{ 0x434D , "HANGUL SYLLABLE KHIEUKH A SSANGSIOS" },
	{ 0x434E , "HANGUL SYLLABLE KHIEUKH A THIEUTH" },
	{ 0x434F , "HANGUL SYLLABLE KHIEUKH A PHIEUPH" },
	{ 0x4350 , "HANGUL SYLLABLE KHIEUKH AE TIKEUT" },
	{ 0x4351 , "HANGUL SYLLABLE KHIEUKH AE THIEUTH" },
	{ 0x4352 , "HANGUL SYLLABLE KHIEUKH AE PHIEUPH" },
	{ 0x4353 , "HANGUL SYLLABLE KHIEUKH YA NIEUN" },
	{ 0x4354 , "HANGUL SYLLABLE KHIEUKH YA RIEUL" },
	{ 0x4355 , "HANGUL SYLLABLE KHIEUKH YA MIEUM" },
	{ 0x4356 , "HANGUL SYLLABLE KHIEUKH YA PIEUP" },
	{ 0x4357 , "HANGUL SYLLABLE KHIEUKH YA SIOS" },
	{ 0x4358 , "HANGUL SYLLABLE KHIEUKH YAE" },
	{ 0x4359 , "HANGUL SYLLABLE KHIEUKH YAE NIEUN" },
	{ 0x435A , "HANGUL SYLLABLE KHIEUKH EO CHIEUCH" },
	{ 0x435B , "HANGUL SYLLABLE KHIEUKH EO THIEUTH" },
	{ 0x435C , "HANGUL SYLLABLE KHIEUKH EO PHIEUPH" },
	{ 0x435D , "HANGUL SYLLABLE KHIEUKH E TIKEUT" },
	{ 0x435E , "HANGUL SYLLABLE KHIEUKH E SSANGSIOS" },
	{ 0x435F , "HANGUL SYLLABLE KHIEUKH E THIEUTH" },
	{ 0x4360 , "HANGUL SYLLABLE KHIEUKH E PHIEUPH" },
	{ 0x4361 , "HANGUL SYLLABLE KHIEUKH YEO KIYEOK" },
	{ 0x4362 , "HANGUL SYLLABLE KHIEUKH YEO TIKEUT" },
	{ 0x4363 , "HANGUL SYLLABLE KHIEUKH YE KIYEOK" },
	{ 0x4364 , "HANGUL SYLLABLE KHIEUKH YE NIEUN" },
	{ 0x4365 , "HANGUL SYLLABLE KHIEUKH YE RIEUL" },
	{ 0x4366 , "HANGUL SYLLABLE KHIEUKH YE MIEUM" },
	{ 0x4367 , "HANGUL SYLLABLE KHIEUKH YE PIEUP" },
	{ 0x4368 , "HANGUL SYLLABLE KHIEUKH YE SIOS" },
	{ 0x4369 , "HANGUL SYLLABLE KHIEUKH YE SSANGSIOS" },
	{ 0x436A , "HANGUL SYLLABLE KHIEUKH YE IEUNG" },
	{ 0x436B , "HANGUL SYLLABLE KHIEUKH O TIKEUT" },
	{ 0x436C , "HANGUL SYLLABLE KHIEUKH O CHIEUCH" },
	{ 0x436D , "HANGUL SYLLABLE KHIEUKH O PHIEUPH" },
	{ 0x436E , "HANGUL SYLLABLE KHIEUKH WA PIEUP" },
	{ 0x436F , "HANGUL SYLLABLE KHIEUKH WA SIOS" },
	{ 0x4370 , "HANGUL SYLLABLE KHIEUKH WAE KIYEOK" },
	{ 0x4371 , "HANGUL SYLLABLE KHIEUKH WAE NIEUN" },
	{ 0x4372 , "HANGUL SYLLABLE KHIEUKH WAE TIKEUT" },
	{ 0x4373 , "HANGUL SYLLABLE KHIEUKH WAE RIEUL" },
	{ 0x4374 , "HANGUL SYLLABLE KHIEUKH WAE MIEUM" },
	{ 0x4375 , "HANGUL SYLLABLE KHIEUKH WAE PIEUP" },
	{ 0x4376 , "HANGUL SYLLABLE KHIEUKH WAE SIOS" },
	{ 0x4377 , "HANGUL SYLLABLE KHIEUKH OE KIYEOK" },
	{ 0x4378 , "HANGUL SYLLABLE KHIEUKH OE NIEUN" },
	{ 0x4379 , "HANGUL SYLLABLE KHIEUKH OE MIEUM" },
	{ 0x437A , "HANGUL SYLLABLE KHIEUKH OE PIEUP" },
	{ 0x437B , "HANGUL SYLLABLE KHIEUKH OE SIOS" },
	{ 0x437C , "HANGUL SYLLABLE KHIEUKH OE IEUNG" },
	{ 0x437D , "HANGUL SYLLABLE KHIEUKH YO KIYEOK" },
	{ 0x437E , "HANGUL SYLLABLE KHIEUKH YO NIEUN" },
	{ 0x437F , "HANGUL SYLLABLE KHIEUKH YO RIEUL" },
	{ 0x4380 , "HANGUL SYLLABLE KHIEUKH YO MIEUM" },
	{ 0x4381 , "HANGUL SYLLABLE KHIEUKH YO PIEUP" },
	{ 0x4382 , "HANGUL SYLLABLE KHIEUKH YO SIOS" },
	{ 0x4383 , "HANGUL SYLLABLE KHIEUKH YO IEUNG" },
	{ 0x4384 , "HANGUL SYLLABLE KHIEUKH U TIKEUT" },
	{ 0x4385 , "HANGUL SYLLABLE KHIEUKH U PHIEUPH" },
	{ 0x4386 , "HANGUL SYLLABLE KHIEUKH WEO KIYEOK" },
	{ 0x4387 , "HANGUL SYLLABLE KHIEUKH WEO TIKEUT" },
	{ 0x4388 , "HANGUL SYLLABLE KHIEUKH WEO MIEUM" },
	{ 0x4389 , "HANGUL SYLLABLE KHIEUKH WEO PIEUP" },
	{ 0x438A , "HANGUL SYLLABLE KHIEUKH WEO SSANGSIOS" },
	{ 0x438B , "HANGUL SYLLABLE KHIEUKH WE KIYEOK" },
	{ 0x438C , "HANGUL SYLLABLE KHIEUKH WE NIEUN" },
	{ 0x438D , "HANGUL SYLLABLE KHIEUKH WE RIEUL" },
	{ 0x438E , "HANGUL SYLLABLE KHIEUKH WE MIEUM" },
	{ 0x438F , "HANGUL SYLLABLE KHIEUKH WE PIEUP" },
	{ 0x4390 , "HANGUL SYLLABLE KHIEUKH WE SIOS" },
	{ 0x4391 , "HANGUL SYLLABLE KHIEUKH WE SSANGSIOS" },
	{ 0x4392 , "HANGUL SYLLABLE KHIEUKH WI TIKEUT" },
	{ 0x4393 , "HANGUL SYLLABLE KHIEUKH YU KIYEOK" },
	{ 0x4394 , "HANGUL SYLLABLE KHIEUKH YU PIEUP" },
	{ 0x4395 , "HANGUL SYLLABLE KHIEUKH YU SIOS" },
	{ 0x4396 , "HANGUL SYLLABLE KHIEUKH YU IEUNG" },
	{ 0x4397 , "HANGUL SYLLABLE KHIEUKH YU KHIEUKH" },
	{ 0x4398 , "HANGUL SYLLABLE KHIEUKH EU NIEUN-HIEUH" },
	{ 0x4399 , "HANGUL SYLLABLE KHIEUKH EU RIEUL-MIEUM" },
	{ 0x439A , "HANGUL SYLLABLE KHIEUKH EU SIOS" },
	{ 0x439B , "HANGUL SYLLABLE KHIEUKH YI" },
	{ 0x439C , "HANGUL SYLLABLE KHIEUKH YI KIYEOK" },
	{ 0x439D , "HANGUL SYLLABLE KHIEUKH YI NIEUN" },
	{ 0x439E , "HANGUL SYLLABLE KHIEUKH YI RIEUL" },
	{ 0x439F , "HANGUL SYLLABLE KHIEUKH YI MIEUM" },
	{ 0x43A0 , "HANGUL SYLLABLE KHIEUKH YI PIEUP" },
	{ 0x43A1 , "HANGUL SYLLABLE KHIEUKH YI SIOS" },
	{ 0x43A2 , "HANGUL SYLLABLE KHIEUKH YI IEUNG" },
	{ 0x43A3 , "HANGUL SYLLABLE KHIEUKH I TIKEUT" },
	{ 0x43A4 , "HANGUL SYLLABLE KHIEUKH I SSANGSIOS" },
	{ 0x43A5 , "HANGUL SYLLABLE KHIEUKH I PHIEUPH" },
	{ 0x43A6 , "HANGUL SYLLABLE THIEUTH A TIKEUT" },
	{ 0x43A7 , "HANGUL SYLLABLE THIEUTH A RIEUL-MIEUM" },
	{ 0x43A8 , "HANGUL SYLLABLE THIEUTH A RIEUL-HIEUH" },
	{ 0x43A9 , "HANGUL SYLLABLE THIEUTH A PHIEUPH" },
	{ 0x43AA , "HANGUL SYLLABLE THIEUTH AE NIEUN-HIEUH" },
	{ 0x43AB , "HANGUL SYLLABLE THIEUTH AE TIKEUT" },
	{ 0x43AC , "HANGUL SYLLABLE THIEUTH AE RIEUL-KIYEOK" },
	{ 0x43AD , "HANGUL SYLLABLE THIEUTH AE PHIEUPH" },
	{ 0x43AE , "HANGUL SYLLABLE THIEUTH YA KIYEOK" },
	{ 0x43AF , "HANGUL SYLLABLE THIEUTH YA NIEUN" },
	{ 0x43B0 , "HANGUL SYLLABLE THIEUTH YA RIEUL" },
	{ 0x43B1 , "HANGUL SYLLABLE THIEUTH YA MIEUM" },
	{ 0x43B2 , "HANGUL SYLLABLE THIEUTH YA PIEUP" },
	{ 0x43B3 , "HANGUL SYLLABLE THIEUTH YA SIOS" },
	{ 0x43B4 , "HANGUL SYLLABLE THIEUTH YAE" },
	{ 0x43B5 , "HANGUL SYLLABLE THIEUTH YAE NIEUN" },
	{ 0x43B6 , "HANGUL SYLLABLE THIEUTH YAE NIEUN-HIEUH" },
	{ 0x43B7 , "HANGUL SYLLABLE THIEUTH EO TIKEUT" },
	{ 0x43B8 , "HANGUL SYLLABLE THIEUTH EO RIEUL-KIYEOK" },
	{ 0x43B9 , "HANGUL SYLLABLE THIEUTH EO RIEUL-PIEUP" },
	{ 0x43BA , "HANGUL SYLLABLE THIEUTH E TIKEUT" },
	{ 0x43BB , "HANGUL SYLLABLE THIEUTH E SSANGSIOS" },
	{ 0x43BC , "HANGUL SYLLABLE THIEUTH E PHIEUPH" },
	{ 0x43BD , "HANGUL SYLLABLE THIEUTH YEO KIYEOK" },
	{ 0x43BE , "HANGUL SYLLABLE THIEUTH YEO TIKEUT" },
	{ 0x43BF , "HANGUL SYLLABLE THIEUTH YEO RIEUL" },
	{ 0x43C0 , "HANGUL SYLLABLE THIEUTH YEO MIEUM" },
	{ 0x43C1 , "HANGUL SYLLABLE THIEUTH YEO PIEUP" },
	{ 0x43C2 , "HANGUL SYLLABLE THIEUTH YEO SIOS" },
	{ 0x43C3 , "HANGUL SYLLABLE THIEUTH YEO IEUNG" },
	{ 0x43C4 , "HANGUL SYLLABLE THIEUTH YE KIYEOK" },
	{ 0x43C5 , "HANGUL SYLLABLE THIEUTH YE RIEUL" },
	{ 0x43C6 , "HANGUL SYLLABLE THIEUTH YE MIEUM" },
	{ 0x43C7 , "HANGUL SYLLABLE THIEUTH YE SIOS" },
	{ 0x43C8 , "HANGUL SYLLABLE THIEUTH YE IEUNG" },
	{ 0x43C9 , "HANGUL SYLLABLE THIEUTH O TIKEUT" },
	{ 0x43CA , "HANGUL SYLLABLE THIEUTH O CHIEUCH" },
	{ 0x43CB , "HANGUL SYLLABLE THIEUTH O KHIEUKH" },
	{ 0x43CC , "HANGUL SYLLABLE THIEUTH O THIEUTH" },
	{ 0x43CD , "HANGUL SYLLABLE THIEUTH WA KIYEOK" },
	{ 0x43CE , "HANGUL SYLLABLE THIEUTH WA RIEUL" },
	{ 0x43CF , "HANGUL SYLLABLE THIEUTH WA MIEUM" },
	{ 0x43D0 , "HANGUL SYLLABLE THIEUTH WA PIEUP" },
	{ 0x43D1 , "HANGUL SYLLABLE THIEUTH WA SIOS" },
	{ 0x43D2 , "HANGUL SYLLABLE THIEUTH WA IEUNG" },
	{ 0x43D3 , "HANGUL SYLLABLE THIEUTH WAE SSANGSIOS" },
	{ 0x43D4 , "HANGUL SYLLABLE THIEUTH WAE IEUNG" },
	{ 0x43D5 , "HANGUL SYLLABLE THIEUTH OE KIYEOK" },
	{ 0x43D6 , "HANGUL SYLLABLE THIEUTH OE TIKEUT" },
	{ 0x43D7 , "HANGUL SYLLABLE THIEUTH OE RIEUL" },
	{ 0x43D8 , "HANGUL SYLLABLE THIEUTH OE MIEUM" },
	{ 0x43D9 , "HANGUL SYLLABLE THIEUTH OE PIEUP" },
	{ 0x43DA , "HANGUL SYLLABLE THIEUTH OE SSANGSIOS" },
	{ 0x43DB , "HANGUL SYLLABLE THIEUTH YO KIYEOK" },
	{ 0x43DC , "HANGUL SYLLABLE THIEUTH YO NIEUN" },
	{ 0x43DD , "HANGUL SYLLABLE THIEUTH YO RIEUL" },
	{ 0x43DE , "HANGUL SYLLABLE THIEUTH YO MIEUM" },
	{ 0x43DF , "HANGUL SYLLABLE THIEUTH YO PIEUP" },
	{ 0x43E0 , "HANGUL SYLLABLE THIEUTH YO SIOS" },
	{ 0x43E1 , "HANGUL SYLLABLE THIEUTH YO IEUNG" },
	{ 0x43E2 , "HANGUL SYLLABLE THIEUTH U TIKEUT" },
	{ 0x43E3 , "HANGUL SYLLABLE THIEUTH U RIEUL-MIEUM" },
	{ 0x43E4 , "HANGUL SYLLABLE THIEUTH WEO KIYEOK" },
	{ 0x43E5 , "HANGUL SYLLABLE THIEUTH WEO NIEUN" },
	{ 0x43E6 , "HANGUL SYLLABLE THIEUTH WEO TIKEUT" },
	{ 0x43E7 , "HANGUL SYLLABLE THIEUTH WEO RIEUL" },
	{ 0x43E8 , "HANGUL SYLLABLE THIEUTH WEO RIEUL-SIOS" },
	{ 0x43E9 , "HANGUL SYLLABLE THIEUTH WEO MIEUM" },
	{ 0x43EA , "HANGUL SYLLABLE THIEUTH WEO PIEUP" },
	{ 0x43EB , "HANGUL SYLLABLE THIEUTH WEO SIOS" },
	{ 0x43EC , "HANGUL SYLLABLE THIEUTH WEO IEUNG" },
	{ 0x43ED , "HANGUL SYLLABLE THIEUTH WE KIYEOK" },
	{ 0x43EE , "HANGUL SYLLABLE THIEUTH WE NIEUN" },
	{ 0x43EF , "HANGUL SYLLABLE THIEUTH WE TIKEUT" },
	{ 0x43F0 , "HANGUL SYLLABLE THIEUTH WE RIEUL" },
	{ 0x43F1 , "HANGUL SYLLABLE THIEUTH WE MIEUM" },
	{ 0x43F2 , "HANGUL SYLLABLE THIEUTH WE PIEUP" },
	{ 0x43F3 , "HANGUL SYLLABLE THIEUTH WE SIOS" },
	{ 0x43F4 , "HANGUL SYLLABLE THIEUTH WE SSANGSIOS" },
	{ 0x43F5 , "HANGUL SYLLABLE THIEUTH WE IEUNG" },
	{ 0x43F6 , "HANGUL SYLLABLE THIEUTH WI TIKEUT" },
	{ 0x43F7 , "HANGUL SYLLABLE THIEUTH WI SIOS" },
	{ 0x43F8 , "HANGUL SYLLABLE THIEUTH YU KIYEOK" },
	{ 0x43F9 , "HANGUL SYLLABLE THIEUTH YU PIEUP" },
	{ 0x43FA , "HANGUL SYLLABLE THIEUTH YU SIOS" },
	{ 0x43FB , "HANGUL SYLLABLE THIEUTH EU KIYEOK-SIOS" },
	{ 0x43FC , "HANGUL SYLLABLE THIEUTH EU RIEUL-KIYEOK" },
	{ 0x43FD , "HANGUL SYLLABLE THIEUTH EU RIEUL-SIOS" },
	{ 0x43FE , "HANGUL SYLLABLE THIEUTH EU IEUNG" },
	{ 0x43FF , "HANGUL SYLLABLE THIEUTH YI KIYEOK" },
	{ 0x4400 , "HANGUL SYLLABLE THIEUTH YI RIEUL-SIOS" },
	{ 0x4401 , "HANGUL SYLLABLE THIEUTH YI SIOS" },
	{ 0x4402 , "HANGUL SYLLABLE THIEUTH YI IEUNG" },
	{ 0x4403 , "HANGUL SYLLABLE THIEUTH I TIKEUT" },
	{ 0x4404 , "HANGUL SYLLABLE THIEUTH I KHIEUKH" },
	{ 0x4405 , "HANGUL SYLLABLE THIEUTH I PHIEUPH" },
	{ 0x4406 , "HANGUL SYLLABLE PHIEUPH A TIKEUT" },
	{ 0x4407 , "HANGUL SYLLABLE PHIEUPH A CHIEUCH" },
	{ 0x4408 , "HANGUL SYLLABLE PHIEUPH A PHIEUPH" },
	{ 0x4409 , "HANGUL SYLLABLE PHIEUPH AE TIKEUT" },
	{ 0x440A , "HANGUL SYLLABLE PHIEUPH AE THIEUTH" },
	{ 0x440B , "HANGUL SYLLABLE PHIEUPH YA NIEUN" },
	{ 0x440C , "HANGUL SYLLABLE PHIEUPH YA RIEUL" },
	{ 0x440D , "HANGUL SYLLABLE PHIEUPH YA MIEUM" },
	{ 0x440E , "HANGUL SYLLABLE PHIEUPH YA PIEUP" },
	{ 0x440F , "HANGUL SYLLABLE PHIEUPH YA SIOS" },
	{ 0x4410 , "HANGUL SYLLABLE PHIEUPH YA IEUNG" },
	{ 0x4411 , "HANGUL SYLLABLE PHIEUPH YAE" },
	{ 0x4412 , "HANGUL SYLLABLE PHIEUPH EO TIKEUT" },
	{ 0x4413 , "HANGUL SYLLABLE PHIEUPH EO PIEUP-SIOS" },
	{ 0x4414 , "HANGUL SYLLABLE PHIEUPH EO THIEUTH" },
	{ 0x4415 , "HANGUL SYLLABLE PHIEUPH E TIKEUT" },
	{ 0x4416 , "HANGUL SYLLABLE PHIEUPH E SSANGSIOS" },
	{ 0x4417 , "HANGUL SYLLABLE PHIEUPH E PHIEUPH" },
	{ 0x4418 , "HANGUL SYLLABLE PHIEUPH YEO KIYEOK" },
	{ 0x4419 , "HANGUL SYLLABLE PHIEUPH YEO NIEUN-CIEUC" },
	{ 0x441A , "HANGUL SYLLABLE PHIEUPH YEO TIKEUT" },
	{ 0x441B , "HANGUL SYLLABLE PHIEUPH YEO SIOS" },
	{ 0x441C , "HANGUL SYLLABLE PHIEUPH YE KIYEOK" },
	{ 0x441D , "HANGUL SYLLABLE PHIEUPH YE NIEUN" },
	{ 0x441E , "HANGUL SYLLABLE PHIEUPH YE TIKEUT" },
	{ 0x441F , "HANGUL SYLLABLE PHIEUPH YE MIEUM" },
	{ 0x4420 , "HANGUL SYLLABLE PHIEUPH YE IEUNG" },
	{ 0x4421 , "HANGUL SYLLABLE PHIEUPH O SSANGKIYEOK" },
	{ 0x4422 , "HANGUL SYLLABLE PHIEUPH O TIKEUT" },
	{ 0x4423 , "HANGUL SYLLABLE PHIEUPH O KHIEUKH" },
	{ 0x4424 , "HANGUL SYLLABLE PHIEUPH O THIEUTH" },
	{ 0x4425 , "HANGUL SYLLABLE PHIEUPH O PHIEUPH" },
	{ 0x4426 , "HANGUL SYLLABLE PHIEUPH WA KIYEOK" },
	{ 0x4427 , "HANGUL SYLLABLE PHIEUPH WA NIEUN" },
	{ 0x4428 , "HANGUL SYLLABLE PHIEUPH WA RIEUL" },
	{ 0x4429 , "HANGUL SYLLABLE PHIEUPH WA MIEUM" },
	{ 0x442A , "HANGUL SYLLABLE PHIEUPH WA PIEUP" },
	{ 0x442B , "HANGUL SYLLABLE PHIEUPH WA SIOS" },
	{ 0x442C , "HANGUL SYLLABLE PHIEUPH WAE" },
	{ 0x442D , "HANGUL SYLLABLE PHIEUPH OE KIYEOK" },
	{ 0x442E , "HANGUL SYLLABLE PHIEUPH OE RIEUL" },
	{ 0x442F , "HANGUL SYLLABLE PHIEUPH OE MIEUM" },
	{ 0x4430 , "HANGUL SYLLABLE PHIEUPH OE PIEUP" },
	{ 0x4431 , "HANGUL SYLLABLE PHIEUPH OE SIOS" },
	{ 0x4432 , "HANGUL SYLLABLE PHIEUPH OE IEUNG" },
	{ 0x4433 , "HANGUL SYLLABLE PHIEUPH YO KIYEOK" },
	{ 0x4434 , "HANGUL SYLLABLE PHIEUPH YO TIKEUT" },
	{ 0x4435 , "HANGUL SYLLABLE PHIEUPH YO MIEUM" },
	{ 0x4436 , "HANGUL SYLLABLE PHIEUPH YO IEUNG" },
	{ 0x4437 , "HANGUL SYLLABLE PHIEUPH WEO KIYEOK" },
	{ 0x4438 , "HANGUL SYLLABLE PHIEUPH WEO NIEUN" },
	{ 0x4439 , "HANGUL SYLLABLE PHIEUPH WEO TIKEUT" },
	{ 0x443A , "HANGUL SYLLABLE PHIEUPH WEO RIEUL" },
	{ 0x443B , "HANGUL SYLLABLE PHIEUPH WEO MIEUM" },
	{ 0x443C , "HANGUL SYLLABLE PHIEUPH WEO PIEUP" },
	{ 0x443D , "HANGUL SYLLABLE PHIEUPH WEO SIOS" },
	{ 0x443E , "HANGUL SYLLABLE PHIEUPH WEO SSANGSIOS" },
	{ 0x443F , "HANGUL SYLLABLE PHIEUPH WE" },
	{ 0x4440 , "HANGUL SYLLABLE PHIEUPH WE KIYEOK" },
	{ 0x4441 , "HANGUL SYLLABLE PHIEUPH WE NIEUN" },
	{ 0x4442 , "HANGUL SYLLABLE PHIEUPH WE RIEUL" },
	{ 0x4443 , "HANGUL SYLLABLE PHIEUPH WE MIEUM" },
	{ 0x4444 , "HANGUL SYLLABLE PHIEUPH WE PIEUP" },
	{ 0x4445 , "HANGUL SYLLABLE PHIEUPH WE SIOS" },
	{ 0x4446 , "HANGUL SYLLABLE PHIEUPH WE IEUNG" },
	{ 0x4447 , "HANGUL SYLLABLE PHIEUPH WI KIYEOK" },
	{ 0x4448 , "HANGUL SYLLABLE PHIEUPH WI PIEUP" },
	{ 0x4449 , "HANGUL SYLLABLE PHIEUPH WI IEUNG" },
	{ 0x444A , "HANGUL SYLLABLE PHIEUPH YU KIYEOK" },
	{ 0x444B , "HANGUL SYLLABLE PHIEUPH YU TIKEUT" },
	{ 0x444C , "HANGUL SYLLABLE PHIEUPH YU PIEUP" },
	{ 0x444D , "HANGUL SYLLABLE PHIEUPH EU KIYEOK" },
	{ 0x444E , "HANGUL SYLLABLE PHIEUPH EU TIKEUT" },
	{ 0x444F , "HANGUL SYLLABLE PHIEUPH EU RIEUL-SIOS" },
	{ 0x4450 , "HANGUL SYLLABLE PHIEUPH EU IEUNG" },
	{ 0x4451 , "HANGUL SYLLABLE PHIEUPH YI" },
	{ 0x4452 , "HANGUL SYLLABLE PHIEUPH YI KIYEOK" },
	{ 0x4453 , "HANGUL SYLLABLE PHIEUPH YI NIEUN" },
	{ 0x4454 , "HANGUL SYLLABLE PHIEUPH YI RIEUL" },
	{ 0x4455 , "HANGUL SYLLABLE PHIEUPH YI MIEUM" },
	{ 0x4456 , "HANGUL SYLLABLE PHIEUPH YI PIEUP" },
	{ 0x4457 , "HANGUL SYLLABLE PHIEUPH YI SIOS" },
	{ 0x4458 , "HANGUL SYLLABLE PHIEUPH YI IEUNG" },
	{ 0x4459 , "HANGUL SYLLABLE PHIEUPH I TIKEUT" },
	{ 0x445A , "HANGUL SYLLABLE PHIEUPH I THIEUTH" },
	{ 0x445B , "HANGUL SYLLABLE PHIEUPH I PHIEUPH" },
	{ 0x445C , "HANGUL SYLLABLE HIEUH A TIKEUT" },
	{ 0x445D , "HANGUL SYLLABLE HIEUH A RIEUL-KIYEOK" },
	{ 0x445E , "HANGUL SYLLABLE HIEUH A RIEUL-PIEUP" },
	{ 0x445F , "HANGUL SYLLABLE HIEUH A RIEUL-SIOS" },
	{ 0x4460 , "HANGUL SYLLABLE HIEUH A RIEUL-HIEUH" },
	{ 0x4461 , "HANGUL SYLLABLE HIEUH A SSANGSIOS" },
	{ 0x4462 , "HANGUL SYLLABLE HIEUH A CIEUC" },
	{ 0x4463 , "HANGUL SYLLABLE HIEUH A THIEUTH" },
	{ 0x4464 , "HANGUL SYLLABLE HIEUH AE TIKEUT" },
	{ 0x4465 , "HANGUL SYLLABLE HIEUH AE THIEUTH" },
	{ 0x4466 , "HANGUL SYLLABLE HIEUH YA KIYEOK" },
	{ 0x4467 , "HANGUL SYLLABLE HIEUH YA NIEUN" },
	{ 0x4468 , "HANGUL SYLLABLE HIEUH YA RIEUL" },
	{ 0x4469 , "HANGUL SYLLABLE HIEUH YA MIEUM" },
	{ 0x446A , "HANGUL SYLLABLE HIEUH YA PIEUP" },
	{ 0x446B , "HANGUL SYLLABLE HIEUH YA SIOS" },
	{ 0x446C , "HANGUL SYLLABLE HIEUH YA HIEUH" },
	{ 0x446D , "HANGUL SYLLABLE HIEUH YAE" },
	{ 0x446E , "HANGUL SYLLABLE HIEUH YAE SSANGSIOS" },
	{ 0x446F , "HANGUL SYLLABLE HIEUH EO TIKEUT" },
	{ 0x4470 , "HANGUL SYLLABLE HIEUH EO RIEUL-KIYEOK" },
	{ 0x4471 , "HANGUL SYLLABLE HIEUH EO RIEUL-HIEUH" },
	{ 0x4472 , "HANGUL SYLLABLE HIEUH EO SSANGSIOS" },
	{ 0x4473 , "HANGUL SYLLABLE HIEUH EO THIEUTH" },
	{ 0x4474 , "HANGUL SYLLABLE HIEUH EO PHIEUPH" },
	{ 0x4475 , "HANGUL SYLLABLE HIEUH E TIKEUT" },
	{ 0x4476 , "HANGUL SYLLABLE HIEUH E SSANGSIOS" },
	{ 0x4477 , "HANGUL SYLLABLE HIEUH E PHIEUPH" },
	{ 0x4478 , "HANGUL SYLLABLE HIEUH YEO TIKEUT" },
	{ 0x4479 , "HANGUL SYLLABLE HIEUH YE KIYEOK" },
	{ 0x447A , "HANGUL SYLLABLE HIEUH YE MIEUM" },
	{ 0x447B , "HANGUL SYLLABLE HIEUH YE SIOS" },
	{ 0x447C , "HANGUL SYLLABLE HIEUH YE SSANGSIOS" },
	{ 0x447D , "HANGUL SYLLABLE HIEUH YE IEUNG" },
	{ 0x447E , "HANGUL SYLLABLE HIEUH O NIEUN-CIEUC" },
	{ 0x447F , "HANGUL SYLLABLE HIEUH O TIKEUT" },
	{ 0x4480 , "HANGUL SYLLABLE HIEUH O RIEUL-KIYEOK" },
	{ 0x4481 , "HANGUL SYLLABLE HIEUH O RIEUL-SIOS" },
	{ 0x4482 , "HANGUL SYLLABLE HIEUH O RIEUL-HIEUH" },
	{ 0x4483 , "HANGUL SYLLABLE HIEUH WA TIKEUT" },
	{ 0x4484 , "HANGUL SYLLABLE HIEUH WA RIEUL-SIOS" },
	{ 0x4485 , "HANGUL SYLLABLE HIEUH WA MIEUM" },
	{ 0x4486 , "HANGUL SYLLABLE HIEUH WA PIEUP" },
	{ 0x4487 , "HANGUL SYLLABLE HIEUH WA PIEUP-SIOS" },
	{ 0x4488 , "HANGUL SYLLABLE HIEUH WA SSANGSIOS" },
	{ 0x4489 , "HANGUL SYLLABLE HIEUH WAE TIKEUT" },
	{ 0x448A , "HANGUL SYLLABLE HIEUH WAE RIEUL" },
	{ 0x448B , "HANGUL SYLLABLE HIEUH WAE MIEUM" },
	{ 0x448C , "HANGUL SYLLABLE HIEUH WAE PIEUP" },
	{ 0x448D , "HANGUL SYLLABLE HIEUH WAE SSANGSIOS" },
	{ 0x448E , "HANGUL SYLLABLE HIEUH OE TIKEUT" },
	{ 0x448F , "HANGUL SYLLABLE HIEUH OE MIEUM" },
	{ 0x4490 , "HANGUL SYLLABLE HIEUH YO KIYEOK" },
	{ 0x4491 , "HANGUL SYLLABLE HIEUH YO TIKEUT" },
	{ 0x4492 , "HANGUL SYLLABLE HIEUH YO MIEUM" },
	{ 0x4493 , "HANGUL SYLLABLE HIEUH YO IEUNG" },
	{ 0x4494 , "HANGUL SYLLABLE HIEUH U TIKEUT" },
	{ 0x4495 , "HANGUL SYLLABLE HIEUH U RIEUL-KIYEOK" },
	{ 0x4496 , "HANGUL SYLLABLE HIEUH U RIEUL-MIEUM" },
	{ 0x4497 , "HANGUL SYLLABLE HIEUH U RIEUL-PHIEUPH" },
	{ 0x4498 , "HANGUL SYLLABLE HIEUH U RIEUL-HIEUH" },
	{ 0x4499 , "HANGUL SYLLABLE HIEUH U PIEUP" },
	{ 0x449A , "HANGUL SYLLABLE HIEUH U KHIEUKH" },
	{ 0x449B , "HANGUL SYLLABLE HIEUH U THIEUTH" },
	{ 0x449C , "HANGUL SYLLABLE HIEUH WEO KIYEOK" },
	{ 0x449D , "HANGUL SYLLABLE HIEUH WEO PIEUP" },
	{ 0x449E , "HANGUL SYLLABLE HIEUH WEO SIOS" },
	{ 0x449F , "HANGUL SYLLABLE HIEUH WEO SSANGSIOS" },
	{ 0x44A0 , "HANGUL SYLLABLE HIEUH WE TIKEUT" },
	{ 0x44A1 , "HANGUL SYLLABLE HIEUH WE MIEUM" },
	{ 0x44A2 , "HANGUL SYLLABLE HIEUH WE PIEUP" },
	{ 0x44A3 , "HANGUL SYLLABLE HIEUH WE SIOS" },
	{ 0x44A4 , "HANGUL SYLLABLE HIEUH WE SSANGSIOS" },
	{ 0x44A5 , "HANGUL SYLLABLE HIEUH WI TIKEUT" },
	{ 0x44A6 , "HANGUL SYLLABLE HIEUH YU NIEUN-HIEUH" },
	{ 0x44A7 , "HANGUL SYLLABLE HIEUH YU TIKEUT" },
	{ 0x44A8 , "HANGUL SYLLABLE HIEUH YU RIEUL-HIEUH" },
	{ 0x44A9 , "HANGUL SYLLABLE HIEUH YU PIEUP" },
	{ 0x44AA , "HANGUL SYLLABLE HIEUH YU THIEUTH" },
	{ 0x44AB , "HANGUL SYLLABLE HIEUH EU NIEUN-CIEUC" },
	{ 0x44AC , "HANGUL SYLLABLE HIEUH EU RIEUL-THIEUTH" },
	{ 0x44AD , "HANGUL SYLLABLE HIEUH EU RIEUL-HIEUH" },
	{ 0x44AE , "HANGUL SYLLABLE HIEUH EU KHIEUKH" },
	{ 0x44AF , "HANGUL SYLLABLE HIEUH EU PHIEUPH" },
	{ 0x44B0 , "HANGUL SYLLABLE HIEUH EU HIEUH" },
	{ 0x44B1 , "HANGUL SYLLABLE HIEUH YI KIYEOK" },
	{ 0x44B2 , "HANGUL SYLLABLE HIEUH YI TIKEUT" },
	{ 0x44B3 , "HANGUL SYLLABLE HIEUH YI SIOS" },
	{ 0x44B4 , "HANGUL SYLLABLE HIEUH I TIKEUT" },
	{ 0x44B5 , "HANGUL SYLLABLE HIEUH I RIEUL-HIEUH" },
	{ 0x44B6 , "HANGUL SYLLABLE HIEUH I SSANGSIOS" },
	{ 0x44B7 , "HANGUL SYLLABLE HIEUH I KHIEUKH" },
	{ 0x44B8 , "HANGUL SYLLABLE MIEUM EU KIYEOK-SIOS" },
	{ 0x44B9 , "HANGUL SYLLABLE HIEUH O KIYEOK-SIOS" },
	{ 0x44BA , "HANGUL SYLLABLE SSANGPIEUP O THIEUTH" },
	{ 0x44BB , "HANGUL SYLLABLE SSANGPIEUP U THIEUTH" },
	{ 0x44BC , "HANGUL SYLLABLE SSANGCIEUC WE SSANGSIOS" },
	{ 0x44BD , "HANGUL SYLLABLE IEUNG WEO RIEUL-SIOS" },
	{ 0x44BE , "HANGUL SYLLABLE KIYEOK A KIYEOK-SIOS" },
	{ 0x44BF , "HANGUL SYLLABLE KIYEOK A NIEUN-CIEUC" },
	{ 0x44C0 , "HANGUL SYLLABLE KIYEOK A NIEUN-HIEUH" },
	{ 0x44C1 , "HANGUL SYLLABLE KIYEOK A RIEUL-THIEUTH" },
	{ 0x44C2 , "HANGUL SYLLABLE KIYEOK A RIEUL-PHIEUPH" },
	{ 0x44C3 , "HANGUL SYLLABLE KIYEOK A RIEUL-HIEUH" },
	{ 0x44C4 , "HANGUL SYLLABLE KIYEOK A KHIEUKH" },
	{ 0x44C5 , "HANGUL SYLLABLE KIYEOK AE SSANGKIYEOK" },
	{ 0x44C6 , "HANGUL SYLLABLE KIYEOK AE KIYEOK-SIOS" },
	{ 0x44C7 , "HANGUL SYLLABLE KIYEOK AE NIEUN-CIEUC" },
	{ 0x44C8 , "HANGUL SYLLABLE KIYEOK AE RIEUL-KIYEOK" },
	{ 0x44C9 , "HANGUL SYLLABLE KIYEOK AE RIEUL-MIEUM" },
	{ 0x44CA , "HANGUL SYLLABLE KIYEOK AE RIEUL-PIEUP" },
	{ 0x44CB , "HANGUL SYLLABLE KIYEOK AE RIEUL-SIOS" },
	{ 0x44CC , "HANGUL SYLLABLE KIYEOK AE RIEUL-THIEUTH" },
	{ 0x44CD , "HANGUL SYLLABLE KIYEOK AE RIEUL-PHIEUPH" },
	{ 0x44CE , "HANGUL SYLLABLE KIYEOK AE RIEUL-HIEUH" },
	{ 0x44CF , "HANGUL SYLLABLE KIYEOK AE PIEUP-SIOS" },
	{ 0x44D0 , "HANGUL SYLLABLE KIYEOK AE CHIEUCH" },
	{ 0x44D1 , "HANGUL SYLLABLE KIYEOK AE KHIEUKH" },
	{ 0x44D2 , "HANGUL SYLLABLE KIYEOK AE HIEUH" },
	{ 0x44D3 , "HANGUL SYLLABLE KIYEOK YA SSANGKIYEOK" },
	{ 0x44D4 , "HANGUL SYLLABLE KIYEOK YA KIYEOK-SIOS" },
	{ 0x44D5 , "HANGUL SYLLABLE KIYEOK YA NIEUN-CIEUC" },
	{ 0x44D6 , "HANGUL SYLLABLE KIYEOK YA NIEUN-HIEUH" },
	{ 0x44D7 , "HANGUL SYLLABLE KIYEOK YA RIEUL-MIEUM" },
	{ 0x44D8 , "HANGUL SYLLABLE KIYEOK YA RIEUL-PIEUP" },
	{ 0x44D9 , "HANGUL SYLLABLE KIYEOK YA RIEUL-SIOS" },
	{ 0x44DA , "HANGUL SYLLABLE KIYEOK YA RIEUL-THIEUTH" },
	{ 0x44DB , "HANGUL SYLLABLE KIYEOK YA RIEUL-PHIEUPH" },
	{ 0x44DC , "HANGUL SYLLABLE KIYEOK YA PIEUP-SIOS" },
	{ 0x44DD , "HANGUL SYLLABLE KIYEOK YA CIEUC" },
	{ 0x44DE , "HANGUL SYLLABLE KIYEOK YA CHIEUCH" },
	{ 0x44DF , "HANGUL SYLLABLE KIYEOK YA KHIEUKH" },
	{ 0x44E0 , "HANGUL SYLLABLE KIYEOK YA THIEUTH" },
	{ 0x44E1 , "HANGUL SYLLABLE KIYEOK YA PHIEUPH" },
	{ 0x44E2 , "HANGUL SYLLABLE KIYEOK YA HIEUH" },
	{ 0x44E3 , "HANGUL SYLLABLE KIYEOK YAE KIYEOK" },
	{ 0x44E4 , "HANGUL SYLLABLE KIYEOK YAE SSANGKIYEOK" },
	{ 0x44E5 , "HANGUL SYLLABLE KIYEOK YAE KIYEOK-SIOS" },
	{ 0x44E6 , "HANGUL SYLLABLE KIYEOK YAE NIEUN-CIEUC" },
	{ 0x44E7 , "HANGUL SYLLABLE KIYEOK YAE NIEUN-HIEUH" },
	{ 0x44E8 , "HANGUL SYLLABLE KIYEOK YAE TIKEUT" },
	{ 0x44E9 , "HANGUL SYLLABLE KIYEOK YAE RIEUL-KIYEOK" },
	{ 0x44EA , "HANGUL SYLLABLE KIYEOK YAE RIEUL-MIEUM" },
	{ 0x44EB , "HANGUL SYLLABLE KIYEOK YAE RIEUL-PIEUP" },
	{ 0x44EC , "HANGUL SYLLABLE KIYEOK YAE RIEUL-SIOS" },
	{ 0x44ED , "HANGUL SYLLABLE KIYEOK YAE RIEUL-THIEUTH" },
	{ 0x44EE , "HANGUL SYLLABLE KIYEOK YAE RIEUL-PHIEUPH" },
	{ 0x44EF , "HANGUL SYLLABLE KIYEOK YAE RIEUL-HIEUH" },
	{ 0x44F0 , "HANGUL SYLLABLE KIYEOK YAE PIEUP-SIOS" },
	{ 0x44F1 , "HANGUL SYLLABLE KIYEOK YAE SIOS" },
	{ 0x44F2 , "HANGUL SYLLABLE KIYEOK YAE SSANGSIOS" },
	{ 0x44F3 , "HANGUL SYLLABLE KIYEOK YAE IEUNG" },
	{ 0x44F4 , "HANGUL SYLLABLE KIYEOK YAE CIEUC" },
	{ 0x44F5 , "HANGUL SYLLABLE KIYEOK YAE CHIEUCH" },
	{ 0x44F6 , "HANGUL SYLLABLE KIYEOK YAE KHIEUKH" },
	{ 0x44F7 , "HANGUL SYLLABLE KIYEOK YAE THIEUTH" },
	{ 0x44F8 , "HANGUL SYLLABLE KIYEOK YAE PHIEUPH" },
	{ 0x44F9 , "HANGUL SYLLABLE KIYEOK YAE HIEUH" },
	{ 0x44FA , "HANGUL SYLLABLE KIYEOK EO SSANGKIYEOK" },
	{ 0x44FB , "HANGUL SYLLABLE KIYEOK EO NIEUN-HIEUH" },
	{ 0x44FC , "HANGUL SYLLABLE KIYEOK EO RIEUL-PIEUP" },
	{ 0x44FD , "HANGUL SYLLABLE KIYEOK EO RIEUL-SIOS" },
	{ 0x44FE , "HANGUL SYLLABLE KIYEOK EO RIEUL-THIEUTH" },
	{ 0x44FF , "HANGUL SYLLABLE KIYEOK EO RIEUL-PHIEUPH" },
	{ 0x4500 , "HANGUL SYLLABLE KIYEOK EO RIEUL-HIEUH" },
	{ 0x4501 , "HANGUL SYLLABLE KIYEOK EO PIEUP-SIOS" },
	{ 0x4502 , "HANGUL SYLLABLE KIYEOK EO KHIEUKH" },
	{ 0x4503 , "HANGUL SYLLABLE KIYEOK E SSANGKIYEOK" },
	{ 0x4504 , "HANGUL SYLLABLE KIYEOK E KIYEOK-SIOS" },
	{ 0x4505 , "HANGUL SYLLABLE KIYEOK E NIEUN-CIEUC" },
	{ 0x4506 , "HANGUL SYLLABLE KIYEOK E NIEUN-HIEUH" },
	{ 0x4507 , "HANGUL SYLLABLE KIYEOK E RIEUL-KIYEOK" },
	{ 0x4508 , "HANGUL SYLLABLE KIYEOK E RIEUL-MIEUM" },
	{ 0x4509 , "HANGUL SYLLABLE KIYEOK E RIEUL-PIEUP" },
	{ 0x450A , "HANGUL SYLLABLE KIYEOK E RIEUL-SIOS" },
	{ 0x450B , "HANGUL SYLLABLE KIYEOK E RIEUL-THIEUTH" },
	{ 0x450C , "HANGUL SYLLABLE KIYEOK E RIEUL-PHIEUPH" },
	{ 0x450D , "HANGUL SYLLABLE KIYEOK E RIEUL-HIEUH" },
	{ 0x450E , "HANGUL SYLLABLE KIYEOK E PIEUP-SIOS" },
	{ 0x450F , "HANGUL SYLLABLE KIYEOK E CIEUC" },
	{ 0x4510 , "HANGUL SYLLABLE KIYEOK E CHIEUCH" },
	{ 0x4511 , "HANGUL SYLLABLE KIYEOK E KHIEUKH" },
	{ 0x4512 , "HANGUL SYLLABLE KIYEOK E PHIEUPH" },
	{ 0x4513 , "HANGUL SYLLABLE KIYEOK YEO KIYEOK-SIOS" },
	{ 0x4514 , "HANGUL SYLLABLE KIYEOK YEO NIEUN-CIEUC" },
	{ 0x4515 , "HANGUL SYLLABLE KIYEOK YEO NIEUN-HIEUH" },
	{ 0x4516 , "HANGUL SYLLABLE KIYEOK YEO RIEUL-MIEUM" },
	{ 0x4517 , "HANGUL SYLLABLE KIYEOK YEO RIEUL-PIEUP" },
	{ 0x4518 , "HANGUL SYLLABLE KIYEOK YEO RIEUL-THIEUTH" },
	{ 0x4519 , "HANGUL SYLLABLE KIYEOK YEO RIEUL-PHIEUPH" },
	{ 0x451A , "HANGUL SYLLABLE KIYEOK YEO PIEUP-SIOS" },
	{ 0x451B , "HANGUL SYLLABLE KIYEOK YEO KHIEUKH" },
	{ 0x451C , "HANGUL SYLLABLE KIYEOK YEO HIEUH" },
	{ 0x451D , "HANGUL SYLLABLE KIYEOK YE SSANGKIYEOK" },
	{ 0x451E , "HANGUL SYLLABLE KIYEOK YE KIYEOK-SIOS" },
	{ 0x451F , "HANGUL SYLLABLE KIYEOK YE NIEUN-CIEUC" },
	{ 0x4520 , "HANGUL SYLLABLE KIYEOK YE NIEUN-HIEUH" },
	{ 0x4521 , "HANGUL SYLLABLE KIYEOK YE RIEUL-KIYEOK" },
	{ 0x4522 , "HANGUL SYLLABLE KIYEOK YE RIEUL-MIEUM" },
	{ 0x4523 , "HANGUL SYLLABLE KIYEOK YE RIEUL-PIEUP" },
	{ 0x4524 , "HANGUL SYLLABLE KIYEOK YE RIEUL-SIOS" },
	{ 0x4525 , "HANGUL SYLLABLE KIYEOK YE RIEUL-THIEUTH" },
	{ 0x4526 , "HANGUL SYLLABLE KIYEOK YE RIEUL-PHIEUPH" },
	{ 0x4527 , "HANGUL SYLLABLE KIYEOK YE RIEUL-HIEUH" },
	{ 0x4528 , "HANGUL SYLLABLE KIYEOK YE PIEUP-SIOS" },
	{ 0x4529 , "HANGUL SYLLABLE KIYEOK YE CIEUC" },
	{ 0x452A , "HANGUL SYLLABLE KIYEOK YE CHIEUCH" },
	{ 0x452B , "HANGUL SYLLABLE KIYEOK YE KHIEUKH" },
	{ 0x452C , "HANGUL SYLLABLE KIYEOK YE THIEUTH" },
	{ 0x452D , "HANGUL SYLLABLE KIYEOK YE PHIEUPH" },
	{ 0x452E , "HANGUL SYLLABLE KIYEOK YE HIEUH" },
	{ 0x452F , "HANGUL SYLLABLE KIYEOK O SSANGKIYEOK" },
	{ 0x4530 , "HANGUL SYLLABLE KIYEOK O KIYEOK-SIOS" },
	{ 0x4531 , "HANGUL SYLLABLE KIYEOK O NIEUN-CIEUC" },
	{ 0x4532 , "HANGUL SYLLABLE KIYEOK O NIEUN-HIEUH" },
	{ 0x4533 , "HANGUL SYLLABLE KIYEOK O RIEUL-THIEUTH" },
	{ 0x4534 , "HANGUL SYLLABLE KIYEOK O PIEUP-SIOS" },
	{ 0x4535 , "HANGUL SYLLABLE KIYEOK O SSANGSIOS" },
	{ 0x4536 , "HANGUL SYLLABLE KIYEOK O KHIEUKH" },
	{ 0x4537 , "HANGUL SYLLABLE KIYEOK WA SSANGKIYEOK" },
	{ 0x4538 , "HANGUL SYLLABLE KIYEOK WA KIYEOK-SIOS" },
	{ 0x4539 , "HANGUL SYLLABLE KIYEOK WA NIEUN-CIEUC" },
	{ 0x453A , "HANGUL SYLLABLE KIYEOK WA NIEUN-HIEUH" },
	{ 0x453B , "HANGUL SYLLABLE KIYEOK WA RIEUL-KIYEOK" },
	{ 0x453C , "HANGUL SYLLABLE KIYEOK WA RIEUL-PIEUP" },
	{ 0x453D , "HANGUL SYLLABLE KIYEOK WA RIEUL-SIOS" },
	{ 0x453E , "HANGUL SYLLABLE KIYEOK WA RIEUL-THIEUTH" },
	{ 0x453F , "HANGUL SYLLABLE KIYEOK WA RIEUL-PHIEUPH" },
	{ 0x4540 , "HANGUL SYLLABLE KIYEOK WA RIEUL-HIEUH" },
	{ 0x4541 , "HANGUL SYLLABLE KIYEOK WA PIEUP-SIOS" },
	{ 0x4542 , "HANGUL SYLLABLE KIYEOK WA CIEUC" },
	{ 0x4543 , "HANGUL SYLLABLE KIYEOK WA CHIEUCH" },
	{ 0x4544 , "HANGUL SYLLABLE KIYEOK WA KHIEUKH" },
	{ 0x4545 , "HANGUL SYLLABLE KIYEOK WA THIEUTH" },
	{ 0x4546 , "HANGUL SYLLABLE KIYEOK WA PHIEUPH" },
	{ 0x4547 , "HANGUL SYLLABLE KIYEOK WA HIEUH" },
	{ 0x4548 , "HANGUL SYLLABLE KIYEOK WAE SSANGKIYEOK" },
	{ 0x4549 , "HANGUL SYLLABLE KIYEOK WAE KIYEOK-SIOS" },
	{ 0x454A , "HANGUL SYLLABLE KIYEOK WAE NIEUN-CIEUC" },
	{ 0x454B , "HANGUL SYLLABLE KIYEOK WAE NIEUN-HIEUH" },
	{ 0x454C , "HANGUL SYLLABLE KIYEOK WAE RIEUL-KIYEOK" },
	{ 0x454D , "HANGUL SYLLABLE KIYEOK WAE RIEUL-PIEUP" },
	{ 0x454E , "HANGUL SYLLABLE KIYEOK WAE RIEUL-SIOS" },
	{ 0x454F , "HANGUL SYLLABLE KIYEOK WAE RIEUL-THIEUTH" },
	{ 0x4550 , "HANGUL SYLLABLE KIYEOK WAE RIEUL-PHIEUPH" },
	{ 0x4551 , "HANGUL SYLLABLE KIYEOK WAE RIEUL-HIEUH" },
	{ 0x4552 , "HANGUL SYLLABLE KIYEOK WAE PIEUP-SIOS" },
	{ 0x4553 , "HANGUL SYLLABLE KIYEOK WAE CIEUC" },
	{ 0x4554 , "HANGUL SYLLABLE KIYEOK WAE CHIEUCH" },
	{ 0x4555 , "HANGUL SYLLABLE KIYEOK WAE KHIEUKH" },
	{ 0x4556 , "HANGUL SYLLABLE KIYEOK WAE THIEUTH" },
	{ 0x4557 , "HANGUL SYLLABLE KIYEOK WAE PHIEUPH" },
	{ 0x4558 , "HANGUL SYLLABLE KIYEOK WAE HIEUH" },
	{ 0x4559 , "HANGUL SYLLABLE KIYEOK OE SSANGKIYEOK" },
	{ 0x455A , "HANGUL SYLLABLE KIYEOK OE KIYEOK-SIOS" },
	{ 0x455B , "HANGUL SYLLABLE KIYEOK OE NIEUN-CIEUC" },
	{ 0x455C , "HANGUL SYLLABLE KIYEOK OE NIEUN-HIEUH" },
	{ 0x455D , "HANGUL SYLLABLE KIYEOK OE RIEUL-KIYEOK" },
	{ 0x455E , "HANGUL SYLLABLE KIYEOK OE RIEUL-PIEUP" },
	{ 0x455F , "HANGUL SYLLABLE KIYEOK OE RIEUL-SIOS" },
	{ 0x4560 , "HANGUL SYLLABLE KIYEOK OE RIEUL-THIEUTH" },
	{ 0x4561 , "HANGUL SYLLABLE KIYEOK OE RIEUL-PHIEUPH" },
	{ 0x4562 , "HANGUL SYLLABLE KIYEOK OE RIEUL-HIEUH" },
	{ 0x4563 , "HANGUL SYLLABLE KIYEOK OE PIEUP-SIOS" },
	{ 0x4564 , "HANGUL SYLLABLE KIYEOK OE CIEUC" },
	{ 0x4565 , "HANGUL SYLLABLE KIYEOK OE CHIEUCH" },
	{ 0x4566 , "HANGUL SYLLABLE KIYEOK OE KHIEUKH" },
	{ 0x4567 , "HANGUL SYLLABLE KIYEOK OE THIEUTH" },
	{ 0x4568 , "HANGUL SYLLABLE KIYEOK OE PHIEUPH" },
	{ 0x4569 , "HANGUL SYLLABLE KIYEOK OE HIEUH" },
	{ 0x456A , "HANGUL SYLLABLE KIYEOK YO SSANGKIYEOK" },
	{ 0x456B , "HANGUL SYLLABLE KIYEOK YO KIYEOK-SIOS" },
	{ 0x456C , "HANGUL SYLLABLE KIYEOK YO NIEUN-CIEUC" },
	{ 0x456D , "HANGUL SYLLABLE KIYEOK YO NIEUN-HIEUH" },
	{ 0x456E , "HANGUL SYLLABLE KIYEOK YO RIEUL-KIYEOK" },
	{ 0x456F , "HANGUL SYLLABLE KIYEOK YO RIEUL-MIEUM" },
	{ 0x4570 , "HANGUL SYLLABLE KIYEOK YO RIEUL-PIEUP" },
	{ 0x4571 , "HANGUL SYLLABLE KIYEOK YO RIEUL-SIOS" },
	{ 0x4572 , "HANGUL SYLLABLE KIYEOK YO RIEUL-THIEUTH" },
	{ 0x4573 , "HANGUL SYLLABLE KIYEOK YO RIEUL-PHIEUPH" },
	{ 0x4574 , "HANGUL SYLLABLE KIYEOK YO RIEUL-HIEUH" },
	{ 0x4575 , "HANGUL SYLLABLE KIYEOK YO PIEUP-SIOS" },
	{ 0x4576 , "HANGUL SYLLABLE KIYEOK YO SSANGSIOS" },
	{ 0x4577 , "HANGUL SYLLABLE KIYEOK YO CIEUC" },
	{ 0x4578 , "HANGUL SYLLABLE KIYEOK YO CHIEUCH" },
	{ 0x4579 , "HANGUL SYLLABLE KIYEOK YO KHIEUKH" },
	{ 0x457A , "HANGUL SYLLABLE KIYEOK YO THIEUTH" },
	{ 0x457B , "HANGUL SYLLABLE KIYEOK YO PHIEUPH" },
	{ 0x457C , "HANGUL SYLLABLE KIYEOK YO HIEUH" },
	{ 0x457D , "HANGUL SYLLABLE KIYEOK U SSANGKIYEOK" },
	{ 0x457E , "HANGUL SYLLABLE KIYEOK U KIYEOK-SIOS" },
	{ 0x457F , "HANGUL SYLLABLE KIYEOK U NIEUN-CIEUC" },
	{ 0x4580 , "HANGUL SYLLABLE KIYEOK U NIEUN-HIEUH" },
	{ 0x4581 , "HANGUL SYLLABLE KIYEOK U RIEUL-PIEUP" },
	{ 0x4582 , "HANGUL SYLLABLE KIYEOK U RIEUL-THIEUTH" },
	{ 0x4583 , "HANGUL SYLLABLE KIYEOK U RIEUL-PHIEUPH" },
	{ 0x4584 , "HANGUL SYLLABLE KIYEOK U PIEUP-SIOS" },
	{ 0x4585 , "HANGUL SYLLABLE KIYEOK U SSANGSIOS" },
	{ 0x4586 , "HANGUL SYLLABLE KIYEOK U KHIEUKH" },
	{ 0x4587 , "HANGUL SYLLABLE KIYEOK U THIEUTH" },
	{ 0x4588 , "HANGUL SYLLABLE KIYEOK U HIEUH" },
	{ 0x4589 , "HANGUL SYLLABLE KIYEOK WEO SSANGKIYEOK" },
	{ 0x458A , "HANGUL SYLLABLE KIYEOK WEO KIYEOK-SIOS" },
	{ 0x458B , "HANGUL SYLLABLE KIYEOK WEO NIEUN-CIEUC" },
	{ 0x458C , "HANGUL SYLLABLE KIYEOK WEO NIEUN-HIEUH" },
	{ 0x458D , "HANGUL SYLLABLE KIYEOK WEO RIEUL-KIYEOK" },
	{ 0x458E , "HANGUL SYLLABLE KIYEOK WEO RIEUL-MIEUM" },
	{ 0x458F , "HANGUL SYLLABLE KIYEOK WEO RIEUL-PIEUP" },
	{ 0x4590 , "HANGUL SYLLABLE KIYEOK WEO RIEUL-SIOS" },
	{ 0x4591 , "HANGUL SYLLABLE KIYEOK WEO RIEUL-THIEUTH" },
	{ 0x4592 , "HANGUL SYLLABLE KIYEOK WEO RIEUL-PHIEUPH" },
	{ 0x4593 , "HANGUL SYLLABLE KIYEOK WEO RIEUL-HIEUH" },
	{ 0x4594 , "HANGUL SYLLABLE KIYEOK WEO MIEUM" },
	{ 0x4595 , "HANGUL SYLLABLE KIYEOK WEO PIEUP-SIOS" },
	{ 0x4596 , "HANGUL SYLLABLE KIYEOK WEO CIEUC" },
	{ 0x4597 , "HANGUL SYLLABLE KIYEOK WEO CHIEUCH" },
	{ 0x4598 , "HANGUL SYLLABLE KIYEOK WEO KHIEUKH" },
	{ 0x4599 , "HANGUL SYLLABLE KIYEOK WEO THIEUTH" },
	{ 0x459A , "HANGUL SYLLABLE KIYEOK WEO PHIEUPH" },
	{ 0x459B , "HANGUL SYLLABLE KIYEOK WEO HIEUH" },
	{ 0x459C , "HANGUL SYLLABLE KIYEOK WE SSANGKIYEOK" },
	{ 0x459D , "HANGUL SYLLABLE KIYEOK WE KIYEOK-SIOS" },
	{ 0x459E , "HANGUL SYLLABLE KIYEOK WE NIEUN-CIEUC" },
	{ 0x459F , "HANGUL SYLLABLE KIYEOK WE NIEUN-HIEUH" },
	{ 0x45A0 , "HANGUL SYLLABLE KIYEOK WE RIEUL-KIYEOK" },
	{ 0x45A1 , "HANGUL SYLLABLE KIYEOK WE RIEUL-MIEUM" },
	{ 0x45A2 , "HANGUL SYLLABLE KIYEOK WE RIEUL-PIEUP" },
	{ 0x45A3 , "HANGUL SYLLABLE KIYEOK WE RIEUL-SIOS" },
	{ 0x45A4 , "HANGUL SYLLABLE KIYEOK WE RIEUL-THIEUTH" },
	{ 0x45A5 , "HANGUL SYLLABLE KIYEOK WE RIEUL-PHIEUPH" },
	{ 0x45A6 , "HANGUL SYLLABLE KIYEOK WE RIEUL-HIEUH" },
	{ 0x45A7 , "HANGUL SYLLABLE KIYEOK WE PIEUP-SIOS" },
	{ 0x45A8 , "HANGUL SYLLABLE KIYEOK WE CIEUC" },
	{ 0x45A9 , "HANGUL SYLLABLE KIYEOK WE CHIEUCH" },
	{ 0x45AA , "HANGUL SYLLABLE KIYEOK WE KHIEUKH" },
	{ 0x45AB , "HANGUL SYLLABLE KIYEOK WE THIEUTH" },
	{ 0x45AC , "HANGUL SYLLABLE KIYEOK WE PHIEUPH" },
	{ 0x45AD , "HANGUL SYLLABLE KIYEOK WE HIEUH" },
	{ 0x45AE , "HANGUL SYLLABLE KIYEOK WI SSANGKIYEOK" },
	{ 0x45AF , "HANGUL SYLLABLE KIYEOK WI KIYEOK-SIOS" },
	{ 0x45B0 , "HANGUL SYLLABLE KIYEOK WI NIEUN-CIEUC" },
	{ 0x45B1 , "HANGUL SYLLABLE KIYEOK WI NIEUN-HIEUH" },
	{ 0x45B2 , "HANGUL SYLLABLE KIYEOK WI RIEUL-KIYEOK" },
	{ 0x45B3 , "HANGUL SYLLABLE KIYEOK WI RIEUL-PIEUP" },
	{ 0x45B4 , "HANGUL SYLLABLE KIYEOK WI RIEUL-SIOS" },
	{ 0x45B5 , "HANGUL SYLLABLE KIYEOK WI RIEUL-THIEUTH" },
	{ 0x45B6 , "HANGUL SYLLABLE KIYEOK WI RIEUL-PHIEUPH" },
	{ 0x45B7 , "HANGUL SYLLABLE KIYEOK WI RIEUL-HIEUH" },
	{ 0x45B8 , "HANGUL SYLLABLE KIYEOK WI PIEUP-SIOS" },
	{ 0x45B9 , "HANGUL SYLLABLE KIYEOK WI CIEUC" },
	{ 0x45BA , "HANGUL SYLLABLE KIYEOK WI CHIEUCH" },
	{ 0x45BB , "HANGUL SYLLABLE KIYEOK WI KHIEUKH" },
	{ 0x45BC , "HANGUL SYLLABLE KIYEOK WI THIEUTH" },
	{ 0x45BD , "HANGUL SYLLABLE KIYEOK WI PHIEUPH" },
	{ 0x45BE , "HANGUL SYLLABLE KIYEOK WI HIEUH" },
	{ 0x45BF , "HANGUL SYLLABLE KIYEOK YU SSANGKIYEOK" },
	{ 0x45C0 , "HANGUL SYLLABLE KIYEOK YU KIYEOK-SIOS" },
	{ 0x45C1 , "HANGUL SYLLABLE KIYEOK YU NIEUN-CIEUC" },
	{ 0x45C2 , "HANGUL SYLLABLE KIYEOK YU NIEUN-HIEUH" },
	{ 0x45C3 , "HANGUL SYLLABLE KIYEOK YU TIKEUT" },
	{ 0x45C4 , "HANGUL SYLLABLE KIYEOK YU RIEUL-KIYEOK" },
	{ 0x45C5 , "HANGUL SYLLABLE KIYEOK YU RIEUL-MIEUM" },
	{ 0x45C6 , "HANGUL SYLLABLE KIYEOK YU RIEUL-PIEUP" },
	{ 0x45C7 , "HANGUL SYLLABLE KIYEOK YU RIEUL-THIEUTH" },
	{ 0x45C8 , "HANGUL SYLLABLE KIYEOK YU RIEUL-PHIEUPH" },
	{ 0x45C9 , "HANGUL SYLLABLE KIYEOK YU RIEUL-HIEUH" },
	{ 0x45CA , "HANGUL SYLLABLE KIYEOK YU PIEUP-SIOS" },
	{ 0x45CB , "HANGUL SYLLABLE KIYEOK YU SSANGSIOS" },
	{ 0x45CC , "HANGUL SYLLABLE KIYEOK YU CIEUC" },
	{ 0x45CD , "HANGUL SYLLABLE KIYEOK YU CHIEUCH" },
	{ 0x45CE , "HANGUL SYLLABLE KIYEOK YU KHIEUKH" },
	{ 0x45CF , "HANGUL SYLLABLE KIYEOK YU THIEUTH" },
	{ 0x45D0 , "HANGUL SYLLABLE KIYEOK YU PHIEUPH" },
	{ 0x45D1 , "HANGUL SYLLABLE KIYEOK YU HIEUH" },
	{ 0x45D2 , "HANGUL SYLLABLE KIYEOK EU SSANGKIYEOK" },
	{ 0x45D3 , "HANGUL SYLLABLE KIYEOK EU KIYEOK-SIOS" },
	{ 0x45D4 , "HANGUL SYLLABLE KIYEOK EU NIEUN-CIEUC" },
	{ 0x45D5 , "HANGUL SYLLABLE KIYEOK EU RIEUL-THIEUTH" },
	{ 0x45D6 , "HANGUL SYLLABLE KIYEOK EU RIEUL-PHIEUPH" },
	{ 0x45D7 , "HANGUL SYLLABLE KIYEOK EU PIEUP-SIOS" },
	{ 0x45D8 , "HANGUL SYLLABLE KIYEOK EU SSANGSIOS" },
	{ 0x45D9 , "HANGUL SYLLABLE KIYEOK EU KHIEUKH" },
	{ 0x45DA , "HANGUL SYLLABLE KIYEOK YI SSANGKIYEOK" },
	{ 0x45DB , "HANGUL SYLLABLE KIYEOK YI KIYEOK-SIOS" },
	{ 0x45DC , "HANGUL SYLLABLE KIYEOK YI NIEUN-CIEUC" },
	{ 0x45DD , "HANGUL SYLLABLE KIYEOK YI NIEUN-HIEUH" },
	{ 0x45DE , "HANGUL SYLLABLE KIYEOK YI TIKEUT" },
	{ 0x45DF , "HANGUL SYLLABLE KIYEOK YI RIEUL-KIYEOK" },
	{ 0x45E0 , "HANGUL SYLLABLE KIYEOK YI RIEUL-MIEUM" },
	{ 0x45E1 , "HANGUL SYLLABLE KIYEOK YI RIEUL-PIEUP" },
	{ 0x45E2 , "HANGUL SYLLABLE KIYEOK YI RIEUL-THIEUTH" },
	{ 0x45E3 , "HANGUL SYLLABLE KIYEOK YI RIEUL-PHIEUPH" },
	{ 0x45E4 , "HANGUL SYLLABLE KIYEOK YI RIEUL-HIEUH" },
	{ 0x45E5 , "HANGUL SYLLABLE KIYEOK YI PIEUP-SIOS" },
	{ 0x45E6 , "HANGUL SYLLABLE KIYEOK YI SSANGSIOS" },
	{ 0x45E7 , "HANGUL SYLLABLE KIYEOK YI CIEUC" },
	{ 0x45E8 , "HANGUL SYLLABLE KIYEOK YI CHIEUCH" },
	{ 0x45E9 , "HANGUL SYLLABLE KIYEOK YI KHIEUKH" },
	{ 0x45EA , "HANGUL SYLLABLE KIYEOK YI THIEUTH" },
	{ 0x45EB , "HANGUL SYLLABLE KIYEOK YI PHIEUPH" },
	{ 0x45EC , "HANGUL SYLLABLE KIYEOK YI HIEUH" },
	{ 0x45ED , "HANGUL SYLLABLE KIYEOK I SSANGKIYEOK" },
	{ 0x45EE , "HANGUL SYLLABLE KIYEOK I KIYEOK-SIOS" },
	{ 0x45EF , "HANGUL SYLLABLE KIYEOK I NIEUN-CIEUC" },
	{ 0x45F0 , "HANGUL SYLLABLE KIYEOK I NIEUN-HIEUH" },
	{ 0x45F1 , "HANGUL SYLLABLE KIYEOK I RIEUL-PIEUP" },
	{ 0x45F2 , "HANGUL SYLLABLE KIYEOK I RIEUL-THIEUTH" },
	{ 0x45F3 , "HANGUL SYLLABLE KIYEOK I RIEUL-PHIEUPH" },
	{ 0x45F4 , "HANGUL SYLLABLE KIYEOK I RIEUL-HIEUH" },
	{ 0x45F5 , "HANGUL SYLLABLE KIYEOK I PIEUP-SIOS" },
	{ 0x45F6 , "HANGUL SYLLABLE KIYEOK I KHIEUKH" },
	{ 0x45F7 , "HANGUL SYLLABLE SSANGKIYEOK A KIYEOK-SIOS" },
	{ 0x45F8 , "HANGUL SYLLABLE SSANGKIYEOK A NIEUN-CIEUC" },
	{ 0x45F9 , "HANGUL SYLLABLE SSANGKIYEOK A NIEUN-HIEUH" },
	{ 0x45FA , "HANGUL SYLLABLE SSANGKIYEOK A RIEUL-KIYEOK" },
	{ 0x45FB , "HANGUL SYLLABLE SSANGKIYEOK A RIEUL-PIEUP" },
	{ 0x45FC , "HANGUL SYLLABLE SSANGKIYEOK A RIEUL-SIOS" },
	{ 0x45FD , "HANGUL SYLLABLE SSANGKIYEOK A RIEUL-THIEUTH" },
	{ 0x45FE , "HANGUL SYLLABLE SSANGKIYEOK A RIEUL-PHIEUPH" },
	{ 0x45FF , "HANGUL SYLLABLE SSANGKIYEOK A RIEUL-HIEUH" },
	{ 0x4600 , "HANGUL SYLLABLE SSANGKIYEOK A PIEUP-SIOS" },
	{ 0x4601 , "HANGUL SYLLABLE SSANGKIYEOK A KHIEUKH" },
	{ 0x4602 , "HANGUL SYLLABLE SSANGKIYEOK A PHIEUPH" },
	{ 0x4603 , "HANGUL SYLLABLE SSANGKIYEOK A HIEUH" },
	{ 0x4604 , "HANGUL SYLLABLE SSANGKIYEOK AE KIYEOK-SIOS" },
	{ 0x4605 , "HANGUL SYLLABLE SSANGKIYEOK AE NIEUN-CIEUC" },
	{ 0x4606 , "HANGUL SYLLABLE SSANGKIYEOK AE NIEUN-HIEUH" },
	{ 0x4607 , "HANGUL SYLLABLE SSANGKIYEOK AE RIEUL-KIYEOK" },
	{ 0x4608 , "HANGUL SYLLABLE SSANGKIYEOK AE RIEUL-MIEUM" },
	{ 0x4609 , "HANGUL SYLLABLE SSANGKIYEOK AE RIEUL-SIOS" },
	{ 0x460A , "HANGUL SYLLABLE SSANGKIYEOK AE RIEUL-THIEUTH" },
	{ 0x460B , "HANGUL SYLLABLE SSANGKIYEOK AE RIEUL-HIEUH" },
	{ 0x460C , "HANGUL SYLLABLE SSANGKIYEOK AE PIEUP-SIOS" },
	{ 0x460D , "HANGUL SYLLABLE SSANGKIYEOK AE CIEUC" },
	{ 0x460E , "HANGUL SYLLABLE SSANGKIYEOK AE CHIEUCH" },
	{ 0x460F , "HANGUL SYLLABLE SSANGKIYEOK AE KHIEUKH" },
	{ 0x4610 , "HANGUL SYLLABLE SSANGKIYEOK AE PHIEUPH" },
	{ 0x4611 , "HANGUL SYLLABLE SSANGKIYEOK AE HIEUH" },
	{ 0x4612 , "HANGUL SYLLABLE SSANGKIYEOK YA SSANGKIYEOK" },
	{ 0x4613 , "HANGUL SYLLABLE SSANGKIYEOK YA KIYEOK-SIOS" },
	{ 0x4614 , "HANGUL SYLLABLE SSANGKIYEOK YA NIEUN-CIEUC" },
	{ 0x4615 , "HANGUL SYLLABLE SSANGKIYEOK YA NIEUN-HIEUH" },
	{ 0x4616 , "HANGUL SYLLABLE SSANGKIYEOK YA RIEUL-MIEUM" },
	{ 0x4617 , "HANGUL SYLLABLE SSANGKIYEOK YA RIEUL-PIEUP" },
	{ 0x4618 , "HANGUL SYLLABLE SSANGKIYEOK YA RIEUL-SIOS" },
	{ 0x4619 , "HANGUL SYLLABLE SSANGKIYEOK YA RIEUL-THIEUTH" },
	{ 0x461A , "HANGUL SYLLABLE SSANGKIYEOK YA RIEUL-PHIEUPH" },
	{ 0x461B , "HANGUL SYLLABLE SSANGKIYEOK YA RIEUL-HIEUH" },
	{ 0x461C , "HANGUL SYLLABLE SSANGKIYEOK YA MIEUM" },
	{ 0x461D , "HANGUL SYLLABLE SSANGKIYEOK YA PIEUP" },
	{ 0x461E , "HANGUL SYLLABLE SSANGKIYEOK YA PIEUP-SIOS" },
	{ 0x461F , "HANGUL SYLLABLE SSANGKIYEOK YA SSANGSIOS" },
	{ 0x4620 , "HANGUL SYLLABLE SSANGKIYEOK YA CIEUC" },
	{ 0x4621 , "HANGUL SYLLABLE SSANGKIYEOK YA CHIEUCH" },
	{ 0x4622 , "HANGUL SYLLABLE SSANGKIYEOK YA KHIEUKH" },
	{ 0x4623 , "HANGUL SYLLABLE SSANGKIYEOK YA THIEUTH" },
	{ 0x4624 , "HANGUL SYLLABLE SSANGKIYEOK YA PHIEUPH" },
	{ 0x4625 , "HANGUL SYLLABLE SSANGKIYEOK YA HIEUH" },
	{ 0x4626 , "HANGUL SYLLABLE SSANGKIYEOK YAE KIYEOK" },
	{ 0x4627 , "HANGUL SYLLABLE SSANGKIYEOK YAE SSANGKIYEOK" },
	{ 0x4628 , "HANGUL SYLLABLE SSANGKIYEOK YAE KIYEOK-SIOS" },
	{ 0x4629 , "HANGUL SYLLABLE SSANGKIYEOK YAE NIEUN" },
	{ 0x462A , "HANGUL SYLLABLE SSANGKIYEOK YAE NIEUN-CIEUC" },
	{ 0x462B , "HANGUL SYLLABLE SSANGKIYEOK YAE NIEUN-HIEUH" },
	{ 0x462C , "HANGUL SYLLABLE SSANGKIYEOK YAE TIKEUT" },
	{ 0x462D , "HANGUL SYLLABLE SSANGKIYEOK YAE RIEUL" },
	{ 0x462E , "HANGUL SYLLABLE SSANGKIYEOK YAE RIEUL-KIYEOK" },
	{ 0x462F , "HANGUL SYLLABLE SSANGKIYEOK YAE RIEUL-MIEUM" },
	{ 0x4630 , "HANGUL SYLLABLE SSANGKIYEOK YAE RIEUL-PIEUP" },
	{ 0x4631 , "HANGUL SYLLABLE SSANGKIYEOK YAE RIEUL-SIOS" },
	{ 0x4632 , "HANGUL SYLLABLE SSANGKIYEOK YAE RIEUL-THIEUTH" },
	{ 0x4633 , "HANGUL SYLLABLE SSANGKIYEOK YAE RIEUL-PHIEUPH" },
	{ 0x4634 , "HANGUL SYLLABLE SSANGKIYEOK YAE RIEUL-HIEUH" },
	{ 0x4635 , "HANGUL SYLLABLE SSANGKIYEOK YAE MIEUM" },
	{ 0x4636 , "HANGUL SYLLABLE SSANGKIYEOK YAE PIEUP" },
	{ 0x4637 , "HANGUL SYLLABLE SSANGKIYEOK YAE PIEUP-SIOS" },
	{ 0x4638 , "HANGUL SYLLABLE SSANGKIYEOK YAE SIOS" },
	{ 0x4639 , "HANGUL SYLLABLE SSANGKIYEOK YAE SSANGSIOS" },
	{ 0x463A , "HANGUL SYLLABLE SSANGKIYEOK YAE CIEUC" },
	{ 0x463B , "HANGUL SYLLABLE SSANGKIYEOK YAE CHIEUCH" },
	{ 0x463C , "HANGUL SYLLABLE SSANGKIYEOK YAE KHIEUKH" },
	{ 0x463D , "HANGUL SYLLABLE SSANGKIYEOK YAE THIEUTH" },
	{ 0x463E , "HANGUL SYLLABLE SSANGKIYEOK YAE PHIEUPH" },
	{ 0x463F , "HANGUL SYLLABLE SSANGKIYEOK YAE HIEUH" },
	{ 0x4640 , "HANGUL SYLLABLE SSANGKIYEOK EO KIYEOK-SIOS" },
	{ 0x4641 , "HANGUL SYLLABLE SSANGKIYEOK EO NIEUN-CIEUC" },
	{ 0x4642 , "HANGUL SYLLABLE SSANGKIYEOK EO RIEUL-KIYEOK" },
	{ 0x4643 , "HANGUL SYLLABLE SSANGKIYEOK EO RIEUL-MIEUM" },
	{ 0x4644 , "HANGUL SYLLABLE SSANGKIYEOK EO RIEUL-PIEUP" },
	{ 0x4645 , "HANGUL SYLLABLE SSANGKIYEOK EO RIEUL-SIOS" },
	{ 0x4646 , "HANGUL SYLLABLE SSANGKIYEOK EO RIEUL-THIEUTH" },
	{ 0x4647 , "HANGUL SYLLABLE SSANGKIYEOK EO RIEUL-PHIEUPH" },
	{ 0x4648 , "HANGUL SYLLABLE SSANGKIYEOK EO RIEUL-HIEUH" },
	{ 0x4649 , "HANGUL SYLLABLE SSANGKIYEOK EO PIEUP-SIOS" },
	{ 0x464A , "HANGUL SYLLABLE SSANGKIYEOK EO KHIEUKH" },
	{ 0x464B , "HANGUL SYLLABLE SSANGKIYEOK EO PHIEUPH" },
	{ 0x464C , "HANGUL SYLLABLE SSANGKIYEOK EO HIEUH" },
	{ 0x464D , "HANGUL SYLLABLE SSANGKIYEOK E KIYEOK-SIOS" },
	{ 0x464E , "HANGUL SYLLABLE SSANGKIYEOK E NIEUN-CIEUC" },
	{ 0x464F , "HANGUL SYLLABLE SSANGKIYEOK E NIEUN-HIEUH" },
	{ 0x4650 , "HANGUL SYLLABLE SSANGKIYEOK E RIEUL-KIYEOK" },
	{ 0x4651 , "HANGUL SYLLABLE SSANGKIYEOK E RIEUL-MIEUM" },
	{ 0x4652 , "HANGUL SYLLABLE SSANGKIYEOK E RIEUL-PIEUP" },
	{ 0x4653 , "HANGUL SYLLABLE SSANGKIYEOK E RIEUL-SIOS" },
	{ 0x4654 , "HANGUL SYLLABLE SSANGKIYEOK E RIEUL-THIEUTH" },
	{ 0x4655 , "HANGUL SYLLABLE SSANGKIYEOK E RIEUL-PHIEUPH" },
	{ 0x4656 , "HANGUL SYLLABLE SSANGKIYEOK E RIEUL-HIEUH" },
	{ 0x4657 , "HANGUL SYLLABLE SSANGKIYEOK E PIEUP-SIOS" },
	{ 0x4658 , "HANGUL SYLLABLE SSANGKIYEOK E CIEUC" },
	{ 0x4659 , "HANGUL SYLLABLE SSANGKIYEOK E CHIEUCH" },
	{ 0x465A , "HANGUL SYLLABLE SSANGKIYEOK E KHIEUKH" },
	{ 0x465B , "HANGUL SYLLABLE SSANGKIYEOK E PHIEUPH" },
	{ 0x465C , "HANGUL SYLLABLE SSANGKIYEOK E HIEUH" },
	{ 0x465D , "HANGUL SYLLABLE SSANGKIYEOK YEO SSANGKIYEOK" },
	{ 0x465E , "HANGUL SYLLABLE SSANGKIYEOK YEO KIYEOK-SIOS" },
	{ 0x465F , "HANGUL SYLLABLE SSANGKIYEOK YEO NIEUN-CIEUC" },
	{ 0x4660 , "HANGUL SYLLABLE SSANGKIYEOK YEO NIEUN-HIEUH" },
	{ 0x4661 , "HANGUL SYLLABLE SSANGKIYEOK YEO RIEUL-KIYEOK" },
	{ 0x4662 , "HANGUL SYLLABLE SSANGKIYEOK YEO RIEUL-MIEUM" },
	{ 0x4663 , "HANGUL SYLLABLE SSANGKIYEOK YEO RIEUL-PIEUP" },
	{ 0x4664 , "HANGUL SYLLABLE SSANGKIYEOK YEO RIEUL-SIOS" },
	{ 0x4665 , "HANGUL SYLLABLE SSANGKIYEOK YEO RIEUL-THIEUTH" },
	{ 0x4666 , "HANGUL SYLLABLE SSANGKIYEOK YEO RIEUL-PHIEUPH" },
	{ 0x4667 , "HANGUL SYLLABLE SSANGKIYEOK YEO RIEUL-HIEUH" },
	{ 0x4668 , "HANGUL SYLLABLE SSANGKIYEOK YEO PIEUP-SIOS" },
	{ 0x4669 , "HANGUL SYLLABLE SSANGKIYEOK YEO CIEUC" },
	{ 0x466A , "HANGUL SYLLABLE SSANGKIYEOK YEO CHIEUCH" },
	{ 0x466B , "HANGUL SYLLABLE SSANGKIYEOK YEO KHIEUKH" },
	{ 0x466C , "HANGUL SYLLABLE SSANGKIYEOK YEO PHIEUPH" },
	{ 0x466D , "HANGUL SYLLABLE SSANGKIYEOK YEO HIEUH" },
	{ 0x466E , "HANGUL SYLLABLE SSANGKIYEOK YE KIYEOK" },
	{ 0x466F , "HANGUL SYLLABLE SSANGKIYEOK YE SSANGKIYEOK" },
	{ 0x4670 , "HANGUL SYLLABLE SSANGKIYEOK YE KIYEOK-SIOS" },
	{ 0x4671 , "HANGUL SYLLABLE SSANGKIYEOK YE NIEUN" },
	{ 0x4672 , "HANGUL SYLLABLE SSANGKIYEOK YE NIEUN-CIEUC" },
	{ 0x4673 , "HANGUL SYLLABLE SSANGKIYEOK YE NIEUN-HIEUH" },
	{ 0x4674 , "HANGUL SYLLABLE SSANGKIYEOK YE RIEUL" },
	{ 0x4675 , "HANGUL SYLLABLE SSANGKIYEOK YE RIEUL-KIYEOK" },
	{ 0x4676 , "HANGUL SYLLABLE SSANGKIYEOK YE RIEUL-MIEUM" },
	{ 0x4677 , "HANGUL SYLLABLE SSANGKIYEOK YE RIEUL-PIEUP" },
	{ 0x4678 , "HANGUL SYLLABLE SSANGKIYEOK YE RIEUL-SIOS" },
	{ 0x4679 , "HANGUL SYLLABLE SSANGKIYEOK YE RIEUL-THIEUTH" },
	{ 0x467A , "HANGUL SYLLABLE SSANGKIYEOK YE RIEUL-PHIEUPH" },
	{ 0x467B , "HANGUL SYLLABLE SSANGKIYEOK YE RIEUL-HIEUH" },
	{ 0x467C , "HANGUL SYLLABLE SSANGKIYEOK YE MIEUM" },
	{ 0x467D , "HANGUL SYLLABLE SSANGKIYEOK YE PIEUP" },
	{ 0x467E , "HANGUL SYLLABLE SSANGKIYEOK YE PIEUP-SIOS" },
	{ 0x467F , "HANGUL SYLLABLE SSANGKIYEOK YE SIOS" },
	{ 0x4680 , "HANGUL SYLLABLE SSANGKIYEOK YE SSANGSIOS" },
	{ 0x4681 , "HANGUL SYLLABLE SSANGKIYEOK YE CIEUC" },
	{ 0x4682 , "HANGUL SYLLABLE SSANGKIYEOK YE CHIEUCH" },
	{ 0x4683 , "HANGUL SYLLABLE SSANGKIYEOK YE KHIEUKH" },
	{ 0x4684 , "HANGUL SYLLABLE SSANGKIYEOK YE THIEUTH" },
	{ 0x4685 , "HANGUL SYLLABLE SSANGKIYEOK YE PHIEUPH" },
	{ 0x4686 , "HANGUL SYLLABLE SSANGKIYEOK YE HIEUH" },
	{ 0x4687 , "HANGUL SYLLABLE SSANGKIYEOK O SSANGKIYEOK" },
	{ 0x4688 , "HANGUL SYLLABLE SSANGKIYEOK O KIYEOK-SIOS" },
	{ 0x4689 , "HANGUL SYLLABLE SSANGKIYEOK O NIEUN-CIEUC" },
	{ 0x468A , "HANGUL SYLLABLE SSANGKIYEOK O RIEUL-KIYEOK" },
	{ 0x468B , "HANGUL SYLLABLE SSANGKIYEOK O RIEUL-PIEUP" },
	{ 0x468C , "HANGUL SYLLABLE SSANGKIYEOK O RIEUL-THIEUTH" },
	{ 0x468D , "HANGUL SYLLABLE SSANGKIYEOK O RIEUL-PHIEUPH" },
	{ 0x468E , "HANGUL SYLLABLE SSANGKIYEOK O PIEUP-SIOS" },
	{ 0x468F , "HANGUL SYLLABLE SSANGKIYEOK O SSANGSIOS" },
	{ 0x4690 , "HANGUL SYLLABLE SSANGKIYEOK O KHIEUKH" },
	{ 0x4691 , "HANGUL SYLLABLE SSANGKIYEOK O PHIEUPH" },
	{ 0x4692 , "HANGUL SYLLABLE SSANGKIYEOK WA SSANGKIYEOK" },
	{ 0x4693 , "HANGUL SYLLABLE SSANGKIYEOK WA KIYEOK-SIOS" },
	{ 0x4694 , "HANGUL SYLLABLE SSANGKIYEOK WA NIEUN-CIEUC" },
	{ 0x4695 , "HANGUL SYLLABLE SSANGKIYEOK WA NIEUN-HIEUH" },
	{ 0x4696 , "HANGUL SYLLABLE SSANGKIYEOK WA RIEUL-KIYEOK" },
	{ 0x4697 , "HANGUL SYLLABLE SSANGKIYEOK WA RIEUL-MIEUM" },
	{ 0x4698 , "HANGUL SYLLABLE SSANGKIYEOK WA RIEUL-PIEUP" },
	{ 0x4699 , "HANGUL SYLLABLE SSANGKIYEOK WA RIEUL-SIOS" },
	{ 0x469A , "HANGUL SYLLABLE SSANGKIYEOK WA RIEUL-THIEUTH" },
	{ 0x469B , "HANGUL SYLLABLE SSANGKIYEOK WA RIEUL-PHIEUPH" },
	{ 0x469C , "HANGUL SYLLABLE SSANGKIYEOK WA RIEUL-HIEUH" },
	{ 0x469D , "HANGUL SYLLABLE SSANGKIYEOK WA PIEUP-SIOS" },
	{ 0x469E , "HANGUL SYLLABLE SSANGKIYEOK WA CIEUC" },
	{ 0x469F , "HANGUL SYLLABLE SSANGKIYEOK WA CHIEUCH" },
	{ 0x46A0 , "HANGUL SYLLABLE SSANGKIYEOK WA KHIEUKH" },
	{ 0x46A1 , "HANGUL SYLLABLE SSANGKIYEOK WA THIEUTH" },
	{ 0x46A2 , "HANGUL SYLLABLE SSANGKIYEOK WA PHIEUPH" },
	{ 0x46A3 , "HANGUL SYLLABLE SSANGKIYEOK WA HIEUH" },
	{ 0x46A4 , "HANGUL SYLLABLE SSANGKIYEOK WAE SSANGKIYEOK" },
	{ 0x46A5 , "HANGUL SYLLABLE SSANGKIYEOK WAE KIYEOK-SIOS" },
	{ 0x46A6 , "HANGUL SYLLABLE SSANGKIYEOK WAE NIEUN-CIEUC" },
	{ 0x46A7 , "HANGUL SYLLABLE SSANGKIYEOK WAE NIEUN-HIEUH" },
	{ 0x46A8 , "HANGUL SYLLABLE SSANGKIYEOK WAE RIEUL-KIYEOK" },
	{ 0x46A9 , "HANGUL SYLLABLE SSANGKIYEOK WAE RIEUL-MIEUM" },
	{ 0x46AA , "HANGUL SYLLABLE SSANGKIYEOK WAE RIEUL-PIEUP" },
	{ 0x46AB , "HANGUL SYLLABLE SSANGKIYEOK WAE RIEUL-SIOS" },
	{ 0x46AC , "HANGUL SYLLABLE SSANGKIYEOK WAE RIEUL-THIEUTH" },
	{ 0x46AD , "HANGUL SYLLABLE SSANGKIYEOK WAE RIEUL-PHIEUPH" },
	{ 0x46AE , "HANGUL SYLLABLE SSANGKIYEOK WAE RIEUL-HIEUH" },
	{ 0x46AF , "HANGUL SYLLABLE SSANGKIYEOK WAE PIEUP-SIOS" },
	{ 0x46B0 , "HANGUL SYLLABLE SSANGKIYEOK WAE SIOS" },
	{ 0x46B1 , "HANGUL SYLLABLE SSANGKIYEOK WAE CIEUC" },
	{ 0x46B2 , "HANGUL SYLLABLE SSANGKIYEOK WAE CHIEUCH" },
	{ 0x46B3 , "HANGUL SYLLABLE SSANGKIYEOK WAE KHIEUKH" },
	{ 0x46B4 , "HANGUL SYLLABLE SSANGKIYEOK WAE THIEUTH" },
	{ 0x46B5 , "HANGUL SYLLABLE SSANGKIYEOK WAE PHIEUPH" },
	{ 0x46B6 , "HANGUL SYLLABLE SSANGKIYEOK WAE HIEUH" },
	{ 0x46B7 , "HANGUL SYLLABLE SSANGKIYEOK OE SSANGKIYEOK" },
	{ 0x46B8 , "HANGUL SYLLABLE SSANGKIYEOK OE KIYEOK-SIOS" },
	{ 0x46B9 , "HANGUL SYLLABLE SSANGKIYEOK OE NIEUN-CIEUC" },
	{ 0x46BA , "HANGUL SYLLABLE SSANGKIYEOK OE NIEUN-HIEUH" },
	{ 0x46BB , "HANGUL SYLLABLE SSANGKIYEOK OE RIEUL-KIYEOK" },
	{ 0x46BC , "HANGUL SYLLABLE SSANGKIYEOK OE RIEUL-MIEUM" },
	{ 0x46BD , "HANGUL SYLLABLE SSANGKIYEOK OE RIEUL-PIEUP" },
	{ 0x46BE , "HANGUL SYLLABLE SSANGKIYEOK OE RIEUL-SIOS" },
	{ 0x46BF , "HANGUL SYLLABLE SSANGKIYEOK OE RIEUL-THIEUTH" },
	{ 0x46C0 , "HANGUL SYLLABLE SSANGKIYEOK OE RIEUL-PHIEUPH" },
	{ 0x46C1 , "HANGUL SYLLABLE SSANGKIYEOK OE RIEUL-HIEUH" },
	{ 0x46C2 , "HANGUL SYLLABLE SSANGKIYEOK OE PIEUP-SIOS" },
	{ 0x46C3 , "HANGUL SYLLABLE SSANGKIYEOK OE CIEUC" },
	{ 0x46C4 , "HANGUL SYLLABLE SSANGKIYEOK OE CHIEUCH" },
	{ 0x46C5 , "HANGUL SYLLABLE SSANGKIYEOK OE KHIEUKH" },
	{ 0x46C6 , "HANGUL SYLLABLE SSANGKIYEOK OE THIEUTH" },
	{ 0x46C7 , "HANGUL SYLLABLE SSANGKIYEOK OE PHIEUPH" },
	{ 0x46C8 , "HANGUL SYLLABLE SSANGKIYEOK OE HIEUH" },
	{ 0x46C9 , "HANGUL SYLLABLE SSANGKIYEOK YO KIYEOK" },
	{ 0x46CA , "HANGUL SYLLABLE SSANGKIYEOK YO SSANGKIYEOK" },
	{ 0x46CB , "HANGUL SYLLABLE SSANGKIYEOK YO KIYEOK-SIOS" },
	{ 0x46CC , "HANGUL SYLLABLE SSANGKIYEOK YO NIEUN-CIEUC" },
	{ 0x46CD , "HANGUL SYLLABLE SSANGKIYEOK YO NIEUN-HIEUH" },
	{ 0x46CE , "HANGUL SYLLABLE SSANGKIYEOK YO RIEUL-KIYEOK" },
	{ 0x46CF , "HANGUL SYLLABLE SSANGKIYEOK YO RIEUL-MIEUM" },
	{ 0x46D0 , "HANGUL SYLLABLE SSANGKIYEOK YO RIEUL-PIEUP" },
	{ 0x46D1 , "HANGUL SYLLABLE SSANGKIYEOK YO RIEUL-SIOS" },
	{ 0x46D2 , "HANGUL SYLLABLE SSANGKIYEOK YO RIEUL-THIEUTH" },
	{ 0x46D3 , "HANGUL SYLLABLE SSANGKIYEOK YO RIEUL-PHIEUPH" },
	{ 0x46D4 , "HANGUL SYLLABLE SSANGKIYEOK YO RIEUL-HIEUH" },
	{ 0x46D5 , "HANGUL SYLLABLE SSANGKIYEOK YO PIEUP-SIOS" },
	{ 0x46D6 , "HANGUL SYLLABLE SSANGKIYEOK YO SIOS" },
	{ 0x46D7 , "HANGUL SYLLABLE SSANGKIYEOK YO SSANGSIOS" },
	{ 0x46D8 , "HANGUL SYLLABLE SSANGKIYEOK YO IEUNG" },
	{ 0x46D9 , "HANGUL SYLLABLE SSANGKIYEOK YO CHIEUCH" },
	{ 0x46DA , "HANGUL SYLLABLE SSANGKIYEOK YO KHIEUKH" },
	{ 0x46DB , "HANGUL SYLLABLE SSANGKIYEOK YO THIEUTH" },
	{ 0x46DC , "HANGUL SYLLABLE SSANGKIYEOK YO PHIEUPH" },
	{ 0x46DD , "HANGUL SYLLABLE SSANGKIYEOK YO HIEUH" },
	{ 0x46DE , "HANGUL SYLLABLE SSANGKIYEOK U SSANGKIYEOK" },
	{ 0x46DF , "HANGUL SYLLABLE SSANGKIYEOK U KIYEOK-SIOS" },
	{ 0x46E0 , "HANGUL SYLLABLE SSANGKIYEOK U NIEUN-CIEUC" },
	{ 0x46E1 , "HANGUL SYLLABLE SSANGKIYEOK U NIEUN-HIEUH" },
	{ 0x46E2 , "HANGUL SYLLABLE SSANGKIYEOK U RIEUL-MIEUM" },
	{ 0x46E3 , "HANGUL SYLLABLE SSANGKIYEOK U RIEUL-PIEUP" },
	{ 0x46E4 , "HANGUL SYLLABLE SSANGKIYEOK U RIEUL-SIOS" },
	{ 0x46E5 , "HANGUL SYLLABLE SSANGKIYEOK U RIEUL-THIEUTH" },
	{ 0x46E6 , "HANGUL SYLLABLE SSANGKIYEOK U RIEUL-PHIEUPH" },
	{ 0x46E7 , "HANGUL SYLLABLE SSANGKIYEOK U PIEUP-SIOS" },
	{ 0x46E8 , "HANGUL SYLLABLE SSANGKIYEOK U SSANGSIOS" },
	{ 0x46E9 , "HANGUL SYLLABLE SSANGKIYEOK U KHIEUKH" },
	{ 0x46EA , "HANGUL SYLLABLE SSANGKIYEOK U THIEUTH" },
	{ 0x46EB , "HANGUL SYLLABLE SSANGKIYEOK U PHIEUPH" },
	{ 0x46EC , "HANGUL SYLLABLE SSANGKIYEOK U HIEUH" },
	{ 0x46ED , "HANGUL SYLLABLE SSANGKIYEOK WEO SSANGKIYEOK" },
	{ 0x46EE , "HANGUL SYLLABLE SSANGKIYEOK WEO KIYEOK-SIOS" },
	{ 0x46EF , "HANGUL SYLLABLE SSANGKIYEOK WEO NIEUN-CIEUC" },
	{ 0x46F0 , "HANGUL SYLLABLE SSANGKIYEOK WEO NIEUN-HIEUH" },
	{ 0x46F1 , "HANGUL SYLLABLE SSANGKIYEOK WEO RIEUL-KIYEOK" },
	{ 0x46F2 , "HANGUL SYLLABLE SSANGKIYEOK WEO RIEUL-MIEUM" },
	{ 0x46F3 , "HANGUL SYLLABLE SSANGKIYEOK WEO RIEUL-PIEUP" },
	{ 0x46F4 , "HANGUL SYLLABLE SSANGKIYEOK WEO RIEUL-SIOS" },
	{ 0x46F5 , "HANGUL SYLLABLE SSANGKIYEOK WEO RIEUL-THIEUTH" },
	{ 0x46F6 , "HANGUL SYLLABLE SSANGKIYEOK WEO RIEUL-PHIEUPH" },
	{ 0x46F7 , "HANGUL SYLLABLE SSANGKIYEOK WEO RIEUL-HIEUH" },
	{ 0x46F8 , "HANGUL SYLLABLE SSANGKIYEOK WEO PIEUP" },
	{ 0x46F9 , "HANGUL SYLLABLE SSANGKIYEOK WEO PIEUP-SIOS" },
	{ 0x46FA , "HANGUL SYLLABLE SSANGKIYEOK WEO CIEUC" },
	{ 0x46FB , "HANGUL SYLLABLE SSANGKIYEOK WEO CHIEUCH" },
	{ 0x46FC , "HANGUL SYLLABLE SSANGKIYEOK WEO KHIEUKH" },
	{ 0x46FD , "HANGUL SYLLABLE SSANGKIYEOK WEO THIEUTH" },
	{ 0x46FE , "HANGUL SYLLABLE SSANGKIYEOK WEO PHIEUPH" },
	{ 0x46FF , "HANGUL SYLLABLE SSANGKIYEOK WEO HIEUH" },
	{ 0x4700 , "HANGUL SYLLABLE SSANGKIYEOK WE SSANGKIYEOK" },
	{ 0x4701 , "HANGUL SYLLABLE SSANGKIYEOK WE KIYEOK-SIOS" },
	{ 0x4702 , "HANGUL SYLLABLE SSANGKIYEOK WE NIEUN-CIEUC" },
	{ 0x4703 , "HANGUL SYLLABLE SSANGKIYEOK WE NIEUN-HIEUH" },
	{ 0x4704 , "HANGUL SYLLABLE SSANGKIYEOK WE RIEUL-KIYEOK" },
	{ 0x4705 , "HANGUL SYLLABLE SSANGKIYEOK WE RIEUL-MIEUM" },
	{ 0x4706 , "HANGUL SYLLABLE SSANGKIYEOK WE RIEUL-PIEUP" },
	{ 0x4707 , "HANGUL SYLLABLE SSANGKIYEOK WE RIEUL-SIOS" },
	{ 0x4708 , "HANGUL SYLLABLE SSANGKIYEOK WE RIEUL-THIEUTH" },
	{ 0x4709 , "HANGUL SYLLABLE SSANGKIYEOK WE RIEUL-PHIEUPH" },
	{ 0x470A , "HANGUL SYLLABLE SSANGKIYEOK WE RIEUL-HIEUH" },
	{ 0x470B , "HANGUL SYLLABLE SSANGKIYEOK WE PIEUP-SIOS" },
	{ 0x470C , "HANGUL SYLLABLE SSANGKIYEOK WE CIEUC" },
	{ 0x470D , "HANGUL SYLLABLE SSANGKIYEOK WE CHIEUCH" },
	{ 0x470E , "HANGUL SYLLABLE SSANGKIYEOK WE KHIEUKH" },
	{ 0x470F , "HANGUL SYLLABLE SSANGKIYEOK WE THIEUTH" },
	{ 0x4710 , "HANGUL SYLLABLE SSANGKIYEOK WE PHIEUPH" },
	{ 0x4711 , "HANGUL SYLLABLE SSANGKIYEOK WE HIEUH" },
	{ 0x4712 , "HANGUL SYLLABLE SSANGKIYEOK WI SSANGKIYEOK" },
	{ 0x4713 , "HANGUL SYLLABLE SSANGKIYEOK WI KIYEOK-SIOS" },
	{ 0x4714 , "HANGUL SYLLABLE SSANGKIYEOK WI NIEUN-CIEUC" },
	{ 0x4715 , "HANGUL SYLLABLE SSANGKIYEOK WI NIEUN-HIEUH" },
	{ 0x4716 , "HANGUL SYLLABLE SSANGKIYEOK WI RIEUL-KIYEOK" },
	{ 0x4717 , "HANGUL SYLLABLE SSANGKIYEOK WI RIEUL-MIEUM" },
	{ 0x4718 , "HANGUL SYLLABLE SSANGKIYEOK WI RIEUL-PIEUP" },
	{ 0x4719 , "HANGUL SYLLABLE SSANGKIYEOK WI RIEUL-SIOS" },
	{ 0x471A , "HANGUL SYLLABLE SSANGKIYEOK WI RIEUL-THIEUTH" },
	{ 0x471B , "HANGUL SYLLABLE SSANGKIYEOK WI RIEUL-PHIEUPH" },
	{ 0x471C , "HANGUL SYLLABLE SSANGKIYEOK WI RIEUL-HIEUH" },
	{ 0x471D , "HANGUL SYLLABLE SSANGKIYEOK WI PIEUP-SIOS" },
	{ 0x471E , "HANGUL SYLLABLE SSANGKIYEOK WI SSANGSIOS" },
	{ 0x471F , "HANGUL SYLLABLE SSANGKIYEOK WI CIEUC" },
	{ 0x4720 , "HANGUL SYLLABLE SSANGKIYEOK WI CHIEUCH" },
	{ 0x4721 , "HANGUL SYLLABLE SSANGKIYEOK WI KHIEUKH" },
	{ 0x4722 , "HANGUL SYLLABLE SSANGKIYEOK WI THIEUTH" },
	{ 0x4723 , "HANGUL SYLLABLE SSANGKIYEOK WI PHIEUPH" },
	{ 0x4724 , "HANGUL SYLLABLE SSANGKIYEOK WI HIEUH" },
	{ 0x4725 , "HANGUL SYLLABLE SSANGKIYEOK YU KIYEOK" },
	{ 0x4726 , "HANGUL SYLLABLE SSANGKIYEOK YU SSANGKIYEOK" },
	{ 0x4727 , "HANGUL SYLLABLE SSANGKIYEOK YU KIYEOK-SIOS" },
	{ 0x4728 , "HANGUL SYLLABLE SSANGKIYEOK YU NIEUN-CIEUC" },
	{ 0x4729 , "HANGUL SYLLABLE SSANGKIYEOK YU NIEUN-HIEUH" },
	{ 0x472A , "HANGUL SYLLABLE SSANGKIYEOK YU TIKEUT" },
	{ 0x472B , "HANGUL SYLLABLE SSANGKIYEOK YU RIEUL-KIYEOK" },
	{ 0x472C , "HANGUL SYLLABLE SSANGKIYEOK YU RIEUL-MIEUM" },
	{ 0x472D , "HANGUL SYLLABLE SSANGKIYEOK YU RIEUL-PIEUP" },
	{ 0x472E , "HANGUL SYLLABLE SSANGKIYEOK YU RIEUL-SIOS" },
	{ 0x472F , "HANGUL SYLLABLE SSANGKIYEOK YU RIEUL-THIEUTH" },
	{ 0x4730 , "HANGUL SYLLABLE SSANGKIYEOK YU RIEUL-PHIEUPH" },
	{ 0x4731 , "HANGUL SYLLABLE SSANGKIYEOK YU RIEUL-HIEUH" },
	{ 0x4732 , "HANGUL SYLLABLE SSANGKIYEOK YU PIEUP-SIOS" },
	{ 0x4733 , "HANGUL SYLLABLE SSANGKIYEOK YU SIOS" },
	{ 0x4734 , "HANGUL SYLLABLE SSANGKIYEOK YU SSANGSIOS" },
	{ 0x4735 , "HANGUL SYLLABLE SSANGKIYEOK YU IEUNG" },
	{ 0x4736 , "HANGUL SYLLABLE SSANGKIYEOK YU CIEUC" },
	{ 0x4737 , "HANGUL SYLLABLE SSANGKIYEOK YU CHIEUCH" },
	{ 0x4738 , "HANGUL SYLLABLE SSANGKIYEOK YU KHIEUKH" },
	{ 0x4739 , "HANGUL SYLLABLE SSANGKIYEOK YU THIEUTH" },
	{ 0x473A , "HANGUL SYLLABLE SSANGKIYEOK YU PHIEUPH" },
	{ 0x473B , "HANGUL SYLLABLE SSANGKIYEOK YU HIEUH" },
	{ 0x473C , "HANGUL SYLLABLE SSANGKIYEOK EU SSANGKIYEOK" },
	{ 0x473D , "HANGUL SYLLABLE SSANGKIYEOK EU KIYEOK-SIOS" },
	{ 0x473E , "HANGUL SYLLABLE SSANGKIYEOK EU NIEUN-CIEUC" },
	{ 0x473F , "HANGUL SYLLABLE SSANGKIYEOK EU RIEUL-PIEUP" },
	{ 0x4740 , "HANGUL SYLLABLE SSANGKIYEOK EU RIEUL-SIOS" },
	{ 0x4741 , "HANGUL SYLLABLE SSANGKIYEOK EU RIEUL-THIEUTH" },
	{ 0x4742 , "HANGUL SYLLABLE SSANGKIYEOK EU RIEUL-PHIEUPH" },
	{ 0x4743 , "HANGUL SYLLABLE SSANGKIYEOK EU PIEUP-SIOS" },
	{ 0x4744 , "HANGUL SYLLABLE SSANGKIYEOK EU SSANGSIOS" },
	{ 0x4745 , "HANGUL SYLLABLE SSANGKIYEOK EU CIEUC" },
	{ 0x4746 , "HANGUL SYLLABLE SSANGKIYEOK EU KHIEUKH" },
	{ 0x4747 , "HANGUL SYLLABLE SSANGKIYEOK EU PHIEUPH" },
	{ 0x4748 , "HANGUL SYLLABLE SSANGKIYEOK YI SSANGKIYEOK" },
	{ 0x4749 , "HANGUL SYLLABLE SSANGKIYEOK YI KIYEOK-SIOS" },
	{ 0x474A , "HANGUL SYLLABLE SSANGKIYEOK YI NIEUN" },
	{ 0x474B , "HANGUL SYLLABLE SSANGKIYEOK YI NIEUN-CIEUC" },
	{ 0x474C , "HANGUL SYLLABLE SSANGKIYEOK YI NIEUN-HIEUH" },
	{ 0x474D , "HANGUL SYLLABLE SSANGKIYEOK YI RIEUL-KIYEOK" },
	{ 0x474E , "HANGUL SYLLABLE SSANGKIYEOK YI RIEUL-MIEUM" },
	{ 0x474F , "HANGUL SYLLABLE SSANGKIYEOK YI RIEUL-SIOS" },
	{ 0x4750 , "HANGUL SYLLABLE SSANGKIYEOK YI RIEUL-THIEUTH" },
	{ 0x4751 , "HANGUL SYLLABLE SSANGKIYEOK YI RIEUL-PHIEUPH" },
	{ 0x4752 , "HANGUL SYLLABLE SSANGKIYEOK YI RIEUL-HIEUH" },
	{ 0x4753 , "HANGUL SYLLABLE SSANGKIYEOK YI MIEUM" },
	{ 0x4754 , "HANGUL SYLLABLE SSANGKIYEOK YI PIEUP" },
	{ 0x4755 , "HANGUL SYLLABLE SSANGKIYEOK YI PIEUP-SIOS" },
	{ 0x4756 , "HANGUL SYLLABLE SSANGKIYEOK YI SSANGSIOS" },
	{ 0x4757 , "HANGUL SYLLABLE SSANGKIYEOK YI CIEUC" },
	{ 0x4758 , "HANGUL SYLLABLE SSANGKIYEOK YI CHIEUCH" },
	{ 0x4759 , "HANGUL SYLLABLE SSANGKIYEOK YI KHIEUKH" },
	{ 0x475A , "HANGUL SYLLABLE SSANGKIYEOK YI THIEUTH" },
	{ 0x475B , "HANGUL SYLLABLE SSANGKIYEOK YI PHIEUPH" },
	{ 0x475C , "HANGUL SYLLABLE SSANGKIYEOK YI HIEUH" },
	{ 0x475D , "HANGUL SYLLABLE SSANGKIYEOK I SSANGKIYEOK" },
	{ 0x475E , "HANGUL SYLLABLE SSANGKIYEOK I KIYEOK-SIOS" },
	{ 0x475F , "HANGUL SYLLABLE SSANGKIYEOK I NIEUN-CIEUC" },
	{ 0x4760 , "HANGUL SYLLABLE SSANGKIYEOK I NIEUN-HIEUH" },
	{ 0x4761 , "HANGUL SYLLABLE SSANGKIYEOK I RIEUL-KIYEOK" },
	{ 0x4762 , "HANGUL SYLLABLE SSANGKIYEOK I RIEUL-MIEUM" },
	{ 0x4763 , "HANGUL SYLLABLE SSANGKIYEOK I RIEUL-PIEUP" },
	{ 0x4764 , "HANGUL SYLLABLE SSANGKIYEOK I RIEUL-SIOS" },
	{ 0x4765 , "HANGUL SYLLABLE SSANGKIYEOK I RIEUL-THIEUTH" },
	{ 0x4766 , "HANGUL SYLLABLE SSANGKIYEOK I RIEUL-PHIEUPH" },
	{ 0x4767 , "HANGUL SYLLABLE SSANGKIYEOK I PIEUP-SIOS" },
	{ 0x4768 , "HANGUL SYLLABLE SSANGKIYEOK I CIEUC" },
	{ 0x4769 , "HANGUL SYLLABLE SSANGKIYEOK I CHIEUCH" },
	{ 0x476A , "HANGUL SYLLABLE SSANGKIYEOK I KHIEUKH" },
	{ 0x476B , "HANGUL SYLLABLE SSANGKIYEOK I PHIEUPH" },
	{ 0x476C , "HANGUL SYLLABLE SSANGKIYEOK I HIEUH" },
	{ 0x476D , "HANGUL SYLLABLE NIEUN A NIEUN-CIEUC" },
	{ 0x476E , "HANGUL SYLLABLE NIEUN A NIEUN-HIEUH" },
	{ 0x476F , "HANGUL SYLLABLE NIEUN A RIEUL-PIEUP" },
	{ 0x4770 , "HANGUL SYLLABLE NIEUN A RIEUL-THIEUTH" },
	{ 0x4771 , "HANGUL SYLLABLE NIEUN A RIEUL-PHIEUPH" },
	{ 0x4772 , "HANGUL SYLLABLE NIEUN A RIEUL-HIEUH" },
	{ 0x4773 , "HANGUL SYLLABLE NIEUN AE SSANGKIYEOK" },
	{ 0x4774 , "HANGUL SYLLABLE NIEUN AE KIYEOK-SIOS" },
	{ 0x4775 , "HANGUL SYLLABLE NIEUN AE NIEUN-CIEUC" },
	{ 0x4776 , "HANGUL SYLLABLE NIEUN AE NIEUN-HIEUH" },
	{ 0x4777 , "HANGUL SYLLABLE NIEUN AE RIEUL-KIYEOK" },
	{ 0x4778 , "HANGUL SYLLABLE NIEUN AE RIEUL-MIEUM" },
	{ 0x4779 , "HANGUL SYLLABLE NIEUN AE RIEUL-PIEUP" },
	{ 0x477A , "HANGUL SYLLABLE NIEUN AE RIEUL-SIOS" },
	{ 0x477B , "HANGUL SYLLABLE NIEUN AE RIEUL-THIEUTH" },
	{ 0x477C , "HANGUL SYLLABLE NIEUN AE RIEUL-PHIEUPH" },
	{ 0x477D , "HANGUL SYLLABLE NIEUN AE RIEUL-HIEUH" },
	{ 0x477E , "HANGUL SYLLABLE NIEUN AE PIEUP-SIOS" },
	{ 0x477F , "HANGUL SYLLABLE NIEUN AE CIEUC" },
	{ 0x4780 , "HANGUL SYLLABLE NIEUN AE CHIEUCH" },
	{ 0x4781 , "HANGUL SYLLABLE NIEUN AE KHIEUKH" },
	{ 0x4782 , "HANGUL SYLLABLE NIEUN AE THIEUTH" },
	{ 0x4783 , "HANGUL SYLLABLE NIEUN AE HIEUH" },
	{ 0x4784 , "HANGUL SYLLABLE NIEUN YA SSANGKIYEOK" },
	{ 0x4785 , "HANGUL SYLLABLE NIEUN YA KIYEOK-SIOS" },
	{ 0x4786 , "HANGUL SYLLABLE NIEUN YA NIEUN-CIEUC" },
	{ 0x4787 , "HANGUL SYLLABLE NIEUN YA NIEUN-HIEUH" },
	{ 0x4788 , "HANGUL SYLLABLE NIEUN YA RIEUL-KIYEOK" },
	{ 0x4789 , "HANGUL SYLLABLE NIEUN YA RIEUL-MIEUM" },
	{ 0x478A , "HANGUL SYLLABLE NIEUN YA RIEUL-PIEUP" },
	{ 0x478B , "HANGUL SYLLABLE NIEUN YA RIEUL-SIOS" },
	{ 0x478C , "HANGUL SYLLABLE NIEUN YA RIEUL-THIEUTH" },
	{ 0x478D , "HANGUL SYLLABLE NIEUN YA RIEUL-PHIEUPH" },
	{ 0x478E , "HANGUL SYLLABLE NIEUN YA RIEUL-HIEUH" },
	{ 0x478F , "HANGUL SYLLABLE NIEUN YA PIEUP-SIOS" },
	{ 0x4790 , "HANGUL SYLLABLE NIEUN YA SSANGSIOS" },
	{ 0x4791 , "HANGUL SYLLABLE NIEUN YA CHIEUCH" },
	{ 0x4792 , "HANGUL SYLLABLE NIEUN YA KHIEUKH" },
	{ 0x4793 , "HANGUL SYLLABLE NIEUN YA PHIEUPH" },
	{ 0x4794 , "HANGUL SYLLABLE NIEUN YA HIEUH" },
	{ 0x4795 , "HANGUL SYLLABLE NIEUN YAE KIYEOK" },
	{ 0x4796 , "HANGUL SYLLABLE NIEUN YAE SSANGKIYEOK" },
	{ 0x4797 , "HANGUL SYLLABLE NIEUN YAE KIYEOK-SIOS" },
	{ 0x4798 , "HANGUL SYLLABLE NIEUN YAE NIEUN" },
	{ 0x4799 , "HANGUL SYLLABLE NIEUN YAE NIEUN-CIEUC" },
	{ 0x479A , "HANGUL SYLLABLE NIEUN YAE NIEUN-HIEUH" },
	{ 0x479B , "HANGUL SYLLABLE NIEUN YAE TIKEUT" },
	{ 0x479C , "HANGUL SYLLABLE NIEUN YAE RIEUL" },
	{ 0x479D , "HANGUL SYLLABLE NIEUN YAE RIEUL-KIYEOK" },
	{ 0x479E , "HANGUL SYLLABLE NIEUN YAE RIEUL-MIEUM" },
	{ 0x479F , "HANGUL SYLLABLE NIEUN YAE RIEUL-PIEUP" },
	{ 0x47A0 , "HANGUL SYLLABLE NIEUN YAE RIEUL-SIOS" },
	{ 0x47A1 , "HANGUL SYLLABLE NIEUN YAE RIEUL-THIEUTH" },
	{ 0x47A2 , "HANGUL SYLLABLE NIEUN YAE RIEUL-PHIEUPH" },
	{ 0x47A3 , "HANGUL SYLLABLE NIEUN YAE RIEUL-HIEUH" },
	{ 0x47A4 , "HANGUL SYLLABLE NIEUN YAE MIEUM" },
	{ 0x47A5 , "HANGUL SYLLABLE NIEUN YAE PIEUP" },
	{ 0x47A6 , "HANGUL SYLLABLE NIEUN YAE PIEUP-SIOS" },
	{ 0x47A7 , "HANGUL SYLLABLE NIEUN YAE SIOS" },
	{ 0x47A8 , "HANGUL SYLLABLE NIEUN YAE SSANGSIOS" },
	{ 0x47A9 , "HANGUL SYLLABLE NIEUN YAE CIEUC" },
	{ 0x47AA , "HANGUL SYLLABLE NIEUN YAE CHIEUCH" },
	{ 0x47AB , "HANGUL SYLLABLE NIEUN YAE KHIEUKH" },
	{ 0x47AC , "HANGUL SYLLABLE NIEUN YAE THIEUTH" },
	{ 0x47AD , "HANGUL SYLLABLE NIEUN YAE PHIEUPH" },
	{ 0x47AE , "HANGUL SYLLABLE NIEUN YAE HIEUH" },
	{ 0x47AF , "HANGUL SYLLABLE NIEUN EO NIEUN-CIEUC" },
	{ 0x47B0 , "HANGUL SYLLABLE NIEUN EO NIEUN-HIEUH" },
	{ 0x47B1 , "HANGUL SYLLABLE NIEUN EO RIEUL-KIYEOK" },
	{ 0x47B2 , "HANGUL SYLLABLE NIEUN EO RIEUL-SIOS" },
	{ 0x47B3 , "HANGUL SYLLABLE NIEUN EO RIEUL-THIEUTH" },
	{ 0x47B4 , "HANGUL SYLLABLE NIEUN EO RIEUL-PHIEUPH" },
	{ 0x47B5 , "HANGUL SYLLABLE NIEUN EO PIEUP-SIOS" },
	{ 0x47B6 , "HANGUL SYLLABLE NIEUN EO CHIEUCH" },
	{ 0x47B7 , "HANGUL SYLLABLE NIEUN E SSANGKIYEOK" },
	{ 0x47B8 , "HANGUL SYLLABLE NIEUN E KIYEOK-SIOS" },
	{ 0x47B9 , "HANGUL SYLLABLE NIEUN E NIEUN-CIEUC" },
	{ 0x47BA , "HANGUL SYLLABLE NIEUN E NIEUN-HIEUH" },
	{ 0x47BB , "HANGUL SYLLABLE NIEUN E RIEUL-KIYEOK" },
	{ 0x47BC , "HANGUL SYLLABLE NIEUN E RIEUL-MIEUM" },
	{ 0x47BD , "HANGUL SYLLABLE NIEUN E RIEUL-PIEUP" },
	{ 0x47BE , "HANGUL SYLLABLE NIEUN E RIEUL-SIOS" },
	{ 0x47BF , "HANGUL SYLLABLE NIEUN E RIEUL-THIEUTH" },
	{ 0x47C0 , "HANGUL SYLLABLE NIEUN E RIEUL-PHIEUPH" },
	{ 0x47C1 , "HANGUL SYLLABLE NIEUN E RIEUL-HIEUH" },
	{ 0x47C2 , "HANGUL SYLLABLE NIEUN E PIEUP-SIOS" },
	{ 0x47C3 , "HANGUL SYLLABLE NIEUN E CIEUC" },
	{ 0x47C4 , "HANGUL SYLLABLE NIEUN E CHIEUCH" },
	{ 0x47C5 , "HANGUL SYLLABLE NIEUN YEO SSANGKIYEOK" },
	{ 0x47C6 , "HANGUL SYLLABLE NIEUN YEO NIEUN-CIEUC" },
	{ 0x47C7 , "HANGUL SYLLABLE NIEUN YEO NIEUN-HIEUH" },
	{ 0x47C8 , "HANGUL SYLLABLE NIEUN YEO RIEUL-KIYEOK" },
	{ 0x47C9 , "HANGUL SYLLABLE NIEUN YEO RIEUL-MIEUM" },
	{ 0x47CA , "HANGUL SYLLABLE NIEUN YEO RIEUL-SIOS" },
	{ 0x47CB , "HANGUL SYLLABLE NIEUN YEO RIEUL-THIEUTH" },
	{ 0x47CC , "HANGUL SYLLABLE NIEUN YEO RIEUL-PHIEUPH" },
	{ 0x47CD , "HANGUL SYLLABLE NIEUN YEO RIEUL-HIEUH" },
	{ 0x47CE , "HANGUL SYLLABLE NIEUN YEO PIEUP-SIOS" },
	{ 0x47CF , "HANGUL SYLLABLE NIEUN YEO CIEUC" },
	{ 0x47D0 , "HANGUL SYLLABLE NIEUN YEO CHIEUCH" },
	{ 0x47D1 , "HANGUL SYLLABLE NIEUN YE SSANGKIYEOK" },
	{ 0x47D2 , "HANGUL SYLLABLE NIEUN YE KIYEOK-SIOS" },
	{ 0x47D3 , "HANGUL SYLLABLE NIEUN YE NIEUN-CIEUC" },
	{ 0x47D4 , "HANGUL SYLLABLE NIEUN YE NIEUN-HIEUH" },
	{ 0x47D5 , "HANGUL SYLLABLE NIEUN YE RIEUL-KIYEOK" },
	{ 0x47D6 , "HANGUL SYLLABLE NIEUN YE RIEUL-MIEUM" },
	{ 0x47D7 , "HANGUL SYLLABLE NIEUN YE RIEUL-PIEUP" },
	{ 0x47D8 , "HANGUL SYLLABLE NIEUN YE RIEUL-SIOS" },
	{ 0x47D9 , "HANGUL SYLLABLE NIEUN YE RIEUL-THIEUTH" },
	{ 0x47DA , "HANGUL SYLLABLE NIEUN YE RIEUL-PHIEUPH" },
	{ 0x47DB , "HANGUL SYLLABLE NIEUN YE RIEUL-HIEUH" },
	{ 0x47DC , "HANGUL SYLLABLE NIEUN YE PIEUP-SIOS" },
	{ 0x47DD , "HANGUL SYLLABLE NIEUN YE SSANGSIOS" },
	{ 0x47DE , "HANGUL SYLLABLE NIEUN YE CIEUC" },
	{ 0x47DF , "HANGUL SYLLABLE NIEUN YE CHIEUCH" },
	{ 0x47E0 , "HANGUL SYLLABLE NIEUN YE KHIEUKH" },
	{ 0x47E1 , "HANGUL SYLLABLE NIEUN YE THIEUTH" },
	{ 0x47E2 , "HANGUL SYLLABLE NIEUN YE PHIEUPH" },
	{ 0x47E3 , "HANGUL SYLLABLE NIEUN YE HIEUH" },
	{ 0x47E4 , "HANGUL SYLLABLE NIEUN O NIEUN-CIEUC" },
	{ 0x47E5 , "HANGUL SYLLABLE NIEUN O RIEUL-KIYEOK" },
	{ 0x47E6 , "HANGUL SYLLABLE NIEUN O RIEUL-PIEUP" },
	{ 0x47E7 , "HANGUL SYLLABLE NIEUN O RIEUL-SIOS" },
	{ 0x47E8 , "HANGUL SYLLABLE NIEUN O RIEUL-THIEUTH" },
	{ 0x47E9 , "HANGUL SYLLABLE NIEUN O RIEUL-PHIEUPH" },
	{ 0x47EA , "HANGUL SYLLABLE NIEUN O RIEUL-HIEUH" },
	{ 0x47EB , "HANGUL SYLLABLE NIEUN O PIEUP-SIOS" },
	{ 0x47EC , "HANGUL SYLLABLE NIEUN O SSANGSIOS" },
	{ 0x47ED , "HANGUL SYLLABLE NIEUN O CHIEUCH" },
	{ 0x47EE , "HANGUL SYLLABLE NIEUN WA SSANGKIYEOK" },
	{ 0x47EF , "HANGUL SYLLABLE NIEUN WA KIYEOK-SIOS" },
	{ 0x47F0 , "HANGUL SYLLABLE NIEUN WA NIEUN-CIEUC" },
	{ 0x47F1 , "HANGUL SYLLABLE NIEUN WA NIEUN-HIEUH" },
	{ 0x47F2 , "HANGUL SYLLABLE NIEUN WA RIEUL-KIYEOK" },
	{ 0x47F3 , "HANGUL SYLLABLE NIEUN WA RIEUL-MIEUM" },
	{ 0x47F4 , "HANGUL SYLLABLE NIEUN WA RIEUL-PIEUP" },
	{ 0x47F5 , "HANGUL SYLLABLE NIEUN WA RIEUL-SIOS" },
	{ 0x47F6 , "HANGUL SYLLABLE NIEUN WA RIEUL-THIEUTH" },
	{ 0x47F7 , "HANGUL SYLLABLE NIEUN WA RIEUL-PHIEUPH" },
	{ 0x47F8 , "HANGUL SYLLABLE NIEUN WA RIEUL-HIEUH" },
	{ 0x47F9 , "HANGUL SYLLABLE NIEUN WA PIEUP-SIOS" },
	{ 0x47FA , "HANGUL SYLLABLE NIEUN WA CIEUC" },
	{ 0x47FB , "HANGUL SYLLABLE NIEUN WA CHIEUCH" },
	{ 0x47FC , "HANGUL SYLLABLE NIEUN WA KHIEUKH" },
	{ 0x47FD , "HANGUL SYLLABLE NIEUN WA THIEUTH" },
	{ 0x47FE , "HANGUL SYLLABLE NIEUN WA PHIEUPH" },
	{ 0x47FF , "HANGUL SYLLABLE NIEUN WA HIEUH" },
	{ 0x4800 , "HANGUL SYLLABLE NIEUN WAE KIYEOK" },
	{ 0x4801 , "HANGUL SYLLABLE NIEUN WAE SSANGKIYEOK" },
	{ 0x4802 , "HANGUL SYLLABLE NIEUN WAE KIYEOK-SIOS" },
	{ 0x4803 , "HANGUL SYLLABLE NIEUN WAE NIEUN-CIEUC" },
	{ 0x4804 , "HANGUL SYLLABLE NIEUN WAE NIEUN-HIEUH" },
	{ 0x4805 , "HANGUL SYLLABLE NIEUN WAE TIKEUT" },
	{ 0x4806 , "HANGUL SYLLABLE NIEUN WAE RIEUL-KIYEOK" },
	{ 0x4807 , "HANGUL SYLLABLE NIEUN WAE RIEUL-MIEUM" },
	{ 0x4808 , "HANGUL SYLLABLE NIEUN WAE RIEUL-PIEUP" },
	{ 0x4809 , "HANGUL SYLLABLE NIEUN WAE RIEUL-SIOS" },
	{ 0x480A , "HANGUL SYLLABLE NIEUN WAE RIEUL-THIEUTH" },
	{ 0x480B , "HANGUL SYLLABLE NIEUN WAE RIEUL-PHIEUPH" },
	{ 0x480C , "HANGUL SYLLABLE NIEUN WAE RIEUL-HIEUH" },
	{ 0x480D , "HANGUL SYLLABLE NIEUN WAE MIEUM" },
	{ 0x480E , "HANGUL SYLLABLE NIEUN WAE PIEUP" },
	{ 0x480F , "HANGUL SYLLABLE NIEUN WAE PIEUP-SIOS" },
	{ 0x4810 , "HANGUL SYLLABLE NIEUN WAE SIOS" },
	{ 0x4811 , "HANGUL SYLLABLE NIEUN WAE IEUNG" },
	{ 0x4812 , "HANGUL SYLLABLE NIEUN WAE CIEUC" },
	{ 0x4813 , "HANGUL SYLLABLE NIEUN WAE CHIEUCH" },
	{ 0x4814 , "HANGUL SYLLABLE NIEUN WAE KHIEUKH" },
	{ 0x4815 , "HANGUL SYLLABLE NIEUN WAE THIEUTH" },
	{ 0x4816 , "HANGUL SYLLABLE NIEUN WAE PHIEUPH" },
	{ 0x4817 , "HANGUL SYLLABLE NIEUN WAE HIEUH" },
	{ 0x4818 , "HANGUL SYLLABLE NIEUN OE SSANGKIYEOK" },
	{ 0x4819 , "HANGUL SYLLABLE NIEUN OE KIYEOK-SIOS" },
	{ 0x481A , "HANGUL SYLLABLE NIEUN OE NIEUN-CIEUC" },
	{ 0x481B , "HANGUL SYLLABLE NIEUN OE NIEUN-HIEUH" },
	{ 0x481C , "HANGUL SYLLABLE NIEUN OE RIEUL-KIYEOK" },
	{ 0x481D , "HANGUL SYLLABLE NIEUN OE RIEUL-MIEUM" },
	{ 0x481E , "HANGUL SYLLABLE NIEUN OE RIEUL-PIEUP" },
	{ 0x481F , "HANGUL SYLLABLE NIEUN OE RIEUL-SIOS" },
	{ 0x4820 , "HANGUL SYLLABLE NIEUN OE RIEUL-THIEUTH" },
	{ 0x4821 , "HANGUL SYLLABLE NIEUN OE RIEUL-PHIEUPH" },
	{ 0x4822 , "HANGUL SYLLABLE NIEUN OE RIEUL-HIEUH" },
	{ 0x4823 , "HANGUL SYLLABLE NIEUN OE PIEUP-SIOS" },
	{ 0x4824 , "HANGUL SYLLABLE NIEUN OE CIEUC" },
	{ 0x4825 , "HANGUL SYLLABLE NIEUN OE CHIEUCH" },
	{ 0x4826 , "HANGUL SYLLABLE NIEUN OE KHIEUKH" },
	{ 0x4827 , "HANGUL SYLLABLE NIEUN OE THIEUTH" },
	{ 0x4828 , "HANGUL SYLLABLE NIEUN OE HIEUH" },
	{ 0x4829 , "HANGUL SYLLABLE NIEUN YO SSANGKIYEOK" },
	{ 0x482A , "HANGUL SYLLABLE NIEUN YO KIYEOK-SIOS" },
	{ 0x482B , "HANGUL SYLLABLE NIEUN YO NIEUN-CIEUC" },
	{ 0x482C , "HANGUL SYLLABLE NIEUN YO NIEUN-HIEUH" },
	{ 0x482D , "HANGUL SYLLABLE NIEUN YO RIEUL-KIYEOK" },
	{ 0x482E , "HANGUL SYLLABLE NIEUN YO RIEUL-MIEUM" },
	{ 0x482F , "HANGUL SYLLABLE NIEUN YO RIEUL-PIEUP" },
	{ 0x4830 , "HANGUL SYLLABLE NIEUN YO RIEUL-SIOS" },
	{ 0x4831 , "HANGUL SYLLABLE NIEUN YO RIEUL-THIEUTH" },
	{ 0x4832 , "HANGUL SYLLABLE NIEUN YO RIEUL-PHIEUPH" },
	{ 0x4833 , "HANGUL SYLLABLE NIEUN YO RIEUL-HIEUH" },
	{ 0x4834 , "HANGUL SYLLABLE NIEUN YO PIEUP-SIOS" },
	{ 0x4835 , "HANGUL SYLLABLE NIEUN YO SSANGSIOS" },
	{ 0x4836 , "HANGUL SYLLABLE NIEUN YO CIEUC" },
	{ 0x4837 , "HANGUL SYLLABLE NIEUN YO CHIEUCH" },
	{ 0x4838 , "HANGUL SYLLABLE NIEUN YO KHIEUKH" },
	{ 0x4839 , "HANGUL SYLLABLE NIEUN YO THIEUTH" },
	{ 0x483A , "HANGUL SYLLABLE NIEUN YO PHIEUPH" },
	{ 0x483B , "HANGUL SYLLABLE NIEUN YO HIEUH" },
	{ 0x483C , "HANGUL SYLLABLE NIEUN U SSANGKIYEOK" },
	{ 0x483D , "HANGUL SYLLABLE NIEUN U KIYEOK-SIOS" },
	{ 0x483E , "HANGUL SYLLABLE NIEUN U NIEUN-CIEUC" },
	{ 0x483F , "HANGUL SYLLABLE NIEUN U NIEUN-HIEUH" },
	{ 0x4840 , "HANGUL SYLLABLE NIEUN U RIEUL-MIEUM" },
	{ 0x4841 , "HANGUL SYLLABLE NIEUN U RIEUL-PIEUP" },
	{ 0x4842 , "HANGUL SYLLABLE NIEUN U RIEUL-THIEUTH" },
	{ 0x4843 , "HANGUL SYLLABLE NIEUN U RIEUL-PHIEUPH" },
	{ 0x4844 , "HANGUL SYLLABLE NIEUN U PIEUP-SIOS" },
	{ 0x4845 , "HANGUL SYLLABLE NIEUN U SSANGSIOS" },
	{ 0x4846 , "HANGUL SYLLABLE NIEUN U CIEUC" },
	{ 0x4847 , "HANGUL SYLLABLE NIEUN U KHIEUKH" },
	{ 0x4848 , "HANGUL SYLLABLE NIEUN U THIEUTH" },
	{ 0x4849 , "HANGUL SYLLABLE NIEUN U HIEUH" },
	{ 0x484A , "HANGUL SYLLABLE NIEUN WEO SSANGKIYEOK" },
	{ 0x484B , "HANGUL SYLLABLE NIEUN WEO KIYEOK-SIOS" },
	{ 0x484C , "HANGUL SYLLABLE NIEUN WEO NIEUN-CIEUC" },
	{ 0x484D , "HANGUL SYLLABLE NIEUN WEO NIEUN-HIEUH" },
	{ 0x484E , "HANGUL SYLLABLE NIEUN WEO RIEUL-KIYEOK" },
	{ 0x484F , "HANGUL SYLLABLE NIEUN WEO RIEUL-MIEUM" },
	{ 0x4850 , "HANGUL SYLLABLE NIEUN WEO RIEUL-PIEUP" },
	{ 0x4851 , "HANGUL SYLLABLE NIEUN WEO RIEUL-SIOS" },
	{ 0x4852 , "HANGUL SYLLABLE NIEUN WEO RIEUL-THIEUTH" },
	{ 0x4853 , "HANGUL SYLLABLE NIEUN WEO RIEUL-PHIEUPH" },
	{ 0x4854 , "HANGUL SYLLABLE NIEUN WEO RIEUL-HIEUH" },
	{ 0x4855 , "HANGUL SYLLABLE NIEUN WEO PIEUP-SIOS" },
	{ 0x4856 , "HANGUL SYLLABLE NIEUN WEO CIEUC" },
	{ 0x4857 , "HANGUL SYLLABLE NIEUN WEO CHIEUCH" },
	{ 0x4858 , "HANGUL SYLLABLE NIEUN WEO KHIEUKH" },
	{ 0x4859 , "HANGUL SYLLABLE NIEUN WEO THIEUTH" },
	{ 0x485A , "HANGUL SYLLABLE NIEUN WEO PHIEUPH" },
	{ 0x485B , "HANGUL SYLLABLE NIEUN WEO HIEUH" },
	{ 0x485C , "HANGUL SYLLABLE NIEUN WE SSANGKIYEOK" },
	{ 0x485D , "HANGUL SYLLABLE NIEUN WE KIYEOK-SIOS" },
	{ 0x485E , "HANGUL SYLLABLE NIEUN WE NIEUN-CIEUC" },
	{ 0x485F , "HANGUL SYLLABLE NIEUN WE NIEUN-HIEUH" },
	{ 0x4860 , "HANGUL SYLLABLE NIEUN WE RIEUL-KIYEOK" },
	{ 0x4861 , "HANGUL SYLLABLE NIEUN WE RIEUL-MIEUM" },
	{ 0x4862 , "HANGUL SYLLABLE NIEUN WE RIEUL-PIEUP" },
	{ 0x4863 , "HANGUL SYLLABLE NIEUN WE RIEUL-SIOS" },
	{ 0x4864 , "HANGUL SYLLABLE NIEUN WE RIEUL-THIEUTH" },
	{ 0x4865 , "HANGUL SYLLABLE NIEUN WE RIEUL-PHIEUPH" },
	{ 0x4866 , "HANGUL SYLLABLE NIEUN WE RIEUL-HIEUH" },
	{ 0x4867 , "HANGUL SYLLABLE NIEUN WE PIEUP-SIOS" },
	{ 0x4868 , "HANGUL SYLLABLE NIEUN WE CIEUC" },
	{ 0x4869 , "HANGUL SYLLABLE NIEUN WE CHIEUCH" },
	{ 0x486A , "HANGUL SYLLABLE NIEUN WE KHIEUKH" },
	{ 0x486B , "HANGUL SYLLABLE NIEUN WE THIEUTH" },
	{ 0x486C , "HANGUL SYLLABLE NIEUN WE PHIEUPH" },
	{ 0x486D , "HANGUL SYLLABLE NIEUN WE HIEUH" },
	{ 0x486E , "HANGUL SYLLABLE NIEUN WI SSANGKIYEOK" },
	{ 0x486F , "HANGUL SYLLABLE NIEUN WI KIYEOK-SIOS" },
	{ 0x4870 , "HANGUL SYLLABLE NIEUN WI NIEUN-CIEUC" },
	{ 0x4871 , "HANGUL SYLLABLE NIEUN WI NIEUN-HIEUH" },
	{ 0x4872 , "HANGUL SYLLABLE NIEUN WI RIEUL-KIYEOK" },
	{ 0x4873 , "HANGUL SYLLABLE NIEUN WI RIEUL-MIEUM" },
	{ 0x4874 , "HANGUL SYLLABLE NIEUN WI RIEUL-PIEUP" },
	{ 0x4875 , "HANGUL SYLLABLE NIEUN WI RIEUL-SIOS" },
	{ 0x4876 , "HANGUL SYLLABLE NIEUN WI RIEUL-THIEUTH" },
	{ 0x4877 , "HANGUL SYLLABLE NIEUN WI RIEUL-PHIEUPH" },
	{ 0x4878 , "HANGUL SYLLABLE NIEUN WI RIEUL-HIEUH" },
	{ 0x4879 , "HANGUL SYLLABLE NIEUN WI PIEUP-SIOS" },
	{ 0x487A , "HANGUL SYLLABLE NIEUN WI SSANGSIOS" },
	{ 0x487B , "HANGUL SYLLABLE NIEUN WI CIEUC" },
	{ 0x487C , "HANGUL SYLLABLE NIEUN WI KHIEUKH" },
	{ 0x487D , "HANGUL SYLLABLE NIEUN WI THIEUTH" },
	{ 0x487E , "HANGUL SYLLABLE NIEUN WI PHIEUPH" },
	{ 0x487F , "HANGUL SYLLABLE NIEUN WI HIEUH" },
	{ 0x4880 , "HANGUL SYLLABLE NIEUN YU SSANGKIYEOK" },
	{ 0x4881 , "HANGUL SYLLABLE NIEUN YU KIYEOK-SIOS" },
	{ 0x4882 , "HANGUL SYLLABLE NIEUN YU NIEUN-CIEUC" },
	{ 0x4883 , "HANGUL SYLLABLE NIEUN YU NIEUN-HIEUH" },
	{ 0x4884 , "HANGUL SYLLABLE NIEUN YU RIEUL-KIYEOK" },
	{ 0x4885 , "HANGUL SYLLABLE NIEUN YU RIEUL-MIEUM" },
	{ 0x4886 , "HANGUL SYLLABLE NIEUN YU RIEUL-PIEUP" },
	{ 0x4887 , "HANGUL SYLLABLE NIEUN YU RIEUL-SIOS" },
	{ 0x4888 , "HANGUL SYLLABLE NIEUN YU RIEUL-THIEUTH" },
	{ 0x4889 , "HANGUL SYLLABLE NIEUN YU RIEUL-PHIEUPH" },
	{ 0x488A , "HANGUL SYLLABLE NIEUN YU RIEUL-HIEUH" },
	{ 0x488B , "HANGUL SYLLABLE NIEUN YU PIEUP-SIOS" },
	{ 0x488C , "HANGUL SYLLABLE NIEUN YU SSANGSIOS" },
	{ 0x488D , "HANGUL SYLLABLE NIEUN YU KHIEUKH" },
	{ 0x488E , "HANGUL SYLLABLE NIEUN YU THIEUTH" },
	{ 0x488F , "HANGUL SYLLABLE NIEUN YU PHIEUPH" },
	{ 0x4890 , "HANGUL SYLLABLE NIEUN YU HIEUH" },
	{ 0x4891 , "HANGUL SYLLABLE NIEUN EU SSANGKIYEOK" },
	{ 0x4892 , "HANGUL SYLLABLE NIEUN EU KIYEOK-SIOS" },
	{ 0x4893 , "HANGUL SYLLABLE NIEUN EU NIEUN-CIEUC" },
	{ 0x4894 , "HANGUL SYLLABLE NIEUN EU NIEUN-HIEUH" },
	{ 0x4895 , "HANGUL SYLLABLE NIEUN EU RIEUL-PIEUP" },
	{ 0x4896 , "HANGUL SYLLABLE NIEUN EU RIEUL-THIEUTH" },
	{ 0x4897 , "HANGUL SYLLABLE NIEUN EU RIEUL-PHIEUPH" },
	{ 0x4898 , "HANGUL SYLLABLE NIEUN EU RIEUL-HIEUH" },
	{ 0x4899 , "HANGUL SYLLABLE NIEUN EU PIEUP-SIOS" },
	{ 0x489A , "HANGUL SYLLABLE NIEUN EU SSANGSIOS" },
	{ 0x489B , "HANGUL SYLLABLE NIEUN EU KHIEUKH" },
	{ 0x489C , "HANGUL SYLLABLE NIEUN EU THIEUTH" },
	{ 0x489D , "HANGUL SYLLABLE NIEUN YI SSANGKIYEOK" },
	{ 0x489E , "HANGUL SYLLABLE NIEUN YI KIYEOK-SIOS" },
	{ 0x489F , "HANGUL SYLLABLE NIEUN YI NIEUN-CIEUC" },
	{ 0x48A0 , "HANGUL SYLLABLE NIEUN YI NIEUN-HIEUH" },
	{ 0x48A1 , "HANGUL SYLLABLE NIEUN YI RIEUL-KIYEOK" },
	{ 0x48A2 , "HANGUL SYLLABLE NIEUN YI RIEUL-MIEUM" },
	{ 0x48A3 , "HANGUL SYLLABLE NIEUN YI RIEUL-PIEUP" },
	{ 0x48A4 , "HANGUL SYLLABLE NIEUN YI RIEUL-SIOS" },
	{ 0x48A5 , "HANGUL SYLLABLE NIEUN YI RIEUL-THIEUTH" },
	{ 0x48A6 , "HANGUL SYLLABLE NIEUN YI RIEUL-PHIEUPH" },
	{ 0x48A7 , "HANGUL SYLLABLE NIEUN YI RIEUL-HIEUH" },
	{ 0x48A8 , "HANGUL SYLLABLE NIEUN YI PIEUP-SIOS" },
	{ 0x48A9 , "HANGUL SYLLABLE NIEUN YI CIEUC" },
	{ 0x48AA , "HANGUL SYLLABLE NIEUN YI CHIEUCH" },
	{ 0x48AB , "HANGUL SYLLABLE NIEUN YI KHIEUKH" },
	{ 0x48AC , "HANGUL SYLLABLE NIEUN YI THIEUTH" },
	{ 0x48AD , "HANGUL SYLLABLE NIEUN YI PHIEUPH" },
	{ 0x48AE , "HANGUL SYLLABLE NIEUN YI HIEUH" },
	{ 0x48AF , "HANGUL SYLLABLE NIEUN I SSANGKIYEOK" },
	{ 0x48B0 , "HANGUL SYLLABLE NIEUN I KIYEOK-SIOS" },
	{ 0x48B1 , "HANGUL SYLLABLE NIEUN I NIEUN-CIEUC" },
	{ 0x48B2 , "HANGUL SYLLABLE NIEUN I NIEUN-HIEUH" },
	{ 0x48B3 , "HANGUL SYLLABLE NIEUN I RIEUL-THIEUTH" },
	{ 0x48B4 , "HANGUL SYLLABLE NIEUN I RIEUL-PHIEUPH" },
	{ 0x48B5 , "HANGUL SYLLABLE NIEUN I RIEUL-HIEUH" },
	{ 0x48B6 , "HANGUL SYLLABLE NIEUN I PIEUP-SIOS" },
	{ 0x48B7 , "HANGUL SYLLABLE NIEUN I SSANGSIOS" },
	{ 0x48B8 , "HANGUL SYLLABLE NIEUN I CHIEUCH" },
	{ 0x48B9 , "HANGUL SYLLABLE NIEUN I THIEUTH" },
	{ 0x48BA , "HANGUL SYLLABLE TIKEUT A KIYEOK-SIOS" },
	{ 0x48BB , "HANGUL SYLLABLE TIKEUT A NIEUN-CIEUC" },
	{ 0x48BC , "HANGUL SYLLABLE TIKEUT A NIEUN-HIEUH" },
	{ 0x48BD , "HANGUL SYLLABLE TIKEUT A RIEUL-THIEUTH" },
	{ 0x48BE , "HANGUL SYLLABLE TIKEUT A PHIEUPH" },
	{ 0x48BF , "HANGUL SYLLABLE TIKEUT AE SSANGKIYEOK" },
	{ 0x48C0 , "HANGUL SYLLABLE TIKEUT AE KIYEOK-SIOS" },
	{ 0x48C1 , "HANGUL SYLLABLE TIKEUT AE NIEUN-CIEUC" },
	{ 0x48C2 , "HANGUL SYLLABLE TIKEUT AE NIEUN-HIEUH" },
	{ 0x48C3 , "HANGUL SYLLABLE TIKEUT AE RIEUL-KIYEOK" },
	{ 0x48C4 , "HANGUL SYLLABLE TIKEUT AE RIEUL-MIEUM" },
	{ 0x48C5 , "HANGUL SYLLABLE TIKEUT AE RIEUL-PIEUP" },
	{ 0x48C6 , "HANGUL SYLLABLE TIKEUT AE RIEUL-SIOS" },
	{ 0x48C7 , "HANGUL SYLLABLE TIKEUT AE RIEUL-THIEUTH" },
	{ 0x48C8 , "HANGUL SYLLABLE TIKEUT AE RIEUL-PHIEUPH" },
	{ 0x48C9 , "HANGUL SYLLABLE TIKEUT AE RIEUL-HIEUH" },
	{ 0x48CA , "HANGUL SYLLABLE TIKEUT AE PIEUP-SIOS" },
	{ 0x48CB , "HANGUL SYLLABLE TIKEUT AE CIEUC" },
	{ 0x48CC , "HANGUL SYLLABLE TIKEUT AE CHIEUCH" },
	{ 0x48CD , "HANGUL SYLLABLE TIKEUT AE KHIEUKH" },
	{ 0x48CE , "HANGUL SYLLABLE TIKEUT AE PHIEUPH" },
	{ 0x48CF , "HANGUL SYLLABLE TIKEUT AE HIEUH" },
	{ 0x48D0 , "HANGUL SYLLABLE TIKEUT YA SSANGKIYEOK" },
	{ 0x48D1 , "HANGUL SYLLABLE TIKEUT YA KIYEOK-SIOS" },
	{ 0x48D2 , "HANGUL SYLLABLE TIKEUT YA NIEUN-CIEUC" },
	{ 0x48D3 , "HANGUL SYLLABLE TIKEUT YA NIEUN-HIEUH" },
	{ 0x48D4 , "HANGUL SYLLABLE TIKEUT YA TIKEUT" },
	{ 0x48D5 , "HANGUL SYLLABLE TIKEUT YA RIEUL-KIYEOK" },
	{ 0x48D6 , "HANGUL SYLLABLE TIKEUT YA RIEUL-MIEUM" },
	{ 0x48D7 , "HANGUL SYLLABLE TIKEUT YA RIEUL-SIOS" },
	{ 0x48D8 , "HANGUL SYLLABLE TIKEUT YA RIEUL-THIEUTH" },
	{ 0x48D9 , "HANGUL SYLLABLE TIKEUT YA RIEUL-PHIEUPH" },
	{ 0x48DA , "HANGUL SYLLABLE TIKEUT YA RIEUL-HIEUH" },
	{ 0x48DB , "HANGUL SYLLABLE TIKEUT YA PIEUP-SIOS" },
	{ 0x48DC , "HANGUL SYLLABLE TIKEUT YA SSANGSIOS" },
	{ 0x48DD , "HANGUL SYLLABLE TIKEUT YA CIEUC" },
	{ 0x48DE , "HANGUL SYLLABLE TIKEUT YA CHIEUCH" },
	{ 0x48DF , "HANGUL SYLLABLE TIKEUT YA KHIEUKH" },
	{ 0x48E0 , "HANGUL SYLLABLE TIKEUT YA THIEUTH" },
	{ 0x48E1 , "HANGUL SYLLABLE TIKEUT YA PHIEUPH" },
	{ 0x48E2 , "HANGUL SYLLABLE TIKEUT YA HIEUH" },
	{ 0x48E3 , "HANGUL SYLLABLE TIKEUT YAE KIYEOK" },
	{ 0x48E4 , "HANGUL SYLLABLE TIKEUT YAE SSANGKIYEOK" },
	{ 0x48E5 , "HANGUL SYLLABLE TIKEUT YAE KIYEOK-SIOS" },
	{ 0x48E6 , "HANGUL SYLLABLE TIKEUT YAE NIEUN-CIEUC" },
	{ 0x48E7 , "HANGUL SYLLABLE TIKEUT YAE NIEUN-HIEUH" },
	{ 0x48E8 , "HANGUL SYLLABLE TIKEUT YAE TIKEUT" },
	{ 0x48E9 , "HANGUL SYLLABLE TIKEUT YAE RIEUL" },
	{ 0x48EA , "HANGUL SYLLABLE TIKEUT YAE RIEUL-KIYEOK" },
	{ 0x48EB , "HANGUL SYLLABLE TIKEUT YAE RIEUL-MIEUM" },
	{ 0x48EC , "HANGUL SYLLABLE TIKEUT YAE RIEUL-PIEUP" },
	{ 0x48ED , "HANGUL SYLLABLE TIKEUT YAE RIEUL-SIOS" },
	{ 0x48EE , "HANGUL SYLLABLE TIKEUT YAE RIEUL-THIEUTH" },
	{ 0x48EF , "HANGUL SYLLABLE TIKEUT YAE RIEUL-PHIEUPH" },
	{ 0x48F0 , "HANGUL SYLLABLE TIKEUT YAE RIEUL-HIEUH" },
	{ 0x48F1 , "HANGUL SYLLABLE TIKEUT YAE MIEUM" },
	{ 0x48F2 , "HANGUL SYLLABLE TIKEUT YAE PIEUP" },
	{ 0x48F3 , "HANGUL SYLLABLE TIKEUT YAE PIEUP-SIOS" },
	{ 0x48F4 , "HANGUL SYLLABLE TIKEUT YAE SIOS" },
	{ 0x48F5 , "HANGUL SYLLABLE TIKEUT YAE SSANGSIOS" },
	{ 0x48F6 , "HANGUL SYLLABLE TIKEUT YAE CIEUC" },
	{ 0x48F7 , "HANGUL SYLLABLE TIKEUT YAE CHIEUCH" },
	{ 0x48F8 , "HANGUL SYLLABLE TIKEUT YAE KHIEUKH" },
	{ 0x48F9 , "HANGUL SYLLABLE TIKEUT YAE THIEUTH" },
	{ 0x48FA , "HANGUL SYLLABLE TIKEUT YAE PHIEUPH" },
	{ 0x48FB , "HANGUL SYLLABLE TIKEUT YAE HIEUH" },
	{ 0x48FC , "HANGUL SYLLABLE TIKEUT EO KIYEOK-SIOS" },
	{ 0x48FD , "HANGUL SYLLABLE TIKEUT EO NIEUN-HIEUH" },
	{ 0x48FE , "HANGUL SYLLABLE TIKEUT EO RIEUL-KIYEOK" },
	{ 0x48FF , "HANGUL SYLLABLE TIKEUT EO RIEUL-SIOS" },
	{ 0x4900 , "HANGUL SYLLABLE TIKEUT EO RIEUL-THIEUTH" },
	{ 0x4901 , "HANGUL SYLLABLE TIKEUT EO RIEUL-PHIEUPH" },
	{ 0x4902 , "HANGUL SYLLABLE TIKEUT EO RIEUL-HIEUH" },
	{ 0x4903 , "HANGUL SYLLABLE TIKEUT EO KHIEUKH" },
	{ 0x4904 , "HANGUL SYLLABLE TIKEUT E SSANGKIYEOK" },
	{ 0x4905 , "HANGUL SYLLABLE TIKEUT E KIYEOK-SIOS" },
	{ 0x4906 , "HANGUL SYLLABLE TIKEUT E NIEUN-CIEUC" },
	{ 0x4907 , "HANGUL SYLLABLE TIKEUT E NIEUN-HIEUH" },
	{ 0x4908 , "HANGUL SYLLABLE TIKEUT E RIEUL-KIYEOK" },
	{ 0x4909 , "HANGUL SYLLABLE TIKEUT E RIEUL-MIEUM" },
	{ 0x490A , "HANGUL SYLLABLE TIKEUT E RIEUL-PIEUP" },
	{ 0x490B , "HANGUL SYLLABLE TIKEUT E RIEUL-SIOS" },
	{ 0x490C , "HANGUL SYLLABLE TIKEUT E RIEUL-THIEUTH" },
	{ 0x490D , "HANGUL SYLLABLE TIKEUT E RIEUL-PHIEUPH" },
	{ 0x490E , "HANGUL SYLLABLE TIKEUT E RIEUL-HIEUH" },
	{ 0x490F , "HANGUL SYLLABLE TIKEUT E PIEUP-SIOS" },
	{ 0x4910 , "HANGUL SYLLABLE TIKEUT E CHIEUCH" },
	{ 0x4911 , "HANGUL SYLLABLE TIKEUT E KHIEUKH" },
	{ 0x4912 , "HANGUL SYLLABLE TIKEUT E THIEUTH" },
	{ 0x4913 , "HANGUL SYLLABLE TIKEUT E HIEUH" },
	{ 0x4914 , "HANGUL SYLLABLE TIKEUT YEO SSANGKIYEOK" },
	{ 0x4915 , "HANGUL SYLLABLE TIKEUT YEO NIEUN-CIEUC" },
	{ 0x4916 , "HANGUL SYLLABLE TIKEUT YEO NIEUN-HIEUH" },
	{ 0x4917 , "HANGUL SYLLABLE TIKEUT YEO RIEUL-KIYEOK" },
	{ 0x4918 , "HANGUL SYLLABLE TIKEUT YEO RIEUL-MIEUM" },
	{ 0x4919 , "HANGUL SYLLABLE TIKEUT YEO RIEUL-PIEUP" },
	{ 0x491A , "HANGUL SYLLABLE TIKEUT YEO RIEUL-THIEUTH" },
	{ 0x491B , "HANGUL SYLLABLE TIKEUT YEO RIEUL-PHIEUPH" },
	{ 0x491C , "HANGUL SYLLABLE TIKEUT YEO RIEUL-HIEUH" },
	{ 0x491D , "HANGUL SYLLABLE TIKEUT YEO PIEUP-SIOS" },
	{ 0x491E , "HANGUL SYLLABLE TIKEUT YEO CIEUC" },
	{ 0x491F , "HANGUL SYLLABLE TIKEUT YEO CHIEUCH" },
	{ 0x4920 , "HANGUL SYLLABLE TIKEUT YEO KHIEUKH" },
	{ 0x4921 , "HANGUL SYLLABLE TIKEUT YEO THIEUTH" },
	{ 0x4922 , "HANGUL SYLLABLE TIKEUT YEO PHIEUPH" },
	{ 0x4923 , "HANGUL SYLLABLE TIKEUT YEO HIEUH" },
	{ 0x4924 , "HANGUL SYLLABLE TIKEUT YE SSANGKIYEOK" },
	{ 0x4925 , "HANGUL SYLLABLE TIKEUT YE KIYEOK-SIOS" },
	{ 0x4926 , "HANGUL SYLLABLE TIKEUT YE NIEUN-CIEUC" },
	{ 0x4927 , "HANGUL SYLLABLE TIKEUT YE NIEUN-HIEUH" },
	{ 0x4928 , "HANGUL SYLLABLE TIKEUT YE TIKEUT" },
	{ 0x4929 , "HANGUL SYLLABLE TIKEUT YE RIEUL-KIYEOK" },
	{ 0x492A , "HANGUL SYLLABLE TIKEUT YE RIEUL-MIEUM" },
	{ 0x492B , "HANGUL SYLLABLE TIKEUT YE RIEUL-PIEUP" },
	{ 0x492C , "HANGUL SYLLABLE TIKEUT YE RIEUL-SIOS" },
	{ 0x492D , "HANGUL SYLLABLE TIKEUT YE RIEUL-THIEUTH" },
	{ 0x492E , "HANGUL SYLLABLE TIKEUT YE RIEUL-PHIEUPH" },
	{ 0x492F , "HANGUL SYLLABLE TIKEUT YE RIEUL-HIEUH" },
	{ 0x4930 , "HANGUL SYLLABLE TIKEUT YE PIEUP-SIOS" },
	{ 0x4931 , "HANGUL SYLLABLE TIKEUT YE SSANGSIOS" },
	{ 0x4932 , "HANGUL SYLLABLE TIKEUT YE CIEUC" },
	{ 0x4933 , "HANGUL SYLLABLE TIKEUT YE CHIEUCH" },
	{ 0x4934 , "HANGUL SYLLABLE TIKEUT YE KHIEUKH" },
	{ 0x4935 , "HANGUL SYLLABLE TIKEUT YE THIEUTH" },
	{ 0x4936 , "HANGUL SYLLABLE TIKEUT YE PHIEUPH" },
	{ 0x4937 , "HANGUL SYLLABLE TIKEUT YE HIEUH" },
	{ 0x4938 , "HANGUL SYLLABLE TIKEUT O NIEUN-CIEUC" },
	{ 0x4939 , "HANGUL SYLLABLE TIKEUT O NIEUN-HIEUH" },
	{ 0x493A , "HANGUL SYLLABLE TIKEUT O RIEUL-KIYEOK" },
	{ 0x493B , "HANGUL SYLLABLE TIKEUT O RIEUL-THIEUTH" },
	{ 0x493C , "HANGUL SYLLABLE TIKEUT O RIEUL-PHIEUPH" },
	{ 0x493D , "HANGUL SYLLABLE TIKEUT O PIEUP-SIOS" },
	{ 0x493E , "HANGUL SYLLABLE TIKEUT O SSANGSIOS" },
	{ 0x493F , "HANGUL SYLLABLE TIKEUT WA SSANGKIYEOK" },
	{ 0x4940 , "HANGUL SYLLABLE TIKEUT WA KIYEOK-SIOS" },
	{ 0x4941 , "HANGUL SYLLABLE TIKEUT WA NIEUN-CIEUC" },
	{ 0x4942 , "HANGUL SYLLABLE TIKEUT WA NIEUN-HIEUH" },
	{ 0x4943 , "HANGUL SYLLABLE TIKEUT WA TIKEUT" },
	{ 0x4944 , "HANGUL SYLLABLE TIKEUT WA RIEUL-KIYEOK" },
	{ 0x4945 , "HANGUL SYLLABLE TIKEUT WA RIEUL-MIEUM" },
	{ 0x4946 , "HANGUL SYLLABLE TIKEUT WA RIEUL-PIEUP" },
	{ 0x4947 , "HANGUL SYLLABLE TIKEUT WA RIEUL-SIOS" },
	{ 0x4948 , "HANGUL SYLLABLE TIKEUT WA RIEUL-THIEUTH" },
	{ 0x4949 , "HANGUL SYLLABLE TIKEUT WA RIEUL-PHIEUPH" },
	{ 0x494A , "HANGUL SYLLABLE TIKEUT WA RIEUL-HIEUH" },
	{ 0x494B , "HANGUL SYLLABLE TIKEUT WA PIEUP-SIOS" },
	{ 0x494C , "HANGUL SYLLABLE TIKEUT WA CIEUC" },
	{ 0x494D , "HANGUL SYLLABLE TIKEUT WA CHIEUCH" },
	{ 0x494E , "HANGUL SYLLABLE TIKEUT WA KHIEUKH" },
	{ 0x494F , "HANGUL SYLLABLE TIKEUT WA THIEUTH" },
	{ 0x4950 , "HANGUL SYLLABLE TIKEUT WA PHIEUPH" },
	{ 0x4951 , "HANGUL SYLLABLE TIKEUT WA HIEUH" },
	{ 0x4952 , "HANGUL SYLLABLE TIKEUT WAE KIYEOK" },
	{ 0x4953 , "HANGUL SYLLABLE TIKEUT WAE SSANGKIYEOK" },
	{ 0x4954 , "HANGUL SYLLABLE TIKEUT WAE KIYEOK-SIOS" },
	{ 0x4955 , "HANGUL SYLLABLE TIKEUT WAE NIEUN-CIEUC" },
	{ 0x4956 , "HANGUL SYLLABLE TIKEUT WAE NIEUN-HIEUH" },
	{ 0x4957 , "HANGUL SYLLABLE TIKEUT WAE RIEUL-KIYEOK" },
	{ 0x4958 , "HANGUL SYLLABLE TIKEUT WAE RIEUL-MIEUM" },
	{ 0x4959 , "HANGUL SYLLABLE TIKEUT WAE RIEUL-PIEUP" },
	{ 0x495A , "HANGUL SYLLABLE TIKEUT WAE RIEUL-SIOS" },
	{ 0x495B , "HANGUL SYLLABLE TIKEUT WAE RIEUL-THIEUTH" },
	{ 0x495C , "HANGUL SYLLABLE TIKEUT WAE RIEUL-PHIEUPH" },
	{ 0x495D , "HANGUL SYLLABLE TIKEUT WAE RIEUL-HIEUH" },
	{ 0x495E , "HANGUL SYLLABLE TIKEUT WAE PIEUP-SIOS" },
	{ 0x495F , "HANGUL SYLLABLE TIKEUT WAE IEUNG" },
	{ 0x4960 , "HANGUL SYLLABLE TIKEUT WAE CIEUC" },
	{ 0x4961 , "HANGUL SYLLABLE TIKEUT WAE CHIEUCH" },
	{ 0x4962 , "HANGUL SYLLABLE TIKEUT WAE KHIEUKH" },
	{ 0x4963 , "HANGUL SYLLABLE TIKEUT WAE THIEUTH" },
	{ 0x4964 , "HANGUL SYLLABLE TIKEUT WAE PHIEUPH" },
	{ 0x4965 , "HANGUL SYLLABLE TIKEUT WAE HIEUH" },
	{ 0x4966 , "HANGUL SYLLABLE TIKEUT OE SSANGKIYEOK" },
	{ 0x4967 , "HANGUL SYLLABLE TIKEUT OE KIYEOK-SIOS" },
	{ 0x4968 , "HANGUL SYLLABLE TIKEUT OE NIEUN-CIEUC" },
	{ 0x4969 , "HANGUL SYLLABLE TIKEUT OE NIEUN-HIEUH" },
	{ 0x496A , "HANGUL SYLLABLE TIKEUT OE RIEUL-KIYEOK" },
	{ 0x496B , "HANGUL SYLLABLE TIKEUT OE RIEUL-MIEUM" },
	{ 0x496C , "HANGUL SYLLABLE TIKEUT OE RIEUL-PIEUP" },
	{ 0x496D , "HANGUL SYLLABLE TIKEUT OE RIEUL-THIEUTH" },
	{ 0x496E , "HANGUL SYLLABLE TIKEUT OE RIEUL-PHIEUPH" },
	{ 0x496F , "HANGUL SYLLABLE TIKEUT OE RIEUL-HIEUH" },
	{ 0x4970 , "HANGUL SYLLABLE TIKEUT OE PIEUP-SIOS" },
	{ 0x4971 , "HANGUL SYLLABLE TIKEUT OE CIEUC" },
	{ 0x4972 , "HANGUL SYLLABLE TIKEUT OE CHIEUCH" },
	{ 0x4973 , "HANGUL SYLLABLE TIKEUT OE KHIEUKH" },
	{ 0x4974 , "HANGUL SYLLABLE TIKEUT OE THIEUTH" },
	{ 0x4975 , "HANGUL SYLLABLE TIKEUT OE PHIEUPH" },
	{ 0x4976 , "HANGUL SYLLABLE TIKEUT OE HIEUH" },
	{ 0x4977 , "HANGUL SYLLABLE TIKEUT YO SSANGKIYEOK" },
	{ 0x4978 , "HANGUL SYLLABLE TIKEUT YO KIYEOK-SIOS" },
	{ 0x4979 , "HANGUL SYLLABLE TIKEUT YO NIEUN-CIEUC" },
	{ 0x497A , "HANGUL SYLLABLE TIKEUT YO NIEUN-HIEUH" },
	{ 0x497B , "HANGUL SYLLABLE TIKEUT YO RIEUL-KIYEOK" },
	{ 0x497C , "HANGUL SYLLABLE TIKEUT YO RIEUL-MIEUM" },
	{ 0x497D , "HANGUL SYLLABLE TIKEUT YO RIEUL-PIEUP" },
	{ 0x497E , "HANGUL SYLLABLE TIKEUT YO RIEUL-SIOS" },
	{ 0x497F , "HANGUL SYLLABLE TIKEUT YO RIEUL-THIEUTH" },
	{ 0x4980 , "HANGUL SYLLABLE TIKEUT YO RIEUL-PHIEUPH" },
	{ 0x4981 , "HANGUL SYLLABLE TIKEUT YO RIEUL-HIEUH" },
	{ 0x4982 , "HANGUL SYLLABLE TIKEUT YO PIEUP-SIOS" },
	{ 0x4983 , "HANGUL SYLLABLE TIKEUT YO SSANGSIOS" },
	{ 0x4984 , "HANGUL SYLLABLE TIKEUT YO CIEUC" },
	{ 0x4985 , "HANGUL SYLLABLE TIKEUT YO CHIEUCH" },
	{ 0x4986 , "HANGUL SYLLABLE TIKEUT YO KHIEUKH" },
	{ 0x4987 , "HANGUL SYLLABLE TIKEUT YO THIEUTH" },
	{ 0x4988 , "HANGUL SYLLABLE TIKEUT YO PHIEUPH" },
	{ 0x4989 , "HANGUL SYLLABLE TIKEUT U SSANGKIYEOK" },
	{ 0x498A , "HANGUL SYLLABLE TIKEUT U KIYEOK-SIOS" },
	{ 0x498B , "HANGUL SYLLABLE TIKEUT U NIEUN-CIEUC" },
	{ 0x498C , "HANGUL SYLLABLE TIKEUT U NIEUN-HIEUH" },
	{ 0x498D , "HANGUL SYLLABLE TIKEUT U RIEUL-SIOS" },
	{ 0x498E , "HANGUL SYLLABLE TIKEUT U RIEUL-THIEUTH" },
	{ 0x498F , "HANGUL SYLLABLE TIKEUT U RIEUL-PHIEUPH" },
	{ 0x4990 , "HANGUL SYLLABLE TIKEUT U RIEUL-HIEUH" },
	{ 0x4991 , "HANGUL SYLLABLE TIKEUT U PIEUP-SIOS" },
	{ 0x4992 , "HANGUL SYLLABLE TIKEUT U SSANGSIOS" },
	{ 0x4993 , "HANGUL SYLLABLE TIKEUT U CIEUC" },
	{ 0x4994 , "HANGUL SYLLABLE TIKEUT U CHIEUCH" },
	{ 0x4995 , "HANGUL SYLLABLE TIKEUT U KHIEUKH" },
	{ 0x4996 , "HANGUL SYLLABLE TIKEUT U THIEUTH" },
	{ 0x4997 , "HANGUL SYLLABLE TIKEUT U HIEUH" },
	{ 0x4998 , "HANGUL SYLLABLE TIKEUT WEO SSANGKIYEOK" },
	{ 0x4999 , "HANGUL SYLLABLE TIKEUT WEO KIYEOK-SIOS" },
	{ 0x499A , "HANGUL SYLLABLE TIKEUT WEO NIEUN-CIEUC" },
	{ 0x499B , "HANGUL SYLLABLE TIKEUT WEO NIEUN-HIEUH" },
	{ 0x499C , "HANGUL SYLLABLE TIKEUT WEO RIEUL-KIYEOK" },
	{ 0x499D , "HANGUL SYLLABLE TIKEUT WEO RIEUL-MIEUM" },
	{ 0x499E , "HANGUL SYLLABLE TIKEUT WEO RIEUL-PIEUP" },
	{ 0x499F , "HANGUL SYLLABLE TIKEUT WEO RIEUL-SIOS" },
	{ 0x49A0 , "HANGUL SYLLABLE TIKEUT WEO RIEUL-THIEUTH" },
	{ 0x49A1 , "HANGUL SYLLABLE TIKEUT WEO RIEUL-PHIEUPH" },
	{ 0x49A2 , "HANGUL SYLLABLE TIKEUT WEO RIEUL-HIEUH" },
	{ 0x49A3 , "HANGUL SYLLABLE TIKEUT WEO PIEUP-SIOS" },
	{ 0x49A4 , "HANGUL SYLLABLE TIKEUT WEO CIEUC" },
	{ 0x49A5 , "HANGUL SYLLABLE TIKEUT WEO CHIEUCH" },
	{ 0x49A6 , "HANGUL SYLLABLE TIKEUT WEO KHIEUKH" },
	{ 0x49A7 , "HANGUL SYLLABLE TIKEUT WEO THIEUTH" },
	{ 0x49A8 , "HANGUL SYLLABLE TIKEUT WEO PHIEUPH" },
	{ 0x49A9 , "HANGUL SYLLABLE TIKEUT WEO HIEUH" },
	{ 0x49AA , "HANGUL SYLLABLE TIKEUT WE SSANGKIYEOK" },
	{ 0x49AB , "HANGUL SYLLABLE TIKEUT WE KIYEOK-SIOS" },
	{ 0x49AC , "HANGUL SYLLABLE TIKEUT WE NIEUN-CIEUC" },
	{ 0x49AD , "HANGUL SYLLABLE TIKEUT WE NIEUN-HIEUH" },
	{ 0x49AE , "HANGUL SYLLABLE TIKEUT WE RIEUL-KIYEOK" },
	{ 0x49AF , "HANGUL SYLLABLE TIKEUT WE RIEUL-MIEUM" },
	{ 0x49B0 , "HANGUL SYLLABLE TIKEUT WE RIEUL-PIEUP" },
	{ 0x49B1 , "HANGUL SYLLABLE TIKEUT WE RIEUL-SIOS" },
	{ 0x49B2 , "HANGUL SYLLABLE TIKEUT WE RIEUL-THIEUTH" },
	{ 0x49B3 , "HANGUL SYLLABLE TIKEUT WE RIEUL-PHIEUPH" },
	{ 0x49B4 , "HANGUL SYLLABLE TIKEUT WE RIEUL-HIEUH" },
	{ 0x49B5 , "HANGUL SYLLABLE TIKEUT WE PIEUP-SIOS" },
	{ 0x49B6 , "HANGUL SYLLABLE TIKEUT WE CIEUC" },
	{ 0x49B7 , "HANGUL SYLLABLE TIKEUT WE CHIEUCH" },
	{ 0x49B8 , "HANGUL SYLLABLE TIKEUT WE KHIEUKH" },
	{ 0x49B9 , "HANGUL SYLLABLE TIKEUT WE THIEUTH" },
	{ 0x49BA , "HANGUL SYLLABLE TIKEUT WE PHIEUPH" },
	{ 0x49BB , "HANGUL SYLLABLE TIKEUT WE HIEUH" },
	{ 0x49BC , "HANGUL SYLLABLE TIKEUT WI SSANGKIYEOK" },
	{ 0x49BD , "HANGUL SYLLABLE TIKEUT WI KIYEOK-SIOS" },
	{ 0x49BE , "HANGUL SYLLABLE TIKEUT WI NIEUN-CIEUC" },
	{ 0x49BF , "HANGUL SYLLABLE TIKEUT WI NIEUN-HIEUH" },
	{ 0x49C0 , "HANGUL SYLLABLE TIKEUT WI RIEUL-KIYEOK" },
	{ 0x49C1 , "HANGUL SYLLABLE TIKEUT WI RIEUL-MIEUM" },
	{ 0x49C2 , "HANGUL SYLLABLE TIKEUT WI RIEUL-PIEUP" },
	{ 0x49C3 , "HANGUL SYLLABLE TIKEUT WI RIEUL-SIOS" },
	{ 0x49C4 , "HANGUL SYLLABLE TIKEUT WI RIEUL-THIEUTH" },
	{ 0x49C5 , "HANGUL SYLLABLE TIKEUT WI RIEUL-PHIEUPH" },
	{ 0x49C6 , "HANGUL SYLLABLE TIKEUT WI RIEUL-HIEUH" },
	{ 0x49C7 , "HANGUL SYLLABLE TIKEUT WI PIEUP-SIOS" },
	{ 0x49C8 , "HANGUL SYLLABLE TIKEUT WI CIEUC" },
	{ 0x49C9 , "HANGUL SYLLABLE TIKEUT WI CHIEUCH" },
	{ 0x49CA , "HANGUL SYLLABLE TIKEUT WI KHIEUKH" },
	{ 0x49CB , "HANGUL SYLLABLE TIKEUT WI THIEUTH" },
	{ 0x49CC , "HANGUL SYLLABLE TIKEUT WI PHIEUPH" },
	{ 0x49CD , "HANGUL SYLLABLE TIKEUT WI HIEUH" },
	{ 0x49CE , "HANGUL SYLLABLE TIKEUT YU SSANGKIYEOK" },
	{ 0x49CF , "HANGUL SYLLABLE TIKEUT YU KIYEOK-SIOS" },
	{ 0x49D0 , "HANGUL SYLLABLE TIKEUT YU NIEUN-CIEUC" },
	{ 0x49D1 , "HANGUL SYLLABLE TIKEUT YU NIEUN-HIEUH" },
	{ 0x49D2 , "HANGUL SYLLABLE TIKEUT YU TIKEUT" },
	{ 0x49D3 , "HANGUL SYLLABLE TIKEUT YU RIEUL-KIYEOK" },
	{ 0x49D4 , "HANGUL SYLLABLE TIKEUT YU RIEUL-MIEUM" },
	{ 0x49D5 , "HANGUL SYLLABLE TIKEUT YU RIEUL-PIEUP" },
	{ 0x49D6 , "HANGUL SYLLABLE TIKEUT YU RIEUL-SIOS" },
	{ 0x49D7 , "HANGUL SYLLABLE TIKEUT YU RIEUL-THIEUTH" },
	{ 0x49D8 , "HANGUL SYLLABLE TIKEUT YU RIEUL-PHIEUPH" },
	{ 0x49D9 , "HANGUL SYLLABLE TIKEUT YU RIEUL-HIEUH" },
	{ 0x49DA , "HANGUL SYLLABLE TIKEUT YU PIEUP-SIOS" },
	{ 0x49DB , "HANGUL SYLLABLE TIKEUT YU SSANGSIOS" },
	{ 0x49DC , "HANGUL SYLLABLE TIKEUT YU CIEUC" },
	{ 0x49DD , "HANGUL SYLLABLE TIKEUT YU CHIEUCH" },
	{ 0x49DE , "HANGUL SYLLABLE TIKEUT YU KHIEUKH" },
	{ 0x49DF , "HANGUL SYLLABLE TIKEUT YU THIEUTH" },
	{ 0x49E0 , "HANGUL SYLLABLE TIKEUT YU PHIEUPH" },
	{ 0x49E1 , "HANGUL SYLLABLE TIKEUT YU HIEUH" },
	{ 0x49E2 , "HANGUL SYLLABLE TIKEUT EU SSANGKIYEOK" },
	{ 0x49E3 , "HANGUL SYLLABLE TIKEUT EU KIYEOK-SIOS" },
	{ 0x49E4 , "HANGUL SYLLABLE TIKEUT EU NIEUN-CIEUC" },
	{ 0x49E5 , "HANGUL SYLLABLE TIKEUT EU NIEUN-HIEUH" },
	{ 0x49E6 , "HANGUL SYLLABLE TIKEUT EU RIEUL-THIEUTH" },
	{ 0x49E7 , "HANGUL SYLLABLE TIKEUT EU RIEUL-PHIEUPH" },
	{ 0x49E8 , "HANGUL SYLLABLE TIKEUT EU RIEUL-HIEUH" },
	{ 0x49E9 , "HANGUL SYLLABLE TIKEUT EU PIEUP-SIOS" },
	{ 0x49EA , "HANGUL SYLLABLE TIKEUT EU SSANGSIOS" },
	{ 0x49EB , "HANGUL SYLLABLE TIKEUT EU CIEUC" },
	{ 0x49EC , "HANGUL SYLLABLE TIKEUT EU CHIEUCH" },
	{ 0x49ED , "HANGUL SYLLABLE TIKEUT EU KHIEUKH" },
	{ 0x49EE , "HANGUL SYLLABLE TIKEUT EU THIEUTH" },
	{ 0x49EF , "HANGUL SYLLABLE TIKEUT EU PHIEUPH" },
	{ 0x49F0 , "HANGUL SYLLABLE TIKEUT EU HIEUH" },
	{ 0x49F1 , "HANGUL SYLLABLE TIKEUT YI SSANGKIYEOK" },
	{ 0x49F2 , "HANGUL SYLLABLE TIKEUT YI KIYEOK-SIOS" },
	{ 0x49F3 , "HANGUL SYLLABLE TIKEUT YI NIEUN-CIEUC" },
	{ 0x49F4 , "HANGUL SYLLABLE TIKEUT YI NIEUN-HIEUH" },
	{ 0x49F5 , "HANGUL SYLLABLE TIKEUT YI RIEUL-KIYEOK" },
	{ 0x49F6 , "HANGUL SYLLABLE TIKEUT YI RIEUL-MIEUM" },
	{ 0x49F7 , "HANGUL SYLLABLE TIKEUT YI RIEUL-PIEUP" },
	{ 0x49F8 , "HANGUL SYLLABLE TIKEUT YI RIEUL-SIOS" },
	{ 0x49F9 , "HANGUL SYLLABLE TIKEUT YI RIEUL-THIEUTH" },
	{ 0x49FA , "HANGUL SYLLABLE TIKEUT YI RIEUL-PHIEUPH" },
	{ 0x49FB , "HANGUL SYLLABLE TIKEUT YI RIEUL-HIEUH" },
	{ 0x49FC , "HANGUL SYLLABLE TIKEUT YI PIEUP-SIOS" },
	{ 0x49FD , "HANGUL SYLLABLE TIKEUT YI SSANGSIOS" },
	{ 0x49FE , "HANGUL SYLLABLE TIKEUT YI CIEUC" },
	{ 0x49FF , "HANGUL SYLLABLE TIKEUT YI CHIEUCH" },
	{ 0x4A00 , "HANGUL SYLLABLE TIKEUT YI KHIEUKH" },
	{ 0x4A01 , "HANGUL SYLLABLE TIKEUT YI THIEUTH" },
	{ 0x4A02 , "HANGUL SYLLABLE TIKEUT YI PHIEUPH" },
	{ 0x4A03 , "HANGUL SYLLABLE TIKEUT YI HIEUH" },
	{ 0x4A04 , "HANGUL SYLLABLE TIKEUT I SSANGKIYEOK" },
	{ 0x4A05 , "HANGUL SYLLABLE TIKEUT I KIYEOK-SIOS" },
	{ 0x4A06 , "HANGUL SYLLABLE TIKEUT I NIEUN-CIEUC" },
	{ 0x4A07 , "HANGUL SYLLABLE TIKEUT I NIEUN-HIEUH" },
	{ 0x4A08 , "HANGUL SYLLABLE TIKEUT I RIEUL-KIYEOK" },
	{ 0x4A09 , "HANGUL SYLLABLE TIKEUT I RIEUL-MIEUM" },
	{ 0x4A0A , "HANGUL SYLLABLE TIKEUT I RIEUL-PIEUP" },
	{ 0x4A0B , "HANGUL SYLLABLE TIKEUT I RIEUL-SIOS" },
	{ 0x4A0C , "HANGUL SYLLABLE TIKEUT I RIEUL-THIEUTH" },
	{ 0x4A0D , "HANGUL SYLLABLE TIKEUT I RIEUL-PHIEUPH" },
	{ 0x4A0E , "HANGUL SYLLABLE TIKEUT I RIEUL-HIEUH" },
	{ 0x4A0F , "HANGUL SYLLABLE TIKEUT I PIEUP-SIOS" },
	{ 0x4A10 , "HANGUL SYLLABLE TIKEUT I KHIEUKH" },
	{ 0x4A11 , "HANGUL SYLLABLE SSANGTIKEUT A KIYEOK-SIOS" },
	{ 0x4A12 , "HANGUL SYLLABLE SSANGTIKEUT A NIEUN-CIEUC" },
	{ 0x4A13 , "HANGUL SYLLABLE SSANGTIKEUT A NIEUN-HIEUH" },
	{ 0x4A14 , "HANGUL SYLLABLE SSANGTIKEUT A RIEUL-SIOS" },
	{ 0x4A15 , "HANGUL SYLLABLE SSANGTIKEUT A RIEUL-THIEUTH" },
	{ 0x4A16 , "HANGUL SYLLABLE SSANGTIKEUT A RIEUL-PHIEUPH" },
	{ 0x4A17 , "HANGUL SYLLABLE SSANGTIKEUT A PIEUP-SIOS" },
	{ 0x4A18 , "HANGUL SYLLABLE SSANGTIKEUT A CIEUC" },
	{ 0x4A19 , "HANGUL SYLLABLE SSANGTIKEUT A CHIEUCH" },
	{ 0x4A1A , "HANGUL SYLLABLE SSANGTIKEUT A KHIEUKH" },
	{ 0x4A1B , "HANGUL SYLLABLE SSANGTIKEUT A THIEUTH" },
	{ 0x4A1C , "HANGUL SYLLABLE SSANGTIKEUT A PHIEUPH" },
	{ 0x4A1D , "HANGUL SYLLABLE SSANGTIKEUT AE KIYEOK-SIOS" },
	{ 0x4A1E , "HANGUL SYLLABLE SSANGTIKEUT AE NIEUN-CIEUC" },
	{ 0x4A1F , "HANGUL SYLLABLE SSANGTIKEUT AE NIEUN-HIEUH" },
	{ 0x4A20 , "HANGUL SYLLABLE SSANGTIKEUT AE RIEUL-KIYEOK" },
	{ 0x4A21 , "HANGUL SYLLABLE SSANGTIKEUT AE RIEUL-MIEUM" },
	{ 0x4A22 , "HANGUL SYLLABLE SSANGTIKEUT AE RIEUL-PIEUP" },
	{ 0x4A23 , "HANGUL SYLLABLE SSANGTIKEUT AE RIEUL-SIOS" },
	{ 0x4A24 , "HANGUL SYLLABLE SSANGTIKEUT AE RIEUL-THIEUTH" },
	{ 0x4A25 , "HANGUL SYLLABLE SSANGTIKEUT AE RIEUL-PHIEUPH" },
	{ 0x4A26 , "HANGUL SYLLABLE SSANGTIKEUT AE RIEUL-HIEUH" },
	{ 0x4A27 , "HANGUL SYLLABLE SSANGTIKEUT AE PIEUP-SIOS" },
	{ 0x4A28 , "HANGUL SYLLABLE SSANGTIKEUT AE CIEUC" },
	{ 0x4A29 , "HANGUL SYLLABLE SSANGTIKEUT AE CHIEUCH" },
	{ 0x4A2A , "HANGUL SYLLABLE SSANGTIKEUT AE KHIEUKH" },
	{ 0x4A2B , "HANGUL SYLLABLE SSANGTIKEUT AE THIEUTH" },
	{ 0x4A2C , "HANGUL SYLLABLE SSANGTIKEUT AE PHIEUPH" },
	{ 0x4A2D , "HANGUL SYLLABLE SSANGTIKEUT YA SSANGKIYEOK" },
	{ 0x4A2E , "HANGUL SYLLABLE SSANGTIKEUT YA KIYEOK-SIOS" },
	{ 0x4A2F , "HANGUL SYLLABLE SSANGTIKEUT YA NIEUN-CIEUC" },
	{ 0x4A30 , "HANGUL SYLLABLE SSANGTIKEUT YA NIEUN-HIEUH" },
	{ 0x4A31 , "HANGUL SYLLABLE SSANGTIKEUT YA TIKEUT" },
	{ 0x4A32 , "HANGUL SYLLABLE SSANGTIKEUT YA RIEUL" },
	{ 0x4A33 , "HANGUL SYLLABLE SSANGTIKEUT YA RIEUL-KIYEOK" },
	{ 0x4A34 , "HANGUL SYLLABLE SSANGTIKEUT YA RIEUL-MIEUM" },
	{ 0x4A35 , "HANGUL SYLLABLE SSANGTIKEUT YA RIEUL-PIEUP" },
	{ 0x4A36 , "HANGUL SYLLABLE SSANGTIKEUT YA RIEUL-SIOS" },
	{ 0x4A37 , "HANGUL SYLLABLE SSANGTIKEUT YA RIEUL-THIEUTH" },
	{ 0x4A38 , "HANGUL SYLLABLE SSANGTIKEUT YA RIEUL-PHIEUPH" },
	{ 0x4A39 , "HANGUL SYLLABLE SSANGTIKEUT YA RIEUL-HIEUH" },
	{ 0x4A3A , "HANGUL SYLLABLE SSANGTIKEUT YA MIEUM" },
	{ 0x4A3B , "HANGUL SYLLABLE SSANGTIKEUT YA PIEUP" },
	{ 0x4A3C , "HANGUL SYLLABLE SSANGTIKEUT YA PIEUP-SIOS" },
	{ 0x4A3D , "HANGUL SYLLABLE SSANGTIKEUT YA SIOS" },
	{ 0x4A3E , "HANGUL SYLLABLE SSANGTIKEUT YA SSANGSIOS" },
	{ 0x4A3F , "HANGUL SYLLABLE SSANGTIKEUT YA CIEUC" },
	{ 0x4A40 , "HANGUL SYLLABLE SSANGTIKEUT YA CHIEUCH" },
	{ 0x4A41 , "HANGUL SYLLABLE SSANGTIKEUT YA KHIEUKH" },
	{ 0x4A42 , "HANGUL SYLLABLE SSANGTIKEUT YA THIEUTH" },
	{ 0x4A43 , "HANGUL SYLLABLE SSANGTIKEUT YA PHIEUPH" },
	{ 0x4A44 , "HANGUL SYLLABLE SSANGTIKEUT YA HIEUH" },
	{ 0x4A45 , "HANGUL SYLLABLE SSANGTIKEUT YAE KIYEOK" },
	{ 0x4A46 , "HANGUL SYLLABLE SSANGTIKEUT YAE SSANGKIYEOK" },
	{ 0x4A47 , "HANGUL SYLLABLE SSANGTIKEUT YAE KIYEOK-SIOS" },
	{ 0x4A48 , "HANGUL SYLLABLE SSANGTIKEUT YAE NIEUN-CIEUC" },
	{ 0x4A49 , "HANGUL SYLLABLE SSANGTIKEUT YAE NIEUN-HIEUH" },
	{ 0x4A4A , "HANGUL SYLLABLE SSANGTIKEUT YAE TIKEUT" },
	{ 0x4A4B , "HANGUL SYLLABLE SSANGTIKEUT YAE RIEUL" },
	{ 0x4A4C , "HANGUL SYLLABLE SSANGTIKEUT YAE RIEUL-KIYEOK" },
	{ 0x4A4D , "HANGUL SYLLABLE SSANGTIKEUT YAE RIEUL-MIEUM" },
	{ 0x4A4E , "HANGUL SYLLABLE SSANGTIKEUT YAE RIEUL-PIEUP" },
	{ 0x4A4F , "HANGUL SYLLABLE SSANGTIKEUT YAE RIEUL-SIOS" },
	{ 0x4A50 , "HANGUL SYLLABLE SSANGTIKEUT YAE RIEUL-THIEUTH" },
	{ 0x4A51 , "HANGUL SYLLABLE SSANGTIKEUT YAE RIEUL-PHIEUPH" },
	{ 0x4A52 , "HANGUL SYLLABLE SSANGTIKEUT YAE RIEUL-HIEUH" },
	{ 0x4A53 , "HANGUL SYLLABLE SSANGTIKEUT YAE MIEUM" },
	{ 0x4A54 , "HANGUL SYLLABLE SSANGTIKEUT YAE PIEUP" },
	{ 0x4A55 , "HANGUL SYLLABLE SSANGTIKEUT YAE PIEUP-SIOS" },
	{ 0x4A56 , "HANGUL SYLLABLE SSANGTIKEUT YAE SIOS" },
	{ 0x4A57 , "HANGUL SYLLABLE SSANGTIKEUT YAE SSANGSIOS" },
	{ 0x4A58 , "HANGUL SYLLABLE SSANGTIKEUT YAE IEUNG" },
	{ 0x4A59 , "HANGUL SYLLABLE SSANGTIKEUT YAE CIEUC" },
	{ 0x4A5A , "HANGUL SYLLABLE SSANGTIKEUT YAE CHIEUCH" },
	{ 0x4A5B , "HANGUL SYLLABLE SSANGTIKEUT YAE KHIEUKH" },
	{ 0x4A5C , "HANGUL SYLLABLE SSANGTIKEUT YAE THIEUTH" },
	{ 0x4A5D , "HANGUL SYLLABLE SSANGTIKEUT YAE PHIEUPH" },
	{ 0x4A5E , "HANGUL SYLLABLE SSANGTIKEUT YAE HIEUH" },
	{ 0x4A5F , "HANGUL SYLLABLE SSANGTIKEUT EO SSANGKIYEOK" },
	{ 0x4A60 , "HANGUL SYLLABLE SSANGTIKEUT EO KIYEOK-SIOS" },
	{ 0x4A61 , "HANGUL SYLLABLE SSANGTIKEUT EO NIEUN-CIEUC" },
	{ 0x4A62 , "HANGUL SYLLABLE SSANGTIKEUT EO NIEUN-HIEUH" },
	{ 0x4A63 , "HANGUL SYLLABLE SSANGTIKEUT EO RIEUL-KIYEOK" },
	{ 0x4A64 , "HANGUL SYLLABLE SSANGTIKEUT EO RIEUL-SIOS" },
	{ 0x4A65 , "HANGUL SYLLABLE SSANGTIKEUT EO RIEUL-THIEUTH" },
	{ 0x4A66 , "HANGUL SYLLABLE SSANGTIKEUT EO RIEUL-PHIEUPH" },
	{ 0x4A67 , "HANGUL SYLLABLE SSANGTIKEUT EO PIEUP-SIOS" },
	{ 0x4A68 , "HANGUL SYLLABLE SSANGTIKEUT EO CIEUC" },
	{ 0x4A69 , "HANGUL SYLLABLE SSANGTIKEUT EO CHIEUCH" },
	{ 0x4A6A , "HANGUL SYLLABLE SSANGTIKEUT EO KHIEUKH" },
	{ 0x4A6B , "HANGUL SYLLABLE SSANGTIKEUT EO THIEUTH" },
	{ 0x4A6C , "HANGUL SYLLABLE SSANGTIKEUT EO PHIEUPH" },
	{ 0x4A6D , "HANGUL SYLLABLE SSANGTIKEUT E SSANGKIYEOK" },
	{ 0x4A6E , "HANGUL SYLLABLE SSANGTIKEUT E KIYEOK-SIOS" },
	{ 0x4A6F , "HANGUL SYLLABLE SSANGTIKEUT E NIEUN-CIEUC" },
	{ 0x4A70 , "HANGUL SYLLABLE SSANGTIKEUT E NIEUN-HIEUH" },
	{ 0x4A71 , "HANGUL SYLLABLE SSANGTIKEUT E RIEUL-KIYEOK" },
	{ 0x4A72 , "HANGUL SYLLABLE SSANGTIKEUT E RIEUL-MIEUM" },
	{ 0x4A73 , "HANGUL SYLLABLE SSANGTIKEUT E RIEUL-PIEUP" },
	{ 0x4A74 , "HANGUL SYLLABLE SSANGTIKEUT E RIEUL-SIOS" },
	{ 0x4A75 , "HANGUL SYLLABLE SSANGTIKEUT E RIEUL-THIEUTH" },
	{ 0x4A76 , "HANGUL SYLLABLE SSANGTIKEUT E RIEUL-PHIEUPH" },
	{ 0x4A77 , "HANGUL SYLLABLE SSANGTIKEUT E RIEUL-HIEUH" },
	{ 0x4A78 , "HANGUL SYLLABLE SSANGTIKEUT E PIEUP-SIOS" },
	{ 0x4A79 , "HANGUL SYLLABLE SSANGTIKEUT E CIEUC" },
	{ 0x4A7A , "HANGUL SYLLABLE SSANGTIKEUT E CHIEUCH" },
	{ 0x4A7B , "HANGUL SYLLABLE SSANGTIKEUT E KHIEUKH" },
	{ 0x4A7C , "HANGUL SYLLABLE SSANGTIKEUT E THIEUTH" },
	{ 0x4A7D , "HANGUL SYLLABLE SSANGTIKEUT E PHIEUPH" },
	{ 0x4A7E , "HANGUL SYLLABLE SSANGTIKEUT E HIEUH" },
	{ 0x4A7F , "HANGUL SYLLABLE SSANGTIKEUT YEO SSANGKIYEOK" },
	{ 0x4A80 , "HANGUL SYLLABLE SSANGTIKEUT YEO KIYEOK-SIOS" },
	{ 0x4A81 , "HANGUL SYLLABLE SSANGTIKEUT YEO NIEUN-CIEUC" },
	{ 0x4A82 , "HANGUL SYLLABLE SSANGTIKEUT YEO NIEUN-HIEUH" },
	{ 0x4A83 , "HANGUL SYLLABLE SSANGTIKEUT YEO RIEUL" },
	{ 0x4A84 , "HANGUL SYLLABLE SSANGTIKEUT YEO RIEUL-KIYEOK" },
	{ 0x4A85 , "HANGUL SYLLABLE SSANGTIKEUT YEO RIEUL-MIEUM" },
	{ 0x4A86 , "HANGUL SYLLABLE SSANGTIKEUT YEO RIEUL-PIEUP" },
	{ 0x4A87 , "HANGUL SYLLABLE SSANGTIKEUT YEO RIEUL-SIOS" },
	{ 0x4A88 , "HANGUL SYLLABLE SSANGTIKEUT YEO RIEUL-THIEUTH" },
	{ 0x4A89 , "HANGUL SYLLABLE SSANGTIKEUT YEO RIEUL-PHIEUPH" },
	{ 0x4A8A , "HANGUL SYLLABLE SSANGTIKEUT YEO RIEUL-HIEUH" },
	{ 0x4A8B , "HANGUL SYLLABLE SSANGTIKEUT YEO PIEUP-SIOS" },
	{ 0x4A8C , "HANGUL SYLLABLE SSANGTIKEUT YEO CIEUC" },
	{ 0x4A8D , "HANGUL SYLLABLE SSANGTIKEUT YEO CHIEUCH" },
	{ 0x4A8E , "HANGUL SYLLABLE SSANGTIKEUT YEO KHIEUKH" },
	{ 0x4A8F , "HANGUL SYLLABLE SSANGTIKEUT YEO THIEUTH" },
	{ 0x4A90 , "HANGUL SYLLABLE SSANGTIKEUT YEO PHIEUPH" },
	{ 0x4A91 , "HANGUL SYLLABLE SSANGTIKEUT YEO HIEUH" },
	{ 0x4A92 , "HANGUL SYLLABLE SSANGTIKEUT YE KIYEOK" },
	{ 0x4A93 , "HANGUL SYLLABLE SSANGTIKEUT YE SSANGKIYEOK" },
	{ 0x4A94 , "HANGUL SYLLABLE SSANGTIKEUT YE KIYEOK-SIOS" },
	{ 0x4A95 , "HANGUL SYLLABLE SSANGTIKEUT YE NIEUN-CIEUC" },
	{ 0x4A96 , "HANGUL SYLLABLE SSANGTIKEUT YE NIEUN-HIEUH" },
	{ 0x4A97 , "HANGUL SYLLABLE SSANGTIKEUT YE TIKEUT" },
	{ 0x4A98 , "HANGUL SYLLABLE SSANGTIKEUT YE RIEUL" },
	{ 0x4A99 , "HANGUL SYLLABLE SSANGTIKEUT YE RIEUL-KIYEOK" },
	{ 0x4A9A , "HANGUL SYLLABLE SSANGTIKEUT YE RIEUL-MIEUM" },
	{ 0x4A9B , "HANGUL SYLLABLE SSANGTIKEUT YE RIEUL-PIEUP" },
	{ 0x4A9C , "HANGUL SYLLABLE SSANGTIKEUT YE RIEUL-SIOS" },
	{ 0x4A9D , "HANGUL SYLLABLE SSANGTIKEUT YE RIEUL-THIEUTH" },
	{ 0x4A9E , "HANGUL SYLLABLE SSANGTIKEUT YE RIEUL-PHIEUPH" },
	{ 0x4A9F , "HANGUL SYLLABLE SSANGTIKEUT YE RIEUL-HIEUH" },
	{ 0x4AA0 , "HANGUL SYLLABLE SSANGTIKEUT YE MIEUM" },
	{ 0x4AA1 , "HANGUL SYLLABLE SSANGTIKEUT YE PIEUP-SIOS" },
	{ 0x4AA2 , "HANGUL SYLLABLE SSANGTIKEUT YE SIOS" },
	{ 0x4AA3 , "HANGUL SYLLABLE SSANGTIKEUT YE SSANGSIOS" },
	{ 0x4AA4 , "HANGUL SYLLABLE SSANGTIKEUT YE IEUNG" },
	{ 0x4AA5 , "HANGUL SYLLABLE SSANGTIKEUT YE CIEUC" },
	{ 0x4AA6 , "HANGUL SYLLABLE SSANGTIKEUT YE CHIEUCH" },
	{ 0x4AA7 , "HANGUL SYLLABLE SSANGTIKEUT YE KHIEUKH" },
	{ 0x4AA8 , "HANGUL SYLLABLE SSANGTIKEUT YE THIEUTH" },
	{ 0x4AA9 , "HANGUL SYLLABLE SSANGTIKEUT YE PHIEUPH" },
	{ 0x4AAA , "HANGUL SYLLABLE SSANGTIKEUT YE HIEUH" },
	{ 0x4AAB , "HANGUL SYLLABLE SSANGTIKEUT O KIYEOK-SIOS" },
	{ 0x4AAC , "HANGUL SYLLABLE SSANGTIKEUT O NIEUN-CIEUC" },
	{ 0x4AAD , "HANGUL SYLLABLE SSANGTIKEUT O NIEUN-HIEUH" },
	{ 0x4AAE , "HANGUL SYLLABLE SSANGTIKEUT O RIEUL-PIEUP" },
	{ 0x4AAF , "HANGUL SYLLABLE SSANGTIKEUT O RIEUL-SIOS" },
	{ 0x4AB0 , "HANGUL SYLLABLE SSANGTIKEUT O RIEUL-THIEUTH" },
	{ 0x4AB1 , "HANGUL SYLLABLE SSANGTIKEUT O RIEUL-PHIEUPH" },
	{ 0x4AB2 , "HANGUL SYLLABLE SSANGTIKEUT O PIEUP-SIOS" },
	{ 0x4AB3 , "HANGUL SYLLABLE SSANGTIKEUT O SSANGSIOS" },
	{ 0x4AB4 , "HANGUL SYLLABLE SSANGTIKEUT O CIEUC" },
	{ 0x4AB5 , "HANGUL SYLLABLE SSANGTIKEUT O CHIEUCH" },
	{ 0x4AB6 , "HANGUL SYLLABLE SSANGTIKEUT O KHIEUKH" },
	{ 0x4AB7 , "HANGUL SYLLABLE SSANGTIKEUT O THIEUTH" },
	{ 0x4AB8 , "HANGUL SYLLABLE SSANGTIKEUT O PHIEUPH" },
	{ 0x4AB9 , "HANGUL SYLLABLE SSANGTIKEUT O HIEUH" },
	{ 0x4ABA , "HANGUL SYLLABLE SSANGTIKEUT WA SSANGKIYEOK" },
	{ 0x4ABB , "HANGUL SYLLABLE SSANGTIKEUT WA KIYEOK-SIOS" },
	{ 0x4ABC , "HANGUL SYLLABLE SSANGTIKEUT WA NIEUN-CIEUC" },
	{ 0x4ABD , "HANGUL SYLLABLE SSANGTIKEUT WA NIEUN-HIEUH" },
	{ 0x4ABE , "HANGUL SYLLABLE SSANGTIKEUT WA TIKEUT" },
	{ 0x4ABF , "HANGUL SYLLABLE SSANGTIKEUT WA RIEUL-KIYEOK" },
	{ 0x4AC0 , "HANGUL SYLLABLE SSANGTIKEUT WA RIEUL-MIEUM" },
	{ 0x4AC1 , "HANGUL SYLLABLE SSANGTIKEUT WA RIEUL-PIEUP" },
	{ 0x4AC2 , "HANGUL SYLLABLE SSANGTIKEUT WA RIEUL-SIOS" },
	{ 0x4AC3 , "HANGUL SYLLABLE SSANGTIKEUT WA RIEUL-THIEUTH" },
	{ 0x4AC4 , "HANGUL SYLLABLE SSANGTIKEUT WA RIEUL-PHIEUPH" },
	{ 0x4AC5 , "HANGUL SYLLABLE SSANGTIKEUT WA RIEUL-HIEUH" },
	{ 0x4AC6 , "HANGUL SYLLABLE SSANGTIKEUT WA MIEUM" },
	{ 0x4AC7 , "HANGUL SYLLABLE SSANGTIKEUT WA PIEUP" },
	{ 0x4AC8 , "HANGUL SYLLABLE SSANGTIKEUT WA PIEUP-SIOS" },
	{ 0x4AC9 , "HANGUL SYLLABLE SSANGTIKEUT WA SIOS" },
	{ 0x4ACA , "HANGUL SYLLABLE SSANGTIKEUT WA SSANGSIOS" },
	{ 0x4ACB , "HANGUL SYLLABLE SSANGTIKEUT WA CIEUC" },
	{ 0x4ACC , "HANGUL SYLLABLE SSANGTIKEUT WA CHIEUCH" },
	{ 0x4ACD , "HANGUL SYLLABLE SSANGTIKEUT WA KHIEUKH" },
	{ 0x4ACE , "HANGUL SYLLABLE SSANGTIKEUT WA THIEUTH" },
	{ 0x4ACF , "HANGUL SYLLABLE SSANGTIKEUT WA PHIEUPH" },
	{ 0x4AD0 , "HANGUL SYLLABLE SSANGTIKEUT WA HIEUH" },
	{ 0x4AD1 , "HANGUL SYLLABLE SSANGTIKEUT WAE SSANGKIYEOK" },
	{ 0x4AD2 , "HANGUL SYLLABLE SSANGTIKEUT WAE KIYEOK-SIOS" },
	{ 0x4AD3 , "HANGUL SYLLABLE SSANGTIKEUT WAE NIEUN-CIEUC" },
	{ 0x4AD4 , "HANGUL SYLLABLE SSANGTIKEUT WAE NIEUN-HIEUH" },
	{ 0x4AD5 , "HANGUL SYLLABLE SSANGTIKEUT WAE RIEUL-KIYEOK" },
	{ 0x4AD6 , "HANGUL SYLLABLE SSANGTIKEUT WAE RIEUL-MIEUM" },
	{ 0x4AD7 , "HANGUL SYLLABLE SSANGTIKEUT WAE RIEUL-PIEUP" },
	{ 0x4AD8 , "HANGUL SYLLABLE SSANGTIKEUT WAE RIEUL-SIOS" },
	{ 0x4AD9 , "HANGUL SYLLABLE SSANGTIKEUT WAE RIEUL-THIEUTH" },
	{ 0x4ADA , "HANGUL SYLLABLE SSANGTIKEUT WAE RIEUL-PHIEUPH" },
	{ 0x4ADB , "HANGUL SYLLABLE SSANGTIKEUT WAE RIEUL-HIEUH" },
	{ 0x4ADC , "HANGUL SYLLABLE SSANGTIKEUT WAE PIEUP-SIOS" },
	{ 0x4ADD , "HANGUL SYLLABLE SSANGTIKEUT WAE IEUNG" },
	{ 0x4ADE , "HANGUL SYLLABLE SSANGTIKEUT WAE CIEUC" },
	{ 0x4ADF , "HANGUL SYLLABLE SSANGTIKEUT WAE CHIEUCH" },
	{ 0x4AE0 , "HANGUL SYLLABLE SSANGTIKEUT WAE KHIEUKH" },
	{ 0x4AE1 , "HANGUL SYLLABLE SSANGTIKEUT WAE THIEUTH" },
	{ 0x4AE2 , "HANGUL SYLLABLE SSANGTIKEUT WAE PHIEUPH" },
	{ 0x4AE3 , "HANGUL SYLLABLE SSANGTIKEUT WAE HIEUH" },
	{ 0x4AE4 , "HANGUL SYLLABLE SSANGTIKEUT OE SSANGKIYEOK" },
	{ 0x4AE5 , "HANGUL SYLLABLE SSANGTIKEUT OE KIYEOK-SIOS" },
	{ 0x4AE6 , "HANGUL SYLLABLE SSANGTIKEUT OE NIEUN-CIEUC" },
	{ 0x4AE7 , "HANGUL SYLLABLE SSANGTIKEUT OE NIEUN-HIEUH" },
	{ 0x4AE8 , "HANGUL SYLLABLE SSANGTIKEUT OE RIEUL-KIYEOK" },
	{ 0x4AE9 , "HANGUL SYLLABLE SSANGTIKEUT OE RIEUL-MIEUM" },
	{ 0x4AEA , "HANGUL SYLLABLE SSANGTIKEUT OE RIEUL-PIEUP" },
	{ 0x4AEB , "HANGUL SYLLABLE SSANGTIKEUT OE RIEUL-SIOS" },
	{ 0x4AEC , "HANGUL SYLLABLE SSANGTIKEUT OE RIEUL-THIEUTH" },
	{ 0x4AED , "HANGUL SYLLABLE SSANGTIKEUT OE RIEUL-PHIEUPH" },
	{ 0x4AEE , "HANGUL SYLLABLE SSANGTIKEUT OE RIEUL-HIEUH" },
	{ 0x4AEF , "HANGUL SYLLABLE SSANGTIKEUT OE PIEUP-SIOS" },
	{ 0x4AF0 , "HANGUL SYLLABLE SSANGTIKEUT OE CIEUC" },
	{ 0x4AF1 , "HANGUL SYLLABLE SSANGTIKEUT OE CHIEUCH" },
	{ 0x4AF2 , "HANGUL SYLLABLE SSANGTIKEUT OE KHIEUKH" },
	{ 0x4AF3 , "HANGUL SYLLABLE SSANGTIKEUT OE THIEUTH" },
	{ 0x4AF4 , "HANGUL SYLLABLE SSANGTIKEUT OE PHIEUPH" },
	{ 0x4AF5 , "HANGUL SYLLABLE SSANGTIKEUT OE HIEUH" },
	{ 0x4AF6 , "HANGUL SYLLABLE SSANGTIKEUT YO SSANGKIYEOK" },
	{ 0x4AF7 , "HANGUL SYLLABLE SSANGTIKEUT YO KIYEOK-SIOS" },
	{ 0x4AF8 , "HANGUL SYLLABLE SSANGTIKEUT YO NIEUN" },
	{ 0x4AF9 , "HANGUL SYLLABLE SSANGTIKEUT YO NIEUN-CIEUC" },
	{ 0x4AFA , "HANGUL SYLLABLE SSANGTIKEUT YO NIEUN-HIEUH" },
	{ 0x4AFB , "HANGUL SYLLABLE SSANGTIKEUT YO TIKEUT" },
	{ 0x4AFC , "HANGUL SYLLABLE SSANGTIKEUT YO RIEUL" },
	{ 0x4AFD , "HANGUL SYLLABLE SSANGTIKEUT YO RIEUL-KIYEOK" },
	{ 0x4AFE , "HANGUL SYLLABLE SSANGTIKEUT YO RIEUL-MIEUM" },
	{ 0x4AFF , "HANGUL SYLLABLE SSANGTIKEUT YO RIEUL-PIEUP" },
	{ 0x4B00 , "HANGUL SYLLABLE SSANGTIKEUT YO RIEUL-SIOS" },
	{ 0x4B01 , "HANGUL SYLLABLE SSANGTIKEUT YO RIEUL-THIEUTH" },
	{ 0x4B02 , "HANGUL SYLLABLE SSANGTIKEUT YO RIEUL-PHIEUPH" },
	{ 0x4B03 , "HANGUL SYLLABLE SSANGTIKEUT YO RIEUL-HIEUH" },
	{ 0x4B04 , "HANGUL SYLLABLE SSANGTIKEUT YO MIEUM" },
	{ 0x4B05 , "HANGUL SYLLABLE SSANGTIKEUT YO PIEUP" },
	{ 0x4B06 , "HANGUL SYLLABLE SSANGTIKEUT YO PIEUP-SIOS" },
	{ 0x4B07 , "HANGUL SYLLABLE SSANGTIKEUT YO SIOS" },
	{ 0x4B08 , "HANGUL SYLLABLE SSANGTIKEUT YO SSANGSIOS" },
	{ 0x4B09 , "HANGUL SYLLABLE SSANGTIKEUT YO IEUNG" },
	{ 0x4B0A , "HANGUL SYLLABLE SSANGTIKEUT YO CIEUC" },
	{ 0x4B0B , "HANGUL SYLLABLE SSANGTIKEUT YO CHIEUCH" },
	{ 0x4B0C , "HANGUL SYLLABLE SSANGTIKEUT YO KHIEUKH" },
	{ 0x4B0D , "HANGUL SYLLABLE SSANGTIKEUT YO THIEUTH" },
	{ 0x4B0E , "HANGUL SYLLABLE SSANGTIKEUT YO PHIEUPH" },
	{ 0x4B0F , "HANGUL SYLLABLE SSANGTIKEUT YO HIEUH" },
	{ 0x4B10 , "HANGUL SYLLABLE SSANGTIKEUT U SSANGKIYEOK" },
	{ 0x4B11 , "HANGUL SYLLABLE SSANGTIKEUT U KIYEOK-SIOS" },
	{ 0x4B12 , "HANGUL SYLLABLE SSANGTIKEUT U NIEUN-CIEUC" },
	{ 0x4B13 , "HANGUL SYLLABLE SSANGTIKEUT U NIEUN-HIEUH" },
	{ 0x4B14 , "HANGUL SYLLABLE SSANGTIKEUT U RIEUL-KIYEOK" },
	{ 0x4B15 , "HANGUL SYLLABLE SSANGTIKEUT U RIEUL-SIOS" },
	{ 0x4B16 , "HANGUL SYLLABLE SSANGTIKEUT U RIEUL-THIEUTH" },
	{ 0x4B17 , "HANGUL SYLLABLE SSANGTIKEUT U RIEUL-PHIEUPH" },
	{ 0x4B18 , "HANGUL SYLLABLE SSANGTIKEUT U PIEUP-SIOS" },
	{ 0x4B19 , "HANGUL SYLLABLE SSANGTIKEUT U SSANGSIOS" },
	{ 0x4B1A , "HANGUL SYLLABLE SSANGTIKEUT U CIEUC" },
	{ 0x4B1B , "HANGUL SYLLABLE SSANGTIKEUT U CHIEUCH" },
	{ 0x4B1C , "HANGUL SYLLABLE SSANGTIKEUT U KHIEUKH" },
	{ 0x4B1D , "HANGUL SYLLABLE SSANGTIKEUT U PHIEUPH" },
	{ 0x4B1E , "HANGUL SYLLABLE SSANGTIKEUT U HIEUH" },
	{ 0x4B1F , "HANGUL SYLLABLE SSANGTIKEUT WEO KIYEOK" },
	{ 0x4B20 , "HANGUL SYLLABLE SSANGTIKEUT WEO SSANGKIYEOK" },
	{ 0x4B21 , "HANGUL SYLLABLE SSANGTIKEUT WEO KIYEOK-SIOS" },
	{ 0x4B22 , "HANGUL SYLLABLE SSANGTIKEUT WEO NIEUN-CIEUC" },
	{ 0x4B23 , "HANGUL SYLLABLE SSANGTIKEUT WEO NIEUN-HIEUH" },
	{ 0x4B24 , "HANGUL SYLLABLE SSANGTIKEUT WEO RIEUL" },
	{ 0x4B25 , "HANGUL SYLLABLE SSANGTIKEUT WEO RIEUL-KIYEOK" },
	{ 0x4B26 , "HANGUL SYLLABLE SSANGTIKEUT WEO RIEUL-MIEUM" },
	{ 0x4B27 , "HANGUL SYLLABLE SSANGTIKEUT WEO RIEUL-PIEUP" },
	{ 0x4B28 , "HANGUL SYLLABLE SSANGTIKEUT WEO RIEUL-SIOS" },
	{ 0x4B29 , "HANGUL SYLLABLE SSANGTIKEUT WEO RIEUL-THIEUTH" },
	{ 0x4B2A , "HANGUL SYLLABLE SSANGTIKEUT WEO RIEUL-PHIEUPH" },
	{ 0x4B2B , "HANGUL SYLLABLE SSANGTIKEUT WEO RIEUL-HIEUH" },
	{ 0x4B2C , "HANGUL SYLLABLE SSANGTIKEUT WEO MIEUM" },
	{ 0x4B2D , "HANGUL SYLLABLE SSANGTIKEUT WEO PIEUP" },
	{ 0x4B2E , "HANGUL SYLLABLE SSANGTIKEUT WEO PIEUP-SIOS" },
	{ 0x4B2F , "HANGUL SYLLABLE SSANGTIKEUT WEO SSANGSIOS" },
	{ 0x4B30 , "HANGUL SYLLABLE SSANGTIKEUT WEO IEUNG" },
	{ 0x4B31 , "HANGUL SYLLABLE SSANGTIKEUT WEO CIEUC" },
	{ 0x4B32 , "HANGUL SYLLABLE SSANGTIKEUT WEO CHIEUCH" },
	{ 0x4B33 , "HANGUL SYLLABLE SSANGTIKEUT WEO KHIEUKH" },
	{ 0x4B34 , "HANGUL SYLLABLE SSANGTIKEUT WEO THIEUTH" },
	{ 0x4B35 , "HANGUL SYLLABLE SSANGTIKEUT WEO PHIEUPH" },
	{ 0x4B36 , "HANGUL SYLLABLE SSANGTIKEUT WEO HIEUH" },
	{ 0x4B37 , "HANGUL SYLLABLE SSANGTIKEUT WE KIYEOK" },
	{ 0x4B38 , "HANGUL SYLLABLE SSANGTIKEUT WE SSANGKIYEOK" },
	{ 0x4B39 , "HANGUL SYLLABLE SSANGTIKEUT WE KIYEOK-SIOS" },
	{ 0x4B3A , "HANGUL SYLLABLE SSANGTIKEUT WE NIEUN-CIEUC" },
	{ 0x4B3B , "HANGUL SYLLABLE SSANGTIKEUT WE NIEUN-HIEUH" },
	{ 0x4B3C , "HANGUL SYLLABLE SSANGTIKEUT WE RIEUL-KIYEOK" },
	{ 0x4B3D , "HANGUL SYLLABLE SSANGTIKEUT WE RIEUL-MIEUM" },
	{ 0x4B3E , "HANGUL SYLLABLE SSANGTIKEUT WE RIEUL-PIEUP" },
	{ 0x4B3F , "HANGUL SYLLABLE SSANGTIKEUT WE RIEUL-SIOS" },
	{ 0x4B40 , "HANGUL SYLLABLE SSANGTIKEUT WE RIEUL-THIEUTH" },
	{ 0x4B41 , "HANGUL SYLLABLE SSANGTIKEUT WE RIEUL-PHIEUPH" },
	{ 0x4B42 , "HANGUL SYLLABLE SSANGTIKEUT WE RIEUL-HIEUH" },
	{ 0x4B43 , "HANGUL SYLLABLE SSANGTIKEUT WE PIEUP-SIOS" },
	{ 0x4B44 , "HANGUL SYLLABLE SSANGTIKEUT WE SIOS" },
	{ 0x4B45 , "HANGUL SYLLABLE SSANGTIKEUT WE CIEUC" },
	{ 0x4B46 , "HANGUL SYLLABLE SSANGTIKEUT WE CHIEUCH" },
	{ 0x4B47 , "HANGUL SYLLABLE SSANGTIKEUT WE KHIEUKH" },
	{ 0x4B48 , "HANGUL SYLLABLE SSANGTIKEUT WE THIEUTH" },
	{ 0x4B49 , "HANGUL SYLLABLE SSANGTIKEUT WE PHIEUPH" },
	{ 0x4B4A , "HANGUL SYLLABLE SSANGTIKEUT WE HIEUH" },
	{ 0x4B4B , "HANGUL SYLLABLE SSANGTIKEUT WI KIYEOK" },
	{ 0x4B4C , "HANGUL SYLLABLE SSANGTIKEUT WI SSANGKIYEOK" },
	{ 0x4B4D , "HANGUL SYLLABLE SSANGTIKEUT WI KIYEOK-SIOS" },
	{ 0x4B4E , "HANGUL SYLLABLE SSANGTIKEUT WI NIEUN-CIEUC" },
	{ 0x4B4F , "HANGUL SYLLABLE SSANGTIKEUT WI NIEUN-HIEUH" },
	{ 0x4B50 , "HANGUL SYLLABLE SSANGTIKEUT WI RIEUL-KIYEOK" },
	{ 0x4B51 , "HANGUL SYLLABLE SSANGTIKEUT WI RIEUL-MIEUM" },
	{ 0x4B52 , "HANGUL SYLLABLE SSANGTIKEUT WI RIEUL-PIEUP" },
	{ 0x4B53 , "HANGUL SYLLABLE SSANGTIKEUT WI RIEUL-SIOS" },
	{ 0x4B54 , "HANGUL SYLLABLE SSANGTIKEUT WI RIEUL-THIEUTH" },
	{ 0x4B55 , "HANGUL SYLLABLE SSANGTIKEUT WI RIEUL-PHIEUPH" },
	{ 0x4B56 , "HANGUL SYLLABLE SSANGTIKEUT WI RIEUL-HIEUH" },
	{ 0x4B57 , "HANGUL SYLLABLE SSANGTIKEUT WI PIEUP-SIOS" },
	{ 0x4B58 , "HANGUL SYLLABLE SSANGTIKEUT WI SSANGSIOS" },
	{ 0x4B59 , "HANGUL SYLLABLE SSANGTIKEUT WI CIEUC" },
	{ 0x4B5A , "HANGUL SYLLABLE SSANGTIKEUT WI CHIEUCH" },
	{ 0x4B5B , "HANGUL SYLLABLE SSANGTIKEUT WI KHIEUKH" },
	{ 0x4B5C , "HANGUL SYLLABLE SSANGTIKEUT WI THIEUTH" },
	{ 0x4B5D , "HANGUL SYLLABLE SSANGTIKEUT WI PHIEUPH" },
	{ 0x4B5E , "HANGUL SYLLABLE SSANGTIKEUT WI HIEUH" },
	{ 0x4B5F , "HANGUL SYLLABLE SSANGTIKEUT YU SSANGKIYEOK" },
	{ 0x4B60 , "HANGUL SYLLABLE SSANGTIKEUT YU KIYEOK-SIOS" },
	{ 0x4B61 , "HANGUL SYLLABLE SSANGTIKEUT YU NIEUN" },
	{ 0x4B62 , "HANGUL SYLLABLE SSANGTIKEUT YU NIEUN-CIEUC" },
	{ 0x4B63 , "HANGUL SYLLABLE SSANGTIKEUT YU NIEUN-HIEUH" },
	{ 0x4B64 , "HANGUL SYLLABLE SSANGTIKEUT YU TIKEUT" },
	{ 0x4B65 , "HANGUL SYLLABLE SSANGTIKEUT YU RIEUL-KIYEOK" },
	{ 0x4B66 , "HANGUL SYLLABLE SSANGTIKEUT YU RIEUL-MIEUM" },
	{ 0x4B67 , "HANGUL SYLLABLE SSANGTIKEUT YU RIEUL-PIEUP" },
	{ 0x4B68 , "HANGUL SYLLABLE SSANGTIKEUT YU RIEUL-SIOS" },
	{ 0x4B69 , "HANGUL SYLLABLE SSANGTIKEUT YU RIEUL-THIEUTH" },
	{ 0x4B6A , "HANGUL SYLLABLE SSANGTIKEUT YU RIEUL-PHIEUPH" },
	{ 0x4B6B , "HANGUL SYLLABLE SSANGTIKEUT YU RIEUL-HIEUH" },
	{ 0x4B6C , "HANGUL SYLLABLE SSANGTIKEUT YU PIEUP" },
	{ 0x4B6D , "HANGUL SYLLABLE SSANGTIKEUT YU PIEUP-SIOS" },
	{ 0x4B6E , "HANGUL SYLLABLE SSANGTIKEUT YU SIOS" },
	{ 0x4B6F , "HANGUL SYLLABLE SSANGTIKEUT YU SSANGSIOS" },
	{ 0x4B70 , "HANGUL SYLLABLE SSANGTIKEUT YU CIEUC" },
	{ 0x4B71 , "HANGUL SYLLABLE SSANGTIKEUT YU CHIEUCH" },
	{ 0x4B72 , "HANGUL SYLLABLE SSANGTIKEUT YU KHIEUKH" },
	{ 0x4B73 , "HANGUL SYLLABLE SSANGTIKEUT YU THIEUTH" },
	{ 0x4B74 , "HANGUL SYLLABLE SSANGTIKEUT YU PHIEUPH" },
	{ 0x4B75 , "HANGUL SYLLABLE SSANGTIKEUT YU HIEUH" },
	{ 0x4B76 , "HANGUL SYLLABLE SSANGTIKEUT EU SSANGKIYEOK" },
	{ 0x4B77 , "HANGUL SYLLABLE SSANGTIKEUT EU KIYEOK-SIOS" },
	{ 0x4B78 , "HANGUL SYLLABLE SSANGTIKEUT EU NIEUN-CIEUC" },
	{ 0x4B79 , "HANGUL SYLLABLE SSANGTIKEUT EU NIEUN-HIEUH" },
	{ 0x4B7A , "HANGUL SYLLABLE SSANGTIKEUT EU RIEUL-KIYEOK" },
	{ 0x4B7B , "HANGUL SYLLABLE SSANGTIKEUT EU RIEUL-SIOS" },
	{ 0x4B7C , "HANGUL SYLLABLE SSANGTIKEUT EU RIEUL-THIEUTH" },
	{ 0x4B7D , "HANGUL SYLLABLE SSANGTIKEUT EU RIEUL-PHIEUPH" },
	{ 0x4B7E , "HANGUL SYLLABLE SSANGTIKEUT EU PIEUP-SIOS" },
	{ 0x4B7F , "HANGUL SYLLABLE SSANGTIKEUT EU SSANGSIOS" },
	{ 0x4B80 , "HANGUL SYLLABLE SSANGTIKEUT EU CIEUC" },
	{ 0x4B81 , "HANGUL SYLLABLE SSANGTIKEUT EU CHIEUCH" },
	{ 0x4B82 , "HANGUL SYLLABLE SSANGTIKEUT EU KHIEUKH" },
	{ 0x4B83 , "HANGUL SYLLABLE SSANGTIKEUT EU THIEUTH" },
	{ 0x4B84 , "HANGUL SYLLABLE SSANGTIKEUT EU PHIEUPH" },
	{ 0x4B85 , "HANGUL SYLLABLE SSANGTIKEUT EU HIEUH" },
	{ 0x4B86 , "HANGUL SYLLABLE SSANGTIKEUT YI KIYEOK" },
	{ 0x4B87 , "HANGUL SYLLABLE SSANGTIKEUT YI SSANGKIYEOK" },
	{ 0x4B88 , "HANGUL SYLLABLE SSANGTIKEUT YI KIYEOK-SIOS" },
	{ 0x4B89 , "HANGUL SYLLABLE SSANGTIKEUT YI NIEUN-CIEUC" },
	{ 0x4B8A , "HANGUL SYLLABLE SSANGTIKEUT YI NIEUN-HIEUH" },
	{ 0x4B8B , "HANGUL SYLLABLE SSANGTIKEUT YI TIKEUT" },
	{ 0x4B8C , "HANGUL SYLLABLE SSANGTIKEUT YI RIEUL-KIYEOK" },
	{ 0x4B8D , "HANGUL SYLLABLE SSANGTIKEUT YI RIEUL-MIEUM" },
	{ 0x4B8E , "HANGUL SYLLABLE SSANGTIKEUT YI RIEUL-PIEUP" },
	{ 0x4B8F , "HANGUL SYLLABLE SSANGTIKEUT YI RIEUL-SIOS" },
	{ 0x4B90 , "HANGUL SYLLABLE SSANGTIKEUT YI RIEUL-THIEUTH" },
	{ 0x4B91 , "HANGUL SYLLABLE SSANGTIKEUT YI RIEUL-PHIEUPH" },
	{ 0x4B92 , "HANGUL SYLLABLE SSANGTIKEUT YI RIEUL-HIEUH" },
	{ 0x4B93 , "HANGUL SYLLABLE SSANGTIKEUT YI PIEUP-SIOS" },
	{ 0x4B94 , "HANGUL SYLLABLE SSANGTIKEUT YI SIOS" },
	{ 0x4B95 , "HANGUL SYLLABLE SSANGTIKEUT YI SSANGSIOS" },
	{ 0x4B96 , "HANGUL SYLLABLE SSANGTIKEUT YI CIEUC" },
	{ 0x4B97 , "HANGUL SYLLABLE SSANGTIKEUT YI CHIEUCH" },
	{ 0x4B98 , "HANGUL SYLLABLE SSANGTIKEUT YI KHIEUKH" },
	{ 0x4B99 , "HANGUL SYLLABLE SSANGTIKEUT YI THIEUTH" },
	{ 0x4B9A , "HANGUL SYLLABLE SSANGTIKEUT YI PHIEUPH" },
	{ 0x4B9B , "HANGUL SYLLABLE SSANGTIKEUT YI HIEUH" },
	{ 0x4B9C , "HANGUL SYLLABLE SSANGTIKEUT I SSANGKIYEOK" },
	{ 0x4B9D , "HANGUL SYLLABLE SSANGTIKEUT I KIYEOK-SIOS" },
	{ 0x4B9E , "HANGUL SYLLABLE SSANGTIKEUT I NIEUN-CIEUC" },
	{ 0x4B9F , "HANGUL SYLLABLE SSANGTIKEUT I NIEUN-HIEUH" },
	{ 0x4BA0 , "HANGUL SYLLABLE SSANGTIKEUT I RIEUL-KIYEOK" },
	{ 0x4BA1 , "HANGUL SYLLABLE SSANGTIKEUT I RIEUL-MIEUM" },
	{ 0x4BA2 , "HANGUL SYLLABLE SSANGTIKEUT I RIEUL-PIEUP" },
	{ 0x4BA3 , "HANGUL SYLLABLE SSANGTIKEUT I RIEUL-SIOS" },
	{ 0x4BA4 , "HANGUL SYLLABLE SSANGTIKEUT I RIEUL-THIEUTH" },
	{ 0x4BA5 , "HANGUL SYLLABLE SSANGTIKEUT I RIEUL-PHIEUPH" },
	{ 0x4BA6 , "HANGUL SYLLABLE SSANGTIKEUT I RIEUL-HIEUH" },
	{ 0x4BA7 , "HANGUL SYLLABLE SSANGTIKEUT I PIEUP-SIOS" },
	{ 0x4BA8 , "HANGUL SYLLABLE SSANGTIKEUT I SSANGSIOS" },
	{ 0x4BA9 , "HANGUL SYLLABLE SSANGTIKEUT I CIEUC" },
	{ 0x4BAA , "HANGUL SYLLABLE SSANGTIKEUT I CHIEUCH" },
	{ 0x4BAB , "HANGUL SYLLABLE SSANGTIKEUT I KHIEUKH" },
	{ 0x4BAC , "HANGUL SYLLABLE SSANGTIKEUT I THIEUTH" },
	{ 0x4BAD , "HANGUL SYLLABLE SSANGTIKEUT I PHIEUPH" },
	{ 0x4BAE , "HANGUL SYLLABLE RIEUL A SSANGKIYEOK" },
	{ 0x4BAF , "HANGUL SYLLABLE RIEUL A KIYEOK-SIOS" },
	{ 0x4BB0 , "HANGUL SYLLABLE RIEUL A NIEUN-CIEUC" },
	{ 0x4BB1 , "HANGUL SYLLABLE RIEUL A NIEUN-HIEUH" },
	{ 0x4BB2 , "HANGUL SYLLABLE RIEUL A RIEUL-KIYEOK" },
	{ 0x4BB3 , "HANGUL SYLLABLE RIEUL A RIEUL-MIEUM" },
	{ 0x4BB4 , "HANGUL SYLLABLE RIEUL A RIEUL-PIEUP" },
	{ 0x4BB5 , "HANGUL SYLLABLE RIEUL A RIEUL-THIEUTH" },
	{ 0x4BB6 , "HANGUL SYLLABLE RIEUL A RIEUL-PHIEUPH" },
	{ 0x4BB7 , "HANGUL SYLLABLE RIEUL A RIEUL-HIEUH" },
	{ 0x4BB8 , "HANGUL SYLLABLE RIEUL A PIEUP-SIOS" },
	{ 0x4BB9 , "HANGUL SYLLABLE RIEUL AE SSANGKIYEOK" },
	{ 0x4BBA , "HANGUL SYLLABLE RIEUL AE KIYEOK-SIOS" },
	{ 0x4BBB , "HANGUL SYLLABLE RIEUL AE NIEUN-CIEUC" },
	{ 0x4BBC , "HANGUL SYLLABLE RIEUL AE NIEUN-HIEUH" },
	{ 0x4BBD , "HANGUL SYLLABLE RIEUL AE RIEUL-KIYEOK" },
	{ 0x4BBE , "HANGUL SYLLABLE RIEUL AE RIEUL-MIEUM" },
	{ 0x4BBF , "HANGUL SYLLABLE RIEUL AE RIEUL-PIEUP" },
	{ 0x4BC0 , "HANGUL SYLLABLE RIEUL AE RIEUL-SIOS" },
	{ 0x4BC1 , "HANGUL SYLLABLE RIEUL AE RIEUL-THIEUTH" },
	{ 0x4BC2 , "HANGUL SYLLABLE RIEUL AE RIEUL-PHIEUPH" },
	{ 0x4BC3 , "HANGUL SYLLABLE RIEUL AE RIEUL-HIEUH" },
	{ 0x4BC4 , "HANGUL SYLLABLE RIEUL AE PIEUP-SIOS" },
	{ 0x4BC5 , "HANGUL SYLLABLE RIEUL AE CIEUC" },
	{ 0x4BC6 , "HANGUL SYLLABLE RIEUL AE CHIEUCH" },
	{ 0x4BC7 , "HANGUL SYLLABLE RIEUL AE HIEUH" },
	{ 0x4BC8 , "HANGUL SYLLABLE RIEUL YA SSANGKIYEOK" },
	{ 0x4BC9 , "HANGUL SYLLABLE RIEUL YA KIYEOK-SIOS" },
	{ 0x4BCA , "HANGUL SYLLABLE RIEUL YA NIEUN-CIEUC" },
	{ 0x4BCB , "HANGUL SYLLABLE RIEUL YA NIEUN-HIEUH" },
	{ 0x4BCC , "HANGUL SYLLABLE RIEUL YA RIEUL-KIYEOK" },
	{ 0x4BCD , "HANGUL SYLLABLE RIEUL YA RIEUL-MIEUM" },
	{ 0x4BCE , "HANGUL SYLLABLE RIEUL YA RIEUL-PIEUP" },
	{ 0x4BCF , "HANGUL SYLLABLE RIEUL YA RIEUL-SIOS" },
	{ 0x4BD0 , "HANGUL SYLLABLE RIEUL YA RIEUL-THIEUTH" },
	{ 0x4BD1 , "HANGUL SYLLABLE RIEUL YA RIEUL-PHIEUPH" },
	{ 0x4BD2 , "HANGUL SYLLABLE RIEUL YA RIEUL-HIEUH" },
	{ 0x4BD3 , "HANGUL SYLLABLE RIEUL YA PIEUP-SIOS" },
	{ 0x4BD4 , "HANGUL SYLLABLE RIEUL YA SSANGSIOS" },
	{ 0x4BD5 , "HANGUL SYLLABLE RIEUL YA CIEUC" },
	{ 0x4BD6 , "HANGUL SYLLABLE RIEUL YA CHIEUCH" },
	{ 0x4BD7 , "HANGUL SYLLABLE RIEUL YA KHIEUKH" },
	{ 0x4BD8 , "HANGUL SYLLABLE RIEUL YA THIEUTH" },
	{ 0x4BD9 , "HANGUL SYLLABLE RIEUL YA PHIEUPH" },
	{ 0x4BDA , "HANGUL SYLLABLE RIEUL YA HIEUH" },
	{ 0x4BDB , "HANGUL SYLLABLE RIEUL YAE KIYEOK" },
	{ 0x4BDC , "HANGUL SYLLABLE RIEUL YAE SSANGKIYEOK" },
	{ 0x4BDD , "HANGUL SYLLABLE RIEUL YAE KIYEOK-SIOS" },
	{ 0x4BDE , "HANGUL SYLLABLE RIEUL YAE NIEUN" },
	{ 0x4BDF , "HANGUL SYLLABLE RIEUL YAE NIEUN-CIEUC" },
	{ 0x4BE0 , "HANGUL SYLLABLE RIEUL YAE NIEUN-HIEUH" },
	{ 0x4BE1 , "HANGUL SYLLABLE RIEUL YAE TIKEUT" },
	{ 0x4BE2 , "HANGUL SYLLABLE RIEUL YAE RIEUL" },
	{ 0x4BE3 , "HANGUL SYLLABLE RIEUL YAE RIEUL-KIYEOK" },
	{ 0x4BE4 , "HANGUL SYLLABLE RIEUL YAE RIEUL-MIEUM" },
	{ 0x4BE5 , "HANGUL SYLLABLE RIEUL YAE RIEUL-PIEUP" },
	{ 0x4BE6 , "HANGUL SYLLABLE RIEUL YAE RIEUL-SIOS" },
	{ 0x4BE7 , "HANGUL SYLLABLE RIEUL YAE RIEUL-THIEUTH" },
	{ 0x4BE8 , "HANGUL SYLLABLE RIEUL YAE RIEUL-PHIEUPH" },
	{ 0x4BE9 , "HANGUL SYLLABLE RIEUL YAE RIEUL-HIEUH" },
	{ 0x4BEA , "HANGUL SYLLABLE RIEUL YAE MIEUM" },
	{ 0x4BEB , "HANGUL SYLLABLE RIEUL YAE PIEUP" },
	{ 0x4BEC , "HANGUL SYLLABLE RIEUL YAE PIEUP-SIOS" },
	{ 0x4BED , "HANGUL SYLLABLE RIEUL YAE SSANGSIOS" },
	{ 0x4BEE , "HANGUL SYLLABLE RIEUL YAE IEUNG" },
	{ 0x4BEF , "HANGUL SYLLABLE RIEUL YAE CIEUC" },
	{ 0x4BF0 , "HANGUL SYLLABLE RIEUL YAE CHIEUCH" },
	{ 0x4BF1 , "HANGUL SYLLABLE RIEUL YAE KHIEUKH" },
	{ 0x4BF2 , "HANGUL SYLLABLE RIEUL YAE THIEUTH" },
	{ 0x4BF3 , "HANGUL SYLLABLE RIEUL YAE PHIEUPH" },
	{ 0x4BF4 , "HANGUL SYLLABLE RIEUL YAE HIEUH" },
	{ 0x4BF5 , "HANGUL SYLLABLE RIEUL EO SSANGKIYEOK" },
	{ 0x4BF6 , "HANGUL SYLLABLE RIEUL EO KIYEOK-SIOS" },
	{ 0x4BF7 , "HANGUL SYLLABLE RIEUL EO NIEUN-CIEUC" },
	{ 0x4BF8 , "HANGUL SYLLABLE RIEUL EO RIEUL-KIYEOK" },
	{ 0x4BF9 , "HANGUL SYLLABLE RIEUL EO RIEUL-MIEUM" },
	{ 0x4BFA , "HANGUL SYLLABLE RIEUL EO RIEUL-PIEUP" },
	{ 0x4BFB , "HANGUL SYLLABLE RIEUL EO RIEUL-SIOS" },
	{ 0x4BFC , "HANGUL SYLLABLE RIEUL EO RIEUL-THIEUTH" },
	{ 0x4BFD , "HANGUL SYLLABLE RIEUL EO RIEUL-PHIEUPH" },
	{ 0x4BFE , "HANGUL SYLLABLE RIEUL EO RIEUL-HIEUH" },
	{ 0x4BFF , "HANGUL SYLLABLE RIEUL EO CHIEUCH" },
	{ 0x4C00 , "HANGUL SYLLABLE RIEUL EO THIEUTH" },
	{ 0x4C01 , "HANGUL SYLLABLE RIEUL E SSANGKIYEOK" },
	{ 0x4C02 , "HANGUL SYLLABLE RIEUL E KIYEOK-SIOS" },
	{ 0x4C03 , "HANGUL SYLLABLE RIEUL E NIEUN-CIEUC" },
	{ 0x4C04 , "HANGUL SYLLABLE RIEUL E NIEUN-HIEUH" },
	{ 0x4C05 , "HANGUL SYLLABLE RIEUL E RIEUL-KIYEOK" },
	{ 0x4C06 , "HANGUL SYLLABLE RIEUL E RIEUL-MIEUM" },
	{ 0x4C07 , "HANGUL SYLLABLE RIEUL E RIEUL-PIEUP" },
	{ 0x4C08 , "HANGUL SYLLABLE RIEUL E RIEUL-SIOS" },
	{ 0x4C09 , "HANGUL SYLLABLE RIEUL E RIEUL-THIEUTH" },
	{ 0x4C0A , "HANGUL SYLLABLE RIEUL E RIEUL-PHIEUPH" },
	{ 0x4C0B , "HANGUL SYLLABLE RIEUL E RIEUL-HIEUH" },
	{ 0x4C0C , "HANGUL SYLLABLE RIEUL E PIEUP-SIOS" },
	{ 0x4C0D , "HANGUL SYLLABLE RIEUL E CIEUC" },
	{ 0x4C0E , "HANGUL SYLLABLE RIEUL E CHIEUCH" },
	{ 0x4C0F , "HANGUL SYLLABLE RIEUL E KHIEUKH" },
	{ 0x4C10 , "HANGUL SYLLABLE RIEUL YEO SSANGKIYEOK" },
	{ 0x4C11 , "HANGUL SYLLABLE RIEUL YEO KIYEOK-SIOS" },
	{ 0x4C12 , "HANGUL SYLLABLE RIEUL YEO NIEUN-CIEUC" },
	{ 0x4C13 , "HANGUL SYLLABLE RIEUL YEO NIEUN-HIEUH" },
	{ 0x4C14 , "HANGUL SYLLABLE RIEUL YEO RIEUL-KIYEOK" },
	{ 0x4C15 , "HANGUL SYLLABLE RIEUL YEO RIEUL-MIEUM" },
	{ 0x4C16 , "HANGUL SYLLABLE RIEUL YEO RIEUL-PIEUP" },
	{ 0x4C17 , "HANGUL SYLLABLE RIEUL YEO RIEUL-THIEUTH" },
	{ 0x4C18 , "HANGUL SYLLABLE RIEUL YEO RIEUL-PHIEUPH" },
	{ 0x4C19 , "HANGUL SYLLABLE RIEUL YEO RIEUL-HIEUH" },
	{ 0x4C1A , "HANGUL SYLLABLE RIEUL YEO PIEUP-SIOS" },
	{ 0x4C1B , "HANGUL SYLLABLE RIEUL YEO CIEUC" },
	{ 0x4C1C , "HANGUL SYLLABLE RIEUL YEO CHIEUCH" },
	{ 0x4C1D , "HANGUL SYLLABLE RIEUL YEO THIEUTH" },
	{ 0x4C1E , "HANGUL SYLLABLE RIEUL YEO HIEUH" },
	{ 0x4C1F , "HANGUL SYLLABLE RIEUL YE SSANGKIYEOK" },
	{ 0x4C20 , "HANGUL SYLLABLE RIEUL YE KIYEOK-SIOS" },
	{ 0x4C21 , "HANGUL SYLLABLE RIEUL YE NIEUN-CIEUC" },
	{ 0x4C22 , "HANGUL SYLLABLE RIEUL YE NIEUN-HIEUH" },
	{ 0x4C23 , "HANGUL SYLLABLE RIEUL YE RIEUL-KIYEOK" },
	{ 0x4C24 , "HANGUL SYLLABLE RIEUL YE RIEUL-MIEUM" },
	{ 0x4C25 , "HANGUL SYLLABLE RIEUL YE RIEUL-PIEUP" },
	{ 0x4C26 , "HANGUL SYLLABLE RIEUL YE RIEUL-SIOS" },
	{ 0x4C27 , "HANGUL SYLLABLE RIEUL YE RIEUL-THIEUTH" },
	{ 0x4C28 , "HANGUL SYLLABLE RIEUL YE RIEUL-PHIEUPH" },
	{ 0x4C29 , "HANGUL SYLLABLE RIEUL YE RIEUL-HIEUH" },
	{ 0x4C2A , "HANGUL SYLLABLE RIEUL YE PIEUP-SIOS" },
	{ 0x4C2B , "HANGUL SYLLABLE RIEUL YE SSANGSIOS" },
	{ 0x4C2C , "HANGUL SYLLABLE RIEUL YE CIEUC" },
	{ 0x4C2D , "HANGUL SYLLABLE RIEUL YE CHIEUCH" },
	{ 0x4C2E , "HANGUL SYLLABLE RIEUL YE KHIEUKH" },
	{ 0x4C2F , "HANGUL SYLLABLE RIEUL YE THIEUTH" },
	{ 0x4C30 , "HANGUL SYLLABLE RIEUL YE PHIEUPH" },
	{ 0x4C31 , "HANGUL SYLLABLE RIEUL YE HIEUH" },
	{ 0x4C32 , "HANGUL SYLLABLE RIEUL O SSANGKIYEOK" },
	{ 0x4C33 , "HANGUL SYLLABLE RIEUL O KIYEOK-SIOS" },
	{ 0x4C34 , "HANGUL SYLLABLE RIEUL O NIEUN-CIEUC" },
	{ 0x4C35 , "HANGUL SYLLABLE RIEUL O NIEUN-HIEUH" },
	{ 0x4C36 , "HANGUL SYLLABLE RIEUL O RIEUL-KIYEOK" },
	{ 0x4C37 , "HANGUL SYLLABLE RIEUL O RIEUL-MIEUM" },
	{ 0x4C38 , "HANGUL SYLLABLE RIEUL O RIEUL-PIEUP" },
	{ 0x4C39 , "HANGUL SYLLABLE RIEUL O RIEUL-THIEUTH" },
	{ 0x4C3A , "HANGUL SYLLABLE RIEUL O RIEUL-PHIEUPH" },
	{ 0x4C3B , "HANGUL SYLLABLE RIEUL O PIEUP-SIOS" },
	{ 0x4C3C , "HANGUL SYLLABLE RIEUL O SSANGSIOS" },
	{ 0x4C3D , "HANGUL SYLLABLE RIEUL O CIEUC" },
	{ 0x4C3E , "HANGUL SYLLABLE RIEUL O CHIEUCH" },
	{ 0x4C3F , "HANGUL SYLLABLE RIEUL O THIEUTH" },
	{ 0x4C40 , "HANGUL SYLLABLE RIEUL WA SSANGKIYEOK" },
	{ 0x4C41 , "HANGUL SYLLABLE RIEUL WA KIYEOK-SIOS" },
	{ 0x4C42 , "HANGUL SYLLABLE RIEUL WA NIEUN-CIEUC" },
	{ 0x4C43 , "HANGUL SYLLABLE RIEUL WA NIEUN-HIEUH" },
	{ 0x4C44 , "HANGUL SYLLABLE RIEUL WA TIKEUT" },
	{ 0x4C45 , "HANGUL SYLLABLE RIEUL WA RIEUL-KIYEOK" },
	{ 0x4C46 , "HANGUL SYLLABLE RIEUL WA RIEUL-MIEUM" },
	{ 0x4C47 , "HANGUL SYLLABLE RIEUL WA RIEUL-PIEUP" },
	{ 0x4C48 , "HANGUL SYLLABLE RIEUL WA RIEUL-SIOS" },
	{ 0x4C49 , "HANGUL SYLLABLE RIEUL WA RIEUL-THIEUTH" },
	{ 0x4C4A , "HANGUL SYLLABLE RIEUL WA RIEUL-PHIEUPH" },
	{ 0x4C4B , "HANGUL SYLLABLE RIEUL WA RIEUL-HIEUH" },
	{ 0x4C4C , "HANGUL SYLLABLE RIEUL WA PIEUP-SIOS" },
	{ 0x4C4D , "HANGUL SYLLABLE RIEUL WA CIEUC" },
	{ 0x4C4E , "HANGUL SYLLABLE RIEUL WA CHIEUCH" },
	{ 0x4C4F , "HANGUL SYLLABLE RIEUL WA KHIEUKH" },
	{ 0x4C50 , "HANGUL SYLLABLE RIEUL WA THIEUTH" },
	{ 0x4C51 , "HANGUL SYLLABLE RIEUL WA PHIEUPH" },
	{ 0x4C52 , "HANGUL SYLLABLE RIEUL WA HIEUH" },
	{ 0x4C53 , "HANGUL SYLLABLE RIEUL WAE KIYEOK" },
	{ 0x4C54 , "HANGUL SYLLABLE RIEUL WAE SSANGKIYEOK" },
	{ 0x4C55 , "HANGUL SYLLABLE RIEUL WAE KIYEOK-SIOS" },
	{ 0x4C56 , "HANGUL SYLLABLE RIEUL WAE NIEUN-CIEUC" },
	{ 0x4C57 , "HANGUL SYLLABLE RIEUL WAE NIEUN-HIEUH" },
	{ 0x4C58 , "HANGUL SYLLABLE RIEUL WAE RIEUL-KIYEOK" },
	{ 0x4C59 , "HANGUL SYLLABLE RIEUL WAE RIEUL-MIEUM" },
	{ 0x4C5A , "HANGUL SYLLABLE RIEUL WAE RIEUL-PIEUP" },
	{ 0x4C5B , "HANGUL SYLLABLE RIEUL WAE RIEUL-SIOS" },
	{ 0x4C5C , "HANGUL SYLLABLE RIEUL WAE RIEUL-THIEUTH" },
	{ 0x4C5D , "HANGUL SYLLABLE RIEUL WAE RIEUL-PHIEUPH" },
	{ 0x4C5E , "HANGUL SYLLABLE RIEUL WAE RIEUL-HIEUH" },
	{ 0x4C5F , "HANGUL SYLLABLE RIEUL WAE MIEUM" },
	{ 0x4C60 , "HANGUL SYLLABLE RIEUL WAE PIEUP" },
	{ 0x4C61 , "HANGUL SYLLABLE RIEUL WAE PIEUP-SIOS" },
	{ 0x4C62 , "HANGUL SYLLABLE RIEUL WAE IEUNG" },
	{ 0x4C63 , "HANGUL SYLLABLE RIEUL WAE CIEUC" },
	{ 0x4C64 , "HANGUL SYLLABLE RIEUL WAE CHIEUCH" },
	{ 0x4C65 , "HANGUL SYLLABLE RIEUL WAE KHIEUKH" },
	{ 0x4C66 , "HANGUL SYLLABLE RIEUL WAE THIEUTH" },
	{ 0x4C67 , "HANGUL SYLLABLE RIEUL WAE PHIEUPH" },
	{ 0x4C68 , "HANGUL SYLLABLE RIEUL WAE HIEUH" },
	{ 0x4C69 , "HANGUL SYLLABLE RIEUL OE SSANGKIYEOK" },
	{ 0x4C6A , "HANGUL SYLLABLE RIEUL OE KIYEOK-SIOS" },
	{ 0x4C6B , "HANGUL SYLLABLE RIEUL OE NIEUN-HIEUH" },
	{ 0x4C6C , "HANGUL SYLLABLE RIEUL OE RIEUL-KIYEOK" },
	{ 0x4C6D , "HANGUL SYLLABLE RIEUL OE RIEUL-MIEUM" },
	{ 0x4C6E , "HANGUL SYLLABLE RIEUL OE RIEUL-PIEUP" },
	{ 0x4C6F , "HANGUL SYLLABLE RIEUL OE RIEUL-SIOS" },
	{ 0x4C70 , "HANGUL SYLLABLE RIEUL OE RIEUL-THIEUTH" },
	{ 0x4C71 , "HANGUL SYLLABLE RIEUL OE RIEUL-PHIEUPH" },
	{ 0x4C72 , "HANGUL SYLLABLE RIEUL OE RIEUL-HIEUH" },
	{ 0x4C73 , "HANGUL SYLLABLE RIEUL OE PIEUP-SIOS" },
	{ 0x4C74 , "HANGUL SYLLABLE RIEUL OE CIEUC" },
	{ 0x4C75 , "HANGUL SYLLABLE RIEUL OE CHIEUCH" },
	{ 0x4C76 , "HANGUL SYLLABLE RIEUL OE KHIEUKH" },
	{ 0x4C77 , "HANGUL SYLLABLE RIEUL OE THIEUTH" },
	{ 0x4C78 , "HANGUL SYLLABLE RIEUL OE PHIEUPH" },
	{ 0x4C79 , "HANGUL SYLLABLE RIEUL OE HIEUH" },
	{ 0x4C7A , "HANGUL SYLLABLE RIEUL YO SSANGKIYEOK" },
	{ 0x4C7B , "HANGUL SYLLABLE RIEUL YO KIYEOK-SIOS" },
	{ 0x4C7C , "HANGUL SYLLABLE RIEUL YO NIEUN-CIEUC" },
	{ 0x4C7D , "HANGUL SYLLABLE RIEUL YO NIEUN-HIEUH" },
	{ 0x4C7E , "HANGUL SYLLABLE RIEUL YO RIEUL-KIYEOK" },
	{ 0x4C7F , "HANGUL SYLLABLE RIEUL YO RIEUL-MIEUM" },
	{ 0x4C80 , "HANGUL SYLLABLE RIEUL YO RIEUL-PIEUP" },
	{ 0x4C81 , "HANGUL SYLLABLE RIEUL YO RIEUL-SIOS" },
	{ 0x4C82 , "HANGUL SYLLABLE RIEUL YO RIEUL-THIEUTH" },
	{ 0x4C83 , "HANGUL SYLLABLE RIEUL YO RIEUL-PHIEUPH" },
	{ 0x4C84 , "HANGUL SYLLABLE RIEUL YO RIEUL-HIEUH" },
	{ 0x4C85 , "HANGUL SYLLABLE RIEUL YO PIEUP-SIOS" },
	{ 0x4C86 , "HANGUL SYLLABLE RIEUL YO SSANGSIOS" },
	{ 0x4C87 , "HANGUL SYLLABLE RIEUL YO CIEUC" },
	{ 0x4C88 , "HANGUL SYLLABLE RIEUL YO CHIEUCH" },
	{ 0x4C89 , "HANGUL SYLLABLE RIEUL YO KHIEUKH" },
	{ 0x4C8A , "HANGUL SYLLABLE RIEUL YO THIEUTH" },
	{ 0x4C8B , "HANGUL SYLLABLE RIEUL YO PHIEUPH" },
	{ 0x4C8C , "HANGUL SYLLABLE RIEUL YO HIEUH" },
	{ 0x4C8D , "HANGUL SYLLABLE RIEUL U SSANGKIYEOK" },
	{ 0x4C8E , "HANGUL SYLLABLE RIEUL U KIYEOK-SIOS" },
	{ 0x4C8F , "HANGUL SYLLABLE RIEUL U NIEUN-CIEUC" },
	{ 0x4C90 , "HANGUL SYLLABLE RIEUL U NIEUN-HIEUH" },
	{ 0x4C91 , "HANGUL SYLLABLE RIEUL U RIEUL-KIYEOK" },
	{ 0x4C92 , "HANGUL SYLLABLE RIEUL U RIEUL-MIEUM" },
	{ 0x4C93 , "HANGUL SYLLABLE RIEUL U RIEUL-SIOS" },
	{ 0x4C94 , "HANGUL SYLLABLE RIEUL U RIEUL-THIEUTH" },
	{ 0x4C95 , "HANGUL SYLLABLE RIEUL U RIEUL-PHIEUPH" },
	{ 0x4C96 , "HANGUL SYLLABLE RIEUL U RIEUL-HIEUH" },
	{ 0x4C97 , "HANGUL SYLLABLE RIEUL U PIEUP-SIOS" },
	{ 0x4C98 , "HANGUL SYLLABLE RIEUL U SSANGSIOS" },
	{ 0x4C99 , "HANGUL SYLLABLE RIEUL U CIEUC" },
	{ 0x4C9A , "HANGUL SYLLABLE RIEUL U CHIEUCH" },
	{ 0x4C9B , "HANGUL SYLLABLE RIEUL U THIEUTH" },
	{ 0x4C9C , "HANGUL SYLLABLE RIEUL U HIEUH" },
	{ 0x4C9D , "HANGUL SYLLABLE RIEUL WEO SSANGKIYEOK" },
	{ 0x4C9E , "HANGUL SYLLABLE RIEUL WEO KIYEOK-SIOS" },
	{ 0x4C9F , "HANGUL SYLLABLE RIEUL WEO NIEUN-CIEUC" },
	{ 0x4CA0 , "HANGUL SYLLABLE RIEUL WEO NIEUN-HIEUH" },
	{ 0x4CA1 , "HANGUL SYLLABLE RIEUL WEO RIEUL-KIYEOK" },
	{ 0x4CA2 , "HANGUL SYLLABLE RIEUL WEO RIEUL-MIEUM" },
	{ 0x4CA3 , "HANGUL SYLLABLE RIEUL WEO RIEUL-PIEUP" },
	{ 0x4CA4 , "HANGUL SYLLABLE RIEUL WEO RIEUL-SIOS" },
	{ 0x4CA5 , "HANGUL SYLLABLE RIEUL WEO RIEUL-THIEUTH" },
	{ 0x4CA6 , "HANGUL SYLLABLE RIEUL WEO RIEUL-PHIEUPH" },
	{ 0x4CA7 , "HANGUL SYLLABLE RIEUL WEO RIEUL-HIEUH" },
	{ 0x4CA8 , "HANGUL SYLLABLE RIEUL WEO PIEUP-SIOS" },
	{ 0x4CA9 , "HANGUL SYLLABLE RIEUL WEO CIEUC" },
	{ 0x4CAA , "HANGUL SYLLABLE RIEUL WEO CHIEUCH" },
	{ 0x4CAB , "HANGUL SYLLABLE RIEUL WEO KHIEUKH" },
	{ 0x4CAC , "HANGUL SYLLABLE RIEUL WEO THIEUTH" },
	{ 0x4CAD , "HANGUL SYLLABLE RIEUL WEO PHIEUPH" },
	{ 0x4CAE , "HANGUL SYLLABLE RIEUL WEO HIEUH" },
	{ 0x4CAF , "HANGUL SYLLABLE RIEUL WE SSANGKIYEOK" },
	{ 0x4CB0 , "HANGUL SYLLABLE RIEUL WE KIYEOK-SIOS" },
	{ 0x4CB1 , "HANGUL SYLLABLE RIEUL WE NIEUN-CIEUC" },
	{ 0x4CB2 , "HANGUL SYLLABLE RIEUL WE NIEUN-HIEUH" },
	{ 0x4CB3 , "HANGUL SYLLABLE RIEUL WE TIKEUT" },
	{ 0x4CB4 , "HANGUL SYLLABLE RIEUL WE RIEUL-KIYEOK" },
	{ 0x4CB5 , "HANGUL SYLLABLE RIEUL WE RIEUL-MIEUM" },
	{ 0x4CB6 , "HANGUL SYLLABLE RIEUL WE RIEUL-PIEUP" },
	{ 0x4CB7 , "HANGUL SYLLABLE RIEUL WE RIEUL-SIOS" },
	{ 0x4CB8 , "HANGUL SYLLABLE RIEUL WE RIEUL-THIEUTH" },
	{ 0x4CB9 , "HANGUL SYLLABLE RIEUL WE RIEUL-PHIEUPH" },
	{ 0x4CBA , "HANGUL SYLLABLE RIEUL WE RIEUL-HIEUH" },
	{ 0x4CBB , "HANGUL SYLLABLE RIEUL WE PIEUP-SIOS" },
	{ 0x4CBC , "HANGUL SYLLABLE RIEUL WE CIEUC" },
	{ 0x4CBD , "HANGUL SYLLABLE RIEUL WE CHIEUCH" },
	{ 0x4CBE , "HANGUL SYLLABLE RIEUL WE KHIEUKH" },
	{ 0x4CBF , "HANGUL SYLLABLE RIEUL WE THIEUTH" },
	{ 0x4CC0 , "HANGUL SYLLABLE RIEUL WE PHIEUPH" },
	{ 0x4CC1 , "HANGUL SYLLABLE RIEUL WE HIEUH" },
	{ 0x4CC2 , "HANGUL SYLLABLE RIEUL WI SSANGKIYEOK" },
	{ 0x4CC3 , "HANGUL SYLLABLE RIEUL WI KIYEOK-SIOS" },
	{ 0x4CC4 , "HANGUL SYLLABLE RIEUL WI NIEUN-CIEUC" },
	{ 0x4CC5 , "HANGUL SYLLABLE RIEUL WI NIEUN-HIEUH" },
	{ 0x4CC6 , "HANGUL SYLLABLE RIEUL WI RIEUL-KIYEOK" },
	{ 0x4CC7 , "HANGUL SYLLABLE RIEUL WI RIEUL-MIEUM" },
	{ 0x4CC8 , "HANGUL SYLLABLE RIEUL WI RIEUL-PIEUP" },
	{ 0x4CC9 , "HANGUL SYLLABLE RIEUL WI RIEUL-SIOS" },
	{ 0x4CCA , "HANGUL SYLLABLE RIEUL WI RIEUL-THIEUTH" },
	{ 0x4CCB , "HANGUL SYLLABLE RIEUL WI RIEUL-PHIEUPH" },
	{ 0x4CCC , "HANGUL SYLLABLE RIEUL WI RIEUL-HIEUH" },
	{ 0x4CCD , "HANGUL SYLLABLE RIEUL WI PIEUP-SIOS" },
	{ 0x4CCE , "HANGUL SYLLABLE RIEUL WI CIEUC" },
	{ 0x4CCF , "HANGUL SYLLABLE RIEUL WI CHIEUCH" },
	{ 0x4CD0 , "HANGUL SYLLABLE RIEUL WI KHIEUKH" },
	{ 0x4CD1 , "HANGUL SYLLABLE RIEUL WI THIEUTH" },
	{ 0x4CD2 , "HANGUL SYLLABLE RIEUL WI PHIEUPH" },
	{ 0x4CD3 , "HANGUL SYLLABLE RIEUL WI HIEUH" },
	{ 0x4CD4 , "HANGUL SYLLABLE RIEUL YU SSANGKIYEOK" },
	{ 0x4CD5 , "HANGUL SYLLABLE RIEUL YU KIYEOK-SIOS" },
	{ 0x4CD6 , "HANGUL SYLLABLE RIEUL YU NIEUN-CIEUC" },
	{ 0x4CD7 , "HANGUL SYLLABLE RIEUL YU NIEUN-HIEUH" },
	{ 0x4CD8 , "HANGUL SYLLABLE RIEUL YU RIEUL-KIYEOK" },
	{ 0x4CD9 , "HANGUL SYLLABLE RIEUL YU RIEUL-MIEUM" },
	{ 0x4CDA , "HANGUL SYLLABLE RIEUL YU RIEUL-PIEUP" },
	{ 0x4CDB , "HANGUL SYLLABLE RIEUL YU RIEUL-SIOS" },
	{ 0x4CDC , "HANGUL SYLLABLE RIEUL YU RIEUL-THIEUTH" },
	{ 0x4CDD , "HANGUL SYLLABLE RIEUL YU RIEUL-PHIEUPH" },
	{ 0x4CDE , "HANGUL SYLLABLE RIEUL YU RIEUL-HIEUH" },
	{ 0x4CDF , "HANGUL SYLLABLE RIEUL YU PIEUP-SIOS" },
	{ 0x4CE0 , "HANGUL SYLLABLE RIEUL YU SSANGSIOS" },
	{ 0x4CE1 , "HANGUL SYLLABLE RIEUL YU CIEUC" },
	{ 0x4CE2 , "HANGUL SYLLABLE RIEUL YU CHIEUCH" },
	{ 0x4CE3 , "HANGUL SYLLABLE RIEUL YU KHIEUKH" },
	{ 0x4CE4 , "HANGUL SYLLABLE RIEUL YU THIEUTH" },
	{ 0x4CE5 , "HANGUL SYLLABLE RIEUL YU PHIEUPH" },
	{ 0x4CE6 , "HANGUL SYLLABLE RIEUL YU HIEUH" },
	{ 0x4CE7 , "HANGUL SYLLABLE RIEUL EU SSANGKIYEOK" },
	{ 0x4CE8 , "HANGUL SYLLABLE RIEUL EU KIYEOK-SIOS" },
	{ 0x4CE9 , "HANGUL SYLLABLE RIEUL EU NIEUN-CIEUC" },
	{ 0x4CEA , "HANGUL SYLLABLE RIEUL EU NIEUN-HIEUH" },
	{ 0x4CEB , "HANGUL SYLLABLE RIEUL EU RIEUL-MIEUM" },
	{ 0x4CEC , "HANGUL SYLLABLE RIEUL EU RIEUL-PIEUP" },
	{ 0x4CED , "HANGUL SYLLABLE RIEUL EU RIEUL-THIEUTH" },
	{ 0x4CEE , "HANGUL SYLLABLE RIEUL EU RIEUL-PHIEUPH" },
	{ 0x4CEF , "HANGUL SYLLABLE RIEUL EU RIEUL-HIEUH" },
	{ 0x4CF0 , "HANGUL SYLLABLE RIEUL EU PIEUP-SIOS" },
	{ 0x4CF1 , "HANGUL SYLLABLE RIEUL EU SSANGSIOS" },
	{ 0x4CF2 , "HANGUL SYLLABLE RIEUL EU KHIEUKH" },
	{ 0x4CF3 , "HANGUL SYLLABLE RIEUL YI SSANGKIYEOK" },
	{ 0x4CF4 , "HANGUL SYLLABLE RIEUL YI KIYEOK-SIOS" },
	{ 0x4CF5 , "HANGUL SYLLABLE RIEUL YI NIEUN-CIEUC" },
	{ 0x4CF6 , "HANGUL SYLLABLE RIEUL YI NIEUN-HIEUH" },
	{ 0x4CF7 , "HANGUL SYLLABLE RIEUL YI TIKEUT" },
	{ 0x4CF8 , "HANGUL SYLLABLE RIEUL YI RIEUL-KIYEOK" },
	{ 0x4CF9 , "HANGUL SYLLABLE RIEUL YI RIEUL-MIEUM" },
	{ 0x4CFA , "HANGUL SYLLABLE RIEUL YI RIEUL-PIEUP" },
	{ 0x4CFB , "HANGUL SYLLABLE RIEUL YI RIEUL-SIOS" },
	{ 0x4CFC , "HANGUL SYLLABLE RIEUL YI RIEUL-THIEUTH" },
	{ 0x4CFD , "HANGUL SYLLABLE RIEUL YI RIEUL-HIEUH" },
	{ 0x4CFE , "HANGUL SYLLABLE RIEUL YI PIEUP-SIOS" },
	{ 0x4CFF , "HANGUL SYLLABLE RIEUL YI SSANGSIOS" },
	{ 0x4D00 , "HANGUL SYLLABLE RIEUL YI CIEUC" },
	{ 0x4D01 , "HANGUL SYLLABLE RIEUL YI CHIEUCH" },
	{ 0x4D02 , "HANGUL SYLLABLE RIEUL YI KHIEUKH" },
	{ 0x4D03 , "HANGUL SYLLABLE RIEUL YI THIEUTH" },
	{ 0x4D04 , "HANGUL SYLLABLE RIEUL YI PHIEUPH" },
	{ 0x4D05 , "HANGUL SYLLABLE RIEUL YI HIEUH" },
	{ 0x4D06 , "HANGUL SYLLABLE RIEUL I SSANGKIYEOK" },
	{ 0x4D07 , "HANGUL SYLLABLE RIEUL I KIYEOK-SIOS" },
	{ 0x4D08 , "HANGUL SYLLABLE RIEUL I NIEUN-CIEUC" },
	{ 0x4D09 , "HANGUL SYLLABLE RIEUL I NIEUN-HIEUH" },
	{ 0x4D0A , "HANGUL SYLLABLE RIEUL I RIEUL-KIYEOK" },
	{ 0x4D0B , "HANGUL SYLLABLE RIEUL I RIEUL-MIEUM" },
	{ 0x4D0C , "HANGUL SYLLABLE RIEUL I RIEUL-PIEUP" },
	{ 0x4D0D , "HANGUL SYLLABLE RIEUL I RIEUL-SIOS" },
	{ 0x4D0E , "HANGUL SYLLABLE RIEUL I RIEUL-THIEUTH" },
	{ 0x4D0F , "HANGUL SYLLABLE RIEUL I RIEUL-PHIEUPH" },
	{ 0x4D10 , "HANGUL SYLLABLE RIEUL I RIEUL-HIEUH" },
	{ 0x4D11 , "HANGUL SYLLABLE RIEUL I CIEUC" },
	{ 0x4D12 , "HANGUL SYLLABLE RIEUL I CHIEUCH" },
	{ 0x4D13 , "HANGUL SYLLABLE RIEUL I THIEUTH" },
	{ 0x4D14 , "HANGUL SYLLABLE RIEUL I HIEUH" },
	{ 0x4D15 , "HANGUL SYLLABLE MIEUM A KIYEOK-SIOS" },
	{ 0x4D16 , "HANGUL SYLLABLE MIEUM A NIEUN-CIEUC" },
	{ 0x4D17 , "HANGUL SYLLABLE MIEUM A RIEUL-PIEUP" },
	{ 0x4D18 , "HANGUL SYLLABLE MIEUM A RIEUL-THIEUTH" },
	{ 0x4D19 , "HANGUL SYLLABLE MIEUM A RIEUL-PHIEUPH" },
	{ 0x4D1A , "HANGUL SYLLABLE MIEUM A RIEUL-HIEUH" },
	{ 0x4D1B , "HANGUL SYLLABLE MIEUM A PIEUP-SIOS" },
	{ 0x4D1C , "HANGUL SYLLABLE MIEUM AE SSANGKIYEOK" },
	{ 0x4D1D , "HANGUL SYLLABLE MIEUM AE KIYEOK-SIOS" },
	{ 0x4D1E , "HANGUL SYLLABLE MIEUM AE NIEUN-CIEUC" },
	{ 0x4D1F , "HANGUL SYLLABLE MIEUM AE NIEUN-HIEUH" },
	{ 0x4D20 , "HANGUL SYLLABLE MIEUM AE RIEUL-MIEUM" },
	{ 0x4D21 , "HANGUL SYLLABLE MIEUM AE RIEUL-PIEUP" },
	{ 0x4D22 , "HANGUL SYLLABLE MIEUM AE RIEUL-SIOS" },
	{ 0x4D23 , "HANGUL SYLLABLE MIEUM AE RIEUL-THIEUTH" },
	{ 0x4D24 , "HANGUL SYLLABLE MIEUM AE RIEUL-PHIEUPH" },
	{ 0x4D25 , "HANGUL SYLLABLE MIEUM AE RIEUL-HIEUH" },
	{ 0x4D26 , "HANGUL SYLLABLE MIEUM AE PIEUP-SIOS" },
	{ 0x4D27 , "HANGUL SYLLABLE MIEUM AE KHIEUKH" },
	{ 0x4D28 , "HANGUL SYLLABLE MIEUM AE HIEUH" },
	{ 0x4D29 , "HANGUL SYLLABLE MIEUM YA SSANGKIYEOK" },
	{ 0x4D2A , "HANGUL SYLLABLE MIEUM YA KIYEOK-SIOS" },
	{ 0x4D2B , "HANGUL SYLLABLE MIEUM YA NIEUN-CIEUC" },
	{ 0x4D2C , "HANGUL SYLLABLE MIEUM YA NIEUN-HIEUH" },
	{ 0x4D2D , "HANGUL SYLLABLE MIEUM YA TIKEUT" },
	{ 0x4D2E , "HANGUL SYLLABLE MIEUM YA RIEUL-KIYEOK" },
	{ 0x4D2F , "HANGUL SYLLABLE MIEUM YA RIEUL-MIEUM" },
	{ 0x4D30 , "HANGUL SYLLABLE MIEUM YA RIEUL-PIEUP" },
	{ 0x4D31 , "HANGUL SYLLABLE MIEUM YA RIEUL-SIOS" },
	{ 0x4D32 , "HANGUL SYLLABLE MIEUM YA RIEUL-THIEUTH" },
	{ 0x4D33 , "HANGUL SYLLABLE MIEUM YA RIEUL-PHIEUPH" },
	{ 0x4D34 , "HANGUL SYLLABLE MIEUM YA RIEUL-HIEUH" },
	{ 0x4D35 , "HANGUL SYLLABLE MIEUM YA PIEUP-SIOS" },
	{ 0x4D36 , "HANGUL SYLLABLE MIEUM YA SSANGSIOS" },
	{ 0x4D37 , "HANGUL SYLLABLE MIEUM YA CIEUC" },
	{ 0x4D38 , "HANGUL SYLLABLE MIEUM YA CHIEUCH" },
	{ 0x4D39 , "HANGUL SYLLABLE MIEUM YA THIEUTH" },
	{ 0x4D3A , "HANGUL SYLLABLE MIEUM YA PHIEUPH" },
	{ 0x4D3B , "HANGUL SYLLABLE MIEUM YA HIEUH" },
	{ 0x4D3C , "HANGUL SYLLABLE MIEUM YAE KIYEOK" },
	{ 0x4D3D , "HANGUL SYLLABLE MIEUM YAE SSANGKIYEOK" },
	{ 0x4D3E , "HANGUL SYLLABLE MIEUM YAE KIYEOK-SIOS" },
	{ 0x4D3F , "HANGUL SYLLABLE MIEUM YAE NIEUN" },
	{ 0x4D40 , "HANGUL SYLLABLE MIEUM YAE NIEUN-CIEUC" },
	{ 0x4D41 , "HANGUL SYLLABLE MIEUM YAE NIEUN-HIEUH" },
	{ 0x4D42 , "HANGUL SYLLABLE MIEUM YAE TIKEUT" },
	{ 0x4D43 , "HANGUL SYLLABLE MIEUM YAE RIEUL" },
	{ 0x4D44 , "HANGUL SYLLABLE MIEUM YAE RIEUL-KIYEOK" },
	{ 0x4D45 , "HANGUL SYLLABLE MIEUM YAE RIEUL-MIEUM" },
	{ 0x4D46 , "HANGUL SYLLABLE MIEUM YAE RIEUL-PIEUP" },
	{ 0x4D47 , "HANGUL SYLLABLE MIEUM YAE RIEUL-SIOS" },
	{ 0x4D48 , "HANGUL SYLLABLE MIEUM YAE RIEUL-THIEUTH" },
	{ 0x4D49 , "HANGUL SYLLABLE MIEUM YAE RIEUL-PHIEUPH" },
	{ 0x4D4A , "HANGUL SYLLABLE MIEUM YAE RIEUL-HIEUH" },
	{ 0x4D4B , "HANGUL SYLLABLE MIEUM YAE MIEUM" },
	{ 0x4D4C , "HANGUL SYLLABLE MIEUM YAE PIEUP" },
	{ 0x4D4D , "HANGUL SYLLABLE MIEUM YAE PIEUP-SIOS" },
	{ 0x4D4E , "HANGUL SYLLABLE MIEUM YAE SIOS" },
	{ 0x4D4F , "HANGUL SYLLABLE MIEUM YAE SSANGSIOS" },
	{ 0x4D50 , "HANGUL SYLLABLE MIEUM YAE IEUNG" },
	{ 0x4D51 , "HANGUL SYLLABLE MIEUM YAE CIEUC" },
	{ 0x4D52 , "HANGUL SYLLABLE MIEUM YAE CHIEUCH" },
	{ 0x4D53 , "HANGUL SYLLABLE MIEUM YAE KHIEUKH" },
	{ 0x4D54 , "HANGUL SYLLABLE MIEUM YAE THIEUTH" },
	{ 0x4D55 , "HANGUL SYLLABLE MIEUM YAE PHIEUPH" },
	{ 0x4D56 , "HANGUL SYLLABLE MIEUM YAE HIEUH" },
	{ 0x4D57 , "HANGUL SYLLABLE MIEUM EO SSANGKIYEOK" },
	{ 0x4D58 , "HANGUL SYLLABLE MIEUM EO NIEUN-CIEUC" },
	{ 0x4D59 , "HANGUL SYLLABLE MIEUM EO NIEUN-HIEUH" },
	{ 0x4D5A , "HANGUL SYLLABLE MIEUM EO RIEUL-PIEUP" },
	{ 0x4D5B , "HANGUL SYLLABLE MIEUM EO RIEUL-SIOS" },
	{ 0x4D5C , "HANGUL SYLLABLE MIEUM EO RIEUL-THIEUTH" },
	{ 0x4D5D , "HANGUL SYLLABLE MIEUM EO RIEUL-PHIEUPH" },
	{ 0x4D5E , "HANGUL SYLLABLE MIEUM EO RIEUL-HIEUH" },
	{ 0x4D5F , "HANGUL SYLLABLE MIEUM EO PIEUP-SIOS" },
	{ 0x4D60 , "HANGUL SYLLABLE MIEUM EO KHIEUKH" },
	{ 0x4D61 , "HANGUL SYLLABLE MIEUM EO THIEUTH" },
	{ 0x4D62 , "HANGUL SYLLABLE MIEUM EO PHIEUPH" },
	{ 0x4D63 , "HANGUL SYLLABLE MIEUM E SSANGKIYEOK" },
	{ 0x4D64 , "HANGUL SYLLABLE MIEUM E KIYEOK-SIOS" },
	{ 0x4D65 , "HANGUL SYLLABLE MIEUM E NIEUN-CIEUC" },
	{ 0x4D66 , "HANGUL SYLLABLE MIEUM E NIEUN-HIEUH" },
	{ 0x4D67 , "HANGUL SYLLABLE MIEUM E RIEUL-KIYEOK" },
	{ 0x4D68 , "HANGUL SYLLABLE MIEUM E RIEUL-MIEUM" },
	{ 0x4D69 , "HANGUL SYLLABLE MIEUM E RIEUL-PIEUP" },
	{ 0x4D6A , "HANGUL SYLLABLE MIEUM E RIEUL-THIEUTH" },
	{ 0x4D6B , "HANGUL SYLLABLE MIEUM E RIEUL-PHIEUPH" },
	{ 0x4D6C , "HANGUL SYLLABLE MIEUM E RIEUL-HIEUH" },
	{ 0x4D6D , "HANGUL SYLLABLE MIEUM E PIEUP-SIOS" },
	{ 0x4D6E , "HANGUL SYLLABLE MIEUM E KHIEUKH" },
	{ 0x4D6F , "HANGUL SYLLABLE MIEUM E PHIEUPH" },
	{ 0x4D70 , "HANGUL SYLLABLE MIEUM E HIEUH" },
	{ 0x4D71 , "HANGUL SYLLABLE MIEUM YEO SSANGKIYEOK" },
	{ 0x4D72 , "HANGUL SYLLABLE MIEUM YEO KIYEOK-SIOS" },
	{ 0x4D73 , "HANGUL SYLLABLE MIEUM YEO NIEUN-CIEUC" },
	{ 0x4D74 , "HANGUL SYLLABLE MIEUM YEO NIEUN-HIEUH" },
	{ 0x4D75 , "HANGUL SYLLABLE MIEUM YEO RIEUL-KIYEOK" },
	{ 0x4D76 , "HANGUL SYLLABLE MIEUM YEO RIEUL-PIEUP" },
	{ 0x4D77 , "HANGUL SYLLABLE MIEUM YEO RIEUL-SIOS" },
	{ 0x4D78 , "HANGUL SYLLABLE MIEUM YEO RIEUL-THIEUTH" },
	{ 0x4D79 , "HANGUL SYLLABLE MIEUM YEO RIEUL-PHIEUPH" },
	{ 0x4D7A , "HANGUL SYLLABLE MIEUM YEO RIEUL-HIEUH" },
	{ 0x4D7B , "HANGUL SYLLABLE MIEUM YEO PIEUP-SIOS" },
	{ 0x4D7C , "HANGUL SYLLABLE MIEUM YEO CIEUC" },
	{ 0x4D7D , "HANGUL SYLLABLE MIEUM YEO KHIEUKH" },
	{ 0x4D7E , "HANGUL SYLLABLE MIEUM YEO PHIEUPH" },
	{ 0x4D7F , "HANGUL SYLLABLE MIEUM YEO HIEUH" },
	{ 0x4D80 , "HANGUL SYLLABLE MIEUM YE SSANGKIYEOK" },
	{ 0x4D81 , "HANGUL SYLLABLE MIEUM YE KIYEOK-SIOS" },
	{ 0x4D82 , "HANGUL SYLLABLE MIEUM YE NIEUN-CIEUC" },
	{ 0x4D83 , "HANGUL SYLLABLE MIEUM YE NIEUN-HIEUH" },
	{ 0x4D84 , "HANGUL SYLLABLE MIEUM YE RIEUL-KIYEOK" },
	{ 0x4D85 , "HANGUL SYLLABLE MIEUM YE RIEUL-MIEUM" },
	{ 0x4D86 , "HANGUL SYLLABLE MIEUM YE RIEUL-PIEUP" },
	{ 0x4D87 , "HANGUL SYLLABLE MIEUM YE RIEUL-SIOS" },
	{ 0x4D88 , "HANGUL SYLLABLE MIEUM YE RIEUL-THIEUTH" },
	{ 0x4D89 , "HANGUL SYLLABLE MIEUM YE RIEUL-PHIEUPH" },
	{ 0x4D8A , "HANGUL SYLLABLE MIEUM YE RIEUL-HIEUH" },
	{ 0x4D8B , "HANGUL SYLLABLE MIEUM YE PIEUP-SIOS" },
	{ 0x4D8C , "HANGUL SYLLABLE MIEUM YE SSANGSIOS" },
	{ 0x4D8D , "HANGUL SYLLABLE MIEUM YE CIEUC" },
	{ 0x4D8E , "HANGUL SYLLABLE MIEUM YE KHIEUKH" },
	{ 0x4D8F , "HANGUL SYLLABLE MIEUM YE HIEUH" },
	{ 0x4D90 , "HANGUL SYLLABLE MIEUM O SSANGKIYEOK" },
	{ 0x4D91 , "HANGUL SYLLABLE MIEUM O NIEUN-CIEUC" },
	{ 0x4D92 , "HANGUL SYLLABLE MIEUM O NIEUN-HIEUH" },
	{ 0x4D93 , "HANGUL SYLLABLE MIEUM O RIEUL-PIEUP" },
	{ 0x4D94 , "HANGUL SYLLABLE MIEUM O RIEUL-THIEUTH" },
	{ 0x4D95 , "HANGUL SYLLABLE MIEUM O RIEUL-PHIEUPH" },
	{ 0x4D96 , "HANGUL SYLLABLE MIEUM O RIEUL-HIEUH" },
	{ 0x4D97 , "HANGUL SYLLABLE MIEUM O PIEUP-SIOS" },
	{ 0x4D98 , "HANGUL SYLLABLE MIEUM O SSANGSIOS" },
	{ 0x4D99 , "HANGUL SYLLABLE MIEUM O CIEUC" },
	{ 0x4D9A , "HANGUL SYLLABLE MIEUM O KHIEUKH" },
	{ 0x4D9B , "HANGUL SYLLABLE MIEUM O THIEUTH" },
	{ 0x4D9C , "HANGUL SYLLABLE MIEUM O PHIEUPH" },
	{ 0x4D9D , "HANGUL SYLLABLE MIEUM WA SSANGKIYEOK" },
	{ 0x4D9E , "HANGUL SYLLABLE MIEUM WA KIYEOK-SIOS" },
	{ 0x4D9F , "HANGUL SYLLABLE MIEUM WA NIEUN-CIEUC" },
	{ 0x4DA0 , "HANGUL SYLLABLE MIEUM WA NIEUN-HIEUH" },
	{ 0x4DA1 , "HANGUL SYLLABLE MIEUM WA RIEUL-KIYEOK" },
	{ 0x4DA2 , "HANGUL SYLLABLE MIEUM WA RIEUL-MIEUM" },
	{ 0x4DA3 , "HANGUL SYLLABLE MIEUM WA RIEUL-PIEUP" },
	{ 0x4DA4 , "HANGUL SYLLABLE MIEUM WA RIEUL-SIOS" },
	{ 0x4DA5 , "HANGUL SYLLABLE MIEUM WA RIEUL-THIEUTH" },
	{ 0x4DA6 , "HANGUL SYLLABLE MIEUM WA RIEUL-PHIEUPH" },
	{ 0x4DA7 , "HANGUL SYLLABLE MIEUM WA RIEUL-HIEUH" },
	{ 0x4DA8 , "HANGUL SYLLABLE MIEUM WA PIEUP-SIOS" },
	{ 0x4DA9 , "HANGUL SYLLABLE MIEUM WA CIEUC" },
	{ 0x4DAA , "HANGUL SYLLABLE MIEUM WA CHIEUCH" },
	{ 0x4DAB , "HANGUL SYLLABLE MIEUM WA KHIEUKH" },
	{ 0x4DAC , "HANGUL SYLLABLE MIEUM WA THIEUTH" },
	{ 0x4DAD , "HANGUL SYLLABLE MIEUM WA PHIEUPH" },
	{ 0x4DAE , "HANGUL SYLLABLE MIEUM WA HIEUH" },
	{ 0x4DAF , "HANGUL SYLLABLE MIEUM WAE KIYEOK" },
	{ 0x4DB0 , "HANGUL SYLLABLE MIEUM WAE SSANGKIYEOK" },
	{ 0x4DB1 , "HANGUL SYLLABLE MIEUM WAE KIYEOK-SIOS" },
	{ 0x4DB2 , "HANGUL SYLLABLE MIEUM WAE NIEUN-CIEUC" },
	{ 0x4DB3 , "HANGUL SYLLABLE MIEUM WAE NIEUN-HIEUH" },
	{ 0x4DB4 , "HANGUL SYLLABLE MIEUM WAE TIKEUT" },
	{ 0x4DB5 , "HANGUL SYLLABLE MIEUM WAE RIEUL-KIYEOK" },
	{ 0x4DB6 , "HANGUL SYLLABLE MIEUM WAE RIEUL-MIEUM" },
	{ 0x4DB7 , "HANGUL SYLLABLE MIEUM WAE RIEUL-PIEUP" },
	{ 0x4DB8 , "HANGUL SYLLABLE MIEUM WAE RIEUL-SIOS" },
	{ 0x4DB9 , "HANGUL SYLLABLE MIEUM WAE RIEUL-THIEUTH" },
	{ 0x4DBA , "HANGUL SYLLABLE MIEUM WAE RIEUL-PHIEUPH" },
	{ 0x4DBB , "HANGUL SYLLABLE MIEUM WAE RIEUL-HIEUH" },
	{ 0x4DBC , "HANGUL SYLLABLE MIEUM WAE MIEUM" },
	{ 0x4DBD , "HANGUL SYLLABLE MIEUM WAE PIEUP-SIOS" },
	{ 0x4DBE , "HANGUL SYLLABLE MIEUM WAE SIOS" },
	{ 0x4DBF , "HANGUL SYLLABLE MIEUM WAE IEUNG" },
	{ 0x4DC0 , "HANGUL SYLLABLE MIEUM WAE CIEUC" },
	{ 0x4DC1 , "HANGUL SYLLABLE MIEUM WAE CHIEUCH" },
	{ 0x4DC2 , "HANGUL SYLLABLE MIEUM WAE KHIEUKH" },
	{ 0x4DC3 , "HANGUL SYLLABLE MIEUM WAE THIEUTH" },
	{ 0x4DC4 , "HANGUL SYLLABLE MIEUM WAE PHIEUPH" },
	{ 0x4DC5 , "HANGUL SYLLABLE MIEUM WAE HIEUH" },
	{ 0x4DC6 , "HANGUL SYLLABLE MIEUM OE SSANGKIYEOK" },
	{ 0x4DC7 , "HANGUL SYLLABLE MIEUM OE KIYEOK-SIOS" },
	{ 0x4DC8 , "HANGUL SYLLABLE MIEUM OE NIEUN-CIEUC" },
	{ 0x4DC9 , "HANGUL SYLLABLE MIEUM OE NIEUN-HIEUH" },
	{ 0x4DCA , "HANGUL SYLLABLE MIEUM OE RIEUL-KIYEOK" },
	{ 0x4DCB , "HANGUL SYLLABLE MIEUM OE RIEUL-MIEUM" },
	{ 0x4DCC , "HANGUL SYLLABLE MIEUM OE RIEUL-PIEUP" },
	{ 0x4DCD , "HANGUL SYLLABLE MIEUM OE RIEUL-SIOS" },
	{ 0x4DCE , "HANGUL SYLLABLE MIEUM OE RIEUL-THIEUTH" },
	{ 0x4DCF , "HANGUL SYLLABLE MIEUM OE RIEUL-PHIEUPH" },
	{ 0x4DD0 , "HANGUL SYLLABLE MIEUM OE RIEUL-HIEUH" },
	{ 0x4DD1 , "HANGUL SYLLABLE MIEUM OE PIEUP-SIOS" },
	{ 0x4DD2 , "HANGUL SYLLABLE MIEUM OE SSANGSIOS" },
	{ 0x4DD3 , "HANGUL SYLLABLE MIEUM OE CIEUC" },
	{ 0x4DD4 , "HANGUL SYLLABLE MIEUM OE CHIEUCH" },
	{ 0x4DD5 , "HANGUL SYLLABLE MIEUM OE KHIEUKH" },
	{ 0x4DD6 , "HANGUL SYLLABLE MIEUM OE THIEUTH" },
	{ 0x4DD7 , "HANGUL SYLLABLE MIEUM OE PHIEUPH" },
	{ 0x4DD8 , "HANGUL SYLLABLE MIEUM OE HIEUH" },
	{ 0x4DD9 , "HANGUL SYLLABLE MIEUM YO SSANGKIYEOK" },
	{ 0x4DDA , "HANGUL SYLLABLE MIEUM YO KIYEOK-SIOS" },
	{ 0x4DDB , "HANGUL SYLLABLE MIEUM YO NIEUN-CIEUC" },
	{ 0x4DDC , "HANGUL SYLLABLE MIEUM YO NIEUN-HIEUH" },
	{ 0x4DDD , "HANGUL SYLLABLE MIEUM YO RIEUL-KIYEOK" },
	{ 0x4DDE , "HANGUL SYLLABLE MIEUM YO RIEUL-MIEUM" },
	{ 0x4DDF , "HANGUL SYLLABLE MIEUM YO RIEUL-PIEUP" },
	{ 0x4DE0 , "HANGUL SYLLABLE MIEUM YO RIEUL-SIOS" },
	{ 0x4DE1 , "HANGUL SYLLABLE MIEUM YO RIEUL-THIEUTH" },
	{ 0x4DE2 , "HANGUL SYLLABLE MIEUM YO RIEUL-PHIEUPH" },
	{ 0x4DE3 , "HANGUL SYLLABLE MIEUM YO RIEUL-HIEUH" },
	{ 0x4DE4 , "HANGUL SYLLABLE MIEUM YO PIEUP-SIOS" },
	{ 0x4DE5 , "HANGUL SYLLABLE MIEUM YO SSANGSIOS" },
	{ 0x4DE6 , "HANGUL SYLLABLE MIEUM YO CIEUC" },
	{ 0x4DE7 , "HANGUL SYLLABLE MIEUM YO CHIEUCH" },
	{ 0x4DE8 , "HANGUL SYLLABLE MIEUM YO KHIEUKH" },
	{ 0x4DE9 , "HANGUL SYLLABLE MIEUM YO THIEUTH" },
	{ 0x4DEA , "HANGUL SYLLABLE MIEUM YO PHIEUPH" },
	{ 0x4DEB , "HANGUL SYLLABLE MIEUM YO HIEUH" },
	{ 0x4DEC , "HANGUL SYLLABLE MIEUM U KIYEOK-SIOS" },
	{ 0x4DED , "HANGUL SYLLABLE MIEUM U NIEUN-CIEUC" },
	{ 0x4DEE , "HANGUL SYLLABLE MIEUM U RIEUL-PIEUP" },
	{ 0x4DEF , "HANGUL SYLLABLE MIEUM U RIEUL-THIEUTH" },
	{ 0x4DF0 , "HANGUL SYLLABLE MIEUM U RIEUL-PHIEUPH" },
	{ 0x4DF1 , "HANGUL SYLLABLE MIEUM U RIEUL-HIEUH" },
	{ 0x4DF2 , "HANGUL SYLLABLE MIEUM U PIEUP-SIOS" },
	{ 0x4DF3 , "HANGUL SYLLABLE MIEUM U SSANGSIOS" },
	{ 0x4DF4 , "HANGUL SYLLABLE MIEUM U CIEUC" },
	{ 0x4DF5 , "HANGUL SYLLABLE MIEUM U KHIEUKH" },
	{ 0x4DF6 , "HANGUL SYLLABLE MIEUM U PHIEUPH" },
	{ 0x4DF7 , "HANGUL SYLLABLE MIEUM WEO SSANGKIYEOK" },
	{ 0x4DF8 , "HANGUL SYLLABLE MIEUM WEO KIYEOK-SIOS" },
	{ 0x4DF9 , "HANGUL SYLLABLE MIEUM WEO NIEUN-CIEUC" },
	{ 0x4DFA , "HANGUL SYLLABLE MIEUM WEO NIEUN-HIEUH" },
	{ 0x4DFB , "HANGUL SYLLABLE MIEUM WEO RIEUL-KIYEOK" },
	{ 0x4DFC , "HANGUL SYLLABLE MIEUM WEO RIEUL-MIEUM" },
	{ 0x4DFD , "HANGUL SYLLABLE MIEUM WEO RIEUL-PIEUP" },
	{ 0x4DFE , "HANGUL SYLLABLE MIEUM WEO RIEUL-SIOS" },
	{ 0x4DFF , "HANGUL SYLLABLE MIEUM WEO RIEUL-THIEUTH" },
	{ 0x4E00 , "<CJK IDEOGRAPH REPRESENTATIVE>" },
#endif
	{ 0xFB00 , "LATIN SMALL LIGATURE FF" },
	{ 0xFB01 , "LATIN SMALL LIGATURE FI" },
	{ 0xFB02 , "LATIN SMALL LIGATURE FL" },
	{ 0xFB03 , "LATIN SMALL LIGATURE FFI" },
	{ 0xFB04 , "LATIN SMALL LIGATURE FFL" },
	{ 0xFB05 , "LATIN SMALL LIGATURE LONG S T" },
	{ 0xFB06 , "LATIN SMALL LIGATURE ST" },
#if 0
	{ 0xFB13 , "ARMENIAN SMALL LIGATURE MEN NOW" },
	{ 0xFB14 , "ARMENIAN SMALL LIGATURE MEN ECH" },
	{ 0xFB15 , "ARMENIAN SMALL LIGATURE MEN INI" },
	{ 0xFB16 , "ARMENIAN SMALL LIGATURE VEW NOW" },
	{ 0xFB17 , "ARMENIAN SMALL LIGATURE MEN XEH" },
	{ 0xFB1E , "HEBREW POINT JUDEO-SPANISH VARIKA" },
	{ 0xFB1F , "HEBREW LIGATURE YIDDISH YOD YOD PATAH" },
	{ 0xFB20 , "HEBREW LETTER ALTERNATIVE AYIN" },
	{ 0xFB21 , "HEBREW LETTER WIDE ALEF" },
	{ 0xFB22 , "HEBREW LETTER WIDE DALET" },
	{ 0xFB23 , "HEBREW LETTER WIDE HE" },
	{ 0xFB24 , "HEBREW LETTER WIDE KAF" },
	{ 0xFB25 , "HEBREW LETTER WIDE LAMED" },
	{ 0xFB26 , "HEBREW LETTER WIDE FINAL MEM" },
	{ 0xFB27 , "HEBREW LETTER WIDE RESH" },
	{ 0xFB28 , "HEBREW LETTER WIDE TAV" },
	{ 0xFB29 , "HEBREW LETTER ALTERNATIVE PLUS SIGN" },
	{ 0xFB2A , "HEBREW LETTER SHIN WITH SHIN DOT" },
	{ 0xFB2B , "HEBREW LETTER SHIN WITH SIN DOT" },
	{ 0xFB2C , "HEBREW LETTER SHIN WITH DAGESH AND SHIN DOT" },
	{ 0xFB2D , "HEBREW LETTER SHIN WITH DAGESH AND SIN DOT" },
	{ 0xFB2E , "HEBREW LETTER ALEF WITH PATAH" },
	{ 0xFB2F , "HEBREW LETTER ALEF WITH QAMATS" },
	{ 0xFB30 , "HEBREW LETTER ALEF WITH MAPIQ" },
	{ 0xFB31 , "HEBREW LETTER BET WITH DAGESH" },
	{ 0xFB32 , "HEBREW LETTER GIMEL WITH DAGESH" },
	{ 0xFB33 , "HEBREW LETTER DALET WITH DAGESH" },
	{ 0xFB34 , "HEBREW LETTER HE WITH MAPIQ" },
	{ 0xFB35 , "HEBREW LETTER VAV WITH DAGESH" },
	{ 0xFB36 , "HEBREW LETTER ZAYIN WITH DAGESH" },
	{ 0xFB38 , "HEBREW LETTER TET WITH DAGESH" },
	{ 0xFB39 , "HEBREW LETTER YOD WITH DAGESH" },
	{ 0xFB3A , "HEBREW LETTER FINAL KAF WITH DAGESH" },
	{ 0xFB3B , "HEBREW LETTER KAF WITH DAGESH" },
	{ 0xFB3C , "HEBREW LETTER LAMED WITH DAGESH" },
	{ 0xFB3E , "HEBREW LETTER MEM WITH DAGESH" },
	{ 0xFB40 , "HEBREW LETTER NUN WITH DAGESH" },
	{ 0xFB41 , "HEBREW LETTER SAMEKH WITH DAGESH" },
	{ 0xFB43 , "HEBREW LETTER FINAL PE WITH DAGESH" },
	{ 0xFB44 , "HEBREW LETTER PE WITH DAGESH" },
	{ 0xFB46 , "HEBREW LETTER TSADI WITH DAGESH" },
	{ 0xFB47 , "HEBREW LETTER QOF WITH DAGESH" },
	{ 0xFB48 , "HEBREW LETTER RESH WITH DAGESH" },
	{ 0xFB49 , "HEBREW LETTER SHIN WITH DAGESH" },
	{ 0xFB4A , "HEBREW LETTER TAV WITH DAGESH" },
	{ 0xFB4B , "HEBREW LETTER VAV WITH HOLAM" },
	{ 0xFB4C , "HEBREW LETTER BET WITH RAFE" },
	{ 0xFB4D , "HEBREW LETTER KAF WITH RAFE" },
	{ 0xFB4E , "HEBREW LETTER PE WITH RAFE" },
	{ 0xFB4F , "HEBREW LIGATURE ALEF LAMED" },
	{ 0xFB50 , "ARABIC LETTER ALEF WASLA ISOLATED FORM" },
	{ 0xFB51 , "ARABIC LETTER ALEF WASLA FINAL FORM" },
	{ 0xFB52 , "ARABIC LETTER BEEH ISOLATED FORM" },
	{ 0xFB53 , "ARABIC LETTER BEEH FINAL FORM" },
	{ 0xFB54 , "ARABIC LETTER BEEH INITIAL FORM" },
	{ 0xFB55 , "ARABIC LETTER BEEH MEDIAL FORM" },
	{ 0xFB56 , "ARABIC LETTER PEH ISOLATED FORM" },
	{ 0xFB57 , "ARABIC LETTER PEH FINAL FORM" },
	{ 0xFB58 , "ARABIC LETTER PEH INITIAL FORM" },
	{ 0xFB59 , "ARABIC LETTER PEH MEDIAL FORM" },
	{ 0xFB5A , "ARABIC LETTER BEHEH ISOLATED FORM" },
	{ 0xFB5B , "ARABIC LETTER BEHEH FINAL FORM" },
	{ 0xFB5C , "ARABIC LETTER BEHEH INITIAL FORM" },
	{ 0xFB5D , "ARABIC LETTER BEHEH MEDIAL FORM" },
	{ 0xFB5E , "ARABIC LETTER TTEHEH ISOLATED FORM" },
	{ 0xFB5F , "ARABIC LETTER TTEHEH FINAL FORM" },
	{ 0xFB60 , "ARABIC LETTER TTEHEH INITIAL FORM" },
	{ 0xFB61 , "ARABIC LETTER TTEHEH MEDIAL FORM" },
	{ 0xFB62 , "ARABIC LETTER TEHEH ISOLATED FORM" },
	{ 0xFB63 , "ARABIC LETTER TEHEH FINAL FORM" },
	{ 0xFB64 , "ARABIC LETTER TEHEH INITIAL FORM" },
	{ 0xFB65 , "ARABIC LETTER TEHEH MEDIAL FORM" },
	{ 0xFB66 , "ARABIC LETTER TTEH ISOLATED FORM" },
	{ 0xFB67 , "ARABIC LETTER TTEH FINAL FORM" },
	{ 0xFB68 , "ARABIC LETTER TTEH INITIAL FORM" },
	{ 0xFB69 , "ARABIC LETTER TTEH MEDIAL FORM" },
	{ 0xFB6A , "ARABIC LETTER VEH ISOLATED FORM" },
	{ 0xFB6B , "ARABIC LETTER VEH FINAL FORM" },
	{ 0xFB6C , "ARABIC LETTER VEH INITIAL FORM" },
	{ 0xFB6D , "ARABIC LETTER VEH MEDIAL FORM" },
	{ 0xFB6E , "ARABIC LETTER PEHEH ISOLATED FORM" },
	{ 0xFB6F , "ARABIC LETTER PEHEH FINAL FORM" },
	{ 0xFB70 , "ARABIC LETTER PEHEH INITIAL FORM" },
	{ 0xFB71 , "ARABIC LETTER PEHEH MEDIAL FORM" },
	{ 0xFB72 , "ARABIC LETTER DYEH ISOLATED FORM" },
	{ 0xFB73 , "ARABIC LETTER DYEH FINAL FORM" },
	{ 0xFB74 , "ARABIC LETTER DYEH INITIAL FORM" },
	{ 0xFB75 , "ARABIC LETTER DYEH MEDIAL FORM" },
	{ 0xFB76 , "ARABIC LETTER NYEH ISOLATED FORM" },
	{ 0xFB77 , "ARABIC LETTER NYEH FINAL FORM" },
	{ 0xFB78 , "ARABIC LETTER NYEH INITIAL FORM" },
	{ 0xFB79 , "ARABIC LETTER NYEH MEDIAL FORM" },
	{ 0xFB7A , "ARABIC LETTER TCHEH ISOLATED FORM" },
	{ 0xFB7B , "ARABIC LETTER TCHEH FINAL FORM" },
	{ 0xFB7C , "ARABIC LETTER TCHEH INITIAL FORM" },
	{ 0xFB7D , "ARABIC LETTER TCHEH MEDIAL FORM" },
	{ 0xFB7E , "ARABIC LETTER TCHEHEH ISOLATED FORM" },
	{ 0xFB7F , "ARABIC LETTER TCHEHEH FINAL FORM" },
	{ 0xFB80 , "ARABIC LETTER TCHEHEH INITIAL FORM" },
	{ 0xFB81 , "ARABIC LETTER TCHEHEH MEDIAL FORM" },
	{ 0xFB82 , "ARABIC LETTER DDAHAL ISOLATED FORM" },
	{ 0xFB83 , "ARABIC LETTER DDAHAL FINAL FORM" },
	{ 0xFB84 , "ARABIC LETTER DAHAL ISOLATED FORM" },
	{ 0xFB85 , "ARABIC LETTER DAHAL FINAL FORM" },
	{ 0xFB86 , "ARABIC LETTER DUL ISOLATED FORM" },
	{ 0xFB87 , "ARABIC LETTER DUL FINAL FORM" },
	{ 0xFB88 , "ARABIC LETTER DDAL ISOLATED FORM" },
	{ 0xFB89 , "ARABIC LETTER DDAL FINAL FORM" },
	{ 0xFB8A , "ARABIC LETTER JEH ISOLATED FORM" },
	{ 0xFB8B , "ARABIC LETTER JEH FINAL FORM" },
	{ 0xFB8C , "ARABIC LETTER RREH ISOLATED FORM" },
	{ 0xFB8D , "ARABIC LETTER RREH FINAL FORM" },
	{ 0xFB8E , "ARABIC LETTER KEHEH ISOLATED FORM" },
	{ 0xFB8F , "ARABIC LETTER KEHEH FINAL FORM" },
	{ 0xFB90 , "ARABIC LETTER KEHEH INITIAL FORM" },
	{ 0xFB91 , "ARABIC LETTER KEHEH MEDIAL FORM" },
	{ 0xFB92 , "ARABIC LETTER GAF ISOLATED FORM" },
	{ 0xFB93 , "ARABIC LETTER GAF FINAL FORM" },
	{ 0xFB94 , "ARABIC LETTER GAF INITIAL FORM" },
	{ 0xFB95 , "ARABIC LETTER GAF MEDIAL FORM" },
	{ 0xFB96 , "ARABIC LETTER GUEH ISOLATED FORM" },
	{ 0xFB97 , "ARABIC LETTER GUEH FINAL FORM" },
	{ 0xFB98 , "ARABIC LETTER GUEH INITIAL FORM" },
	{ 0xFB99 , "ARABIC LETTER GUEH MEDIAL FORM" },
	{ 0xFB9A , "ARABIC LETTER NGOEH ISOLATED FORM" },
	{ 0xFB9B , "ARABIC LETTER NGOEH FINAL FORM" },
	{ 0xFB9C , "ARABIC LETTER NGOEH INITIAL FORM" },
	{ 0xFB9D , "ARABIC LETTER NGOEH MEDIAL FORM" },
	{ 0xFB9E , "ARABIC LETTER NOON GHUNNA ISOLATED FORM" },
	{ 0xFB9F , "ARABIC LETTER NOON GHUNNA FINAL FORM" },
	{ 0xFBA0 , "ARABIC LETTER RNOON ISOLATED FORM" },
	{ 0xFBA1 , "ARABIC LETTER RNOON FINAL FORM" },
	{ 0xFBA2 , "ARABIC LETTER RNOON INITIAL FORM" },
	{ 0xFBA3 , "ARABIC LETTER RNOON MEDIAL FORM" },
	{ 0xFBA4 , "ARABIC LETTER HEH WITH YEH ABOVE ISOLATED FORM" },
	{ 0xFBA5 , "ARABIC LETTER HEH WITH YEH ABOVE FINAL FORM" },
	{ 0xFBA6 , "ARABIC LETTER HEH GOAL ISOLATED FORM" },
	{ 0xFBA7 , "ARABIC LETTER HEH GOAL FINAL FORM" },
	{ 0xFBA8 , "ARABIC LETTER HEH GOAL INITIAL FORM" },
	{ 0xFBA9 , "ARABIC LETTER HEH GOAL MEDIAL FORM" },
	{ 0xFBAA , "ARABIC LETTER HEH DOACHASHMEE ISOLATED FORM" },
	{ 0xFBAB , "ARABIC LETTER HEH DOACHASHMEE FINAL FORM" },
	{ 0xFBAC , "ARABIC LETTER HEH DOACHASHMEE INITIAL FORM" },
	{ 0xFBAD , "ARABIC LETTER HEH DOACHASHMEE MEDIAL FORM" },
	{ 0xFBAE , "ARABIC LETTER YEH BARREE ISOLATED FORM" },
	{ 0xFBAF , "ARABIC LETTER YEH BARREE FINAL FORM" },
	{ 0xFBB0 , "ARABIC LETTER YEH BARREE WITH HAMZA ABOVE ISOLATED FORM" },
	{ 0xFBB1 , "ARABIC LETTER YEH BARREE WITH HAMZA ABOVE FINAL FORM" },
	{ 0xFBD3 , "ARABIC LETTER NG ISOLATED FORM" },
	{ 0xFBD4 , "ARABIC LETTER NG FINAL FORM" },
	{ 0xFBD5 , "ARABIC LETTER NG INITIAL FORM" },
	{ 0xFBD6 , "ARABIC LETTER NG MEDIAL FORM" },
	{ 0xFBD7 , "ARABIC LETTER U ISOLATED FORM" },
	{ 0xFBD8 , "ARABIC LETTER U FINAL FORM" },
	{ 0xFBD9 , "ARABIC LETTER OE ISOLATED FORM" },
	{ 0xFBDA , "ARABIC LETTER OE FINAL FORM" },
	{ 0xFBDB , "ARABIC LETTER YU ISOLATED FORM" },
	{ 0xFBDC , "ARABIC LETTER YU FINAL FORM" },
	{ 0xFBDD , "ARABIC LETTER U WITH HAMZA ABOVE ISOLATED FORM" },
	{ 0xFBDE , "ARABIC LETTER VE ISOLATED FORM" },
	{ 0xFBDF , "ARABIC LETTER VE FINAL FORM" },
	{ 0xFBE0 , "ARABIC LETTER KIRGHIZ OE ISOLATED FORM" },
	{ 0xFBE1 , "ARABIC LETTER KIRGHIZ OE FINAL FORM" },
	{ 0xFBE2 , "ARABIC LETTER KIRGHIZ YU ISOLATED FORM" },
	{ 0xFBE3 , "ARABIC LETTER KIRGHIZ YU FINAL FORM" },
	{ 0xFBE4 , "ARABIC LETTER E ISOLATED FORM" },
	{ 0xFBE5 , "ARABIC LETTER E FINAL FORM" },
	{ 0xFBE6 , "ARABIC LETTER E INITIAL FORM" },
	{ 0xFBE7 , "ARABIC LETTER E MEDIAL FORM" },
	{ 0xFBE8 , "ARABIC LETTER UIGHUR KAZAKH KIRGHIZ ALEF MAKSURA INITIAL FORM" },
	{ 0xFBE9 , "ARABIC LETTER UIGHUR KAZAKH KIRGHIZ ALEF MAKSURA MEDIAL FORM" },
	{ 0xFBEA , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH ALEF ISOLATED FORM" },
	{ 0xFBEB , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH ALEF FINAL FORM" },
	{ 0xFBEC , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH AE ISOLATED FORM" },
	{ 0xFBED , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH AE FINAL FORM" },
	{ 0xFBEE , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH WAW ISOLATED FORM" },
	{ 0xFBEF , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH WAW FINAL FORM" },
	{ 0xFBF0 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH U ISOLATED FORM" },
	{ 0xFBF1 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH U FINAL FORM" },
	{ 0xFBF2 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH OE ISOLATED FORM" },
	{ 0xFBF3 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH OE FINAL FORM" },
	{ 0xFBF4 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH YU ISOLATED FORM" },
	{ 0xFBF5 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH YU FINAL FORM" },
	{ 0xFBF6 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH E ISOLATED FORM" },
	{ 0xFBF7 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH E FINAL FORM" },
	{ 0xFBF8 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH E INITIAL FORM" },
	{ 0xFBF9 , "ARABIC LIGATURE UIGHUR KIRGHIZ YEH WITH HAMZA ABOVE WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFBFA , "ARABIC LIGATURE UIGHUR KIRGHIZ YEH WITH HAMZA ABOVE WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFBFB , "ARABIC LIGATURE UIGHUR KIRGHIZ YEH WITH HAMZA ABOVE WITH ALEF MAKSURA INITIAL FORM" },
	{ 0xFBFC , "ARABIC LETTER FARSI YEH ISOLATED FORM" },
	{ 0xFBFD , "ARABIC LETTER FARSI YEH FINAL FORM" },
	{ 0xFBFE , "ARABIC LETTER FARSI YEH INITIAL FORM" },
	{ 0xFBFF , "ARABIC LETTER FARSI YEH MEDIAL FORM" },
	{ 0xFC00 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH JEEM ISOLATED FORM" },
	{ 0xFC01 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH HAH ISOLATED FORM" },
	{ 0xFC02 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH MEEM ISOLATED FORM" },
	{ 0xFC03 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFC04 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH YEH ISOLATED FORM" },
	{ 0xFC05 , "ARABIC LIGATURE BEH WITH JEEM ISOLATED FORM" },
	{ 0xFC06 , "ARABIC LIGATURE BEH WITH HAH ISOLATED FORM" },
	{ 0xFC07 , "ARABIC LIGATURE BEH WITH KHAH ISOLATED FORM" },
	{ 0xFC08 , "ARABIC LIGATURE BEH WITH MEEM ISOLATED FORM" },
	{ 0xFC09 , "ARABIC LIGATURE BEH WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFC0A , "ARABIC LIGATURE BEH WITH YEH ISOLATED FORM" },
	{ 0xFC0B , "ARABIC LIGATURE TEH WITH JEEM ISOLATED FORM" },
	{ 0xFC0C , "ARABIC LIGATURE TEH WITH HAH ISOLATED FORM" },
	{ 0xFC0D , "ARABIC LIGATURE TEH WITH KHAH ISOLATED FORM" },
	{ 0xFC0E , "ARABIC LIGATURE TEH WITH MEEM ISOLATED FORM" },
	{ 0xFC0F , "ARABIC LIGATURE TEH WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFC10 , "ARABIC LIGATURE TEH WITH YEH ISOLATED FORM" },
	{ 0xFC11 , "ARABIC LIGATURE THEH WITH JEEM ISOLATED FORM" },
	{ 0xFC12 , "ARABIC LIGATURE THEH WITH MEEM ISOLATED FORM" },
	{ 0xFC13 , "ARABIC LIGATURE THEH WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFC14 , "ARABIC LIGATURE THEH WITH YEH ISOLATED FORM" },
	{ 0xFC15 , "ARABIC LIGATURE JEEM WITH HAH ISOLATED FORM" },
	{ 0xFC16 , "ARABIC LIGATURE JEEM WITH MEEM ISOLATED FORM" },
	{ 0xFC17 , "ARABIC LIGATURE HAH WITH JEEM ISOLATED FORM" },
	{ 0xFC18 , "ARABIC LIGATURE HAH WITH MEEM ISOLATED FORM" },
	{ 0xFC19 , "ARABIC LIGATURE KHAH WITH JEEM ISOLATED FORM" },
	{ 0xFC1A , "ARABIC LIGATURE KHAH WITH HAH ISOLATED FORM" },
	{ 0xFC1B , "ARABIC LIGATURE KHAH WITH MEEM ISOLATED FORM" },
	{ 0xFC1C , "ARABIC LIGATURE SEEN WITH JEEM ISOLATED FORM" },
	{ 0xFC1D , "ARABIC LIGATURE SEEN WITH HAH ISOLATED FORM" },
	{ 0xFC1E , "ARABIC LIGATURE SEEN WITH KHAH ISOLATED FORM" },
	{ 0xFC1F , "ARABIC LIGATURE SEEN WITH MEEM ISOLATED FORM" },
	{ 0xFC20 , "ARABIC LIGATURE SAD WITH HAH ISOLATED FORM" },
	{ 0xFC21 , "ARABIC LIGATURE SAD WITH MEEM ISOLATED FORM" },
	{ 0xFC22 , "ARABIC LIGATURE DAD WITH JEEM ISOLATED FORM" },
	{ 0xFC23 , "ARABIC LIGATURE DAD WITH HAH ISOLATED FORM" },
	{ 0xFC24 , "ARABIC LIGATURE DAD WITH KHAH ISOLATED FORM" },
	{ 0xFC25 , "ARABIC LIGATURE DAD WITH MEEM ISOLATED FORM" },
	{ 0xFC26 , "ARABIC LIGATURE TAH WITH HAH ISOLATED FORM" },
	{ 0xFC27 , "ARABIC LIGATURE TAH WITH MEEM ISOLATED FORM" },
	{ 0xFC28 , "ARABIC LIGATURE ZAH WITH MEEM ISOLATED FORM" },
	{ 0xFC29 , "ARABIC LIGATURE AIN WITH JEEM ISOLATED FORM" },
	{ 0xFC2A , "ARABIC LIGATURE AIN WITH MEEM ISOLATED FORM" },
	{ 0xFC2B , "ARABIC LIGATURE GHAIN WITH JEEM ISOLATED FORM" },
	{ 0xFC2C , "ARABIC LIGATURE GHAIN WITH MEEM ISOLATED FORM" },
	{ 0xFC2D , "ARABIC LIGATURE FEH WITH JEEM ISOLATED FORM" },
	{ 0xFC2E , "ARABIC LIGATURE FEH WITH HAH ISOLATED FORM" },
	{ 0xFC2F , "ARABIC LIGATURE FEH WITH KHAH ISOLATED FORM" },
	{ 0xFC30 , "ARABIC LIGATURE FEH WITH MEEM ISOLATED FORM" },
	{ 0xFC31 , "ARABIC LIGATURE FEH WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFC32 , "ARABIC LIGATURE FEH WITH YEH ISOLATED FORM" },
	{ 0xFC33 , "ARABIC LIGATURE QAF WITH HAH ISOLATED FORM" },
	{ 0xFC34 , "ARABIC LIGATURE QAF WITH MEEM ISOLATED FORM" },
	{ 0xFC35 , "ARABIC LIGATURE QAF WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFC36 , "ARABIC LIGATURE QAF WITH YEH ISOLATED FORM" },
	{ 0xFC37 , "ARABIC LIGATURE KAF WITH ALEF ISOLATED FORM" },
	{ 0xFC38 , "ARABIC LIGATURE KAF WITH JEEM ISOLATED FORM" },
	{ 0xFC39 , "ARABIC LIGATURE KAF WITH HAH ISOLATED FORM" },
	{ 0xFC3A , "ARABIC LIGATURE KAF WITH KHAH ISOLATED FORM" },
	{ 0xFC3B , "ARABIC LIGATURE KAF WITH LAM ISOLATED FORM" },
	{ 0xFC3C , "ARABIC LIGATURE KAF WITH MEEM ISOLATED FORM" },
	{ 0xFC3D , "ARABIC LIGATURE KAF WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFC3E , "ARABIC LIGATURE KAF WITH YEH ISOLATED FORM" },
	{ 0xFC3F , "ARABIC LIGATURE LAM WITH JEEM ISOLATED FORM" },
	{ 0xFC40 , "ARABIC LIGATURE LAM WITH HAH ISOLATED FORM" },
	{ 0xFC41 , "ARABIC LIGATURE LAM WITH KHAH ISOLATED FORM" },
	{ 0xFC42 , "ARABIC LIGATURE LAM WITH MEEM ISOLATED FORM" },
	{ 0xFC43 , "ARABIC LIGATURE LAM WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFC44 , "ARABIC LIGATURE LAM WITH YEH ISOLATED FORM" },
	{ 0xFC45 , "ARABIC LIGATURE MEEM WITH JEEM ISOLATED FORM" },
	{ 0xFC46 , "ARABIC LIGATURE MEEM WITH HAH ISOLATED FORM" },
	{ 0xFC47 , "ARABIC LIGATURE MEEM WITH KHAH ISOLATED FORM" },
	{ 0xFC48 , "ARABIC LIGATURE MEEM WITH MEEM ISOLATED FORM" },
	{ 0xFC49 , "ARABIC LIGATURE MEEM WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFC4A , "ARABIC LIGATURE MEEM WITH YEH ISOLATED FORM" },
	{ 0xFC4B , "ARABIC LIGATURE NOON WITH JEEM ISOLATED FORM" },
	{ 0xFC4C , "ARABIC LIGATURE NOON WITH HAH ISOLATED FORM" },
	{ 0xFC4D , "ARABIC LIGATURE NOON WITH KHAH ISOLATED FORM" },
	{ 0xFC4E , "ARABIC LIGATURE NOON WITH MEEM ISOLATED FORM" },
	{ 0xFC4F , "ARABIC LIGATURE NOON WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFC50 , "ARABIC LIGATURE NOON WITH YEH ISOLATED FORM" },
	{ 0xFC51 , "ARABIC LIGATURE HEH WITH JEEM ISOLATED FORM" },
	{ 0xFC52 , "ARABIC LIGATURE HEH WITH MEEM ISOLATED FORM" },
	{ 0xFC53 , "ARABIC LIGATURE HEH WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFC54 , "ARABIC LIGATURE HEH WITH YEH ISOLATED FORM" },
	{ 0xFC55 , "ARABIC LIGATURE YEH WITH JEEM ISOLATED FORM" },
	{ 0xFC56 , "ARABIC LIGATURE YEH WITH HAH ISOLATED FORM" },
	{ 0xFC57 , "ARABIC LIGATURE YEH WITH KHAH ISOLATED FORM" },
	{ 0xFC58 , "ARABIC LIGATURE YEH WITH MEEM ISOLATED FORM" },
	{ 0xFC59 , "ARABIC LIGATURE YEH WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFC5A , "ARABIC LIGATURE YEH WITH YEH ISOLATED FORM" },
	{ 0xFC5B , "ARABIC LIGATURE THAL WITH SUPERSCRIPT ALEF ISOLATED FORM" },
	{ 0xFC5C , "ARABIC LIGATURE REH WITH SUPERSCRIPT ALEF ISOLATED FORM" },
	{ 0xFC5D , "ARABIC LIGATURE ALEF MAKSURA WITH SUPERSCRIPT ALEF ISOLATED FORM" },
	{ 0xFC5E , "ARABIC LIGATURE SHADDA WITH DAMMATAN ISOLATED FORM" },
	{ 0xFC5F , "ARABIC LIGATURE SHADDA WITH KASRATAN ISOLATED FORM" },
	{ 0xFC60 , "ARABIC LIGATURE SHADDA WITH FATHA ISOLATED FORM" },
	{ 0xFC61 , "ARABIC LIGATURE SHADDA WITH DAMMA ISOLATED FORM" },
	{ 0xFC62 , "ARABIC LIGATURE SHADDA WITH KASRA ISOLATED FORM" },
	{ 0xFC63 , "ARABIC LIGATURE SHADDA WITH SUPERSCRIPT ALEF ISOLATED FORM" },
	{ 0xFC64 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH REH FINAL FORM" },
	{ 0xFC65 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH ZAIN FINAL FORM" },
	{ 0xFC66 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH MEEM FINAL FORM" },
	{ 0xFC67 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH NOON FINAL FORM" },
	{ 0xFC68 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFC69 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH YEH FINAL FORM" },
	{ 0xFC6A , "ARABIC LIGATURE BEH WITH REH FINAL FORM" },
	{ 0xFC6B , "ARABIC LIGATURE BEH WITH ZAIN FINAL FORM" },
	{ 0xFC6C , "ARABIC LIGATURE BEH WITH MEEM FINAL FORM" },
	{ 0xFC6D , "ARABIC LIGATURE BEH WITH NOON FINAL FORM" },
	{ 0xFC6E , "ARABIC LIGATURE BEH WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFC6F , "ARABIC LIGATURE BEH WITH YEH FINAL FORM" },
	{ 0xFC70 , "ARABIC LIGATURE TEH WITH REH FINAL FORM" },
	{ 0xFC71 , "ARABIC LIGATURE TEH WITH ZAIN FINAL FORM" },
	{ 0xFC72 , "ARABIC LIGATURE TEH WITH MEEM FINAL FORM" },
	{ 0xFC73 , "ARABIC LIGATURE TEH WITH NOON FINAL FORM" },
	{ 0xFC74 , "ARABIC LIGATURE TEH WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFC75 , "ARABIC LIGATURE TEH WITH YEH FINAL FORM" },
	{ 0xFC76 , "ARABIC LIGATURE THEH WITH REH FINAL FORM" },
	{ 0xFC77 , "ARABIC LIGATURE THEH WITH ZAIN FINAL FORM" },
	{ 0xFC78 , "ARABIC LIGATURE THEH WITH MEEM FINAL FORM" },
	{ 0xFC79 , "ARABIC LIGATURE THEH WITH NOON FINAL FORM" },
	{ 0xFC7A , "ARABIC LIGATURE THEH WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFC7B , "ARABIC LIGATURE THEH WITH YEH FINAL FORM" },
	{ 0xFC7C , "ARABIC LIGATURE FEH WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFC7D , "ARABIC LIGATURE FEH WITH YEH FINAL FORM" },
	{ 0xFC7E , "ARABIC LIGATURE QAF WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFC7F , "ARABIC LIGATURE QAF WITH YEH FINAL FORM" },
	{ 0xFC80 , "ARABIC LIGATURE KAF WITH ALEF FINAL FORM" },
	{ 0xFC81 , "ARABIC LIGATURE KAF WITH LAM FINAL FORM" },
	{ 0xFC82 , "ARABIC LIGATURE KAF WITH MEEM FINAL FORM" },
	{ 0xFC83 , "ARABIC LIGATURE KAF WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFC84 , "ARABIC LIGATURE KAF WITH YEH FINAL FORM" },
	{ 0xFC85 , "ARABIC LIGATURE LAM WITH MEEM FINAL FORM" },
	{ 0xFC86 , "ARABIC LIGATURE LAM WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFC87 , "ARABIC LIGATURE LAM WITH YEH FINAL FORM" },
	{ 0xFC88 , "ARABIC LIGATURE MEEM WITH ALEF FINAL FORM" },
	{ 0xFC89 , "ARABIC LIGATURE MEEM WITH MEEM FINAL FORM" },
	{ 0xFC8A , "ARABIC LIGATURE NOON WITH REH FINAL FORM" },
	{ 0xFC8B , "ARABIC LIGATURE NOON WITH ZAIN FINAL FORM" },
	{ 0xFC8C , "ARABIC LIGATURE NOON WITH MEEM FINAL FORM" },
	{ 0xFC8D , "ARABIC LIGATURE NOON WITH NOON FINAL FORM" },
	{ 0xFC8E , "ARABIC LIGATURE NOON WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFC8F , "ARABIC LIGATURE NOON WITH YEH FINAL FORM" },
	{ 0xFC90 , "ARABIC LIGATURE ALEF MAKSURA WITH SUPERSCRIPT ALEF FINAL FORM" },
	{ 0xFC91 , "ARABIC LIGATURE YEH WITH REH FINAL FORM" },
	{ 0xFC92 , "ARABIC LIGATURE YEH WITH ZAIN FINAL FORM" },
	{ 0xFC93 , "ARABIC LIGATURE YEH WITH MEEM FINAL FORM" },
	{ 0xFC94 , "ARABIC LIGATURE YEH WITH NOON FINAL FORM" },
	{ 0xFC95 , "ARABIC LIGATURE YEH WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFC96 , "ARABIC LIGATURE YEH WITH YEH FINAL FORM" },
	{ 0xFC97 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH JEEM INITIAL FORM" },
	{ 0xFC98 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH HAH INITIAL FORM" },
	{ 0xFC99 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH KHAH INITIAL FORM" },
	{ 0xFC9A , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH MEEM INITIAL FORM" },
	{ 0xFC9B , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH HEH INITIAL FORM" },
	{ 0xFC9C , "ARABIC LIGATURE BEH WITH JEEM INITIAL FORM" },
	{ 0xFC9D , "ARABIC LIGATURE BEH WITH HAH INITIAL FORM" },
	{ 0xFC9E , "ARABIC LIGATURE BEH WITH KHAH INITIAL FORM" },
	{ 0xFC9F , "ARABIC LIGATURE BEH WITH MEEM INITIAL FORM" },
	{ 0xFCA0 , "ARABIC LIGATURE BEH WITH HEH INITIAL FORM" },
	{ 0xFCA1 , "ARABIC LIGATURE TEH WITH JEEM INITIAL FORM" },
	{ 0xFCA2 , "ARABIC LIGATURE TEH WITH HAH INITIAL FORM" },
	{ 0xFCA3 , "ARABIC LIGATURE TEH WITH KHAH INITIAL FORM" },
	{ 0xFCA4 , "ARABIC LIGATURE TEH WITH MEEM INITIAL FORM" },
	{ 0xFCA5 , "ARABIC LIGATURE TEH WITH HEH INITIAL FORM" },
	{ 0xFCA6 , "ARABIC LIGATURE THEH WITH MEEM INITIAL FORM" },
	{ 0xFCA7 , "ARABIC LIGATURE JEEM WITH HAH INITIAL FORM" },
	{ 0xFCA8 , "ARABIC LIGATURE JEEM WITH MEEM INITIAL FORM" },
	{ 0xFCA9 , "ARABIC LIGATURE HAH WITH JEEM INITIAL FORM" },
	{ 0xFCAA , "ARABIC LIGATURE HAH WITH MEEM INITIAL FORM" },
	{ 0xFCAB , "ARABIC LIGATURE KHAH WITH JEEM INITIAL FORM" },
	{ 0xFCAC , "ARABIC LIGATURE KHAH WITH MEEM INITIAL FORM" },
	{ 0xFCAD , "ARABIC LIGATURE SEEN WITH JEEM INITIAL FORM" },
	{ 0xFCAE , "ARABIC LIGATURE SEEN WITH HAH INITIAL FORM" },
	{ 0xFCAF , "ARABIC LIGATURE SEEN WITH KHAH INITIAL FORM" },
	{ 0xFCB0 , "ARABIC LIGATURE SEEN WITH MEEM INITIAL FORM" },
	{ 0xFCB1 , "ARABIC LIGATURE SAD WITH HAH INITIAL FORM" },
	{ 0xFCB2 , "ARABIC LIGATURE SAD WITH KHAH INITIAL FORM" },
	{ 0xFCB3 , "ARABIC LIGATURE SAD WITH MEEM INITIAL FORM" },
	{ 0xFCB4 , "ARABIC LIGATURE DAD WITH JEEM INITIAL FORM" },
	{ 0xFCB5 , "ARABIC LIGATURE DAD WITH HAH INITIAL FORM" },
	{ 0xFCB6 , "ARABIC LIGATURE DAD WITH KHAH INITIAL FORM" },
	{ 0xFCB7 , "ARABIC LIGATURE DAD WITH MEEM INITIAL FORM" },
	{ 0xFCB8 , "ARABIC LIGATURE TAH WITH HAH INITIAL FORM" },
	{ 0xFCB9 , "ARABIC LIGATURE ZAH WITH MEEM INITIAL FORM" },
	{ 0xFCBA , "ARABIC LIGATURE AIN WITH JEEM INITIAL FORM" },
	{ 0xFCBB , "ARABIC LIGATURE AIN WITH MEEM INITIAL FORM" },
	{ 0xFCBC , "ARABIC LIGATURE GHAIN WITH JEEM INITIAL FORM" },
	{ 0xFCBD , "ARABIC LIGATURE GHAIN WITH MEEM INITIAL FORM" },
	{ 0xFCBE , "ARABIC LIGATURE FEH WITH JEEM INITIAL FORM" },
	{ 0xFCBF , "ARABIC LIGATURE FEH WITH HAH INITIAL FORM" },
	{ 0xFCC0 , "ARABIC LIGATURE FEH WITH KHAH INITIAL FORM" },
	{ 0xFCC1 , "ARABIC LIGATURE FEH WITH MEEM INITIAL FORM" },
	{ 0xFCC2 , "ARABIC LIGATURE QAF WITH HAH INITIAL FORM" },
	{ 0xFCC3 , "ARABIC LIGATURE QAF WITH MEEM INITIAL FORM" },
	{ 0xFCC4 , "ARABIC LIGATURE KAF WITH JEEM INITIAL FORM" },
	{ 0xFCC5 , "ARABIC LIGATURE KAF WITH HAH INITIAL FORM" },
	{ 0xFCC6 , "ARABIC LIGATURE KAF WITH KHAH INITIAL FORM" },
	{ 0xFCC7 , "ARABIC LIGATURE KAF WITH LAM INITIAL FORM" },
	{ 0xFCC8 , "ARABIC LIGATURE KAF WITH MEEM INITIAL FORM" },
	{ 0xFCC9 , "ARABIC LIGATURE LAM WITH JEEM INITIAL FORM" },
	{ 0xFCCA , "ARABIC LIGATURE LAM WITH HAH INITIAL FORM" },
	{ 0xFCCB , "ARABIC LIGATURE LAM WITH KHAH INITIAL FORM" },
	{ 0xFCCC , "ARABIC LIGATURE LAM WITH MEEM INITIAL FORM" },
	{ 0xFCCD , "ARABIC LIGATURE LAM WITH HEH INITIAL FORM" },
	{ 0xFCCE , "ARABIC LIGATURE MEEM WITH JEEM INITIAL FORM" },
	{ 0xFCCF , "ARABIC LIGATURE MEEM WITH HAH INITIAL FORM" },
	{ 0xFCD0 , "ARABIC LIGATURE MEEM WITH KHAH INITIAL FORM" },
	{ 0xFCD1 , "ARABIC LIGATURE MEEM WITH MEEM INITIAL FORM" },
	{ 0xFCD2 , "ARABIC LIGATURE NOON WITH JEEM INITIAL FORM" },
	{ 0xFCD3 , "ARABIC LIGATURE NOON WITH HAH INITIAL FORM" },
	{ 0xFCD4 , "ARABIC LIGATURE NOON WITH KHAH INITIAL FORM" },
	{ 0xFCD5 , "ARABIC LIGATURE NOON WITH MEEM INITIAL FORM" },
	{ 0xFCD6 , "ARABIC LIGATURE NOON WITH HEH INITIAL FORM" },
	{ 0xFCD7 , "ARABIC LIGATURE HEH WITH JEEM INITIAL FORM" },
	{ 0xFCD8 , "ARABIC LIGATURE HEH WITH MEEM INITIAL FORM" },
	{ 0xFCD9 , "ARABIC LIGATURE HEH WITH SUPERSCRIPT ALEF INITIAL FORM" },
	{ 0xFCDA , "ARABIC LIGATURE YEH WITH JEEM INITIAL FORM" },
	{ 0xFCDB , "ARABIC LIGATURE YEH WITH HAH INITIAL FORM" },
	{ 0xFCDC , "ARABIC LIGATURE YEH WITH KHAH INITIAL FORM" },
	{ 0xFCDD , "ARABIC LIGATURE YEH WITH MEEM INITIAL FORM" },
	{ 0xFCDE , "ARABIC LIGATURE YEH WITH HEH INITIAL FORM" },
	{ 0xFCDF , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH MEEM MEDIAL FORM" },
	{ 0xFCE0 , "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH HEH MEDIAL FORM" },
	{ 0xFCE1 , "ARABIC LIGATURE BEH WITH MEEM MEDIAL FORM" },
	{ 0xFCE2 , "ARABIC LIGATURE BEH WITH HEH MEDIAL FORM" },
	{ 0xFCE3 , "ARABIC LIGATURE TEH WITH MEEM MEDIAL FORM" },
	{ 0xFCE4 , "ARABIC LIGATURE TEH WITH HEH MEDIAL FORM" },
	{ 0xFCE5 , "ARABIC LIGATURE THEH WITH MEEM MEDIAL FORM" },
	{ 0xFCE6 , "ARABIC LIGATURE THEH WITH HEH MEDIAL FORM" },
	{ 0xFCE7 , "ARABIC LIGATURE SEEN WITH MEEM MEDIAL FORM" },
	{ 0xFCE8 , "ARABIC LIGATURE SEEN WITH HEH MEDIAL FORM" },
	{ 0xFCE9 , "ARABIC LIGATURE SHEEN WITH MEEM MEDIAL FORM" },
	{ 0xFCEA , "ARABIC LIGATURE SHEEN WITH HEH MEDIAL FORM" },
	{ 0xFCEB , "ARABIC LIGATURE KAF WITH LAM MEDIAL FORM" },
	{ 0xFCEC , "ARABIC LIGATURE KAF WITH MEEM MEDIAL FORM" },
	{ 0xFCED , "ARABIC LIGATURE LAM WITH MEEM MEDIAL FORM" },
	{ 0xFCEE , "ARABIC LIGATURE NOON WITH MEEM MEDIAL FORM" },
	{ 0xFCEF , "ARABIC LIGATURE NOON WITH HEH MEDIAL FORM" },
	{ 0xFCF0 , "ARABIC LIGATURE YEH WITH MEEM MEDIAL FORM" },
	{ 0xFCF1 , "ARABIC LIGATURE YEH WITH HEH MEDIAL FORM" },
	{ 0xFCF2 , "ARABIC LIGATURE SHADDA WITH FATHA MEDIAL FORM" },
	{ 0xFCF3 , "ARABIC LIGATURE SHADDA WITH DAMMA MEDIAL FORM" },
	{ 0xFCF4 , "ARABIC LIGATURE SHADDA WITH KASRA MEDIAL FORM" },
	{ 0xFCF5 , "ARABIC LIGATURE TAH WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFCF6 , "ARABIC LIGATURE TAH WITH YEH ISOLATED FORM" },
	{ 0xFCF7 , "ARABIC LIGATURE AIN WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFCF8 , "ARABIC LIGATURE AIN WITH YEH ISOLATED FORM" },
	{ 0xFCF9 , "ARABIC LIGATURE GHAIN WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFCFA , "ARABIC LIGATURE GHAIN WITH YEH ISOLATED FORM" },
	{ 0xFCFB , "ARABIC LIGATURE SEEN WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFCFC , "ARABIC LIGATURE SEEN WITH YEH ISOLATED FORM" },
	{ 0xFCFD , "ARABIC LIGATURE SHEEN WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFCFE , "ARABIC LIGATURE SHEEN WITH YEH ISOLATED FORM" },
	{ 0xFCFF , "ARABIC LIGATURE HAH WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFD00 , "ARABIC LIGATURE HAH WITH YEH ISOLATED FORM" },
	{ 0xFD01 , "ARABIC LIGATURE JEEM WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFD02 , "ARABIC LIGATURE JEEM WITH YEH ISOLATED FORM" },
	{ 0xFD03 , "ARABIC LIGATURE KHAH WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFD04 , "ARABIC LIGATURE KHAH WITH YEH ISOLATED FORM" },
	{ 0xFD05 , "ARABIC LIGATURE SAD WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFD06 , "ARABIC LIGATURE SAD WITH YEH ISOLATED FORM" },
	{ 0xFD07 , "ARABIC LIGATURE DAD WITH ALEF MAKSURA ISOLATED FORM" },
	{ 0xFD08 , "ARABIC LIGATURE DAD WITH YEH ISOLATED FORM" },
	{ 0xFD09 , "ARABIC LIGATURE SHEEN WITH JEEM ISOLATED FORM" },
	{ 0xFD0A , "ARABIC LIGATURE SHEEN WITH HAH ISOLATED FORM" },
	{ 0xFD0B , "ARABIC LIGATURE SHEEN WITH KHAH ISOLATED FORM" },
	{ 0xFD0C , "ARABIC LIGATURE SHEEN WITH MEEM ISOLATED FORM" },
	{ 0xFD0D , "ARABIC LIGATURE SHEEN WITH REH ISOLATED FORM" },
	{ 0xFD0E , "ARABIC LIGATURE SEEN WITH REH ISOLATED FORM" },
	{ 0xFD0F , "ARABIC LIGATURE SAD WITH REH ISOLATED FORM" },
	{ 0xFD10 , "ARABIC LIGATURE DAD WITH REH ISOLATED FORM" },
	{ 0xFD11 , "ARABIC LIGATURE TAH WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD12 , "ARABIC LIGATURE TAH WITH YEH FINAL FORM" },
	{ 0xFD13 , "ARABIC LIGATURE AIN WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD14 , "ARABIC LIGATURE AIN WITH YEH FINAL FORM" },
	{ 0xFD15 , "ARABIC LIGATURE GHAIN WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD16 , "ARABIC LIGATURE GHAIN WITH YEH FINAL FORM" },
	{ 0xFD17 , "ARABIC LIGATURE SEEN WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD18 , "ARABIC LIGATURE SEEN WITH YEH FINAL FORM" },
	{ 0xFD19 , "ARABIC LIGATURE SHEEN WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD1A , "ARABIC LIGATURE SHEEN WITH YEH FINAL FORM" },
	{ 0xFD1B , "ARABIC LIGATURE HAH WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD1C , "ARABIC LIGATURE HAH WITH YEH FINAL FORM" },
	{ 0xFD1D , "ARABIC LIGATURE JEEM WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD1E , "ARABIC LIGATURE JEEM WITH YEH FINAL FORM" },
	{ 0xFD1F , "ARABIC LIGATURE KHAH WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD20 , "ARABIC LIGATURE KHAH WITH YEH FINAL FORM" },
	{ 0xFD21 , "ARABIC LIGATURE SAD WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD22 , "ARABIC LIGATURE SAD WITH YEH FINAL FORM" },
	{ 0xFD23 , "ARABIC LIGATURE DAD WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD24 , "ARABIC LIGATURE DAD WITH YEH FINAL FORM" },
	{ 0xFD25 , "ARABIC LIGATURE SHEEN WITH JEEM FINAL FORM" },
	{ 0xFD26 , "ARABIC LIGATURE SHEEN WITH HAH FINAL FORM" },
	{ 0xFD27 , "ARABIC LIGATURE SHEEN WITH KHAH FINAL FORM" },
	{ 0xFD28 , "ARABIC LIGATURE SHEEN WITH MEEM FINAL FORM" },
	{ 0xFD29 , "ARABIC LIGATURE SHEEN WITH REH FINAL FORM" },
	{ 0xFD2A , "ARABIC LIGATURE SEEN WITH REH FINAL FORM" },
	{ 0xFD2B , "ARABIC LIGATURE SAD WITH REH FINAL FORM" },
	{ 0xFD2C , "ARABIC LIGATURE DAD WITH REH FINAL FORM" },
	{ 0xFD2D , "ARABIC LIGATURE SHEEN WITH JEEM INITIAL FORM" },
	{ 0xFD2E , "ARABIC LIGATURE SHEEN WITH HAH INITIAL FORM" },
	{ 0xFD2F , "ARABIC LIGATURE SHEEN WITH KHAH INITIAL FORM" },
	{ 0xFD30 , "ARABIC LIGATURE SHEEN WITH MEEM INITIAL FORM" },
	{ 0xFD31 , "ARABIC LIGATURE SEEN WITH HEH INITIAL FORM" },
	{ 0xFD32 , "ARABIC LIGATURE SHEEN WITH HEH INITIAL FORM" },
	{ 0xFD33 , "ARABIC LIGATURE TAH WITH MEEM INITIAL FORM" },
	{ 0xFD34 , "ARABIC LIGATURE SEEN WITH JEEM MEDIAL FORM" },
	{ 0xFD35 , "ARABIC LIGATURE SEEN WITH HAH MEDIAL FORM" },
	{ 0xFD36 , "ARABIC LIGATURE SEEN WITH KHAH MEDIAL FORM" },
	{ 0xFD37 , "ARABIC LIGATURE SHEEN WITH JEEM MEDIAL FORM" },
	{ 0xFD38 , "ARABIC LIGATURE SHEEN WITH HAH MEDIAL FORM" },
	{ 0xFD39 , "ARABIC LIGATURE SHEEN WITH KHAH MEDIAL FORM" },
	{ 0xFD3A , "ARABIC LIGATURE TAH WITH MEEM MEDIAL FORM" },
	{ 0xFD3B , "ARABIC LIGATURE ZAH WITH MEEM MEDIAL FORM" },
	{ 0xFD3C , "ARABIC LIGATURE ALEF WITH FATHATAN FINAL FORM" },
	{ 0xFD3D , "ARABIC LIGATURE ALEF WITH FATHATAN ISOLATED FORM" },
	{ 0xFD3E , "ORNATE LEFT PARENTHESIS" },
	{ 0xFD3F , "ORNATE RIGHT PARENTHESIS" },
	{ 0xFD50 , "ARABIC LIGATURE TEH WITH JEEM WITH MEEM INITIAL FORM" },
	{ 0xFD51 , "ARABIC LIGATURE TEH WITH HAH WITH JEEM FINAL FORM" },
	{ 0xFD52 , "ARABIC LIGATURE TEH WITH HAH WITH JEEM INITIAL FORM" },
	{ 0xFD53 , "ARABIC LIGATURE TEH WITH HAH WITH MEEM INITIAL FORM" },
	{ 0xFD54 , "ARABIC LIGATURE TEH WITH KHAH WITH MEEM INITIAL FORM" },
	{ 0xFD55 , "ARABIC LIGATURE TEH WITH MEEM WITH JEEM INITIAL FORM" },
	{ 0xFD56 , "ARABIC LIGATURE TEH WITH MEEM WITH HAH INITIAL FORM" },
	{ 0xFD57 , "ARABIC LIGATURE TEH WITH MEEM WITH KHAH INITIAL FORM" },
	{ 0xFD58 , "ARABIC LIGATURE JEEM WITH MEEM WITH HAH FINAL FORM" },
	{ 0xFD59 , "ARABIC LIGATURE JEEM WITH MEEM WITH HAH INITIAL FORM" },
	{ 0xFD5A , "ARABIC LIGATURE HAH WITH MEEM WITH YEH FINAL FORM" },
	{ 0xFD5B , "ARABIC LIGATURE HAH WITH MEEM WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD5C , "ARABIC LIGATURE SEEN WITH HAH WITH JEEM INITIAL FORM" },
	{ 0xFD5D , "ARABIC LIGATURE SEEN WITH JEEM WITH HAH INITIAL FORM" },
	{ 0xFD5E , "ARABIC LIGATURE SEEN WITH JEEM WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD5F , "ARABIC LIGATURE SEEN WITH MEEM WITH HAH FINAL FORM" },
	{ 0xFD60 , "ARABIC LIGATURE SEEN WITH MEEM WITH HAH INITIAL FORM" },
	{ 0xFD61 , "ARABIC LIGATURE SEEN WITH MEEM WITH JEEM INITIAL FORM" },
	{ 0xFD62 , "ARABIC LIGATURE SEEN WITH MEEM WITH MEEM FINAL FORM" },
	{ 0xFD63 , "ARABIC LIGATURE SEEN WITH MEEM WITH MEEM INITIAL FORM" },
	{ 0xFD64 , "ARABIC LIGATURE SAD WITH HAH WITH HAH FINAL FORM" },
	{ 0xFD65 , "ARABIC LIGATURE SAD WITH HAH WITH HAH INITIAL FORM" },
	{ 0xFD66 , "ARABIC LIGATURE SAD WITH MEEM WITH MEEM FINAL FORM" },
	{ 0xFD67 , "ARABIC LIGATURE SHEEN WITH HAH WITH MEEM FINAL FORM" },
	{ 0xFD68 , "ARABIC LIGATURE SHEEN WITH HAH WITH MEEM INITIAL FORM" },
	{ 0xFD69 , "ARABIC LIGATURE SHEEN WITH JEEM WITH YEH FINAL FORM" },
	{ 0xFD6A , "ARABIC LIGATURE SHEEN WITH MEEM WITH KHAH FINAL FORM" },
	{ 0xFD6B , "ARABIC LIGATURE SHEEN WITH MEEM WITH KHAH INITIAL FORM" },
	{ 0xFD6C , "ARABIC LIGATURE SHEEN WITH MEEM WITH MEEM FINAL FORM" },
	{ 0xFD6D , "ARABIC LIGATURE SHEEN WITH MEEM WITH MEEM INITIAL FORM" },
	{ 0xFD6E , "ARABIC LIGATURE DAD WITH HAH WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD6F , "ARABIC LIGATURE DAD WITH KHAH WITH MEEM FINAL FORM" },
	{ 0xFD70 , "ARABIC LIGATURE DAD WITH KHAH WITH MEEM INITIAL FORM" },
	{ 0xFD71 , "ARABIC LIGATURE TAH WITH MEEM WITH HAH FINAL FORM" },
	{ 0xFD72 , "ARABIC LIGATURE TAH WITH MEEM WITH HAH INITIAL FORM" },
	{ 0xFD73 , "ARABIC LIGATURE TAH WITH MEEM WITH MEEM INITIAL FORM" },
	{ 0xFD74 , "ARABIC LIGATURE TAH WITH MEEM WITH YEH FINAL FORM" },
	{ 0xFD75 , "ARABIC LIGATURE AIN WITH JEEM WITH MEEM FINAL FORM" },
	{ 0xFD76 , "ARABIC LIGATURE AIN WITH MEEM WITH MEEM FINAL FORM" },
	{ 0xFD77 , "ARABIC LIGATURE AIN WITH MEEM WITH MEEM INITIAL FORM" },
	{ 0xFD78 , "ARABIC LIGATURE AIN WITH MEEM WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD79 , "ARABIC LIGATURE GHAIN WITH MEEM WITH MEEM FINAL FORM" },
	{ 0xFD7A , "ARABIC LIGATURE GHAIN WITH MEEM WITH YEH FINAL FORM" },
	{ 0xFD7B , "ARABIC LIGATURE GHAIN WITH MEEM WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD7C , "ARABIC LIGATURE FEH WITH KHAH WITH MEEM FINAL FORM" },
	{ 0xFD7D , "ARABIC LIGATURE FEH WITH KHAH WITH MEEM INITIAL FORM" },
	{ 0xFD7E , "ARABIC LIGATURE QAF WITH MEEM WITH HAH FINAL FORM" },
	{ 0xFD7F , "ARABIC LIGATURE QAF WITH MEEM WITH MEEM FINAL FORM" },
	{ 0xFD80 , "ARABIC LIGATURE LAM WITH HAH WITH MEEM FINAL FORM" },
	{ 0xFD81 , "ARABIC LIGATURE LAM WITH HAH WITH YEH FINAL FORM" },
	{ 0xFD82 , "ARABIC LIGATURE LAM WITH HAH WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD83 , "ARABIC LIGATURE LAM WITH JEEM WITH JEEM INITIAL FORM" },
	{ 0xFD84 , "ARABIC LIGATURE LAM WITH JEEM WITH JEEM FINAL FORM" },
	{ 0xFD85 , "ARABIC LIGATURE LAM WITH KHAH WITH MEEM FINAL FORM" },
	{ 0xFD86 , "ARABIC LIGATURE LAM WITH KHAH WITH MEEM INITIAL FORM" },
	{ 0xFD87 , "ARABIC LIGATURE LAM WITH MEEM WITH HAH FINAL FORM" },
	{ 0xFD88 , "ARABIC LIGATURE LAM WITH MEEM WITH HAH INITIAL FORM" },
	{ 0xFD89 , "ARABIC LIGATURE MEEM WITH HAH WITH JEEM INITIAL FORM" },
	{ 0xFD8A , "ARABIC LIGATURE MEEM WITH HAH WITH MEEM INITIAL FORM" },
	{ 0xFD8B , "ARABIC LIGATURE MEEM WITH HAH WITH YEH FINAL FORM" },
	{ 0xFD8C , "ARABIC LIGATURE MEEM WITH JEEM WITH HAH INITIAL FORM" },
	{ 0xFD8D , "ARABIC LIGATURE MEEM WITH JEEM WITH MEEM INITIAL FORM" },
	{ 0xFD8E , "ARABIC LIGATURE MEEM WITH KHAH WITH JEEM INITIAL FORM" },
	{ 0xFD8F , "ARABIC LIGATURE MEEM WITH KHAH WITH MEEM INITIAL FORM" },
	{ 0xFD92 , "ARABIC LIGATURE MEEM WITH JEEM WITH KHAH INITIAL FORM" },
	{ 0xFD93 , "ARABIC LIGATURE HEH WITH MEEM WITH JEEM INITIAL FORM" },
	{ 0xFD94 , "ARABIC LIGATURE HEH WITH MEEM WITH MEEM INITIAL FORM" },
	{ 0xFD95 , "ARABIC LIGATURE NOON WITH HAH WITH MEEM INITIAL FORM" },
	{ 0xFD96 , "ARABIC LIGATURE NOON WITH HAH WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD97 , "ARABIC LIGATURE NOON WITH JEEM WITH MEEM FINAL FORM" },
	{ 0xFD98 , "ARABIC LIGATURE NOON WITH JEEM WITH MEEM INITIAL FORM" },
	{ 0xFD99 , "ARABIC LIGATURE NOON WITH JEEM WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD9A , "ARABIC LIGATURE NOON WITH MEEM WITH YEH FINAL FORM" },
	{ 0xFD9B , "ARABIC LIGATURE NOON WITH MEEM WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFD9C , "ARABIC LIGATURE YEH WITH MEEM WITH MEEM FINAL FORM" },
	{ 0xFD9D , "ARABIC LIGATURE YEH WITH MEEM WITH MEEM INITIAL FORM" },
	{ 0xFD9E , "ARABIC LIGATURE BEH WITH KHAH WITH YEH FINAL FORM" },
	{ 0xFD9F , "ARABIC LIGATURE TEH WITH JEEM WITH YEH FINAL FORM" },
	{ 0xFDA0 , "ARABIC LIGATURE TEH WITH JEEM WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFDA1 , "ARABIC LIGATURE TEH WITH KHAH WITH YEH FINAL FORM" },
	{ 0xFDA2 , "ARABIC LIGATURE TEH WITH KHAH WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFDA3 , "ARABIC LIGATURE TEH WITH MEEM WITH YEH FINAL FORM" },
	{ 0xFDA4 , "ARABIC LIGATURE TEH WITH MEEM WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFDA5 , "ARABIC LIGATURE JEEM WITH MEEM WITH YEH FINAL FORM" },
	{ 0xFDA6 , "ARABIC LIGATURE JEEM WITH HAH WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFDA7 , "ARABIC LIGATURE JEEM WITH MEEM WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFDA8 , "ARABIC LIGATURE SEEN WITH KHAH WITH ALEF MAKSURA FINAL FORM" },
	{ 0xFDA9 , "ARABIC LIGATURE SAD WITH HAH WITH YEH FINAL FORM" },
	{ 0xFDAA , "ARABIC LIGATURE SHEEN WITH HAH WITH YEH FINAL FORM" },
	{ 0xFDAB , "ARABIC LIGATURE DAD WITH HAH WITH YEH FINAL FORM" },
	{ 0xFDAC , "ARABIC LIGATURE LAM WITH JEEM WITH YEH FINAL FORM" },
	{ 0xFDAD , "ARABIC LIGATURE LAM WITH MEEM WITH YEH FINAL FORM" },
	{ 0xFDAE , "ARABIC LIGATURE YEH WITH HAH WITH YEH FINAL FORM" },
	{ 0xFDAF , "ARABIC LIGATURE YEH WITH JEEM WITH YEH FINAL FORM" },
	{ 0xFDB0 , "ARABIC LIGATURE YEH WITH MEEM WITH YEH FINAL FORM" },
	{ 0xFDB1 , "ARABIC LIGATURE MEEM WITH MEEM WITH YEH FINAL FORM" },
	{ 0xFDB2 , "ARABIC LIGATURE QAF WITH MEEM WITH YEH FINAL FORM" },
	{ 0xFDB3 , "ARABIC LIGATURE NOON WITH HAH WITH YEH FINAL FORM" },
	{ 0xFDB4 , "ARABIC LIGATURE QAF WITH MEEM WITH HAH INITIAL FORM" },
	{ 0xFDB5 , "ARABIC LIGATURE LAM WITH HAH WITH MEEM INITIAL FORM" },
	{ 0xFDB6 , "ARABIC LIGATURE AIN WITH MEEM WITH YEH FINAL FORM" },
	{ 0xFDB7 , "ARABIC LIGATURE KAF WITH MEEM WITH YEH FINAL FORM" },
	{ 0xFDB8 , "ARABIC LIGATURE NOON WITH JEEM WITH HAH INITIAL FORM" },
	{ 0xFDB9 , "ARABIC LIGATURE MEEM WITH KHAH WITH YEH FINAL FORM" },
	{ 0xFDBA , "ARABIC LIGATURE LAM WITH JEEM WITH MEEM INITIAL FORM" },
	{ 0xFDBB , "ARABIC LIGATURE KAF WITH MEEM WITH MEEM FINAL FORM" },
	{ 0xFDBC , "ARABIC LIGATURE LAM WITH JEEM WITH MEEM FINAL FORM" },
	{ 0xFDBD , "ARABIC LIGATURE NOON WITH JEEM WITH HAH FINAL FORM" },
	{ 0xFDBE , "ARABIC LIGATURE JEEM WITH HAH WITH YEH FINAL FORM" },
	{ 0xFDBF , "ARABIC LIGATURE HAH WITH JEEM WITH YEH FINAL FORM" },
	{ 0xFDC0 , "ARABIC LIGATURE MEEM WITH JEEM WITH YEH FINAL FORM" },
	{ 0xFDC1 , "ARABIC LIGATURE FEH WITH MEEM WITH YEH FINAL FORM" },
	{ 0xFDC2 , "ARABIC LIGATURE BEH WITH HAH WITH YEH FINAL FORM" },
	{ 0xFDC3 , "ARABIC LIGATURE KAF WITH MEEM WITH MEEM INITIAL FORM" },
	{ 0xFDC4 , "ARABIC LIGATURE AIN WITH JEEM WITH MEEM INITIAL FORM" },
	{ 0xFDC5 , "ARABIC LIGATURE SAD WITH MEEM WITH MEEM INITIAL FORM" },
	{ 0xFDC6 , "ARABIC LIGATURE SEEN WITH KHAH WITH YEH FINAL FORM" },
	{ 0xFDC7 , "ARABIC LIGATURE NOON WITH JEEM WITH YEH FINAL FORM" },
	{ 0xFDF0 , "ARABIC LIGATURE SALLA USED AS KORANIC STOP SIGN ISOLATED FORM" },
	{ 0xFDF1 , "ARABIC LIGATURE QALA USED AS KORANIC STOP SIGN ISOLATED FORM" },
	{ 0xFDF2 , "ARABIC LIGATURE ALLAH ISOLATED FORM" },
	{ 0xFDF3 , "ARABIC LIGATURE AKBAR ISOLATED FORM" },
	{ 0xFDF4 , "ARABIC LIGATURE MOHAMMAD ISOLATED FORM" },
	{ 0xFDF5 , "ARABIC LIGATURE SALAM ISOLATED FORM" },
	{ 0xFDF6 , "ARABIC LIGATURE RASOUL ISOLATED FORM" },
	{ 0xFDF7 , "ARABIC LIGATURE ALAYHE ISOLATED FORM" },
	{ 0xFDF8 , "ARABIC LIGATURE WASALLAM ISOLATED FORM" },
	{ 0xFDF9 , "ARABIC LIGATURE SALLA ISOLATED FORM" },
	{ 0xFDFA , "ARABIC LIGATURE SALLALLAHOU ALAYHE WASALLAM" },
	{ 0xFDFB , "ARABIC LIGATURE JALLAJALALOUHOU" },
	{ 0xFE20 , "COMBINING LIGATURE LEFT HALF" },
	{ 0xFE21 , "COMBINING LIGATURE RIGHT HALF" },
	{ 0xFE22 , "COMBINING DOUBLE TILDE LEFT HALF" },
	{ 0xFE23 , "COMBINING DOUBLE TILDE RIGHT HALF" },
	{ 0xFE30 , "PRESENTATION FORM FOR VERTICAL TWO DOT LEADER" },
	{ 0xFE31 , "PRESENTATION FORM FOR VERTICAL EM DASH" },
	{ 0xFE32 , "PRESENTATION FORM FOR VERTICAL EN DASH" },
	{ 0xFE33 , "PRESENTATION FORM FOR VERTICAL LOW LINE" },
	{ 0xFE34 , "PRESENTATION FORM FOR VERTICAL WAVY LOW LINE" },
	{ 0xFE35 , "PRESENTATION FORM FOR VERTICAL LEFT PARENTHESIS" },
	{ 0xFE36 , "PRESENTATION FORM FOR VERTICAL RIGHT PARENTHESIS" },
	{ 0xFE37 , "PRESENTATION FORM FOR VERTICAL LEFT CURLY BRACKET" },
	{ 0xFE38 , "PRESENTATION FORM FOR VERTICAL RIGHT CURLY BRACKET" },
	{ 0xFE39 , "PRESENTATION FORM FOR VERTICAL LEFT TORTOISE SHELL BRACKET" },
	{ 0xFE3A , "PRESENTATION FORM FOR VERTICAL RIGHT TORTOISE SHELL BRACKET" },
	{ 0xFE3B , "PRESENTATION FORM FOR VERTICAL LEFT BLACK LENTICULAR BRACKET" },
	{ 0xFE3C , "PRESENTATION FORM FOR VERTICAL RIGHT BLACK LENTICULAR BRACKET" },
	{ 0xFE3D , "PRESENTATION FORM FOR VERTICAL LEFT DOUBLE ANGLE BRACKET" },
	{ 0xFE3E , "PRESENTATION FORM FOR VERTICAL RIGHT DOUBLE ANGLE BRACKET" },
	{ 0xFE3F , "PRESENTATION FORM FOR VERTICAL LEFT ANGLE BRACKET" },
	{ 0xFE40 , "PRESENTATION FORM FOR VERTICAL RIGHT ANGLE BRACKET" },
	{ 0xFE41 , "PRESENTATION FORM FOR VERTICAL LEFT CORNER BRACKET" },
	{ 0xFE42 , "PRESENTATION FORM FOR VERTICAL RIGHT CORNER BRACKET" },
	{ 0xFE43 , "PRESENTATION FORM FOR VERTICAL LEFT WHITE CORNER BRACKET" },
	{ 0xFE44 , "PRESENTATION FORM FOR VERTICAL RIGHT WHITE CORNER BRACKET" },
	{ 0xFE49 , "DASHED OVERLINE" },
	{ 0xFE4A , "CENTRELINE OVERLINE" },
	{ 0xFE4B , "WAVY OVERLINE" },
	{ 0xFE4C , "DOUBLE WAVY OVERLINE" },
	{ 0xFE4D , "DASHED LOW LINE" },
	{ 0xFE4E , "CENTRELINE LOW LINE" },
	{ 0xFE4F , "WAVY LOW LINE" },
	{ 0xFE50 , "SMALL COMMA" },
	{ 0xFE51 , "SMALL IDEOGRAPHIC COMMA" },
	{ 0xFE52 , "SMALL FULL STOP" },
	{ 0xFE54 , "SMALL SEMICOLON" },
	{ 0xFE55 , "SMALL COLON" },
	{ 0xFE56 , "SMALL QUESTION MARK" },
	{ 0xFE57 , "SMALL EXCLAMATION MARK" },
	{ 0xFE58 , "SMALL EM DASH" },
	{ 0xFE59 , "SMALL LEFT PARENTHESIS" },
	{ 0xFE5A , "SMALL RIGHT PARENTHESIS" },
	{ 0xFE5B , "SMALL LEFT CURLY BRACKET" },
	{ 0xFE5C , "SMALL RIGHT CURLY BRACKET" },
	{ 0xFE5D , "SMALL LEFT TORTOISE SHELL BRACKET" },
	{ 0xFE5E , "SMALL RIGHT TORTOISE SHELL BRACKET" },
	{ 0xFE5F , "SMALL NUMBER SIGN" },
	{ 0xFE60 , "SMALL AMPERSAND" },
	{ 0xFE61 , "SMALL ASTERISK" },
	{ 0xFE62 , "SMALL PLUS SIGN" },
	{ 0xFE63 , "SMALL HYPHEN-MINUS" },
	{ 0xFE64 , "SMALL LESS-THAN SIGN" },
	{ 0xFE65 , "SMALL GREATER-THAN SIGN" },
	{ 0xFE66 , "SMALL EQUALS SIGN" },
	{ 0xFE68 , "SMALL REVERSE SOLIDUS" },
	{ 0xFE69 , "SMALL DOLLAR SIGN" },
	{ 0xFE6A , "SMALL PERCENT SIGN" },
	{ 0xFE6B , "SMALL COMMERCIAL AT" },
	{ 0xFE70 , "ARABIC FATHATAN ISOLATED FORM" },
	{ 0xFE71 , "ARABIC TATWEEL WITH FATHATAN ABOVE" },
	{ 0xFE72 , "ARABIC DAMMATAN ISOLATED FORM" },
	{ 0xFE74 , "ARABIC KASRATAN ISOLATED FORM" },
	{ 0xFE76 , "ARABIC FATHA ISOLATED FORM" },
	{ 0xFE77 , "ARABIC FATHA MEDIAL FORM" },
	{ 0xFE78 , "ARABIC DAMMA ISOLATED FORM" },
	{ 0xFE79 , "ARABIC DAMMA MEDIAL FORM" },
	{ 0xFE7A , "ARABIC KASRA ISOLATED FORM" },
	{ 0xFE7B , "ARABIC KASRA MEDIAL FORM" },
	{ 0xFE7C , "ARABIC SHADDA ISOLATED FORM" },
	{ 0xFE7D , "ARABIC SHADDA MEDIAL FORM" },
	{ 0xFE7E , "ARABIC SUKUN ISOLATED FORM" },
	{ 0xFE7F , "ARABIC SUKUN MEDIAL FORM" },
	{ 0xFE80 , "ARABIC LETTER HAMZA ISOLATED FORM" },
	{ 0xFE81 , "ARABIC LETTER ALEF WITH MADDA ABOVE ISOLATED FORM" },
	{ 0xFE82 , "ARABIC LETTER ALEF WITH MADDA ABOVE FINAL FORM" },
	{ 0xFE83 , "ARABIC LETTER ALEF WITH HAMZA ABOVE ISOLATED FORM" },
	{ 0xFE84 , "ARABIC LETTER ALEF WITH HAMZA ABOVE FINAL FORM" },
	{ 0xFE85 , "ARABIC LETTER WAW WITH HAMZA ABOVE ISOLATED FORM" },
	{ 0xFE86 , "ARABIC LETTER WAW WITH HAMZA ABOVE FINAL FORM" },
	{ 0xFE87 , "ARABIC LETTER ALEF WITH HAMZA BELOW ISOLATED FORM" },
	{ 0xFE88 , "ARABIC LETTER ALEF WITH HAMZA BELOW FINAL FORM" },
	{ 0xFE89 , "ARABIC LETTER YEH WITH HAMZA ABOVE ISOLATED FORM" },
	{ 0xFE8A , "ARABIC LETTER YEH WITH HAMZA ABOVE FINAL FORM" },
	{ 0xFE8B , "ARABIC LETTER YEH WITH HAMZA ABOVE INITIAL FORM" },
	{ 0xFE8C , "ARABIC LETTER YEH WITH HAMZA ABOVE MEDIAL FORM" },
	{ 0xFE8D , "ARABIC LETTER ALEF ISOLATED FORM" },
	{ 0xFE8E , "ARABIC LETTER ALEF FINAL FORM" },
	{ 0xFE8F , "ARABIC LETTER BEH ISOLATED FORM" },
	{ 0xFE90 , "ARABIC LETTER BEH FINAL FORM" },
	{ 0xFE91 , "ARABIC LETTER BEH INITIAL FORM" },
	{ 0xFE92 , "ARABIC LETTER BEH MEDIAL FORM" },
	{ 0xFE93 , "ARABIC LETTER TEH MARBUTA ISOLATED FORM" },
	{ 0xFE94 , "ARABIC LETTER TEH MARBUTA FINAL FORM" },
	{ 0xFE95 , "ARABIC LETTER TEH ISOLATED FORM" },
	{ 0xFE96 , "ARABIC LETTER TEH FINAL FORM" },
	{ 0xFE97 , "ARABIC LETTER TEH INITIAL FORM" },
	{ 0xFE98 , "ARABIC LETTER TEH MEDIAL FORM" },
	{ 0xFE99 , "ARABIC LETTER THEH ISOLATED FORM" },
	{ 0xFE9A , "ARABIC LETTER THEH FINAL FORM" },
	{ 0xFE9B , "ARABIC LETTER THEH INITIAL FORM" },
	{ 0xFE9C , "ARABIC LETTER THEH MEDIAL FORM" },
	{ 0xFE9D , "ARABIC LETTER JEEM ISOLATED FORM" },
	{ 0xFE9E , "ARABIC LETTER JEEM FINAL FORM" },
	{ 0xFE9F , "ARABIC LETTER JEEM INITIAL FORM" },
	{ 0xFEA0 , "ARABIC LETTER JEEM MEDIAL FORM" },
	{ 0xFEA1 , "ARABIC LETTER HAH ISOLATED FORM" },
	{ 0xFEA2 , "ARABIC LETTER HAH FINAL FORM" },
	{ 0xFEA3 , "ARABIC LETTER HAH INITIAL FORM" },
	{ 0xFEA4 , "ARABIC LETTER HAH MEDIAL FORM" },
	{ 0xFEA5 , "ARABIC LETTER KHAH ISOLATED FORM" },
	{ 0xFEA6 , "ARABIC LETTER KHAH FINAL FORM" },
	{ 0xFEA7 , "ARABIC LETTER KHAH INITIAL FORM" },
	{ 0xFEA8 , "ARABIC LETTER KHAH MEDIAL FORM" },
	{ 0xFEA9 , "ARABIC LETTER DAL ISOLATED FORM" },
	{ 0xFEAA , "ARABIC LETTER DAL FINAL FORM" },
	{ 0xFEAB , "ARABIC LETTER THAL ISOLATED FORM" },
	{ 0xFEAC , "ARABIC LETTER THAL FINAL FORM" },
	{ 0xFEAD , "ARABIC LETTER REH ISOLATED FORM" },
	{ 0xFEAE , "ARABIC LETTER REH FINAL FORM" },
	{ 0xFEAF , "ARABIC LETTER ZAIN ISOLATED FORM" },
	{ 0xFEB0 , "ARABIC LETTER ZAIN FINAL FORM" },
	{ 0xFEB1 , "ARABIC LETTER SEEN ISOLATED FORM" },
	{ 0xFEB2 , "ARABIC LETTER SEEN FINAL FORM" },
	{ 0xFEB3 , "ARABIC LETTER SEEN INITIAL FORM" },
	{ 0xFEB4 , "ARABIC LETTER SEEN MEDIAL FORM" },
	{ 0xFEB5 , "ARABIC LETTER SHEEN ISOLATED FORM" },
	{ 0xFEB6 , "ARABIC LETTER SHEEN FINAL FORM" },
	{ 0xFEB7 , "ARABIC LETTER SHEEN INITIAL FORM" },
	{ 0xFEB8 , "ARABIC LETTER SHEEN MEDIAL FORM" },
	{ 0xFEB9 , "ARABIC LETTER SAD ISOLATED FORM" },
	{ 0xFEBA , "ARABIC LETTER SAD FINAL FORM" },
	{ 0xFEBB , "ARABIC LETTER SAD INITIAL FORM" },
	{ 0xFEBC , "ARABIC LETTER SAD MEDIAL FORM" },
	{ 0xFEBD , "ARABIC LETTER DAD ISOLATED FORM" },
	{ 0xFEBE , "ARABIC LETTER DAD FINAL FORM" },
	{ 0xFEBF , "ARABIC LETTER DAD INITIAL FORM" },
	{ 0xFEC0 , "ARABIC LETTER DAD MEDIAL FORM" },
	{ 0xFEC1 , "ARABIC LETTER TAH ISOLATED FORM" },
	{ 0xFEC2 , "ARABIC LETTER TAH FINAL FORM" },
	{ 0xFEC3 , "ARABIC LETTER TAH INITIAL FORM" },
	{ 0xFEC4 , "ARABIC LETTER TAH MEDIAL FORM" },
	{ 0xFEC5 , "ARABIC LETTER ZAH ISOLATED FORM" },
	{ 0xFEC6 , "ARABIC LETTER ZAH FINAL FORM" },
	{ 0xFEC7 , "ARABIC LETTER ZAH INITIAL FORM" },
	{ 0xFEC8 , "ARABIC LETTER ZAH MEDIAL FORM" },
	{ 0xFEC9 , "ARABIC LETTER AIN ISOLATED FORM" },
	{ 0xFECA , "ARABIC LETTER AIN FINAL FORM" },
	{ 0xFECB , "ARABIC LETTER AIN INITIAL FORM" },
	{ 0xFECC , "ARABIC LETTER AIN MEDIAL FORM" },
	{ 0xFECD , "ARABIC LETTER GHAIN ISOLATED FORM" },
	{ 0xFECE , "ARABIC LETTER GHAIN FINAL FORM" },
	{ 0xFECF , "ARABIC LETTER GHAIN INITIAL FORM" },
	{ 0xFED0 , "ARABIC LETTER GHAIN MEDIAL FORM" },
	{ 0xFED1 , "ARABIC LETTER FEH ISOLATED FORM" },
	{ 0xFED2 , "ARABIC LETTER FEH FINAL FORM" },
	{ 0xFED3 , "ARABIC LETTER FEH INITIAL FORM" },
	{ 0xFED4 , "ARABIC LETTER FEH MEDIAL FORM" },
	{ 0xFED5 , "ARABIC LETTER QAF ISOLATED FORM" },
	{ 0xFED6 , "ARABIC LETTER QAF FINAL FORM" },
	{ 0xFED7 , "ARABIC LETTER QAF INITIAL FORM" },
	{ 0xFED8 , "ARABIC LETTER QAF MEDIAL FORM" },
	{ 0xFED9 , "ARABIC LETTER KAF ISOLATED FORM" },
	{ 0xFEDA , "ARABIC LETTER KAF FINAL FORM" },
	{ 0xFEDB , "ARABIC LETTER KAF INITIAL FORM" },
	{ 0xFEDC , "ARABIC LETTER KAF MEDIAL FORM" },
	{ 0xFEDD , "ARABIC LETTER LAM ISOLATED FORM" },
	{ 0xFEDE , "ARABIC LETTER LAM FINAL FORM" },
	{ 0xFEDF , "ARABIC LETTER LAM INITIAL FORM" },
	{ 0xFEE0 , "ARABIC LETTER LAM MEDIAL FORM" },
	{ 0xFEE1 , "ARABIC LETTER MEEM ISOLATED FORM" },
	{ 0xFEE2 , "ARABIC LETTER MEEM FINAL FORM" },
	{ 0xFEE3 , "ARABIC LETTER MEEM INITIAL FORM" },
	{ 0xFEE4 , "ARABIC LETTER MEEM MEDIAL FORM" },
	{ 0xFEE5 , "ARABIC LETTER NOON ISOLATED FORM" },
	{ 0xFEE6 , "ARABIC LETTER NOON FINAL FORM" },
	{ 0xFEE7 , "ARABIC LETTER NOON INITIAL FORM" },
	{ 0xFEE8 , "ARABIC LETTER NOON MEDIAL FORM" },
	{ 0xFEE9 , "ARABIC LETTER HEH ISOLATED FORM" },
	{ 0xFEEA , "ARABIC LETTER HEH FINAL FORM" },
	{ 0xFEEB , "ARABIC LETTER HEH INITIAL FORM" },
	{ 0xFEEC , "ARABIC LETTER HEH MEDIAL FORM" },
	{ 0xFEED , "ARABIC LETTER WAW ISOLATED FORM" },
	{ 0xFEEE , "ARABIC LETTER WAW FINAL FORM" },
	{ 0xFEEF , "ARABIC LETTER ALEF MAKSURA ISOLATED FORM" },
	{ 0xFEF0 , "ARABIC LETTER ALEF MAKSURA FINAL FORM" },
	{ 0xFEF1 , "ARABIC LETTER YEH ISOLATED FORM" },
	{ 0xFEF2 , "ARABIC LETTER YEH FINAL FORM" },
	{ 0xFEF3 , "ARABIC LETTER YEH INITIAL FORM" },
	{ 0xFEF4 , "ARABIC LETTER YEH MEDIAL FORM" },
	{ 0xFEF5 , "ARABIC LIGATURE LAM WITH ALEF WITH MADDA ABOVE ISOLATED FORM" },
	{ 0xFEF6 , "ARABIC LIGATURE LAM WITH ALEF WITH MADDA ABOVE FINAL FORM" },
	{ 0xFEF7 , "ARABIC LIGATURE LAM WITH ALEF WITH HAMZA ABOVE ISOLATED FORM" },
	{ 0xFEF8 , "ARABIC LIGATURE LAM WITH ALEF WITH HAMZA ABOVE FINAL FORM" },
	{ 0xFEF9 , "ARABIC LIGATURE LAM WITH ALEF WITH HAMZA BELOW ISOLATED FORM" },
	{ 0xFEFA , "ARABIC LIGATURE LAM WITH ALEF WITH HAMZA BELOW FINAL FORM" },
	{ 0xFEFB , "ARABIC LIGATURE LAM WITH ALEF ISOLATED FORM" },
	{ 0xFEFC , "ARABIC LIGATURE LAM WITH ALEF FINAL FORM" },
#endif
	{ 0xFEFF , "ZERO WIDTH NO-BREAK SPACE" },
	{ 0xFF01 , "FULLWIDTH EXCLAMATION MARK" },
	{ 0xFF02 , "FULLWIDTH QUOTATION MARK" },
	{ 0xFF03 , "FULLWIDTH NUMBER SIGN" },
	{ 0xFF04 , "FULLWIDTH DOLLAR SIGN" },
	{ 0xFF05 , "FULLWIDTH PERCENT SIGN" },
	{ 0xFF06 , "FULLWIDTH AMPERSAND" },
	{ 0xFF07 , "FULLWIDTH APOSTROPHE" },
	{ 0xFF08 , "FULLWIDTH LEFT PARENTHESIS" },
	{ 0xFF09 , "FULLWIDTH RIGHT PARENTHESIS" },
	{ 0xFF0A , "FULLWIDTH ASTERISK" },
	{ 0xFF0B , "FULLWIDTH PLUS SIGN" },
	{ 0xFF0C , "FULLWIDTH COMMA" },
	{ 0xFF0D , "FULLWIDTH HYPHEN-MINUS" },
	{ 0xFF0E , "FULLWIDTH FULL STOP" },
	{ 0xFF0F , "FULLWIDTH SOLIDUS" },
	{ 0xFF10 , "FULLWIDTH DIGIT ZERO" },
	{ 0xFF11 , "FULLWIDTH DIGIT ONE" },
	{ 0xFF12 , "FULLWIDTH DIGIT TWO" },
	{ 0xFF13 , "FULLWIDTH DIGIT THREE" },
	{ 0xFF14 , "FULLWIDTH DIGIT FOUR" },
	{ 0xFF15 , "FULLWIDTH DIGIT FIVE" },
	{ 0xFF16 , "FULLWIDTH DIGIT SIX" },
	{ 0xFF17 , "FULLWIDTH DIGIT SEVEN" },
	{ 0xFF18 , "FULLWIDTH DIGIT EIGHT" },
	{ 0xFF19 , "FULLWIDTH DIGIT NINE" },
	{ 0xFF1A , "FULLWIDTH COLON" },
	{ 0xFF1B , "FULLWIDTH SEMICOLON" },
	{ 0xFF1C , "FULLWIDTH LESS-THAN SIGN" },
	{ 0xFF1D , "FULLWIDTH EQUALS SIGN" },
	{ 0xFF1E , "FULLWIDTH GREATER-THAN SIGN" },
	{ 0xFF1F , "FULLWIDTH QUESTION MARK" },
	{ 0xFF20 , "FULLWIDTH COMMERCIAL AT" },
	{ 0xFF21 , "FULLWIDTH LATIN CAPITAL LETTER A" },
	{ 0xFF22 , "FULLWIDTH LATIN CAPITAL LETTER B" },
	{ 0xFF23 , "FULLWIDTH LATIN CAPITAL LETTER C" },
	{ 0xFF24 , "FULLWIDTH LATIN CAPITAL LETTER D" },
	{ 0xFF25 , "FULLWIDTH LATIN CAPITAL LETTER E" },
	{ 0xFF26 , "FULLWIDTH LATIN CAPITAL LETTER F" },
	{ 0xFF27 , "FULLWIDTH LATIN CAPITAL LETTER G" },
	{ 0xFF28 , "FULLWIDTH LATIN CAPITAL LETTER H" },
	{ 0xFF29 , "FULLWIDTH LATIN CAPITAL LETTER I" },
	{ 0xFF2A , "FULLWIDTH LATIN CAPITAL LETTER J" },
	{ 0xFF2B , "FULLWIDTH LATIN CAPITAL LETTER K" },
	{ 0xFF2C , "FULLWIDTH LATIN CAPITAL LETTER L" },
	{ 0xFF2D , "FULLWIDTH LATIN CAPITAL LETTER M" },
	{ 0xFF2E , "FULLWIDTH LATIN CAPITAL LETTER N" },
	{ 0xFF2F , "FULLWIDTH LATIN CAPITAL LETTER O" },
	{ 0xFF30 , "FULLWIDTH LATIN CAPITAL LETTER P" },
	{ 0xFF31 , "FULLWIDTH LATIN CAPITAL LETTER Q" },
	{ 0xFF32 , "FULLWIDTH LATIN CAPITAL LETTER R" },
	{ 0xFF33 , "FULLWIDTH LATIN CAPITAL LETTER S" },
	{ 0xFF34 , "FULLWIDTH LATIN CAPITAL LETTER T" },
	{ 0xFF35 , "FULLWIDTH LATIN CAPITAL LETTER U" },
	{ 0xFF36 , "FULLWIDTH LATIN CAPITAL LETTER V" },
	{ 0xFF37 , "FULLWIDTH LATIN CAPITAL LETTER W" },
	{ 0xFF38 , "FULLWIDTH LATIN CAPITAL LETTER X" },
	{ 0xFF39 , "FULLWIDTH LATIN CAPITAL LETTER Y" },
	{ 0xFF3A , "FULLWIDTH LATIN CAPITAL LETTER Z" },
	{ 0xFF3B , "FULLWIDTH LEFT SQUARE BRACKET" },
	{ 0xFF3C , "FULLWIDTH REVERSE SOLIDUS" },
	{ 0xFF3D , "FULLWIDTH RIGHT SQUARE BRACKET" },
	{ 0xFF3E , "FULLWIDTH CIRCUMFLEX ACCENT" },
	{ 0xFF3F , "FULLWIDTH LOW LINE" },
	{ 0xFF40 , "FULLWIDTH GRAVE ACCENT" },
	{ 0xFF41 , "FULLWIDTH LATIN SMALL LETTER A" },
	{ 0xFF42 , "FULLWIDTH LATIN SMALL LETTER B" },
	{ 0xFF43 , "FULLWIDTH LATIN SMALL LETTER C" },
	{ 0xFF44 , "FULLWIDTH LATIN SMALL LETTER D" },
	{ 0xFF45 , "FULLWIDTH LATIN SMALL LETTER E" },
	{ 0xFF46 , "FULLWIDTH LATIN SMALL LETTER F" },
	{ 0xFF47 , "FULLWIDTH LATIN SMALL LETTER G" },
	{ 0xFF48 , "FULLWIDTH LATIN SMALL LETTER H" },
	{ 0xFF49 , "FULLWIDTH LATIN SMALL LETTER I" },
	{ 0xFF4A , "FULLWIDTH LATIN SMALL LETTER J" },
	{ 0xFF4B , "FULLWIDTH LATIN SMALL LETTER K" },
	{ 0xFF4C , "FULLWIDTH LATIN SMALL LETTER L" },
	{ 0xFF4D , "FULLWIDTH LATIN SMALL LETTER M" },
	{ 0xFF4E , "FULLWIDTH LATIN SMALL LETTER N" },
	{ 0xFF4F , "FULLWIDTH LATIN SMALL LETTER O" },
	{ 0xFF50 , "FULLWIDTH LATIN SMALL LETTER P" },
	{ 0xFF51 , "FULLWIDTH LATIN SMALL LETTER Q" },
	{ 0xFF52 , "FULLWIDTH LATIN SMALL LETTER R" },
	{ 0xFF53 , "FULLWIDTH LATIN SMALL LETTER S" },
	{ 0xFF54 , "FULLWIDTH LATIN SMALL LETTER T" },
	{ 0xFF55 , "FULLWIDTH LATIN SMALL LETTER U" },
	{ 0xFF56 , "FULLWIDTH LATIN SMALL LETTER V" },
	{ 0xFF57 , "FULLWIDTH LATIN SMALL LETTER W" },
	{ 0xFF58 , "FULLWIDTH LATIN SMALL LETTER X" },
	{ 0xFF59 , "FULLWIDTH LATIN SMALL LETTER Y" },
	{ 0xFF5A , "FULLWIDTH LATIN SMALL LETTER Z" },
	{ 0xFF5B , "FULLWIDTH LEFT CURLY BRACKET" },
	{ 0xFF5C , "FULLWIDTH VERTICAL LINE" },
	{ 0xFF5D , "FULLWIDTH RIGHT CURLY BRACKET" },
	{ 0xFF5E , "FULLWIDTH TILDE" },
	{ 0xFF61 , "HALFWIDTH IDEOGRAPHIC FULL STOP" },
	{ 0xFF62 , "HALFWIDTH LEFT CORNER BRACKET" },
	{ 0xFF63 , "HALFWIDTH RIGHT CORNER BRACKET" },
	{ 0xFF64 , "HALFWIDTH IDEOGRAPHIC COMMA" },
	{ 0xFF65 , "HALFWIDTH KATAKANA MIDDLE DOT" },
	{ 0xFF66 , "HALFWIDTH KATAKANA LETTER WO" },
	{ 0xFF67 , "HALFWIDTH KATAKANA LETTER SMALL A" },
	{ 0xFF68 , "HALFWIDTH KATAKANA LETTER SMALL I" },
	{ 0xFF69 , "HALFWIDTH KATAKANA LETTER SMALL U" },
	{ 0xFF6A , "HALFWIDTH KATAKANA LETTER SMALL E" },
	{ 0xFF6B , "HALFWIDTH KATAKANA LETTER SMALL O" },
	{ 0xFF6C , "HALFWIDTH KATAKANA LETTER SMALL YA" },
	{ 0xFF6D , "HALFWIDTH KATAKANA LETTER SMALL YU" },
	{ 0xFF6E , "HALFWIDTH KATAKANA LETTER SMALL YO" },
	{ 0xFF6F , "HALFWIDTH KATAKANA LETTER SMALL TU" },
	{ 0xFF70 , "HALFWIDTH KATAKANA-HIRAGANA PROLONGED SOUND MARK" },
	{ 0xFF71 , "HALFWIDTH KATAKANA LETTER A" },
	{ 0xFF72 , "HALFWIDTH KATAKANA LETTER I" },
	{ 0xFF73 , "HALFWIDTH KATAKANA LETTER U" },
	{ 0xFF74 , "HALFWIDTH KATAKANA LETTER E" },
	{ 0xFF75 , "HALFWIDTH KATAKANA LETTER O" },
	{ 0xFF76 , "HALFWIDTH KATAKANA LETTER KA" },
	{ 0xFF77 , "HALFWIDTH KATAKANA LETTER KI" },
	{ 0xFF78 , "HALFWIDTH KATAKANA LETTER KU" },
	{ 0xFF79 , "HALFWIDTH KATAKANA LETTER KE" },
	{ 0xFF7A , "HALFWIDTH KATAKANA LETTER KO" },
	{ 0xFF7B , "HALFWIDTH KATAKANA LETTER SA" },
	{ 0xFF7C , "HALFWIDTH KATAKANA LETTER SI" },
	{ 0xFF7D , "HALFWIDTH KATAKANA LETTER SU" },
	{ 0xFF7E , "HALFWIDTH KATAKANA LETTER SE" },
	{ 0xFF7F , "HALFWIDTH KATAKANA LETTER SO" },
	{ 0xFF80 , "HALFWIDTH KATAKANA LETTER TA" },
	{ 0xFF81 , "HALFWIDTH KATAKANA LETTER TI" },
	{ 0xFF82 , "HALFWIDTH KATAKANA LETTER TU" },
	{ 0xFF83 , "HALFWIDTH KATAKANA LETTER TE" },
	{ 0xFF84 , "HALFWIDTH KATAKANA LETTER TO" },
	{ 0xFF85 , "HALFWIDTH KATAKANA LETTER NA" },
	{ 0xFF86 , "HALFWIDTH KATAKANA LETTER NI" },
	{ 0xFF87 , "HALFWIDTH KATAKANA LETTER NU" },
	{ 0xFF88 , "HALFWIDTH KATAKANA LETTER NE" },
	{ 0xFF89 , "HALFWIDTH KATAKANA LETTER NO" },
	{ 0xFF8A , "HALFWIDTH KATAKANA LETTER HA" },
	{ 0xFF8B , "HALFWIDTH KATAKANA LETTER HI" },
	{ 0xFF8C , "HALFWIDTH KATAKANA LETTER HU" },
	{ 0xFF8D , "HALFWIDTH KATAKANA LETTER HE" },
	{ 0xFF8E , "HALFWIDTH KATAKANA LETTER HO" },
	{ 0xFF8F , "HALFWIDTH KATAKANA LETTER MA" },
	{ 0xFF90 , "HALFWIDTH KATAKANA LETTER MI" },
	{ 0xFF91 , "HALFWIDTH KATAKANA LETTER MU" },
	{ 0xFF92 , "HALFWIDTH KATAKANA LETTER ME" },
	{ 0xFF93 , "HALFWIDTH KATAKANA LETTER MO" },
	{ 0xFF94 , "HALFWIDTH KATAKANA LETTER YA" },
	{ 0xFF95 , "HALFWIDTH KATAKANA LETTER YU" },
	{ 0xFF96 , "HALFWIDTH KATAKANA LETTER YO" },
	{ 0xFF97 , "HALFWIDTH KATAKANA LETTER RA" },
	{ 0xFF98 , "HALFWIDTH KATAKANA LETTER RI" },
	{ 0xFF99 , "HALFWIDTH KATAKANA LETTER RU" },
	{ 0xFF9A , "HALFWIDTH KATAKANA LETTER RE" },
	{ 0xFF9B , "HALFWIDTH KATAKANA LETTER RO" },
	{ 0xFF9C , "HALFWIDTH KATAKANA LETTER WA" },
	{ 0xFF9D , "HALFWIDTH KATAKANA LETTER N" },
	{ 0xFF9E , "HALFWIDTH KATAKANA VOICED SOUND MARK" },
	{ 0xFF9F , "HALFWIDTH KATAKANA SEMI-VOICED SOUND MARK" },
	{ 0xFFA0 , "HALFWIDTH HANGUL FILLER" },
	{ 0xFFA1 , "HALFWIDTH HANGUL LETTER KIYEOK" },
	{ 0xFFA2 , "HALFWIDTH HANGUL LETTER SSANGKIYEOK" },
	{ 0xFFA3 , "HALFWIDTH HANGUL LETTER KIYEOK-SIOS" },
	{ 0xFFA4 , "HALFWIDTH HANGUL LETTER NIEUN" },
	{ 0xFFA5 , "HALFWIDTH HANGUL LETTER NIEUN-CIEUC" },
	{ 0xFFA6 , "HALFWIDTH HANGUL LETTER NIEUN-HIEUH" },
	{ 0xFFA7 , "HALFWIDTH HANGUL LETTER TIKEUT" },
	{ 0xFFA8 , "HALFWIDTH HANGUL LETTER SSANGTIKEUT" },
	{ 0xFFA9 , "HALFWIDTH HANGUL LETTER RIEUL" },
	{ 0xFFAA , "HALFWIDTH HANGUL LETTER RIEUL-KIYEOK" },
	{ 0xFFAB , "HALFWIDTH HANGUL LETTER RIEUL-MIEUM" },
	{ 0xFFAC , "HALFWIDTH HANGUL LETTER RIEUL-PIEUP" },
	{ 0xFFAD , "HALFWIDTH HANGUL LETTER RIEUL-SIOS" },
	{ 0xFFAE , "HALFWIDTH HANGUL LETTER RIEUL-THIEUTH" },
	{ 0xFFAF , "HALFWIDTH HANGUL LETTER RIEUL-PHIEUPH" },
	{ 0xFFB0 , "HALFWIDTH HANGUL LETTER RIEUL-HIEUH" },
	{ 0xFFB1 , "HALFWIDTH HANGUL LETTER MIEUM" },
	{ 0xFFB2 , "HALFWIDTH HANGUL LETTER PIEUP" },
	{ 0xFFB3 , "HALFWIDTH HANGUL LETTER SSANGPIEUP" },
	{ 0xFFB4 , "HALFWIDTH HANGUL LETTER PIEUP-SIOS" },
	{ 0xFFB5 , "HALFWIDTH HANGUL LETTER SIOS" },
	{ 0xFFB6 , "HALFWIDTH HANGUL LETTER SSANGSIOS" },
	{ 0xFFB7 , "HALFWIDTH HANGUL LETTER IEUNG" },
	{ 0xFFB8 , "HALFWIDTH HANGUL LETTER CIEUC" },
	{ 0xFFB9 , "HALFWIDTH HANGUL LETTER SSANGCIEUC" },
	{ 0xFFBA , "HALFWIDTH HANGUL LETTER CHIEUCH" },
	{ 0xFFBB , "HALFWIDTH HANGUL LETTER KHIEUKH" },
	{ 0xFFBC , "HALFWIDTH HANGUL LETTER THIEUTH" },
	{ 0xFFBD , "HALFWIDTH HANGUL LETTER PHIEUPH" },
	{ 0xFFBE , "HALFWIDTH HANGUL LETTER HIEUH" },
	{ 0xFFC2 , "HALFWIDTH HANGUL LETTER A" },
	{ 0xFFC3 , "HALFWIDTH HANGUL LETTER AE" },
	{ 0xFFC4 , "HALFWIDTH HANGUL LETTER YA" },
	{ 0xFFC5 , "HALFWIDTH HANGUL LETTER YAE" },
	{ 0xFFC6 , "HALFWIDTH HANGUL LETTER EO" },
	{ 0xFFC7 , "HALFWIDTH HANGUL LETTER E" },
	{ 0xFFCA , "HALFWIDTH HANGUL LETTER YEO" },
	{ 0xFFCB , "HALFWIDTH HANGUL LETTER YE" },
	{ 0xFFCC , "HALFWIDTH HANGUL LETTER O" },
	{ 0xFFCD , "HALFWIDTH HANGUL LETTER WA" },
	{ 0xFFCE , "HALFWIDTH HANGUL LETTER WAE" },
	{ 0xFFCF , "HALFWIDTH HANGUL LETTER OE" },
	{ 0xFFD2 , "HALFWIDTH HANGUL LETTER YO" },
	{ 0xFFD3 , "HALFWIDTH HANGUL LETTER U" },
	{ 0xFFD4 , "HALFWIDTH HANGUL LETTER WEO" },
	{ 0xFFD5 , "HALFWIDTH HANGUL LETTER WE" },
	{ 0xFFD6 , "HALFWIDTH HANGUL LETTER WI" },
	{ 0xFFD7 , "HALFWIDTH HANGUL LETTER YU" },
	{ 0xFFDA , "HALFWIDTH HANGUL LETTER EU" },
	{ 0xFFDB , "HALFWIDTH HANGUL LETTER YI" },
	{ 0xFFDC , "HALFWIDTH HANGUL LETTER I" },
	{ 0xFFE0 , "FULLWIDTH CENT SIGN" },
	{ 0xFFE1 , "FULLWIDTH POUND SIGN" },
	{ 0xFFE2 , "FULLWIDTH NOT SIGN" },
	{ 0xFFE3 , "FULLWIDTH MACRON" },
	{ 0xFFE4 , "FULLWIDTH BROKEN BAR" },
	{ 0xFFE5 , "FULLWIDTH YEN SIGN" },
	{ 0xFFE6 , "FULLWIDTH WON SIGN" },
	{ 0xFFE8 , "HALFWIDTH FORMS LIGHT VERTICAL" },
	{ 0xFFE9 , "HALFWIDTH LEFTWARDS ARROW" },
	{ 0xFFEA , "HALFWIDTH UPWARDS ARROW" },
	{ 0xFFEB , "HALFWIDTH RIGHTWARDS ARROW" },
	{ 0xFFEC , "HALFWIDTH DOWNWARDS ARROW" },
	{ 0xFFED , "HALFWIDTH BLACK SQUARE" },
	{ 0xFFEE , "HALFWIDTH WHITE CIRCLE" },
	{ 0xFFFD , "REPLACEMENT CHARACTER" },
};

static int4 
findUniSlot(UnicodeNameTable *t, short c, int4 start, int4 end)
{
	int4 si;
	int4 ei;
	int4 mi;
	int4 midpoint;

	si = t[start].index;
	ei = t[end].index;

	if(si > c || ei < c) return -1;

	if(si == c) return start;
	if(ei == c) return end;
	if(end == start + 1) return -1;

	mi = t[midpoint = ((start + end) >> 1)].index;

	if(mi < c) return findUniSlot(t, c, midpoint, end);
	else return findUniSlot(t, c, start, midpoint);
}
	
Dll void APIENTRY
uc2subs(iUCode uc, char as[MAXUNILENGTH + 2])
{
   int4 index;

   index = findUniSlot(uni_table, uc, 0, MAXUNIVALUE);
   if(index < 0) strcpy(as,"NO UNICODE ");
   else {
	   strcpy(as, uni_table[index].name);
	   if(uc > MAXASCII) strcat(as," ");
   }
}
    
Dll void APIENTRY
ucs2subs(iUCode *s, char *string)
{
	unsigned long ucslength,i;
	char buffer[MAXUNILENGTH+2];

	ucslength = ucslen((const unsigned short*)s);
	string[0] = '\0';

	string[0] = 0;
	for(i=0;i<ucslength;i++) {
		uc2subs(s[i], buffer);
		strcat(string,buffer);
	}
}

