/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "g3states.h"

/*
// global data
// This is obviously missing the original TIFF disclaimer
//
// it contains several large tables
// the question is, what can be done to make it more palatable to
// DOS
//
// answer -- first, we can declare it huge,
// where does the huge keyword go?
// second, it has a lot of double index arrays, they can be broken up
// into two levels of arrays
// unsigned char	RlMode2D[20][256]
// unsigned char	RlNextState2D[20][256]
// unsigned char	RlUncompAction[20][256]
// unsigned char	RlUncompNextState[20][256]
//
// very important calculation 20 * 256 ~= 5000
// 
*/

unsigned char	RlMode2D[20][256] = {
/* prefix 0 */ {
    12,12,11, 0, 3, 3, 9, 9, 4, 4, 4, 4, 8, 8, 8, 8, /*   0- 15 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  16- 31 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  32- 47 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  48- 63 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  64- 79 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  80- 95 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /*  96-111 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 112-127 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 128-143 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 144-159 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 160-175 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 176-191 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 192-207 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 208-223 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 224-239 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 240-255 */
},
/* prefix 1 */ {
    12, 0, 3, 9, 4, 4, 8, 8, 1, 1, 1, 1, 1, 1, 1, 1, /*   0- 15 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  16- 31 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  32- 47 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /*  48- 63 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  64- 79 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  80- 95 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  96-111 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 112-127 */
    12, 0, 3, 9, 4, 4, 8, 8, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 144-159 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 160-175 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 176-191 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 192-207 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 208-223 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 224-239 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 240-255 */
},
/* prefix 2 */ {
     0, 0, 4, 8, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, /*   0- 15 */
     5, 5, 5, 5, 5, 5, 5, 5, 7, 7, 7, 7, 7, 7, 7, 7, /*  16- 31 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  32- 47 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  48- 63 */
     0, 0, 4, 8, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     5, 5, 5, 5, 5, 5, 5, 5, 7, 7, 7, 7, 7, 7, 7, 7, /*  80- 95 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  96-111 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 112-127 */
     0, 0, 4, 8, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, /* 128-143 */
     5, 5, 5, 5, 5, 5, 5, 5, 7, 7, 7, 7, 7, 7, 7, 7, /* 144-159 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 160-175 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 176-191 */
     0, 0, 4, 8, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, /* 192-207 */
     5, 5, 5, 5, 5, 5, 5, 5, 7, 7, 7, 7, 7, 7, 7, 7, /* 208-223 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 224-239 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 240-255 */
},
/* prefix 3 */ {
     0, 0, 1, 1, 2, 2, 2, 2, 5, 5, 5, 5, 7, 7, 7, 7, /*   0- 15 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  16- 31 */
     0, 0, 1, 1, 2, 2, 2, 2, 5, 5, 5, 5, 7, 7, 7, 7, /*  32- 47 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  48- 63 */
     0, 0, 1, 1, 2, 2, 2, 2, 5, 5, 5, 5, 7, 7, 7, 7, /*  64- 79 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  80- 95 */
     0, 0, 1, 1, 2, 2, 2, 2, 5, 5, 5, 5, 7, 7, 7, 7, /*  96-111 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 112-127 */
     0, 0, 1, 1, 2, 2, 2, 2, 5, 5, 5, 5, 7, 7, 7, 7, /* 128-143 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 144-159 */
     0, 0, 1, 1, 2, 2, 2, 2, 5, 5, 5, 5, 7, 7, 7, 7, /* 160-175 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 176-191 */
     0, 0, 1, 1, 2, 2, 2, 2, 5, 5, 5, 5, 7, 7, 7, 7, /* 192-207 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 208-223 */
     0, 0, 1, 1, 2, 2, 2, 2, 5, 5, 5, 5, 7, 7, 7, 7, /* 224-239 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 240-255 */
},
/* prefix 4 */ {
     0, 1, 2, 2, 5, 5, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /*   0- 15 */
     0, 1, 2, 2, 5, 5, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /*  16- 31 */
     0, 1, 2, 2, 5, 5, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /*  32- 47 */
     0, 1, 2, 2, 5, 5, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /*  48- 63 */
     0, 1, 2, 2, 5, 5, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /*  64- 79 */
     0, 1, 2, 2, 5, 5, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /*  80- 95 */
     0, 1, 2, 2, 5, 5, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /*  96-111 */
     0, 1, 2, 2, 5, 5, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /* 112-127 */
     0, 1, 2, 2, 5, 5, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /* 128-143 */
     0, 1, 2, 2, 5, 5, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /* 144-159 */
     0, 1, 2, 2, 5, 5, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /* 160-175 */
     0, 1, 2, 2, 5, 5, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /* 176-191 */
     0, 1, 2, 2, 5, 5, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /* 192-207 */
     0, 1, 2, 2, 5, 5, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /* 208-223 */
     0, 1, 2, 2, 5, 5, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /* 224-239 */
     0, 1, 2, 2, 5, 5, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /* 240-255 */
},
/* prefix 5 */ {
     0, 2, 5, 7, 6, 6, 6, 6, 0, 2, 5, 7, 6, 6, 6, 6, /*   0- 15 */
     0, 2, 5, 7, 6, 6, 6, 6, 0, 2, 5, 7, 6, 6, 6, 6, /*  16- 31 */
     0, 2, 5, 7, 6, 6, 6, 6, 0, 2, 5, 7, 6, 6, 6, 6, /*  32- 47 */
     0, 2, 5, 7, 6, 6, 6, 6, 0, 2, 5, 7, 6, 6, 6, 6, /*  48- 63 */
     0, 2, 5, 7, 6, 6, 6, 6, 0, 2, 5, 7, 6, 6, 6, 6, /*  64- 79 */
     0, 2, 5, 7, 6, 6, 6, 6, 0, 2, 5, 7, 6, 6, 6, 6, /*  80- 95 */
     0, 2, 5, 7, 6, 6, 6, 6, 0, 2, 5, 7, 6, 6, 6, 6, /*  96-111 */
     0, 2, 5, 7, 6, 6, 6, 6, 0, 2, 5, 7, 6, 6, 6, 6, /* 112-127 */
     0, 2, 5, 7, 6, 6, 6, 6, 0, 2, 5, 7, 6, 6, 6, 6, /* 128-143 */
     0, 2, 5, 7, 6, 6, 6, 6, 0, 2, 5, 7, 6, 6, 6, 6, /* 144-159 */
     0, 2, 5, 7, 6, 6, 6, 6, 0, 2, 5, 7, 6, 6, 6, 6, /* 160-175 */
     0, 2, 5, 7, 6, 6, 6, 6, 0, 2, 5, 7, 6, 6, 6, 6, /* 176-191 */
     0, 2, 5, 7, 6, 6, 6, 6, 0, 2, 5, 7, 6, 6, 6, 6, /* 192-207 */
     0, 2, 5, 7, 6, 6, 6, 6, 0, 2, 5, 7, 6, 6, 6, 6, /* 208-223 */
     0, 2, 5, 7, 6, 6, 6, 6, 0, 2, 5, 7, 6, 6, 6, 6, /* 224-239 */
     0, 2, 5, 7, 6, 6, 6, 6, 0, 2, 5, 7, 6, 6, 6, 6, /* 240-255 */
},
/* prefix 6 */ {
     0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, /*   0- 15 */
     0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, /*  16- 31 */
     0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, /*  32- 47 */
     0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, /*  48- 63 */
     0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, /*  64- 79 */
     0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, /*  80- 95 */
     0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, /*  96-111 */
     0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, /* 112-127 */
     0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, /* 128-143 */
     0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, /* 144-159 */
     0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, /* 160-175 */
     0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, /* 176-191 */
     0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, /* 192-207 */
     0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, /* 208-223 */
     0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, /* 224-239 */
     0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, 0, 0, 6, 6, /* 240-255 */
},
/* prefix 7 */ {
     0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, /*   0- 15 */
     0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, /*  16- 31 */
     0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, /*  32- 47 */
     0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, /*  48- 63 */
     0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, /*  64- 79 */
     0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, /*  80- 95 */
     0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, /*  96-111 */
     0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, /* 112-127 */
     0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, /* 128-143 */
     0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, /* 144-159 */
     0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, /* 160-175 */
     0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, /* 176-191 */
     0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, /* 192-207 */
     0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, /* 208-223 */
     0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, /* 224-239 */
     0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, 0, 6, /* 240-255 */
},
/* prefix 8 */ {
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*   0- 15 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  16- 31 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  32- 47 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  48- 63 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  64- 79 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  80- 95 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  96-111 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 112-127 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 128-143 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 144-159 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 160-175 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 176-191 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 192-207 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 208-223 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 224-239 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 240-255 */
},
/* prefix 9 */ {
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*   0- 15 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  16- 31 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  32- 47 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  48- 63 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  64- 79 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  80- 95 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  96-111 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 112-127 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 128-143 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 144-159 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 160-175 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 176-191 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 192-207 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 208-223 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 224-239 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 240-255 */
},
/* prefix 10 */ {
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*   0- 15 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*  16- 31 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*  32- 47 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*  48- 63 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*  64- 79 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*  80- 95 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*  96-111 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /* 112-127 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 128-143 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 144-159 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 160-175 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 176-191 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 192-207 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 208-223 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 224-239 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 240-255 */
},
/* prefix 11 */ {
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*   0- 15 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  48- 63 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  64- 79 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  80- 95 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  96-111 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 112-127 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 128-143 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 144-159 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 160-175 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 176-191 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 192-207 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 208-223 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 224-239 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 240-255 */
},
/* prefix 12 */ {
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*   0- 15 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*  16- 31 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*  32- 47 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*  48- 63 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  64- 79 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  80- 95 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  96-111 */
    11,11,11,11,11,11,11,11,10,10,10,10,10,10,10,10, /* 112-127 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 128-143 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 144-159 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 160-175 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 176-191 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 192-207 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 208-223 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 224-239 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 240-255 */
},
/* prefix 13 */ {
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  16- 31 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  32- 47 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  48- 63 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  64- 79 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  80- 95 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  96-111 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 112-127 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 128-143 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 144-159 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 160-175 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 176-191 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 192-207 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 208-223 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 224-239 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 240-255 */
},
/* prefix 14 */ {
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*   0- 15 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*  16- 31 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  32- 47 */
    11,11,11,11,11,11,11,11,11,11,11,11,10,10,10,10, /*  48- 63 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  64- 79 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  80- 95 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /*  96-111 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 112-127 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 128-143 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 144-159 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 160-175 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 176-191 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 192-207 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 208-223 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 224-239 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 240-255 */
},
/* prefix 15 */ {
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*   0- 15 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,10,10, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /*  48- 63 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  64- 79 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  80- 95 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /*  96-111 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 16 */ {
    12,12,12,12,12,12,12,12,11,11,11,11,11,11,11,10, /*   0- 15 */
     3, 3, 3, 3, 3, 3, 3, 3, 9, 9, 9, 9, 9, 9, 9, 9, /*  16- 31 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  32- 47 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /*  48- 63 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  64- 79 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  80- 95 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  96-111 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 112-127 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 128-143 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 144-159 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 160-175 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 176-191 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 192-207 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 208-223 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 224-239 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 240-255 */
},
/* prefix 17 */ {
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*   0- 15 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  16- 31 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  32- 47 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  48- 63 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  64- 79 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  80- 95 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  96-111 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 112-127 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 128-143 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 144-159 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 160-175 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 176-191 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 192-207 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 208-223 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 224-239 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 240-255 */
},
/* prefix 18 */ {
    12,12,12,12,11,11,11, 0, 3, 3, 3, 3, 9, 9, 9, 9, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 8, 8, 8, 8, 8, 8, 8, 8, /*  16- 31 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  32- 47 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  48- 63 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 128-143 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 144-159 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 160-175 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 176-191 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 192-207 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 208-223 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 224-239 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 240-255 */
},
/* prefix 19 */ {
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*   0- 15 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  16- 31 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  32- 47 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  48- 63 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  64- 79 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  80- 95 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  96-111 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 112-127 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 128-143 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 144-159 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 160-175 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 176-191 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 192-207 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 208-223 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 224-239 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 240-255 */
}
};
unsigned char	RlNextState2D[20][256] = {
/* prefix 0 */ {
     7, 7, 0, 8, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  48- 63 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  64- 79 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  80- 95 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  96-111 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 1 */ {
     0, 9, 0, 0, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  16- 31 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  32- 47 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  48- 63 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     0, 9, 0, 0, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /* 128-143 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 144-159 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 160-175 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 176-191 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 192-207 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 208-223 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 224-239 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 240-255 */
},
/* prefix 2 */ {
    10,11, 0, 0, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*   0- 15 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  48- 63 */
    10,11, 0, 0, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*  64- 79 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  80- 95 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  96-111 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 112-127 */
    10,11, 0, 0, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 128-143 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 144-159 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 160-175 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 176-191 */
    10,11, 0, 0, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 192-207 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 208-223 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 224-239 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 240-255 */
},
/* prefix 3 */ {
    12,13, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  16- 31 */
    12,13, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  32- 47 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  48- 63 */
    12,13, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  64- 79 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  80- 95 */
    12,13, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  96-111 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 112-127 */
    12,13, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 128-143 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 144-159 */
    12,13, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 160-175 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 176-191 */
    12,13, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 192-207 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 208-223 */
    12,13, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 224-239 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 240-255 */
},
/* prefix 4 */ {
    14, 0, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /*   0- 15 */
    14, 0, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /*  16- 31 */
    14, 0, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /*  32- 47 */
    14, 0, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /*  48- 63 */
    14, 0, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /*  64- 79 */
    14, 0, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /*  80- 95 */
    14, 0, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /*  96-111 */
    14, 0, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /* 112-127 */
    14, 0, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /* 128-143 */
    14, 0, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /* 144-159 */
    14, 0, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /* 160-175 */
    14, 0, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /* 176-191 */
    14, 0, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /* 192-207 */
    14, 0, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /* 208-223 */
    14, 0, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /* 224-239 */
    14, 0, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 5, 5, 5, /* 240-255 */
},
/* prefix 5 */ {
    15, 0, 0, 0, 6, 6, 6, 6,15, 0, 0, 0, 6, 6, 6, 6, /*   0- 15 */
    15, 0, 0, 0, 6, 6, 6, 6,15, 0, 0, 0, 6, 6, 6, 6, /*  16- 31 */
    15, 0, 0, 0, 6, 6, 6, 6,15, 0, 0, 0, 6, 6, 6, 6, /*  32- 47 */
    15, 0, 0, 0, 6, 6, 6, 6,15, 0, 0, 0, 6, 6, 6, 6, /*  48- 63 */
    15, 0, 0, 0, 6, 6, 6, 6,15, 0, 0, 0, 6, 6, 6, 6, /*  64- 79 */
    15, 0, 0, 0, 6, 6, 6, 6,15, 0, 0, 0, 6, 6, 6, 6, /*  80- 95 */
    15, 0, 0, 0, 6, 6, 6, 6,15, 0, 0, 0, 6, 6, 6, 6, /*  96-111 */
    15, 0, 0, 0, 6, 6, 6, 6,15, 0, 0, 0, 6, 6, 6, 6, /* 112-127 */
    15, 0, 0, 0, 6, 6, 6, 6,15, 0, 0, 0, 6, 6, 6, 6, /* 128-143 */
    15, 0, 0, 0, 6, 6, 6, 6,15, 0, 0, 0, 6, 6, 6, 6, /* 144-159 */
    15, 0, 0, 0, 6, 6, 6, 6,15, 0, 0, 0, 6, 6, 6, 6, /* 160-175 */
    15, 0, 0, 0, 6, 6, 6, 6,15, 0, 0, 0, 6, 6, 6, 6, /* 176-191 */
    15, 0, 0, 0, 6, 6, 6, 6,15, 0, 0, 0, 6, 6, 6, 6, /* 192-207 */
    15, 0, 0, 0, 6, 6, 6, 6,15, 0, 0, 0, 6, 6, 6, 6, /* 208-223 */
    15, 0, 0, 0, 6, 6, 6, 6,15, 0, 0, 0, 6, 6, 6, 6, /* 224-239 */
    15, 0, 0, 0, 6, 6, 6, 6,15, 0, 0, 0, 6, 6, 6, 6, /* 240-255 */
},
/* prefix 6 */ {
    16,17, 7, 7,16,17, 7, 7,16,17, 7, 7,16,17, 7, 7, /*   0- 15 */
    16,17, 7, 7,16,17, 7, 7,16,17, 7, 7,16,17, 7, 7, /*  16- 31 */
    16,17, 7, 7,16,17, 7, 7,16,17, 7, 7,16,17, 7, 7, /*  32- 47 */
    16,17, 7, 7,16,17, 7, 7,16,17, 7, 7,16,17, 7, 7, /*  48- 63 */
    16,17, 7, 7,16,17, 7, 7,16,17, 7, 7,16,17, 7, 7, /*  64- 79 */
    16,17, 7, 7,16,17, 7, 7,16,17, 7, 7,16,17, 7, 7, /*  80- 95 */
    16,17, 7, 7,16,17, 7, 7,16,17, 7, 7,16,17, 7, 7, /*  96-111 */
    16,17, 7, 7,16,17, 7, 7,16,17, 7, 7,16,17, 7, 7, /* 112-127 */
    16,17, 7, 7,16,17, 7, 7,16,17, 7, 7,16,17, 7, 7, /* 128-143 */
    16,17, 7, 7,16,17, 7, 7,16,17, 7, 7,16,17, 7, 7, /* 144-159 */
    16,17, 7, 7,16,17, 7, 7,16,17, 7, 7,16,17, 7, 7, /* 160-175 */
    16,17, 7, 7,16,17, 7, 7,16,17, 7, 7,16,17, 7, 7, /* 176-191 */
    16,17, 7, 7,16,17, 7, 7,16,17, 7, 7,16,17, 7, 7, /* 192-207 */
    16,17, 7, 7,16,17, 7, 7,16,17, 7, 7,16,17, 7, 7, /* 208-223 */
    16,17, 7, 7,16,17, 7, 7,16,17, 7, 7,16,17, 7, 7, /* 224-239 */
    16,17, 7, 7,16,17, 7, 7,16,17, 7, 7,16,17, 7, 7, /* 240-255 */
},
/* prefix 7 */ {
    18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0, /*   0- 15 */
    18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0, /*  16- 31 */
    18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0, /*  32- 47 */
    18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0, /*  48- 63 */
    18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0, /*  64- 79 */
    18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0, /*  80- 95 */
    18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0, /*  96-111 */
    18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0, /* 112-127 */
    18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0, /* 128-143 */
    18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0, /* 144-159 */
    18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0, /* 160-175 */
    18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0, /* 176-191 */
    18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0, /* 192-207 */
    18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0, /* 208-223 */
    18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0, /* 224-239 */
    18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0,18, 0, /* 240-255 */
},
/* prefix 8 */ {
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*   0- 15 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  16- 31 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  32- 47 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  48- 63 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  64- 79 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  80- 95 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  96-111 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 112-127 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 128-143 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 144-159 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 160-175 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 176-191 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 192-207 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 208-223 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 224-239 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 240-255 */
},
/* prefix 9 */ {
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*   0- 15 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  16- 31 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  32- 47 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  48- 63 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  64- 79 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  80- 95 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  96-111 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 112-127 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 128-143 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 144-159 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 160-175 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 176-191 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 192-207 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 208-223 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 224-239 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 240-255 */
},
/* prefix 10 */ {
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*   0- 15 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  16- 31 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  32- 47 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  48- 63 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  64- 79 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  80- 95 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  96-111 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 112-127 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 128-143 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 144-159 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 160-175 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 176-191 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 192-207 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 208-223 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 224-239 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 240-255 */
},
/* prefix 11 */ {
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*   0- 15 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  16- 31 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  32- 47 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  48- 63 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  64- 79 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  80- 95 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  96-111 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 12 */ {
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*   0- 15 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  16- 31 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  32- 47 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  48- 63 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  64- 79 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  80- 95 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  96-111 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 112-127 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 128-143 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 144-159 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 160-175 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 176-191 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 192-207 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 208-223 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 224-239 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 240-255 */
},
/* prefix 13 */ {
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*   0- 15 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  16- 31 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  32- 47 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  48- 63 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  64- 79 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  80- 95 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  96-111 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 14 */ {
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*   0- 15 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  16- 31 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  32- 47 */
     5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, /*  48- 63 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  64- 79 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  80- 95 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  96-111 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 112-127 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 128-143 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 144-159 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 160-175 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 176-191 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 192-207 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 208-223 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 224-239 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 240-255 */
},
/* prefix 15 */ {
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*   0- 15 */
     5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 7, /*  16- 31 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  32- 47 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  48- 63 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  64- 79 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  80- 95 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  96-111 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 16 */ {
     5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 7, 7, 0, 0, /*   0- 15 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  16- 31 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  32- 47 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  48- 63 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 17 */ {
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*   0- 15 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  16- 31 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  32- 47 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  48- 63 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  64- 79 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  80- 95 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  96-111 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 18 */ {
     6, 6, 6, 6, 7, 7, 0,19, 6, 6, 6, 6, 6, 6, 6, 6, /*   0- 15 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  48- 63 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 128-143 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 144-159 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 160-175 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 176-191 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 192-207 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 208-223 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 224-239 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 240-255 */
},
/* prefix 19 */ {
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*   0- 15 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  16- 31 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  32- 47 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  48- 63 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  64- 79 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  80- 95 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  96-111 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
}
};
unsigned char	RlUncompAction[20][256] = {
/* prefix 0 */ {
     0, 9, 8, 8, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, /*   0- 15 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  16- 31 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  32- 47 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  48- 63 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  64- 79 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  80- 95 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  96-111 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 112-127 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 128-143 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 144-159 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 160-175 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 176-191 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 192-207 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 208-223 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 224-239 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 240-255 */
},
/* prefix 1 */ {
     0, 8, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  48- 63 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     0, 8, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 128-143 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 144-159 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 160-175 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 176-191 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 192-207 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 208-223 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 224-239 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 240-255 */
},
/* prefix 2 */ {
     0, 7, 6, 6, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, /*   0- 15 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  16- 31 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  32- 47 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  48- 63 */
     0, 7, 6, 6, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, /*  64- 79 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     0, 7, 6, 6, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, /* 128-143 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 144-159 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 160-175 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 176-191 */
     0, 7, 6, 6, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, /* 192-207 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 208-223 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 224-239 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 240-255 */
},
/* prefix 3 */ {
     0, 6, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, /*   0- 15 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  16- 31 */
     0, 6, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  48- 63 */
     0, 6, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     0, 6, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     0, 6, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, /* 128-143 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 144-159 */
     0, 6, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, /* 160-175 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 176-191 */
     0, 6, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, /* 192-207 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 208-223 */
     0, 6, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, /* 224-239 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 240-255 */
},
/* prefix 4 */ {
     0, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, /*   0- 15 */
     0, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, /*  16- 31 */
     0, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, /*  32- 47 */
     0, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, /*  48- 63 */
     0, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     0, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     0, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     0, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     0, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, /* 128-143 */
     0, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, /* 144-159 */
     0, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, /* 160-175 */
     0, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, /* 176-191 */
     0, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, /* 192-207 */
     0, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, /* 208-223 */
     0, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, /* 224-239 */
     0, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, /* 240-255 */
},
/* prefix 5 */ {
     0, 4, 3, 3, 2, 2, 2, 2, 0, 4, 3, 3, 2, 2, 2, 2, /*   0- 15 */
     0, 4, 3, 3, 2, 2, 2, 2, 0, 4, 3, 3, 2, 2, 2, 2, /*  16- 31 */
     0, 4, 3, 3, 2, 2, 2, 2, 0, 4, 3, 3, 2, 2, 2, 2, /*  32- 47 */
     0, 4, 3, 3, 2, 2, 2, 2, 0, 4, 3, 3, 2, 2, 2, 2, /*  48- 63 */
     0, 4, 3, 3, 2, 2, 2, 2, 0, 4, 3, 3, 2, 2, 2, 2, /*  64- 79 */
     0, 4, 3, 3, 2, 2, 2, 2, 0, 4, 3, 3, 2, 2, 2, 2, /*  80- 95 */
     0, 4, 3, 3, 2, 2, 2, 2, 0, 4, 3, 3, 2, 2, 2, 2, /*  96-111 */
     0, 4, 3, 3, 2, 2, 2, 2, 0, 4, 3, 3, 2, 2, 2, 2, /* 112-127 */
     0, 4, 3, 3, 2, 2, 2, 2, 0, 4, 3, 3, 2, 2, 2, 2, /* 128-143 */
     0, 4, 3, 3, 2, 2, 2, 2, 0, 4, 3, 3, 2, 2, 2, 2, /* 144-159 */
     0, 4, 3, 3, 2, 2, 2, 2, 0, 4, 3, 3, 2, 2, 2, 2, /* 160-175 */
     0, 4, 3, 3, 2, 2, 2, 2, 0, 4, 3, 3, 2, 2, 2, 2, /* 176-191 */
     0, 4, 3, 3, 2, 2, 2, 2, 0, 4, 3, 3, 2, 2, 2, 2, /* 192-207 */
     0, 4, 3, 3, 2, 2, 2, 2, 0, 4, 3, 3, 2, 2, 2, 2, /* 208-223 */
     0, 4, 3, 3, 2, 2, 2, 2, 0, 4, 3, 3, 2, 2, 2, 2, /* 224-239 */
     0, 4, 3, 3, 2, 2, 2, 2, 0, 4, 3, 3, 2, 2, 2, 2, /* 240-255 */
},
/* prefix 6 */ {
     0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, /*   0- 15 */
     0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, /*  16- 31 */
     0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, /*  32- 47 */
     0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, /*  48- 63 */
     0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, /*  64- 79 */
     0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, /*  80- 95 */
     0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, /*  96-111 */
     0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, /* 112-127 */
     0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, /* 128-143 */
     0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, /* 144-159 */
     0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, /* 160-175 */
     0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, /* 176-191 */
     0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, /* 192-207 */
     0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, /* 208-223 */
     0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, /* 224-239 */
     0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, 0, 3, 2, 2, /* 240-255 */
},
/* prefix 7 */ {
     0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, /*   0- 15 */
     0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, /*  16- 31 */
     0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, /*  32- 47 */
     0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, /*  48- 63 */
     0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, /*  64- 79 */
     0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, /*  80- 95 */
     0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, /*  96-111 */
     0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, /* 112-127 */
     0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, /* 128-143 */
     0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, /* 144-159 */
     0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, /* 160-175 */
     0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, /* 176-191 */
     0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, /* 192-207 */
     0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, /* 208-223 */
     0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, /* 224-239 */
     0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, /* 240-255 */
},
/* prefix 8 */ {
    14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14, /*   0- 15 */
    14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14, /*  16- 31 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*  32- 47 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*  48- 63 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  64- 79 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  80- 95 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  96-111 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 112-127 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 128-143 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 144-159 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 160-175 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 176-191 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 192-207 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 208-223 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 224-239 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 240-255 */
},
/* prefix 9 */ {
    14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14, /*   0- 15 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*  16- 31 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  32- 47 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  48- 63 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /*  64- 79 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /*  80- 95 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /*  96-111 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /* 112-127 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 128-143 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 144-159 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 160-175 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 176-191 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 192-207 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 208-223 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 224-239 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 240-255 */
},
/* prefix 10 */ {
    14,14,14,14,14,14,14,14,12,12,12,12,12,12,12,12, /*   0- 15 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /*  16- 31 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /*  32- 47 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /*  48- 63 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /*  64- 79 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /*  80- 95 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /*  96-111 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /* 112-127 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 128-143 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 144-159 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 160-175 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 176-191 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 192-207 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 208-223 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 224-239 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 240-255 */
},
/* prefix 11 */ {
    14,14,14,14,12,12,12,12,11,11,11,11,11,11,11,11, /*   0- 15 */
    10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10, /*  16- 31 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /*  32- 47 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /*  48- 63 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /*  64- 79 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /*  80- 95 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /*  96-111 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /* 112-127 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 128-143 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 144-159 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 160-175 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 176-191 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 192-207 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 208-223 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 224-239 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 240-255 */
},
/* prefix 12 */ {
    14,14,12,12,11,11,11,11,10,10,10,10,10,10,10,10, /*   0- 15 */
     9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, /*  16- 31 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /*  32- 47 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /*  48- 63 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /*  64- 79 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /*  80- 95 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /*  96-111 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 112-127 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 128-143 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 144-159 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 160-175 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 176-191 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 192-207 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 208-223 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 224-239 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 240-255 */
},
/* prefix 13 */ {
    14,12,11,11,10,10,10,10, 9, 9, 9, 9, 9, 9, 9, 9, /*   0- 15 */
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, /*  16- 31 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /*  32- 47 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /*  48- 63 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  64- 79 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  80- 95 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  96-111 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 112-127 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 128-143 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 144-159 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 160-175 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 176-191 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 192-207 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 208-223 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 224-239 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 240-255 */
},
/* prefix 14 */ {
     0,11,10,10, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 8, 8, /*   0- 15 */
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /*  16- 31 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  32- 47 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  48- 63 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  64- 79 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  80- 95 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  96-111 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 112-127 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 128-143 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 144-159 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 160-175 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 176-191 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 192-207 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 208-223 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 224-239 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 240-255 */
},
/* prefix 15 */ {
     0,10, 9, 9, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 7, 7, /*   0- 15 */
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /*  16- 31 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  32- 47 */
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*  48- 63 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  64- 79 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  80- 95 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  96-111 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 112-127 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 128-143 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 144-159 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 160-175 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 176-191 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 192-207 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 208-223 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 224-239 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 240-255 */
},
/* prefix 16 */ {
    14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14, /*   0- 15 */
    14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14, /*  16- 31 */
    14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14, /*  32- 47 */
    14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14, /*  48- 63 */
    14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14, /*  64- 79 */
    14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14, /*  80- 95 */
    14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14, /*  96-111 */
    14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14, /* 112-127 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /* 128-143 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /* 144-159 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /* 160-175 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /* 176-191 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /* 192-207 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /* 208-223 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /* 224-239 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /* 240-255 */
},
/* prefix 17 */ {
    14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14, /*   0- 15 */
    14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14, /*  16- 31 */
    14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14, /*  32- 47 */
    14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14, /*  48- 63 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*  64- 79 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*  80- 95 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /*  96-111 */
    12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12, /* 112-127 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 128-143 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 144-159 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 160-175 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 176-191 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 192-207 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 208-223 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 224-239 */
    11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11, /* 240-255 */
},
/* prefix 18 */ {
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*   0- 15 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  16- 31 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  32- 47 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  48- 63 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  64- 79 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  80- 95 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  96-111 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 112-127 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 128-143 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 144-159 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 160-175 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 176-191 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 192-207 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 208-223 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 224-239 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 240-255 */
},
/* prefix 19 */ {
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*   0- 15 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  16- 31 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  32- 47 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  48- 63 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  64- 79 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  80- 95 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  96-111 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 112-127 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 128-143 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 144-159 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 160-175 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 176-191 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 192-207 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 208-223 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 224-239 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 240-255 */
}
};
unsigned char	RlUncompNextState[20][256] = {
/* prefix 0 */ {
     8, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  48- 63 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 1 */ {
     9, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  48- 63 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     9, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 128-143 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 144-159 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 160-175 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 176-191 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 192-207 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 208-223 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 224-239 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 240-255 */
},
/* prefix 2 */ {
    10, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  48- 63 */
    10, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*  64- 79 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  80- 95 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  96-111 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 112-127 */
    10, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 128-143 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 144-159 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 160-175 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 176-191 */
    10, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 192-207 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 208-223 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 224-239 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 240-255 */
},
/* prefix 3 */ {
    11, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  16- 31 */
    11, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*  32- 47 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  48- 63 */
    11, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*  64- 79 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  80- 95 */
    11, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*  96-111 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 112-127 */
    11, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 128-143 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 144-159 */
    11, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 160-175 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 176-191 */
    11, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 192-207 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 208-223 */
    11, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 224-239 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 240-255 */
},
/* prefix 4 */ {
    12, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*   0- 15 */
    12, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*  16- 31 */
    12, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*  32- 47 */
    12, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*  48- 63 */
    12, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*  64- 79 */
    12, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*  80- 95 */
    12, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*  96-111 */
    12, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 112-127 */
    12, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 128-143 */
    12, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 144-159 */
    12, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 160-175 */
    12, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 176-191 */
    12, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 192-207 */
    12, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 208-223 */
    12, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 224-239 */
    12, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /* 240-255 */
},
/* prefix 5 */ {
    13, 0, 7, 7, 6, 6, 6, 6,13, 0, 7, 7, 6, 6, 6, 6, /*   0- 15 */
    13, 0, 7, 7, 6, 6, 6, 6,13, 0, 7, 7, 6, 6, 6, 6, /*  16- 31 */
    13, 0, 7, 7, 6, 6, 6, 6,13, 0, 7, 7, 6, 6, 6, 6, /*  32- 47 */
    13, 0, 7, 7, 6, 6, 6, 6,13, 0, 7, 7, 6, 6, 6, 6, /*  48- 63 */
    13, 0, 7, 7, 6, 6, 6, 6,13, 0, 7, 7, 6, 6, 6, 6, /*  64- 79 */
    13, 0, 7, 7, 6, 6, 6, 6,13, 0, 7, 7, 6, 6, 6, 6, /*  80- 95 */
    13, 0, 7, 7, 6, 6, 6, 6,13, 0, 7, 7, 6, 6, 6, 6, /*  96-111 */
    13, 0, 7, 7, 6, 6, 6, 6,13, 0, 7, 7, 6, 6, 6, 6, /* 112-127 */
    13, 0, 7, 7, 6, 6, 6, 6,13, 0, 7, 7, 6, 6, 6, 6, /* 128-143 */
    13, 0, 7, 7, 6, 6, 6, 6,13, 0, 7, 7, 6, 6, 6, 6, /* 144-159 */
    13, 0, 7, 7, 6, 6, 6, 6,13, 0, 7, 7, 6, 6, 6, 6, /* 160-175 */
    13, 0, 7, 7, 6, 6, 6, 6,13, 0, 7, 7, 6, 6, 6, 6, /* 176-191 */
    13, 0, 7, 7, 6, 6, 6, 6,13, 0, 7, 7, 6, 6, 6, 6, /* 192-207 */
    13, 0, 7, 7, 6, 6, 6, 6,13, 0, 7, 7, 6, 6, 6, 6, /* 208-223 */
    13, 0, 7, 7, 6, 6, 6, 6,13, 0, 7, 7, 6, 6, 6, 6, /* 224-239 */
    13, 0, 7, 7, 6, 6, 6, 6,13, 0, 7, 7, 6, 6, 6, 6, /* 240-255 */
},
/* prefix 6 */ {
    14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7, /*   0- 15 */
    14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7, /*  16- 31 */
    14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7, /*  32- 47 */
    14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7, /*  48- 63 */
    14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7, /*  64- 79 */
    14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7, /*  80- 95 */
    14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7, /*  96-111 */
    14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7, /* 112-127 */
    14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7, /* 128-143 */
    14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7, /* 144-159 */
    14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7, /* 160-175 */
    14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7, /* 176-191 */
    14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7, /* 192-207 */
    14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7, /* 208-223 */
    14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7, /* 224-239 */
    14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7,14, 0, 7, 7, /* 240-255 */
},
/* prefix 7 */ {
    15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0, /*   0- 15 */
    15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0, /*  16- 31 */
    15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0, /*  32- 47 */
    15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0, /*  48- 63 */
    15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0, /*  64- 79 */
    15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0, /*  80- 95 */
    15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0, /*  96-111 */
    15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0, /* 112-127 */
    15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0, /* 128-143 */
    15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0, /* 144-159 */
    15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0, /* 160-175 */
    15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0, /* 176-191 */
    15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0, /* 192-207 */
    15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0, /* 208-223 */
    15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0, /* 224-239 */
    15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0,15, 0, /* 240-255 */
},
/* prefix 8 */ {
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*   0- 15 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  48- 63 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 9 */ {
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  48- 63 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 10 */ {
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  48- 63 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 11 */ {
     6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  48- 63 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 12 */ {
     7, 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  48- 63 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 13 */ {
     0, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  48- 63 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 14 */ {
    16, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  48- 63 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 15 */ {
    17, 0, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, /*   0- 15 */
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /*  16- 31 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  32- 47 */
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /*  48- 63 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 16 */ {
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*   0- 15 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  16- 31 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  32- 47 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  48- 63 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  64- 79 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  80- 95 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  96-111 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 17 */ {
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*   0- 15 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  16- 31 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  32- 47 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  48- 63 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  64- 79 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  80- 95 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /*  96-111 */
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 112-127 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 128-143 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 144-159 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 160-175 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 176-191 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 192-207 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 208-223 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 224-239 */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* 240-255 */
},
/* prefix 18 */ {
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*   0- 15 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  16- 31 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  32- 47 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  48- 63 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  64- 79 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  80- 95 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  96-111 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 112-127 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 128-143 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 144-159 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 160-175 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 176-191 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 192-207 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 208-223 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 224-239 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 240-255 */
},
/* prefix 19 */ {
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*   0- 15 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  16- 31 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  32- 47 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  48- 63 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  64- 79 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  80- 95 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /*  96-111 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 112-127 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 128-143 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 144-159 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 160-175 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 176-191 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 192-207 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 208-223 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 224-239 */
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 240-255 */
}
};
